(* ::Package:: *)

BeginPackage["ControllableSystem`"]


generate::usage = "generate[initialState, inputs, step, max, \[CapitalDelta]t] runs the simulation for the specified duration and time granularity with the step function supplied. 'inputs' must be a function generating the inputs, given the current time.";
plotResults::usage = "plotResults[results, components] extracts and scales the components specified and provides a plot.";
animateResults::usage = "animateResults[results, \[CapitalDelta]t, drawFrame] uses a custom drawing function to provide an animated view of the results.";
keepEveryNthFrame::usage = "keepEveryNthFrame[result, n] drops frames to increase playback speed.";


Begin["`Private`"]


generate[initialState_,inputs_,step_,max_,\[CapitalDelta]t_]:=
FoldList[
{#2[[1]],step[#2[[1]],\[CapitalDelta]t,#1[[2]],#2[[2]]],#2[[2]]}&,
{0,initialState},
Table[{t,inputs[t]},{t,0,max-\[CapitalDelta]t,\[CapitalDelta]t}]
]//Rest


prepareAndScaleResults[results_,components_]:=Function[component,{#[[1]],#[[2]][component[[1]]]*component[[2]]}&/@results]/@components


plotResults[results_,components_]:=ListLinePlot[prepareAndScaleResults[results,components],PlotRange->{-1,1}]


animateResults[results_,\[CapitalDelta]t_,drawFrame_]:=ListAnimate[drawFrame/@results,1/\[CapitalDelta]t]


keepEveryNthFrame[result_,n_]:=Take[result,{1,-1,n}]


End[]


EndPackage[]
