(* ::Package:: *)

SetDirectory[NotebookDirectory[]];


Needs["ControllableSystem`"]


BeginPackage["Unicycle`"]


rw=0.5;
rcom=1.0;
rs=1.5;
mwreal=3;
mw=1.5*mwreal;
ms=75;
g=9.80665;
is=1/12*ms*rs^2;
maxTorque=200;(* maximum torque the motor can deliver *)


arcArrow[a_,r_,{start_,end_}]:={Circle[a,r,{start,end}],Arrow[{a+r {Cos[end+If[end<start,.01,-.01]],Sin[end+If[end<start,.01,-.01]]},a+r {Cos[end],Sin[end]}}]}


torquescale=0.01;
forcescale=0.01;


computeForces[t_,state_]:=Module[{
fwd,fwg,fw,\[Tau]sg,\[Tau]s,\[Tau]sw,\[Beta],fw1,fw2
},
fwd=-state["motorTorque"]/rw;
fwg=ms*g*Cos[state["\[Theta]"]]*Sin[state["\[Theta]"]];
fw=fwd+fwg;
\[Tau]sg=ms*g*Sin[state["\[Theta]"]]*rcom;
\[Beta]=(ms/mw)*(1+3 Cos[state["\[Theta]"]]^2);
fw1=fw*\[Beta]/(1+\[Beta]);
fw2=fw*1/(1+\[Beta]);
\[Tau]sw=fw1*rcom*Cos[state ["\[Theta]"]];
\[Tau]s=-state["motorTorque"]+\[Tau]sg+\[Tau]sw;
<|
"fwd"->fwd,"fwg"->fwg,"fw"->fw,"\[Tau]sg"->\[Tau]sg,"\[Tau]s"->\[Tau]s,"\[Tau]sw"->\[Tau]sw,"\[Beta]"->\[Beta],"fw1"->fw1,"fw2"->fw2
|>
]


drawFrame[{t_,state_ ,inputs_},options_:<||>]:=Module[{
wcenter,top,com,forces
},
wcenter={state["xw"],rw};
top=wcenter+RotationMatrix[state["\[Theta]"]].{0,rs};
com={state["xw"],rw}+RotationMatrix[state["\[Theta]"]].{0,rcom};
forces=computeForces[t,state];
Graphics[{
If[Lookup[options,"hideDetails",False],{},Inset[Grid[{{"t",t},{"driveIntent",inputs["driveIntent"]},{"motorTorque",state["motorTorque"]},{"vw",state["vw"]},{"controllerState",state["controllerState"]}}],{-1,-0.5}]],
{Black,Line[{{-5,0},{5,0}}]}, (* ground *)
{Gray,Circle[wcenter,rw],Disk[wcenter,0.1]} (* wheel *)
}~Join~If[state["fallen"],{},{
{Gray,Disk[top,0.1]}, (* saddle *)
{Gray,Thickness[0.01],Line[{wcenter, top}]}, (* rod *)
{Green, Disk[com,0.05]} (* com *)
}~Join~If[Lookup[options,"hideForces",False],{},{
{Blue,Arrowheads[0.01],arcArrow[wcenter,rw*2/3,{0,state["motorTorque"]*torquescale/(rw*(2/3))}-\[Pi]/2]},(* \[Tau]sd *)
{Red,Arrowheads[0.02],Opacity[0.5],Arrow[{wcenter,wcenter+{forces["fwd"] forcescale,0}}]}, (* fwd *)
{Red,Arrowheads[0.02],Opacity[0.5],Arrow[{wcenter,wcenter+{forces["fwg"] forcescale,0}}]}, (* fwg *)
{Red,Arrowheads[0.04],Thickness[0.005],Arrow[{wcenter,wcenter+{forces["fw"] forcescale,0}}]}, (* fw *)
{Red,Opacity[0.5],Arrowheads[0.02],arcArrow[wcenter,rs+0.1,{state["\[Theta]"]+\[Pi]/2,state["\[Theta]"]+\[Pi]/2+(forces["\[Tau]sg"] torquescale)/(rs+0.1)}]}, (* \[Tau]sg *)
{Red,Opacity[0.5],Arrowheads[0.02],Dashed,Opacity[0.5],arcArrow[wcenter,rs+0.2,{state["\[Theta]"]+\[Pi]/2,state["\[Theta]"]+\[Pi]/2+(forces["\[Tau]sw"] torquescale)/(rs+0.2)}]}, (* \[Tau]sw *)
{Blue,Opacity[0.5],Arrowheads[0.02],arcArrow[wcenter,rs+0.3,{state["\[Theta]"]+\[Pi]/2,state["\[Theta]"]+\[Pi]/2+(-state["motorTorque"] torquescale)/(rs+0.3)}]} ,(* \[Tau]sd *)
{Red,Thickness[0.003],Arrowheads[0.025],arcArrow[wcenter,rs,{state["\[Theta]"]+\[Pi]/2,state["\[Theta]"]+\[Pi]/2+(forces["\[Tau]s"] torquescale)/rs}]}, (* \[Tau]s *)
If[Lookup[options,"hideDriveIntent",False],{},{Blue, Arrow[{{0,2.8},{3*inputs["driveIntent"],2.8}}]}](* driveIntent *)
}]],
PlotRange->{If[Lookup[options,"thin",False],{-1.5,1.5},{-3,3}],{If[Lookup[options,"hideDetails",False],-0.2,-1],3}},
Frame->True,
FrameTicks->Automatic,
ImageSize->If[Lookup[options,"thin",False],400,800]
]]


limitAbs[value_,max_]:=Sign[value]*Min[Abs[value],Abs[max]]


step[observer_,controller_]:=Function[{t,\[CapitalDelta]t,state,inputs},Module[{
forces,xw2,vw2,\[Theta]2,\[Omega]s2,fallen2,state2,controllerOutput
},
forces=computeForces[t,state];
vw2=state["vw"]+(forces["fw2"] \[CapitalDelta]t)/mw;
\[Omega]s2=state["\[Omega]s"]+(forces["\[Tau]s"] \[CapitalDelta]t)/is;
xw2=state["xw"]+vw2*\[CapitalDelta]t;
\[Theta]2=state["\[Theta]"]+\[Omega]s2*\[CapitalDelta]t;
fallen2=state["fallen"]\[Or]Abs[\[Theta]2]>=\[Pi]/2;
state2=<|"xw"->xw2,"vw"->vw2,"\[Theta]"->\[Theta]2,"\[Omega]s"->\[Omega]s2,"fallen"->fallen2|>;
controllerOutput=controller[\[CapitalDelta]t,state["controllerState"],observer[state,inputs]]; (* note the built-in one-step delay *)
Append[state2,{"motorTorque"->limitAbs[controllerOutput["motorTorque"],maxTorque],"controllerState"->controllerOutput["controllerState"]}]
]]


EndPackage[]
