(* Content-type: application/vnd.wolfram.mathematica *)

(*** Wolfram Notebook File ***)
(* http://www.wolfram.com/nb *)

(* CreatedBy='Mathematica 10.0' *)

(*CacheID: 234*)
(* Internal cache information:
NotebookFileLineBreakTest
NotebookFileLineBreakTest
NotebookDataPosition[       158,          7]
NotebookDataLength[   3516068,      63017]
NotebookOptionsPosition[   3514915,      62977]
NotebookOutlinePosition[   3515294,      62993]
CellTagsIndexPosition[   3515251,      62990]
WindowFrame->Normal*)

(* Beginning of Notebook Content *)
Notebook[{
Cell[BoxData[
 RowBox[{
  RowBox[{"SetDirectory", "[", 
   RowBox[{"NotebookDirectory", "[", "]"}], "]"}], ";"}]], "Input",
 CellChangeTimes->{{3.6426595944234076`*^9, 3.6426595981610937`*^9}}],

Cell[BoxData[
 RowBox[{"<<", "Unicycle`"}]], "Input",
 CellChangeTimes->{{3.6426589282393456`*^9, 3.6426589443407955`*^9}, {
  3.6426589764629793`*^9, 3.6426589915225773`*^9}, {3.642659601333319*^9, 
  3.642659606752175*^9}, {3.6426602351178184`*^9, 3.642660236295658*^9}}],

Cell[BoxData[
 RowBox[{
  RowBox[{"observer", "[", 
   RowBox[{"state_", ",", "inputs_"}], "]"}], ":=", 
  RowBox[{"<|", 
   RowBox[{
    RowBox[{"\"\<\[Theta]\>\"", "\[Rule]", 
     RowBox[{"state", "[", "\"\<\[Theta]\>\"", "]"}]}], ",", 
    RowBox[{"\"\<\[Omega]s\>\"", "\[Rule]", 
     RowBox[{"state", "[", "\"\<\[Omega]s\>\"", "]"}]}], ",", 
    RowBox[{"\"\<vw\>\"", "\[Rule]", 
     RowBox[{"state", "[", "\"\<vw\>\"", "]"}]}], ",", 
    RowBox[{"\"\<driveIntent\>\"", "\[Rule]", 
     RowBox[{"inputs", "[", "\"\<driveIntent\>\"", "]"}]}]}], 
   "|>"}]}]], "Input",
 CellChangeTimes->{{3.641970370532409*^9, 3.6419703827369113`*^9}, {
  3.6419733100925593`*^9, 3.641973351933449*^9}, {3.6419757653174095`*^9, 
  3.6419757766604886`*^9}, {3.6419781572264924`*^9, 3.6419781630472355`*^9}}],

Cell[BoxData[
 RowBox[{
  RowBox[{"noController", "[", 
   RowBox[{"controllerState_", ",", "observations_"}], "]"}], ":=", 
  RowBox[{"<|", 
   RowBox[{
    RowBox[{"\"\<motorTorque\>\"", "\[Rule]", "0"}], ",", 
    RowBox[{"\"\<controllerState\>\"", "\[Rule]", 
     RowBox[{"<|", "|>"}]}]}], "|>"}]}]], "Input",
 CellChangeTimes->{{3.641970385917512*^9, 3.641970420583209*^9}, {
  3.6419712385382075`*^9, 3.641971260573309*^9}, {3.6419732767992043`*^9, 
  3.6419732777243567`*^9}}],

Cell[BoxData[
 RowBox[{
  RowBox[{"initialControllerState", "=", 
   RowBox[{"<|", "|>"}]}], ";"}]], "Input",
 CellChangeTimes->{{3.6419710158858643`*^9, 3.641971033416424*^9}}],

Cell[BoxData[
 RowBox[{
  RowBox[{"initialState", "=", 
   RowBox[{"<|", 
    RowBox[{
     RowBox[{"\"\<xw\>\"", "\[Rule]", 
      RowBox[{"-", "2"}]}], ",", 
     RowBox[{"\"\<vw\>\"", "\[Rule]", "0"}], ",", 
     RowBox[{"\"\<\[Theta]\>\"", "\[Rule]", "0.0"}], ",", 
     RowBox[{"\"\<\[Omega]s\>\"", "\[Rule]", "0"}], ",", 
     RowBox[{"\"\<fallen\>\"", "\[Rule]", "False"}], ",", 
     RowBox[{"\"\<motorTorque\>\"", "\[Rule]", "0"}], ",", 
     RowBox[{
     "\"\<controllerState\>\"", "\[Rule]", "initialControllerState"}]}], 
    "|>"}]}], ";"}]], "Input",
 CellChangeTimes->{
  3.6419710352041054`*^9, {3.6419733789753513`*^9, 3.641973381403776*^9}, 
   3.641974517623592*^9, {3.6419749860849133`*^9, 3.6419750090145206`*^9}, {
   3.641978257027047*^9, 3.641978271753752*^9}, {3.641979205860104*^9, 
   3.6419792059450383`*^9}}],

Cell[BoxData[
 RowBox[{
  RowBox[{"slowOscillatingDriver", "[", "t_", "]"}], ":=", 
  RowBox[{"<|", 
   RowBox[{"\"\<driveIntent\>\"", "\[Rule]", 
    RowBox[{"Piecewise", "[", 
     RowBox[{
      RowBox[{"{", 
       RowBox[{
        RowBox[{"{", 
         RowBox[{"0", ",", 
          RowBox[{
           RowBox[{"Mod", "[", 
            RowBox[{"t", ",", "50"}], "]"}], "<", "2"}]}], "}"}], ",", 
        RowBox[{"{", 
         RowBox[{"0.5", ",", 
          RowBox[{
           RowBox[{"Mod", "[", 
            RowBox[{"t", ",", "50"}], "]"}], "<", "7"}]}], "}"}], ",", 
        RowBox[{"{", 
         RowBox[{"0", ",", 
          RowBox[{
           RowBox[{"Mod", "[", 
            RowBox[{"t", ",", "50"}], "]"}], "<", "12"}]}], "}"}], ",", 
        RowBox[{"{", 
         RowBox[{
          RowBox[{"-", "0.5"}], ",", 
          RowBox[{
           RowBox[{"Mod", "[", 
            RowBox[{"t", ",", "50"}], "]"}], "<", "17"}]}], "}"}]}], "}"}], 
      ",", "0"}], "]"}]}], "|>"}]}]], "Input",
 CellChangeTimes->{{3.6419715804970117`*^9, 3.6419715974458914`*^9}, {
   3.641971674655006*^9, 3.641971688974321*^9}, 3.641971882725773*^9, {
   3.6419721957244*^9, 3.641972200524847*^9}, 3.6419729718365493`*^9, {
   3.641979161465256*^9, 3.641979179569099*^9}, {3.6419793142503963`*^9, 
   3.641979320423047*^9}, {3.6497453570673485`*^9, 3.6497453596451015`*^9}, {
   3.6497454187276077`*^9, 3.649745434946433*^9}, {3.6497689164624233`*^9, 
   3.649768962629446*^9}}],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"slowOscillatingDriver", "[", "0.5", "]"}]], "Input",
 CellChangeTimes->{{3.6419729436891336`*^9, 3.6419729463311405`*^9}}],

Cell[BoxData[
 RowBox[{"\[LeftAssociation]", 
  RowBox[{"\<\"driveIntent\"\>", "\[Rule]", "0"}], 
  "\[RightAssociation]"}]], "Output",
 CellChangeTimes->{
  3.641972946565504*^9, 3.6419730069542556`*^9, {3.641973118756719*^9, 
   3.6419731329679003`*^9}, 3.6419733824208755`*^9, 3.6419741177985153`*^9, {
   3.641974343002222*^9, 3.6419745195034466`*^9}, {3.6419746183513412`*^9, 
   3.6419746454443226`*^9}, {3.641974955350365*^9, 3.6419750272584457`*^9}, {
   3.6419758125723395`*^9, 3.6419758607366023`*^9}, 3.64197718489893*^9, 
   3.6419774353027587`*^9, {3.641977483941815*^9, 3.641977498760721*^9}, 
   3.641977587699587*^9, {3.641977638526926*^9, 3.6419776476683607`*^9}, {
   3.6419779813288136`*^9, 3.641978004804963*^9}, 3.6419780485273304`*^9, {
   3.6419781712312016`*^9, 3.6419783093216343`*^9}, 3.641978340416995*^9, 
   3.641978438235708*^9, {3.641978623703939*^9, 3.6419786324098244`*^9}, 
   3.6419786642024736`*^9, 3.6419787741518197`*^9, {3.6419788225560255`*^9, 
   3.641978886953824*^9}, {3.6419789285054703`*^9, 3.6419791345418415`*^9}, 
   3.64197918238476*^9, 3.641979276790465*^9, 3.6419793241964836`*^9, 
   3.6419811921143084`*^9, {3.6419812257779846`*^9, 3.641981254511924*^9}, {
   3.641981287184244*^9, 3.641981358397845*^9}, {3.64198139250191*^9, 
   3.641981444175412*^9}, 3.641981604551755*^9, {3.641981713700585*^9, 
   3.6419817399255342`*^9}, {3.6419818364044256`*^9, 3.6419818662800913`*^9}, 
   3.6419845480091877`*^9, 3.6423070278269053`*^9, 3.6423219974199533`*^9, 
   3.642441828516202*^9, 3.6426602450152426`*^9, 3.6426603721768856`*^9, {
   3.6426675698454*^9, 3.642667582391677*^9}, 3.642667624197134*^9, 
   3.642667668492997*^9, 3.6455975905085754`*^9, 3.645597720411602*^9, 
   3.649745374914149*^9, 3.649745436285436*^9, 3.649768291082699*^9, 
   3.649768354846021*^9, 3.649768939098403*^9, 3.649768970492677*^9, 
   3.649770535961728*^9, 3.654163549813106*^9, 3.6541662468817096`*^9, 
   3.6543208976688786`*^9, 3.654595913203909*^9, 3.654596085580839*^9}]
}, Open  ]],

Cell[BoxData[
 RowBox[{
  RowBox[{"pidController", "[", 
   RowBox[{
   "\[CapitalDelta]t_", ",", "controllerState_", ",", "observations_"}], 
   "]"}], ":=", 
  RowBox[{"Module", "[", 
   RowBox[{
    RowBox[{"{", "\[IndentingNewLine]", 
     RowBox[{
     "error", ",", "\[IndentingNewLine]", "diff", ",", "\[IndentingNewLine]", 
      "int", ",", "\[IndentingNewLine]", 
      RowBox[{"p", "=", "700"}], ",", "\[IndentingNewLine]", 
      RowBox[{"d", "=", "100"}], ",", "\[IndentingNewLine]", 
      RowBox[{"i", "=", "200"}], ",", "\[IndentingNewLine]", 
      RowBox[{"intentFactor", "=", "0.3"}]}], "\[IndentingNewLine]", "}"}], 
    ",", "\[IndentingNewLine]", 
    RowBox[{
     RowBox[{"error", "=", 
      RowBox[{
       RowBox[{"observations", "[", "\"\<\[Theta]\>\"", "]"}], "+", 
       RowBox[{
        RowBox[{"(", 
         RowBox[{
          RowBox[{"observations", "[", "\"\<driveIntent\>\"", "]"}], "-", 
          RowBox[{"observations", "[", "\"\<vw\>\"", "]"}]}], ")"}], "*", 
        "intentFactor"}]}]}], ";", "\[IndentingNewLine]", 
     RowBox[{"diff", "=", 
      RowBox[{
       RowBox[{"(", 
        RowBox[{"error", "-", 
         RowBox[{"controllerState", "[", "\"\<lastError\>\"", "]"}]}], ")"}], 
       "/", "\[CapitalDelta]t"}]}], ";", "\[IndentingNewLine]", 
     RowBox[{"int", "=", 
      RowBox[{
       RowBox[{"Lookup", "[", 
        RowBox[{"controllerState", ",", "\"\<lastInt\>\"", ",", "0"}], "]"}], 
       "+", 
       RowBox[{"error", "*", "\[CapitalDelta]t"}]}]}], ";", 
     "\[IndentingNewLine]", 
     RowBox[{"<|", "\[IndentingNewLine]", 
      RowBox[{
       RowBox[{"\"\<controllerState\>\"", "\[Rule]", 
        RowBox[{"<|", 
         RowBox[{
          RowBox[{"\"\<lastError\>\"", "\[Rule]", "error"}], ",", 
          RowBox[{"\"\<lastInt\>\"", "\[Rule]", "int"}], ",", 
          RowBox[{"\"\<initialized\>\"", "\[Rule]", "True"}]}], "|>"}]}], ",",
        "\[IndentingNewLine]", 
       RowBox[{"\"\<motorTorque\>\"", "\[Rule]", 
        RowBox[{"If", "[", 
         RowBox[{
          RowBox[{"Lookup", "[", 
           RowBox[{
           "controllerState", ",", "\"\<initialized\>\"", ",", "False"}], 
           "]"}], ",", 
          RowBox[{
           RowBox[{"p", "*", "error"}], "+", 
           RowBox[{"d", "*", "diff"}], "+", 
           RowBox[{"i", "*", "int"}]}], ",", "0"}], "]"}]}]}], 
      "\[IndentingNewLine]", "|>"}]}]}], "\[IndentingNewLine]", 
   "]"}]}]], "Input",
 CellChangeTimes->{{3.641977165632508*^9, 3.64197716841586*^9}, {
   3.6419773778693347`*^9, 3.6419774060698595`*^9}, {3.6419775340221415`*^9, 
   3.641977582602208*^9}, 3.6419776317846527`*^9, {3.6419776871960583`*^9, 
   3.6419779329484434`*^9}, {3.6419779675713058`*^9, 
   3.6419779757462234`*^9}, {3.6419781064126606`*^9, 3.641978145699026*^9}, 
   3.641978239806724*^9, 3.641978275395651*^9, {3.6419783080087013`*^9, 
   3.6419783396982927`*^9}, {3.6419785572324777`*^9, 3.641978619330909*^9}, {
   3.641978832933446*^9, 3.6419788857347608`*^9}, {3.6419789252397733`*^9, 
   3.6419791337670145`*^9}, 3.6419811906751785`*^9, {3.6419812246457033`*^9, 
   3.6419812534453707`*^9}, {3.6419812857142615`*^9, 
   3.6419813574108877`*^9}, {3.6419813915651183`*^9, 3.641981470763973*^9}, {
   3.6419818345467815`*^9, 3.641981834796847*^9}, 3.641981865373411*^9, {
   3.649745363746827*^9, 3.649745367934815*^9}}],

Cell[BoxData[
 RowBox[{
  RowBox[{"results", "=", 
   RowBox[{"generate", "[", 
    RowBox[{"initialState", ",", "slowOscillatingDriver", ",", 
     RowBox[{"step", "[", 
      RowBox[{"observer", ",", "pidController"}], "]"}], ",", "30", ",", 
     "0.001"}], "]"}]}], ";"}]], "Input",
 CellChangeTimes->{{3.6410275892469788`*^9, 3.6410276363652124`*^9}, 
   3.641027862064164*^9, 3.64102834183352*^9, {3.6410283961024985`*^9, 
   3.641028404688478*^9}, {3.641028527744836*^9, 3.641028528072959*^9}, 
   3.6410286442787924`*^9, {3.641029449092619*^9, 3.6410294615040493`*^9}, {
   3.641030442937172*^9, 3.6410304437496285`*^9}, {3.6410305073930655`*^9, 
   3.641030547329398*^9}, {3.6410306736253033`*^9, 3.641030716079194*^9}, {
   3.6410325288262563`*^9, 3.641032547752393*^9}, {3.6419686230028467`*^9, 
   3.641968630136175*^9}, {3.6419706364801097`*^9, 3.641970639685331*^9}, {
   3.641970882394908*^9, 3.6419709197355223`*^9}, {3.641970962876346*^9, 
   3.6419709644159937`*^9}, {3.6419710388721633`*^9, 
   3.6419710462118945`*^9}, {3.641971892906275*^9, 3.6419718937345343`*^9}, 
   3.6419725700574293`*^9, 3.6419726133237977`*^9, 3.6419729025827894`*^9, 
   3.641972939576999*^9, 3.6419730772290816`*^9, 3.641973132264871*^9, {
   3.6419732852651653`*^9, 3.6419732860620527`*^9}, {3.641973406896188*^9, 
   3.6419734083272796`*^9}, 3.64197411656363*^9, 3.6419744869228354`*^9, 
   3.6419771710088964`*^9, {3.6419774755542374`*^9, 3.6419774922576065`*^9}, 
   3.641977535506946*^9, 3.6419779984379377`*^9, {3.641978039150938*^9, 
   3.641978060856901*^9}, {3.6419781769808846`*^9, 3.641978204950899*^9}, 
   3.641978294969345*^9, {3.6419786288468585`*^9, 3.6419786615925107`*^9}, {
   3.6419787066010103`*^9, 3.6419787457295003`*^9}, {3.6419787809729056`*^9, 
   3.6419788035782876`*^9}, {3.6419792101746535`*^9, 3.641979211351735*^9}, {
   3.6426603686803846`*^9, 3.6426603943430567`*^9}, {3.6426675784815035`*^9, 
   3.6426675807835336`*^9}}],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"plotResults", "[", 
  RowBox[{"results", ",", 
   RowBox[{"{", 
    RowBox[{
     RowBox[{"{", 
      RowBox[{"\"\<vw\>\"", ",", "1"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"\"\<\[Theta]\>\"", ",", 
       RowBox[{"-", "1"}]}], "}"}]}], "}"}]}], "]"}]], "Input",
 CellChangeTimes->{{3.6410300641036634`*^9, 3.6410300894607706`*^9}, {
   3.641030144109178*^9, 3.6410301469378614`*^9}, {3.6419710838065524`*^9, 
   3.6419711434183607`*^9}, {3.6419739036762314`*^9, 3.64197390580604*^9}, 
   3.641974132303665*^9, {3.641977195029725*^9, 3.6419772575675845`*^9}, {
   3.6497454724970946`*^9, 3.6497454760671296`*^9}, {3.6497455856766806`*^9, 
   3.6497455920429964`*^9}}],

Cell[BoxData[
 GraphicsBox[{{}, {{}, {}, 
    {RGBColor[0.368417, 0.506779, 0.709798], PointSize[0.002777777777777778], 
     AbsoluteThickness[1.6], LineBox[CompressedData["
1:eJxc3GV4VPf29vFxKdoixYpD6SnuVmYhheLuFNfiFHd3d5Lg7sE1TELcJzi0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      "]]}, 
    {RGBColor[0.880722, 0.611041, 0.142051], PointSize[0.002777777777777778], 
     AbsoluteThickness[1.6], LineBox[CompressedData["
1:eJxc3XWYU4eXxvG4DK7FHYq1UKA45OBe3Ir8cHd3aaFYi7u7U5xixWHcEhyK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      "]]}}, {}},
  AspectRatio->NCache[GoldenRatio^(-1), 0.6180339887498948],
  Axes->{True, True},
  AxesLabel->{None, None},
  AxesOrigin->{0, 0},
  DisplayFunction->Identity,
  Frame->{{False, False}, {False, False}},
  FrameLabel->{{None, None}, {None, None}},
  FrameTicks->{{Automatic, Automatic}, {Automatic, Automatic}},
  GridLines->{None, None},
  GridLinesStyle->Directive[
    GrayLevel[0.5, 0.4]],
  Method->{},
  PlotRange->{{0, 29.999000000000002`}, {-1, 1}},
  PlotRangeClipping->True,
  PlotRangePadding->{{
     Scaled[0.02], 
     Scaled[0.02]}, {0, 0}},
  Ticks->{Automatic, Automatic}]], "Output",
 CellChangeTimes->{{3.6497454774191284`*^9, 3.649745483806176*^9}, {
   3.6497455876417246`*^9, 3.64974559335297*^9}, 3.6497682974134827`*^9, 
   3.6497683648029003`*^9, 3.649768958116469*^9, 3.6497689930271893`*^9, 
   3.6497705567692175`*^9, 3.654163555426861*^9, 3.654166265536311*^9, 
   3.654320913855856*^9, 3.654595922952411*^9, 
   3.6545960932281284`*^9},ImageCache->GraphicsData["CompressedBitmap", "\<\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\
\>"]]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"animateResults", "[", 
  RowBox[{
   RowBox[{"keepEveryNthFrame", "[", 
    RowBox[{"results", ",", "50"}], "]"}], ",", "0.05", ",", "drawFrame"}], 
  "]"}]], "Input",
 CellChangeTimes->{{3.641027667812115*^9, 3.641027700997234*^9}, 
   3.641028020265662*^9, 3.6410284599871244`*^9, 3.6410287329084983`*^9, {
   3.641029466311617*^9, 3.641029466983533*^9}, 3.6410297623919353`*^9, 
   3.6410301596821704`*^9, 3.641030303423514*^9, 3.641030460602631*^9, {
   3.6410305093243*^9, 3.641030518620326*^9}, 3.6410305666650877`*^9, 
   3.641030662957533*^9, 3.641030718726774*^9, 3.6411154726998677`*^9, 
   3.641115507566936*^9, 3.6415548260871706`*^9, {3.64155537381778*^9, 
   3.6415553773031797`*^9}, {3.6419710509405656`*^9, 3.6419710526749706`*^9}, 
   3.641973028580598*^9, {3.641973087911451*^9, 3.641973117324768*^9}, 
   3.641974094983841*^9, 3.641974489928064*^9, {3.641978191938369*^9, 
   3.6419782107759905`*^9}, 3.6419782971088724`*^9, {3.642667620423341*^9, 
   3.642667622950262*^9}, {3.6426676625445986`*^9, 3.6426676644739695`*^9}}],

Cell[BoxData[
 TagBox[
  StyleBox[
   DynamicModuleBox[{$CellContext`i3$$ = 441, Typeset`show$$ = True, 
    Typeset`bookmarkList$$ = {
    "\"min\"" :> {$CellContext`i3$$ = 1}, 
     "\"max\"" :> {$CellContext`i3$$ = 600}}, Typeset`bookmarkMode$$ = "Menu",
     Typeset`animator$$, Typeset`animvar$$ = 1, Typeset`name$$ = 
    "\"untitled\"", Typeset`specs$$ = {{{
       Hold[$CellContext`i3$$], 1, ""}, 1, 600, 1}}, Typeset`size$$ = 
    Automatic, Typeset`update$$ = 0, Typeset`initDone$$, 
    Typeset`skipInitDone$$ = True, $CellContext`i3$91829$$ = 0}, 
    PaneBox[
     PanelBox[
      DynamicWrapperBox[GridBox[{
         {
          ItemBox[
           ItemBox[
            TagBox[
             StyleBox[GridBox[{
                {"\<\"\"\>", 
                 AnimatorBox[Dynamic[$CellContext`i3$$], {1, 600, 1},
                  AnimationRate->20.,
                  AnimationRunTime->182.3514404296875,
                  AnimationTimeIndex->21.98222904205123,
                  
                  AppearanceElements->{
                   "ProgressSlider", "PlayPauseButton", "FasterSlowerButtons",
                     "DirectionButton"},
                  AutoAction->False,
                  DisplayAllSteps->True]}
               },
               AutoDelete->False,
               
               GridBoxAlignment->{
                "Columns" -> {Right, {Left}}, "ColumnsIndexed" -> {}, 
                 "Rows" -> {{Baseline}}, "RowsIndexed" -> {}},
               
               GridBoxItemSize->{
                "Columns" -> {{Automatic}}, "Rows" -> {{Automatic}}}], 
              "ListAnimateLabel",
              StripOnInput->False],
             {"ControlArea", Top}],
            Alignment->{Automatic, Inherited},
            StripOnInput->False],
           Background->None,
           StripOnInput->False]},
         {
          ItemBox[
           TagBox[
            StyleBox[
             PaneBox[
              TagBox[
               PaneSelectorBox[{1->
                GraphicsBox[{InsetBox[
                   TagBox[GridBox[{
                    {"\<\"t\"\>", "0.`"},
                    {"\<\"driveIntent\"\>", "0"},
                    {"\<\"motorTorque\"\>", "0"},
                    {"\<\"vw\"\>", "0.`"},
                    {"\<\"controllerState\"\>", 
                    RowBox[{"\[LeftAssociation]", 
                    RowBox[{
                    RowBox[{"\<\"lastError\"\>", "\[Rule]", "0.`"}], ",", 
                    RowBox[{"\<\"lastInt\"\>", "\[Rule]", "0.`"}], ",", 
                    RowBox[{"\<\"initialized\"\>", "\[Rule]", "True"}]}], 
                    "\[RightAssociation]"}]}
                    },
                    AutoDelete->False,
                    
                    GridBoxItemSize->{
                    "Columns" -> {{Automatic}}, "Rows" -> {{Automatic}}}],
                    "Grid"], {-1, -0.5}], 
                  {GrayLevel[0], LineBox[{{-5, 0}, {5, 0}}]}, 
                  {GrayLevel[0.5], CircleBox[{-2., 0.5}, 0.5], 
                   DiskBox[{-2., 0.5}, 0.1]}, 
                  {GrayLevel[0.5], DiskBox[{-2., 2.}, 0.1]}, 
                  {GrayLevel[0.5], Thickness[0.01], 
                   LineBox[{{-2., 0.5}, {-2., 2.}}]}, 
                  {RGBColor[0, 1, 0], DiskBox[{-2., 1.5}, 0.05]}, 
                  {RGBColor[0, 0, 1], Arrowheads[0.01], 
                   CircleBox[{-2., 0.5}, 0.3333333333333333, 
                    NCache[{
                    Rational[-1, 2] 
                    Pi, -1.5707963267948966`}, {-1.5707963267948966`, \
-1.5707963267948966`}]], 
                   ArrowBox[{{-2.0033332777780557`, 
                    0.16668333319444495`}, {-2., 0.16666666666666669`}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-2., 0.5}, {-2., 0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-2., 0.5}, {-2., 0.5}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.005], Arrowheads[0.04], 
                   ArrowBox[{{-2., 0.5}, {-2., 0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-2., 0.5}, 1.6, {1.5707963267948966`, 
                    1.5707963267948966`}], 
                   ArrowBox[{{-1.9840002666653331`, 
                    2.0999200006666645`}, {-2., 2.1}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Dashing[{Small, Small}], 
                   Arrowheads[0.02], 
                   CircleBox[{-2., 0.5}, 1.7, {1.5707963267948966`, 
                    1.5707963267948966`}], 
                   ArrowBox[{{-1.9830002833319165`, 2.199915000708331}, {-2., 
                    2.2}}]}, 
                  {RGBColor[0, 0, 1], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-2., 0.5}, 1.8, {1.5707963267948966`, 
                    1.5707963267948966`}], 
                   ArrowBox[{{-1.9820002999984998`, 
                    2.2999100007499975`}, {-2., 2.3}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.003], Arrowheads[0.025], 
                   CircleBox[{-2., 0.5}, 1.5, {1.5707963267948966`, 
                    1.5707963267948966`}], 
                   ArrowBox[{{-1.9850002499987498`, 1.999925000624998}, {-2., 
                    2.}}]}, 
                  {RGBColor[0, 0, 1], ArrowBox[{{0, 2.8}, {0, 2.8}}]}},
                 Frame->True,
                 FrameTicks->Automatic,
                 ImageSize->800,
                 PlotRange->{{-3, 3}, {-1, 3}}], 2->
                GraphicsBox[{InsetBox[
                   TagBox[GridBox[{
                    {"\<\"t\"\>", "0.05`"},
                    {"\<\"driveIntent\"\>", "0"},
                    {"\<\"motorTorque\"\>", "0.`"},
                    {"\<\"vw\"\>", "0.`"},
                    {"\<\"controllerState\"\>", 
                    RowBox[{"\[LeftAssociation]", 
                    RowBox[{
                    RowBox[{"\<\"lastError\"\>", "\[Rule]", "0.`"}], ",", 
                    RowBox[{"\<\"lastInt\"\>", "\[Rule]", "0.`"}], ",", 
                    RowBox[{"\<\"initialized\"\>", "\[Rule]", "True"}]}], 
                    "\[RightAssociation]"}]}
                    },
                    AutoDelete->False,
                    
                    GridBoxItemSize->{
                    "Columns" -> {{Automatic}}, "Rows" -> {{Automatic}}}],
                    "Grid"], {-1, -0.5}], 
                  {GrayLevel[0], LineBox[{{-5, 0}, {5, 0}}]}, 
                  {GrayLevel[0.5], CircleBox[{-2., 0.5}, 0.5], 
                   DiskBox[{-2., 0.5}, 0.1]}, 
                  {GrayLevel[0.5], DiskBox[{-2., 2.}, 0.1]}, 
                  {GrayLevel[0.5], Thickness[0.01], 
                   LineBox[{{-2., 0.5}, {-2., 2.}}]}, 
                  {RGBColor[0, 1, 0], DiskBox[{-2., 1.5}, 0.05]}, 
                  {RGBColor[0, 0, 1], Arrowheads[0.01], 
                   CircleBox[{-2., 0.5}, 0.3333333333333333, 
                    NCache[{
                    Rational[-1, 2] 
                    Pi, -1.5707963267948966`}, {-1.5707963267948966`, \
-1.5707963267948966`}]], 
                   ArrowBox[{{-2.0033332777780557`, 
                    0.16668333319444495`}, {-2., 0.16666666666666669`}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-2., 0.5}, {-2., 0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-2., 0.5}, {-2., 0.5}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.005], Arrowheads[0.04], 
                   ArrowBox[{{-2., 0.5}, {-2., 0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-2., 0.5}, 1.6, {1.5707963267948966`, 
                    1.5707963267948966`}], 
                   ArrowBox[{{-1.9840002666653331`, 
                    2.0999200006666645`}, {-2., 2.1}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Dashing[{Small, Small}], 
                   Arrowheads[0.02], 
                   CircleBox[{-2., 0.5}, 1.7, {1.5707963267948966`, 
                    1.5707963267948966`}], 
                   ArrowBox[{{-1.9830002833319165`, 2.199915000708331}, {-2., 
                    2.2}}]}, 
                  {RGBColor[0, 0, 1], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-2., 0.5}, 1.8, {1.5707963267948966`, 
                    1.5707963267948966`}], 
                   ArrowBox[{{-1.9820002999984998`, 
                    2.2999100007499975`}, {-2., 2.3}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.003], Arrowheads[0.025], 
                   CircleBox[{-2., 0.5}, 1.5, {1.5707963267948966`, 
                    1.5707963267948966`}], 
                   ArrowBox[{{-1.9850002499987498`, 1.999925000624998}, {-2., 
                    2.}}]}, 
                  {RGBColor[0, 0, 1], ArrowBox[{{0, 2.8}, {0, 2.8}}]}},
                 Frame->True,
                 FrameTicks->Automatic,
                 ImageSize->800,
                 PlotRange->{{-3, 3}, {-1, 3}}], 3->
                GraphicsBox[{InsetBox[
                   TagBox[GridBox[{
                    {"\<\"t\"\>", "0.1`"},
                    {"\<\"driveIntent\"\>", "0"},
                    {"\<\"motorTorque\"\>", "0.`"},
                    {"\<\"vw\"\>", "0.`"},
                    {"\<\"controllerState\"\>", 
                    RowBox[{"\[LeftAssociation]", 
                    RowBox[{
                    RowBox[{"\<\"lastError\"\>", "\[Rule]", "0.`"}], ",", 
                    RowBox[{"\<\"lastInt\"\>", "\[Rule]", "0.`"}], ",", 
                    RowBox[{"\<\"initialized\"\>", "\[Rule]", "True"}]}], 
                    "\[RightAssociation]"}]}
                    },
                    AutoDelete->False,
                    
                    GridBoxItemSize->{
                    "Columns" -> {{Automatic}}, "Rows" -> {{Automatic}}}],
                    "Grid"], {-1, -0.5}], 
                  {GrayLevel[0], LineBox[{{-5, 0}, {5, 0}}]}, 
                  {GrayLevel[0.5], CircleBox[{-2., 0.5}, 0.5], 
                   DiskBox[{-2., 0.5}, 0.1]}, 
                  {GrayLevel[0.5], DiskBox[{-2., 2.}, 0.1]}, 
                  {GrayLevel[0.5], Thickness[0.01], 
                   LineBox[{{-2., 0.5}, {-2., 2.}}]}, 
                  {RGBColor[0, 1, 0], DiskBox[{-2., 1.5}, 0.05]}, 
                  {RGBColor[0, 0, 1], Arrowheads[0.01], 
                   CircleBox[{-2., 0.5}, 0.3333333333333333, 
                    NCache[{
                    Rational[-1, 2] 
                    Pi, -1.5707963267948966`}, {-1.5707963267948966`, \
-1.5707963267948966`}]], 
                   ArrowBox[{{-2.0033332777780557`, 
                    0.16668333319444495`}, {-2., 0.16666666666666669`}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-2., 0.5}, {-2., 0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-2., 0.5}, {-2., 0.5}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.005], Arrowheads[0.04], 
                   ArrowBox[{{-2., 0.5}, {-2., 0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-2., 0.5}, 1.6, {1.5707963267948966`, 
                    1.5707963267948966`}], 
                   ArrowBox[{{-1.9840002666653331`, 
                    2.0999200006666645`}, {-2., 2.1}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Dashing[{Small, Small}], 
                   Arrowheads[0.02], 
                   CircleBox[{-2., 0.5}, 1.7, {1.5707963267948966`, 
                    1.5707963267948966`}], 
                   ArrowBox[{{-1.9830002833319165`, 2.199915000708331}, {-2., 
                    2.2}}]}, 
                  {RGBColor[0, 0, 1], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-2., 0.5}, 1.8, {1.5707963267948966`, 
                    1.5707963267948966`}], 
                   ArrowBox[{{-1.9820002999984998`, 
                    2.2999100007499975`}, {-2., 2.3}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.003], Arrowheads[0.025], 
                   CircleBox[{-2., 0.5}, 1.5, {1.5707963267948966`, 
                    1.5707963267948966`}], 
                   ArrowBox[{{-1.9850002499987498`, 1.999925000624998}, {-2., 
                    2.}}]}, 
                  {RGBColor[0, 0, 1], ArrowBox[{{0, 2.8}, {0, 2.8}}]}},
                 Frame->True,
                 FrameTicks->Automatic,
                 ImageSize->800,
                 PlotRange->{{-3, 3}, {-1, 3}}], 4->
                GraphicsBox[{InsetBox[
                   TagBox[GridBox[{
                    {"\<\"t\"\>", "0.15`"},
                    {"\<\"driveIntent\"\>", "0"},
                    {"\<\"motorTorque\"\>", "0.`"},
                    {"\<\"vw\"\>", "0.`"},
                    {"\<\"controllerState\"\>", 
                    RowBox[{"\[LeftAssociation]", 
                    RowBox[{
                    RowBox[{"\<\"lastError\"\>", "\[Rule]", "0.`"}], ",", 
                    RowBox[{"\<\"lastInt\"\>", "\[Rule]", "0.`"}], ",", 
                    RowBox[{"\<\"initialized\"\>", "\[Rule]", "True"}]}], 
                    "\[RightAssociation]"}]}
                    },
                    AutoDelete->False,
                    
                    GridBoxItemSize->{
                    "Columns" -> {{Automatic}}, "Rows" -> {{Automatic}}}],
                    "Grid"], {-1, -0.5}], 
                  {GrayLevel[0], LineBox[{{-5, 0}, {5, 0}}]}, 
                  {GrayLevel[0.5], CircleBox[{-2., 0.5}, 0.5], 
                   DiskBox[{-2., 0.5}, 0.1]}, 
                  {GrayLevel[0.5], DiskBox[{-2., 2.}, 0.1]}, 
                  {GrayLevel[0.5], Thickness[0.01], 
                   LineBox[{{-2., 0.5}, {-2., 2.}}]}, 
                  {RGBColor[0, 1, 0], DiskBox[{-2., 1.5}, 0.05]}, 
                  {RGBColor[0, 0, 1], Arrowheads[0.01], 
                   CircleBox[{-2., 0.5}, 0.3333333333333333, 
                    NCache[{
                    Rational[-1, 2] 
                    Pi, -1.5707963267948966`}, {-1.5707963267948966`, \
-1.5707963267948966`}]], 
                   ArrowBox[{{-2.0033332777780557`, 
                    0.16668333319444495`}, {-2., 0.16666666666666669`}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-2., 0.5}, {-2., 0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-2., 0.5}, {-2., 0.5}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.005], Arrowheads[0.04], 
                   ArrowBox[{{-2., 0.5}, {-2., 0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-2., 0.5}, 1.6, {1.5707963267948966`, 
                    1.5707963267948966`}], 
                   ArrowBox[{{-1.9840002666653331`, 
                    2.0999200006666645`}, {-2., 2.1}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Dashing[{Small, Small}], 
                   Arrowheads[0.02], 
                   CircleBox[{-2., 0.5}, 1.7, {1.5707963267948966`, 
                    1.5707963267948966`}], 
                   ArrowBox[{{-1.9830002833319165`, 2.199915000708331}, {-2., 
                    2.2}}]}, 
                  {RGBColor[0, 0, 1], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-2., 0.5}, 1.8, {1.5707963267948966`, 
                    1.5707963267948966`}], 
                   ArrowBox[{{-1.9820002999984998`, 
                    2.2999100007499975`}, {-2., 2.3}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.003], Arrowheads[0.025], 
                   CircleBox[{-2., 0.5}, 1.5, {1.5707963267948966`, 
                    1.5707963267948966`}], 
                   ArrowBox[{{-1.9850002499987498`, 1.999925000624998}, {-2., 
                    2.}}]}, 
                  {RGBColor[0, 0, 1], ArrowBox[{{0, 2.8}, {0, 2.8}}]}},
                 Frame->True,
                 FrameTicks->Automatic,
                 ImageSize->800,
                 PlotRange->{{-3, 3}, {-1, 3}}], 5->
                GraphicsBox[{InsetBox[
                   TagBox[GridBox[{
                    {"\<\"t\"\>", "0.2`"},
                    {"\<\"driveIntent\"\>", "0"},
                    {"\<\"motorTorque\"\>", "0.`"},
                    {"\<\"vw\"\>", "0.`"},
                    {"\<\"controllerState\"\>", 
                    RowBox[{"\[LeftAssociation]", 
                    RowBox[{
                    RowBox[{"\<\"lastError\"\>", "\[Rule]", "0.`"}], ",", 
                    RowBox[{"\<\"lastInt\"\>", "\[Rule]", "0.`"}], ",", 
                    RowBox[{"\<\"initialized\"\>", "\[Rule]", "True"}]}], 
                    "\[RightAssociation]"}]}
                    },
                    AutoDelete->False,
                    
                    GridBoxItemSize->{
                    "Columns" -> {{Automatic}}, "Rows" -> {{Automatic}}}],
                    "Grid"], {-1, -0.5}], 
                  {GrayLevel[0], LineBox[{{-5, 0}, {5, 0}}]}, 
                  {GrayLevel[0.5], CircleBox[{-2., 0.5}, 0.5], 
                   DiskBox[{-2., 0.5}, 0.1]}, 
                  {GrayLevel[0.5], DiskBox[{-2., 2.}, 0.1]}, 
                  {GrayLevel[0.5], Thickness[0.01], 
                   LineBox[{{-2., 0.5}, {-2., 2.}}]}, 
                  {RGBColor[0, 1, 0], DiskBox[{-2., 1.5}, 0.05]}, 
                  {RGBColor[0, 0, 1], Arrowheads[0.01], 
                   CircleBox[{-2., 0.5}, 0.3333333333333333, 
                    NCache[{
                    Rational[-1, 2] 
                    Pi, -1.5707963267948966`}, {-1.5707963267948966`, \
-1.5707963267948966`}]], 
                   ArrowBox[{{-2.0033332777780557`, 
                    0.16668333319444495`}, {-2., 0.16666666666666669`}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-2., 0.5}, {-2., 0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-2., 0.5}, {-2., 0.5}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.005], Arrowheads[0.04], 
                   ArrowBox[{{-2., 0.5}, {-2., 0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-2., 0.5}, 1.6, {1.5707963267948966`, 
                    1.5707963267948966`}], 
                   ArrowBox[{{-1.9840002666653331`, 
                    2.0999200006666645`}, {-2., 2.1}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Dashing[{Small, Small}], 
                   Arrowheads[0.02], 
                   CircleBox[{-2., 0.5}, 1.7, {1.5707963267948966`, 
                    1.5707963267948966`}], 
                   ArrowBox[{{-1.9830002833319165`, 2.199915000708331}, {-2., 
                    2.2}}]}, 
                  {RGBColor[0, 0, 1], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-2., 0.5}, 1.8, {1.5707963267948966`, 
                    1.5707963267948966`}], 
                   ArrowBox[{{-1.9820002999984998`, 
                    2.2999100007499975`}, {-2., 2.3}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.003], Arrowheads[0.025], 
                   CircleBox[{-2., 0.5}, 1.5, {1.5707963267948966`, 
                    1.5707963267948966`}], 
                   ArrowBox[{{-1.9850002499987498`, 1.999925000624998}, {-2., 
                    2.}}]}, 
                  {RGBColor[0, 0, 1], ArrowBox[{{0, 2.8}, {0, 2.8}}]}},
                 Frame->True,
                 FrameTicks->Automatic,
                 ImageSize->800,
                 PlotRange->{{-3, 3}, {-1, 3}}], 6->
                GraphicsBox[{InsetBox[
                   TagBox[GridBox[{
                    {"\<\"t\"\>", "0.25`"},
                    {"\<\"driveIntent\"\>", "0"},
                    {"\<\"motorTorque\"\>", "0.`"},
                    {"\<\"vw\"\>", "0.`"},
                    {"\<\"controllerState\"\>", 
                    RowBox[{"\[LeftAssociation]", 
                    RowBox[{
                    RowBox[{"\<\"lastError\"\>", "\[Rule]", "0.`"}], ",", 
                    RowBox[{"\<\"lastInt\"\>", "\[Rule]", "0.`"}], ",", 
                    RowBox[{"\<\"initialized\"\>", "\[Rule]", "True"}]}], 
                    "\[RightAssociation]"}]}
                    },
                    AutoDelete->False,
                    
                    GridBoxItemSize->{
                    "Columns" -> {{Automatic}}, "Rows" -> {{Automatic}}}],
                    "Grid"], {-1, -0.5}], 
                  {GrayLevel[0], LineBox[{{-5, 0}, {5, 0}}]}, 
                  {GrayLevel[0.5], CircleBox[{-2., 0.5}, 0.5], 
                   DiskBox[{-2., 0.5}, 0.1]}, 
                  {GrayLevel[0.5], DiskBox[{-2., 2.}, 0.1]}, 
                  {GrayLevel[0.5], Thickness[0.01], 
                   LineBox[{{-2., 0.5}, {-2., 2.}}]}, 
                  {RGBColor[0, 1, 0], DiskBox[{-2., 1.5}, 0.05]}, 
                  {RGBColor[0, 0, 1], Arrowheads[0.01], 
                   CircleBox[{-2., 0.5}, 0.3333333333333333, 
                    NCache[{
                    Rational[-1, 2] 
                    Pi, -1.5707963267948966`}, {-1.5707963267948966`, \
-1.5707963267948966`}]], 
                   ArrowBox[{{-2.0033332777780557`, 
                    0.16668333319444495`}, {-2., 0.16666666666666669`}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-2., 0.5}, {-2., 0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-2., 0.5}, {-2., 0.5}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.005], Arrowheads[0.04], 
                   ArrowBox[{{-2., 0.5}, {-2., 0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-2., 0.5}, 1.6, {1.5707963267948966`, 
                    1.5707963267948966`}], 
                   ArrowBox[{{-1.9840002666653331`, 
                    2.0999200006666645`}, {-2., 2.1}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Dashing[{Small, Small}], 
                   Arrowheads[0.02], 
                   CircleBox[{-2., 0.5}, 1.7, {1.5707963267948966`, 
                    1.5707963267948966`}], 
                   ArrowBox[{{-1.9830002833319165`, 2.199915000708331}, {-2., 
                    2.2}}]}, 
                  {RGBColor[0, 0, 1], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-2., 0.5}, 1.8, {1.5707963267948966`, 
                    1.5707963267948966`}], 
                   ArrowBox[{{-1.9820002999984998`, 
                    2.2999100007499975`}, {-2., 2.3}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.003], Arrowheads[0.025], 
                   CircleBox[{-2., 0.5}, 1.5, {1.5707963267948966`, 
                    1.5707963267948966`}], 
                   ArrowBox[{{-1.9850002499987498`, 1.999925000624998}, {-2., 
                    2.}}]}, 
                  {RGBColor[0, 0, 1], ArrowBox[{{0, 2.8}, {0, 2.8}}]}},
                 Frame->True,
                 FrameTicks->Automatic,
                 ImageSize->800,
                 PlotRange->{{-3, 3}, {-1, 3}}], 7->
                GraphicsBox[{InsetBox[
                   TagBox[GridBox[{
                    {"\<\"t\"\>", "0.3`"},
                    {"\<\"driveIntent\"\>", "0"},
                    {"\<\"motorTorque\"\>", "0.`"},
                    {"\<\"vw\"\>", "0.`"},
                    {"\<\"controllerState\"\>", 
                    RowBox[{"\[LeftAssociation]", 
                    RowBox[{
                    RowBox[{"\<\"lastError\"\>", "\[Rule]", "0.`"}], ",", 
                    RowBox[{"\<\"lastInt\"\>", "\[Rule]", "0.`"}], ",", 
                    RowBox[{"\<\"initialized\"\>", "\[Rule]", "True"}]}], 
                    "\[RightAssociation]"}]}
                    },
                    AutoDelete->False,
                    
                    GridBoxItemSize->{
                    "Columns" -> {{Automatic}}, "Rows" -> {{Automatic}}}],
                    "Grid"], {-1, -0.5}], 
                  {GrayLevel[0], LineBox[{{-5, 0}, {5, 0}}]}, 
                  {GrayLevel[0.5], CircleBox[{-2., 0.5}, 0.5], 
                   DiskBox[{-2., 0.5}, 0.1]}, 
                  {GrayLevel[0.5], DiskBox[{-2., 2.}, 0.1]}, 
                  {GrayLevel[0.5], Thickness[0.01], 
                   LineBox[{{-2., 0.5}, {-2., 2.}}]}, 
                  {RGBColor[0, 1, 0], DiskBox[{-2., 1.5}, 0.05]}, 
                  {RGBColor[0, 0, 1], Arrowheads[0.01], 
                   CircleBox[{-2., 0.5}, 0.3333333333333333, 
                    NCache[{
                    Rational[-1, 2] 
                    Pi, -1.5707963267948966`}, {-1.5707963267948966`, \
-1.5707963267948966`}]], 
                   ArrowBox[{{-2.0033332777780557`, 
                    0.16668333319444495`}, {-2., 0.16666666666666669`}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-2., 0.5}, {-2., 0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-2., 0.5}, {-2., 0.5}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.005], Arrowheads[0.04], 
                   ArrowBox[{{-2., 0.5}, {-2., 0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-2., 0.5}, 1.6, {1.5707963267948966`, 
                    1.5707963267948966`}], 
                   ArrowBox[{{-1.9840002666653331`, 
                    2.0999200006666645`}, {-2., 2.1}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Dashing[{Small, Small}], 
                   Arrowheads[0.02], 
                   CircleBox[{-2., 0.5}, 1.7, {1.5707963267948966`, 
                    1.5707963267948966`}], 
                   ArrowBox[{{-1.9830002833319165`, 2.199915000708331}, {-2., 
                    2.2}}]}, 
                  {RGBColor[0, 0, 1], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-2., 0.5}, 1.8, {1.5707963267948966`, 
                    1.5707963267948966`}], 
                   ArrowBox[{{-1.9820002999984998`, 
                    2.2999100007499975`}, {-2., 2.3}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.003], Arrowheads[0.025], 
                   CircleBox[{-2., 0.5}, 1.5, {1.5707963267948966`, 
                    1.5707963267948966`}], 
                   ArrowBox[{{-1.9850002499987498`, 1.999925000624998}, {-2., 
                    2.}}]}, 
                  {RGBColor[0, 0, 1], ArrowBox[{{0, 2.8}, {0, 2.8}}]}},
                 Frame->True,
                 FrameTicks->Automatic,
                 ImageSize->800,
                 PlotRange->{{-3, 3}, {-1, 3}}], 8->
                GraphicsBox[{InsetBox[
                   TagBox[GridBox[{
                    {"\<\"t\"\>", "0.35000000000000003`"},
                    {"\<\"driveIntent\"\>", "0"},
                    {"\<\"motorTorque\"\>", "0.`"},
                    {"\<\"vw\"\>", "0.`"},
                    {"\<\"controllerState\"\>", 
                    RowBox[{"\[LeftAssociation]", 
                    RowBox[{
                    RowBox[{"\<\"lastError\"\>", "\[Rule]", "0.`"}], ",", 
                    RowBox[{"\<\"lastInt\"\>", "\[Rule]", "0.`"}], ",", 
                    RowBox[{"\<\"initialized\"\>", "\[Rule]", "True"}]}], 
                    "\[RightAssociation]"}]}
                    },
                    AutoDelete->False,
                    
                    GridBoxItemSize->{
                    "Columns" -> {{Automatic}}, "Rows" -> {{Automatic}}}],
                    "Grid"], {-1, -0.5}], 
                  {GrayLevel[0], LineBox[{{-5, 0}, {5, 0}}]}, 
                  {GrayLevel[0.5], CircleBox[{-2., 0.5}, 0.5], 
                   DiskBox[{-2., 0.5}, 0.1]}, 
                  {GrayLevel[0.5], DiskBox[{-2., 2.}, 0.1]}, 
                  {GrayLevel[0.5], Thickness[0.01], 
                   LineBox[{{-2., 0.5}, {-2., 2.}}]}, 
                  {RGBColor[0, 1, 0], DiskBox[{-2., 1.5}, 0.05]}, 
                  {RGBColor[0, 0, 1], Arrowheads[0.01], 
                   CircleBox[{-2., 0.5}, 0.3333333333333333, 
                    NCache[{
                    Rational[-1, 2] 
                    Pi, -1.5707963267948966`}, {-1.5707963267948966`, \
-1.5707963267948966`}]], 
                   ArrowBox[{{-2.0033332777780557`, 
                    0.16668333319444495`}, {-2., 0.16666666666666669`}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-2., 0.5}, {-2., 0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-2., 0.5}, {-2., 0.5}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.005], Arrowheads[0.04], 
                   ArrowBox[{{-2., 0.5}, {-2., 0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-2., 0.5}, 1.6, {1.5707963267948966`, 
                    1.5707963267948966`}], 
                   ArrowBox[{{-1.9840002666653331`, 
                    2.0999200006666645`}, {-2., 2.1}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Dashing[{Small, Small}], 
                   Arrowheads[0.02], 
                   CircleBox[{-2., 0.5}, 1.7, {1.5707963267948966`, 
                    1.5707963267948966`}], 
                   ArrowBox[{{-1.9830002833319165`, 2.199915000708331}, {-2., 
                    2.2}}]}, 
                  {RGBColor[0, 0, 1], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-2., 0.5}, 1.8, {1.5707963267948966`, 
                    1.5707963267948966`}], 
                   ArrowBox[{{-1.9820002999984998`, 
                    2.2999100007499975`}, {-2., 2.3}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.003], Arrowheads[0.025], 
                   CircleBox[{-2., 0.5}, 1.5, {1.5707963267948966`, 
                    1.5707963267948966`}], 
                   ArrowBox[{{-1.9850002499987498`, 1.999925000624998}, {-2., 
                    2.}}]}, 
                  {RGBColor[0, 0, 1], ArrowBox[{{0, 2.8}, {0, 2.8}}]}},
                 Frame->True,
                 FrameTicks->Automatic,
                 ImageSize->800,
                 PlotRange->{{-3, 3}, {-1, 3}}], 9->
                GraphicsBox[{InsetBox[
                   TagBox[GridBox[{
                    {"\<\"t\"\>", "0.4`"},
                    {"\<\"driveIntent\"\>", "0"},
                    {"\<\"motorTorque\"\>", "0.`"},
                    {"\<\"vw\"\>", "0.`"},
                    {"\<\"controllerState\"\>", 
                    RowBox[{"\[LeftAssociation]", 
                    RowBox[{
                    RowBox[{"\<\"lastError\"\>", "\[Rule]", "0.`"}], ",", 
                    RowBox[{"\<\"lastInt\"\>", "\[Rule]", "0.`"}], ",", 
                    RowBox[{"\<\"initialized\"\>", "\[Rule]", "True"}]}], 
                    "\[RightAssociation]"}]}
                    },
                    AutoDelete->False,
                    
                    GridBoxItemSize->{
                    "Columns" -> {{Automatic}}, "Rows" -> {{Automatic}}}],
                    "Grid"], {-1, -0.5}], 
                  {GrayLevel[0], LineBox[{{-5, 0}, {5, 0}}]}, 
                  {GrayLevel[0.5], CircleBox[{-2., 0.5}, 0.5], 
                   DiskBox[{-2., 0.5}, 0.1]}, 
                  {GrayLevel[0.5], DiskBox[{-2., 2.}, 0.1]}, 
                  {GrayLevel[0.5], Thickness[0.01], 
                   LineBox[{{-2., 0.5}, {-2., 2.}}]}, 
                  {RGBColor[0, 1, 0], DiskBox[{-2., 1.5}, 0.05]}, 
                  {RGBColor[0, 0, 1], Arrowheads[0.01], 
                   CircleBox[{-2., 0.5}, 0.3333333333333333, 
                    NCache[{
                    Rational[-1, 2] 
                    Pi, -1.5707963267948966`}, {-1.5707963267948966`, \
-1.5707963267948966`}]], 
                   ArrowBox[{{-2.0033332777780557`, 
                    0.16668333319444495`}, {-2., 0.16666666666666669`}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-2., 0.5}, {-2., 0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-2., 0.5}, {-2., 0.5}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.005], Arrowheads[0.04], 
                   ArrowBox[{{-2., 0.5}, {-2., 0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-2., 0.5}, 1.6, {1.5707963267948966`, 
                    1.5707963267948966`}], 
                   ArrowBox[{{-1.9840002666653331`, 
                    2.0999200006666645`}, {-2., 2.1}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Dashing[{Small, Small}], 
                   Arrowheads[0.02], 
                   CircleBox[{-2., 0.5}, 1.7, {1.5707963267948966`, 
                    1.5707963267948966`}], 
                   ArrowBox[{{-1.9830002833319165`, 2.199915000708331}, {-2., 
                    2.2}}]}, 
                  {RGBColor[0, 0, 1], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-2., 0.5}, 1.8, {1.5707963267948966`, 
                    1.5707963267948966`}], 
                   ArrowBox[{{-1.9820002999984998`, 
                    2.2999100007499975`}, {-2., 2.3}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.003], Arrowheads[0.025], 
                   CircleBox[{-2., 0.5}, 1.5, {1.5707963267948966`, 
                    1.5707963267948966`}], 
                   ArrowBox[{{-1.9850002499987498`, 1.999925000624998}, {-2., 
                    2.}}]}, 
                  {RGBColor[0, 0, 1], ArrowBox[{{0, 2.8}, {0, 2.8}}]}},
                 Frame->True,
                 FrameTicks->Automatic,
                 ImageSize->800,
                 PlotRange->{{-3, 3}, {-1, 3}}], 10->
                GraphicsBox[{InsetBox[
                   TagBox[GridBox[{
                    {"\<\"t\"\>", "0.45`"},
                    {"\<\"driveIntent\"\>", "0"},
                    {"\<\"motorTorque\"\>", "0.`"},
                    {"\<\"vw\"\>", "0.`"},
                    {"\<\"controllerState\"\>", 
                    RowBox[{"\[LeftAssociation]", 
                    RowBox[{
                    RowBox[{"\<\"lastError\"\>", "\[Rule]", "0.`"}], ",", 
                    RowBox[{"\<\"lastInt\"\>", "\[Rule]", "0.`"}], ",", 
                    RowBox[{"\<\"initialized\"\>", "\[Rule]", "True"}]}], 
                    "\[RightAssociation]"}]}
                    },
                    AutoDelete->False,
                    
                    GridBoxItemSize->{
                    "Columns" -> {{Automatic}}, "Rows" -> {{Automatic}}}],
                    "Grid"], {-1, -0.5}], 
                  {GrayLevel[0], LineBox[{{-5, 0}, {5, 0}}]}, 
                  {GrayLevel[0.5], CircleBox[{-2., 0.5}, 0.5], 
                   DiskBox[{-2., 0.5}, 0.1]}, 
                  {GrayLevel[0.5], DiskBox[{-2., 2.}, 0.1]}, 
                  {GrayLevel[0.5], Thickness[0.01], 
                   LineBox[{{-2., 0.5}, {-2., 2.}}]}, 
                  {RGBColor[0, 1, 0], DiskBox[{-2., 1.5}, 0.05]}, 
                  {RGBColor[0, 0, 1], Arrowheads[0.01], 
                   CircleBox[{-2., 0.5}, 0.3333333333333333, 
                    NCache[{
                    Rational[-1, 2] 
                    Pi, -1.5707963267948966`}, {-1.5707963267948966`, \
-1.5707963267948966`}]], 
                   ArrowBox[{{-2.0033332777780557`, 
                    0.16668333319444495`}, {-2., 0.16666666666666669`}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-2., 0.5}, {-2., 0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-2., 0.5}, {-2., 0.5}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.005], Arrowheads[0.04], 
                   ArrowBox[{{-2., 0.5}, {-2., 0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-2., 0.5}, 1.6, {1.5707963267948966`, 
                    1.5707963267948966`}], 
                   ArrowBox[{{-1.9840002666653331`, 
                    2.0999200006666645`}, {-2., 2.1}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Dashing[{Small, Small}], 
                   Arrowheads[0.02], 
                   CircleBox[{-2., 0.5}, 1.7, {1.5707963267948966`, 
                    1.5707963267948966`}], 
                   ArrowBox[{{-1.9830002833319165`, 2.199915000708331}, {-2., 
                    2.2}}]}, 
                  {RGBColor[0, 0, 1], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-2., 0.5}, 1.8, {1.5707963267948966`, 
                    1.5707963267948966`}], 
                   ArrowBox[{{-1.9820002999984998`, 
                    2.2999100007499975`}, {-2., 2.3}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.003], Arrowheads[0.025], 
                   CircleBox[{-2., 0.5}, 1.5, {1.5707963267948966`, 
                    1.5707963267948966`}], 
                   ArrowBox[{{-1.9850002499987498`, 1.999925000624998}, {-2., 
                    2.}}]}, 
                  {RGBColor[0, 0, 1], ArrowBox[{{0, 2.8}, {0, 2.8}}]}},
                 Frame->True,
                 FrameTicks->Automatic,
                 ImageSize->800,
                 PlotRange->{{-3, 3}, {-1, 3}}], 11->
                GraphicsBox[{InsetBox[
                   TagBox[GridBox[{
                    {"\<\"t\"\>", "0.5`"},
                    {"\<\"driveIntent\"\>", "0"},
                    {"\<\"motorTorque\"\>", "0.`"},
                    {"\<\"vw\"\>", "0.`"},
                    {"\<\"controllerState\"\>", 
                    RowBox[{"\[LeftAssociation]", 
                    RowBox[{
                    RowBox[{"\<\"lastError\"\>", "\[Rule]", "0.`"}], ",", 
                    RowBox[{"\<\"lastInt\"\>", "\[Rule]", "0.`"}], ",", 
                    RowBox[{"\<\"initialized\"\>", "\[Rule]", "True"}]}], 
                    "\[RightAssociation]"}]}
                    },
                    AutoDelete->False,
                    
                    GridBoxItemSize->{
                    "Columns" -> {{Automatic}}, "Rows" -> {{Automatic}}}],
                    "Grid"], {-1, -0.5}], 
                  {GrayLevel[0], LineBox[{{-5, 0}, {5, 0}}]}, 
                  {GrayLevel[0.5], CircleBox[{-2., 0.5}, 0.5], 
                   DiskBox[{-2., 0.5}, 0.1]}, 
                  {GrayLevel[0.5], DiskBox[{-2., 2.}, 0.1]}, 
                  {GrayLevel[0.5], Thickness[0.01], 
                   LineBox[{{-2., 0.5}, {-2., 2.}}]}, 
                  {RGBColor[0, 1, 0], DiskBox[{-2., 1.5}, 0.05]}, 
                  {RGBColor[0, 0, 1], Arrowheads[0.01], 
                   CircleBox[{-2., 0.5}, 0.3333333333333333, 
                    NCache[{
                    Rational[-1, 2] 
                    Pi, -1.5707963267948966`}, {-1.5707963267948966`, \
-1.5707963267948966`}]], 
                   ArrowBox[{{-2.0033332777780557`, 
                    0.16668333319444495`}, {-2., 0.16666666666666669`}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-2., 0.5}, {-2., 0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-2., 0.5}, {-2., 0.5}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.005], Arrowheads[0.04], 
                   ArrowBox[{{-2., 0.5}, {-2., 0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-2., 0.5}, 1.6, {1.5707963267948966`, 
                    1.5707963267948966`}], 
                   ArrowBox[{{-1.9840002666653331`, 
                    2.0999200006666645`}, {-2., 2.1}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Dashing[{Small, Small}], 
                   Arrowheads[0.02], 
                   CircleBox[{-2., 0.5}, 1.7, {1.5707963267948966`, 
                    1.5707963267948966`}], 
                   ArrowBox[{{-1.9830002833319165`, 2.199915000708331}, {-2., 
                    2.2}}]}, 
                  {RGBColor[0, 0, 1], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-2., 0.5}, 1.8, {1.5707963267948966`, 
                    1.5707963267948966`}], 
                   ArrowBox[{{-1.9820002999984998`, 
                    2.2999100007499975`}, {-2., 2.3}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.003], Arrowheads[0.025], 
                   CircleBox[{-2., 0.5}, 1.5, {1.5707963267948966`, 
                    1.5707963267948966`}], 
                   ArrowBox[{{-1.9850002499987498`, 1.999925000624998}, {-2., 
                    2.}}]}, 
                  {RGBColor[0, 0, 1], ArrowBox[{{0, 2.8}, {0, 2.8}}]}},
                 Frame->True,
                 FrameTicks->Automatic,
                 ImageSize->800,
                 PlotRange->{{-3, 3}, {-1, 3}}], 12->
                GraphicsBox[{InsetBox[
                   TagBox[GridBox[{
                    {"\<\"t\"\>", "0.55`"},
                    {"\<\"driveIntent\"\>", "0"},
                    {"\<\"motorTorque\"\>", "0.`"},
                    {"\<\"vw\"\>", "0.`"},
                    {"\<\"controllerState\"\>", 
                    RowBox[{"\[LeftAssociation]", 
                    RowBox[{
                    RowBox[{"\<\"lastError\"\>", "\[Rule]", "0.`"}], ",", 
                    RowBox[{"\<\"lastInt\"\>", "\[Rule]", "0.`"}], ",", 
                    RowBox[{"\<\"initialized\"\>", "\[Rule]", "True"}]}], 
                    "\[RightAssociation]"}]}
                    },
                    AutoDelete->False,
                    
                    GridBoxItemSize->{
                    "Columns" -> {{Automatic}}, "Rows" -> {{Automatic}}}],
                    "Grid"], {-1, -0.5}], 
                  {GrayLevel[0], LineBox[{{-5, 0}, {5, 0}}]}, 
                  {GrayLevel[0.5], CircleBox[{-2., 0.5}, 0.5], 
                   DiskBox[{-2., 0.5}, 0.1]}, 
                  {GrayLevel[0.5], DiskBox[{-2., 2.}, 0.1]}, 
                  {GrayLevel[0.5], Thickness[0.01], 
                   LineBox[{{-2., 0.5}, {-2., 2.}}]}, 
                  {RGBColor[0, 1, 0], DiskBox[{-2., 1.5}, 0.05]}, 
                  {RGBColor[0, 0, 1], Arrowheads[0.01], 
                   CircleBox[{-2., 0.5}, 0.3333333333333333, 
                    NCache[{
                    Rational[-1, 2] 
                    Pi, -1.5707963267948966`}, {-1.5707963267948966`, \
-1.5707963267948966`}]], 
                   ArrowBox[{{-2.0033332777780557`, 
                    0.16668333319444495`}, {-2., 0.16666666666666669`}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-2., 0.5}, {-2., 0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-2., 0.5}, {-2., 0.5}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.005], Arrowheads[0.04], 
                   ArrowBox[{{-2., 0.5}, {-2., 0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-2., 0.5}, 1.6, {1.5707963267948966`, 
                    1.5707963267948966`}], 
                   ArrowBox[{{-1.9840002666653331`, 
                    2.0999200006666645`}, {-2., 2.1}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Dashing[{Small, Small}], 
                   Arrowheads[0.02], 
                   CircleBox[{-2., 0.5}, 1.7, {1.5707963267948966`, 
                    1.5707963267948966`}], 
                   ArrowBox[{{-1.9830002833319165`, 2.199915000708331}, {-2., 
                    2.2}}]}, 
                  {RGBColor[0, 0, 1], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-2., 0.5}, 1.8, {1.5707963267948966`, 
                    1.5707963267948966`}], 
                   ArrowBox[{{-1.9820002999984998`, 
                    2.2999100007499975`}, {-2., 2.3}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.003], Arrowheads[0.025], 
                   CircleBox[{-2., 0.5}, 1.5, {1.5707963267948966`, 
                    1.5707963267948966`}], 
                   ArrowBox[{{-1.9850002499987498`, 1.999925000624998}, {-2., 
                    2.}}]}, 
                  {RGBColor[0, 0, 1], ArrowBox[{{0, 2.8}, {0, 2.8}}]}},
                 Frame->True,
                 FrameTicks->Automatic,
                 ImageSize->800,
                 PlotRange->{{-3, 3}, {-1, 3}}], 13->
                GraphicsBox[{InsetBox[
                   TagBox[GridBox[{
                    {"\<\"t\"\>", "0.6`"},
                    {"\<\"driveIntent\"\>", "0"},
                    {"\<\"motorTorque\"\>", "0.`"},
                    {"\<\"vw\"\>", "0.`"},
                    {"\<\"controllerState\"\>", 
                    RowBox[{"\[LeftAssociation]", 
                    RowBox[{
                    RowBox[{"\<\"lastError\"\>", "\[Rule]", "0.`"}], ",", 
                    RowBox[{"\<\"lastInt\"\>", "\[Rule]", "0.`"}], ",", 
                    RowBox[{"\<\"initialized\"\>", "\[Rule]", "True"}]}], 
                    "\[RightAssociation]"}]}
                    },
                    AutoDelete->False,
                    
                    GridBoxItemSize->{
                    "Columns" -> {{Automatic}}, "Rows" -> {{Automatic}}}],
                    "Grid"], {-1, -0.5}], 
                  {GrayLevel[0], LineBox[{{-5, 0}, {5, 0}}]}, 
                  {GrayLevel[0.5], CircleBox[{-2., 0.5}, 0.5], 
                   DiskBox[{-2., 0.5}, 0.1]}, 
                  {GrayLevel[0.5], DiskBox[{-2., 2.}, 0.1]}, 
                  {GrayLevel[0.5], Thickness[0.01], 
                   LineBox[{{-2., 0.5}, {-2., 2.}}]}, 
                  {RGBColor[0, 1, 0], DiskBox[{-2., 1.5}, 0.05]}, 
                  {RGBColor[0, 0, 1], Arrowheads[0.01], 
                   CircleBox[{-2., 0.5}, 0.3333333333333333, 
                    NCache[{
                    Rational[-1, 2] 
                    Pi, -1.5707963267948966`}, {-1.5707963267948966`, \
-1.5707963267948966`}]], 
                   ArrowBox[{{-2.0033332777780557`, 
                    0.16668333319444495`}, {-2., 0.16666666666666669`}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-2., 0.5}, {-2., 0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-2., 0.5}, {-2., 0.5}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.005], Arrowheads[0.04], 
                   ArrowBox[{{-2., 0.5}, {-2., 0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-2., 0.5}, 1.6, {1.5707963267948966`, 
                    1.5707963267948966`}], 
                   ArrowBox[{{-1.9840002666653331`, 
                    2.0999200006666645`}, {-2., 2.1}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Dashing[{Small, Small}], 
                   Arrowheads[0.02], 
                   CircleBox[{-2., 0.5}, 1.7, {1.5707963267948966`, 
                    1.5707963267948966`}], 
                   ArrowBox[{{-1.9830002833319165`, 2.199915000708331}, {-2., 
                    2.2}}]}, 
                  {RGBColor[0, 0, 1], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-2., 0.5}, 1.8, {1.5707963267948966`, 
                    1.5707963267948966`}], 
                   ArrowBox[{{-1.9820002999984998`, 
                    2.2999100007499975`}, {-2., 2.3}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.003], Arrowheads[0.025], 
                   CircleBox[{-2., 0.5}, 1.5, {1.5707963267948966`, 
                    1.5707963267948966`}], 
                   ArrowBox[{{-1.9850002499987498`, 1.999925000624998}, {-2., 
                    2.}}]}, 
                  {RGBColor[0, 0, 1], ArrowBox[{{0, 2.8}, {0, 2.8}}]}},
                 Frame->True,
                 FrameTicks->Automatic,
                 ImageSize->800,
                 PlotRange->{{-3, 3}, {-1, 3}}], 14->
                GraphicsBox[{InsetBox[
                   TagBox[GridBox[{
                    {"\<\"t\"\>", "0.65`"},
                    {"\<\"driveIntent\"\>", "0"},
                    {"\<\"motorTorque\"\>", "0.`"},
                    {"\<\"vw\"\>", "0.`"},
                    {"\<\"controllerState\"\>", 
                    RowBox[{"\[LeftAssociation]", 
                    RowBox[{
                    RowBox[{"\<\"lastError\"\>", "\[Rule]", "0.`"}], ",", 
                    RowBox[{"\<\"lastInt\"\>", "\[Rule]", "0.`"}], ",", 
                    RowBox[{"\<\"initialized\"\>", "\[Rule]", "True"}]}], 
                    "\[RightAssociation]"}]}
                    },
                    AutoDelete->False,
                    
                    GridBoxItemSize->{
                    "Columns" -> {{Automatic}}, "Rows" -> {{Automatic}}}],
                    "Grid"], {-1, -0.5}], 
                  {GrayLevel[0], LineBox[{{-5, 0}, {5, 0}}]}, 
                  {GrayLevel[0.5], CircleBox[{-2., 0.5}, 0.5], 
                   DiskBox[{-2., 0.5}, 0.1]}, 
                  {GrayLevel[0.5], DiskBox[{-2., 2.}, 0.1]}, 
                  {GrayLevel[0.5], Thickness[0.01], 
                   LineBox[{{-2., 0.5}, {-2., 2.}}]}, 
                  {RGBColor[0, 1, 0], DiskBox[{-2., 1.5}, 0.05]}, 
                  {RGBColor[0, 0, 1], Arrowheads[0.01], 
                   CircleBox[{-2., 0.5}, 0.3333333333333333, 
                    NCache[{
                    Rational[-1, 2] 
                    Pi, -1.5707963267948966`}, {-1.5707963267948966`, \
-1.5707963267948966`}]], 
                   ArrowBox[{{-2.0033332777780557`, 
                    0.16668333319444495`}, {-2., 0.16666666666666669`}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-2., 0.5}, {-2., 0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-2., 0.5}, {-2., 0.5}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.005], Arrowheads[0.04], 
                   ArrowBox[{{-2., 0.5}, {-2., 0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-2., 0.5}, 1.6, {1.5707963267948966`, 
                    1.5707963267948966`}], 
                   ArrowBox[{{-1.9840002666653331`, 
                    2.0999200006666645`}, {-2., 2.1}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Dashing[{Small, Small}], 
                   Arrowheads[0.02], 
                   CircleBox[{-2., 0.5}, 1.7, {1.5707963267948966`, 
                    1.5707963267948966`}], 
                   ArrowBox[{{-1.9830002833319165`, 2.199915000708331}, {-2., 
                    2.2}}]}, 
                  {RGBColor[0, 0, 1], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-2., 0.5}, 1.8, {1.5707963267948966`, 
                    1.5707963267948966`}], 
                   ArrowBox[{{-1.9820002999984998`, 
                    2.2999100007499975`}, {-2., 2.3}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.003], Arrowheads[0.025], 
                   CircleBox[{-2., 0.5}, 1.5, {1.5707963267948966`, 
                    1.5707963267948966`}], 
                   ArrowBox[{{-1.9850002499987498`, 1.999925000624998}, {-2., 
                    2.}}]}, 
                  {RGBColor[0, 0, 1], ArrowBox[{{0, 2.8}, {0, 2.8}}]}},
                 Frame->True,
                 FrameTicks->Automatic,
                 ImageSize->800,
                 PlotRange->{{-3, 3}, {-1, 3}}], 15->
                GraphicsBox[{InsetBox[
                   TagBox[GridBox[{
                    {"\<\"t\"\>", "0.7000000000000001`"},
                    {"\<\"driveIntent\"\>", "0"},
                    {"\<\"motorTorque\"\>", "0.`"},
                    {"\<\"vw\"\>", "0.`"},
                    {"\<\"controllerState\"\>", 
                    RowBox[{"\[LeftAssociation]", 
                    RowBox[{
                    RowBox[{"\<\"lastError\"\>", "\[Rule]", "0.`"}], ",", 
                    RowBox[{"\<\"lastInt\"\>", "\[Rule]", "0.`"}], ",", 
                    RowBox[{"\<\"initialized\"\>", "\[Rule]", "True"}]}], 
                    "\[RightAssociation]"}]}
                    },
                    AutoDelete->False,
                    
                    GridBoxItemSize->{
                    "Columns" -> {{Automatic}}, "Rows" -> {{Automatic}}}],
                    "Grid"], {-1, -0.5}], 
                  {GrayLevel[0], LineBox[{{-5, 0}, {5, 0}}]}, 
                  {GrayLevel[0.5], CircleBox[{-2., 0.5}, 0.5], 
                   DiskBox[{-2., 0.5}, 0.1]}, 
                  {GrayLevel[0.5], DiskBox[{-2., 2.}, 0.1]}, 
                  {GrayLevel[0.5], Thickness[0.01], 
                   LineBox[{{-2., 0.5}, {-2., 2.}}]}, 
                  {RGBColor[0, 1, 0], DiskBox[{-2., 1.5}, 0.05]}, 
                  {RGBColor[0, 0, 1], Arrowheads[0.01], 
                   CircleBox[{-2., 0.5}, 0.3333333333333333, 
                    NCache[{
                    Rational[-1, 2] 
                    Pi, -1.5707963267948966`}, {-1.5707963267948966`, \
-1.5707963267948966`}]], 
                   ArrowBox[{{-2.0033332777780557`, 
                    0.16668333319444495`}, {-2., 0.16666666666666669`}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-2., 0.5}, {-2., 0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-2., 0.5}, {-2., 0.5}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.005], Arrowheads[0.04], 
                   ArrowBox[{{-2., 0.5}, {-2., 0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-2., 0.5}, 1.6, {1.5707963267948966`, 
                    1.5707963267948966`}], 
                   ArrowBox[{{-1.9840002666653331`, 
                    2.0999200006666645`}, {-2., 2.1}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Dashing[{Small, Small}], 
                   Arrowheads[0.02], 
                   CircleBox[{-2., 0.5}, 1.7, {1.5707963267948966`, 
                    1.5707963267948966`}], 
                   ArrowBox[{{-1.9830002833319165`, 2.199915000708331}, {-2., 
                    2.2}}]}, 
                  {RGBColor[0, 0, 1], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-2., 0.5}, 1.8, {1.5707963267948966`, 
                    1.5707963267948966`}], 
                   ArrowBox[{{-1.9820002999984998`, 
                    2.2999100007499975`}, {-2., 2.3}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.003], Arrowheads[0.025], 
                   CircleBox[{-2., 0.5}, 1.5, {1.5707963267948966`, 
                    1.5707963267948966`}], 
                   ArrowBox[{{-1.9850002499987498`, 1.999925000624998}, {-2., 
                    2.}}]}, 
                  {RGBColor[0, 0, 1], ArrowBox[{{0, 2.8}, {0, 2.8}}]}},
                 Frame->True,
                 FrameTicks->Automatic,
                 ImageSize->800,
                 PlotRange->{{-3, 3}, {-1, 3}}], 16->
                GraphicsBox[{InsetBox[
                   TagBox[GridBox[{
                    {"\<\"t\"\>", "0.75`"},
                    {"\<\"driveIntent\"\>", "0"},
                    {"\<\"motorTorque\"\>", "0.`"},
                    {"\<\"vw\"\>", "0.`"},
                    {"\<\"controllerState\"\>", 
                    RowBox[{"\[LeftAssociation]", 
                    RowBox[{
                    RowBox[{"\<\"lastError\"\>", "\[Rule]", "0.`"}], ",", 
                    RowBox[{"\<\"lastInt\"\>", "\[Rule]", "0.`"}], ",", 
                    RowBox[{"\<\"initialized\"\>", "\[Rule]", "True"}]}], 
                    "\[RightAssociation]"}]}
                    },
                    AutoDelete->False,
                    
                    GridBoxItemSize->{
                    "Columns" -> {{Automatic}}, "Rows" -> {{Automatic}}}],
                    "Grid"], {-1, -0.5}], 
                  {GrayLevel[0], LineBox[{{-5, 0}, {5, 0}}]}, 
                  {GrayLevel[0.5], CircleBox[{-2., 0.5}, 0.5], 
                   DiskBox[{-2., 0.5}, 0.1]}, 
                  {GrayLevel[0.5], DiskBox[{-2., 2.}, 0.1]}, 
                  {GrayLevel[0.5], Thickness[0.01], 
                   LineBox[{{-2., 0.5}, {-2., 2.}}]}, 
                  {RGBColor[0, 1, 0], DiskBox[{-2., 1.5}, 0.05]}, 
                  {RGBColor[0, 0, 1], Arrowheads[0.01], 
                   CircleBox[{-2., 0.5}, 0.3333333333333333, 
                    NCache[{
                    Rational[-1, 2] 
                    Pi, -1.5707963267948966`}, {-1.5707963267948966`, \
-1.5707963267948966`}]], 
                   ArrowBox[{{-2.0033332777780557`, 
                    0.16668333319444495`}, {-2., 0.16666666666666669`}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-2., 0.5}, {-2., 0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-2., 0.5}, {-2., 0.5}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.005], Arrowheads[0.04], 
                   ArrowBox[{{-2., 0.5}, {-2., 0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-2., 0.5}, 1.6, {1.5707963267948966`, 
                    1.5707963267948966`}], 
                   ArrowBox[{{-1.9840002666653331`, 
                    2.0999200006666645`}, {-2., 2.1}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Dashing[{Small, Small}], 
                   Arrowheads[0.02], 
                   CircleBox[{-2., 0.5}, 1.7, {1.5707963267948966`, 
                    1.5707963267948966`}], 
                   ArrowBox[{{-1.9830002833319165`, 2.199915000708331}, {-2., 
                    2.2}}]}, 
                  {RGBColor[0, 0, 1], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-2., 0.5}, 1.8, {1.5707963267948966`, 
                    1.5707963267948966`}], 
                   ArrowBox[{{-1.9820002999984998`, 
                    2.2999100007499975`}, {-2., 2.3}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.003], Arrowheads[0.025], 
                   CircleBox[{-2., 0.5}, 1.5, {1.5707963267948966`, 
                    1.5707963267948966`}], 
                   ArrowBox[{{-1.9850002499987498`, 1.999925000624998}, {-2., 
                    2.}}]}, 
                  {RGBColor[0, 0, 1], ArrowBox[{{0, 2.8}, {0, 2.8}}]}},
                 Frame->True,
                 FrameTicks->Automatic,
                 ImageSize->800,
                 PlotRange->{{-3, 3}, {-1, 3}}], 17->
                GraphicsBox[{InsetBox[
                   TagBox[GridBox[{
                    {"\<\"t\"\>", "0.8`"},
                    {"\<\"driveIntent\"\>", "0"},
                    {"\<\"motorTorque\"\>", "0.`"},
                    {"\<\"vw\"\>", "0.`"},
                    {"\<\"controllerState\"\>", 
                    RowBox[{"\[LeftAssociation]", 
                    RowBox[{
                    RowBox[{"\<\"lastError\"\>", "\[Rule]", "0.`"}], ",", 
                    RowBox[{"\<\"lastInt\"\>", "\[Rule]", "0.`"}], ",", 
                    RowBox[{"\<\"initialized\"\>", "\[Rule]", "True"}]}], 
                    "\[RightAssociation]"}]}
                    },
                    AutoDelete->False,
                    
                    GridBoxItemSize->{
                    "Columns" -> {{Automatic}}, "Rows" -> {{Automatic}}}],
                    "Grid"], {-1, -0.5}], 
                  {GrayLevel[0], LineBox[{{-5, 0}, {5, 0}}]}, 
                  {GrayLevel[0.5], CircleBox[{-2., 0.5}, 0.5], 
                   DiskBox[{-2., 0.5}, 0.1]}, 
                  {GrayLevel[0.5], DiskBox[{-2., 2.}, 0.1]}, 
                  {GrayLevel[0.5], Thickness[0.01], 
                   LineBox[{{-2., 0.5}, {-2., 2.}}]}, 
                  {RGBColor[0, 1, 0], DiskBox[{-2., 1.5}, 0.05]}, 
                  {RGBColor[0, 0, 1], Arrowheads[0.01], 
                   CircleBox[{-2., 0.5}, 0.3333333333333333, 
                    NCache[{
                    Rational[-1, 2] 
                    Pi, -1.5707963267948966`}, {-1.5707963267948966`, \
-1.5707963267948966`}]], 
                   ArrowBox[{{-2.0033332777780557`, 
                    0.16668333319444495`}, {-2., 0.16666666666666669`}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-2., 0.5}, {-2., 0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-2., 0.5}, {-2., 0.5}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.005], Arrowheads[0.04], 
                   ArrowBox[{{-2., 0.5}, {-2., 0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-2., 0.5}, 1.6, {1.5707963267948966`, 
                    1.5707963267948966`}], 
                   ArrowBox[{{-1.9840002666653331`, 
                    2.0999200006666645`}, {-2., 2.1}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Dashing[{Small, Small}], 
                   Arrowheads[0.02], 
                   CircleBox[{-2., 0.5}, 1.7, {1.5707963267948966`, 
                    1.5707963267948966`}], 
                   ArrowBox[{{-1.9830002833319165`, 2.199915000708331}, {-2., 
                    2.2}}]}, 
                  {RGBColor[0, 0, 1], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-2., 0.5}, 1.8, {1.5707963267948966`, 
                    1.5707963267948966`}], 
                   ArrowBox[{{-1.9820002999984998`, 
                    2.2999100007499975`}, {-2., 2.3}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.003], Arrowheads[0.025], 
                   CircleBox[{-2., 0.5}, 1.5, {1.5707963267948966`, 
                    1.5707963267948966`}], 
                   ArrowBox[{{-1.9850002499987498`, 1.999925000624998}, {-2., 
                    2.}}]}, 
                  {RGBColor[0, 0, 1], ArrowBox[{{0, 2.8}, {0, 2.8}}]}},
                 Frame->True,
                 FrameTicks->Automatic,
                 ImageSize->800,
                 PlotRange->{{-3, 3}, {-1, 3}}], 18->
                GraphicsBox[{InsetBox[
                   TagBox[GridBox[{
                    {"\<\"t\"\>", "0.85`"},
                    {"\<\"driveIntent\"\>", "0"},
                    {"\<\"motorTorque\"\>", "0.`"},
                    {"\<\"vw\"\>", "0.`"},
                    {"\<\"controllerState\"\>", 
                    RowBox[{"\[LeftAssociation]", 
                    RowBox[{
                    RowBox[{"\<\"lastError\"\>", "\[Rule]", "0.`"}], ",", 
                    RowBox[{"\<\"lastInt\"\>", "\[Rule]", "0.`"}], ",", 
                    RowBox[{"\<\"initialized\"\>", "\[Rule]", "True"}]}], 
                    "\[RightAssociation]"}]}
                    },
                    AutoDelete->False,
                    
                    GridBoxItemSize->{
                    "Columns" -> {{Automatic}}, "Rows" -> {{Automatic}}}],
                    "Grid"], {-1, -0.5}], 
                  {GrayLevel[0], LineBox[{{-5, 0}, {5, 0}}]}, 
                  {GrayLevel[0.5], CircleBox[{-2., 0.5}, 0.5], 
                   DiskBox[{-2., 0.5}, 0.1]}, 
                  {GrayLevel[0.5], DiskBox[{-2., 2.}, 0.1]}, 
                  {GrayLevel[0.5], Thickness[0.01], 
                   LineBox[{{-2., 0.5}, {-2., 2.}}]}, 
                  {RGBColor[0, 1, 0], DiskBox[{-2., 1.5}, 0.05]}, 
                  {RGBColor[0, 0, 1], Arrowheads[0.01], 
                   CircleBox[{-2., 0.5}, 0.3333333333333333, 
                    NCache[{
                    Rational[-1, 2] 
                    Pi, -1.5707963267948966`}, {-1.5707963267948966`, \
-1.5707963267948966`}]], 
                   ArrowBox[{{-2.0033332777780557`, 
                    0.16668333319444495`}, {-2., 0.16666666666666669`}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-2., 0.5}, {-2., 0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-2., 0.5}, {-2., 0.5}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.005], Arrowheads[0.04], 
                   ArrowBox[{{-2., 0.5}, {-2., 0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-2., 0.5}, 1.6, {1.5707963267948966`, 
                    1.5707963267948966`}], 
                   ArrowBox[{{-1.9840002666653331`, 
                    2.0999200006666645`}, {-2., 2.1}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Dashing[{Small, Small}], 
                   Arrowheads[0.02], 
                   CircleBox[{-2., 0.5}, 1.7, {1.5707963267948966`, 
                    1.5707963267948966`}], 
                   ArrowBox[{{-1.9830002833319165`, 2.199915000708331}, {-2., 
                    2.2}}]}, 
                  {RGBColor[0, 0, 1], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-2., 0.5}, 1.8, {1.5707963267948966`, 
                    1.5707963267948966`}], 
                   ArrowBox[{{-1.9820002999984998`, 
                    2.2999100007499975`}, {-2., 2.3}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.003], Arrowheads[0.025], 
                   CircleBox[{-2., 0.5}, 1.5, {1.5707963267948966`, 
                    1.5707963267948966`}], 
                   ArrowBox[{{-1.9850002499987498`, 1.999925000624998}, {-2., 
                    2.}}]}, 
                  {RGBColor[0, 0, 1], ArrowBox[{{0, 2.8}, {0, 2.8}}]}},
                 Frame->True,
                 FrameTicks->Automatic,
                 ImageSize->800,
                 PlotRange->{{-3, 3}, {-1, 3}}], 19->
                GraphicsBox[{InsetBox[
                   TagBox[GridBox[{
                    {"\<\"t\"\>", "0.9`"},
                    {"\<\"driveIntent\"\>", "0"},
                    {"\<\"motorTorque\"\>", "0.`"},
                    {"\<\"vw\"\>", "0.`"},
                    {"\<\"controllerState\"\>", 
                    RowBox[{"\[LeftAssociation]", 
                    RowBox[{
                    RowBox[{"\<\"lastError\"\>", "\[Rule]", "0.`"}], ",", 
                    RowBox[{"\<\"lastInt\"\>", "\[Rule]", "0.`"}], ",", 
                    RowBox[{"\<\"initialized\"\>", "\[Rule]", "True"}]}], 
                    "\[RightAssociation]"}]}
                    },
                    AutoDelete->False,
                    
                    GridBoxItemSize->{
                    "Columns" -> {{Automatic}}, "Rows" -> {{Automatic}}}],
                    "Grid"], {-1, -0.5}], 
                  {GrayLevel[0], LineBox[{{-5, 0}, {5, 0}}]}, 
                  {GrayLevel[0.5], CircleBox[{-2., 0.5}, 0.5], 
                   DiskBox[{-2., 0.5}, 0.1]}, 
                  {GrayLevel[0.5], DiskBox[{-2., 2.}, 0.1]}, 
                  {GrayLevel[0.5], Thickness[0.01], 
                   LineBox[{{-2., 0.5}, {-2., 2.}}]}, 
                  {RGBColor[0, 1, 0], DiskBox[{-2., 1.5}, 0.05]}, 
                  {RGBColor[0, 0, 1], Arrowheads[0.01], 
                   CircleBox[{-2., 0.5}, 0.3333333333333333, 
                    NCache[{
                    Rational[-1, 2] 
                    Pi, -1.5707963267948966`}, {-1.5707963267948966`, \
-1.5707963267948966`}]], 
                   ArrowBox[{{-2.0033332777780557`, 
                    0.16668333319444495`}, {-2., 0.16666666666666669`}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-2., 0.5}, {-2., 0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-2., 0.5}, {-2., 0.5}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.005], Arrowheads[0.04], 
                   ArrowBox[{{-2., 0.5}, {-2., 0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-2., 0.5}, 1.6, {1.5707963267948966`, 
                    1.5707963267948966`}], 
                   ArrowBox[{{-1.9840002666653331`, 
                    2.0999200006666645`}, {-2., 2.1}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Dashing[{Small, Small}], 
                   Arrowheads[0.02], 
                   CircleBox[{-2., 0.5}, 1.7, {1.5707963267948966`, 
                    1.5707963267948966`}], 
                   ArrowBox[{{-1.9830002833319165`, 2.199915000708331}, {-2., 
                    2.2}}]}, 
                  {RGBColor[0, 0, 1], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-2., 0.5}, 1.8, {1.5707963267948966`, 
                    1.5707963267948966`}], 
                   ArrowBox[{{-1.9820002999984998`, 
                    2.2999100007499975`}, {-2., 2.3}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.003], Arrowheads[0.025], 
                   CircleBox[{-2., 0.5}, 1.5, {1.5707963267948966`, 
                    1.5707963267948966`}], 
                   ArrowBox[{{-1.9850002499987498`, 1.999925000624998}, {-2., 
                    2.}}]}, 
                  {RGBColor[0, 0, 1], ArrowBox[{{0, 2.8}, {0, 2.8}}]}},
                 Frame->True,
                 FrameTicks->Automatic,
                 ImageSize->800,
                 PlotRange->{{-3, 3}, {-1, 3}}], 20->
                GraphicsBox[{InsetBox[
                   TagBox[GridBox[{
                    {"\<\"t\"\>", "0.9500000000000001`"},
                    {"\<\"driveIntent\"\>", "0"},
                    {"\<\"motorTorque\"\>", "0.`"},
                    {"\<\"vw\"\>", "0.`"},
                    {"\<\"controllerState\"\>", 
                    RowBox[{"\[LeftAssociation]", 
                    RowBox[{
                    RowBox[{"\<\"lastError\"\>", "\[Rule]", "0.`"}], ",", 
                    RowBox[{"\<\"lastInt\"\>", "\[Rule]", "0.`"}], ",", 
                    RowBox[{"\<\"initialized\"\>", "\[Rule]", "True"}]}], 
                    "\[RightAssociation]"}]}
                    },
                    AutoDelete->False,
                    
                    GridBoxItemSize->{
                    "Columns" -> {{Automatic}}, "Rows" -> {{Automatic}}}],
                    "Grid"], {-1, -0.5}], 
                  {GrayLevel[0], LineBox[{{-5, 0}, {5, 0}}]}, 
                  {GrayLevel[0.5], CircleBox[{-2., 0.5}, 0.5], 
                   DiskBox[{-2., 0.5}, 0.1]}, 
                  {GrayLevel[0.5], DiskBox[{-2., 2.}, 0.1]}, 
                  {GrayLevel[0.5], Thickness[0.01], 
                   LineBox[{{-2., 0.5}, {-2., 2.}}]}, 
                  {RGBColor[0, 1, 0], DiskBox[{-2., 1.5}, 0.05]}, 
                  {RGBColor[0, 0, 1], Arrowheads[0.01], 
                   CircleBox[{-2., 0.5}, 0.3333333333333333, 
                    NCache[{
                    Rational[-1, 2] 
                    Pi, -1.5707963267948966`}, {-1.5707963267948966`, \
-1.5707963267948966`}]], 
                   ArrowBox[{{-2.0033332777780557`, 
                    0.16668333319444495`}, {-2., 0.16666666666666669`}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-2., 0.5}, {-2., 0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-2., 0.5}, {-2., 0.5}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.005], Arrowheads[0.04], 
                   ArrowBox[{{-2., 0.5}, {-2., 0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-2., 0.5}, 1.6, {1.5707963267948966`, 
                    1.5707963267948966`}], 
                   ArrowBox[{{-1.9840002666653331`, 
                    2.0999200006666645`}, {-2., 2.1}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Dashing[{Small, Small}], 
                   Arrowheads[0.02], 
                   CircleBox[{-2., 0.5}, 1.7, {1.5707963267948966`, 
                    1.5707963267948966`}], 
                   ArrowBox[{{-1.9830002833319165`, 2.199915000708331}, {-2., 
                    2.2}}]}, 
                  {RGBColor[0, 0, 1], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-2., 0.5}, 1.8, {1.5707963267948966`, 
                    1.5707963267948966`}], 
                   ArrowBox[{{-1.9820002999984998`, 
                    2.2999100007499975`}, {-2., 2.3}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.003], Arrowheads[0.025], 
                   CircleBox[{-2., 0.5}, 1.5, {1.5707963267948966`, 
                    1.5707963267948966`}], 
                   ArrowBox[{{-1.9850002499987498`, 1.999925000624998}, {-2., 
                    2.}}]}, 
                  {RGBColor[0, 0, 1], ArrowBox[{{0, 2.8}, {0, 2.8}}]}},
                 Frame->True,
                 FrameTicks->Automatic,
                 ImageSize->800,
                 PlotRange->{{-3, 3}, {-1, 3}}], 21->
                GraphicsBox[{InsetBox[
                   TagBox[GridBox[{
                    {"\<\"t\"\>", "1.`"},
                    {"\<\"driveIntent\"\>", "0"},
                    {"\<\"motorTorque\"\>", "0.`"},
                    {"\<\"vw\"\>", "0.`"},
                    {"\<\"controllerState\"\>", 
                    RowBox[{"\[LeftAssociation]", 
                    RowBox[{
                    RowBox[{"\<\"lastError\"\>", "\[Rule]", "0.`"}], ",", 
                    RowBox[{"\<\"lastInt\"\>", "\[Rule]", "0.`"}], ",", 
                    RowBox[{"\<\"initialized\"\>", "\[Rule]", "True"}]}], 
                    "\[RightAssociation]"}]}
                    },
                    AutoDelete->False,
                    
                    GridBoxItemSize->{
                    "Columns" -> {{Automatic}}, "Rows" -> {{Automatic}}}],
                    "Grid"], {-1, -0.5}], 
                  {GrayLevel[0], LineBox[{{-5, 0}, {5, 0}}]}, 
                  {GrayLevel[0.5], CircleBox[{-2., 0.5}, 0.5], 
                   DiskBox[{-2., 0.5}, 0.1]}, 
                  {GrayLevel[0.5], DiskBox[{-2., 2.}, 0.1]}, 
                  {GrayLevel[0.5], Thickness[0.01], 
                   LineBox[{{-2., 0.5}, {-2., 2.}}]}, 
                  {RGBColor[0, 1, 0], DiskBox[{-2., 1.5}, 0.05]}, 
                  {RGBColor[0, 0, 1], Arrowheads[0.01], 
                   CircleBox[{-2., 0.5}, 0.3333333333333333, 
                    NCache[{
                    Rational[-1, 2] 
                    Pi, -1.5707963267948966`}, {-1.5707963267948966`, \
-1.5707963267948966`}]], 
                   ArrowBox[{{-2.0033332777780557`, 
                    0.16668333319444495`}, {-2., 0.16666666666666669`}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-2., 0.5}, {-2., 0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-2., 0.5}, {-2., 0.5}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.005], Arrowheads[0.04], 
                   ArrowBox[{{-2., 0.5}, {-2., 0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-2., 0.5}, 1.6, {1.5707963267948966`, 
                    1.5707963267948966`}], 
                   ArrowBox[{{-1.9840002666653331`, 
                    2.0999200006666645`}, {-2., 2.1}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Dashing[{Small, Small}], 
                   Arrowheads[0.02], 
                   CircleBox[{-2., 0.5}, 1.7, {1.5707963267948966`, 
                    1.5707963267948966`}], 
                   ArrowBox[{{-1.9830002833319165`, 2.199915000708331}, {-2., 
                    2.2}}]}, 
                  {RGBColor[0, 0, 1], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-2., 0.5}, 1.8, {1.5707963267948966`, 
                    1.5707963267948966`}], 
                   ArrowBox[{{-1.9820002999984998`, 
                    2.2999100007499975`}, {-2., 2.3}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.003], Arrowheads[0.025], 
                   CircleBox[{-2., 0.5}, 1.5, {1.5707963267948966`, 
                    1.5707963267948966`}], 
                   ArrowBox[{{-1.9850002499987498`, 1.999925000624998}, {-2., 
                    2.}}]}, 
                  {RGBColor[0, 0, 1], ArrowBox[{{0, 2.8}, {0, 2.8}}]}},
                 Frame->True,
                 FrameTicks->Automatic,
                 ImageSize->800,
                 PlotRange->{{-3, 3}, {-1, 3}}], 22->
                GraphicsBox[{InsetBox[
                   TagBox[GridBox[{
                    {"\<\"t\"\>", "1.05`"},
                    {"\<\"driveIntent\"\>", "0"},
                    {"\<\"motorTorque\"\>", "0.`"},
                    {"\<\"vw\"\>", "0.`"},
                    {"\<\"controllerState\"\>", 
                    RowBox[{"\[LeftAssociation]", 
                    RowBox[{
                    RowBox[{"\<\"lastError\"\>", "\[Rule]", "0.`"}], ",", 
                    RowBox[{"\<\"lastInt\"\>", "\[Rule]", "0.`"}], ",", 
                    RowBox[{"\<\"initialized\"\>", "\[Rule]", "True"}]}], 
                    "\[RightAssociation]"}]}
                    },
                    AutoDelete->False,
                    
                    GridBoxItemSize->{
                    "Columns" -> {{Automatic}}, "Rows" -> {{Automatic}}}],
                    "Grid"], {-1, -0.5}], 
                  {GrayLevel[0], LineBox[{{-5, 0}, {5, 0}}]}, 
                  {GrayLevel[0.5], CircleBox[{-2., 0.5}, 0.5], 
                   DiskBox[{-2., 0.5}, 0.1]}, 
                  {GrayLevel[0.5], DiskBox[{-2., 2.}, 0.1]}, 
                  {GrayLevel[0.5], Thickness[0.01], 
                   LineBox[{{-2., 0.5}, {-2., 2.}}]}, 
                  {RGBColor[0, 1, 0], DiskBox[{-2., 1.5}, 0.05]}, 
                  {RGBColor[0, 0, 1], Arrowheads[0.01], 
                   CircleBox[{-2., 0.5}, 0.3333333333333333, 
                    NCache[{
                    Rational[-1, 2] 
                    Pi, -1.5707963267948966`}, {-1.5707963267948966`, \
-1.5707963267948966`}]], 
                   ArrowBox[{{-2.0033332777780557`, 
                    0.16668333319444495`}, {-2., 0.16666666666666669`}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-2., 0.5}, {-2., 0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-2., 0.5}, {-2., 0.5}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.005], Arrowheads[0.04], 
                   ArrowBox[{{-2., 0.5}, {-2., 0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-2., 0.5}, 1.6, {1.5707963267948966`, 
                    1.5707963267948966`}], 
                   ArrowBox[{{-1.9840002666653331`, 
                    2.0999200006666645`}, {-2., 2.1}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Dashing[{Small, Small}], 
                   Arrowheads[0.02], 
                   CircleBox[{-2., 0.5}, 1.7, {1.5707963267948966`, 
                    1.5707963267948966`}], 
                   ArrowBox[{{-1.9830002833319165`, 2.199915000708331}, {-2., 
                    2.2}}]}, 
                  {RGBColor[0, 0, 1], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-2., 0.5}, 1.8, {1.5707963267948966`, 
                    1.5707963267948966`}], 
                   ArrowBox[{{-1.9820002999984998`, 
                    2.2999100007499975`}, {-2., 2.3}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.003], Arrowheads[0.025], 
                   CircleBox[{-2., 0.5}, 1.5, {1.5707963267948966`, 
                    1.5707963267948966`}], 
                   ArrowBox[{{-1.9850002499987498`, 1.999925000624998}, {-2., 
                    2.}}]}, 
                  {RGBColor[0, 0, 1], ArrowBox[{{0, 2.8}, {0, 2.8}}]}},
                 Frame->True,
                 FrameTicks->Automatic,
                 ImageSize->800,
                 PlotRange->{{-3, 3}, {-1, 3}}], 23->
                GraphicsBox[{InsetBox[
                   TagBox[GridBox[{
                    {"\<\"t\"\>", "1.1`"},
                    {"\<\"driveIntent\"\>", "0"},
                    {"\<\"motorTorque\"\>", "0.`"},
                    {"\<\"vw\"\>", "0.`"},
                    {"\<\"controllerState\"\>", 
                    RowBox[{"\[LeftAssociation]", 
                    RowBox[{
                    RowBox[{"\<\"lastError\"\>", "\[Rule]", "0.`"}], ",", 
                    RowBox[{"\<\"lastInt\"\>", "\[Rule]", "0.`"}], ",", 
                    RowBox[{"\<\"initialized\"\>", "\[Rule]", "True"}]}], 
                    "\[RightAssociation]"}]}
                    },
                    AutoDelete->False,
                    
                    GridBoxItemSize->{
                    "Columns" -> {{Automatic}}, "Rows" -> {{Automatic}}}],
                    "Grid"], {-1, -0.5}], 
                  {GrayLevel[0], LineBox[{{-5, 0}, {5, 0}}]}, 
                  {GrayLevel[0.5], CircleBox[{-2., 0.5}, 0.5], 
                   DiskBox[{-2., 0.5}, 0.1]}, 
                  {GrayLevel[0.5], DiskBox[{-2., 2.}, 0.1]}, 
                  {GrayLevel[0.5], Thickness[0.01], 
                   LineBox[{{-2., 0.5}, {-2., 2.}}]}, 
                  {RGBColor[0, 1, 0], DiskBox[{-2., 1.5}, 0.05]}, 
                  {RGBColor[0, 0, 1], Arrowheads[0.01], 
                   CircleBox[{-2., 0.5}, 0.3333333333333333, 
                    NCache[{
                    Rational[-1, 2] 
                    Pi, -1.5707963267948966`}, {-1.5707963267948966`, \
-1.5707963267948966`}]], 
                   ArrowBox[{{-2.0033332777780557`, 
                    0.16668333319444495`}, {-2., 0.16666666666666669`}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-2., 0.5}, {-2., 0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-2., 0.5}, {-2., 0.5}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.005], Arrowheads[0.04], 
                   ArrowBox[{{-2., 0.5}, {-2., 0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-2., 0.5}, 1.6, {1.5707963267948966`, 
                    1.5707963267948966`}], 
                   ArrowBox[{{-1.9840002666653331`, 
                    2.0999200006666645`}, {-2., 2.1}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Dashing[{Small, Small}], 
                   Arrowheads[0.02], 
                   CircleBox[{-2., 0.5}, 1.7, {1.5707963267948966`, 
                    1.5707963267948966`}], 
                   ArrowBox[{{-1.9830002833319165`, 2.199915000708331}, {-2., 
                    2.2}}]}, 
                  {RGBColor[0, 0, 1], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-2., 0.5}, 1.8, {1.5707963267948966`, 
                    1.5707963267948966`}], 
                   ArrowBox[{{-1.9820002999984998`, 
                    2.2999100007499975`}, {-2., 2.3}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.003], Arrowheads[0.025], 
                   CircleBox[{-2., 0.5}, 1.5, {1.5707963267948966`, 
                    1.5707963267948966`}], 
                   ArrowBox[{{-1.9850002499987498`, 1.999925000624998}, {-2., 
                    2.}}]}, 
                  {RGBColor[0, 0, 1], ArrowBox[{{0, 2.8}, {0, 2.8}}]}},
                 Frame->True,
                 FrameTicks->Automatic,
                 ImageSize->800,
                 PlotRange->{{-3, 3}, {-1, 3}}], 24->
                GraphicsBox[{InsetBox[
                   TagBox[GridBox[{
                    {"\<\"t\"\>", "1.1500000000000001`"},
                    {"\<\"driveIntent\"\>", "0"},
                    {"\<\"motorTorque\"\>", "0.`"},
                    {"\<\"vw\"\>", "0.`"},
                    {"\<\"controllerState\"\>", 
                    RowBox[{"\[LeftAssociation]", 
                    RowBox[{
                    RowBox[{"\<\"lastError\"\>", "\[Rule]", "0.`"}], ",", 
                    RowBox[{"\<\"lastInt\"\>", "\[Rule]", "0.`"}], ",", 
                    RowBox[{"\<\"initialized\"\>", "\[Rule]", "True"}]}], 
                    "\[RightAssociation]"}]}
                    },
                    AutoDelete->False,
                    
                    GridBoxItemSize->{
                    "Columns" -> {{Automatic}}, "Rows" -> {{Automatic}}}],
                    "Grid"], {-1, -0.5}], 
                  {GrayLevel[0], LineBox[{{-5, 0}, {5, 0}}]}, 
                  {GrayLevel[0.5], CircleBox[{-2., 0.5}, 0.5], 
                   DiskBox[{-2., 0.5}, 0.1]}, 
                  {GrayLevel[0.5], DiskBox[{-2., 2.}, 0.1]}, 
                  {GrayLevel[0.5], Thickness[0.01], 
                   LineBox[{{-2., 0.5}, {-2., 2.}}]}, 
                  {RGBColor[0, 1, 0], DiskBox[{-2., 1.5}, 0.05]}, 
                  {RGBColor[0, 0, 1], Arrowheads[0.01], 
                   CircleBox[{-2., 0.5}, 0.3333333333333333, 
                    NCache[{
                    Rational[-1, 2] 
                    Pi, -1.5707963267948966`}, {-1.5707963267948966`, \
-1.5707963267948966`}]], 
                   ArrowBox[{{-2.0033332777780557`, 
                    0.16668333319444495`}, {-2., 0.16666666666666669`}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-2., 0.5}, {-2., 0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-2., 0.5}, {-2., 0.5}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.005], Arrowheads[0.04], 
                   ArrowBox[{{-2., 0.5}, {-2., 0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-2., 0.5}, 1.6, {1.5707963267948966`, 
                    1.5707963267948966`}], 
                   ArrowBox[{{-1.9840002666653331`, 
                    2.0999200006666645`}, {-2., 2.1}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Dashing[{Small, Small}], 
                   Arrowheads[0.02], 
                   CircleBox[{-2., 0.5}, 1.7, {1.5707963267948966`, 
                    1.5707963267948966`}], 
                   ArrowBox[{{-1.9830002833319165`, 2.199915000708331}, {-2., 
                    2.2}}]}, 
                  {RGBColor[0, 0, 1], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-2., 0.5}, 1.8, {1.5707963267948966`, 
                    1.5707963267948966`}], 
                   ArrowBox[{{-1.9820002999984998`, 
                    2.2999100007499975`}, {-2., 2.3}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.003], Arrowheads[0.025], 
                   CircleBox[{-2., 0.5}, 1.5, {1.5707963267948966`, 
                    1.5707963267948966`}], 
                   ArrowBox[{{-1.9850002499987498`, 1.999925000624998}, {-2., 
                    2.}}]}, 
                  {RGBColor[0, 0, 1], ArrowBox[{{0, 2.8}, {0, 2.8}}]}},
                 Frame->True,
                 FrameTicks->Automatic,
                 ImageSize->800,
                 PlotRange->{{-3, 3}, {-1, 3}}], 25->
                GraphicsBox[{InsetBox[
                   TagBox[GridBox[{
                    {"\<\"t\"\>", "1.2`"},
                    {"\<\"driveIntent\"\>", "0"},
                    {"\<\"motorTorque\"\>", "0.`"},
                    {"\<\"vw\"\>", "0.`"},
                    {"\<\"controllerState\"\>", 
                    RowBox[{"\[LeftAssociation]", 
                    RowBox[{
                    RowBox[{"\<\"lastError\"\>", "\[Rule]", "0.`"}], ",", 
                    RowBox[{"\<\"lastInt\"\>", "\[Rule]", "0.`"}], ",", 
                    RowBox[{"\<\"initialized\"\>", "\[Rule]", "True"}]}], 
                    "\[RightAssociation]"}]}
                    },
                    AutoDelete->False,
                    
                    GridBoxItemSize->{
                    "Columns" -> {{Automatic}}, "Rows" -> {{Automatic}}}],
                    "Grid"], {-1, -0.5}], 
                  {GrayLevel[0], LineBox[{{-5, 0}, {5, 0}}]}, 
                  {GrayLevel[0.5], CircleBox[{-2., 0.5}, 0.5], 
                   DiskBox[{-2., 0.5}, 0.1]}, 
                  {GrayLevel[0.5], DiskBox[{-2., 2.}, 0.1]}, 
                  {GrayLevel[0.5], Thickness[0.01], 
                   LineBox[{{-2., 0.5}, {-2., 2.}}]}, 
                  {RGBColor[0, 1, 0], DiskBox[{-2., 1.5}, 0.05]}, 
                  {RGBColor[0, 0, 1], Arrowheads[0.01], 
                   CircleBox[{-2., 0.5}, 0.3333333333333333, 
                    NCache[{
                    Rational[-1, 2] 
                    Pi, -1.5707963267948966`}, {-1.5707963267948966`, \
-1.5707963267948966`}]], 
                   ArrowBox[{{-2.0033332777780557`, 
                    0.16668333319444495`}, {-2., 0.16666666666666669`}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-2., 0.5}, {-2., 0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-2., 0.5}, {-2., 0.5}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.005], Arrowheads[0.04], 
                   ArrowBox[{{-2., 0.5}, {-2., 0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-2., 0.5}, 1.6, {1.5707963267948966`, 
                    1.5707963267948966`}], 
                   ArrowBox[{{-1.9840002666653331`, 
                    2.0999200006666645`}, {-2., 2.1}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Dashing[{Small, Small}], 
                   Arrowheads[0.02], 
                   CircleBox[{-2., 0.5}, 1.7, {1.5707963267948966`, 
                    1.5707963267948966`}], 
                   ArrowBox[{{-1.9830002833319165`, 2.199915000708331}, {-2., 
                    2.2}}]}, 
                  {RGBColor[0, 0, 1], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-2., 0.5}, 1.8, {1.5707963267948966`, 
                    1.5707963267948966`}], 
                   ArrowBox[{{-1.9820002999984998`, 
                    2.2999100007499975`}, {-2., 2.3}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.003], Arrowheads[0.025], 
                   CircleBox[{-2., 0.5}, 1.5, {1.5707963267948966`, 
                    1.5707963267948966`}], 
                   ArrowBox[{{-1.9850002499987498`, 1.999925000624998}, {-2., 
                    2.}}]}, 
                  {RGBColor[0, 0, 1], ArrowBox[{{0, 2.8}, {0, 2.8}}]}},
                 Frame->True,
                 FrameTicks->Automatic,
                 ImageSize->800,
                 PlotRange->{{-3, 3}, {-1, 3}}], 26->
                GraphicsBox[{InsetBox[
                   TagBox[GridBox[{
                    {"\<\"t\"\>", "1.25`"},
                    {"\<\"driveIntent\"\>", "0"},
                    {"\<\"motorTorque\"\>", "0.`"},
                    {"\<\"vw\"\>", "0.`"},
                    {"\<\"controllerState\"\>", 
                    RowBox[{"\[LeftAssociation]", 
                    RowBox[{
                    RowBox[{"\<\"lastError\"\>", "\[Rule]", "0.`"}], ",", 
                    RowBox[{"\<\"lastInt\"\>", "\[Rule]", "0.`"}], ",", 
                    RowBox[{"\<\"initialized\"\>", "\[Rule]", "True"}]}], 
                    "\[RightAssociation]"}]}
                    },
                    AutoDelete->False,
                    
                    GridBoxItemSize->{
                    "Columns" -> {{Automatic}}, "Rows" -> {{Automatic}}}],
                    "Grid"], {-1, -0.5}], 
                  {GrayLevel[0], LineBox[{{-5, 0}, {5, 0}}]}, 
                  {GrayLevel[0.5], CircleBox[{-2., 0.5}, 0.5], 
                   DiskBox[{-2., 0.5}, 0.1]}, 
                  {GrayLevel[0.5], DiskBox[{-2., 2.}, 0.1]}, 
                  {GrayLevel[0.5], Thickness[0.01], 
                   LineBox[{{-2., 0.5}, {-2., 2.}}]}, 
                  {RGBColor[0, 1, 0], DiskBox[{-2., 1.5}, 0.05]}, 
                  {RGBColor[0, 0, 1], Arrowheads[0.01], 
                   CircleBox[{-2., 0.5}, 0.3333333333333333, 
                    NCache[{
                    Rational[-1, 2] 
                    Pi, -1.5707963267948966`}, {-1.5707963267948966`, \
-1.5707963267948966`}]], 
                   ArrowBox[{{-2.0033332777780557`, 
                    0.16668333319444495`}, {-2., 0.16666666666666669`}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-2., 0.5}, {-2., 0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-2., 0.5}, {-2., 0.5}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.005], Arrowheads[0.04], 
                   ArrowBox[{{-2., 0.5}, {-2., 0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-2., 0.5}, 1.6, {1.5707963267948966`, 
                    1.5707963267948966`}], 
                   ArrowBox[{{-1.9840002666653331`, 
                    2.0999200006666645`}, {-2., 2.1}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Dashing[{Small, Small}], 
                   Arrowheads[0.02], 
                   CircleBox[{-2., 0.5}, 1.7, {1.5707963267948966`, 
                    1.5707963267948966`}], 
                   ArrowBox[{{-1.9830002833319165`, 2.199915000708331}, {-2., 
                    2.2}}]}, 
                  {RGBColor[0, 0, 1], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-2., 0.5}, 1.8, {1.5707963267948966`, 
                    1.5707963267948966`}], 
                   ArrowBox[{{-1.9820002999984998`, 
                    2.2999100007499975`}, {-2., 2.3}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.003], Arrowheads[0.025], 
                   CircleBox[{-2., 0.5}, 1.5, {1.5707963267948966`, 
                    1.5707963267948966`}], 
                   ArrowBox[{{-1.9850002499987498`, 1.999925000624998}, {-2., 
                    2.}}]}, 
                  {RGBColor[0, 0, 1], ArrowBox[{{0, 2.8}, {0, 2.8}}]}},
                 Frame->True,
                 FrameTicks->Automatic,
                 ImageSize->800,
                 PlotRange->{{-3, 3}, {-1, 3}}], 27->
                GraphicsBox[{InsetBox[
                   TagBox[GridBox[{
                    {"\<\"t\"\>", "1.3`"},
                    {"\<\"driveIntent\"\>", "0"},
                    {"\<\"motorTorque\"\>", "0.`"},
                    {"\<\"vw\"\>", "0.`"},
                    {"\<\"controllerState\"\>", 
                    RowBox[{"\[LeftAssociation]", 
                    RowBox[{
                    RowBox[{"\<\"lastError\"\>", "\[Rule]", "0.`"}], ",", 
                    RowBox[{"\<\"lastInt\"\>", "\[Rule]", "0.`"}], ",", 
                    RowBox[{"\<\"initialized\"\>", "\[Rule]", "True"}]}], 
                    "\[RightAssociation]"}]}
                    },
                    AutoDelete->False,
                    
                    GridBoxItemSize->{
                    "Columns" -> {{Automatic}}, "Rows" -> {{Automatic}}}],
                    "Grid"], {-1, -0.5}], 
                  {GrayLevel[0], LineBox[{{-5, 0}, {5, 0}}]}, 
                  {GrayLevel[0.5], CircleBox[{-2., 0.5}, 0.5], 
                   DiskBox[{-2., 0.5}, 0.1]}, 
                  {GrayLevel[0.5], DiskBox[{-2., 2.}, 0.1]}, 
                  {GrayLevel[0.5], Thickness[0.01], 
                   LineBox[{{-2., 0.5}, {-2., 2.}}]}, 
                  {RGBColor[0, 1, 0], DiskBox[{-2., 1.5}, 0.05]}, 
                  {RGBColor[0, 0, 1], Arrowheads[0.01], 
                   CircleBox[{-2., 0.5}, 0.3333333333333333, 
                    NCache[{
                    Rational[-1, 2] 
                    Pi, -1.5707963267948966`}, {-1.5707963267948966`, \
-1.5707963267948966`}]], 
                   ArrowBox[{{-2.0033332777780557`, 
                    0.16668333319444495`}, {-2., 0.16666666666666669`}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-2., 0.5}, {-2., 0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-2., 0.5}, {-2., 0.5}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.005], Arrowheads[0.04], 
                   ArrowBox[{{-2., 0.5}, {-2., 0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-2., 0.5}, 1.6, {1.5707963267948966`, 
                    1.5707963267948966`}], 
                   ArrowBox[{{-1.9840002666653331`, 
                    2.0999200006666645`}, {-2., 2.1}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Dashing[{Small, Small}], 
                   Arrowheads[0.02], 
                   CircleBox[{-2., 0.5}, 1.7, {1.5707963267948966`, 
                    1.5707963267948966`}], 
                   ArrowBox[{{-1.9830002833319165`, 2.199915000708331}, {-2., 
                    2.2}}]}, 
                  {RGBColor[0, 0, 1], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-2., 0.5}, 1.8, {1.5707963267948966`, 
                    1.5707963267948966`}], 
                   ArrowBox[{{-1.9820002999984998`, 
                    2.2999100007499975`}, {-2., 2.3}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.003], Arrowheads[0.025], 
                   CircleBox[{-2., 0.5}, 1.5, {1.5707963267948966`, 
                    1.5707963267948966`}], 
                   ArrowBox[{{-1.9850002499987498`, 1.999925000624998}, {-2., 
                    2.}}]}, 
                  {RGBColor[0, 0, 1], ArrowBox[{{0, 2.8}, {0, 2.8}}]}},
                 Frame->True,
                 FrameTicks->Automatic,
                 ImageSize->800,
                 PlotRange->{{-3, 3}, {-1, 3}}], 28->
                GraphicsBox[{InsetBox[
                   TagBox[GridBox[{
                    {"\<\"t\"\>", "1.35`"},
                    {"\<\"driveIntent\"\>", "0"},
                    {"\<\"motorTorque\"\>", "0.`"},
                    {"\<\"vw\"\>", "0.`"},
                    {"\<\"controllerState\"\>", 
                    RowBox[{"\[LeftAssociation]", 
                    RowBox[{
                    RowBox[{"\<\"lastError\"\>", "\[Rule]", "0.`"}], ",", 
                    RowBox[{"\<\"lastInt\"\>", "\[Rule]", "0.`"}], ",", 
                    RowBox[{"\<\"initialized\"\>", "\[Rule]", "True"}]}], 
                    "\[RightAssociation]"}]}
                    },
                    AutoDelete->False,
                    
                    GridBoxItemSize->{
                    "Columns" -> {{Automatic}}, "Rows" -> {{Automatic}}}],
                    "Grid"], {-1, -0.5}], 
                  {GrayLevel[0], LineBox[{{-5, 0}, {5, 0}}]}, 
                  {GrayLevel[0.5], CircleBox[{-2., 0.5}, 0.5], 
                   DiskBox[{-2., 0.5}, 0.1]}, 
                  {GrayLevel[0.5], DiskBox[{-2., 2.}, 0.1]}, 
                  {GrayLevel[0.5], Thickness[0.01], 
                   LineBox[{{-2., 0.5}, {-2., 2.}}]}, 
                  {RGBColor[0, 1, 0], DiskBox[{-2., 1.5}, 0.05]}, 
                  {RGBColor[0, 0, 1], Arrowheads[0.01], 
                   CircleBox[{-2., 0.5}, 0.3333333333333333, 
                    NCache[{
                    Rational[-1, 2] 
                    Pi, -1.5707963267948966`}, {-1.5707963267948966`, \
-1.5707963267948966`}]], 
                   ArrowBox[{{-2.0033332777780557`, 
                    0.16668333319444495`}, {-2., 0.16666666666666669`}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-2., 0.5}, {-2., 0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-2., 0.5}, {-2., 0.5}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.005], Arrowheads[0.04], 
                   ArrowBox[{{-2., 0.5}, {-2., 0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-2., 0.5}, 1.6, {1.5707963267948966`, 
                    1.5707963267948966`}], 
                   ArrowBox[{{-1.9840002666653331`, 
                    2.0999200006666645`}, {-2., 2.1}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Dashing[{Small, Small}], 
                   Arrowheads[0.02], 
                   CircleBox[{-2., 0.5}, 1.7, {1.5707963267948966`, 
                    1.5707963267948966`}], 
                   ArrowBox[{{-1.9830002833319165`, 2.199915000708331}, {-2., 
                    2.2}}]}, 
                  {RGBColor[0, 0, 1], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-2., 0.5}, 1.8, {1.5707963267948966`, 
                    1.5707963267948966`}], 
                   ArrowBox[{{-1.9820002999984998`, 
                    2.2999100007499975`}, {-2., 2.3}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.003], Arrowheads[0.025], 
                   CircleBox[{-2., 0.5}, 1.5, {1.5707963267948966`, 
                    1.5707963267948966`}], 
                   ArrowBox[{{-1.9850002499987498`, 1.999925000624998}, {-2., 
                    2.}}]}, 
                  {RGBColor[0, 0, 1], ArrowBox[{{0, 2.8}, {0, 2.8}}]}},
                 Frame->True,
                 FrameTicks->Automatic,
                 ImageSize->800,
                 PlotRange->{{-3, 3}, {-1, 3}}], 29->
                GraphicsBox[{InsetBox[
                   TagBox[GridBox[{
                    {"\<\"t\"\>", "1.4000000000000001`"},
                    {"\<\"driveIntent\"\>", "0"},
                    {"\<\"motorTorque\"\>", "0.`"},
                    {"\<\"vw\"\>", "0.`"},
                    {"\<\"controllerState\"\>", 
                    RowBox[{"\[LeftAssociation]", 
                    RowBox[{
                    RowBox[{"\<\"lastError\"\>", "\[Rule]", "0.`"}], ",", 
                    RowBox[{"\<\"lastInt\"\>", "\[Rule]", "0.`"}], ",", 
                    RowBox[{"\<\"initialized\"\>", "\[Rule]", "True"}]}], 
                    "\[RightAssociation]"}]}
                    },
                    AutoDelete->False,
                    
                    GridBoxItemSize->{
                    "Columns" -> {{Automatic}}, "Rows" -> {{Automatic}}}],
                    "Grid"], {-1, -0.5}], 
                  {GrayLevel[0], LineBox[{{-5, 0}, {5, 0}}]}, 
                  {GrayLevel[0.5], CircleBox[{-2., 0.5}, 0.5], 
                   DiskBox[{-2., 0.5}, 0.1]}, 
                  {GrayLevel[0.5], DiskBox[{-2., 2.}, 0.1]}, 
                  {GrayLevel[0.5], Thickness[0.01], 
                   LineBox[{{-2., 0.5}, {-2., 2.}}]}, 
                  {RGBColor[0, 1, 0], DiskBox[{-2., 1.5}, 0.05]}, 
                  {RGBColor[0, 0, 1], Arrowheads[0.01], 
                   CircleBox[{-2., 0.5}, 0.3333333333333333, 
                    NCache[{
                    Rational[-1, 2] 
                    Pi, -1.5707963267948966`}, {-1.5707963267948966`, \
-1.5707963267948966`}]], 
                   ArrowBox[{{-2.0033332777780557`, 
                    0.16668333319444495`}, {-2., 0.16666666666666669`}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-2., 0.5}, {-2., 0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-2., 0.5}, {-2., 0.5}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.005], Arrowheads[0.04], 
                   ArrowBox[{{-2., 0.5}, {-2., 0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-2., 0.5}, 1.6, {1.5707963267948966`, 
                    1.5707963267948966`}], 
                   ArrowBox[{{-1.9840002666653331`, 
                    2.0999200006666645`}, {-2., 2.1}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Dashing[{Small, Small}], 
                   Arrowheads[0.02], 
                   CircleBox[{-2., 0.5}, 1.7, {1.5707963267948966`, 
                    1.5707963267948966`}], 
                   ArrowBox[{{-1.9830002833319165`, 2.199915000708331}, {-2., 
                    2.2}}]}, 
                  {RGBColor[0, 0, 1], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-2., 0.5}, 1.8, {1.5707963267948966`, 
                    1.5707963267948966`}], 
                   ArrowBox[{{-1.9820002999984998`, 
                    2.2999100007499975`}, {-2., 2.3}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.003], Arrowheads[0.025], 
                   CircleBox[{-2., 0.5}, 1.5, {1.5707963267948966`, 
                    1.5707963267948966`}], 
                   ArrowBox[{{-1.9850002499987498`, 1.999925000624998}, {-2., 
                    2.}}]}, 
                  {RGBColor[0, 0, 1], ArrowBox[{{0, 2.8}, {0, 2.8}}]}},
                 Frame->True,
                 FrameTicks->Automatic,
                 ImageSize->800,
                 PlotRange->{{-3, 3}, {-1, 3}}], 30->
                GraphicsBox[{InsetBox[
                   TagBox[GridBox[{
                    {"\<\"t\"\>", "1.45`"},
                    {"\<\"driveIntent\"\>", "0"},
                    {"\<\"motorTorque\"\>", "0.`"},
                    {"\<\"vw\"\>", "0.`"},
                    {"\<\"controllerState\"\>", 
                    RowBox[{"\[LeftAssociation]", 
                    RowBox[{
                    RowBox[{"\<\"lastError\"\>", "\[Rule]", "0.`"}], ",", 
                    RowBox[{"\<\"lastInt\"\>", "\[Rule]", "0.`"}], ",", 
                    RowBox[{"\<\"initialized\"\>", "\[Rule]", "True"}]}], 
                    "\[RightAssociation]"}]}
                    },
                    AutoDelete->False,
                    
                    GridBoxItemSize->{
                    "Columns" -> {{Automatic}}, "Rows" -> {{Automatic}}}],
                    "Grid"], {-1, -0.5}], 
                  {GrayLevel[0], LineBox[{{-5, 0}, {5, 0}}]}, 
                  {GrayLevel[0.5], CircleBox[{-2., 0.5}, 0.5], 
                   DiskBox[{-2., 0.5}, 0.1]}, 
                  {GrayLevel[0.5], DiskBox[{-2., 2.}, 0.1]}, 
                  {GrayLevel[0.5], Thickness[0.01], 
                   LineBox[{{-2., 0.5}, {-2., 2.}}]}, 
                  {RGBColor[0, 1, 0], DiskBox[{-2., 1.5}, 0.05]}, 
                  {RGBColor[0, 0, 1], Arrowheads[0.01], 
                   CircleBox[{-2., 0.5}, 0.3333333333333333, 
                    NCache[{
                    Rational[-1, 2] 
                    Pi, -1.5707963267948966`}, {-1.5707963267948966`, \
-1.5707963267948966`}]], 
                   ArrowBox[{{-2.0033332777780557`, 
                    0.16668333319444495`}, {-2., 0.16666666666666669`}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-2., 0.5}, {-2., 0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-2., 0.5}, {-2., 0.5}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.005], Arrowheads[0.04], 
                   ArrowBox[{{-2., 0.5}, {-2., 0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-2., 0.5}, 1.6, {1.5707963267948966`, 
                    1.5707963267948966`}], 
                   ArrowBox[{{-1.9840002666653331`, 
                    2.0999200006666645`}, {-2., 2.1}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Dashing[{Small, Small}], 
                   Arrowheads[0.02], 
                   CircleBox[{-2., 0.5}, 1.7, {1.5707963267948966`, 
                    1.5707963267948966`}], 
                   ArrowBox[{{-1.9830002833319165`, 2.199915000708331}, {-2., 
                    2.2}}]}, 
                  {RGBColor[0, 0, 1], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-2., 0.5}, 1.8, {1.5707963267948966`, 
                    1.5707963267948966`}], 
                   ArrowBox[{{-1.9820002999984998`, 
                    2.2999100007499975`}, {-2., 2.3}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.003], Arrowheads[0.025], 
                   CircleBox[{-2., 0.5}, 1.5, {1.5707963267948966`, 
                    1.5707963267948966`}], 
                   ArrowBox[{{-1.9850002499987498`, 1.999925000624998}, {-2., 
                    2.}}]}, 
                  {RGBColor[0, 0, 1], ArrowBox[{{0, 2.8}, {0, 2.8}}]}},
                 Frame->True,
                 FrameTicks->Automatic,
                 ImageSize->800,
                 PlotRange->{{-3, 3}, {-1, 3}}], 31->
                GraphicsBox[{InsetBox[
                   TagBox[GridBox[{
                    {"\<\"t\"\>", "1.5`"},
                    {"\<\"driveIntent\"\>", "0"},
                    {"\<\"motorTorque\"\>", "0.`"},
                    {"\<\"vw\"\>", "0.`"},
                    {"\<\"controllerState\"\>", 
                    RowBox[{"\[LeftAssociation]", 
                    RowBox[{
                    RowBox[{"\<\"lastError\"\>", "\[Rule]", "0.`"}], ",", 
                    RowBox[{"\<\"lastInt\"\>", "\[Rule]", "0.`"}], ",", 
                    RowBox[{"\<\"initialized\"\>", "\[Rule]", "True"}]}], 
                    "\[RightAssociation]"}]}
                    },
                    AutoDelete->False,
                    
                    GridBoxItemSize->{
                    "Columns" -> {{Automatic}}, "Rows" -> {{Automatic}}}],
                    "Grid"], {-1, -0.5}], 
                  {GrayLevel[0], LineBox[{{-5, 0}, {5, 0}}]}, 
                  {GrayLevel[0.5], CircleBox[{-2., 0.5}, 0.5], 
                   DiskBox[{-2., 0.5}, 0.1]}, 
                  {GrayLevel[0.5], DiskBox[{-2., 2.}, 0.1]}, 
                  {GrayLevel[0.5], Thickness[0.01], 
                   LineBox[{{-2., 0.5}, {-2., 2.}}]}, 
                  {RGBColor[0, 1, 0], DiskBox[{-2., 1.5}, 0.05]}, 
                  {RGBColor[0, 0, 1], Arrowheads[0.01], 
                   CircleBox[{-2., 0.5}, 0.3333333333333333, 
                    NCache[{
                    Rational[-1, 2] 
                    Pi, -1.5707963267948966`}, {-1.5707963267948966`, \
-1.5707963267948966`}]], 
                   ArrowBox[{{-2.0033332777780557`, 
                    0.16668333319444495`}, {-2., 0.16666666666666669`}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-2., 0.5}, {-2., 0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-2., 0.5}, {-2., 0.5}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.005], Arrowheads[0.04], 
                   ArrowBox[{{-2., 0.5}, {-2., 0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-2., 0.5}, 1.6, {1.5707963267948966`, 
                    1.5707963267948966`}], 
                   ArrowBox[{{-1.9840002666653331`, 
                    2.0999200006666645`}, {-2., 2.1}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Dashing[{Small, Small}], 
                   Arrowheads[0.02], 
                   CircleBox[{-2., 0.5}, 1.7, {1.5707963267948966`, 
                    1.5707963267948966`}], 
                   ArrowBox[{{-1.9830002833319165`, 2.199915000708331}, {-2., 
                    2.2}}]}, 
                  {RGBColor[0, 0, 1], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-2., 0.5}, 1.8, {1.5707963267948966`, 
                    1.5707963267948966`}], 
                   ArrowBox[{{-1.9820002999984998`, 
                    2.2999100007499975`}, {-2., 2.3}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.003], Arrowheads[0.025], 
                   CircleBox[{-2., 0.5}, 1.5, {1.5707963267948966`, 
                    1.5707963267948966`}], 
                   ArrowBox[{{-1.9850002499987498`, 1.999925000624998}, {-2., 
                    2.}}]}, 
                  {RGBColor[0, 0, 1], ArrowBox[{{0, 2.8}, {0, 2.8}}]}},
                 Frame->True,
                 FrameTicks->Automatic,
                 ImageSize->800,
                 PlotRange->{{-3, 3}, {-1, 3}}], 32->
                GraphicsBox[{InsetBox[
                   TagBox[GridBox[{
                    {"\<\"t\"\>", "1.55`"},
                    {"\<\"driveIntent\"\>", "0"},
                    {"\<\"motorTorque\"\>", "0.`"},
                    {"\<\"vw\"\>", "0.`"},
                    {"\<\"controllerState\"\>", 
                    RowBox[{"\[LeftAssociation]", 
                    RowBox[{
                    RowBox[{"\<\"lastError\"\>", "\[Rule]", "0.`"}], ",", 
                    RowBox[{"\<\"lastInt\"\>", "\[Rule]", "0.`"}], ",", 
                    RowBox[{"\<\"initialized\"\>", "\[Rule]", "True"}]}], 
                    "\[RightAssociation]"}]}
                    },
                    AutoDelete->False,
                    
                    GridBoxItemSize->{
                    "Columns" -> {{Automatic}}, "Rows" -> {{Automatic}}}],
                    "Grid"], {-1, -0.5}], 
                  {GrayLevel[0], LineBox[{{-5, 0}, {5, 0}}]}, 
                  {GrayLevel[0.5], CircleBox[{-2., 0.5}, 0.5], 
                   DiskBox[{-2., 0.5}, 0.1]}, 
                  {GrayLevel[0.5], DiskBox[{-2., 2.}, 0.1]}, 
                  {GrayLevel[0.5], Thickness[0.01], 
                   LineBox[{{-2., 0.5}, {-2., 2.}}]}, 
                  {RGBColor[0, 1, 0], DiskBox[{-2., 1.5}, 0.05]}, 
                  {RGBColor[0, 0, 1], Arrowheads[0.01], 
                   CircleBox[{-2., 0.5}, 0.3333333333333333, 
                    NCache[{
                    Rational[-1, 2] 
                    Pi, -1.5707963267948966`}, {-1.5707963267948966`, \
-1.5707963267948966`}]], 
                   ArrowBox[{{-2.0033332777780557`, 
                    0.16668333319444495`}, {-2., 0.16666666666666669`}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-2., 0.5}, {-2., 0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-2., 0.5}, {-2., 0.5}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.005], Arrowheads[0.04], 
                   ArrowBox[{{-2., 0.5}, {-2., 0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-2., 0.5}, 1.6, {1.5707963267948966`, 
                    1.5707963267948966`}], 
                   ArrowBox[{{-1.9840002666653331`, 
                    2.0999200006666645`}, {-2., 2.1}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Dashing[{Small, Small}], 
                   Arrowheads[0.02], 
                   CircleBox[{-2., 0.5}, 1.7, {1.5707963267948966`, 
                    1.5707963267948966`}], 
                   ArrowBox[{{-1.9830002833319165`, 2.199915000708331}, {-2., 
                    2.2}}]}, 
                  {RGBColor[0, 0, 1], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-2., 0.5}, 1.8, {1.5707963267948966`, 
                    1.5707963267948966`}], 
                   ArrowBox[{{-1.9820002999984998`, 
                    2.2999100007499975`}, {-2., 2.3}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.003], Arrowheads[0.025], 
                   CircleBox[{-2., 0.5}, 1.5, {1.5707963267948966`, 
                    1.5707963267948966`}], 
                   ArrowBox[{{-1.9850002499987498`, 1.999925000624998}, {-2., 
                    2.}}]}, 
                  {RGBColor[0, 0, 1], ArrowBox[{{0, 2.8}, {0, 2.8}}]}},
                 Frame->True,
                 FrameTicks->Automatic,
                 ImageSize->800,
                 PlotRange->{{-3, 3}, {-1, 3}}], 33->
                GraphicsBox[{InsetBox[
                   TagBox[GridBox[{
                    {"\<\"t\"\>", "1.6`"},
                    {"\<\"driveIntent\"\>", "0"},
                    {"\<\"motorTorque\"\>", "0.`"},
                    {"\<\"vw\"\>", "0.`"},
                    {"\<\"controllerState\"\>", 
                    RowBox[{"\[LeftAssociation]", 
                    RowBox[{
                    RowBox[{"\<\"lastError\"\>", "\[Rule]", "0.`"}], ",", 
                    RowBox[{"\<\"lastInt\"\>", "\[Rule]", "0.`"}], ",", 
                    RowBox[{"\<\"initialized\"\>", "\[Rule]", "True"}]}], 
                    "\[RightAssociation]"}]}
                    },
                    AutoDelete->False,
                    
                    GridBoxItemSize->{
                    "Columns" -> {{Automatic}}, "Rows" -> {{Automatic}}}],
                    "Grid"], {-1, -0.5}], 
                  {GrayLevel[0], LineBox[{{-5, 0}, {5, 0}}]}, 
                  {GrayLevel[0.5], CircleBox[{-2., 0.5}, 0.5], 
                   DiskBox[{-2., 0.5}, 0.1]}, 
                  {GrayLevel[0.5], DiskBox[{-2., 2.}, 0.1]}, 
                  {GrayLevel[0.5], Thickness[0.01], 
                   LineBox[{{-2., 0.5}, {-2., 2.}}]}, 
                  {RGBColor[0, 1, 0], DiskBox[{-2., 1.5}, 0.05]}, 
                  {RGBColor[0, 0, 1], Arrowheads[0.01], 
                   CircleBox[{-2., 0.5}, 0.3333333333333333, 
                    NCache[{
                    Rational[-1, 2] 
                    Pi, -1.5707963267948966`}, {-1.5707963267948966`, \
-1.5707963267948966`}]], 
                   ArrowBox[{{-2.0033332777780557`, 
                    0.16668333319444495`}, {-2., 0.16666666666666669`}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-2., 0.5}, {-2., 0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-2., 0.5}, {-2., 0.5}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.005], Arrowheads[0.04], 
                   ArrowBox[{{-2., 0.5}, {-2., 0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-2., 0.5}, 1.6, {1.5707963267948966`, 
                    1.5707963267948966`}], 
                   ArrowBox[{{-1.9840002666653331`, 
                    2.0999200006666645`}, {-2., 2.1}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Dashing[{Small, Small}], 
                   Arrowheads[0.02], 
                   CircleBox[{-2., 0.5}, 1.7, {1.5707963267948966`, 
                    1.5707963267948966`}], 
                   ArrowBox[{{-1.9830002833319165`, 2.199915000708331}, {-2., 
                    2.2}}]}, 
                  {RGBColor[0, 0, 1], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-2., 0.5}, 1.8, {1.5707963267948966`, 
                    1.5707963267948966`}], 
                   ArrowBox[{{-1.9820002999984998`, 
                    2.2999100007499975`}, {-2., 2.3}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.003], Arrowheads[0.025], 
                   CircleBox[{-2., 0.5}, 1.5, {1.5707963267948966`, 
                    1.5707963267948966`}], 
                   ArrowBox[{{-1.9850002499987498`, 1.999925000624998}, {-2., 
                    2.}}]}, 
                  {RGBColor[0, 0, 1], ArrowBox[{{0, 2.8}, {0, 2.8}}]}},
                 Frame->True,
                 FrameTicks->Automatic,
                 ImageSize->800,
                 PlotRange->{{-3, 3}, {-1, 3}}], 34->
                GraphicsBox[{InsetBox[
                   TagBox[GridBox[{
                    {"\<\"t\"\>", "1.6500000000000001`"},
                    {"\<\"driveIntent\"\>", "0"},
                    {"\<\"motorTorque\"\>", "0.`"},
                    {"\<\"vw\"\>", "0.`"},
                    {"\<\"controllerState\"\>", 
                    RowBox[{"\[LeftAssociation]", 
                    RowBox[{
                    RowBox[{"\<\"lastError\"\>", "\[Rule]", "0.`"}], ",", 
                    RowBox[{"\<\"lastInt\"\>", "\[Rule]", "0.`"}], ",", 
                    RowBox[{"\<\"initialized\"\>", "\[Rule]", "True"}]}], 
                    "\[RightAssociation]"}]}
                    },
                    AutoDelete->False,
                    
                    GridBoxItemSize->{
                    "Columns" -> {{Automatic}}, "Rows" -> {{Automatic}}}],
                    "Grid"], {-1, -0.5}], 
                  {GrayLevel[0], LineBox[{{-5, 0}, {5, 0}}]}, 
                  {GrayLevel[0.5], CircleBox[{-2., 0.5}, 0.5], 
                   DiskBox[{-2., 0.5}, 0.1]}, 
                  {GrayLevel[0.5], DiskBox[{-2., 2.}, 0.1]}, 
                  {GrayLevel[0.5], Thickness[0.01], 
                   LineBox[{{-2., 0.5}, {-2., 2.}}]}, 
                  {RGBColor[0, 1, 0], DiskBox[{-2., 1.5}, 0.05]}, 
                  {RGBColor[0, 0, 1], Arrowheads[0.01], 
                   CircleBox[{-2., 0.5}, 0.3333333333333333, 
                    NCache[{
                    Rational[-1, 2] 
                    Pi, -1.5707963267948966`}, {-1.5707963267948966`, \
-1.5707963267948966`}]], 
                   ArrowBox[{{-2.0033332777780557`, 
                    0.16668333319444495`}, {-2., 0.16666666666666669`}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-2., 0.5}, {-2., 0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-2., 0.5}, {-2., 0.5}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.005], Arrowheads[0.04], 
                   ArrowBox[{{-2., 0.5}, {-2., 0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-2., 0.5}, 1.6, {1.5707963267948966`, 
                    1.5707963267948966`}], 
                   ArrowBox[{{-1.9840002666653331`, 
                    2.0999200006666645`}, {-2., 2.1}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Dashing[{Small, Small}], 
                   Arrowheads[0.02], 
                   CircleBox[{-2., 0.5}, 1.7, {1.5707963267948966`, 
                    1.5707963267948966`}], 
                   ArrowBox[{{-1.9830002833319165`, 2.199915000708331}, {-2., 
                    2.2}}]}, 
                  {RGBColor[0, 0, 1], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-2., 0.5}, 1.8, {1.5707963267948966`, 
                    1.5707963267948966`}], 
                   ArrowBox[{{-1.9820002999984998`, 
                    2.2999100007499975`}, {-2., 2.3}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.003], Arrowheads[0.025], 
                   CircleBox[{-2., 0.5}, 1.5, {1.5707963267948966`, 
                    1.5707963267948966`}], 
                   ArrowBox[{{-1.9850002499987498`, 1.999925000624998}, {-2., 
                    2.}}]}, 
                  {RGBColor[0, 0, 1], ArrowBox[{{0, 2.8}, {0, 2.8}}]}},
                 Frame->True,
                 FrameTicks->Automatic,
                 ImageSize->800,
                 PlotRange->{{-3, 3}, {-1, 3}}], 35->
                GraphicsBox[{InsetBox[
                   TagBox[GridBox[{
                    {"\<\"t\"\>", "1.7`"},
                    {"\<\"driveIntent\"\>", "0"},
                    {"\<\"motorTorque\"\>", "0.`"},
                    {"\<\"vw\"\>", "0.`"},
                    {"\<\"controllerState\"\>", 
                    RowBox[{"\[LeftAssociation]", 
                    RowBox[{
                    RowBox[{"\<\"lastError\"\>", "\[Rule]", "0.`"}], ",", 
                    RowBox[{"\<\"lastInt\"\>", "\[Rule]", "0.`"}], ",", 
                    RowBox[{"\<\"initialized\"\>", "\[Rule]", "True"}]}], 
                    "\[RightAssociation]"}]}
                    },
                    AutoDelete->False,
                    
                    GridBoxItemSize->{
                    "Columns" -> {{Automatic}}, "Rows" -> {{Automatic}}}],
                    "Grid"], {-1, -0.5}], 
                  {GrayLevel[0], LineBox[{{-5, 0}, {5, 0}}]}, 
                  {GrayLevel[0.5], CircleBox[{-2., 0.5}, 0.5], 
                   DiskBox[{-2., 0.5}, 0.1]}, 
                  {GrayLevel[0.5], DiskBox[{-2., 2.}, 0.1]}, 
                  {GrayLevel[0.5], Thickness[0.01], 
                   LineBox[{{-2., 0.5}, {-2., 2.}}]}, 
                  {RGBColor[0, 1, 0], DiskBox[{-2., 1.5}, 0.05]}, 
                  {RGBColor[0, 0, 1], Arrowheads[0.01], 
                   CircleBox[{-2., 0.5}, 0.3333333333333333, 
                    NCache[{
                    Rational[-1, 2] 
                    Pi, -1.5707963267948966`}, {-1.5707963267948966`, \
-1.5707963267948966`}]], 
                   ArrowBox[{{-2.0033332777780557`, 
                    0.16668333319444495`}, {-2., 0.16666666666666669`}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-2., 0.5}, {-2., 0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-2., 0.5}, {-2., 0.5}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.005], Arrowheads[0.04], 
                   ArrowBox[{{-2., 0.5}, {-2., 0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-2., 0.5}, 1.6, {1.5707963267948966`, 
                    1.5707963267948966`}], 
                   ArrowBox[{{-1.9840002666653331`, 
                    2.0999200006666645`}, {-2., 2.1}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Dashing[{Small, Small}], 
                   Arrowheads[0.02], 
                   CircleBox[{-2., 0.5}, 1.7, {1.5707963267948966`, 
                    1.5707963267948966`}], 
                   ArrowBox[{{-1.9830002833319165`, 2.199915000708331}, {-2., 
                    2.2}}]}, 
                  {RGBColor[0, 0, 1], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-2., 0.5}, 1.8, {1.5707963267948966`, 
                    1.5707963267948966`}], 
                   ArrowBox[{{-1.9820002999984998`, 
                    2.2999100007499975`}, {-2., 2.3}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.003], Arrowheads[0.025], 
                   CircleBox[{-2., 0.5}, 1.5, {1.5707963267948966`, 
                    1.5707963267948966`}], 
                   ArrowBox[{{-1.9850002499987498`, 1.999925000624998}, {-2., 
                    2.}}]}, 
                  {RGBColor[0, 0, 1], ArrowBox[{{0, 2.8}, {0, 2.8}}]}},
                 Frame->True,
                 FrameTicks->Automatic,
                 ImageSize->800,
                 PlotRange->{{-3, 3}, {-1, 3}}], 36->
                GraphicsBox[{InsetBox[
                   TagBox[GridBox[{
                    {"\<\"t\"\>", "1.75`"},
                    {"\<\"driveIntent\"\>", "0"},
                    {"\<\"motorTorque\"\>", "0.`"},
                    {"\<\"vw\"\>", "0.`"},
                    {"\<\"controllerState\"\>", 
                    RowBox[{"\[LeftAssociation]", 
                    RowBox[{
                    RowBox[{"\<\"lastError\"\>", "\[Rule]", "0.`"}], ",", 
                    RowBox[{"\<\"lastInt\"\>", "\[Rule]", "0.`"}], ",", 
                    RowBox[{"\<\"initialized\"\>", "\[Rule]", "True"}]}], 
                    "\[RightAssociation]"}]}
                    },
                    AutoDelete->False,
                    
                    GridBoxItemSize->{
                    "Columns" -> {{Automatic}}, "Rows" -> {{Automatic}}}],
                    "Grid"], {-1, -0.5}], 
                  {GrayLevel[0], LineBox[{{-5, 0}, {5, 0}}]}, 
                  {GrayLevel[0.5], CircleBox[{-2., 0.5}, 0.5], 
                   DiskBox[{-2., 0.5}, 0.1]}, 
                  {GrayLevel[0.5], DiskBox[{-2., 2.}, 0.1]}, 
                  {GrayLevel[0.5], Thickness[0.01], 
                   LineBox[{{-2., 0.5}, {-2., 2.}}]}, 
                  {RGBColor[0, 1, 0], DiskBox[{-2., 1.5}, 0.05]}, 
                  {RGBColor[0, 0, 1], Arrowheads[0.01], 
                   CircleBox[{-2., 0.5}, 0.3333333333333333, 
                    NCache[{
                    Rational[-1, 2] 
                    Pi, -1.5707963267948966`}, {-1.5707963267948966`, \
-1.5707963267948966`}]], 
                   ArrowBox[{{-2.0033332777780557`, 
                    0.16668333319444495`}, {-2., 0.16666666666666669`}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-2., 0.5}, {-2., 0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-2., 0.5}, {-2., 0.5}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.005], Arrowheads[0.04], 
                   ArrowBox[{{-2., 0.5}, {-2., 0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-2., 0.5}, 1.6, {1.5707963267948966`, 
                    1.5707963267948966`}], 
                   ArrowBox[{{-1.9840002666653331`, 
                    2.0999200006666645`}, {-2., 2.1}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Dashing[{Small, Small}], 
                   Arrowheads[0.02], 
                   CircleBox[{-2., 0.5}, 1.7, {1.5707963267948966`, 
                    1.5707963267948966`}], 
                   ArrowBox[{{-1.9830002833319165`, 2.199915000708331}, {-2., 
                    2.2}}]}, 
                  {RGBColor[0, 0, 1], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-2., 0.5}, 1.8, {1.5707963267948966`, 
                    1.5707963267948966`}], 
                   ArrowBox[{{-1.9820002999984998`, 
                    2.2999100007499975`}, {-2., 2.3}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.003], Arrowheads[0.025], 
                   CircleBox[{-2., 0.5}, 1.5, {1.5707963267948966`, 
                    1.5707963267948966`}], 
                   ArrowBox[{{-1.9850002499987498`, 1.999925000624998}, {-2., 
                    2.}}]}, 
                  {RGBColor[0, 0, 1], ArrowBox[{{0, 2.8}, {0, 2.8}}]}},
                 Frame->True,
                 FrameTicks->Automatic,
                 ImageSize->800,
                 PlotRange->{{-3, 3}, {-1, 3}}], 37->
                GraphicsBox[{InsetBox[
                   TagBox[GridBox[{
                    {"\<\"t\"\>", "1.8`"},
                    {"\<\"driveIntent\"\>", "0"},
                    {"\<\"motorTorque\"\>", "0.`"},
                    {"\<\"vw\"\>", "0.`"},
                    {"\<\"controllerState\"\>", 
                    RowBox[{"\[LeftAssociation]", 
                    RowBox[{
                    RowBox[{"\<\"lastError\"\>", "\[Rule]", "0.`"}], ",", 
                    RowBox[{"\<\"lastInt\"\>", "\[Rule]", "0.`"}], ",", 
                    RowBox[{"\<\"initialized\"\>", "\[Rule]", "True"}]}], 
                    "\[RightAssociation]"}]}
                    },
                    AutoDelete->False,
                    
                    GridBoxItemSize->{
                    "Columns" -> {{Automatic}}, "Rows" -> {{Automatic}}}],
                    "Grid"], {-1, -0.5}], 
                  {GrayLevel[0], LineBox[{{-5, 0}, {5, 0}}]}, 
                  {GrayLevel[0.5], CircleBox[{-2., 0.5}, 0.5], 
                   DiskBox[{-2., 0.5}, 0.1]}, 
                  {GrayLevel[0.5], DiskBox[{-2., 2.}, 0.1]}, 
                  {GrayLevel[0.5], Thickness[0.01], 
                   LineBox[{{-2., 0.5}, {-2., 2.}}]}, 
                  {RGBColor[0, 1, 0], DiskBox[{-2., 1.5}, 0.05]}, 
                  {RGBColor[0, 0, 1], Arrowheads[0.01], 
                   CircleBox[{-2., 0.5}, 0.3333333333333333, 
                    NCache[{
                    Rational[-1, 2] 
                    Pi, -1.5707963267948966`}, {-1.5707963267948966`, \
-1.5707963267948966`}]], 
                   ArrowBox[{{-2.0033332777780557`, 
                    0.16668333319444495`}, {-2., 0.16666666666666669`}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-2., 0.5}, {-2., 0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-2., 0.5}, {-2., 0.5}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.005], Arrowheads[0.04], 
                   ArrowBox[{{-2., 0.5}, {-2., 0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-2., 0.5}, 1.6, {1.5707963267948966`, 
                    1.5707963267948966`}], 
                   ArrowBox[{{-1.9840002666653331`, 
                    2.0999200006666645`}, {-2., 2.1}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Dashing[{Small, Small}], 
                   Arrowheads[0.02], 
                   CircleBox[{-2., 0.5}, 1.7, {1.5707963267948966`, 
                    1.5707963267948966`}], 
                   ArrowBox[{{-1.9830002833319165`, 2.199915000708331}, {-2., 
                    2.2}}]}, 
                  {RGBColor[0, 0, 1], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-2., 0.5}, 1.8, {1.5707963267948966`, 
                    1.5707963267948966`}], 
                   ArrowBox[{{-1.9820002999984998`, 
                    2.2999100007499975`}, {-2., 2.3}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.003], Arrowheads[0.025], 
                   CircleBox[{-2., 0.5}, 1.5, {1.5707963267948966`, 
                    1.5707963267948966`}], 
                   ArrowBox[{{-1.9850002499987498`, 1.999925000624998}, {-2., 
                    2.}}]}, 
                  {RGBColor[0, 0, 1], ArrowBox[{{0, 2.8}, {0, 2.8}}]}},
                 Frame->True,
                 FrameTicks->Automatic,
                 ImageSize->800,
                 PlotRange->{{-3, 3}, {-1, 3}}], 38->
                GraphicsBox[{InsetBox[
                   TagBox[GridBox[{
                    {"\<\"t\"\>", "1.85`"},
                    {"\<\"driveIntent\"\>", "0"},
                    {"\<\"motorTorque\"\>", "0.`"},
                    {"\<\"vw\"\>", "0.`"},
                    {"\<\"controllerState\"\>", 
                    RowBox[{"\[LeftAssociation]", 
                    RowBox[{
                    RowBox[{"\<\"lastError\"\>", "\[Rule]", "0.`"}], ",", 
                    RowBox[{"\<\"lastInt\"\>", "\[Rule]", "0.`"}], ",", 
                    RowBox[{"\<\"initialized\"\>", "\[Rule]", "True"}]}], 
                    "\[RightAssociation]"}]}
                    },
                    AutoDelete->False,
                    
                    GridBoxItemSize->{
                    "Columns" -> {{Automatic}}, "Rows" -> {{Automatic}}}],
                    "Grid"], {-1, -0.5}], 
                  {GrayLevel[0], LineBox[{{-5, 0}, {5, 0}}]}, 
                  {GrayLevel[0.5], CircleBox[{-2., 0.5}, 0.5], 
                   DiskBox[{-2., 0.5}, 0.1]}, 
                  {GrayLevel[0.5], DiskBox[{-2., 2.}, 0.1]}, 
                  {GrayLevel[0.5], Thickness[0.01], 
                   LineBox[{{-2., 0.5}, {-2., 2.}}]}, 
                  {RGBColor[0, 1, 0], DiskBox[{-2., 1.5}, 0.05]}, 
                  {RGBColor[0, 0, 1], Arrowheads[0.01], 
                   CircleBox[{-2., 0.5}, 0.3333333333333333, 
                    NCache[{
                    Rational[-1, 2] 
                    Pi, -1.5707963267948966`}, {-1.5707963267948966`, \
-1.5707963267948966`}]], 
                   ArrowBox[{{-2.0033332777780557`, 
                    0.16668333319444495`}, {-2., 0.16666666666666669`}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-2., 0.5}, {-2., 0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-2., 0.5}, {-2., 0.5}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.005], Arrowheads[0.04], 
                   ArrowBox[{{-2., 0.5}, {-2., 0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-2., 0.5}, 1.6, {1.5707963267948966`, 
                    1.5707963267948966`}], 
                   ArrowBox[{{-1.9840002666653331`, 
                    2.0999200006666645`}, {-2., 2.1}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Dashing[{Small, Small}], 
                   Arrowheads[0.02], 
                   CircleBox[{-2., 0.5}, 1.7, {1.5707963267948966`, 
                    1.5707963267948966`}], 
                   ArrowBox[{{-1.9830002833319165`, 2.199915000708331}, {-2., 
                    2.2}}]}, 
                  {RGBColor[0, 0, 1], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-2., 0.5}, 1.8, {1.5707963267948966`, 
                    1.5707963267948966`}], 
                   ArrowBox[{{-1.9820002999984998`, 
                    2.2999100007499975`}, {-2., 2.3}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.003], Arrowheads[0.025], 
                   CircleBox[{-2., 0.5}, 1.5, {1.5707963267948966`, 
                    1.5707963267948966`}], 
                   ArrowBox[{{-1.9850002499987498`, 1.999925000624998}, {-2., 
                    2.}}]}, 
                  {RGBColor[0, 0, 1], ArrowBox[{{0, 2.8}, {0, 2.8}}]}},
                 Frame->True,
                 FrameTicks->Automatic,
                 ImageSize->800,
                 PlotRange->{{-3, 3}, {-1, 3}}], 39->
                GraphicsBox[{InsetBox[
                   TagBox[GridBox[{
                    {"\<\"t\"\>", "1.9000000000000001`"},
                    {"\<\"driveIntent\"\>", "0"},
                    {"\<\"motorTorque\"\>", "0.`"},
                    {"\<\"vw\"\>", "0.`"},
                    {"\<\"controllerState\"\>", 
                    RowBox[{"\[LeftAssociation]", 
                    RowBox[{
                    RowBox[{"\<\"lastError\"\>", "\[Rule]", "0.`"}], ",", 
                    RowBox[{"\<\"lastInt\"\>", "\[Rule]", "0.`"}], ",", 
                    RowBox[{"\<\"initialized\"\>", "\[Rule]", "True"}]}], 
                    "\[RightAssociation]"}]}
                    },
                    AutoDelete->False,
                    
                    GridBoxItemSize->{
                    "Columns" -> {{Automatic}}, "Rows" -> {{Automatic}}}],
                    "Grid"], {-1, -0.5}], 
                  {GrayLevel[0], LineBox[{{-5, 0}, {5, 0}}]}, 
                  {GrayLevel[0.5], CircleBox[{-2., 0.5}, 0.5], 
                   DiskBox[{-2., 0.5}, 0.1]}, 
                  {GrayLevel[0.5], DiskBox[{-2., 2.}, 0.1]}, 
                  {GrayLevel[0.5], Thickness[0.01], 
                   LineBox[{{-2., 0.5}, {-2., 2.}}]}, 
                  {RGBColor[0, 1, 0], DiskBox[{-2., 1.5}, 0.05]}, 
                  {RGBColor[0, 0, 1], Arrowheads[0.01], 
                   CircleBox[{-2., 0.5}, 0.3333333333333333, 
                    NCache[{
                    Rational[-1, 2] 
                    Pi, -1.5707963267948966`}, {-1.5707963267948966`, \
-1.5707963267948966`}]], 
                   ArrowBox[{{-2.0033332777780557`, 
                    0.16668333319444495`}, {-2., 0.16666666666666669`}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-2., 0.5}, {-2., 0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-2., 0.5}, {-2., 0.5}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.005], Arrowheads[0.04], 
                   ArrowBox[{{-2., 0.5}, {-2., 0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-2., 0.5}, 1.6, {1.5707963267948966`, 
                    1.5707963267948966`}], 
                   ArrowBox[{{-1.9840002666653331`, 
                    2.0999200006666645`}, {-2., 2.1}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Dashing[{Small, Small}], 
                   Arrowheads[0.02], 
                   CircleBox[{-2., 0.5}, 1.7, {1.5707963267948966`, 
                    1.5707963267948966`}], 
                   ArrowBox[{{-1.9830002833319165`, 2.199915000708331}, {-2., 
                    2.2}}]}, 
                  {RGBColor[0, 0, 1], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-2., 0.5}, 1.8, {1.5707963267948966`, 
                    1.5707963267948966`}], 
                   ArrowBox[{{-1.9820002999984998`, 
                    2.2999100007499975`}, {-2., 2.3}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.003], Arrowheads[0.025], 
                   CircleBox[{-2., 0.5}, 1.5, {1.5707963267948966`, 
                    1.5707963267948966`}], 
                   ArrowBox[{{-1.9850002499987498`, 1.999925000624998}, {-2., 
                    2.}}]}, 
                  {RGBColor[0, 0, 1], ArrowBox[{{0, 2.8}, {0, 2.8}}]}},
                 Frame->True,
                 FrameTicks->Automatic,
                 ImageSize->800,
                 PlotRange->{{-3, 3}, {-1, 3}}], 40->
                GraphicsBox[{InsetBox[
                   TagBox[GridBox[{
                    {"\<\"t\"\>", "1.95`"},
                    {"\<\"driveIntent\"\>", "0"},
                    {"\<\"motorTorque\"\>", "0.`"},
                    {"\<\"vw\"\>", "0.`"},
                    {"\<\"controllerState\"\>", 
                    RowBox[{"\[LeftAssociation]", 
                    RowBox[{
                    RowBox[{"\<\"lastError\"\>", "\[Rule]", "0.`"}], ",", 
                    RowBox[{"\<\"lastInt\"\>", "\[Rule]", "0.`"}], ",", 
                    RowBox[{"\<\"initialized\"\>", "\[Rule]", "True"}]}], 
                    "\[RightAssociation]"}]}
                    },
                    AutoDelete->False,
                    
                    GridBoxItemSize->{
                    "Columns" -> {{Automatic}}, "Rows" -> {{Automatic}}}],
                    "Grid"], {-1, -0.5}], 
                  {GrayLevel[0], LineBox[{{-5, 0}, {5, 0}}]}, 
                  {GrayLevel[0.5], CircleBox[{-2., 0.5}, 0.5], 
                   DiskBox[{-2., 0.5}, 0.1]}, 
                  {GrayLevel[0.5], DiskBox[{-2., 2.}, 0.1]}, 
                  {GrayLevel[0.5], Thickness[0.01], 
                   LineBox[{{-2., 0.5}, {-2., 2.}}]}, 
                  {RGBColor[0, 1, 0], DiskBox[{-2., 1.5}, 0.05]}, 
                  {RGBColor[0, 0, 1], Arrowheads[0.01], 
                   CircleBox[{-2., 0.5}, 0.3333333333333333, 
                    NCache[{
                    Rational[-1, 2] 
                    Pi, -1.5707963267948966`}, {-1.5707963267948966`, \
-1.5707963267948966`}]], 
                   ArrowBox[{{-2.0033332777780557`, 
                    0.16668333319444495`}, {-2., 0.16666666666666669`}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-2., 0.5}, {-2., 0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-2., 0.5}, {-2., 0.5}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.005], Arrowheads[0.04], 
                   ArrowBox[{{-2., 0.5}, {-2., 0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-2., 0.5}, 1.6, {1.5707963267948966`, 
                    1.5707963267948966`}], 
                   ArrowBox[{{-1.9840002666653331`, 
                    2.0999200006666645`}, {-2., 2.1}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Dashing[{Small, Small}], 
                   Arrowheads[0.02], 
                   CircleBox[{-2., 0.5}, 1.7, {1.5707963267948966`, 
                    1.5707963267948966`}], 
                   ArrowBox[{{-1.9830002833319165`, 2.199915000708331}, {-2., 
                    2.2}}]}, 
                  {RGBColor[0, 0, 1], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-2., 0.5}, 1.8, {1.5707963267948966`, 
                    1.5707963267948966`}], 
                   ArrowBox[{{-1.9820002999984998`, 
                    2.2999100007499975`}, {-2., 2.3}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.003], Arrowheads[0.025], 
                   CircleBox[{-2., 0.5}, 1.5, {1.5707963267948966`, 
                    1.5707963267948966`}], 
                   ArrowBox[{{-1.9850002499987498`, 1.999925000624998}, {-2., 
                    2.}}]}, 
                  {RGBColor[0, 0, 1], ArrowBox[{{0, 2.8}, {0, 2.8}}]}},
                 Frame->True,
                 FrameTicks->Automatic,
                 ImageSize->800,
                 PlotRange->{{-3, 3}, {-1, 3}}], 41->
                GraphicsBox[{InsetBox[
                   TagBox[GridBox[{
                    {"\<\"t\"\>", "2.`"},
                    {"\<\"driveIntent\"\>", "0.5`"},
                    {"\<\"motorTorque\"\>", "200"},
                    {"\<\"vw\"\>", "0.`"},
                    {"\<\"controllerState\"\>", 
                    RowBox[{"\[LeftAssociation]", 
                    RowBox[{
                    RowBox[{"\<\"lastError\"\>", "\[Rule]", "0.15`"}], ",", 
                    RowBox[{"\<\"lastInt\"\>", "\[Rule]", "0.00015`"}], ",", 
                    RowBox[{"\<\"initialized\"\>", "\[Rule]", "True"}]}], 
                    "\[RightAssociation]"}]}
                    },
                    AutoDelete->False,
                    
                    GridBoxItemSize->{
                    "Columns" -> {{Automatic}}, "Rows" -> {{Automatic}}}],
                    "Grid"], {-1, -0.5}], 
                  {GrayLevel[0], LineBox[{{-5, 0}, {5, 0}}]}, 
                  {GrayLevel[0.5], CircleBox[{-2., 0.5}, 0.5], 
                   DiskBox[{-2., 0.5}, 0.1]}, 
                  {GrayLevel[0.5], DiskBox[{-2., 2.}, 0.1]}, 
                  {GrayLevel[0.5], Thickness[0.01], 
                   LineBox[{{-2., 0.5}, {-2., 2.}}]}, 
                  {RGBColor[0, 1, 0], DiskBox[{-2., 1.5}, 0.05]}, 
                  {RGBColor[0, 0, 1], Arrowheads[0.01], 
                   CircleBox[{-2., 0.5}, 0.3333333333333333, 
                    NCache[{
                    Rational[-1, 2] Pi, 
                    4.429203673205103}, {-1.5707963267948966`, 
                    4.429203673205103}]], 
                   ArrowBox[{{-2.0963343567931205`, 
                    0.18089060995895645`}, {-2.093138499399642, 
                    0.1799432377832113}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-2., 0.5}, {-6., 0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-2., 0.5}, {-2., 0.5}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.005], Arrowheads[0.04], 
                   ArrowBox[{{-2., 0.5}, {-6., 0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-2., 0.5}, 1.6, {1.5707963267948966`, 
                    1.5707963267948966`}], 
                   ArrowBox[{{-1.9840002666653331`, 
                    2.0999200006666645`}, {-2., 2.1}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Dashing[{Small, Small}], 
                   Arrowheads[0.02], 
                   CircleBox[{-2., 0.5}, 1.7, {
                    1.5707963267948966`, -0.7473723199741564}], 
                   ArrowBox[{{-0.741595705802204, -0.6429867157340663}, \
{-0.7530883021530879, -0.6555134000834999}}]}, 
                  {RGBColor[0, 0, 1], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-2., 0.5}, 1.8, {1.5707963267948966`, 
                    0.4596852156837854}], 
                   ArrowBox[{{-0.39492055006352844`, 
                    1.3146901002170297`}, {-0.38685403814607855`, 
                    1.2985988390640113`}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.003], Arrowheads[0.025], 
                   CircleBox[{-2., 0.5}, 1.5, {
                    1.5707963267948966`, -2.389794806210029}], 
                   ArrowBox[{{-3.085394396549022, -0.5353352133198137}, \
{-3.0956933068596326`, -0.5244296839232074}}]}, 
                  {RGBColor[0, 0, 1], ArrowBox[{{0, 2.8}, {1.5, 2.8}}]}},
                 Frame->True,
                 FrameTicks->Automatic,
                 ImageSize->800,
                 PlotRange->{{-3, 3}, {-1, 3}}], 42->
                GraphicsBox[{InsetBox[
                   TagBox[GridBox[{
                    {"\<\"t\"\>", "2.05`"},
                    {"\<\"driveIntent\"\>", "0.5`"},
                    {"\<\"motorTorque\"\>", "20.94095840065306`"},
                    {"\<\"vw\"\>", 
                    RowBox[{"-", "0.024733802623653788`"}]},
                    {"\<\"controllerState\"\>", 
                    RowBox[{"\[LeftAssociation]", 
                    RowBox[{
                    
                    RowBox[{"\<\"lastError\"\>", "\[Rule]", 
                    "0.13291066238111304`"}], ",", 
                    
                    RowBox[{"\<\"lastInt\"\>", "\[Rule]", 
                    "0.007417415904331226`"}], ",", 
                    RowBox[{"\<\"initialized\"\>", "\[Rule]", "True"}]}], 
                    "\[RightAssociation]"}]}
                    },
                    AutoDelete->False,
                    
                    GridBoxItemSize->{
                    "Columns" -> {{Automatic}}, "Rows" -> {{Automatic}}}],
                    "Grid"], {-1, -0.5}], 
                  {GrayLevel[0], LineBox[{{-5, 0}, {5, 0}}]}, 
                  {GrayLevel[0.5], 
                   CircleBox[{-2.0007801516288737`, 0.5}, 0.5], 
                   DiskBox[{-2.0007801516288737`, 0.5}, 0.1]}, 
                  {GrayLevel[0.5], 
                   DiskBox[{-1.9629017736850436`, 1.9995216665604882`}, 0.1]}, 
                  {GrayLevel[0.5], Thickness[0.01], 
                   LineBox[{{-2.0007801516288737`, 
                    0.5}, {-1.9629017736850436`, 1.9995216665604882`}}]}, 
                  {RGBColor[0, 1, 0], 
                   DiskBox[{-1.9755278996663204`, 1.4996811110403256`}, 
                    0.05]}, 
                  {RGBColor[0, 0, 1], Arrowheads[0.01], 
                   CircleBox[{-2.0007801516288737`, 0.5}, 0.3333333333333333, 
                    NCache[{
                    Rational[-1, 2] 
                    Pi, -0.9425675747753048}, {-1.5707963267948966`, \
-0.9425675747753048}]], 
                   ArrowBox[{{-1.8075825952047997`, 
                    0.22836455440631864`}, {-1.8048759458185333`, 
                    0.2303100794302263}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-2.0007801516288737`, 0.5}, {-2.419599319641935,
                     0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-2.0007801516288737`, 
                    0.5}, {-2.1864509219146115`, 0.5}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.005], Arrowheads[0.04], 
                   ArrowBox[{{-2.0007801516288737`, 
                    0.5}, {-2.6052700899276724`, 0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-2.0007801516288737`, 0.5}, 1.6, {
                    1.5455413902684563`, 1.429460141811312}], 
                   ArrowBox[{{-1.7912458530458808`, 
                    2.0862204694547772`}, {-1.7753943893477129`, 
                    2.084045851028606}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Dashing[{Small, Small}], 
                   Arrowheads[0.02], 
                   CircleBox[{-2.0007801516288737`, 0.5}, 1.7, {
                    1.5455413902684563`, 1.1953281766827781`}], 
                   ArrowBox[{{-1.3932229902973217`, 
                    2.0877261400237592`}, {-1.3773763711217026`, 
                    2.0815712840240135`}}]}, 
                  {RGBColor[0, 0, 1], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-2.0007801516288737`, 0.5}, 1.8, {
                    1.5455413902684563`, 1.4292027324870504`}], 
                   ArrowBox[{{-1.764594727106064, 
                    2.2844372909247834`}, {-1.7467624607743601`, 
                    2.281986254922563}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.003], Arrowheads[0.025], 
                   CircleBox[{-2.0007801516288737`, 0.5}, 1.5, {
                    1.5455413902684563`, 0.8852066938460467}], 
                   ArrowBox[{{-1.062743843706792, 
                    1.6705075330897747`}, {-1.0510858648840566`, 
                    1.6610688014602122`}}]}, 
                  {RGBColor[0, 0, 1], ArrowBox[{{0, 2.8}, {1.5, 2.8}}]}},
                 Frame->True,
                 FrameTicks->Automatic,
                 ImageSize->800,
                 PlotRange->{{-3, 3}, {-1, 3}}], 43->
                GraphicsBox[{InsetBox[
                   TagBox[GridBox[{
                    {"\<\"t\"\>", "2.1`"},
                    {"\<\"driveIntent\"\>", "0.5`"},
                    {"\<\"motorTorque\"\>", 
                    RowBox[{"-", "33.16392495716223`"}]},
                    {"\<\"vw\"\>", 
                    RowBox[{"-", "0.027684189010418102`"}]},
                    {"\<\"controllerState\"\>", 
                    RowBox[{"\[LeftAssociation]", 
                    RowBox[{
                    
                    RowBox[{"\<\"lastError\"\>", "\[Rule]", 
                    "0.0882920071946943`"}], ",", 
                    
                    RowBox[{"\<\"lastInt\"\>", "\[Rule]", 
                    "0.012972616605768638`"}], ",", 
                    RowBox[{"\<\"initialized\"\>", "\[Rule]", "True"}]}], 
                    "\[RightAssociation]"}]}
                    },
                    AutoDelete->False,
                    
                    GridBoxItemSize->{
                    "Columns" -> {{Automatic}}, "Rows" -> {{Automatic}}}],
                    "Grid"], {-1, -0.5}], 
                  {GrayLevel[0], LineBox[{{-5, 0}, {5, 0}}]}, 
                  {GrayLevel[0.5], 
                   CircleBox[{-2.0021428226549474`, 0.5}, 0.5], 
                   DiskBox[{-2.0021428226549474`, 0.5}, 0.1]}, 
                  {GrayLevel[0.5], 
                   DiskBox[{-1.8957469039822512`, 1.9962218780948877`}, 0.1]}, 
                  {GrayLevel[0.5], Thickness[0.01], 
                   LineBox[{{-2.0021428226549474`, 
                    0.5}, {-1.8957469039822512`, 1.9962218780948877`}}]}, 
                  {RGBColor[0, 1, 0], 
                   DiskBox[{-1.9312122102064833`, 1.4974812520632583`}, 
                    0.05]}, 
                  {RGBColor[0, 0, 1], Arrowheads[0.01], 
                   CircleBox[{-2.0021428226549474`, 0.5}, 0.3333333333333333, 
                    NCache[{
                    Rational[-1, 2] 
                    Pi, -2.5657140755097636`}, {-1.5707963267948966`, \
-2.5657140755097636`}]], 
                   ArrowBox[{{-2.2798850272900824`, 
                    0.3156894499071581}, {-2.281714215078233, 
                    0.3184760411140822}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-2.0021428226549474`, 
                    0.5}, {-1.3388643235117028`, 0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-2.0021428226549474`, 0.5}, {-2.522522575479171,
                     0.5}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.005], Arrowheads[0.04], 
                   ArrowBox[{{-2.0021428226549474`, 
                    0.5}, {-1.8592440763359261`, 0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-2.0021428226549474`, 0.5}, 1.6, {
                    1.4998061021716385`, 1.173747497218015}], 
                   ArrowBox[{{-1.398211303907448, 
                    1.9816432501325476`}, {-1.3834253146697293`, 
                    1.9755299540546651`}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Dashing[{Small, Small}], 
                   Arrowheads[0.02], 
                   CircleBox[{-2.0021428226549474`, 0.5}, 1.7, {
                    1.4998061021716385`, 1.5824087345694346`}], 
                   ArrowBox[{{-2.0048839146839197`, 
                    2.199997790120472}, {-2.0218834722001167`, 
                    2.1998853804758527`}}]}, 
                  {RGBColor[0, 0, 1], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-2.0021428226549474`, 0.5}, 1.8, {
                    1.4998061021716385`, 1.6840501297114288`}], 
                   ArrowBox[{{-2.187669596492263, 
                    2.290413308761281}, {-2.2055641549177603`, 
                    2.2884685520244474`}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.003], Arrowheads[0.025], 
                   CircleBox[{-2.0021428226549474`, 0.5}, 1.5, {
                    1.4998061021716385`, 1.4667194066530238`}], 
                   ArrowBox[{{-1.8612355065146051`, 
                    1.9933670440511688`}, {-1.846309130274452, 
                    1.9918833266442324`}}]}, 
                  {RGBColor[0, 0, 1], ArrowBox[{{0, 2.8}, {1.5, 2.8}}]}},
                 Frame->True,
                 FrameTicks->Automatic,
                 ImageSize->800,
                 PlotRange->{{-3, 3}, {-1, 3}}], 44->
                GraphicsBox[{InsetBox[
                   TagBox[GridBox[{
                    {"\<\"t\"\>", "2.15`"},
                    {"\<\"driveIntent\"\>", "0.5`"},
                    {"\<\"motorTorque\"\>", 
                    RowBox[{"-", "66.08363873078383`"}]},
                    {"\<\"vw\"\>", 
                    RowBox[{"-", "0.022447440669817305`"}]},
                    {"\<\"controllerState\"\>", 
                    RowBox[{"\[LeftAssociation]", 
                    RowBox[{
                    
                    RowBox[{"\<\"lastError\"\>", "\[Rule]", 
                    "0.03918868990811447`"}], ",", 
                    
                    RowBox[{"\<\"lastInt\"\>", "\[Rule]", 
                    "0.016132591650513818`"}], ",", 
                    RowBox[{"\<\"initialized\"\>", "\[Rule]", "True"}]}], 
                    "\[RightAssociation]"}]}
                    },
                    AutoDelete->False,
                    
                    GridBoxItemSize->{
                    "Columns" -> {{Automatic}}, "Rows" -> {{Automatic}}}],
                    "Grid"], {-1, -0.5}], 
                  {GrayLevel[0], LineBox[{{-5, 0}, {5, 0}}]}, 
                  {GrayLevel[0.5], 
                   CircleBox[{-2.0034152195170334`, 0.5}, 0.5], 
                   DiskBox[{-2.0034152195170334`, 0.5}, 0.1]}, 
                  {GrayLevel[0.5], 
                   DiskBox[{-1.8260634521161303`, 1.9894785498958272`}, 0.1]}, 
                  {GrayLevel[0.5], Thickness[0.01], 
                   LineBox[{{-2.0034152195170334`, 
                    0.5}, {-1.8260634521161303`, 1.9894785498958272`}}]}, 
                  {RGBColor[0, 1, 0], 
                   DiskBox[{-1.8851807079164313`, 1.4929856999305513`}, 
                    0.05]}, 
                  {RGBColor[0, 0, 1], Arrowheads[0.01], 
                   CircleBox[{-2.0034152195170334`, 0.5}, 0.3333333333333333, 
                    NCache[{
                    Rational[-1, 2] 
                    Pi, -3.5533054887184115`}, {-1.5707963267948966`, \
-3.5533054887184115`}]], 
                   ArrowBox[{{-2.3102127641192434`, 
                    0.6303317986416438}, {-2.308894128105287, 
                    0.63339320641937}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-2.0034152195170334`, 
                    0.5}, {-0.6817424449013567, 0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-2.0034152195170334`, 0.5}, {-2.866928845392461,
                     0.5}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.005], Arrowheads[0.04], 
                   ArrowBox[{{-2.0034152195170334`, 
                    0.5}, {-1.5452560707767846`, 0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-2.0034152195170334`, 0.5}, 1.6, {
                    1.452284593116276, 0.9087762463093803}], 
                   ArrowBox[{{-1.032545902414142, 
                    1.7717754397330392`}, {-1.0198769030396526`, 
                    1.7620033201306722`}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Dashing[{Small, Small}], 
                   Arrowheads[0.02], 
                   CircleBox[{-2.0034152195170334`, 0.5}, 1.7, {
                    1.452284593116276, 1.7159034028144258`}], 
                   ArrowBox[{{-2.2323991199270288`, 
                    2.1845077540198576`}, {-2.2492324676177278`, 
                    2.1821337284937252`}}]}, 
                  {RGBColor[0, 0, 1], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-2.0034152195170334`, 0.5}, 1.8, {
                    1.452284593116276, 1.8194159193984085`}], 
                   ArrowBox[{{-2.428866024374109, 
                    2.2489973163634263`}, {-2.4463344336766757`, 
                    2.2446554300858965`}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.003], Arrowheads[0.025], 
                   CircleBox[{-2.0034152195170334`, 0.5}, 1.5, {
                    1.452284593116276, 1.6118679323853824`}], 
                   ArrowBox[{{-2.050015128785738, 
                    1.9992759747478608`}, {-2.06500530867909, 
                    1.9987350202477454`}}]}, 
                  {RGBColor[0, 0, 1], ArrowBox[{{0, 2.8}, {1.5, 2.8}}]}},
                 Frame->True,
                 FrameTicks->Automatic,
                 ImageSize->800,
                 PlotRange->{{-3, 3}, {-1, 3}}], 45->
                GraphicsBox[{InsetBox[
                   TagBox[GridBox[{
                    {"\<\"t\"\>", "2.2`"},
                    {"\<\"driveIntent\"\>", "0.5`"},
                    {"\<\"motorTorque\"\>", 
                    RowBox[{"-", "89.94892120231587`"}]},
                    {"\<\"vw\"\>", 
                    RowBox[{"-", "0.013310854074773129`"}]},
                    {"\<\"controllerState\"\>", 
                    RowBox[{"\[LeftAssociation]", 
                    RowBox[{
                    RowBox[{"\<\"lastError\"\>", "\[Rule]", 
                    RowBox[{"-", "0.0071883155461822945`"}]}], ",", 
                    
                    RowBox[{"\<\"lastInt\"\>", "\[Rule]", 
                    "0.016891579015443674`"}], ",", 
                    RowBox[{"\<\"initialized\"\>", "\[Rule]", "True"}]}], 
                    "\[RightAssociation]"}]}
                    },
                    AutoDelete->False,
                    
                    GridBoxItemSize->{
                    "Columns" -> {{Automatic}}, "Rows" -> {{Automatic}}}],
                    "Grid"], {-1, -0.5}], 
                  {GrayLevel[0], LineBox[{{-5, 0}, {5, 0}}]}, 
                  {GrayLevel[0.5], 
                   CircleBox[{-2.0043166922333016`, 0.5}, 0.5], 
                   DiskBox[{-2.0043166922333016`, 0.5}, 0.1]}, 
                  {GrayLevel[0.5], 
                   DiskBox[{-1.7622856299409166`, 1.9803448803861956`}, 0.1]}, 
                  {GrayLevel[0.5], Thickness[0.01], 
                   LineBox[{{-2.0043166922333016`, 
                    0.5}, {-1.7622856299409166`, 1.9803448803861956`}}]}, 
                  {RGBColor[0, 1, 0], 
                   DiskBox[{-1.842962650705045, 1.4868965869241304`}, 0.05]}, 
                  {RGBColor[0, 0, 1], Arrowheads[0.01], 
                   CircleBox[{-2.0043166922333016`, 0.5}, 0.3333333333333333, 
                    NCache[{
                    Rational[-1, 2] 
                    Pi, -4.269263962864373}, {-1.5707963267948966`, \
-4.269263962864373}]], 
                   ArrowBox[{{-2.1502424651082954`, 
                    0.7996944776300471}, {-2.1472382740529836`, 
                    0.8011387264389472}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-2.0043166922333016`, 
                    0.5}, {-0.20533826818698397`, 0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-2.0043166922333016`, 
                    0.5}, {-3.1755230841000275`, 0.5}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.005], Arrowheads[0.04], 
                   ArrowBox[{{-2.0043166922333016`, 
                    0.5}, {-1.3765446600537097`, 0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-2.0043166922333016`, 0.5}, 1.6, {
                    1.4087338075954297`, 0.6670107085236747}], 
                   ArrowBox[{{-0.7571984723303231, 
                    1.5023453225241423`}, {-0.7472375415531647, 
                    1.489824231328626}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Dashing[{Small, Small}], 
                   Arrowheads[0.02], 
                   CircleBox[{-2.0043166922333016`, 0.5}, 1.7, {
                    1.4087338075954297`, 1.767681236614983}], 
                   ArrowBox[{{-2.3201749103815796`, 
                    2.170399229534124}, {-2.336862831499141, 
                    2.167157180729935}}]}, 
                  {RGBColor[0, 0, 1], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-2.0043166922333016`, 0.5}, 1.8, {
                    1.4087338075954297`, 1.9084500364971846`}], 
                   ArrowBox[{{-2.5835970995673696`, 
                    2.2042400680886702`}, {-2.6006102524306645`, 
                    2.198362149268275}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.003], Arrowheads[0.025], 
                   CircleBox[{-2.0043166922333016`, 0.5}, 1.5, {
                    1.4087338075954297`, 1.6240290628231575`}], 
                   ArrowBox[{{-2.0691455969167913`, 
                    1.9985984162268218`}, {-2.0841280898956835`, 
                    1.9978752086883529`}}]}, 
                  {RGBColor[0, 0, 1], ArrowBox[{{0, 2.8}, {1.5, 2.8}}]}},
                 Frame->True,
                 FrameTicks->Automatic,
                 ImageSize->800,
                 PlotRange->{{-3, 3}, {-1, 3}}], 46->
                GraphicsBox[{InsetBox[
                   TagBox[GridBox[{
                    {"\<\"t\"\>", "2.25`"},
                    {"\<\"driveIntent\"\>", "0.5`"},
                    {"\<\"motorTorque\"\>", 
                    RowBox[{"-", "108.61719425716336`"}]},
                    {"\<\"vw\"\>", 
                    RowBox[{"-", "0.0017644829062387387`"}]},
                    {"\<\"controllerState\"\>", 
                    RowBox[{"\[LeftAssociation]", 
                    RowBox[{
                    RowBox[{"\<\"lastError\"\>", "\[Rule]", 
                    RowBox[{"-", "0.048661103899543534`"}]}], ",", 
                    
                    RowBox[{"\<\"lastInt\"\>", "\[Rule]", 
                    "0.015452353075532029`"}], ",", 
                    RowBox[{"\<\"initialized\"\>", "\[Rule]", "True"}]}], 
                    "\[RightAssociation]"}]}
                    },
                    AutoDelete->False,
                    
                    GridBoxItemSize->{
                    "Columns" -> {{Automatic}}, "Rows" -> {{Automatic}}}],
                    "Grid"], {-1, -0.5}], 
                  {GrayLevel[0], LineBox[{{-5, 0}, {5, 0}}]}, 
                  {GrayLevel[0.5], 
                   CircleBox[{-2.0046961900313525`, 0.5}, 0.5], 
                   DiskBox[{-2.0046961900313525`, 0.5}, 0.1]}, 
                  {GrayLevel[0.5], 
                   DiskBox[{-1.7067441004083852`, 1.970110387790423}, 0.1]}, 
                  {GrayLevel[0.5], Thickness[0.01], 
                   LineBox[{{-2.0046961900313525`, 
                    0.5}, {-1.7067441004083852`, 1.970110387790423}}]}, 
                  {RGBColor[0, 1, 0], 
                   DiskBox[{-1.8060614636160408`, 1.4800735918602819`}, 
                    0.05]}, 
                  {RGBColor[0, 0, 1], Arrowheads[0.01], 
                   CircleBox[{-2.0046961900313525`, 0.5}, 0.3333333333333333, 
                    NCache[{
                    Rational[-1, 2] 
                    Pi, -4.829312154509798}, {-1.5707963267948966`, \
-4.829312154509798}]], 
                   ArrowBox[{{-1.9691230046332113`, 
                    0.8314297204412129}, {-1.9658105413112577`, 
                    0.8310574231681394}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-2.0046961900313525`, 0.5}, {0.1676476951119148,
                     0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-2.0046961900313525`, 
                    0.5}, {-3.4365405157495457`, 0.5}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.005], Arrowheads[0.04], 
                   ArrowBox[{{-2.0046961900313525`, 
                    0.5}, {-1.2641966306062784`, 0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-2.0046961900313525`, 0.5}, 1.6, {
                    1.3708316348614122`, 0.45773417870482724`}], 
                   ArrowBox[{{-0.5765487371094578, 
                    1.221383984236346}, {-0.5694064242747439, 
                    1.2070666788318765`}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Dashing[{Small, Small}], 
                   Arrowheads[0.02], 
                   CircleBox[{-2.0046961900313525`, 0.5}, 1.7, {
                    1.3708316348614122`, 1.7912414589550947`}], 
                   ArrowBox[{{-2.3598180847596857`, 
                    2.1624946435656742`}, {-2.3764249981675176`, 
                    2.158860359765603}}]}, 
                  {RGBColor[0, 0, 1], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-2.0046961900313525`, 0.5}, 1.8, {
                    1.3708316348614122`, 1.9742604918456532`}], 
                   ArrowBox[{{-2.6947984872894466`, 
                    2.162455659354288}, {-2.71138826198111, 
                    2.1554716293079017`}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.003], Arrowheads[0.025], 
                   CircleBox[{-2.0046961900313525`, 0.5}, 1.5, {
                    1.3708316348614122`, 1.5974401106483171`}], 
                   ArrowBox[{{-2.029660713180751, 
                    1.9997922431403368`}, {-2.044657137432275, 
                    1.9994676130823301`}}]}, 
                  {RGBColor[0, 0, 1], ArrowBox[{{0, 2.8}, {1.5, 2.8}}]}},
                 Frame->True,
                 FrameTicks->Automatic,
                 ImageSize->800,
                 PlotRange->{{-3, 3}, {-1, 3}}], 47->
                GraphicsBox[{InsetBox[
                   TagBox[GridBox[{
                    {"\<\"t\"\>", "2.3000000000000003`"},
                    {"\<\"driveIntent\"\>", "0.5`"},
                    {"\<\"motorTorque\"\>", 
                    RowBox[{"-", "123.41216119755113`"}]},
                    {"\<\"vw\"\>", "0.01154894075248287`"},
                    {"\<\"controllerState\"\>", 
                    RowBox[{"\[LeftAssociation]", 
                    RowBox[{
                    RowBox[{"\<\"lastError\"\>", "\[Rule]", 
                    RowBox[{"-", "0.08465858886853475`"}]}], ",", 
                    
                    RowBox[{"\<\"lastInt\"\>", "\[Rule]", 
                    "0.0120782716684118`"}], ",", 
                    RowBox[{"\<\"initialized\"\>", "\[Rule]", "True"}]}], 
                    "\[RightAssociation]"}]}
                    },
                    AutoDelete->False,
                    
                    GridBoxItemSize->{
                    "Columns" -> {{Automatic}}, "Rows" -> {{Automatic}}}],
                    "Grid"], {-1, -0.5}], 
                  {GrayLevel[0], LineBox[{{-5, 0}, {5, 0}}]}, 
                  {GrayLevel[0.5], 
                   CircleBox[{-2.0044513660748065`, 0.5}, 0.5], 
                   DiskBox[{-2.0044513660748065`, 0.5}, 0.1]}, 
                  {GrayLevel[0.5], 
                   DiskBox[{-1.659772986059386, 1.9598619161927422`}, 0.1]}, 
                  {GrayLevel[0.5], Thickness[0.01], 
                   LineBox[{{-2.0044513660748065`, 0.5}, {-1.659772986059386, 
                    1.9598619161927422`}}]}, 
                  {RGBColor[0, 1, 0], 
                   DiskBox[{-1.7746657793978593`, 1.4732412774618282`}, 
                    0.05]}, 
                  {RGBColor[0, 0, 1], Arrowheads[0.01], 
                   CircleBox[{-2.0044513660748065`, 0.5}, 0.3333333333333333, 
                    NCache[{
                    Rational[-1, 2] 
                    Pi, -5.273161162721431}, {-1.5707963267948966`, \
-5.273161162721431}]], 
                   ArrowBox[{{-1.8300029069372656`, 
                    0.7840402193627671}, {-1.827171274433708, 
                    0.7822815619533711}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-2.0044513660748065`, 0.5}, {
                    0.46379185787621635`, 0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-2.0044513660748065`, 0.5}, {-3.64929736641462, 
                    0.5}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.005], Arrowheads[0.04], 
                   ArrowBox[{{-2.0044513660748065`, 
                    0.5}, {-1.1810541424635974`, 0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-2.0044513660748065`, 0.5}, 1.6, {
                    1.3389389581609987`, 0.2826451346053038}], 
                   ArrowBox[{{-0.4724767389062148, 
                    0.9615774493101403}, {-0.4679376394353414, 
                    0.9462348796861401}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Dashing[{Small, Small}], 
                   Arrowheads[0.02], 
                   CircleBox[{-2.0044513660748065`, 0.5}, 1.7, {
                    1.3389389581609987`, 1.803080445644445}], 
                   ArrowBox[{{-2.379230166043148, 
                    2.158173950794756}, {-2.395792890406311, 
                    2.154343317251257}}]}, 
                  {RGBColor[0, 0, 1], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-2.0044513660748065`, 0.5}, 1.8, {
                    1.3389389581609987`, 2.024562075925172}], 
                   ArrowBox[{{-2.777269674553856, 
                    2.1256542873808013`}, {-2.79348730589322, 
                    2.117844951061403}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.003], Arrowheads[0.025], 
                   CircleBox[{-2.0044513660748065`, 0.5}, 1.5, {
                    1.3389389581609987`, 1.5610003068331708`}], 
                   ArrowBox[{{-1.9747592755222012`, 
                    1.9997060977933696`}, {-1.9897575711425244`, 
                    1.9999280290702246`}}]}, 
                  {RGBColor[0, 0, 1], ArrowBox[{{0, 2.8}, {1.5, 2.8}}]}},
                 Frame->True,
                 FrameTicks->Automatic,
                 ImageSize->800,
                 PlotRange->{{-3, 3}, {-1, 3}}], 48->
                GraphicsBox[{InsetBox[
                   TagBox[GridBox[{
                    {"\<\"t\"\>", "2.35`"},
                    {"\<\"driveIntent\"\>", "0.5`"},
                    {"\<\"motorTorque\"\>", 
                    RowBox[{"-", "134.9138527654103`"}]},
                    {"\<\"vw\"\>", "0.026230185283051217`"},
                    {"\<\"controllerState\"\>", 
                    RowBox[{"\[LeftAssociation]", 
                    RowBox[{
                    RowBox[{"\<\"lastError\"\>", "\[Rule]", 
                    RowBox[{"-", "0.11515727583598959`"}]}], ",", 
                    
                    RowBox[{"\<\"lastInt\"\>", "\[Rule]", 
                    "0.007045021039686777`"}], ",", 
                    RowBox[{"\<\"initialized\"\>", "\[Rule]", "True"}]}], 
                    "\[RightAssociation]"}]}
                    },
                    AutoDelete->False,
                    
                    GridBoxItemSize->{
                    "Columns" -> {{Automatic}}, "Rows" -> {{Automatic}}}],
                    "Grid"], {-1, -0.5}], 
                  {GrayLevel[0], LineBox[{{-5, 0}, {5, 0}}]}, 
                  {GrayLevel[0.5], CircleBox[{-2.003504525658701, 0.5}, 0.5], 
                   DiskBox[{-2.003504525658701, 0.5}, 0.1]}, 
                  {GrayLevel[0.5], 
                   DiskBox[{-1.621011023582772, 1.9504132931236158`}, 0.1]}, 
                  {GrayLevel[0.5], Thickness[0.01], 
                   LineBox[{{-2.003504525658701, 0.5}, {-1.621011023582772, 
                    1.9504132931236158`}}]}, 
                  {RGBColor[0, 1, 0], 
                   DiskBox[{-1.7485088576080816`, 1.4669421954157438`}, 
                    0.05]}, 
                  {RGBColor[0, 0, 1], Arrowheads[0.01], 
                   CircleBox[{-2.003504525658701, 0.5}, 0.3333333333333333, 
                    NCache[{
                    Rational[-1, 2] 
                    Pi, -5.618211909757205}, {-1.5707963267948966`, \
-5.618211909757205}]], 
                   ArrowBox[{{-1.7432633340801793`, 
                    0.708292182563106}, {-1.741193458920884, 
                    0.705679399498296}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-2.003504525658701, 0.5}, {0.6947725296495051, 
                    0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-2.003504525658701, 0.5}, {-3.816994896423065, 
                    0.5}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.005], Arrowheads[0.04], 
                   ArrowBox[{{-2.003504525658701, 0.5}, {-1.1187178411148586`,
                     0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-2.003504525658701, 0.5}, 1.6, {
                    1.3129531036404622`, 0.14077188422392806`}], 
                   ArrowBox[{{-0.4216558305337419, 
                    0.740322087481498}, {-0.4193317414880626, 
                    0.7244918481661391}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Dashing[{Small, Small}], 
                   Arrowheads[0.02], 
                   CircleBox[{-2.003504525658701, 0.5}, 1.7, {
                    1.3129531036404622`, 1.8083978309629583`}], 
                   ArrowBox[{{-2.3870951355530488`, 
                    2.1561576748609665`}, {-2.4036372569060935`, 
                    2.152239025499792}}]}, 
                  {RGBColor[0, 0, 1], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-2.003504525658701, 0.5}, 1.8, {
                    1.3129531036404622`, 2.062474507892742}], 
                   ArrowBox[{{-2.8373852385255116`, 
                    2.095193705074321}, {-2.8532952160237723`, 
                    2.086775277904485}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.003], Arrowheads[0.025], 
                   CircleBox[{-2.003504525658701, 0.5}, 1.5, {
                    1.3129531036404622`, 1.5235561789977232`}], 
                   ArrowBox[{{-1.917691182181652, 
                    1.997543344975794}, {-1.9326706566735188`, 
                    1.9983265875651377`}}]}, 
                  {RGBColor[0, 0, 1], ArrowBox[{{0, 2.8}, {1.5, 2.8}}]}},
                 Frame->True,
                 FrameTicks->Automatic,
                 ImageSize->800,
                 PlotRange->{{-3, 3}, {-1, 3}}], 49->
                GraphicsBox[{InsetBox[
                   TagBox[GridBox[{
                    {"\<\"t\"\>", "2.4`"},
                    {"\<\"driveIntent\"\>", "0.5`"},
                    {"\<\"motorTorque\"\>", 
                    RowBox[{"-", "143.50174650602585`"}]},
                    {"\<\"vw\"\>", "0.04194469307188092`"},
                    {"\<\"controllerState\"\>", 
                    RowBox[{"\[LeftAssociation]", 
                    RowBox[{
                    RowBox[{"\<\"lastError\"\>", "\[Rule]", 
                    RowBox[{"-", "0.1403528318401673`"}]}], ",", 
                    
                    RowBox[{"\<\"lastInt\"\>", "\[Rule]", 
                    "0.0006231585616262247`"}], ",", 
                    RowBox[{"\<\"initialized\"\>", "\[Rule]", "True"}]}], 
                    "\[RightAssociation]"}]}
                    },
                    AutoDelete->False,
                    
                    GridBoxItemSize->{
                    "Columns" -> {{Automatic}}, "Rows" -> {{Automatic}}}],
                    "Grid"], {-1, -0.5}], 
                  {GrayLevel[0], LineBox[{{-5, 0}, {5, 0}}]}, 
                  {GrayLevel[0.5], CircleBox[{-2.001795934220734, 0.5}, 0.5], 
                   DiskBox[{-2.001795934220734, 0.5}, 0.1]}, 
                  {GrayLevel[0.5], 
                   DiskBox[{-1.5898273791638733`, 1.9423182414586464`}, 0.1]}, 
                  {GrayLevel[0.5], Thickness[0.01], 
                   LineBox[{{-2.001795934220734, 0.5}, {-1.5898273791638733`, 
                    1.9423182414586464`}}]}, 
                  {RGBColor[0, 1, 0], 
                   DiskBox[{-1.7271502308494937`, 1.4615454943057644`}, 
                    0.05]}, 
                  {RGBColor[0, 0, 1], Arrowheads[0.01], 
                   CircleBox[{-2.001795934220734, 0.5}, 0.3333333333333333, 
                    NCache[{
                    Rational[-1, 2] 
                    Pi, -5.875848721975672}, {-1.5707963267948966`, \
-5.875848721975672}]], 
                   ArrowBox[{{-1.697072022344449, 
                    0.6351090250202599}, {-1.6957361906809307`, 
                    0.632055081293606}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-2.001795934220734, 0.5}, {0.8682389958997829, 
                    0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-2.001795934220734, 0.5}, {-3.9441329436213817`,
                     0.5}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.005], Arrowheads[0.04], 
                   ArrowBox[{{-2.001795934220734, 0.5}, {-1.0740980135008646`,
                     0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-2.001795934220734, 0.5}, 1.6, {
                    1.2925751096020814`, 0.030065287586968248`}], 
                   ArrowBox[{{-0.4030799442618276, 
                    0.5640873111443576}, {-0.40251901696491754`, 
                    0.5480972133572469}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Dashing[{Small, Small}], 
                   Arrowheads[0.02], 
                   CircleBox[{-2.001795934220734, 0.5}, 1.7, {
                    1.2925751096020814`, 1.8090827647276506`}], 
                   ArrowBox[{{-2.3865208123598087`, 
                    2.1558945522408344`}, {-2.4030602458175663`, 
                    2.151964573542278}}]}, 
                  {RGBColor[0, 0, 1], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-2.001795934220734, 0.5}, 1.8, {
                    1.2925751096020814`, 2.0898070346355584`}], 
                   ArrowBox[{{-2.8789604301254554`, 
                    2.0718086547427506`}, {-2.8946343968467723`, 
                    2.0629585661992365`}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.003], Arrowheads[0.025], 
                   CircleBox[{-2.001795934220734, 0.5}, 1.5, {
                    1.2925751096020814`, 1.4879516186351114`}], 
                   ArrowBox[{{-1.8627288693531059`, 
                    1.993539537966472}, {-1.8776709691060578`, 
                    1.9948555090828277`}}]}, 
                  {RGBColor[0, 0, 1], ArrowBox[{{0, 2.8}, {1.5, 2.8}}]}},
                 Frame->True,
                 FrameTicks->Automatic,
                 ImageSize->800,
                 PlotRange->{{-3, 3}, {-1, 3}}], 50->
                GraphicsBox[{InsetBox[
                   TagBox[GridBox[{
                    {"\<\"t\"\>", "2.45`"},
                    {"\<\"driveIntent\"\>", "0.5`"},
                    {"\<\"motorTorque\"\>", 
                    RowBox[{"-", "149.5083667783029`"}]},
                    {"\<\"vw\"\>", "0.058378241467570174`"},
                    {"\<\"controllerState\"\>", 
                    RowBox[{"\[LeftAssociation]", 
                    RowBox[{
                    RowBox[{"\<\"lastError\"\>", "\[Rule]", 
                    RowBox[{"-", "0.160552125463986`"}]}], ",", 
                    RowBox[{"\<\"lastInt\"\>", "\[Rule]", 
                    RowBox[{"-", "0.006929635448024599`"}]}], ",", 
                    RowBox[{"\<\"initialized\"\>", "\[Rule]", "True"}]}], 
                    "\[RightAssociation]"}]}
                    },
                    AutoDelete->False,
                    
                    GridBoxItemSize->{
                    "Columns" -> {{Automatic}}, "Rows" -> {{Automatic}}}],
                    "Grid"], {-1, -0.5}], 
                  {GrayLevel[0], LineBox[{{-5, 0}, {5, 0}}]}, 
                  {GrayLevel[0.5], 
                   CircleBox[{-1.9992820050981681`, 0.5}, 0.5], 
                   DiskBox[{-1.9992820050981681`, 0.5}, 0.1]}, 
                  {GrayLevel[0.5], 
                   DiskBox[{-1.5654774946112084`, 1.9359016841974834`}, 0.1]}, 
                  {GrayLevel[0.5], Thickness[0.01], 
                   LineBox[{{-1.9992820050981681`, 
                    0.5}, {-1.5654774946112084`, 1.9359016841974834`}}]}, 
                  {RGBColor[0, 1, 0], 
                   DiskBox[{-1.7100789981068616`, 1.4572677894649888`}, 
                    0.05]}, 
                  {RGBColor[0, 0, 1], Arrowheads[0.01], 
                   CircleBox[{-1.9992820050981681`, 0.5}, 0.3333333333333333, 
                    NCache[{
                    Rational[-1, 2] 
                    Pi, -6.056047330143984}, {-1.5707963267948966`, \
-6.056047330143984}]], 
                   ArrowBox[{{-1.6752772363976232`, 
                    0.5783072217002847}, {-1.6745103773351924`, 
                    0.5750633126853469}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-1.9992820050981681`, 0.5}, {0.99088533046789, 
                    0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-1.9992820050981681`, 
                    0.5}, {-4.0354714837427625`, 0.5}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.005], Arrowheads[0.04], 
                   ArrowBox[{{-1.9992820050981681`, 0.5}, {-1.045304148176704,
                     0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-1.9992820050981681`, 0.5}, 1.6, {
                    1.2774021644673432`, -0.052025648897326526`}], 
                   ArrowBox[{{-0.40069472128899886`, 
                    0.43277875303430674`}, {-0.4014468512532854, 
                    0.41679650766043636`}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Dashing[{Small, Small}], 
                   Arrowheads[0.02], 
                   CircleBox[{-1.9992820050981681`, 0.5}, 1.7, {
                    1.2774021644673432`, 1.8061242371202693`}], 
                   ArrowBox[{{-2.3791061969110725`, 
                    2.157025522831098}, {-2.395657184918513, 
                    2.153144493630968}}]}, 
                  {RGBColor[0, 0, 1], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-1.9992820050981681`, 0.5}, 1.8, {
                    1.2774021644673432`, 2.1080042021245813`}], 
                   ArrowBox[{{-2.90490216106432, 
                    2.055587391665233}, {-2.9204124950872448`, 
                    2.0464535623200857`}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.003], Arrowheads[0.025], 
                   CircleBox[{-1.9992820050981681`, 0.5}, 1.5, {
                    1.2774021644673432`, 1.455286624407031}], 
                   ArrowBox[{{-1.8115113410653871`, 
                    1.9882009870070267`}, {-1.8264024913581638`, 
                    1.9900042529231317`}}]}, 
                  {RGBColor[0, 0, 1], ArrowBox[{{0, 2.8}, {1.5, 2.8}}]}},
                 Frame->True,
                 FrameTicks->Automatic,
                 ImageSize->800,
                 PlotRange->{{-3, 3}, {-1, 3}}], 51->
                GraphicsBox[{InsetBox[
                   TagBox[GridBox[{
                    {"\<\"t\"\>", "2.5`"},
                    {"\<\"driveIntent\"\>", "0.5`"},
                    {"\<\"motorTorque\"\>", 
                    RowBox[{"-", "153.25344672388476`"}]},
                    {"\<\"vw\"\>", "0.07523411134846149`"},
                    {"\<\"controllerState\"\>", 
                    RowBox[{"\[LeftAssociation]", 
                    RowBox[{
                    RowBox[{"\<\"lastError\"\>", "\[Rule]", 
                    RowBox[{"-", "0.1761292508550266`"}]}], ",", 
                    RowBox[{"\<\"lastInt\"\>", "\[Rule]", 
                    RowBox[{"-", "0.01537286816345687`"}]}], ",", 
                    RowBox[{"\<\"initialized\"\>", "\[Rule]", "True"}]}], 
                    "\[RightAssociation]"}]}
                    },
                    AutoDelete->False,
                    
                    GridBoxItemSize->{
                    "Columns" -> {{Automatic}}, "Rows" -> {{Automatic}}}],
                    "Grid"], {-1, -0.5}], 
                  {GrayLevel[0], LineBox[{{-5, 0}, {5, 0}}]}, 
                  {GrayLevel[0.5], CircleBox[{-1.995934439383422, 0.5}, 0.5], 
                   DiskBox[{-1.995934439383422, 0.5}, 0.1]}, 
                  {GrayLevel[0.5], 
                   DiskBox[{-1.5471743991512783`, 1.931298161212696}, 0.1]}, 
                  {GrayLevel[0.5], Thickness[0.01], 
                   LineBox[{{-1.995934439383422, 0.5}, {-1.5471743991512783`, 
                    1.931298161212696}}]}, 
                  {RGBColor[0, 1, 0], 
                   DiskBox[{-1.6967610792286596`, 1.4541987741417972`}, 
                    0.05]}, 
                  {RGBColor[0, 0, 1], Arrowheads[0.01], 
                   CircleBox[{-1.995934439383422, 0.5}, 0.3333333333333333, 
                    NCache[{
                    Rational[-1, 2] 
                    Pi, -6.16839972851144}, {-1.5707963267948966`, \
-6.16839972851144}]], 
                   ArrowBox[{{-1.6651929802692176`, 
                    0.5414873273918448}, {-1.6647946508449663`, 
                    0.5381778935749205}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-1.995934439383422, 0.5}, {1.0691344950942734`, 
                    0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-1.995934439383422, 0.5}, {-4.095568998604673, 
                    0.5}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.005], Arrowheads[0.04], 
                   ArrowBox[{{-1.995934439383422, 0.5}, {-1.0305000641269773`,
                     0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-1.995934439383422, 0.5}, 1.6, {
                    1.2669701088591871`, -0.10829009381036014`}], 
                   ArrowBox[{{-0.40365697316578486`, 
                    0.3429889476006823}, {-0.40530667073126425`, 
                    0.3270742888039557}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Dashing[{Small, Small}], 
                   Arrowheads[0.02], 
                   CircleBox[{-1.995934439383422, 0.5}, 1.7, {
                    1.2669701088591871`, 1.800286646340285}], 
                   ArrowBox[{{-2.3660791775160854`, 
                    2.1592145349027962`}, {-2.3826525392480598`, 
                    2.155430189176543}}]}, 
                  {RGBColor[0, 0, 1], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-1.995934439383422, 0.5}, 1.8, {
                    1.2669701088591871`, 2.1183781462141025`}], 
                   ArrowBox[{{-2.9176431521256068`, 
                    2.0461090029021705`}, {-2.93305789941949, 
                    2.0368147645861647`}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.003], Arrowheads[0.025], 
                   CircleBox[{-1.995934439383422, 0.5}, 1.5, {
                    1.2669701088591871`, 1.4261409466494792`}], 
                   ArrowBox[{{-1.7648750367726542`, 
                    1.9820970118265386`}, {-1.7797073127498422`, 
                    1.9843334631098855`}}]}, 
                  {RGBColor[0, 0, 1], ArrowBox[{{0, 2.8}, {1.5, 2.8}}]}},
                 Frame->True,
                 FrameTicks->Automatic,
                 ImageSize->800,
                 PlotRange->{{-3, 3}, {-1, 3}}], 52->
                GraphicsBox[{InsetBox[
                   TagBox[GridBox[{
                    {"\<\"t\"\>", "2.5500000000000003`"},
                    {"\<\"driveIntent\"\>", "0.5`"},
                    {"\<\"motorTorque\"\>", 
                    RowBox[{"-", "155.0439481823086`"}]},
                    {"\<\"vw\"\>", "0.09224090826959848`"},
                    {"\<\"controllerState\"\>", 
                    RowBox[{"\[LeftAssociation]", 
                    RowBox[{
                    RowBox[{"\<\"lastError\"\>", "\[Rule]", 
                    RowBox[{"-", "0.1874993062455126`"}]}], ",", 
                    RowBox[{"\<\"lastInt\"\>", "\[Rule]", 
                    RowBox[{"-", "0.02448589121758619`"}]}], ",", 
                    RowBox[{"\<\"initialized\"\>", "\[Rule]", "True"}]}], 
                    "\[RightAssociation]"}]}
                    },
                    AutoDelete->False,
                    
                    GridBoxItemSize->{
                    "Columns" -> {{Automatic}}, "Rows" -> {{Automatic}}}],
                    "Grid"], {-1, -0.5}], 
                  {GrayLevel[0], LineBox[{{-5, 0}, {5, 0}}]}, 
                  {GrayLevel[0.5], CircleBox[{-1.991739165371897, 0.5}, 0.5], 
                   DiskBox[{-1.991739165371897, 0.5}, 0.1]}, 
                  {GrayLevel[0.5], 
                   DiskBox[{-1.5341305009915174`, 1.9284937207716402`}, 0.1]}, 
                  {GrayLevel[0.5], Thickness[0.01], 
                   LineBox[{{-1.991739165371897, 0.5}, {-1.5341305009915174`, 
                    1.9284937207716402`}}]}, 
                  {RGBColor[0, 1, 0], 
                   DiskBox[{-1.6866667224516438`, 1.4523291471810935`}, 
                    0.05]}, 
                  {RGBColor[0, 0, 1], Arrowheads[0.01], 
                   CircleBox[{-1.991739165371897, 0.5}, 0.3333333333333333, 
                    NCache[{
                    Rational[-1, 2] 
                    Pi, -6.222114772264155}, {-1.5707963267948966`, \
-6.222114772264155}]], 
                   ArrowBox[{{-1.6592473145757949`, 
                    0.5236702400768087}, {-1.6590272405740492`, 
                    0.5203441934817379}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-1.991739165371897, 0.5}, {1.1091397982742754`, 
                    0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-1.991739165371897, 0.5}, {-4.128579119202655, 
                    0.5}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.005], Arrowheads[0.04], 
                   ArrowBox[{{-1.991739165371897, 0.5}, {-1.0277001555564824`,
                     0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-1.991739165371897, 0.5}, 1.6, {
                    1.2607818292393076`, -0.14159567343127089`}], 
                   ArrowBox[{{-0.4055731210498721, 
                    0.2900541025887372}, {-0.40775185267449254`, 
                    0.2742032037125959}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Dashing[{Small, Small}], 
                   Arrowheads[0.02], 
                   CircleBox[{-1.991739165371897, 0.5}, 1.7, {
                    1.2607818292393076`, 1.792259926348729}], 
                   ArrowBox[{{-2.3485540722144114`, 
                    2.162132101324972}, {-2.36515727561036, 
                    2.1584809058128873`}}]}, 
                  {RGBColor[0, 0, 1], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-1.991739165371897, 0.5}, 1.8, {
                    1.2607818292393076`, 2.1221370969188}], 
                   ArrowBox[{{-2.91925310022278, 
                    2.042633430422611}, {-2.93463280211244, 
                    2.0332813146302255`}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.003], Arrowheads[0.025], 
                   CircleBox[{-1.991739165371897, 0.5}, 1.5, {
                    1.2607818292393076`, 1.4008806576634258`}], 
                   ArrowBox[{{-1.7233192587775807`, 
                    1.975788180513687}, {-1.7380903155027392`, 
                    1.9783985460490867`}}]}, 
                  {RGBColor[0, 0, 1], ArrowBox[{{0, 2.8}, {1.5, 2.8}}]}},
                 Frame->True,
                 FrameTicks->Automatic,
                 ImageSize->800,
                 PlotRange->{{-3, 3}, {-1, 3}}], 53->
                GraphicsBox[{InsetBox[
                   TagBox[GridBox[{
                    {"\<\"t\"\>", "2.6`"},
                    {"\<\"driveIntent\"\>", "0.5`"},
                    {"\<\"motorTorque\"\>", 
                    RowBox[{"-", "155.16695471080013`"}]},
                    {"\<\"vw\"\>", "0.10916047628703306`"},
                    {"\<\"controllerState\"\>", 
                    RowBox[{"\[LeftAssociation]", 
                    RowBox[{
                    RowBox[{"\<\"lastError\"\>", "\[Rule]", 
                    RowBox[{"-", "0.19509763257220297`"}]}], ",", 
                    RowBox[{"\<\"lastInt\"\>", "\[Rule]", 
                    RowBox[{"-", "0.03406940701089108`"}]}], ",", 
                    RowBox[{"\<\"initialized\"\>", "\[Rule]", "True"}]}], 
                    "\[RightAssociation]"}]}
                    },
                    AutoDelete->False,
                    GridBoxItemSize->{
                    "Columns" -> {{Automatic}}, "Rows" -> {{Automatic}}}],
                    "Grid"], {-1, -0.5}], 
                  {GrayLevel[0], LineBox[{{-5, 0}, {5, 0}}]}, 
                  {GrayLevel[0.5], 
                   CircleBox[{-1.9866948607907142`, 0.5}, 0.5], 
                   DiskBox[{-1.9866948607907142`, 0.5}, 0.1]}, 
                  {GrayLevel[0.5], 
                   DiskBox[{-1.5255856135914811`, 1.9273675988151604`}, 0.1]}, 
                  {GrayLevel[0.5], Thickness[0.01], 
                   LineBox[{{-1.9866948607907142`, 
                    0.5}, {-1.5255856135914811`, 1.9273675988151604`}}]}, 
                  {RGBColor[0, 1, 0], 
                   DiskBox[{-1.6792886959912257`, 1.4515783992101068`}, 
                    0.05]}, 
                  {RGBColor[0, 0, 1], Arrowheads[0.01], 
                   CircleBox[{-1.9866948607907142`, 0.5}, 0.3333333333333333, 
                    NCache[{
                    Rational[-1, 2] 
                    Pi, -6.225804968118901}, {-1.5707963267948966`, \
-6.225804968118901}]], 
                   ArrowBox[{{-1.654117926229919, 
                    0.5224431216468576}, {-1.6539101274621069`, 
                    0.5191162855837306}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-1.9866948607907142`, 0.5}, {
                    1.1166442334252884`, 0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-1.9866948607907142`, 0.5}, {-4.138183646231457,
                     0.5}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.005], Arrowheads[0.04], 
                   ArrowBox[{{-1.9866948607907142`, 
                    0.5}, {-1.0348445520154552`, 0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-1.9866948607907142`, 0.5}, 1.6, {
                    1.2583303228992273`, -0.15477498930275946`}], 
                   ArrowBox[{{-0.40343343172635304`, 
                    0.2691683573747379}, {-0.40582087209262063`, 
                    0.25334754844563145`}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Dashing[{Small, Small}], 
                   Arrowheads[0.02], 
                   CircleBox[{-1.9866948607907142`, 0.5}, 1.7, {
                    1.2583303228992273`, 1.7826654636368355`}], 
                   ArrowBox[{{-2.327546324888348, 
                    2.165478993990167}, {-2.3441837948186217`, 
                    2.161987262901732}}]}, 
                  {RGBColor[0, 0, 1], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-1.9866948607907142`, 0.5}, 1.8, {
                    1.2583303228992273`, 2.1203689601814504`}], 
                   ArrowBox[{{-2.9114797603756974`, 
                    2.044270989658095}, {-2.9268759740354136`, 
                    2.034946081886252}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.003], Arrowheads[0.025], 
                   CircleBox[{-1.9866948607907142`, 0.5}, 1.5, {
                    1.2583303228992273`, 1.3797108474583983`}], 
                   ArrowBox[{{-1.6870952774474157`, 
                    1.9697755235615138`}, {-1.7018077675760028`, 
                    1.9726979812981589`}}]}, 
                  {RGBColor[0, 0, 1], ArrowBox[{{0, 2.8}, {1.5, 2.8}}]}},
                 Frame->True,
                 FrameTicks->Automatic,
                 ImageSize->800,
                 PlotRange->{{-3, 3}, {-1, 3}}], 54->
                GraphicsBox[{InsetBox[
                   TagBox[GridBox[{
                    {"\<\"t\"\>", "2.65`"},
                    {"\<\"driveIntent\"\>", "0.5`"},
                    {"\<\"motorTorque\"\>", 
                    RowBox[{"-", "153.88366653669468`"}]},
                    {"\<\"vw\"\>", "0.1257925815170101`"},
                    {"\<\"controllerState\"\>", 
                    RowBox[{"\[LeftAssociation]", 
                    RowBox[{
                    RowBox[{"\<\"lastError\"\>", "\[Rule]", 
                    RowBox[{"-", "0.19936220849611597`"}]}], ",", 
                    RowBox[{"\<\"lastInt\"\>", "\[Rule]", 
                    RowBox[{"-", "0.04394601902039317`"}]}], ",", 
                    RowBox[{"\<\"initialized\"\>", "\[Rule]", "True"}]}], 
                    "\[RightAssociation]"}]}
                    },
                    AutoDelete->False,
                    
                    GridBoxItemSize->{
                    "Columns" -> {{Automatic}}, "Rows" -> {{Automatic}}}],
                    "Grid"], {-1, -0.5}], 
                  {GrayLevel[0], LineBox[{{-5, 0}, {5, 0}}]}, 
                  {GrayLevel[0.5], 
                   CircleBox[{-1.9808111557259034`, 0.5}, 0.5], 
                   DiskBox[{-1.9808111557259034`, 0.5}, 0.1]}, 
                  {GrayLevel[0.5], 
                   DiskBox[{-1.5208258505169683`, 1.9277301982489}, 0.1]}, 
                  {GrayLevel[0.5], Thickness[0.01], 
                   LineBox[{{-1.9808111557259034`, 
                    0.5}, {-1.5208258505169683`, 1.9277301982489}}]}, 
                  {RGBColor[0, 1, 0], 
                   DiskBox[{-1.6741542855866134`, 1.4518201321659332`}, 
                    0.05]}, 
                  {RGBColor[0, 0, 1], Arrowheads[0.01], 
                   CircleBox[{-1.9808111557259034`, 0.5}, 0.3333333333333333, 
                    NCache[{
                    Rational[-1, 2] 
                    Pi, -6.187306322895736}, {-1.5707963267948966`, \
-6.187306322895736}]], 
                   ArrowBox[{{-1.6493444708168954`, 
                    0.5352270905205581}, {-1.6490087789789771`, 
                    0.5319107175757912}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-1.9808111557259034`, 0.5}, {1.09686217500799, 
                    0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-1.9808111557259034`, 0.5}, {-4.127600960703891,
                     0.5}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.005], Arrowheads[0.04], 
                   ArrowBox[{{-1.9808111557259034`, 0.5}, {-1.049927629969997,
                     0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-1.9808111557259034`, 0.5}, 1.6, {
                    1.2591176457947126`, -0.15054325837003835`}], 
                   ArrowBox[{{-0.39658708831741607`, 
                    0.2758703405532654}, {-0.39890755810040623`, 
                    0.2600395703047902}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Dashing[{Small, Small}], 
                   Arrowheads[0.02], 
                   CircleBox[{-1.9808111557259034`, 0.5}, 1.7, {
                    1.2591176457947126`, 1.7720363394010117`}], 
                   ArrowBox[{{-2.3039411157918077`, 
                    2.169007797737269}, {-2.3206147592392385`, 
                    2.1656931022968644`}}]}, 
                  {RGBColor[0, 0, 1], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-1.9808111557259034`, 0.5}, 1.8, {
                    1.2591176457947126`, 2.1140269043319053`}], 
                   ArrowBox[{{-2.8957836699664035`, 
                    2.050104931346397}, {-2.9112387126858623`, 
                    2.040877854097956}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.003], Arrowheads[0.025], 
                   CircleBox[{-1.9808111557259034`, 0.5}, 1.5, {
                    1.2591176457947126`, 1.3626783110174152`}], 
                   ArrowBox[{{-1.6562222345866497`, 
                    1.9644596383218131`}, {-1.670882816205294, 
                    1.967632251063596}}]}, 
                  {RGBColor[0, 0, 1], ArrowBox[{{0, 2.8}, {1.5, 2.8}}]}},
                 Frame->True,
                 FrameTicks->Automatic,
                 ImageSize->800,
                 PlotRange->{{-3, 3}, {-1, 3}}], 55->
                GraphicsBox[{InsetBox[
                   TagBox[GridBox[{
                    {"\<\"t\"\>", "2.7`"},
                    {"\<\"driveIntent\"\>", "0.5`"},
                    {"\<\"motorTorque\"\>", 
                    RowBox[{"-", "151.42643330798757`"}]},
                    {"\<\"vw\"\>", "0.1419759179549393`"},
                    {"\<\"controllerState\"\>", 
                    RowBox[{"\[LeftAssociation]", 
                    RowBox[{
                    RowBox[{"\<\"lastError\"\>", "\[Rule]", 
                    RowBox[{"-", "0.2007193677457365`"}]}], ",", 
                    RowBox[{"\<\"lastInt\"\>", "\[Rule]", 
                    RowBox[{"-", "0.05395998469443248`"}]}], ",", 
                    RowBox[{"\<\"initialized\"\>", "\[Rule]", "True"}]}], 
                    "\[RightAssociation]"}]}
                    },
                    AutoDelete->False,
                    
                    GridBoxItemSize->{
                    "Columns" -> {{Automatic}}, "Rows" -> {{Automatic}}}],
                    "Grid"], {-1, -0.5}], 
                  {GrayLevel[0], LineBox[{{-5, 0}, {5, 0}}]}, 
                  {GrayLevel[0.5], 
                   CircleBox[{-1.9741066965468144`, 0.5}, 0.5], 
                   DiskBox[{-1.9741066965468144`, 0.5}, 0.1]}, 
                  {GrayLevel[0.5], 
                   DiskBox[{-1.5191952659096926`, 1.9293549560118675`}, 0.1]}, 
                  {GrayLevel[0.5], Thickness[0.01], 
                   LineBox[{{-1.9741066965468144`, 
                    0.5}, {-1.5191952659096926`, 1.9293549560118675`}}]}, 
                  {RGBColor[0, 1, 0], 
                   DiskBox[{-1.6708324094554, 1.4529033040079116`}, 0.05]}, 
                  {RGBColor[0, 0, 1], Arrowheads[0.01], 
                   CircleBox[{-1.9741066965468144`, 0.5}, 0.3333333333333333, 
                    NCache[{
                    Rational[-1, 2] 
                    Pi, -6.113589326034524}, {-1.5707963267948966`, \
-6.113589326034524}]], 
                   ArrowBox[{{-1.6461347152760144`, 
                    0.559544022474277}, {-1.6455556834376348`, 
                    0.5562613801469789}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-1.9741066965468144`, 0.5}, {1.054421969612937, 
                    0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-1.9741066965468144`, 0.5}, {-4.099632405406275,
                     0.5}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.005], Arrowheads[0.04], 
                   ArrowBox[{{-1.9741066965468144`, 0.5}, {-1.071103739246524,
                     0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-1.9741066965468144`, 0.5}, 1.6, {
                    1.2626694259929854`, -0.13144219314999273`}], 
                   ArrowBox[{{-0.3858907680559891, 
                    0.30616975342319175`}, {-0.3879084483515478, 
                    0.29029755027118126`}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Dashing[{Small, Small}], 
                   Arrowheads[0.02], 
                   CircleBox[{-1.9741066965468144`, 0.5}, 1.7, {
                    1.2626694259929854`, 1.760805238548397}], 
                   ArrowBox[{{-2.2784718765718335`, 
                    2.1725315653787636`}, {-2.295181695339571, 
                    2.169404338424407}}]}, 
                  {RGBColor[0, 0, 1], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-1.9741066965468144`, 0.5}, 1.8, {
                    1.2626694259929854`, 2.103927388815139}], 
                   ArrowBox[{{-2.8733775048206773`, 
                    2.059266498513475}, {-2.888924946763643, 
                    2.050195977633222}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.003], Arrowheads[0.025], 
                   CircleBox[{-1.9741066965468144`, 0.5}, 1.5, {
                    1.2626694259929854`, 1.3496805085231922`}], 
                   ArrowBox[{{-1.6305109723336715`, 
                    1.9601171111600761`}, {-1.6451290797366835`, 
                    1.963480005889362}}]}, 
                  {RGBColor[0, 0, 1], ArrowBox[{{0, 2.8}, {1.5, 2.8}}]}},
                 Frame->True,
                 FrameTicks->Automatic,
                 ImageSize->800,
                 PlotRange->{{-3, 3}, {-1, 3}}], 56->
                GraphicsBox[{InsetBox[
                   TagBox[GridBox[{
                    {"\<\"t\"\>", "2.75`"},
                    {"\<\"driveIntent\"\>", "0.5`"},
                    {"\<\"motorTorque\"\>", 
                    RowBox[{"-", "147.9986149344869`"}]},
                    {"\<\"vw\"\>", "0.15758614802695806`"},
                    {"\<\"controllerState\"\>", 
                    RowBox[{"\[LeftAssociation]", 
                    RowBox[{
                    RowBox[{"\<\"lastError\"\>", "\[Rule]", 
                    RowBox[{"-", "0.19957318945344515`"}]}], ",", 
                    RowBox[{"\<\"lastInt\"\>", "\[Rule]", 
                    RowBox[{"-", "0.06397634709107798`"}]}], ",", 
                    RowBox[{"\<\"initialized\"\>", "\[Rule]", "True"}]}], 
                    "\[RightAssociation]"}]}
                    },
                    AutoDelete->False,
                    
                    GridBoxItemSize->{
                    "Columns" -> {{Automatic}}, "Rows" -> {{Automatic}}}],
                    "Grid"], {-1, -0.5}], 
                  {GrayLevel[0], LineBox[{{-5, 0}, {5, 0}}]}, 
                  {GrayLevel[0.5], 
                   CircleBox[{-1.9666072392680927`, 0.5}, 0.5], 
                   DiskBox[{-1.9666072392680927`, 0.5}, 0.1]}, 
                  {GrayLevel[0.5], 
                   DiskBox[{-1.5201016226365471`, 1.932003049688262}, 0.1]}, 
                  {GrayLevel[0.5], Thickness[0.01], 
                   LineBox[{{-1.9666072392680927`, 
                    0.5}, {-1.5201016226365471`, 1.932003049688262}}]}, 
                  {RGBColor[0, 1, 0], 
                   DiskBox[{-1.6689368281803958`, 1.4546686997921747`}, 
                    0.05]}, 
                  {RGBColor[0, 0, 1], Arrowheads[0.01], 
                   CircleBox[{-1.9666072392680927`, 0.5}, 0.3333333333333333, 
                    NCache[{
                    Rational[-1, 2] 
                    Pi, -6.010754774829504}, {-1.5707963267948966`, \
-6.010754774829504}]], 
                   ArrowBox[{{-1.646480269628745, 
                    0.5928969021046409}, {-1.6455673223055343`, 
                    0.5896910409560773}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-1.9666072392680927`, 0.5}, {0.9933650594216457,
                     0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-1.9666072392680927`, 0.5}, {-4.056722758931636,
                     0.5}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.005], Arrowheads[0.04], 
                   ArrowBox[{{-1.9666072392680927`, 
                    0.5}, {-1.0967504602418985`, 0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-1.9666072392680927`, 0.5}, 1.6, {
                    1.2685448109032194`, -0.09980653451545107}], 
                   ArrowBox[{{-0.37305507483568867`, 
                    0.3565026159374308}, {-0.3745697017044469, 
                    0.340574534693219}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Dashing[{Small, Small}], 
                   Arrowheads[0.02], 
                   CircleBox[{-1.9666072392680927`, 0.5}, 1.7, {
                    1.2685448109032194`, 1.749305365394312}], 
                   ArrowBox[{{-2.2517188168983795`, 
                    2.1759210567031992`}, {-2.2684634926865463`, 
                    2.1729861930907184`}}]}, 
                  {RGBColor[0, 0, 1], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-1.9666072392680927`, 0.5}, 1.8, {
                    1.2685448109032194`, 2.0907593383170355`}], 
                   ArrowBox[{{-2.8452681763726124`, 
                    2.0709726151676255`}, {-2.86093370801608, 
                    2.062107604263149}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.003], Arrowheads[0.025], 
                   CircleBox[{-1.9666072392680927`, 0.5}, 1.5, {
                    1.2685448109032194`, 1.3404894371097886`}], 
                   ArrowBox[{{-1.6096061760169087`, 
                    1.9568974709421125`}, {-1.6241927578157114`, 
                    1.9603945778082363`}}]}, 
                  {RGBColor[0, 0, 1], ArrowBox[{{0, 2.8}, {1.5, 2.8}}]}},
                 Frame->True,
                 FrameTicks->Automatic,
                 ImageSize->800,
                 PlotRange->{{-3, 3}, {-1, 3}}], 57->
                GraphicsBox[{InsetBox[
                   TagBox[GridBox[{
                    {"\<\"t\"\>", "2.8000000000000003`"},
                    {"\<\"driveIntent\"\>", "0.5`"},
                    {"\<\"motorTorque\"\>", 
                    RowBox[{"-", "143.776422439822`"}]},
                    {"\<\"vw\"\>", "0.17253203425441233`"},
                    {"\<\"controllerState\"\>", 
                    RowBox[{"\[LeftAssociation]", 
                    RowBox[{
                    RowBox[{"\<\"lastError\"\>", "\[Rule]", 
                    RowBox[{"-", "0.19629852957592325`"}]}], ",", 
                    RowBox[{"\<\"lastInt\"\>", "\[Rule]", 
                    RowBox[{"-", "0.07387962990969829`"}]}], ",", 
                    RowBox[{"\<\"initialized\"\>", "\[Rule]", "True"}]}], 
                    "\[RightAssociation]"}]}
                    },
                    AutoDelete->False,
                    
                    GridBoxItemSize->{
                    "Columns" -> {{Automatic}}, "Rows" -> {{Automatic}}}],
                    "Grid"], {-1, -0.5}], 
                  {GrayLevel[0], LineBox[{{-5, 0}, {5, 0}}]}, 
                  {GrayLevel[0.5], CircleBox[{-1.958343894027679, 0.5}, 0.5], 
                   DiskBox[{-1.958343894027679, 0.5}, 0.1]}, 
                  {GrayLevel[0.5], 
                   DiskBox[{-1.5230176309809886`, 1.9354410627754117`}, 0.1]}, 
                  {GrayLevel[0.5], Thickness[0.01], 
                   LineBox[{{-1.958343894027679, 0.5}, {-1.5230176309809886`, 
                    1.9354410627754117`}}]}, 
                  {RGBColor[0, 1, 0], 
                   DiskBox[{-1.6681263853298853`, 1.4569607085169411`}, 
                    0.05]}, 
                  {RGBColor[0, 0, 1], Arrowheads[0.01], 
                   CircleBox[{-1.958343894027679, 0.5}, 0.3333333333333333, 
                    NCache[{
                    Rational[-1, 2] 
                    Pi, -5.884088999989556}, {-1.5707963267948966`, \
-5.884088999989556}]], 
                   ArrowBox[{{-1.652517004473338, 
                    0.6325934566131677}, {-1.6512063832230555`, 
                    0.6295286090709341}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-1.958343894027679, 0.5}, {0.9171845547687612, 
                    0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-1.958343894027679, 0.5}, {-4.001020688900853, 
                    0.5}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.005], Arrowheads[0.04], 
                   ArrowBox[{{-1.958343894027679, 0.5}, {-1.1254922401044134`,
                     0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-1.958343894027679, 0.5}, 1.6, {
                    1.2763422057293212`, -0.05774913724156239}], 
                   ArrowBox[{{-0.36016753158591386`, 
                    0.4236304083262966}, {-0.3610111229266668, 
                    0.40765272951103737`}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Dashing[{Small, Small}], 
                   Arrowheads[0.02], 
                   CircleBox[{-1.958343894027679, 0.5}, 1.7, {
                    1.2763422057293212`, 1.7377811457483383`}], 
                   ArrowBox[{{-2.2241232845009953`, 
                    2.1790953860932474`}, {-2.2409006696553133`, 
                    2.176353682415174}}]}, 
                  {RGBColor[0, 0, 1], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-1.958343894027679, 0.5}, 1.8, {
                    1.2763422057293212`, 2.0751001081727765`}], 
                   ArrowBox[{{-2.812297888087915, 
                    2.084538600359288}, {-2.8281003126591706`, 
                    2.075919976473844}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.003], Arrowheads[0.025], 
                   CircleBox[{-1.958343894027679, 0.5}, 1.5, {
                    1.2763422057293212`, 1.3347850548474116`}], 
                   ArrowBox[{{-1.5930379841441848`, 
                    1.9548373078128676`}, {-1.607604379893925, 
                    1.9584175647684783`}}]}, 
                  {RGBColor[0, 0, 1], ArrowBox[{{0, 2.8}, {1.5, 2.8}}]}},
                 Frame->True,
                 FrameTicks->Automatic,
                 ImageSize->800,
                 PlotRange->{{-3, 3}, {-1, 3}}], 58->
                GraphicsBox[{InsetBox[
                   TagBox[GridBox[{
                    {"\<\"t\"\>", "2.85`"},
                    {"\<\"driveIntent\"\>", "0.5`"},
                    {"\<\"motorTorque\"\>", 
                    RowBox[{"-", "138.9118488142305`"}]},
                    {"\<\"vw\"\>", "0.18675065215960285`"},
                    {"\<\"controllerState\"\>", 
                    RowBox[{"\[LeftAssociation]", 
                    RowBox[{
                    RowBox[{"\<\"lastError\"\>", "\[Rule]", 
                    RowBox[{"-", "0.19123727706835708`"}]}], ",", 
                    RowBox[{"\<\"lastInt\"\>", "\[Rule]", 
                    RowBox[{"-", "0.08357226866804661`"}]}], ",", 
                    RowBox[{"\<\"initialized\"\>", "\[Rule]", "True"}]}], 
                    "\[RightAssociation]"}]}
                    },
                    AutoDelete->False,
                    
                    GridBoxItemSize->{
                    "Columns" -> {{Automatic}}, "Rows" -> {{Automatic}}}],
                    "Grid"], {-1, -0.5}], 
                  {GrayLevel[0], LineBox[{{-5, 0}, {5, 0}}]}, 
                  {GrayLevel[0.5], 
                   CircleBox[{-1.9493515901006735`, 0.5}, 0.5], 
                   DiskBox[{-1.9493515901006735`, 0.5}, 0.1]}, 
                  {GrayLevel[0.5], 
                   DiskBox[{-1.5274789453833293`, 1.9394524902334198`}, 0.1]}, 
                  {GrayLevel[0.5], Thickness[0.01], 
                   LineBox[{{-1.9493515901006735`, 
                    0.5}, {-1.5274789453833293`, 1.9394524902334198`}}]}, 
                  {RGBColor[0, 1, 0], 
                   DiskBox[{-1.6681031602891105`, 1.4596349934889465`}, 
                    0.05]}, 
                  {RGBColor[0, 0, 1], Arrowheads[0.01], 
                   CircleBox[{-1.9493515901006735`, 0.5}, 0.3333333333333333, 
                    NCache[{
                    Rational[-1, 2] 
                    Pi, -5.738151791221811}, {-1.5707963267948966`, \
-5.738151791221811}]], 
                   ArrowBox[{{-1.6660573305063133`, 
                    0.6756572617115342}, {-1.6643149517602027`, 
                    0.6728155835411691}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-1.9493515901006735`, 0.5}, {0.8288853861839363,
                     0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-1.9493515901006735`, 0.5}, {-3.934432083644107,
                     0.5}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.005], Arrowheads[0.04], 
                   ArrowBox[{{-1.9493515901006735`, 0.5}, {-1.156195107359497,
                     0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-1.9493515901006735`, 0.5}, 1.6, {
                    1.2857015227776194`, -0.007160155759050824}], 
                   ArrowBox[{{-0.34935804186858754`, 
                    0.5045437446781784}, {-0.34939260418984275`, 
                    0.4885438486747749}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Dashing[{Small, Small}], 
                   Arrowheads[0.02], 
                   CircleBox[{-1.9493515901006735`, 0.5}, 1.7, {
                    1.2857015227776194`, 1.726403913344877}], 
                   ArrowBox[{{-2.1960107330210183`, 
                    2.1820104836813004`}, {-2.212818224669781, 
                    2.1794598335383997`}}]}, 
                  {RGBColor[0, 0, 1], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-1.9493515901006735`, 0.5}, 1.8, {
                    1.2857015227776194`, 2.057434016190011}], 
                   ArrowBox[{{-2.7751811833835855`, 
                    2.09937659194393}, {-2.7911333916060226`, 
                    2.091038465485486}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.003], Arrowheads[0.025], 
                   CircleBox[{-1.9493515901006735`, 0.5}, 1.5, {
                    1.2857015227776194`, 1.3321907670762663`}], 
                   ArrowBox[{{-1.5802726471279864`, 
                    1.95388470445705}, {-1.5948297056530172`, 
                    1.9575027387444879`}}]}, 
                  {RGBColor[0, 0, 1], ArrowBox[{{0, 2.8}, {1.5, 2.8}}]}},
                 Frame->True,
                 FrameTicks->Automatic,
                 ImageSize->800,
                 PlotRange->{{-3, 3}, {-1, 3}}], 59->
                GraphicsBox[{InsetBox[
                   TagBox[GridBox[{
                    {"\<\"t\"\>", "2.9`"},
                    {"\<\"driveIntent\"\>", "0.5`"},
                    {"\<\"motorTorque\"\>", 
                    RowBox[{"-", "133.53598434244986`"}]},
                    {"\<\"vw\"\>", "0.20020243532477047`"},
                    {"\<\"controllerState\"\>", 
                    RowBox[{"\[LeftAssociation]", 
                    RowBox[{
                    RowBox[{"\<\"lastError\"\>", "\[Rule]", 
                    RowBox[{"-", "0.1846971911024795`"}]}], ",", 
                    RowBox[{"\<\"lastInt\"\>", "\[Rule]", 
                    RowBox[{"-", "0.09297292342986463`"}]}], ",", 
                    RowBox[{"\<\"initialized\"\>", "\[Rule]", "True"}]}], 
                    "\[RightAssociation]"}]}
                    },
                    AutoDelete->False,
                    
                    GridBoxItemSize->{
                    "Columns" -> {{Automatic}}, "Rows" -> {{Automatic}}}],
                    "Grid"], {-1, -0.5}], 
                  {GrayLevel[0], LineBox[{{-5, 0}, {5, 0}}]}, 
                  {GrayLevel[0.5], 
                   CircleBox[{-1.9396677871000039`, 0.5}, 0.5], 
                   DiskBox[{-1.9396677871000039`, 0.5}, 0.1]}, 
                  {GrayLevel[0.5], 
                   DiskBox[{-1.5330800608956356`, 1.94384432017443}, 0.1]}, 
                  {GrayLevel[0.5], Thickness[0.01], 
                   LineBox[{{-1.9396677871000039`, 
                    0.5}, {-1.5330800608956356`, 1.94384432017443}}]}, 
                  {RGBColor[0, 1, 0], 
                   DiskBox[{-1.6686093029637585`, 1.4625628801162867`}, 
                    0.05]}, 
                  {RGBColor[0, 0, 1], Arrowheads[0.01], 
                   CircleBox[{-1.9396677871000039`, 0.5}, 0.3333333333333333, 
                    NCache[{
                    Rational[-1, 2] 
                    Pi, -5.5768758570683925`}, {-1.5707963267948966`, \
-5.5768758570683925`}]], 
                   ArrowBox[{{-1.6882564218353857`, 
                    0.7188685370876589}, {-1.6860803434059253`, 
                    0.7163435220010379}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-1.9396677871000039`, 0.5}, {0.7310518997489934,
                     0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-1.9396677871000039`, 0.5}, {-3.858663718390938,
                     0.5}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.005], Arrowheads[0.04], 
                   ArrowBox[{{-1.9396677871000039`, 
                    0.5}, {-1.1879440315419405`, 0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-1.9396677871000039`, 0.5}, 1.6, {
                    1.2963038133762748`, 0.05028396175687888}], 
                   ArrowBox[{{-0.3425742315700626, 
                    0.5963959277393529}, {-0.3416901423708998, 
                    0.5804204386087597}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Dashing[{Small, Small}], 
                   Arrowheads[0.02], 
                   CircleBox[{-1.9396677871000039`, 0.5}, 1.7, {
                    1.2963038133762748`, 1.7152886327550658`}], 
                   ArrowBox[{{-2.167616059371959, 
                    2.184648208133743}, {-2.1844508633613633`, 
                    2.18228453169374}}]}, 
                  {RGBColor[0, 0, 1], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-1.9396677871000039`, 0.5}, 1.8, {
                    1.2963038133762748`, 2.038170393056552}], 
                   ArrowBox[{{-2.734536275303836, 
                    2.1149873332198474`}, {-2.750646136379608, 
                    2.1069580321214767`}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.003], Arrowheads[0.025], 
                   CircleBox[{-1.9396677871000039`, 0.5}, 1.5, {
                    1.2963038133762748`, 1.3323053292278813`}], 
                   ArrowBox[{{-1.570755406708881, 
                    1.9539269773940353`}, {-1.5853128796281766`, 
                    1.9575433439697885`}}]}, 
                  {RGBColor[0, 0, 1], ArrowBox[{{0, 2.8}, {1.5, 2.8}}]}},
                 Frame->True,
                 FrameTicks->Automatic,
                 ImageSize->800,
                 PlotRange->{{-3, 3}, {-1, 3}}], 60->
                GraphicsBox[{InsetBox[
                   TagBox[GridBox[{
                    {"\<\"t\"\>", "2.95`"},
                    {"\<\"driveIntent\"\>", "0.5`"},
                    {"\<\"motorTorque\"\>", 
                    RowBox[{"-", "127.7622480491768`"}]},
                    {"\<\"vw\"\>", "0.21286653364922015`"},
                    {"\<\"controllerState\"\>", 
                    RowBox[{"\[LeftAssociation]", 
                    RowBox[{
                    RowBox[{"\<\"lastError\"\>", "\[Rule]", 
                    RowBox[{"-", "0.1769526114792821`"}]}], ",", 
                    RowBox[{"\<\"lastInt\"\>", "\[Rule]", 
                    RowBox[{"-", "0.10201478406583235`"}]}], ",", 
                    RowBox[{"\<\"initialized\"\>", "\[Rule]", "True"}]}], 
                    "\[RightAssociation]"}]}
                    },
                    AutoDelete->False,
                    
                    GridBoxItemSize->{
                    "Columns" -> {{Automatic}}, "Rows" -> {{Automatic}}}],
                    "Grid"], {-1, -0.5}], 
                  {GrayLevel[0], LineBox[{{-5, 0}, {5, 0}}]}, 
                  {GrayLevel[0.5], 
                   CircleBox[{-1.9293314274189264`, 0.5}, 0.5], 
                   DiskBox[{-1.9293314274189264`, 0.5}, 0.1]}, 
                  {GrayLevel[0.5], 
                   DiskBox[{-1.5394690437401375`, 1.9484499721399744`}, 0.1]}, 
                  {GrayLevel[0.5], Thickness[0.01], 
                   LineBox[{{-1.9293314274189264`, 
                    0.5}, {-1.5394690437401375`, 1.9484499721399744`}}]}, 
                  {RGBColor[0, 1, 0], 
                   DiskBox[{-1.669423171633067, 1.465633314759983}, 0.05]}, 
                  {RGBColor[0, 0, 1], Arrowheads[0.01], 
                   CircleBox[{-1.9293314274189264`, 0.5}, 0.3333333333333333, 
                    NCache[{
                    Rational[-1, 2] 
                    Pi, -5.4036637682702}, {-1.5707963267948966`, \
-5.4036637682702}]], 
                   ArrowBox[{{-1.7194035097151794`, 
                    0.7589235031426849}, {-1.7168248141458697`, 
                    0.7568113128861862}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-1.9293314274189264`, 0.5}, {0.6259135335646095,
                     0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-1.9293314274189264`, 
                    0.5}, {-3.7752572892455727`, 0.5}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.005], Arrowheads[0.04], 
                   ArrowBox[{{-1.9293314274189264`, 
                    0.5}, {-1.2200123282620368`, 0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-1.9293314274189264`, 0.5}, 1.6, {
                    1.3078691344727174`, 0.11310540169034367`}], 
                   ArrowBox[{{-0.3414400756116489, 
                    0.696471511562509}, {-0.3395547871470832, 
                    0.6805830391979564}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Dashing[{Small, Small}], 
                   Arrowheads[0.02], 
                   CircleBox[{-1.9293314274189264`, 0.5}, 1.7, {
                    1.3078691344727174`, 1.704509164074459}], 
                   ArrowBox[{{-2.139107195455143, 
                    2.187007447270111}, {-2.1559665000603463`, 
                    2.184825374882756}}]}, 
                  {RGBColor[0, 0, 1], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-1.9293314274189264`, 0.5}, 1.8, {
                    1.3078691344727174`, 2.0176594014125886`}], 
                   ArrowBox[{{-2.6909100514613504`, 
                    2.130950029707118}, {-2.7071812013208962`, 
                    2.123252823573905}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.003], Arrowheads[0.025], 
                   CircleBox[{-1.9293314274189264`, 0.5}, 1.5, {
                    1.3078691344727174`, 1.3347281733813379`}], 
                   ArrowBox[{{-1.5639427648474835`, 
                    1.9548165263235953`}, {-1.5785089569233945`, 
                    1.958397611831359}}]}, 
                  {RGBColor[0, 0, 1], ArrowBox[{{0, 2.8}, {1.5, 2.8}}]}},
                 Frame->True,
                 FrameTicks->Automatic,
                 ImageSize->800,
                 PlotRange->{{-3, 3}, {-1, 3}}], 61->
                GraphicsBox[{InsetBox[
                   TagBox[GridBox[{
                    {"\<\"t\"\>", "3.`"},
                    {"\<\"driveIntent\"\>", "0.5`"},
                    {"\<\"motorTorque\"\>", 
                    RowBox[{"-", "121.68927600071132`"}]},
                    {"\<\"vw\"\>", "0.22473673746448003`"},
                    {"\<\"controllerState\"\>", 
                    RowBox[{"\[LeftAssociation]", 
                    RowBox[{
                    RowBox[{"\<\"lastError\"\>", "\[Rule]", 
                    RowBox[{"-", "0.1682463871768773`"}]}], ",", 
                    RowBox[{"\<\"lastInt\"\>", "\[Rule]", 
                    RowBox[{"-", "0.1106439446143106`"}]}], ",", 
                    RowBox[{"\<\"initialized\"\>", "\[Rule]", "True"}]}], 
                    "\[RightAssociation]"}]}
                    },
                    AutoDelete->False,
                    
                    GridBoxItemSize->{
                    "Columns" -> {{Automatic}}, "Rows" -> {{Automatic}}}],
                    "Grid"], {-1, -0.5}], 
                  {GrayLevel[0], LineBox[{{-5, 0}, {5, 0}}]}, 
                  {GrayLevel[0.5], 
                   CircleBox[{-1.9183821069926057`, 0.5}, 0.5], 
                   DiskBox[{-1.9183821069926057`, 0.5}, 0.1]}, 
                  {GrayLevel[0.5], 
                   DiskBox[{-1.5463418045799404`, 1.9531297304028614`}, 0.1]}, 
                  {GrayLevel[0.5], Thickness[0.01], 
                   LineBox[{{-1.9183821069926057`, 
                    0.5}, {-1.5463418045799404`, 1.9531297304028614`}}]}, 
                  {RGBColor[0, 1, 0], 
                   DiskBox[{-1.6703552387174954`, 1.4687531536019076`}, 
                    0.05]}, 
                  {RGBColor[0, 0, 1], Arrowheads[0.01], 
                   CircleBox[{-1.9183821069926057`, 0.5}, 0.3333333333333333, 
                    NCache[{
                    Rational[-1, 2] 
                    Pi, -5.221474606816236}, {-1.5707963267948966`, \
-5.221474606816236}]], 
                   ArrowBox[{{-1.7588411403264432`, 
                    0.7926735230018894}, {-1.755922430856959, 
                    0.7910635063710528}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-1.9183821069926057`, 0.5}, {0.5154034130216205,
                     0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-1.9183821069926057`, 
                    0.5}, {-3.6856150471100255`, 0.5}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.005], Arrowheads[0.04], 
                   ArrowBox[{{-1.9183821069926057`, 
                    0.5}, {-1.2518295270957989`, 0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-1.9183821069926057`, 0.5}, 1.6, {
                    1.3201533789022721`, 0.1800068065100331}], 
                   ArrowBox[{{-0.3471773875933106, 
                    0.8021849263900874}, {-0.34423414827461585`, 
                    0.7864580319418288}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Dashing[{Small, Small}], 
                   Arrowheads[0.02], 
                   CircleBox[{-1.9183821069926057`, 0.5}, 1.7, {
                    1.3201533789022721`, 1.6941108779714906`}], 
                   ArrowBox[{{-2.110604864219375, 
                    2.1890975139417317`}, {-2.127485946786179, 
                    2.1870908642345213`}}]}, 
                  {RGBColor[0, 0, 1], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-1.9183821069926057`, 0.5}, 1.8, {
                    1.3201533789022721`, 1.9962049122395573`}], 
                   ArrowBox[{{-2.6447969468652905`, 
                    2.146912711837741}, {-2.6612294790602666`, 
                    2.139566339558168}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.003], Arrowheads[0.025], 
                   CircleBox[{-1.9183821069926057`, 0.5}, 1.5, {
                    1.3201533789022721`, 1.3390773739670736`}], 
                   ArrowBox[{{-1.5593241691197843`, 
                    1.9563919105963603`}, {-1.573905798242263, 
                    1.9599096111436962`}}]}, 
                  {RGBColor[0, 0, 1], ArrowBox[{{0, 2.8}, {1.5, 2.8}}]}},
                 Frame->True,
                 FrameTicks->Automatic,
                 ImageSize->800,
                 PlotRange->{{-3, 3}, {-1, 3}}], 62->
                GraphicsBox[{InsetBox[
                   TagBox[GridBox[{
                    {"\<\"t\"\>", "3.0500000000000003`"},
                    {"\<\"driveIntent\"\>", "0.5`"},
                    {"\<\"motorTorque\"\>", 
                    RowBox[{"-", "115.4033598277029`"}]},
                    {"\<\"vw\"\>", "0.2358180570808383`"},
                    {"\<\"controllerState\"\>", 
                    RowBox[{"\[LeftAssociation]", 
                    RowBox[{
                    RowBox[{"\<\"lastError\"\>", "\[Rule]", 
                    RowBox[{"-", "0.1587924834513234`"}]}], ",", 
                    RowBox[{"\<\"lastInt\"\>", "\[Rule]", 
                    RowBox[{"-", "0.11881789334096671`"}]}], ",", 
                    RowBox[{"\<\"initialized\"\>", "\[Rule]", "True"}]}], 
                    "\[RightAssociation]"}]}
                    },
                    AutoDelete->False,
                    
                    GridBoxItemSize->{
                    "Columns" -> {{Automatic}}, "Rows" -> {{Automatic}}}],
                    "Grid"], {-1, -0.5}], 
                  {GrayLevel[0], LineBox[{{-5, 0}, {5, 0}}]}, 
                  {GrayLevel[0.5], CircleBox[{-1.906859433260333, 0.5}, 0.5], 
                   DiskBox[{-1.906859433260333, 0.5}, 0.1]}, 
                  {GrayLevel[0.5], 
                   DiskBox[{-1.5534364113264854`, 1.9577695865832663`}, 0.1]}, 
                  {GrayLevel[0.5], Thickness[0.01], 
                   LineBox[{{-1.906859433260333, 0.5}, {-1.5534364113264854`, 
                    1.9577695865832663`}}]}, 
                  {RGBColor[0, 1, 0], 
                   DiskBox[{-1.6712440853044346`, 1.4718463910555108`}, 
                    0.05]}, 
                  {RGBColor[0, 0, 1], Arrowheads[0.01], 
                   CircleBox[{-1.906859433260333, 0.5}, 0.3333333333333333, 
                    NCache[{
                    Rational[-1, 2] 
                    Pi, -5.032897121625984}, {-1.5707963267948966`, \
-5.032897121625984}]], 
                   ArrowBox[{{-1.805011927593423, 
                    0.8173928113561173}, {-1.8018431447112464`, 
                    0.8163584837656264}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-1.906859433260333, 0.5}, {0.4012077632937252, 
                    0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-1.906859433260333, 0.5}, {-3.591018633687683, 
                    0.5}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.005], Arrowheads[0.04], 
                   ArrowBox[{{-1.906859433260333, 0.5}, {-1.282951437133625, 
                    0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-1.906859433260333, 0.5}, 1.6, {
                    1.332944633157469, 0.24985217126760406`}], 
                   ArrowBox[{{-0.36057467891303197`, 
                    0.9111003022050785}, {-0.3565410580007897, 
                    0.8956171575306237}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Dashing[{Small, Small}], 
                   Arrowheads[0.02], 
                   CircleBox[{-1.906859433260333, 0.5}, 1.7, {
                    1.332944633157469, 1.684120342782223}], 
                   ArrowBox[{{-2.08219789017906, 
                    2.1909335958355545`}, {-2.0990981774667703`, 
                    2.189095694514059}}]}, 
                  {RGBColor[0, 0, 1], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-1.906859433260333, 0.5}, 1.8, {
                    1.332944633157469, 1.9740744099780407`}], 
                   ArrowBox[{{-2.5966523657183966`, 
                    2.1625840460953865`}, {-2.613243439724185, 
                    2.155603103226156}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.003], Arrowheads[0.025], 
                   CircleBox[{-1.906859433260333, 0.5}, 1.5, {
                    1.332944633157469, 1.3450008769010202`}], 
                   ArrowBox[{{-1.5564346859665585`, 
                    1.958493228124181}, {-1.571036896258166, 
                    1.9619244931395836`}}]}, 
                  {RGBColor[0, 0, 1], ArrowBox[{{0, 2.8}, {1.5, 2.8}}]}},
                 Frame->True,
                 FrameTicks->Automatic,
                 ImageSize->800,
                 PlotRange->{{-3, 3}, {-1, 3}}], 63->
                GraphicsBox[{InsetBox[
                   TagBox[GridBox[{
                    {"\<\"t\"\>", "3.1`"},
                    {"\<\"driveIntent\"\>", "0.5`"},
                    {"\<\"motorTorque\"\>", 
                    RowBox[{"-", "108.98042496449818`"}]},
                    {"\<\"vw\"\>", "0.24612394621784967`"},
                    {"\<\"controllerState\"\>", 
                    RowBox[{"\[LeftAssociation]", 
                    RowBox[{
                    RowBox[{"\<\"lastError\"\>", "\[Rule]", 
                    RowBox[{"-", "0.1487788627168537`"}]}], ",", 
                    RowBox[{"\<\"lastInt\"\>", "\[Rule]", 
                    RowBox[{"-", "0.12650414154365947`"}]}], ",", 
                    RowBox[{"\<\"initialized\"\>", "\[Rule]", "True"}]}], 
                    "\[RightAssociation]"}]}
                    },
                    AutoDelete->False,
                    
                    GridBoxItemSize->{
                    "Columns" -> {{Automatic}}, "Rows" -> {{Automatic}}}],
                    "Grid"], {-1, -0.5}], 
                  {GrayLevel[0], LineBox[{{-5, 0}, {5, 0}}]}, 
                  {GrayLevel[0.5], 
                   CircleBox[{-1.8948025375417705`, 0.5}, 0.5], 
                   DiskBox[{-1.8948025375417705`, 0.5}, 0.1]}, 
                  {GrayLevel[0.5], 
                   DiskBox[{-1.5605277588566542`, 1.962279170451052}, 0.1]}, 
                  {GrayLevel[0.5], Thickness[0.01], 
                   LineBox[{{-1.8948025375417705`, 
                    0.5}, {-1.5605277588566542`, 1.962279170451052}}]}, 
                  {RGBColor[0, 1, 0], 
                   DiskBox[{-1.6719526850850264`, 1.4748527803007012`}, 
                    0.05]}, 
                  {RGBColor[0, 0, 1], Arrowheads[0.01], 
                   CircleBox[{-1.8948025375417705`, 0.5}, 0.3333333333333333, 
                    NCache[{
                    Rational[-1, 2] 
                    Pi, -4.840209075729842}, {-1.5707963267948966`, \
-4.840209075729842}]], 
                   ArrowBox[{{-1.8556199720747177`, 
                    0.8310224126468346}, {-1.852311762230323, 
                    0.8306140425398522}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-1.8948025375417705`, 0.5}, {
                    0.28480596174819306`, 0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-1.8948025375417705`, 0.5}, {-3.492642668149955,
                     0.5}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.005], Arrowheads[0.04], 
                   ArrowBox[{{-1.8948025375417705`, 
                    0.5}, {-1.3130341688599914`, 0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-1.8948025375417705`, 0.5}, 1.6, {
                    1.3460594612888226`, 0.32164828679119895`}], 
                   ArrowBox[{{-0.3819914290621138, 
                    1.0209630985593439`}, {-0.37685752482835744`, 
                    1.0058091916706107`}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Dashing[{Small, Small}], 
                   Arrowheads[0.02], 
                   CircleBox[{-1.8948025375417705`, 0.5}, 1.7, {
                    1.3460594612888226`, 1.6745523288999753`}], 
                   ArrowBox[{{-2.0539543394107964`, 
                    2.1925338117632514`}, {-2.0708714379170905`, 
                    2.190857694284361}}]}, 
                  {RGBColor[0, 0, 1], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-1.8948025375417705`, 0.5}, 1.8, {
                    1.3460594612888226`, 1.9515062666471459`}], 
                   ArrowBox[{{-2.5469015643534956`, 
                    2.177726693842356}, {-2.5636459469925676`, 
                    2.1711219266212245`}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.003], Arrowheads[0.025], 
                   CircleBox[{-1.8948025375417705`, 0.5}, 1.5, {
                    1.3460594612888226`, 1.3521822915473185`}], 
                   ArrowBox[{{-1.554860780988258, 
                    1.9609721428389086`}, {-1.5694872558686932`, 
                    1.9642984557493621`}}]}, 
                  {RGBColor[0, 0, 1], ArrowBox[{{0, 2.8}, {1.5, 2.8}}]}},
                 Frame->True,
                 FrameTicks->Automatic,
                 ImageSize->800,
                 PlotRange->{{-3, 3}, {-1, 3}}], 64->
                GraphicsBox[{InsetBox[
                   TagBox[GridBox[{
                    {"\<\"t\"\>", "3.15`"},
                    {"\<\"driveIntent\"\>", "0.5`"},
                    {"\<\"motorTorque\"\>", 
                    RowBox[{"-", "102.48758988601007`"}]},
                    {"\<\"vw\"\>", "0.2556741042575181`"},
                    {"\<\"controllerState\"\>", 
                    RowBox[{"\[LeftAssociation]", 
                    RowBox[{
                    RowBox[{"\<\"lastError\"\>", "\[Rule]", 
                    RowBox[{"-", "0.13837036109779688`"}]}], ",", 
                    RowBox[{"\<\"lastInt\"\>", "\[Rule]", 
                    RowBox[{"-", "0.13367899721139373`"}]}], ",", 
                    RowBox[{"\<\"initialized\"\>", "\[Rule]", "True"}]}], 
                    "\[RightAssociation]"}]}
                    },
                    AutoDelete->False,
                    
                    GridBoxItemSize->{
                    "Columns" -> {{Automatic}}, "Rows" -> {{Automatic}}}],
                    "Grid"], {-1, -0.5}], 
                  {GrayLevel[0], LineBox[{{-5, 0}, {5, 0}}]}, 
                  {GrayLevel[0.5], 
                   CircleBox[{-1.8822497111467744`, 0.5}, 0.5], 
                   DiskBox[{-1.8822497111467744`, 0.5}, 0.1]}, 
                  {GrayLevel[0.5], 
                   DiskBox[{-1.5674227725728556`, 1.966589239953769}, 0.1]}, 
                  {GrayLevel[0.5], Thickness[0.01], 
                   LineBox[{{-1.8822497111467744`, 
                    0.5}, {-1.5674227725728556`, 1.966589239953769}}]}, 
                  {RGBColor[0, 1, 0], 
                   DiskBox[{-1.6723650854308285`, 1.4777261599691793`}, 
                    0.05]}, 
                  {RGBColor[0, 0, 1], Arrowheads[0.01], 
                   CircleBox[{-1.8822497111467744`, 0.5}, 0.3333333333333333, 
                    NCache[{
                    Rational[-1, 2] 
                    Pi, -4.645424023375199}, {-1.5707963267948966`, \
-4.645424023375199}]], 
                   ArrowBox[{{-1.9078793759854116`, 
                    0.8323465531510297}, {-1.9045546843721535`, 
                    0.8325862283386464}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-1.8822497111467744`, 0.5}, {
                    0.16750208657342713`, 0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-1.8822497111467744`, 
                    0.5}, {-3.3915644096343285`, 0.5}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.005], Arrowheads[0.04], 
                   ArrowBox[{{-1.8822497111467744`, 0.5}, {-1.341812611914127,
                     0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-1.8822497111467744`, 0.5}, 1.6, {
                    1.3593393711880297`, 0.3945276220736791}], 
                   ArrowBox[{{-0.41138823854015105`, 
                    1.1297352843866029`}, {-0.4051645331124143, 
                    1.114995428301059}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Dashing[{Small, Small}], 
                   Arrowheads[0.02], 
                   CircleBox[{-1.8822497111467744`, 0.5}, 1.7, {
                    1.3593393711880297`, 1.6654146044882026`}], 
                   ArrowBox[{{-2.0259291763312115`, 
                    2.193917415721414}, {-2.042860884256875, 
                    2.192395949851038}}]}, 
                  {RGBColor[0, 0, 1], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-1.8822497111467744`, 0.5}, 1.8, {
                    1.3593393711880297`, 1.9287148705547525`}], 
                   ArrowBox[{{-2.495944956325484, 
                    2.192151927589554}, {-2.512835509070915, 
                    2.185930470528478}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.003], Arrowheads[0.025], 
                   CircleBox[{-1.8822497111467744`, 0.5}, 1.5, {
                    1.3593393711880297`, 1.3603427024463186`}], 
                   ArrowBox[{{-1.554241273960805, 
                    1.963697531983578}, {-1.5688944056174612`, 
                    1.966904377420911}}]}, 
                  {RGBColor[0, 0, 1], ArrowBox[{{0, 2.8}, {1.5, 2.8}}]}},
                 Frame->True,
                 FrameTicks->Automatic,
                 ImageSize->800,
                 PlotRange->{{-3, 3}, {-1, 3}}], 65->
                GraphicsBox[{InsetBox[
                   TagBox[GridBox[{
                    {"\<\"t\"\>", "3.2`"},
                    {"\<\"driveIntent\"\>", "0.5`"},
                    {"\<\"motorTorque\"\>", 
                    RowBox[{"-", "95.98436959792566`"}]},
                    {"\<\"vw\"\>", "0.26449277055678505`"},
                    {"\<\"controllerState\"\>", 
                    RowBox[{"\[LeftAssociation]", 
                    RowBox[{
                    RowBox[{"\<\"lastError\"\>", "\[Rule]", 
                    RowBox[{"-", "0.12771138760860698`"}]}], ",", 
                    RowBox[{"\<\"lastInt\"\>", "\[Rule]", 
                    RowBox[{"-", "0.1403264785233238`"}]}], ",", 
                    RowBox[{"\<\"initialized\"\>", "\[Rule]", "True"}]}], 
                    "\[RightAssociation]"}]}
                    },
                    AutoDelete->False,
                    
                    GridBoxItemSize->{
                    "Columns" -> {{Automatic}}, "Rows" -> {{Automatic}}}],
                    "Grid"], {-1, -0.5}], 
                  {GrayLevel[0], LineBox[{{-5, 0}, {5, 0}}]}, 
                  {GrayLevel[0.5], 
                   CircleBox[{-1.8692381384474288`, 0.5}, 0.5], 
                   DiskBox[{-1.8692381384474288`, 0.5}, 0.1]}, 
                  {GrayLevel[0.5], 
                   DiskBox[{-1.5739562202433874`, 1.9706490365759406`}, 0.1]}, 
                  {GrayLevel[0.5], Thickness[0.01], 
                   LineBox[{{-1.8692381384474288`, 
                    0.5}, {-1.5739562202433874`, 1.9706490365759406`}}]}, 
                  {RGBColor[0, 1, 0], 
                   DiskBox[{-1.6723835263114013`, 1.480432691050627}, 0.05]}, 
                  {RGBColor[0, 0, 1], Arrowheads[0.01], 
                   CircleBox[{-1.8692381384474288`, 0.5}, 0.3333333333333333, 
                    NCache[{
                    Rational[-1, 2] 
                    Pi, -4.450327414732667}, {-1.5707963267948966`, \
-4.450327414732667}]], 
                   ArrowBox[{{-1.958810720262978, 
                    0.8210729881164216}, {-1.9555955653019421`, 
                    0.8219526454902623}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-1.8692381384474288`, 0.5}, {
                    0.050449253511084535`, 0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-1.8692381384474288`, 
                    0.5}, {-3.2887705632478847`, 0.5}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.005], Arrowheads[0.04], 
                   ArrowBox[{{-1.8692381384474288`, 
                    0.5}, {-1.3690831712893714`, 0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-1.8692381384474288`, 0.5}, 1.6, {
                    1.3726476084245696`, 0.4677331011884256}], 
                   ArrowBox[{{-0.44837501926376966`, 
                    1.2356276208406554`}, {-0.4410899082232884, 
                    1.221382445383623}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Dashing[{Small, Small}], 
                   Arrowheads[0.02], 
                   CircleBox[{-1.8692381384474288`, 0.5}, 1.7, {
                    1.3726476084245696`, 1.6567110401495009`}], 
                   ArrowBox[{{-1.9981692285844728`, 
                    2.195103764964279}, {-2.0151135372174482`, 
                    2.1937297210693587`}}]}, 
                  {RGBColor[0, 0, 1], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-1.8692381384474288`, 0.5}, 1.8, {
                    1.3726476084245696`, 1.9058941061908232`}], 
                   ArrowBox[{{-2.44416073953024, 
                    2.2057151001161284`}, {-2.4611888603564687`, 
                    2.1998806848809616`}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.003], Arrowheads[0.025], 
                   CircleBox[{-1.8692381384474288`, 0.5}, 1.5, {
                    1.3726476084245696`, 1.3692398206471093`}], 
                   ArrowBox[{{-1.5836573966276613`, 
                    1.9725636284730352`}, {-1.5689462846870748`, 
                    1.9696342410835324`}}]}, 
                  {RGBColor[0, 0, 1], ArrowBox[{{0, 2.8}, {1.5, 2.8}}]}},
                 Frame->True,
                 FrameTicks->Automatic,
                 ImageSize->800,
                 PlotRange->{{-3, 3}, {-1, 3}}], 66->
                GraphicsBox[{InsetBox[
                   TagBox[GridBox[{
                    {"\<\"t\"\>", "3.25`"},
                    {"\<\"driveIntent\"\>", "0.5`"},
                    {"\<\"motorTorque\"\>", 
                    RowBox[{"-", "89.52359104519174`"}]},
                    {"\<\"vw\"\>", "0.2726074212147546`"},
                    {"\<\"controllerState\"\>", 
                    RowBox[{"\[LeftAssociation]", 
                    RowBox[{
                    RowBox[{"\<\"lastError\"\>", "\[Rule]", 
                    RowBox[{"-", "0.1169283526612671`"}]}], ",", 
                    RowBox[{"\<\"lastInt\"\>", "\[Rule]", 
                    RowBox[{"-", "0.14643735565956556`"}]}], ",", 
                    RowBox[{"\<\"initialized\"\>", "\[Rule]", "True"}]}], 
                    "\[RightAssociation]"}]}
                    },
                    AutoDelete->False,
                    
                    GridBoxItemSize->{
                    "Columns" -> {{Automatic}}, "Rows" -> {{Automatic}}}],
                    "Grid"], {-1, -0.5}], 
                  {GrayLevel[0], LineBox[{{-5, 0}, {5, 0}}]}, 
                  {GrayLevel[0.5], 
                   CircleBox[{-1.8558037046082847`, 0.5}, 0.5], 
                   DiskBox[{-1.8558037046082847`, 0.5}, 0.1]}, 
                  {GrayLevel[0.5], 
                   DiskBox[{-1.5799871367666256`, 1.9744236911091897`}, 0.1]}, 
                  {GrayLevel[0.5], Thickness[0.01], 
                   LineBox[{{-1.8558037046082847`, 
                    0.5}, {-1.5799871367666256`, 1.9744236911091897`}}]}, 
                  {RGBColor[0, 1, 0], 
                   DiskBox[{-1.6719259927138452`, 1.4829491274061266`}, 
                    0.05]}, 
                  {RGBColor[0, 0, 1], Arrowheads[0.01], 
                   CircleBox[{-1.8558037046082847`, 0.5}, 0.3333333333333333, 
                    NCache[{
                    Rational[-1, 2] 
                    Pi, -4.256504058150648}, {-1.5707963267948966`, \
-4.256504058150648}]], 
                   ArrowBox[{{-2.0055415673889416`, 
                    0.7978081321270328}, {-2.002556048871363, 
                    0.7992905955161337}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-1.8558037046082847`, 
                    0.5}, {-0.06533188370444987, 0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-1.8558037046082847`, 0.5}, {-3.185162065462256,
                     0.5}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.005], Arrowheads[0.04], 
                   ArrowBox[{{-1.8558037046082847`, 
                    0.5}, {-1.3946902445584213`, 0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-1.8558037046082847`, 0.5}, 1.6, {
                    1.3858663427117375`, 0.5406049223916103}], 
                   ArrowBox[{{-0.4922706151881451, 
                    1.3371245511011902`}, {-0.4839676852835282, 
                    1.3234475915826134`}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Dashing[{Small, Small}], 
                   Arrowheads[0.02], 
                   CircleBox[{-1.8558037046082847`, 0.5}, 1.7, {
                    1.3858663427117375`, 1.6484434877762801`}], 
                   ArrowBox[{{-1.970716188891405, 
                    2.196111765467088}, {-1.987671278285861, 
                    2.1948778548946795`}}]}, 
                  {RGBColor[0, 0, 1], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-1.8558037046082847`, 0.5}, 1.8, {
                    1.3858663427117375`, 1.883219626296136}], 
                   ArrowBox[{{-2.3919056302422996`, 
                    2.21831159145584}, {-2.4090616549001176`, 
                    2.212864746685762}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.003], Arrowheads[0.025], 
                   CircleBox[{-1.8558037046082847`, 0.5}, 1.5, {
                    1.3858663427117375`, 1.3786655324113617`}], 
                   ArrowBox[{{-1.5841154035604665`, 
                    1.9751899766042849`}, {-1.5693773339600399`, 
                    1.972399379990794}}]}, 
                  {RGBColor[0, 0, 1], ArrowBox[{{0, 2.8}, {1.5, 2.8}}]}},
                 Frame->True,
                 FrameTicks->Automatic,
                 ImageSize->800,
                 PlotRange->{{-3, 3}, {-1, 3}}], 67->
                GraphicsBox[{InsetBox[
                   TagBox[GridBox[{
                    {"\<\"t\"\>", "3.3000000000000003`"},
                    {"\<\"driveIntent\"\>", "0.5`"},
                    {"\<\"motorTorque\"\>", 
                    RowBox[{"-", "83.15208354744996`"}]},
                    {"\<\"vw\"\>", "0.28004778563967664`"},
                    {"\<\"controllerState\"\>", 
                    RowBox[{"\[LeftAssociation]", 
                    RowBox[{
                    RowBox[{"\<\"lastError\"\>", "\[Rule]", 
                    RowBox[{"-", "0.10613178883437734`"}]}], ",", 
                    RowBox[{"\<\"lastInt\"\>", "\[Rule]", 
                    RowBox[{"-", "0.1520083062545833`"}]}], ",", 
                    RowBox[{"\<\"initialized\"\>", "\[Rule]", "True"}]}], 
                    "\[RightAssociation]"}]}
                    },
                    AutoDelete->False,
                    
                    GridBoxItemSize->{
                    "Columns" -> {{Automatic}}, "Rows" -> {{Automatic}}}],
                    "Grid"], {-1, -0.5}], 
                  {GrayLevel[0], LineBox[{{-5, 0}, {5, 0}}]}, 
                  {GrayLevel[0.5], 
                   CircleBox[{-1.8419808600005254`, 0.5}, 0.5], 
                   DiskBox[{-1.8419808600005254`, 0.5}, 0.1]}, 
                  {GrayLevel[0.5], 
                   DiskBox[{-1.585395822640708, 1.9778917817631512`}, 0.1]}, 
                  {GrayLevel[0.5], Thickness[0.01], 
                   LineBox[{{-1.8419808600005254`, 0.5}, {-1.585395822640708, 
                    1.9778917817631512`}}]}, 
                  {RGBColor[0, 1, 0], 
                   DiskBox[{-1.6709241684273137`, 1.4852611878421007`}, 
                    0.05]}, 
                  {RGBColor[0, 0, 1], Arrowheads[0.01], 
                   CircleBox[{-1.8419808600005254`, 0.5}, 0.3333333333333333, 
                    NCache[{
                    Rational[-1, 2] 
                    Pi, -4.065358833218395}, {-1.5707963267948966`, \
-4.065358833218395}]], 
                   ArrowBox[{{-2.045570190301767, 
                    0.7639365372558393}, {-2.042920689536725, 
                    0.7659591999105766}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-1.8419808600005254`, 
                    0.5}, {-0.17893918905152617`, 0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-1.8419808600005254`, 
                    0.5}, {-3.0815574964912282`, 0.5}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.005], Arrowheads[0.04], 
                   ArrowBox[{{-1.8419808600005254`, 0.5}, {-1.418515825542229,
                     0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-1.8419808600005254`, 0.5}, 1.6, {
                    1.3988942583591535`, 0.6125693656639495}], 
                   ArrowBox[{{-0.5421682519216204, 
                    1.4329990267299935`}, {-0.5329034072421979, 
                    1.4199544677209677`}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Dashing[{Small, Small}], 
                   Arrowheads[0.02], 
                   CircleBox[{-1.8419808600005254`, 0.5}, 1.7, {
                    1.3988942583591535`, 1.6406128100103459`}], 
                   ArrowBox[{{-1.943608252159937, 
                    2.1969595967974245`}, {-1.9605724839754626`, 
                    2.1958584925408684`}}]}, 
                  {RGBColor[0, 0, 1], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-1.8419808600005254`, 0.5}, 1.8, {
                    1.3988942583591535`, 1.860850278067209}], 
                   ArrowBox[{{-2.3395143569865127`, 
                    2.2298729489118254`}, {-2.3567879216973697`, 
                    2.224811203937133}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.003], Arrowheads[0.025], 
                   CircleBox[{-1.8419808600005254`, 0.5}, 1.5, {
                    1.3988942583591535`, 1.3884426216719536`}], 
                   ArrowBox[{{-1.5847283786835993`, 
                    1.9777757478238316`}, {-1.569963730016965, 
                    1.975129377714208}}]}, 
                  {RGBColor[0, 0, 1], ArrowBox[{{0, 2.8}, {1.5, 2.8}}]}},
                 Frame->True,
                 FrameTicks->Automatic,
                 ImageSize->800,
                 PlotRange->{{-3, 3}, {-1, 3}}], 68->
                GraphicsBox[{InsetBox[
                   TagBox[GridBox[{
                    {"\<\"t\"\>", "3.35`"},
                    {"\<\"driveIntent\"\>", "0.5`"},
                    {"\<\"motorTorque\"\>", 
                    RowBox[{"-", "76.91119913199452`"}]},
                    {"\<\"vw\"\>", "0.28684511133437435`"},
                    {"\<\"controllerState\"\>", 
                    RowBox[{"\[LeftAssociation]", 
                    RowBox[{
                    RowBox[{"\<\"lastError\"\>", "\[Rule]", 
                    RowBox[{"-", "0.09541816378143075`"}]}], ",", 
                    RowBox[{"\<\"lastInt\"\>", "\[Rule]", 
                    RowBox[{"-", "0.15704116898313647`"}]}], ",", 
                    RowBox[{"\<\"initialized\"\>", "\[Rule]", "True"}]}], 
                    "\[RightAssociation]"}]}
                    },
                    AutoDelete->False,
                    
                    GridBoxItemSize->{
                    "Columns" -> {{Automatic}}, "Rows" -> {{Automatic}}}],
                    "Grid"], {-1, -0.5}], 
                  {GrayLevel[0], LineBox[{{-5, 0}, {5, 0}}]}, 
                  {GrayLevel[0.5], 
                   CircleBox[{-1.8278025271851734`, 0.5}, 0.5], 
                   DiskBox[{-1.8278025271851734`, 0.5}, 0.1]}, 
                  {GrayLevel[0.5], 
                   DiskBox[{-1.590081352336601, 1.9810430929006133`}, 0.1]}, 
                  {GrayLevel[0.5], Thickness[0.01], 
                   LineBox[{{-1.8278025271851734`, 0.5}, {-1.590081352336601, 
                    1.9810430929006133`}}]}, 
                  {RGBColor[0, 1, 0], 
                   DiskBox[{-1.6693217439527919`, 1.4873620619337422`}, 
                    0.05]}, 
                  {RGBColor[0, 0, 1], Arrowheads[0.01], 
                   CircleBox[{-1.8278025271851734`, 0.5}, 0.3333333333333333, 
                    NCache[{
                    Rational[-1, 2] 
                    Pi, -3.878132300754732}, {-1.5707963267948966`, \
-3.878132300754732}]], 
                   ArrowBox[{{-2.076961707385674, 
                    0.7214290270784878}, {-2.074734996164348, 
                    0.7239095059950782}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-1.8278025271851734`, 0.5}, {-0.289578544545283,
                     0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-1.8278025271851734`, 0.5}, {-2.978695620658419,
                     0.5}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.005], Arrowheads[0.04], 
                   ArrowBox[{{-1.8278025271851734`, 
                    0.5}, {-1.4404716380185283`, 0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-1.8278025271851734`, 0.5}, 1.6, {
                    1.411644526785084, 0.6831294144948489}], 
                   ArrowBox[{{-0.5970021222728501, 
                    1.522316175783041}, {-0.586840670407611, 
                    1.5099572614834673`}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Dashing[{Small, Small}], 
                   Arrowheads[0.02], 
                   CircleBox[{-1.8278025271851734`, 0.5}, 1.7, {
                    1.411644526785084, 1.6332193452755948`}], 
                   ArrowBox[{{-1.9168808450810848`, 
                    2.1976645879798618`}, {-1.9338527541394213`, 
                    2.196688937125174}}]}, 
                  {RGBColor[0, 0, 1], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-1.8278025271851734`, 0.5}, 1.8, {
                    1.411644526785084, 1.8389289664072759`}], 
                   ArrowBox[{{-2.2872984386196524`, 
                    2.2403630389591127`}, {-2.3046788043460715`, 
                    2.2356811390002376`}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.003], Arrowheads[0.025], 
                   CircleBox[{-1.8278025271851734`, 0.5}, 1.5, {
                    1.411644526785084, 1.3984211955113757`}], 
                   ArrowBox[{{-1.5853087030285518`, 
                    1.9802691462183144`}, {-1.5705183828668283`, 
                    1.9777702355516493`}}]}, 
                  {RGBColor[0, 0, 1], ArrowBox[{{0, 2.8}, {1.5, 2.8}}]}},
                 Frame->True,
                 FrameTicks->Automatic,
                 ImageSize->800,
                 PlotRange->{{-3, 3}, {-1, 3}}], 69->
                GraphicsBox[{InsetBox[
                   TagBox[GridBox[{
                    {"\<\"t\"\>", "3.4`"},
                    {"\<\"driveIntent\"\>", "0.5`"},
                    {"\<\"motorTorque\"\>", 
                    RowBox[{"-", "70.83720857411255`"}]},
                    {"\<\"vw\"\>", "0.29303161732328836`"},
                    {"\<\"controllerState\"\>", 
                    RowBox[{"\[LeftAssociation]", 
                    RowBox[{
                    RowBox[{"\<\"lastError\"\>", "\[Rule]", 
                    RowBox[{"-", "0.08487140750121303`"}]}], ",", 
                    RowBox[{"\<\"lastInt\"\>", "\[Rule]", 
                    RowBox[{"-", "0.16154228038391522`"}]}], ",", 
                    RowBox[{"\<\"initialized\"\>", "\[Rule]", "True"}]}], 
                    "\[RightAssociation]"}]}
                    },
                    AutoDelete->False,
                    
                    GridBoxItemSize->{
                    "Columns" -> {{Automatic}}, "Rows" -> {{Automatic}}}],
                    "Grid"], {-1, -0.5}], 
                  {GrayLevel[0], LineBox[{{-5, 0}, {5, 0}}]}, 
                  {GrayLevel[0.5], 
                   CircleBox[{-1.8133000396230001`, 0.5}, 0.5], 
                   DiskBox[{-1.8133000396230001`, 0.5}, 0.1]}, 
                  {GrayLevel[0.5], 
                   DiskBox[{-1.5939595177579138`, 1.9838765903766902`}, 0.1]}, 
                  {GrayLevel[0.5], Thickness[0.01], 
                   LineBox[{{-1.8133000396230001`, 
                    0.5}, {-1.5939595177579138`, 1.9838765903766902`}}]}, 
                  {RGBColor[0, 1, 0], 
                   DiskBox[{-1.667073025046276, 1.4892510602511266`}, 0.05]}, 
                  {RGBColor[0, 0, 1], Arrowheads[0.01], 
                   CircleBox[{-1.8133000396230001`, 0.5}, 0.3333333333333333, 
                    NCache[{
                    Rational[-1, 2] 
                    Pi, -3.695912584018273}, {-1.5707963267948966`, \
-3.695912584018273}]], 
                   ArrowBox[{{-2.098459924081718, 
                    0.6726121415387744}, {-2.09671957355947, 
                    0.6754550623217963}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-1.8133000396230001`, 
                    0.5}, {-0.39655586814074906`, 0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-1.8133000396230001`, 0.5}, {-2.877237442252925,
                     0.5}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.005], Arrowheads[0.04], 
                   ArrowBox[{{-1.8133000396230001`, 0.5}, {-1.460493270770674,
                     0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-1.8133000396230001`, 0.5}, 1.6, {
                    1.4240431206643915`, 0.7518569554305636}], 
                   ArrowBox[{{-0.6556112950826061, 
                    1.6044259915288512`}, {-0.6446251031922541, 
                    1.5927940761912063`}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Dashing[{Small, Small}], 
                   Arrowheads[0.02], 
                   CircleBox[{-1.8133000396230001`, 0.5}, 1.7, {
                    1.4240431206643915`, 1.6262630146814896`}], 
                   ArrowBox[{{-1.8905667813865352`, 
                    2.198243165926849}, {-1.9075450667017975`, 
                    2.1973855999362417`}}]}, 
                  {RGBColor[0, 0, 1], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-1.8133000396230001`, 0.5}, 1.8, {
                    1.4240431206643915`, 1.81758316829835}], 
                   ArrowBox[{{-2.2355446549795834`, 
                    2.2497741239378213`}, {-2.253020992536566, 
                    2.2454642601808796`}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.003], Arrowheads[0.025], 
                   CircleBox[{-1.8133000396230001`, 0.5}, 1.5, {
                    1.4240431206643915`, 1.4084751481284368`}], 
                   ArrowBox[{{-1.5857007763705657`, 
                    1.9826323129376848`}, {-1.5708860802136686`, 
                    1.9802822272402958`}}]}, 
                  {RGBColor[0, 0, 1], ArrowBox[{{0, 2.8}, {1.5, 2.8}}]}},
                 Frame->True,
                 FrameTicks->Automatic,
                 ImageSize->800,
                 PlotRange->{{-3, 3}, {-1, 3}}], 70->
                GraphicsBox[{InsetBox[
                   TagBox[GridBox[{
                    {"\<\"t\"\>", "3.45`"},
                    {"\<\"driveIntent\"\>", "0.5`"},
                    {"\<\"motorTorque\"\>", 
                    RowBox[{"-", "64.96161040271055`"}]},
                    {"\<\"vw\"\>", "0.29864008798166475`"},
                    {"\<\"controllerState\"\>", 
                    RowBox[{"\[LeftAssociation]", 
                    RowBox[{
                    RowBox[{"\<\"lastError\"\>", "\[Rule]", 
                    RowBox[{"-", "0.07456418814195909`"}]}], ",", 
                    RowBox[{"\<\"lastInt\"\>", "\[Rule]", 
                    RowBox[{"-", "0.16552188147746752`"}]}], ",", 
                    RowBox[{"\<\"initialized\"\>", "\[Rule]", "True"}]}], 
                    "\[RightAssociation]"}]}
                    },
                    AutoDelete->False,
                    
                    GridBoxItemSize->{
                    "Columns" -> {{Automatic}}, "Rows" -> {{Automatic}}}],
                    "Grid"], {-1, -0.5}], 
                  {GrayLevel[0], LineBox[{{-5, 0}, {5, 0}}]}, 
                  {GrayLevel[0.5], 
                   CircleBox[{-1.7985031039538855`, 0.5}, 0.5], 
                   DiskBox[{-1.7985031039538855`, 0.5}, 0.1]}, 
                  {GrayLevel[0.5], 
                   DiskBox[{-1.5969611300317652`, 1.9863986116609418`}, 0.1]}, 
                  {GrayLevel[0.5], Thickness[0.01], 
                   LineBox[{{-1.7985031039538855`, 
                    0.5}, {-1.5969611300317652`, 1.9863986116609418`}}]}, 
                  {RGBColor[0, 1, 0], 
                   DiskBox[{-1.6641417880058054`, 1.4909324077739612`}, 
                    0.05]}, 
                  {RGBColor[0, 0, 1], Arrowheads[0.01], 
                   CircleBox[{-1.7985031039538855`, 0.5}, 0.3333333333333333, 
                    NCache[{
                    Rational[-1, 2] 
                    Pi, -3.519644638876213}, {-1.5707963267948966`, \
-3.519644638876213}]], 
                   ArrowBox[{{-2.109513106741658, 
                    0.6199328532015345}, {-2.108298247827799, 
                    0.6230369048019826}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-1.7985031039538855`, 
                    0.5}, {-0.49927089589967455`, 0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-1.7985031039538855`, 
                    0.5}, {-2.7777680746604276`, 0.5}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.005], Arrowheads[0.04], 
                   ArrowBox[{{-1.7985031039538855`, 
                    0.5}, {-1.4785358666062167`, 0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-1.7985031039538855`, 0.5}, 1.6, {
                    1.4360274209111639`, 0.818386296360114}], 
                   ArrowBox[{{-0.7167980105256604, 
                    1.6789461780978108`}, {-0.7050628300386919, 
                    1.6680703606291303`}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Dashing[{Small, Small}], 
                   Arrowheads[0.02], 
                   CircleBox[{-1.7985031039538855`, 0.5}, 1.7, {
                    1.4360274209111639`, 1.6197432157797327`}], 
                   ArrowBox[{{-1.864696078018667, 
                    2.198710831832322}, {-1.8816795935988109`, 
                    2.197963978289984}}]}, 
                  {RGBColor[0, 0, 1], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-1.7985031039538855`, 0.5}, 1.8, {
                    1.4360274209111639`, 1.7969252564817781`}], 
                   ArrowBox[{{-2.184513517535505, 
                    2.2581228513976344`}, {-2.2020751526706293`, 
                    2.254174906186544}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.003], Arrowheads[0.025], 
                   CircleBox[{-1.7985031039538855`, 0.5}, 1.5, {
                    1.4360274209111639`, 1.418498858259159}], 
                   ArrowBox[{{-1.585776502266289, 
                    1.9848391808322026`}, {-1.570938994171375, 
                    1.9826377089292224`}}]}, 
                  {RGBColor[0, 0, 1], ArrowBox[{{0, 2.8}, {1.5, 2.8}}]}},
                 Frame->True,
                 FrameTicks->Automatic,
                 ImageSize->800,
                 PlotRange->{{-3, 3}, {-1, 3}}], 71->
                GraphicsBox[{InsetBox[
                   TagBox[GridBox[{
                    {"\<\"t\"\>", "3.5`"},
                    {"\<\"driveIntent\"\>", "0.5`"},
                    {"\<\"motorTorque\"\>", 
                    RowBox[{"-", "59.31138256275949`"}]},
                    {"\<\"vw\"\>", "0.30370356903507917`"},
                    {"\<\"controllerState\"\>", 
                    RowBox[{"\[LeftAssociation]", 
                    RowBox[{
                    RowBox[{"\<\"lastError\"\>", "\[Rule]", 
                    RowBox[{"-", "0.06455897541719276`"}]}], ",", 
                    RowBox[{"\<\"lastInt\"\>", "\[Rule]", 
                    RowBox[{"-", "0.16899358259223127`"}]}], ",", 
                    RowBox[{"\<\"initialized\"\>", "\[Rule]", "True"}]}], 
                    "\[RightAssociation]"}]}
                    },
                    AutoDelete->False,
                    
                    GridBoxItemSize->{
                    "Columns" -> {{Automatic}}, "Rows" -> {{Automatic}}}],
                    "Grid"], {-1, -0.5}], 
                  {GrayLevel[0], LineBox[{{-5, 0}, {5, 0}}]}, 
                  {GrayLevel[0.5], CircleBox[{-1.783439779839006, 0.5}, 0.5], 
                   DiskBox[{-1.783439779839006, 0.5}, 0.1]}, 
                  {GrayLevel[0.5], 
                   DiskBox[{-1.5990306066810283`, 1.98862126037995}, 0.1]}, 
                  {GrayLevel[0.5], Thickness[0.01], 
                   LineBox[{{-1.783439779839006, 0.5}, {-1.5990306066810283`, 
                    1.98862126037995}}]}, 
                  {RGBColor[0, 1, 0], 
                   DiskBox[{-1.6605003310670208`, 1.4924141735866332`}, 
                    0.05]}, 
                  {RGBColor[0, 0, 1], Arrowheads[0.01], 
                   CircleBox[{-1.783439779839006, 0.5}, 0.3333333333333333, 
                    NCache[{
                    Rational[-1, 2] 
                    Pi, -3.3501378036776814`}, {-1.5707963267948966`, \
-3.3501378036776814`}]], 
                   ArrowBox[{{-2.1102246380387397`, 
                    0.5657477570909548}, {-2.1095508323189844`, 
                    0.5690122638486217}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-1.783439779839006, 0.5}, {-0.5972121285838161, 
                    0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-1.783439779839006, 0.5}, {-2.680798647199342, 
                    0.5}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.005], Arrowheads[0.04], 
                   ArrowBox[{{-1.783439779839006, 0.5}, {-1.4945709959441522`,
                     0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-1.783439779839006, 0.5}, 1.6, {
                    1.4475450659038074`, 0.8824087477945314}], 
                   ArrowBox[{{-0.7793782877085846, 
                    1.745736938531098}, {-0.7669713286013078, 
                    1.7356342046246676`}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Dashing[{Small, Small}], 
                   Arrowheads[0.02], 
                   CircleBox[{-1.783439779839006, 0.5}, 1.7, {
                    1.4475450659038074`, 1.6136586031839992`}], 
                   ArrowBox[{{-1.8392955950459655`, 
                    2.199082142778143}, {-1.8562833405273187`, 
                    2.198438640536139}}]}, 
                  {RGBColor[0, 0, 1], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-1.783439779839006, 0.5}, 1.8, {
                    1.4475450659038074`, 1.7770527468080268`}], 
                   ArrowBox[{{-2.1344379611638677`, 
                    2.265446197624453}, {-2.152074579137733, 
                    2.261848002736328}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.003], Arrowheads[0.025], 
                   CircleBox[{-1.783439779839006, 0.5}, 1.5, {
                    1.4475450659038074`, 1.428404219256527}], 
                   ArrowBox[{{-1.5854312416235412`, 
                    1.9868734373825414`}, {-1.5705726554051238`, 
                    1.9848190419493044`}}]}, 
                  {RGBColor[0, 0, 1], ArrowBox[{{0, 2.8}, {1.5, 2.8}}]}},
                 Frame->True,
                 FrameTicks->Automatic,
                 ImageSize->800,
                 PlotRange->{{-3, 3}, {-1, 3}}], 72->
                GraphicsBox[{InsetBox[
                   TagBox[GridBox[{
                    {"\<\"t\"\>", "3.5500000000000003`"},
                    {"\<\"driveIntent\"\>", "0.5`"},
                    {"\<\"motorTorque\"\>", 
                    RowBox[{"-", "53.909199909873834`"}]},
                    {"\<\"vw\"\>", "0.30825513601153404`"},
                    {"\<\"controllerState\"\>", 
                    RowBox[{"\[LeftAssociation]", 
                    RowBox[{
                    RowBox[{"\<\"lastError\"\>", "\[Rule]", 
                    RowBox[{"-", "0.054908931066588194`"}]}], ",", 
                    RowBox[{"\<\"lastInt\"\>", "\[Rule]", 
                    RowBox[{"-", "0.17197387679023945`"}]}], ",", 
                    RowBox[{"\<\"initialized\"\>", "\[Rule]", "True"}]}], 
                    "\[RightAssociation]"}]}
                    },
                    AutoDelete->False,
                    
                    GridBoxItemSize->{
                    "Columns" -> {{Automatic}}, "Rows" -> {{Automatic}}}],
                    "Grid"], {-1, -0.5}], 
                  {GrayLevel[0], LineBox[{{-5, 0}, {5, 0}}]}, 
                  {GrayLevel[0.5], CircleBox[{-1.768136473047532, 0.5}, 0.5], 
                   DiskBox[{-1.768136473047532, 0.5}, 0.1]}, 
                  {GrayLevel[0.5], 
                   DiskBox[{-1.6001247784172317`, 1.9905609918642897`}, 0.1]}, 
                  {GrayLevel[0.5], Thickness[0.01], 
                   LineBox[{{-1.768136473047532, 0.5}, {-1.6001247784172317`, 
                    1.9905609918642897`}}]}, 
                  {RGBColor[0, 1, 0], 
                   DiskBox[{-1.6561286766273318`, 1.4937073279095263`}, 
                    0.05]}, 
                  {RGBColor[0, 0, 1], Arrowheads[0.01], 
                   CircleBox[{-1.768136473047532, 0.5}, 0.3333333333333333, 
                    NCache[{
                    Rational[-1, 2] 
                    Pi, -3.1880723240911117`}, {-1.5707963267948966`, \
-3.1880723240911117`}]], 
                   ArrowBox[{{-2.1012480365828363`, 
                    0.5121571933510917}, {-2.1011098112361335`, 
                    0.5154876456135263}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-1.768136473047532, 0.5}, {-0.6899524748500552, 
                    0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-1.768136473047532, 0.5}, {-2.586768412031132, 
                    0.5}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.005], Arrowheads[0.04], 
                   ArrowBox[{{-1.768136473047532, 0.5}, {-1.5085844138336553`,
                     0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-1.768136473047532, 0.5}, 1.6, {
                    1.4585529946300944`, 0.9436680305218963}], 
                   ArrowBox[{{-0.8422235928110137, 
                    1.8048698548943933`}, {-0.8292214069975073, 
                    1.795545637461019}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Dashing[{Small, Small}], 
                   Arrowheads[0.02], 
                   CircleBox[{-1.768136473047532, 0.5}, 1.7, {
                    1.4585529946300944`, 1.608006824263621}], 
                   ArrowBox[{{-1.8143886106513893`, 
                    2.1993706893338705`}, {-1.8313797217300873`, 
                    2.1988232078400847`}}]}, 
                  {RGBColor[0, 0, 1], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-1.768136473047532, 0.5}, 1.8, {
                    1.4585529946300944`, 1.7580485496849492`}], 
                   ArrowBox[{{-2.0855224050653978`, 
                    2.2717974404985326`}, {-2.1032242150079288`, 
                    2.2685350449419666`}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.003], Arrowheads[0.025], 
                   CircleBox[{-1.768136473047532, 0.5}, 1.5, {
                    1.4585529946300944`, 1.438118039231839}], 
                   ArrowBox[{{-1.584580270391243, 
                    1.9887266775558243`}, {-1.5697024294692086`, 
                    1.986816710408232}}]}, 
                  {RGBColor[0, 0, 1], ArrowBox[{{0, 2.8}, {1.5, 2.8}}]}},
                 Frame->True,
                 FrameTicks->Automatic,
                 ImageSize->800,
                 PlotRange->{{-3, 3}, {-1, 3}}], 73->
                GraphicsBox[{InsetBox[
                   TagBox[GridBox[{
                    {"\<\"t\"\>", "3.6`"},
                    {"\<\"driveIntent\"\>", "0.5`"},
                    {"\<\"motorTorque\"\>", 
                    RowBox[{"-", "48.7736351498084`"}]},
                    {"\<\"vw\"\>", "0.3123277125201935`"},
                    {"\<\"controllerState\"\>", 
                    RowBox[{"\[LeftAssociation]", 
                    RowBox[{
                    RowBox[{"\<\"lastError\"\>", "\[Rule]", 
                    RowBox[{"-", "0.04565866336098477`"}]}], ",", 
                    RowBox[{"\<\"lastInt\"\>", "\[Rule]", 
                    RowBox[{"-", "0.17448169420241338`"}]}], ",", 
                    RowBox[{"\<\"initialized\"\>", "\[Rule]", "True"}]}], 
                    "\[RightAssociation]"}]}
                    },
                    AutoDelete->False,
                    
                    GridBoxItemSize->{
                    "Columns" -> {{Automatic}}, "Rows" -> {{Automatic}}}],
                    "Grid"], {-1, -0.5}], 
                  {GrayLevel[0], LineBox[{{-5, 0}, {5, 0}}]}, 
                  {GrayLevel[0.5], 
                   CircleBox[{-1.7526179387689373`, 0.5}, 0.5], 
                   DiskBox[{-1.7526179387689373`, 0.5}, 0.1]}, 
                  {GrayLevel[0.5], 
                   DiskBox[{-1.6002118587405225`, 1.9922373761470968`}, 0.1]}, 
                  {GrayLevel[0.5], Thickness[0.01], 
                   LineBox[{{-1.7526179387689373`, 
                    0.5}, {-1.6002118587405225`, 1.9922373761470968`}}]}, 
                  {RGBColor[0, 1, 0], 
                   DiskBox[{-1.6510138854166607`, 1.494824917431398}, 0.05]}, 
                  {RGBColor[0, 0, 1], Arrowheads[0.01], 
                   CircleBox[{-1.7526179387689373`, 0.5}, 0.3333333333333333, 
                    NCache[{
                    Rational[-1, 2] 
                    Pi, -3.0340053812891483`}, {-1.5707963267948966`, \
-3.0340053812891483`}]], 
                   ArrowBox[{{-2.0836494650480786`, 
                    0.46089450524013975`}, {-2.0840239620397423`, 
                    0.46420672058972995`}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-1.7526179387689373`, 
                    0.5}, {-0.7771452357727694, 0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-1.7526179387689373`, 
                    0.5}, {-2.4960471598143448`, 0.5}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.005], Arrowheads[0.04], 
                   ArrowBox[{{-1.7526179387689373`, 
                    0.5}, {-1.5205744568181767`, 0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-1.7526179387689373`, 0.5}, 1.6, {
                    1.4690166400333635`, 1.001956301061284}], 
                   ArrowBox[{{-0.9042930641301161, 
                    1.8565931251001637`}, {-0.8907697748671026, 
                    1.8480421886495382`}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Dashing[{Small, Small}], 
                   Arrowheads[0.02], 
                   CircleBox[{-1.7526179387689373`, 0.5}, 1.7, {
                    1.4690166400333635`, 1.6027842559838144`}], 
                   ArrowBox[{{-1.7899944064927806`, 
                    2.1995890678809653`}, {-1.806988145100349, 
                    2.199130330687873}}]}, 
                  {RGBColor[0, 0, 1], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-1.7526179387689373`, 0.5}, 1.8, {
                    1.4690166400333635`, 1.739981279754521}], 
                   ArrowBox[{{-2.037942269567237, 
                    2.2772422531142182`}, {-2.055700129795163, 
                    2.274300195987912}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.003], Arrowheads[0.025], 
                   CircleBox[{-1.7526179387689373`, 0.5}, 1.5, {
                    1.4690166400333635`, 1.447579810872379}], 
                   ArrowBox[{{-1.583155734181251, 
                    1.9903967797926434`}, {-1.5682604878211663`, 
                    1.988627666772333}}]}, 
                  {RGBColor[0, 0, 1], ArrowBox[{{0, 2.8}, {1.5, 2.8}}]}},
                 Frame->True,
                 FrameTicks->Automatic,
                 ImageSize->800,
                 PlotRange->{{-3, 3}, {-1, 3}}], 74->
                GraphicsBox[{InsetBox[
                   TagBox[GridBox[{
                    {"\<\"t\"\>", "3.65`"},
                    {"\<\"driveIntent\"\>", "0.5`"},
                    {"\<\"motorTorque\"\>", 
                    RowBox[{"-", "43.91935609669121`"}]},
                    {"\<\"vw\"\>", "0.3159539215610666`"},
                    {"\<\"controllerState\"\>", 
                    RowBox[{"\[LeftAssociation]", 
                    RowBox[{
                    RowBox[{"\<\"lastError\"\>", "\[Rule]", 
                    RowBox[{"-", "0.03684487863006069`"}]}], ",", 
                    RowBox[{"\<\"lastInt\"\>", "\[Rule]", 
                    RowBox[{"-", "0.17653799133547107`"}]}], ",", 
                    RowBox[{"\<\"initialized\"\>", "\[Rule]", "True"}]}], 
                    "\[RightAssociation]"}]}
                    },
                    AutoDelete->False,
                    
                    GridBoxItemSize->{
                    "Columns" -> {{Automatic}}, "Rows" -> {{Automatic}}}],
                    "Grid"], {-1, -0.5}], 
                  {GrayLevel[0], LineBox[{{-5, 0}, {5, 0}}]}, 
                  {GrayLevel[0.5], CircleBox[{-1.736907293109776, 0.5}, 0.5], 
                   DiskBox[{-1.736907293109776, 0.5}, 0.1]}, 
                  {GrayLevel[0.5], 
                   DiskBox[{-1.5992705291431046`, 1.9936720259832086`}, 0.1]}, 
                  {GrayLevel[0.5], Thickness[0.01], 
                   LineBox[{{-1.736907293109776, 0.5}, {-1.5992705291431046`, 
                    1.9936720259832086`}}]}, 
                  {RGBColor[0, 1, 0], 
                   DiskBox[{-1.6451494504653283`, 1.4957813506554722`}, 
                    0.05]}, 
                  {RGBColor[0, 0, 1], Arrowheads[0.01], 
                   CircleBox[{-1.736907293109776, 0.5}, 0.3333333333333333, 
                    NCache[{
                    Rational[-1, 2] 
                    Pi, -2.888377009695633}, {-1.5707963267948966`, \
-2.888377009695633}]], 
                   ArrowBox[{{-2.058760060990991, 
                    0.4132710721943824}, {-2.0596112433100084`, 
                    0.41649388264158804`}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-1.736907293109776, 0.5}, {-0.8585201711759518, 
                    0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-1.736907293109776, 0.5}, {-2.408938006058473, 
                    0.5}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.005], Arrowheads[0.04], 
                   ArrowBox[{{-1.736907293109776, 0.5}, {-1.5305508841246485`,
                     0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-1.736907293109776, 0.5}, 1.6, {
                    1.4789092346331985`, 1.0571106185851495`}], 
                   ArrowBox[{{-0.9646564320331769, 
                    1.9012953320290664`}, {-0.9506823244822241, 
                    1.8935028879434033`}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Dashing[{Small, Small}], 
                   Arrowheads[0.02], 
                   CircleBox[{-1.736907293109776, 0.5}, 1.7, {
                    1.4789092346331985`, 1.5979857734632699`}], 
                   ArrowBox[{{-1.7661279133924928`, 
                    2.1997488506689136`}, {-1.7831236575902842`, 
                    2.1993716626018585`}}]}, 
                  {RGBColor[0, 0, 1], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-1.736907293109776, 0.5}, 1.8, {
                    1.4789092346331985`, 1.722905657392594}], 
                   ArrowBox[{{-1.9918439851209178`, 
                    2.2818550117970924`}, {-2.009649491536161, 
                    2.279216595358064}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.003], Arrowheads[0.025], 
                   CircleBox[{-1.736907293109776, 0.5}, 1.5, {
                    1.4789092346331985`, 1.4567398288339684`}], 
                   ArrowBox[{{-1.58110406827302, 
                    1.9918865087969886`}, {-1.5661932419278823`, 
                    1.9902539088118727`}}]}, 
                  {RGBColor[0, 0, 1], ArrowBox[{{0, 2.8}, {1.5, 2.8}}]}},
                 Frame->True,
                 FrameTicks->Automatic,
                 ImageSize->800,
                 PlotRange->{{-3, 3}, {-1, 3}}], 75->
                GraphicsBox[{InsetBox[
                   TagBox[GridBox[{
                    {"\<\"t\"\>", "3.7`"},
                    {"\<\"driveIntent\"\>", "0.5`"},
                    {"\<\"motorTorque\"\>", 
                    RowBox[{"-", "39.35732809738633`"}]},
                    {"\<\"vw\"\>", "0.31916595781476104`"},
                    {"\<\"controllerState\"\>", 
                    RowBox[{"\[LeftAssociation]", 
                    RowBox[{
                    RowBox[{"\<\"lastError\"\>", "\[Rule]", 
                    RowBox[{"-", "0.028496957988320477`"}]}], ",", 
                    RowBox[{"\<\"lastInt\"\>", "\[Rule]", 
                    RowBox[{"-", "0.17816537094091434`"}]}], ",", 
                    RowBox[{"\<\"initialized\"\>", "\[Rule]", "True"}]}], 
                    "\[RightAssociation]"}]}
                    },
                    AutoDelete->False,
                    
                    GridBoxItemSize->{
                    "Columns" -> {{Automatic}}, "Rows" -> {{Automatic}}}],
                    "Grid"], {-1, -0.5}], 
                  {GrayLevel[0], LineBox[{{-5, 0}, {5, 0}}]}, 
                  {GrayLevel[0.5], 
                   CircleBox[{-1.7210260314483978`, 0.5}, 0.5], 
                   DiskBox[{-1.7210260314483978`, 0.5}, 0.1]}, 
                  {GrayLevel[0.5], 
                   DiskBox[{-1.5972891022559141`, 1.994887678842131}, 0.1]}, 
                  {GrayLevel[0.5], Thickness[0.01], 
                   LineBox[{{-1.7210260314483978`, 
                    0.5}, {-1.5972891022559141`, 1.994887678842131}}]}, 
                  {RGBColor[0, 1, 0], 
                   DiskBox[{-1.6385347453200754`, 1.4965917858947542`}, 
                    0.05]}, 
                  {RGBColor[0, 0, 1], Arrowheads[0.01], 
                   CircleBox[{-1.7210260314483978`, 0.5}, 0.3333333333333333, 
                    NCache[{
                    Rational[-1, 2] 
                    Pi, -2.7515161697164867`}, {-1.5707963267948966`, \
-2.7515161697164867`}]], 
                   ArrowBox[{{-2.028036433810316, 
                    0.370170404172684}, {-2.029319357738234, 
                    0.3732469484538544}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-1.7210260314483978`, 
                    0.5}, {-0.9338794695006712, 0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-1.7210260314483978`, 
                    0.5}, {-2.3256805700133314`, 0.5}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.005], Arrowheads[0.04], 
                   ArrowBox[{{-1.7210260314483978`, 
                    0.5}, {-1.5385340080656047`, 0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-1.7210260314483978`, 0.5}, 1.6, {
                    1.4882111967296987`, 1.1090097102717391`}], 
                   ArrowBox[{{-1.0225082830929373`, 
                    1.939469678469268}, {-1.0081487518150278`, 
                    1.9324126445206096`}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Dashing[{Small, Small}], 
                   Arrowheads[0.02], 
                   CircleBox[{-1.7210260314483978`, 0.5}, 1.7, {
                    1.4882111967296987`, 1.5936045696800467`}], 
                   ArrowBox[{{-1.742799449016293, 
                    2.1998605584833757`}, {-1.7597966826306441`, 
                    2.19955783561693}}]}, 
                  {RGBColor[0, 0, 1], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-1.7210260314483978`, 0.5}, 1.8, {
                    1.4882111967296987`, 1.7068630194929562`}], 
                   ArrowBox[{{-1.9473454895193842`, 
                    2.2857154036683602`}, {-1.965191030059718, 
                    2.2833629617812337`}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.003], Arrowheads[0.025], 
                   CircleBox[{-1.7210260314483978`, 0.5}, 1.5, {
                    1.4882111967296987`, 1.4655576211675116`}], 
                   ArrowBox[{{-1.578383838636528, 
                    1.9932023321806798`}, {-1.563459196230739, 
                    1.9917012745316975`}}]}, 
                  {RGBColor[0, 0, 1], ArrowBox[{{0, 2.8}, {1.5, 2.8}}]}},
                 Frame->True,
                 FrameTicks->Automatic,
                 ImageSize->800,
                 PlotRange->{{-3, 3}, {-1, 3}}], 76->
                GraphicsBox[{InsetBox[
                   TagBox[GridBox[{
                    {"\<\"t\"\>", "3.75`"},
                    {"\<\"driveIntent\"\>", "0.5`"},
                    {"\<\"motorTorque\"\>", 
                    RowBox[{"-", "35.09502707749889`"}]},
                    {"\<\"vw\"\>", "0.3219954726808846`"},
                    {"\<\"controllerState\"\>", 
                    RowBox[{"\[LeftAssociation]", 
                    RowBox[{
                    RowBox[{"\<\"lastError\"\>", "\[Rule]", 
                    RowBox[{"-", "0.020637482381364133`"}]}], ",", 
                    RowBox[{"\<\"lastInt\"\>", "\[Rule]", 
                    RowBox[{"-", "0.1793877293166757`"}]}], ",", 
                    RowBox[{"\<\"initialized\"\>", "\[Rule]", "True"}]}], 
                    "\[RightAssociation]"}]}
                    },
                    AutoDelete->False,
                    
                    GridBoxItemSize->{
                    "Columns" -> {{Automatic}}, "Rows" -> {{Automatic}}}],
                    "Grid"], {-1, -0.5}], 
                  {GrayLevel[0], LineBox[{{-5, 0}, {5, 0}}]}, 
                  {GrayLevel[0.5], 
                   CircleBox[{-1.7049940528226606`, 0.5}, 0.5], 
                   DiskBox[{-1.7049940528226606`, 0.5}, 0.1]}, 
                  {GrayLevel[0.5], 
                   DiskBox[{-1.594264734273875, 1.9959074229422493`}, 0.1]}, 
                  {GrayLevel[0.5], Thickness[0.01], 
                   LineBox[{{-1.7049940528226606`, 0.5}, {-1.594264734273875, 
                    1.9959074229422493`}}]}, 
                  {RGBColor[0, 1, 0], 
                   DiskBox[{-1.63117450712347, 1.4972716152948329`}, 0.05]}, 
                  {RGBColor[0, 0, 1], Arrowheads[0.01], 
                   CircleBox[{-1.7049940528226606`, 0.5}, 0.3333333333333333, 
                    NCache[{
                    Rational[-1, 2] 
                    Pi, -2.6236471391198632`}, {-1.5707963267948966`, \
-2.6236471391198632`}]], 
                   ArrowBox[{{-1.992942006090457, 
                    0.332080116364982}, {-1.9946067796626143`, 
                    0.33496794383078965`}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-1.7049940528226606`, 
                    0.5}, {-1.0030935112726826`, 0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-1.7049940528226606`, 
                    0.5}, {-2.2464545344950935`, 0.5}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.005], Arrowheads[0.04], 
                   ArrowBox[{{-1.7049940528226606`, 
                    0.5}, {-1.5445539929451158`, 0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-1.7049940528226606`, 0.5}, 1.6, {
                    1.4969095717009409`, 1.1575709242801753`}], 
                   ArrowBox[{{-1.0771746864177059`, 
                    1.9716802788536922`}, {-1.0624895196147186`, 
                    1.965328606424936}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Dashing[{Small, Small}], 
                   Arrowheads[0.02], 
                   CircleBox[{-1.7049940528226606`, 0.5}, 1.7, {
                    1.4969095717009409`, 1.5896320374405648`}], 
                   ArrowBox[{{-1.7200145654771568`, 
                    2.1999336411164983`}, {-1.7370128675480907`, 
                    2.1996984425196073`}}]}, 
                  {RGBColor[0, 0, 1], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-1.7049940528226606`, 0.5}, 1.8, {
                    1.4969095717009409`, 1.6918819443537125`}], 
                   ArrowBox[{{-1.9045371786519816`, 
                    2.288905403014442}, {-1.9224159574595676`, 
                    2.286820560488394}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.003], Arrowheads[0.025], 
                   CircleBox[{-1.7049940528226606`, 0.5}, 1.5, {
                    1.4969095717009409`, 1.4740006561403574`}], 
                   ArrowBox[{{-1.5749639538229494`, 
                    1.9943534298666181`}, {-1.560027170032714, 
                    1.9929784334993477`}}]}, 
                  {RGBColor[0, 0, 1], ArrowBox[{{0, 2.8}, {1.5, 2.8}}]}},
                 Frame->True,
                 FrameTicks->Automatic,
                 ImageSize->800,
                 PlotRange->{{-3, 3}, {-1, 3}}], 77->
                GraphicsBox[{InsetBox[
                   TagBox[GridBox[{
                    {"\<\"t\"\>", "3.8000000000000003`"},
                    {"\<\"driveIntent\"\>", "0.5`"},
                    {"\<\"motorTorque\"\>", 
                    RowBox[{"-", "31.136665856120842`"}]},
                    {"\<\"vw\"\>", "0.3244734668624664`"},
                    {"\<\"controllerState\"\>", 
                    RowBox[{"\[LeftAssociation]", 
                    RowBox[{
                    RowBox[{"\<\"lastError\"\>", "\[Rule]", 
                    RowBox[{"-", "0.013282724119427908`"}]}], ",", 
                    RowBox[{"\<\"lastInt\"\>", "\[Rule]", 
                    RowBox[{"-", "0.1802299289407926`"}]}], ",", 
                    RowBox[{"\<\"initialized\"\>", "\[Rule]", "True"}]}], 
                    "\[RightAssociation]"}]}
                    },
                    AutoDelete->False,
                    
                    GridBoxItemSize->{
                    "Columns" -> {{Automatic}}, "Rows" -> {{Automatic}}}],
                    "Grid"], {-1, -0.5}], 
                  {GrayLevel[0], LineBox[{{-5, 0}, {5, 0}}]}, 
                  {GrayLevel[0.5], 
                   CircleBox[{-1.6888296898570145`, 0.5}, 0.5], 
                   DiskBox[{-1.6888296898570145`, 0.5}, 0.1]}, 
                  {GrayLevel[0.5], 
                   DiskBox[{-1.590202666132517, 1.9967540580172973`}, 0.1]}, 
                  {GrayLevel[0.5], Thickness[0.01], 
                   LineBox[{{-1.6888296898570145`, 0.5}, {-1.590202666132517, 
                    1.9967540580172973`}}]}, 
                  {RGBColor[0, 1, 0], 
                   DiskBox[{-1.6230783407073495`, 1.4978360386781984`}, 
                    0.05]}, 
                  {RGBColor[0, 0, 1], Arrowheads[0.01], 
                   CircleBox[{-1.6888296898570145`, 0.5}, 0.3333333333333333, 
                    NCache[{
                    Rational[-1, 2] 
                    Pi, -2.504896302478522}, {-1.5707963267948966`, \
-2.504896302478522}]], 
                   ArrowBox[{{-1.9548559478721808`, 
                    0.29914895778821726`}, {-1.9568511236172355`, 
                    0.3018192184993038}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-1.6888296898570145`, 
                    0.5}, {-1.0660963727345978`, 0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-1.6888296898570145`, 0.5}, {-2.171383548505937,
                     0.5}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.005], Arrowheads[0.04], 
                   ArrowBox[{{-1.6888296898570145`, 
                    0.5}, {-1.5486502313835204`, 0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-1.6888296898570145`, 0.5}, 1.6, {
                    1.5049975087622203`, 1.2027472893222693`}], 
                   ArrowBox[{{-1.128113426851601, 
                    1.9985317055041725`}, {-1.113156395130114, 
                    1.9928497103654725`}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Dashing[{Small, Small}], 
                   Arrowheads[0.02], 
                   CircleBox[{-1.6888296898570145`, 0.5}, 1.7, {
                    1.5049975087622203`, 1.5860577182828286`}], 
                   ArrowBox[{{-1.6977740141198765`, 
                    2.199976470149949}, {-1.7147730482808947`, 
                    2.1998020302828474`}}]}, 
                  {RGBColor[0, 0, 1], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-1.6888296898570145`, 0.5}, 1.8, {
                    1.5049975087622203`, 1.6779789857406695`}], 
                   ArrowBox[{{-1.8634832543689008`, 
                    2.2915066654643774`}, {-1.8813892898351399`, 
                    2.2896705843412257`}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.003], Arrowheads[0.025], 
                   CircleBox[{-1.6888296898570145`, 0.5}, 1.5, {
                    1.5049975087622203`, 1.4820432845237679`}], 
                   ArrowBox[{{-1.5708221960025408`, 
                    1.9953508723353814`}, {-1.5558748368286093`, 
                    1.994096050144098}}]}, 
                  {RGBColor[0, 0, 1], ArrowBox[{{0, 2.8}, {1.5, 2.8}}]}},
                 Frame->True,
                 FrameTicks->Automatic,
                 ImageSize->800,
                 PlotRange->{{-3, 3}, {-1, 3}}], 78->
                GraphicsBox[{InsetBox[
                   TagBox[GridBox[{
                    {"\<\"t\"\>", "3.85`"},
                    {"\<\"driveIntent\"\>", "0.5`"},
                    {"\<\"motorTorque\"\>", 
                    RowBox[{"-", "27.48343411638743`"}]},
                    {"\<\"vw\"\>", "0.32663018764271873`"},
                    {"\<\"controllerState\"\>", 
                    RowBox[{"\[LeftAssociation]", 
                    RowBox[{
                    RowBox[{"\<\"lastError\"\>", "\[Rule]", 
                    RowBox[{"-", "0.006443118437733875`"}]}], ",", 
                    RowBox[{"\<\"lastInt\"\>", "\[Rule]", 
                    RowBox[{"-", "0.18071749512510257`"}]}], ",", 
                    RowBox[{"\<\"initialized\"\>", "\[Rule]", "True"}]}], 
                    "\[RightAssociation]"}]}
                    },
                    AutoDelete->False,
                    
                    GridBoxItemSize->{
                    "Columns" -> {{Automatic}}, "Rows" -> {{Automatic}}}],
                    "Grid"], {-1, -0.5}], 
                  {GrayLevel[0], LineBox[{{-5, 0}, {5, 0}}]}, 
                  {GrayLevel[0.5], 
                   CircleBox[{-1.6725497439328953`, 0.5}, 0.5], 
                   DiskBox[{-1.6725497439328953`, 0.5}, 0.1]}, 
                  {GrayLevel[0.5], 
                   DiskBox[{-1.5851154799987728`, 1.997449581618793}, 0.1]}, 
                  {GrayLevel[0.5], Thickness[0.01], 
                   LineBox[{{-1.6725497439328953`, 
                    0.5}, {-1.5851154799987728`, 1.997449581618793}}]}, 
                  {RGBColor[0, 1, 0], 
                   DiskBox[{-1.6142602346434802`, 1.4982997210791953`}, 
                    0.05]}, 
                  {RGBColor[0, 0, 1], Arrowheads[0.01], 
                   CircleBox[{-1.6725497439328953`, 0.5}, 0.3333333333333333, 
                    NCache[{
                    Rational[-1, 2] 
                    Pi, -2.3952993502865194`}, {-1.5707963267948966`, \
-2.3952993502865194`}]], 
                   ArrowBox[{{-1.9150112933546644`, 
                    0.2712566762436418}, {-1.9172865654920856`, 
                    0.27369268839868166`}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-1.6725497439328953`, 
                    0.5}, {-1.1228810616051468`, 0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-1.6725497439328953`, 0.5}, {-2.100539414918388,
                     0.5}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.005], Arrowheads[0.04], 
                   ArrowBox[{{-1.6725497439328953`, 
                    0.5}, {-1.5508707325906397`, 0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-1.6725497439328953`, 0.5}, 1.6, {
                    1.5124737588795185`, 1.2445246262515302`}], 
                   ArrowBox[{{-1.174909186089089, 
                    2.020642586273614}, {-1.1597278954860593`, 
                    2.0155902321391395`}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Dashing[{Small, Small}], 
                   Arrowheads[0.02], 
                   CircleBox[{-1.6725497439328953`, 0.5}, 1.7, {
                    1.5124737588795185`, 1.5828693178871809`}], 
                   ArrowBox[{{-1.6760738262657755`, 
                    2.1999963473030495`}, {-1.6930733302048497`, 
                    2.1998761079580293`}}]}, 
                  {RGBColor[0, 0, 1], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-1.6725497439328953`, 0.5}, 1.8, {
                    1.5124737588795185`, 1.6651595039705598`}], 
                   ArrowBox[{{-1.8242233994317427`, 
                    2.2935983670341633`}, {-1.8421515005509401`, 
                    2.2919919765869694`}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.003], Arrowheads[0.025], 
                   CircleBox[{-1.6725497439328953`, 0.5}, 1.5, {
                    1.5124737588795185`, 1.4896658783942645`}], 
                   ArrowBox[{{-1.5659440204233666`, 
                    1.9962069441474366`}, {-1.550987530590226, 
                    1.995066094956083}}]}, 
                  {RGBColor[0, 0, 1], ArrowBox[{{0, 2.8}, {1.5, 2.8}}]}},
                 Frame->True,
                 FrameTicks->Automatic,
                 ImageSize->800,
                 PlotRange->{{-3, 3}, {-1, 3}}], 79->
                GraphicsBox[{InsetBox[
                   TagBox[GridBox[{
                    {"\<\"t\"\>", "3.9`"},
                    {"\<\"driveIntent\"\>", "0.5`"},
                    {"\<\"motorTorque\"\>", 
                    RowBox[{"-", "24.13375067738646`"}]},
                    {"\<\"vw\"\>", "0.3284950297623843`"},
                    {"\<\"controllerState\"\>", 
                    RowBox[{"\[LeftAssociation]", 
                    RowBox[{
                    RowBox[{"\<\"lastError\"\>", "\[Rule]", 
                    RowBox[{"-", "0.00012372447004326687`"}]}], ",", 
                    RowBox[{"\<\"lastInt\"\>", "\[Rule]", 
                    RowBox[{"-", "0.18087633594582853`"}]}], ",", 
                    RowBox[{"\<\"initialized\"\>", "\[Rule]", "True"}]}], 
                    "\[RightAssociation]"}]}
                    },
                    AutoDelete->False,
                    
                    GridBoxItemSize->{
                    "Columns" -> {{Automatic}}, "Rows" -> {{Automatic}}}],
                    "Grid"], {-1, -0.5}], 
                  {GrayLevel[0], LineBox[{{-5, 0}, {5, 0}}]}, 
                  {GrayLevel[0.5], 
                   CircleBox[{-1.6561695254019067`, 0.5}, 0.5], 
                   DiskBox[{-1.6561695254019067`, 0.5}, 0.1]}, 
                  {GrayLevel[0.5], 
                   DiskBox[{-1.5790223635837486`, 1.9980147914568143`}, 0.1]}, 
                  {GrayLevel[0.5], Thickness[0.01], 
                   LineBox[{{-1.6561695254019067`, 
                    0.5}, {-1.5790223635837486`, 1.9980147914568143`}}]}, 
                  {RGBColor[0, 1, 0], 
                   DiskBox[{-1.6047380841898011`, 1.4986765276378762`}, 
                    0.05]}, 
                  {RGBColor[0, 0, 1], Arrowheads[0.01], 
                   CircleBox[{-1.6561695254019067`, 0.5}, 0.3333333333333333, 
                    NCache[{
                    Rational[-1, 2] 
                    Pi, -2.2948088471164905`}, {-1.5707963267948966`, \
-2.2948088471164905`}]], 
                   ArrowBox[{{-1.874460011380505, 
                    0.2480865727232101}, {-1.8769681892345667`, 
                    0.25028203676783056`}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-1.6561695254019067`, 
                    0.5}, {-1.1734945118541775`, 0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-1.6561695254019067`, 
                    0.5}, {-2.0339464926655584`, 0.5}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.005], Arrowheads[0.04], 
                   ArrowBox[{{-1.6561695254019067`, 0.5}, {-1.551271479117829,
                     0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-1.6561695254019067`, 0.5}, 1.6, {
                    1.519342184200813, 1.2829186796870504`}], 
                   ArrowBox[{{-1.2172649991604427`, 
                    2.0386236761615093`}, {-1.2019009638782605`, 
                    2.0341577735067657`}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Dashing[{Small, Small}], 
                   Arrowheads[0.02], 
                   CircleBox[{-1.6561695254019067`, 0.5}, 1.7, {
                    1.519342184200813, 1.5800527835667033`}], 
                   ArrowBox[{{-1.6549055020304486`, 
                    2.1999995300719695`}, {-1.6719052771999718`, 
                    2.1999271708268413`}}]}, 
                  {RGBColor[0, 0, 1], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-1.6561695254019067`, 0.5}, 1.8, {
                    1.519342184200813, 1.6534185768529601`}], 
                   ArrowBox[{{-1.7867747030700212`, 
                    2.2952554936738894`}, {-1.8047204285945424`, 
                    2.293859701637966}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.003], Arrowheads[0.025], 
                   CircleBox[{-1.6561695254019067`, 0.5}, 1.5, {
                    1.519342184200813, 1.4968541298500515`}], 
                   ArrowBox[{{-1.5603215752140118`, 
                    1.9969345912379668`}, {-1.5453572711470562`, 
                    1.9959012816048252`}}]}, 
                  {RGBColor[0, 0, 1], ArrowBox[{{0, 2.8}, {1.5, 2.8}}]}},
                 Frame->True,
                 FrameTicks->Automatic,
                 ImageSize->800,
                 PlotRange->{{-3, 3}, {-1, 3}}], 80->
                GraphicsBox[{InsetBox[
                   TagBox[GridBox[{
                    {"\<\"t\"\>", "3.95`"},
                    {"\<\"driveIntent\"\>", "0.5`"},
                    {"\<\"motorTorque\"\>", 
                    RowBox[{"-", "21.083525455352625`"}]},
                    {"\<\"vw\"\>", "0.3300964400619083`"},
                    {"\<\"controllerState\"\>", 
                    RowBox[{"\[LeftAssociation]", 
                    RowBox[{
                    
                    RowBox[{"\<\"lastError\"\>", "\[Rule]", 
                    "0.0056753185688431865`"}], ",", 
                    RowBox[{"\<\"lastInt\"\>", "\[Rule]", 
                    RowBox[{"-", "0.18073248508321613`"}]}], ",", 
                    RowBox[{"\<\"initialized\"\>", "\[Rule]", "True"}]}], 
                    "\[RightAssociation]"}]}
                    },
                    AutoDelete->False,
                    GridBoxItemSize->{
                    "Columns" -> {{Automatic}}, "Rows" -> {{Automatic}}}],
                    "Grid"], {-1, -0.5}], 
                  {GrayLevel[0], LineBox[{{-5, 0}, {5, 0}}]}, 
                  {GrayLevel[0.5], 
                   CircleBox[{-1.6397028986618216`, 0.5}, 0.5], 
                   DiskBox[{-1.6397028986618216`, 0.5}, 0.1]}, 
                  {GrayLevel[0.5], 
                   DiskBox[{-1.5719483795591789`, 1.9984689937203137`}, 0.1]}, 
                  {GrayLevel[0.5], Thickness[0.01], 
                   LineBox[{{-1.6397028986618216`, 
                    0.5}, {-1.5719483795591789`, 1.9984689937203137`}}]}, 
                  {RGBColor[0, 1, 0], 
                   DiskBox[{-1.5945332192600596`, 1.4989793291468758`}, 
                    0.05]}, 
                  {RGBColor[0, 0, 1], Arrowheads[0.01], 
                   CircleBox[{-1.6397028986618216`, 0.5}, 0.3333333333333333, 
                    NCache[{
                    Rational[-1, 2] 
                    Pi, -2.2033020904554754`}, {-1.5707963267948966`, \
-2.2033020904554754`}]], 
                   ArrowBox[{{-1.8340604724445462`, 
                    0.22919334457143814`}, {-1.8367587760669077`, 
                    0.23115042813643405`}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-1.6397028986618216`, 
                    0.5}, {-1.2180323895547691`, 0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-1.6397028986618216`, 
                    0.5}, {-1.9715862362924077`, 0.5}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.005], Arrowheads[0.04], 
                   ArrowBox[{{-1.6397028986618216`, 
                    0.5}, {-1.5499157271853552`, 0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-1.6397028986618216`, 0.5}, 1.6, {
                    1.5256112733244693`, 1.3179722562126155`}], 
                   ArrowBox[{{-1.254991215625017, 
                    2.0530605013762306`}, {-1.239480104877233, 
                    2.0491357962861936`}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Dashing[{Small, Small}], 
                   Arrowheads[0.02], 
                   CircleBox[{-1.6397028986618216`, 0.5}, 1.7, {
                    1.5256112733244693`, 1.5775924363935627`}], 
                   ArrowBox[{{-1.634256294297719, 
                    2.1999912748308157`}, {-1.6512561960435135`, 
                    2.1999607411112794`}}]}, 
                  {RGBColor[0, 0, 1], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-1.6397028986618216`, 0.5}, 1.8, {
                    1.5256112733244693`, 1.6427419702986505`}], 
                   ArrowBox[{{-1.751133760136102, 
                    2.2965475677284752`}, {-1.7690933648936447`, 
                    2.295343451055622}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.003], Arrowheads[0.025], 
                   CircleBox[{-1.6397028986618216`, 0.5}, 1.5, {
                    1.5256112733244693`, 1.503598476252482}], 
                   ArrowBox[{{-1.5539528972338912`, 
                    1.997546973305048}, {-1.5389819645550973`, 
                    1.9966146108576746`}}]}, 
                  {RGBColor[0, 0, 1], ArrowBox[{{0, 2.8}, {1.5, 2.8}}]}},
                 Frame->True,
                 FrameTicks->Automatic,
                 ImageSize->800,
                 PlotRange->{{-3, 3}, {-1, 3}}], 81->
                GraphicsBox[{InsetBox[
                   TagBox[GridBox[{
                    {"\<\"t\"\>", "4.`"},
                    {"\<\"driveIntent\"\>", "0.5`"},
                    {"\<\"motorTorque\"\>", 
                    RowBox[{"-", "18.32642768595989`"}]},
                    {"\<\"vw\"\>", "0.3314618268772098`"},
                    {"\<\"controllerState\"\>", 
                    RowBox[{"\[LeftAssociation]", 
                    RowBox[{
                    
                    RowBox[{"\<\"lastError\"\>", "\[Rule]", 
                    "0.010958337180429684`"}], ",", 
                    RowBox[{"\<\"lastInt\"\>", "\[Rule]", 
                    RowBox[{"-", "0.1803118674135035`"}]}], ",", 
                    RowBox[{"\<\"initialized\"\>", "\[Rule]", "True"}]}], 
                    "\[RightAssociation]"}]}
                    },
                    AutoDelete->False,
                    
                    GridBoxItemSize->{
                    "Columns" -> {{Automatic}}, "Rows" -> {{Automatic}}}],
                    "Grid"], {-1, -0.5}], 
                  {GrayLevel[0], LineBox[{{-5, 0}, {5, 0}}]}, 
                  {GrayLevel[0.5], 
                   CircleBox[{-1.6231623318845996`, 0.5}, 0.5], 
                   DiskBox[{-1.6231623318845996`, 0.5}, 0.1]}, 
                  {GrayLevel[0.5], 
                   DiskBox[{-1.5639237409949054`, 1.9988298066656547`}, 0.1]}, 
                  {GrayLevel[0.5], Thickness[0.01], 
                   LineBox[{{-1.6231623318845996`, 
                    0.5}, {-1.5639237409949054`, 1.9988298066656547`}}]}, 
                  {RGBColor[0, 1, 0], 
                   DiskBox[{-1.5836699379581367`, 1.4992198711104363`}, 
                    0.05]}, 
                  {RGBColor[0, 0, 1], Arrowheads[0.01], 
                   CircleBox[{-1.6231623318845996`, 0.5}, 0.3333333333333333, 
                    NCache[{
                    Rational[-1, 2] 
                    Pi, -2.1205891573736935`}, {-1.5707963267948966`, \
-2.1205891573736935`}]], 
                   ArrowBox[{{-1.7944817617546531`, 
                    0.21406160792906648`}, {-1.7973325321190914`, 
                    0.21578907047513435`}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-1.6231623318845996`, 
                    0.5}, {-1.2566337781654018`, 0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-1.6231623318845996`, 
                    0.5}, {-1.9134017945910995`, 0.5}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.005], Arrowheads[0.04], 
                   ArrowBox[{{-1.6231623318845996`, 0.5}, {-1.546873240871902,
                     0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-1.6231623318845996`, 0.5}, 1.6, {
                    1.53129365994354, 1.3497523701471588`}], 
                   ArrowBox[{{-1.2879930391786316`, 
                    2.0645004139427323`}, {-1.2723650541129514`, 
                    2.0610705525081183`}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Dashing[{Small, Small}], 
                   Arrowheads[0.02], 
                   CircleBox[{-1.6231623318845996`, 0.5}, 1.7, {
                    1.53129365994354, 1.575471148443491}], 
                   ArrowBox[{{-1.6141095714729972`, 
                    2.199975896161157}, {-1.631109499740958, 
                    2.1999814241700006`}}]}, 
                  {RGBColor[0, 0, 1], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-1.6231623318845996`, 0.5}, 1.8, {
                    1.53129365994354, 1.6331071470877616`}], 
                   ArrowBox[{{-1.7172788709443427`, 
                    2.2975377818214047`}, {-1.7352492433866864`, 
                    2.296506755976699}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.003], Arrowheads[0.025], 
                   CircleBox[{-1.6231623318845996`, 0.5}, 1.5, {
                    1.53129365994354, 1.509893622367077}], 
                   ArrowBox[{{-1.5468412442838235`, 
                    1.9980571055829062`}, {-1.53186473892551, 
                    1.9972190051959253`}}]}, 
                  {RGBColor[0, 0, 1], ArrowBox[{{0, 2.8}, {1.5, 2.8}}]}},
                 Frame->True,
                 FrameTicks->Automatic,
                 ImageSize->800,
                 PlotRange->{{-3, 3}, {-1, 3}}], 82->
                GraphicsBox[{InsetBox[
                   TagBox[GridBox[{
                    {"\<\"t\"\>", "4.05`"},
                    {"\<\"driveIntent\"\>", "0.5`"},
                    {"\<\"motorTorque\"\>", 
                    RowBox[{"-", "15.854156551822825`"}]},
                    {"\<\"vw\"\>", "0.33261747564199184`"},
                    {"\<\"controllerState\"\>", 
                    RowBox[{"\[LeftAssociation]", 
                    RowBox[{
                    
                    RowBox[{"\<\"lastError\"\>", "\[Rule]", 
                    "0.015733670918957013`"}], ",", 
                    RowBox[{"\<\"lastInt\"\>", "\[Rule]", 
                    RowBox[{"-", "0.17964008727408168`"}]}], ",", 
                    RowBox[{"\<\"initialized\"\>", "\[Rule]", "True"}]}], 
                    "\[RightAssociation]"}]}
                    },
                    AutoDelete->False,
                    
                    GridBoxItemSize->{
                    "Columns" -> {{Automatic}}, "Rows" -> {{Automatic}}}],
                    "Grid"], {-1, -0.5}], 
                  {GrayLevel[0], LineBox[{{-5, 0}, {5, 0}}]}, 
                  {GrayLevel[0.5], 
                   CircleBox[{-1.6065589511229508`, 0.5}, 0.5], 
                   DiskBox[{-1.6065589511229508`, 0.5}, 0.1]}, 
                  {GrayLevel[0.5], 
                   DiskBox[{-1.5549830963110391`, 1.9991130481722919`}, 0.1]}, 
                  {GrayLevel[0.5], Thickness[0.01], 
                   LineBox[{{-1.6065589511229508`, 
                    0.5}, {-1.5549830963110391`, 1.9991130481722919`}}]}, 
                  {RGBColor[0, 1, 0], 
                   DiskBox[{-1.5721750479150098`, 1.499408698781528}, 0.05]}, 
                  {RGBColor[0, 0, 1], Arrowheads[0.01], 
                   CircleBox[{-1.6065589511229508`, 0.5}, 0.3333333333333333, 
                    NCache[{
                    Rational[-1, 2] 
                    Pi, -2.0464210233495814`}, {-1.5707963267948966`, \
-2.0464210233495814`}]], 
                   ArrowBox[{{-1.756219311511052, 
                    0.20215291231973626`}, {-1.7591902497912604`, 
                    0.20366438321063002`}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-1.6065589511229508`, 
                    0.5}, {-1.2894758200864942`, 0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-1.6065589511229508`, 
                    0.5}, {-1.8593025933740974`, 0.5}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.005], Arrowheads[0.04], 
                   ArrowBox[{{-1.6065589511229508`, 
                    0.5}, {-1.5422194623376408`, 0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-1.6065589511229508`, 0.5}, 1.6, {
                    1.536405644902343, 1.378347408467583}], 
                   ArrowBox[{{-1.3162575318746645`, 
                    2.0734437028322406`}, {-1.300537872035652, 
                    2.070462065493543}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Dashing[{Small, Small}], 
                   Arrowheads[0.02], 
                   CircleBox[{-1.6065589511229508`, 0.5}, 1.7, {
                    1.536405644902343, 1.5736705544578498`}], 
                   ArrowBox[{{-1.594445240666147, 
                    2.199956840045937}, {-1.6114451314223583`, 
                    2.1999929779978746`}}]}, 
                  {RGBColor[0, 0, 1], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-1.6065589511229508`, 0.5}, 1.8, {
                    1.536405644902343, 1.6244842924124698`}], 
                   ArrowBox[{{-1.6851722762638879`, 
                    2.2982824986943196`}, {-1.7031508709050782`, 
                    2.297406465169413}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.003], Arrowheads[0.025], 
                   CircleBox[{-1.6065589511229508`, 0.5}, 1.5, {
                    1.536405644902343, 1.5157381338803253`}], 
                   ArrowBox[{{-1.5389945291744813`, 
                    1.998477577038365}, {-1.5240133813420575`, 
                    1.9977270208250726`}}]}, 
                  {RGBColor[0, 0, 1], ArrowBox[{{0, 2.8}, {1.5, 2.8}}]}},
                 Frame->True,
                 FrameTicks->Automatic,
                 ImageSize->800,
                 PlotRange->{{-3, 3}, {-1, 3}}], 83->
                GraphicsBox[{InsetBox[
                   TagBox[GridBox[{
                    {"\<\"t\"\>", "4.1`"},
                    {"\<\"driveIntent\"\>", "0.5`"},
                    {"\<\"motorTorque\"\>", 
                    RowBox[{"-", "13.656710259188891`"}]},
                    {"\<\"vw\"\>", "0.33358847232221034`"},
                    {"\<\"controllerState\"\>", 
                    RowBox[{"\[LeftAssociation]", 
                    RowBox[{
                    
                    RowBox[{"\<\"lastError\"\>", "\[Rule]", 
                    "0.020013222603668248`"}], ",", 
                    RowBox[{"\<\"lastInt\"\>", "\[Rule]", 
                    RowBox[{"-", "0.17874223929702437`"}]}], ",", 
                    RowBox[{"\<\"initialized\"\>", "\[Rule]", "True"}]}], 
                    "\[RightAssociation]"}]}
                    },
                    AutoDelete->False,
                    
                    GridBoxItemSize->{
                    "Columns" -> {{Automatic}}, "Rows" -> {{Automatic}}}],
                    "Grid"], {-1, -0.5}], 
                  {GrayLevel[0], LineBox[{{-5, 0}, {5, 0}}]}, 
                  {GrayLevel[0.5], CircleBox[{-1.589902598443882, 0.5}, 0.5], 
                   DiskBox[{-1.589902598443882, 0.5}, 0.1]}, 
                  {GrayLevel[0.5], 
                   DiskBox[{-1.545164828869872, 1.9993326955594422`}, 0.1]}, 
                  {GrayLevel[0.5], Thickness[0.01], 
                   LineBox[{{-1.589902598443882, 0.5}, {-1.545164828869872, 
                    1.9993326955594422`}}]}, 
                  {RGBColor[0, 1, 0], 
                   DiskBox[{-1.5600774187278752`, 1.4995551303729613`}, 
                    0.05]}, 
                  {RGBColor[0, 0, 1], Arrowheads[0.01], 
                   CircleBox[{-1.589902598443882, 0.5}, 0.3333333333333333, 
                    NCache[{
                    Rational[-1, 2] 
                    Pi, -1.9804976345705634`}, {-1.5707963267948966`, \
-1.9804976345705634`}]], 
                   ArrowBox[{{-1.7196170021719883`, 
                    0.19294091028505134`}, {-1.7226810562267398`, 
                    0.19425338552991805`}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-1.589902598443882, 0.5}, {-1.3167683932601042`,
                     0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-1.589902598443882, 0.5}, {-1.8091688341377972`,
                     0.5}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.005], Arrowheads[0.04], 
                   ArrowBox[{{-1.589902598443882, 0.5}, {-1.5360346289540192`,
                     0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-1.589902598443882, 0.5}, 1.6, {
                    1.5409667235195248`, 1.4038643335217234`}], 
                   ArrowBox[{{-1.3398407718589256`, 
                    2.080338281155334}, {-1.324050155422906, 
                    2.0777586883106616`}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Dashing[{Small, Small}], 
                   Arrowheads[0.02], 
                   CircleBox[{-1.589902598443882, 0.5}, 1.7, {
                    1.5409667235195248`, 1.572171286860588}], 
                   ArrowBox[{{-1.575240212349496, 
                    2.199936767775207}, {-1.5922400298190653`, 
                    2.199998393062348}}]}, 
                  {RGBColor[0, 0, 1], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-1.589902598443882, 0.5}, 1.8, {
                    1.5409667235195248`, 1.6168373360705743`}], 
                   ArrowBox[{{-1.6547623713645152`, 
                    2.2988310676260526`}, {-1.6727471392754756`, 
                    2.2980925399028838`}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.003], Arrowheads[0.025], 
                   CircleBox[{-1.589902598443882, 0.5}, 1.5, {
                    1.5409667235195248`, 1.521134081036334}], 
                   ArrowBox[{{-1.5304248266868354`, 
                    1.998820334351925}, {-1.5154398470092616`, 
                    1.998150626155056}}]}, 
                  {RGBColor[0, 0, 1], ArrowBox[{{0, 2.8}, {1.5, 2.8}}]}},
                 Frame->True,
                 FrameTicks->Automatic,
                 ImageSize->800,
                 PlotRange->{{-3, 3}, {-1, 3}}], 84->
                GraphicsBox[{InsetBox[
                   TagBox[GridBox[{
                    {"\<\"t\"\>", "4.15`"},
                    {"\<\"driveIntent\"\>", "0.5`"},
                    {"\<\"motorTorque\"\>", 
                    RowBox[{"-", "11.722649768552522`"}]},
                    {"\<\"vw\"\>", "0.3343986362559718`"},
                    {"\<\"controllerState\"\>", 
                    RowBox[{"\[LeftAssociation]", 
                    RowBox[{
                    
                    RowBox[{"\<\"lastError\"\>", "\[Rule]", 
                    "0.023812013972644437`"}], ",", 
                    RowBox[{"\<\"lastInt\"\>", "\[Rule]", 
                    RowBox[{"-", "0.17764274160582155`"}]}], ",", 
                    RowBox[{"\<\"initialized\"\>", "\[Rule]", "True"}]}], 
                    "\[RightAssociation]"}]}
                    },
                    AutoDelete->False,
                    
                    GridBoxItemSize->{
                    "Columns" -> {{Automatic}}, "Rows" -> {{Automatic}}}],
                    "Grid"], {-1, -0.5}], 
                  {GrayLevel[0], LineBox[{{-5, 0}, {5, 0}}]}, 
                  {GrayLevel[0.5], 
                   CircleBox[{-1.5732018936568761`, 0.5}, 0.5], 
                   DiskBox[{-1.5732018936568761`, 0.5}, 0.1]}, 
                  {GrayLevel[0.5], 
                   DiskBox[{-1.5345103771568702`, 1.9995009058186426`}, 0.1]}, 
                  {GrayLevel[0.5], Thickness[0.01], 
                   LineBox[{{-1.5732018936568761`, 
                    0.5}, {-1.5345103771568702`, 1.9995009058186426`}}]}, 
                  {RGBColor[0, 1, 0], 
                   DiskBox[{-1.5474075493235389`, 1.4996672705457619`}, 
                    0.05]}, 
                  {RGBColor[0, 0, 1], Arrowheads[0.01], 
                   CircleBox[{-1.5732018936568761`, 0.5}, 0.3333333333333333, 
                    NCache[{
                    Rational[-1, 2] 
                    Pi, -1.9224758198514722`}, {-1.5707963267948966`, \
-1.9224758198514722`}]], 
                   ArrowBox[{{-1.6848918839829987`, 
                    0.18593558435877877`}, {-1.6880268913426244`, 
                    0.18706816873705706`}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-1.5732018936568761`, 
                    0.5}, {-1.3387488982858258`, 0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-1.5732018936568761`, 
                    0.5}, {-1.7628558493387319`, 0.5}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.005], Arrowheads[0.04], 
                   ArrowBox[{{-1.5732018936568761`, 
                    0.5}, {-1.5284028539676815`, 0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-1.5732018936568761`, 0.5}, 1.6, {
                    1.5449991212347125`, 1.4264259461457178`}], 
                   ArrowBox[{{-1.3588556637072635`, 
                    2.085577400730216}, {-1.343010871183727, 
                    2.0833546959455314`}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Dashing[{Small, Small}], 
                   Arrowheads[0.02], 
                   CircleBox[{-1.5732018936568761`, 0.5}, 1.7, {
                    1.5449991212347125`, 1.5709532234102732`}], 
                   ArrowBox[{{-1.5564688881069748`, 
                    2.199917646983308}, {-1.573468617901922, 
                    2.199999979075934}}]}, 
                  {RGBColor[0, 0, 1], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-1.5732018936568761`, 0.5}, 1.8, {
                    1.5449991212347125`, 1.6101249532822264`}], 
                   ArrowBox[{{-1.6259858533928477`, 
                    2.2992259039916556`}, {-1.643975173387286, 
                    2.298608112646054}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.003], Arrowheads[0.025], 
                   CircleBox[{-1.5732018936568761`, 0.5}, 1.5, {
                    1.5449991212347125`, 1.5260867153957707`}], 
                   ArrowBox[{{-1.5211479300913655`, 
                    1.9990965228687312`}, {-1.5061598173873385`, 
                    1.9985010377071712`}}]}, 
                  {RGBColor[0, 0, 1], ArrowBox[{{0, 2.8}, {1.5, 2.8}}]}},
                 Frame->True,
                 FrameTicks->Automatic,
                 ImageSize->800,
                 PlotRange->{{-3, 3}, {-1, 3}}], 85->
                GraphicsBox[{InsetBox[
                   TagBox[GridBox[{
                    {"\<\"t\"\>", "4.2`"},
                    {"\<\"driveIntent\"\>", "0.5`"},
                    {"\<\"motorTorque\"\>", 
                    RowBox[{"-", "10.039353738564635`"}]},
                    {"\<\"vw\"\>", "0.3350704637573324`"},
                    {"\<\"controllerState\"\>", 
                    RowBox[{"\[LeftAssociation]", 
                    RowBox[{
                    
                    RowBox[{"\<\"lastError\"\>", "\[Rule]", 
                    "0.027147750086334226`"}], ",", 
                    RowBox[{"\<\"lastInt\"\>", "\[Rule]", 
                    RowBox[{"-", "0.17636519102105308`"}]}], ",", 
                    RowBox[{"\<\"initialized\"\>", "\[Rule]", "True"}]}], 
                    "\[RightAssociation]"}]}
                    },
                    AutoDelete->False,
                    
                    GridBoxItemSize->{
                    "Columns" -> {{Automatic}}, "Rows" -> {{Automatic}}}],
                    "Grid"], {-1, -0.5}], 
                  {GrayLevel[0], LineBox[{{-5, 0}, {5, 0}}]}, 
                  {GrayLevel[0.5], 
                   CircleBox[{-1.5564642991306594`, 0.5}, 0.5], 
                   DiskBox[{-1.5564642991306594`, 0.5}, 0.1]}, 
                  {GrayLevel[0.5], 
                   DiskBox[{-1.523063581667481, 1.9996280845839562`}, 0.1]}, 
                  {GrayLevel[0.5], Thickness[0.01], 
                   LineBox[{{-1.5564642991306594`, 0.5}, {-1.523063581667481, 
                    1.9996280845839562`}}]}, 
                  {RGBColor[0, 1, 0], 
                   DiskBox[{-1.534197154155207, 1.499752056389304}, 0.05]}, 
                  {RGBColor[0, 0, 1], Arrowheads[0.01], 
                   CircleBox[{-1.5564642991306594`, 0.5}, 0.3333333333333333, 
                    NCache[{
                    Rational[-1, 2] 
                    Pi, -1.8719769389518355`}, {-1.5707963267948966`, \
-1.8719769389518355`}]], 
                   ArrowBox[{{-1.6521587457720726`, 
                    0.1806981303012699}, {-1.6553469265698881`, 
                    0.18167102377913225`}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-1.5564642991306594`, 
                    0.5}, {-1.3556772243593667`, 0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-1.5564642991306594`, 0.5}, {-1.720198265226841,
                     0.5}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.005], Arrowheads[0.04], 
                   ArrowBox[{{-1.5564642991306594`, 
                    0.5}, {-1.5194111904555483`, 0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-1.5564642991306594`, 0.5}, 1.6, {
                    1.5485273413047969`, 1.4461682332078343`}], 
                   ArrowBox[{{-1.3734607302094486`, 
                    2.0894998250273886`}, {-1.3575751469766824`, 
                    2.087590345509657}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Dashing[{Small, Small}], 
                   Arrowheads[0.02], 
                   CircleBox[{-1.5564642991306594`, 0.5}, 1.7, {
                    1.5485273413047969`, 1.5699957376769582`}], 
                   ArrowBox[{{-1.5381036546048927`, 
                    2.199900846147386}, {-1.5551032977755515`, 
                    2.199999455198534}}]}, 
                  {RGBColor[0, 0, 1], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-1.5564642991306594`, 0.5}, 1.8, {
                    1.5485273413047969`, 1.6043015287412672`}], 
                   ArrowBox[{{-1.5987697667931797`, 
                    2.2995027778266572`}, {-1.6167623794000603`, 
                    2.2989897558118066`}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.003], Arrowheads[0.025], 
                   CircleBox[{-1.5564642991306594`, 0.5}, 1.5, {
                    1.5485273413047969`, 1.5306041668135837`}], 
                   ArrowBox[{{-1.5111829393856775`, 
                    1.9993163770397646`}, {-1.4961922895492132`, 
                    1.9987886057950315`}}]}, 
                  {RGBColor[0, 0, 1], ArrowBox[{{0, 2.8}, {1.5, 2.8}}]}},
                 Frame->True,
                 FrameTicks->Automatic,
                 ImageSize->800,
                 PlotRange->{{-3, 3}, {-1, 3}}], 86->
                GraphicsBox[{InsetBox[
                   TagBox[GridBox[{
                    {"\<\"t\"\>", "4.25`"},
                    {"\<\"driveIntent\"\>", "0.5`"},
                    {"\<\"motorTorque\"\>", 
                    RowBox[{"-", "8.59326172659081`"}]},
                    {"\<\"vw\"\>", "0.3356250835223028`"},
                    {"\<\"controllerState\"\>", 
                    RowBox[{"\[LeftAssociation]", 
                    RowBox[{
                    
                    RowBox[{"\<\"lastError\"\>", "\[Rule]", 
                    "0.030040396118172056`"}], ",", 
                    RowBox[{"\<\"lastInt\"\>", "\[Rule]", 
                    RowBox[{"-", "0.17493223974538044`"}]}], ",", 
                    RowBox[{"\<\"initialized\"\>", "\[Rule]", "True"}]}], 
                    "\[RightAssociation]"}]}
                    },
                    AutoDelete->False,
                    
                    GridBoxItemSize->{
                    "Columns" -> {{Automatic}}, "Rows" -> {{Automatic}}}],
                    "Grid"], {-1, -0.5}], 
                  {GrayLevel[0], LineBox[{{-5, 0}, {5, 0}}]}, 
                  {GrayLevel[0.5], 
                   CircleBox[{-1.5396961871323815`, 0.5}, 0.5], 
                   DiskBox[{-1.5396961871323815`, 0.5}, 0.1]}, 
                  {GrayLevel[0.5], 
                   DiskBox[{-1.5108700642787696`, 1.9997229926360496`}, 0.1]}, 
                  {GrayLevel[0.5], Thickness[0.01], 
                   LineBox[{{-1.5396961871323815`, 
                    0.5}, {-1.5108700642787696`, 1.9997229926360496`}}]}, 
                  {RGBColor[0, 1, 0], 
                   DiskBox[{-1.5204787718966402`, 1.4998153284240332`}, 
                    0.05]}, 
                  {RGBColor[0, 0, 1], Arrowheads[0.01], 
                   CircleBox[{-1.5396961871323815`, 0.5}, 0.3333333333333333, 
                    NCache[{
                    Rational[-1, 2] 
                    Pi, -1.828594178592621}, {-1.5707963267948966`, \
-1.828594178592621}]], 
                   ArrowBox[{{-1.6214527445804736`, 
                    0.1768483692980215}, {-1.6246801192353504`, 
                    0.17768207869336666`}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-1.5396961871323815`, 
                    0.5}, {-1.3678309526005654`, 0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-1.5396961871323815`, 
                    0.5}, {-1.6810139337822503`, 0.5}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.005], Arrowheads[0.04], 
                   ArrowBox[{{-1.5396961871323815`, 
                    0.5}, {-1.5091486992504342`, 0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-1.5396961871323815`, 0.5}, 1.6, {
                    1.5515777285016448`, 1.4632378229759024`}], 
                   ArrowBox[{{-1.3838500709049464`, 
                    2.0923919078093887`}, {-1.3679342094650524`, 
                    2.090753853689441}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Dashing[{Small, Small}], 
                   Arrowheads[0.02], 
                   CircleBox[{-1.5396961871323815`, 0.5}, 1.7, {
                    1.5515777285016448`, 1.5692779438463744`}], 
                   ArrowBox[{{-1.5201153691013831`, 
                    2.199887229072928}, {-1.5371149371117336`, 
                    2.199998040336615}}]}, 
                  {RGBColor[0, 0, 1], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-1.5396961871323815`, 0.5}, 1.8, {
                    1.5515777285016448`, 1.599318071427149}], 
                   ArrowBox[{{-1.5730334213098662`, 
                    2.2996912593046055`}, {-1.5910283671080503`, 
                    2.299267908705912}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.003], Arrowheads[0.025], 
                   CircleBox[{-1.5396961871323815`, 0.5}, 1.5, {
                    1.5515777285016448`, 1.5346971515088188`}], 
                   ArrowBox[{{-1.5005518685258066`, 
                    1.9994891537856574`}, {-1.4855591841011797`, 
                    1.999022743290708}}]}, 
                  {RGBColor[0, 0, 1], ArrowBox[{{0, 2.8}, {1.5, 2.8}}]}},
                 Frame->True,
                 FrameTicks->Automatic,
                 ImageSize->800,
                 PlotRange->{{-3, 3}, {-1, 3}}], 87->
                GraphicsBox[{InsetBox[
                   TagBox[GridBox[{
                    {"\<\"t\"\>", "4.3`"},
                    {"\<\"driveIntent\"\>", "0.5`"},
                    {"\<\"motorTorque\"\>", 
                    RowBox[{"-", "7.370103244902726`"}]},
                    {"\<\"vw\"\>", "0.3360822244995361`"},
                    {"\<\"controllerState\"\>", 
                    RowBox[{"\[LeftAssociation]", 
                    RowBox[{
                    
                    RowBox[{"\<\"lastError\"\>", "\[Rule]", 
                    "0.03251177016508132`"}], ",", 
                    RowBox[{"\<\"lastInt\"\>", "\[Rule]", 
                    RowBox[{"-", "0.1733654928152562`"}]}], ",", 
                    RowBox[{"\<\"initialized\"\>", "\[Rule]", "True"}]}], 
                    "\[RightAssociation]"}]}
                    },
                    AutoDelete->False,
                    
                    GridBoxItemSize->{
                    "Columns" -> {{Automatic}}, "Rows" -> {{Automatic}}}],
                    "Grid"], {-1, -0.5}], 
                  {GrayLevel[0], LineBox[{{-5, 0}, {5, 0}}]}, 
                  {GrayLevel[0.5], 
                   CircleBox[{-1.5229029090805943`, 0.5}, 0.5], 
                   DiskBox[{-1.5229029090805943`, 0.5}, 0.1]}, 
                  {GrayLevel[0.5], 
                   DiskBox[{-1.4979766451870304`, 1.9997928794898008`}, 0.1]}, 
                  {GrayLevel[0.5], Thickness[0.01], 
                   LineBox[{{-1.5229029090805943`, 
                    0.5}, {-1.4979766451870304`, 1.9997928794898008`}}]}, 
                  {RGBColor[0, 1, 0], 
                   DiskBox[{-1.5062853998182184`, 1.4998619196598673`}, 
                    0.05]}, 
                  {RGBColor[0, 0, 1], Arrowheads[0.01], 
                   CircleBox[{-1.5229029090805943`, 0.5}, 0.3333333333333333, 
                    NCache[{
                    Rational[-1, 2] 
                    Pi, -1.7918994241419783`}, {-1.5707963267948966`, \
-1.7918994241419783`}]], 
                   ArrowBox[{{-1.5927491211156715`, 
                    0.17406654394575855`}, {-1.5960049090727435`, 
                    0.17478129096212974`}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-1.5229029090805943`, 
                    0.5}, {-1.3755008441825398`, 0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-1.5229029090805943`, 
                    0.5}, {-1.6451076055901455`, 0.5}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.005], Arrowheads[0.04], 
                   ArrowBox[{{-1.5229029090805943`, 0.5}, {-1.497705540692091,
                     0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-1.5229029090805943`, 0.5}, 1.6, {
                    1.55417805263983, 1.4777895695736365`}], 
                   ArrowBox[{{-1.3902445587700354`, 
                    2.0944910667961993`}, {-1.3743065467114972`, 
                    2.0930847815137374`}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Dashing[{Small, Small}], 
                   Arrowheads[0.02], 
                   CircleBox[{-1.5229029090805943`, 0.5}, 1.7, {
                    1.55417805263983, 1.5687789289081944`}], 
                   ArrowBox[{{-1.50247382440224, 
                    2.1998772463031573`}, {-1.5194733349995355`, 
                    2.1999965405910746`}}]}, 
                  {RGBColor[0, 0, 1], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-1.5229029090805943`, 0.5}, 1.8, {
                    1.55417805263983, 1.5951230706670674`}], 
                   ArrowBox[{{-1.5486901658662755`, 
                    2.299815273128737}, {-1.566686729277756, 
                    2.2994674148449987`}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.003], Arrowheads[0.025], 
                   CircleBox[{-1.5229029090805943`, 0.5}, 1.5, {
                    1.55417805263983, 1.5383786854393882`}], 
                   ArrowBox[{{-1.4892792634765855`, 
                    1.9996231028015994`}, {-1.474284963552774, 
                    1.999211891419173}}]}, 
                  {RGBColor[0, 0, 1], ArrowBox[{{0, 2.8}, {1.5, 2.8}}]}},
                 Frame->True,
                 FrameTicks->Automatic,
                 ImageSize->800,
                 PlotRange->{{-3, 3}, {-1, 3}}], 88->
                GraphicsBox[{InsetBox[
                   TagBox[GridBox[{
                    {"\<\"t\"\>", "4.3500000000000005`"},
                    {"\<\"driveIntent\"\>", "0.5`"},
                    {"\<\"motorTorque\"\>", 
                    RowBox[{"-", "6.355110849968053`"}]},
                    {"\<\"vw\"\>", "0.3364601964978711`"},
                    {"\<\"controllerState\"\>", 
                    RowBox[{"\[LeftAssociation]", 
                    RowBox[{
                    
                    RowBox[{"\<\"lastError\"\>", "\[Rule]", 
                    "0.034585155451823654`"}], ",", 
                    RowBox[{"\<\"lastInt\"\>", "\[Rule]", 
                    RowBox[{"-", "0.17168542543081522`"}]}], ",", 
                    RowBox[{"\<\"initialized\"\>", "\[Rule]", "True"}]}], 
                    "\[RightAssociation]"}]}
                    },
                    AutoDelete->False,
                    
                    GridBoxItemSize->{
                    "Columns" -> {{Automatic}}, "Rows" -> {{Automatic}}}],
                    "Grid"], {-1, -0.5}], 
                  {GrayLevel[0], LineBox[{{-5, 0}, {5, 0}}]}, 
                  {GrayLevel[0.5], 
                   CircleBox[{-1.5060888660802403`, 0.5}, 0.5], 
                   DiskBox[{-1.5060888660802403`, 0.5}, 0.1]}, 
                  {GrayLevel[0.5], 
                   DiskBox[{-1.4844308015649732`, 1.9998436345971045`}, 0.1]}, 
                  {GrayLevel[0.5], Thickness[0.01], 
                   LineBox[{{-1.5060888660802403`, 
                    0.5}, {-1.4844308015649732`, 1.9998436345971045`}}]}, 
                  {RGBColor[0, 1, 0], 
                   DiskBox[{-1.4916501564033955`, 1.4998957563980695`}, 
                    0.05]}, 
                  {RGBColor[0, 0, 1], Arrowheads[0.01], 
                   CircleBox[{-1.5060888660802403`, 0.5}, 0.3333333333333333, 
                    NCache[{
                    Rational[-1, 2] 
                    Pi, -1.7614496522939382`}, {-1.5707963267948966`, \
-1.7614496522939382`}]], 
                   ArrowBox[{{-1.565979634528139, 
                    0.17209116058600787`}, {-1.5692556737576109`, 
                    0.17270645359408449`}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-1.5060888660802403`, 
                    0.5}, {-1.3789866490808793`, 0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-1.5060888660802403`, 
                    0.5}, {-1.6122743249608475`, 0.5}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.005], Arrowheads[0.04], 
                   ArrowBox[{{-1.5060888660802403`, 
                    0.5}, {-1.4851721079614864`, 0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-1.5060888660802403`, 0.5}, 1.6, {
                    1.5563571153827627`, 1.4899842846394362`}], 
                   ArrowBox[{{-1.3928842618364927`, 
                    2.0959901997123973`}, {-1.3769302860194492`, 
                    2.094778373692307}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Dashing[{Small, Small}], 
                   Arrowheads[0.02], 
                   CircleBox[{-1.5060888660802403`, 0.5}, 1.7, {
                    1.5563571153827627`, 1.5684779669306164`}], 
                   ArrowBox[{{-1.4851481839171359`, 
                    2.19987102093969}, {-1.5021476578414954`, 
                    2.199995431428455}}]}, 
                  {RGBColor[0, 0, 1], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-1.5060888660802403`, 0.5}, 1.8, {
                    1.5563571153827627`, 1.5916632867714742`}], 
                   ArrowBox[{{-1.5256490090538914`, 
                    2.2998937193086846`}, {-1.543646668267193, 
                    2.2996081272029434`}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.003], Arrowheads[0.025], 
                   CircleBox[{-1.5060888660802403`, 0.5}, 1.5, {
                    1.5563571153827627`, 1.541663800010569}], 
                   ArrowBox[{{-1.4773918267541897`, 
                    1.9997254681887346`}, {-1.4623962568653064`, 
                    1.999363516929764}}]}, 
                  {RGBColor[0, 0, 1], ArrowBox[{{0, 2.8}, {1.5, 2.8}}]}},
                 Frame->True,
                 FrameTicks->Automatic,
                 ImageSize->800,
                 PlotRange->{{-3, 3}, {-1, 3}}], 89->
                GraphicsBox[{InsetBox[
                   TagBox[GridBox[{
                    {"\<\"t\"\>", "4.4`"},
                    {"\<\"driveIntent\"\>", "0.5`"},
                    {"\<\"motorTorque\"\>", 
                    RowBox[{"-", "5.533216005060929`"}]},
                    {"\<\"vw\"\>", "0.33677588343008885`"},
                    {"\<\"controllerState\"\>", 
                    RowBox[{"\[LeftAssociation]", 
                    RowBox[{
                    
                    RowBox[{"\<\"lastError\"\>", "\[Rule]", 
                    "0.03628493486555193`"}], ",", 
                    RowBox[{"\<\"lastInt\"\>", "\[Rule]", 
                    RowBox[{"-", "0.16991131911727703`"}]}], ",", 
                    RowBox[{"\<\"initialized\"\>", "\[Rule]", "True"}]}], 
                    "\[RightAssociation]"}]}
                    },
                    AutoDelete->False,
                    
                    GridBoxItemSize->{
                    "Columns" -> {{Automatic}}, "Rows" -> {{Automatic}}}],
                    "Grid"], {-1, -0.5}], 
                  {GrayLevel[0], LineBox[{{-5, 0}, {5, 0}}]}, 
                  {GrayLevel[0.5], 
                   CircleBox[{-1.4892575801038883`, 0.5}, 0.5], 
                   DiskBox[{-1.4892575801038883`, 0.5}, 0.1]}, 
                  {GrayLevel[0.5], 
                   DiskBox[{-1.470280171046863, 1.9998799478443208`}, 0.1]}, 
                  {GrayLevel[0.5], Thickness[0.01], 
                   LineBox[{{-1.4892575801038883`, 0.5}, {-1.470280171046863, 
                    1.9998799478443208`}}]}, 
                  {RGBColor[0, 1, 0], 
                   DiskBox[{-1.4766059740658712`, 1.4999199652295472`}, 
                    0.05]}, 
                  {RGBColor[0, 0, 1], Arrowheads[0.01], 
                   CircleBox[{-1.4892575801038883`, 0.5}, 0.3333333333333333, 
                    NCache[{
                    Rational[-1, 2] 
                    Pi, -1.7367928069467244`}, {-1.5707963267948966`, \
-1.7367928069467244`}]], 
                   ArrowBox[{{-1.541045765557487, 
                    0.17071426547975876`}, {-1.5443359786343136`, 
                    0.17124860285249777`}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-1.4892575801038883`, 
                    0.5}, {-1.3785932600026698`, 0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-1.4892575801038883`, 
                    0.5}, {-1.5823025369406123`, 0.5}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.005], Arrowheads[0.04], 
                   ArrowBox[{{-1.4892575801038883`, 
                    0.5}, {-1.4716382168393936`, 0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-1.4892575801038883`, 0.5}, 1.6, {
                    1.5581443832232789`, 1.4999866305579417`}], 
                   ArrowBox[{{-1.3920220185749255`, 
                    2.097042656153601}, {-1.3760567199233962`, 
                    2.095990465276781}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Dashing[{Small, Small}], 
                   Arrowheads[0.02], 
                   CircleBox[{-1.4892575801038883`, 0.5}, 1.7, {
                    1.5581443832232789`, 1.5683547117327623`}], 
                   ArrowBox[{{-1.468107380161366, 
                    2.1998684269796858`}, {-1.4851068386223594`, 
                    2.1999949327410224`}}]}, 
                  {RGBColor[0, 0, 1], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-1.4892575801038883`, 0.5}, 1.8, {
                    1.5581443832232789`, 1.588884472140284}], 
                   ArrowBox[{{-1.5038160829927862`, 
                    2.2999411240353487`}, {-1.5218144663253739`, 
                    2.2997055451266357`}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.003], Arrowheads[0.025], 
                   CircleBox[{-1.4892575801038883`, 0.5}, 1.5, {
                    1.5581443832232789`, 1.5445692593914067`}], 
                   ArrowBox[{{-1.464918047203173, 
                    1.9998025160460877`}, {-1.449921488975052, 
                    1.9994841352727624`}}]}, 
                  {RGBColor[0, 0, 1], ArrowBox[{{0, 2.8}, {1.5, 2.8}}]}},
                 Frame->True,
                 FrameTicks->Automatic,
                 ImageSize->800,
                 PlotRange->{{-3, 3}, {-1, 3}}], 90->
                GraphicsBox[{InsetBox[
                   TagBox[GridBox[{
                    {"\<\"t\"\>", "4.45`"},
                    {"\<\"driveIntent\"\>", "0.5`"},
                    {"\<\"motorTorque\"\>", 
                    RowBox[{"-", "4.889226975557605`"}]},
                    {"\<\"vw\"\>", "0.3370447487596735`"},
                    {"\<\"controllerState\"\>", 
                    RowBox[{"\[LeftAssociation]", 
                    RowBox[{
                    
                    RowBox[{"\<\"lastError\"\>", "\[Rule]", 
                    "0.03763625020834217`"}], ",", 
                    RowBox[{"\<\"lastInt\"\>", "\[Rule]", 
                    RowBox[{"-", "0.1680612155380246`"}]}], ",", 
                    RowBox[{"\<\"initialized\"\>", "\[Rule]", "True"}]}], 
                    "\[RightAssociation]"}]}
                    },
                    AutoDelete->False,
                    
                    GridBoxItemSize->{
                    "Columns" -> {{Automatic}}, "Rows" -> {{Automatic}}}],
                    "Grid"], {-1, -0.5}], 
                  {GrayLevel[0], LineBox[{{-5, 0}, {5, 0}}]}, 
                  {GrayLevel[0.5], 
                   CircleBox[{-1.4724117651971782`, 0.5}, 0.5], 
                   DiskBox[{-1.4724117651971782`, 0.5}, 0.1]}, 
                  {GrayLevel[0.5], 
                   DiskBox[{-1.4555721020776775`, 1.9999054722701768`}, 0.1]}, 
                  {GrayLevel[0.5], Thickness[0.01], 
                   LineBox[{{-1.4724117651971782`, 
                    0.5}, {-1.4555721020776775`, 1.9999054722701768`}}]}, 
                  {RGBColor[0, 1, 0], 
                   DiskBox[{-1.4611853231175111`, 1.4999369815134511`}, 
                    0.05]}, 
                  {RGBColor[0, 0, 1], Arrowheads[0.01], 
                   CircleBox[{-1.4724117651971782`, 0.5}, 0.3333333333333333, 
                    NCache[{
                    Rational[-1, 2] 
                    Pi, -1.7174731360616247`}, {-1.5707963267948966`, \
-1.7174731360616247`}]], 
                   ArrowBox[{{-1.5178289897139354`, 
                    0.16977524800780674`}, {-1.5211289113543718`, 
                    0.17024592378348113`}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-1.4724117651971782`, 
                    0.5}, {-1.3746272256860261`, 0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-1.4724117651971782`, 0.5}, {-1.55497690290467, 
                    0.5}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.005], Arrowheads[0.04], 
                   ArrowBox[{{-1.4724117651971782`, 0.5}, {-1.457192363393518,
                     0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-1.4724117651971782`, 0.5}, 1.6, {
                    1.5595696488848216`, 1.5079631856564304`}], 
                   ArrowBox[{{-1.3879180606346357`, 
                    2.097767446748524}, {-1.3719448771104161`, 
                    2.0968426360785095`}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Dashing[{Small, Small}], 
                   Arrowheads[0.02], 
                   CircleBox[{-1.4724117651971782`, 0.5}, 1.7, {
                    1.5595696488848216`, 1.5683893657431835`}], 
                   ArrowBox[{{-1.4513204725255269`, 
                    2.1998691588982484`}, {-1.4683199353602452`, 
                    2.1999950755600985`}}]}, 
                  {RGBColor[0, 0, 1], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-1.4724117651971782`, 0.5}, 1.8, {
                    1.5595696488848216`, 1.5867320209712528`}], 
                   ArrowBox[{{-1.4830959519759879`, 
                    2.29996829087428}, {-1.5010948006866283`, 
                    2.299771453122621}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.003], Arrowheads[0.025], 
                   CircleBox[{-1.4724117651971782`, 0.5}, 1.5, {
                    1.5595696488848216`, 1.5471132803843985`}], 
                   ArrowBox[{{-1.4518878360301233`, 
                    1.9998595828715253`}, {-1.4368905163646621`, 
                    1.9995793546462883`}}]}, 
                  {RGBColor[0, 0, 1], ArrowBox[{{0, 2.8}, {1.5, 2.8}}]}},
                 Frame->True,
                 FrameTicks->Automatic,
                 ImageSize->800,
                 PlotRange->{{-3, 3}, {-1, 3}}], 91->
                GraphicsBox[{InsetBox[
                   TagBox[GridBox[{
                    {"\<\"t\"\>", "4.5`"},
                    {"\<\"driveIntent\"\>", "0.5`"},
                    {"\<\"motorTorque\"\>", 
                    RowBox[{"-", "4.407988473689059`"}]},
                    {"\<\"vw\"\>", "0.33728085243930883`"},
                    {"\<\"controllerState\"\>", 
                    RowBox[{"\[LeftAssociation]", 
                    RowBox[{
                    
                    RowBox[{"\<\"lastError\"\>", "\[Rule]", 
                    "0.038664687953674816`"}], ",", 
                    RowBox[{"\<\"lastInt\"\>", "\[Rule]", 
                    RowBox[{"-", "0.166151886675368`"}]}], ",", 
                    RowBox[{"\<\"initialized\"\>", "\[Rule]", "True"}]}], 
                    "\[RightAssociation]"}]}
                    },
                    AutoDelete->False,
                    GridBoxItemSize->{
                    "Columns" -> {{Automatic}}, "Rows" -> {{Automatic}}}],
                    "Grid"], {-1, -0.5}], 
                  {GrayLevel[0], LineBox[{{-5, 0}, {5, 0}}]}, 
                  {GrayLevel[0.5], CircleBox[{-1.455553398115426, 0.5}, 0.5], 
                   DiskBox[{-1.455553398115426, 0.5}, 0.1]}, 
                  {GrayLevel[0.5], 
                   DiskBox[{-1.4403532521318645`, 1.9999229832101642`}, 0.1]}, 
                  {GrayLevel[0.5], Thickness[0.01], 
                   LineBox[{{-1.455553398115426, 0.5}, {-1.4403532521318645`, 
                    1.9999229832101642`}}]}, 
                  {RGBColor[0, 1, 0], 
                   DiskBox[{-1.4454199674597183`, 1.499948655473443}, 0.05]}, 
                  {RGBColor[0, 0, 1], Arrowheads[0.01], 
                   CircleBox[{-1.455553398115426, 0.5}, 0.3333333333333333, 
                    NCache[{
                    Rational[-1, 2] 
                    Pi, -1.7030359810055684`}, {-1.5707963267948966`, \
-1.7030359810055684`}]], 
                   ArrowBox[{{-1.4961985491922773`, 
                    0.16915398928647934`}, {-1.499504921918068, 
                    0.1695769761857735}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-1.455553398115426, 0.5}, {-1.3673936286416448`,
                     0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-1.455553398115426, 0.5}, {-1.5300808271482302`,
                     0.5}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.005], Arrowheads[0.04], 
                   ArrowBox[{{-1.455553398115426, 0.5}, {-1.4419210576744492`,
                     0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-1.455553398115426, 0.5}, 1.6, {
                    1.5606627227035603`, 1.5140806878255308`}], 
                   ArrowBox[{{-1.3808355594428243`, 
                    2.0982544367478213`}, {-1.364857017310555, 
                    2.0974273587581056`}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Dashing[{Small, Small}], 
                   Arrowheads[0.02], 
                   CircleBox[{-1.455553398115426, 0.5}, 1.7, {
                    1.5606627227035603`, 1.5685628241112004`}], 
                   ArrowBox[{{-1.4347569622897363`, 
                    2.19987279178677}, {-1.4517564467100146`, 
                    2.1999957597476603`}}]}, 
                  {RGBColor[0, 0, 1], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-1.455553398115426, 0.5}, 1.8, {
                    1.5606627227035603`, 1.5851515475573885`}], 
                   ArrowBox[{{-1.463392770705055, 
                    2.2999829288738827`}, {-1.4813919080327758`, 
                    2.2998145380580888`}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.003], Arrowheads[0.025], 
                   CircleBox[{-1.455553398115426, 0.5}, 1.5, {
                    1.5606627227035603`, 1.549315256920248}], 
                   ArrowBox[{{-1.4383321716451667`, 
                    1.9999011398618445`}, {-1.4233342712829697`, 
                    1.9996539360352954`}}]}, 
                  {RGBColor[0, 0, 1], ArrowBox[{{0, 2.8}, {1.5, 2.8}}]}},
                 Frame->True,
                 FrameTicks->Automatic,
                 ImageSize->800,
                 PlotRange->{{-3, 3}, {-1, 3}}], 92->
                GraphicsBox[{InsetBox[
                   TagBox[GridBox[{
                    {"\<\"t\"\>", "4.55`"},
                    {"\<\"driveIntent\"\>", "0.5`"},
                    {"\<\"motorTorque\"\>", 
                    RowBox[{"-", "4.074523155925803`"}]},
                    {"\<\"vw\"\>", "0.33749687841314335`"},
                    {"\<\"controllerState\"\>", 
                    RowBox[{"\[LeftAssociation]", 
                    RowBox[{
                    
                    RowBox[{"\<\"lastError\"\>", "\[Rule]", 
                    "0.039395992684350185`"}], ",", 
                    RowBox[{"\<\"lastInt\"\>", "\[Rule]", 
                    RowBox[{"-", "0.1641988200212703`"}]}], ",", 
                    RowBox[{"\<\"initialized\"\>", "\[Rule]", "True"}]}], 
                    "\[RightAssociation]"}]}
                    },
                    AutoDelete->False,
                    
                    GridBoxItemSize->{
                    "Columns" -> {{Automatic}}, "Rows" -> {{Automatic}}}],
                    "Grid"], {-1, -0.5}], 
                  {GrayLevel[0], LineBox[{{-5, 0}, {5, 0}}]}, 
                  {GrayLevel[0.5], 
                   CircleBox[{-1.4386837878396956`, 0.5}, 0.5], 
                   DiskBox[{-1.4386837878396956`, 0.5}, 0.1]}, 
                  {GrayLevel[0.5], 
                   DiskBox[{-1.4246692338670313`, 1.9999345293301796`}, 0.1]}, 
                  {GrayLevel[0.5], Thickness[0.01], 
                   LineBox[{{-1.4386837878396956`, 
                    0.5}, {-1.4246692338670313`, 1.9999345293301796`}}]}, 
                  {RGBColor[0, 1, 0], 
                   DiskBox[{-1.4293407518579193`, 1.4999563528867865`}, 
                    0.05]}, 
                  {RGBColor[0, 0, 1], Arrowheads[0.01], 
                   CircleBox[{-1.4386837878396956`, 0.5}, 0.3333333333333333, 
                    NCache[{
                    Rational[-1, 2] 
                    Pi, -1.6930320214726706`}, {-1.5707963267948966`, \
-1.6930320214726706`}]], 
                   ArrowBox[{{-1.4760171901697312`, 
                    0.16876393888712116`}, {-1.4793276289205648`, 
                    0.16915382835325998`}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-1.4386837878396956`, 
                    0.5}, {-1.3571933247211796`, 0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-1.4386837878396956`, 
                    0.5}, {-1.5073987013591879`, 0.5}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.005], Arrowheads[0.04], 
                   ArrowBox[{{-1.4386837878396956`, 
                    0.5}, {-1.4259082382406718`, 0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-1.4386837878396956`, 0.5}, 1.6, {
                    1.561453154878564, 1.518504459342305}], 
                   ArrowBox[{{-1.371036969663931, 
                    2.0985693316183367`}, {-1.3550549250873603`, 
                    2.0978129469104796`}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Dashing[{Small, Small}], 
                   Arrowheads[0.02], 
                   CircleBox[{-1.4386837878396956`, 0.5}, 1.7, {
                    1.561453154878564, 1.5688567941844078`}], 
                   ArrowBox[{{-1.4183870646344507`, 
                    2.199878831866298}, {-1.435386584469095, 
                    2.199996802482267}}]}, 
                  {RGBColor[0, 0, 1], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-1.4386837878396956`, 0.5}, 1.8, {
                    1.561453154878564, 1.5840893946337073`}], 
                   ArrowBox[{{-1.4446112992362603`, 
                    2.2999902401426082`}, {-1.462610605267713, 
                    2.2998409672545423`}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.003], Arrowheads[0.025], 
                   CircleBox[{-1.4386837878396956`, 0.5}, 1.5, {
                    1.561453154878564, 1.5511954918926827`}], 
                   ArrowBox[{{-1.4242827567270686`, 
                    1.99993086850791}, {-1.4092844180747728`, 
                    1.9997118646784873`}}]}, 
                  {RGBColor[0, 0, 1], ArrowBox[{{0, 2.8}, {1.5, 2.8}}]}},
                 Frame->True,
                 FrameTicks->Automatic,
                 ImageSize->800,
                 PlotRange->{{-3, 3}, {-1, 3}}], 93->
                GraphicsBox[{InsetBox[
                   TagBox[GridBox[{
                    {"\<\"t\"\>", "4.6000000000000005`"},
                    {"\<\"driveIntent\"\>", "0.5`"},
                    {"\<\"motorTorque\"\>", 
                    RowBox[{"-", "3.8741553891155505`"}]},
                    {"\<\"vw\"\>", "0.337704171600365`"},
                    {"\<\"controllerState\"\>", 
                    RowBox[{"\[LeftAssociation]", 
                    RowBox[{
                    
                    RowBox[{"\<\"lastError\"\>", "\[Rule]", 
                    "0.0398558088093089`"}], ",", 
                    RowBox[{"\<\"lastInt\"\>", "\[Rule]", 
                    RowBox[{"-", "0.1622162173764087`"}]}], ",", 
                    RowBox[{"\<\"initialized\"\>", "\[Rule]", "True"}]}], 
                    "\[RightAssociation]"}]}
                    },
                    AutoDelete->False,
                    
                    GridBoxItemSize->{
                    "Columns" -> {{Automatic}}, "Rows" -> {{Automatic}}}],
                    "Grid"], {-1, -0.5}], 
                  {GrayLevel[0], LineBox[{{-5, 0}, {5, 0}}]}, 
                  {GrayLevel[0.5], 
                   CircleBox[{-1.4218036434712318`, 0.5}, 0.5], 
                   DiskBox[{-1.4218036434712318`, 0.5}, 0.1]}, 
                  {GrayLevel[0.5], 
                   DiskBox[{-1.4085643084579673`, 1.9999415721981997`}, 0.1]}, 
                  {GrayLevel[0.5], Thickness[0.01], 
                   LineBox[{{-1.4218036434712318`, 
                    0.5}, {-1.4085643084579673`, 1.9999415721981997`}}]}, 
                  {RGBColor[0, 1, 0], 
                   DiskBox[{-1.4129774201290555`, 1.499961048132133}, 0.05]}, 
                  {RGBColor[0, 0, 1], Arrowheads[0.01], 
                   CircleBox[{-1.4218036434712318`, 0.5}, 0.3333333333333333, 
                    NCache[{
                    Rational[-1, 2] 
                    Pi, -1.6870209884683631`}, {-1.5707963267948966`, \
-1.6870209884683631`}]], 
                   ArrowBox[{{-1.4571453124231992`, 
                    0.16854551210339508`}, {-1.460458034991305, 
                    0.1689154954889553}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-1.4218036434712318`, 
                    0.5}, {-1.3443205356889207`, 0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-1.4218036434712318`, 0.5}, {-1.486717877195997,
                     0.5}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.005], Arrowheads[0.04], 
                   ArrowBox[{{-1.4218036434712318`, 0.5}, {-1.409234769413686,
                     0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-1.4218036434712318`, 0.5}, 1.6, {
                    1.5619699888516396`, 1.5213970123804428`}], 
                   ArrowBox[{{-1.358781048419964, 
                    2.0987583158542145`}, {-1.3427968828233021`, 
                    2.0980481631577694`}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Dashing[{Small, Small}], 
                   Arrowheads[0.02], 
                   CircleBox[{-1.4218036434712318`, 0.5}, 1.7, {
                    1.5619699888516396`, 1.5692538912704457`}], 
                   ArrowBox[{{-1.4021819387792538`, 
                    2.1998867576121004`}, {-1.4191815041193907`, 
                    2.199997977759156}}]}, 
                  {RGBColor[0, 0, 1], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-1.4218036434712318`, 0.5}, 1.8, {
                    1.5619699888516396`, 1.583493074346726}], 
                   ArrowBox[{{-1.4266577831809402`, 
                    2.299993454801344}, {-1.4446571750265818`, 
                    2.299854915290521}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.003], Arrowheads[0.025], 
                   CircleBox[{-1.4218036434712318`, 0.5}, 1.5, {
                    1.5619699888516396`, 1.552774939284447}], 
                   ArrowBox[{{-1.4097716912344898`, 
                    1.9999517432655534`}, {-1.39477302538514, 
                    1.9997564287863159`}}]}, 
                  {RGBColor[0, 0, 1], ArrowBox[{{0, 2.8}, {1.5, 2.8}}]}},
                 Frame->True,
                 FrameTicks->Automatic,
                 ImageSize->800,
                 PlotRange->{{-3, 3}, {-1, 3}}], 94->
                GraphicsBox[{InsetBox[
                   TagBox[GridBox[{
                    {"\<\"t\"\>", "4.65`"},
                    {"\<\"driveIntent\"\>", "0.5`"},
                    {"\<\"motorTorque\"\>", 
                    RowBox[{"-", "3.7926179439234886`"}]},
                    {"\<\"vw\"\>", "0.3379127831817511`"},
                    {"\<\"controllerState\"\>", 
                    RowBox[{"\[LeftAssociation]", 
                    RowBox[{
                    
                    RowBox[{"\<\"lastError\"\>", "\[Rule]", 
                    "0.040069450629696024`"}], ",", 
                    RowBox[{"\<\"lastInt\"\>", "\[Rule]", 
                    RowBox[{"-", "0.16021700583977047`"}]}], ",", 
                    RowBox[{"\<\"initialized\"\>", "\[Rule]", "True"}]}], 
                    "\[RightAssociation]"}]}
                    },
                    AutoDelete->False,
                    
                    GridBoxItemSize->{
                    "Columns" -> {{Automatic}}, "Rows" -> {{Automatic}}}],
                    "Grid"], {-1, -0.5}], 
                  {GrayLevel[0], LineBox[{{-5, 0}, {5, 0}}]}, 
                  {GrayLevel[0.5], CircleBox[{-1.404913140059951, 0.5}, 0.5], 
                   DiskBox[{-1.404913140059951, 0.5}, 0.1]}, 
                  {GrayLevel[0.5], 
                   DiskBox[{-1.392081124671594, 1.999945112122798}, 0.1]}, 
                  {GrayLevel[0.5], Thickness[0.01], 
                   LineBox[{{-1.404913140059951, 0.5}, {-1.392081124671594, 
                    1.999945112122798}}]}, 
                  {RGBColor[0, 1, 0], 
                   DiskBox[{-1.3963584631343797`, 1.4999634080818653`}, 
                    0.05]}, 
                  {RGBColor[0, 0, 1], Arrowheads[0.01], 
                   CircleBox[{-1.404913140059951, 0.5}, 0.3333333333333333, 
                    NCache[{
                    Rational[-1, 2] 
                    Pi, -1.6845748651126011`}, {-1.5707963267948966`, \
-1.6845748651126011`}]], 
                   ArrowBox[{{-1.4394439255225244`, 
                    0.16846005373944933`}, {-1.442757543203863, 
                    0.16882193269814477`}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-1.404913140059951, 0.5}, {-1.3290607811814812`,
                     0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-1.404913140059951, 0.5}, {-1.467830379567342, 
                    0.5}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.005], Arrowheads[0.04], 
                   ArrowBox[{{-1.404913140059951, 0.5}, {-1.3919780206888723`,
                     0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-1.404913140059951, 0.5}, 1.6, {
                    1.5622415455237855`, 1.522916831864963}], 
                   ArrowBox[{{-1.3443204408991318`, 
                    2.09885225233866}, {-1.3283352144594995`, 
                    2.098166393499354}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Dashing[{Small, Small}], 
                   Arrowheads[0.02], 
                   CircleBox[{-1.404913140059951, 0.5}, 1.7, {
                    1.5622415455237855`, 1.5697377121757266`}], 
                   ArrowBox[{{-1.3861138783823799`, 
                    2.1998960520456476`}, {-1.403113495543495, 
                    2.199999047434914}}]}, 
                  {RGBColor[0, 0, 1], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-1.404913140059951, 0.5}, 1.8, {
                    1.5622415455237855`, 1.583311645212249}], 
                   ArrowBox[{{-1.4094407084369918`, 
                    2.2999943058589354`}, {-1.4274401251214979`, 
                    2.2998590319644583`}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.003], Arrowheads[0.025], 
                   CircleBox[{-1.404913140059951, 0.5}, 1.5, {
                    1.5622415455237855`, 1.5540749594527317`}], 
                   ArrowBox[{{-1.3948311649589638`, 
                    1.9999661175433474`}, {-1.3798322578711875`, 
                    1.9997903017917649`}}]}, 
                  {RGBColor[0, 0, 1], ArrowBox[{{0, 2.8}, {1.5, 2.8}}]}},
                 Frame->True,
                 FrameTicks->Automatic,
                 ImageSize->800,
                 PlotRange->{{-3, 3}, {-1, 3}}], 95->
                GraphicsBox[{InsetBox[
                   TagBox[GridBox[{
                    {"\<\"t\"\>", "4.7`"},
                    {"\<\"driveIntent\"\>", "0.5`"},
                    {"\<\"motorTorque\"\>", 
                    RowBox[{"-", "3.816142452556086`"}]},
                    {"\<\"vw\"\>", "0.3381315229669579`"},
                    {"\<\"controllerState\"\>", 
                    RowBox[{"\[LeftAssociation]", 
                    RowBox[{
                    
                    RowBox[{"\<\"lastError\"\>", "\[Rule]", 
                    "0.04006170036518412`"}], ",", 
                    RowBox[{"\<\"lastInt\"\>", "\[Rule]", 
                    RowBox[{"-", "0.15821285957947986`"}]}], ",", 
                    RowBox[{"\<\"initialized\"\>", "\[Rule]", "True"}]}], 
                    "\[RightAssociation]"}]}
                    },
                    AutoDelete->False,
                    
                    GridBoxItemSize->{
                    "Columns" -> {{Automatic}}, "Rows" -> {{Automatic}}}],
                    "Grid"], {-1, -0.5}], 
                  {GrayLevel[0], LineBox[{{-5, 0}, {5, 0}}]}, 
                  {GrayLevel[0.5], CircleBox[{-1.388011981982932, 0.5}, 0.5], 
                   DiskBox[{-1.388011981982932, 0.5}, 0.1]}, 
                  {GrayLevel[0.5], 
                   DiskBox[{-1.3752605016965669`, 1.9999457989375837`}, 0.1]}, 
                  {GrayLevel[0.5], Thickness[0.01], 
                   LineBox[{{-1.388011981982932, 0.5}, {-1.3752605016965669`, 
                    1.9999457989375837`}}]}, 
                  {RGBColor[0, 1, 0], 
                   DiskBox[{-1.3795109951253552`, 1.4999638659583892`}, 
                    0.05]}, 
                  {RGBColor[0, 0, 1], Arrowheads[0.01], 
                   CircleBox[{-1.388011981982932, 0.5}, 0.3333333333333333, 
                    NCache[{
                    Rational[-1, 2] 
                    Pi, -1.6852806003715792`}, {-1.5707963267948966`, \
-1.6852806003715792`}]], 
                   ArrowBox[{{-1.4227767382566783`, 
                    0.16848450589401087`}, {-1.4260900997221038`, 
                    0.16884872329922557`}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-1.388011981982932, 0.5}, {-1.3116891329318103`,
                     0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-1.388011981982932, 0.5}, {-1.4505343747896937`,
                     0.5}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.005], Arrowheads[0.04], 
                   ArrowBox[{{-1.388011981982932, 0.5}, {-1.3742115257385719`,
                     0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-1.388011981982932, 0.5}, 1.6, {
                    1.562295237544169, 1.5232173299972054`}], 
                   ArrowBox[{{-1.3278997376661752`, 
                    2.0988703881438306`}, {-1.3119143058489717`, 
                    2.0981893328661068`}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Dashing[{Small, Small}], 
                   Arrowheads[0.02], 
                   CircleBox[{-1.388011981982932, 0.5}, 1.7, {
                    1.562295237544169, 1.5702928883838851`}], 
                   ArrowBox[{{-1.370156464998479, 
                    2.199906226976423}, {-1.3871561367203649`, 
                    2.199999784567306}}]}, 
                  {RGBColor[0, 0, 1], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-1.388011981982932, 0.5}, 1.8, {
                    1.562295237544169, 1.5834960289472582`}], 
                   ArrowBox[{{-1.3928714399542388`, 
                    2.299993440451444}, {-1.4108708313904728`, 
                    2.2998548477596086`}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.003], Arrowheads[0.025], 
                   CircleBox[{-1.388011981982932, 0.5}, 1.5, {
                    1.562295237544169, 1.5551170901294604`}], 
                   ArrowBox[{{-1.379493172778844, 
                    1.9999758097681923`}, {-1.3644940906127911`, 
                    1.9998156249304453`}}]}, 
                  {RGBColor[0, 0, 1], ArrowBox[{{0, 2.8}, {1.5, 2.8}}]}},
                 Frame->True,
                 FrameTicks->Automatic,
                 ImageSize->800,
                 PlotRange->{{-3, 3}, {-1, 3}}], 96->
                GraphicsBox[{InsetBox[
                   TagBox[GridBox[{
                    {"\<\"t\"\>", "4.75`"},
                    {"\<\"driveIntent\"\>", "0.5`"},
                    {"\<\"motorTorque\"\>", 
                    RowBox[{"-", "3.931534590585919`"}]},
                    {"\<\"vw\"\>", "0.3383680176200696`"},
                    {"\<\"controllerState\"\>", 
                    RowBox[{"\[LeftAssociation]", 
                    RowBox[{
                    
                    RowBox[{"\<\"lastError\"\>", "\[Rule]", 
                    "0.03985663336705809`"}], ",", 
                    RowBox[{"\<\"lastInt\"\>", "\[Rule]", 
                    RowBox[{"-", "0.15621423100512402`"}]}], ",", 
                    RowBox[{"\<\"initialized\"\>", "\[Rule]", "True"}]}], 
                    "\[RightAssociation]"}]}
                    },
                    AutoDelete->False,
                    
                    GridBoxItemSize->{
                    "Columns" -> {{Automatic}}, "Rows" -> {{Automatic}}}],
                    "Grid"], {-1, -0.5}], 
                  {GrayLevel[0], LineBox[{{-5, 0}, {5, 0}}]}, 
                  {GrayLevel[0.5], 
                   CircleBox[{-1.3710994635501041`, 0.5}, 0.5], 
                   DiskBox[{-1.3710994635501041`, 0.5}, 0.1]}, 
                  {GrayLevel[0.5], 
                   DiskBox[{-1.3581412533262285`, 1.9999440272182805`}, 0.1]}, 
                  {GrayLevel[0.5], Thickness[0.01], 
                   LineBox[{{-1.3710994635501041`, 
                    0.5}, {-1.3581412533262285`, 1.9999440272182805`}}]}, 
                  {RGBColor[0, 1, 0], 
                   DiskBox[{-1.362460656734187, 1.499962684812187}, 0.05]}, 
                  {RGBColor[0, 0, 1], Arrowheads[0.01], 
                   CircleBox[{-1.3710994635501041`, 0.5}, 0.3333333333333333, 
                    NCache[{
                    Rational[-1, 2] 
                    Pi, -1.6887423645124742`}, {-1.5707963267948966`, \
-1.6887423645124742`}]], 
                   ArrowBox[{{-1.4070116376744128`, 
                    0.16860683944779964`}, {-1.4103237184542747`, 
                    0.16898252472365793`}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-1.3710994635501041`, 
                    0.5}, {-1.2924687717383858`, 0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-1.3710994635501041`, 
                    0.5}, {-1.4346354087518887`, 0.5}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.005], Arrowheads[0.04], 
                   ArrowBox[{{-1.3710994635501041`, 
                    0.5}, {-1.3560047169401703`, 0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-1.3710994635501041`, 0.5}, 1.6, {
                    1.562157412524476, 1.5224459649332354`}], 
                   ArrowBox[{{-1.3097539244799352`, 
                    2.0988235439960814`}, {-1.2937690227606256`, 
                    2.09813015831856}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Dashing[{Small, Small}], 
                   Arrowheads[0.02], 
                   CircleBox[{-1.3710994635501041`, 0.5}, 1.7, {
                    1.562157412524476, 1.5709051209339266`}], 
                   ArrowBox[{{-1.354284687771188, 
                    2.199916840117629}, {-1.3712844135860904`, 
                    2.19999998993926}}]}, 
                  {RGBColor[0, 0, 1], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-1.3710994635501041`, 0.5}, 1.8, {
                    1.562157412524476, 1.5839992713610644`}], 
                   ArrowBox[{{-1.3768647539116492`, 
                    2.2999907670393887`}, {-1.394864073322966, 
                    2.299843116308292}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.003], Arrowheads[0.025], 
                   CircleBox[{-1.3710994635501041`, 0.5}, 1.5, {
                    1.562157412524476, 1.5559228352284362`}], 
                   ArrowBox[{{-1.3637892551378563`, 
                    1.9999821868452203`}, {-1.34879004877256, 
                    1.9998340874951082`}}]}, 
                  {RGBColor[0, 0, 1], ArrowBox[{{0, 2.8}, {1.5, 2.8}}]}},
                 Frame->True,
                 FrameTicks->Automatic,
                 ImageSize->800,
                 PlotRange->{{-3, 3}, {-1, 3}}], 97->
                GraphicsBox[{InsetBox[
                   TagBox[GridBox[{
                    {"\<\"t\"\>", "4.8`"},
                    {"\<\"driveIntent\"\>", "0.5`"},
                    {"\<\"motorTorque\"\>", 
                    RowBox[{"-", "4.126235019387114`"}]},
                    {"\<\"vw\"\>", "0.3386287735554319`"},
                    {"\<\"controllerState\"\>", 
                    RowBox[{"\[LeftAssociation]", 
                    RowBox[{
                    
                    RowBox[{"\<\"lastError\"\>", "\[Rule]", 
                    "0.03947746943568661`"}], ",", 
                    RowBox[{"\<\"lastInt\"\>", "\[Rule]", 
                    RowBox[{"-", "0.15423039000870886`"}]}], ",", 
                    RowBox[{"\<\"initialized\"\>", "\[Rule]", "True"}]}], 
                    "\[RightAssociation]"}]}
                    },
                    AutoDelete->False,
                    
                    GridBoxItemSize->{
                    "Columns" -> {{Automatic}}, "Rows" -> {{Automatic}}}],
                    "Grid"], {-1, -0.5}], 
                  {GrayLevel[0], LineBox[{{-5, 0}, {5, 0}}]}, 
                  {GrayLevel[0.5], 
                   CircleBox[{-1.3541745265747336`, 0.5}, 0.5], 
                   DiskBox[{-1.3541745265747336`, 0.5}, 0.1]}, 
                  {GrayLevel[0.5], 
                   DiskBox[{-1.3407600507986466`, 1.9999400160805274`}, 0.1]}, 
                  {GrayLevel[0.5], Thickness[0.01], 
                   LineBox[{{-1.3541745265747336`, 
                    0.5}, {-1.3407600507986466`, 1.9999400160805274`}}]}, 
                  {RGBColor[0, 1, 0], 
                   DiskBox[{-1.345231542724009, 1.4999600107203515`}, 0.05]}, 
                  {RGBColor[0, 0, 1], Arrowheads[0.01], 
                   CircleBox[{-1.3541745265747336`, 0.5}, 0.3333333333333333, 
                    NCache[{
                    Rational[-1, 2] 
                    Pi, -1.69458337737651}, {-1.5707963267948966`, \
-1.69458337737651}]], 
                   ArrowBox[{{-1.3920217487923843`, 
                    0.16882225485169017`}, {-1.3953315787025111`, 
                    0.16921727951529497`}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-1.3541745265747336`, 
                    0.5}, {-1.2716498261869913`, 0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-1.3541745265747336`, 
                    0.5}, {-1.4199474306932744`, 0.5}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.005], Arrowheads[0.04], 
                   ArrowBox[{{-1.3541745265747336`, 
                    0.5}, {-1.3374227303055322`, 0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-1.3541745265747336`, 0.5}, 1.6, {
                    1.5618532237344378`, 1.5207435147126993`}], 
                   ArrowBox[{{-1.2901071602243228`, 
                    2.098716789356052}, {-1.274123462123851, 
                    2.0979961911970504`}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Dashing[{Small, Small}], 
                   Arrowheads[0.02], 
                   CircleBox[{-1.3541745265747336`, 0.5}, 1.7, {
                    1.5618532237344378`, 1.5715611991196132`}], 
                   ArrowBox[{{-1.338475032690954, 
                    2.1999275060695362`}, {-1.3554748093999678`, 
                    2.1999995027248023`}}]}, 
                  {RGBColor[0, 0, 1], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-1.3541745265747336`, 0.5}, 1.8, {
                    1.5618532237344378`, 1.5847767516199216`}], 
                   ArrowBox[{{-1.3613392723402988`, 
                    2.299985740615218}, {-1.3793384715160246`, 
                    2.2998240958146416`}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.003], Arrowheads[0.025], 
                   CircleBox[{-1.3541745265747336`, 0.5}, 1.5, {
                    1.5618532237344378`, 1.5565134730103631`}], 
                   ArrowBox[{{-1.347750265537837, 
                    1.999986242893624}, {-1.332750974315733, 
                    1.9998470026668074`}}]}, 
                  {RGBColor[0, 0, 1], ArrowBox[{{0, 2.8}, {1.5, 2.8}}]}},
                 Frame->True,
                 FrameTicks->Automatic,
                 ImageSize->800,
                 PlotRange->{{-3, 3}, {-1, 3}}], 98->
                GraphicsBox[{InsetBox[
                   TagBox[GridBox[{
                    {"\<\"t\"\>", "4.8500000000000005`"},
                    {"\<\"driveIntent\"\>", "0.5`"},
                    {"\<\"motorTorque\"\>", 
                    RowBox[{"-", "4.388367165442084`"}]},
                    {"\<\"vw\"\>", "0.3389192433762527`"},
                    {"\<\"controllerState\"\>", 
                    RowBox[{"\[LeftAssociation]", 
                    RowBox[{
                    
                    RowBox[{"\<\"lastError\"\>", "\[Rule]", 
                    "0.03894644892309443`"}], ",", 
                    RowBox[{"\<\"lastInt\"\>", "\[Rule]", 
                    RowBox[{"-", "0.15226947000182303`"}]}], ",", 
                    RowBox[{"\<\"initialized\"\>", "\[Rule]", "True"}]}], 
                    "\[RightAssociation]"}]}
                    },
                    AutoDelete->False,
                    
                    GridBoxItemSize->{
                    "Columns" -> {{Automatic}}, "Rows" -> {{Automatic}}}],
                    "Grid"], {-1, -0.5}], 
                  {GrayLevel[0], LineBox[{{-5, 0}, {5, 0}}]}, 
                  {GrayLevel[0.5], CircleBox[{-1.337235814704231, 0.5}, 0.5], 
                   DiskBox[{-1.337235814704231, 0.5}, 0.1]}, 
                  {GrayLevel[0.5], 
                   DiskBox[{-1.3231513214071862`, 1.9999338742253157`}, 0.1]}, 
                  {GrayLevel[0.5], Thickness[0.01], 
                   LineBox[{{-1.337235814704231, 0.5}, {-1.3231513214071862`, 
                    1.9999338742253157`}}]}, 
                  {RGBColor[0, 1, 0], 
                   DiskBox[{-1.327846152506201, 1.4999559161502105`}, 0.05]}, 
                  {RGBColor[0, 0, 1], Arrowheads[0.01], 
                   CircleBox[{-1.337235814704231, 0.5}, 0.3333333333333333, 
                    NCache[{
                    Rational[-1, 2] 
                    Pi, -1.7024473417581591`}, {-1.5707963267948966`, \
-1.7024473417581591`}]], 
                   ArrowBox[{{-1.3776862098038798`, 
                    0.16913012127515537`}, {-1.3809928309435235`, 
                    0.16955116184052732`}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-1.337235814704231, 0.5}, {-1.2494684713953892`,
                     0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-1.337235814704231, 0.5}, {-1.4062936183319086`,
                     0.5}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.005], Arrowheads[0.04], 
                   ArrowBox[{{-1.337235814704231, 0.5}, {-1.3185262750230669`,
                     0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-1.337235814704231, 0.5}, 1.6, {
                    1.5614065266169477`, 1.5182434965571154`}], 
                   ArrowBox[{{-1.2691718317315313`, 
                    2.098551623883912}, {-1.2531899850874193`, 
                    2.0977910684829926`}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Dashing[{Small, Small}], 
                   Arrowheads[0.02], 
                   CircleBox[{-1.337235814704231, 0.5}, 1.7, {
                    1.5614065266169477`, 1.5722490050922622`}], 
                   ArrowBox[{{-1.3227055447335427`, 
                    2.1999379021760115`}, {-1.33970536694118, 
                    2.199998206267215}}]}, 
                  {RGBColor[0, 0, 1], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-1.337235814704231, 0.5}, 1.8, {
                    1.5614065266169477`, 1.5857863442027371`}], 
                   ArrowBox[{{-1.3462178087625505`, 
                    2.299977589800144}, {-1.3642168355698265`, 
                    2.2997977732270507`}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.003], Arrowheads[0.025], 
                   CircleBox[{-1.337235814704231, 0.5}, 1.5, {
                    1.5614065266169477`, 1.5569098845947638`}], 
                   ArrowBox[{{-1.331406166079647, 
                    1.9999886716895279`}, {-1.3164068208396158`, 
                    1.999855377366294}}]}, 
                  {RGBColor[0, 0, 1], ArrowBox[{{0, 2.8}, {1.5, 2.8}}]}},
                 Frame->True,
                 FrameTicks->Automatic,
                 ImageSize->800,
                 PlotRange->{{-3, 3}, {-1, 3}}], 99->
                GraphicsBox[{InsetBox[
                   TagBox[GridBox[{
                    {"\<\"t\"\>", "4.9`"},
                    {"\<\"driveIntent\"\>", "0.5`"},
                    {"\<\"motorTorque\"\>", 
                    RowBox[{"-", "4.706772924123783`"}]},
                    {"\<\"vw\"\>", "0.3392438948069448`"},
                    {"\<\"controllerState\"\>", 
                    RowBox[{"\[LeftAssociation]", 
                    RowBox[{
                    
                    RowBox[{"\<\"lastError\"\>", "\[Rule]", 
                    "0.038284732129762`"}], ",", 
                    RowBox[{"\<\"lastInt\"\>", "\[Rule]", 
                    RowBox[{"-", "0.15033851954718064`"}]}], ",", 
                    RowBox[{"\<\"initialized\"\>", "\[Rule]", "True"}]}], 
                    "\[RightAssociation]"}]}
                    },
                    AutoDelete->False,
                    
                    GridBoxItemSize->{
                    "Columns" -> {{Automatic}}, "Rows" -> {{Automatic}}}],
                    "Grid"], {-1, -0.5}], 
                  {GrayLevel[0], LineBox[{{-5, 0}, {5, 0}}]}, 
                  {GrayLevel[0.5], 
                   CircleBox[{-1.3202817243612501`, 0.5}, 0.5], 
                   DiskBox[{-1.3202817243612501`, 0.5}, 0.1]}, 
                  {GrayLevel[0.5], 
                   DiskBox[{-1.30534717989295, 1.9999256512845975`}, 0.1]}, 
                  {GrayLevel[0.5], Thickness[0.01], 
                   LineBox[{{-1.3202817243612501`, 0.5}, {-1.30534717989295, 
                    1.9999256512845975`}}]}, 
                  {RGBColor[0, 1, 0], 
                   DiskBox[{-1.3103253613823833`, 1.4999504341897318`}, 
                    0.05]}, 
                  {RGBColor[0, 0, 1], Arrowheads[0.01], 
                   CircleBox[{-1.3202817243612501`, 0.5}, 0.3333333333333333, 
                    NCache[{
                    Rational[-1, 2] 
                    Pi, -1.71199951451861}, {-1.5707963267948966`, \
-1.71199951451861}]], 
                   ArrowBox[{{-1.363890752226766, 
                    0.16953159939301948`}, {-1.3671932007218215`, 
                    0.1699842056858753}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-1.3202817243612501`, 
                    0.5}, {-1.2261462658787745`, 0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-1.3202817243612501`, 
                    0.5}, {-1.3935070199652626`, 0.5}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.005], Arrowheads[0.04], 
                   ArrowBox[{{-1.3202817243612501`, 0.5}, {-1.299371561482787,
                     0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-1.3202817243612501`, 0.5}, 1.6, {
                    1.5608397993143688`, 1.5150717210299764`}], 
                   ArrowBox[{{-1.2471478453372844`, 
                    2.098327699734603}, {-1.2311684913900351`, 
                    2.0975164574142635`}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Dashing[{Small, Small}], 
                   Arrowheads[0.02], 
                   CircleBox[{-1.3202817243612501`, 0.5}, 1.7, {
                    1.5608397993143688`, 1.5729575063409071`}], 
                   ArrowBox[{{-1.3069558660630178`, 
                    2.1999477702272547`}, {-1.323955726729441, 
                    2.1999960299090695`}}]}, 
                  {RGBColor[0, 0, 1], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-1.3202817243612501`, 0.5}, 1.8, {
                    1.5608397993143688`, 1.5869885377817232`}], 
                   ArrowBox[{{-1.331427632908405, 
                    2.2999654909810516`}, {-1.3494264305346837`, 
                    2.299764036228656}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.003], Arrowheads[0.025], 
                   CircleBox[{-1.3202817243612501`, 0.5}, 1.5, {
                    1.5608397993143688`, 1.5571324032685856`}], 
                   ArrowBox[{{-1.31478585136821, 
                    1.9999899317595582`}, {-1.2997864768390461`, 
                    1.999859975074008}}]}, 
                  {RGBColor[0, 0, 1], ArrowBox[{{0, 2.8}, {1.5, 2.8}}]}},
                 Frame->True,
                 FrameTicks->Automatic,
                 ImageSize->800,
                 PlotRange->{{-3, 3}, {-1, 3}}], 100->
                GraphicsBox[{InsetBox[
                   TagBox[GridBox[{
                    {"\<\"t\"\>", "4.95`"},
                    {"\<\"driveIntent\"\>", "0.5`"},
                    {"\<\"motorTorque\"\>", 
                    RowBox[{"-", "5.071037365846436`"}]},
                    {"\<\"vw\"\>", "0.3396062811597899`"},
                    {"\<\"controllerState\"\>", 
                    RowBox[{"\[LeftAssociation]", 
                    RowBox[{
                    
                    RowBox[{"\<\"lastError\"\>", "\[Rule]", 
                    "0.037512320390230704`"}], ",", 
                    RowBox[{"\<\"lastInt\"\>", "\[Rule]", 
                    RowBox[{"-", "0.1484435584604029`"}]}], ",", 
                    RowBox[{"\<\"initialized\"\>", "\[Rule]", "True"}]}], 
                    "\[RightAssociation]"}]}
                    },
                    AutoDelete->False,
                    
                    GridBoxItemSize->{
                    "Columns" -> {{Automatic}}, "Rows" -> {{Automatic}}}],
                    "Grid"], {-1, -0.5}], 
                  {GrayLevel[0], LineBox[{{-5, 0}, {5, 0}}]}, 
                  {GrayLevel[0.5], 
                   CircleBox[{-1.3033104521952472`, 0.5}, 0.5], 
                   DiskBox[{-1.3033104521952472`, 0.5}, 0.1]}, 
                  {GrayLevel[0.5], 
                   DiskBox[{-1.2873773895999525`, 1.999915376785082}, 0.1]}, 
                  {GrayLevel[0.5], Thickness[0.01], 
                   LineBox[{{-1.3033104521952472`, 
                    0.5}, {-1.2873773895999525`, 1.999915376785082}}]}, 
                  {RGBColor[0, 1, 0], 
                   DiskBox[{-1.2926884104650507`, 1.499943584523388}, 0.05]}, 
                  {RGBColor[0, 0, 1], Arrowheads[0.01], 
                   CircleBox[{-1.3033104521952472`, 0.5}, 0.3333333333333333, 
                    NCache[{
                    Rational[-1, 2] 
                    Pi, -1.7229274477702896`}, {-1.5707963267948966`, \
-1.7229274477702896`}]], 
                   ArrowBox[{{-1.3505281409441954`, 
                    0.170027878480138}, {-1.3538254462995518`, 
                    0.17051654596664012`}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-1.3033104521952472`, 
                    0.5}, {-1.2018897048783186`, 0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-1.3033104521952472`, 
                    0.5}, {-1.3814310288871048`, 0.5}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.005], Arrowheads[0.04], 
                   ArrowBox[{{-1.3033104521952472`, 
                    0.5}, {-1.2800102815701762`, 0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-1.3033104521952472`, 0.5}, 1.6, {
                    1.56017408531101, 1.511345970217214}], 
                   ArrowBox[{{-1.2242221238150799`, 
                    2.0980441284000984`}, {-1.208245903253901, 
                    2.0971733567570485`}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Dashing[{Small, Small}], 
                   Arrowheads[0.02], 
                   CircleBox[{-1.3033104521952472`, 0.5}, 1.7, {
                    1.56017408531101, 1.5736767378731258`}], 
                   ArrowBox[{{-1.291207253277768, 
                    2.199956914917541}, {-1.3082071442571106`, 
                    2.199992947752093}}]}, 
                  {RGBColor[0, 0, 1], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-1.3033104521952472`, 0.5}, 1.8, {
                    1.56017408531101, 1.5883465151212681`}], 
                   ArrowBox[{{-1.3169006620627597`, 
                    2.299948695434333}, {-1.334899169522323, 
                    2.299722798915886}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.003], Arrowheads[0.025], 
                   CircleBox[{-1.3033104521952472`, 0.5}, 1.5, {
                    1.56017408531101, 1.5572006845536683`}], 
                   ArrowBox[{{-1.297917000455091, 
                    1.9999903035281017`}, {-1.2829176170872838`, 
                    1.9998613710194217`}}]}, 
                  {RGBColor[0, 0, 1], ArrowBox[{{0, 2.8}, {1.5, 2.8}}]}},
                 Frame->True,
                 FrameTicks->Automatic,
                 ImageSize->800,
                 PlotRange->{{-3, 3}, {-1, 3}}], 101->
                GraphicsBox[{InsetBox[
                   TagBox[GridBox[{
                    {"\<\"t\"\>", "5.`"},
                    {"\<\"driveIntent\"\>", "0.5`"},
                    {"\<\"motorTorque\"\>", 
                    RowBox[{"-", "5.471503497998029`"}]},
                    {"\<\"vw\"\>", "0.3400091124716204`"},
                    {"\<\"controllerState\"\>", 
                    RowBox[{"\[LeftAssociation]", 
                    RowBox[{
                    
                    RowBox[{"\<\"lastError\"\>", "\[Rule]", 
                    "0.03664799717565606`"}], ",", 
                    RowBox[{"\<\"lastInt\"\>", "\[Rule]", 
                    RowBox[{"-", "0.14658963734003969`"}]}], ",", 
                    RowBox[{"\<\"initialized\"\>", "\[Rule]", "True"}]}], 
                    "\[RightAssociation]"}]}
                    },
                    AutoDelete->False,
                    
                    GridBoxItemSize->{
                    "Columns" -> {{Automatic}}, "Rows" -> {{Automatic}}}],
                    "Grid"], {-1, -0.5}], 
                  {GrayLevel[0], LineBox[{{-5, 0}, {5, 0}}]}, 
                  {GrayLevel[0.5], CircleBox[{-1.286320038990218, 0.5}, 0.5], 
                   DiskBox[{-1.286320038990218, 0.5}, 0.1]}, 
                  {GrayLevel[0.5], 
                   DiskBox[{-1.269269350399791, 1.999903088208899}, 0.1]}, 
                  {GrayLevel[0.5], Thickness[0.01], 
                   LineBox[{{-1.286320038990218, 0.5}, {-1.269269350399791, 
                    1.999903088208899}}]}, 
                  {RGBColor[0, 1, 0], 
                   DiskBox[{-1.2749529132632667`, 1.499935392139266}, 0.05]}, 
                  {RGBColor[0, 0, 1], Arrowheads[0.01], 
                   CircleBox[{-1.286320038990218, 0.5}, 0.3333333333333333, 
                    NCache[{
                    Rational[-1, 2] 
                    Pi, -1.7349414317348375`}, {-1.5707963267948966`, \
-1.7349414317348375`}]], 
                   ArrowBox[{{-1.3374985045912466`, 
                    0.17061895049982695`}, {-1.3407897012877255`, 
                    0.17114719554136898`}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-1.286320038990218, 0.5}, {-1.1768899690302574`,
                     0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-1.286320038990218, 0.5}, {-1.3699197050783083`,
                     0.5}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.005], Arrowheads[0.04], 
                   ArrowBox[{{-1.286320038990218, 0.5}, {-1.2604896351183477`,
                     0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-1.286320038990218, 0.5}, 1.6, {
                    1.5594289562597266`, 1.5071757889893163`}], 
                   ArrowBox[{{-1.2005682839880554`, 
                    2.097700421391335}, {-1.1845958336082345`, 
                    2.0967630337778376`}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Dashing[{Small, Small}], 
                   Arrowheads[0.02], 
                   CircleBox[{-1.286320038990218, 0.5}, 1.7, {
                    1.5594289562597266`, 1.5743977757500998`}], 
                   ArrowBox[{{-1.2754425764376094`, 
                    2.1999651998815786`}, {-1.292442488978904, 
                    2.199988975142526}}]}, 
                  {RGBColor[0, 0, 1], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-1.286320038990218, 0.5}, 1.8, {
                    1.5594289562597266`, 1.589826197915271}], 
                   ArrowBox[{{-1.3025735861229524`, 
                    2.2999266157834333`}, {-1.320571739623933, 
                    2.299674087440195}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.003], Arrowheads[0.025], 
                   CircleBox[{-1.286320038990218, 0.5}, 1.5, {
                    1.5594289562597266`, 1.5571335965803654`}], 
                   ArrowBox[{{-1.2808259559528372`, 
                    1.9999899383167803`}, {-1.2658265812686051`, 
                    1.9998599995301602`}}]}, 
                  {RGBColor[0, 0, 1], ArrowBox[{{0, 2.8}, {1.5, 2.8}}]}},
                 Frame->True,
                 FrameTicks->Automatic,
                 ImageSize->800,
                 PlotRange->{{-3, 3}, {-1, 3}}], 102->
                GraphicsBox[{InsetBox[
                   TagBox[GridBox[{
                    {"\<\"t\"\>", "5.05`"},
                    {"\<\"driveIntent\"\>", "0.5`"},
                    {"\<\"motorTorque\"\>", 
                    RowBox[{"-", "5.899278101418652`"}]},
                    {"\<\"vw\"\>", "0.3404543265424889`"},
                    {"\<\"controllerState\"\>", 
                    RowBox[{"\[LeftAssociation]", 
                    RowBox[{
                    
                    RowBox[{"\<\"lastError\"\>", "\[Rule]", 
                    "0.03570928751435769`"}], ",", 
                    RowBox[{"\<\"lastInt\"\>", "\[Rule]", 
                    RowBox[{"-", "0.14478089956826617`"}]}], ",", 
                    RowBox[{"\<\"initialized\"\>", "\[Rule]", "True"}]}], 
                    "\[RightAssociation]"}]}
                    },
                    AutoDelete->False,
                    
                    GridBoxItemSize->{
                    "Columns" -> {{Automatic}}, "Rows" -> {{Automatic}}}],
                    "Grid"], {-1, -0.5}], 
                  {GrayLevel[0], LineBox[{{-5, 0}, {5, 0}}]}, 
                  {GrayLevel[0.5], 
                   CircleBox[{-1.2693084100151026`, 0.5}, 0.5], 
                   DiskBox[{-1.2693084100151026`, 0.5}, 0.1]}, 
                  {GrayLevel[0.5], 
                   DiskBox[{-1.2510481104613618`, 1.9998888497019396`}, 0.1]}, 
                  {GrayLevel[0.5], Thickness[0.01], 
                   LineBox[{{-1.2693084100151026`, 
                    0.5}, {-1.2510481104613618`, 1.9998888497019396`}}]}, 
                  {RGBColor[0, 1, 0], 
                   DiskBox[{-1.2571348769792754`, 1.4999258998012932`}, 
                    0.05]}, 
                  {RGBColor[0, 0, 1], Arrowheads[0.01], 
                   CircleBox[{-1.2693084100151026`, 0.5}, 0.3333333333333333, 
                    NCache[{
                    Rational[-1, 2] 
                    Pi, -1.747774669837456}, {-1.5707963267948966`, \
-1.747774669837456}]], 
                   ArrowBox[{{-1.3247095707415057`, 
                    0.17130284074656443`}, {-1.3279937175165015`, 
                    0.17187327784135387`}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-1.2693084100151026`, 
                    0.5}, {-1.1513228479867297`, 0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-1.2693084100151026`, 0.5}, {-1.358837958675474,
                     0.5}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.005], Arrowheads[0.04], 
                   ArrowBox[{{-1.2693084100151026`, 0.5}, {-1.240852396647101,
                     0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-1.2693084100151026`, 0.5}, 1.6, {
                    1.5586224930630843`, 1.502662378494743}], 
                   ArrowBox[{{-1.1763464750641113`, 
                    2.097297116584816}, {-1.1603784181711316`, 
                    2.0962876485385937`}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Dashing[{Small, Small}], 
                   Arrowheads[0.02], 
                   CircleBox[{-1.2693084100151026`, 0.5}, 1.7, {
                    1.5586224930630843`, 1.575112703454016}], 
                   ArrowBox[{{-1.2596463023559037`, 
                    2.199972542035777}, {-1.2766462275502781`, 
                    2.1999841635832436`}}]}, 
                  {RGBColor[0, 0, 1], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-1.2693084100151026`, 0.5}, 1.8, {
                    1.5586224930630843`, 1.591396260293188}], 
                   ArrowBox[{{-1.288387933015959, 
                    2.299898878215679}, {-1.3063856678482693`, 
                    2.2996180919716194`}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.003], Arrowheads[0.025], 
                   CircleBox[{-1.2693084100151026`, 0.5}, 1.5, {
                    1.5586224930630843`, 1.5569491299760339`}], 
                   ArrowBox[{{-1.2635376290223144`, 
                    1.9999888993211694`}, {-1.2485382785626475`, 
                    1.9998561936530608`}}]}, 
                  {RGBColor[0, 0, 1], ArrowBox[{{0, 2.8}, {1.5, 2.8}}]}},
                 Frame->True,
                 FrameTicks->Automatic,
                 ImageSize->800,
                 PlotRange->{{-3, 3}, {-1, 3}}], 103->
                GraphicsBox[{InsetBox[
                   TagBox[GridBox[{
                    {"\<\"t\"\>", "5.1000000000000005`"},
                    {"\<\"driveIntent\"\>", "0.5`"},
                    {"\<\"motorTorque\"\>", 
                    RowBox[{"-", "6.34622961913454`"}]},
                    {"\<\"vw\"\>", "0.3409431592025167`"},
                    {"\<\"controllerState\"\>", 
                    RowBox[{"\[LeftAssociation]", 
                    RowBox[{
                    
                    RowBox[{"\<\"lastError\"\>", "\[Rule]", 
                    "0.034712434035556905`"}], ",", 
                    RowBox[{"\<\"lastInt\"\>", "\[Rule]", 
                    RowBox[{"-", "0.14302064490964572`"}]}], ",", 
                    RowBox[{"\<\"initialized\"\>", "\[Rule]", "True"}]}], 
                    "\[RightAssociation]"}]}
                    },
                    AutoDelete->False,
                    
                    GridBoxItemSize->{
                    "Columns" -> {{Automatic}}, "Rows" -> {{Automatic}}}],
                    "Grid"], {-1, -0.5}], 
                  {GrayLevel[0], LineBox[{{-5, 0}, {5, 0}}]}, 
                  {GrayLevel[0.5], 
                   CircleBox[{-1.2522734118393228`, 0.5}, 0.5], 
                   DiskBox[{-1.2522734118393228`, 0.5}, 0.1]}, 
                  {GrayLevel[0.5], 
                   DiskBox[{-1.232736399041407, 1.9998727629805584`}, 0.1]}, 
                  {GrayLevel[0.5], Thickness[0.01], 
                   LineBox[{{-1.2522734118393228`, 0.5}, {-1.232736399041407, 
                    1.9998727629805584`}}]}, 
                  {RGBColor[0, 1, 0], 
                   DiskBox[{-1.2392487366407123`, 1.499915175320372}, 0.05]}, 
                  {RGBColor[0, 0, 1], Arrowheads[0.01], 
                   CircleBox[{-1.2522734118393228`, 0.5}, 0.3333333333333333, 
                    NCache[{
                    Rational[-1, 2] 
                    Pi, -1.7611832153689329`}, {-1.5707963267948966`, \
-1.7611832153689329`}]], 
                   ArrowBox[{{-1.3120768111396177`, 
                    0.17207521511291557`}, {-1.315353014189584, 
                    0.17268963524134462`}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-1.2522734118393228`, 
                    0.5}, {-1.1253488194566321`, 0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-1.2522734118393228`, 
                    0.5}, {-1.3480616092242315`, 0.5}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.005], Arrowheads[0.04], 
                   ArrowBox[{{-1.2522734118393228`, 
                    0.5}, {-1.2211370168415407`, 0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-1.2522734118393228`, 0.5}, 1.6, {
                    1.55777128331249, 1.4978985812641523`}], 
                   ArrowBox[{{-1.1517033609142255`, 
                    2.0968361421438715`}, {-1.1357402940914554`, 
                    2.095750617254451}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Dashing[{Small, Small}], 
                   Arrowheads[0.02], 
                   CircleBox[{-1.2522734118393228`, 0.5}, 1.7, {
                    1.55777128331249, 1.5758145723973003`}], 
                   ArrowBox[{{-1.2438044643937263`, 
                    2.1999789048482814`}, {-1.2608043935576518`, 
                    2.1999785946743335`}}]}, 
                  {RGBColor[0, 0, 1], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-1.2522734118393228`, 0.5}, 1.8, {
                    1.55777128331249, 1.593028114529904}], 
                   ArrowBox[{{-1.2742900807427808`, 
                    2.299865346710802}, {-1.2922873334095593`, 
                    2.2995551911737997`}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.003], Arrowheads[0.025], 
                   CircleBox[{-1.2522734118393228`, 0.5}, 1.5, {
                    1.55777128331249, 1.5566643262179451`}], 
                   ArrowBox[{{-1.246075428610735, 
                    1.9999871949466428`}, {-1.2310761165544633`, 
                    1.9998502174125943`}}]}, 
                  {RGBColor[0, 0, 1], ArrowBox[{{0, 2.8}, {1.5, 2.8}}]}},
                 Frame->True,
                 FrameTicks->Automatic,
                 ImageSize->800,
                 PlotRange->{{-3, 3}, {-1, 3}}], 104->
                GraphicsBox[{InsetBox[
                   TagBox[GridBox[{
                    {"\<\"t\"\>", "5.15`"},
                    {"\<\"driveIntent\"\>", "0.5`"},
                    {"\<\"motorTorque\"\>", 
                    RowBox[{"-", "6.804979030340043`"}]},
                    {"\<\"vw\"\>", "0.34147621322313665`"},
                    {"\<\"controllerState\"\>", 
                    RowBox[{"\[LeftAssociation]", 
                    RowBox[{
                    
                    RowBox[{"\<\"lastError\"\>", "\[Rule]", 
                    "0.03367238796972608`"}], ",", 
                    RowBox[{"\<\"lastInt\"\>", "\[Rule]", 
                    RowBox[{"-", "0.141311393919454`"}]}], ",", 
                    RowBox[{"\<\"initialized\"\>", "\[Rule]", "True"}]}], 
                    "\[RightAssociation]"}]}
                    },
                    AutoDelete->False,
                    
                    GridBoxItemSize->{
                    "Columns" -> {{Automatic}}, "Rows" -> {{Automatic}}}],
                    "Grid"], {-1, -0.5}], 
                  {GrayLevel[0], LineBox[{{-5, 0}, {5, 0}}]}, 
                  {GrayLevel[0.5], 
                   CircleBox[{-1.2352128456672924`, 0.5}, 0.5], 
                   DiskBox[{-1.2352128456672924`, 0.5}, 0.1]}, 
                  {GrayLevel[0.5], 
                   DiskBox[{-1.2143546775940854`, 1.9998549719304963`}, 0.1]}, 
                  {GrayLevel[0.5], Thickness[0.01], 
                   LineBox[{{-1.2352128456672924`, 
                    0.5}, {-1.2143546775940854`, 1.9998549719304963`}}]}, 
                  {RGBColor[0, 1, 0], 
                   DiskBox[{-1.2213074002851545`, 1.4999033146203309`}, 
                    0.05]}, 
                  {RGBColor[0, 0, 1], Arrowheads[0.01], 
                   CircleBox[{-1.2352128456672924`, 0.5}, 0.3333333333333333, 
                    NCache[{
                    Rational[-1, 2] 
                    Pi, -1.774945697705098}, {-1.5707963267948966`, \
-1.774945697705098}]], 
                   ArrowBox[{{-1.2995234980923105`, 
                    0.17292928731575552`}, {-1.3027909352018148`, 
                    0.1735887365209724}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-1.2352128456672924`, 
                    0.5}, {-1.0991132650604916`, 0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-1.2352128456672924`, 0.5}, {-1.337477334197882,
                     0.5}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.005], Arrowheads[0.04], 
                   ArrowBox[{{-1.2352128456672924`, 
                    0.5}, {-1.2013777535910812`, 0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-1.2352128456672924`, 0.5}, 1.6, {
                    1.5568904332443374`, 1.492968947639614}], 
                   ArrowBox[{{-1.1267722314739095`, 
                    2.096320968099261}, {-1.1108147098306074`, 
                    2.095156764647399}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Dashing[{Small, Small}], 
                   Arrowheads[0.02], 
                   CircleBox[{-1.2352128456672924`, 0.5}, 1.7, {
                    1.5568904332443374`, 1.5764973577299004`}], 
                   ArrowBox[{{-1.2279046207675626`, 
                    2.1999842910594247`}, {-1.2449045457570582`, 
                    2.199972373584162}}]}, 
                  {RGBColor[0, 0, 1], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-1.2352128456672924`, 0.5}, 1.8, {
                    1.5568904332443374`, 1.594695872301782}], 
                   ArrowBox[{{-1.2602312219807967`, 
                    2.299826125170606}, {-1.2782279323545904`, 
                    2.2994859550208453`}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.003], Arrowheads[0.025], 
                   CircleBox[{-1.2352128456672924`, 0.5}, 1.5, {
                    1.5568904332443374`, 1.5562952232185376`}], 
                   ArrowBox[{{-1.2284612131007455`, 
                    1.9999848050756004`}, {-1.213461952625022, 
                    1.999842291259939}}]}, 
                  {RGBColor[0, 0, 1], ArrowBox[{{0, 2.8}, {1.5, 2.8}}]}},
                 Frame->True,
                 FrameTicks->Automatic,
                 ImageSize->800,
                 PlotRange->{{-3, 3}, {-1, 3}}], 105->
                GraphicsBox[{InsetBox[
                   TagBox[GridBox[{
                    {"\<\"t\"\>", "5.2`"},
                    {"\<\"driveIntent\"\>", "0.5`"},
                    {"\<\"motorTorque\"\>", 
                    RowBox[{"-", "7.268884595477012`"}]},
                    {"\<\"vw\"\>", "0.3420535253734271`"},
                    {"\<\"controllerState\"\>", 
                    RowBox[{"\[LeftAssociation]", 
                    RowBox[{
                    
                    RowBox[{"\<\"lastError\"\>", "\[Rule]", 
                    "0.03260281348528012`"}], ",", 
                    RowBox[{"\<\"lastInt\"\>", "\[Rule]", 
                    RowBox[{"-", "0.13965495245527723`"}]}], ",", 
                    RowBox[{"\<\"initialized\"\>", "\[Rule]", "True"}]}], 
                    "\[RightAssociation]"}]}
                    },
                    AutoDelete->False,
                    
                    GridBoxItemSize->{
                    "Columns" -> {{Automatic}}, "Rows" -> {{Automatic}}}],
                    "Grid"], {-1, -0.5}], 
                  {GrayLevel[0], LineBox[{{-5, 0}, {5, 0}}]}, 
                  {GrayLevel[0.5], 
                   CircleBox[{-1.2181244972727587`, 0.5}, 0.5], 
                   DiskBox[{-1.2181244972727587`, 0.5}, 0.1]}, 
                  {GrayLevel[0.5], 
                   DiskBox[{-1.195921206635099, 1.9998356622926592`}, 0.1]}, 
                  {GrayLevel[0.5], Thickness[0.01], 
                   LineBox[{{-1.2181244972727587`, 0.5}, {-1.195921206635099, 
                    1.9998356622926592`}}]}, 
                  {RGBColor[0, 1, 0], 
                   DiskBox[{-1.203322303514319, 1.4998904415284393`}, 0.05]}, 
                  {RGBColor[0, 0, 1], Arrowheads[0.01], 
                   CircleBox[{-1.2181244972727587`, 0.5}, 0.3333333333333333, 
                    NCache[{
                    Rational[-1, 2] 
                    Pi, -1.7888628646592069`}, {-1.5707963267948966`, \
-1.7888628646592069`}]], 
                   ArrowBox[{{-1.2869806724870707`, 
                    0.17385595475930105`}, {-1.290238615802338, 
                    0.1745608121018415}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-1.2181244972727587`, 
                    0.5}, {-1.0727468053632185`, 0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-1.2181244972727587`, 0.5}, {-1.326982519713333,
                     0.5}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.005], Arrowheads[0.04], 
                   ArrowBox[{{-1.2181244972727587`, 
                    0.5}, {-1.1816048278037927`, 0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-1.2181244972727587`, 0.5}, 1.6, {
                    1.5559935924441914`, 1.4879498736530026`}], 
                   ArrowBox[{{-1.1016732295393177`, 
                    2.0957565924173007`}, {-1.0857217520881124`, 
                    2.0945123119836895`}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Dashing[{Small, Small}], 
                   Arrowheads[0.02], 
                   CircleBox[{-1.2181244972727587`, 0.5}, 1.7, {
                    1.5559935924441914`, 1.5771559104604338`}], 
                   ArrowBox[{{-1.2119358031736063`, 
                    2.1999887352760155`}, {-1.2289357166284536`, 
                    2.1999656224571256`}}]}, 
                  {RGBColor[0, 0, 1], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-1.2181244972727587`, 0.5}, 1.8, {
                    1.5559935924441914`, 1.5963762846412861`}], 
                   ArrowBox[{{-1.2461672868693048`, 
                    2.2997815428411426`}, {-1.2641634002078301`, 
                    2.2994111312917163`}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.003], Arrowheads[0.025], 
                   CircleBox[{-1.2181244972727587`, 0.5}, 1.5, {
                    1.5559935924441914`, 1.5558568167885114`}], 
                   ArrowBox[{{-1.210715262392623, 
                    1.9999817009678789`}, {-1.195716065837301, 
                    1.9998326113938858`}}]}, 
                  {RGBColor[0, 0, 1], ArrowBox[{{0, 2.8}, {1.5, 2.8}}]}},
                 Frame->True,
                 FrameTicks->Automatic,
                 ImageSize->800,
                 PlotRange->{{-3, 3}, {-1, 3}}], 106->
                GraphicsBox[{InsetBox[
                   TagBox[GridBox[{
                    {"\<\"t\"\>", "5.25`"},
                    {"\<\"driveIntent\"\>", "0.5`"},
                    {"\<\"motorTorque\"\>", 
                    RowBox[{"-", "7.73202131054807`"}]},
                    {"\<\"vw\"\>", "0.3426746312004984`"},
                    {"\<\"controllerState\"\>", 
                    RowBox[{"\[LeftAssociation]", 
                    RowBox[{
                    
                    RowBox[{"\<\"lastError\"\>", "\[Rule]", 
                    "0.031516103800811116`"}], ",", 
                    RowBox[{"\<\"lastInt\"\>", "\[Rule]", 
                    RowBox[{"-", "0.1380524756651362`"}]}], ",", 
                    RowBox[{"\<\"initialized\"\>", "\[Rule]", "True"}]}], 
                    "\[RightAssociation]"}]}
                    },
                    AutoDelete->False,
                    
                    GridBoxItemSize->{
                    "Columns" -> {{Automatic}}, "Rows" -> {{Automatic}}}],
                    "Grid"], {-1, -0.5}], 
                  {GrayLevel[0], LineBox[{{-5, 0}, {5, 0}}]}, 
                  {GrayLevel[0.5], 
                   CircleBox[{-1.2010061636367582`, 0.5}, 0.5], 
                   DiskBox[{-1.2010061636367582`, 0.5}, 0.1]}, 
                  {GrayLevel[0.5], 
                   DiskBox[{-1.1774521259425783`, 1.9998150577015492`}, 0.1]}, 
                  {GrayLevel[0.5], Thickness[0.01], 
                   LineBox[{{-1.2010061636367582`, 
                    0.5}, {-1.1774521259425783`, 1.9998150577015492`}}]}, 
                  {RGBColor[0, 1, 0], 
                   DiskBox[{-1.1853034718406383`, 1.499876705134366}, 0.05]}, 
                  {RGBColor[0, 0, 1], Arrowheads[0.01], 
                   CircleBox[{-1.2010061636367582`, 0.5}, 0.3333333333333333, 
                    NCache[{
                    Rational[-1, 2] 
                    Pi, -1.8027569661113387`}, {-1.5707963267948966`, \
-1.8027569661113387`}]], 
                   ArrowBox[{{-1.274387025408139, 
                    0.1748440985668558}, {-1.277634861217578, 
                    0.17559415261407763`}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-1.2010061636367582`, 
                    0.5}, {-1.0463657374257966`, 0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-1.2010061636367582`, 
                    0.5}, {-1.3164850258070953`, 0.5}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.005], Arrowheads[0.04], 
                   ArrowBox[{{-1.2010061636367582`, 
                    0.5}, {-1.1618445995961337`, 0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-1.2010061636367582`, 0.5}, 1.6, {
                    1.555092989613189, 1.48290980094018}], 
                   ArrowBox[{{-1.0765136797600126`, 
                    2.095149404118059}, {-1.0605686761480588`, 
                    2.093824743222372}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Dashing[{Small, Small}], 
                   Arrowheads[0.02], 
                   CircleBox[{-1.2010061636367582`, 0.5}, 1.7, {
                    1.555092989613189, 1.577785906789199}], 
                   ArrowBox[{{-1.195888457357059, 
                    2.199992296771499}, {-1.2128883528773229`, 
                    2.1999584740748377`}}]}, 
                  {RGBColor[0, 0, 1], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-1.2010061636367582`, 0.5}, 1.8, {
                    1.555092989613189, 1.5980486635606783`}], 
                   ArrowBox[{{-1.2320588293262684`, 
                    2.299732127832799}, {-1.2500542980303957`, 
                    2.299331620494817}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.003], Arrowheads[0.025], 
                   CircleBox[{-1.2010061636367582`, 0.5}, 1.5, {
                    1.555092989613189, 1.555363036565111}], 
                   ArrowBox[{{-1.1628603410543246`, 
                    1.9995148869616166`}, {-1.1778571472812778`, 
                    1.999821363710283}}]}, 
                  {RGBColor[0, 0, 1], ArrowBox[{{0, 2.8}, {1.5, 2.8}}]}},
                 Frame->True,
                 FrameTicks->Automatic,
                 ImageSize->800,
                 PlotRange->{{-3, 3}, {-1, 3}}], 107->
                GraphicsBox[{InsetBox[
                   TagBox[GridBox[{
                    {"\<\"t\"\>", "5.3`"},
                    {"\<\"driveIntent\"\>", "0.5`"},
                    {"\<\"motorTorque\"\>", 
                    RowBox[{"-", "8.189155860304862`"}]},
                    {"\<\"vw\"\>", "0.3433386271847407`"},
                    {"\<\"controllerState\"\>", 
                    RowBox[{"\[LeftAssociation]", 
                    RowBox[{
                    
                    RowBox[{"\<\"lastError\"\>", "\[Rule]", 
                    "0.030423407580839005`"}], ",", 
                    RowBox[{"\<\"lastInt\"\>", "\[Rule]", 
                    RowBox[{"-", "0.13650453090170556`"}]}], ",", 
                    RowBox[{"\<\"initialized\"\>", "\[Rule]", "True"}]}], 
                    "\[RightAssociation]"}]}
                    },
                    AutoDelete->False,
                    
                    GridBoxItemSize->{
                    "Columns" -> {{Automatic}}, "Rows" -> {{Automatic}}}],
                    "Grid"], {-1, -0.5}], 
                  {GrayLevel[0], LineBox[{{-5, 0}, {5, 0}}]}, 
                  {GrayLevel[0.5], CircleBox[{-1.183855676412185, 0.5}, 0.5], 
                   DiskBox[{-1.183855676412185, 0.5}, 0.1]}, 
                  {GrayLevel[0.5], 
                   DiskBox[{-1.1589615458289884`, 1.9997934131947996`}, 0.1]}, 
                  {GrayLevel[0.5], Thickness[0.01], 
                   LineBox[{{-1.183855676412185, 0.5}, {-1.1589615458289884`, 
                    1.9997934131947996`}}]}, 
                  {RGBColor[0, 1, 0], 
                   DiskBox[{-1.1672595893567206`, 1.4998622754631996`}, 
                    0.05]}, 
                  {RGBColor[0, 0, 1], Arrowheads[0.01], 
                   CircleBox[{-1.183855676412185, 0.5}, 0.3333333333333333, 
                    NCache[{
                    Rational[-1, 2] 
                    Pi, -1.8164710026040425`}, {-1.5707963267948966`, \
-1.8164710026040425`}]], 
                   ArrowBox[{{-1.2616886978685797`, 
                    0.17588099117441525`}, {-1.2649259423186283`, 
                    0.1766755142322659}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-1.183855676412185, 0.5}, {-1.0200725592060877`,
                     0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-1.183855676412185, 0.5}, {-1.305902878044409, 
                    0.5}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.005], Arrowheads[0.04], 
                   ArrowBox[{{-1.183855676412185, 0.5}, {-1.142119760838312, 
                    0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-1.183855676412185, 0.5}, 1.6, {
                    1.5541994778013228`, 1.477909469775165}], 
                   ArrowBox[{{-1.0513885068549642`, 
                    2.0945069610978493`}, {-1.0354503262971007`, 
                    2.0931025867963493`}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Dashing[{Small, Small}], 
                   Arrowheads[0.02], 
                   CircleBox[{-1.183855676412185, 0.5}, 1.7, {
                    1.5541994778013228`, 1.578383795445241}], 
                   ArrowBox[{{-1.179754377096244, 
                    2.1999950527410137`}, {-1.1967542493558188`, 
                    2.199951066006318}}]}, 
                  {RGBColor[0, 0, 1], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-1.183855676412185, 0.5}, 1.8, {
                    1.5541994778013228`, 1.5996947881363497`}], 
                   ArrowBox[{{-1.2178708819768813`, 
                    2.299678572909727}, {-1.2358656670149442`, 
                    2.299248443344471}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.003], Arrowheads[0.025], 
                   CircleBox[{-1.183855676412185, 0.5}, 1.5, {
                    1.5541994778013228`, 1.5548267349718943`}], 
                   ArrowBox[{{-1.144905667131111, 
                    1.9994942136523917`}, {-1.1599023068374172`, 
                    1.9998087331676713`}}]}, 
                  {RGBColor[0, 0, 1], ArrowBox[{{0, 2.8}, {1.5, 2.8}}]}},
                 Frame->True,
                 FrameTicks->Automatic,
                 ImageSize->800,
                 PlotRange->{{-3, 3}, {-1, 3}}], 108->
                GraphicsBox[{InsetBox[
                   TagBox[GridBox[{
                    {"\<\"t\"\>", "5.3500000000000005`"},
                    {"\<\"driveIntent\"\>", "0.5`"},
                    {"\<\"motorTorque\"\>", 
                    RowBox[{"-", "8.635717811649993`"}]},
                    {"\<\"vw\"\>", "0.34404422998629874`"},
                    {"\<\"controllerState\"\>", 
                    RowBox[{"\[LeftAssociation]", 
                    RowBox[{
                    
                    RowBox[{"\<\"lastError\"\>", "\[Rule]", 
                    "0.029334664198179625`"}], ",", 
                    RowBox[{"\<\"lastInt\"\>", "\[Rule]", 
                    RowBox[{"-", "0.13501115908491`"}]}], ",", 
                    RowBox[{"\<\"initialized\"\>", "\[Rule]", "True"}]}], 
                    "\[RightAssociation]"}]}
                    },
                    AutoDelete->False,
                    
                    GridBoxItemSize->{
                    "Columns" -> {{Automatic}}, "Rows" -> {{Automatic}}}],
                    "Grid"], {-1, -0.5}], 
                  {GrayLevel[0], LineBox[{{-5, 0}, {5, 0}}]}, 
                  {GrayLevel[0.5], 
                   CircleBox[{-1.1666709223537184`, 0.5}, 0.5], 
                   DiskBox[{-1.1666709223537184`, 0.5}, 0.1]}, 
                  {GrayLevel[0.5], 
                   DiskBox[{-1.1404616473727844`, 1.9997710071557502`}, 0.1]}, 
                  {GrayLevel[0.5], Thickness[0.01], 
                   LineBox[{{-1.1666709223537184`, 
                    0.5}, {-1.1404616473727844`, 1.9997710071557502`}}]}, 
                  {RGBColor[0, 1, 0], 
                   DiskBox[{-1.149198072366429, 1.4998473381038335`}, 0.05]}, 
                  {RGBColor[0, 0, 1], Arrowheads[0.01], 
                   CircleBox[{-1.1666709223537184`, 0.5}, 0.3333333333333333, 
                    NCache[{
                    Rational[-1, 2] 
                    Pi, -1.8298678611443964`}, {-1.5707963267948966`, \
-1.8298678611443964`}]], 
                   ArrowBox[{{-1.2488390059686443`, 
                    0.1769527632896984}, {-1.2520653161248663`, 
                    0.17779058265846848`}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-1.1666709223537184`, 
                    0.5}, {-0.9939565661207186, 0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-1.1666709223537184`, 
                    0.5}, {-1.2951638966234404`, 0.5}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.005], Arrowheads[0.04], 
                   ArrowBox[{{-1.1666709223537184`, 
                    0.5}, {-1.1224495403904404`, 0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-1.1666709223537184`, 0.5}, 1.6, {
                    1.5533225876071715`, 1.4730022168284913`}], 
                   ArrowBox[{{-1.0263807301939838`, 
                    2.093837715071326}, {-1.0104496331327155`, 
                    2.0923551453096536`}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Dashing[{Small, Small}], 
                   Arrowheads[0.02], 
                   CircleBox[{-1.1666709223537184`, 0.5}, 1.7, {
                    1.5533225876071715`, 1.5789467437316729`}], 
                   ArrowBox[{{-1.1635266329389855`, 
                    2.1999970921869476`}, {-1.1805264777425835`, 
                    2.1999435354107693`}}]}, 
                  {RGBColor[0, 0, 1], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-1.1666709223537184`, 0.5}, 1.8, {
                    1.5533225876071715`, 1.6012987976718935`}], 
                   ArrowBox[{{-1.2035727845144926`, 
                    2.2996216970711005`}, {-1.2215668564720215`, 
                    2.2991627042647584`}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.003], Arrowheads[0.025], 
                   CircleBox[{-1.1666709223537184`, 0.5}, 1.5, {
                    1.5533225876071715`, 1.5542596877953474`}], 
                   ArrowBox[{{-1.1268706354201532`, 
                    1.999471886085233}, {-1.1418670943680373`, 
                    1.9997949093516942`}}]}, 
                  {RGBColor[0, 0, 1], ArrowBox[{{0, 2.8}, {1.5, 2.8}}]}},
                 Frame->True,
                 FrameTicks->Automatic,
                 ImageSize->800,
                 PlotRange->{{-3, 3}, {-1, 3}}], 109->
                GraphicsBox[{InsetBox[
                   TagBox[GridBox[{
                    {"\<\"t\"\>", "5.4`"},
                    {"\<\"driveIntent\"\>", "0.5`"},
                    {"\<\"motorTorque\"\>", 
                    RowBox[{"-", "9.067767740289089`"}]},
                    {"\<\"vw\"\>", "0.3447898325587143`"},
                    {"\<\"controllerState\"\>", 
                    RowBox[{"\[LeftAssociation]", 
                    RowBox[{
                    
                    RowBox[{"\<\"lastError\"\>", "\[Rule]", 
                    "0.028258646525322947`"}], ",", 
                    RowBox[{"\<\"lastInt\"\>", "\[Rule]", 
                    RowBox[{"-", "0.13357193410400925`"}]}], ",", 
                    RowBox[{"\<\"initialized\"\>", "\[Rule]", "True"}]}], 
                    "\[RightAssociation]"}]}
                    },
                    AutoDelete->False,
                    
                    GridBoxItemSize->{
                    "Columns" -> {{Automatic}}, "Rows" -> {{Automatic}}}],
                    "Grid"], {-1, -0.5}], 
                  {GrayLevel[0], LineBox[{{-5, 0}, {5, 0}}]}, 
                  {GrayLevel[0.5], 
                   CircleBox[{-1.1494498608644919`, 0.5}, 0.5], 
                   DiskBox[{-1.1494498608644919`, 0.5}, 0.1]}, 
                  {GrayLevel[0.5], 
                   DiskBox[{-1.1219627896534734`, 1.9997481324929998`}, 0.1]}, 
                  {GrayLevel[0.5], Thickness[0.01], 
                   LineBox[{{-1.1494498608644919`, 
                    0.5}, {-1.1219627896534734`, 1.9997481324929998`}}]}, 
                  {RGBColor[0, 1, 0], 
                   DiskBox[{-1.1311251467238128`, 1.4998320883286667`}, 
                    0.05]}, 
                  {RGBColor[0, 0, 1], Arrowheads[0.01], 
                   CircleBox[{-1.1494498608644919`, 0.5}, 0.3333333333333333, 
                    NCache[{
                    Rational[-1, 2] 
                    Pi, -1.8428293590035691`}, {-1.5707963267948966`, \
-1.8428293590035691`}]], 
                   ArrowBox[{{-1.2357981012662804`, 
                    0.17804489056605105`}, {-1.2390132813256616`, 
                    0.17892445620009167`}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-1.1494498608644919`, 
                    0.5}, {-0.9680945060587101, 0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-1.1494498608644919`, 
                    0.5}, {-1.2842052735037246`, 0.5}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.005], Arrowheads[0.04], 
                   ArrowBox[{{-1.1494498608644919`, 0.5}, {-1.102849918697943,
                     0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-1.1494498608644919`, 0.5}, 1.6, {
                    1.5524705869408661`, 1.4682343097872619`}], 
                   ArrowBox[{{-1.0015620220338157`, 
                    2.093150710738313}, {-0.9856381747805452, 
                    2.0915922001261302`}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Dashing[{Small, Small}], 
                   Arrowheads[0.02], 
                   CircleBox[{-1.1494498608644919`, 0.5}, 1.7, {
                    1.5524705869408661`, 1.5794725829070524`}], 
                   ArrowBox[{{-1.1471994969123749`, 
                    2.199998510547019}, {-1.1641993112034366`, 
                    2.199936014594284}}]}, 
                  {RGBColor[0, 0, 1], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-1.1494498608644919`, 0.5}, 1.8, {
                    1.5524705869408661`, 1.6028470743869168`}], 
                   ArrowBox[{{-1.1891379900516459`, 
                    2.2995624058091524`}, {-1.2071313297942465`, 
                    2.2990755537614604`}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.003], Arrowheads[0.025], 
                   CircleBox[{-1.1494498608644919`, 0.5}, 1.5, {
                    1.5524705869408661`, 1.5536726050072933`}], 
                   ArrowBox[{{-1.1087692667049107`, 
                    1.9994482616145257`}, {-1.1237655334270198`, 
                    1.9997800889877437`}}]}, 
                  {RGBColor[0, 0, 1], ArrowBox[{{0, 2.8}, {1.5, 2.8}}]}},
                 Frame->True,
                 FrameTicks->Automatic,
                 ImageSize->800,
                 PlotRange->{{-3, 3}, {-1, 3}}], 110->
                GraphicsBox[{InsetBox[
                   TagBox[GridBox[{
                    {"\<\"t\"\>", "5.45`"},
                    {"\<\"driveIntent\"\>", "0.5`"},
                    {"\<\"motorTorque\"\>", 
                    RowBox[{"-", "9.481962935352982`"}]},
                    {"\<\"vw\"\>", "0.3455735569596931`"},
                    {"\<\"controllerState\"\>", 
                    RowBox[{"\[LeftAssociation]", 
                    RowBox[{
                    
                    RowBox[{"\<\"lastError\"\>", "\[Rule]", 
                    "0.027203009999075645`"}], ",", 
                    RowBox[{"\<\"lastInt\"\>", "\[Rule]", 
                    RowBox[{"-", "0.1321860199150885`"}]}], ",", 
                    RowBox[{"\<\"initialized\"\>", "\[Rule]", "True"}]}], 
                    "\[RightAssociation]"}]}
                    },
                    AutoDelete->False,
                    
                    GridBoxItemSize->{
                    "Columns" -> {{Automatic}}, "Rows" -> {{Automatic}}}],
                    "Grid"], {-1, -0.5}], 
                  {GrayLevel[0], LineBox[{{-5, 0}, {5, 0}}]}, 
                  {GrayLevel[0.5], CircleBox[{-1.13219053882071, 0.5}, 0.5], 
                   DiskBox[{-1.13219053882071, 0.5}, 0.1]}, 
                  {GrayLevel[0.5], 
                   DiskBox[{-1.1034736221877497`, 1.9997250877074424`}, 0.1]}, 
                  {GrayLevel[0.5], Thickness[0.01], 
                   LineBox[{{-1.13219053882071, 0.5}, {-1.1034736221877497`, 
                    1.9997250877074424`}}]}, 
                  {RGBColor[0, 1, 0], 
                   DiskBox[{-1.11304592773207, 1.4998167251382948`}, 0.05]}, 
                  {RGBColor[0, 0, 1], Arrowheads[0.01], 
                   CircleBox[{-1.13219053882071, 0.5}, 0.3333333333333333, 
                    NCache[{
                    Rational[-1, 2] 
                    Pi, -1.855255214855486}, {-1.5707963267948966`, \
-1.855255214855486}]], 
                   ArrowBox[{{-1.222532577974557, 
                    0.17914266866310213`}, {-1.2257365807476561`, 
                    0.18006211673058603`}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-1.13219053882071, 0.5}, {-0.9425512801136504, 
                    0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-1.13219053882071, 0.5}, {-1.2729731074345199`, 
                    0.5}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.005], Arrowheads[0.04], 
                   ArrowBox[{{-1.13219053882071, 0.5}, {-1.0833338487274602`, 
                    0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-1.13219053882071, 0.5}, 1.6, {
                    1.5516505460451915`, 1.4636453115143726`}], 
                   ArrowBox[{{-0.9769933169653852, 
                    2.092455281107884}, {-0.9610767893586203, 
                    2.0908237126548705`}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Dashing[{Small, Small}], 
                   Arrowheads[0.02], 
                   CircleBox[{-1.13219053882071, 0.5}, 1.7, {
                    1.5516505460451915`, 1.579959753462607}], 
                   ArrowBox[{{-1.130768364321704, 
                    2.1999994051233354`}, {-1.1477681461492522`, 
                    2.1999286273693723`}}]}, 
                  {RGBColor[0, 0, 1], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-1.13219053882071, 0.5}, 1.8, {
                    1.5516505460451915`, 1.6043281179082636`}], 
                   ArrowBox[{{-1.1745438537481403`, 
                    2.299501652323403}, {-1.1925364527078326`, 
                    2.2989881519001534`}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.003], Arrowheads[0.025], 
                   CircleBox[{-1.13219053882071, 0.5}, 1.5, {
                    1.5516505460451915`, 1.5530751505210754`}], 
                   ArrowBox[{{-1.0906140998841785`, 
                    1.9994236891970716`}, {-1.1056101656780737`, 
                    1.9997644760973632`}}]}, 
                  {RGBColor[0, 0, 1], ArrowBox[{{0, 2.8}, {1.5, 2.8}}]}},
                 Frame->True,
                 FrameTicks->Automatic,
                 ImageSize->800,
                 PlotRange->{{-3, 3}, {-1, 3}}], 111->
                GraphicsBox[{InsetBox[
                   TagBox[GridBox[{
                    {"\<\"t\"\>", "5.5`"},
                    {"\<\"driveIntent\"\>", "0.5`"},
                    {"\<\"motorTorque\"\>", 
                    RowBox[{"-", "9.875521278676036`"}]},
                    {"\<\"vw\"\>", "0.34639330373785165`"},
                    {"\<\"controllerState\"\>", 
                    RowBox[{"\[LeftAssociation]", 
                    RowBox[{
                    RowBox[{"\<\"lastError\"\>", "\[Rule]", 
                    "0.026174346786004634`"}], ",", 
                    RowBox[{"\<\"lastInt\"\>", "\[Rule]", 
                    RowBox[{"-", "0.13085222505053715`"}]}], ",", 
                    RowBox[{"\<\"initialized\"\>", "\[Rule]", "True"}]}], 
                    "\[RightAssociation]"}]}
                    },
                    AutoDelete->False,
                    
                    GridBoxItemSize->{
                    "Columns" -> {{Automatic}}, "Rows" -> {{Automatic}}}],
                    "Grid"], {-1, -0.5}], 
                  {GrayLevel[0], LineBox[{{-5, 0}, {5, 0}}]}, 
                  {GrayLevel[0.5], 
                   CircleBox[{-1.1148911028426198`, 0.5}, 0.5], 
                   DiskBox[{-1.1148911028426198`, 0.5}, 0.1]}, 
                  {GrayLevel[0.5], 
                   DiskBox[{-1.085001200916424, 1.9997021683530507`}, 0.1]}, 
                  {GrayLevel[0.5], Thickness[0.01], 
                   LineBox[{{-1.1148911028426198`, 0.5}, {-1.085001200916424, 
                    1.9997021683530507`}}]}, 
                  {RGBColor[0, 1, 0], 
                   DiskBox[{-1.0949645015584892`, 1.4998014455687005`}, 
                    0.05]}, 
                  {RGBColor[0, 0, 1], Arrowheads[0.01], 
                   CircleBox[{-1.1148911028426198`, 0.5}, 0.3333333333333333, 
                    NCache[{
                    Rational[-1, 2] 
                    Pi, -1.8670619651551776`}, {-1.5707963267948966`, \
-1.8670619651551776`}]], 
                   ArrowBox[{{-1.2090150396382418`, 
                    0.18023165317185497`}, {-1.2122079636544436`, 
                    0.18118886513667187`}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-1.1148911028426198`, 
                    0.5}, {-0.9173806772690991, 0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-1.1148911028426198`, 
                    0.5}, {-1.2614219060866207`, 0.5}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.005], Arrowheads[0.04], 
                   ArrowBox[{{-1.1148911028426198`, 0.5}, {-1.0639114805131, 
                    0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-1.1148911028426198`, 0.5}, 1.6, {
                    1.550868406567684, 1.4592684669662688`}], 
                   ArrowBox[{{-0.9527254605618394, 
                    2.091760756038313}, {-0.9368162265081343, 
                    2.090059539268407}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Dashing[{Small, Small}], 
                   Arrowheads[0.02], 
                   CircleBox[{-1.1148911028426198`, 0.5}, 1.7, {
                    1.550868406567684, 1.58040725079613}], 
                   ArrowBox[{{-1.1142296736614044`, 
                    2.199999871326889}, {-1.1312294221139616`, 
                    2.1999214862232277`}}]}, 
                  {RGBColor[0, 0, 1], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-1.1148911028426198`, 0.5}, 1.8, {
                    1.550868406567684, 1.6057324136714397`}], 
                   ArrowBox[{{-1.159771407724317, 
                    2.299440401411985}, {-1.1777632678366587`, 
                    2.298901634572875}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.003], Arrowheads[0.025], 
                   CircleBox[{-1.1148911028426198`, 0.5}, 1.5, {
                    1.550868406567684, 1.55247596964292}], 
                   ArrowBox[{{-1.0724162454203656`, 
                    1.999398508231537}, {-1.0874121043293523`, 
                    1.999748280425988}}]}, 
                  {RGBColor[0, 0, 1], ArrowBox[{{0, 2.8}, {1.5, 2.8}}]}},
                 Frame->True,
                 FrameTicks->Automatic,
                 ImageSize->800,
                 PlotRange->{{-3, 3}, {-1, 3}}], 112->
                GraphicsBox[{InsetBox[
                   TagBox[GridBox[{
                    {"\<\"t\"\>", "5.55`"},
                    {"\<\"driveIntent\"\>", "0.5`"},
                    {"\<\"motorTorque\"\>", 
                    RowBox[{"-", "10.246183847486384`"}]},
                    {"\<\"vw\"\>", "0.3472467978185024`"},
                    {"\<\"controllerState\"\>", 
                    RowBox[{"\[LeftAssociation]", 
                    RowBox[{
                    
                    RowBox[{"\<\"lastError\"\>", "\[Rule]", 
                    "0.025178243960086905`"}], ",", 
                    RowBox[{"\<\"lastInt\"\>", "\[Rule]", 
                    RowBox[{"-", "0.12956905431358584`"}]}], ",", 
                    RowBox[{"\<\"initialized\"\>", "\[Rule]", "True"}]}], 
                    "\[RightAssociation]"}]}
                    },
                    AutoDelete->False,
                    
                    GridBoxItemSize->{
                    "Columns" -> {{Automatic}}, "Rows" -> {{Automatic}}}],
                    "Grid"], {-1, -0.5}], 
                  {GrayLevel[0], LineBox[{{-5, 0}, {5, 0}}]}, 
                  {GrayLevel[0.5], 
                   CircleBox[{-1.0975498091851925`, 0.5}, 0.5], 
                   DiskBox[{-1.0975498091851925`, 0.5}, 0.1]}, 
                  {GrayLevel[0.5], 
                   DiskBox[{-1.0665511062413646`, 1.9996796592658714`}, 0.1]}, 
                  {GrayLevel[0.5], Thickness[0.01], 
                   LineBox[{{-1.0975498091851925`, 
                    0.5}, {-1.0665511062413646`, 1.9996796592658714`}}]}, 
                  {RGBColor[0, 1, 0], 
                   DiskBox[{-1.0768840072226404`, 1.499786439510581}, 0.05]}, 
                  {RGBColor[0, 0, 1], Arrowheads[0.01], 
                   CircleBox[{-1.0975498091851925`, 0.5}, 0.3333333333333333, 
                    NCache[{
                    Rational[-1, 2] 
                    Pi, -1.878181842219488}, {-1.5707963267948966`, \
-1.878181842219488}]], 
                   ArrowBox[{{-1.1952236381764398`, 
                    0.18129804794871485`}, {-1.198405720929474, 
                    0.18229070492454774`}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-1.0975498091851925`, 
                    0.5}, {-0.8926261322354648, 0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-1.0975498091851925`, 
                    0.5}, {-1.2495140638043682`, 0.5}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.005], Arrowheads[0.04], 
                   ArrowBox[{{-1.0975498091851925`, 
                    0.5}, {-1.0445903868546407`, 0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-1.0975498091851925`, 0.5}, 1.6, {
                    1.5501290535737355`, 1.4551311066287074`}], 
                   ArrowBox[{{-0.9287998858276971, 
                    2.091076196593626}, {-0.9128978264656568, 
                    2.089309172338014}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Dashing[{Small, Small}], 
                   Arrowheads[0.02], 
                   CircleBox[{-1.0975498091851925`, 0.5}, 1.7, {
                    1.5501290535737355`, 1.5808145717287132`}], 
                   ArrowBox[{{-1.0975808255726789`, 
                    2.199999999717054}, {-1.1145805406871265`, 
                    2.1999146902666933`}}]}, 
                  {RGBColor[0, 0, 1], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-1.0975498091851925`, 0.5}, 1.8, {
                    1.5501290535737355`, 1.6070522971708823`}], 
                   ArrowBox[{{-1.144805125978536, 
                    2.2993795972597777`}, {-1.1627962593098837`, 
                    2.2988170837375677`}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.003], Arrowheads[0.025], 
                   CircleBox[{-1.0975498091851925`, 0.5}, 1.5, {
                    1.5501290535737355`, 1.5518827230579226`}], 
                   ArrowBox[{{-1.0541854462448883`, 
                    1.999373046318621}, {-1.069181095013895, 
                    1.999731714693087}}]}, 
                  {RGBColor[0, 0, 1], ArrowBox[{{0, 2.8}, {1.5, 2.8}}]}},
                 Frame->True,
                 FrameTicks->Automatic,
                 ImageSize->800,
                 PlotRange->{{-3, 3}, {-1, 3}}], 113->
                GraphicsBox[{InsetBox[
                   TagBox[GridBox[{
                    {"\<\"t\"\>", "5.6000000000000005`"},
                    {"\<\"driveIntent\"\>", "0.5`"},
                    {"\<\"motorTorque\"\>", 
                    RowBox[{"-", "10.592176741598083`"}]},
                    {"\<\"vw\"\>", "0.34813163085142435`"},
                    {"\<\"controllerState\"\>", 
                    RowBox[{"\[LeftAssociation]", 
                    RowBox[{
                    
                    RowBox[{"\<\"lastError\"\>", "\[Rule]", 
                    "0.024219344687802388`"}], ",", 
                    RowBox[{"\<\"lastInt\"\>", "\[Rule]", 
                    RowBox[{"-", "0.12833475748325918`"}]}], ",", 
                    RowBox[{"\<\"initialized\"\>", "\[Rule]", "True"}]}], 
                    "\[RightAssociation]"}]}
                    },
                    AutoDelete->False,
                    
                    GridBoxItemSize->{
                    "Columns" -> {{Automatic}}, "Rows" -> {{Automatic}}}],
                    "Grid"], {-1, -0.5}], 
                  {GrayLevel[0], LineBox[{{-5, 0}, {5, 0}}]}, 
                  {GrayLevel[0.5], 
                   CircleBox[{-1.0801650314246551`, 0.5}, 0.5], 
                   DiskBox[{-1.0801650314246551`, 0.5}, 0.1]}, 
                  {GrayLevel[0.5], 
                   DiskBox[{-1.04812756175788, 1.9996578278185162`}, 0.1]}, 
                  {GrayLevel[0.5], Thickness[0.01], 
                   LineBox[{{-1.0801650314246551`, 0.5}, {-1.04812756175788, 
                    1.9996578278185162`}}]}, 
                  {RGBColor[0, 1, 0], 
                   DiskBox[{-1.0588067183134717`, 1.4997718852123443`}, 
                    0.05]}, 
                  {RGBColor[0, 0, 1], Arrowheads[0.01], 
                   CircleBox[{-1.0801650314246551`, 0.5}, 0.3333333333333333, 
                    NCache[{
                    Rational[-1, 2] 
                    Pi, -1.888561629042839}, {-1.5707963267948966`, \
-1.888561629042839}]], 
                   ArrowBox[{{-1.1811415976963329`, 
                    0.1823290315828458}, {-1.1843132056493675`, 
                    0.1833546638322774}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-1.0801650314246551`, 
                    0.5}, {-0.8683214965926935, 0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-1.0801650314246551`, 
                    0.5}, {-1.2372193227977704`, 0.5}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.005], Arrowheads[0.04], 
                   ArrowBox[{{-1.0801650314246551`, 
                    0.5}, {-1.0253757879658085`, 0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-1.0801650314246551`, 0.5}, 1.6, {
                    1.5494363894864909`, 1.451255060765341}], 
                   ArrowBox[{{-0.9052493068243294, 
                    2.090410163853194}, {-0.8893542159661814, 
                    2.088581515914142}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Dashing[{Small, Small}], 
                   Arrowheads[0.02], 
                   CircleBox[{-1.0801650314246551`, 0.5}, 1.7, {
                    1.5494363894864909`, 1.5811816622590826`}], 
                   ArrowBox[{{-1.0808201016975603`, 
                    2.1999998737890945`}, {-1.0978197843503148`, 
                    2.1999083239101846`}}]}, 
                  {RGBColor[0, 0, 1], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-1.0801650314246551`, 0.5}, 1.8, {
                    1.5494363894864909`, 1.6082818158287024`}], 
                   ArrowBox[{{-1.12963268272961, 
                    2.299320135905329}, {-1.1476231108415196`, 
                    2.298735502379766}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.003], Arrowheads[0.025], 
                   CircleBox[{-1.0801650314246551`, 0.5}, 1.5, {
                    1.5494363894864909`, 1.551302126270189}], 
                   ArrowBox[{{-1.0359301446678555`, 
                    1.9993476163964155`}, {-1.0509255826677064`, 
                    1.9997149911354457`}}]}, 
                  {RGBColor[0, 0, 1], ArrowBox[{{0, 2.8}, {1.5, 2.8}}]}},
                 Frame->True,
                 FrameTicks->Automatic,
                 ImageSize->800,
                 PlotRange->{{-3, 3}, {-1, 3}}], 114->
                GraphicsBox[{InsetBox[
                   TagBox[GridBox[{
                    {"\<\"t\"\>", "5.65`"},
                    {"\<\"driveIntent\"\>", "0.5`"},
                    {"\<\"motorTorque\"\>", 
                    RowBox[{"-", "10.912172589081012`"}]},
                    {"\<\"vw\"\>", "0.34904530001944356`"},
                    {"\<\"controllerState\"\>", 
                    RowBox[{"\[LeftAssociation]", 
                    RowBox[{
                    
                    RowBox[{"\<\"lastError\"\>", "\[Rule]", 
                    "0.023301411499200553`"}], ",", 
                    RowBox[{"\<\"lastInt\"\>", "\[Rule]", 
                    RowBox[{"-", "0.1271473749032171`"}]}], ",", 
                    RowBox[{"\<\"initialized\"\>", "\[Rule]", "True"}]}], 
                    "\[RightAssociation]"}]}
                    },
                    AutoDelete->False,
                    
                    GridBoxItemSize->{
                    "Columns" -> {{Automatic}}, "Rows" -> {{Automatic}}}],
                    "Grid"], {-1, -0.5}], 
                  {GrayLevel[0], LineBox[{{-5, 0}, {5, 0}}]}, 
                  {GrayLevel[0.5], CircleBox[{-1.062735266117951, 0.5}, 0.5], 
                   DiskBox[{-1.062735266117951, 0.5}, 0.1]}, 
                  {GrayLevel[0.5], 
                   DiskBox[{-1.0297335524704951`, 1.999636918356017}, 0.1]}, 
                  {GrayLevel[0.5], Thickness[0.01], 
                   LineBox[{{-1.062735266117951, 0.5}, {-1.0297335524704951`, 
                    1.999636918356017}}]}, 
                  {RGBColor[0, 1, 0], 
                   DiskBox[{-1.0407341236863137`, 1.499757945570678}, 0.05]}, 
                  {RGBColor[0, 0, 1], Arrowheads[0.01], 
                   CircleBox[{-1.062735266117951, 0.5}, 0.3333333333333333, 
                    NCache[{
                    Rational[-1, 2] 
                    Pi, -1.898161504467327}, {-1.5707963267948966`, \
-1.898161504467327}]], 
                   ArrowBox[{{-1.1667567344282777`, 
                    0.18331301693679486`}, {-1.1699183504479365`, 
                    0.18436904850027414`}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-1.062735266117951, 0.5}, {-0.8444918143363307, 
                    0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-1.062735266117951, 0.5}, {-1.2245142248938394`,
                     0.5}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.005], Arrowheads[0.04], 
                   ArrowBox[{{-1.062735266117951, 0.5}, {-1.0062707731122194`,
                     0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-1.062735266117951, 0.5}, 1.6, {
                    1.5487934090333757`, 1.4476570793018684`}], 
                   ArrowBox[{{-0.8820984197569581, 
                    2.0897705273833562`}, {-0.8662100112106073, 
                    2.0878847011617703`}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Dashing[{Small, Small}], 
                   Arrowheads[0.02], 
                   CircleBox[{-1.062735266117951, 0.5}, 1.7, {
                    1.5487934090333757`, 1.581508866902412}], 
                   ArrowBox[{{-1.0639465841982267`, 
                    2.199999568443624}, {-1.080946235985419, 
                    2.199902456194615}}]}, 
                  {RGBColor[0, 0, 1], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-1.062735266117951, 0.5}, 1.8, {
                    1.5487934090333757`, 1.6094165900838258`}], 
                   ArrowBox[{{-1.114244707301587, 
                    2.299262842796724}, {-1.132234460403316, 
                    2.298657794577302}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.003], Arrowheads[0.025], 
                   CircleBox[{-1.062735266117951, 0.5}, 1.5, {
                    1.5487934090333757`, 1.550739993498549}], 
                   ArrowBox[{{-1.0176575539632458`, 
                    1.9993225136263704`}, {-1.0326527830804988`, 
                    1.9996983177340373`}}]}, 
                  {RGBColor[0, 0, 1], ArrowBox[{{0, 2.8}, {1.5, 2.8}}]}},
                 Frame->True,
                 FrameTicks->Automatic,
                 ImageSize->800,
                 PlotRange->{{-3, 3}, {-1, 3}}], 115->
                GraphicsBox[{InsetBox[
                   TagBox[GridBox[{
                    {"\<\"t\"\>", "5.7`"},
                    {"\<\"driveIntent\"\>", "0.5`"},
                    {"\<\"motorTorque\"\>", 
                    RowBox[{"-", "11.205252137860978`"}]},
                    {"\<\"vw\"\>", "0.3499852433387758`"},
                    {"\<\"controllerState\"\>", 
                    RowBox[{"\[LeftAssociation]", 
                    RowBox[{
                    
                    RowBox[{"\<\"lastError\"\>", "\[Rule]", 
                    "0.022427390805794156`"}], ",", 
                    RowBox[{"\<\"lastInt\"\>", "\[Rule]", 
                    RowBox[{"-", "0.12600477987192274`"}]}], ",", 
                    RowBox[{"\<\"initialized\"\>", "\[Rule]", "True"}]}], 
                    "\[RightAssociation]"}]}
                    },
                    AutoDelete->False,
                    
                    GridBoxItemSize->{
                    "Columns" -> {{Automatic}}, "Rows" -> {{Automatic}}}],
                    "Grid"], {-1, -0.5}], 
                  {GrayLevel[0], LineBox[{{-5, 0}, {5, 0}}]}, 
                  {GrayLevel[0.5], 
                   CircleBox[{-1.0452591366114197`, 0.5}, 0.5], 
                   DiskBox[{-1.0452591366114197`, 0.5}, 0.1]}, 
                  {GrayLevel[0.5], 
                   DiskBox[{-1.0113709414182814`, 1.999617147883603}, 0.1]}, 
                  {GrayLevel[0.5], Thickness[0.01], 
                   LineBox[{{-1.0452591366114197`, 
                    0.5}, {-1.0113709414182814`, 1.999617147883603}}]}, 
                  {RGBColor[0, 1, 0], 
                   DiskBox[{-1.0226670064826608`, 1.4997447652557354`}, 
                    0.05]}, 
                  {RGBColor[0, 0, 1], Arrowheads[0.01], 
                   CircleBox[{-1.0452591366114197`, 0.5}, 0.3333333333333333, 
                    NCache[{
                    Rational[-1, 2] 
                    Pi, -1.9069538909307258`}, {-1.5707963267948966`, \
-1.9069538909307258`}]], 
                   ArrowBox[{{-1.1520609826702781`, 
                    0.18423984293528894`}, {-1.1552131915666932`, 
                    0.1853236314719458}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-1.0452591366114197`, 
                    0.5}, {-0.8211540938542001, 0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-1.0452591366114197`, 
                    0.5}, {-1.2113815602677247`, 0.5}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.005], Arrowheads[0.04], 
                   ArrowBox[{{-1.0452591366114197`, 0.5}, {-0.987276517510505,
                     0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-1.0452591366114197`, 0.5}, 1.6, {
                    1.5482022743710298`, 1.4443492526864083`}], 
                   ArrowBox[{{-0.8593646028150278, 
                    2.0891643157032638`}, {-0.8434825191666241, 
                    2.0872259437939316`}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Dashing[{Small, Small}], 
                   Arrowheads[0.02], 
                   CircleBox[{-1.0452591366114197`, 0.5}, 1.7, {
                    1.5482022743710298`, 1.5817968799133082`}], 
                   ArrowBox[{{-1.0469600766289155`, 
                    2.1999991490595097`}, {-1.0639596997414438`, 
                    2.199897140693701}}]}, 
                  {RGBColor[0, 0, 1], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-1.0452591366114197`, 0.5}, 1.8, {
                    1.5482022743710298`, 1.610453675136924}], 
                   ArrowBox[{{-1.0986345383610538`, 
                    2.2992084555403984`}, {-1.1166236543020334`, 
                    2.2985847507456487`}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.003], Arrowheads[0.025], 
                   CircleBox[{-1.0452591366114197`, 0.5}, 1.5, {
                    1.5482022743710298`, 1.5502012851077556`}], 
                   ArrowBox[{{-0.9993737334181124, 
                    1.9992980123290325`}, {-1.014368757910691, 
                    1.9996818944375256`}}]}, 
                  {RGBColor[0, 0, 1], ArrowBox[{{0, 2.8}, {1.5, 2.8}}]}},
                 Frame->True,
                 FrameTicks->Automatic,
                 ImageSize->800,
                 PlotRange->{{-3, 3}, {-1, 3}}], 116->
                GraphicsBox[{InsetBox[
                   TagBox[GridBox[{
                    {"\<\"t\"\>", "5.75`"},
                    {"\<\"driveIntent\"\>", "0.5`"},
                    {"\<\"motorTorque\"\>", 
                    RowBox[{"-", "11.470866295156476`"}]},
                    {"\<\"vw\"\>", "0.3509488715106874`"},
                    {"\<\"controllerState\"\>", 
                    RowBox[{"\[LeftAssociation]", 
                    RowBox[{
                    
                    RowBox[{"\<\"lastError\"\>", "\[Rule]", 
                    "0.02159947790707863`"}], ",", 
                    RowBox[{"\<\"lastInt\"\>", "\[Rule]", 
                    RowBox[{"-", "0.12490471779146456`"}]}], ",", 
                    RowBox[{"\<\"initialized\"\>", "\[Rule]", "True"}]}], 
                    "\[RightAssociation]"}]}
                    },
                    AutoDelete->False,
                    
                    GridBoxItemSize->{
                    "Columns" -> {{Automatic}}, "Rows" -> {{Automatic}}}],
                    "Grid"], {-1, -0.5}], 
                  {GrayLevel[0], LineBox[{{-5, 0}, {5, 0}}]}, 
                  {GrayLevel[0.5], 
                   CircleBox[{-1.0277353951727022`, 0.5}, 0.5], 
                   DiskBox[{-1.0277353951727022`, 0.5}, 0.1]}, 
                  {GrayLevel[0.5], 
                   DiskBox[{-0.9930405837693502, 1.9995987030074698`}, 0.1]}, 
                  {GrayLevel[0.5], Thickness[0.01], 
                   LineBox[{{-1.0277353951727022`, 0.5}, {-0.9930405837693502,
                     1.9995987030074698`}}]}, 
                  {RGBColor[0, 1, 0], 
                   DiskBox[{-1.004605520903801, 1.4997324686716467`}, 0.05]}, 
                  {RGBColor[0, 0, 1], Arrowheads[0.01], 
                   CircleBox[{-1.0277353951727022`, 0.5}, 0.3333333333333333, 
                    NCache[{
                    Rational[-1, 2] 
                    Pi, -1.9149223156495907`}, {-1.5707963267948966`, \
-1.9149223156495907`}]], 
                   ArrowBox[{{-1.1370499349293683`, 
                    0.18510090106686655`}, {-1.1401934077543388`, 
                    0.18620977306917352`}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-1.0277353951727022`, 
                    0.5}, {-0.7983180692695726, 0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-1.0277353951727022`, 0.5}, {-1.197809818884552,
                     0.5}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.005], Arrowheads[0.04], 
                   ArrowBox[{{-1.0277353951727022`, 
                    0.5}, {-0.9683924929814225, 0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-1.0277353951727022`, 0.5}, 1.6, {
                    1.5476643896499107`, 1.4413394295721977`}], 
                   ArrowBox[{{-0.8370586071202667, 
                    2.0885976087411207`}, {-0.821182429557754, 
                    2.0866114434213783`}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Dashing[{Small, Small}], 
                   Arrowheads[0.02], 
                   CircleBox[{-1.0277353951727022`, 0.5}, 1.7, {
                    1.5476643896499107`, 1.582046698647299}], 
                   ArrowBox[{{-1.0298610267679067`, 
                    2.1999986710848694`}, {-1.0468606238663662`, 
                    2.1998924158979634`}}]}, 
                  {RGBColor[0, 0, 1], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-1.0277353951727022`, 0.5}, 1.8, {
                    1.5476643896499107`, 1.6113914246230023`}], 
                   ArrowBox[{{-1.0827979800111147`, 
                    2.299157611703575}, {-1.1007865031637487`, 
                    2.298517037901304}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.003], Arrowheads[0.025], 
                   CircleBox[{-1.0277353951727022`, 0.5}, 1.5, {
                    1.5476643896499107`, 1.5496901577317668`}], 
                   ArrowBox[{{-0.9810836657477036, 
                    1.9992743631976293`}, {-0.9960784920689019, 
                    1.9996659096231655`}}]}, 
                  {RGBColor[0, 0, 1], ArrowBox[{{0, 2.8}, {1.5, 2.8}}]}},
                 Frame->True,
                 FrameTicks->Automatic,
                 ImageSize->800,
                 PlotRange->{{-3, 3}, {-1, 3}}], 117->
                GraphicsBox[{InsetBox[
                   TagBox[GridBox[{
                    {"\<\"t\"\>", "5.8`"},
                    {"\<\"driveIntent\"\>", "0.5`"},
                    {"\<\"motorTorque\"\>", 
                    RowBox[{"-", "11.708798932330136`"}]},
                    {"\<\"vw\"\>", "0.35193359640921573`"},
                    {"\<\"controllerState\"\>", 
                    RowBox[{"\[LeftAssociation]", 
                    RowBox[{
                    
                    RowBox[{"\<\"lastError\"\>", "\[Rule]", 
                    "0.02081918180666133`"}], ",", 
                    RowBox[{"\<\"lastInt\"\>", "\[Rule]", 
                    RowBox[{"-", "0.12384484206824094`"}]}], ",", 
                    RowBox[{"\<\"initialized\"\>", "\[Rule]", "True"}]}], 
                    "\[RightAssociation]"}]}
                    },
                    AutoDelete->False,
                    
                    GridBoxItemSize->{
                    "Columns" -> {{Automatic}}, "Rows" -> {{Automatic}}}],
                    "Grid"], {-1, -0.5}], 
                  {GrayLevel[0], LineBox[{{-5, 0}, {5, 0}}]}, 
                  {GrayLevel[0.5], 
                   CircleBox[{-1.0101629236162513`, 0.5}, 0.5], 
                   DiskBox[{-1.0101629236162513`, 0.5}, 0.1]}, 
                  {GrayLevel[0.5], 
                   DiskBox[{-0.9747424375722219, 1.9995817380751222`}, 0.1]}, 
                  {GrayLevel[0.5], Thickness[0.01], 
                   LineBox[{{-1.0101629236162513`, 0.5}, {-0.9747424375722219,
                     1.9995817380751222`}}]}, 
                  {RGBColor[0, 1, 0], 
                   DiskBox[{-0.986549266253565, 1.4997211587167483`}, 0.05]}, 
                  {RGBColor[0, 0, 1], Arrowheads[0.01], 
                   CircleBox[{-1.0101629236162513`, 0.5}, 0.3333333333333333, 
                    NCache[{
                    Rational[-1, 2] 
                    Pi, -1.9220602947648007`}, {-1.5707963267948966`, \
-1.9220602947648007`}]], 
                   ArrowBox[{{-1.1217224026603414`, 
                    0.1858892014805567}, {-1.1248578803665288`, 
                    0.187020483086891}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-1.0101629236162513`, 
                    0.5}, {-0.7759869449696486, 0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-1.0101629236162513`, 
                    0.5}, {-1.1837926497085536`, 0.5}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.005], Arrowheads[0.04], 
                   ArrowBox[{{-1.0101629236162513`, 
                    0.5}, {-0.9496166710619507, 0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-1.0101629236162513`, 0.5}, 1.6, {
                    1.547180474366538, 1.4386316276591375`}], 
                   ArrowBox[{{-0.8151852322153901, 
                    2.0880754704534636`}, {-0.7993144909921067, 
                    2.086046322923748}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Dashing[{Small, Small}], 
                   Arrowheads[0.02], 
                   CircleBox[{-1.0101629236162513`, 0.5}, 1.7, {
                    1.547180474366538, 1.58225957927104}], 
                   ArrowBox[{{-1.0126504519380173`, 
                    2.199998180058687}, {-1.029650026031611, 
                    2.199888305989383}}]}, 
                  {RGBColor[0, 0, 1], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-1.0101629236162513`, 0.5}, 1.8, {
                    1.547180474366538, 1.6122293573239275`}], 
                   ArrowBox[{{-1.0667330619446311`, 
                    2.2991108413462213`}, {-1.0847210420244402`, 
                    2.298455194598806}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.003], Arrowheads[0.025], 
                   CircleBox[{-1.0101629236162513`, 0.5}, 1.5, {
                    1.547180474366538, 1.549210016319281}], 
                   ArrowBox[{{-0.9627913358859086, 
                    1.9992517909530427`}, {-0.9777859724810383, 
                    1.9996505369702597`}}]}, 
                  {RGBColor[0, 0, 1], ArrowBox[{{0, 2.8}, {1.5, 2.8}}]}},
                 Frame->True,
                 FrameTicks->Automatic,
                 ImageSize->800,
                 PlotRange->{{-3, 3}, {-1, 3}}], 118->
                GraphicsBox[{InsetBox[
                   TagBox[GridBox[{
                    {"\<\"t\"\>", "5.8500000000000005`"},
                    {"\<\"driveIntent\"\>", "0.5`"},
                    {"\<\"motorTorque\"\>", 
                    RowBox[{"-", "11.919130729922898`"}]},
                    {"\<\"vw\"\>", "0.3529368563116628`"},
                    {"\<\"controllerState\"\>", 
                    RowBox[{"\[LeftAssociation]", 
                    RowBox[{
                    
                    RowBox[{"\<\"lastError\"\>", "\[Rule]", 
                    "0.020087389236003016`"}], ",", 
                    RowBox[{"\<\"lastInt\"\>", "\[Rule]", 
                    RowBox[{"-", "0.12282274679069512`"}]}], ",", 
                    RowBox[{"\<\"initialized\"\>", "\[Rule]", "True"}]}], 
                    "\[RightAssociation]"}]}
                    },
                    AutoDelete->False,
                    
                    GridBoxItemSize->{
                    "Columns" -> {{Automatic}}, "Rows" -> {{Automatic}}}],
                    "Grid"], {-1, -0.5}], 
                  {GrayLevel[0], LineBox[{{-5, 0}, {5, 0}}]}, 
                  {GrayLevel[0.5], CircleBox[{-0.9925407325879952, 0.5}, 0.5],
                    DiskBox[{-0.9925407325879952, 0.5}, 0.1]}, 
                  {GrayLevel[0.5], 
                   DiskBox[{-0.9564756704740208, 1.999566374421191}, 0.1]}, 
                  {GrayLevel[0.5], Thickness[0.01], 
                   LineBox[{{-0.9925407325879952, 0.5}, {-0.9564756704740208, 
                    1.999566374421191}}]}, 
                  {RGBColor[0, 1, 0], 
                   DiskBox[{-0.9684973578453456, 1.4997109162807938`}, 0.05]}, 
                  {RGBColor[0, 0, 1], Arrowheads[0.01], 
                   CircleBox[{-0.9925407325879952, 0.5}, 0.3333333333333333, 
                    NCache[{
                    Rational[-1, 2] 
                    Pi, -1.9283702486925836`}, {-1.5707963267948966`, \
-1.9283702486925836`}]], 
                   ArrowBox[{{-1.1060800022545458`, 
                    0.18659938520370223`}, {-1.1092082792531588`, 
                    0.18775042887740745`}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-0.9925407325879952, 0.5}, {-0.7541581179895372,
                     0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-0.9925407325879952, 0.5}, {-1.169328332082922, 
                    0.5}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.005], Arrowheads[0.04], 
                   ArrowBox[{{-0.9925407325879952, 0.5}, {-0.9309457174844641,
                     0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-0.9925407325879952, 0.5}, 1.6, {
                    1.5467506349349107`, 1.4362264345036586`}], 
                   ArrowBox[{{-0.7937439805861706, 
                    2.087601918427137}, {-0.7778781657556644, 
                    2.0855346046053223`}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Dashing[{Small, Small}], 
                   Arrowheads[0.02], 
                   CircleBox[{-0.9925407325879952, 0.5}, 1.7, {
                    1.5467506349349107`, 1.5824369949889012`}], 
                   ArrowBox[{{-0.9953298672665075, 
                    2.1999977119772085`}, {-1.0123294215991725`, 
                    2.1998848219180083`}}]}, 
                  {RGBColor[0, 0, 1], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-0.9925407325879952, 0.5}, 1.8, {
                    1.5467506349349107`, 1.612968027878927}], 
                   ArrowBox[{{-1.0504398055659043`, 
                    2.299068563826376}, {-1.0684272964313823`, 
                    2.298399630067812}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.003], Arrowheads[0.025], 
                   CircleBox[{-0.9925407325879952, 0.5}, 1.5, {
                    1.5467506349349107`, 1.5487635674022504`}], 
                   ArrowBox[{{-0.9444998102624331, 
                    1.9992304925467963`}, {-0.959494267343501, 
                    1.9996359328633215`}}]}, 
                  {RGBColor[0, 0, 1], ArrowBox[{{0, 2.8}, {1.5, 2.8}}]}},
                 Frame->True,
                 FrameTicks->Automatic,
                 ImageSize->800,
                 PlotRange->{{-3, 3}, {-1, 3}}], 119->
                GraphicsBox[{InsetBox[
                   TagBox[GridBox[{
                    {"\<\"t\"\>", "5.9`"},
                    {"\<\"driveIntent\"\>", "0.5`"},
                    {"\<\"motorTorque\"\>", 
                    RowBox[{"-", "12.102204296588596`"}]},
                    {"\<\"vw\"\>", "0.3539561379973992`"},
                    {"\<\"controllerState\"\>", 
                    RowBox[{"\[LeftAssociation]", 
                    RowBox[{
                    
                    RowBox[{"\<\"lastError\"\>", "\[Rule]", 
                    "0.01940442735825078`"}], ",", 
                    RowBox[{"\<\"lastInt\"\>", "\[Rule]", 
                    RowBox[{"-", "0.12183599623747386`"}]}], ",", 
                    RowBox[{"\<\"initialized\"\>", "\[Rule]", "True"}]}], 
                    "\[RightAssociation]"}]}
                    },
                    AutoDelete->False,
                    
                    GridBoxItemSize->{
                    "Columns" -> {{Automatic}}, "Rows" -> {{Automatic}}}],
                    "Grid"], {-1, -0.5}], 
                  {GrayLevel[0], LineBox[{{-5, 0}, {5, 0}}]}, 
                  {GrayLevel[0.5], CircleBox[{-0.9748679596688942, 0.5}, 0.5],
                    DiskBox[{-0.9748679596688942, 0.5}, 0.1]}, 
                  {GrayLevel[0.5], 
                   DiskBox[{-0.9382387618313566, 1.9995527005963405`}, 0.1]}, 
                  {GrayLevel[0.5], Thickness[0.01], 
                   LineBox[{{-0.9748679596688942, 0.5}, {-0.9382387618313566, 
                    1.9995527005963405`}}]}, 
                  {RGBColor[0, 1, 0], 
                   DiskBox[{-0.9504484944438691, 1.4997018003975604`}, 0.05]}, 
                  {RGBColor[0, 0, 1], Arrowheads[0.01], 
                   CircleBox[{-0.9748679596688942, 0.5}, 0.3333333333333333, 
                    NCache[{
                    Rational[-1, 2] 
                    Pi, -1.9338624556925543`}, {-1.5707963267948966`, \
-1.9338624556925543`}]], 
                   ArrowBox[{{-1.0901267693182537`, 
                    0.18722768998946887`}, {-1.0932486773974432`, 
                    0.18839589736143608`}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-0.9748679596688942, 0.5}, {-0.7328238737371222,
                     0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-0.9748679596688942, 
                    0.5}, {-1.1544192630573464`, 0.5}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.005], Arrowheads[0.04], 
                   ArrowBox[{{-0.9748679596688942, 0.5}, {-0.9123751771255745,
                     0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-0.9748679596688942, 0.5}, 1.6, {
                    1.5463744339887002`, 1.4341213955594854`}], 
                   ArrowBox[{{-0.7727296867219047, 
                    2.0871799263505095`}, {-0.7568682588164879, 
                    2.0850792189755882`}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Dashing[{Small, Small}], 
                   Arrowheads[0.02], 
                   CircleBox[{-0.9748679596688942, 0.5}, 1.7, {
                    1.5463744339887002`, 1.5825805969124092`}], 
                   ArrowBox[{{-0.9779012172592083, 
                    2.1999972939238432`}, {-0.9949007552053655, 
                    2.1998819626971144`}}]}, 
                  {RGBColor[0, 0, 1], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-0.9748679596688942, 0.5}, 1.8, {
                    1.5463744339887002`, 1.6136089023030813`}], 
                   ArrowBox[{{-1.0339199977075806`, 
                    2.29903108833713}, {-1.0519070561766393`, 
                    2.298350626993877}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.003], Arrowheads[0.025], 
                   CircleBox[{-0.9748679596688942, 0.5}, 1.5, {
                    1.5463744339887002`, 1.5483528729549985`}], 
                   ArrowBox[{{-0.9262113157737307, 
                    1.9992106359697623`}, {-0.9412056050781596, 
                    1.9996222343921843`}}]}, 
                  {RGBColor[0, 0, 1], ArrowBox[{{0, 2.8}, {1.5, 2.8}}]}},
                 Frame->True,
                 FrameTicks->Automatic,
                 ImageSize->800,
                 PlotRange->{{-3, 3}, {-1, 3}}], 120->
                GraphicsBox[{InsetBox[
                   TagBox[GridBox[{
                    {"\<\"t\"\>", "5.95`"},
                    {"\<\"driveIntent\"\>", "0.5`"},
                    {"\<\"motorTorque\"\>", 
                    RowBox[{"-", "12.2585907566706`"}]},
                    {"\<\"vw\"\>", "0.35498899585632027`"},
                    {"\<\"controllerState\"\>", 
                    RowBox[{"\[LeftAssociation]", 
                    RowBox[{
                    
                    RowBox[{"\<\"lastError\"\>", "\[Rule]", 
                    "0.018770124696199642`"}], ",", 
                    RowBox[{"\<\"lastInt\"\>", "\[Rule]", 
                    RowBox[{"-", "0.12088215129392582`"}]}], ",", 
                    RowBox[{"\<\"initialized\"\>", "\[Rule]", "True"}]}], 
                    "\[RightAssociation]"}]}
                    },
                    AutoDelete->False,
                    
                    GridBoxItemSize->{
                    "Columns" -> {{Automatic}}, "Rows" -> {{Automatic}}}],
                    "Grid"], {-1, -0.5}], 
                  {GrayLevel[0], LineBox[{{-5, 0}, {5, 0}}]}, 
                  {GrayLevel[0.5], CircleBox[{-0.9571438664504313, 0.5}, 0.5],
                    DiskBox[{-0.9571438664504313, 0.5}, 0.1]}, 
                  {GrayLevel[0.5], 
                   DiskBox[{-0.9200295997488387, 1.9995407734393897`}, 0.1]}, 
                  {GrayLevel[0.5], Thickness[0.01], 
                   LineBox[{{-0.9571438664504313, 0.5}, {-0.9200295997488387, 
                    1.9995407734393897`}}]}, 
                  {RGBColor[0, 1, 0], 
                   DiskBox[{-0.9324010219827029, 1.499693848959593}, 0.05]}, 
                  {RGBColor[0, 0, 1], Arrowheads[0.01], 
                   CircleBox[{-0.9571438664504313, 0.5}, 0.3333333333333333, 
                    NCache[{
                    Rational[-1, 2] 
                    Pi, -1.9385540494950146`}, {-1.5707963267948966`, \
-1.9385540494950146`}]], 
                   ArrowBox[{{-1.0738688028663284`, 
                    0.18777187774030363`}, {-1.0769851958529795`, 
                    0.18895471892642218`}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-0.9571438664504313, 0.5}, {-0.7119720513170192,
                     0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-0.9571438664504313, 
                    0.5}, {-1.1390714638447812`, 0.5}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.005], Arrowheads[0.04], 
                   ArrowBox[{{-0.9571438664504313, 0.5}, {-0.8938996487113693,
                     0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-0.9571438664504313, 0.5}, 1.6, {
                    1.5460509570020542`, 1.4323113871429378`}], 
                   ArrowBox[{{-0.7521331171623749, 
                    2.0868114546713956`}, {-0.7362755175349579, 
                    2.084682041435238}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Dashing[{Small, Small}], 
                   Arrowheads[0.02], 
                   CircleBox[{-0.9571438664504313, 0.5}, 1.7, {
                    1.5460509570020542`, 1.5826921776602716`}], 
                   ArrowBox[{{-0.9603668109908895, 
                    2.199996944887987}, {-0.9773663359624778, 
                    2.1998797168408224`}}]}, 
                  {RGBColor[0, 0, 1], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-0.9571438664504313, 0.5}, 1.8, {
                    1.5460509570020542`, 1.6141542389835577`}], 
                   ArrowBox[{{-1.0171769733014735`, 
                    2.2989986175875208`}, {-1.035163658015416, 
                    2.298308347343179}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.003], Arrowheads[0.025], 
                   CircleBox[{-0.9571438664504313, 0.5}, 1.5, {
                    1.5460509570020542`, 1.5479794042761137`}], 
                   ArrowBox[{{-0.9079273177459846, 
                    1.9991923596835808`}, {-0.9229214522856048, 
                    1.9996095579745854`}}]}, 
                  {RGBColor[0, 0, 1], ArrowBox[{{0, 2.8}, {1.5, 2.8}}]}},
                 Frame->True,
                 FrameTicks->Automatic,
                 ImageSize->800,
                 PlotRange->{{-3, 3}, {-1, 3}}], 121->
                GraphicsBox[{InsetBox[
                   TagBox[GridBox[{
                    {"\<\"t\"\>", "6.`"},
                    {"\<\"driveIntent\"\>", "0.5`"},
                    {"\<\"motorTorque\"\>", 
                    RowBox[{"-", "12.389057964446224`"}]},
                    {"\<\"vw\"\>", "0.35603306816122093`"},
                    {"\<\"controllerState\"\>", 
                    RowBox[{"\[LeftAssociation]", 
                    RowBox[{
                    
                    RowBox[{"\<\"lastError\"\>", "\[Rule]", 
                    "0.018183869896700247`"}], ",", 
                    RowBox[{"\<\"lastInt\"\>", "\[Rule]", 
                    RowBox[{"-", "0.11995879287589475`"}]}], ",", 
                    RowBox[{"\<\"initialized\"\>", "\[Rule]", "True"}]}], 
                    "\[RightAssociation]"}]}
                    },
                    AutoDelete->False,
                    
                    GridBoxItemSize->{
                    "Columns" -> {{Automatic}}, "Rows" -> {{Automatic}}}],
                    "Grid"], {-1, -0.5}], 
                  {GrayLevel[0], LineBox[{{-5, 0}, {5, 0}}]}, 
                  {GrayLevel[0.5], CircleBox[{-0.9393678347276616, 0.5}, 0.5],
                    DiskBox[{-0.9393678347276616, 0.5}, 0.1]}, 
                  {GrayLevel[0.5], 
                   DiskBox[{-0.9018455726820593, 1.999530619844417}, 0.1]}, 
                  {GrayLevel[0.5], Thickness[0.01], 
                   LineBox[{{-0.9393678347276616, 0.5}, {-0.9018455726820593, 
                    1.999530619844417}}]}, 
                  {RGBColor[0, 1, 0], 
                   DiskBox[{-0.9143529933639267, 1.4996870798962778`}, 0.05]}, 
                  {RGBColor[0, 0, 1], Arrowheads[0.01], 
                   CircleBox[{-0.9393678347276616, 0.5}, 0.3333333333333333, 
                    NCache[{
                    Rational[-1, 2] 
                    Pi, -1.9424680657282831`}, {-1.5707963267948966`, \
-1.9424680657282831`}]], 
                   ArrowBox[{{-1.057313939878287, 
                    0.18823113145968418`}, {-1.0604256793463582`, 
                    0.18942616116712824`}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-0.9393678347276616, 0.5}, {-0.6915866754387372,
                     0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-0.9393678347276616, 
                    0.5}, {-1.1232941080283836`, 0.5}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.005], Arrowheads[0.04], 
                   ArrowBox[{{-0.9393678347276616, 0.5}, {-0.8755129487394591,
                     0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-0.9393678347276616, 0.5}, 1.6, {
                    1.5457788758889428`, 1.4307889724234721`}], 
                   ArrowBox[{{-0.7319415388386283, 
                    2.086497504496542}, {-0.7160871994369586, 
                    2.0843439518943416`}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Dashing[{Small, Small}], 
                   Arrowheads[0.02], 
                   CircleBox[{-0.9393678347276616, 0.5}, 1.7, {
                    1.5457788758889428`, 1.5827736377403383`}], 
                   ArrowBox[{{-0.9427292611445172, 
                    2.199996676706294}, {-0.9597287765102971, 
                    2.199878063876855}}]}, 
                  {RGBColor[0, 0, 1], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-0.9393678347276616, 0.5}, 1.8, {
                    1.5457788758889428`, 1.6146069756914219`}], 
                   ArrowBox[{{-1.0002154081099501`, 
                    2.298971254026447}, {-1.0182017784698554`, 
                    2.298272840620703}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.003], Arrowheads[0.025], 
                   CircleBox[{-0.9393678347276616, 0.5}, 1.5, {
                    1.5457788758889428`, 1.5476440953869972`}], 
                   ArrowBox[{{-0.889648596274696, 
                    1.9991757726589825`}, {-0.9046425900811127, 
                    1.9995979985930354`}}]}, 
                  {RGBColor[0, 0, 1], ArrowBox[{{0, 2.8}, {1.5, 2.8}}]}},
                 Frame->True,
                 FrameTicks->Automatic,
                 ImageSize->800,
                 PlotRange->{{-3, 3}, {-1, 3}}], 122->
                GraphicsBox[{InsetBox[
                   TagBox[GridBox[{
                    {"\<\"t\"\>", "6.05`"},
                    {"\<\"driveIntent\"\>", "0.5`"},
                    {"\<\"motorTorque\"\>", 
                    RowBox[{"-", "12.494540468786251`"}]},
                    {"\<\"vw\"\>", "0.357086090668499`"},
                    {"\<\"controllerState\"\>", 
                    RowBox[{"\[LeftAssociation]", 
                    RowBox[{
                    
                    RowBox[{"\<\"lastError\"\>", "\[Rule]", 
                    "0.017644668008524416`"}], ",", 
                    RowBox[{"\<\"lastInt\"\>", "\[Rule]", 
                    RowBox[{"-", "0.1190635424774838`"}]}], ",", 
                    RowBox[{"\<\"initialized\"\>", "\[Rule]", "True"}]}], 
                    "\[RightAssociation]"}]}
                    },
                    AutoDelete->False,
                    
                    GridBoxItemSize->{
                    "Columns" -> {{Automatic}}, "Rows" -> {{Automatic}}}],
                    "Grid"], {-1, -0.5}], 
                  {GrayLevel[0], LineBox[{{-5, 0}, {5, 0}}]}, 
                  {GrayLevel[0.5], CircleBox[{-0.9215393619474773, 0.5}, 0.5],
                    DiskBox[{-0.9215393619474773, 0.5}, 0.1]}, 
                  {GrayLevel[0.5], 
                   DiskBox[{-0.8836836553362988, 1.9995222390738219`}, 0.1]}, 
                  {GrayLevel[0.5], Thickness[0.01], 
                   LineBox[{{-0.9215393619474773, 0.5}, {-0.8836836553362988, 
                    1.9995222390738219`}}]}, 
                  {RGBColor[0, 1, 0], 
                   DiskBox[{-0.8963022242066917, 1.4996814927158812`}, 0.05]}, 
                  {RGBColor[0, 0, 1], Arrowheads[0.01], 
                   CircleBox[{-0.9215393619474773, 0.5}, 0.3333333333333333, 
                    NCache[{
                    Rational[-1, 2] 
                    Pi, -1.945632540858484}, {-1.5707963267948966`, \
-1.945632540858484}]], 
                   ArrowBox[{{-1.0404714597324496`, 
                    0.1886059293634882}, {-1.043579401984728, 
                    0.1898108000932055}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-0.9215393619474773, 0.5}, {-0.6716485525717523,
                     0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-0.9215393619474773, 
                    0.5}, {-1.1070990736164754`, 0.5}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.005], Arrowheads[0.04], 
                   ArrowBox[{{-0.9215393619474773, 0.5}, {-0.8572082642407503,
                     0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-0.9215393619474773, 0.5}, 1.6, {
                    1.5455565093086334`, 1.4295447389215978`}], 
                   ArrowBox[{{-0.7121392537821005, 
                    2.0862381897748925`}, {-0.6962876061742199, 
                    2.084064912344524}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Dashing[{Small, Small}], 
                   Arrowheads[0.02], 
                   CircleBox[{-0.9215393619474773, 0.5}, 1.7, {
                    1.5455565093086334`, 1.5828269547311842`}], 
                   ArrowBox[{{-0.9249914270667622, 
                    2.199996495068861}, {-0.9419909360843006, 
                    2.1998769758765855`}}]}, 
                  {RGBColor[0, 0, 1], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-0.9215393619474773, 0.5}, 1.8, {
                    1.5455565093086334`, 1.6149706230241128`}], 
                   ArrowBox[{{-0.983041122390239, 
                    2.2989490080217507`}, {-1.0010272375847244`, 
                    2.298244053966724}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.003], Arrowheads[0.025], 
                   CircleBox[{-0.9215393619474773, 0.5}, 1.5, {
                    1.5455565093086334`, 1.547347395499928}], 
                   ArrowBox[{{-0.8713753204070056, 
                    1.9991609549799265`}, {-0.8863691882790704, 
                    1.9995876296115989`}}]}, 
                  {RGBColor[0, 0, 1], ArrowBox[{{0, 2.8}, {1.5, 2.8}}]}},
                 Frame->True,
                 FrameTicks->Automatic,
                 ImageSize->800,
                 PlotRange->{{-3, 3}, {-1, 3}}], 123->
                GraphicsBox[{InsetBox[
                   TagBox[GridBox[{
                    {"\<\"t\"\>", "6.1000000000000005`"},
                    {"\<\"driveIntent\"\>", "0.5`"},
                    {"\<\"motorTorque\"\>", 
                    RowBox[{"-", "12.576111320262282`"}]},
                    {"\<\"vw\"\>", "0.3581459077191314`"},
                    {"\<\"controllerState\"\>", 
                    RowBox[{"\[LeftAssociation]", 
                    RowBox[{
                    
                    RowBox[{"\<\"lastError\"\>", "\[Rule]", 
                    "0.017151194011527454`"}], ",", 
                    RowBox[{"\<\"lastInt\"\>", "\[Rule]", 
                    RowBox[{"-", "0.11819407997404813`"}]}], ",", 
                    RowBox[{"\<\"initialized\"\>", "\[Rule]", "True"}]}], 
                    "\[RightAssociation]"}]}
                    },
                    AutoDelete->False,
                    
                    GridBoxItemSize->{
                    "Columns" -> {{Automatic}}, "Rows" -> {{Automatic}}}],
                    "Grid"], {-1, -0.5}], 
                  {GrayLevel[0], LineBox[{{-5, 0}, {5, 0}}]}, 
                  {GrayLevel[0.5], CircleBox[{-0.9036580560413054, 0.5}, 0.5],
                    DiskBox[{-0.9036580560413054, 0.5}, 0.1]}, 
                  {GrayLevel[0.5], 
                   DiskBox[{-0.8655404886788818, 1.9995156054734378`}, 0.1]}, 
                  {GrayLevel[0.5], Thickness[0.01], 
                   LineBox[{{-0.9036580560413054, 0.5}, {-0.8655404886788818, 
                    1.9995156054734378`}}]}, 
                  {RGBColor[0, 1, 0], 
                   DiskBox[{-0.8782463444663563, 1.4996770703156252`}, 0.05]}, 
                  {RGBColor[0, 0, 1], Arrowheads[0.01], 
                   CircleBox[{-0.9036580560413054, 0.5}, 0.3333333333333333, 
                    NCache[{
                    Rational[-1, 2] 
                    Pi, -1.9480796664027649`}, {-1.5707963267948966`, \
-1.9480796664027649`}]], 
                   ArrowBox[{{-1.0233518173426173`, 
                    0.18889790322684147`}, {-1.026456801822066, 
                    0.19011037586620738`}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-0.9036580560413054, 0.5}, {-0.6521358296360598,
                     0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-0.9036580560413054, 
                    0.5}, {-1.0905005205596712`, 0.5}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.005], Arrowheads[0.04], 
                   ArrowBox[{{-0.9036580560413054, 0.5}, {-0.8389782941544255,
                     0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-0.9036580560413054, 0.5}, 1.6, {
                    1.545381879467896, 1.4285676163507985`}], 
                   ArrowBox[{{-0.6927080989489838, 
                    2.0860328229903526`}, {-0.6768585824665215, 
                    2.0838440575972754`}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Dashing[{Small, Small}], 
                   Arrowheads[0.02], 
                   CircleBox[{-0.9036580560413054, 0.5}, 1.7, {
                    1.545381879467896, 1.5828541552502582`}], 
                   ArrowBox[{{-0.9071563619463974, 
                    2.199996400542011}, {-0.9241558677066631, 
                    2.199876418954311}}]}, 
                  {RGBColor[0, 0, 1], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-0.9036580560413054, 0.5}, 1.8, {
                    1.545381879467896, 1.615249164580464}], 
                   ArrowBox[{{-0.9656608961482595, 
                    2.298931807439813}, {-0.9836468142860179, 
                    2.2982218435316786`}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.003], Arrowheads[0.025], 
                   CircleBox[{-0.9036580560413054, 0.5}, 1.5, {
                    1.545381879467896, 1.547089320164751}], 
                   ArrowBox[{{-0.8531071197097227, 
                    1.9991479589540186`}, {-0.8681008769682729, 
                    1.999578503118916}}]}, 
                  {RGBColor[0, 0, 1], ArrowBox[{{0, 2.8}, {1.5, 2.8}}]}},
                 Frame->True,
                 FrameTicks->Automatic,
                 ImageSize->800,
                 PlotRange->{{-3, 3}, {-1, 3}}], 124->
                GraphicsBox[{InsetBox[
                   TagBox[GridBox[{
                    {"\<\"t\"\>", "6.15`"},
                    {"\<\"driveIntent\"\>", "0.5`"},
                    {"\<\"motorTorque\"\>", 
                    RowBox[{"-", "12.63495578373539`"}]},
                    {"\<\"vw\"\>", "0.359210481016936`"},
                    {"\<\"controllerState\"\>", 
                    RowBox[{"\[LeftAssociation]", 
                    RowBox[{
                    
                    RowBox[{"\<\"lastError\"\>", "\[Rule]", 
                    "0.016701843391689997`"}], ",", 
                    RowBox[{"\<\"lastInt\"\>", "\[Rule]", 
                    RowBox[{"-", "0.11734815882331621`"}]}], ",", 
                    RowBox[{"\<\"initialized\"\>", "\[Rule]", "True"}]}], 
                    "\[RightAssociation]"}]}
                    },
                    AutoDelete->False,
                    
                    GridBoxItemSize->{
                    "Columns" -> {{Automatic}}, "Rows" -> {{Automatic}}}],
                    "Grid"], {-1, -0.5}], 
                  {GrayLevel[0], LineBox[{{-5, 0}, {5, 0}}]}, 
                  {GrayLevel[0.5], CircleBox[{-0.8857236297627324, 0.5}, 0.5],
                    DiskBox[{-0.8857236297627324, 0.5}, 0.1]}, 
                  {GrayLevel[0.5], 
                   DiskBox[{-0.8474124539620007, 1.9995106714554471`}, 0.1]}, 
                  {GrayLevel[0.5], Thickness[0.01], 
                   LineBox[{{-0.8857236297627324, 0.5}, {-0.8474124539620007, 
                    1.9995106714554471`}}]}, 
                  {RGBColor[0, 1, 0], 
                   DiskBox[{-0.8601828458955779, 1.499673780970298}, 0.05]}, 
                  {RGBColor[0, 0, 1], Arrowheads[0.01], 
                   CircleBox[{-0.8857236297627324, 0.5}, 0.3333333333333333, 
                    NCache[{
                    Rational[-1, 2] 
                    Pi, -1.9498450003069583`}, {-1.5707963267948966`, \
-1.9498450003069583`}]], 
                   ArrowBox[{{-1.005966403350892, 
                    0.1891096873317894}, {-1.0090692425742025`, 
                    0.19032763941340514`}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-0.8857236297627324, 0.5}, {-0.6330245140880246,
                     0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-0.8857236297627324, 
                    0.5}, {-1.0735144949010327`, 0.5}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.005], Arrowheads[0.04], 
                   ArrowBox[{{-0.8857236297627324, 0.5}, {-0.8208153792263249,
                     0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-0.8857236297627324, 0.5}, 1.6, {
                    1.5452527652687116`, 1.42784517396676}], 
                   ArrowBox[{{-0.6736279104860922, 
                    2.08588000992021}, {-0.6577799793964945, 
                    2.083679794736838}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Dashing[{Small, Small}], 
                   Arrowheads[0.02], 
                   CircleBox[{-0.8857236297627324, 0.5}, 1.7, {
                    1.5452527652687116`, 1.5828572896686695`}], 
                   ArrowBox[{{-0.8892272641678242, 
                    2.199996389568506}, {-0.9062267695519338, 
                    2.199876354697242}}]}, 
                  {RGBColor[0, 0, 1], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-0.8857236297627324, 0.5}, 1.8, {
                    1.5452527652687116`, 1.6154469640672415`}], 
                   ArrowBox[{{-0.9480822964427061, 
                    2.298919508118664}, {-0.9660680737981225, 
                    2.298205986619065}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.003], Arrowheads[0.025], 
                   CircleBox[{-0.8857236297627324, 0.5}, 1.5, {
                    1.5452527652687116`, 1.5468695007581512`}], 
                   ArrowBox[{{-0.8348431528403861, 
                    1.9991368106574379`}, {-0.8498368150947206, 
                    1.9995706507307296`}}]}, 
                  {RGBColor[0, 0, 1], ArrowBox[{{0, 2.8}, {1.5, 2.8}}]}},
                 Frame->True,
                 FrameTicks->Automatic,
                 ImageSize->800,
                 PlotRange->{{-3, 3}, {-1, 3}}], 125->
                GraphicsBox[{InsetBox[
                   TagBox[GridBox[{
                    {"\<\"t\"\>", "6.2`"},
                    {"\<\"driveIntent\"\>", "0.5`"},
                    {"\<\"motorTorque\"\>", 
                    RowBox[{"-", "12.672346993069784`"}]},
                    {"\<\"vw\"\>", "0.3602778962639269`"},
                    {"\<\"controllerState\"\>", 
                    RowBox[{"\[LeftAssociation]", 
                    RowBox[{
                    
                    RowBox[{"\<\"lastError\"\>", "\[Rule]", 
                    "0.01629477960911301`"}], ",", 
                    RowBox[{"\<\"lastInt\"\>", "\[Rule]", 
                    RowBox[{"-", "0.11652361881645562`"}]}], ",", 
                    RowBox[{"\<\"initialized\"\>", "\[Rule]", "True"}]}], 
                    "\[RightAssociation]"}]}
                    },
                    AutoDelete->False,
                    
                    GridBoxItemSize->{
                    "Columns" -> {{Automatic}}, "Rows" -> {{Automatic}}}],
                    "Grid"], {-1, -0.5}], 
                  {GrayLevel[0], LineBox[{{-5, 0}, {5, 0}}]}, 
                  {GrayLevel[0.5], CircleBox[{-0.8677358946416122, 0.5}, 0.5],
                    DiskBox[{-0.8677358946416122, 0.5}, 0.1]}, 
                  {GrayLevel[0.5], 
                   DiskBox[{-0.8292957407238543, 1.9995073706276936`}, 0.1]}, 
                  {GrayLevel[0.5], Thickness[0.01], 
                   LineBox[{{-0.8677358946416122, 0.5}, {-0.8292957407238543, 
                    1.9995073706276936`}}]}, 
                  {RGBColor[0, 1, 0], 
                   DiskBox[{-0.8421091253631069, 1.4996715804184624`}, 0.05]}, 
                  {RGBColor[0, 0, 1], Arrowheads[0.01], 
                   CircleBox[{-0.8677358946416122, 0.5}, 0.3333333333333333, 
                    NCache[{
                    Rational[-1, 2] 
                    Pi, -1.95096673658699}, {-1.5707963267948966`, \
-1.95096673658699}]], 
                   ArrowBox[{{-0.9883273294491973, 
                    0.18924476358046832`}, {-0.9914288004996181, 
                    0.19046619546241228`}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-0.8677358946416122, 0.5}, {-0.6142889547802165,
                     0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-0.8677358946416122, 
                    0.5}, {-1.0561585603275492`, 0.5}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.005], Arrowheads[0.04], 
                   ArrowBox[{{-0.8677358946416122, 0.5}, {-0.8027116204661535,
                     0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-0.8677358946416122, 0.5}, 1.6, {
                    1.5451667517035135`, 1.4273638968855193`}], 
                   ArrowBox[{{-0.6548769522556784, 
                    2.0857777494486234`}, {-0.6390300819145907, 
                    2.0835699072743434`}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Dashing[{Small, Small}], 
                   Arrowheads[0.02], 
                   CircleBox[{-0.8677358946416122, 0.5}, 1.7, {
                    1.5451667517035135`, 1.5828384095081356`}], 
                   ArrowBox[{{-0.8712074328413393, 
                    2.199996455414695}, {-0.8882069404886969, 
                    2.199876741496843}}]}, 
                  {RGBColor[0, 0, 1], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-0.8677358946416122, 0.5}, 1.8, {
                    1.5451667517035135`, 1.61556867944279}], 
                   ArrowBox[{{-0.9303135170226469, 
                    2.2989119047849833`}, {-0.9482992073982999, 
                    2.2981961941450297`}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.003], Arrowheads[0.025], 
                   CircleBox[{-0.8677358946416122, 0.5}, 1.5, {
                    1.5451667517035135`, 1.5466872320300236`}], 
                   ArrowBox[{{-0.8165821728061836, 
                    1.9991275118355953`}, {-0.8315757557359814, 
                    1.9995640847774148`}}]}, 
                  {RGBColor[0, 0, 1], ArrowBox[{{0, 2.8}, {1.5, 2.8}}]}},
                 Frame->True,
                 FrameTicks->Automatic,
                 ImageSize->800,
                 PlotRange->{{-3, 3}, {-1, 3}}], 126->
                GraphicsBox[{InsetBox[
                   TagBox[GridBox[{
                    {"\<\"t\"\>", "6.25`"},
                    {"\<\"driveIntent\"\>", "0.5`"},
                    {"\<\"motorTorque\"\>", 
                    RowBox[{"-", "12.689623561000909`"}]},
                    {"\<\"vw\"\>", "0.3613463678332725`"},
                    {"\<\"controllerState\"\>", 
                    RowBox[{"\[LeftAssociation]", 
                    RowBox[{
                    
                    RowBox[{"\<\"lastError\"\>", "\[Rule]", 
                    "0.015927978354179802`"}], ",", 
                    RowBox[{"\<\"lastInt\"\>", "\[Rule]", 
                    RowBox[{"-", "0.11571839653729878`"}]}], ",", 
                    RowBox[{"\<\"initialized\"\>", "\[Rule]", "True"}]}], 
                    "\[RightAssociation]"}]}
                    },
                    AutoDelete->False,
                    
                    GridBoxItemSize->{
                    "Columns" -> {{Automatic}}, "Rows" -> {{Automatic}}}],
                    "Grid"], {-1, -0.5}], 
                  {GrayLevel[0], LineBox[{{-5, 0}, {5, 0}}]}, 
                  {GrayLevel[0.5], CircleBox[{-0.8496947546571971, 0.5}, 0.5],
                    DiskBox[{-0.8496947546571971, 0.5}, 0.1]}, 
                  {GrayLevel[0.5], 
                   DiskBox[{-0.8111864087992355, 1.9995056209628836`}, 0.1]}, 
                  {GrayLevel[0.5], Thickness[0.01], 
                   LineBox[{{-0.8496947546571971, 0.5}, {-0.8111864087992355, 
                    1.9995056209628836`}}]}, 
                  {RGBColor[0, 1, 0], 
                   DiskBox[{-0.8240225240852227, 1.499670413975256}, 0.05]}, 
                  {RGBColor[0, 0, 1], Arrowheads[0.01], 
                   CircleBox[{-0.8496947546571971, 0.5}, 0.3333333333333333, 
                    NCache[{
                    Rational[-1, 2] 
                    Pi, -1.9514850336249239`}, {-1.5707963267948966`, \
-1.9514850336249239`}]], 
                   ArrowBox[{{-0.970447236778782, 
                    0.1893073075005023}, {-0.9735480743481277, 
                    0.1905303467015752}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-0.8496947546571971, 0.5}, {-0.595902283437179, 
                    0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-0.8496947546571971, 0.5}, {-1.038451457529019, 
                    0.5}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.005], Arrowheads[0.04], 
                   ArrowBox[{{-0.8496947546571971, 0.5}, {-0.7846589863090007,
                     0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-0.8496947546571971, 0.5}, 1.6, {
                    1.5451212754482366`, 1.4271094411019931`}], 
                   ArrowBox[{{-0.6364323088469517, 
                    2.085723534922476}, {-0.6205860008170927, 
                    2.083511660491906}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Dashing[{Small, Small}], 
                   Arrowheads[0.02], 
                   CircleBox[{-0.8496947546571971, 0.5}, 1.7, {
                    1.5451212754482366`, 1.5827995474354168`}], 
                   ArrowBox[{{-0.8531002274684475, 
                    2.1999965890421462`}, {-0.8700997397552991, 
                    2.1998775357604874`}}]}, 
                  {RGBColor[0, 0, 1], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-0.8496947546571971, 0.5}, 1.8, {
                    1.5451212754482366`, 1.615619184120464}], 
                   ArrowBox[{{-0.9123632304242991, 
                    2.2989087420280745`}, {-0.9303488846302788, 
                    2.2981921230275386`}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.003], Arrowheads[0.025], 
                   CircleBox[{-0.8496947546571971, 0.5}, 1.5, {
                    1.5451212754482366`, 1.546541517471054}], 
                   ArrowBox[{{-0.798322588661382, 
                    1.9991200420783182`}, {-0.8133161078169688, 
                    1.999558799798819}}]}, 
                  {RGBColor[0, 0, 1], ArrowBox[{{0, 2.8}, {1.5, 2.8}}]}},
                 Frame->True,
                 FrameTicks->Automatic,
                 ImageSize->800,
                 PlotRange->{{-3, 3}, {-1, 3}}], 127->
                GraphicsBox[{InsetBox[
                   TagBox[GridBox[{
                    {"\<\"t\"\>", "6.3`"},
                    {"\<\"driveIntent\"\>", "0.5`"},
                    {"\<\"motorTorque\"\>", 
                    RowBox[{"-", "12.688169136190101`"}]},
                    {"\<\"vw\"\>", "0.3624142416591761`"},
                    {"\<\"controllerState\"\>", 
                    RowBox[{"\[LeftAssociation]", 
                    RowBox[{
                    
                    RowBox[{"\<\"lastError\"\>", "\[Rule]", 
                    "0.015599268530839434`"}], ",", 
                    RowBox[{"\<\"lastInt\"\>", "\[Rule]", 
                    RowBox[{"-", "0.11493053369204981`"}]}], ",", 
                    RowBox[{"\<\"initialized\"\>", "\[Rule]", "True"}]}], 
                    "\[RightAssociation]"}]}
                    },
                    AutoDelete->False,
                    
                    GridBoxItemSize->{
                    "Columns" -> {{Automatic}}, "Rows" -> {{Automatic}}}],
                    "Grid"], {-1, -0.5}], 
                  {GrayLevel[0], LineBox[{{-5, 0}, {5, 0}}]}, 
                  {GrayLevel[0.5], CircleBox[{-0.831600199723831, 0.5}, 0.5], 
                   DiskBox[{-0.831600199723831, 0.5}, 0.1]}, 
                  {GrayLevel[0.5], 
                   DiskBox[{-0.793080444426408, 1.999505327917119}, 0.1]}, 
                  {GrayLevel[0.5], Thickness[0.01], 
                   LineBox[{{-0.831600199723831, 0.5}, {-0.793080444426408, 
                    1.999505327917119}}]}, 
                  {RGBColor[0, 1, 0], 
                   DiskBox[{-0.8059203628588824, 1.4996702186114126`}, 0.05]}, 
                  {RGBColor[0, 0, 1], Arrowheads[0.01], 
                   CircleBox[{-0.831600199723831, 0.5}, 0.3333333333333333, 
                    NCache[{
                    Rational[-1, 2] 
                    Pi, -1.9514414008805996`}, {-1.5707963267948966`, \
-1.9514414008805996`}]], 
                   ArrowBox[{{-0.9523391253556597, 
                    0.18930203903407594`}, {-0.9554400162866105, 
                    0.1905249429359318}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-0.831600199723831, 0.5}, {-0.577836817000029, 
                    0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-0.831600199723831, 0.5}, {-1.0204127914476544`,
                     0.5}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.005], Arrowheads[0.04], 
                   ArrowBox[{{-0.831600199723831, 0.5}, {-0.7666494087238523, 
                    0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-0.831600199723831, 0.5}, 1.6, {
                    1.5451136666467646`, 1.427066867181929}], 
                   ArrowBox[{{-0.6182702436398597, 
                    2.085714454067065}, {-0.6024240297925271, 
                    2.083501904999049}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Dashing[{Small, Small}], 
                   Arrowheads[0.02], 
                   CircleBox[{-0.831600199723831, 0.5}, 1.7, {
                    1.5451136666467646`, 1.5827426997522644`}], 
                   ArrowBox[{{-0.8349090316621768, 
                    2.1999967798884805`}, {-0.8519085506894634, 
                    2.199878692989902}}]}, 
                  {RGBColor[0, 0, 1], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-0.831600199723831, 0.5}, 1.8, {
                    1.5451136666467646`, 1.615603495181154}], 
                   ArrowBox[{{-0.8942404525131438, 
                    2.298909725008593}, {-0.9122261179599018, 
                    2.298193388183982}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.003], Arrowheads[0.025], 
                   CircleBox[{-0.831600199723831, 0.5}, 1.5, {
                    1.5451136666467646`, 1.546431112311774}], 
                   ArrowBox[{{-0.7800625234544221, 
                    1.999114361189549}, {-0.7950559940775128, 
                    1.9995547742692428`}}]}, 
                  {RGBColor[0, 0, 1], ArrowBox[{{0, 2.8}, {1.5, 2.8}}]}},
                 Frame->True,
                 FrameTicks->Automatic,
                 ImageSize->800,
                 PlotRange->{{-3, 3}, {-1, 3}}], 128->
                GraphicsBox[{InsetBox[
                   TagBox[GridBox[{
                    {"\<\"t\"\>", "6.3500000000000005`"},
                    {"\<\"driveIntent\"\>", "0.5`"},
                    {"\<\"motorTorque\"\>", 
                    RowBox[{"-", "12.66939388106027`"}]},
                    {"\<\"vw\"\>", "0.3634799965201404`"},
                    {"\<\"controllerState\"\>", 
                    RowBox[{"\[LeftAssociation]", 
                    RowBox[{
                    
                    RowBox[{"\<\"lastError\"\>", "\[Rule]", 
                    "0.015306369945322314`"}], ",", 
                    RowBox[{"\<\"lastInt\"\>", "\[Rule]", 
                    RowBox[{"-", "0.11415818347382517`"}]}], ",", 
                    RowBox[{"\<\"initialized\"\>", "\[Rule]", "True"}]}], 
                    "\[RightAssociation]"}]}
                    },
                    AutoDelete->False,
                    
                    GridBoxItemSize->{
                    "Columns" -> {{Automatic}}, "Rows" -> {{Automatic}}}],
                    "Grid"], {-1, -0.5}], 
                  {GrayLevel[0], LineBox[{{-5, 0}, {5, 0}}]}, 
                  {GrayLevel[0.5], CircleBox[{-0.8134522990738469, 0.5}, 0.5],
                    DiskBox[{-0.8134522990738469, 0.5}, 0.1]}, 
                  {GrayLevel[0.5], 
                   DiskBox[{-0.7749738105854859, 1.9995063874234251`}, 0.1]}, 
                  {GrayLevel[0.5], Thickness[0.01], 
                   LineBox[{{-0.8134522990738469, 0.5}, {-0.7749738105854859, 
                    1.9995063874234251`}}]}, 
                  {RGBColor[0, 1, 0], 
                   DiskBox[{-0.7877999734149395, 1.49967092494895}, 0.05]}, 
                  {RGBColor[0, 0, 1], Arrowheads[0.01], 
                   CircleBox[{-0.8134522990738469, 0.5}, 0.3333333333333333, 
                    NCache[{
                    Rational[-1, 2] 
                    Pi, -1.9508781432267046`}, {-1.5707963267948966`, \
-1.9508781432267046`}]], 
                   ArrowBox[{{-0.9340162025575993, 
                    0.1892340811995763}, {-0.9371177818066033, 
                    0.19045523830698508`}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-0.8134522990738469, 0.5}, {-0.5600644214526415,
                     0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-0.8134522990738469, 
                    0.5}, {-1.0020627462170952`, 0.5}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.005], Arrowheads[0.04], 
                   ArrowBox[{{-0.8134522990738469, 0.5}, {-0.7486748685958899,
                     0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-0.8134522990738469, 0.5}, 1.6, {
                    1.5451411869187583`, 1.4272208528142627`}], 
                   ArrowBox[{{-0.6003665227610153, 
                    2.085747285015225}, {-0.584519968400786, 
                    2.0835371760626904`}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Dashing[{Small, Small}], 
                   Arrowheads[0.02], 
                   CircleBox[{-0.8134522990738469, 0.5}, 1.7, {
                    1.5451411869187583`, 1.58266981126469}], 
                   ArrowBox[{{-0.8166372208093453, 
                    2.199997016548423}, {-0.8336367483986503, 
                    2.199880168719388}}]}, 
                  {RGBColor[0, 0, 1], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-0.8134522990738469, 0.5}, 1.8, {
                    1.5451411869187583`, 1.6155267084802043`}], 
                   ArrowBox[{{-0.875954419335536, 
                    2.298914529643583}, {-0.893940139734412, 
                    2.2981995738809973`}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.003], Arrowheads[0.025], 
                   CircleBox[{-0.8134522990738469, 0.5}, 1.5, {
                    1.5451411869187583`, 1.5463545640064207`}], 
                   ArrowBox[{{-0.7617998682916662, 
                    1.999110411675635}, {-0.776793305157954, 
                    1.999551972478805}}]}, 
                  {RGBColor[0, 0, 1], ArrowBox[{{0, 2.8}, {1.5, 2.8}}]}},
                 Frame->True,
                 FrameTicks->Automatic,
                 ImageSize->800,
                 PlotRange->{{-3, 3}, {-1, 3}}], 129->
                GraphicsBox[{InsetBox[
                   TagBox[GridBox[{
                    {"\<\"t\"\>", "6.4`"},
                    {"\<\"driveIntent\"\>", "0.5`"},
                    {"\<\"motorTorque\"\>", 
                    RowBox[{"-", "12.634717828042623`"}]},
                    {"\<\"vw\"\>", "0.3645422438877324`"},
                    {"\<\"controllerState\"\>", 
                    RowBox[{"\[LeftAssociation]", 
                    RowBox[{
                    
                    RowBox[{"\<\"lastError\"\>", "\[Rule]", 
                    "0.0150469277136383`"}], ",", 
                    RowBox[{"\<\"lastInt\"\>", "\[Rule]", 
                    RowBox[{"-", "0.11339961512655178`"}]}], ",", 
                    RowBox[{"\<\"initialized\"\>", "\[Rule]", "True"}]}], 
                    "\[RightAssociation]"}]}
                    },
                    AutoDelete->False,
                    
                    GridBoxItemSize->{
                    "Columns" -> {{Automatic}}, "Rows" -> {{Automatic}}}],
                    "Grid"], {-1, -0.5}], 
                  {GrayLevel[0], LineBox[{{-5, 0}, {5, 0}}]}, 
                  {GrayLevel[0.5], CircleBox[{-0.7952511946135841, 0.5}, 0.5],
                    DiskBox[{-0.7952511946135841, 0.5}, 0.1]}, 
                  {GrayLevel[0.5], 
                   DiskBox[{-0.7568624917448707, 1.9995086887017552`}, 0.1]}, 
                  {GrayLevel[0.5], Thickness[0.01], 
                   LineBox[{{-0.7952511946135841, 0.5}, {-0.7568624917448707, 
                    1.9995086887017552`}}]}, 
                  {RGBColor[0, 1, 0], 
                   DiskBox[{-0.7696587260344419, 1.4996724591345036`}, 0.05]}, 
                  {RGBColor[0, 0, 1], Arrowheads[0.01], 
                   CircleBox[{-0.7952511946135841, 0.5}, 0.3333333333333333, 
                    NCache[{
                    Rational[-1, 2] 
                    Pi, -1.9498378616361753`}, {-1.5707963267948966`, \
-1.9498378616361753`}]], 
                   ArrowBox[{{-0.9154917488550879, 
                    0.18910882896613634`}, {-0.9185945967728785, 
                    0.19032675889757333`}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-0.7952511946135841, 0.5}, {-0.5425568380527316,
                     0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-0.7952511946135841, 0.5}, {-0.9834218273412851,
                     0.5}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.005], Arrowheads[0.04], 
                   ArrowBox[{{-0.7952511946135841, 0.5}, {-0.7307274707804327,
                     0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-0.7952511946135841, 0.5}, 1.6, {
                    1.5452010636565152`, 1.4275558845979317`}], 
                   ArrowBox[{{-0.5826967059912049, 
                    2.0858185865244736`}, {-0.5668494120636035, 
                    2.08361378679525}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Dashing[{Small, Small}], 
                   Arrowheads[0.02], 
                   CircleBox[{-0.7952511946135841, 0.5}, 1.7, {
                    1.5452010636565152`, 1.582582762404163}], 
                   ArrowBox[{{-0.7982881335340128, 
                    2.199997287351363}, {-0.8152876712303813, 
                    2.199881919312275}}]}, 
                  {RGBColor[0, 0, 1], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-0.7952511946135841, 0.5}, 1.8, {
                    1.5452010636565152`, 1.615393940478974}], 
                   ArrowBox[{{-0.857514476038617, 
                    2.2989228120700975`}, {-0.8755002912022201, 
                    2.298210244241954}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.003], Arrowheads[0.025], 
                   CircleBox[{-0.7952511946135841, 0.5}, 1.5, {
                    1.5452010636565152`, 1.546310250094978}], 
                   ArrowBox[{{-0.7435323324361148, 
                    1.9991081212824737`}, {-0.7585257497203947, 
                    1.9995503465030433`}}]}, 
                  {RGBColor[0, 0, 1], ArrowBox[{{0, 2.8}, {1.5, 2.8}}]}},
                 Frame->True,
                 FrameTicks->Automatic,
                 ImageSize->800,
                 PlotRange->{{-3, 3}, {-1, 3}}], 130->
                GraphicsBox[{InsetBox[
                   TagBox[GridBox[{
                    {"\<\"t\"\>", "6.45`"},
                    {"\<\"driveIntent\"\>", "0.5`"},
                    {"\<\"motorTorque\"\>", 
                    RowBox[{"-", "12.585556058317005`"}]},
                    {"\<\"vw\"\>", "0.36559972650736483`"},
                    {"\<\"controllerState\"\>", 
                    RowBox[{"\[LeftAssociation]", 
                    RowBox[{
                    
                    RowBox[{"\<\"lastError\"\>", "\[Rule]", 
                    "0.014818543432042176`"}], ",", 
                    RowBox[{"\<\"lastInt\"\>", "\[Rule]", 
                    RowBox[{"-", "0.11265321687127795`"}]}], ",", 
                    RowBox[{"\<\"initialized\"\>", "\[Rule]", "True"}]}], 
                    "\[RightAssociation]"}]}
                    },
                    AutoDelete->False,
                    
                    GridBoxItemSize->{
                    "Columns" -> {{Automatic}}, "Rows" -> {{Automatic}}}],
                    "Grid"], {-1, -0.5}], 
                  {GrayLevel[0], LineBox[{{-5, 0}, {5, 0}}]}, 
                  {GrayLevel[0.5], CircleBox[{-0.7769970943199819, 0.5}, 0.5],
                    DiskBox[{-0.7769970943199819, 0.5}, 0.1]}, 
                  {GrayLevel[0.5], 
                   DiskBox[{-0.7387425332270152, 1.9995121168418697`}, 0.1]}, 
                  {GrayLevel[0.5], Thickness[0.01], 
                   LineBox[{{-0.7769970943199819, 0.5}, {-0.7387425332270152, 
                    1.9995121168418697`}}]}, 
                  {RGBColor[0, 1, 0], 
                   DiskBox[{-0.7514940535913374, 1.4996747445612464`}, 0.05]}, 
                  {RGBColor[0, 0, 1], Arrowheads[0.01], 
                   CircleBox[{-0.7769970943199819, 0.5}, 0.3333333333333333, 
                    NCache[{
                    Rational[-1, 2] 
                    Pi, -1.9483630085444066`}, {-1.5707963267948966`, \
-1.9483630085444066`}]], 
                   ArrowBox[{{-0.8967789991498099, 
                    0.1889318300011248}, {-0.8998836399600303, 
                    0.19014518236476075`}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-0.7769970943199819, 0.5}, {-0.5252859731536418,
                     0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-0.7769970943199819, 0.5}, {-0.9645106304499689,
                     0.5}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.005], Arrowheads[0.04], 
                   ArrowBox[{{-0.7769970943199819, 0.5}, {-0.7127995092836288,
                     0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-0.7769970943199819, 0.5}, 1.6, {
                    1.5452905207055687`, 1.428056429598587}], 
                   ArrowBox[{{-0.5652364058571508, 
                    2.0859247809469865`}, {-0.549388010313346, 
                    2.0837279137773823`}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Dashing[{Small, Small}], 
                   Arrowheads[0.02], 
                   CircleBox[{-0.7769970943199819, 0.5}, 1.7, {
                    1.5452905207055687`, 1.5824833584657942`}], 
                   ArrowBox[{{-0.7798650468001056, 
                    2.1999975808360936`}, {-0.796864595880524, 
                    2.199883902618571}}]}, 
                  {RGBColor[0, 0, 1], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-0.7769970943199819, 0.5}, 1.8, {
                    1.5452905207055687`, 1.6152102765851077`}], 
                   ArrowBox[{{-0.8389299775282968, 
                    2.2989342172457294`}, {-0.8569159232615284, 
                    2.298224952774433}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.003], Arrowheads[0.025], 
                   CircleBox[{-0.7769970943199819, 0.5}, 1.5, {
                    1.5452905207055687`, 1.546296413375157}], 
                   ArrowBox[{{-0.725257489408409, 
                    1.9991074055195628`}, {-0.740250900572307, 
                    1.999549838199804}}]}, 
                  {RGBColor[0, 0, 1], ArrowBox[{{0, 2.8}, {1.5, 2.8}}]}},
                 Frame->True,
                 FrameTicks->Automatic,
                 ImageSize->800,
                 PlotRange->{{-3, 3}, {-1, 3}}], 131->
                GraphicsBox[{InsetBox[
                   TagBox[GridBox[{
                    {"\<\"t\"\>", "6.5`"},
                    {"\<\"driveIntent\"\>", "0.5`"},
                    {"\<\"motorTorque\"\>", 
                    RowBox[{"-", "12.52330563564804`"}]},
                    {"\<\"vw\"\>", "0.3666513158709589`"},
                    {"\<\"controllerState\"\>", 
                    RowBox[{"\[LeftAssociation]", 
                    RowBox[{
                    RowBox[{"\<\"lastError\"\>", "\[Rule]", 
                    "0.014618803181432043`"}], ",", 
                    RowBox[{"\<\"lastInt\"\>", "\[Rule]", 
                    RowBox[{"-", "0.11191749735504188`"}]}], ",", 
                    RowBox[{"\<\"initialized\"\>", "\[Rule]", "True"}]}], 
                    "\[RightAssociation]"}]}
                    },
                    AutoDelete->False,
                    
                    GridBoxItemSize->{
                    "Columns" -> {{Automatic}}, "Rows" -> {{Automatic}}}],
                    "Grid"], {-1, -0.5}], 
                  {GrayLevel[0], LineBox[{{-5, 0}, {5, 0}}]}, 
                  {GrayLevel[0.5], CircleBox[{-0.7586902657370436, 0.5}, 0.5],
                    DiskBox[{-0.7586902657370436, 0.5}, 0.1]}, 
                  {GrayLevel[0.5], 
                   DiskBox[{-0.720610075433236, 1.999516555129161}, 0.1]}, 
                  {GrayLevel[0.5], Thickness[0.01], 
                   LineBox[{{-0.7586902657370436, 0.5}, {-0.720610075433236, 
                    1.999516555129161}}]}, 
                  {RGBColor[0, 1, 0], 
                   DiskBox[{-0.7333034722011718, 1.4996777034194406`}, 0.05]}, 
                  {RGBColor[0, 0, 1], Arrowheads[0.01], 
                   CircleBox[{-0.7586902657370436, 0.5}, 0.3333333333333333, 
                    NCache[{
                    Rational[-1, 2] 
                    Pi, -1.9464954958643377`}, {-1.5707963267948966`, \
-1.9464954958643377`}]], 
                   ArrowBox[{{-0.8778910382768637, 
                    0.1887086783461171}, {-0.8809979396228147, 
                    0.18991623064119761`}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-0.7586902657370436, 0.5}, {-0.5082241530240827,
                     0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-0.7586902657370436, 0.5}, {-0.9453496357862559,
                     0.5}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.005], Arrowheads[0.04], 
                   ArrowBox[{{-0.7586902657370436, 0.5}, {-0.694883523073295, 
                    0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-0.7586902657370436, 0.5}, 1.6, {
                    1.5454068055484529`, 1.428707087347567}], 
                   ArrowBox[{{-0.5479615162744507, 
                    2.08606222896516}, {-0.5321116946768519, 
                    2.0838756741412254`}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Dashing[{Small, Small}], 
                   Arrowheads[0.02], 
                   CircleBox[{-0.7586902657370436, 0.5}, 1.7, {
                    1.5454068055484529`, 1.5823733208219495`}], 
                   ArrowBox[{{-0.7613711544718449, 
                    2.199997886126801}, {-0.7783707159582288, 
                    2.199886078500289}}]}, 
                  {RGBColor[0, 0, 1], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-0.7586902657370436, 0.5}, 1.8, {
                    1.5454068055484529`, 1.6149807257464974`}], 
                   ArrowBox[{{-0.8202102004590786, 
                    2.298948386594734}, {-0.8381963085306912, 
                    2.2982432508310144`}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.003], Arrowheads[0.025], 
                   CircleBox[{-0.7586902657370436, 0.5}, 1.5, {
                    1.5454068055484529`, 1.5463111943484575`}], 
                   ArrowBox[{{-0.7069728190976572, 
                    1.999108170117521}, {-0.7219662367995031, 
                    1.9995503811805038`}}]}, 
                  {RGBColor[0, 0, 1], ArrowBox[{{0, 2.8}, {1.5, 2.8}}]}},
                 Frame->True,
                 FrameTicks->Automatic,
                 ImageSize->800,
                 PlotRange->{{-3, 3}, {-1, 3}}], 132->
                GraphicsBox[{InsetBox[
                   TagBox[GridBox[{
                    {"\<\"t\"\>", "6.55`"},
                    {"\<\"driveIntent\"\>", "0.5`"},
                    {"\<\"motorTorque\"\>", 
                    RowBox[{"-", "12.449334218614377`"}]},
                    {"\<\"vw\"\>", "0.3676960087338161`"},
                    {"\<\"controllerState\"\>", 
                    RowBox[{"\[LeftAssociation]", 
                    RowBox[{
                    
                    RowBox[{"\<\"lastError\"\>", "\[Rule]", 
                    "0.014445302459569626`"}], ",", 
                    RowBox[{"\<\"lastInt\"\>", "\[Rule]", 
                    RowBox[{"-", "0.11119108577829423`"}]}], ",", 
                    RowBox[{"\<\"initialized\"\>", "\[Rule]", "True"}]}], 
                    "\[RightAssociation]"}]}
                    },
                    AutoDelete->False,
                    
                    GridBoxItemSize->{
                    "Columns" -> {{Automatic}}, "Rows" -> {{Automatic}}}],
                    "Grid"], {-1, -0.5}], 
                  {GrayLevel[0], LineBox[{{-5, 0}, {5, 0}}]}, 
                  {GrayLevel[0.5], CircleBox[{-0.7403310296236559, 0.5}, 0.5],
                    DiskBox[{-0.7403310296236559, 0.5}, 0.1]}, 
                  {GrayLevel[0.5], 
                   DiskBox[{-0.7024613831899663, 1.9995218870956795`}, 0.1]}, 
                  {GrayLevel[0.5], Thickness[0.01], 
                   LineBox[{{-0.7403310296236559, 0.5}, {-0.7024613831899663, 
                    1.9995218870956795`}}]}, 
                  {RGBColor[0, 1, 0], 
                   DiskBox[{-0.7150845986678629, 1.4996812580637862`}, 0.05]}, 
                  {RGBColor[0, 0, 1], Arrowheads[0.01], 
                   CircleBox[{-0.7403310296236559, 0.5}, 0.3333333333333333, 
                    NCache[{
                    Rational[-1, 2] 
                    Pi, -1.9442763533533278`}, {-1.5707963267948966`, \
-1.9442763533533278`}]], 
                   ArrowBox[{{-0.8588407094177165, 
                    0.1884449215528331}, {-0.8619502828419906, 
                    0.1896455762233673}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-0.7403310296236559, 0.5}, {-0.4913443452513684,
                     0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-0.7403310296236559, 0.5}, {-0.9259590274305372,
                     0.5}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.005], Arrowheads[0.04], 
                   ArrowBox[{{-0.7403310296236559, 0.5}, {-0.6769723430582496,
                     0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-0.7403310296236559, 0.5}, 1.6, {
                    1.5455472131316355`, 1.4294927230694658`}], 
                   ArrowBox[{{-0.5308484122104725, 
                    2.0862272955039334`}, {-0.5149968776689138, 
                    2.084053193539547}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Dashing[{Small, Small}], 
                   Arrowheads[0.02], 
                   CircleBox[{-0.7403310296236559, 0.5}, 1.7, {
                    1.5455472131316355`, 1.582254279966857}], 
                   ArrowBox[{{-0.7428095491379203, 
                    2.1999981932169863`}, {-0.7598091238135316, 
                    2.1998884092335977`}}]}, 
                  {RGBColor[0, 0, 1], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-0.7403310296236559, 0.5}, 1.8, {
                    1.5455472131316355`, 1.6147101810128264`}], 
                   ArrowBox[{{-0.80136426608803, 
                    2.298964964652365}, {-0.8193505642721659, 
                    2.29826469496121}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.003], Arrowheads[0.025], 
                   CircleBox[{-0.7403310296236559, 0.5}, 1.5, {
                    1.5455472131316355`, 1.5463526609360017`}], 
                   ArrowBox[{{-0.6886757459288896, 
                    1.9991103133747075`}, {-0.7036691819548287, 
                    1.9995519027114426`}}]}, 
                  {RGBColor[0, 0, 1], ArrowBox[{{0, 2.8}, {1.5, 2.8}}]}},
                 Frame->True,
                 FrameTicks->Automatic,
                 ImageSize->800,
                 PlotRange->{{-3, 3}, {-1, 3}}], 133->
                GraphicsBox[{InsetBox[
                   TagBox[GridBox[{
                    {"\<\"t\"\>", "6.6000000000000005`"},
                    {"\<\"driveIntent\"\>", "0.5`"},
                    {"\<\"motorTorque\"\>", 
                    RowBox[{"-", "12.364970267092996`"}]},
                    {"\<\"vw\"\>", "0.36873292281983117`"},
                    {"\<\"controllerState\"\>", 
                    RowBox[{"\[LeftAssociation]", 
                    RowBox[{
                    
                    RowBox[{"\<\"lastError\"\>", "\[Rule]", 
                    "0.01429566815427697`"}], ",", 
                    RowBox[{"\<\"lastInt\"\>", "\[Rule]", 
                    RowBox[{"-", "0.11047273085167214`"}]}], ",", 
                    RowBox[{"\<\"initialized\"\>", "\[Rule]", "True"}]}], 
                    "\[RightAssociation]"}]}
                    },
                    AutoDelete->False,
                    GridBoxItemSize->{
                    "Columns" -> {{Automatic}}, "Rows" -> {{Automatic}}}],
                    "Grid"], {-1, -0.5}], 
                  {GrayLevel[0], LineBox[{{-5, 0}, {5, 0}}]}, 
                  {GrayLevel[0.5], CircleBox[{-0.7219197537968447, 0.5}, 0.5],
                    DiskBox[{-0.7219197537968447, 0.5}, 0.1]}, 
                  {GrayLevel[0.5], 
                   DiskBox[{-0.6842928704961588, 1.9995279982891538`}, 0.1]}, 
                  {GrayLevel[0.5], Thickness[0.01], 
                   LineBox[{{-0.7219197537968447, 0.5}, {-0.6842928704961588, 
                    1.9995279982891538`}}]}, 
                  {RGBColor[0, 1, 0], 
                   DiskBox[{-0.6968351649297208, 1.4996853321927692`}, 0.05]}, 
                  {RGBColor[0, 0, 1], Arrowheads[0.01], 
                   CircleBox[{-0.7219197537968447, 0.5}, 0.3333333333333333, 
                    NCache[{
                    Rational[-1, 2] 
                    Pi, -1.9417454348076864`}, {-1.5707963267948966`, \
-1.9417454348076864`}]], 
                   ArrowBox[{{-0.8396405343471329, 
                    0.18814598136669458`}, {-0.8427531365680787, 
                    0.18933876212315842`}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-0.7219197537968447, 0.5}, {-0.4746203484549848,
                     0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-0.7219197537968447, 0.5}, {-0.9063585361418642,
                     0.5}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.005], Arrowheads[0.04], 
                   ArrowBox[{{-0.7219197537968447, 0.5}, {-0.6590591308000043,
                     0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-0.7219197537968447, 0.5}, 1.6, {
                    1.5457091064923671`, 1.4303985830171573`}], 
                   ArrowBox[{{-0.5138741219109298, 
                    2.0864164065759}, {-0.4980206244414921, 
                    2.0842566647718153`}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Dashing[{Small, Small}], 
                   Arrowheads[0.02], 
                   CircleBox[{-0.7219197537968447, 0.5}, 1.7, {
                    1.5457091064923671`, 1.5821277702465555`}], 
                   ArrowBox[{{-0.7241832069959117, 
                    2.199998493169807}, {-0.7411827954242274, 
                    2.199890859798729}}]}, 
                  {RGBColor[0, 0, 1], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-0.7219197537968447, 0.5}, 1.8, {
                    1.5457091064923671`, 1.6144033857539948`}], 
                   ArrowBox[{{-0.7824010734755915, 
                    2.298983604697363}, {-0.8003875856526801, 
                    2.2982888531500842`}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.003], Arrowheads[0.025], 
                   CircleBox[{-0.7219197537968447, 0.5}, 1.5, {
                    1.5457091064923671`, 1.5464188354875101`}], 
                   ArrowBox[{{-0.6703636731677562, 
                    1.999113728357581}, {-0.6853571383828432, 
                    1.9995543255094455`}}]}, 
                  {RGBColor[0, 0, 1], ArrowBox[{{0, 2.8}, {1.5, 2.8}}]}},
                 Frame->True,
                 FrameTicks->Automatic,
                 ImageSize->800,
                 PlotRange->{{-3, 3}, {-1, 3}}], 134->
                GraphicsBox[{InsetBox[
                   TagBox[GridBox[{
                    {"\<\"t\"\>", "6.65`"},
                    {"\<\"driveIntent\"\>", "0.5`"},
                    {"\<\"motorTorque\"\>", 
                    RowBox[{"-", "12.271494753106213`"}]},
                    {"\<\"vw\"\>", "0.3697612918504513`"},
                    {"\<\"controllerState\"\>", 
                    RowBox[{"\[LeftAssociation]", 
                    RowBox[{
                    
                    RowBox[{"\<\"lastError\"\>", "\[Rule]", 
                    "0.014167577686617722`"}], ",", 
                    RowBox[{"\<\"lastInt\"\>", "\[Rule]", 
                    RowBox[{"-", "0.1097612987268445`"}]}], ",", 
                    RowBox[{"\<\"initialized\"\>", "\[Rule]", "True"}]}], 
                    "\[RightAssociation]"}]}
                    },
                    AutoDelete->False,
                    
                    GridBoxItemSize->{
                    "Columns" -> {{Automatic}}, "Rows" -> {{Automatic}}}],
                    "Grid"], {-1, -0.5}], 
                  {GrayLevel[0], LineBox[{{-5, 0}, {5, 0}}]}, 
                  {GrayLevel[0.5], CircleBox[{-0.7034568472075615, 0.5}, 0.5],
                    DiskBox[{-0.7034568472075615, 0.5}, 0.1]}, 
                  {GrayLevel[0.5], 
                   DiskBox[{-0.6661011209639994, 1.999534777761695}, 0.1]}, 
                  {GrayLevel[0.5], Thickness[0.01], 
                   LineBox[{{-0.7034568472075615, 0.5}, {-0.6661011209639994, 
                    1.999534777761695}}]}, 
                  {RGBColor[0, 1, 0], 
                   DiskBox[{-0.6785530297118535, 1.4996898518411301`}, 0.05]}, 
                  {RGBColor[0, 0, 1], Arrowheads[0.01], 
                   CircleBox[{-0.7034568472075615, 0.5}, 0.3333333333333333, 
                    NCache[{
                    Rational[-1, 2] 
                    Pi, -1.938941169388083}, {-1.5707963267948966`, \
-1.938941169388083}]], 
                   ArrowBox[{{-0.8203026445914365, 
                    0.18781708767961447`}, {-0.8234185794432316, 
                    0.18900113519479128`}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-0.7034568472075615, 
                    0.5}, {-0.45802695214543726`, 0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-0.7034568472075615, 0.5}, {-0.8865673046003415,
                     0.5}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.005], Arrowheads[0.04], 
                   ArrowBox[{{-0.7034568472075615, 0.5}, {-0.6411374095382172,
                     0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-0.7034568472075615, 0.5}, 1.6, {
                    1.5458899343553583`, 1.4314103928658497`}], 
                   ArrowBox[{{-0.49701647328636855`, 
                    2.0866260971051993`}, {-0.4811607986843567, 
                    2.0844823971287814`}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Dashing[{Small, Small}], 
                   Arrowheads[0.02], 
                   CircleBox[{-0.7034568472075615, 0.5}, 1.7, {
                    1.5458899343553583`, 1.581995226128471}], 
                   ArrowBox[{{-0.7054949755863842, 
                    2.199998778244476}, {-0.7224945781315462, 
                    2.1998933980697926`}}]}, 
                  {RGBColor[0, 0, 1], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-0.7034568472075615, 0.5}, 1.8, {
                    1.5458899343553583`, 1.6140649052059484`}], 
                   ArrowBox[{{-0.7633292424768465, 
                    2.299003973393255}, {-0.781315988783466, 
                    2.2983153099701576`}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.003], Arrowheads[0.025], 
                   CircleBox[{-0.7034568472075615, 0.5}, 1.5, {
                    1.5458899343553583`, 1.5465077191301182`}], 
                   ArrowBox[{{-0.6520340134708121, 
                    1.9991183049280943`}, {-0.6670275177885523, 
                    1.9995575694044165`}}]}, 
                  {RGBColor[0, 0, 1], ArrowBox[{{0, 2.8}, {1.5, 2.8}}]}},
                 Frame->True,
                 FrameTicks->Automatic,
                 ImageSize->800,
                 PlotRange->{{-3, 3}, {-1, 3}}], 135->
                GraphicsBox[{InsetBox[
                   TagBox[GridBox[{
                    {"\<\"t\"\>", "6.7`"},
                    {"\<\"driveIntent\"\>", "0.5`"},
                    {"\<\"motorTorque\"\>", 
                    RowBox[{"-", "12.17013428203564`"}]},
                    {"\<\"vw\"\>", "0.3707804600237266`"},
                    {"\<\"controllerState\"\>", 
                    RowBox[{"\[LeftAssociation]", 
                    RowBox[{
                    
                    RowBox[{"\<\"lastError\"\>", "\[Rule]", 
                    "0.014058775465746874`"}], ",", 
                    RowBox[{"\<\"lastInt\"\>", "\[Rule]", 
                    RowBox[{"-", "0.10905577003936326`"}]}], ",", 
                    RowBox[{"\<\"initialized\"\>", "\[Rule]", "True"}]}], 
                    "\[RightAssociation]"}]}
                    },
                    AutoDelete->False,
                    
                    GridBoxItemSize->{
                    "Columns" -> {{Automatic}}, "Rows" -> {{Automatic}}}],
                    "Grid"], {-1, -0.5}], 
                  {GrayLevel[0], LineBox[{{-5, 0}, {5, 0}}]}, 
                  {GrayLevel[0.5], CircleBox[{-0.6849427542795501, 0.5}, 0.5],
                    DiskBox[{-0.6849427542795501, 0.5}, 0.1]}, 
                  {GrayLevel[0.5], 
                   DiskBox[{-0.6478829042531269, 1.9995421192870906`}, 0.1]}, 
                  {GrayLevel[0.5], Thickness[0.01], 
                   LineBox[{{-0.6849427542795501, 0.5}, {-0.6478829042531269, 
                    1.9995421192870906`}}]}, 
                  {RGBColor[0, 1, 0], 
                   DiskBox[{-0.6602361875952679, 1.4996947461913939`}, 0.05]}, 
                  {RGBColor[0, 0, 1], Arrowheads[0.01], 
                   CircleBox[{-0.6849427542795501, 0.5}, 0.3333333333333333, 
                    NCache[{
                    Rational[-1, 2] 
                    Pi, -1.9359003552559657`}, {-1.5707963267948966`, \
-1.9359003552559657`}]], 
                   ArrowBox[{{-0.8008387227049065, 
                    0.18746322518164382`}, {-0.8039582436137567, 
                    0.18863779225851696`}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-0.6849427542795501, 0.5}, {-0.4415400686388373,
                     0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-0.6849427542795501, 0.5}, {-0.8666037737599681,
                     0.5}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.005], Arrowheads[0.04], 
                   ArrowBox[{{-0.6849427542795501, 0.5}, {-0.6232010881192553,
                     0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-0.6849427542795501, 0.5}, 1.6, {
                    1.5460872458790942`, 1.4325144401723369`}], 
                   ArrowBox[{{-0.4802542160885869, 
                    2.0868530500122713`}, {-0.4643961844042731, 
                    2.0847268567536332`}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Dashing[{Small, Small}], 
                   Arrowheads[0.02], 
                   CircleBox[{-0.6849427542795501, 0.5}, 1.7, {
                    1.5460872458790942`, 1.581857979867176}], 
                   ArrowBox[{{-0.6867475641633891, 
                    2.1999990419589306`}, {-0.7037471810114793, 
                    2.1998959949171253`}}]}, 
                  {RGBColor[0, 0, 1], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-0.6849427542795501, 0.5}, 1.8, {
                    1.5460872458790942`, 1.6136991030015144`}], 
                   ArrowBox[{{-0.7441570659384588, 
                    2.2990257544834543`}, {-0.7621440629562564, 
                    2.298343670697735}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.003], Arrowheads[0.025], 
                   CircleBox[{-0.6849427542795501, 0.5}, 1.5, {
                    1.5460872458790942`, 1.5466173135252832`}], 
                   ArrowBox[{{-0.6336842158151993, 
                    1.9991239315795406`}, {-0.648677768183821, 
                    1.9995615528491908`}}]}, 
                  {RGBColor[0, 0, 1], ArrowBox[{{0, 2.8}, {1.5, 2.8}}]}},
                 Frame->True,
                 FrameTicks->Automatic,
                 ImageSize->800,
                 PlotRange->{{-3, 3}, {-1, 3}}], 136->
                GraphicsBox[{InsetBox[
                   TagBox[GridBox[{
                    {"\<\"t\"\>", "6.75`"},
                    {"\<\"driveIntent\"\>", "0.5`"},
                    {"\<\"motorTorque\"\>", 
                    RowBox[{"-", "12.062055527496412`"}]},
                    {"\<\"vw\"\>", "0.3717898760604957`"},
                    {"\<\"controllerState\"\>", 
                    RowBox[{"\[LeftAssociation]", 
                    RowBox[{
                    
                    RowBox[{"\<\"lastError\"\>", "\[Rule]", 
                    "0.013967086806874779`"}], ",", 
                    RowBox[{"\<\"lastInt\"\>", "\[Rule]", 
                    RowBox[{"-", "0.10835523619411061`"}]}], ",", 
                    RowBox[{"\<\"initialized\"\>", "\[Rule]", "True"}]}], 
                    "\[RightAssociation]"}]}
                    },
                    AutoDelete->False,
                    
                    GridBoxItemSize->{
                    "Columns" -> {{Automatic}}, "Rows" -> {{Automatic}}}],
                    "Grid"], {-1, -0.5}], 
                  {GrayLevel[0], LineBox[{{-5, 0}, {5, 0}}]}, 
                  {GrayLevel[0.5], CircleBox[{-0.6663779495357669, 0.5}, 0.5],
                    DiskBox[{-0.6663779495357669, 0.5}, 0.1]}, 
                  {GrayLevel[0.5], 
                   DiskBox[{-0.6296351888026402, 1.9995499223212638`}, 0.1]}, 
                  {GrayLevel[0.5], Thickness[0.01], 
                   LineBox[{{-0.6663779495357669, 0.5}, {-0.6296351888026402, 
                    1.9995499223212638`}}]}, 
                  {RGBColor[0, 1, 0], 
                   DiskBox[{-0.6418827757136825, 1.4996999482141757`}, 0.05]}, 
                  {RGBColor[0, 0, 1], Arrowheads[0.01], 
                   CircleBox[{-0.6663779495357669, 0.5}, 0.3333333333333333, 
                    NCache[{
                    Rational[-1, 2] 
                    Pi, -1.9326579926197889`}, {-1.5707963267948966`, \
-1.9326579926197889`}]], 
                   ArrowBox[{{-0.7812599529738515, 
                    0.18708909191726314`}, {-0.7843832658508172, 
                    0.1882535382197642}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-0.6663779495357669, 
                    0.5}, {-0.42513683898583865`, 0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-0.6663779495357669, 0.5}, {-0.8464855889685219,
                     0.5}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.005], Arrowheads[0.04], 
                   ArrowBox[{{-0.6663779495357669, 0.5}, {-0.6052444784185936,
                     0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-0.6663779495357669, 0.5}, 1.6, {
                    1.5462987027385142`, 1.4336976419436651`}], 
                   ArrowBox[{{-0.4635671215248519, 
                    2.0870941270263486`}, {-0.44770658522584994`, 
                    2.084986698502811}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Dashing[{Small, Small}], 
                   Arrowheads[0.02], 
                   CircleBox[{-0.6663779495357669, 0.5}, 1.7, {
                    1.5462987027385142`, 1.581717260426075}], 
                   ArrowBox[{{-0.6679435364874694, 
                    2.199999279099111}, {-0.6849431676679688, 
                    2.1998986242348995`}}]}, 
                  {RGBColor[0, 0, 1], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-0.6663779495357669, 0.5}, 1.8, {
                    1.5462987027385142`, 1.6133101223357165`}], 
                   ArrowBox[{{-0.7248924704968169, 
                    2.299048651603591}, {-0.742879731471243, 
                    2.2983735644633727`}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.003], Arrowheads[0.025], 
                   CircleBox[{-0.6663779495357669, 0.5}, 1.5, {
                    1.5462987027385142`, 1.546745640119887}], 
                   ArrowBox[{{-0.6153117889619721, 
                    1.999130497069635}, {-0.6303053973655861, 
                    1.9995661942641711`}}]}, 
                  {RGBColor[0, 0, 1], ArrowBox[{{0, 2.8}, {1.5, 2.8}}]}},
                 Frame->True,
                 FrameTicks->Automatic,
                 ImageSize->800,
                 PlotRange->{{-3, 3}, {-1, 3}}], 137->
                GraphicsBox[{InsetBox[
                   TagBox[GridBox[{
                    {"\<\"t\"\>", "6.8`"},
                    {"\<\"driveIntent\"\>", "0.5`"},
                    {"\<\"motorTorque\"\>", 
                    RowBox[{"-", "11.948360881720145`"}]},
                    {"\<\"vw\"\>", "0.37278908692540014`"},
                    {"\<\"controllerState\"\>", 
                    RowBox[{"\[LeftAssociation]", 
                    RowBox[{
                    
                    RowBox[{"\<\"lastError\"\>", "\[Rule]", 
                    "0.013890429470878554`"}], ",", 
                    RowBox[{"\<\"lastInt\"\>", "\[Rule]", 
                    RowBox[{"-", "0.10765889501616713`"}]}], ",", 
                    RowBox[{"\<\"initialized\"\>", "\[Rule]", "True"}]}], 
                    "\[RightAssociation]"}]}
                    },
                    AutoDelete->False,
                    
                    GridBoxItemSize->{
                    "Columns" -> {{Automatic}}, "Rows" -> {{Automatic}}}],
                    "Grid"], {-1, -0.5}], 
                  {GrayLevel[0], LineBox[{{-5, 0}, {5, 0}}]}, 
                  {GrayLevel[0.5], CircleBox[{-0.647762932531214, 0.5}, 0.5], 
                   DiskBox[{-0.647762932531214, 0.5}, 0.1]}, 
                  {GrayLevel[0.5], 
                   DiskBox[{-0.6113551511657559, 1.9995580927246683`}, 0.1]}, 
                  {GrayLevel[0.5], Thickness[0.01], 
                   LineBox[{{-0.647762932531214, 0.5}, {-0.6113551511657559, 
                    1.9995580927246683`}}]}, 
                  {RGBColor[0, 1, 0], 
                   DiskBox[{-0.6234910782875752, 1.4997053951497787`}, 0.05]}, 
                  {RGBColor[0, 0, 1], Arrowheads[0.01], 
                   CircleBox[{-0.647762932531214, 0.5}, 0.3333333333333333, 
                    NCache[{
                    Rational[-1, 2] 
                    Pi, -1.929247153246501}, {-1.5707963267948966`, \
-1.929247153246501}]], 
                   ArrowBox[{{-0.7615769809351924, 
                    0.18669906879007214`}, {-0.7647042473757331, 
                    0.18785285522121753`}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-0.647762932531214, 0.5}, {-0.4087957148968111, 
                    0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-0.647762932531214, 0.5}, {-0.8262295244773686, 
                    0.5}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.005], Arrowheads[0.04], 
                   ArrowBox[{{-0.647762932531214, 0.5}, {-0.5872623068429658, 
                    0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-0.647762932531214, 0.5}, 1.6, {
                    1.546522088735014, 1.4349475983826485`}], 
                   ArrowBox[{{-0.4469360609619657, 
                    2.0873463918299966`}, {-0.4310729028599758, 
                    2.0852587899270825`}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Dashing[{Small, Small}], 
                   Arrowheads[0.02], 
                   CircleBox[{-0.647762932531214, 0.5}, 1.7, {
                    1.546522088735014, 1.5815741935189573`}], 
                   ArrowBox[{{-0.649085305828761, 
                    2.1999994856848817`}, {-0.6660849512356649, 
                    2.199901262906347}}]}, 
                  {RGBColor[0, 0, 1], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-0.647762932531214, 0.5}, 1.8, {
                    1.546522088735014, 1.6129018714112369`}], 
                   ArrowBox[{{-0.7055429853611921, 
                    2.299072390287552}, {-0.7235305204416019, 
                    2.2984046465193093`}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.003], Arrowheads[0.025], 
                   CircleBox[{-0.647762932531214, 0.5}, 1.5, {
                    1.546522088735014, 1.5468907569924277`}], 
                   ArrowBox[{{-0.5969143216236197, 
                    1.9991378918460998`}, {-0.6119079930963732, 
                    1.999571413210496}}]}, 
                  {RGBColor[0, 0, 1], ArrowBox[{{0, 2.8}, {1.5, 2.8}}]}},
                 Frame->True,
                 FrameTicks->Automatic,
                 ImageSize->800,
                 PlotRange->{{-3, 3}, {-1, 3}}], 138->
                GraphicsBox[{InsetBox[
                   TagBox[GridBox[{
                    {"\<\"t\"\>", "6.8500000000000005`"},
                    {"\<\"driveIntent\"\>", "0.5`"},
                    {"\<\"motorTorque\"\>", 
                    RowBox[{"-", "11.8300852229924`"}]},
                    {"\<\"vw\"\>", "0.37377773132105513`"},
                    {"\<\"controllerState\"\>", 
                    RowBox[{"\[LeftAssociation]", 
                    RowBox[{
                    
                    RowBox[{"\<\"lastError\"\>", "\[Rule]", 
                    "0.013826822988779585`"}], ",", 
                    RowBox[{"\<\"lastInt\"\>", "\[Rule]", 
                    RowBox[{"-", "0.10696604588187224`"}]}], ",", 
                    RowBox[{"\<\"initialized\"\>", "\[Rule]", "True"}]}], 
                    "\[RightAssociation]"}]}
                    },
                    AutoDelete->False,
                    
                    GridBoxItemSize->{
                    "Columns" -> {{Automatic}}, "Rows" -> {{Automatic}}}],
                    "Grid"], {-1, -0.5}], 
                  {GrayLevel[0], LineBox[{{-5, 0}, {5, 0}}]}, 
                  {GrayLevel[0.5], CircleBox[{-0.6290982231058937, 0.5}, 0.5],
                    DiskBox[{-0.6290982231058937, 0.5}, 0.1]}, 
                  {GrayLevel[0.5], 
                   DiskBox[{-0.593040182249465, 1.999566543268286}, 0.1]}, 
                  {GrayLevel[0.5], Thickness[0.01], 
                   LineBox[{{-0.6290982231058937, 0.5}, {-0.593040182249465, 
                    1.999566543268286}}]}, 
                  {RGBColor[0, 1, 0], 
                   DiskBox[{-0.6050595292016079, 1.499711028845524}, 0.05]}, 
                  {RGBColor[0, 0, 1], Arrowheads[0.01], 
                   CircleBox[{-0.6290982231058937, 0.5}, 0.3333333333333333, 
                    NCache[{
                    Rational[-1, 2] 
                    Pi, -1.9256988834846687`}, {-1.5707963267948966`, \
-1.9256988834846687`}]], 
                   ArrowBox[{{-0.7417998811509063, 
                    0.18629719895254976`}, {-0.7449312218418986, 
                    0.18743988175882742`}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-0.6290982231058937, 0.5}, {-0.3924965186460457,
                     0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-0.6290982231058937, 0.5}, {-0.8058514249475247,
                     0.5}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.005], Arrowheads[0.04], 
                   ArrowBox[{{-0.6290982231058937, 0.5}, {-0.5692497204876766,
                     0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-0.6290982231058937, 0.5}, 1.6, {
                    1.5467553171265565`, 1.4362526338875887`}], 
                   ArrowBox[{{-0.43034306536448114`, 
                    2.087607126234695}, {-0.4144771963770716, 
                    2.085540228088164}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Dashing[{Small, Small}], 
                   Arrowheads[0.02], 
                   CircleBox[{-0.6290982231058937, 0.5}, 1.7, {
                    1.5467553171265565`, 1.5814298026405362`}], 
                   ArrowBox[{{-0.6301751319714552, 
                    2.1999996589021116`}, {-0.647174791383622, 
                    2.1999038907183257`}}]}, 
                  {RGBColor[0, 0, 1], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-0.6290982231058937, 0.5}, 1.8, {
                    1.5467553171265565`, 1.6124780128098475`}], 
                   ArrowBox[{{-0.6861157184612894, 
                    2.2990967192520246`}, {-0.7041035349548451, 
                    2.2984365997149974`}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.003], Arrowheads[0.025], 
                   CircleBox[{-0.6290982231058937, 0.5}, 1.5, {
                    1.5467553171265565`, 1.5470507734074508`}], 
                   ArrowBox[{{-0.5784894995193499, 
                    1.9991460092655888`}, {-0.5934832401706793, 
                    1.9995771313908879`}}]}, 
                  {RGBColor[0, 0, 1], ArrowBox[{{0, 2.8}, {1.5, 2.8}}]}},
                 Frame->True,
                 FrameTicks->Automatic,
                 ImageSize->800,
                 PlotRange->{{-3, 3}, {-1, 3}}], 139->
                GraphicsBox[{InsetBox[
                   TagBox[GridBox[{
                    {"\<\"t\"\>", "6.9`"},
                    {"\<\"driveIntent\"\>", "0.5`"},
                    {"\<\"motorTorque\"\>", 
                    RowBox[{"-", "11.708193702373263`"}]},
                    {"\<\"vw\"\>", "0.3747555330444891`"},
                    {"\<\"controllerState\"\>", 
                    RowBox[{"\[LeftAssociation]", 
                    RowBox[{
                    
                    RowBox[{"\<\"lastError\"\>", "\[Rule]", 
                    "0.013774395936816251`"}], ",", 
                    RowBox[{"\<\"lastInt\"\>", "\[Rule]", 
                    RowBox[{"-", "0.10627608443661386`"}]}], ",", 
                    RowBox[{"\<\"initialized\"\>", "\[Rule]", "True"}]}], 
                    "\[RightAssociation]"}]}
                    },
                    AutoDelete->False,
                    
                    GridBoxItemSize->{
                    "Columns" -> {{Automatic}}, "Rows" -> {{Automatic}}}],
                    "Grid"], {-1, -0.5}], 
                  {GrayLevel[0], LineBox[{{-5, 0}, {5, 0}}]}, 
                  {GrayLevel[0.5], CircleBox[{-0.6103843569669236, 0.5}, 0.5],
                    DiskBox[{-0.6103843569669236, 0.5}, 0.1]}, 
                  {GrayLevel[0.5], 
                   DiskBox[{-0.5746878907563694, 1.999575193946632}, 0.1]}, 
                  {GrayLevel[0.5], Thickness[0.01], 
                   LineBox[{{-0.6103843569669236, 0.5}, {-0.5746878907563694, 
                    1.999575193946632}}]}, 
                  {RGBColor[0, 1, 0], 
                   DiskBox[{-0.5865867128265541, 1.4997167959644213`}, 0.05]}, 
                  {RGBColor[0, 0, 1], Arrowheads[0.01], 
                   CircleBox[{-0.6103843569669236, 0.5}, 0.3333333333333333, 
                    NCache[{
                    Rational[-1, 2] 
                    Pi, -1.9220421378660943`}, {-1.5707963267948966`, \
-1.9220421378660943`}]], 
                   ArrowBox[{{-0.7219381327146738, 
                    0.185887175958173}, {-0.7250736309609097, 
                    0.18701840063376973`}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-0.6103843569669236, 0.5}, {-0.3762204829194583,
                     0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-0.6103843569669236, 0.5}, {-0.7853661625569852,
                     0.5}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.005], Arrowheads[0.04], 
                   ArrowBox[{{-0.6103843569669236, 0.5}, {-0.5512022885095199,
                     0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-0.6103843569669236, 0.5}, 1.6, {
                    1.5469964358703865`, 1.43760182638172}], 
                   ArrowBox[{{-0.4137713670930111, 
                    2.087873840143744}, {-0.3979027239034624, 
                    2.0858283499832777`}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Dashing[{Small, Small}], 
                   Arrowheads[0.02], 
                   CircleBox[{-0.6103843569669236, 0.5}, 1.7, {
                    1.5469964358703865`, 1.5812850109609198`}], 
                   ArrowBox[{{-0.6112151200160969, 
                    2.199999797009622}, {-0.628214793116504, 
                    2.1999064902360708`}}]}, 
                  {RGBColor[0, 0, 1], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-0.6103843569669236, 0.5}, 1.8, {
                    1.5469964358703865`, 1.6120419564391268`}], 
                   ArrowBox[{{-0.6666173393404082, 
                    2.2991214110485663`}, {-0.6846054419731364, 
                    2.298469135276027}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.003], Arrowheads[0.025], 
                   CircleBox[{-0.6103843569669236, 0.5}, 1.5, {
                    1.5469964358703865`, 1.547223862200902}], 
                   ArrowBox[{{-0.560035119511147, 
                    1.9991547466114437`}, {-0.5750289345602804, 
                    1.999583273481819}}]}, 
                  {RGBColor[0, 0, 1], ArrowBox[{{0, 2.8}, {1.5, 2.8}}]}},
                 Frame->True,
                 FrameTicks->Automatic,
                 ImageSize->800,
                 PlotRange->{{-3, 3}, {-1, 3}}], 140->
                GraphicsBox[{InsetBox[
                   TagBox[GridBox[{
                    {"\<\"t\"\>", "6.95`"},
                    {"\<\"driveIntent\"\>", "0.5`"},
                    {"\<\"motorTorque\"\>", 
                    RowBox[{"-", "11.583580453494493`"}]},
                    {"\<\"vw\"\>", "0.3757222942859428`"},
                    {"\<\"controllerState\"\>", 
                    RowBox[{"\[LeftAssociation]", 
                    RowBox[{
                    
                    RowBox[{"\<\"lastError\"\>", "\[Rule]", 
                    "0.013731391328430344`"}], ",", 
                    RowBox[{"\<\"lastInt\"\>", "\[Rule]", 
                    RowBox[{"-", "0.10558849699757378`"}]}], ",", 
                    RowBox[{"\<\"initialized\"\>", "\[Rule]", "True"}]}], 
                    "\[RightAssociation]"}]}
                    },
                    AutoDelete->False,
                    
                    GridBoxItemSize->{
                    "Columns" -> {{Automatic}}, "Rows" -> {{Automatic}}}],
                    "Grid"], {-1, -0.5}], 
                  {GrayLevel[0], LineBox[{{-5, 0}, {5, 0}}]}, 
                  {GrayLevel[0.5], CircleBox[{-0.5916218816046083, 0.5}, 0.5],
                    DiskBox[{-0.5916218816046083, 0.5}, 0.1]}, 
                  {GrayLevel[0.5], 
                   DiskBox[{-0.5562961041183778, 1.9995839721219326`}, 0.1]}, 
                  {GrayLevel[0.5], Thickness[0.01], 
                   LineBox[{{-0.5916218816046083, 0.5}, {-0.5562961041183778, 
                    1.9995839721219326`}}]}, 
                  {RGBColor[0, 1, 0], 
                   DiskBox[{-0.5680713632804546, 1.4997226480812884`}, 0.05]}, 
                  {RGBColor[0, 0, 1], Arrowheads[0.01], 
                   CircleBox[{-0.5916218816046083, 0.5}, 0.3333333333333333, 
                    NCache[{
                    Rational[-1, 2] 
                    Pi, -1.9183037403997314`}, {-1.5707963267948966`, \
-1.9183037403997314`}]], 
                   ArrowBox[{{-0.7020006019865801, 
                    0.18547233953442255`}, {-0.7051403072801932, 
                    0.18659183459385137`}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-0.5916218816046083, 0.5}, {-0.3599502725347184,
                     0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-0.5916218816046083, 0.5}, {-0.7647876083264068,
                     0.5}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.005], Arrowheads[0.04], 
                   ArrowBox[{{-0.5916218816046083, 0.5}, {-0.5331159992565169,
                     0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-0.5916218816046083, 0.5}, 1.6, {
                    1.547243630968686, 1.4389850260357666`}], 
                   ArrowBox[{{-0.3972054256607099, 
                    2.08814427608395}, {-0.3813339683310502, 
                    2.0861207373750124`}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Dashing[{Small, Small}], 
                   Arrowheads[0.02], 
                   CircleBox[{-0.5916218816046083, 0.5}, 1.7, {
                    1.547243630968686, 1.5811406439654971`}], 
                   ArrowBox[{{-0.5922072207830634, 
                    2.199999899228834}, {-0.6092069071767369, 
                    2.199909046647975}}]}, 
                  {RGBColor[0, 0, 1], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-0.5916218816046083, 0.5}, 1.8, {
                    1.547243630968686, 1.611596855710322}], 
                   ArrowBox[{{-0.6470540681829792, 
                    2.299146262173018}, {-0.665042459362266, 
                    2.298501992982419}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.003], Arrowheads[0.025], 
                   CircleBox[{-0.5916218816046083, 0.5}, 1.5, {
                    1.547243630968686, 1.547408270126588}], 
                   ArrowBox[{{-0.541549101020465, 
                    1.99916400591949}, {-0.5565429948384072, 
                    1.9995897678042631`}}]}, 
                  {RGBColor[0, 0, 1], ArrowBox[{{0, 2.8}, {1.5, 2.8}}]}},
                 Frame->True,
                 FrameTicks->Automatic,
                 ImageSize->800,
                 PlotRange->{{-3, 3}, {-1, 3}}], 141->
                GraphicsBox[{InsetBox[
                   TagBox[GridBox[{
                    {"\<\"t\"\>", "7.`"},
                    {"\<\"driveIntent\"\>", "0"},
                    {"\<\"motorTorque\"\>", 
                    RowBox[{"-", "200"}]},
                    {"\<\"vw\"\>", "0.37667788894135673`"},
                    {"\<\"controllerState\"\>", 
                    RowBox[{"\[LeftAssociation]", 
                    RowBox[{
                    RowBox[{"\<\"lastError\"\>", "\[Rule]", 
                    RowBox[{"-", "0.13630382971162103`"}]}], ",", 
                    RowBox[{"\<\"lastInt\"\>", "\[Rule]", 
                    RowBox[{"-", "0.10505285473136432`"}]}], ",", 
                    RowBox[{"\<\"initialized\"\>", "\[Rule]", "True"}]}], 
                    "\[RightAssociation]"}]}
                    },
                    AutoDelete->False,
                    
                    GridBoxItemSize->{
                    "Columns" -> {{Automatic}}, "Rows" -> {{Automatic}}}],
                    "Grid"], {-1, -0.5}], 
                  {GrayLevel[0], LineBox[{{-5, 0}, {5, 0}}]}, 
                  {GrayLevel[0.5], CircleBox[{-0.572811352543507, 0.5}, 0.5], 
                   DiskBox[{-0.572811352543507, 0.5}, 0.1]}, 
                  {GrayLevel[0.5], 
                   DiskBox[{-0.5378628672025433, 1.9995928125235771`}, 0.1]}, 
                  {GrayLevel[0.5], Thickness[0.01], 
                   LineBox[{{-0.572811352543507, 0.5}, {-0.5378628672025433, 
                    1.9995928125235771`}}]}, 
                  {RGBColor[0, 1, 0], 
                   DiskBox[{-0.5495123623161978, 1.499728541682385}, 0.05]}, 
                  {RGBColor[0, 0, 1], Arrowheads[0.01], 
                   CircleBox[{-0.572811352543507, 0.5}, 0.3333333333333333, 
                    NCache[{
                    Rational[-1, 2] 
                    Pi, -7.570796326794897}, {-1.5707963267948966`, \
-7.570796326794897}]], 
                   ArrowBox[{{-0.4764769957503866, 
                    0.18089060995895645`}, {-0.47967285314386504`, 
                    0.17994323778321136`}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-0.572811352543507, 0.5}, {3.427188647456493, 
                    0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-0.572811352543507, 0.5}, {-0.7441286163040575, 
                    0.5}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.005], Arrowheads[0.04], 
                   ArrowBox[{{-0.572811352543507, 0.5}, {3.2558713836959425`, 
                    0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-0.572811352543507, 0.5}, 1.6, {
                    1.5474952281037453`, 1.4403928644259445`}], 
                   ArrowBox[{{-0.38063093900958606`, 
                    2.0884164090860846`}, {-0.36475664859407153`, 
                    2.086415216822038}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Dashing[{Small, Small}], 
                   Arrowheads[0.02], 
                   CircleBox[{-0.572811352543507, 0.5}, 1.7, {
                    1.5474952281037453`, 3.7657626110760156`}], 
                   ArrowBox[{{-1.962139864032219, -0.47967662377162}, \
{-1.9522737952264342`, -0.49352069390992137`}}]}, 
                  {RGBColor[0, 0, 1], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-0.572811352543507, 0.5}, 1.8, {
                    1.5474952281037453`, 2.6586063392148565`}], 
                   ArrowBox[{{-2.158473633629618, 
                    1.3518656762311712`}, {-2.166912865961663, 
                    1.3359667247671672`}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.003], Arrowheads[0.025], 
                   CircleBox[{-0.572811352543507, 0.5}, 1.5, {
                    1.5474952281037453`, 5.28062240754933}], 
                   ArrowBox[{{0.22172159542400438`, -0.7722882513778297}, {
                    0.23440453957445662`, -0.7642794404372002}}]}, 
                  {RGBColor[0, 0, 1], ArrowBox[{{0, 2.8}, {0, 2.8}}]}},
                 Frame->True,
                 FrameTicks->Automatic,
                 ImageSize->800,
                 PlotRange->{{-3, 3}, {-1, 3}}], 142->
                GraphicsBox[{InsetBox[
                   TagBox[GridBox[{
                    {"\<\"t\"\>", "7.05`"},
                    {"\<\"driveIntent\"\>", "0"},
                    {"\<\"motorTorque\"\>", 
                    RowBox[{"-", "32.434488655024865`"}]},
                    {"\<\"vw\"\>", "0.40233982686263414`"},
                    {"\<\"controllerState\"\>", 
                    RowBox[{"\[LeftAssociation]", 
                    RowBox[{
                    RowBox[{"\<\"lastError\"\>", "\[Rule]", 
                    RowBox[{"-", "0.11932508424646393`"}]}], ",", 
                    RowBox[{"\<\"lastInt\"\>", "\[Rule]", 
                    RowBox[{"-", "0.11163811968216329`"}]}], ",", 
                    RowBox[{"\<\"initialized\"\>", "\[Rule]", "True"}]}], 
                    "\[RightAssociation]"}]}
                    },
                    AutoDelete->False,
                    
                    GridBoxItemSize->{
                    "Columns" -> {{Automatic}}, "Rows" -> {{Automatic}}}],
                    "Grid"], {-1, -0.5}], 
                  {GrayLevel[0], LineBox[{{-5, 0}, {5, 0}}]}, 
                  {GrayLevel[0.5], CircleBox[{-0.5531755464679811, 0.5}, 0.5],
                    DiskBox[{-0.5531755464679811, 0.5}, 0.1]}, 
                  {GrayLevel[0.5], 
                   DiskBox[{-0.5563566913123954, 1.9999966267687}, 0.1]}, 
                  {GrayLevel[0.5], Thickness[0.01], 
                   LineBox[{{-0.5531755464679811, 0.5}, {-0.5563566913123954, 
                    1.9999966267687}}]}, 
                  {RGBColor[0, 1, 0], 
                   DiskBox[{-0.5552963096975906, 1.4999977511791334`}, 0.05]}, 
                  {RGBColor[0, 0, 1], Arrowheads[0.01], 
                   CircleBox[{-0.5531755464679811, 0.5}, 0.3333333333333333, 
                    NCache[{
                    Rational[-1, 2] 
                    Pi, -2.5438309864456423`}, {-1.5707963267948966`, \
-2.5438309864456423`}]], 
                   ArrowBox[{{-0.8268182903214525, 
                    0.3096562061756293}, {-0.8287080145127138, 
                    0.31240210511764954`}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-0.5531755464679811, 0.5}, {0.0955142266325163, 
                    0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-0.5531755464679811, 0.5}, {-0.5375773945012718,
                     0.5}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.005], Arrowheads[0.04], 
                   ArrowBox[{{-0.5531755464679811, 0.5}, {
                    0.11111237859922551`, 0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-0.5531755464679811, 0.5}, 1.6, {
                    1.5729170916142463`, 1.582665958516895}], 
                   ArrowBox[{{-0.5561669554804213, 
                    2.099997203582594}, {-0.5721665112821758, 
                    2.0998872907975192`}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Dashing[{Small, Small}], 
                   Arrowheads[0.02], 
                   CircleBox[{-0.5531755464679811, 0.5}, 1.7, {
                    1.5729170916142463`, 1.957899066744607}], 
                   ArrowBox[{{-1.179163713341122, 
                    2.080550162106475}, {-1.1949376523909607`, 
                    2.0742113579190384`}}]}, 
                  {RGBColor[0, 0, 1], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-0.5531755464679811, 0.5}, 1.8, {
                    1.5729170916142463`, 1.7531086952532735`}], 
                   ArrowBox[{{-0.8618052202000999, 
                    2.2733436566249665`}, {-0.8795229298554599, 
                    2.270168744881726}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.003], Arrowheads[0.025], 
                   CircleBox[{-0.5531755464679811, 0.5}, 1.5, {
                    1.5729170916142463`, 2.2358587124916456`}], 
                   ArrowBox[{{-1.4669886219552557`, 
                    1.6895148856018944`}, {-1.4788378822867647`, 
                    1.6803174318997893`}}]}, 
                  {RGBColor[0, 0, 1], ArrowBox[{{0, 2.8}, {0, 2.8}}]}},
                 Frame->True,
                 FrameTicks->Automatic,
                 ImageSize->800,
                 PlotRange->{{-3, 3}, {-1, 3}}], 143->
                GraphicsBox[{InsetBox[
                   TagBox[GridBox[{
                    {"\<\"t\"\>", "7.1000000000000005`"},
                    {"\<\"driveIntent\"\>", "0"},
                    {"\<\"motorTorque\"\>", "21.87927990231364`"},
                    {"\<\"vw\"\>", "0.4062556954237965`"},
                    {"\<\"controllerState\"\>", 
                    RowBox[{"\[LeftAssociation]", 
                    RowBox[{
                    RowBox[{"\<\"lastError\"\>", "\[Rule]", 
                    RowBox[{"-", "0.07475405783552842`"}]}], ",", 
                    RowBox[{"\<\"lastInt\"\>", "\[Rule]", 
                    RowBox[{"-", "0.11651546916477983`"}]}], ",", 
                    RowBox[{"\<\"initialized\"\>", "\[Rule]", "True"}]}], 
                    "\[RightAssociation]"}]}
                    },
                    AutoDelete->False,
                    GridBoxItemSize->{
                    "Columns" -> {{Automatic}}, "Rows" -> {{Automatic}}}],
                    "Grid"], {-1, -0.5}], 
                  {GrayLevel[0], LineBox[{{-5, 0}, {5, 0}}]}, 
                  {GrayLevel[0.5], CircleBox[{-0.5329079158255796, 0.5}, 0.5],
                    DiskBox[{-0.5329079158255796, 0.5}, 0.1]}, 
                  {GrayLevel[0.5], 
                   DiskBox[{-0.6050287676070776, 1.9982651910587494`}, 0.1]}, 
                  {GrayLevel[0.5], Thickness[0.01], 
                   LineBox[{{-0.5329079158255796, 0.5}, {-0.6050287676070776, 
                    1.9982651910587494`}}]}, 
                  {RGBColor[0, 1, 0], 
                   DiskBox[{-0.5809884836799115, 1.498843460705833}, 0.05]}, 
                  {RGBColor[0, 0, 1], Arrowheads[0.01], 
                   CircleBox[{-0.5329079158255796, 0.5}, 0.3333333333333333, 
                    NCache[{
                    Rational[-1, 2] 
                    Pi, -0.9144179297254873}, {-1.5707963267948966`, \
-0.9144179297254873}]], 
                   ArrowBox[{{-0.3321414678888442, 
                    0.23390989403215834`}, {-0.32949064941603945`, 
                    0.23593082944504623`}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-0.5329079158255796, 0.5}, {-0.9704935138718525,
                     0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-0.5329079158255796, 
                    0.5}, {-0.17968492953947712`, 0.5}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.005], Arrowheads[0.04], 
                   ArrowBox[{{-0.5329079158255796, 0.5}, {-0.61727052758575, 
                    0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-0.5329079158255796, 0.5}, 1.6, {
                    1.6188954389170516`, 1.8399154236429975`}], 
                   ArrowBox[{{-0.942874565871249, 
                    2.046585705950476}, {-0.9583196670060748, 
                    2.0424087791365637`}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Dashing[{Small, Small}], 
                   Arrowheads[0.02], 
                   CircleBox[{-0.5329079158255796, 0.5}, 1.7, {
                    1.6188954389170516`, 1.5700615471242165`}], 
                   ArrowBox[{{-0.5486585650325703, 
                    2.1999270328603986`}, {-0.5316587904978242, 
                    2.199999541084031}}]}, 
                  {RGBColor[0, 0, 1], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-0.5329079158255796, 0.5}, 1.8, {
                    1.6188954389170516`, 1.4973438839041981`}], 
                   ArrowBox[{{-0.41877014510075544`, 
                    2.2963776243579654`}, {-0.4008123750929285, 
                    2.295146447540854}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.003], Arrowheads[0.025], 
                   CircleBox[{-0.5329079158255796, 0.5}, 1.5, {
                    1.6188954389170516`, 1.6534431459107564`}], 
                   ArrowBox[{{-0.6417823202961412, 
                    1.9960435702382402`}, {-0.656737062984316, 
                    1.9948800427840154`}}]}, 
                  {RGBColor[0, 0, 1], ArrowBox[{{0, 2.8}, {0, 2.8}}]}},
                 Frame->True,
                 FrameTicks->Automatic,
                 ImageSize->800,
                 PlotRange->{{-3, 3}, {-1, 3}}], 144->
                GraphicsBox[{InsetBox[
                   TagBox[GridBox[{
                    {"\<\"t\"\>", "7.15`"},
                    {"\<\"driveIntent\"\>", "0"},
                    {"\<\"motorTorque\"\>", "54.9874745789061`"},
                    {"\<\"vw\"\>", "0.40200541463206024`"},
                    {"\<\"controllerState\"\>", 
                    RowBox[{"\[LeftAssociation]", 
                    RowBox[{
                    RowBox[{"\<\"lastError\"\>", "\[Rule]", 
                    RowBox[{"-", "0.025661831375735383`"}]}], ",", 
                    RowBox[{"\<\"lastInt\"\>", "\[Rule]", 
                    RowBox[{"-", "0.11899895441988904`"}]}], ",", 
                    RowBox[{"\<\"initialized\"\>", "\[Rule]", "True"}]}], 
                    "\[RightAssociation]"}]}
                    },
                    AutoDelete->False,
                    
                    GridBoxItemSize->{
                    "Columns" -> {{Automatic}}, "Rows" -> {{Automatic}}}],
                    "Grid"], {-1, -0.5}], 
                  {GrayLevel[0], LineBox[{{-5, 0}, {5, 0}}]}, 
                  {GrayLevel[0.5], CircleBox[{-0.512681969399285, 0.5}, 0.5], 
                   DiskBox[{-0.512681969399285, 0.5}, 0.1]}, 
                  {GrayLevel[0.5], 
                   DiskBox[{-0.6563206574120601, 1.9931068037169908`}, 0.1]}, 
                  {GrayLevel[0.5], Thickness[0.01], 
                   LineBox[{{-0.512681969399285, 0.5}, {-0.6563206574120601, 
                    1.9931068037169908`}}]}, 
                  {RGBColor[0, 1, 0], 
                   DiskBox[{-0.6084410947411352, 1.4954045358113273`}, 0.05]}, 
                  {RGBColor[0, 0, 1], Arrowheads[0.01], 
                   CircleBox[{-0.512681969399285, 0.5}, 0.3333333333333333, 
                    NCache[{
                    Rational[-1, 2] Pi, 
                    0.07882791057228644}, {-1.5707963267948966`, 
                    0.07882791057228644}]], 
                   ArrowBox[{{-0.18013787130198738`, 
                    0.5229245268602427}, {-0.1803837398161997, 
                    0.5262487662006854}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-0.512681969399285, 0.5}, {-1.612431460977407, 
                    0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-0.512681969399285, 0.5}, {0.1883885821238539, 
                    0.5}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.005], Arrowheads[0.04], 
                   ArrowBox[{{-0.512681969399285, 0.5}, {-0.9113609094542682, 
                    0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-0.512681969399285, 0.5}, 1.6, {
                    1.6667024081818378`, 2.1068943893694883`}], 
                   ArrowBox[{{-1.3161425808415859`, 
                    1.8836368909004864`}, {-1.329938546449797, 
                    1.8755332374274643`}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Dashing[{Small, Small}], 
                   Arrowheads[0.02], 
                   CircleBox[{-0.512681969399285, 0.5}, 1.7, {
                    1.6667024081818378`, 1.436736475841695}], 
                   ArrowBox[{{-0.3023207993535242, 
                    2.1869345506382216`}, {-0.28546225297234673`, 
                    2.184746627973139}}]}, 
                  {RGBColor[0, 0, 1], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-0.512681969399285, 0.5}, 1.8, {
                    1.6667024081818378`, 1.3612164382990262`}], 
                   ArrowBox[{{-0.15581833254336125`, 
                    2.2642699183203128`}, {-0.1381937704368244, 
                    2.260613128667923}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.003], Arrowheads[0.025], 
                   CircleBox[{-0.512681969399285, 0.5}, 1.5, {
                    1.6667024081818378`, 1.5090293009371292`}], 
                   ArrowBox[{{-0.43506610760797315`, 
                    1.9979905800766544`}, {-0.4200903322318047, 
                    1.9971395354898105`}}]}, 
                  {RGBColor[0, 0, 1], ArrowBox[{{0, 2.8}, {0, 2.8}}]}},
                 Frame->True,
                 FrameTicks->Automatic,
                 ImageSize->800,
                 PlotRange->{{-3, 3}, {-1, 3}}], 145->
                GraphicsBox[{InsetBox[
                   TagBox[GridBox[{
                    {"\<\"t\"\>", "7.2`"},
                    {"\<\"driveIntent\"\>", "0"},
                    {"\<\"motorTorque\"\>", "79.09162754582977`"},
                    {"\<\"vw\"\>", "0.3939059406783106`"},
                    {"\<\"controllerState\"\>", 
                    RowBox[{"\[LeftAssociation]", 
                    RowBox[{
                    
                    RowBox[{"\<\"lastError\"\>", "\[Rule]", 
                    "0.020739877314918304`"}], ",", 
                    RowBox[{"\<\"lastInt\"\>", "\[Rule]", 
                    RowBox[{"-", "0.11908114921606804`"}]}], ",", 
                    RowBox[{"\<\"initialized\"\>", "\[Rule]", "True"}]}], 
                    "\[RightAssociation]"}]}
                    },
                    AutoDelete->False,
                    
                    GridBoxItemSize->{
                    "Columns" -> {{Automatic}}, "Rows" -> {{Automatic}}}],
                    "Grid"], {-1, -0.5}], 
                  {GrayLevel[0], LineBox[{{-5, 0}, {5, 0}}]}, 
                  {GrayLevel[0.5], CircleBox[{-0.4927763781331333, 0.5}, 0.5],
                    DiskBox[{-0.4927763781331333, 0.5}, 0.1]}, 
                  {GrayLevel[0.5], 
                   DiskBox[{-0.7017844345334676, 1.9853671708906706`}, 0.1]}, 
                  {GrayLevel[0.5], Thickness[0.01], 
                   LineBox[{{-0.4927763781331333, 0.5}, {-0.7017844345334676, 
                    1.9853671708906706`}}]}, 
                  {RGBColor[0, 1, 0], 
                   DiskBox[{-0.6321150824000228, 1.4902447805937804`}, 0.05]}, 
                  {RGBColor[0, 0, 1], Arrowheads[0.01], 
                   CircleBox[{-0.4927763781331333, 0.5}, 0.3333333333333333, 
                    NCache[{
                    Rational[-1, 2] Pi, 
                    0.8019524995799965}, {-1.5707963267948966`, 
                    0.8019524995799965}]], 
                   ArrowBox[{{-0.2586240413167711, 
                    0.737242058401431}, {-0.2610081298799177, 
                    0.7395716807403308}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-0.4927763781331333, 
                    0.5}, {-2.0746089290497287`, 0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-0.4927763781331333, 0.5}, {0.5220605653143933, 
                    0.5}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.005], Arrowheads[0.04], 
                   ArrowBox[{{-0.4927763781331333, 0.5}, {-1.059771985602202, 
                    0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-0.4927763781331333, 0.5}, 1.6, {
                    1.7105898997057918`, 2.351111417299023}], 
                   ArrowBox[{{-1.6069546355756312`, 
                    1.6483060613975697`}, {-1.6183817963575873`, 
                    1.6371070496939821`}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Dashing[{Small, Small}], 
                   Arrowheads[0.02], 
                   CircleBox[{-0.4927763781331333, 0.5}, 1.7, {
                    1.7105898997057918`, 1.3852686255369486`}], 
                   ArrowBox[{{-0.1959091950394879, 
                    2.1738786920210327`}, {-0.17918553033312506`, 
                    2.1708263764305586`}}]}, 
                  {RGBColor[0, 0, 1], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-0.4927763781331333, 0.5}, 1.8, {
                    1.7105898997057918`, 1.2711919688956264`}], 
                   ArrowBox[{{0.021255167872740788`, 2.2250424834510047`}, {
                    0.03847960383848653, 2.2198160022570352`}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.003], Arrowheads[0.025], 
                   CircleBox[{-0.4927763781331333, 0.5}, 1.5, {
                    1.7105898997057918`, 1.4978378907750172`}], 
                   ArrowBox[{{-0.3984010998459241, 
                    1.9970281583351102`}, {-0.38343578649061, 
                    1.9960095704972156`}}]}, 
                  {RGBColor[0, 0, 1], ArrowBox[{{0, 2.8}, {0, 2.8}}]}},
                 Frame->True,
                 FrameTicks->Automatic,
                 ImageSize->800,
                 PlotRange->{{-3, 3}, {-1, 3}}], 146->
                GraphicsBox[{InsetBox[
                   TagBox[GridBox[{
                    {"\<\"t\"\>", "7.25`"},
                    {"\<\"driveIntent\"\>", "0"},
                    {"\<\"motorTorque\"\>", "98.06461382653859`"},
                    {"\<\"vw\"\>", "0.383448396409057`"},
                    {"\<\"controllerState\"\>", 
                    RowBox[{"\[LeftAssociation]", 
                    RowBox[{
                    
                    RowBox[{"\<\"lastError\"\>", "\[Rule]", 
                    "0.06228852825159033`"}], ",", 
                    RowBox[{"\<\"lastInt\"\>", "\[Rule]", 
                    RowBox[{"-", "0.11696265289917594`"}]}], ",", 
                    RowBox[{"\<\"initialized\"\>", "\[Rule]", "True"}]}], 
                    "\[RightAssociation]"}]}
                    },
                    AutoDelete->False,
                    
                    GridBoxItemSize->{
                    "Columns" -> {{Automatic}}, "Rows" -> {{Automatic}}}],
                    "Grid"], {-1, -0.5}], 
                  {GrayLevel[0], LineBox[{{-5, 0}, {5, 0}}]}, 
                  {GrayLevel[0.5], 
                   CircleBox[{-0.47333955906066677`, 0.5}, 0.5], 
                   DiskBox[{-0.47333955906066677`, 0.5}, 0.1]}, 
                  {GrayLevel[0.5], 
                   DiskBox[{-0.7390786227452646, 1.9762732640104383`}, 0.1]}, 
                  {GrayLevel[0.5], Thickness[0.01], 
                   LineBox[{{-0.47333955906066677`, 
                    0.5}, {-0.7390786227452646, 1.9762732640104383`}}]}, 
                  {RGBColor[0, 1, 0], 
                   DiskBox[{-0.6504989348503987, 1.484182176006959}, 0.05]}, 
                  {RGBColor[0, 0, 1], Arrowheads[0.01], 
                   CircleBox[{-0.47333955906066677`, 0.5}, 0.3333333333333333,
                     NCache[{
                    Rational[-1, 2] Pi, 
                    1.3711420880012612`}, {-1.5707963267948966`, 
                    1.3711420880012612`}]], 
                   ArrowBox[{{-0.40396565164629156`, 
                    0.8260343112023224}, {-0.40722940908599964`, 
                    0.8267117371344931}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-0.47333955906066677`, 
                    0.5}, {-2.4346318355914387`, 0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-0.47333955906066677`, 0.5}, {0.809054731716488,
                     0.5}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.005], Arrowheads[0.04], 
                   ArrowBox[{{-0.47333955906066677`, 
                    0.5}, {-1.1522375448142839`, 0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-0.47333955906066677`, 0.5}, 1.6, {
                    1.7488957449958626`, 2.563273866521663}], 
                   ArrowBox[{{-1.8043393441630564`, 
                    1.387941198535913}, {-1.8131520587242815`, 
                    1.3745870258271249`}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Dashing[{Small, Small}], 
                   Arrowheads[0.02], 
                   CircleBox[{-0.47333955906066677`, 0.5}, 1.7, {
                    1.7488957449958626`, 1.361807163614988}], 
                   ArrowBox[{{-0.13728603752214258`, 
                    2.1664537289290564`}, {-0.12063858050980653`, 
                    2.163009927730218}}]}, 
                  {RGBColor[0, 0, 1], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-0.47333955906066677`, 0.5}, 1.8, {
                    1.7488957449958626`, 1.204092334848426}], 
                   ArrowBox[{{0.1551981243567962, 2.186695105976243}, {
                    0.17203336767983218`, 2.1803254998453134`}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.003], Arrowheads[0.025], 
                   CircleBox[{-0.47333955906066677`, 0.5}, 1.5, {
                    1.7488957449958626`, 1.5251012568814677`}], 
                   ArrowBox[{{-0.419808323577582, 
                    1.9990444979478275`}, {-0.40482080497707373`, 
                    1.9984342429145272`}}]}, 
                  {RGBColor[0, 0, 1], ArrowBox[{{0, 2.8}, {0, 2.8}}]}},
                 Frame->True,
                 FrameTicks->Automatic,
                 ImageSize->800,
                 PlotRange->{{-3, 3}, {-1, 3}}], 147->
                GraphicsBox[{InsetBox[
                   TagBox[GridBox[{
                    {"\<\"t\"\>", "7.3`"},
                    {"\<\"driveIntent\"\>", "0"},
                    {"\<\"motorTorque\"\>", "113.22018778122646`"},
                    {"\<\"vw\"\>", "0.37126449976702136`"},
                    {"\<\"controllerState\"\>", 
                    RowBox[{"\[LeftAssociation]", 
                    RowBox[{
                    
                    RowBox[{"\<\"lastError\"\>", "\[Rule]", 
                    "0.09842358956323502`"}], ",", 
                    RowBox[{"\<\"lastInt\"\>", "\[Rule]", 
                    RowBox[{"-", "0.1129039580817279`"}]}], ",", 
                    RowBox[{"\<\"initialized\"\>", "\[Rule]", "True"}]}], 
                    "\[RightAssociation]"}]}
                    },
                    AutoDelete->False,
                    
                    GridBoxItemSize->{
                    "Columns" -> {{Automatic}}, "Rows" -> {{Automatic}}}],
                    "Grid"], {-1, -0.5}], 
                  {GrayLevel[0], LineBox[{{-5, 0}, {5, 0}}]}, 
                  {GrayLevel[0.5], 
                   CircleBox[{-0.45447152264827395`, 0.5}, 0.5], 
                   DiskBox[{-0.45447152264827395`, 0.5}, 0.1]}, 
                  {GrayLevel[0.5], 
                   DiskBox[{-0.7678505611914879, 1.9668993074515138`}, 0.1]}, 
                  {GrayLevel[0.5], Thickness[0.01], 
                   LineBox[{{-0.45447152264827395`, 
                    0.5}, {-0.7678505611914879, 1.9668993074515138`}}]}, 
                  {RGBColor[0, 1, 0], 
                   DiskBox[{-0.6633908816770832, 1.4779328716343425`}, 0.05]}, 
                  {RGBColor[0, 0, 1], Arrowheads[0.01], 
                   CircleBox[{-0.45447152264827395`, 0.5}, 0.3333333333333333,
                     NCache[{
                    Rational[-1, 2] Pi, 
                    1.8258093066418972`}, {-1.5707963267948966`, 
                    1.8258093066418972`}]], 
                   ArrowBox[{{-0.5353278283156696, 
                    0.823378058848976}, {-0.5385575122264923, 
                    0.8225533405000836}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-0.45447152264827395`, 
                    0.5}, {-2.718875278272803, 0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-0.45447152264827395`, 0.5}, {1.048219418087057,
                     0.5}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.005], Arrowheads[0.04], 
                   ArrowBox[{{-0.45447152264827395`, 
                    0.5}, {-1.2161843375374721`, 0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-0.45447152264827395`, 0.5}, 1.6, {
                    1.781266130149604, 2.7416406765026693`}], 
                   ArrowBox[{{-1.921895465272426, 
                    1.1377044555382918`}, {-1.9281990329585579`, 
                    1.122998575724417}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Dashing[{Small, Small}], 
                   Arrowheads[0.02], 
                   CircleBox[{-0.45447152264827395`, 0.5}, 1.7, {
                    1.781266130149604, 1.3497786631097581`}], 
                   ArrowBox[{{-0.09839785558157349, 
                    2.1622910526203505`}, {-0.08179302563748331, 
                    2.158647261434986}}]}, 
                  {RGBColor[0, 0, 1], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-0.45447152264827395`, 0.5}, 1.8, {
                    1.781266130149604, 1.1522650869205682`}], 
                   ArrowBox[{{0.2605998397923983, 2.1518695307490967`}, {
                    0.2770825065195006, 2.144636343514394}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.003], Arrowheads[0.025], 
                   CircleBox[{-0.45447152264827395`, 0.5}, 1.5, {
                    1.781266130149604, 1.5618452650728716`}], 
                   ArrowBox[{{-0.4560449297767072, 
                    1.9999991747964423`}, {-0.4410451093576544, 
                    1.9999399092717514`}}]}, 
                  {RGBColor[0, 0, 1], ArrowBox[{{0, 2.8}, {0, 2.8}}]}},
                 Frame->True,
                 FrameTicks->Automatic,
                 ImageSize->800,
                 PlotRange->{{-3, 3}, {-1, 3}}], 148->
                GraphicsBox[{InsetBox[
                   TagBox[GridBox[{
                    {"\<\"t\"\>", "7.3500000000000005`"},
                    {"\<\"driveIntent\"\>", "0"},
                    {"\<\"motorTorque\"\>", "125.11954695056869`"},
                    {"\<\"vw\"\>", "0.3577373680045059`"},
                    {"\<\"controllerState\"\>", 
                    RowBox[{"\[LeftAssociation]", 
                    RowBox[{
                    
                    RowBox[{"\<\"lastError\"\>", "\[Rule]", 
                    "0.12912218841639322`"}], ",", 
                    RowBox[{"\<\"lastInt\"\>", "\[Rule]", 
                    RowBox[{"-", "0.10717760788894534`"}]}], ",", 
                    RowBox[{"\<\"initialized\"\>", "\[Rule]", "True"}]}], 
                    "\[RightAssociation]"}]}
                    },
                    AutoDelete->False,
                    
                    GridBoxItemSize->{
                    "Columns" -> {{Automatic}}, "Rows" -> {{Automatic}}}],
                    "Grid"], {-1, -0.5}], 
                  {GrayLevel[0], LineBox[{{-5, 0}, {5, 0}}]}, 
                  {GrayLevel[0.5], CircleBox[{-0.4362483268941677, 0.5}, 0.5],
                    DiskBox[{-0.4362483268941677, 0.5}, 0.1]}, 
                  {GrayLevel[0.5], 
                   DiskBox[{-0.7884340092047706, 1.9580690124872058`}, 0.1]}, 
                  {GrayLevel[0.5], Thickness[0.01], 
                   LineBox[{{-0.4362483268941677, 0.5}, {-0.7884340092047706, 
                    1.9580690124872058`}}]}, 
                  {RGBColor[0, 1, 0], 
                   DiskBox[{-0.671038781767903, 1.4720460083248037`}, 0.05]}, 
                  {RGBColor[0, 0, 1], Arrowheads[0.01], 
                   CircleBox[{-0.4362483268941677, 0.5}, 0.3333333333333333, 
                    NCache[{
                    Rational[-1, 2] Pi, 
                    2.182790081722164}, {-1.5707963267948966`, 
                    2.182790081722164}]], 
                   ArrowBox[{{-0.625010615897708, 
                    0.7747360721879365}, {-0.6277484927946714, 
                    0.7728347440689889}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-0.4362483268941677, 
                    0.5}, {-2.9386392659055414`, 0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-0.4362483268941677, 0.5}, {1.242359320616969, 
                    0.5}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.005], Arrowheads[0.04], 
                   ArrowBox[{{-0.4362483268941677, 0.5}, {-1.260031618394405, 
                    0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-0.4362483268941677, 0.5}, 1.6, {
                    1.8077993180731686`, 2.8870998560204417`}], 
                   ArrowBox[{{-1.9806086753036904`, 
                    0.9182715795513943}, {-1.9847141040136835`, 
                    0.9028073200547055}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Dashing[{Small, Small}], 
                   Arrowheads[0.02], 
                   CircleBox[{-0.4362483268941677, 0.5}, 1.7, {
                    1.8077993180731686`, 1.3440234758621443`}], 
                   ArrowBox[{{-0.07061381332352401, 
                    2.1602142640291824`}, {-0.054030228957557325`, 
                    2.1564749698108088`}}]}, 
                  {RGBColor[0, 0, 1], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-0.4362483268941677, 0.5}, 1.8, {
                    1.8077993180731686`, 1.1126907239033426`}], 
                   ArrowBox[{{0.3436177835658153, 2.1222850704349154`}, {
                    0.35980137091009146`, 2.1144054257297786`}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.003], Arrowheads[0.025], 
                   CircleBox[{-0.4362483268941677, 0.5}, 1.5, {
                    1.8077993180731686`, 1.5993102910406414`}], 
                   ArrowBox[{{-0.49400499213624516`, 
                    1.9988876434276568`}, {-0.47901347770612984`, 
                    1.999390256696378}}]}, 
                  {RGBColor[0, 0, 1], ArrowBox[{{0, 2.8}, {0, 2.8}}]}},
                 Frame->True,
                 FrameTicks->Automatic,
                 ImageSize->800,
                 PlotRange->{{-3, 3}, {-1, 3}}], 149->
                GraphicsBox[{InsetBox[
                   TagBox[GridBox[{
                    {"\<\"t\"\>", "7.4`"},
                    {"\<\"driveIntent\"\>", "0"},
                    {"\<\"motorTorque\"\>", "134.1217121887603`"},
                    {"\<\"vw\"\>", "0.343183855408456`"},
                    {"\<\"controllerState\"\>", 
                    RowBox[{"\[LeftAssociation]", 
                    RowBox[{
                    
                    RowBox[{"\<\"lastError\"\>", "\[Rule]", 
                    "0.15457172237992642`"}], ",", 
                    RowBox[{"\<\"lastInt\"\>", "\[Rule]", 
                    RowBox[{"-", "0.10005122210301508`"}]}], ",", 
                    RowBox[{"\<\"initialized\"\>", "\[Rule]", "True"}]}], 
                    "\[RightAssociation]"}]}
                    },
                    AutoDelete->False,
                    
                    GridBoxItemSize->{
                    "Columns" -> {{Automatic}}, "Rows" -> {{Automatic}}}],
                    "Grid"], {-1, -0.5}], 
                  {GrayLevel[0], LineBox[{{-5, 0}, {5, 0}}]}, 
                  {GrayLevel[0.5], 
                   CircleBox[{-0.41872892736073664`, 0.5}, 0.5], 
                   DiskBox[{-0.41872892736073664`, 0.5}, 0.1]}, 
                  {GrayLevel[0.5], 
                   DiskBox[{-0.8014269174532962, 1.950359351464014}, 0.1]}, 
                  {GrayLevel[0.5], Thickness[0.01], 
                   LineBox[{{-0.41872892736073664`, 
                    0.5}, {-0.8014269174532962, 1.950359351464014}}]}, 
                  {RGBColor[0, 1, 0], 
                   DiskBox[{-0.6738609207557763, 1.4669062343093426`}, 0.05]}, 
                  {RGBColor[0, 0, 1], Arrowheads[0.01], 
                   CircleBox[{-0.41872892736073664`, 0.5}, 0.3333333333333333,
                     NCache[{
                    Rational[-1, 2] Pi, 
                    2.452855038867913}, {-1.5707963267948966`, 
                    2.452855038867913}]], 
                   ArrowBox[{{-0.6739472035474838, 
                    0.7144172162196318}, {-0.6760785791661873, 
                    0.7118543552224603}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-0.41872892736073664`, 
                    0.5}, {-3.1011631711359433`, 0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-0.41872892736073664`, 0.5}, {
                    1.3956634877481922`, 0.5}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.005], Arrowheads[0.04], 
                   ArrowBox[{{-0.41872892736073664`, 
                    0.5}, {-1.286770756027014, 0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-0.41872892736073664`, 0.5}, 1.6, {
                    1.828780538603973, 3.0015884275230977`}], 
                   ArrowBox[{{-2.0007616415874536`, 
                    0.7391076977356583}, {-2.0030735777371933`, 
                    0.7232756789789333}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Dashing[{Small, Small}], 
                   Arrowheads[0.02], 
                   CircleBox[{-0.41872892736073664`, 0.5}, 1.7, {
                    1.828780538603973, 1.3427307436434965`}], 
                   ArrowBox[{{-0.05094850743512064, 
                    2.15974020940608}, {-0.03436977083079995, 
                    2.1559794801843397`}}]}, 
                  {RGBColor[0, 0, 1], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-0.41872892736073664`, 0.5}, 1.8, {
                    1.828780538603973, 1.0836599153330824`}], 
                   ArrowBox[{{0.40789820758678175`, 2.098964533618072}, {
                    0.42384625541788556`, 2.090618452478649}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.003], Arrowheads[0.025], 
                   CircleBox[{-0.41872892736073664`, 0.5}, 1.5, {
                    1.828780538603973, 1.6347744379040972`}], 
                   ArrowBox[{{-0.5295949055885982, 
                    1.995897300910587}, {-0.5146306386417453, 
                    1.9969311479735379`}}]}, 
                  {RGBColor[0, 0, 1], ArrowBox[{{0, 2.8}, {0, 2.8}}]}},
                 Frame->True,
                 FrameTicks->Automatic,
                 ImageSize->800,
                 PlotRange->{{-3, 3}, {-1, 3}}], 150->
                GraphicsBox[{InsetBox[
                   TagBox[GridBox[{
                    {"\<\"t\"\>", "7.45`"},
                    {"\<\"driveIntent\"\>", "0"},
                    {"\<\"motorTorque\"\>", "140.54037726495844`"},
                    {"\<\"vw\"\>", "0.3279012559577667`"},
                    {"\<\"controllerState\"\>", 
                    RowBox[{"\[LeftAssociation]", 
                    RowBox[{
                    
                    RowBox[{"\<\"lastError\"\>", "\[Rule]", 
                    "0.17506469095705496`"}], ",", 
                    RowBox[{"\<\"lastInt\"\>", "\[Rule]", 
                    RowBox[{"-", "0.09178012504373424`"}]}], ",", 
                    RowBox[{"\<\"initialized\"\>", "\[Rule]", "True"}]}], 
                    "\[RightAssociation]"}]}
                    },
                    AutoDelete->False,
                    
                    GridBoxItemSize->{
                    "Columns" -> {{Automatic}}, "Rows" -> {{Automatic}}}],
                    "Grid"], {-1, -0.5}], 
                  {GrayLevel[0], LineBox[{{-5, 0}, {5, 0}}]}, 
                  {GrayLevel[0.5], 
                   CircleBox[{-0.40195700552080965`, 0.5}, 0.5], 
                   DiskBox[{-0.40195700552080965`, 0.5}, 0.1]}, 
                  {GrayLevel[0.5], 
                   DiskBox[{-0.8075400414252321, 1.9441268645747685`}, 0.1]}, 
                  {GrayLevel[0.5], Thickness[0.01], 
                   LineBox[{{-0.40195700552080965`, 
                    0.5}, {-0.8075400414252321, 1.9441268645747685`}}]}, 
                  {RGBColor[0, 1, 0], 
                   DiskBox[{-0.6723456961237579, 1.4627512430498457`}, 0.05]}, 
                  {RGBColor[0, 0, 1], Arrowheads[0.01], 
                   CircleBox[{-0.40195700552080965`, 0.5}, 0.3333333333333333,
                     NCache[{
                    Rational[-1, 2] Pi, 
                    2.6454149911538565`}, {-1.5707963267948966`, 
                    2.6454149911538565`}]], 
                   ArrowBox[{{-0.6934917074622866, 
                    0.6616125882318742}, {-0.6950932297956838, 
                    0.6586892092392603}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-0.40195700552080965`, 
                    0.5}, {-3.2127645508199785`, 0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-0.40195700552080965`, 0.5}, {
                    1.5126716284428865`, 0.5}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.005], Arrowheads[0.04], 
                   ArrowBox[{{-0.40195700552080965`, 
                    0.5}, {-1.2981359168562823`, 0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-0.40195700552080965`, 0.5}, 1.6, {
                    1.8445930643936486`, 3.087533964097431}], 
                   ArrowBox[{{-1.9986753153970922`, 
                    0.6024238200118903}, {-1.9996197012764632`, 
                    0.5864517818831914}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Dashing[{Small, Small}], 
                   Arrowheads[0.02], 
                   CircleBox[{-0.40195700552080965`, 0.5}, 1.7, {
                    1.8445930643936486`, 1.3449939447254726`}], 
                   ArrowBox[{{-0.037933850126027846`, 
                    2.1605683190812797`}, {-0.0213466447013117, 
                    2.156845120473501}}]}, 
                  {RGBColor[0, 0, 1], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-0.40195700552080965`, 0.5}, 1.8, {
                    1.8445930643936486`, 1.0638131906994348`}], 
                   ArrowBox[{{0.45623945919260617`, 2.0822448697826124`}, {
                    0.4720187347186175, 2.073583936583284}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.003], Arrowheads[0.025], 
                   CircleBox[{-0.40195700552080965`, 0.5}, 1.5, {
                    1.8445930643936486`, 1.667248506687361}], 
                   ArrowBox[{{-0.5613338653919269, 
                    1.9915089729993656`}, {-0.5464110554689323, 
                    1.9930281402082095`}}]}, 
                  {RGBColor[0, 0, 1], ArrowBox[{{0, 2.8}, {0, 2.8}}]}},
                 Frame->True,
                 FrameTicks->Automatic,
                 ImageSize->800,
                 PlotRange->{{-3, 3}, {-1, 3}}], 151->
                GraphicsBox[{InsetBox[
                   TagBox[GridBox[{
                    {"\<\"t\"\>", "7.5`"},
                    {"\<\"driveIntent\"\>", "0"},
                    {"\<\"motorTorque\"\>", "144.6796218140634`"},
                    {"\<\"vw\"\>", "0.31217182533632926`"},
                    {"\<\"controllerState\"\>", 
                    RowBox[{"\[LeftAssociation]", 
                    RowBox[{
                    
                    RowBox[{"\<\"lastError\"\>", "\[Rule]", 
                    "0.1909576516470109`"}], ",", 
                    RowBox[{"\<\"lastInt\"\>", "\[Rule]", 
                    RowBox[{"-", "0.08260326433893579`"}]}], ",", 
                    RowBox[{"\<\"initialized\"\>", "\[Rule]", "True"}]}], 
                    "\[RightAssociation]"}]}
                    },
                    AutoDelete->False,
                    
                    GridBoxItemSize->{
                    "Columns" -> {{Automatic}}, "Rows" -> {{Automatic}}}],
                    "Grid"], {-1, -0.5}], 
                  {GrayLevel[0], LineBox[{{-5, 0}, {5, 0}}]}, 
                  {GrayLevel[0.5], CircleBox[{-0.3859617443097894, 0.5}, 0.5],
                    DiskBox[{-0.3859617443097894, 0.5}, 0.1]}, 
                  {GrayLevel[0.5], 
                   DiskBox[{-0.8075292335713329, 1.9395418896293775`}, 0.1]}, 
                  {GrayLevel[0.5], Thickness[0.01], 
                   LineBox[{{-0.3859617443097894, 0.5}, {-0.8075292335713329, 
                    1.9395418896293775`}}]}, 
                  {RGBColor[0, 1, 0], 
                   DiskBox[{-0.6670067371508184, 1.4596945930862515`}, 0.05]}, 
                  {RGBColor[0, 0, 1], Arrowheads[0.01], 
                   CircleBox[{-0.3859617443097894, 0.5}, 0.3333333333333333, 
                    NCache[{
                    Rational[-1, 2] Pi, 
                    2.769592327627006}, {-1.5707963267948966`, 
                    2.769592327627006}]], 
                   ArrowBox[{{-0.6952686830738213, 
                    0.6242591193576335}, {-0.696495788339587, 
                    0.6211598886166985}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-0.3859617443097894, 
                    0.5}, {-3.2795541805910577`, 0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-0.3859617443097894, 0.5}, {1.5978060673431513`,
                     0.5}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.005], Arrowheads[0.04], 
                   ArrowBox[{{-0.3859617443097894, 0.5}, {-1.295786368938117, 
                    0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-0.3859617443097894, 0.5}, 1.6, {
                    1.855679143281533, 3.1476056490836317`}], 
                   ArrowBox[{{-1.9859490273626896`, 
                    0.5063791903090074}, {-1.9859328195050931`, 
                    0.49037926518483504`}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Dashing[{Small, Small}], 
                   Arrowheads[0.02], 
                   CircleBox[{-0.3859617443097894, 0.5}, 1.7, {
                    1.855679143281533, 1.3501201872447028`}], 
                   ArrowBox[{{-0.030455810511816006`, 
                    2.162412563425346}, {-0.013849737093500736`, 
                    2.1587744434025575`}}]}, 
                  {RGBColor[0, 0, 1], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-0.3859617443097894, 0.5}, 1.8, {
                    1.855679143281533, 1.0519034665367362`}], 
                   ArrowBox[{{0.4910175116541912, 2.0719120155431296`}, {
                    0.506692521228205, 2.0630637742001516`}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.003], Arrowheads[0.025], 
                   CircleBox[{-0.3859617443097894, 0.5}, 1.5, {
                    1.855679143281533, 1.6962364538682748`}], 
                   ArrowBox[{{-0.5885013750002535, 
                    1.9862629975881694`}, {-0.5736288658364894, 
                    1.9882140476080323`}}]}, 
                  {RGBColor[0, 0, 1], ArrowBox[{{0, 2.8}, {0, 2.8}}]}},
                 Frame->True,
                 FrameTicks->Automatic,
                 ImageSize->800,
                 PlotRange->{{-3, 3}, {-1, 3}}], 152->
                GraphicsBox[{InsetBox[
                   TagBox[GridBox[{
                    {"\<\"t\"\>", "7.55`"},
                    {"\<\"driveIntent\"\>", "0"},
                    {"\<\"motorTorque\"\>", "146.83457131790502`"},
                    {"\<\"vw\"\>", "0.2962559617443036`"},
                    {"\<\"controllerState\"\>", 
                    RowBox[{"\[LeftAssociation]", 
                    RowBox[{
                    
                    RowBox[{"\<\"lastError\"\>", "\[Rule]", 
                    "0.20264751118001462`"}], ",", 
                    RowBox[{"\<\"lastInt\"\>", "\[Rule]", 
                    RowBox[{"-", "0.07274064594912874`"}]}], ",", 
                    RowBox[{"\<\"initialized\"\>", "\[Rule]", "True"}]}], 
                    "\[RightAssociation]"}]}
                    },
                    AutoDelete->False,
                    
                    GridBoxItemSize->{
                    "Columns" -> {{Automatic}}, "Rows" -> {{Automatic}}}],
                    "Grid"], {-1, -0.5}], 
                  {GrayLevel[0], LineBox[{{-5, 0}, {5, 0}}]}, 
                  {GrayLevel[0.5], CircleBox[{-0.3707587369157254, 0.5}, 0.5],
                    DiskBox[{-0.3707587369157254, 0.5}, 0.1]}, 
                  {GrayLevel[0.5], 
                   DiskBox[{-0.8021559498860873, 1.936626758988362}, 0.1]}, 
                  {GrayLevel[0.5], Thickness[0.01], 
                   LineBox[{{-0.3707587369157254, 0.5}, {-0.8021559498860873, 
                    1.936626758988362}}]}, 
                  {RGBColor[0, 1, 0], 
                   DiskBox[{-0.6583568788959666, 1.4577511726589079`}, 0.05]}, 
                  {RGBColor[0, 0, 1], Arrowheads[0.01], 
                   CircleBox[{-0.3707587369157254, 0.5}, 0.3333333333333333, 
                    NCache[{
                    Rational[-1, 2] Pi, 
                    2.8342408127422543`}, {-1.5707963267948966`, 
                    2.8342408127422543`}]], 
                   ArrowBox[{{-0.6874471072668737, 
                    0.604017244702237}, {-0.6884714280912108, 
                    0.600845212960962}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-0.3707587369157254, 0.5}, {-3.307450163273826, 
                    0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-0.3707587369157254, 0.5}, {1.6551538716400616`,
                     0.5}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.005], Arrowheads[0.04], 
                   ArrowBox[{{-0.3707587369157254, 
                    0.5}, {-1.2815375547180388`, 0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-0.3707587369157254, 0.5}, 1.6, {
                    1.8625144068829513`, 3.1845648689378883`}], 
                   ArrowBox[{{-1.9698890821203714`, 
                    0.4472540139378348}, {-1.9692816751997455`, 
                    0.43126561428350596`}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Dashing[{Small, Small}], 
                   Arrowheads[0.02], 
                   CircleBox[{-0.3707587369157254, 0.5}, 1.7, {
                    1.8625144068829513`, 1.3574737062927722`}], 
                   ArrowBox[{{-0.027486887210147193`, 
                    2.164981812873555}, {-0.010854510026448771`, 
                    2.161465903191285}}]}, 
                  {RGBColor[0, 0, 1], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-0.3707587369157254, 0.5}, 1.8, {
                    1.8625144068829513`, 1.0467667884501457`}], 
                   ArrowBox[{{0.5142833198188945, 2.0673865374600338`}, {
                    0.5299126722296412, 2.0584578957251622`}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.003], Arrowheads[0.025], 
                   CircleBox[{-0.3707587369157254, 0.5}, 1.5, {
                    1.8625144068829513`, 1.7214249636199812`}], 
                   ArrowBox[{{-0.6106669106933341, 
                    1.9806904025334577`}, {-0.5958482581397719, 
                    1.9830154103835635`}}]}, 
                  {RGBColor[0, 0, 1], ArrowBox[{{0, 2.8}, {0, 2.8}}]}},
                 Frame->True,
                 FrameTicks->Automatic,
                 ImageSize->800,
                 PlotRange->{{-3, 3}, {-1, 3}}], 153->
                GraphicsBox[{InsetBox[
                   TagBox[GridBox[{
                    {"\<\"t\"\>", "7.6000000000000005`"},
                    {"\<\"driveIntent\"\>", "0"},
                    {"\<\"motorTorque\"\>", "147.2843168895607`"},
                    {"\<\"vw\"\>", "0.2803846104996343`"},
                    {"\<\"controllerState\"\>", 
                    RowBox[{"\[LeftAssociation]", 
                    RowBox[{
                    
                    RowBox[{"\<\"lastError\"\>", "\[Rule]", 
                    "0.21055266222947672`"}], ",", 
                    RowBox[{"\<\"lastInt\"\>", "\[Rule]", 
                    RowBox[{"-", "0.06239180947199278`"}]}], ",", 
                    RowBox[{"\<\"initialized\"\>", "\[Rule]", "True"}]}], 
                    "\[RightAssociation]"}]}
                    },
                    AutoDelete->False,
                    
                    GridBoxItemSize->{
                    "Columns" -> {{Automatic}}, "Rows" -> {{Automatic}}}],
                    "Grid"], {-1, -0.5}], 
                  {GrayLevel[0], LineBox[{{-5, 0}, {5, 0}}]}, 
                  {GrayLevel[0.5], 
                   CircleBox[{-0.35635128040487946`, 0.5}, 0.5], 
                   DiskBox[{-0.35635128040487946`, 0.5}, 0.1]}, 
                  {GrayLevel[0.5], 
                   DiskBox[{-0.7921603126090586, 1.9352945646971764`}, 0.1]}, 
                  {GrayLevel[0.5], Thickness[0.01], 
                   LineBox[{{-0.35635128040487946`, 
                    0.5}, {-0.7921603126090586, 1.9352945646971764`}}]}, 
                  {RGBColor[0, 1, 0], 
                   DiskBox[{-0.6468906352076655, 1.4568630431314509`}, 0.05]}, 
                  {RGBColor[0, 0, 1], Arrowheads[0.01], 
                   CircleBox[{-0.35635128040487946`, 0.5}, 0.3333333333333333,
                     NCache[{
                    Rational[-1, 2] Pi, 
                    2.8477331798919243`}, {-1.5707963267948966`, 
                    2.8477331798919243`}]], 
                   ArrowBox[{{-0.6744142218635606, 
                    0.5997350308656024}, {-0.6753956525352474, 
                    0.5965494677512537}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-0.35635128040487946`, 
                    0.5}, {-3.3020376181960933`, 0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-0.35635128040487946`, 0.5}, {
                    1.6883821045888214`, 0.5}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.005], Arrowheads[0.04], 
                   ArrowBox[{{-0.35635128040487946`, 
                    0.5}, {-1.2573042332023925`, 0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-0.35635128040487946`, 0.5}, 1.6, {
                    1.8655867861483433`, 3.2011576129186916`}], 
                   ArrowBox[{{-1.954386334570788, 
                    0.42072853188591913`}, {-1.9535137320146299`, 
                    0.4047524112224804}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Dashing[{Small, Small}], 
                   Arrowheads[0.02], 
                   CircleBox[{-0.35635128040487946`, 0.5}, 1.7, {
                    1.8655867861483433`, 1.3664686488912956`}], 
                   ArrowBox[{{-0.028069531376149037`, 
                    2.1680021262740157`}, {-0.011406202063041049`, 
                    2.1646359640857646`}}]}, 
                  {RGBColor[0, 0, 1], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-0.35635128040487946`, 0.5}, 1.8, {
                    1.8655867861483433`, 1.0473405812063394`}], 
                   ArrowBox[{{0.5277912756429501, 2.0678941101315518`}, {
                    0.5434257486704763, 2.05897443787521}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.003], Arrowheads[0.025], 
                   CircleBox[{-0.35635128040487946`, 0.5}, 1.5, {
                    1.8655867861483433`, 1.742632999881656}], 
                   ArrowBox[{{-0.6276056849185288, 
                    1.9752698221111775`}, {-0.6128396699672878, 
                    1.977908558071061}}]}, 
                  {RGBColor[0, 0, 1], ArrowBox[{{0, 2.8}, {0, 2.8}}]}},
                 Frame->True,
                 FrameTicks->Automatic,
                 ImageSize->800,
                 PlotRange->{{-3, 3}, {-1, 3}}], 154->
                GraphicsBox[{InsetBox[
                   TagBox[GridBox[{
                    {"\<\"t\"\>", "7.65`"},
                    {"\<\"driveIntent\"\>", "0"},
                    {"\<\"motorTorque\"\>", "146.28552038354766`"},
                    {"\<\"vw\"\>", "0.26475434387850394`"},
                    {"\<\"controllerState\"\>", 
                    RowBox[{"\[LeftAssociation]", 
                    RowBox[{
                    
                    RowBox[{"\<\"lastError\"\>", "\[Rule]", 
                    "0.21509662257282106`"}], ",", 
                    RowBox[{"\<\"lastInt\"\>", "\[Rule]", 
                    RowBox[{"-", "0.051735181277262066`"}]}], ",", 
                    RowBox[{"\<\"initialized\"\>", "\[Rule]", "True"}]}], 
                    "\[RightAssociation]"}]}
                    },
                    AutoDelete->False,
                    
                    GridBoxItemSize->{
                    "Columns" -> {{Automatic}}, "Rows" -> {{Automatic}}}],
                    "Grid"], {-1, -0.5}], 
                  {GrayLevel[0], LineBox[{{-5, 0}, {5, 0}}]}, 
                  {GrayLevel[0.5], 
                   CircleBox[{-0.34273198763299245`, 0.5}, 0.5], 
                   DiskBox[{-0.34273198763299245`, 0.5}, 0.1]}, 
                  {GrayLevel[0.5], 
                   DiskBox[{-0.7782423798093577, 1.935385209031495}, 0.1]}, 
                  {GrayLevel[0.5], Thickness[0.01], 
                   LineBox[{{-0.34273198763299245`, 
                    0.5}, {-0.7782423798093577, 1.935385209031495}}]}, 
                  {RGBColor[0, 1, 0], 
                   DiskBox[{-0.6330722490839027, 1.4569234726876634`}, 0.05]}, 
                  {RGBColor[0, 0, 1], Arrowheads[0.01], 
                   CircleBox[{-0.34273198763299245`, 0.5}, 0.3333333333333333,
                     NCache[{
                    Rational[-1, 2] Pi, 
                    2.8177692847115337`}, {-1.5707963267948966`, 
                    2.8177692847115337`}]], 
                   ArrowBox[{{-0.6576641529100176, 
                    0.6092192399947722}, {-0.658740580629807, 
                    0.606064509913942}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-0.34273198763299245`, 
                    0.5}, {-3.268442395303946, 0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-0.34273198763299245`, 0.5}, {1.700729279183181,
                     0.5}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.005], Arrowheads[0.04], 
                   ArrowBox[{{-0.34273198763299245`, 
                    0.5}, {-1.2249811284877723`, 0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-0.34273198763299245`, 0.5}, 1.6, {
                    1.865378723912026, 3.200034344985886}], 
                   ArrowBox[{{-1.9408550767333386`, 
                    0.42252360304991354`}, {-1.9400004201879333`, 
                    0.40654651229899263`}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Dashing[{Small, Small}], 
                   Arrowheads[0.02], 
                   CircleBox[{-0.34273198763299245`, 0.5}, 1.7, {
                    1.865378723912026, 1.3765907148970307`}], 
                   ArrowBox[{{-0.03135039503014758, 
                    2.171239511197637}, {-0.014653847406585263`, 
                    2.1680421858890684`}}]}, 
                  {RGBColor[0, 0, 1], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-0.34273198763299245`, 0.5}, 1.8, {
                    1.865378723912026, 1.0526813884478723`}], 
                   ArrowBox[{{0.5330241782975875, 2.072593761222068}, {
                    0.5487060663687593, 2.0637577164885776`}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.003], Arrowheads[0.025], 
                   CircleBox[{-0.34273198763299245`, 0.5}, 1.5, {
                    1.865378723912026, 1.7598148402834977`}], 
                   ArrowBox[{{-0.6392929573316435, 
                    1.9703916455323718`}, {-0.6245744580154504, 
                    1.9732836868331616`}}]}, 
                  {RGBColor[0, 0, 1], ArrowBox[{{0, 2.8}, {0, 2.8}}]}},
                 Frame->True,
                 FrameTicks->Automatic,
                 ImageSize->800,
                 PlotRange->{{-3, 3}, {-1, 3}}], 155->
                GraphicsBox[{InsetBox[
                   TagBox[GridBox[{
                    {"\<\"t\"\>", "7.7`"},
                    {"\<\"driveIntent\"\>", "0"},
                    {"\<\"motorTorque\"\>", "144.0688239713291`"},
                    {"\<\"vw\"\>", "0.24952576809887417`"},
                    {"\<\"controllerState\"\>", 
                    RowBox[{"\[LeftAssociation]", 
                    RowBox[{
                    
                    RowBox[{"\<\"lastError\"\>", "\[Rule]", 
                    "0.21669439675755361`"}], ",", 
                    RowBox[{"\<\"lastInt\"\>", "\[Rule]", 
                    RowBox[{"-", "0.04092817851793974`"}]}], ",", 
                    RowBox[{"\<\"initialized\"\>", "\[Rule]", "True"}]}], 
                    "\[RightAssociation]"}]}
                    },
                    AutoDelete->False,
                    
                    GridBoxItemSize->{
                    "Columns" -> {{Automatic}}, "Rows" -> {{Automatic}}}],
                    "Grid"], {-1, -0.5}], 
                  {GrayLevel[0], LineBox[{{-5, 0}, {5, 0}}]}, 
                  {GrayLevel[0.5], CircleBox[{-0.3298845598357399, 0.5}, 0.5],
                    DiskBox[{-0.3298845598357399, 0.5}, 0.1]}, 
                  {GrayLevel[0.5], 
                   DiskBox[{-0.761050056171398, 1.9366963196060696`}, 0.1]}, 
                  {GrayLevel[0.5], Thickness[0.01], 
                   LineBox[{{-0.3298845598357399, 0.5}, {-0.761050056171398, 
                    1.9366963196060696`}}]}, 
                  {RGBColor[0, 1, 0], 
                   DiskBox[{-0.6173282240595119, 1.4577975464040462`}, 0.05]}, 
                  {RGBColor[0, 0, 1], Arrowheads[0.01], 
                   CircleBox[{-0.3298845598357399, 0.5}, 0.3333333333333333, 
                    NCache[{
                    Rational[-1, 2] Pi, 
                    2.7512683923449766`}, {-1.5707963267948966`, 
                    2.7512683923449766`}]], 
                   ArrowBox[{{-0.6368627839377481, 
                    0.6299056620717064}, {-0.6381464701243317, 
                    0.6268294357644915}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-0.3298845598357399, 
                    0.5}, {-3.2112610392623218`, 0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-0.3298845598357399, 0.5}, {1.695037909908864, 
                    0.5}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.005], Arrowheads[0.04], 
                   ArrowBox[{{-0.3298845598357399, 0.5}, {-1.186338569517718, 
                    0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-0.3298845598357399, 0.5}, 1.6, {
                    1.8623531186269056`, 3.1836934669519312`}], 
                   ArrowBox[{{-1.9290602608486587`, 
                    0.4486475191461999}, {-1.928466786480095, 
                    0.43265859626666775`}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Dashing[{Small, Small}], 
                   Arrowheads[0.02], 
                   CircleBox[{-0.3298845598357399, 0.5}, 1.7, {
                    1.8623531186269056`, 1.3874128238633299`}], 
                   ArrowBox[{{-0.036607184317880515`, 
                    2.174511385750887}, {-0.019877013290784507`, 
                    2.17149493600345}}]}, 
                  {RGBColor[0, 0, 1], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-0.3298845598357399, 0.5}, 1.8, {
                    1.8623531186269056`, 1.0619707632306328`}], 
                   ArrowBox[{{0.5312256180485946, 2.0806610204417675`}, {
                    0.5469889096607266, 2.071971029788154}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.003], Arrowheads[0.025], 
                   CircleBox[{-0.3298845598357399, 0.5}, 1.5, {
                    1.8623531186269056`, 1.7730583296326863`}], 
                   ArrowBox[{{-0.645892069868824, 
                    1.9663353141770439`}, {-0.6312131608712184, 
                    1.969422020454982}}]}, 
                  {RGBColor[0, 0, 1], ArrowBox[{{0, 2.8}, {0, 2.8}}]}},
                 Frame->True,
                 FrameTicks->Automatic,
                 ImageSize->800,
                 PlotRange->{{-3, 3}, {-1, 3}}], 156->
                GraphicsBox[{InsetBox[
                   TagBox[GridBox[{
                    {"\<\"t\"\>", "7.75`"},
                    {"\<\"driveIntent\"\>", "0"},
                    {"\<\"motorTorque\"\>", "140.83801530692696`"},
                    {"\<\"vw\"\>", "0.2348247508777999`"},
                    {"\<\"controllerState\"\>", 
                    RowBox[{"\[LeftAssociation]", 
                    RowBox[{
                    
                    RowBox[{"\<\"lastError\"\>", "\[Rule]", 
                    "0.2157420244471336`"}], ",", 
                    RowBox[{"\<\"lastInt\"\>", "\[Rule]", 
                    RowBox[{"-", "0.030107915136035772`"}]}], ",", 
                    RowBox[{"\<\"initialized\"\>", "\[Rule]", "True"}]}], 
                    "\[RightAssociation]"}]}
                    },
                    AutoDelete->False,
                    
                    GridBoxItemSize->{
                    "Columns" -> {{Automatic}}, "Rows" -> {{Automatic}}}],
                    "Grid"], {-1, -0.5}], 
                  {GrayLevel[0], LineBox[{{-5, 0}, {5, 0}}]}, 
                  {GrayLevel[0.5], 
                   CircleBox[{-0.31778556361182153`, 0.5}, 0.5], 
                   DiskBox[{-0.31778556361182153`, 0.5}, 0.1]}, 
                  {GrayLevel[0.5], 
                   DiskBox[{-0.7411725325147686, 1.93900780906956}, 0.1]}, 
                  {GrayLevel[0.5], Thickness[0.01], 
                   LineBox[{{-0.31778556361182153`, 
                    0.5}, {-0.7411725325147686, 1.93900780906956}}]}, 
                  {RGBColor[0, 1, 0], 
                   DiskBox[{-0.6000435428804529, 1.4593385393797065`}, 0.05]}, 
                  {RGBColor[0, 0, 1], Arrowheads[0.01], 
                   CircleBox[{-0.31778556361182153`, 0.5}, 0.3333333333333333,
                     NCache[{
                    Rational[-1, 2] Pi, 
                    2.6543441324129127`}, {-1.5707963267948966`, 
                    2.6543441324129127`}]], 
                   ArrowBox[{{-0.6107516861170941, 
                    0.6590030256798176}, {-0.612327041689464, 
                    0.656065463197175}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-0.31778556361182153`, 
                    0.5}, {-3.1345458697503608`, 0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-0.31778556361182153`, 0.5}, {
                    1.6738049944788047`, 0.5}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.005], Arrowheads[0.04], 
                   ArrowBox[{{-0.31778556361182153`, 
                    0.5}, {-1.1429553116597344`, 0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-0.31778556361182153`, 0.5}, 1.6, {
                    1.856943307245709, 3.1544457505557357`}], 
                   ArrowBox[{{-1.9177790514864013`, 
                    0.4954350510477351}, {-1.9176534037499677`, 
                    0.4794356110787832}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Dashing[{Small, Small}], 
                   Arrowheads[0.02], 
                   CircleBox[{-0.31778556361182153`, 0.5}, 1.7, {
                    1.856943307245709, 1.398598254592663}], 
                   ArrowBox[{{-0.04325627498175533, 
                    2.1776869999151423`}, {-0.026493410945750473`, 
                    2.1748578691325324`}}]}, 
                  {RGBColor[0, 0, 1], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-0.31778556361182153`, 0.5}, 1.8, {
                    1.856943307245709, 1.0745098888738926`}], 
                   ArrowBox[{{0.5234373313701435, 2.0913340444288124`}, {
                    0.5393083457991762, 2.08284238964294}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.003], Arrowheads[0.025], 
                   CircleBox[{-0.31778556361182153`, 0.5}, 1.5, {
                    1.856943307245709, 1.7825680850567722`}], 
                   ArrowBox[{{-0.64772306464129, 
                    1.9632639014936537`}, {-0.633074172764873, 
                    1.9664900589292567`}}]}, 
                  {RGBColor[0, 0, 1], ArrowBox[{{0, 2.8}, {0, 2.8}}]}},
                 Frame->True,
                 FrameTicks->Automatic,
                 ImageSize->800,
                 PlotRange->{{-3, 3}, {-1, 3}}], 157->
                GraphicsBox[{InsetBox[
                   TagBox[GridBox[{
                    {"\<\"t\"\>", "7.8`"},
                    {"\<\"driveIntent\"\>", "0"},
                    {"\<\"motorTorque\"\>", "136.7712172551961`"},
                    {"\<\"vw\"\>", "0.22074557027095618`"},
                    {"\<\"controllerState\"\>", 
                    RowBox[{"\[LeftAssociation]", 
                    RowBox[{
                    
                    RowBox[{"\<\"lastError\"\>", "\[Rule]", 
                    "0.21260942391648277`"}], ",", 
                    RowBox[{"\<\"lastInt\"\>", "\[Rule]", 
                    RowBox[{"-", "0.01939234599537714`"}]}], ",", 
                    RowBox[{"\<\"initialized\"\>", "\[Rule]", "True"}]}], 
                    "\[RightAssociation]"}]}
                    },
                    AutoDelete->False,
                    
                    GridBoxItemSize->{
                    "Columns" -> {{Automatic}}, "Rows" -> {{Automatic}}}],
                    "Grid"], {-1, -0.5}], 
                  {GrayLevel[0], LineBox[{{-5, 0}, {5, 0}}]}, 
                  {GrayLevel[0.5], CircleBox[{-0.3064060935883109, 0.5}, 0.5],
                    DiskBox[{-0.3064060935883109, 0.5}, 0.1]}, 
                  {GrayLevel[0.5], 
                   DiskBox[{-0.719138116006099, 1.9420999541193122`}, 0.1]}, 
                  {GrayLevel[0.5], Thickness[0.01], 
                   LineBox[{{-0.3064060935883109, 0.5}, {-0.719138116006099, 
                    1.9420999541193122`}}]}, 
                  {RGBColor[0, 1, 0], 
                   DiskBox[{-0.5815607752001697, 1.461399969412875}, 0.05]}, 
                  {RGBColor[0, 0, 1], Arrowheads[0.01], 
                   CircleBox[{-0.3064060935883109, 0.5}, 0.3333333333333333, 
                    NCache[{
                    Rational[-1, 2] Pi, 
                    2.532340190860986}, {-1.5707963267948966`, 
                    2.532340190860986}]], 
                   ArrowBox[{{-0.5778436184742957, 
                    0.6934755312562327}, {-0.5797647697779515, 
                    0.690751527550786}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-0.3064060935883109, 0.5}, {-3.041830438692233, 
                    0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-0.3064060935883109, 0.5}, {1.6392359815208952`,
                     0.5}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.005], Arrowheads[0.04], 
                   ArrowBox[{{-0.3064060935883109, 
                    0.5}, {-1.0961883635830267`, 0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-0.3064060935883109, 0.5}, 1.6, {
                    1.8495469175329684`, 3.1143964449215313`}], 
                   ArrowBox[{{-1.905299374846496, 
                    0.5595002113225994}, {-1.9058144230452103`, 
                    0.5435085700051246}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Dashing[{Small, Small}], 
                   Arrowheads[0.02], 
                   CircleBox[{-0.3064060935883109, 0.5}, 1.7, {
                    1.8495469175329684`, 1.4098936414092091`}], 
                   ArrowBox[{{-0.05084403844917351, 
                    2.1806808251339875`}, {-0.03405028830617707, 
                    2.1780412138350833`}}]}, 
                  {RGBColor[0, 0, 1], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-0.3064060935883109, 0.5}, 1.8, {
                    1.8495469175329684`, 1.089706821670768}], 
                   ArrowBox[{{0.5105371987143008, 2.103933806976388}, {
                    0.526535422638877, 2.095684314187851}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.003], Arrowheads[0.025], 
                   CircleBox[{-0.3064060935883109, 0.5}, 1.5, {
                    1.8495469175329684`, 1.7886379187725345`}], 
                   ArrowBox[{{-0.6452192307384608, 
                    1.9612342926767337`}, {-0.6305901908338387, 
                    1.9645493064738728`}}]}, 
                  {RGBColor[0, 0, 1], ArrowBox[{{0, 2.8}, {0, 2.8}}]}},
                 Frame->True,
                 FrameTicks->Automatic,
                 ImageSize->800,
                 PlotRange->{{-3, 3}, {-1, 3}}], 158->
                GraphicsBox[{InsetBox[
                   TagBox[GridBox[{
                    {"\<\"t\"\>", "7.8500000000000005`"},
                    {"\<\"driveIntent\"\>", "0"},
                    {"\<\"motorTorque\"\>", "132.02327779528187`"},
                    {"\<\"vw\"\>", "0.20735508687806253`"},
                    {"\<\"controllerState\"\>", 
                    RowBox[{"\[LeftAssociation]", 
                    RowBox[{
                    
                    RowBox[{"\<\"lastError\"\>", "\[Rule]", 
                    "0.2076362523863854`"}], ",", 
                    RowBox[{"\<\"lastInt\"\>", "\[Rule]", 
                    RowBox[{"-", "0.008881690105893016`"}]}], ",", 
                    RowBox[{"\<\"initialized\"\>", "\[Rule]", "True"}]}], 
                    "\[RightAssociation]"}]}
                    },
                    AutoDelete->False,
                    
                    GridBoxItemSize->{
                    "Columns" -> {{Automatic}}, "Rows" -> {{Automatic}}}],
                    "Grid"], {-1, -0.5}], 
                  {GrayLevel[0], LineBox[{{-5, 0}, {5, 0}}]}, 
                  {GrayLevel[0.5], 
                   CircleBox[{-0.29571324811144906`, 0.5}, 0.5], 
                   DiskBox[{-0.29571324811144906`, 0.5}, 0.1]}, 
                  {GrayLevel[0.5], 
                   DiskBox[{-0.6954153030115835, 1.9457656336033893`}, 0.1]}, 
                  {GrayLevel[0.5], Thickness[0.01], 
                   LineBox[{{-0.29571324811144906`, 
                    0.5}, {-0.6954153030115835, 1.9457656336033893`}}]}, 
                  {RGBColor[0, 1, 0], 
                   DiskBox[{-0.5621812847115386, 1.463843755735593}, 0.05]}, 
                  {RGBColor[0, 0, 1], Arrowheads[0.01], 
                   CircleBox[{-0.29571324811144906`, 0.5}, 0.3333333333333333,
                     NCache[{
                    Rational[-1, 2] Pi, 
                    2.3899020070635597`}, {-1.5707963267948966`, 
                    2.3899020070635597`}]], 
                   ArrowBox[{{-0.5369366656960444, 
                    0.7300486338145025}, {-0.5392250526225723, 
                    0.727624937506521}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-0.29571324811144906`, 
                    0.5}, {-2.9361788040170866`, 0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-0.29571324811144906`, 0.5}, {
                    1.5932943251089173`, 0.5}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.005], Arrowheads[0.04], 
                   ArrowBox[{{-0.29571324811144906`, 0.5}, {-1.04717123079672,
                     0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-0.29571324811144906`, 0.5}, 1.6, {
                    1.8405230355675084`, 3.0654412095320094`}], 
                   ArrowBox[{{-1.889779262628569, 
                    0.6376718611826884}, {-1.8910762556586664`, 
                    0.6217245831781621}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Dashing[{Small, Small}], 
                   Arrowheads[0.02], 
                   CircleBox[{-0.29571324811144906`, 0.5}, 1.7, {
                    1.8405230355675084`, 1.4211161947044102`}], 
                   ArrowBox[{{-0.05902842048802076, 
                    2.183442987562355}, {-0.042206105327588395`, 
                    2.180992007285449}}]}, 
                  {RGBColor[0, 0, 1], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-0.29571324811144906`, 0.5}, 1.8, {
                    1.8405230355675084`, 1.1070603811492759`}], 
                   ArrowBox[{{0.4932744739393693, 2.1178684663634}, {
                    0.5094134399022484, 2.10989782789098}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.003], Arrowheads[0.025], 
                   CircleBox[{-0.29571324811144906`, 0.5}, 1.5, {
                    1.8405230355675084`, 1.7916194828495857`}], 
                   ArrowBox[{{-0.6388816365051495, 
                    1.9602176061146743`}, {-0.6242625455359535, 
                    1.9635762225322821`}}]}, 
                  {RGBColor[0, 0, 1], ArrowBox[{{0, 2.8}, {0, 2.8}}]}},
                 Frame->True,
                 FrameTicks->Automatic,
                 ImageSize->800,
                 PlotRange->{{-3, 3}, {-1, 3}}], 159->
                GraphicsBox[{InsetBox[
                   TagBox[GridBox[{
                    {"\<\"t\"\>", "7.9`"},
                    {"\<\"driveIntent\"\>", "0"},
                    {"\<\"motorTorque\"\>", "126.72867587636743`"},
                    {"\<\"vw\"\>", "0.1946972242068678`"},
                    {"\<\"controllerState\"\>", 
                    RowBox[{"\[LeftAssociation]", 
                    RowBox[{
                    
                    RowBox[{"\<\"lastError\"\>", "\[Rule]", 
                    "0.20113025484494182`"}], ",", 
                    
                    RowBox[{"\<\"lastInt\"\>", "\[Rule]", 
                    "0.001340005133408011`"}], ",", 
                    RowBox[{"\<\"initialized\"\>", "\[Rule]", "True"}]}], 
                    "\[RightAssociation]"}]}
                    },
                    AutoDelete->False,
                    
                    GridBoxItemSize->{
                    "Columns" -> {{Automatic}}, "Rows" -> {{Automatic}}}],
                    "Grid"], {-1, -0.5}], 
                  {GrayLevel[0], LineBox[{{-5, 0}, {5, 0}}]}, 
                  {GrayLevel[0.5], CircleBox[{-0.2856713859715741, 0.5}, 0.5],
                    DiskBox[{-0.2856713859715741, 0.5}, 0.1]}, 
                  {GrayLevel[0.5], 
                   DiskBox[{-0.6704160365588581, 1.9498177657362559`}, 0.1]}, 
                  {GrayLevel[0.5], Thickness[0.01], 
                   LineBox[{{-0.2856713859715741, 0.5}, {-0.6704160365588581, 
                    1.9498177657362559`}}]}, 
                  {RGBColor[0, 1, 0], 
                   DiskBox[{-0.5421678196964301, 1.4665451771575038`}, 0.05]}, 
                  {RGBColor[0, 0, 1], Arrowheads[0.01], 
                   CircleBox[{-0.2856713859715741, 0.5}, 0.3333333333333333, 
                    NCache[{
                    Rational[-1, 2] Pi, 
                    2.231063949496126}, {-1.5707963267948966`, 
                    2.231063949496126}]], 
                   ArrowBox[{{-0.4874711973574961, 
                    0.7653072694739392}, {-0.490114135928092, 
                    0.7632760397402844}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-0.2856713859715741, 
                    0.5}, {-2.8202449034989225`, 0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-0.2856713859715741, 0.5}, {1.5377432157061257`,
                     0.5}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.005], Arrowheads[0.04], 
                   ArrowBox[{{-0.2856713859715741, 0.5}, {-0.996830301821223, 
                    0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-0.2856713859715741, 0.5}, 1.6, {
                    1.830191942164694, 3.0092719820652527`}], 
                   ArrowBox[{{-1.8694945801470868`, 
                    0.7269451246263638}, {-1.8716848030695676`, 
                    0.7110958094921505}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Dashing[{Small, Small}], 
                   Arrowheads[0.02], 
                   CircleBox[{-0.2856713859715741, 0.5}, 1.7, {
                    1.830191942164694, 1.4321389850304227`}], 
                   ArrowBox[{{-0.06755680034170577, 
                    2.1859495922285195`}, {-0.05070849104801503, 
                    2.1836841859473353`}}]}, 
                  {RGBColor[0, 0, 1], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-0.2856713859715741, 0.5}, 1.8, {
                    1.830191942164694, 1.1261437428515415`}], 
                   ArrowBox[{{0.4723001805376063, 2.132629506153683}, {
                    0.488588305233081, 2.124968286021506}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.003], Arrowheads[0.025], 
                   CircleBox[{-0.2856713859715741, 0.5}, 1.5, {
                    1.830191942164694, 1.7918927941306664`}], 
                   ArrowBox[{{-0.6292388554877122, 
                    1.9601237597855456`}, {-0.6146206830122769, 
                    1.9634863716401418`}}]}, 
                  {RGBColor[0, 0, 1], ArrowBox[{{0, 2.8}, {0, 2.8}}]}},
                 Frame->True,
                 FrameTicks->Automatic,
                 ImageSize->800,
                 PlotRange->{{-3, 3}, {-1, 3}}], 160->
                GraphicsBox[{InsetBox[
                   TagBox[GridBox[{
                    {"\<\"t\"\>", "7.95`"},
                    {"\<\"driveIntent\"\>", "0"},
                    {"\<\"motorTorque\"\>", "121.00446644848628`"},
                    {"\<\"vw\"\>", "0.1827972739412368`"},
                    {"\<\"controllerState\"\>", 
                    RowBox[{"\[LeftAssociation]", 
                    RowBox[{
                    
                    RowBox[{"\<\"lastError\"\>", "\[Rule]", 
                    "0.1933674769491578`"}], ",", 
                    
                    RowBox[{"\<\"lastInt\"\>", "\[Rule]", 
                    "0.011203267294239105`"}], ",", 
                    RowBox[{"\<\"initialized\"\>", "\[Rule]", "True"}]}], 
                    "\[RightAssociation]"}]}
                    },
                    AutoDelete->False,
                    
                    GridBoxItemSize->{
                    "Columns" -> {{Automatic}}, "Rows" -> {{Automatic}}}],
                    "Grid"], {-1, -0.5}], 
                  {GrayLevel[0], LineBox[{{-5, 0}, {5, 0}}]}, 
                  {GrayLevel[0.5], 
                   CircleBox[{-0.27624316203111965`, 0.5}, 0.5], 
                   DiskBox[{-0.27624316203111965`, 0.5}, 0.1]}, 
                  {GrayLevel[0.5], 
                   DiskBox[{-0.6445002548488301, 1.9540930897261868`}, 0.1]}, 
                  {GrayLevel[0.5], Thickness[0.01], 
                   LineBox[{{-0.27624316203111965`, 
                    0.5}, {-0.6445002548488301, 1.9540930897261868`}}]}, 
                  {RGBColor[0, 1, 0], 
                   DiskBox[{-0.52174789057626, 1.469395393150791}, 0.05]}, 
                  {RGBColor[0, 0, 1], Arrowheads[0.01], 
                   CircleBox[{-0.27624316203111965`, 0.5}, 0.3333333333333333,
                     NCache[{
                    Rational[-1, 2] Pi, 
                    2.059337666659692}, {-1.5707963267948966`, 
                    2.059337666659692}]], 
                   ArrowBox[{{-0.4297381155661417, 
                    0.7958891859301257}, {-0.43268928342710444`, 
                    0.7943394676411302}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-0.27624316203111965`, 
                    0.5}, {-2.6963324910008453`, 0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-0.27624316203111965`, 0.5}, {
                    1.4741787922794116`, 0.5}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.005], Arrowheads[0.04], 
                   ArrowBox[{{-0.27624316203111965`, 
                    0.5}, {-0.945910536690314, 0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-0.27624316203111965`, 0.5}, 1.6, {
                    1.8188366482680773`, 2.9473892792933274`}], 
                   ArrowBox[{{-1.8429997071065913`, 
                    0.8244594403976737}, {-1.8461659102598245`, 
                    0.8087759132348755}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Dashing[{Small, Small}], 
                   Arrowheads[0.02], 
                   CircleBox[{-0.27624316203111965`, 0.5}, 1.7, {
                    1.8188366482680773`, 1.4428768822046143`}], 
                   ArrowBox[{{-0.07624435631884965, 
                    2.188194443099984}, {-0.05937269310913654, 
                    2.186110079357086}}]}, 
                  {RGBColor[0, 0, 1], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-0.27624316203111965`, 0.5}, 1.8, {
                    1.8188366482680773`, 1.1465896124431536`}], 
                   ArrowBox[{{0.448191776886261, 2.1477845791473382`}, {
                    0.4646331266032455, 2.1404579619543345`}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.003], Arrowheads[0.025], 
                   CircleBox[{-0.27624316203111965`, 0.5}, 1.5, {
                    1.8188366482680773`, 1.7898419434998443`}], 
                   ArrowBox[{{-0.6168154153924315, 
                    1.9608252942225495`}, {-0.6021903774491075, 
                    1.9641579193383683`}}]}, 
                  {RGBColor[0, 0, 1], ArrowBox[{{0, 2.8}, {0, 2.8}}]}},
                 Frame->True,
                 FrameTicks->Automatic,
                 ImageSize->800,
                 PlotRange->{{-3, 3}, {-1, 3}}], 161->
                GraphicsBox[{InsetBox[
                   TagBox[GridBox[{
                    {"\<\"t\"\>", "8.`"},
                    {"\<\"driveIntent\"\>", "0"},
                    {"\<\"motorTorque\"\>", "114.95298213779024`"},
                    {"\<\"vw\"\>", "0.17166575171685258`"},
                    {"\<\"controllerState\"\>", 
                    RowBox[{"\[LeftAssociation]", 
                    RowBox[{
                    
                    RowBox[{"\<\"lastError\"\>", "\[Rule]", 
                    "0.1845937384806671`"}], ",", 
                    
                    RowBox[{"\<\"lastInt\"\>", "\[Rule]", 
                    "0.020651647430434204`"}], ",", 
                    RowBox[{"\<\"initialized\"\>", "\[Rule]", "True"}]}], 
                    "\[RightAssociation]"}]}
                    },
                    AutoDelete->False,
                    
                    GridBoxItemSize->{
                    "Columns" -> {{Automatic}}, "Rows" -> {{Automatic}}}],
                    "Grid"], {-1, -0.5}], 
                  {GrayLevel[0], LineBox[{{-5, 0}, {5, 0}}]}, 
                  {GrayLevel[0.5], 
                   CircleBox[{-0.26739035827867424`, 0.5}, 0.5], 
                   DiskBox[{-0.26739035827867424`, 0.5}, 0.1]}, 
                  {GrayLevel[0.5], 
                   DiskBox[{-0.6179810332319146, 1.9584533515460245`}, 0.1]}, 
                  {GrayLevel[0.5], Thickness[0.01], 
                   LineBox[{{-0.26739035827867424`, 
                    0.5}, {-0.6179810332319146, 1.9584533515460245`}}]}, 
                  {RGBColor[0, 1, 0], 
                   DiskBox[{-0.5011174749141678, 1.4723022343640162`}, 0.05]}, 
                  {RGBColor[0, 0, 1], Arrowheads[0.01], 
                   CircleBox[{-0.26739035827867424`, 0.5}, 0.3333333333333333,
                     NCache[{
                    Rational[-1, 2] Pi, 
                    1.8777931373388106`}, {-1.5707963267948966`, 
                    1.8777931373388106`}]], 
                   ArrowBox[{{-0.3649402988899556, 
                    0.8187398942677973}, {-0.3681227672531986, 
                    0.817748474258021}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-0.26739035827867424`, 
                    0.5}, {-2.566450001034479, 0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-0.26739035827867424`, 0.5}, {1.404055541403233,
                     0.5}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.005], Arrowheads[0.04], 
                   ArrowBox[{{-0.26739035827867424`, 
                    0.5}, {-0.8950041013525718, 0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-0.26739035827867424`, 0.5}, 1.6, {
                    1.806705544736802, 2.8811180580274875`}], 
                   ArrowBox[{{-1.8092210755172822`, 
                    0.9275020928363751}, {-1.8134189343022185`, 
                    0.9120626677079742}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Dashing[{Small, Small}], 
                   Arrowheads[0.02], 
                   CircleBox[{-0.26739035827867424`, 0.5}, 1.7, {
                    1.806705544736802, 1.4532745297407865`}], 
                   ArrowBox[{{-0.08495529802199708, 
                    2.190182667284558}, {-0.06806287472185268, 
                    2.1882738386585587`}}]}, 
                  {RGBColor[0, 0, 1], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-0.26739035827867424`, 0.5}, 1.8, {
                    1.806705544736802, 1.1680778661935227`}], 
                   ArrowBox[{{0.4214720466378801, 2.162969809435091}, {
                    0.4380670247387612, 2.1559981523981815`}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.003], Arrowheads[0.025], 
                   CircleBox[{-0.26739035827867424`, 0.5}, 1.5, {
                    1.806705544736802, 1.7858371943327809`}], 
                   ArrowBox[{{-0.6021096573494591, 
                    1.9621774826708154`}, {-0.5874713903922919, 
                    1.9654515116103575`}}]}, 
                  {RGBColor[0, 0, 1], ArrowBox[{{0, 2.8}, {0, 2.8}}]}},
                 Frame->True,
                 FrameTicks->Automatic,
                 ImageSize->800,
                 PlotRange->{{-3, 3}, {-1, 3}}], 162->
                GraphicsBox[{InsetBox[
                   TagBox[GridBox[{
                    {"\<\"t\"\>", "8.05`"},
                    {"\<\"driveIntent\"\>", "0"},
                    {"\<\"motorTorque\"\>", "108.66415793212938`"},
                    {"\<\"vw\"\>", "0.16130168594909042`"},
                    {"\<\"controllerState\"\>", 
                    RowBox[{"\[LeftAssociation]", 
                    RowBox[{
                    
                    RowBox[{"\<\"lastError\"\>", "\[Rule]", 
                    "0.17502685301878754`"}], ",", 
                    
                    RowBox[{"\<\"lastInt\"\>", "\[Rule]", 
                    "0.029640263198665383`"}], ",", 
                    RowBox[{"\<\"initialized\"\>", "\[Rule]", "True"}]}], 
                    "\[RightAssociation]"}]}
                    },
                    AutoDelete->False,
                    
                    GridBoxItemSize->{
                    "Columns" -> {{Automatic}}, "Rows" -> {{Automatic}}}],
                    "Grid"], {-1, -0.5}], 
                  {GrayLevel[0], LineBox[{{-5, 0}, {5, 0}}]}, 
                  {GrayLevel[0.5], 
                   CircleBox[{-0.25907453631472677`, 0.5}, 0.5], 
                   DiskBox[{-0.25907453631472677`, 0.5}, 0.1]}, 
                  {GrayLevel[0.5], 
                   DiskBox[{-0.591129815293115, 1.9627847728570957`}, 0.1]}, 
                  {GrayLevel[0.5], Thickness[0.01], 
                   LineBox[{{-0.25907453631472677`, 0.5}, {-0.591129815293115,
                     1.9627847728570957`}}]}, 
                  {RGBColor[0, 1, 0], 
                   DiskBox[{-0.48044472230031887`, 1.475189848571397}, 0.05]}, 
                  {RGBColor[0, 0, 1], Arrowheads[0.01], 
                   CircleBox[{-0.25907453631472677`, 0.5}, 0.3333333333333333,
                     NCache[{
                    Rational[-1, 2] Pi, 
                    1.6891284111689848`}, {-1.5707963267948966`, 
                    1.6891284111689848`}]], 
                   ArrowBox[{{-0.2951146409814476, 
                    0.8313792720837001}, {-0.2984265764824654, 
                    0.831002308218157}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-0.25907453631472677`, 
                    0.5}, {-2.4323576949573145`, 0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-0.25907453631472677`, 0.5}, {
                    1.3287051474004543`, 0.5}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.005], Arrowheads[0.04], 
                   ArrowBox[{{-0.25907453631472677`, 
                    0.5}, {-0.8445780112421336, 0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-0.25907453631472677`, 0.5}, 1.6, {
                    1.7940156191307555`, 2.8116249633786965`}], 
                   ArrowBox[{{-1.767499240208826, 
                    1.0335306108200344`}, {-1.772759036789016, 
                    1.0184199388757165`}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Dashing[{Small, Small}], 
                   Arrowheads[0.02], 
                   CircleBox[{-0.25907453631472677`, 0.5}, 1.7, {
                    1.7940156191307555`, 1.4632968341479502`}], 
                   ArrowBox[{{-0.09358788008122301, 
                    2.1919261705549253`}, {-0.07667717462581805, 
                    2.190186735970001}}]}, 
                  {RGBColor[0, 0, 1], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-0.25907453631472677`, 0.5}, 1.8, {
                    1.7940156191307555`, 1.190325852841148}], 
                   ArrowBox[{{0.39262271347370226`, 2.177882801216521}, {
                    0.40936867724918247`, 2.171282043893356}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.003], Arrowheads[0.025], 
                   CircleBox[{-0.25907453631472677`, 0.5}, 1.5, {
                    1.7940156191307555`, 1.780223243800517}], 
                   ArrowBox[{{-0.5855800118793433, 
                    1.964033529133238}, {-0.5709235954546443, 
                    1.9672253284051322`}}]}, 
                  {RGBColor[0, 0, 1], ArrowBox[{{0, 2.8}, {0, 2.8}}]}},
                 Frame->True,
                 FrameTicks->Automatic,
                 ImageSize->800,
                 PlotRange->{{-3, 3}, {-1, 3}}], 163->
                GraphicsBox[{InsetBox[
                   TagBox[GridBox[{
                    {"\<\"t\"\>", "8.1`"},
                    {"\<\"driveIntent\"\>", "0"},
                    {"\<\"motorTorque\"\>", "102.2174440941621`"},
                    {"\<\"vw\"\>", "0.151695324946319`"},
                    {"\<\"controllerState\"\>", 
                    RowBox[{"\[LeftAssociation]", 
                    RowBox[{
                    
                    RowBox[{"\<\"lastError\"\>", "\[Rule]", 
                    "0.16485919569852053`"}], ",", 
                    
                    RowBox[{"\<\"lastInt\"\>", "\[Rule]", 
                    "0.038134463248768735`"}], ",", 
                    RowBox[{"\<\"initialized\"\>", "\[Rule]", "True"}]}], 
                    "\[RightAssociation]"}]}
                    },
                    AutoDelete->False,
                    
                    GridBoxItemSize->{
                    "Columns" -> {{Automatic}}, "Rows" -> {{Automatic}}}],
                    "Grid"], {-1, -0.5}], 
                  {GrayLevel[0], LineBox[{{-5, 0}, {5, 0}}]}, 
                  {GrayLevel[0.5], 
                   CircleBox[{-0.25125754031869374`, 0.5}, 0.5], 
                   DiskBox[{-0.25125754031869374`, 0.5}, 0.1]}, 
                  {GrayLevel[0.5], 
                   DiskBox[{-0.564181398374263, 1.9669964754762086`}, 0.1]}, 
                  {GrayLevel[0.5], Thickness[0.01], 
                   LineBox[{{-0.25125754031869374`, 0.5}, {-0.564181398374263,
                     1.9669964754762086`}}]}, 
                  {RGBColor[0, 1, 0], 
                   DiskBox[{-0.4598734456890732, 1.4779976503174723`}, 0.05]}, 
                  {RGBColor[0, 0, 1], Arrowheads[0.01], 
                   CircleBox[{-0.25125754031869374`, 0.5}, 0.3333333333333333,
                     NCache[{
                    Rational[-1, 2] Pi, 
                    1.4957269960299668`}, {-1.5707963267948966`, 
                    1.4957269960299668`}]], 
                   ArrowBox[{{-0.22293528596648676`, 
                    0.8321279286954351}, {-0.22625792599997982`, 
                    0.8323945402605564}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-0.25125754031869374`, 
                    0.5}, {-2.295606422201936, 0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-0.25125754031869374`, 0.5}, {
                    1.2493501484268075`, 0.5}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.005], Arrowheads[0.04], 
                   ArrowBox[{{-0.25125754031869374`, 
                    0.5}, {-0.7949987334564347, 0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-0.25125754031869374`, 0.5}, 1.6, {
                    1.7809558393160705`, 2.739935449503773}], 
                   ArrowBox[{{-1.7175919191284446`, 
                    1.1402058180933863`}, {-1.7239205545009724`, 
                    1.125510708668977}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Dashing[{Small, Small}], 
                   Arrowheads[0.02], 
                   CircleBox[{-0.25125754031869374`, 0.5}, 1.7, {
                    1.7809558393160705`, 1.4729218943662519`}], 
                   ArrowBox[{{-0.10206318934967185`, 
                    2.1934405940684583`}, {-0.08513652530305937, 
                    2.191864004100263}}]}, 
                  {RGBColor[0, 0, 1], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-0.25125754031869374`, 0.5}, 1.8, {
                    1.7809558393160705`, 1.2130811499040588`}], 
                   ArrowBox[{{0.3620935645323414, 2.192276697877157}, {
                    0.3789853821667266, 2.1860586759235385`}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.003], Arrowheads[0.025], 
                   CircleBox[{-0.25125754031869374`, 0.5}, 1.5, {
                    1.7809558393160705`, 1.7733126586120775`}], 
                   ArrowBox[{{-0.5676379716914643, 
                    1.966254896886752}, {-0.5529598482074471, 
                    1.9693453363367646`}}]}, 
                  {RGBColor[0, 0, 1], ArrowBox[{{0, 2.8}, {0, 2.8}}]}},
                 Frame->True,
                 FrameTicks->Automatic,
                 ImageSize->800,
                 PlotRange->{{-3, 3}, {-1, 3}}], 164->
                GraphicsBox[{InsetBox[
                   TagBox[GridBox[{
                    {"\<\"t\"\>", "8.15`"},
                    {"\<\"driveIntent\"\>", "0"},
                    {"\<\"motorTorque\"\>", "95.6833296867935`"},
                    {"\<\"vw\"\>", "0.1428303059673467`"},
                    {"\<\"controllerState\"\>", 
                    RowBox[{"\[LeftAssociation]", 
                    RowBox[{
                    
                    RowBox[{"\<\"lastError\"\>", "\[Rule]", 
                    "0.15426033628864694`"}], ",", 
                    
                    RowBox[{"\<\"lastInt\"\>", "\[Rule]", 
                    "0.04610862262271266`"}], ",", 
                    RowBox[{"\<\"initialized\"\>", "\[Rule]", "True"}]}], 
                    "\[RightAssociation]"}]}
                    },
                    AutoDelete->False,
                    
                    GridBoxItemSize->{
                    "Columns" -> {{Automatic}}, "Rows" -> {{Automatic}}}],
                    "Grid"], {-1, -0.5}], 
                  {GrayLevel[0], LineBox[{{-5, 0}, {5, 0}}]}, 
                  {GrayLevel[0.5], 
                   CircleBox[{-0.24390187863826765`, 0.5}, 0.5], 
                   DiskBox[{-0.24390187863826765`, 0.5}, 0.1]}, 
                  {GrayLevel[0.5], 
                   DiskBox[{-0.5373384757624554, 1.9710183423289382`}, 0.1]}, 
                  {GrayLevel[0.5], Thickness[0.01], 
                   LineBox[{{-0.24390187863826765`, 
                    0.5}, {-0.5373384757624554, 1.9710183423289382`}}]}, 
                  {RGBColor[0, 1, 0], 
                   DiskBox[{-0.4395262767210595, 1.4806788948859588`}, 0.05]}, 
                  {RGBColor[0, 0, 1], Arrowheads[0.01], 
                   CircleBox[{-0.24390187863826765`, 0.5}, 0.3333333333333333,
                     NCache[{
                    Rational[-1, 2] Pi, 
                    1.2997035638089085`}, {-1.5707963267948966`, 
                    1.2997035638089085`}]], 
                   ArrowBox[{{-0.1514333155838779, 
                    0.8202509577780641}, {-0.15464039517639008`, 
                    0.8211596155828065}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-0.24390187863826765`, 
                    0.5}, {-2.157568472374138, 0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-0.24390187863826765`, 0.5}, {
                    1.1671136280509131`, 0.5}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.005], Arrowheads[0.04], 
                   ArrowBox[{{-0.24390187863826765`, 
                    0.5}, {-0.746552965684957, 0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-0.24390187863826765`, 0.5}, 1.6, {
                    1.7676904364007962`, 2.66694981302202}], 
                   ArrowBox[{{-1.6596481240656529`, 
                    1.2454277755478813`}, {-1.667031490861411, 
                    1.231233277971952}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Dashing[{Small, Small}], 
                   Arrowheads[0.02], 
                   CircleBox[{-0.24390187863826765`, 0.5}, 1.7, {
                    1.7676904364007962`, 1.4821360222792235`}], 
                   ArrowBox[{{-0.11031721839273634`, 
                    2.1947433842759456`}, {-0.09337674618314748, 
                    2.193322823474416}}]}, 
                  {RGBColor[0, 0, 1], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-0.24390187863826765`, 0.5}, 1.8, {
                    1.7676904364007962`, 1.2361163825852766`}], 
                   ArrowBox[{{0.33030796449185107`, 2.205955174104081}, {
                    0.3473385216555479, 2.200127874326049}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.003], Arrowheads[0.025], 
                   CircleBox[{-0.24390187863826765`, 0.5}, 1.5, {
                    1.7676904364007962`, 1.765383237547029}], 
                   ArrowBox[{{-0.5486459331111798, 
                    1.9687174885809084`}, {-0.5339437659346475, 
                    1.9716914430727501`}}]}, 
                  {RGBColor[0, 0, 1], ArrowBox[{{0, 2.8}, {0, 2.8}}]}},
                 Frame->True,
                 FrameTicks->Automatic,
                 ImageSize->800,
                 PlotRange->{{-3, 3}, {-1, 3}}], 165->
                GraphicsBox[{InsetBox[
                   TagBox[GridBox[{
                    {"\<\"t\"\>", "8.2`"},
                    {"\<\"driveIntent\"\>", "0"},
                    {"\<\"motorTorque\"\>", "89.12452673982442`"},
                    {"\<\"vw\"\>", "0.1346853571683896`"},
                    {"\<\"controllerState\"\>", 
                    RowBox[{"\[LeftAssociation]", 
                    RowBox[{
                    
                    RowBox[{"\<\"lastError\"\>", "\[Rule]", 
                    "0.14337955432459087`"}], ",", 
                    
                    RowBox[{"\<\"lastInt\"\>", "\[Rule]", 
                    "0.05354506738365892`"}], ",", 
                    RowBox[{"\<\"initialized\"\>", "\[Rule]", "True"}]}], 
                    "\[RightAssociation]"}]}
                    },
                    AutoDelete->False,
                    
                    GridBoxItemSize->{
                    "Columns" -> {{Automatic}}, "Rows" -> {{Automatic}}}],
                    "Grid"], {-1, -0.5}], 
                  {GrayLevel[0], LineBox[{{-5, 0}, {5, 0}}]}, 
                  {GrayLevel[0.5], 
                   CircleBox[{-0.23697100916001423`, 0.5}, 0.5], 
                   DiskBox[{-0.23697100916001423`, 0.5}, 0.1]}, 
                  {GrayLevel[0.5], 
                   DiskBox[{-0.5107756406906018, 1.97479863837488}, 0.1]}, 
                  {GrayLevel[0.5], Thickness[0.01], 
                   LineBox[{{-0.23697100916001423`, 
                    0.5}, {-0.5107756406906018, 1.97479863837488}}]}, 
                  {RGBColor[0, 1, 0], 
                   DiskBox[{-0.4195074301804059, 1.4831990922499199`}, 0.05]}, 
                  {RGBColor[0, 0, 1], Arrowheads[0.01], 
                   CircleBox[{-0.23697100916001423`, 0.5}, 0.3333333333333333,
                     NCache[{
                    Rational[-1, 2] Pi, 
                    1.102939475399836}, {-1.5707963267948966`, 
                    1.102939475399836}]], 
                   ArrowBox[{{-0.08367862430333445, 
                    0.7959941821321201}, {-0.08664618134790986, 
                    0.7975122808463078}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-0.23697100916001423`, 
                    0.5}, {-2.0194615439565027`, 0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-0.23697100916001423`, 0.5}, {1.083025975042709,
                     0.5}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.005], Arrowheads[0.04], 
                   ArrowBox[{{-0.23697100916001423`, 
                    0.5}, {-0.6994645597537794, 0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-0.23697100916001423`, 0.5}, 1.6, {
                    1.7543619267969717`, 2.5934576111092955`}], 
                   ArrowBox[{{-1.5941618519261418`, 
                    1.3473682884742435`}, {-1.6025675346068984`, 
                    1.3337542381825704`}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Dashing[{Small, Small}], 
                   Arrowheads[0.02], 
                   CircleBox[{-0.23697100916001423`, 0.5}, 1.7, {
                    1.7543619267969717`, 1.4909304179525278`}], 
                   ArrowBox[{{-0.11829556646080386`, 
                    2.1958526290041083`}, {-0.10134325653414142`, 
                    2.1945811024314104`}}]}, 
                  {RGBColor[0, 0, 1], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-0.23697100916001423`, 0.5}, 1.8, {
                    1.7543619267969717`, 1.2592256671312805`}], 
                   ArrowBox[{{0.29766562086357984`, 2.218768068657611}, {
                    0.3148262834815069, 2.213335853775753}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.003], Arrowheads[0.025], 
                   CircleBox[{-0.23697100916001423`, 0.5}, 1.5, {
                    1.7543619267969717`, 1.7566781017742512`}], 
                   ArrowBox[{{-0.49943557468655253`, 
                    1.9768589478494436`}, {-0.5141907949042033, 
                    1.9741605036060172`}}]}, 
                  {RGBColor[0, 0, 1], ArrowBox[{{0, 2.8}, {0, 2.8}}]}},
                 Frame->True,
                 FrameTicks->Automatic,
                 ImageSize->800,
                 PlotRange->{{-3, 3}, {-1, 3}}], 166->
                GraphicsBox[{InsetBox[
                   TagBox[GridBox[{
                    {"\<\"t\"\>", "8.25`"},
                    {"\<\"driveIntent\"\>", "0"},
                    {"\<\"motorTorque\"\>", "82.5968743050094`"},
                    {"\<\"vw\"\>", "0.12723561117866025`"},
                    {"\<\"controllerState\"\>", 
                    RowBox[{"\[LeftAssociation]", 
                    RowBox[{
                    
                    RowBox[{"\<\"lastError\"\>", "\[Rule]", 
                    "0.1323481311025713`"}], ",", 
                    
                    RowBox[{"\<\"lastInt\"\>", "\[Rule]", 
                    "0.06043311961605495`"}], ",", 
                    RowBox[{"\<\"initialized\"\>", "\[Rule]", "True"}]}], 
                    "\[RightAssociation]"}]}
                    },
                    AutoDelete->False,
                    
                    GridBoxItemSize->{
                    "Columns" -> {{Automatic}}, "Rows" -> {{Automatic}}}],
                    "Grid"], {-1, -0.5}], 
                  {GrayLevel[0], LineBox[{{-5, 0}, {5, 0}}]}, 
                  {GrayLevel[0.5], CircleBox[{-0.2304295498133501, 0.5}, 0.5],
                    DiskBox[{-0.2304295498133501, 0.5}, 0.1]}, 
                  {GrayLevel[0.5], 
                   DiskBox[{-0.4846428296794477, 1.9783015958659185`}, 0.1]}, 
                  {GrayLevel[0.5], Thickness[0.01], 
                   LineBox[{{-0.2304295498133501, 0.5}, {-0.4846428296794477, 
                    1.9783015958659185`}}]}, 
                  {RGBColor[0, 1, 0], 
                   DiskBox[{-0.3999050697240818, 1.4855343972439456`}, 0.05]}, 
                  {RGBColor[0, 0, 1], Arrowheads[0.01], 
                   CircleBox[{-0.2304295498133501, 0.5}, 0.3333333333333333, 
                    NCache[{
                    Rational[-1, 2] Pi, 
                    0.9071099023553852}, {-1.5707963267948966`, 
                    0.9071099023553852}]], 
                   ArrowBox[{{-0.022472463037470664`, 
                    0.760509042397381}, {-0.025087907811178117`, 
                    0.7625755532622238}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-0.2304295498133501, 0.5}, {-1.882367035913538, 
                    0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-0.2304295498133501, 0.5}, {0.9980295467258152, 
                    0.5}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.005], Arrowheads[0.04], 
                   ArrowBox[{{-0.2304295498133501, 0.5}, {-0.6539079393743727,
                     0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-0.2304295498133501, 0.5}, 1.6, {
                    1.7410937931753219`, 2.5201502497374673`}], 
                   ArrowBox[{{-1.521912563215351, 
                    1.4444954346602676`}, {-1.5312927865342807`, 
                    1.431533595394146}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Dashing[{Small, Small}], 
                   Arrowheads[0.02], 
                   CircleBox[{-0.2304295498133501, 0.5}, 1.7, {
                    1.7410937931753219`, 1.499299081105889}], 
                   ArrowBox[{{-0.1259501165341348, 
                    2.1967863884477783`}, {-0.1089877593741054, 
                    2.195656772915707}}]}, 
                  {RGBColor[0, 0, 1], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-0.2304295498133501, 0.5}, 1.8, {
                    1.7410937931753219`, 1.282222269258603}], 
                   ArrowBox[{{0.26454337514553916`, 2.2306073510642563`}, {
                    0.2818244117830555, 2.225571174663273}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.003], Arrowheads[0.025], 
                   CircleBox[{-0.2304295498133501, 0.5}, 1.5, {
                    1.7410937931753219`, 1.7474075111295242`}], 
                   ArrowBox[{{-0.47919167818119, 
                    1.9792286515241322`}, {-0.4939712801567876, 
                    1.976667110884368}}]}, 
                  {RGBColor[0, 0, 1], ArrowBox[{{0, 2.8}, {0, 2.8}}]}},
                 Frame->True,
                 FrameTicks->Automatic,
                 ImageSize->800,
                 PlotRange->{{-3, 3}, {-1, 3}}], 167->
                GraphicsBox[{InsetBox[
                   TagBox[GridBox[{
                    {"\<\"t\"\>", "8.3`"},
                    {"\<\"driveIntent\"\>", "0"},
                    {"\<\"motorTorque\"\>", "76.15002069766291`"},
                    {"\<\"vw\"\>", "0.12045360597084592`"},
                    {"\<\"controllerState\"\>", 
                    RowBox[{"\[LeftAssociation]", 
                    RowBox[{
                    
                    RowBox[{"\<\"lastError\"\>", "\[Rule]", 
                    "0.12128137019992367`"}], ",", 
                    
                    RowBox[{"\<\"lastInt\"\>", "\[Rule]", 
                    "0.06676825021119925`"}], ",", 
                    RowBox[{"\<\"initialized\"\>", "\[Rule]", "True"}]}], 
                    "\[RightAssociation]"}]}
                    },
                    AutoDelete->False,
                    
                    GridBoxItemSize->{
                    "Columns" -> {{Automatic}}, "Rows" -> {{Automatic}}}],
                    "Grid"], {-1, -0.5}], 
                  {GrayLevel[0], LineBox[{{-5, 0}, {5, 0}}]}, 
                  {GrayLevel[0.5], CircleBox[{-0.2242434316700287, 0.5}, 0.5],
                    DiskBox[{-0.2242434316700287, 0.5}, 0.1]}, 
                  {GrayLevel[0.5], 
                   DiskBox[{-0.4590682300974668, 1.9815050840424115`}, 0.1]}, 
                  {GrayLevel[0.5], Thickness[0.01], 
                   LineBox[{{-0.2242434316700287, 0.5}, {-0.4590682300974668, 
                    1.9815050840424115`}}]}, 
                  {RGBColor[0, 1, 0], 
                   DiskBox[{-0.38079329728832073`, 1.4876700560282745`}, 
                    0.05]}, 
                  {RGBColor[0, 0, 1], Arrowheads[0.01], 
                   CircleBox[{-0.2242434316700287, 0.5}, 0.3333333333333333, 
                    NCache[{
                    Rational[-1, 2] Pi, 
                    0.7137042941349909}, {-1.5707963267948966`, 
                    0.7137042941349909}]], 
                   ArrowBox[{{0.029906759477864142`, 0.7156821537601121}, {
                    0.0277372664834471, 0.7182128292956169}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-0.2242434316700287, 0.5}, {-1.747243845623287, 
                    0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-0.2242434316700287, 0.5}, {0.9129819005738361, 
                    0.5}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.005], Arrowheads[0.04], 
                   ArrowBox[{{-0.2242434316700287, 0.5}, {-0.6100185133794223,
                     0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-0.2242434316700287, 0.5}, 1.6, {
                    1.7279927996165367`, 2.447631740084798}], 
                   ArrowBox[{{-1.4438986375663871`, 
                    1.535587359294189}, {-1.4541933563101916`, 
                    1.5233392315736054`}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Dashing[{Small, Small}], 
                   Arrowheads[0.02], 
                   CircleBox[{-0.2242434316700287, 0.5}, 1.7, {
                    1.7279927996165367`, 1.5072376050205745`}], 
                   ArrowBox[{{-0.13323712837213142`, 
                    2.19756232662016}, {-0.11626633830881532`, 
                    2.1965674013458067`}}]}, 
                  {RGBColor[0, 0, 1], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-0.2242434316700287, 0.5}, 1.8, {
                    1.7279927996165367`, 1.304937129073965}], 
                   ArrowBox[{{0.23129468147313514`, 2.2414031777489}, {
                    0.24868564630236217`, 2.236760803106802}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.003], Arrowheads[0.025], 
                   CircleBox[{-0.2242434316700287, 0.5}, 1.5, {
                    1.7279927996165367`, 1.737751644256172}], 
                   ArrowBox[{{-0.4587109503802804, 
                    1.981561670221614}, {-0.47351459687854514`, 
                    1.97914295664604}}]}, 
                  {RGBColor[0, 0, 1], ArrowBox[{{0, 2.8}, {0, 2.8}}]}},
                 Frame->True,
                 FrameTicks->Automatic,
                 ImageSize->800,
                 PlotRange->{{-3, 3}, {-1, 3}}], 168->
                GraphicsBox[{InsetBox[
                   TagBox[GridBox[{
                    {"\<\"t\"\>", "8.35`"},
                    {"\<\"driveIntent\"\>", "0"},
                    {"\<\"motorTorque\"\>", "69.82793646004218`"},
                    {"\<\"vw\"\>", "0.11431004042129182`"},
                    {"\<\"controllerState\"\>", 
                    RowBox[{"\[LeftAssociation]", 
                    RowBox[{
                    
                    RowBox[{"\<\"lastError\"\>", "\[Rule]", 
                    "0.11028033687044224`"}], ",", 
                    
                    RowBox[{"\<\"lastInt\"\>", "\[Rule]", 
                    "0.0725513254900738`"}], ",", 
                    RowBox[{"\<\"initialized\"\>", "\[Rule]", "True"}]}], 
                    "\[RightAssociation]"}]}
                    },
                    AutoDelete->False,
                    
                    GridBoxItemSize->{
                    "Columns" -> {{Automatic}}, "Rows" -> {{Automatic}}}],
                    "Grid"], {-1, -0.5}], 
                  {GrayLevel[0], LineBox[{{-5, 0}, {5, 0}}]}, 
                  {GrayLevel[0.5], 
                   CircleBox[{-0.21838000851569367`, 0.5}, 0.5], 
                   DiskBox[{-0.21838000851569367`, 0.5}, 0.1]}, 
                  {GrayLevel[0.5], 
                   DiskBox[{-0.434160704964766, 1.9843984273233226`}, 0.1]}, 
                  {GrayLevel[0.5], Thickness[0.01], 
                   LineBox[{{-0.21838000851569367`, 0.5}, {-0.434160704964766,
                     1.9843984273233226`}}]}, 
                  {RGBColor[0, 1, 0], 
                   DiskBox[{-0.36223380614840855`, 1.4895989515488817`}, 
                    0.05]}, 
                  {RGBColor[0, 0, 1], Arrowheads[0.01], 
                   CircleBox[{-0.21838000851569367`, 0.5}, 0.3333333333333333,
                     NCache[{
                    Rational[-1, 2] Pi, 
                    0.5240417670063691}, {-1.5707963267948966`, 
                    0.5240417670063691}]], 
                   ArrowBox[{{0.07187475342190092, 0.6639002266125993}, {
                    0.07022126585518496, 0.6667945309133844}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-0.21838000851569367`, 
                    0.5}, {-1.6149387377165374`, 0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-0.21838000851569367`, 0.5}, {
                    0.8286581196066813, 0.5}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.005], Arrowheads[0.04], 
                   ArrowBox[{{-0.21838000851569367`, 
                    0.5}, {-0.5679006095941621, 0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-0.21838000851569367`, 0.5}, 1.6, {
                    1.7151509521415522`, 2.3764277542766443`}], 
                   ArrowBox[{{-1.361269035144557, 
                    1.6197341973929926`}, {-1.3724090465219243`, 
                    1.608249511363941}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Dashing[{Small, Small}], 
                   Arrowheads[0.02], 
                   CircleBox[{-0.21838000851569367`, 0.5}, 1.7, {
                    1.7151509521415522`, 1.5147425753815866`}], 
                   ArrowBox[{{-0.14011630350195636`, 
                    2.198197512799239}, {-0.12313852455810852`, 
                    2.1973299796249277`}}]}, 
                  {RGBColor[0, 0, 1], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-0.21838000851569367`, 0.5}, 1.8, {
                    1.7151509521415522`, 1.3272179718079844`}], 
                   ArrowBox[{{0.198248316416962, 2.2511198813513054`}, {
                    0.21573839213596874`, 2.246866112275249}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.003], Arrowheads[0.025], 
                   CircleBox[{-0.21838000851569367`, 0.5}, 1.5, {
                    1.7151509521415522`, 1.7278638043574084`}], 
                   ArrowBox[{{-0.438186859562331, 
                    1.9838075846392487`}, {-0.45301369785772944`, 
                    1.9815353630020935`}}]}, 
                  {RGBColor[0, 0, 1], ArrowBox[{{0, 2.8}, {0, 2.8}}]}},
                 Frame->True,
                 FrameTicks->Automatic,
                 ImageSize->800,
                 PlotRange->{{-3, 3}, {-1, 3}}], 169->
                GraphicsBox[{InsetBox[
                   TagBox[GridBox[{
                    {"\<\"t\"\>", "8.4`"},
                    {"\<\"driveIntent\"\>", "0"},
                    {"\<\"motorTorque\"\>", "63.66930309297147`"},
                    {"\<\"vw\"\>", "0.10877434185731982`"},
                    {"\<\"controllerState\"\>", 
                    RowBox[{"\[LeftAssociation]", 
                    RowBox[{
                    
                    RowBox[{"\<\"lastError\"\>", "\[Rule]", 
                    "0.09943333110438252`"}], ",", 
                    
                    RowBox[{"\<\"lastInt\"\>", "\[Rule]", 
                    "0.07778793390703083`"}], ",", 
                    RowBox[{"\<\"initialized\"\>", "\[Rule]", "True"}]}], 
                    "\[RightAssociation]"}]}
                    },
                    AutoDelete->False,
                    
                    GridBoxItemSize->{
                    "Columns" -> {{Automatic}}, "Rows" -> {{Automatic}}}],
                    "Grid"], {-1, -0.5}], 
                  {GrayLevel[0], LineBox[{{-5, 0}, {5, 0}}]}, 
                  {GrayLevel[0.5], 
                   CircleBox[{-0.21280813365491405`, 0.5}, 0.5], 
                   DiskBox[{-0.21280813365491405`, 0.5}, 0.1]}, 
                  {GrayLevel[0.5], 
                   DiskBox[{-0.41001180222995315`, 1.9869804010478909`}, 
                    0.1]}, 
                  {GrayLevel[0.5], Thickness[0.01], 
                   LineBox[{{-0.21280813365491405`, 
                    0.5}, {-0.41001180222995315`, 1.9869804010478909`}}]}, 
                  {RGBColor[0, 1, 0], 
                   DiskBox[{-0.3442772460382734, 1.4913202673652606`}, 0.05]}, 
                  {RGBColor[0, 0, 1], Arrowheads[0.01], 
                   CircleBox[{-0.21280813365491405`, 0.5}, 0.3333333333333333,
                     NCache[{
                    Rational[-1, 2] Pi, 
                    0.33928276599424767`}, {-1.5707963267948966`, 
                    0.33928276599424767`}]], 
                   ArrowBox[{{0.10261670462976144`, 0.6077881371218439}, {
                    0.10152307011265502`, 0.6109369435722026}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-0.21280813365491405`, 
                    0.5}, {-1.4861941955143436`, 0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-0.21280813365491405`, 0.5}, {
                    0.7457526451636991, 0.5}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.005], Arrowheads[0.04], 
                   ArrowBox[{{-0.21280813365491405`, 
                    0.5}, {-0.5276334166957303, 0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-0.21280813365491405`, 0.5}, 1.6, {
                    1.702647136969978, 2.3069931858547927`}], 
                   ArrowBox[{{-1.275257438939213, 
                    1.696328330225825}, {-1.2871674008318346`, 
                    1.6856441983289383`}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Dashing[{Small, Small}], 
                   Arrowheads[0.02], 
                   CircleBox[{-0.21280813365491405`, 0.5}, 1.7, {
                    1.702647136969978, 1.5218113665807123`}], 
                   ArrowBox[{{-0.14655048762844364`, 
                    2.198708310553354}, {-0.1295670004942403, 
                    2.19796081042824}}]}, 
                  {RGBColor[0, 0, 1], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-0.21280813365491405`, 0.5}, 1.8, {
                    1.702647136969978, 1.3489287864534698`}], 
                   ArrowBox[{{0.1657067517539588, 2.2597518238444616`}, {
                    0.18328505111534274`, 2.2558787512179066`}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.003], Arrowheads[0.025], 
                   CircleBox[{-0.21280813365491405`, 0.5}, 1.5, {
                    1.702647136969978, 1.7178736953861358`}], 
                   ArrowBox[{{-0.41778086328462294`, 
                    1.9859293994359712`}, {-0.4326296610742446, 
                    1.9838054104507883`}}]}, 
                  {RGBColor[0, 0, 1], ArrowBox[{{0, 2.8}, {0, 2.8}}]}},
                 Frame->True,
                 FrameTicks->Automatic,
                 ImageSize->800,
                 PlotRange->{{-3, 3}, {-1, 3}}], 170->
                GraphicsBox[{InsetBox[
                   TagBox[GridBox[{
                    {"\<\"t\"\>", "8.45`"},
                    {"\<\"driveIntent\"\>", "0"},
                    {"\<\"motorTorque\"\>", "57.707814961414535`"},
                    {"\<\"vw\"\>", "0.10381509277974876`"},
                    {"\<\"controllerState\"\>", 
                    RowBox[{"\[LeftAssociation]", 
                    RowBox[{
                    
                    RowBox[{"\<\"lastError\"\>", "\[Rule]", 
                    "0.08881712310415854`"}], ",", 
                    
                    RowBox[{"\<\"lastInt\"\>", "\[Rule]", 
                    "0.08248778020997108`"}], ",", 
                    RowBox[{"\<\"initialized\"\>", "\[Rule]", "True"}]}], 
                    "\[RightAssociation]"}]}
                    },
                    AutoDelete->False,
                    
                    GridBoxItemSize->{
                    "Columns" -> {{Automatic}}, "Rows" -> {{Automatic}}}],
                    "Grid"], {-1, -0.5}], 
                  {GrayLevel[0], LineBox[{{-5, 0}, {5, 0}}]}, 
                  {GrayLevel[0.5], 
                   CircleBox[{-0.20749821210630254`, 0.5}, 0.5], 
                   DiskBox[{-0.20749821210630254`, 0.5}, 0.1]}, 
                  {GrayLevel[0.5], 
                   DiskBox[{-0.3866974202887833, 1.9892574135409808`}, 0.1]}, 
                  {GrayLevel[0.5], Thickness[0.01], 
                   LineBox[{{-0.20749821210630254`, 
                    0.5}, {-0.3866974202887833, 1.9892574135409808`}}]}, 
                  {RGBColor[0, 1, 0], 
                   DiskBox[{-0.326964350894623, 1.492838275693987}, 0.05]}, 
                  {RGBColor[0, 0, 1], Arrowheads[0.01], 
                   CircleBox[{-0.20749821210630254`, 0.5}, 0.3333333333333333,
                     NCache[{
                    Rational[-1, 2] Pi, 
                    0.1604381220475397}, {-1.5707963267948966`, 
                    0.1604381220475397}]], 
                   ArrowBox[{{0.12207029152696253`, 0.5499571068421866}, {
                    0.12155425049682161`, 0.5532502391161832}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-0.20749821210630254`, 
                    0.5}, {-1.3616545113345933`, 0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-0.20749821210630254`, 0.5}, {0.664880939040297,
                     0.5}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.005], Arrowheads[0.04], 
                   ArrowBox[{{-0.20749821210630254`, 
                    0.5}, {-0.48927536018799384`, 0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-0.20749821210630254`, 0.5}, 1.6, {
                    1.6905484796092682`, 2.2397184530226197`}], 
                   ArrowBox[{{-1.1871221866159023`, 
                    1.7650442160517612`}, {-1.199723437146221, 
                    1.755184887892809}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Dashing[{Small, Small}], 
                   Arrowheads[0.02], 
                   CircleBox[{-0.20749821210630254`, 0.5}, 1.7, {
                    1.6905484796092682`, 1.5284421850015442`}], 
                   ArrowBox[{{-0.15250576649096342`, 
                    2.19911030569685}, {-0.13551769621699727`, 
                    2.198475435598734}}]}, 
                  {RGBColor[0, 0, 1], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-0.20749821210630254`, 0.5}, 1.8, {
                    1.6905484796092682`, 1.3699495076014099`}], 
                   ArrowBox[{{0.13394452084628689`, 2.2673191166605613`}, {
                    0.15160034546679885`, 2.2638163810184206`}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.003], Arrowheads[0.025], 
                   CircleBox[{-0.20749821210630254`, 0.5}, 1.5, {
                    1.6905484796092682`, 1.7078905509519922`}], 
                   ArrowBox[{{-0.39762672629451973`, 
                    1.987901592207153}, {-0.412495987887743, 
                    1.9859259442935446`}}]}, 
                  {RGBColor[0, 0, 1], ArrowBox[{{0, 2.8}, {0, 2.8}}]}},
                 Frame->True,
                 FrameTicks->Automatic,
                 ImageSize->800,
                 PlotRange->{{-3, 3}, {-1, 3}}], 171->
                GraphicsBox[{InsetBox[
                   TagBox[GridBox[{
                    {"\<\"t\"\>", "8.5`"},
                    {"\<\"driveIntent\"\>", "0"},
                    {"\<\"motorTorque\"\>", "51.972424669895865`"},
                    {"\<\"vw\"\>", "0.09940035468184512`"},
                    {"\<\"controllerState\"\>", 
                    RowBox[{"\[LeftAssociation]", 
                    RowBox[{
                    
                    RowBox[{"\<\"lastError\"\>", "\[Rule]", 
                    "0.07849798654963783`"}], ",", 
                    
                    RowBox[{"\<\"lastInt\"\>", "\[Rule]", 
                    "0.0866641360640752`"}], ",", 
                    RowBox[{"\<\"initialized\"\>", "\[Rule]", "True"}]}], 
                    "\[RightAssociation]"}]}
                    },
                    AutoDelete->False,
                    
                    GridBoxItemSize->{
                    "Columns" -> {{Automatic}}, "Rows" -> {{Automatic}}}],
                    "Grid"], {-1, -0.5}], 
                  {GrayLevel[0], LineBox[{{-5, 0}, {5, 0}}]}, 
                  {GrayLevel[0.5], 
                   CircleBox[{-0.20242223422567468`, 0.5}, 0.5], 
                   DiskBox[{-0.20242223422567468`, 0.5}, 0.1]}, 
                  {GrayLevel[0.5], 
                   DiskBox[{-0.36427919966849276`, 1.9912418726476406`}, 
                    0.1]}, 
                  {GrayLevel[0.5], Thickness[0.01], 
                   LineBox[{{-0.20242223422567468`, 
                    0.5}, {-0.36427919966849276`, 1.9912418726476406`}}]}, 
                  {RGBColor[0, 1, 0], 
                   DiskBox[{-0.3103268778542201, 1.4941612484317606`}, 0.05]}, 
                  {RGBColor[0, 0, 1], Arrowheads[0.01], 
                   CircleBox[{-0.20242223422567468`, 0.5}, 0.3333333333333333,
                     NCache[{
                    Rational[-1, 2] 
                    Pi, -0.01162358669802055}, {-1.5707963267948966`, \
-0.01162358669802055}]], 
                   ArrowBox[{{0.13083317222719948`, 0.49279269946229964`}, {
                    0.13088858139989906`, 0.4961255583465622}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-0.20242223422567468`, 
                    0.5}, {-1.241870727623592, 0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-0.20242223422567468`, 0.5}, {
                    0.5865812197945828, 0.5}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.005], Arrowheads[0.04], 
                   ArrowBox[{{-0.20242223422567468`, 
                    0.5}, {-0.4528672736033346, 0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-0.20242223422567468`, 0.5}, 1.6, {
                    1.678911471600283, 2.1749347872752245`}], 
                   ArrowBox[{{-1.0980943364788038`, 
                    1.8258097470019825`}, {-1.1113074297497199`, 
                    1.8167868853224574`}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Dashing[{Small, Small}], 
                   Arrowheads[0.02], 
                   CircleBox[{-0.20242223422567468`, 0.5}, 1.7, {
                    1.678911471600283, 1.5346342520729275`}], 
                   ArrowBox[{{-0.1579517806012342, 
                    2.1994182471523707`}, {-0.14096010486882074`, 
                    2.1988885798235627`}}]}, 
                  {RGBColor[0, 0, 1], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-0.20242223422567468`, 0.5}, 1.8, {
                    1.678911471600283, 1.3901757789897504`}], 
                   ArrowBox[{{0.1032068198254687, 2.2738632645499495`}, {
                    0.12092987550321144`, 2.2707183325232387`}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.003], Arrowheads[0.025], 
                   CircleBox[{-0.20242223422567468`, 0.5}, 1.5, {
                    1.678911471600283, 1.6980059950565785`}], 
                   ArrowBox[{{-0.3778344533266901, 
                    1.9897082108218567`}, {-0.39272251661358143`, 
                    1.987879633076239}}]}, 
                  {RGBColor[0, 0, 1], ArrowBox[{{0, 2.8}, {0, 2.8}}]}},
                 Frame->True,
                 FrameTicks->Automatic,
                 ImageSize->800,
                 PlotRange->{{-3, 3}, {-1, 3}}], 172->
                GraphicsBox[{InsetBox[
                   TagBox[GridBox[{
                    {"\<\"t\"\>", "8.55`"},
                    {"\<\"driveIntent\"\>", "0"},
                    {"\<\"motorTorque\"\>", "46.48755584185717`"},
                    {"\<\"vw\"\>", "0.0954979187758939`"},
                    {"\<\"controllerState\"\>", 
                    RowBox[{"\[LeftAssociation]", 
                    RowBox[{
                    
                    RowBox[{"\<\"lastError\"\>", "\[Rule]", 
                    "0.06853256672790245`"}], ",", 
                    
                    RowBox[{"\<\"lastInt\"\>", "\[Rule]", 
                    "0.09033333797459547`"}], ",", 
                    RowBox[{"\<\"initialized\"\>", "\[Rule]", "True"}]}], 
                    "\[RightAssociation]"}]}
                    },
                    AutoDelete->False,
                    
                    GridBoxItemSize->{
                    "Columns" -> {{Automatic}}, "Rows" -> {{Automatic}}}],
                    "Grid"], {-1, -0.5}], 
                  {GrayLevel[0], LineBox[{{-5, 0}, {5, 0}}]}, 
                  {GrayLevel[0.5], 
                   CircleBox[{-0.19755379511098922`, 0.5}, 0.5], 
                   DiskBox[{-0.19755379511098922`, 0.5}, 0.1]}, 
                  {GrayLevel[0.5], 
                   DiskBox[{-0.3428057049730018, 1.9929507301587142`}, 0.1]}, 
                  {GrayLevel[0.5], Thickness[0.01], 
                   LineBox[{{-0.19755379511098922`, 
                    0.5}, {-0.3428057049730018, 1.9929507301587142`}}]}, 
                  {RGBColor[0, 1, 0], 
                   DiskBox[{-0.29438840168566427`, 1.495300486772476}, 0.05]}, 
                  {RGBColor[0, 0, 1], Arrowheads[0.01], 
                   CircleBox[{-0.19755379511098922`, 0.5}, 0.3333333333333333,
                     NCache[{
                    Rational[-1, 2] 
                    Pi, -0.1761696515391813}, {-1.5707963267948966`, \
-0.1761696515391813}]], 
                   ArrowBox[{{0.13001967987676535`, 0.438301299887615}, {
                    0.13062027805756307`, 0.4415800649514843}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-0.19755379511098922`, 
                    0.5}, {-1.1273049119481326`, 0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-0.19755379511098922`, 0.5}, {
                    0.5113164510926088, 0.5}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.005], Arrowheads[0.04], 
                   ArrowBox[{{-0.19755379511098922`, 
                    0.5}, {-0.4184346657445346, 0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-0.19755379511098922`, 0.5}, 1.6, {
                    1.66778291093437, 2.1129187365119657`}], 
                   ArrowBox[{{-1.009335285226666, 
                    1.8787714866161005`}, {-1.0230821815624631`, 
                    1.8705848690113378`}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Dashing[{Small, Small}], 
                   Arrowheads[0.02], 
                   CircleBox[{-0.19755379511098922`, 0.5}, 1.7, {
                    1.66778291093437, 1.5403880514605013`}], 
                   ArrowBox[{{-0.1628621353256357, 
                    2.199645989240506}, {-0.14586769327468027`, 
                    2.199214096833288}}]}, 
                  {RGBColor[0, 0, 1], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-0.19755379511098922`, 0.5}, 1.8, {
                    1.66778291093437, 1.4095187118129413`}], 
                   ArrowBox[{{0.07370850692503758, 2.2794428238901396`}, {
                    0.0914890755895417, 2.276641274680175}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.003], Arrowheads[0.025], 
                   CircleBox[{-0.19755379511098922`, 0.5}, 1.5, {
                    1.66778291093437, 1.6882965785343729`}], 
                   ArrowBox[{{-0.35849377572597285`, 
                    1.9913411154526817`}, {-0.37339889139291726`, 
                    1.9896571760353452`}}]}, 
                  {RGBColor[0, 0, 1], ArrowBox[{{0, 2.8}, {0, 2.8}}]}},
                 Frame->True,
                 FrameTicks->Automatic,
                 ImageSize->800,
                 PlotRange->{{-3, 3}, {-1, 3}}], 173->
                GraphicsBox[{InsetBox[
                   TagBox[GridBox[{
                    {"\<\"t\"\>", "8.6`"},
                    {"\<\"driveIntent\"\>", "0"},
                    {"\<\"motorTorque\"\>", "41.27330163657858`"},
                    {"\<\"vw\"\>", "0.09207550652472486`"},
                    {"\<\"controllerState\"\>", 
                    RowBox[{"\[LeftAssociation]", 
                    RowBox[{
                    
                    RowBox[{"\<\"lastError\"\>", "\[Rule]", 
                    "0.05896861914549634`"}], ",", 
                    
                    RowBox[{"\<\"lastInt\"\>", "\[Rule]", 
                    "0.0935143251710266`"}], ",", 
                    RowBox[{"\<\"initialized\"\>", "\[Rule]", "True"}]}], 
                    "\[RightAssociation]"}]}
                    },
                    AutoDelete->False,
                    
                    GridBoxItemSize->{
                    "Columns" -> {{Automatic}}, "Rows" -> {{Automatic}}}],
                    "Grid"], {-1, -0.5}], 
                  {GrayLevel[0], LineBox[{{-5, 0}, {5, 0}}]}, 
                  {GrayLevel[0.5], 
                   CircleBox[{-0.19286810283341677`, 0.5}, 0.5], 
                   DiskBox[{-0.19286810283341677`, 0.5}, 0.1]}, 
                  {GrayLevel[0.5], 
                   DiskBox[{-0.32231345308133236`, 1.9944041960926082`}, 
                    0.1]}, 
                  {GrayLevel[0.5], Thickness[0.01], 
                   LineBox[{{-0.19286810283341677`, 
                    0.5}, {-0.32231345308133236`, 1.9944041960926082`}}]}, 
                  {RGBColor[0, 1, 0], 
                   DiskBox[{-0.27916500299869385`, 1.4962694640617387`}, 
                    0.05]}, 
                  {RGBColor[0, 0, 1], Arrowheads[0.01], 
                   CircleBox[{-0.19286810283341677`, 0.5}, 0.3333333333333333,
                     NCache[{
                    Rational[-1, 2] 
                    Pi, -0.3325972776975392}, {-1.5707963267948966`, \
-0.3325972776975392}]], 
                   ArrowBox[{{0.12109367319180664`, 0.38802181325723056`}, {
                    0.12219773843831222`, 0.39116697755346147`}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-0.19286810283341677`, 
                    0.5}, {-1.0183341355649884`, 0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-0.19286810283341677`, 0.5}, {
                    0.4394767009240887, 0.5}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.005], Arrowheads[0.04], 
                   ArrowBox[{{-0.19286810283341677`, 
                    0.5}, {-0.38598933180748296`, 0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-0.19286810283341677`, 0.5}, 1.6, {
                    1.6572006985771182`, 2.0538960873298437`}], 
                   ArrowBox[{{-0.9219045082199668, 
                    1.9242562689421654`}, {-0.936110382018025, 
                    1.9168948141737512`}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Dashing[{Small, Small}], 
                   Arrowheads[0.02], 
                   CircleBox[{-0.19286810283341677`, 0.5}, 1.7, {
                    1.6572006985771182`, 1.5457055874791805`}], 
                   ArrowBox[{{-0.16721481969469681`, 
                    2.199806432822339}, {-0.15021832131959717`, 
                    2.1994649146530865`}}]}, 
                  {RGBColor[0, 0, 1], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-0.19286810283341677`, 0.5}, 1.8, {
                    1.6572006985771182`, 1.4279045783739037`}], 
                   ArrowBox[{{0.04563359663734212, 2.28412918236028}, {
                    0.06346266612196982, 2.2816549993999233`}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.003], Arrowheads[0.025], 
                   CircleBox[{-0.19286810283341677`, 0.5}, 1.5, {
                    1.6572006985771182`, 1.678825976425172}], 
                   ArrowBox[{{-0.3396771787831615, 
                    1.9927984107771493`}, {-0.35459757369981465`, 
                    1.991255705187167}}]}, 
                  {RGBColor[0, 0, 1], ArrowBox[{{0, 2.8}, {0, 2.8}}]}},
                 Frame->True,
                 FrameTicks->Automatic,
                 ImageSize->800,
                 PlotRange->{{-3, 3}, {-1, 3}}], 174->
                GraphicsBox[{InsetBox[
                   TagBox[GridBox[{
                    {"\<\"t\"\>", "8.65`"},
                    {"\<\"driveIntent\"\>", "0"},
                    {"\<\"motorTorque\"\>", "36.345622444461185`"},
                    {"\<\"vw\"\>", "0.08910093707686602`"},
                    {"\<\"controllerState\"\>", 
                    RowBox[{"\[LeftAssociation]", 
                    RowBox[{
                    
                    RowBox[{"\<\"lastError\"\>", "\[Rule]", 
                    "0.04984565087141693`"}], ",", 
                    
                    RowBox[{"\<\"lastInt\"\>", "\[Rule]", 
                    "0.09622821181587353`"}], ",", 
                    RowBox[{"\<\"initialized\"\>", "\[Rule]", "True"}]}], 
                    "\[RightAssociation]"}]}
                    },
                    AutoDelete->False,
                    
                    GridBoxItemSize->{
                    "Columns" -> {{Automatic}}, "Rows" -> {{Automatic}}}],
                    "Grid"], {-1, -0.5}], 
                  {GrayLevel[0], LineBox[{{-5, 0}, {5, 0}}]}, 
                  {GrayLevel[0.5], CircleBox[{-0.1883419775462907, 0.5}, 0.5],
                    DiskBox[{-0.1883419775462907, 0.5}, 0.1]}, 
                  {GrayLevel[0.5], 
                   DiskBox[{-0.3028278344135068, 1.9956246148607542`}, 0.1]}, 
                  {GrayLevel[0.5], Thickness[0.01], 
                   LineBox[{{-0.1883419775462907, 0.5}, {-0.3028278344135068, 
                    1.9956246148607542`}}]}, 
                  {RGBColor[0, 1, 0], 
                   DiskBox[{-0.26466588212443476`, 1.497083076573836}, 0.05]}, 
                  {RGBColor[0, 0, 1], Arrowheads[0.01], 
                   CircleBox[{-0.1883419775462907, 0.5}, 0.3333333333333333, 
                    NCache[{
                    Rational[-1, 2] 
                    Pi, -0.480427653461061}, {-1.5707963267948966`, \
-0.480427653461061}]], 
                   ArrowBox[{{0.10570186016787544`, 0.34299894072511417`}, {
                    0.10725714252454407`, 0.3459471800827414}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-0.1883419775462907, 0.5}, {-0.9152544264355144,
                     0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-0.1883419775462907, 0.5}, {0.3713819384635967, 
                    0.5}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.005], Arrowheads[0.04], 
                   ArrowBox[{{-0.1883419775462907, 0.5}, {-0.355530510425627, 
                    0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-0.1883419775462907, 0.5}, 1.6, {
                    1.647194528396175, 1.9980453809733265`}], 
                   ArrowBox[{{-0.8367372657361305, 
                    1.962731537314081}, {-0.851331917827655, 
                    1.9561745565301274`}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Dashing[{Small, Small}], 
                   Arrowheads[0.02], 
                   CircleBox[{-0.1883419775462907, 0.5}, 1.7, {
                    1.647194528396175, 1.5505906182829912`}], 
                   ArrowBox[{{-0.17099257425519207`, 
                    2.1999114677551423`}, {-0.15399461035773765`, 
                    2.1996529817486907`}}]}, 
                  {RGBColor[0, 0, 1], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-0.1883419775462907, 0.5}, 1.8, {
                    1.647194528396175, 1.445274403704724}], 
                   ArrowBox[{{0.019135289604315986`, 2.288002568123356}, {
                    0.03700464350970256, 2.285838430647814}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.003], Arrowheads[0.025], 
                   CircleBox[{-0.1883419775462907, 0.5}, 1.5, {
                    1.647194528396175, 1.6696468567204534`}], 
                   ArrowBox[{{-0.3214424858620929, 
                    1.994083081587525}, {-0.33637641269540886`, 
                    1.9926773951561274`}}]}, 
                  {RGBColor[0, 0, 1], ArrowBox[{{0, 2.8}, {0, 2.8}}]}},
                 Frame->True,
                 FrameTicks->Automatic,
                 ImageSize->800,
                 PlotRange->{{-3, 3}, {-1, 3}}], 175->
                GraphicsBox[{InsetBox[
                   TagBox[GridBox[{
                    {"\<\"t\"\>", "8.700000000000001`"},
                    {"\<\"driveIntent\"\>", "0"},
                    {"\<\"motorTorque\"\>", "31.716551960322484`"},
                    {"\<\"vw\"\>", "0.08654227390131719`"},
                    {"\<\"controllerState\"\>", 
                    RowBox[{"\[LeftAssociation]", 
                    RowBox[{
                    
                    RowBox[{"\<\"lastError\"\>", "\[Rule]", 
                    "0.04119549243841261`"}], ",", 
                    
                    RowBox[{"\<\"lastInt\"\>", "\[Rule]", 
                    "0.09849788940389366`"}], ",", 
                    RowBox[{"\<\"initialized\"\>", "\[Rule]", "True"}]}], 
                    "\[RightAssociation]"}]}
                    },
                    AutoDelete->False,
                    
                    GridBoxItemSize->{
                    "Columns" -> {{Automatic}}, "Rows" -> {{Automatic}}}],
                    "Grid"], {-1, -0.5}], 
                  {GrayLevel[0], LineBox[{{-5, 0}, {5, 0}}]}, 
                  {GrayLevel[0.5], 
                   CircleBox[{-0.18395384279512111`, 0.5}, 0.5], 
                   DiskBox[{-0.18395384279512111`, 0.5}, 0.1]}, 
                  {GrayLevel[0.5], 
                   DiskBox[{-0.28436396466040686`, 1.996635495846266}, 0.1]}, 
                  {GrayLevel[0.5], Thickness[0.01], 
                   LineBox[{{-0.18395384279512111`, 
                    0.5}, {-0.28436396466040686`, 1.996635495846266}}]}, 
                  {RGBColor[0, 1, 0], 
                   DiskBox[{-0.25089392403864497`, 1.497756997230844}, 0.05]}, 
                  {RGBColor[0, 0, 1], Arrowheads[0.01], 
                   CircleBox[{-0.18395384279512111`, 0.5}, 0.3333333333333333,
                     NCache[{
                    Rational[-1, 2] 
                    Pi, -0.6192997679852219}, {-1.5707963267948966`, \
-0.6192997679852219}]], 
                   ArrowBox[{{0.0855261066053134, 0.3038070643925976}, {
                    0.08747452937754124, 0.3065116285385352}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-0.18395384279512111`, 
                    0.5}, {-0.8182848820015709, 0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-0.18395384279512111`, 0.5}, {
                    0.3072852902540419, 0.5}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.005], Arrowheads[0.04], 
                   ArrowBox[{{-0.18395384279512111`, 
                    0.5}, {-0.3270457489524078, 0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-0.18395384279512111`, 0.5}, 1.6, {
                    1.6377865019147826`, 1.9455011649117218`}], 
                   ArrowBox[{{-0.7546315407209516, 
                    1.9947665252774678`}, {-0.7695504232001031, 
                    1.9889851057072372`}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Dashing[{Small, Small}], 
                   Arrowheads[0.02], 
                   CircleBox[{-0.18395384279512111`, 0.5}, 1.7, {
                    1.6377865019147826`, 1.5550488396713589`}], 
                   ArrowBox[{{-0.1741831684787522, 
                    2.1999719215102944`}, {-0.1571842211205309, 
                    2.1997892185078123`}}]}, 
                  {RGBColor[0, 0, 1], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-0.18395384279512111`, 0.5}, 1.8, {
                    1.6377865019147826`, 1.4615834354685466`}], 
                   ArrowBox[{{-0.005663465886840857, 2.2911483862321136`}, {
                    0.012238805007683873`, 2.2892759555049427`}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.003], Arrowheads[0.025], 
                   CircleBox[{-0.18395384279512111`, 0.5}, 1.5, {
                    1.6377865019147826`, 1.660802445500154}], 
                   ArrowBox[{{-0.3038350324503975, 
                    1.9952018259642528`}, {-0.3187808075014493, 
                    1.9939282745794988`}}]}, 
                  {RGBColor[0, 0, 1], ArrowBox[{{0, 2.8}, {0, 2.8}}]}},
                 Frame->True,
                 FrameTicks->Automatic,
                 ImageSize->800,
                 PlotRange->{{-3, 3}, {-1, 3}}], 176->
                GraphicsBox[{InsetBox[
                   TagBox[GridBox[{
                    {"\<\"t\"\>", "8.75`"},
                    {"\<\"driveIntent\"\>", "0"},
                    {"\<\"motorTorque\"\>", "27.394417206301398`"},
                    {"\<\"vw\"\>", "0.08436795899241967`"},
                    {"\<\"controllerState\"\>", 
                    RowBox[{"\[LeftAssociation]", 
                    RowBox[{
                    
                    RowBox[{"\<\"lastError\"\>", "\[Rule]", 
                    "0.03304282325118621`"}], ",", 
                    
                    RowBox[{"\<\"lastInt\"\>", "\[Rule]", 
                    "0.1003476564855996`"}], ",", 
                    RowBox[{"\<\"initialized\"\>", "\[Rule]", "True"}]}], 
                    "\[RightAssociation]"}]}
                    },
                    AutoDelete->False,
                    
                    GridBoxItemSize->{
                    "Columns" -> {{Automatic}}, "Rows" -> {{Automatic}}}],
                    "Grid"], {-1, -0.5}], 
                  {GrayLevel[0], LineBox[{{-5, 0}, {5, 0}}]}, 
                  {GrayLevel[0.5], CircleBox[{-0.1796837098405753, 0.5}, 0.5],
                    DiskBox[{-0.1796837098405753, 0.5}, 0.1]}, 
                  {GrayLevel[0.5], 
                   DiskBox[{-0.26692749529673354`, 1.9974606912701516`}, 
                    0.1]}, 
                  {GrayLevel[0.5], Thickness[0.01], 
                   LineBox[{{-0.1796837098405753, 
                    0.5}, {-0.26692749529673354`, 1.9974606912701516`}}]}, 
                  {RGBColor[0, 1, 0], 
                   DiskBox[{-0.23784623347801412`, 1.4983071275134343`}, 
                    0.05]}, 
                  {RGBColor[0, 0, 1], Arrowheads[0.01], 
                   CircleBox[{-0.1796837098405753, 0.5}, 0.3333333333333333, 
                    NCache[{
                    Rational[-1, 2] 
                    Pi, -0.7489638106058547}, {-1.5707963267948966`, \
-0.7489638106058547}]], 
                   ArrowBox[{{0.06216611502075464, 0.2706099973284469}, {
                    0.0644478844255216, 0.27303992467351235`}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-0.1796837098405753, 0.5}, {-0.7275720539666033,
                     0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-0.1796837098405753, 0.5}, {
                    0.24737673964362272`, 0.5}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.005], Arrowheads[0.04], 
                   ArrowBox[{{-0.1796837098405753, 0.5}, {-0.3005116044824052,
                     0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-0.1796837098405753, 0.5}, 1.6, {
                    1.6289916932515651`, 1.896357089702701}], 
                   ArrowBox[{{-0.6762430599688241, 
                    2.020995993354424}, {-0.6914279386436961, 
                    2.015954433446698}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Dashing[{Small, Small}], 
                   Arrowheads[0.02], 
                   CircleBox[{-0.1796837098405753, 0.5}, 1.7, {
                    1.6289916932515651`, 1.5590880048590425`}], 
                   ArrowBox[{{-0.1767795639621829, 
                    2.199997519391342}, {-0.15978001730585656`, 
                    2.1998834792489403`}}]}, 
                  {RGBColor[0, 0, 1], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-0.1796837098405753, 0.5}, 1.8, {
                    1.6289916932515651`, 1.4768004865498907`}], 
                   ArrowBox[{{-0.028668919479038713`, 
                    2.293653961357112}, {-0.010740229482895214`, 
                    2.2920541566717882`}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.003], Arrowheads[0.025], 
                   CircleBox[{-0.1796837098405753, 0.5}, 1.5, {
                    1.6289916932515651`, 1.652327821245908}], 
                   ArrowBox[{{-0.28688947264639725`, 
                    1.9961640700208054`}, {-0.30184550374370245`, 
                    1.9950172226801837`}}]}, 
                  {RGBColor[0, 0, 1], ArrowBox[{{0, 2.8}, {0, 2.8}}]}},
                 Frame->True,
                 FrameTicks->Automatic,
                 ImageSize->800,
                 PlotRange->{{-3, 3}, {-1, 3}}], 177->
                GraphicsBox[{InsetBox[
                   TagBox[GridBox[{
                    {"\<\"t\"\>", "8.8`"},
                    {"\<\"driveIntent\"\>", "0"},
                    {"\<\"motorTorque\"\>", "23.38407503948575`"},
                    {"\<\"vw\"\>", "0.0825469398623373`"},
                    {"\<\"controllerState\"\>", 
                    RowBox[{"\[LeftAssociation]", 
                    RowBox[{
                    
                    RowBox[{"\<\"lastError\"\>", "\[Rule]", 
                    "0.025405668520809596`"}], ",", 
                    
                    RowBox[{"\<\"lastInt\"\>", "\[Rule]", 
                    "0.10180287387008677`"}], ",", 
                    RowBox[{"\<\"initialized\"\>", "\[Rule]", "True"}]}], 
                    "\[RightAssociation]"}]}
                    },
                    AutoDelete->False,
                    
                    GridBoxItemSize->{
                    "Columns" -> {{Automatic}}, "Rows" -> {{Automatic}}}],
                    "Grid"], {-1, -0.5}], 
                  {GrayLevel[0], LineBox[{{-5, 0}, {5, 0}}]}, 
                  {GrayLevel[0.5], 
                   CircleBox[{-0.17551315547122898`, 0.5}, 0.5], 
                   DiskBox[{-0.17551315547122898`, 0.5}, 0.1]}, 
                  {GrayLevel[0.5], 
                   DiskBox[{-0.2505154028568333, 1.9981237141461679`}, 0.1]}, 
                  {GrayLevel[0.5], Thickness[0.01], 
                   LineBox[{{-0.17551315547122898`, 
                    0.5}, {-0.2505154028568333, 1.9981237141461679`}}]}, 
                  {RGBColor[0, 1, 0], 
                   DiskBox[{-0.22551465372829854`, 1.4987491427641118`}, 
                    0.05]}, 
                  {RGBColor[0, 0, 1], Arrowheads[0.01], 
                   CircleBox[{-0.17551315547122898`, 0.5}, 0.3333333333333333,
                     NCache[{
                    Rational[-1, 2] 
                    Pi, -0.869274075610324}, {-1.5707963267948966`, \
-0.869274075610324}]], 
                   ArrowBox[{{0.037057002669153544`, 0.2432412825641866}, {
                    0.03961391863126967, 0.24537978654629722`}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-0.17551315547122898`, 
                    0.5}, {-0.6431946562609441, 0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-0.17551315547122898`, 0.5}, {
                    0.1917872234400534, 0.5}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.005], Arrowheads[0.04], 
                   ArrowBox[{{-0.17551315547122898`, 
                    0.5}, {-0.27589427734966165`, 0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-0.17551315547122898`, 0.5}, 1.6, {
                    1.6208186837341327`, 1.850668930397894}], 
                   ArrowBox[{{-0.6020870012071102, 
                    2.042087790670201}, {-0.6174862935859173, 
                    2.0377450195611284`}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Dashing[{Small, Small}], 
                   Arrowheads[0.02], 
                   CircleBox[{-0.17551315547122898`, 0.5}, 1.7, {
                    1.6208186837341327`, 1.562717973157473}], 
                   ArrowBox[{{-0.17877995227704435`, 
                    2.1999968611849345`}, {-0.16178010365810938`, 
                    2.1999445294737994`}}]}, 
                  {RGBColor[0, 0, 1], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-0.17551315547122898`, 0.5}, 1.8, {
                    1.6208186837341327`, 1.4909071557369895`}], 
                   ArrowBox[{{-0.04981503457613842, 
                    2.2956057424733984`}, {-0.03186556127120255, 
                    2.294259002675072}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.003], Arrowheads[0.025], 
                   CircleBox[{-0.17551315547122898`, 0.5}, 1.5, {
                    1.6208186837341327`, 1.6442509745920257`}], 
                   ArrowBox[{{-0.2706312651121644, 
                    1.996981143908745}, {-0.2855960711897924, 
                    1.9959551302318193`}}]}, 
                  {RGBColor[0, 0, 1], ArrowBox[{{0, 2.8}, {0, 2.8}}]}},
                 Frame->True,
                 FrameTicks->Automatic,
                 ImageSize->800,
                 PlotRange->{{-3, 3}, {-1, 3}}], 178->
                GraphicsBox[{InsetBox[
                   TagBox[GridBox[{
                    {"\<\"t\"\>", "8.85`"},
                    {"\<\"driveIntent\"\>", "0"},
                    {"\<\"motorTorque\"\>", "19.687165220370467`"},
                    {"\<\"vw\"\>", "0.08104879206916883`"},
                    {"\<\"controllerState\"\>", 
                    RowBox[{"\[LeftAssociation]", 
                    RowBox[{
                    
                    RowBox[{"\<\"lastError\"\>", "\[Rule]", 
                    "0.01829588104603261`"}], ",", 
                    
                    RowBox[{"\<\"lastInt\"\>", "\[Rule]", 
                    "0.1028896442419812`"}], ",", 
                    RowBox[{"\<\"initialized\"\>", "\[Rule]", "True"}]}], 
                    "\[RightAssociation]"}]}
                    },
                    AutoDelete->False,
                    
                    GridBoxItemSize->{
                    "Columns" -> {{Automatic}}, "Rows" -> {{Automatic}}}],
                    "Grid"], {-1, -0.5}], 
                  {GrayLevel[0], LineBox[{{-5, 0}, {5, 0}}]}, 
                  {GrayLevel[0.5], 
                   CircleBox[{-0.17142529358766345`, 0.5}, 0.5], 
                   DiskBox[{-0.17142529358766345`, 0.5}, 0.1]}, 
                  {GrayLevel[0.5], 
                   DiskBox[{-0.2351167696129026, 1.9986471885943424`}, 0.1]}, 
                  {GrayLevel[0.5], Thickness[0.01], 
                   LineBox[{{-0.17142529358766345`, 
                    0.5}, {-0.2351167696129026, 1.9986471885943424`}}]}, 
                  {RGBColor[0, 1, 0], 
                   DiskBox[{-0.21388627760448953`, 1.4990981257295615`}, 
                    0.05]}, 
                  {RGBColor[0, 0, 1], Arrowheads[0.01], 
                   CircleBox[{-0.17142529358766345`, 0.5}, 0.3333333333333333,
                     NCache[{
                    Rational[-1, 2] 
                    Pi, -0.9801813701837825}, {-1.5707963267948966`, \
-0.9801813701837825}]], 
                   ArrowBox[{{0.011420780203142328`, 0.221291506389923}, {
                    0.014198676460555837`, 0.2231338719621902}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-0.17142529358766345`, 
                    0.5}, {-0.5651685979950728, 0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-0.17142529358766345`, 0.5}, {
                    0.14059305775310849`, 0.5}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.005], Arrowheads[0.04], 
                   ArrowBox[{{-0.17142529358766345`, 
                    0.5}, {-0.25315024665430086`, 0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-0.17142529358766345`, 0.5}, 1.6, {
                    1.6132700802414035`, 1.8084575844173134`}], 
                   ArrowBox[{{-0.5325448892126219, 
                    2.058715059802678}, {-0.548113724196789, 
                    2.0550259889291995`}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Dashing[{Small, Small}], 
                   Arrowheads[0.02], 
                   CircleBox[{-0.17142529358766345`, 0.5}, 1.7, {
                    1.6132700802414035`, 1.565950686675167}], 
                   ArrowBox[{{-0.18018766658505364`, 
                    2.1999774177380873`}, {-0.16318773762082686`, 
                    2.1999800418451074`}}]}, 
                  {RGBColor[0, 0, 1], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-0.17142529358766345`, 0.5}, 1.8, {
                    1.6132700802414035`, 1.5038969401282343`}], 
                   ArrowBox[{{-0.06906165285791335, 
                    2.297086999857422}, {-0.05109620051172664, 
                    2.295973526909437}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.003], Arrowheads[0.025], 
                   CircleBox[{-0.17142529358766345`, 0.5}, 1.5, {
                    1.6132700802414035`, 1.6365936705181694`}], 
                   ArrowBox[{{-0.25507788685727495`, 
                    1.9976655980689644`}, {-0.2700501106334713, 
                    1.9967542034224193`}}]}, 
                  {RGBColor[0, 0, 1], ArrowBox[{{0, 2.8}, {0, 2.8}}]}},
                 Frame->True,
                 FrameTicks->Automatic,
                 ImageSize->800,
                 PlotRange->{{-3, 3}, {-1, 3}}], 179->
                GraphicsBox[{InsetBox[
                   TagBox[GridBox[{
                    {"\<\"t\"\>", "8.9`"},
                    {"\<\"driveIntent\"\>", "0"},
                    {"\<\"motorTorque\"\>", "16.302378220390864`"},
                    {"\<\"vw\"\>", "0.07984383816494897`"},
                    {"\<\"controllerState\"\>", 
                    RowBox[{"\[LeftAssociation]", 
                    RowBox[{
                    
                    RowBox[{"\<\"lastError\"\>", "\[Rule]", 
                    "0.011719616885388175`"}], ",", 
                    
                    RowBox[{"\<\"lastInt\"\>", "\[Rule]", 
                    "0.10363451568222198`"}], ",", 
                    RowBox[{"\<\"initialized\"\>", "\[Rule]", "True"}]}], 
                    "\[RightAssociation]"}]}
                    },
                    AutoDelete->False,
                    
                    GridBoxItemSize->{
                    "Columns" -> {{Automatic}}, "Rows" -> {{Automatic}}}],
                    "Grid"], {-1, -0.5}], 
                  {GrayLevel[0], LineBox[{{-5, 0}, {5, 0}}]}, 
                  {GrayLevel[0.5], CircleBox[{-0.1674047407519831, 0.5}, 0.5],
                    DiskBox[{-0.1674047407519831, 0.5}, 0.1]}, 
                  {GrayLevel[0.5], 
                   DiskBox[{-0.22071356210753568`, 1.999052423888398}, 0.1]}, 
                  {GrayLevel[0.5], Thickness[0.01], 
                   LineBox[{{-0.1674047407519831, 
                    0.5}, {-0.22071356210753568`, 1.999052423888398}}]}, 
                  {RGBColor[0, 1, 0], 
                   DiskBox[{-0.20294395498901815`, 1.4993682825922654`}, 
                    0.05]}, 
                  {RGBColor[0, 0, 1], Arrowheads[0.01], 
                   CircleBox[{-0.1674047407519831, 0.5}, 0.3333333333333333, 
                    NCache[{
                    Rational[-1, 2] 
                    Pi, -1.0817249801831705`}, {-1.5707963267948966`, \
-1.0817249801831705`}]], 
                   ArrowBox[{{-0.01375298551833104, 
                    0.20419220898575208`}, {-0.010802639432980876`, 
                    0.20574349119588864`}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-0.1674047407519831, 
                    0.5}, {-0.49345230515980043`, 0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-0.1674047407519831, 0.5}, {0.09382061080702755,
                     0.5}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.005], Arrowheads[0.04], 
                   ArrowBox[{{-0.1674047407519831, 
                    0.5}, {-0.23222695360078974`, 0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-0.1674047407519831, 0.5}, 1.6, {
                    1.6063430265035437`, 1.769712074299303}], 
                   ArrowBox[{{-0.4678752105614782, 
                    2.0715334857305656`}, {-0.483575260099551, 
                    2.0684502550911485`}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Dashing[{Small, Small}], 
                   Arrowheads[0.02], 
                   CircleBox[{-0.1674047407519831, 0.5}, 1.7, {
                    1.6063430265035437`, 1.568800079069293}], 
                   ArrowBox[{{-0.18101097434803626`, 
                    2.199945549247778}, {-0.1640111218723895, 
                    2.19999661274689}}]}, 
                  {RGBColor[0, 0, 1], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-0.1674047407519831, 0.5}, 1.8, {
                    1.6063430265035437`, 1.5157742586124834`}], 
                   ArrowBox[{{-0.08639239298804703, 
                    2.2981760201687083`}, {-0.06841498306449799, 
                    2.2972760021412886`}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.003], Arrowheads[0.025], 
                   CircleBox[{-0.1674047407519831, 0.5}, 1.5, {
                    1.6063430265035437`, 1.6293721489242636`}], 
                   ArrowBox[{{-0.24023982232164467`, 
                    1.9982306400860785`}, {-0.25521823729491666`, 
                    1.997427390501758}}]}, 
                  {RGBColor[0, 0, 1], ArrowBox[{{0, 2.8}, {0, 2.8}}]}},
                 Frame->True,
                 FrameTicks->Automatic,
                 ImageSize->800,
                 PlotRange->{{-3, 3}, {-1, 3}}], 180->
                GraphicsBox[{InsetBox[
                   TagBox[GridBox[{
                    {"\<\"t\"\>", "8.950000000000001`"},
                    {"\<\"driveIntent\"\>", "0"},
                    {"\<\"motorTorque\"\>", "13.22573458315242`"},
                    {"\<\"vw\"\>", "0.07890326261803975`"},
                    {"\<\"controllerState\"\>", 
                    RowBox[{"\[LeftAssociation]", 
                    RowBox[{
                    
                    RowBox[{"\<\"lastError\"\>", "\[Rule]", 
                    "0.005677810233081431`"}], ",", 
                    
                    RowBox[{"\<\"lastInt\"\>", "\[Rule]", 
                    "0.1040642089369894`"}], ",", 
                    RowBox[{"\<\"initialized\"\>", "\[Rule]", "True"}]}], 
                    "\[RightAssociation]"}]}
                    },
                    AutoDelete->False,
                    
                    GridBoxItemSize->{
                    "Columns" -> {{Automatic}}, "Rows" -> {{Automatic}}}],
                    "Grid"], {-1, -0.5}], 
                  {GrayLevel[0], LineBox[{{-5, 0}, {5, 0}}]}, 
                  {GrayLevel[0.5], 
                   CircleBox[{-0.16343757588860044`, 0.5}, 0.5], 
                   DiskBox[{-0.16343757588860044`, 0.5}, 0.1]}, 
                  {GrayLevel[0.5], 
                   DiskBox[{-0.20728140902279496`, 1.999359102515505}, 0.1]}, 
                  {GrayLevel[0.5], Thickness[0.01], 
                   LineBox[{{-0.16343757588860044`, 
                    0.5}, {-0.20728140902279496`, 1.999359102515505}}]}, 
                  {RGBColor[0, 1, 0], 
                   DiskBox[{-0.19266679797806346`, 1.4995727350103367`}, 
                    0.05]}, 
                  {RGBColor[0, 0, 1], Arrowheads[0.01], 
                   CircleBox[{-0.16343757588860044`, 0.5}, 0.3333333333333333,
                     NCache[{
                    Rational[-1, 2] 
                    Pi, -1.1740242893003239`}, {-1.5707963267948966`, \
-1.1740242893003239`}]], 
                   ArrowBox[{{-0.037703953284947994`, 
                    0.19128950899252972`}, {-0.03462318645510587, 
                    0.19256225965898865`}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-0.16343757588860044`, 
                    0.5}, {-0.42795226755164884`, 0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-0.16343757588860044`, 0.5}, {
                    0.05145113354605199, 0.5}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.005], Arrowheads[0.04], 
                   ArrowBox[{{-0.16343757588860044`, 
                    0.5}, {-0.2130635581169964, 0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-0.16343757588860044`, 0.5}, 1.6, {
                    1.6000297124703746`, 1.7343925644095775`}], 
                   ArrowBox[{{-0.40822639863539534`, 
                    2.081163632347468}, {-0.42402553209377336`, 
                    2.0786367273951316`}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Dashing[{Small, Small}], 
                   Arrowheads[0.02], 
                   CircleBox[{-0.16343757588860044`, 0.5}, 1.7, {
                    1.6000297124703746`, 1.5712819241417824`}], 
                   ArrowBox[{{-0.18126276473421002`, 
                    2.1999065452673032`}, {-0.16426309134586278`, 
                    2.199999799565938}}]}, 
                  {RGBColor[0, 0, 1], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-0.16343757588860044`, 0.5}, 1.8, {
                    1.6000297124703746`, 1.526553409230639}], 
                   ArrowBox[{{-0.10181236930813602`, 
                    2.2989447834533205`}, {-0.08382630253088619, 
                    2.298238595168718}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.003], Arrowheads[0.025], 
                   CircleBox[{-0.16343757588860044`, 0.5}, 1.5, {
                    1.6000297124703746`, 1.6225976972121035`}], 
                   ArrowBox[{{-0.226121372655824, 
                    1.9986896748903173`}, {-0.2411048854606434, 
                    1.9979879135105998`}}]}, 
                  {RGBColor[0, 0, 1], ArrowBox[{{0, 2.8}, {0, 2.8}}]}},
                 Frame->True,
                 FrameTicks->Automatic,
                 ImageSize->800,
                 PlotRange->{{-3, 3}, {-1, 3}}], 181->
                GraphicsBox[{InsetBox[
                   TagBox[GridBox[{
                    {"\<\"t\"\>", "9.`"},
                    {"\<\"driveIntent\"\>", "0"},
                    {"\<\"motorTorque\"\>", "10.45087178329373`"},
                    {"\<\"vw\"\>", "0.07819922139856622`"},
                    {"\<\"controllerState\"\>", 
                    RowBox[{"\[LeftAssociation]", 
                    RowBox[{
                    
                    RowBox[{"\<\"lastError\"\>", "\[Rule]", 
                    "0.00016664969731856535`"}], ",", 
                    
                    RowBox[{"\<\"lastInt\"\>", "\[Rule]", 
                    "0.10420536846267094`"}], ",", 
                    RowBox[{"\<\"initialized\"\>", "\[Rule]", "True"}]}], 
                    "\[RightAssociation]"}]}
                    },
                    AutoDelete->False,
                    
                    GridBoxItemSize->{
                    "Columns" -> {{Automatic}}, "Rows" -> {{Automatic}}}],
                    "Grid"], {-1, -0.5}], 
                  {GrayLevel[0], LineBox[{{-5, 0}, {5, 0}}]}, 
                  {GrayLevel[0.5], 
                   CircleBox[{-0.15951129436827338`, 0.5}, 0.5], 
                   DiskBox[{-0.15951129436827338`, 0.5}, 0.1]}, 
                  {GrayLevel[0.5], 
                   DiskBox[{-0.1947903761070961, 1.9995850714086432`}, 0.1]}, 
                  {GrayLevel[0.5], Thickness[0.01], 
                   LineBox[{{-0.15951129436827338`, 
                    0.5}, {-0.1947903761070961, 1.9995850714086432`}}]}, 
                  {RGBColor[0, 1, 0], 
                   DiskBox[{-0.1830306821941552, 1.4997233809390953`}, 0.05]}, 
                  {RGBColor[0, 0, 1], Arrowheads[0.01], 
                   CircleBox[{-0.15951129436827338`, 0.5}, 0.3333333333333333,
                     NCache[{
                    Rational[-1, 2] 
                    Pi, -1.2572701732960847`}, {-1.5707963267948966`, \
-1.2572701732960847`}]], 
                   ArrowBox[{{-0.05988228584406574, 
                    0.18190383251036413`}, {-0.056706358593846484`, 
                    0.1829160106666894}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-0.15951129436827338`, 
                    0.5}, {-0.368528730034148, 0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-0.15951129436827338`, 0.5}, {
                    0.013425658204853669`, 0.5}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.005], Arrowheads[0.04], 
                   ArrowBox[{{-0.15951129436827338`, 
                    0.5}, {-0.1955917774610209, 0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-0.15951129436827338`, 0.5}, 1.6, {
                    1.5943178834977525`, 1.7024333856646356`}], 
                   ArrowBox[{{-0.3536510261030169, 
                    2.0881781274661737`}, {-0.3695228357769521, 
                    2.086157354260652}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Dashing[{Small, Small}], 
                   Arrowheads[0.02], 
                   CircleBox[{-0.15951129436827338`, 0.5}, 1.7, {
                    1.5943178834977525`, 1.5734136346996916`}], 
                   ArrowBox[{{-0.18096014869891264`, 
                    2.199864684805206}, {-0.16396071272644527`, 
                    2.199994177247756}}]}, 
                  {RGBColor[0, 0, 1], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-0.15951129436827338`, 0.5}, 1.8, {
                    1.5943178834977525`, 1.5362574847016761`}], 
                   ArrowBox[{{-0.11534581132126959`, 
                    2.299458087899583}, {-0.09735373860620564, 
                    2.298926468275367}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.003], Arrowheads[0.025], 
                   CircleBox[{-0.15951129436827338`, 0.5}, 1.5, {
                    1.5943178834977525`, 1.6162771252826673`}], 
                   ArrowBox[{{-0.212721326223246, 
                    1.9990559337496359`}, {-0.22770897523991396`, 
                    1.9984488901272977`}}]}, 
                  {RGBColor[0, 0, 1], ArrowBox[{{0, 2.8}, {0, 2.8}}]}},
                 Frame->True,
                 FrameTicks->Automatic,
                 ImageSize->800,
                 PlotRange->{{-3, 3}, {-1, 3}}], 182->
                GraphicsBox[{InsetBox[
                   TagBox[GridBox[{
                    {"\<\"t\"\>", "9.05`"},
                    {"\<\"driveIntent\"\>", "0"},
                    {"\<\"motorTorque\"\>", "7.969334093139272`"},
                    {"\<\"vw\"\>", "0.0777049444426959`"},
                    {"\<\"controllerState\"\>", 
                    RowBox[{"\[LeftAssociation]", 
                    RowBox[{
                    RowBox[{"\<\"lastError\"\>", "\[Rule]", 
                    RowBox[{"-", "0.004821944291480713`"}]}], ",", 
                    RowBox[{"\<\"lastInt\"\>", "\[Rule]", 
                    "0.10408433731899218`"}], ",", 
                    RowBox[{"\<\"initialized\"\>", "\[Rule]", "True"}]}], 
                    "\[RightAssociation]"}]}
                    },
                    AutoDelete->False,
                    
                    GridBoxItemSize->{
                    "Columns" -> {{Automatic}}, "Rows" -> {{Automatic}}}],
                    "Grid"], {-1, -0.5}], 
                  {GrayLevel[0], LineBox[{{-5, 0}, {5, 0}}]}, 
                  {GrayLevel[0.5], 
                   CircleBox[{-0.15561475678649672`, 0.5}, 0.5], 
                   DiskBox[{-0.15561475678649672`, 0.5}, 0.1]}, 
                  {GrayLevel[0.5], 
                   DiskBox[{-0.18320573326006012`, 1.9997462245384168`}, 
                    0.1]}, 
                  {GrayLevel[0.5], Thickness[0.01], 
                   LineBox[{{-0.15561475678649672`, 
                    0.5}, {-0.18320573326006012`, 1.9997462245384168`}}]}, 
                  {RGBColor[0, 1, 0], 
                   DiskBox[{-0.17400874110220566`, 1.4998308163589447`}, 
                    0.05]}, 
                  {RGBColor[0, 0, 1], Arrowheads[0.01], 
                   CircleBox[{-0.15561475678649672`, 0.5}, 0.3333333333333333,
                     NCache[{
                    Rational[-1, 2] 
                    Pi, -1.3317163040007185`}, {-1.5707963267948966`, \
-1.3317163040007185`}]], 
                   ArrowBox[{{-0.07992086468757585, 
                    0.17537476097809646`}, {-0.07667845106435872, 
                    0.17614791841019106`}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-0.15561475678649672`, 
                    0.5}, {-0.3150014386492822, 0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-0.15561475678649672`, 
                    0.5}, {-0.02035012050528251, 0.5}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.005], Arrowheads[0.04], 
                   ArrowBox[{{-0.15561475678649672`, 
                    0.5}, {-0.17973680236806797`, 0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-0.15561475678649672`, 0.5}, 1.6, {
                    1.589191348501223, 1.673746051449495}], 
                   ArrowBox[{{-0.30412026119997304`, 
                    2.093093253754751}, {-0.3200435030099622, 
                    2.0915285694625085`}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Dashing[{Small, Small}], 
                   Arrowheads[0.02], 
                   CircleBox[{-0.15561475678649672`, 0.5}, 1.7, {
                    1.589191348501223, 1.5752140236332643`}], 
                   ArrowBox[{{-0.18012399226873868`, 
                    2.1998233135758776`}, {-0.163124816983887, 
                    2.199983411388426}}]}, 
                  {RGBColor[0, 0, 1], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-0.15561475678649672`, 0.5}, 1.8, {
                    1.589191348501223, 1.5449172702060048`}], 
                   ArrowBox[{{-0.12703365605608902`, 
                    2.299773074773883}, {-0.10903765431215726`, 
                    2.299397280626237}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.003], Arrowheads[0.025], 
                   CircleBox[{-0.15561475678649672`, 0.5}, 1.5, {
                    1.589191348501223, 1.6104131695080983`}], 
                   ArrowBox[{{-0.20003352648315265`, 
                    1.999342180057186}, {-0.21502447747463355`, 
                    1.9988230332790318`}}]}, 
                  {RGBColor[0, 0, 1], ArrowBox[{{0, 2.8}, {0, 2.8}}]}},
                 Frame->True,
                 FrameTicks->Automatic,
                 ImageSize->800,
                 PlotRange->{{-3, 3}, {-1, 3}}], 183->
                GraphicsBox[{InsetBox[
                   TagBox[GridBox[{
                    {"\<\"t\"\>", "9.1`"},
                    {"\<\"driveIntent\"\>", "0"},
                    {"\<\"motorTorque\"\>", "5.7708609022899235`"},
                    {"\<\"vw\"\>", "0.07739482905956674`"},
                    {"\<\"controllerState\"\>", 
                    RowBox[{"\[LeftAssociation]", 
                    RowBox[{
                    RowBox[{"\<\"lastError\"\>", "\[Rule]", 
                    RowBox[{"-", "0.009299843053654175`"}]}], ",", 
                    
                    RowBox[{"\<\"lastInt\"\>", "\[Rule]", 
                    "0.10372695591935634`"}], ",", 
                    RowBox[{"\<\"initialized\"\>", "\[Rule]", "True"}]}], 
                    "\[RightAssociation]"}]}
                    },
                    AutoDelete->False,
                    
                    GridBoxItemSize->{
                    "Columns" -> {{Automatic}}, "Rows" -> {{Automatic}}}],
                    "Grid"], {-1, -0.5}], 
                  {GrayLevel[0], LineBox[{{-5, 0}, {5, 0}}]}, 
                  {GrayLevel[0.5], 
                   CircleBox[{-0.15173813284161572`, 0.5}, 0.5], 
                   DiskBox[{-0.15173813284161572`, 0.5}, 0.1]}, 
                  {GrayLevel[0.5], 
                   DiskBox[{-0.1724887072785066, 1.9998564643526857`}, 0.1]}, 
                  {GrayLevel[0.5], Thickness[0.01], 
                   LineBox[{{-0.15173813284161572`, 
                    0.5}, {-0.1724887072785066, 1.9998564643526857`}}]}, 
                  {RGBColor[0, 1, 0], 
                   DiskBox[{-0.16557184913287631`, 1.499904309568457}, 0.05]}, 
                  {RGBColor[0, 0, 1], Arrowheads[0.01], 
                   CircleBox[{-0.15173813284161572`, 0.5}, 0.3333333333333333,
                     NCache[{
                    Rational[-1, 2] 
                    Pi, -1.3976704997261988`}, {-1.5707963267948966`, \
-1.3976704997261988`}]], 
                   ArrowBox[{{-0.09760369116755849, 
                    0.1710918466566248}, {-0.09431737115140396, 
                    0.17164962732162614`}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-0.15173813284161572`, 
                    0.5}, {-0.2671553508874142, 0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-0.15173813284161572`, 
                    0.5}, {-0.05000105863704306, 0.5}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.005], Arrowheads[0.04], 
                   ArrowBox[{{-0.15173813284161572`, 
                    0.5}, {-0.16541827668284154`, 0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-0.15173813284161572`, 0.5}, 1.6, {
                    1.5846304843544745`, 1.6482222408549545`}], 
                   ArrowBox[{{-0.25953787116383087`, 
                    2.096364374576701}, {-0.2754958589081994, 
                    2.095206577606433}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Dashing[{Small, Small}], 
                   Arrowheads[0.02], 
                   CircleBox[{-0.15173813284161572`, 0.5}, 1.7, {
                    1.5846304843544745`, 1.5767030396927106`}], 
                   ArrowBox[{{-0.17877840443014262`, 
                    2.1997849345468437`}, {-0.1617794863784341, 
                    2.1999703442175536`}}]}, 
                  {RGBColor[0, 0, 1], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-0.15173813284161572`, 0.5}, 1.8, {
                    1.5846304843544745`, 1.5525701460084194`}], 
                   ArrowBox[{{-0.136931174425212, 
                    2.299939097298143}, {-0.11893282378233141`, 
                    2.2997010339769006`}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.003], Arrowheads[0.025], 
                   CircleBox[{-0.15173813284161572`, 0.5}, 1.5, {
                    1.5846304843544745`, 1.6050048479897214`}], 
                   ArrowBox[{{-0.1880473678717015, 
                    1.999560482091846}, {-0.2030409073204999, 
                    1.9991224183937644`}}]}, 
                  {RGBColor[0, 0, 1], ArrowBox[{{0, 2.8}, {0, 2.8}}]}},
                 Frame->True,
                 FrameTicks->Automatic,
                 ImageSize->800,
                 PlotRange->{{-3, 3}, {-1, 3}}], 184->
                GraphicsBox[{InsetBox[
                   TagBox[GridBox[{
                    {"\<\"t\"\>", "9.15`"},
                    {"\<\"driveIntent\"\>", "0"},
                    {"\<\"motorTorque\"\>", "3.843669163380138`"},
                    {"\<\"vw\"\>", "0.0772445224414781`"},
                    {"\<\"controllerState\"\>", 
                    RowBox[{"\[LeftAssociation]", 
                    RowBox[{
                    RowBox[{"\<\"lastError\"\>", "\[Rule]", 
                    RowBox[{"-", "0.013282237571877244`"}]}], ",", 
                    
                    RowBox[{"\<\"lastInt\"\>", "\[Rule]", 
                    "0.10315838450775532`"}], ",", 
                    RowBox[{"\<\"initialized\"\>", "\[Rule]", "True"}]}], 
                    "\[RightAssociation]"}]}
                    },
                    AutoDelete->False,
                    
                    GridBoxItemSize->{
                    "Columns" -> {{Automatic}}, "Rows" -> {{Automatic}}}],
                    "Grid"], {-1, -0.5}], 
                  {GrayLevel[0], LineBox[{{-5, 0}, {5, 0}}]}, 
                  {GrayLevel[0.5], 
                   CircleBox[{-0.14787284081326285`, 0.5}, 0.5], 
                   DiskBox[{-0.14787284081326285`, 0.5}, 0.1]}, 
                  {GrayLevel[0.5], 
                   DiskBox[{-0.1625972130464497, 1.9999277292130908`}, 0.1]}, 
                  {GrayLevel[0.5], Thickness[0.01], 
                   LineBox[{{-0.14787284081326285`, 
                    0.5}, {-0.1625972130464497, 1.9999277292130908`}}]}, 
                  {RGBColor[0, 1, 0], 
                   DiskBox[{-0.15768908896872075`, 1.4999518194753938`}, 
                    0.05]}, 
                  {RGBColor[0, 0, 1], Arrowheads[0.01], 
                   CircleBox[{-0.14787284081326285`, 0.5}, 0.3333333333333333,
                     NCache[{
                    Rational[-1, 2] 
                    Pi, -1.4554862518934923`}, {-1.5707963267948966`, \
-1.4554862518934923`}]], 
                   ArrowBox[{{-0.11283433049030808`, 
                    0.16851332771066152`}, {-0.10952127092583398`, 
                    0.16888028116966403`}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-0.14787284081326285`, 
                    0.5}, {-0.22474622408086561`, 0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-0.14787284081326285`, 
                    0.5}, {-0.07567793688891561, 0.5}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.005], Arrowheads[0.04], 
                   ArrowBox[{{-0.14787284081326285`, 
                    0.5}, {-0.15255132015651837`, 0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-0.14787284081326285`, 0.5}, 1.6, {
                    1.5806127326040547`, 1.6257367216542364`}], 
                   ArrowBox[{{-0.2197532714646971, 
                    2.0983845606390736`}, {-0.23573325668341055`, 
                    2.097585849750531}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Dashing[{Small, Small}], 
                   Arrowheads[0.02], 
                   CircleBox[{-0.14787284081326285`, 0.5}, 1.7, {
                    1.5806127326040547`, 1.5779014901034096`}], 
                   ArrowBox[{{-0.17695020044830925`, 
                    2.1997513074437256`}, {-0.1599515168084739, 
                    2.1999570893367286`}}]}, 
                  {RGBColor[0, 0, 1], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-0.14787284081326285`, 0.5}, 1.8, {
                    1.5806127326040547`, 1.5592590150297205`}], 
                   ArrowBox[{{-0.1451056807258971, 
                    2.299997873005702}, {-0.12710614035083698`, 
                    2.2998802027223655`}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.003], Arrowheads[0.025], 
                   CircleBox[{-0.14787284081326285`, 0.5}, 1.5, {
                    1.5806127326040547`, 1.6000477850009833`}], 
                   ArrowBox[{{-0.1767482444181176, 
                    1.9997220446024846`}, {-0.19174377115356928`, 
                    1.9993583099016312`}}]}, 
                  {RGBColor[0, 0, 1], ArrowBox[{{0, 2.8}, {0, 2.8}}]}},
                 Frame->True,
                 FrameTicks->Automatic,
                 ImageSize->800,
                 PlotRange->{{-3, 3}, {-1, 3}}], 185->
                GraphicsBox[{InsetBox[
                   TagBox[GridBox[{
                    {"\<\"t\"\>", "9.200000000000001`"},
                    {"\<\"driveIntent\"\>", "0"},
                    {"\<\"motorTorque\"\>", "2.1747260807787754`"},
                    {"\<\"vw\"\>", "0.07723099171394208`"},
                    {"\<\"controllerState\"\>", 
                    RowBox[{"\[LeftAssociation]", 
                    RowBox[{
                    RowBox[{"\<\"lastError\"\>", "\[Rule]", 
                    RowBox[{"-", "0.016787174816530366`"}]}], ",", 
                    
                    RowBox[{"\<\"lastInt\"\>", "\[Rule]", 
                    "0.10240294904339162`"}], ",", 
                    RowBox[{"\<\"initialized\"\>", "\[Rule]", "True"}]}], 
                    "\[RightAssociation]"}]}
                    },
                    AutoDelete->False,
                    
                    GridBoxItemSize->{
                    "Columns" -> {{Automatic}}, "Rows" -> {{Automatic}}}],
                    "Grid"], {-1, -0.5}], 
                  {GrayLevel[0], LineBox[{{-5, 0}, {5, 0}}]}, 
                  {GrayLevel[0.5], 
                   CircleBox[{-0.14401148322725588`, 0.5}, 0.5], 
                   DiskBox[{-0.14401148322725588`, 0.5}, 0.1]}, 
                  {GrayLevel[0.5], 
                   DiskBox[{-0.15348655594313704`, 1.9999700740338218`}, 
                    0.1]}, 
                  {GrayLevel[0.5], Thickness[0.01], 
                   LineBox[{{-0.14401148322725588`, 
                    0.5}, {-0.15348655594313704`, 1.9999700740338218`}}]}, 
                  {RGBColor[0, 1, 0], 
                   DiskBox[{-0.15032819837117664`, 1.499980049355881}, 0.05]}, 
                  {RGBColor[0, 0, 1], Arrowheads[0.01], 
                   CircleBox[{-0.14401148322725588`, 0.5}, 0.3333333333333333,
                     NCache[{
                    Rational[-1, 2] 
                    Pi, -1.5055545443715332`}, {-1.5707963267948966`, \
-1.5055545443715332`}]], 
                   ArrowBox[{{-0.12560691981474414`, 
                    0.16717514642577258`}, {-0.12227964697003553`, 
                    0.16737583009648793`}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-0.14401148322725588`, 
                    0.5}, {-0.1875060048428314, 0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-0.14401148322725588`, 
                    0.5}, {-0.09755304919683366, 0.5}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.005], Arrowheads[0.04], 
                   ArrowBox[{{-0.14401148322725588`, 
                    0.5}, {-0.14104757081240915`, 0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-0.14401148322725588`, 0.5}, 1.6, {
                    1.5771130839466645`, 1.6061501845245383`}], 
                   ArrowBox[{{-0.18457330962271543`, 
                    2.099485773065664}, {-0.20056587269932374`, 
                    2.099000187939777}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Dashing[{Small, Small}], 
                   Arrowheads[0.02], 
                   CircleBox[{-0.14401148322725588`, 0.5}, 1.7, {
                    1.5771130839466645`, 1.5788307611254988`}], 
                   ArrowBox[{{-0.14067002374087326`, 
                    2.1999967160699168`}, {-0.15766987464178514`, 
                    2.1999451309803995`}}]}, 
                  {RGBColor[0, 0, 1], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-0.14401148322725588`, 0.5}, 1.8, {
                    1.5771130839466645`, 1.5650312723867825`}], 
                   ArrowBox[{{-0.1516343625068458, 
                    2.2999838587363746`}, {-0.13363444277450406`, 
                    2.299970087815751}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.003], Arrowheads[0.025], 
                   CircleBox[{-0.14401148322725588`, 0.5}, 1.5, {
                    1.5771130839466645`, 1.5955345181605503`}], 
                   ArrowBox[{{-0.16611796994800457`, 
                    1.9998370922353086`}, {-0.1811149855836338, 
                    1.9995410398228153`}}]}, 
                  {RGBColor[0, 0, 1], ArrowBox[{{0, 2.8}, {0, 2.8}}]}},
                 Frame->True,
                 FrameTicks->Automatic,
                 ImageSize->800,
                 PlotRange->{{-3, 3}, {-1, 3}}], 186->
                GraphicsBox[{InsetBox[
                   TagBox[GridBox[{
                    {"\<\"t\"\>", "9.25`"},
                    {"\<\"driveIntent\"\>", "0"},
                    {"\<\"motorTorque\"\>", "0.7500087434717635`"},
                    {"\<\"vw\"\>", "0.07733258035290126`"},
                    {"\<\"controllerState\"\>", 
                    RowBox[{"\[LeftAssociation]", 
                    RowBox[{
                    RowBox[{"\<\"lastError\"\>", "\[Rule]", 
                    RowBox[{"-", "0.01983510682834784`"}]}], ",", 
                    
                    RowBox[{"\<\"lastInt\"\>", "\[Rule]", 
                    "0.10148400996186008`"}], ",", 
                    RowBox[{"\<\"initialized\"\>", "\[Rule]", "True"}]}], 
                    "\[RightAssociation]"}]}
                    },
                    AutoDelete->False,
                    
                    GridBoxItemSize->{
                    "Columns" -> {{Automatic}}, "Rows" -> {{Automatic}}}],
                    "Grid"], {-1, -0.5}], 
                  {GrayLevel[0], LineBox[{{-5, 0}, {5, 0}}]}, 
                  {GrayLevel[0.5], 
                   CircleBox[{-0.14014777936165662`, 0.5}, 0.5], 
                   DiskBox[{-0.14014777936165662`, 0.5}, 0.1]}, 
                  {GrayLevel[0.5], 
                   DiskBox[{-0.14511009877629663`, 1.999991791772884}, 0.1]}, 
                  {GrayLevel[0.5], Thickness[0.01], 
                   LineBox[{{-0.14014777936165662`, 
                    0.5}, {-0.14511009877629663`, 1.999991791772884}}]}, 
                  {RGBColor[0, 1, 0], 
                   DiskBox[{-0.14345599230474998`, 1.4999945278485893`}, 
                    0.05]}, 
                  {RGBColor[0, 0, 1], Arrowheads[0.01], 
                   CircleBox[{-0.14014777936165662`, 0.5}, 0.3333333333333333,
                     NCache[{
                    Rational[-1, 2] 
                    Pi, -1.5482960644907438`}, {-1.5707963267948966`, \
-1.5482960644907438`}]], 
                   ArrowBox[{{-0.13598113377319995`, 
                    0.16669270908716788`}, {-0.13264832474555235`, 
                    0.1667510400742831}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-0.14014777936165662`, 
                    0.5}, {-0.1551479542310919, 0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-0.14014777936165662`, 
                    0.5}, {-0.11581604766551537`, 0.5}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.005], Arrowheads[0.04], 
                   ArrowBox[{{-0.14014777936165662`, 
                    0.5}, {-0.13081622253495065`, 0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-0.14014777936165662`, 0.5}, 1.6, {
                    1.5741045457723506`, 1.589311961299719}], 
                   ArrowBox[{{-0.15377262989796536`, 
                    2.099941987525755}, {-0.16977110188070849`, 
                    2.0997257448584534`}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Dashing[{Small, Small}], 
                   Arrowheads[0.02], 
                   CircleBox[{-0.14014777936165662`, 0.5}, 1.7, {
                    1.5741045457723506`, 1.5795125461380766`}], 
                   ArrowBox[{{-0.1379653528445362, 
                    2.1999985991213338`}, {-0.15496516462448287`, 
                    2.1999354238011435`}}]}, 
                  {RGBColor[0, 0, 1], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-0.14014777936165662`, 0.5}, 1.8, {
                    1.5741045457723506`, 1.5699378305308407`}], 
                   ArrowBox[{{-0.1566022569086524, 
                    2.299924790142259}, {-0.13860248627617353`, 
                    2.299999336685789}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.003], Arrowheads[0.025], 
                   CircleBox[{-0.14014777936165662`, 0.5}, 1.5, {
                    1.5741045457723506`, 1.5914547977928883`}], 
                   ArrowBox[{{-0.1561351831517827, 
                    1.9999147985535883`}, {-0.17113328178924067`, 
                    1.9996799320652756`}}]}, 
                  {RGBColor[0, 0, 1], ArrowBox[{{0, 2.8}, {0, 2.8}}]}},
                 Frame->True,
                 FrameTicks->Automatic,
                 ImageSize->800,
                 PlotRange->{{-3, 3}, {-1, 3}}], 187->
                GraphicsBox[{InsetBox[
                   TagBox[GridBox[{
                    {"\<\"t\"\>", "9.3`"},
                    {"\<\"driveIntent\"\>", "0"},
                    {"\<\"motorTorque\"\>", 
                    RowBox[{"-", "0.4452519381343194`"}]},
                    {"\<\"vw\"\>", "0.07752905024512208`"},
                    {"\<\"controllerState\"\>", 
                    RowBox[{"\[LeftAssociation]", 
                    RowBox[{
                    RowBox[{"\<\"lastError\"\>", "\[Rule]", 
                    RowBox[{"-", "0.022448456808195728`"}]}], ",", 
                    
                    RowBox[{"\<\"lastInt\"\>", "\[Rule]", 
                    "0.1004238530656267`"}], ",", 
                    RowBox[{"\<\"initialized\"\>", "\[Rule]", "True"}]}], 
                    "\[RightAssociation]"}]}
                    },
                    AutoDelete->False,
                    
                    GridBoxItemSize->{
                    "Columns" -> {{Automatic}}, "Rows" -> {{Automatic}}}],
                    "Grid"], {-1, -0.5}], 
                  {GrayLevel[0], LineBox[{{-5, 0}, {5, 0}}]}, 
                  {GrayLevel[0.5], 
                   CircleBox[{-0.13627649529601604`, 0.5}, 0.5], 
                   DiskBox[{-0.13627649529601604`, 0.5}, 0.1]}, 
                  {GrayLevel[0.5], 
                   DiskBox[{-0.13741988746019612`, 1.9999995642180564`}, 
                    0.1]}, 
                  {GrayLevel[0.5], Thickness[0.01], 
                   LineBox[{{-0.13627649529601604`, 
                    0.5}, {-0.13741988746019612`, 1.9999995642180564`}}]}, 
                  {RGBColor[0, 1, 0], 
                   DiskBox[{-0.13703875673880275`, 1.499999709478704}, 0.05]}, 
                  {RGBColor[0, 0, 1], Arrowheads[0.01], 
                   CircleBox[{-0.13627649529601604`, 0.5}, 0.3333333333333333,
                     NCache[{
                    Rational[-1, 2] 
                    Pi, -1.584153884938926}, {-1.5707963267948966`, \
-1.584153884938926}]], 
                   ArrowBox[{{-0.1373956792412263, 
                    0.16666854553101673`}, {-0.14072888227222055`, 
                    0.16669640361777488`}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-0.13627649529601604`, 
                    0.5}, {-0.12737145653332965`, 0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-0.13627649529601604`, 
                    0.5}, {-0.13067007354137322`, 0.5}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.005], Arrowheads[0.04], 
                   ArrowBox[{{-0.13627649529601604`, 
                    0.5}, {-0.12176503477868682`, 0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-0.13627649529601604`, 0.5}, 1.6, {
                    1.571558588311501, 1.5750626029261436`}], 
                   ArrowBox[{{-0.12710258737247573`, 
                    2.099973699600531}, {-0.14310251639910485`, 
                    2.099985439132463}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Dashing[{Small, Small}], 
                   Arrowheads[0.02], 
                   CircleBox[{-0.13627649529601604`, 0.5}, 1.7, {
                    1.571558588311501, 1.5799685890115793`}], 
                   ArrowBox[{{-0.13486934122506186`, 
                    2.1999994176226707`}, {-0.15186912242658623`, 
                    2.1999284896663056`}}]}, 
                  {RGBColor[0, 0, 1], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-0.13627649529601604`, 0.5}, 1.8, {
                    1.571558588311501, 1.574032210190025}], 
                   ArrowBox[{{-0.12410117825117778`, 
                    2.299958822210902}, {-0.14210107524242896`, 
                    2.2999905761610107`}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.003], Arrowheads[0.025], 
                   CircleBox[{-0.13627649529601604`, 0.5}, 1.5, {
                    1.571558588311501, 1.5877958842814375`}], 
                   ArrowBox[{{-0.14677574579229868`, 
                    1.9999632547962687`}, {-0.1617746033894854, 
                    1.9997832665034152`}}]}, 
                  {RGBColor[0, 0, 1], ArrowBox[{{0, 2.8}, {0, 2.8}}]}},
                 Frame->True,
                 FrameTicks->Automatic,
                 ImageSize->800,
                 PlotRange->{{-3, 3}, {-1, 3}}], 188->
                GraphicsBox[{InsetBox[
                   TagBox[GridBox[{
                    {"\<\"t\"\>", "9.35`"},
                    {"\<\"driveIntent\"\>", "0"},
                    {"\<\"motorTorque\"\>", 
                    RowBox[{"-", "1.4263449555033887`"}]},
                    {"\<\"vw\"\>", "0.07780160912734417`"},
                    {"\<\"controllerState\"\>", 
                    RowBox[{"\[LeftAssociation]", 
                    RowBox[{
                    RowBox[{"\<\"lastError\"\>", "\[Rule]", 
                    RowBox[{"-", "0.024651206097252246`"}]}], ",", 
                    
                    RowBox[{"\<\"lastInt\"\>", "\[Rule]", 
                    "0.0992436015924454`"}], ",", 
                    RowBox[{"\<\"initialized\"\>", "\[Rule]", "True"}]}], 
                    "\[RightAssociation]"}]}
                    },
                    AutoDelete->False,
                    
                    GridBoxItemSize->{
                    "Columns" -> {{Automatic}}, "Rows" -> {{Automatic}}}],
                    "Grid"], {-1, -0.5}], 
                  {GrayLevel[0], LineBox[{{-5, 0}, {5, 0}}]}, 
                  {GrayLevel[0.5], CircleBox[{-0.132393372230276, 0.5}, 0.5], 
                   DiskBox[{-0.132393372230276, 0.5}, 0.1]}, 
                  {GrayLevel[0.5], 
                   DiskBox[{-0.1303672308029168, 1.9999986315830145`}, 0.1]}, 
                  {GrayLevel[0.5], Thickness[0.01], 
                   LineBox[{{-0.132393372230276, 0.5}, {-0.1303672308029168, 
                    1.9999986315830145`}}]}, 
                  {RGBColor[0, 1, 0], 
                   DiskBox[{-0.13104261127870317`, 1.4999990877220097`}, 
                    0.05]}, 
                  {RGBColor[0, 0, 1], Arrowheads[0.01], 
                   CircleBox[{-0.132393372230276, 0.5}, 0.3333333333333333, 
                    NCache[{
                    Rational[-1, 2] 
                    Pi, -1.6135866754599981`}, {-1.5707963267948966`, \
-1.6135866754599981`}]], 
                   ArrowBox[{{-0.14332152986775762`, 
                    0.1668458517716439}, {-0.14665246942129895`, 
                    0.1669717890952659}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-0.132393372230276, 
                    0.5}, {-0.10386647312020822`, 0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-0.132393372230276, 0.5}, {-0.1423281930812555, 
                    0.5}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.005], Arrowheads[0.04], 
                   ArrowBox[{{-0.132393372230276, 
                    0.5}, {-0.11380129397118773`, 0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-0.132393372230276, 0.5}, 1.6, {
                    1.5694455654325672`, 1.5632362967361257`}], 
                   ArrowBox[{{-0.13629732026257785`, 
                    2.0999952372397117`}, {-0.1202974393589453, 
                    2.099954276974181}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Dashing[{Small, Small}], 
                   Arrowheads[0.02], 
                   CircleBox[{-0.132393372230276, 0.5}, 1.7, {
                    1.5694455654325672`, 1.5802204486095932`}], 
                   ArrowBox[{{-0.1314143793693717, 
                    2.1999997181096758`}, {-0.1484141421678339, 
                    2.1999245085975456`}}]}, 
                  {RGBColor[0, 0, 1], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-0.132393372230276, 0.5}, 1.8, {
                    1.5694455654325672`, 1.5773697040742527`}], 
                   ArrowBox[{{-0.12622546340346713`, 
                    2.2999894324413974`}, {-0.14422536612401374`, 
                    2.299961111780057}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.003], Arrowheads[0.025], 
                   CircleBox[{-0.132393372230276, 0.5}, 1.5, {
                    1.5694455654325672`, 1.5845428461270152`}], 
                   ArrowBox[{{-0.1380131380815455, 
                    1.9999894727069845`}, {-0.1530125018256689, 
                    1.9998582771364528`}}]}, 
                  {RGBColor[0, 0, 1], ArrowBox[{{0, 2.8}, {0, 2.8}}]}},
                 Frame->True,
                 FrameTicks->Automatic,
                 ImageSize->800,
                 PlotRange->{{-3, 3}, {-1, 3}}], 189->
                GraphicsBox[{InsetBox[
                   TagBox[GridBox[{
                    {"\<\"t\"\>", "9.4`"},
                    {"\<\"driveIntent\"\>", "0"},
                    {"\<\"motorTorque\"\>", 
                    RowBox[{"-", "2.2088718993053362`"}]},
                    {"\<\"vw\"\>", "0.07813292357372198`"},
                    {"\<\"controllerState\"\>", 
                    RowBox[{"\[LeftAssociation]", 
                    RowBox[{
                    RowBox[{"\<\"lastError\"\>", "\[Rule]", 
                    RowBox[{"-", "0.026468505376227434`"}]}], ",", 
                    
                    RowBox[{"\<\"lastInt\"\>", "\[Rule]", 
                    "0.09796314832971476`"}], ",", 
                    RowBox[{"\<\"initialized\"\>", "\[Rule]", "True"}]}], 
                    "\[RightAssociation]"}]}
                    },
                    AutoDelete->False,
                    
                    GridBoxItemSize->{
                    "Columns" -> {{Automatic}}, "Rows" -> {{Automatic}}}],
                    "Grid"], {-1, -0.5}], 
                  {GrayLevel[0], LineBox[{{-5, 0}, {5, 0}}]}, 
                  {GrayLevel[0.5], 
                   CircleBox[{-0.12849505380178042`, 0.5}, 0.5], 
                   DiskBox[{-0.12849505380178042`, 0.5}, 0.1]}, 
                  {GrayLevel[0.5], 
                   DiskBox[{-0.12390323101895892`, 1.9999929717047114`}, 
                    0.1]}, 
                  {GrayLevel[0.5], Thickness[0.01], 
                   LineBox[{{-0.12849505380178042`, 
                    0.5}, {-0.12390323101895892`, 1.9999929717047114`}}]}, 
                  {RGBColor[0, 1, 0], 
                   DiskBox[{-0.12543383861323276`, 1.4999953144698077`}, 
                    0.05]}, 
                  {RGBColor[0, 0, 1], Arrowheads[0.01], 
                   CircleBox[{-0.12849505380178042`, 0.5}, 0.3333333333333333,
                     NCache[{
                    Rational[-1, 2] 
                    Pi, -1.6370624837740566`}, {-1.5707963267948966`, \
-1.6370624837740566`}]], 
                   ArrowBox[{{-0.14724054480985244`, 
                    0.1671941742126839}, {-0.1505676103336255, 
                    0.1673982661511525}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-0.12849505380178042`, 
                    0.5}, {-0.0843176158156737, 0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-0.12849505380178042`, 
                    0.5}, {-0.15101014775271185`, 0.5}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.005], Arrowheads[0.04], 
                   ArrowBox[{{-0.12849505380178042`, 
                    0.5}, {-0.10683270976660514`, 0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-0.12849505380178042`, 0.5}, 1.6, {
                    1.5677351068252012`, 1.5536631071710898`}], 
                   ArrowBox[{{-0.1170819991923355, 
                    2.099959293914843}, {-0.10108324355971113`, 
                    2.0997651679728664`}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Dashing[{Small, Small}], 
                   Arrowheads[0.02], 
                   CircleBox[{-0.12849505380178042`, 0.5}, 1.7, {
                    1.5677351068252012`, 1.5802892883874256`}], 
                   ArrowBox[{{-0.12763308854601307`, 
                    2.19999978147525}, {-0.14463284612678898`, 
                    2.199923401703405}}]}, 
                  {RGBColor[0, 0, 1], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-0.12849505380178042`, 0.5}, 1.8, {
                    1.5677351068252012`, 1.5800066173768974`}], 
                   ArrowBox[{{-0.12707357699713034`, 
                    2.2999994387231606`}, {-0.14507334245920347`, 
                    2.299923654032357}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.003], Arrowheads[0.025], 
                   CircleBox[{-0.12849505380178042`, 0.5}, 1.5, {
                    1.5677351068252012`, 1.5816788589600388`}], 
                   ArrowBox[{{-0.12981885187765074`, 
                    1.9999994158527712`}, {-0.14481852984817362`, 
                    1.9999111787468493`}}]}, 
                  {RGBColor[0, 0, 1], ArrowBox[{{0, 2.8}, {0, 2.8}}]}},
                 Frame->True,
                 FrameTicks->Automatic,
                 ImageSize->800,
                 PlotRange->{{-3, 3}, {-1, 3}}], 190->
                GraphicsBox[{InsetBox[
                   TagBox[GridBox[{
                    {"\<\"t\"\>", "9.450000000000001`"},
                    {"\<\"driveIntent\"\>", "0"},
                    {"\<\"motorTorque\"\>", 
                    RowBox[{"-", "2.8085593523754717`"}]},
                    {"\<\"vw\"\>", "0.07850711808368693`"},
                    {"\<\"controllerState\"\>", 
                    RowBox[{"\[LeftAssociation]", 
                    RowBox[{
                    RowBox[{"\<\"lastError\"\>", "\[Rule]", 
                    RowBox[{"-", "0.027926312745853086`"}]}], ",", 
                    
                    RowBox[{"\<\"lastInt\"\>", "\[Rule]", 
                    "0.09660110649301774`"}], ",", 
                    RowBox[{"\<\"initialized\"\>", "\[Rule]", "True"}]}], 
                    "\[RightAssociation]"}]}
                    },
                    AutoDelete->False,
                    
                    GridBoxItemSize->{
                    "Columns" -> {{Automatic}}, "Rows" -> {{Automatic}}}],
                    "Grid"], {-1, -0.5}], 
                  {GrayLevel[0], LineBox[{{-5, 0}, {5, 0}}]}, 
                  {GrayLevel[0.5], 
                   CircleBox[{-0.12457901311036622`, 0.5}, 0.5], 
                   DiskBox[{-0.12457901311036622`, 0.5}, 0.1]}, 
                  {GrayLevel[0.5], 
                   DiskBox[{-0.1179792628411248, 1.9999854810285278`}, 0.1]}, 
                  {GrayLevel[0.5], Thickness[0.01], 
                   LineBox[{{-0.12457901311036622`, 
                    0.5}, {-0.1179792628411248, 1.9999854810285278`}}]}, 
                  {RGBColor[0, 1, 0], 
                   DiskBox[{-0.1201791795975386, 1.499990320685685}, 0.05]}, 
                  {RGBColor[0, 0, 1], Arrowheads[0.01], 
                   CircleBox[{-0.12457901311036622`, 0.5}, 0.3333333333333333,
                     NCache[{
                    Rational[-1, 2] 
                    Pi, -1.6550531073661607`}, {-1.5707963267948966`, \
-1.6550531073661607`}]], 
                   ArrowBox[{{-0.14930853195789337`, 
                    0.16758525602993946`}, {-0.1526313875297752, 
                    0.1678491676955417}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-0.12457901311036622`, 
                    0.5}, {-0.06840782606285678, 0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-0.12457901311036622`, 
                    0.5}, {-0.15693942036970934`, 0.5}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.005], Arrowheads[0.04], 
                   ArrowBox[{{-0.12457901311036622`, 
                    0.5}, {-0.10076823332219992`, 0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-0.12457901311036622`, 0.5}, 1.6, {
                    1.5663964790862235`, 1.5461710287806434`}], 
                   ArrowBox[{{-0.1011793705030716, 
                    2.099828883576569}, {-0.08518251827650052, 
                    2.0995149002728315`}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Dashing[{Small, Small}], 
                   Arrowheads[0.02], 
                   CircleBox[{-0.12457901311036622`, 0.5}, 1.7, {
                    1.5663964790862235`, 1.5801956933471102`}], 
                   ArrowBox[{{-0.12355793631052331`, 
                    2.1999996933535515`}, {-0.14055770096560782`, 
                    2.1999249046750347`}}]}, 
                  {RGBColor[0, 0, 1], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-0.12457901311036622`, 0.5}, 1.8, {
                    1.5663964790862235`, 1.5819995865994205`}], 
                   ArrowBox[{{-0.12674488023587308`, 
                    2.299998696949416}, {-0.14474445891463059`, 
                    2.299887039454286}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.003], Arrowheads[0.025], 
                   CircleBox[{-0.12457901311036622`, 0.5}, 1.5, {
                    1.5663964790862235`, 1.5791855039384461`}], 
                   ArrowBox[{{-0.12216277987061112`, 
                    1.9999980539377147`}, {-0.13716263122221783`, 
                    1.999947216589709}}]}, 
                  {RGBColor[0, 0, 1], ArrowBox[{{0, 2.8}, {0, 2.8}}]}},
                 Frame->True,
                 FrameTicks->Automatic,
                 ImageSize->800,
                 PlotRange->{{-3, 3}, {-1, 3}}], 191->
                GraphicsBox[{InsetBox[
                   TagBox[GridBox[{
                    {"\<\"t\"\>", "9.5`"},
                    {"\<\"driveIntent\"\>", "0"},
                    {"\<\"motorTorque\"\>", 
                    RowBox[{"-", "3.2410911846787407`"}]},
                    {"\<\"vw\"\>", "0.078909761137841`"},
                    {"\<\"controllerState\"\>", 
                    RowBox[{"\[LeftAssociation]", 
                    RowBox[{
                    RowBox[{"\<\"lastError\"\>", "\[Rule]", 
                    RowBox[{"-", "0.029051060637478893`"}]}], ",", 
                    
                    RowBox[{"\<\"lastInt\"\>", "\[Rule]", 
                    "0.09517477797209052`"}], ",", 
                    RowBox[{"\<\"initialized\"\>", "\[Rule]", "True"}]}], 
                    "\[RightAssociation]"}]}
                    },
                    AutoDelete->False,
                    
                    GridBoxItemSize->{
                    "Columns" -> {{Automatic}}, "Rows" -> {{Automatic}}}],
                    "Grid"], {-1, -0.5}], 
                  {GrayLevel[0], LineBox[{{-5, 0}, {5, 0}}]}, 
                  {GrayLevel[0.5], CircleBox[{-0.1206434801255501, 0.5}, 0.5],
                    DiskBox[{-0.1206434801255501, 0.5}, 0.1]}, 
                  {GrayLevel[0.5], 
                   DiskBox[{-0.11254740029393956`, 1.9999781510046606`}, 
                    0.1]}, 
                  {GrayLevel[0.5], Thickness[0.01], 
                   LineBox[{{-0.1206434801255501, 
                    0.5}, {-0.11254740029393956`, 1.9999781510046606`}}]}, 
                  {RGBColor[0, 1, 0], 
                   DiskBox[{-0.11524609357114307`, 1.499985434003107}, 0.05]}, 
                  {RGBColor[0, 0, 1], Arrowheads[0.01], 
                   CircleBox[{-0.1206434801255501, 0.5}, 0.3333333333333333, 
                    NCache[{
                    Rational[-1, 2] 
                    Pi, -1.6680290623352587`}, {-1.5707963267948966`, \
-1.6680290623352587`}]], 
                   ArrowBox[{{-0.14968419478567777`, 
                    0.1679341209893131}, {-0.15300334620811545`, 
                    0.1682411264514097}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-0.1206434801255501, 
                    0.5}, {-0.055821656431975286`, 0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-0.1206434801255501, 0.5}, {-0.160340612529152, 
                    0.5}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.005], Arrowheads[0.04], 
                   ArrowBox[{{-0.1206434801255501, 
                    0.5}, {-0.09551878883557718, 0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-0.1206434801255501, 0.5}, 1.6, {
                    1.5653989140342315`, 1.5405878448840242`}], 
                   ArrowBox[{{-0.08831210976861395, 
                    2.099673304925366}, {-0.0723172598852912, 
                    2.099270013611614}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Dashing[{Small, Small}], 
                   Arrowheads[0.02], 
                   CircleBox[{-0.1206434801255501, 0.5}, 1.7, {
                    1.5653989140342315`, 1.5799595151418813`}], 
                   ArrowBox[{{-0.11922090048145179`, 
                    2.1999994047843536`}, {-0.13622068232586731`, 
                    2.1999286310817903`}}]}, 
                  {RGBColor[0, 0, 1], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-0.1206434801255501, 0.5}, 1.8, {
                    1.5653989140342315`, 1.5834049761713358`}], 
                   ArrowBox[{{-0.12533904367754437`, 
                    2.299993875457061}, {-0.1433384476584146, 
                    2.299856921660353}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.003], Arrowheads[0.025], 
                   CircleBox[{-0.1206434801255501, 0.5}, 1.5, {
                    1.5653989140342315`, 1.5770430627605383`}], 
                   ArrowBox[{{-0.11501359729205236`, 
                    1.9999894347692189`}, {-0.13001352313455133`, 
                    1.9999707338125}}]}, 
                  {RGBColor[0, 0, 1], ArrowBox[{{0, 2.8}, {0, 2.8}}]}},
                 Frame->True,
                 FrameTicks->Automatic,
                 ImageSize->800,
                 PlotRange->{{-3, 3}, {-1, 3}}], 192->
                GraphicsBox[{InsetBox[
                   TagBox[GridBox[{
                    {"\<\"t\"\>", "9.55`"},
                    {"\<\"driveIntent\"\>", "0"},
                    {"\<\"motorTorque\"\>", 
                    RowBox[{"-", "3.5219605282811486`"}]},
                    {"\<\"vw\"\>", "0.07932783933085695`"},
                    {"\<\"controllerState\"\>", 
                    RowBox[{"\[LeftAssociation]", 
                    RowBox[{
                    RowBox[{"\<\"lastError\"\>", "\[Rule]", 
                    RowBox[{"-", "0.029869352794023325`"}]}], ",", 
                    
                    RowBox[{"\<\"lastInt\"\>", "\[Rule]", 
                    "0.0937001374682023`"}], ",", 
                    RowBox[{"\<\"initialized\"\>", "\[Rule]", "True"}]}], 
                    "\[RightAssociation]"}]}
                    },
                    AutoDelete->False,
                    
                    GridBoxItemSize->{
                    "Columns" -> {{Automatic}}, "Rows" -> {{Automatic}}}],
                    "Grid"], {-1, -0.5}], 
                  {GrayLevel[0], LineBox[{{-5, 0}, {5, 0}}]}, 
                  {GrayLevel[0.5], 
                   CircleBox[{-0.11668737009997893`, 0.5}, 0.5], 
                   DiskBox[{-0.11668737009997893`, 0.5}, 0.1]}, 
                  {GrayLevel[0.5], 
                   DiskBox[{-0.10756079125788777`, 1.9999722349292466`}, 
                    0.1]}, 
                  {GrayLevel[0.5], Thickness[0.01], 
                   LineBox[{{-0.11668737009997893`, 
                    0.5}, {-0.10756079125788777`, 1.9999722349292466`}}]}, 
                  {RGBColor[0, 1, 0], 
                   DiskBox[{-0.11060298420525148`, 1.4999814899528312`}, 
                    0.05]}, 
                  {RGBColor[0, 0, 1], Arrowheads[0.01], 
                   CircleBox[{-0.11668737009997893`, 0.5}, 0.3333333333333333,
                     NCache[{
                    Rational[-1, 2] 
                    Pi, -1.676455142643331}, {-1.5707963267948966`, \
-1.676455142643331}]], 
                   ArrowBox[{{-0.1485250344927824, 
                    0.16819060556228627`}, {-0.15184148126591632`, 
                    0.16852556723143186`}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-0.11668737009997893`, 
                    0.5}, {-0.04624815953435596, 0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-0.11668737009997893`, 
                    0.5}, {-0.1614371239654882, 0.5}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.005], Arrowheads[0.04], 
                   ArrowBox[{{-0.11668737009997893`, 
                    0.5}, {-0.09099791339986521, 0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-0.11668737009997893`, 0.5}, 1.6, {
                    1.5647119033591346`, 1.5367427894835741`}], 
                   ArrowBox[{{-0.07820542141957251, 
                    2.0995371641902416`}, {-0.06221224044726443, 
                    2.099072374925325}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Dashing[{Small, Small}], 
                   Arrowheads[0.02], 
                   CircleBox[{-0.11668737009997893`, 0.5}, 1.7, {
                    1.5647119033591346`, 1.5795997449390435`}], 
                   ArrowBox[{{-0.11465318143045673`, 
                    2.1999987829632284`}, {-0.1316529876369611, 
                    2.199934125280076}}]}, 
                  {RGBColor[0, 0, 1], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-0.11668737009997893`, 0.5}, 1.8, {
                    1.5647119033591346`, 1.5842783507384743`}], 
                   ArrowBox[{{-0.12295500053309637`, 
                    2.299989087969356}, {-0.14095427803719845`, 
                    2.299836414005219}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.003], Arrowheads[0.025], 
                   CircleBox[{-0.11668737009997893`, 0.5}, 1.5, {
                    1.5647119033591346`, 1.5752308058709745`}], 
                   ArrowBox[{{-0.1083391318120639, 
                    1.9999767687925998`}, {-0.12333906691354773`, 
                    1.9999852515706613`}}]}, 
                  {RGBColor[0, 0, 1], ArrowBox[{{0, 2.8}, {0, 2.8}}]}},
                 Frame->True,
                 FrameTicks->Automatic,
                 ImageSize->800,
                 PlotRange->{{-3, 3}, {-1, 3}}], 193->
                GraphicsBox[{InsetBox[
                   TagBox[GridBox[{
                    {"\<\"t\"\>", "9.6`"},
                    {"\<\"driveIntent\"\>", "0"},
                    {"\<\"motorTorque\"\>", 
                    RowBox[{"-", "3.666340869866655`"}]},
                    {"\<\"vw\"\>", "0.07974972085755216`"},
                    {"\<\"controllerState\"\>", 
                    RowBox[{"\[LeftAssociation]", 
                    RowBox[{
                    RowBox[{"\<\"lastError\"\>", "\[Rule]", 
                    RowBox[{"-", "0.03040769189630343`"}]}], ",", 
                    
                    RowBox[{"\<\"lastInt\"\>", "\[Rule]", 
                    "0.09219183100212432`"}], ",", 
                    RowBox[{"\<\"initialized\"\>", "\[Rule]", "True"}]}], 
                    "\[RightAssociation]"}]}
                    },
                    AutoDelete->False,
                    
                    GridBoxItemSize->{
                    "Columns" -> {{Automatic}}, "Rows" -> {{Automatic}}}],
                    "Grid"], {-1, -0.5}], 
                  {GrayLevel[0], LineBox[{{-5, 0}, {5, 0}}]}, 
                  {GrayLevel[0.5], CircleBox[{-0.1127102135532993, 0.5}, 0.5],
                    DiskBox[{-0.1127102135532993, 0.5}, 0.1]}, 
                  {GrayLevel[0.5], 
                   DiskBox[{-0.10297398086872547`, 1.9999684015915509`}, 
                    0.1]}, 
                  {GrayLevel[0.5], Thickness[0.01], 
                   LineBox[{{-0.1127102135532993, 
                    0.5}, {-0.10297398086872547`, 1.9999684015915509`}}]}, 
                  {RGBColor[0, 1, 0], 
                   DiskBox[{-0.10621939176358342`, 1.4999789343943672`}, 
                    0.05]}, 
                  {RGBColor[0, 0, 1], Arrowheads[0.01], 
                   CircleBox[{-0.1127102135532993, 0.5}, 0.3333333333333333, 
                    NCache[{
                    Rational[-1, 2] 
                    Pi, -1.6807865528908963`}, {-1.5707963267948966`, \
-1.6807865528908963`}]], 
                   ArrowBox[{{-0.14598477740892213`, 
                    0.16833161966909}, {-0.14929974222011577`, 
                    0.16868094304273518`}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-0.1127102135532993, 0.5}, {-0.0393833961559662,
                     0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-0.1127102135532993, 0.5}, {-0.1604491210112044,
                     0.5}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.005], Arrowheads[0.04], 
                   ArrowBox[{{-0.1127102135532993, 
                    0.5}, {-0.08712230361387129, 0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-0.1127102135532993, 0.5}, 1.6, {
                    1.5643054594270993`, 1.5344680137220443`}], 
                   ArrowBox[{{-0.07058977920456337, 
                    2.099445487977091}, {-0.05459769690187489, 
                    2.0989443190456756`}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Dashing[{Small, Small}], 
                   Arrowheads[0.02], 
                   CircleBox[{-0.1127102135532993, 0.5}, 1.7, {
                    1.5643054594270993`, 1.5791344127541278`}], 
                   ArrowBox[{{-0.109884960984531, 
                    2.199997652336003}, {-0.12688479543781692`, 
                    2.199940905216531}}]}, 
                  {RGBColor[0, 0, 1], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-0.1127102135532993, 0.5}, 1.8, {
                    1.5643054594270993`, 1.5846740198152474`}], 
                   ArrowBox[{{-0.11969004349786076`, 
                    2.2999864671641133`}, {-0.1376892591846684, 
                    2.2998266714546007`}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.003], Arrowheads[0.025], 
                   CircleBox[{-0.1127102135532993, 0.5}, 1.5, {
                    1.5643054594270993`, 1.573727270244784}], 
                   ArrowBox[{{-0.10210671704085646`, 
                    1.9999625214856904`}, {-0.11710662243361727`, 
                    1.9999935571824823`}}]}, 
                  {RGBColor[0, 0, 1], ArrowBox[{{0, 2.8}, {0, 2.8}}]}},
                 Frame->True,
                 FrameTicks->Automatic,
                 ImageSize->800,
                 PlotRange->{{-3, 3}, {-1, 3}}], 194->
                GraphicsBox[{InsetBox[
                   TagBox[GridBox[{
                    {"\<\"t\"\>", "9.65`"},
                    {"\<\"driveIntent\"\>", "0"},
                    {"\<\"motorTorque\"\>", 
                    RowBox[{"-", "3.688975441241926`"}]},
                    {"\<\"vw\"\>", "0.08016510972648282`"},
                    {"\<\"controllerState\"\>", 
                    RowBox[{"\[LeftAssociation]", 
                    RowBox[{
                    RowBox[{"\<\"lastError\"\>", "\[Rule]", 
                    RowBox[{"-", "0.03069223781404347`"}]}], ",", 
                    RowBox[{"\<\"lastInt\"\>", "\[Rule]", 
                    "0.09066318725863504`"}], ",", 
                    RowBox[{"\<\"initialized\"\>", "\[Rule]", "True"}]}], 
                    "\[RightAssociation]"}]}
                    },
                    AutoDelete->False,
                    
                    GridBoxItemSize->{
                    "Columns" -> {{Automatic}}, "Rows" -> {{Automatic}}}],
                    "Grid"], {-1, -0.5}], 
                  {GrayLevel[0], LineBox[{{-5, 0}, {5, 0}}]}, 
                  {GrayLevel[0.5], CircleBox[{-0.1087120883240162, 0.5}, 0.5],
                    DiskBox[{-0.1087120883240162, 0.5}, 0.1]}, 
                  {GrayLevel[0.5], 
                   DiskBox[{-0.09874318554580018, 1.99996687329334}, 0.1]}, 
                  {GrayLevel[0.5], Thickness[0.01], 
                   LineBox[{{-0.1087120883240162, 0.5}, {-0.09874318554580018,
                     1.99996687329334}}]}, 
                  {RGBColor[0, 1, 0], 
                   DiskBox[{-0.10206615313853885`, 1.4999779155288933`}, 
                    0.05]}, 
                  {RGBColor[0, 0, 1], Arrowheads[0.01], 
                   CircleBox[{-0.1087120883240162, 0.5}, 0.3333333333333333, 
                    NCache[{
                    Rational[-1, 2] 
                    Pi, -1.6814655900321542`}, {-1.5707963267948966`, \
-1.6814655900321542`}]], 
                   ArrowBox[{{-0.14221185963984911`, 
                    0.1683542907967932}, {-0.1455265864832651, 
                    0.16870586507395913`}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-0.1087120883240162, 
                    0.5}, {-0.03493257949917769, 0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-0.1087120883240162, 0.5}, {-0.1575917790330548,
                     0.5}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.005], Arrowheads[0.04], 
                   ArrowBox[{{-0.1087120883240162, 
                    0.5}, {-0.08381227020821629, 0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-0.1087120883240162, 0.5}, 1.6, {
                    1.5641503426849992`, 1.5335998613006265`}], 
                   ArrowBox[{{-0.0652031075492363, 
                    2.0994083182827143`}, {-0.04921146636403941, 
                    2.0988932659769306`}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Dashing[{Small, Small}], 
                   Arrowheads[0.02], 
                   CircleBox[{-0.1087120883240162, 0.5}, 1.7, {
                    1.5641503426849992`, 1.5785805113471874`}], 
                   ArrowBox[{{-0.10494520514537663`, 
                    2.1999958266393236`}, {-0.12194506842313817`, 
                    2.199948495760297}}]}, 
                  {RGBColor[0, 0, 1], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-0.1087120883240162, 0.5}, 1.8, {
                    1.5641503426849992`, 1.5846446506918987`}], 
                   ArrowBox[{{-0.11563905424099508`, 
                    2.2999866713793145`}, {-0.13363827461310004`, 
                    2.2998274042910563`}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.003], Arrowheads[0.025], 
                   CircleBox[{-0.1087120883240162, 0.5}, 1.5, {
                    1.5641503426849992`, 1.572510523300428}], 
                   ArrowBox[{{-0.09628352529633093, 
                    1.9999485093899279`}, {-0.11128338182303478`, 
                    1.999997796148295}}]}, 
                  {RGBColor[0, 0, 1], ArrowBox[{{0, 2.8}, {0, 2.8}}]}},
                 Frame->True,
                 FrameTicks->Automatic,
                 ImageSize->800,
                 PlotRange->{{-3, 3}, {-1, 3}}], 195->
                GraphicsBox[{InsetBox[
                   TagBox[GridBox[{
                    {"\<\"t\"\>", "9.700000000000001`"},
                    {"\<\"driveIntent\"\>", "0"},
                    {"\<\"motorTorque\"\>", 
                    RowBox[{"-", "3.604083905686604`"}]},
                    {"\<\"vw\"\>", "0.0805649921131103`"},
                    {"\<\"controllerState\"\>", 
                    RowBox[{"\[LeftAssociation]", 
                    RowBox[{
                    RowBox[{"\<\"lastError\"\>", "\[Rule]", 
                    RowBox[{"-", "0.030748595949818616`"}]}], ",", 
                    
                    RowBox[{"\<\"lastInt\"\>", "\[Rule]", 
                    "0.08912624024794447`"}], ",", 
                    RowBox[{"\<\"initialized\"\>", "\[Rule]", "True"}]}], 
                    "\[RightAssociation]"}]}
                    },
                    AutoDelete->False,
                    
                    GridBoxItemSize->{
                    "Columns" -> {{Automatic}}, "Rows" -> {{Automatic}}}],
                    "Grid"], {-1, -0.5}], 
                  {GrayLevel[0], LineBox[{{-5, 0}, {5, 0}}]}, 
                  {GrayLevel[0.5], 
                   CircleBox[{-0.10469355411700276`, 0.5}, 0.5], 
                   DiskBox[{-0.10469355411700276`, 0.5}, 0.1]}, 
                  {GrayLevel[0.5], 
                   DiskBox[{-0.09482652002797196, 1.9999675468616933`}, 0.1]}, 
                  {GrayLevel[0.5], Thickness[0.01], 
                   LineBox[{{-0.10469355411700276`, 
                    0.5}, {-0.09482652002797196, 1.9999675468616933`}}]}, 
                  {RGBColor[0, 1, 0], 
                   DiskBox[{-0.09811553139098222, 1.4999783645744622`}, 
                    0.05]}, 
                  {RGBColor[0, 0, 1], Arrowheads[0.01], 
                   CircleBox[{-0.10469355411700276`, 0.5}, 0.3333333333333333,
                     NCache[{
                    Rational[-1, 2] 
                    Pi, -1.6789188439654947`}, {-1.5707963267948966`, \
-1.6789188439654947`}]], 
                   ArrowBox[{{-0.1373486003021324, 
                    0.16827005099063164`}, {-0.14066421176547467`, 
                    0.1686131823692325}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-0.10469355411700276`, 
                    0.5}, {-0.03261187600327067, 0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-0.10469355411700276`, 
                    0.5}, {-0.15307378229212873`, 0.5}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.005], Arrowheads[0.04], 
                   ArrowBox[{{-0.10469355411700276`, 
                    0.5}, {-0.08099210417839664, 0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-0.10469355411700276`, 0.5}, 1.6, {
                    1.5642182566290252`, 1.533979959801152}], 
                   ArrowBox[{{-0.061792509174068866`, 
                    2.0994247404435145`}, {-0.045800673373462705`, 
                    2.0989157665736267`}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Dashing[{Small, Small}], 
                   Arrowheads[0.02], 
                   CircleBox[{-0.10469355411700276`, 0.5}, 1.7, {
                    1.5642182566290252`, 1.577953942389888}], 
                   ArrowBox[{{-0.09986150713496568, 
                    2.1999931327278834`}, {-0.11686139673180815`, 
                    2.1999564534440585`}}]}, 
                  {RGBColor[0, 0, 1], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-0.10469355411700276`, 0.5}, 1.8, {
                    1.5642182566290252`, 1.5842409449939507`}], 
                   ArrowBox[{{-0.11089385461378125`, 
                    2.2999893211554756`}, {-0.12889313781617442`, 
                    2.2998373204678213`}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.003], Arrowheads[0.025], 
                   CircleBox[{-0.10469355411700276`, 0.5}, 1.5, {
                    1.5642182566290252`, 1.5715584099128492`}], 
                   ArrowBox[{{-0.09083687588198806, 
                    1.9999359961239318`}, {-0.1058366786832828, 
                    1.999999564422012}}]}, 
                  {RGBColor[0, 0, 1], ArrowBox[{{0, 2.8}, {0, 2.8}}]}},
                 Frame->True,
                 FrameTicks->Automatic,
                 ImageSize->800,
                 PlotRange->{{-3, 3}, {-1, 3}}], 196->
                GraphicsBox[{InsetBox[
                   TagBox[GridBox[{
                    {"\<\"t\"\>", "9.75`"},
                    {"\<\"driveIntent\"\>", "0"},
                    {"\<\"motorTorque\"\>", 
                    RowBox[{"-", "3.4252852191820935`"}]},
                    {"\<\"vw\"\>", "0.08094157625221478`"},
                    {"\<\"controllerState\"\>", 
                    RowBox[{"\[LeftAssociation]", 
                    RowBox[{
                    RowBox[{"\<\"lastError\"\>", "\[Rule]", 
                    RowBox[{"-", "0.03060163472397904`"}]}], ",", 
                    
                    RowBox[{"\<\"lastInt\"\>", "\[Rule]", 
                    "0.08759176180210687`"}], ",", 
                    RowBox[{"\<\"initialized\"\>", "\[Rule]", "True"}]}], 
                    "\[RightAssociation]"}]}
                    },
                    AutoDelete->False,
                    GridBoxItemSize->{
                    "Columns" -> {{Automatic}}, "Rows" -> {{Automatic}}}],
                    "Grid"], {-1, -0.5}], 
                  {GrayLevel[0], LineBox[{{-5, 0}, {5, 0}}]}, 
                  {GrayLevel[0.5], 
                   CircleBox[{-0.10065558990384868`, 0.5}, 0.5], 
                   DiskBox[{-0.10065558990384868`, 0.5}, 0.1]}, 
                  {GrayLevel[0.5], 
                   DiskBox[{-0.0911841802254079, 1.9999700971681746`}, 0.1]}, 
                  {GrayLevel[0.5], Thickness[0.01], 
                   LineBox[{{-0.10065558990384868`, 
                    0.5}, {-0.0911841802254079, 1.9999700971681746`}}]}, 
                  {RGBColor[0, 1, 0], 
                   DiskBox[{-0.09434131678488816, 1.499980064778783}, 0.05]}, 
                  {RGBColor[0, 0, 1], Arrowheads[0.01], 
                   CircleBox[{-0.10065558990384868`, 0.5}, 0.3333333333333333,
                     NCache[{
                    Rational[-1, 2] 
                    Pi, -1.6735548833703593`}, {-1.5707963267948966`, \
-1.6735548833703593`}]], 
                   ArrowBox[{{-0.1315307884714617, 
                    0.16809966371737195`}, {-0.13484819277077809`, 
                    0.1684250054357302}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-0.10065558990384868`, 
                    0.5}, {-0.03214988552020681, 0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-0.10065558990384868`, 
                    0.5}, {-0.14709606394580943`, 0.5}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.005], Arrowheads[0.04], 
                   ArrowBox[{{-0.10065558990384868`, 
                    0.5}, {-0.07859035956216756, 0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-0.10065558990384868`, 0.5}, 1.6, {
                    1.564482011716791, 1.5354561368033282`}], 
                   ArrowBox[{{-0.060115624864775126`, 
                    2.0994863273046853`}, {-0.044123055152810266`, 
                    2.099000960761007}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Dashing[{Small, Small}], 
                   Arrowheads[0.02], 
                   CircleBox[{-0.10065558990384868`, 0.5}, 1.7, {
                    1.564482011716791, 1.5772694824299223`}], 
                   ArrowBox[{{-0.09465996691296748, 
                    2.1999894271744607`}, {-0.11165987763320893`, 
                    2.1999643836420715`}}]}, 
                  {RGBColor[0, 0, 1], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-0.10065558990384868`, 0.5}, 1.8, {
                    1.564482011716791, 1.5835113740455804`}], 
                   ArrowBox[{{-0.10554266895090544`, 
                    2.2999933656706593`}, {-0.1235420582583017, 
                    2.299854496776409}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.003], Arrowheads[0.025], 
                   CircleBox[{-0.10065558990384868`, 0.5}, 1.5, {
                    1.564482011716791, 1.5708487800785267`}], 
                   ArrowBox[{{-0.08573451591467895, 
                    1.9999257853477317`}, {-0.1007342698292577, 
                    1.9999999979364897`}}]}, 
                  {RGBColor[0, 0, 1], ArrowBox[{{0, 2.8}, {0, 2.8}}]}},
                 Frame->True,
                 FrameTicks->Automatic,
                 ImageSize->800,
                 PlotRange->{{-3, 3}, {-1, 3}}], 197->
                GraphicsBox[{InsetBox[
                   TagBox[GridBox[{
                    {"\<\"t\"\>", "9.8`"},
                    {"\<\"driveIntent\"\>", "0"},
                    {"\<\"motorTorque\"\>", 
                    RowBox[{"-", "3.1655354787709378`"}]},
                    {"\<\"vw\"\>", "0.08128822721771617`"},
                    {"\<\"controllerState\"\>", 
                    RowBox[{"\[LeftAssociation]", 
                    RowBox[{
                    RowBox[{"\<\"lastError\"\>", "\[Rule]", 
                    RowBox[{"-", "0.030275330918390072`"}]}], ",", 
                    
                    RowBox[{"\<\"lastInt\"\>", "\[Rule]", 
                    "0.08606930248087387`"}], ",", 
                    RowBox[{"\<\"initialized\"\>", "\[Rule]", "True"}]}], 
                    "\[RightAssociation]"}]}
                    },
                    AutoDelete->False,
                    
                    GridBoxItemSize->{
                    "Columns" -> {{Automatic}}, "Rows" -> {{Automatic}}}],
                    "Grid"], {-1, -0.5}], 
                  {GrayLevel[0], LineBox[{{-5, 0}, {5, 0}}]}, 
                  {GrayLevel[0.5], 
                   CircleBox[{-0.09659953446442565, 0.5}, 0.5], 
                   DiskBox[{-0.09659953446442565, 0.5}, 0.1]}, 
                  {GrayLevel[0.5], 
                   DiskBox[{-0.0877785849865511, 1.9999740633925338`}, 0.1]}, 
                  {GrayLevel[0.5], Thickness[0.01], 
                   LineBox[{{-0.09659953446442565, 0.5}, {-0.0877785849865511,
                     1.9999740633925338`}}]}, 
                  {RGBColor[0, 1, 0], 
                   DiskBox[{-0.09071890147917595, 1.4999827089283557`}, 
                    0.05]}, 
                  {RGBColor[0, 0, 1], Arrowheads[0.01], 
                   CircleBox[{-0.09659953446442565, 0.5}, 0.3333333333333333, 
                    NCache[{
                    Rational[-1, 2] 
                    Pi, -1.6657623911580246`}, {-1.5707963267948966`, \
-1.6657623911580246`}]], 
                   ArrowBox[{{-0.12488749100954576`, 
                    0.16786914833817002`}, {-0.12820732978525823`, 
                    0.16816862959318107`}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-0.09659953446442565, 
                    0.5}, {-0.03328882488900689, 0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-0.09659953446442565, 
                    0.5}, {-0.13985076868990365`, 0.5}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.005], Arrowheads[0.04], 
                   ArrowBox[{{-0.09659953446442565, 
                    0.5}, {-0.0765400591144849, 0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-0.09659953446442565, 0.5}, 1.6, {
                    1.5649156599152636`, 1.5378831711036391`}], 
                   ArrowBox[{{-0.059941693193623635`, 
                    2.099580008212582}, {-0.043947992583622336`, 
                    2.0991334575755625`}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Dashing[{Small, Small}], 
                   Arrowheads[0.02], 
                   CircleBox[{-0.09659953446442565, 0.5}, 1.7, {
                    1.5649156599152636`, 1.5765407661519075`}], 
                   ArrowBox[{{-0.08936510320706982, 
                    2.1999846066963027`}, {-0.10636502766323025`, 
                    2.1999719512811335`}}]}, 
                  {RGBColor[0, 0, 1], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-0.09659953446442565, 0.5}, 1.8, {
                    1.5649156599152636`, 1.5825019681306578`}], 
                   ArrowBox[{{-0.09966968738017412, 
                    2.299997381709505}, {-0.11766920769283899`, 
                    2.2998766815729486`}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.003], Arrowheads[0.025], 
                   CircleBox[{-0.09659953446442565, 0.5}, 1.5, {
                    1.5649156599152636`, 1.570359695442867}], 
                   ArrowBox[{{-0.08094487163284733, 
                    1.9999183082860312`}, {-0.0959445874571917, 
                    1.9999998570147992`}}]}, 
                  {RGBColor[0, 0, 1], ArrowBox[{{0, 2.8}, {0, 2.8}}]}},
                 Frame->True,
                 FrameTicks->Automatic,
                 ImageSize->800,
                 PlotRange->{{-3, 3}, {-1, 3}}], 198->
                GraphicsBox[{InsetBox[
                   TagBox[GridBox[{
                    {"\<\"t\"\>", "9.85`"},
                    {"\<\"driveIntent\"\>", "0"},
                    {"\<\"motorTorque\"\>", 
                    RowBox[{"-", "2.8370795443456736`"}]},
                    {"\<\"vw\"\>", "0.08159939785797046`"},
                    {"\<\"controllerState\"\>", 
                    RowBox[{"\[LeftAssociation]", 
                    RowBox[{
                    RowBox[{"\<\"lastError\"\>", "\[Rule]", 
                    RowBox[{"-", "0.02979264135081375`"}]}], ",", 
                    
                    RowBox[{"\<\"lastInt\"\>", "\[Rule]", 
                    "0.08456723953215486`"}], ",", 
                    RowBox[{"\<\"initialized\"\>", "\[Rule]", "True"}]}], 
                    "\[RightAssociation]"}]}
                    },
                    AutoDelete->False,
                    
                    GridBoxItemSize->{
                    "Columns" -> {{Automatic}}, "Rows" -> {{Automatic}}}],
                    "Grid"], {-1, -0.5}], 
                  {GrayLevel[0], LineBox[{{-5, 0}, {5, 0}}]}, 
                  {GrayLevel[0.5], 
                   CircleBox[{-0.09252703029258295, 0.5}, 0.5], 
                   DiskBox[{-0.09252703029258295, 0.5}, 0.1]}, 
                  {GrayLevel[0.5], 
                   DiskBox[{-0.08457448005164124, 1.9999789188334165`}, 0.1]}, 
                  {GrayLevel[0.5], Thickness[0.01], 
                   LineBox[{{-0.09252703029258295, 
                    0.5}, {-0.08457448005164124, 1.9999789188334165`}}]}, 
                  {RGBColor[0, 1, 0], 
                   DiskBox[{-0.08722533013195514, 1.4999859458889444`}, 
                    0.05]}, 
                  {RGBColor[0, 0, 1], Arrowheads[0.01], 
                   CircleBox[{-0.09252703029258295, 0.5}, 0.3333333333333333, 
                    NCache[{
                    Rational[-1, 2] 
                    Pi, -1.6559087131252668`}, {-1.5707963267948966`, \
-1.6559087131252668`}]], 
                   ArrowBox[{{-0.11754095602300074`, 
                    0.167606536420067}, {-0.12086358457430238`, 
                    0.16787329104308657`}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-0.09252703029258295, 
                    0.5}, {-0.03578543940566948, 0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-0.09252703029258295, 
                    0.5}, {-0.1315204206776076, 0.5}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.005], Arrowheads[0.04], 
                   ArrowBox[{{-0.09252703029258295, 
                    0.5}, {-0.07477882979069411, 0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-0.09252703029258295, 0.5}, 1.6, {
                    1.5654946017972349`, 1.5411233902908814`}], 
                   ArrowBox[{{-0.06105236222208222, 
                    2.0996903904411788`}, {-0.045057298651691544`, 
                    2.0992956651532393`}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Dashing[{Small, Small}], 
                   Arrowheads[0.02], 
                   CircleBox[{-0.09252703029258295, 0.5}, 1.7, {
                    1.5654946017972349`, 1.575780284519552}], 
                   ArrowBox[{{-0.08399979418311221, 
                    2.19997861346675}, {-0.10099972334767993`, 
                    2.1999788861842946`}}]}, 
                  {RGBColor[0, 0, 1], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-0.09252703029258295, 0.5}, 1.8, {
                    1.5654946017972349`, 1.5812561548213775`}], 
                   ArrowBox[{{-0.09335472071108045, 
                    2.2999998097023706`}, {-0.11135437742545982`, 
                    2.299901533695646}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.003], Arrowheads[0.025], 
                   CircleBox[{-0.09252703029258295, 0.5}, 1.5, {
                    1.5654946017972349`, 1.5700696135713885`}], 
                   ArrowBox[{{-0.0764372690158247, 
                    1.9999137040450217`}, {-0.09143696055326732, 
                    1.9999996039159353`}}]}, 
                  {RGBColor[0, 0, 1], ArrowBox[{{0, 2.8}, {0, 2.8}}]}},
                 Frame->True,
                 FrameTicks->Automatic,
                 ImageSize->800,
                 PlotRange->{{-3, 3}, {-1, 3}}], 199->
                GraphicsBox[{InsetBox[
                   TagBox[GridBox[{
                    {"\<\"t\"\>", "9.9`"},
                    {"\<\"driveIntent\"\>", "0"},
                    {"\<\"motorTorque\"\>", 
                    RowBox[{"-", "2.451415224742451`"}]},
                    {"\<\"vw\"\>", "0.08187055705544527`"},
                    {"\<\"controllerState\"\>", 
                    RowBox[{"\[LeftAssociation]", 
                    RowBox[{
                    RowBox[{"\<\"lastError\"\>", "\[Rule]", 
                    RowBox[{"-", "0.02917539918096612`"}]}], ",", 
                    
                    RowBox[{"\<\"lastInt\"\>", "\[Rule]", 
                    "0.0830928306333085`"}], ",", 
                    RowBox[{"\<\"initialized\"\>", "\[Rule]", "True"}]}], 
                    "\[RightAssociation]"}]}
                    },
                    AutoDelete->False,
                    
                    GridBoxItemSize->{
                    "Columns" -> {{Automatic}}, "Rows" -> {{Automatic}}}],
                    "Grid"], {-1, -0.5}], 
                  {GrayLevel[0], LineBox[{{-5, 0}, {5, 0}}]}, 
                  {GrayLevel[0.5], 
                   CircleBox[{-0.08843997102795176, 0.5}, 0.5], 
                   DiskBox[{-0.08843997102795176, 0.5}, 0.1]}, 
                  {GrayLevel[0.5], 
                   DiskBox[{-0.0815390075380667, 1.999984125483637}, 0.1]}, 
                  {GrayLevel[0.5], Thickness[0.01], 
                   LineBox[{{-0.08843997102795176, 0.5}, {-0.0815390075380667,
                     1.999984125483637}}]}, 
                  {RGBColor[0, 1, 0], 
                   DiskBox[{-0.08383932870136171, 1.4999894169890915`}, 
                    0.05]}, 
                  {RGBColor[0, 0, 1], Arrowheads[0.01], 
                   CircleBox[{-0.08843997102795176, 0.5}, 0.3333333333333333, 
                    NCache[{
                    Rational[-1, 2] 
                    Pi, -1.6443387835371701`}, {-1.5707963267948966`, \
-1.6443387835371701`}]], 
                   ArrowBox[{{-0.10960653938555494`, 
                    0.16733938090739686`}, {-0.11293203181372316`, 
                    0.16756767595557565`}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-0.08843997102795176, 
                    0.5}, {-0.039411666533102734`, 0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-0.08843997102795176, 
                    0.5}, {-0.12227727972759554`, 0.5}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.005], Arrowheads[0.04], 
                   ArrowBox[{{-0.08843997102795176, 
                    0.5}, {-0.07324897523274651, 0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-0.08843997102795176, 0.5}, 1.6, {
                    1.5661956682386886`, 1.5450471264861632`}], 
                   ArrowBox[{{-0.06324229222486857, 
                    2.099801574253175}, {-0.0472458029880405, 
                    2.099469612252605}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Dashing[{Small, Small}], 
                   Arrowheads[0.02], 
                   CircleBox[{-0.08843997102795176, 0.5}, 1.7, {
                    1.5661956682386886`, 1.5749993955437844`}], 
                   ArrowBox[{{-0.07858524309500023, 
                    2.1999714363298484`}, {-0.09558516686343317, 
                    2.199984984103234}}]}, 
                  {RGBColor[0, 0, 1], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-0.08843997102795176, 0.5}, 1.8, {
                    1.5661956682386886`, 1.5798146417094798`}], 
                   ArrowBox[{{-0.08667293815801813, 
                    2.2999991326650235`}, {-0.10467271783722017`, 
                    2.299926803492583}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.003], Arrowheads[0.025], 
                   CircleBox[{-0.08843997102795176, 0.5}, 1.5, {
                    1.5661956682386886`, 1.569957549480053}], 
                   ArrowBox[{{-0.07218212338625044, 
                    1.9999118915423197`}, {-0.0871818052032163, 
                    1.9999994723394932`}}]}, 
                  {RGBColor[0, 0, 1], ArrowBox[{{0, 2.8}, {0, 2.8}}]}},
                 Frame->True,
                 FrameTicks->Automatic,
                 ImageSize->800,
                 PlotRange->{{-3, 3}, {-1, 3}}], 200->
                GraphicsBox[{InsetBox[
                   TagBox[GridBox[{
                    {"\<\"t\"\>", "9.950000000000001`"},
                    {"\<\"driveIntent\"\>", "0"},
                    {"\<\"motorTorque\"\>", 
                    RowBox[{"-", "2.0192688453878844`"}]},
                    {"\<\"vw\"\>", "0.08209811636954427`"},
                    {"\<\"controllerState\"\>", 
                    RowBox[{"\[LeftAssociation]", 
                    RowBox[{
                    RowBox[{"\<\"lastError\"\>", "\[Rule]", 
                    RowBox[{"-", "0.02844423304307989`"}]}], ",", 
                    
                    RowBox[{"\<\"lastInt\"\>", "\[Rule]", 
                    "0.08165227222739012`"}], ",", 
                    RowBox[{"\<\"initialized\"\>", "\[Rule]", "True"}]}], 
                    "\[RightAssociation]"}]}
                    },
                    AutoDelete->False,
                    
                    GridBoxItemSize->{
                    "Columns" -> {{Automatic}}, "Rows" -> {{Automatic}}}],
                    "Grid"], {-1, -0.5}], 
                  {GrayLevel[0], LineBox[{{-5, 0}, {5, 0}}]}, 
                  {GrayLevel[0.5], 
                   CircleBox[{-0.08434045252016745, 0.5}, 0.5], 
                   DiskBox[{-0.08434045252016745, 0.5}, 0.1]}, 
                  {GrayLevel[0.5], 
                   DiskBox[{-0.07864174429870133, 1.9999891748691412`}, 0.1]}, 
                  {GrayLevel[0.5], Thickness[0.01], 
                   LineBox[{{-0.08434045252016745, 
                    0.5}, {-0.07864174429870133, 1.9999891748691412`}}]}, 
                  {RGBColor[0, 1, 0], 
                   DiskBox[{-0.0805413137058567, 1.499992783246094}, 0.05]}, 
                  {RGBColor[0, 0, 1], Arrowheads[0.01], 
                   CircleBox[{-0.08434045252016745, 0.5}, 0.3333333333333333, 
                    NCache[{
                    Rational[-1, 2] 
                    Pi, -1.6313743921565331`}, {-1.5707963267948966`, \
-1.6313743921565331`}]], 
                   ArrowBox[{{-0.1011926204596593, 
                    0.1670929325669846}, {-0.10452079304838038`, 
                    0.1672780966523597}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-0.08434045252016745, 
                    0.5}, {-0.04395507561240976, 0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-0.08434045252016745, 
                    0.5}, {-0.11228286935518668`, 0.5}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.005], Arrowheads[0.04], 
                   ArrowBox[{{-0.08434045252016745, 
                    0.5}, {-0.07189749244742899, 0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-0.08434045252016745, 0.5}, 1.6, {
                    1.5669971788414094`, 1.5495330422851468`}], 
                   ArrowBox[{{-0.06631957833629631, 
                    2.0998985118105615`}, {-0.05032176090280993, 
                    2.0996383118132185`}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Dashing[{Small, Small}], 
                   Arrowheads[0.02], 
                   CircleBox[{-0.08434045252016745, 0.5}, 1.7, {
                    1.5669971788414094`, 1.5742083456245834`}], 
                   ArrowBox[{{-0.07314096554345993, 
                    2.1999631088619127`}, {-0.0901408732759933, 
                    2.19999010441798}}]}, 
                  {RGBColor[0, 0, 1], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-0.08434045252016745, 0.5}, 1.8, {
                    1.5669971788414094`, 1.5782153390935645`}], 
                   ArrowBox[{{-0.07969467981574081, 
                    2.299994004655565}, {-0.0976945521514952, 
                    2.2999504626580802`}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.003], Arrowheads[0.025], 
                   CircleBox[{-0.08434045252016745, 0.5}, 1.5, {
                    1.5669971788414094`, 1.5700032145049123`}], 
                   ArrowBox[{{-0.06815109840920748, 
                    1.999912632393457}, {-0.08315078420991327, 
                    1.9999995282296963`}}]}, 
                  {RGBColor[0, 0, 1], ArrowBox[{{0, 2.8}, {0, 2.8}}]}},
                 Frame->True,
                 FrameTicks->Automatic,
                 ImageSize->800,
                 PlotRange->{{-3, 3}, {-1, 3}}], 201->
                GraphicsBox[{InsetBox[
                   TagBox[GridBox[{
                    {"\<\"t\"\>", "10.`"},
                    {"\<\"driveIntent\"\>", "0"},
                    {"\<\"motorTorque\"\>", 
                    RowBox[{"-", "1.5505810556750959`"}]},
                    {"\<\"vw\"\>", "0.08227935600676639`"},
                    {"\<\"controllerState\"\>", 
                    RowBox[{"\[LeftAssociation]", 
                    RowBox[{
                    RowBox[{"\<\"lastError\"\>", "\[Rule]", 
                    RowBox[{"-", "0.027618507147076574`"}]}], ",", 
                    
                    RowBox[{"\<\"lastInt\"\>", "\[Rule]", 
                    "0.080250761360284`"}], ",", 
                    RowBox[{"\<\"initialized\"\>", "\[Rule]", "True"}]}], 
                    "\[RightAssociation]"}]}
                    },
                    AutoDelete->False,
                    
                    GridBoxItemSize->{
                    "Columns" -> {{Automatic}}, "Rows" -> {{Automatic}}}],
                    "Grid"], {-1, -0.5}], 
                  {GrayLevel[0], LineBox[{{-5, 0}, {5, 0}}]}, 
                  {GrayLevel[0.5], 
                   CircleBox[{-0.08023072758179688, 0.5}, 0.5], 
                   DiskBox[{-0.08023072758179688, 0.5}, 0.1]}, 
                  {GrayLevel[0.5], 
                   DiskBox[{-0.075854712430626, 1.999993616816884}, 0.1]}, 
                  {GrayLevel[0.5], Thickness[0.01], 
                   LineBox[{{-0.08023072758179688, 0.5}, {-0.075854712430626, 
                    1.999993616816884}}]}, 
                  {RGBColor[0, 1, 0], 
                   DiskBox[{-0.07731338414768295, 1.4999957445445893`}, 
                    0.05]}, 
                  {RGBColor[0, 0, 1], Arrowheads[0.01], 
                   CircleBox[{-0.08023072758179688, 0.5}, 0.3333333333333333, 
                    NCache[{
                    Rational[-1, 2] 
                    Pi, -1.6173137584651494`}, {-1.5707963267948966`, \
-1.6173137584651494`}]], 
                   ArrowBox[{{-0.09240049960624262, 
                    0.16688889577202004`}, {-0.09573094664675215, 
                    0.1670272468803949}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-0.08023072758179688, 
                    0.5}, {-0.04921910646829496, 0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-0.08023072758179688, 
                    0.5}, {-0.10168766076350087`, 0.5}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.005], Arrowheads[0.04], 
                   ArrowBox[{{-0.08023072758179688, 
                    0.5}, {-0.07067603964999895, 0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-0.08023072758179688, 0.5}, 1.6, {
                    1.5678789792225674`, 1.5544683389156204`}], 
                   ArrowBox[{{-0.07010601454664461, 
                    2.099967965362418}, {-0.05410710778444781, 
                    2.0997867221878934`}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Dashing[{Small, Small}], 
                   Arrowheads[0.02], 
                   CircleBox[{-0.08023072758179688, 0.5}, 1.7, {
                    1.5678789792225674`, 1.5734162996345809`}], 
                   ArrowBox[{{-0.0676847952952677, 
                    2.1999537051293667`}, {-0.08468467631374736, 
                    2.199994165384309}}]}, 
                  {RGBColor[0, 0, 1], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-0.08023072758179688, 0.5}, 1.8, {
                    1.5678789792225674`, 1.5764933184207623`}], 
                   ArrowBox[{{-0.07248533641053043, 
                    2.299983335732752}, {-0.09048525703846665, 
                    2.2999707899367765`}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.003], Arrowheads[0.025], 
                   CircleBox[{-0.08023072758179688, 0.5}, 1.5, {
                    1.5678789792225674`, 1.5701871330666066`}], 
                   ArrowBox[{{-0.06431723551710947, 
                    1.999915584548113}, {-0.07931693704588241, 
                    1.9999997216622596`}}]}, 
                  {RGBColor[0, 0, 1], ArrowBox[{{0, 2.8}, {0, 2.8}}]}},
                 Frame->True,
                 FrameTicks->Automatic,
                 ImageSize->800,
                 PlotRange->{{-3, 3}, {-1, 3}}], 202->
                GraphicsBox[{InsetBox[
                   TagBox[GridBox[{
                    {"\<\"t\"\>", "10.05`"},
                    {"\<\"driveIntent\"\>", "0"},
                    {"\<\"motorTorque\"\>", 
                    RowBox[{"-", "1.0545017842784574`"}]},
                    {"\<\"vw\"\>", "0.08241235094833739`"},
                    {"\<\"controllerState\"\>", 
                    RowBox[{"\[LeftAssociation]", 
                    RowBox[{
                    RowBox[{"\<\"lastError\"\>", "\[Rule]", 
                    RowBox[{"-", "0.026716280480503108`"}]}], ",", 
                    
                    RowBox[{"\<\"lastInt\"\>", "\[Rule]", 
                    "0.07889256001795443`"}], ",", 
                    RowBox[{"\<\"initialized\"\>", "\[Rule]", "True"}]}], 
                    "\[RightAssociation]"}]}
                    },
                    AutoDelete->False,
                    
                    GridBoxItemSize->{
                    "Columns" -> {{Automatic}}, "Rows" -> {{Automatic}}}],
                    "Grid"], {-1, -0.5}], 
                  {GrayLevel[0], LineBox[{{-5, 0}, {5, 0}}]}, 
                  {GrayLevel[0.5], 
                   CircleBox[{-0.07611316444196654, 0.5}, 0.5], 
                   DiskBox[{-0.07611316444196654, 0.5}, 0.1]}, 
                  {GrayLevel[0.5], 
                   DiskBox[{-0.07315236510423445, 1.9999970778862477`}, 0.1]}, 
                  {GrayLevel[0.5], Thickness[0.01], 
                   LineBox[{{-0.07611316444196654, 
                    0.5}, {-0.07315236510423445, 1.9999970778862477`}}]}, 
                  {RGBColor[0, 1, 0], 
                   DiskBox[{-0.07413929821681181, 1.4999980519241651`}, 
                    0.05]}, 
                  {RGBColor[0, 0, 1], Arrowheads[0.01], 
                   CircleBox[{-0.07611316444196654, 0.5}, 0.3333333333333333, 
                    NCache[{
                    Rational[-1, 2] 
                    Pi, -1.6024313803232504`}, {-1.5707963267948966`, \
-1.6024313803232504`}]], 
                   ArrowBox[{{-0.08332428636239642, 
                    0.16674467621392786`}, {-0.08665642350488949, 
                    0.16683344885861784`}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-0.07611316444196654, 
                    0.5}, {-0.055023128756397394`, 0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-0.07611316444196654, 
                    0.5}, {-0.09063089757295159, 0.5}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.005], Arrowheads[0.04], 
                   ArrowBox[{{-0.07611316444196654, 
                    0.5}, {-0.06954086188738244, 0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-0.07611316444196654, 0.5}, 1.6, {
                    1.5688224592879938`, 1.5597488584050656`}], 
                   ArrowBox[{{-0.07443721532470943, 
                    2.0999991222480583`}, {-0.05843757456417342, 
                    2.099902363746761}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Dashing[{Small, Small}], 
                   Arrowheads[0.02], 
                   CircleBox[{-0.07611316444196654, 0.5}, 1.7, {
                    1.5688224592879938`, 1.572631378133211}], 
                   ArrowBox[{{-0.06223290594274728, 
                    2.1999433338861607`}, {-0.07923274996627808, 
                    2.199997137699401}}]}, 
                  {RGBColor[0, 0, 1], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-0.07611316444196654, 0.5}, 1.8, {
                    1.5688224592879938`, 1.5746808025339851`}], 
                   ArrowBox[{{-0.06510528938771373, 
                    2.2999663404316175`}, {-0.08310520318830593, 
                    2.2999864197804856`}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.003], Arrowheads[0.025], 
                   CircleBox[{-0.07611316444196654, 0.5}, 1.5, {
                    1.5688224592879938`, 1.570490738265973}], 
                   ArrowBox[{{-0.060655055273784605`, 
                    1.9999203468387727`}, {-0.07565478165571536, 
                    1.9999999299617388`}}]}, 
                  {RGBColor[0, 0, 1], ArrowBox[{{0, 2.8}, {0, 2.8}}]}},
                 Frame->True,
                 FrameTicks->Automatic,
                 ImageSize->800,
                 PlotRange->{{-3, 3}, {-1, 3}}], 203->
                GraphicsBox[{InsetBox[
                   TagBox[GridBox[{
                    {"\<\"t\"\>", "10.1`"},
                    {"\<\"driveIntent\"\>", "0"},
                    {"\<\"motorTorque\"\>", 
                    RowBox[{"-", "0.5393933054978177`"}]},
                    {"\<\"vw\"\>", "0.08249589795549254`"},
                    {"\<\"controllerState\"\>", 
                    RowBox[{"\[LeftAssociation]", 
                    RowBox[{
                    RowBox[{"\<\"lastError\"\>", "\[Rule]", 
                    RowBox[{"-", "0.025754283266479407`"}]}], ",", 
                    
                    RowBox[{"\<\"lastInt\"\>", "\[Rule]", 
                    "0.07758106105597148`"}], ",", 
                    RowBox[{"\<\"initialized\"\>", "\[Rule]", "True"}]}], 
                    "\[RightAssociation]"}]}
                    },
                    AutoDelete->False,
                    
                    GridBoxItemSize->{
                    "Columns" -> {{Automatic}}, "Rows" -> {{Automatic}}}],
                    "Grid"], {-1, -0.5}], 
                  {GrayLevel[0], LineBox[{{-5, 0}, {5, 0}}]}, 
                  {GrayLevel[0.5], 
                   CircleBox[{-0.07199020887400726, 0.5}, 0.5], 
                   DiskBox[{-0.07199020887400726, 0.5}, 0.1]}, 
                  {GrayLevel[0.5], 
                   DiskBox[{-0.07051155074497693, 1.9999992711898686`}, 0.1]}, 
                  {GrayLevel[0.5], Thickness[0.01], 
                   LineBox[{{-0.07199020887400726, 
                    0.5}, {-0.07051155074497693, 1.9999992711898686`}}]}, 
                  {RGBColor[0, 1, 0], 
                   DiskBox[{-0.07100443678798704, 1.4999995141265792`}, 
                    0.05]}, 
                  {RGBColor[0, 0, 1], Arrowheads[0.01], 
                   CircleBox[{-0.07199020887400726, 0.5}, 0.3333333333333333, 
                    NCache[{
                    Rational[-1, 2] 
                    Pi, -1.586978125959831}, {-1.5707963267948966`, \
-1.586978125959831}]], 
                   ArrowBox[{{-0.07405079547149739, 
                    0.16667303575320325`}, {-0.07738390653127779, 
                    0.16671030748507532`}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-0.07199020887400726, 
                    0.5}, {-0.0612023427640509, 0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-0.07199020887400726, 
                    0.5}, {-0.07924054672178671, 0.5}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.005], Arrowheads[0.04], 
                   ArrowBox[{{-0.07199020887400726, 
                    0.5}, {-0.06845268061183035, 0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-0.07199020887400726, 0.5}, 1.6, {
                    1.569810554549223, 1.565279091192643}], 
                   ArrowBox[{{-0.07916260788857443, 
                    2.0999839238856044`}, {-0.06316267669541012, 
                    2.09997564815082}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Dashing[{Small, Small}], 
                   Arrowheads[0.02], 
                   CircleBox[{-0.07199020887400726, 0.5}, 1.7, {
                    1.569810554549223, 1.5718607003071392`}], 
                   ArrowBox[{{-0.056799845993521905`, 
                    2.1999321318439624`}, {-0.0737996435031709, 
                    2.1999990370427636`}}]}, 
                  {RGBColor[0, 0, 1], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-0.07199020887400726, 0.5}, 1.8, {
                    1.569810554549223, 1.5728071840242106`}], 
                   ArrowBox[{{-0.057609904861756156`, 
                    2.2999425565435456`}, {-0.07560974944747433, 
                    2.299996360809109}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.003], Arrowheads[0.025], 
                   CircleBox[{-0.07199020887400726, 0.5}, 1.5, {
                    1.569810554549223, 1.5708964475311618`}], 
                   ArrowBox[{{-0.057140632543091374`, 
                    1.9999264948932638`}, {-0.07214038997815406, 
                    1.9999999924818788`}}]}, 
                  {RGBColor[0, 0, 1], ArrowBox[{{0, 2.8}, {0, 2.8}}]}},
                 Frame->True,
                 FrameTicks->Automatic,
                 ImageSize->800,
                 PlotRange->{{-3, 3}, {-1, 3}}], 204->
                GraphicsBox[{InsetBox[
                   TagBox[GridBox[{
                    {"\<\"t\"\>", "10.15`"},
                    {"\<\"driveIntent\"\>", "0"},
                    {"\<\"motorTorque\"\>", 
                    RowBox[{"-", "0.012840436795910293`"}]},
                    {"\<\"vw\"\>", "0.08252944406905519`"},
                    {"\<\"controllerState\"\>", 
                    RowBox[{"\[LeftAssociation]", 
                    RowBox[{
                    RowBox[{"\<\"lastError\"\>", "\[Rule]", 
                    RowBox[{"-", "0.024747908880599922`"}]}], ",", 
                    
                    RowBox[{"\<\"lastInt\"\>", "\[Rule]", 
                    "0.07631885490457625`"}], ",", 
                    RowBox[{"\<\"initialized\"\>", "\[Rule]", "True"}]}], 
                    "\[RightAssociation]"}]}
                    },
                    AutoDelete->False,
                    
                    GridBoxItemSize->{
                    "Columns" -> {{Automatic}}, "Rows" -> {{Automatic}}}],
                    "Grid"], {-1, -0.5}], 
                  {GrayLevel[0], LineBox[{{-5, 0}, {5, 0}}]}, 
                  {GrayLevel[0.5], 
                   CircleBox[{-0.06786434993709267, 0.5}, 0.5], 
                   DiskBox[{-0.06786434993709267, 0.5}, 0.1]}, 
                  {GrayLevel[0.5], 
                   DiskBox[{-0.06791145844252237, 1.9999999992602628`}, 0.1]}, 
                  {GrayLevel[0.5], Thickness[0.01], 
                   LineBox[{{-0.06786434993709267, 
                    0.5}, {-0.06791145844252237, 1.9999999992602628`}}]}, 
                  {RGBColor[0, 1, 0], 
                   DiskBox[{-0.06789575560737914, 1.499999999506842}, 0.05]}, 
                  {RGBColor[0, 0, 1], Arrowheads[0.01], 
                   CircleBox[{-0.06786434993709267, 0.5}, 0.3333333333333333, 
                    NCache[{
                    Rational[-1, 2] 
                    Pi, -1.5711815398987738`}, {-1.5707963267948966`, \
-1.5711815398987738`}]], 
                   ArrowBox[{{-0.06465947035096692, 
                    0.1666820739024834}, {-0.06799275430187611, 
                    0.16666669139818896`}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-0.06786434993709267, 
                    0.5}, {-0.06760754120117446, 0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-0.06786434993709267, 
                    0.5}, {-0.0676333616248205, 0.5}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.005], Arrowheads[0.04], 
                   ArrowBox[{{-0.06786434993709267, 
                    0.5}, {-0.06737655288890229, 0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-0.06786434993709267, 0.5}, 1.6, {
                    1.5708277324651883`, 1.5709721001604295`}], 
                   ArrowBox[{{-0.05214584017124457, 
                    2.099922788277841}, {-0.06814558732049715, 
                    2.099999975282979}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Dashing[{Small, Small}], 
                   Arrowheads[0.02], 
                   CircleBox[{-0.06786434993709267, 0.5}, 1.7, {
                    1.5708277324651883`, 1.571110431420821}], 
                   ArrowBox[{{-0.051398585264240695`, 
                    2.199920256539623}, {-0.06839832779238372, 
                    2.199999916137542}}]}, 
                  {RGBColor[0, 0, 1], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-0.06786434993709267, 0.5}, 1.8, {
                    1.5708277324651883`, 1.5708990682251656`}], 
                   ArrowBox[{{-0.05004957535810038, 
                    2.2999118405651706`}, {-0.06804928451125142, 
                    2.2999999904997788`}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.003], Arrowheads[0.025], 
                   CircleBox[{-0.06786434993709267, 0.5}, 1.5, {
                    1.5708277324651883`, 1.5713877197351356`}], 
                   ArrowBox[{{-0.05375164756344179, 
                    1.9999336090746527`}, {-0.06875143929574185, 
                    1.9999997376908003`}}]}, 
                  {RGBColor[0, 0, 1], ArrowBox[{{0, 2.8}, {0, 2.8}}]}},
                 Frame->True,
                 FrameTicks->Automatic,
                 ImageSize->800,
                 PlotRange->{{-3, 3}, {-1, 3}}], 205->
                GraphicsBox[{InsetBox[
                   TagBox[GridBox[{
                    {"\<\"t\"\>", "10.200000000000001`"},
                    {"\<\"driveIntent\"\>", "0"},
                    {"\<\"motorTorque\"\>", "0.5183330550563436`"},
                    {"\<\"vw\"\>", "0.08251301712418978`"},
                    {"\<\"controllerState\"\>", 
                    RowBox[{"\[LeftAssociation]", 
                    RowBox[{
                    RowBox[{"\<\"lastError\"\>", "\[Rule]", 
                    RowBox[{"-", "0.023711219495021238`"}]}], ",", 
                    
                    RowBox[{"\<\"lastInt\"\>", "\[Rule]", 
                    "0.07510779632079878`"}], ",", 
                    RowBox[{"\<\"initialized\"\>", "\[Rule]", "True"}]}], 
                    "\[RightAssociation]"}]}
                    },
                    AutoDelete->False,
                    
                    GridBoxItemSize->{
                    "Columns" -> {{Automatic}}, "Rows" -> {{Automatic}}}],
                    "Grid"], {-1, -0.5}], 
                  {GrayLevel[0], LineBox[{{-5, 0}, {5, 0}}]}, 
                  {GrayLevel[0.5], 
                   CircleBox[{-0.06373808924349361, 0.5}, 0.5], 
                   DiskBox[{-0.06373808924349361, 0.5}, 0.1]}, 
                  {GrayLevel[0.5], 
                   DiskBox[{-0.06533354729306816, 1.9999991515042974`}, 0.1]}, 
                  {GrayLevel[0.5], Thickness[0.01], 
                   LineBox[{{-0.06373808924349361, 
                    0.5}, {-0.06533354729306816, 1.9999991515042974`}}]}, 
                  {RGBColor[0, 1, 0], 
                   DiskBox[{-0.06480172794320999, 1.4999994343361982`}, 
                    0.05]}, 
                  {RGBColor[0, 0, 1], Arrowheads[0.01], 
                   CircleBox[{-0.06373808924349361, 0.5}, 0.3333333333333333, 
                    NCache[{
                    Rational[-1, 2] 
                    Pi, -1.5552463351432062`}, {-1.5707963267948966`, \
-1.5552463351432062`}]], 
                   ArrowBox[{{-0.06188810152364349, 
                    0.16667180038804474`}, {-0.058554967580561265`, 
                    0.1667069662280078}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-0.06373808924349361, 
                    0.5}, {-0.07410475034462048, 0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-0.06373808924349361, 
                    0.5}, {-0.0559150443277793, 0.5}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.005], Arrowheads[0.04], 
                   ArrowBox[{{-0.06373808924349361, 
                    0.5}, {-0.06628170542890617, 0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-0.06373808924349361, 0.5}, 1.6, {
                    1.5718599656951668`, 1.5767493715332483`}], 
                   ArrowBox[{{-0.05726297849961886, 
                    2.0999868977403704`}, {-0.07326290456671218, 
                    2.0999716490904015`}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Dashing[{Small, Small}], 
                   Arrowheads[0.02], 
                   CircleBox[{-0.06373808924349361, 0.5}, 1.7, {
                    1.5718599656951668`, 1.5703858337249057`}], 
                   ArrowBox[{{-0.08004000117154488, 
                    2.1999218357523063`}, {-0.0630402510441072, 
                    2.1999998567711256`}}]}, 
                  {RGBColor[0, 0, 1], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-0.06373808924349361, 0.5}, 1.8, {
                    1.5718599656951668`, 1.5689803376115206`}], 
                   ArrowBox[{{-0.07846914426928335, 
                    2.2999397201067118`}, {-0.06046931051005625, 
                    2.299997031965773}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.003], Arrowheads[0.025], 
                   CircleBox[{-0.06373808924349361, 0.5}, 1.5, {
                    1.5718599656951668`, 1.5719490953224486`}], 
                   ArrowBox[{{-0.0504674151600977, 
                    1.999941295254375}, {-0.06546724165185028, 
                    1.9999990033436517`}}]}, 
                  {RGBColor[0, 0, 1], ArrowBox[{{0, 2.8}, {0, 2.8}}]}},
                 Frame->True,
                 FrameTicks->Automatic,
                 ImageSize->800,
                 PlotRange->{{-3, 3}, {-1, 3}}], 206->
                GraphicsBox[{InsetBox[
                   TagBox[GridBox[{
                    {"\<\"t\"\>", "10.25`"},
                    {"\<\"driveIntent\"\>", "0"},
                    {"\<\"motorTorque\"\>", "1.0480427052942698`"},
                    {"\<\"vw\"\>", "0.08244715871378115`"},
                    {"\<\"controllerState\"\>", 
                    RowBox[{"\[LeftAssociation]", 
                    RowBox[{
                    RowBox[{"\<\"lastError\"\>", "\[Rule]", 
                    RowBox[{"-", "0.022656963795242874`"}]}], ",", 
                    RowBox[{"\<\"lastInt\"\>", "\[Rule]", 
                    "0.07394907054379886`"}], ",", 
                    RowBox[{"\<\"initialized\"\>", "\[Rule]", "True"}]}], 
                    "\[RightAssociation]"}]}
                    },
                    AutoDelete->False,
                    
                    GridBoxItemSize->{
                    "Columns" -> {{Automatic}}, "Rows" -> {{Automatic}}}],
                    "Grid"], {-1, -0.5}], 
                  {GrayLevel[0], LineBox[{{-5, 0}, {5, 0}}]}, 
                  {GrayLevel[0.5], 
                   CircleBox[{-0.05961391363928804, 0.5}, 0.5], 
                   DiskBox[{-0.05961391363928804, 0.5}, 0.1]}, 
                  {GrayLevel[0.5], 
                   DiskBox[{-0.0627614622058802, 1.9999966976423718`}, 0.1]}, 
                  {GrayLevel[0.5], Thickness[0.01], 
                   LineBox[{{-0.05961391363928804, 0.5}, {-0.0627614622058802,
                     1.9999966976423718`}}]}, 
                  {RGBColor[0, 1, 0], 
                   DiskBox[{-0.06171227935034948, 1.499997798428248}, 0.05]}, 
                  {RGBColor[0, 0, 1], Arrowheads[0.01], 
                   CircleBox[{-0.05961391363928804, 0.5}, 0.3333333333333333, 
                    NCache[{
                    Rational[-1, 2] 
                    Pi, -1.5393550456360685`}, {-1.5707963267948966`, \
-1.5393550456360685`}]], 
                   ArrowBox[{{-0.05246736752758176, 
                    0.16674328515424575`}, {-0.04913521324596018, 
                    0.16683141212116354`}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-0.05961391363928804, 
                    0.5}, {-0.08057476774517343, 0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-0.05961391363928804, 
                    0.5}, {-0.04418049404185792, 0.5}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.005], Arrowheads[0.04], 
                   ArrowBox[{{-0.05961391363928804, 
                    0.5}, {-0.06514134814774332, 0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-0.05961391363928804, 0.5}, 1.6, {
                    1.5728946940458604`, 1.5825406025304136`}], 
                   ArrowBox[{{-0.062404753400927396`, 
                    2.0999975660023438`}, {-0.07840432285519794, 
                    2.0998896588582285`}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Dashing[{Small, Small}], 
                   Arrowheads[0.02], 
                   CircleBox[{-0.05961391363928804, 0.5}, 1.7, {
                    1.5728946940458604`, 1.5696913199110272`}], 
                   ArrowBox[{{-0.07473520253319854, 
                    2.199932747676268}, {-0.05773540231899967, 
                    2.199998962115924}}]}, 
                  {RGBColor[0, 0, 1], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-0.05961391363928804, 0.5}, 1.8, {
                    1.5728946940458604`, 1.5670722345720032`}], 
                   ArrowBox[{{-0.07091047348143731, 
                    2.299964551799766}, {-0.05291056313274642, 
                    2.2999875180378297`}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.003], Arrowheads[0.025], 
                   CircleBox[{-0.05961391363928804, 0.5}, 1.5, {
                    1.5728946940458604`, 1.572566221041278}], 
                   ArrowBox[{{-0.07726834738961305, 
                    1.9998961033914833`}, {-0.06226875362280066, 
                    1.9999976506063808`}}]}, 
                  {RGBColor[0, 0, 1], ArrowBox[{{0, 2.8}, {0, 2.8}}]}},
                 Frame->True,
                 FrameTicks->Automatic,
                 ImageSize->800,
                 PlotRange->{{-3, 3}, {-1, 3}}], 207->
                GraphicsBox[{InsetBox[
                   TagBox[GridBox[{
                    {"\<\"t\"\>", "10.3`"},
                    {"\<\"driveIntent\"\>", "0"},
                    {"\<\"motorTorque\"\>", "1.5709170707845033`"},
                    {"\<\"vw\"\>", "0.08233285995483015`"},
                    {"\<\"controllerState\"\>", 
                    RowBox[{"\[LeftAssociation]", 
                    RowBox[{
                    RowBox[{"\<\"lastError\"\>", "\[Rule]", 
                    RowBox[{"-", "0.021596605200035163`"}]}], ",", 
                    
                    RowBox[{"\<\"lastInt\"\>", "\[Rule]", 
                    "0.07284325829018506`"}], ",", 
                    RowBox[{"\<\"initialized\"\>", "\[Rule]", "True"}]}], 
                    "\[RightAssociation]"}]}
                    },
                    AutoDelete->False,
                    
                    GridBoxItemSize->{
                    "Columns" -> {{Automatic}}, "Rows" -> {{Automatic}}}],
                    "Grid"], {-1, -0.5}], 
                  {GrayLevel[0], LineBox[{{-5, 0}, {5, 0}}]}, 
                  {GrayLevel[0.5], 
                   CircleBox[{-0.05549427116674827, 0.5}, 0.5], 
                   DiskBox[{-0.05549427116674827, 0.5}, 0.1]}, 
                  {GrayLevel[0.5], 
                   DiskBox[{-0.06018093852912241, 1.9999926783651427`}, 0.1]}, 
                  {GrayLevel[0.5], Thickness[0.01], 
                   LineBox[{{-0.05549427116674827, 
                    0.5}, {-0.06018093852912241, 1.9999926783651427`}}]}, 
                  {RGBColor[0, 1, 0], 
                   DiskBox[{-0.0586187160749977, 1.4999951189100953`}, 0.05]}, 
                  {RGBColor[0, 0, 1], Arrowheads[0.01], 
                   CircleBox[{-0.05549427116674827, 0.5}, 0.3333333333333333, 
                    NCache[{
                    Rational[-1, 2] 
                    Pi, -1.5236688146713615`}, {-1.5707963267948966`, \
-1.5236688146713615`}]], 
                   ArrowBox[{{-0.04312127684623167, 
                    0.16689638230326054`}, {-0.039790914830816926`, 
                    0.16703676522640243`}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-0.05549427116674827, 
                    0.5}, {-0.08691261258243835, 0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-0.05549427116674827, 
                    0.5}, {-0.03251413009081723, 0.5}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.005], Arrowheads[0.04], 
                   ArrowBox[{{-0.05549427116674827, 
                    0.5}, {-0.0639324715065073, 0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-0.05549427116674827, 0.5}, 1.6, {
                    1.5739207767867214`, 1.5882834350646047`}], 
                   ArrowBox[{{-0.0674735324777261, 
                    2.0999551547772968`}, {-0.0834722184095657, 
                    2.099755367069627}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Dashing[{Small, Small}], 
                   Arrowheads[0.02], 
                   CircleBox[{-0.05549427116674827, 0.5}, 1.7, {
                    1.5739207767867214`, 1.5690305083201213`}], 
                   ArrowBox[{{-0.06949222157746729, 
                    2.1999423688420436`}, {-0.05249238131967029, 
                    2.1999973496030356`}}]}, 
                  {RGBColor[0, 0, 1], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-0.05549427116674827, 0.5}, 1.8, {
                    1.5739207767867214`, 1.5651934597268076`}], 
                   ArrowBox[{{-0.0634090849389359, 
                    2.2999825987278184`}, {-0.045409163209866724`, 
                    2.2999717471664654`}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.003], Arrowheads[0.025], 
                   CircleBox[{-0.05549427116674827, 0.5}, 1.5, {
                    1.5739207767867214`, 1.57322586088242}], 
                   ArrowBox[{{-0.0741380922317521, 
                    1.9998841315035292`}, {-0.059138568712870467`, 
                    1.9999955730252656`}}]}, 
                  {RGBColor[0, 0, 1], ArrowBox[{{0, 2.8}, {0, 2.8}}]}},
                 Frame->True,
                 FrameTicks->Automatic,
                 ImageSize->800,
                 PlotRange->{{-3, 3}, {-1, 3}}], 208->
                GraphicsBox[{InsetBox[
                   TagBox[GridBox[{
                    {"\<\"t\"\>", "10.35`"},
                    {"\<\"driveIntent\"\>", "0"},
                    {"\<\"motorTorque\"\>", "2.082267675020658`"},
                    {"\<\"vw\"\>", "0.08217150034121071`"},
                    {"\<\"controllerState\"\>", 
                    RowBox[{"\[LeftAssociation]", 
                    RowBox[{
                    RowBox[{"\<\"lastError\"\>", "\[Rule]", 
                    RowBox[{"-", "0.020540359104388353`"}]}], ",", 
                    
                    RowBox[{"\<\"lastInt\"\>", "\[Rule]", 
                    "0.07179039910227453`"}], ",", 
                    RowBox[{"\<\"initialized\"\>", "\[Rule]", "True"}]}], 
                    "\[RightAssociation]"}]}
                    },
                    AutoDelete->False,
                    
                    GridBoxItemSize->{
                    "Columns" -> {{Automatic}}, "Rows" -> {{Automatic}}}],
                    "Grid"], {-1, -0.5}], 
                  {GrayLevel[0], LineBox[{{-5, 0}, {5, 0}}]}, 
                  {GrayLevel[0.5], 
                   CircleBox[{-0.05138155016074792, 0.5}, 0.5], 
                   DiskBox[{-0.05138155016074792, 0.5}, 0.1]}, 
                  {GrayLevel[0.5], 
                   DiskBox[{-0.057579697675426394`, 1.9999871942677998`}, 
                    0.1]}, 
                  {GrayLevel[0.5], Thickness[0.01], 
                   LineBox[{{-0.05138155016074792, 
                    0.5}, {-0.057579697675426394`, 1.9999871942677998`}}]}, 
                  {RGBColor[0, 1, 0], 
                   DiskBox[{-0.055513648503866905`, 1.4999914628452}, 0.05]}, 
                  {RGBColor[0, 0, 1], Arrowheads[0.01], 
                   CircleBox[{-0.05138155016074792, 0.5}, 0.3333333333333333, 
                    NCache[{
                    Rational[-1, 2] 
                    Pi, -1.5083282965442768`}, {-1.5707963267948966`, \
-1.5083282965442768`}]], 
                   ArrowBox[{{-0.03390023002477048, 
                    0.16712537712013165`}, {-0.030572413333521155`, 
                    0.16731683100039935`}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-0.05138155016074792, 
                    0.5}, {-0.09302690366116108, 0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-0.05138155016074792, 
                    0.5}, {-0.020990277955547507`, 0.5}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.005], Arrowheads[0.04], 
                   ArrowBox[{{-0.05138155016074792, 
                    0.5}, {-0.06263563145596067, 0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-0.05138155016074792, 0.5}, 1.6, {
                    1.5749284368968435`, 1.5939231441858501`}], 
                   ArrowBox[{{-0.07238385481128892, 
                    2.099862151311595}, {-0.08838115957556462, 
                    2.099572139324498}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Dashing[{Small, Small}], 
                   Arrowheads[0.02], 
                   CircleBox[{-0.05138155016074792, 0.5}, 1.7, {
                    1.5749284368968435`, 1.5684062792106221`}], 
                   ArrowBox[{{-0.0643183444020456, 
                    2.1999507755681513`}, {-0.04731847313575496, 
                    2.1999951445239745`}}]}, 
                  {RGBColor[0, 0, 1], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-0.05138155016074792, 0.5}, 1.8, {
                    1.5749284368968435`, 1.5633602831467286`}], 
                   ArrowBox[{{-0.05599666653751075, 
                    2.2999940835182846`}, {-0.03799679494594594, 
                    2.299950234958689}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.003], Arrowheads[0.025], 
                   CircleBox[{-0.05138155016074792, 0.5}, 1.5, {
                    1.5749284368968435`, 1.573915894793929}], 
                   ArrowBox[{{-0.07106033761909274, 
                    1.9998709095532754`}, {-0.05606089456962156, 
                    1.9999927012275438`}}]}, 
                  {RGBColor[0, 0, 1], ArrowBox[{{0, 2.8}, {0, 2.8}}]}},
                 Frame->True,
                 FrameTicks->Automatic,
                 ImageSize->800,
                 PlotRange->{{-3, 3}, {-1, 3}}], 209->
                GraphicsBox[{InsetBox[
                   TagBox[GridBox[{
                    {"\<\"t\"\>", "10.4`"},
                    {"\<\"driveIntent\"\>", "0"},
                    {"\<\"motorTorque\"\>", "2.578055883121838`"},
                    {"\<\"vw\"\>", "0.08196478990254877`"},
                    {"\<\"controllerState\"\>", 
                    RowBox[{"\[LeftAssociation]", 
                    RowBox[{
                    RowBox[{"\<\"lastError\"\>", "\[Rule]", 
                    RowBox[{"-", "0.019497237756971895`"}]}], ",", 
                    
                    RowBox[{"\<\"lastInt\"\>", "\[Rule]", 
                    "0.07079005263393474`"}], ",", 
                    RowBox[{"\<\"initialized\"\>", "\[Rule]", "True"}]}], 
                    "\[RightAssociation]"}]}
                    },
                    AutoDelete->False,
                    
                    GridBoxItemSize->{
                    "Columns" -> {{Automatic}}, "Rows" -> {{Automatic}}}],
                    "Grid"], {-1, -0.5}], 
                  {GrayLevel[0], LineBox[{{-5, 0}, {5, 0}}]}, 
                  {GrayLevel[0.5], 
                   CircleBox[{-0.04727806131901714, 0.5}, 0.5], 
                   DiskBox[{-0.04727806131901714, 0.5}, 0.1]}, 
                  {GrayLevel[0.5], 
                   DiskBox[{-0.05494733575623665, 1.999980393948403}, 0.1]}, 
                  {GrayLevel[0.5], Thickness[0.01], 
                   LineBox[{{-0.04727806131901714, 
                    0.5}, {-0.05494733575623665, 1.999980393948403}}]}, 
                  {RGBColor[0, 1, 0], 
                   DiskBox[{-0.052390910943830145`, 1.4999869292989354`}, 
                    0.05]}, 
                  {RGBColor[0, 0, 1], Arrowheads[0.01], 
                   CircleBox[{-0.04727806131901714, 0.5}, 0.3333333333333333, 
                    NCache[{
                    Rational[-1, 2] 
                    Pi, -1.4934546503012414`}, {-1.5707963267948966`, \
-1.4934546503012414`}]], 
                   ArrowBox[{{-0.02484779796695542, 
                    0.16742219797907698`}, {-0.0215231968799244, 
                    0.16766312562576652`}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-0.04727806131901714, 
                    0.5}, {-0.0988391789814539, 0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-0.04727806131901714, 
                    0.5}, {-0.009673607762133489, 0.5}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.005], Arrowheads[0.04], 
                   ArrowBox[{{-0.04727806131901714, 
                    0.5}, {-0.061234725424570244`, 0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-0.04727806131901714, 0.5}, 1.6, {
                    1.5759091986960023`, 1.599412289370927}], 
                   ArrowBox[{{-0.07706188108715306, 
                    2.09972276475645}, {-0.09305735293701135, 
                    2.0993449460510245`}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Dashing[{Small, Small}], 
                   Arrowheads[0.02], 
                   CircleBox[{-0.04727806131901714, 0.5}, 1.7, {
                    1.5759091986960023`, 1.5678208314758912`}], 
                   ArrowBox[{{-0.05921962106941999, 
                    2.199958058056353}, {-0.04221972674077094, 
                    2.199992474469018}}]}, 
                  {RGBColor[0, 0, 1], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-0.04727806131901714, 0.5}, 1.8, {
                    1.5759091986960023`, 1.5615866660119921`}], 
                   ArrowBox[{{-0.048700671761686784`, 
                    2.299999437827559}, {-0.030700906251888108`, 
                    2.299923664472991}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.003], Arrowheads[0.025], 
                   CircleBox[{-0.04727806131901714, 0.5}, 1.5, {
                    1.5759091986960023`, 1.5746253066789837`}], 
                   ArrowBox[{{-0.06802086998557463, 
                    1.999856571772322}, {-0.0530215171109063, 
                    1.9999890041982198`}}]}, 
                  {RGBColor[0, 0, 1], ArrowBox[{{0, 2.8}, {0, 2.8}}]}},
                 Frame->True,
                 FrameTicks->Automatic,
                 ImageSize->800,
                 PlotRange->{{-3, 3}, {-1, 3}}], 210->
                GraphicsBox[{InsetBox[
                   TagBox[GridBox[{
                    {"\<\"t\"\>", "10.450000000000001`"},
                    {"\<\"driveIntent\"\>", "0"},
                    {"\<\"motorTorque\"\>", "3.0548573601587723`"},
                    {"\<\"vw\"\>", "0.08171471483218834`"},
                    {"\<\"controllerState\"\>", 
                    RowBox[{"\[LeftAssociation]", 
                    RowBox[{
                    RowBox[{"\<\"lastError\"\>", "\[Rule]", 
                    RowBox[{"-", "0.018475101475839566`"}]}], ",", 
                    
                    RowBox[{"\<\"lastInt\"\>", "\[Rule]", 
                    "0.0698413575257199`"}], ",", 
                    RowBox[{"\<\"initialized\"\>", "\[Rule]", "True"}]}], 
                    "\[RightAssociation]"}]}
                    },
                    AutoDelete->False,
                    
                    GridBoxItemSize->{
                    "Columns" -> {{Automatic}}, "Rows" -> {{Automatic}}}],
                    "Grid"], {-1, -0.5}], 
                  {GrayLevel[0], LineBox[{{-5, 0}, {5, 0}}]}, 
                  {GrayLevel[0.5], CircleBox[{-0.043186022576561, 0.5}, 0.5], 
                   DiskBox[{-0.043186022576561, 0.5}, 0.1]}, 
                  {GrayLevel[0.5], 
                   DiskBox[{-0.0522752070667661, 1.999972461988987}, 0.1]}, 
                  {GrayLevel[0.5], Thickness[0.01], 
                   LineBox[{{-0.043186022576561, 0.5}, {-0.0522752070667661, 
                    1.999972461988987}}]}, 
                  {RGBColor[0, 1, 0], 
                   DiskBox[{-0.04924547890336439, 1.4999816413259914`}, 
                    0.05]}, 
                  {RGBColor[0, 0, 1], Arrowheads[0.01], 
                   CircleBox[{-0.043186022576561, 0.5}, 0.3333333333333333, 
                    NCache[{
                    Rational[-1, 2] 
                    Pi, -1.4791506059901334`}, {-1.5707963267948966`, \
-1.4791506059901334`}]], 
                   ArrowBox[{{-0.016001008471686354`, 
                    0.1677770535931799}, {-0.01268019361721056, 
                    0.1680655102123102}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-0.043186022576561, 
                    0.5}, {-0.10428316977973645`, 0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-0.043186022576561, 0.5}, {
                    0.0013803847687087409`, 0.5}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.005], Arrowheads[0.04], 
                   ArrowBox[{{-0.043186022576561, 
                    0.5}, {-0.059716762434466705`, 0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-0.043186022576561, 0.5}, 1.6, {
                    1.5768558202031666`, 1.6047103361659385`}], 
                   ArrowBox[{{-0.08144479075718564, 
                    2.09954251792733}, {-0.09743803642490431, 
                    2.0990799601625305`}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Dashing[{Small, Small}], 
                   Arrowheads[0.02], 
                   CircleBox[{-0.043186022576561, 0.5}, 1.7, {
                    1.5768558202031666`, 1.5672757392712626`}], 
                   ArrowBox[{{-0.054200946713138506`, 
                    2.1999643147567145`}, {-0.03720103614992322, 
                    2.1999894646548466`}}]}, 
                  {RGBColor[0, 0, 1], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-0.043186022576561, 0.5}, 1.8, {
                    1.5768558202031666`, 1.5598843904245068`}], 
                   ArrowBox[{{-0.04154453733737681, 
                    2.299999251534903}, {-0.02354492689398104, 
                    2.299892837743511}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.003], Arrowheads[0.025], 
                   CircleBox[{-0.043186022576561, 0.5}, 1.5, {
                    1.5768558202031666`, 1.5753441631062401`}], 
                   ArrowBox[{{-0.06500700732737036, 
                    1.99984127314343}, {-0.05000775352808621, 
                    1.9999844879154}}]}, 
                  {RGBColor[0, 0, 1], ArrowBox[{{0, 2.8}, {0, 2.8}}]}},
                 Frame->True,
                 FrameTicks->Automatic,
                 ImageSize->800,
                 PlotRange->{{-3, 3}, {-1, 3}}], 211->
                GraphicsBox[{InsetBox[
                   TagBox[GridBox[{
                    {"\<\"t\"\>", "10.5`"},
                    {"\<\"driveIntent\"\>", "0"},
                    {"\<\"motorTorque\"\>", "3.5098247159191285`"},
                    {"\<\"vw\"\>", "0.08142348669653808`"},
                    {"\<\"controllerState\"\>", 
                    RowBox[{"\[LeftAssociation]", 
                    RowBox[{
                    RowBox[{"\<\"lastError\"\>", "\[Rule]", 
                    RowBox[{"-", "0.017480714997966326`"}]}], ",", 
                    
                    RowBox[{"\<\"lastInt\"\>", "\[Rule]", 
                    "0.0689430875834829`"}], ",", 
                    RowBox[{"\<\"initialized\"\>", "\[Rule]", "True"}]}], 
                    "\[RightAssociation]"}]}
                    },
                    AutoDelete->False,
                    
                    GridBoxItemSize->{
                    "Columns" -> {{Automatic}}, "Rows" -> {{Automatic}}}],
                    "Grid"], {-1, -0.5}], 
                  {GrayLevel[0], LineBox[{{-5, 0}, {5, 0}}]}, 
                  {GrayLevel[0.5], 
                   CircleBox[{-0.03910754660772399, 0.5}, 0.5], 
                   DiskBox[{-0.03910754660772399, 0.5}, 0.1]}, 
                  {GrayLevel[0.5], 
                   DiskBox[{-0.04955630410097919, 1.999963607380808}, 0.1]}, 
                  {GrayLevel[0.5], Thickness[0.01], 
                   LineBox[{{-0.03910754660772399, 
                    0.5}, {-0.04955630410097919, 1.999963607380808}}]}, 
                  {RGBColor[0, 1, 0], 
                   DiskBox[{-0.04607338493656079, 1.499975738253872}, 0.05]}, 
                  {RGBColor[0, 0, 1], Arrowheads[0.01], 
                   CircleBox[{-0.03910754660772399, 0.5}, 0.3333333333333333, 
                    NCache[{
                    Rational[-1, 2] 
                    Pi, -1.4655015853173228`}, {-1.5707963267948966`, \
-1.4655015853173228`}]], 
                   ArrowBox[{{-0.007390687618820814, 
                    0.1681790362756007}, {-0.004074119114528388, 
                    0.1685127904893009}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-0.03910754660772399, 
                    0.5}, {-0.10930404092610656`, 0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-0.03910754660772399, 0.5}, {
                    0.012124864209988997`, 0.5}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.005], Arrowheads[0.04], 
                   ArrowBox[{{-0.03910754660772399, 
                    0.5}, {-0.05807163010839357, 0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-0.03910754660772399, 0.5}, 1.6, {
                    1.577762221458747, 1.6097832551060067`}], 
                   ArrowBox[{{-0.08548013723569627, 
                    2.0993278534554607`}, {-0.10147083060673186`, 
                    2.0987841695516187`}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Dashing[{Small, Small}], 
                   Arrowheads[0.02], 
                   CircleBox[{-0.03910754660772399, 0.5}, 1.7, {
                    1.577762221458747, 1.566772008071558}], 
                   ArrowBox[{{-0.04926614431929703, 
                    2.1999696476386084`}, {-0.03226622324411606, 
                    2.1999862341485694`}}]}, 
                  {RGBColor[0, 0, 1], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-0.03910754660772399, 0.5}, 1.8, {
                    1.577762221458747, 1.5582631952591963`}], 
                   ArrowBox[{{-0.034547914719807525`, 
                    2.299994224923249}, {-0.01654850044780707, 
                    2.2998586304030533`}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.003], Arrowheads[0.025], 
                   CircleBox[{-0.03910754660772399, 0.5}, 1.5, {
                    1.577762221458747, 1.576063584070882}], 
                   ArrowBox[{{-0.06200754287408543, 
                    1.9998251865370846`}, {-0.04700839598805746, 
                    1.9999791920486996`}}]}, 
                  {RGBColor[0, 0, 1], ArrowBox[{{0, 2.8}, {0, 2.8}}]}},
                 Frame->True,
                 FrameTicks->Automatic,
                 ImageSize->800,
                 PlotRange->{{-3, 3}, {-1, 3}}], 212->
                GraphicsBox[{InsetBox[
                   TagBox[GridBox[{
                    {"\<\"t\"\>", "10.55`"},
                    {"\<\"driveIntent\"\>", "0"},
                    {"\<\"motorTorque\"\>", "3.940648889811646`"},
                    {"\<\"vw\"\>", "0.08109349529289675`"},
                    {"\<\"controllerState\"\>", 
                    RowBox[{"\[LeftAssociation]", 
                    RowBox[{
                    RowBox[{"\<\"lastError\"\>", "\[Rule]", 
                    RowBox[{"-", "0.016519807848994372`"}]}], ",", 
                    
                    RowBox[{"\<\"lastInt\"\>", "\[Rule]", 
                    "0.06809370503254429`"}], ",", 
                    RowBox[{"\<\"initialized\"\>", "\[Rule]", "True"}]}], 
                    "\[RightAssociation]"}]}
                    },
                    AutoDelete->False,
                    
                    GridBoxItemSize->{
                    "Columns" -> {{Automatic}}, "Rows" -> {{Automatic}}}],
                    "Grid"], {-1, -0.5}], 
                  {GrayLevel[0], LineBox[{{-5, 0}, {5, 0}}]}, 
                  {GrayLevel[0.5], 
                   CircleBox[{-0.035044630774943006`, 0.5}, 0.5], 
                   DiskBox[{-0.035044630774943006`, 0.5}, 0.1]}, 
                  {GrayLevel[0.5], 
                   DiskBox[{-0.046785135618670017`, 1.9999540528116235`}, 
                    0.1]}, 
                  {GrayLevel[0.5], Thickness[0.01], 
                   LineBox[{{-0.035044630774943006`, 
                    0.5}, {-0.046785135618670017`, 1.9999540528116235`}}]}, 
                  {RGBColor[0, 1, 0], 
                   DiskBox[{-0.04287163400409435, 1.4999693685410822`}, 
                    0.05]}, 
                  {RGBColor[0, 0, 1], Arrowheads[0.01], 
                   CircleBox[{-0.035044630774943006`, 0.5}, 
                    0.3333333333333333, 
                    NCache[{
                    Rational[-1, 2] 
                    Pi, -1.452576860100547}, {-1.5707963267948966`, \
-1.452576860100547}]], 
                   ArrowBox[{{0.0009581544985774598, 0.16861667126474228`}, {
                    0.004270132431388837, 0.1689932621454041}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-0.035044630774943006`, 
                    0.5}, {-0.11385760857117593`, 0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-0.035044630774943006`, 0.5}, {
                    0.02252111676107922, 0.5}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.005], Arrowheads[0.04], 
                   ArrowBox[{{-0.035044630774943006`, 
                    0.5}, {-0.0562918610351537, 0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-0.035044630774943006`, 0.5}, 1.6, {
                    1.5786234099425365`, 1.6146031042630788`}], 
                   ArrowBox[{{-0.08912517192474864, 
                    2.099085768515543}, {-0.1051130590924179, 
                    2.098465018495281}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Dashing[{Small, Small}], 
                   Arrowheads[0.02], 
                   CircleBox[{-0.035044630774943006`, 0.5}, 1.7, {
                    1.5786234099425365`, 1.5663101297343576`}], 
                   ArrowBox[{{-0.044418048276715985`, 
                    2.199974158346043}, {-0.027418121353895784`, 
                    2.1999828929592353`}}]}, 
                  {RGBColor[0, 0, 1], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-0.035044630774943006`, 0.5}, 1.8, {
                    1.5786234099425365`, 1.5567309161102494`}], 
                   ArrowBox[{{-0.02772691169996197, 
                    2.2999851252128556`}, {-0.00972772632675899, 
                    2.29982195073545}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.003], Arrowheads[0.025], 
                   CircleBox[{-0.035044630774943006`, 0.5}, 1.5, {
                    1.5786234099425365`, 1.5767757070497679`}], 
                   ArrowBox[{{-0.05901268112416396, 
                    1.999808498629894}, {-0.04401364771216747, 
                    1.999973185338718}}]}, 
                  {RGBColor[0, 0, 1], ArrowBox[{{0, 2.8}, {0, 2.8}}]}},
                 Frame->True,
                 FrameTicks->Automatic,
                 ImageSize->800,
                 PlotRange->{{-3, 3}, {-1, 3}}], 213->
                GraphicsBox[{InsetBox[
                   TagBox[GridBox[{
                    {"\<\"t\"\>", "10.6`"},
                    {"\<\"driveIntent\"\>", "0"},
                    {"\<\"motorTorque\"\>", "4.345519781025706`"},
                    {"\<\"vw\"\>", "0.08072726518257972`"},
                    {"\<\"controllerState\"\>", 
                    RowBox[{"\[LeftAssociation]", 
                    RowBox[{
                    RowBox[{"\<\"lastError\"\>", "\[Rule]", 
                    RowBox[{"-", "0.015597137708879799`"}]}], ",", 
                    
                    RowBox[{"\<\"lastInt\"\>", "\[Rule]", 
                    "0.06729141067289597`"}], ",", 
                    RowBox[{"\<\"initialized\"\>", "\[Rule]", "True"}]}], 
                    "\[RightAssociation]"}]}
                    },
                    AutoDelete->False,
                    
                    GridBoxItemSize->{
                    "Columns" -> {{Automatic}}, "Rows" -> {{Automatic}}}],
                    "Grid"], {-1, -0.5}], 
                  {GrayLevel[0], LineBox[{{-5, 0}, {5, 0}}]}, 
                  {GrayLevel[0.5], 
                   CircleBox[{-0.03099914934092269, 0.5}, 0.5], 
                   DiskBox[{-0.03099914934092269, 0.5}, 0.1]}, 
                  {GrayLevel[0.5], 
                   DiskBox[{-0.04395760413458569, 1.9999440251053906`}, 0.1]}, 
                  {GrayLevel[0.5], Thickness[0.01], 
                   LineBox[{{-0.03099914934092269, 
                    0.5}, {-0.04395760413458569, 1.9999440251053906`}}]}, 
                  {RGBColor[0, 1, 0], 
                   DiskBox[{-0.039638119203364686`, 1.4999626834035937`}, 
                    0.05]}, 
                  {RGBColor[0, 0, 1], Arrowheads[0.01], 
                   CircleBox[{-0.03099914934092269, 0.5}, 0.3333333333333333, 
                    NCache[{
                    Rational[-1, 2] 
                    Pi, -1.4404307333641253`}, {-1.5707963267948966`, \
-1.4404307333641253`}]], 
                   ArrowBox[{{0.009025905191022199, 0.16907839883013692`}, {
                    0.01233306481334686, 0.16949518864682223`}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-0.03099914934092269, 
                    0.5}, {-0.1179095449614368, 0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-0.03099914934092269, 0.5}, {
                    0.03253799493176473, 0.5}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.005], Arrowheads[0.04], 
                   ArrowBox[{{-0.03099914934092269, 
                    0.5}, {-0.054372400688749385`, 0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-0.03099914934092269, 0.5}, 1.6, {
                    1.5794354041179266`, 1.6191476012123875`}], 
                   ArrowBox[{{-0.09234614742681174, 
                    2.0988234880141867`}, {-0.1083310485133615, 
                    2.09813008774955}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Dashing[{Small, Small}], 
                   Arrowheads[0.02], 
                   CircleBox[{-0.03099914934092269, 0.5}, 1.7, {
                    1.5794354041179266`, 1.5658901361938773`}], 
                   ArrowBox[{{-0.039658587871487266`, 
                    2.199977945187624}, {-0.02265865877953346, 
                    2.199979539940759}}]}, 
                  {RGBColor[0, 0, 1], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-0.03099914934092269, 0.5}, 1.8, {
                    1.5794354041179266`, 1.5552936275566727`}], 
                   ArrowBox[{{-0.02109434069806674, 
                    2.2999727483397487`}, {-0.0030954084449324737`, 
                    2.299783704016683}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.003], Arrowheads[0.025], 
                   CircleBox[{-0.03099914934092269, 0.5}, 1.5, {
                    1.5794354041179266`, 1.5774736454979246`}], 
                   ArrowBox[{{-0.05601396778359079, 
                    1.9997914051154848`}, {-0.04101505296592096, 
                    1.999966560185451}}]}, 
                  {RGBColor[0, 0, 1], ArrowBox[{{0, 2.8}, {0, 2.8}}]}},
                 Frame->True,
                 FrameTicks->Automatic,
                 ImageSize->800,
                 PlotRange->{{-3, 3}, {-1, 3}}], 214->
                GraphicsBox[{InsetBox[
                   TagBox[GridBox[{
                    {"\<\"t\"\>", "10.65`"},
                    {"\<\"driveIntent\"\>", "0"},
                    {"\<\"motorTorque\"\>", "4.723086581378416`"},
                    {"\<\"vw\"\>", "0.08032741589032795`"},
                    {"\<\"controllerState\"\>", 
                    RowBox[{"\[LeftAssociation]", 
                    RowBox[{
                    RowBox[{"\<\"lastError\"\>", "\[Rule]", 
                    RowBox[{"-", "0.01471655583668046`"}]}], ",", 
                    
                    RowBox[{"\<\"lastInt\"\>", "\[Rule]", 
                    "0.06653419080989237`"}], ",", 
                    RowBox[{"\<\"initialized\"\>", "\[Rule]", "True"}]}], 
                    "\[RightAssociation]"}]}
                    },
                    AutoDelete->False,
                    
                    GridBoxItemSize->{
                    "Columns" -> {{Automatic}}, "Rows" -> {{Automatic}}}],
                    "Grid"], {-1, -0.5}], 
                  {GrayLevel[0], LineBox[{{-5, 0}, {5, 0}}]}, 
                  {GrayLevel[0.5], 
                   CircleBox[{-0.02697284776055763, 0.5}, 0.5], 
                   DiskBox[{-0.02697284776055763, 0.5}, 0.1]}, 
                  {GrayLevel[0.5], 
                   DiskBox[{-0.04107088405279221, 1.9999337469944138`}, 0.1]}, 
                  {GrayLevel[0.5], Thickness[0.01], 
                   LineBox[{{-0.02697284776055763, 
                    0.5}, {-0.04107088405279221, 1.9999337469944138`}}]}, 
                  {RGBColor[0, 1, 0], 
                   DiskBox[{-0.03637153862204735, 1.4999558313296093`}, 
                    0.05]}, 
                  {RGBColor[0, 0, 1], Arrowheads[0.01], 
                   CircleBox[{-0.02697284776055763, 0.5}, 0.3333333333333333, 
                    NCache[{
                    Rational[-1, 2] 
                    Pi, -1.429103729353544}, {-1.5707963267948966`, \
-1.429103729353544}]], 
                   ArrowBox[{{0.016797909322473544`, 0.16955298165138577`}, {
                    0.020100135912115614`, 0.17000720414035803`}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-0.02697284776055763, 
                    0.5}, {-0.12143457938812595`, 0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-0.02697284776055763, 0.5}, {
                    0.04215135278317528, 0.5}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.005], Arrowheads[0.04], 
                   ArrowBox[{{-0.02697284776055763, 
                    0.5}, {-0.05231037884439304, 0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-0.02697284776055763, 0.5}, 1.6, {
                    1.580195156034724, 1.6233996896613623`}], 
                   ArrowBox[{{-0.09511760966196339, 
                    2.098548182390947}, {-0.11109941785280641`, 
                    2.0977868193862763`}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Dashing[{Small, Small}], 
                   Arrowheads[0.02], 
                   CircleBox[{-0.02697284776055763, 0.5}, 1.7, {
                    1.580195156034724, 1.5655116514559875`}], 
                   ArrowBox[{{-0.03498886997924573, 
                    2.199981100891357}, {-0.017988941501328043`, 
                    2.199976261430825}}]}, 
                  {RGBColor[0, 0, 1], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-0.02697284776055763, 0.5}, 1.8, {
                    1.580195156034724, 1.5539557861381772`}], 
                   ArrowBox[{{-0.01465997060505647, 2.29995788646739}, {
                    0.003338692629408904, 2.299744762603628}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.003], Arrowheads[0.025], 
                   CircleBox[{-0.02697284776055763, 0.5}, 1.5, {
                    1.580195156034724, 1.5781514428380472`}], 
                   ArrowBox[{{-0.05300421500449695, 
                    1.9997741056303817`}, {-0.03800542235177842, 
                    1.9999594268839038`}}]}, 
                  {RGBColor[0, 0, 1], ArrowBox[{{0, 2.8}, {0, 2.8}}]}},
                 Frame->True,
                 FrameTicks->Automatic,
                 ImageSize->800,
                 PlotRange->{{-3, 3}, {-1, 3}}], 215->
                GraphicsBox[{InsetBox[
                   TagBox[GridBox[{
                    {"\<\"t\"\>", "10.700000000000001`"},
                    {"\<\"driveIntent\"\>", "0"},
                    {"\<\"motorTorque\"\>", "5.072418221407043`"},
                    {"\<\"vw\"\>", "0.07989662572918821`"},
                    {"\<\"controllerState\"\>", 
                    RowBox[{"\[LeftAssociation]", 
                    RowBox[{
                    RowBox[{"\<\"lastError\"\>", "\[Rule]", 
                    RowBox[{"-", "0.013881073703292862`"}]}], ",", 
                    
                    RowBox[{"\<\"lastInt\"\>", "\[Rule]", 
                    "0.06581986087953018`"}], ",", 
                    RowBox[{"\<\"initialized\"\>", "\[Rule]", "True"}]}], 
                    "\[RightAssociation]"}]}
                    },
                    AutoDelete->False,
                    
                    GridBoxItemSize->{
                    "Columns" -> {{Automatic}}, "Rows" -> {{Automatic}}}],
                    "Grid"], {-1, -0.5}], 
                  {GrayLevel[0], LineBox[{{-5, 0}, {5, 0}}]}, 
                  {GrayLevel[0.5], 
                   CircleBox[{-0.022967338870203288`, 0.5}, 0.5], 
                   DiskBox[{-0.022967338870203288`, 0.5}, 0.1]}, 
                  {GrayLevel[0.5], 
                   DiskBox[{-0.03812330152825161, 1.999923430310997}, 0.1]}, 
                  {GrayLevel[0.5], Thickness[0.01], 
                   LineBox[{{-0.022967338870203288`, 
                    0.5}, {-0.03812330152825161, 1.999923430310997}}]}, 
                  {RGBColor[0, 1, 0], 
                   DiskBox[{-0.03307131397556884, 1.4999489535406645`}, 
                    0.05]}, 
                  {RGBColor[0, 0, 1], Arrowheads[0.01], 
                   CircleBox[{-0.022967338870203288`, 0.5}, 
                    0.3333333333333333, 
                    NCache[{
                    Rational[-1, 2] 
                    Pi, -1.4186237801526853`}, {-1.5707963267948966`, \
-1.4186237801526853`}]], 
                   ArrowBox[{{0.024264019153391547`, 0.17002983478750977`}, {
                    0.02756130426254238, 0.17051863886666557`}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-0.022967338870203288`, 
                    0.5}, {-0.12441570329834414`, 0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-0.022967338870203288`, 0.5}, {
                    0.051343478232323464`, 0.5}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.005], Arrowheads[0.04], 
                   ArrowBox[{{-0.022967338870203288`, 
                    0.5}, {-0.05010488619581739, 0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-0.022967338870203288`, 0.5}, 1.6, {
                    1.580900473827824, 1.6273471054529958`}], 
                   ArrowBox[{{-0.09742168780469794, 
                    2.0982667330344897`}, {-0.11340036607216172`, 
                    2.0974422892834315`}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Dashing[{Small, Small}], 
                   Arrowheads[0.02], 
                   CircleBox[{-0.022967338870203288`, 0.5}, 1.7, {
                    1.580900473827824, 1.565173941609997}], 
                   ArrowBox[{{-0.030409260286929758`, 
                    2.1999837110412637`}, {-0.013409334412748913`, 
                    2.199973130537889}}]}, 
                  {RGBColor[0, 0, 1], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-0.022967338870203288`, 0.5}, 1.8, {
                    1.580900473827824, 1.552720372597785}], 
                   ArrowBox[{{-0.008430779331517292, 2.2999413013864585`}, {
                    0.0095676068717103, 2.2997059418987233`}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.003], Arrowheads[0.025], 
                   CircleBox[{-0.022967338870203288`, 0.5}, 1.5, {
                    1.580900473827824, 1.5788040229050895`}], 
                   ArrowBox[{{-0.049977423188208046`, 
                    1.9997567987327594`}, {-0.03497875466613536, 
                    1.9999519078592412`}}]}, 
                  {RGBColor[0, 0, 1], ArrowBox[{{0, 2.8}, {0, 2.8}}]}},
                 Frame->True,
                 FrameTicks->Automatic,
                 ImageSize->800,
                 PlotRange->{{-3, 3}, {-1, 3}}], 216->
                GraphicsBox[{InsetBox[
                   TagBox[GridBox[{
                    {"\<\"t\"\>", "10.75`"},
                    {"\<\"driveIntent\"\>", "0"},
                    {"\<\"motorTorque\"\>", "5.392964294624953`"},
                    {"\<\"vw\"\>", "0.07943759918199556`"},
                    {"\<\"controllerState\"\>", 
                    RowBox[{"\[LeftAssociation]", 
                    RowBox[{
                    RowBox[{"\<\"lastError\"\>", "\[Rule]", 
                    RowBox[{"-", "0.013092930064325655`"}]}], ",", 
                    
                    RowBox[{"\<\"lastInt\"\>", "\[Rule]", 
                    "0.06514610572783995`"}], ",", 
                    RowBox[{"\<\"initialized\"\>", "\[Rule]", "True"}]}], 
                    "\[RightAssociation]"}]}
                    },
                    AutoDelete->False,
                    
                    GridBoxItemSize->{
                    "Columns" -> {{Automatic}}, "Rows" -> {{Automatic}}}],
                    "Grid"], {-1, -0.5}], 
                  {GrayLevel[0], LineBox[{{-5, 0}, {5, 0}}]}, 
                  {GrayLevel[0.5], 
                   CircleBox[{-0.018984100794669112`, 0.5}, 0.5], 
                   DiskBox[{-0.018984100794669112`, 0.5}, 0.1]}, 
                  {GrayLevel[0.5], 
                   DiskBox[{-0.03511421700208858, 1.999913270609716}, 0.1]}, 
                  {GrayLevel[0.5], Thickness[0.01], 
                   LineBox[{{-0.018984100794669112`, 
                    0.5}, {-0.03511421700208858, 1.999913270609716}}]}, 
                  {RGBColor[0, 1, 0], 
                   DiskBox[{-0.02973751159961542, 1.4999421804064772`}, 
                    0.05]}, 
                  {RGBColor[0, 0, 1], Arrowheads[0.01], 
                   CircleBox[{-0.018984100794669112`, 0.5}, 
                    0.3333333333333333, 
                    NCache[{
                    Rational[-1, 2] 
                    Pi, -1.409007397956148}, {-1.5707963267948966`, \
-1.409007397956148}]], 
                   ArrowBox[{{0.03141814370717312, 0.17049927942372023`}, {
                    0.03471057590519952, 0.17101976836714383`}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-0.018984100794669112`, 
                    0.5}, {-0.12684338668716816`, 0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-0.018984100794669112`, 0.5}, {
                    0.060102528236922026`, 0.5}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.005], Arrowheads[0.04], 
                   ArrowBox[{{-0.018984100794669112`, 
                    0.5}, {-0.047756757655577034`, 0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-0.018984100794669112`, 0.5}, 1.6, {
                    1.58154994485725, 1.6309819461402157`}], 
                   ArrowBox[{{-0.09924739003959854, 
                    2.097985545741758}, {-0.11522296603640418`, 
                    2.0971030276149953`}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Dashing[{Small, Small}], 
                   Arrowheads[0.02], 
                   CircleBox[{-0.018984100794669112`, 0.5}, 1.7, {
                    1.58154994485725, 1.5648759626148385`}], 
                   ArrowBox[{{-0.025919462450500457`, 
                    2.1999858531054612`}, {-0.008919540483811628, 
                    2.1999702069818015`}}]}, 
                  {RGBColor[0, 0, 1], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-0.018984100794669112`, 0.5}, 1.8, {
                    1.58154994485725, 1.5515890321093337`}], 
                   ArrowBox[{{-0.0024112045221840878`, 2.2999237036911153`}, {
                    0.015586903891034787`, 2.299667982055307}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.003], Arrowheads[0.025], 
                   CircleBox[{-0.018984100794669112`, 0.5}, 1.5, {
                    1.58154994485725, 1.5794271377201805`}], 
                   ArrowBox[{{-0.04692870048889249, 
                    1.9997396771933218`}, {-0.03193015645398117, 
                    1.999944132173884}}]}, 
                  {RGBColor[0, 0, 1], ArrowBox[{{0, 2.8}, {0, 2.8}}]}},
                 Frame->True,
                 FrameTicks->Automatic,
                 ImageSize->800,
                 PlotRange->{{-3, 3}, {-1, 3}}], 217->
                GraphicsBox[{InsetBox[
                   TagBox[GridBox[{
                    {"\<\"t\"\>", "10.8`"},
                    {"\<\"driveIntent\"\>", "0"},
                    {"\<\"motorTorque\"\>", "5.684516780421745`"},
                    {"\<\"vw\"\>", "0.078953037746015`"},
                    {"\<\"controllerState\"\>", 
                    RowBox[{"\[LeftAssociation]", 
                    RowBox[{
                    RowBox[{"\<\"lastError\"\>", "\[Rule]", 
                    RowBox[{"-", "0.012353657786277266`"}]}], ",", 
                    
                    RowBox[{"\<\"lastInt\"\>", "\[Rule]", 
                    "0.06451051654022978`"}], ",", 
                    RowBox[{"\<\"initialized\"\>", "\[Rule]", "True"}]}], 
                    "\[RightAssociation]"}]}
                    },
                    AutoDelete->False,
                    GridBoxItemSize->{
                    "Columns" -> {{Automatic}}, "Rows" -> {{Automatic}}}],
                    "Grid"], {-1, -0.5}], 
                  {GrayLevel[0], LineBox[{{-5, 0}, {5, 0}}]}, 
                  {GrayLevel[0.5], 
                   CircleBox[{-0.015024476396390854`, 0.5}, 0.5], 
                   DiskBox[{-0.015024476396390854`, 0.5}, 0.1]}, 
                  {GrayLevel[0.5], 
                   DiskBox[{-0.03204391122753633, 1.9999034431716023`}, 0.1]}, 
                  {GrayLevel[0.5], Thickness[0.01], 
                   LineBox[{{-0.015024476396390854`, 
                    0.5}, {-0.03204391122753633, 1.9999034431716023`}}]}, 
                  {RGBColor[0, 1, 0], 
                   DiskBox[{-0.026370766283821172`, 1.4999356287810683`}, 
                    0.05]}, 
                  {RGBColor[0, 0, 1], Arrowheads[0.01], 
                   CircleBox[{-0.015024476396390854`, 0.5}, 
                    0.3333333333333333, 
                    NCache[{
                    Rational[-1, 2] 
                    Pi, -1.4002608233822442`}, {-1.5707963267948966`, \
-1.4002608233822442`}]], 
                   ArrowBox[{{0.03825780604525422, 0.1709527245380691}, {
                    0.04154555986701396, 0.17150199070882}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-0.015024476396390854`, 
                    0.5}, {-0.12871481200482576`, 0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-0.015024476396390854`, 0.5}, {
                    0.06842197200164116, 0.5}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.005], Arrowheads[0.04], 
                   ArrowBox[{{-0.015024476396390854`, 
                    0.5}, {-0.045268363606793746`, 0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-0.015024476396390854`, 0.5}, 1.6, {
                    1.582142860146766, 1.6343002478301574`}], 
                   ArrowBox[{{-0.10058991215280574`, 
                    2.0977104106200892`}, {-0.11656247173913396`, 
                    2.0967748856685393`}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Dashing[{Small, Small}], 
                   Arrowheads[0.02], 
                   CircleBox[{-0.015024476396390854`, 0.5}, 1.7, {
                    1.582142860146766, 1.564616405659302}], 
                   ArrowBox[{{-0.02151859467107228, 
                    2.1999875959629334`}, {-0.004518677338084406, 
                    2.1999675373918604`}}]}, 
                  {RGBColor[0, 0, 1], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-0.015024476396390854`, 0.5}, 1.8, {
                    1.582142860146766, 1.550562211366645}], 
                   ArrowBox[{{0.0033966098086175064`, 2.2999057374159984`}, {
                    0.02139444615335303, 2.299631535087201}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.003], Arrowheads[0.025], 
                   CircleBox[{-0.015024476396390854`, 0.5}, 1.5, {
                    1.582142860146766, 1.5800173133871127`}], 
                   ArrowBox[{{-0.04385418103679064, 
                    1.999722923786373}, {-0.02885576027827808, 
                    1.9999362305065433`}}]}, 
                  {RGBColor[0, 0, 1], ArrowBox[{{0, 2.8}, {0, 2.8}}]}},
                 Frame->True,
                 FrameTicks->Automatic,
                 ImageSize->800,
                 PlotRange->{{-3, 3}, {-1, 3}}], 218->
                GraphicsBox[{InsetBox[
                   TagBox[GridBox[{
                    {"\<\"t\"\>", "10.85`"},
                    {"\<\"driveIntent\"\>", "0"},
                    {"\<\"motorTorque\"\>", "5.947172843228613`"},
                    {"\<\"vw\"\>", "0.07844561412601614`"},
                    {"\<\"controllerState\"\>", 
                    RowBox[{"\[LeftAssociation]", 
                    RowBox[{
                    RowBox[{"\<\"lastError\"\>", "\[Rule]", 
                    RowBox[{"-", "0.011664149817531628`"}]}], ",", 
                    
                    RowBox[{"\<\"lastInt\"\>", "\[Rule]", 
                    "0.0639106244489516`"}], ",", 
                    RowBox[{"\<\"initialized\"\>", "\[Rule]", "True"}]}], 
                    "\[RightAssociation]"}]}
                    },
                    AutoDelete->False,
                    
                    GridBoxItemSize->{
                    "Columns" -> {{Automatic}}, "Rows" -> {{Automatic}}}],
                    "Grid"], {-1, -0.5}], 
                  {GrayLevel[0], LineBox[{{-5, 0}, {5, 0}}]}, 
                  {GrayLevel[0.5], 
                   CircleBox[{-0.011089674096467947`, 0.5}, 0.5], 
                   DiskBox[{-0.011089674096467947`, 0.5}, 0.1]}, 
                  {GrayLevel[0.5], 
                   DiskBox[{-0.028913475480374665`, 1.9998941002964932`}, 
                    0.1]}, 
                  {GrayLevel[0.5], Thickness[0.01], 
                   LineBox[{{-0.011089674096467947`, 
                    0.5}, {-0.028913475480374665`, 1.9998941002964932`}}]}, 
                  {RGBColor[0, 1, 0], 
                   DiskBox[{-0.022972208352405758`, 1.4999294001976622`}, 
                    0.05]}, 
                  {RGBColor[0, 0, 1], Arrowheads[0.01], 
                   CircleBox[{-0.011089674096467947`, 0.5}, 
                    0.3333333333333333, 
                    NCache[{
                    Rational[-1, 2] 
                    Pi, -1.392381141498038}, {-1.5707963267948966`, \
-1.392381141498038}]], 
                   ArrowBox[{{0.04478371524643931, 0.17138278274800178`}, {
                    0.04806703900351029, 0.17195793805318504`}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-0.011089674096467947`, 
                    0.5}, {-0.13003313096104022`, 0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-0.011089674096467947`, 0.5}, {
                    0.07630004669165233, 0.5}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.005], Arrowheads[0.04], 
                   ArrowBox[{{-0.011089674096467947`, 
                    0.5}, {-0.042643410172919935`, 0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-0.011089674096467947`, 0.5}, 1.6, {
                    1.582679140693588, 1.6373015725190534`}], 
                   ArrowBox[{{-0.10144996497047176`, 
                    2.0974464053085367`}, {-0.11741964480692708`, 
                    2.096462945805105}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Dashing[{Small, Small}], 
                   Arrowheads[0.02], 
                   CircleBox[{-0.011089674096467947`, 0.5}, 1.7, {
                    1.582679140693588, 1.5643937399362635`}], 
                   ArrowBox[{{-0.017205263246076585`, 
                    2.1999889998377498`}, {-0.00020535080087306402`, 
                    2.1999651559683207`}}]}, 
                  {RGBColor[0, 0, 1], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-0.011089674096467947`, 0.5}, 1.8, {
                    1.582679140693588, 1.5496392915645403`}], 
                   ArrowBox[{{0.008992572674333675, 2.299887969670512}, {
                    0.026990148287239467`, 2.2995971569012954`}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.003], Arrowheads[0.025], 
                   CircleBox[{-0.011089674096467947`, 0.5}, 1.5, {
                    1.582679140693588, 1.5805717948275924`}], 
                   ArrowBox[{{-0.0407509427895168, 
                    1.999706707706383}, {-0.02575264261124532, 
                    1.9999283307392839`}}]}, 
                  {RGBColor[0, 0, 1], ArrowBox[{{0, 2.8}, {0, 2.8}}]}},
                 Frame->True,
                 FrameTicks->Automatic,
                 ImageSize->800,
                 PlotRange->{{-3, 3}, {-1, 3}}], 219->
                GraphicsBox[{InsetBox[
                   TagBox[GridBox[{
                    {"\<\"t\"\>", "10.9`"},
                    {"\<\"driveIntent\"\>", "0"},
                    {"\<\"motorTorque\"\>", "6.181298944495377`"},
                    {"\<\"vw\"\>", "0.0779179496428927`"},
                    {"\<\"controllerState\"\>", 
                    RowBox[{"\[LeftAssociation]", 
                    RowBox[{
                    RowBox[{"\<\"lastError\"\>", "\[Rule]", 
                    RowBox[{"-", "0.011024723771143876`"}]}], ",", 
                    
                    RowBox[{"\<\"lastInt\"\>", "\[Rule]", 
                    "0.06334393087535108`"}], ",", 
                    RowBox[{"\<\"initialized\"\>", "\[Rule]", "True"}]}], 
                    "\[RightAssociation]"}]}
                    },
                    AutoDelete->False,
                    
                    GridBoxItemSize->{
                    "Columns" -> {{Automatic}}, "Rows" -> {{Automatic}}}],
                    "Grid"], {-1, -0.5}], 
                  {GrayLevel[0], LineBox[{{-5, 0}, {5, 0}}]}, 
                  {GrayLevel[0.5], 
                   CircleBox[{-0.007180769903463291, 0.5}, 0.5], 
                   DiskBox[{-0.007180769903463291, 0.5}, 0.1]}, 
                  {GrayLevel[0.5], 
                   DiskBox[{-0.025724706531131954`, 1.9998853697580856`}, 
                    0.1]}, 
                  {GrayLevel[0.5], Thickness[0.01], 
                   LineBox[{{-0.007180769903463291, 
                    0.5}, {-0.025724706531131954`, 1.9998853697580856`}}]}, 
                  {RGBColor[0, 1, 0], 
                   DiskBox[{-0.019543394321909068`, 1.4999235798387236`}, 
                    0.05]}, 
                  {RGBColor[0, 0, 1], Arrowheads[0.01], 
                   CircleBox[{-0.007180769903463291, 0.5}, 0.3333333333333333,
                     NCache[{
                    Rational[-1, 2] 
                    Pi, -1.3853573584600352`}, {-1.5707963267948966`, \
-1.3853573584600352`}]], 
                   ArrowBox[{{0.05099935828789131, 0.17178332797563645`}, {
                    0.054278561323461864`, 0.17238153025775516`}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-0.007180769903463291, 
                    0.5}, {-0.13080674879337084`, 0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-0.007180769903463291, 0.5}, {
                    0.0837392295093647, 0.5}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.005], Arrowheads[0.04], 
                   ArrowBox[{{-0.007180769903463291, 
                    0.5}, {-0.039886749380542846`, 0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-0.007180769903463291, 0.5}, 1.6, {
                    1.5831592661408864`, 1.639988608681426}], 
                   ArrowBox[{{-0.10183312566285126`, 
                    2.0971978373229785`}, {-0.11780010525942287`, 
                    2.096171470314329}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Dashing[{Small, Small}], 
                   Arrowheads[0.02], 
                   CircleBox[{-0.007180769903463291, 0.5}, 1.7, {
                    1.5831592661408864`, 1.564206252709642}], 
                   ArrowBox[{{-0.01297763272458675, 2.1999901165540443`}, {
                    0.004022274951409975, 2.1999630854186156`}}]}, 
                  {RGBColor[0, 0, 1], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-0.007180769903463291, 0.5}, 1.8, {
                    1.5831592661408864`, 1.5488187164492453`}], 
                   ArrowBox[{{0.01437841321869543, 2.2998708847089855`}, {
                    0.03237574413863125, 2.2995653036766512`}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.003], Arrowheads[0.025], 
                   CircleBox[{-0.007180769903463291, 0.5}, 1.5, {
                    1.5831592661408864`, 1.5810884899987492`}], 
                   ArrowBox[{{-0.03761692581666404, 
                    1.999691181681491}, {-0.022618742152015682`, 
                    1.999920554233741}}]}, 
                  {RGBColor[0, 0, 1], ArrowBox[{{0, 2.8}, {0, 2.8}}]}},
                 Frame->True,
                 FrameTicks->Automatic,
                 ImageSize->800,
                 PlotRange->{{-3, 3}, {-1, 3}}], 220->
                GraphicsBox[{InsetBox[
                   TagBox[GridBox[{
                    {"\<\"t\"\>", "10.950000000000001`"},
                    {"\<\"driveIntent\"\>", "0"},
                    {"\<\"motorTorque\"\>", "6.387496464926802`"},
                    {"\<\"vw\"\>", "0.0773725947097543`"},
                    {"\<\"controllerState\"\>", 
                    RowBox[{"\[LeftAssociation]", 
                    RowBox[{
                    RowBox[{"\<\"lastError\"\>", "\[Rule]", 
                    RowBox[{"-", "0.01043518465869106`"}]}], ",", 
                    
                    RowBox[{"\<\"lastInt\"\>", "\[Rule]", 
                    "0.06280793468835252`"}], ",", 
                    RowBox[{"\<\"initialized\"\>", "\[Rule]", "True"}]}], 
                    "\[RightAssociation]"}]}
                    },
                    AutoDelete->False,
                    
                    GridBoxItemSize->{
                    "Columns" -> {{Automatic}}, "Rows" -> {{Automatic}}}],
                    "Grid"], {-1, -0.5}], 
                  {GrayLevel[0], LineBox[{{-5, 0}, {5, 0}}]}, 
                  {GrayLevel[0.5], 
                   CircleBox[{-0.0032987104929070557`, 0.5}, 0.5], 
                   DiskBox[{-0.0032987104929070557`, 0.5}, 0.1]}, 
                  {GrayLevel[0.5], 
                   DiskBox[{-0.022480006846740477`, 1.9998773542760708`}, 
                    0.1]}, 
                  {GrayLevel[0.5], Thickness[0.01], 
                   LineBox[{{-0.0032987104929070557`, 
                    0.5}, {-0.022480006846740477`, 1.9998773542760708`}}]}, 
                  {RGBColor[0, 1, 0], 
                   DiskBox[{-0.01608624139546267, 1.4999182361840473`}, 
                    0.05]}, 
                  {RGBColor[0, 0, 1], Arrowheads[0.01], 
                   CircleBox[{-0.0032987104929070557`, 0.5}, 
                    0.3333333333333333, 
                    NCache[{
                    Rational[-1, 2] 
                    Pi, -1.3791714328470925`}, {-1.5707963267948966`, \
-1.3791714328470925`}]], 
                   ArrowBox[{{0.05691061552051435, 0.1721495033218332}, {
                    0.060186055404606253`, 0.17276797893535994`}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-0.0032987104929070557`, 
                    0.5}, {-0.1310486397914431, 0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-0.0032987104929070557`, 0.5}, {
                    0.09074572939021047, 0.5}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.005], Arrowheads[0.04], 
                   ArrowBox[{{-0.0032987104929070557`, 
                    0.5}, {-0.03700419990832556, 0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-0.0032987104929070557`, 0.5}, 1.6, {
                    1.5835842062279588`, 1.642366787443059}], 
                   ArrowBox[{{-0.10174921698250855`, 
                    2.096968220651852}, {-0.1177137105446842, 
                    2.09590388424966}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Dashing[{Small, Small}], 
                   Arrowheads[0.02], 
                   CircleBox[{-0.0032987104929070557`, 0.5}, 1.7, {
                    1.5835842062279588`, 1.564052086590201}], 
                   ArrowBox[{{-0.008833492366823766, 2.199990990031891}, {
                    0.008166410939867684, 2.1999613380869967`}}]}, 
                  {RGBColor[0, 0, 1], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-0.0032987104929070557`, 0.5}, 1.8, {
                    1.5835842062279588`, 1.5480981147561432`}], 
                   ArrowBox[{{0.019557456926406987`, 2.2998548818198925`}, {
                    0.03755456297144457, 2.299536331960888}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.003], Arrowheads[0.025], 
                   CircleBox[{-0.0032987104929070557`, 0.5}, 1.5, {
                    1.5835842062279588`, 1.5815659141684282`}], 
                   ArrowBox[{{-0.03445085172738248, 
                    1.999676479810398}, {-0.019452779280027012`, 
                    1.9999130128316178`}}]}, 
                  {RGBColor[0, 0, 1], ArrowBox[{{0, 2.8}, {0, 2.8}}]}},
                 Frame->True,
                 FrameTicks->Automatic,
                 ImageSize->800,
                 PlotRange->{{-3, 3}, {-1, 3}}], 221->
                GraphicsBox[{InsetBox[
                   TagBox[GridBox[{
                    {"\<\"t\"\>", "11.`"},
                    {"\<\"driveIntent\"\>", "0"},
                    {"\<\"motorTorque\"\>", "6.566568997579652`"},
                    {"\<\"vw\"\>", "0.07681201221507196`"},
                    {"\<\"controllerState\"\>", 
                    RowBox[{"\[LeftAssociation]", 
                    RowBox[{
                    RowBox[{"\<\"lastError\"\>", "\[Rule]", 
                    RowBox[{"-", "0.009894885383345334`"}]}], ",", 
                    
                    RowBox[{"\<\"lastInt\"\>", "\[Rule]", 
                    "0.06230015628189824`"}], ",", 
                    RowBox[{"\<\"initialized\"\>", "\[Rule]", "True"}]}], 
                    "\[RightAssociation]"}]}
                    },
                    AutoDelete->False,
                    
                    GridBoxItemSize->{
                    "Columns" -> {{Automatic}}, "Rows" -> {{Automatic}}}],
                    "Grid"], {-1, -0.5}], 
                  {GrayLevel[0], LineBox[{{-5, 0}, {5, 0}}]}, 
                  {GrayLevel[0.5], 
                   CircleBox[{0.0005556828118242399, 0.5}, 0.5], 
                   DiskBox[{0.0005556828118242399, 0.5}, 0.1]}, 
                  {GrayLevel[0.5], 
                   DiskBox[{-0.019182290386998927`, 1.999870131849422}, 0.1]}, 
                  {GrayLevel[0.5], Thickness[0.01], 
                   LineBox[{{0.0005556828118242399, 
                    0.5}, {-0.019182290386998927`, 1.999870131849422}}]}, 
                  {RGBColor[0, 1, 0], 
                   DiskBox[{-0.012602965987391206`, 1.499913421232948}, 
                    0.05]}, 
                  {RGBColor[0, 0, 1], Arrowheads[0.01], 
                   CircleBox[{0.0005556828118242399, 0.5}, 0.3333333333333333,
                     NCache[{
                    Rational[-1, 2] 
                    Pi, -1.373799256867507}, {-1.5707963267948966`, \
-1.373799256867507}]], 
                   ArrowBox[{{0.06252540208985918, 0.17247768777728645`}, {
                    0.06579747216516395, 0.1731137506209759}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{0.0005556828118242399, 
                    0.5}, {-0.1307756971397688, 0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{0.0005556828118242399, 0.5}, {
                    0.09732900100690672, 0.5}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.005], Arrowheads[0.04], 
                   ArrowBox[{{0.0005556828118242399, 
                    0.5}, {-0.034002378944686315`, 0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{0.0005556828118242399, 0.5}, 1.6, {
                    1.5839553553604604`, 1.64444391625741}], 
                   ArrowBox[{{-0.10121171761872681`, 
                    2.096760281385283}, {-0.11717396597957812`, 
                    2.0956627869933717`}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Dashing[{Small, Small}], 
                   Arrowheads[0.02], 
                   CircleBox[{0.0005556828118242399, 0.5}, 1.7, {
                    1.5839553553604604`, 1.5639292739711466`}], 
                   ArrowBox[{{-0.00477031867502214, 2.199991656952516}, {
                    0.012229580861832052`, 2.1999599172052022`}}]}, 
                  {RGBColor[0, 0, 1], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{0.0005556828118242399, 0.5}, 1.8, {
                    1.5839553553604604`, 1.5474744164850178`}], 
                   ArrowBox[{{0.024534412090907252`, 2.2998402763973695`}, {
                    0.042531315956528586`, 2.299510501837125}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.003], Arrowheads[0.025], 
                   CircleBox[{0.0005556828118242399, 0.5}, 1.5, {
                    1.5839553553604604`, 1.5820031347587868`}], 
                   ArrowBox[{{-0.031252144860060074`, 
                    1.9996627161127918`}, {-0.01625417726333807, 
                    1.9999058065772841`}}]}, 
                  {RGBColor[0, 0, 1], ArrowBox[{{0, 2.8}, {0, 2.8}}]}},
                 Frame->True,
                 FrameTicks->Automatic,
                 ImageSize->800,
                 PlotRange->{{-3, 3}, {-1, 3}}], 222->
                GraphicsBox[{InsetBox[
                   TagBox[GridBox[{
                    {"\<\"t\"\>", "11.05`"},
                    {"\<\"driveIntent\"\>", "0"},
                    {"\<\"motorTorque\"\>", "6.719491437902841`"},
                    {"\<\"vw\"\>", "0.07623856364281087`"},
                    {"\<\"controllerState\"\>", 
                    RowBox[{"\[LeftAssociation]", 
                    RowBox[{
                    RowBox[{"\<\"lastError\"\>", "\[Rule]", 
                    RowBox[{"-", "0.009402784665535673`"}]}], ",", 
                    
                    RowBox[{"\<\"lastInt\"\>", "\[Rule]", 
                    "0.06181815869197301`"}], ",", 
                    RowBox[{"\<\"initialized\"\>", "\[Rule]", "True"}]}], 
                    "\[RightAssociation]"}]}
                    },
                    AutoDelete->False,
                    
                    GridBoxItemSize->{
                    "Columns" -> {{Automatic}}, "Rows" -> {{Automatic}}}],
                    "Grid"], {-1, -0.5}], 
                  {GrayLevel[0], LineBox[{{-5, 0}, {5, 0}}]}, 
                  {GrayLevel[0.5], 
                   CircleBox[{0.004381709298304103, 0.5}, 0.5], 
                   DiskBox[{0.004381709298304103, 0.5}, 0.1]}, 
                  {GrayLevel[0.5], 
                   DiskBox[{-0.01583489426634814, 1.9998637567926993`}, 0.1]}, 
                  {GrayLevel[0.5], Thickness[0.01], 
                   LineBox[{{0.004381709298304103, 
                    0.5}, {-0.01583489426634814, 1.9998637567926993`}}]}, 
                  {RGBColor[0, 1, 0], 
                   DiskBox[{-0.00909602641146406, 1.499909171195133}, 0.05]}, 
                  {RGBColor[0, 0, 1], Arrowheads[0.01], 
                   CircleBox[{0.004381709298304103, 0.5}, 0.3333333333333333, 
                    NCache[{
                    Rational[-1, 2] 
                    Pi, -1.3692115836578114`}, {-1.5707963267948966`, \
-1.3692115836578114`}]], 
                   ArrowBox[{{0.06785333651371052, 0.1727654302343049}, {
                    0.07112245411763087, 0.17341649752004828`}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{0.004381709298304103, 
                    0.5}, {-0.1300081194597527, 0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{0.004381709298304103, 0.5}, {0.1035012832417143,
                     0.5}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.005], Arrowheads[0.04], 
                   ArrowBox[{{0.004381709298304103, 
                    0.5}, {-0.03088854551634252, 0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{0.004381709298304103, 0.5}, 1.6, {
                    1.584274470575033, 1.646229831621063}], 
                   ArrowBox[{{-0.10023720604174478`, 
                    2.0965759870902083`}, {-0.11619746891580357`, 
                    2.0954499872390895`}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Dashing[{Small, Small}], 
                   Arrowheads[0.02], 
                   CircleBox[{0.004381709298304103, 0.5}, 1.7, {
                    1.584274470575033, 1.5638357686075282`}], 
                   ArrowBox[{{-0.0007853338274590197, 2.1999921475304927`}, {
                    0.01621456266724135, 2.199958818201532}}]}, 
                  {RGBColor[0, 0, 1], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{0.004381709298304103, 0.5}, 1.8, {
                    1.584274470575033, 1.5469439625866837`}], 
                   ArrowBox[{{0.02931516745002359, 2.299827303566816}, {
                    0.047311893853455675`, 2.2994879825256023`}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.003], Arrowheads[0.025], 
                   CircleBox[{0.004381709298304103, 0.5}, 1.5, {
                    1.584274470575033, 1.5823997172082742`}], 
                   ArrowBox[{{-0.02802085577002124, 
                    1.999649983755207}, {-0.013022985758131015`, 
                    1.9998990221311541`}}]}, 
                  {RGBColor[0, 0, 1], ArrowBox[{{0, 2.8}, {0, 2.8}}]}},
                 Frame->True,
                 FrameTicks->Automatic,
                 ImageSize->800,
                 PlotRange->{{-3, 3}, {-1, 3}}], 223->
                GraphicsBox[{InsetBox[
                   TagBox[GridBox[{
                    {"\<\"t\"\>", "11.1`"},
                    {"\<\"driveIntent\"\>", "0"},
                    {"\<\"motorTorque\"\>", "6.847380964898209`"},
                    {"\<\"vw\"\>", "0.07565449775238545`"},
                    {"\<\"controllerState\"\>", 
                    RowBox[{"\[LeftAssociation]", 
                    RowBox[{
                    RowBox[{"\<\"lastError\"\>", "\[Rule]", 
                    RowBox[{"-", "0.008957502135870966`"}]}], ",", 
                    
                    RowBox[{"\<\"lastInt\"\>", "\[Rule]", 
                    "0.061359565888598144`"}], ",", 
                    RowBox[{"\<\"initialized\"\>", "\[Rule]", "True"}]}], 
                    "\[RightAssociation]"}]}
                    },
                    AutoDelete->False,
                    
                    GridBoxItemSize->{
                    "Columns" -> {{Automatic}}, "Rows" -> {{Automatic}}}],
                    "Grid"], {-1, -0.5}], 
                  {GrayLevel[0], LineBox[{{-5, 0}, {5, 0}}]}, 
                  {GrayLevel[0.5], 
                   CircleBox[{0.008178783495883996, 0.5}, 0.5], 
                   DiskBox[{0.008178783495883996, 0.5}, 0.1]}, 
                  {GrayLevel[0.5], 
                   DiskBox[{-0.012441496464276412`, 1.999858261321504}, 0.1]}, 
                  {GrayLevel[0.5], Thickness[0.01], 
                   LineBox[{{0.008178783495883996, 
                    0.5}, {-0.012441496464276412`, 1.999858261321504}}]}, 
                  {RGBColor[0, 1, 0], 
                   DiskBox[{-0.005568069810889611, 1.4999055075476693`}, 
                    0.05]}, 
                  {RGBColor[0, 0, 1], Arrowheads[0.01], 
                   CircleBox[{0.008178783495883996, 0.5}, 0.3333333333333333, 
                    NCache[{
                    Rational[-1, 2] 
                    Pi, -1.3653748978479503`}, {-1.5707963267948966`, \
-1.3653748978479503`}]], 
                   ArrowBox[{{0.07290543670580728, 0.17301135879765883`}, {
                    0.07617203231430517, 0.17367496383785175`}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{0.008178783495883996, 
                    0.5}, {-0.12876883580208018`, 0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{0.008178783495883996, 0.5}, {
                    0.10927716379488152`, 0.5}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.005], Arrowheads[0.04], 
                   ArrowBox[{{0.008178783495883996, 
                    0.5}, {-0.027670455503082662`, 0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{0.008178783495883996, 0.5}, 1.6, {
                    1.5845436131093302`, 1.6477360720066137`}], 
                   ArrowBox[{{-0.09884483948760328, 
                    2.096416594790811}, {-0.11480338823085312`, 
                    2.0952665562335895`}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Dashing[{Small, Small}], 
                   Arrowheads[0.02], 
                   CircleBox[{0.008178783495883996, 0.5}, 1.7, {
                    1.5845436131093302`, 1.5637694743526567`}], 
                   ArrowBox[{{0.0031244400941006136`, 2.1999924863400944`}, {
                    0.020124334341901293`, 2.199958030015737}}]}, 
                  {RGBColor[0, 0, 1], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{0.008178783495883996, 0.5}, 1.8, {
                    1.5845436131093302`, 1.5465026077487847`}], 
                   ArrowBox[{{0.033906601681180704`, 2.299816123766932}, {
                    0.05190317657080654, 2.29946885981676}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.003], Arrowheads[0.025], 
                   CircleBox[{0.008178783495883996, 0.5}, 1.5, {
                    1.5845436131093302`, 1.5827556722428813`}], 
                   ArrowBox[{{-0.02475758747026774, 
                    1.9996383548934658`}, {-0.009759807054984325, 
                    1.999892731820862}}]}, 
                  {RGBColor[0, 0, 1], ArrowBox[{{0, 2.8}, {0, 2.8}}]}},
                 Frame->True,
                 FrameTicks->Automatic,
                 ImageSize->800,
                 PlotRange->{{-3, 3}, {-1, 3}}], 224->
                GraphicsBox[{InsetBox[
                   TagBox[GridBox[{
                    {"\<\"t\"\>", "11.15`"},
                    {"\<\"driveIntent\"\>", "0"},
                    {"\<\"motorTorque\"\>", "6.951469978154694`"},
                    {"\<\"vw\"\>", "0.07506194163656442`"},
                    {"\<\"controllerState\"\>", 
                    RowBox[{"\[LeftAssociation]", 
                    RowBox[{
                    RowBox[{"\<\"lastError\"\>", "\[Rule]", 
                    RowBox[{"-", "0.008557370387214847`"}]}], ",", 
                    
                    RowBox[{"\<\"lastInt\"\>", "\[Rule]", 
                    "0.060922078389967445`"}], ",", 
                    RowBox[{"\<\"initialized\"\>", "\[Rule]", "True"}]}], 
                    "\[RightAssociation]"}]}
                    },
                    AutoDelete->False,
                    
                    GridBoxItemSize->{
                    "Columns" -> {{Automatic}}, "Rows" -> {{Automatic}}}],
                    "Grid"], {-1, -0.5}], 
                  {GrayLevel[0], LineBox[{{-5, 0}, {5, 0}}]}, 
                  {GrayLevel[0.5], 
                   CircleBox[{0.011946429301921052`, 0.5}, 0.5], 
                   DiskBox[{0.011946429301921052`, 0.5}, 0.1]}, 
                  {GrayLevel[0.5], 
                   DiskBox[{-0.009006039688231916, 1.9998536575423673`}, 
                    0.1]}, 
                  {GrayLevel[0.5], Thickness[0.01], 
                   LineBox[{{0.011946429301921052`, 
                    0.5}, {-0.009006039688231916, 1.9998536575423673`}}]}, 
                  {RGBColor[0, 1, 0], 
                   DiskBox[{-0.0020218833581809263`, 1.4999024383615782`}, 
                    0.05]}, 
                  {RGBColor[0, 0, 1], Arrowheads[0.01], 
                   CircleBox[{0.011946429301921052`, 0.5}, 0.3333333333333333,
                     NCache[{
                    Rational[-1, 2] 
                    Pi, -1.3622522274502558`}, {-1.5707963267948966`, \
-1.3622522274502558`}]], 
                   ArrowBox[{{0.07769384302585647, 0.17321507271643516`}, {
                    0.08095835049151837, 0.17388887500603123`}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{0.011946429301921052`, 
                    0.5}, {-0.12708297026117282`, 0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{0.011946429301921052`, 0.5}, {
                    0.11467317114594221`, 0.5}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.005], Arrowheads[0.04], 
                   ArrowBox[{{0.011946429301921052`, 
                    0.5}, {-0.024356228417151668`, 0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{0.011946429301921052`, 0.5}, 1.6, {
                    1.584765093729883, 1.6489755718618464`}], 
                   ArrowBox[{{-0.09705586909962952, 
                    2.0962827127245287`}, {-0.11301298011158364`, 
                    2.0951128944369515`}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Dashing[{Small, Small}], 
                   Arrowheads[0.02], 
                   CircleBox[{0.011946429301921052`, 0.5}, 1.7, {
                    1.584765093729883, 1.5637282710924743`}], 
                   ArrowBox[{{0.006962131137131083, 2.199992693152475}, {
                    0.023962023950788433`, 2.199957536377081}}]}, 
                  {RGBColor[0, 0, 1], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{0.011946429301921052`, 0.5}, 1.8, {
                    1.584765093729883, 1.546145816073468}], 
                   ArrowBox[{{0.03831640524607452, 2.2998068297372094`}, {
                    0.05631285508933176, 2.2994531447811726`}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.003], Arrowheads[0.025], 
                   CircleBox[{0.011946429301921052`, 0.5}, 1.5, {
                    1.584765093729883, 1.583071404893883}], 
                   ArrowBox[{{-0.021463424808032566`, 
                    1.999627881058615}, {-0.006465725455392118, 
                    1.9998869932622234`}}]}, 
                  {RGBColor[0, 0, 1], ArrowBox[{{0, 2.8}, {0, 2.8}}]}},
                 Frame->True,
                 FrameTicks->Automatic,
                 ImageSize->800,
                 PlotRange->{{-3, 3}, {-1, 3}}], 225->
                GraphicsBox[{InsetBox[
                   TagBox[GridBox[{
                    {"\<\"t\"\>", "11.200000000000001`"},
                    {"\<\"driveIntent\"\>", "0"},
                    {"\<\"motorTorque\"\>", "7.033081028878235`"},
                    {"\<\"vw\"\>", "0.07446289397296502`"},
                    {"\<\"controllerState\"\>", 
                    RowBox[{"\[LeftAssociation]", 
                    RowBox[{
                    RowBox[{"\<\"lastError\"\>", "\[Rule]", 
                    RowBox[{"-", "0.008200483830782942`"}]}], ",", 
                    
                    RowBox[{"\<\"lastInt\"\>", "\[Rule]", 
                    "0.060503486354935604`"}], ",", 
                    RowBox[{"\<\"initialized\"\>", "\[Rule]", "True"}]}], 
                    "\[RightAssociation]"}]}
                    },
                    AutoDelete->False,
                    
                    GridBoxItemSize->{
                    "Columns" -> {{Automatic}}, "Rows" -> {{Automatic}}}],
                    "Grid"], {-1, -0.5}], 
                  {GrayLevel[0], LineBox[{{-5, 0}, {5, 0}}]}, 
                  {GrayLevel[0.5], 
                   CircleBox[{0.015684273717607857`, 0.5}, 0.5], 
                   DiskBox[{0.015684273717607857`, 0.5}, 0.1]}, 
                  {GrayLevel[0.5], 
                   DiskBox[{-0.00553266142124163, 1.9998499397150749`}, 0.1]}, 
                  {GrayLevel[0.5], Thickness[0.01], 
                   LineBox[{{0.015684273717607857`, 
                    0.5}, {-0.00553266142124163, 1.9998499397150749`}}]}, 
                  {RGBColor[0, 1, 0], 
                   DiskBox[{0.0015396502917081993`, 1.49989995981005}, 0.05]}, 
                  {RGBColor[0, 0, 1], Arrowheads[0.01], 
                   CircleBox[{0.015684273717607857`, 0.5}, 0.3333333333333333,
                     NCache[{
                    Rational[-1, 2] 
                    Pi, -1.3598038959285494`}, {-1.5707963267948966`, \
-1.3598038959285494`}]], 
                   ArrowBox[{{0.08223156742485088, 0.1733770234485133}, {
                    0.08549441541651777, 0.1740588163071607}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{0.015684273717607857`, 
                    0.5}, {-0.12497734685995684`, 0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{0.015684273717607857`, 0.5}, {
                    0.11970739467335574`, 0.5}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.005], Arrowheads[0.04], 
                   ArrowBox[{{0.015684273717607857`, 
                    0.5}, {-0.020954225904208968`, 0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{0.015684273717607857`, 0.5}, 1.6, {
                    1.5849414219166098`, 1.6499623772226717`}], 
                   ArrowBox[{{-0.09489319169421026, 
                    2.096174371471707}, {-0.11084914055399878`, 
                    2.094988807193571}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Dashing[{Small, Small}], 
                   Arrowheads[0.02], 
                   CircleBox[{0.015684273717607857`, 0.5}, 1.7, {
                    1.5849414219166098`, 1.5637100379417939`}], 
                   ArrowBox[{{0.010730971776574495`, 2.199992783749355}, {
                    0.027730863946387272`, 2.19995731701236}}]}, 
                  {RGBColor[0, 0, 1], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{0.015684273717607857`, 0.5}, 1.8, {
                    1.5849414219166098`, 1.545868749533953}], 
                   ArrowBox[{{0.042552914893335125`, 2.2997994544174554`}, {
                    0.060549266051570225`, 2.299440783260976}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.003], Arrowheads[0.025], 
                   CircleBox[{0.015684273717607857`, 0.5}, 1.5, {
                    1.5849414219166098`, 1.5833476655457628`}], 
                   ArrowBox[{{-0.018139867288155243`, 
                    1.9996185940048963`}, {-0.003142240090331979, 
                    1.9998818494727642`}}]}, 
                  {RGBColor[0, 0, 1], ArrowBox[{{0, 2.8}, {0, 2.8}}]}},
                 Frame->True,
                 FrameTicks->Automatic,
                 ImageSize->800,
                 PlotRange->{{-3, 3}, {-1, 3}}], 226->
                GraphicsBox[{InsetBox[
                   TagBox[GridBox[{
                    {"\<\"t\"\>", "11.25`"},
                    {"\<\"driveIntent\"\>", "0"},
                    {"\<\"motorTorque\"\>", "7.093603758943127`"},
                    {"\<\"vw\"\>", "0.07385922028432387`"},
                    {"\<\"controllerState\"\>", 
                    RowBox[{"\[LeftAssociation]", 
                    RowBox[{
                    RowBox[{"\<\"lastError\"\>", "\[Rule]", 
                    RowBox[{"-", "0.00788474424978517`"}]}], ",", 
                    
                    RowBox[{"\<\"lastInt\"\>", "\[Rule]", 
                    "0.060101680316566215`"}], ",", 
                    RowBox[{"\<\"initialized\"\>", "\[Rule]", "True"}]}], 
                    "\[RightAssociation]"}]}
                    },
                    AutoDelete->False,
                    
                    GridBoxItemSize->{
                    "Columns" -> {{Automatic}}, "Rows" -> {{Automatic}}}],
                    "Grid"], {-1, -0.5}], 
                  {GrayLevel[0], LineBox[{{-5, 0}, {5, 0}}]}, 
                  {GrayLevel[0.5], 
                   CircleBox[{0.019392040298368933`, 0.5}, 0.5], 
                   DiskBox[{0.019392040298368933`, 0.5}, 0.1]}, 
                  {GrayLevel[0.5], 
                   DiskBox[{-0.0020256301224411584`, 1.9998470866704197`}, 
                    0.1]}, 
                  {GrayLevel[0.5], Thickness[0.01], 
                   LineBox[{{0.019392040298368933`, 
                    0.5}, {-0.0020256301224411584`, 1.9998470866704197`}}]}, 
                  {RGBColor[0, 1, 0], 
                   DiskBox[{0.005113593351162205, 1.4998980577802798`}, 
                    0.05]}, 
                  {RGBColor[0, 0, 1], Arrowheads[0.01], 
                   CircleBox[{0.019392040298368933`, 0.5}, 0.3333333333333333,
                     NCache[{
                    Rational[-1, 2] 
                    Pi, -1.3579882140266029`}, {-1.5707963267948966`, \
-1.3579882140266029`}]], 
                   ArrowBox[{{0.08653226741382164, 0.17349839048789284`}, {
                    0.08979387210889911, 0.17418610651349914`}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{0.019392040298368933`, 
                    0.5}, {-0.12248003488049361`, 0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{0.019392040298368933`, 0.5}, {
                    0.12439913336696615`, 0.5}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.005], Arrowheads[0.04], 
                   ArrowBox[{{0.019392040298368933`, 
                    0.5}, {-0.017472941811896403`, 0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{0.019392040298368933`, 0.5}, 1.6, {
                    1.5850752589540809`, 1.650711383214155}], 
                   ArrowBox[{{-0.09238093815181894, 
                    2.096091100560483}, {-0.10833599454121987`, 
                    2.0948935855147255`}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Dashing[{Small, Small}], 
                   Arrowheads[0.02], 
                   CircleBox[{0.019392040298368933`, 0.5}, 1.7, {
                    1.5850752589540809`, 1.5637126737876048`}], 
                   ArrowBox[{{0.01443425743849509, 2.199992770687309}, {
                    0.031434149701733666`, 2.1999573487594084`}}]}, 
                  {RGBColor[0, 0, 1], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{0.019392040298368933`, 0.5}, 1.8, {
                    1.5850752589540809`, 1.5456663491821745`}], 
                   ArrowBox[{{0.04662496096396468, 2.2997939793298627`}, {
                    0.06462123915841365, 2.299431665712949}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.003], Arrowheads[0.025], 
                   CircleBox[{0.019392040298368933`, 0.5}, 1.5, {
                    1.5850752589540809`, 1.5835855032498665`}], 
                   ArrowBox[{{-0.014788765590799662`, 1.9996105069346397`}, {
                    0.00020879857076444228`, 1.9998773293962477`}}]}, 
                  {RGBColor[0, 0, 1], ArrowBox[{{0, 2.8}, {0, 2.8}}]}},
                 Frame->True,
                 FrameTicks->Automatic,
                 ImageSize->800,
                 PlotRange->{{-3, 3}, {-1, 3}}], 227->
                GraphicsBox[{InsetBox[
                   TagBox[GridBox[{
                    {"\<\"t\"\>", "11.3`"},
                    {"\<\"driveIntent\"\>", "0"},
                    {"\<\"motorTorque\"\>", "7.134473840633974`"},
                    {"\<\"vw\"\>", "0.07325265002412451`"},
                    {"\<\"controllerState\"\>", 
                    RowBox[{"\[LeftAssociation]", 
                    RowBox[{
                    RowBox[{"\<\"lastError\"\>", "\[Rule]", 
                    RowBox[{"-", "0.007607902988413422`"}]}], ",", 
                    
                    RowBox[{"\<\"lastInt\"\>", "\[Rule]", 
                    "0.059714659723623956`"}], ",", 
                    RowBox[{"\<\"initialized\"\>", "\[Rule]", "True"}]}], 
                    "\[RightAssociation]"}]}
                    },
                    AutoDelete->False,
                    
                    GridBoxItemSize->{
                    "Columns" -> {{Automatic}}, "Rows" -> {{Automatic}}}],
                    "Grid"], {-1, -0.5}], 
                  {GrayLevel[0], LineBox[{{-5, 0}, {5, 0}}]}, 
                  {GrayLevel[0.5], CircleBox[{0.02306954241457612, 0.5}, 0.5],
                    DiskBox[{0.02306954241457612, 0.5}, 0.1]}, 
                  {GrayLevel[0.5], 
                   DiskBox[{0.0015107125077457179`, 1.9998450642826573`}, 
                    0.1]}, 
                  {GrayLevel[0.5], Thickness[0.01], 
                   LineBox[{{0.02306954241457612, 0.5}, {
                    0.0015107125077457179`, 1.9998450642826573`}}]}, 
                  {RGBColor[0, 1, 0], 
                   DiskBox[{0.008696989143355852, 1.4998967095217717`}, 
                    0.05]}, 
                  {RGBColor[0, 0, 1], Arrowheads[0.01], 
                   CircleBox[{0.02306954241457612, 0.5}, 0.3333333333333333, 
                    NCache[{
                    Rational[-1, 2] 
                    Pi, -1.3567621115758772`}, {-1.5707963267948966`, \
-1.3567621115758772`}]], 
                   ArrowBox[{{0.09061004338645524, 0.17358095668362317`}, {
                    0.09387080241981033, 0.17427267125280643`}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{0.02306954241457612, 
                    0.5}, {-0.11961993439810337`, 0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{0.02306954241457612, 0.5}, {
                    0.12876857323623211`, 0.5}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.005], Arrowheads[0.04], 
                   ArrowBox[{{0.02306954241457612, 
                    0.5}, {-0.01392090357644735, 0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{0.02306954241457612, 0.5}, 1.6, {
                    1.5851693749358633`, 1.6512380934689315`}], 
                   ArrowBox[{{-0.0895440980427281, 
                    2.0960320071925103`}, {-0.10549852147554806`, 
                    2.094826088621435}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Dashing[{Small, Small}], 
                   Arrowheads[0.02], 
                   CircleBox[{0.02306954241457612, 0.5}, 1.7, {
                    1.5851693749358633`, 1.5637341152814488`}], 
                   ArrowBox[{{0.01807530917131721, 2.199992663994263}, {
                    0.03507520219014709, 2.199957606569456}}]}, 
                  {RGBColor[0, 0, 1], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{0.02306954241457612, 0.5}, 1.8, {
                    1.5851693749358633`, 1.5455334091545634`}], 
                   ArrowBox[{{0.05054172750012913, 2.2997903430807227`}, {
                    0.06853795736957144, 2.2994256370412978`}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.003], Arrowheads[0.025], 
                   CircleBox[{0.02306954241457612, 0.5}, 1.5, {
                    1.5851693749358633`, 1.5837862214919065`}], 
                   ArrowBox[{{-0.011412261968733118`, 1.9996036160154027`}, {
                    0.003585248334586382, 1.999873448756329}}]}, 
                  {RGBColor[0, 0, 1], ArrowBox[{{0, 2.8}, {0, 2.8}}]}},
                 Frame->True,
                 FrameTicks->Automatic,
                 ImageSize->800,
                 PlotRange->{{-3, 3}, {-1, 3}}], 228->
                GraphicsBox[{InsetBox[
                   TagBox[GridBox[{
                    {"\<\"t\"\>", "11.35`"},
                    {"\<\"driveIntent\"\>", "0"},
                    {"\<\"motorTorque\"\>", "7.157153890862689`"},
                    {"\<\"vw\"\>", "0.07264477530720069`"},
                    {"\<\"controllerState\"\>", 
                    RowBox[{"\[LeftAssociation]", 
                    RowBox[{
                    RowBox[{"\<\"lastError\"\>", "\[Rule]", 
                    RowBox[{"-", "0.007367599753891917`"}]}], ",", 
                    
                    RowBox[{"\<\"lastInt\"\>", "\[Rule]", 
                    "0.059340539458968694`"}], ",", 
                    RowBox[{"\<\"initialized\"\>", "\[Rule]", "True"}]}], 
                    "\[RightAssociation]"}]}
                    },
                    AutoDelete->False,
                    
                    GridBoxItemSize->{
                    "Columns" -> {{Automatic}}, "Rows" -> {{Automatic}}}],
                    "Grid"], {-1, -0.5}], 
                  {GrayLevel[0], LineBox[{{-5, 0}, {5, 0}}]}, 
                  {GrayLevel[0.5], 
                   CircleBox[{0.026716676409235248`, 0.5}, 0.5], 
                   DiskBox[{0.026716676409235248`, 0.5}, 0.1]}, 
                  {GrayLevel[0.5], 
                   DiskBox[{0.005072003335208346, 1.9998438279125992`}, 0.1]}, 
                  {GrayLevel[0.5], Thickness[0.01], 
                   LineBox[{{0.026716676409235248`, 0.5}, {
                    0.005072003335208346, 1.9998438279125992`}}]}, 
                  {RGBColor[0, 1, 0], 
                   DiskBox[{0.01228689435988398, 1.4998958852750661`}, 0.05]}, 
                  {RGBColor[0, 0, 1], Arrowheads[0.01], 
                   CircleBox[{0.026716676409235248`, 0.5}, 0.3333333333333333,
                     NCache[{
                    Rational[-1, 2] 
                    Pi, -1.356081710069016}, {-1.5707963267948966`, \
-1.356081710069016}]], 
                   ArrowBox[{{0.09447925773910976, 0.17362698689593953`}, {
                    0.097739545374088, 0.17432091993019805`}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{0.026716676409235248`, 
                    0.5}, {-0.11642640140801852`, 0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{0.026716676409235248`, 0.5}, {
                    0.13283649322395508`, 0.5}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.005], Arrowheads[0.04], 
                   ArrowBox[{{0.026716676409235248`, 
                    0.5}, {-0.010306584593298692`, 0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{0.026716676409235248`, 0.5}, 1.6, {
                    1.5852266096493672`, 1.6515584012748068`}], 
                   ArrowBox[{{-0.0864081797727421, 
                    2.095995854290921}, {-0.10236221612199836`, 
                    2.0947848254554313`}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Dashing[{Small, Small}], 
                   Arrowheads[0.02], 
                   CircleBox[{0.026716676409235248`, 0.5}, 1.7, {
                    1.5852266096493672`, 1.56377235239766}], 
                   ArrowBox[{{0.021657440352591626`, 2.1999924717864263`}, {
                    0.03865733469948875, 2.1999580643885297`}}]}, 
                  {RGBColor[0, 0, 1], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{0.026716676409235248`, 0.5}, 1.8, {
                    1.5852266096493672`, 1.5454646435890191`}], 
                   ArrowBox[{{0.05431262503053984, 2.2997884496850434`}, {
                    0.0723088297782157, 2.2994225061255564`}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.003], Arrowheads[0.025], 
                   CircleBox[{0.026716676409235248`, 0.5}, 1.5, {
                    1.5852266096493672`, 1.5839513365588165`}], 
                   ArrowBox[{{-0.008012734654307258, 1.999597902108155}, {
                    0.0069847308911229165`, 1.999870211160309}}]}, 
                  {RGBColor[0, 0, 1], ArrowBox[{{0, 2.8}, {0, 2.8}}]}},
                 Frame->True,
                 FrameTicks->Automatic,
                 ImageSize->800,
                 PlotRange->{{-3, 3}, {-1, 3}}], 229->
                GraphicsBox[{InsetBox[
                   TagBox[GridBox[{
                    {"\<\"t\"\>", "11.4`"},
                    {"\<\"driveIntent\"\>", "0"},
                    {"\<\"motorTorque\"\>", "7.163116317309293`"},
                    {"\<\"vw\"\>", "0.07203705110941938`"},
                    {"\<\"controllerState\"\>", 
                    RowBox[{"\[LeftAssociation]", 
                    RowBox[{
                    RowBox[{"\<\"lastError\"\>", "\[Rule]", 
                    RowBox[{"-", "0.00716139804491946`"}]}], ",", 
                    
                    RowBox[{"\<\"lastInt\"\>", "\[Rule]", 
                    "0.058977554503998435`"}], ",", 
                    RowBox[{"\<\"initialized\"\>", "\[Rule]", "True"}]}], 
                    "\[RightAssociation]"}]}
                    },
                    AutoDelete->False,
                    
                    GridBoxItemSize->{
                    "Columns" -> {{Automatic}}, "Rows" -> {{Automatic}}}],
                    "Grid"], {-1, -0.5}], 
                  {GrayLevel[0], LineBox[{{-5, 0}, {5, 0}}]}, 
                  {GrayLevel[0.5], CircleBox[{0.03033341473038602, 0.5}, 0.5],
                    DiskBox[{0.03033341473038602, 0.5}, 0.1]}, 
                  {GrayLevel[0.5], 
                   DiskBox[{0.008653903852126273, 1.999843324753649}, 0.1]}, 
                  {GrayLevel[0.5], Thickness[0.01], 
                   LineBox[{{0.03033341473038602, 0.5}, {0.008653903852126273,
                     1.999843324753649}}]}, 
                  {RGBColor[0, 1, 0], 
                   DiskBox[{0.015880407478212857`, 1.4998955498357658`}, 
                    0.05]}, 
                  {RGBColor[0, 0, 1], Arrowheads[0.01], 
                   CircleBox[{0.03033341473038602, 0.5}, 0.3333333333333333, 
                    NCache[{
                    Rational[-1, 2] 
                    Pi, -1.3559028372756177`}, {-1.5707963267948966`, \
-1.3559028372756177`}]], 
                   ArrowBox[{{0.09815437422844492, 0.17363911299931523`}, {
                    0.1014145376855263, 0.17433362919922585`}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{0.03033341473038602, 
                    0.5}, {-0.11292891161579983`, 0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{0.03033341473038602, 0.5}, {
                    0.13662399917879275`, 0.5}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.005], Arrowheads[0.04], 
                   ArrowBox[{{0.03033341473038602, 
                    0.5}, {-0.0066383271673931266`, 0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{0.03033341473038602, 0.5}, 1.6, {
                    1.5852498372744217`, 1.6516883920726362`}], 
                   ArrowBox[{{-0.08299890526966901, 
                    2.0959811356163973`}, {-0.09895278406152867, 
                    2.0947680329132314`}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Dashing[{Small, Small}], 
                   Arrowheads[0.02], 
                   CircleBox[{0.03033341473038602, 0.5}, 1.7, {
                    1.5852498372744217`, 1.5638254416852144`}], 
                   ArrowBox[{{0.025183927291739603`, 2.1999922007994974`}, {
                    0.0421838234413441, 2.199958695913928}}]}, 
                  {RGBColor[0, 0, 1], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{0.03033341473038602, 0.5}, 1.8, {
                    1.5852498372744217`, 1.5454547466227033`}], 
                   ArrowBox[{{0.05794717579600093, 2.2997881764807246`}, {
                    0.07594337692106434, 2.299422054813424}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.003], Arrowheads[0.025], 
                   CircleBox[{0.03033341473038602, 0.5}, 1.5, {
                    1.5852498372744217`, 1.5840825386093536`}], 
                   ArrowBox[{{-0.004592746353414036, 1.9995933326311999`}, {
                    0.010404683335427205`, 1.999867609379237}}]}, 
                  {RGBColor[0, 0, 1], ArrowBox[{{0, 2.8}, {0, 2.8}}]}},
                 Frame->True,
                 FrameTicks->Automatic,
                 ImageSize->800,
                 PlotRange->{{-3, 3}, {-1, 3}}], 230->
                GraphicsBox[{InsetBox[
                   TagBox[GridBox[{
                    {"\<\"t\"\>", "11.450000000000001`"},
                    {"\<\"driveIntent\"\>", "0"},
                    {"\<\"motorTorque\"\>", "7.153828039877254`"},
                    {"\<\"vw\"\>", "0.07143079676627939`"},
                    {"\<\"controllerState\"\>", 
                    RowBox[{"\[LeftAssociation]", 
                    RowBox[{
                    RowBox[{"\<\"lastError\"\>", "\[Rule]", 
                    RowBox[{"-", "0.0069868172512464535`"}]}], ",", 
                    
                    RowBox[{"\<\"lastInt\"\>", "\[Rule]", 
                    "0.058624062916802284`"}], ",", 
                    RowBox[{"\<\"initialized\"\>", "\[Rule]", "True"}]}], 
                    "\[RightAssociation]"}]}
                    },
                    AutoDelete->False,
                    
                    GridBoxItemSize->{
                    "Columns" -> {{Automatic}}, "Rows" -> {{Automatic}}}],
                    "Grid"], {-1, -0.5}], 
                  {GrayLevel[0], LineBox[{{-5, 0}, {5, 0}}]}, 
                  {GrayLevel[0.5], CircleBox[{0.03391979910730371, 0.5}, 0.5],
                    DiskBox[{0.03391979910730371, 0.5}, 0.1]}, 
                  {GrayLevel[0.5], 
                   DiskBox[{0.012252140704426838`, 1.9998434960286144`}, 
                    0.1]}, 
                  {GrayLevel[0.5], Thickness[0.01], 
                   LineBox[{{0.03391979910730371, 0.5}, {
                    0.012252140704426838`, 1.9998434960286144`}}]}, 
                  {RGBColor[0, 1, 0], 
                   DiskBox[{0.019474693505385793`, 1.4998956640190761`}, 
                    0.05]}, 
                  {RGBColor[0, 0, 1], Arrowheads[0.01], 
                   CircleBox[{0.03391979910730371, 0.5}, 0.3333333333333333, 
                    NCache[{
                    Rational[-1, 2] 
                    Pi, -1.356181485598579}, {-1.5707963267948966`, \
-1.356181485598579}]], 
                   ArrowBox[{{0.10164981605972115`, 0.17362022747306116`}, {
                    0.10491017291600689`, 0.17431383520694083`}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{0.03391979910730371, 
                    0.5}, {-0.10915676169024137`, 0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{0.03391979910730371, 0.5}, {0.1401522852183784, 
                    0.5}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.005], Arrowheads[0.04], 
                   ArrowBox[{{0.03391979910730371, 
                    0.5}, {-0.0029242755791667047`, 0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{0.03391979910730371, 0.5}, 1.6, {
                    1.5852419347992028`, 1.6516441667606316`}], 
                   ArrowBox[{{-0.07934193801826622, 
                    2.0959861462128355`}, {-0.09529587044436948, 
                    2.094773749076123}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Dashing[{Small, Small}], 
                   Arrowheads[0.02], 
                   CircleBox[{0.03391979910730371, 0.5}, 1.7, {
                    1.5852419347992028`, 1.5638915173500199`}], 
                   ArrowBox[{{0.028657983565148468`, 2.1999918568326144`}, {
                    0.04565788189149984, 2.199959475226557}}]}, 
                  {RGBColor[0, 0, 1], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{0.03391979910730371, 0.5}, 1.8, {
                    1.5852419347992028`, 1.5454984456887737`}], 
                   ArrowBox[{{0.061454911084141195`, 2.299789381457848}, {
                    0.07945112819119668, 2.299424046208079}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.003], Arrowheads[0.025], 
                   CircleBox[{0.03391979910730371, 0.5}, 1.5, {
                    1.5852419347992028`, 1.5841816555164712`}], 
                   ArrowBox[{{-0.0011549968577574538`, 1.9995898634920182`}, {
                    0.013842405571952283`, 1.999865626737551}}]}, 
                  {RGBColor[0, 0, 1], ArrowBox[{{0, 2.8}, {0, 2.8}}]}},
                 Frame->True,
                 FrameTicks->Automatic,
                 ImageSize->800,
                 PlotRange->{{-3, 3}, {-1, 3}}], 231->
                GraphicsBox[{InsetBox[
                   TagBox[GridBox[{
                    {"\<\"t\"\>", "11.5`"},
                    {"\<\"driveIntent\"\>", "0"},
                    {"\<\"motorTorque\"\>", "7.1307370191179915`"},
                    {"\<\"vw\"\>", "0.07082719860703977`"},
                    {"\<\"controllerState\"\>", 
                    RowBox[{"\[LeftAssociation]", 
                    RowBox[{
                    RowBox[{"\<\"lastError\"\>", "\[Rule]", 
                    RowBox[{"-", "0.006841361496484923`"}]}], ",", 
                    
                    RowBox[{"\<\"lastInt\"\>", "\[Rule]", 
                    "0.05827854728873437`"}], ",", 
                    RowBox[{"\<\"initialized\"\>", "\[Rule]", "True"}]}], 
                    "\[RightAssociation]"}]}
                    },
                    AutoDelete->False,
                    
                    GridBoxItemSize->{
                    "Columns" -> {{Automatic}}, "Rows" -> {{Automatic}}}],
                    "Grid"], {-1, -0.5}], 
                  {GrayLevel[0], LineBox[{{-5, 0}, {5, 0}}]}, 
                  {GrayLevel[0.5], CircleBox[{0.03747593383125203, 0.5}, 0.5],
                    DiskBox[{0.03747593383125203, 0.5}, 0.1]}, 
                  {GrayLevel[0.5], 
                   DiskBox[{0.015862540931072262`, 1.9998442789994375`}, 
                    0.1]}, 
                  {GrayLevel[0.5], Thickness[0.01], 
                   LineBox[{{0.03747593383125203, 0.5}, {
                    0.015862540931072262`, 1.9998442789994375`}}]}, 
                  {RGBColor[0, 1, 0], 
                   DiskBox[{0.023067005231132184`, 1.4998961859996252`}, 
                    0.05]}, 
                  {RGBColor[0, 0, 1], Arrowheads[0.01], 
                   CircleBox[{0.03747593383125203, 0.5}, 0.3333333333333333, 
                    NCache[{
                    Rational[-1, 2] 
                    Pi, -1.3568742162213567`}, {-1.5707963267948966`, \
-1.3568742162213567`}]], 
                   ArrowBox[{{0.10497984128767307`, 0.1735733871308498}, {
                    0.10824067784495484`, 0.17426473614945187`}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{0.03747593383125203, 0.5}, {-0.1051388065511078,
                     0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{0.03747593383125203, 0.5}, {0.1434424216263661, 
                    0.5}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.005], Arrowheads[0.04], 
                   ArrowBox[{{0.03747593383125203, 0.5}, {
                    0.0008276812440062981, 0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{0.03747593383125203, 0.5}, 1.6, {
                    1.5852057540318958`, 1.651441685120817}], 
                   ArrowBox[{{-0.07546264308272084, 
                    2.0960090469181702`}, {-0.09141682066993662, 
                    2.0947998801846888`}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Dashing[{Small, Small}], 
                   Arrowheads[0.02], 
                   CircleBox[{0.03747593383125203, 0.5}, 1.7, {
                    1.5852057540318958`, 1.563968800309919}], 
                   ArrowBox[{{0.03208273790244057, 2.199991445107202}, {
                    0.04908263868057102, 2.199960377303701}}]}, 
                  {RGBColor[0, 0, 1], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{0.03747593383125203, 0.5}, 1.8, {
                    1.5852057540318958`, 1.545590548370129}], 
                   ArrowBox[{{0.06484528026362778, 2.299791909881769}, {
                    0.08284153094270923, 2.2994282321333968`}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.003], Arrowheads[0.025], 
                   CircleBox[{0.03747593383125203, 0.5}, 1.5, {
                    1.5852057540318958`, 1.5842506195143848`}], 
                   ArrowBox[{{0.0022977202374640515`, 1.9995874410278147`}, {
                    0.017295103613773847`, 1.999864238553521}}]}, 
                  {RGBColor[0, 0, 1], ArrowBox[{{0, 2.8}, {0, 2.8}}]}},
                 Frame->True,
                 FrameTicks->Automatic,
                 ImageSize->800,
                 PlotRange->{{-3, 3}, {-1, 3}}], 232->
                GraphicsBox[{InsetBox[
                   TagBox[GridBox[{
                    {"\<\"t\"\>", "11.55`"},
                    {"\<\"driveIntent\"\>", "0"},
                    {"\<\"motorTorque\"\>", "7.095260513517935`"},
                    {"\<\"vw\"\>", "0.07022731356863698`"},
                    {"\<\"controllerState\"\>", 
                    RowBox[{"\[LeftAssociation]", 
                    RowBox[{
                    RowBox[{"\<\"lastError\"\>", "\[Rule]", 
                    RowBox[{"-", "0.006722545319727808`"}]}], ",", 
                    
                    RowBox[{"\<\"lastInt\"\>", "\[Rule]", 
                    "0.057939614839899636`"}], ",", 
                    RowBox[{"\<\"initialized\"\>", "\[Rule]", "True"}]}], 
                    "\[RightAssociation]"}]}
                    },
                    AutoDelete->False,
                    
                    GridBoxItemSize->{
                    "Columns" -> {{Automatic}}, "Rows" -> {{Automatic}}}],
                    "Grid"], {-1, -0.5}], 
                  {GrayLevel[0], LineBox[{{-5, 0}, {5, 0}}]}, 
                  {GrayLevel[0.5], 
                   CircleBox[{0.041001979193558774`, 0.5}, 0.5], 
                   DiskBox[{0.041001979193558774`, 0.5}, 0.1]}, 
                  {GrayLevel[0.5], 
                   DiskBox[{0.019481062182926962`, 1.9998456087647893`}, 
                    0.1]}, 
                  {GrayLevel[0.5], Thickness[0.01], 
                   LineBox[{{0.041001979193558774`, 0.5}, {
                    0.019481062182926962`, 1.9998456087647893`}}]}, 
                  {RGBColor[0, 1, 0], 
                   DiskBox[{0.026654701186470897`, 1.4998970725098597`}, 
                    0.05]}, 
                  {RGBColor[0, 0, 1], Arrowheads[0.01], 
                   CircleBox[{0.041001979193558774`, 0.5}, 0.3333333333333333,
                     NCache[{
                    Rational[-1, 2] 
                    Pi, -1.3579385113893585`}, {-1.5707963267948966`, \
-1.3579385113893585`}]], 
                   ArrowBox[{{0.10815843421713209`, 0.1735017279375312}, {
                    0.11142000472688077`, 0.17418960607264306`}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{0.041001979193558774`, 
                    0.5}, {-0.10090323107679994`, 0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{0.041001979193558774`, 0.5}, {
                    0.14651516826905775`, 0.5}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.005], Arrowheads[0.04], 
                   ArrowBox[{{0.041001979193558774`, 0.5}, {
                    0.004609957998699052, 0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{0.041001979193558774`, 0.5}, 1.6, {
                    1.5851440970653567`, 1.6510966285660793`}], 
                   ArrowBox[{{-0.07138587762482207, 
                    2.0960479221000137`}, {-0.08734047149315244, 
                    2.0948442605319584`}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Dashing[{Small, Small}], 
                   Arrowheads[0.02], 
                   CircleBox[{0.041001979193558774`, 0.5}, 1.7, {
                    1.5851440970653567`, 1.5640556053690282`}], 
                   ArrowBox[{{0.035461215427374175`, 2.1999909705456933`}, {
                    0.05246111883829902, 2.1999613784196987`}}]}, 
                  {RGBColor[0, 0, 1], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{0.041001979193558774`, 0.5}, 1.8, {
                    1.5851440970653567`, 1.545725983101368}], 
                   ArrowBox[{{0.06812757104206671, 2.2997956001354343`}, {
                    0.0861238708106974, 2.299434359707764}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.003], Arrowheads[0.025], 
                   CircleBox[{0.041001979193558774`, 0.5}, 1.5, {
                    1.5851440970653567`, 1.5842914366535024`}], 
                   ArrowBox[{{0.005762556759892616, 1.9995860039045916`}, {
                    0.020759928825626182`, 1.9998634135803506`}}]}, 
                  {RGBColor[0, 0, 1], ArrowBox[{{0, 2.8}, {0, 2.8}}]}},
                 Frame->True,
                 FrameTicks->Automatic,
                 ImageSize->800,
                 PlotRange->{{-3, 3}, {-1, 3}}], 233->
                GraphicsBox[{InsetBox[
                   TagBox[GridBox[{
                    {"\<\"t\"\>", "11.6`"},
                    {"\<\"driveIntent\"\>", "0"},
                    {"\<\"motorTorque\"\>", "7.048774979798602`"},
                    {"\<\"vw\"\>", "0.06963207364197388`"},
                    {"\<\"controllerState\"\>", 
                    RowBox[{"\[LeftAssociation]", 
                    RowBox[{
                    RowBox[{"\<\"lastError\"\>", "\[Rule]", 
                    RowBox[{"-", "0.006627916311352847`"}]}], ",", 
                    
                    RowBox[{"\<\"lastInt\"\>", "\[Rule]", 
                    "0.05760599630874405`"}], ",", 
                    RowBox[{"\<\"initialized\"\>", "\[Rule]", "True"}]}], 
                    "\[RightAssociation]"}]}
                    },
                    AutoDelete->False,
                    
                    GridBoxItemSize->{
                    "Columns" -> {{Automatic}}, "Rows" -> {{Automatic}}}],
                    "Grid"], {-1, -0.5}], 
                  {GrayLevel[0], LineBox[{{-5, 0}, {5, 0}}]}, 
                  {GrayLevel[0.5], 
                   CircleBox[{0.044498145126208706`, 0.5}, 0.5], 
                   DiskBox[{0.044498145126208706`, 0.5}, 0.1]}, 
                  {GrayLevel[0.5], 
                   DiskBox[{0.023103818206955146`, 1.9998474198316547`}, 
                    0.1]}, 
                  {GrayLevel[0.5], Thickness[0.01], 
                   LineBox[{{0.044498145126208706`, 0.5}, {
                    0.023103818206955146`, 1.9998474198316547`}}]}, 
                  {RGBColor[0, 1, 0], 
                   DiskBox[{0.030235260513372997`, 1.4998982798877698`}, 
                    0.05]}, 
                  {RGBColor[0, 0, 1], Arrowheads[0.01], 
                   CircleBox[{0.044498145126208706`, 0.5}, 0.3333333333333333,
                     NCache[{
                    Rational[-1, 2] 
                    Pi, -1.3593330774009384`}, {-1.5707963267948966`, \
-1.3593330774009384`}]], 
                   ArrowBox[{{0.11119921160090218`, 0.17340839134746644`}, {
                    0.1144617382302332, 0.17409172033977782`}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{0.044498145126208706`, 
                    0.5}, {-0.09647735446976335, 0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{0.044498145126208706`, 0.5}, {
                    0.1493908123882388, 0.5}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.005], Arrowheads[0.04], 
                   ArrowBox[{{0.044498145126208706`, 0.5}, {
                    0.00841531279226674, 0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{0.044498145126208706`, 0.5}, 1.6, {
                    1.585059695034818, 1.650624281310661}], 
                   ArrowBox[{{-0.06713581032715027, 
                    2.0961008301450876`}, {-0.08309097096186754, 
                    2.094904704819653}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Dashing[{Small, Small}], 
                   Arrowheads[0.02], 
                   CircleBox[{0.044498145126208706`, 0.5}, 1.7, {
                    1.585059695034818, 1.5641503466593105`}], 
                   ArrowBox[{{0.03879632204711399, 2.1999904379771005`}, {
                    0.055796228185340205`, 2.1999624564440206`}}]}, 
                  {RGBColor[0, 0, 1], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{0.044498145126208706`, 0.5}, 1.8, {
                    1.585059695034818, 1.545899834035937}], 
                   ArrowBox[{{0.07131084041925254, 2.2998002887462605`}, {
                    0.08930720271790746, 2.2994421769975673`}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.003], Arrowheads[0.025], 
                   CircleBox[{0.044498145126208706`, 0.5}, 1.5, {
                    1.585059695034818, 1.5843061590381515`}], 
                   ArrowBox[{{0.009236645214398397, 1.999585484933743}, {
                    0.02423401319437468, 1.9998631154065525`}}]}, 
                  {RGBColor[0, 0, 1], ArrowBox[{{0, 2.8}, {0, 2.8}}]}},
                 Frame->True,
                 FrameTicks->Automatic,
                 ImageSize->800,
                 PlotRange->{{-3, 3}, {-1, 3}}], 234->
                GraphicsBox[{InsetBox[
                   TagBox[GridBox[{
                    {"\<\"t\"\>", "11.65`"},
                    {"\<\"driveIntent\"\>", "0"},
                    {"\<\"motorTorque\"\>", "6.992607524351754`"},
                    {"\<\"vw\"\>", "0.06904229101200532`"},
                    {"\<\"controllerState\"\>", 
                    RowBox[{"\[LeftAssociation]", 
                    RowBox[{
                    RowBox[{"\<\"lastError\"\>", "\[Rule]", 
                    RowBox[{"-", "0.006555074834733795`"}]}], ",", 
                    
                    RowBox[{"\<\"lastInt\"\>", "\[Rule]", 
                    "0.05727654378474235`"}], ",", 
                    RowBox[{"\<\"initialized\"\>", "\[Rule]", "True"}]}], 
                    "\[RightAssociation]"}]}
                    },
                    AutoDelete->False,
                    
                    GridBoxItemSize->{
                    "Columns" -> {{Automatic}}, "Rows" -> {{Automatic}}}],
                    "Grid"], {-1, -0.5}], 
                  {GrayLevel[0], LineBox[{{-5, 0}, {5, 0}}]}, 
                  {GrayLevel[0.5], 
                   CircleBox[{0.047964685083001016`, 0.5}, 0.5], 
                   DiskBox[{0.047964685083001016`, 0.5}, 0.1]}, 
                  {GrayLevel[0.5], 
                   DiskBox[{0.026727099894551798`, 1.999849647456492}, 0.1]}, 
                  {GrayLevel[0.5], Thickness[0.01], 
                   LineBox[{{0.047964685083001016`, 0.5}, {
                    0.026727099894551798`, 1.999849647456492}}]}, 
                  {RGBColor[0, 1, 0], 
                   DiskBox[{0.0338062949573682, 1.4998997649709946`}, 0.05]}, 
                  {RGBColor[0, 0, 1], Arrowheads[0.01], 
                   CircleBox[{0.047964685083001016`, 0.5}, 0.3333333333333333,
                     NCache[{
                    Rational[-1, 2] 
                    Pi, -1.3610181010643438`}, {-1.5707963267948966`, \
-1.3610181010643438`}]], 
                   ArrowBox[{{0.11411534253695887`, 0.17329646217171102`}, {
                    0.11737901595961438`, 0.17397429276196158`}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{0.047964685083001016`, 
                    0.5}, {-0.09188746540403407, 0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{0.047964685083001016`, 0.5}, {
                    0.15208902952422004`, 0.5}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.005], Arrowheads[0.04], 
                   ArrowBox[{{0.047964685083001016`, 0.5}, {
                    0.012236879037184956`, 0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{0.047964685083001016`, 0.5}, 1.6, {
                    1.5849551899950451`, 1.6500394289913907`}], 
                   ArrowBox[{{-0.06273576805895337, 
                    2.096165846544201}, {-0.07869162552155239, 
                    2.0949790528355043`}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Dashing[{Small, Small}], 
                   Arrowheads[0.02], 
                   CircleBox[{0.047964685083001016`, 0.5}, 1.7, {
                    1.5849551899950451`, 1.5642515414974472`}], 
                   ArrowBox[{{0.042090831776189434`, 2.199989852277751}, {
                    0.05909074065895947, 2.1999635910475615`}}]}, 
                  {RGBColor[0, 0, 1], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{0.047964685083001016`, 0.5}, 1.8, {
                    1.5849551899950451`, 1.5461073704153132`}], 
                   ArrowBox[{{0.07440385576576988, 2.2998058145960103`}, {
                    0.09240029199807621, 2.2994514377548754`}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.003], Arrowheads[0.025], 
                   CircleBox[{0.047964685083001016`, 0.5}, 1.5, {
                    1.5849551899950451`, 1.584296859798191}], 
                   ArrowBox[{{0.012717130177980839`, 1.9995858127740531`}, {
                    0.027714500739061062`, 1.999863303782727}}]}, 
                  {RGBColor[0, 0, 1], ArrowBox[{{0, 2.8}, {0, 2.8}}]}},
                 Frame->True,
                 FrameTicks->Automatic,
                 ImageSize->800,
                 PlotRange->{{-3, 3}, {-1, 3}}], 235->
                GraphicsBox[{InsetBox[
                   TagBox[GridBox[{
                    {"\<\"t\"\>", "11.700000000000001`"},
                    {"\<\"driveIntent\"\>", "0"},
                    {"\<\"motorTorque\"\>", "6.928028809564379`"},
                    {"\<\"vw\"\>", "0.0684586637622465`"},
                    {"\<\"controllerState\"\>", 
                    RowBox[{"\[LeftAssociation]", 
                    RowBox[{
                    RowBox[{"\<\"lastError\"\>", "\[Rule]", 
                    RowBox[{"-", "0.006501690978712069`"}]}], ",", 
                    
                    RowBox[{"\<\"lastInt\"\>", "\[Rule]", 
                    "0.056950227626242494`"}], ",", 
                    RowBox[{"\<\"initialized\"\>", "\[Rule]", "True"}]}], 
                    "\[RightAssociation]"}]}
                    },
                    AutoDelete->False,
                    
                    GridBoxItemSize->{
                    "Columns" -> {{Automatic}}, "Rows" -> {{Automatic}}}],
                    "Grid"], {-1, -0.5}], 
                  {GrayLevel[0], LineBox[{{-5, 0}, {5, 0}}]}, 
                  {GrayLevel[0.5], CircleBox[{0.05140189019262306, 0.5}, 0.5],
                    DiskBox[{0.05140189019262306, 0.5}, 0.1]}, 
                  {GrayLevel[0.5], 
                   DiskBox[{0.03034739220136896, 1.9998522287593323`}, 0.1]}, 
                  {GrayLevel[0.5], Thickness[0.01], 
                   LineBox[{{0.05140189019262306, 0.5}, {0.03034739220136896, 
                    1.9998522287593323`}}]}, 
                  {RGBColor[0, 1, 0], 
                   DiskBox[{0.03736555819845366, 1.4999014858395547`}, 0.05]}, 
                  {RGBColor[0, 0, 1], Arrowheads[0.01], 
                   CircleBox[{0.05140189019262306, 0.5}, 0.3333333333333333, 
                    NCache[{
                    Rational[-1, 2] 
                    Pi, -1.3629554625079652`}, {-1.5707963267948966`, \
-1.3629554625079652`}]], 
                   ArrowBox[{{0.11691948106095565`, 0.1731689176379932}, {
                    0.12018446156075652`, 0.1738404240450716}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{0.05140189019262306, 
                    0.5}, {-0.08715868599866453, 0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{0.05140189019262306, 0.5}, {
                    0.15462876624463978`, 0.5}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.005], Arrowheads[0.04], 
                   ArrowBox[{{0.05140189019262306, 0.5}, {
                    0.016068190053352194`, 0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{0.05140189019262306, 0.5}, 1.6, {
                    1.5848331197330505`, 1.6493562737099043`}], 
                   ArrowBox[{{-0.05820810807394122, 
                    2.096241099671351}, {-0.07416477257755905, 
                    2.0950652065670416`}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Dashing[{Small, Small}], 
                   Arrowheads[0.02], 
                   CircleBox[{0.05140189019262306, 0.5}, 1.7, {
                    1.5848331197330505`, 1.5643578128031772`}], 
                   ArrowBox[{{0.0453473767779906, 2.199989218456197}, {
                    0.06234728835558086, 2.1999647638286666`}}]}, 
                  {RGBColor[0, 0, 1], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{0.05140189019262306, 0.5}, 1.8, {
                    1.5848331197330505`, 1.546344070791026}], 
                   ArrowBox[{{0.07741504542619028, 2.2998120223386644`}, {
                    0.0954115650354835, 2.2994619052706353`}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.003], Arrowheads[0.025], 
                   CircleBox[{0.05140189019262306, 0.5}, 1.5, {
                    1.5848331197330505`, 1.5842656107240756`}], 
                   ArrowBox[{{0.016201195973010812`, 1.9995869134953324`}, {
                    0.031198575198105625`, 1.9998639358499264`}}]}, 
                  {RGBColor[0, 0, 1], ArrowBox[{{0, 2.8}, {0, 2.8}}]}},
                 Frame->True,
                 FrameTicks->Automatic,
                 ImageSize->800,
                 PlotRange->{{-3, 3}, {-1, 3}}], 236->
                GraphicsBox[{InsetBox[
                   TagBox[GridBox[{
                    {"\<\"t\"\>", "11.75`"},
                    {"\<\"driveIntent\"\>", "0"},
                    {"\<\"motorTorque\"\>", "6.856247315849661`"},
                    {"\<\"vw\"\>", "0.0678817820237565`"},
                    {"\<\"controllerState\"\>", 
                    RowBox[{"\[LeftAssociation]", 
                    RowBox[{
                    RowBox[{"\<\"lastError\"\>", "\[Rule]", 
                    RowBox[{"-", "0.006465518895844543`"}]}], ",", 
                    
                    RowBox[{"\<\"lastInt\"\>", "\[Rule]", 
                    "0.0566261325980122`"}], ",", 
                    RowBox[{"\<\"initialized\"\>", "\[Rule]", "True"}]}], 
                    "\[RightAssociation]"}]}
                    },
                    AutoDelete->False,
                    
                    GridBoxItemSize->{
                    "Columns" -> {{Automatic}}, "Rows" -> {{Automatic}}}],
                    "Grid"], {-1, -0.5}], 
                  {GrayLevel[0], LineBox[{{-5, 0}, {5, 0}}]}, 
                  {GrayLevel[0.5], CircleBox[{0.05481008370876342, 0.5}, 0.5],
                    DiskBox[{0.05481008370876342, 0.5}, 0.1]}, 
                  {GrayLevel[0.5], 
                   DiskBox[{0.03396138725051833, 1.9998551036203436`}, 0.1]}, 
                  {GrayLevel[0.5], Thickness[0.01], 
                   LineBox[{{0.05481008370876342, 0.5}, {0.03396138725051833, 
                    1.9998551036203436`}}]}, 
                  {RGBColor[0, 1, 0], 
                   DiskBox[{0.04091095273660003, 1.4999034024135625`}, 0.05]}, 
                  {RGBColor[0, 0, 1], Arrowheads[0.01], 
                   CircleBox[{0.05481008370876342, 0.5}, 0.3333333333333333, 
                    NCache[{
                    Rational[-1, 2] 
                    Pi, -1.3651089073194067`}, {-1.5707963267948966`, \
-1.3651089073194067`}]], 
                   ArrowBox[{{0.1196237105094258, 0.17302858704153423`}, {
                    0.12289012948971301`, 0.17369306094173403`}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{0.05481008370876342, 
                    0.5}, {-0.08231486260822979, 0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{0.05481008370876342, 0.5}, {0.1570281432981429, 
                    0.5}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.005], Arrowheads[0.04], 
                   ArrowBox[{{0.05481008370876342, 0.5}, {
                    0.019903196981149694`, 0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{0.05481008370876342, 0.5}, 1.6, {
                    1.5846959053251883`, 1.6485883644258912`}], 
                   ArrowBox[{{-0.053574114000643186`, 
                    2.0963247995589396`}, {-0.06953167677870406, 
                    2.095161160070943}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Dashing[{Small, Small}], 
                   Arrowheads[0.02], 
                   CircleBox[{0.05481008370876342, 0.5}, 1.7, {
                    1.5846959053251883`, 1.5644678902224654`}], 
                   ArrowBox[{{0.048568439905204916`, 2.1999885416915697`}, {
                    0.06556835407170317, 2.199965958370578}}]}, 
                  {RGBColor[0, 0, 1], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{0.05481008370876342, 0.5}, 1.8, {
                    1.5846959053251883`, 1.5466056424593568`}], 
                   ArrowBox[{{0.0803524582244276, 2.2998187650716675`}, {
                    0.0983490687987668, 2.2994733553952202`}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.003], Arrowheads[0.025], 
                   CircleBox[{0.05481008370876342, 0.5}, 1.5, {
                    1.5846959053251883`, 1.5842144624771877`}], 
                   ArrowBox[{{0.01968609077684408, 1.999588711987563}, {
                    0.034683484151529026`, 1.9998649672521398`}}]}, 
                  {RGBColor[0, 0, 1], ArrowBox[{{0, 2.8}, {0, 2.8}}]}},
                 Frame->True,
                 FrameTicks->Automatic,
                 ImageSize->800,
                 PlotRange->{{-3, 3}, {-1, 3}}], 237->
                GraphicsBox[{InsetBox[
                   TagBox[GridBox[{
                    {"\<\"t\"\>", "11.8`"},
                    {"\<\"driveIntent\"\>", "0"},
                    {"\<\"motorTorque\"\>", "6.77840485860098`"},
                    {"\<\"vw\"\>", "0.0673121344581699`"},
                    {"\<\"controllerState\"\>", 
                    RowBox[{"\[LeftAssociation]", 
                    RowBox[{
                    RowBox[{"\<\"lastError\"\>", "\[Rule]", 
                    RowBox[{"-", "0.006444408688885307`"}]}], ",", 
                    
                    RowBox[{"\<\"lastInt\"\>", "\[Rule]", 
                    "0.056303453355082224`"}], ",", 
                    RowBox[{"\<\"initialized\"\>", "\[Rule]", "True"}]}], 
                    "\[RightAssociation]"}]}
                    },
                    AutoDelete->False,
                    
                    GridBoxItemSize->{
                    "Columns" -> {{Automatic}}, "Rows" -> {{Automatic}}}],
                    "Grid"], {-1, -0.5}], 
                  {GrayLevel[0], LineBox[{{-5, 0}, {5, 0}}]}, 
                  {GrayLevel[0.5], CircleBox[{0.0581896157766326, 0.5}, 0.5], 
                   DiskBox[{0.0581896157766326, 0.5}, 0.1]}, 
                  {GrayLevel[0.5], 
                   DiskBox[{0.03756599393194687, 1.9998582153730424`}, 0.1]}, 
                  {GrayLevel[0.5], Thickness[0.01], 
                   LineBox[{{0.0581896157766326, 0.5}, {0.03756599393194687, 
                    1.9998582153730424`}}]}, 
                  {RGBColor[0, 1, 0], 
                   DiskBox[{0.04444053454684212, 1.4999054769153615`}, 0.05]}, 
                  {RGBColor[0, 0, 1], Arrowheads[0.01], 
                   CircleBox[{0.0581896157766326, 0.5}, 0.3333333333333333, 
                    NCache[{
                    Rational[-1, 2] 
                    Pi, -1.3674441810368672`}, {-1.5707963267948966`, \
-1.3674441810368672`}]], 
                   ArrowBox[{{0.1222394987936872, 0.1728781211893452}, {
                    0.12550746059428972`, 0.17353496530222884`}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{0.0581896157766326, 0.5}, {-0.077378481395387, 
                    0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{0.0581896157766326, 0.5}, {0.1593043777755129, 
                    0.5}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.005], Arrowheads[0.04], 
                   ArrowBox[{{0.0581896157766326, 0.5}, {
                    0.023736280603493316`, 0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{0.0581896157766326, 0.5}, 1.6, {
                    1.584545841242923, 1.647748541606419}], 
                   ArrowBox[{{-0.048853913874593265`, 
                    2.0964152601249486`}, {-0.06481244827608187, 
                    2.0952650225711}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Dashing[{Small, Small}], 
                   Arrowheads[0.02], 
                   CircleBox[{0.0581896157766326, 0.5}, 1.7, {
                    1.584545841242923, 1.5645806100946493`}], 
                   ArrowBox[{{0.05175634952199863, 2.199987827334448}, {
                    0.06875626612608744, 2.199967160241748}}]}, 
                  {RGBColor[0, 0, 1], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{0.0581896157766326, 0.5}, 1.8, {
                    1.584545841242923, 1.5468880364729176`}], 
                   ArrowBox[{{0.08322373123584555, 2.299825906320713}, {
                    0.10122043863422603`, 2.2994855787930613`}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.003], Arrowheads[0.025], 
                   CircleBox[{0.0581896157766326, 0.5}, 1.5, {
                    1.584545841242923, 1.5841454272719353`}], 
                   ArrowBox[{{0.023169147342855224`, 1.999591133206208}, {
                    0.038166559753141316`, 1.9998663531219973`}}]}, 
                  {RGBColor[0, 0, 1], ArrowBox[{{0, 2.8}, {0, 2.8}}]}},
                 Frame->True,
                 FrameTicks->Automatic,
                 ImageSize->800,
                 PlotRange->{{-3, 3}, {-1, 3}}], 238->
                GraphicsBox[{InsetBox[
                   TagBox[GridBox[{
                    {"\<\"t\"\>", "11.85`"},
                    {"\<\"driveIntent\"\>", "0"},
                    {"\<\"motorTorque\"\>", "6.695573258853814`"},
                    {"\<\"vw\"\>", "0.06675011497385513`"},
                    {"\<\"controllerState\"\>", 
                    RowBox[{"\[LeftAssociation]", 
                    RowBox[{
                    RowBox[{"\<\"lastError\"\>", "\[Rule]", 
                    RowBox[{"-", "0.006436316012932288`"}]}], ",", 
                    
                    RowBox[{"\<\"lastInt\"\>", "\[Rule]", 
                    "0.055981489391220496`"}], ",", 
                    RowBox[{"\<\"initialized\"\>", "\[Rule]", "True"}]}], 
                    "\[RightAssociation]"}]}
                    },
                    AutoDelete->False,
                    
                    GridBoxItemSize->{
                    "Columns" -> {{Automatic}}, "Rows" -> {{Automatic}}}],
                    "Grid"], {-1, -0.5}], 
                  {GrayLevel[0], LineBox[{{-5, 0}, {5, 0}}]}, 
                  {GrayLevel[0.5], 
                   CircleBox[{0.061540858529980494`, 0.5}, 0.5], 
                   DiskBox[{0.061540858529980494`, 0.5}, 0.1]}, 
                  {GrayLevel[0.5], 
                   DiskBox[{0.041158344308520625`, 1.9998615113116318`}, 
                    0.1]}, 
                  {GrayLevel[0.5], Thickness[0.01], 
                   LineBox[{{0.061540858529980494`, 0.5}, {
                    0.041158344308520625`, 1.9998615113116318`}}]}, 
                  {RGBColor[0, 1, 0], 
                   DiskBox[{0.04795251571567392, 1.4999076742077544`}, 0.05]}, 
                  {RGBColor[0, 0, 1], Arrowheads[0.01], 
                   CircleBox[{0.061540858529980494`, 0.5}, 0.3333333333333333,
                     NCache[{
                    Rational[-1, 2] 
                    Pi, -1.369929129029282}, {-1.5707963267948966`, \
-1.369929129029282}]], 
                   ArrowBox[{{0.12477766377475462`, 0.17271997070773537`}, {
                    0.1280472477073648, 0.17336869208586664`}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{0.061540858529980494`, 
                    0.5}, {-0.07237060664709578, 0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{0.061540858529980494`, 0.5}, {
                    0.16147372284213962`, 0.5}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.005], Arrowheads[0.04], 
                   ArrowBox[{{0.061540858529980494`, 0.5}, {
                    0.027562257665063344`, 0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{0.061540858529980494`, 0.5}, 1.6, {
                    1.584385087809486, 1.6468488950250733`}], 
                   ArrowBox[{{-0.04406641846521657, 
                    2.09651091541701}, {-0.060025981215717644`, 
                    2.095375035367623}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Dashing[{Small, Small}], 
                   Arrowheads[0.02], 
                   CircleBox[{0.061540858529980494`, 0.5}, 1.7, {
                    1.584385087809486, 1.5646949143976272`}], 
                   ArrowBox[{{0.05491327639412216, 2.199987080878861}, {
                    0.0719131952494926, 2.1999683569499098`}}]}, 
                  {RGBColor[0, 0, 1], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{0.061540858529980494`, 0.5}, 1.8, {
                    1.584385087809486, 1.5471874585936314`}], 
                   ArrowBox[{{0.08603606518526047, 2.2998333214081006`}, {
                    0.10403287367849465`, 2.2994983825079194`}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.003], Arrowheads[0.025], 
                   CircleBox[{0.061540858529980494`, 0.5}, 1.5, {
                    1.584385087809486, 1.584060463913647}], 
                   ArrowBox[{{0.026647800521188668`, 1.999594103250208}, {
                    0.041645236260951596`, 1.9998680489344816`}}]}, 
                  {RGBColor[0, 0, 1], ArrowBox[{{0, 2.8}, {0, 2.8}}]}},
                 Frame->True,
                 FrameTicks->Automatic,
                 ImageSize->800,
                 PlotRange->{{-3, 3}, {-1, 3}}], 239->
                GraphicsBox[{InsetBox[
                   TagBox[GridBox[{
                    {"\<\"t\"\>", "11.9`"},
                    {"\<\"driveIntent\"\>", "0"},
                    {"\<\"motorTorque\"\>", "6.6087520670182105`"},
                    {"\<\"vw\"\>", "0.06619602958367364`"},
                    {"\<\"controllerState\"\>", 
                    RowBox[{"\[LeftAssociation]", 
                    RowBox[{
                    RowBox[{"\<\"lastError\"\>", "\[Rule]", 
                    RowBox[{"-", "0.006439309563419121`"}]}], ",", 
                    
                    RowBox[{"\<\"lastInt\"\>", "\[Rule]", 
                    "0.05565963956192186`"}], ",", 
                    RowBox[{"\<\"initialized\"\>", "\[Rule]", "True"}]}], 
                    "\[RightAssociation]"}]}
                    },
                    AutoDelete->False,
                    
                    GridBoxItemSize->{
                    "Columns" -> {{Automatic}}, "Rows" -> {{Automatic}}}],
                    "Grid"], {-1, -0.5}], 
                  {GrayLevel[0], LineBox[{{-5, 0}, {5, 0}}]}, 
                  {GrayLevel[0.5], CircleBox[{0.0648642015278946, 0.5}, 0.5], 
                   DiskBox[{0.0648642015278946, 0.5}, 0.1]}, 
                  {GrayLevel[0.5], 
                   DiskBox[{0.044735797134318904`, 1.9998649430320614`}, 
                    0.1]}, 
                  {GrayLevel[0.5], Thickness[0.01], 
                   LineBox[{{0.0648642015278946, 0.5}, {0.044735797134318904`,
                     1.9998649430320614`}}]}, 
                  {RGBColor[0, 1, 0], 
                   DiskBox[{0.0514452652655108, 1.4999099620213743`}, 0.05]}, 
                  {RGBColor[0, 0, 1], Arrowheads[0.01], 
                   CircleBox[{0.0648642015278946, 0.5}, 0.3333333333333333, 
                    NCache[{
                    Rational[-1, 2] 
                    Pi, -1.3725337647843503`}, {-1.5707963267948966`, \
-1.3725337647843503`}]], 
                   ArrowBox[{{0.12724834796715273`, 0.17255637220407494`}, {
                    0.13051961049011823`, 0.1731965753161172}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{0.0648642015278946, 0.5}, {-0.0673108398124696, 
                    0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{0.0648642015278946, 0.5}, {0.16355142360291902`,
                     0.5}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.005], Arrowheads[0.04], 
                   ArrowBox[{{0.0648642015278946, 0.5}, {0.03137638226255479, 
                    0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{0.0648642015278946, 0.5}, 1.6, {
                    1.584215665809751, 1.645900733605457}], 
                   ArrowBox[{{-0.03922927816712091, 
                    2.096610330508037}, {-0.055189910741197445`, 
                    2.09548958320864}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Dashing[{Small, Small}], 
                   Arrowheads[0.02], 
                   CircleBox[{0.0648642015278946, 0.5}, 1.7, {
                    1.584215665809751, 1.5648098487995794`}], 
                   ArrowBox[{{0.058041232437774065`, 2.19998630791333}, {
                    0.07504115333288432, 2.199969537860005}}]}, 
                  {RGBColor[0, 0, 1], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{0.0648642015278946, 0.5}, 1.8, {
                    1.584215665809751, 1.547500376548539}], 
                   ArrowBox[{{0.08879620683099987, 2.2998408982802263`}, {
                    0.10679311925152546`, 2.29951159092086}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.003], Arrowheads[0.025], 
                   CircleBox[{0.0648642015278946, 0.5}, 1.5, {
                    1.584215665809751, 1.5839614650668548`}], 
                   ArrowBox[{{0.030119601776728053`, 1.9995975502741163`}, {
                    0.04511706456330454, 1.999870011228207}}]}, 
                  {RGBColor[0, 0, 1], ArrowBox[{{0, 2.8}, {0, 2.8}}]}},
                 Frame->True,
                 FrameTicks->Automatic,
                 ImageSize->800,
                 PlotRange->{{-3, 3}, {-1, 3}}], 240->
                GraphicsBox[{InsetBox[
                   TagBox[GridBox[{
                    {"\<\"t\"\>", "11.950000000000001`"},
                    {"\<\"driveIntent\"\>", "0"},
                    {"\<\"motorTorque\"\>", "6.518867240574024`"},
                    {"\<\"vw\"\>", "0.06565010332201271`"},
                    {"\<\"controllerState\"\>", 
                    RowBox[{"\[LeftAssociation]", 
                    RowBox[{
                    RowBox[{"\<\"lastError\"\>", "\[Rule]", 
                    RowBox[{"-", "0.006451576620897018`"}]}], ",", 
                    
                    RowBox[{"\<\"lastInt\"\>", "\[Rule]", 
                    "0.05533739628326113`"}], ",", 
                    RowBox[{"\<\"initialized\"\>", "\[Rule]", "True"}]}], 
                    "\[RightAssociation]"}]}
                    },
                    AutoDelete->False,
                    
                    GridBoxItemSize->{
                    "Columns" -> {{Automatic}}, "Rows" -> {{Automatic}}}],
                    "Grid"], {-1, -0.5}], 
                  {GrayLevel[0], LineBox[{{-5, 0}, {5, 0}}]}, 
                  {GrayLevel[0.5], CircleBox[{0.06816004753632157, 0.5}, 0.5],
                    DiskBox[{0.06816004753632157, 0.5}, 0.1]}, 
                  {GrayLevel[0.5], 
                   DiskBox[{0.04829593878321758, 1.9998684666274722`}, 0.1]}, 
                  {GrayLevel[0.5], Thickness[0.01], 
                   LineBox[{{0.06816004753632157, 0.5}, {0.04829593878321758, 
                    1.9998684666274722`}}]}, 
                  {RGBColor[0, 1, 0], 
                   DiskBox[{0.05491730836758557, 1.4999123110849815`}, 0.05]}, 
                  {RGBColor[0, 0, 1], Arrowheads[0.01], 
                   CircleBox[{0.06816004753632157, 0.5}, 0.3333333333333333, 
                    NCache[{
                    Rational[-1, 2] 
                    Pi, -1.375230309577676}, {-1.5707963267948966`, \
-1.375230309577676}]], 
                   ArrowBox[{{0.12966100182753537`, 0.17238934124119049`}, {
                    0.13293397879153074`, 0.17302072093042875`}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{0.06816004753632157, 0.5}, {-0.0622172972751589,
                     0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{0.06816004753632157, 0.5}, {
                    0.16555168767193618`, 0.5}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.005], Arrowheads[0.04], 
                   ArrowBox[{{0.06816004753632157, 0.5}, {0.03517434286045569,
                     0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{0.06816004753632157, 0.5}, 1.6, {
                    1.5840394530576867`, 1.6449145662150702`}], 
                   ArrowBox[{{-0.03435885675399335, 
                    2.096712207714062}, {-0.050320586811264834`, 
                    2.0956072008124664`}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Dashing[{Small, Small}], 
                   Arrowheads[0.02], 
                   CircleBox[{0.06816004753632157, 0.5}, 1.7, {
                    1.5840394530576867`, 1.5649245599389923`}], 
                   ArrowBox[{{0.06114207112515222, 2.1999855140580147`}, {
                    0.07814199383212644, 2.199970694085091}}]}, 
                  {RGBColor[0, 0, 1], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{0.06816004753632157, 0.5}, 1.8, {
                    1.5840394530576867`, 1.5478235239433866`}], 
                   ArrowBox[{{0.0915104377025889, 2.299848537871752}, {
                    0.10950745559827099`, 2.299525046184842}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.003], Arrowheads[0.025], 
                   CircleBox[{0.06816004753632157, 0.5}, 1.5, {
                    1.5840394530576867`, 1.5838502466205486`}], 
                   ArrowBox[{{0.03358223090932456, 1.9996014052398423`}, {
                    0.04857972390582909, 1.9998721981977414`}}]}, 
                  {RGBColor[0, 0, 1], ArrowBox[{{0, 2.8}, {0, 2.8}}]}},
                 Frame->True,
                 FrameTicks->Automatic,
                 ImageSize->800,
                 PlotRange->{{-3, 3}, {-1, 3}}], 241->
                GraphicsBox[{InsetBox[
                   TagBox[GridBox[{
                    {"\<\"t\"\>", "12.`"},
                    {"\<\"driveIntent\"\>", 
                    RowBox[{"-", "0.5`"}]},
                    {"\<\"motorTorque\"\>", 
                    RowBox[{"-", "200"}]},
                    {"\<\"vw\"\>", "0.0651124871476991`"},
                    {"\<\"controllerState\"\>", 
                    RowBox[{"\[LeftAssociation]", 
                    RowBox[{
                    RowBox[{"\<\"lastError\"\>", "\[Rule]", 
                    RowBox[{"-", "0.1564714268225616`"}]}], ",", 
                    
                    RowBox[{"\<\"lastInt\"\>", "\[Rule]", 
                    "0.054864339499428795`"}], ",", 
                    RowBox[{"\<\"initialized\"\>", "\[Rule]", "True"}]}], 
                    "\[RightAssociation]"}]}
                    },
                    AutoDelete->False,
                    
                    GridBoxItemSize->{
                    "Columns" -> {{Automatic}}, "Rows" -> {{Automatic}}}],
                    "Grid"], {-1, -0.5}], 
                  {GrayLevel[0], LineBox[{{-5, 0}, {5, 0}}]}, 
                  {GrayLevel[0.5], CircleBox[{0.0714288086553682, 0.5}, 0.5], 
                   DiskBox[{0.0714288086553682, 0.5}, 0.1]}, 
                  {GrayLevel[0.5], 
                   DiskBox[{0.05183658187638969, 1.9998720427589285`}, 0.1]}, 
                  {GrayLevel[0.5], Thickness[0.01], 
                   LineBox[{{0.0714288086553682, 0.5}, {0.05183658187638969, 
                    1.9998720427589285`}}]}, 
                  {RGBColor[0, 1, 0], 
                   DiskBox[{0.05836732413604919, 1.499914695172619}, 0.05]}, 
                  {RGBColor[0, 0, 1], Arrowheads[0.01], 
                   CircleBox[{0.0714288086553682, 0.5}, 0.3333333333333333, 
                    NCache[{
                    Rational[-1, 2] 
                    Pi, -7.570796326794897}, {-1.5707963267948966`, \
-7.570796326794897}]], 
                   ArrowBox[{{0.16776316544848857`, 0.18089060995895645`}, {
                    0.1645673080550102, 0.17994323778321136`}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{0.0714288086553682, 0.5}, {4.0714288086553685`, 
                    0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{0.0714288086553682, 0.5}, {0.16748766904282758`,
                     0.5}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.005], Arrowheads[0.04], 
                   ArrowBox[{{0.0714288086553682, 0.5}, {4.167487669042828, 
                    0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{0.0714288086553682, 0.5}, 1.6, {
                    1.5838581827294498`, 1.6439000923363465`}], 
                   ArrowBox[{{-0.029470220332191363`, 
                    2.0968153888127983`}, {-0.045433063176343644`, 
                    2.0957265752352403`}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Dashing[{Small, Small}], 
                   Arrowheads[0.02], 
                   CircleBox[{0.0714288086553682, 0.5}, 1.7, {
                    1.5838581827294498`, 3.957490000281783}], 
                   ArrowBox[{{-1.1057601971438016`, -0.7264689334123233}, \
{-1.0934368532603491`, -0.7381793043374845}}]}, 
                  {RGBColor[0, 0, 1], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{0.0714288086553682, 0.5}, 1.8, {
                    1.5838581827294498`, 2.694969293840561}], 
                   ArrowBox[{{-1.544154782387721, 
                    1.2936558828310387`}, {-1.5520104304343179`, 
                    1.2774606337197412`}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.003], Arrowheads[0.025], 
                   CircleBox[{0.0714288086553682, 0.5}, 1.5, {
                    1.5838581827294498`, 5.671352279536117}], 
                   ArrowBox[{{1.2906468633099313`, -0.373789067913042}, {
                    1.2993236479635515`, -0.3615534014791717}}]}, 
                  {RGBColor[0, 0, 1], ArrowBox[{{0, 2.8}, {-1.5, 2.8}}]}},
                 Frame->True,
                 FrameTicks->Automatic,
                 ImageSize->800,
                 PlotRange->{{-3, 3}, {-1, 3}}], 242->
                GraphicsBox[{InsetBox[
                   TagBox[GridBox[{
                    {"\<\"t\"\>", "12.05`"},
                    {"\<\"driveIntent\"\>", 
                    RowBox[{"-", "0.5`"}]},
                    {"\<\"motorTorque\"\>", 
                    RowBox[{"-", "14.524706602443302`"}]},
                    {"\<\"vw\"\>", "0.08933698525634585`"},
                    {"\<\"controllerState\"\>", 
                    RowBox[{"\[LeftAssociation]", 
                    RowBox[{
                    RowBox[{"\<\"lastError\"\>", "\[Rule]", 
                    RowBox[{"-", "0.13936677268008799`"}]}], ",", 
                    
                    RowBox[{"\<\"lastInt\"\>", "\[Rule]", 
                    "0.0472736482893189`"}], ",", 
                    RowBox[{"\<\"initialized\"\>", "\[Rule]", "True"}]}], 
                    "\[RightAssociation]"}]}
                    },
                    AutoDelete->False,
                    
                    GridBoxItemSize->{
                    "Columns" -> {{Automatic}}, "Rows" -> {{Automatic}}}],
                    "Grid"], {-1, -0.5}], 
                  {GrayLevel[0], LineBox[{{-5, 0}, {5, 0}}]}, 
                  {GrayLevel[0.5], CircleBox[{0.07545272844839226, 0.5}, 0.5],
                    DiskBox[{0.07545272844839226, 0.5}, 0.1]}, 
                  {GrayLevel[0.5], 
                   DiskBox[{0.0181970392352035, 1.9989068637019187`}, 0.1]}, 
                  {GrayLevel[0.5], Thickness[0.01], 
                   LineBox[{{0.07545272844839226, 0.5}, {0.0181970392352035, 
                    1.9989068637019187`}}]}, 
                  {RGBColor[0, 1, 0], 
                   DiskBox[{0.03728226897293309, 1.499271242467946}, 0.05]}, 
                  {RGBColor[0, 0, 1], Arrowheads[0.01], 
                   CircleBox[{0.07545272844839226, 0.5}, 0.3333333333333333, 
                    NCache[{
                    Rational[-1, 2] 
                    Pi, -2.006537524868196}, {-1.5707963267948966`, \
-2.006537524868196}]], 
                   ArrowBox[{{-0.062212581076904055`, 
                    0.1964223762125814}, {-0.06524142351064469, 
                    0.19781418511842996`}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{0.07545272844839226, 0.5}, {
                    0.36594686049725833`, 0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{0.07545272844839226, 0.5}, {0.3559913869999558, 
                    0.5}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.005], Arrowheads[0.04], 
                   ArrowBox[{{0.07545272844839226, 0.5}, {0.6464855190488219, 
                    0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{0.07545272844839226, 0.5}, 1.6, {
                    1.6089760613107291`, 1.784440594005266}], 
                   ArrowBox[{{-0.24813067659525248`, 
                    2.0669377077536812`}, {-0.2637836134823842, 
                    2.0636235814010435`}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Dashing[{Small, Small}], 
                   Arrowheads[0.02], 
                   CircleBox[{0.07545272844839226, 0.5}, 1.7, {
                    1.6089760613107291`, 1.939667119801995}], 
                   ArrowBox[{{-0.5216164905869581, 
                    2.091699829641417}, {-0.53750337038922, 
                    2.0856496526338266`}}]}, 
                  {RGBColor[0, 0, 1], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{0.07545272844839226, 0.5}, 1.8, {
                    1.6089760613107291`, 1.6896688757687475`}], 
                   ArrowBox[{{-0.12013094161541918`, 
                    2.2893426245424244`}, {-0.1380142905365535, 
                    2.287297354053231}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.003], Arrowheads[0.025], 
                   CircleBox[{0.07545272844839226, 0.5}, 1.5, {
                    1.6089760613107291`, 2.2677528064912913`}], 
                   ArrowBox[{{-0.8758276737746687, 
                    1.659769630722555}, {-0.8873776131641765, 
                    1.650198997247965}}]}, 
                  {RGBColor[0, 0, 1], ArrowBox[{{0, 2.8}, {-1.5, 2.8}}]}},
                 Frame->True,
                 FrameTicks->Automatic,
                 ImageSize->800,
                 PlotRange->{{-3, 3}, {-1, 3}}], 243->
                GraphicsBox[{InsetBox[
                   TagBox[GridBox[{
                    {"\<\"t\"\>", "12.1`"},
                    {"\<\"driveIntent\"\>", 
                    RowBox[{"-", "0.5`"}]},
                    {"\<\"motorTorque\"\>", "39.43984048420173`"},
                    {"\<\"vw\"\>", "0.09174540652009666`"},
                    {"\<\"controllerState\"\>", 
                    RowBox[{"\[LeftAssociation]", 
                    RowBox[{
                    RowBox[{"\<\"lastError\"\>", "\[Rule]", 
                    RowBox[{"-", "0.0947676229474259`"}]}], ",", 
                    RowBox[{"\<\"lastInt\"\>", "\[Rule]", 
                    "0.04139526915872063`"}], ",", 
                    RowBox[{"\<\"initialized\"\>", "\[Rule]", "True"}]}], 
                    "\[RightAssociation]"}]}
                    },
                    AutoDelete->False,
                    
                    GridBoxItemSize->{
                    "Columns" -> {{Automatic}}, "Rows" -> {{Automatic}}}],
                    "Grid"], {-1, -0.5}], 
                  {GrayLevel[0], LineBox[{{-5, 0}, {5, 0}}]}, 
                  {GrayLevel[0.5], CircleBox[{0.08003175161374303, 0.5}, 0.5],
                    DiskBox[{0.08003175161374303, 0.5}, 0.1]}, 
                  {GrayLevel[0.5], 
                   DiskBox[{-0.045420013616922314`, 1.9947447456340197`}, 
                    0.1]}, 
                  {GrayLevel[0.5], Thickness[0.01], 
                   LineBox[{{0.08003175161374303, 
                    0.5}, {-0.045420013616922314`, 1.9947447456340197`}}]}, 
                  {RGBColor[0, 1, 0], 
                   DiskBox[{-0.0036027585400338624`, 1.4964964970893464`}, 
                    0.05]}, 
                  {RGBColor[0, 0, 1], Arrowheads[0.01], 
                   CircleBox[{0.08003175161374303, 0.5}, 0.3333333333333333, 
                    NCache[{
                    Rational[-1, 2] 
                    Pi, -0.3876011122688445}, {-1.5707963267948966`, \
-0.3876011122688445}]], 
                   ArrowBox[{{0.38736258954518427`, 0.37093076599177754`}, {
                    0.3886378939599926, 0.37401047655746344`}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{0.08003175161374303, 0.5}, {-0.7087650580702917,
                     0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{0.08003175161374303, 0.5}, {0.6930074158966201, 
                    0.5}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.005], Arrowheads[0.04], 
                   ArrowBox[{{0.08003175161374303, 
                    0.5}, {-0.09578939378741459, 0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{0.08003175161374303, 0.5}, 1.6, {
                    1.654528645281036, 2.0389853807495686`}], 
                   ArrowBox[{{-0.6276877318363074, 
                    1.9349679901465378`}, {-0.6420017868983463, 
                    1.927819165463089}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Dashing[{Small, Small}], 
                   Arrowheads[0.02], 
                   CircleBox[{0.08003175161374303, 0.5}, 1.7, {
                    1.654528645281036, 1.5529977836963682`}], 
                   ArrowBox[{{0.09328914050057807, 2.199948305578703}, {
                    0.1102876773691437, 2.199730737192419}}]}, 
                  {RGBColor[0, 0, 1], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{0.08003175161374303, 0.5}, 1.8, {
                    1.654528645281036, 1.435418420368804}], 
                   ArrowBox[{{0.30512117984032294`, 2.2858708657964693`}, {
                    0.32296833647700723`, 2.283530716229741}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.003], Arrowheads[0.025], 
                   CircleBox[{0.08003175161374303, 0.5}, 1.5, {
                    1.654528645281036, 1.686615250090169}], 
                   ArrowBox[{{-0.07840056846740955, 
                    1.9916096003826547`}, {-0.09330849432088824, 
                    1.9899507237285807`}}]}, 
                  {RGBColor[0, 0, 1], ArrowBox[{{0, 2.8}, {-1.5, 2.8}}]}},
                 Frame->True,
                 FrameTicks->Automatic,
                 ImageSize->800,
                 PlotRange->{{-3, 3}, {-1, 3}}], 244->
                GraphicsBox[{InsetBox[
                   TagBox[GridBox[{
                    {"\<\"t\"\>", "12.15`"},
                    {"\<\"driveIntent\"\>", 
                    RowBox[{"-", "0.5`"}]},
                    {"\<\"motorTorque\"\>", "72.22206833157314`"},
                    {"\<\"vw\"\>", "0.0859493585173567`"},
                    {"\<\"controllerState\"\>", 
                    RowBox[{"\[LeftAssociation]", 
                    RowBox[{
                    RowBox[{"\<\"lastError\"\>", "\[Rule]", 
                    RowBox[{"-", "0.04570720299817016`"}]}], ",", 
                    
                    RowBox[{"\<\"lastInt\"\>", "\[Rule]", 
                    "0.03791050706868803`"}], ",", 
                    RowBox[{"\<\"initialized\"\>", "\[Rule]", "True"}]}], 
                    "\[RightAssociation]"}]}
                    },
                    AutoDelete->False,
                    
                    GridBoxItemSize->{
                    "Columns" -> {{Automatic}}, "Rows" -> {{Automatic}}}],
                    "Grid"], {-1, -0.5}], 
                  {GrayLevel[0], LineBox[{{-5, 0}, {5, 0}}]}, 
                  {GrayLevel[0.5], CircleBox[{0.08449306154715758, 0.5}, 0.5],
                    DiskBox[{0.08449306154715758, 0.5}, 0.1]}, 
                  {GrayLevel[0.5], 
                   DiskBox[{-0.11150891951861314`, 1.9871392750574148`}, 
                    0.1]}, 
                  {GrayLevel[0.5], Thickness[0.01], 
                   LineBox[{{0.08449306154715758, 
                    0.5}, {-0.11150891951861314`, 1.9871392750574148`}}]}, 
                  {RGBColor[0, 1, 0], 
                   DiskBox[{-0.04617492583002289, 1.49142618337161}, 0.05]}, 
                  {RGBColor[0, 0, 1], Arrowheads[0.01], 
                   CircleBox[{0.08449306154715758, 0.5}, 0.3333333333333333, 
                    NCache[{
                    Rational[-1, 2] Pi, 
                    0.5958657231522979}, {-1.5707963267948966`, 
                    0.5958657231522979}]], 
                   ArrowBox[{{0.3622376354581456, 0.6843069797216449}, {
                    0.3603807092656365, 0.6870751638980321}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{0.08449306154715758, 
                    0.5}, {-1.3599483050843053`, 0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{0.08449306154715758, 0.5}, {1.0373145110258553`,
                     0.5}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.005], Arrowheads[0.04], 
                   ArrowBox[{{0.08449306154715758, 0.5}, {-0.4071268556056076,
                     0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{0.08449306154715758, 0.5}, 1.6, {
                    1.7018390407433674`, 2.3025024243741923`}], 
                   ArrowBox[{{-0.972573123454167, 
                    1.7010874574845687`}, {-0.9845309449799631, 
                    1.6904569179389495`}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Dashing[{Small, Small}], 
                   Arrowheads[0.02], 
                   CircleBox[{0.08449306154715758, 0.5}, 1.7, {
                    1.7018390407433674`, 1.4194215150923144`}], 
                   ArrowBox[{{0.3240304444420654, 2.1830394654302845`}, {
                    0.3408485818218554, 2.180559980252026}}]}, 
                  {RGBColor[0, 0, 1], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{0.08449306154715758, 0.5}, 1.8, {
                    1.7018390407433674`, 1.3006053277901835`}], 
                   ArrowBox[{{0.5475702896535839, 2.23941354508043}, {
                    0.564940981535123, 2.234695880026017}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.003], Arrowheads[0.025], 
                   CircleBox[{0.08449306154715758, 0.5}, 1.5, {
                    1.7018390407433674`, 1.5409929986678998`}], 
                   ArrowBox[{{0.11419611219899027`, 1.9997058807586154`}, {
                    0.1291914359166897, 1.9993338705334183`}}]}, 
                  {RGBColor[0, 0, 1], ArrowBox[{{0, 2.8}, {-1.5, 2.8}}]}},
                 Frame->True,
                 FrameTicks->Automatic,
                 ImageSize->800,
                 PlotRange->{{-3, 3}, {-1, 3}}], 245->
                GraphicsBox[{InsetBox[
                   TagBox[GridBox[{
                    {"\<\"t\"\>", "12.200000000000001`"},
                    {"\<\"driveIntent\"\>", 
                    RowBox[{"-", "0.5`"}]},
                    {"\<\"motorTorque\"\>", "95.91730314409064`"},
                    {"\<\"vw\"\>", "0.07622361937568548`"},
                    {"\<\"controllerState\"\>", 
                    RowBox[{"\[LeftAssociation]", 
                    RowBox[{
                    
                    RowBox[{"\<\"lastError\"\>", "\[Rule]", 
                    "0.0006023180364603808`"}], ",", 
                    
                    RowBox[{"\<\"lastInt\"\>", "\[Rule]", 
                    "0.036823993024686065`"}], ",", 
                    RowBox[{"\<\"initialized\"\>", "\[Rule]", "True"}]}], 
                    "\[RightAssociation]"}]}
                    },
                    AutoDelete->False,
                    
                    GridBoxItemSize->{
                    "Columns" -> {{Automatic}}, "Rows" -> {{Automatic}}}],
                    "Grid"], {-1, -0.5}], 
                  {GrayLevel[0], LineBox[{{-5, 0}, {5, 0}}]}, 
                  {GrayLevel[0.5], CircleBox[{0.0885547337342312, 0.5}, 0.5], 
                   DiskBox[{0.0885547337342312, 0.5}, 0.1]}, 
                  {GrayLevel[0.5], 
                   DiskBox[{-0.17164537802902238`, 1.9772595918924982`}, 
                    0.1]}, 
                  {GrayLevel[0.5], Thickness[0.01], 
                   LineBox[{{0.0885547337342312, 0.5}, {-0.17164537802902238`,
                     1.9772595918924982`}}]}, 
                  {RGBColor[0, 1, 0], 
                   DiskBox[{-0.0849120074412712, 1.4848397279283323`}, 0.05]}, 
                  {RGBColor[0, 0, 1], Arrowheads[0.01], 
                   CircleBox[{0.0885547337342312, 0.5}, 0.3333333333333333, 
                    NCache[{
                    Rational[-1, 2] Pi, 
                    1.3067227675278223`}, {-1.5707963267948966`, 
                    1.3067227675278223`}]], 
                   ArrowBox[{{0.178773130615848, 0.8208921189048154}, {
                    0.17555975202629906`, 0.8217782433650667}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{0.0885547337342312, 0.5}, {-1.8297913291475816`,
                     0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{0.0885547337342312, 0.5}, {1.3450582786150604`, 
                    0.5}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.005], Arrowheads[0.04], 
                   ArrowBox[{{0.0885547337342312, 0.5}, {-0.5732877842667523, 
                    0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{0.0885547337342312, 0.5}, 1.6, {
                    1.7451450195439338`, 2.542548590176156}], 
                   ArrowBox[{{-1.2237576954227658`, 
                    1.4153338671108275`}, {-1.2328452664643295`, 
                    1.4021651952249345`}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Dashing[{Small, Small}], 
                   Arrowheads[0.02], 
                   CircleBox[{0.0885547337342312, 0.5}, 1.7, {
                    1.7451450195439338`, 1.3675231869751796`}], 
                   ArrowBox[{{0.4150773287898809, 2.168347384365176}, {
                    0.43174419858332336`, 2.1649987961607042`}}]}, 
                  {RGBColor[0, 0, 1], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{0.0885547337342312, 0.5}, 1.8, {
                    1.7451450195439338`, 1.2122711131878745`}], 
                   ArrowBox[{{0.7032764434665183, 2.1917793058149795`}, {
                    0.7201632187335069, 2.1855476029103778`}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.003], Arrowheads[0.025], 
                   CircleBox[{0.0885547337342312, 0.5}, 1.5, {
                    1.7451450195439338`, 1.5282887303464452`}], 
                   ArrowBox[{{0.1373075408101972, 1.99920751192162}, {
                    0.15229692844270387`, 1.998645032225359}}]}, 
                  {RGBColor[0, 0, 1], ArrowBox[{{0, 2.8}, {-1.5, 2.8}}]}},
                 Frame->True,
                 FrameTicks->Automatic,
                 ImageSize->800,
                 PlotRange->{{-3, 3}, {-1, 3}}], 246->
                GraphicsBox[{InsetBox[
                   TagBox[GridBox[{
                    {"\<\"t\"\>", "12.25`"},
                    {"\<\"driveIntent\"\>", 
                    RowBox[{"-", "0.5`"}]},
                    {"\<\"motorTorque\"\>", "114.37800902593082`"},
                    {"\<\"vw\"\>", "0.06406016823931489`"},
                    {"\<\"controllerState\"\>", 
                    RowBox[{"\[LeftAssociation]", 
                    RowBox[{
                    
                    RowBox[{"\<\"lastError\"\>", "\[Rule]", 
                    "0.0419750519624901`"}], ",", 
                    
                    RowBox[{"\<\"lastInt\"\>", "\[Rule]", 
                    "0.03793151655947725`"}], ",", 
                    RowBox[{"\<\"initialized\"\>", "\[Rule]", "True"}]}], 
                    "\[RightAssociation]"}]}
                    },
                    AutoDelete->False,
                    
                    GridBoxItemSize->{
                    "Columns" -> {{Automatic}}, "Rows" -> {{Automatic}}}],
                    "Grid"], {-1, -0.5}], 
                  {GrayLevel[0], LineBox[{{-5, 0}, {5, 0}}]}, 
                  {GrayLevel[0.5], CircleBox[{0.09206423692813288, 0.5}, 0.5],
                    DiskBox[{0.09206423692813288, 0.5}, 0.1]}, 
                  {GrayLevel[0.5], 
                   DiskBox[{-0.2235077144975528, 1.9664291130066207`}, 0.1]}, 
                  {GrayLevel[0.5], Thickness[0.01], 
                   LineBox[{{0.09206423692813288, 0.5}, {-0.2235077144975528, 
                    1.9664291130066207`}}]}, 
                  {RGBColor[0, 1, 0], 
                   DiskBox[{-0.11831706402232424`, 1.4776194086710805`}, 
                    0.05]}, 
                  {RGBColor[0, 0, 1], Arrowheads[0.01], 
                   CircleBox[{0.09206423692813288, 0.5}, 0.3333333333333333, 
                    NCache[{
                    Rational[-1, 2] Pi, 
                    1.8605439439830285`}, {-1.5707963267948966`, 
                    1.8605439439830285`}]], 
                   ArrowBox[{{0.000026541631038809466`, 
                    0.8203750517058255}, {-0.0031725536440285335`, 
                    0.8194386714732976}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{0.09206423692813288, 
                    0.5}, {-2.1954959435904837`, 0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{0.09206423692813288, 0.5}, {1.6047854265079415`,
                     0.5}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.005], Arrowheads[0.04], 
                   ArrowBox[{{0.09206423692813288, 0.5}, {-0.6827747540106749,
                     0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{0.09206423692813288, 0.5}, 1.6, {
                    1.7827613002496516`, 2.7498561994523705`}], 
                   ArrowBox[{{-1.380549200873801, 
                    1.125627415324145}, {-1.3867317406980246`, 
                    1.110870245270383}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Dashing[{Small, Small}], 
                   Arrowheads[0.02], 
                   CircleBox[{0.09206423692813288, 0.5}, 1.7, {
                    1.7827613002496516`, 1.343982125175255}], 
                   ArrowBox[{{0.4577674011863625, 2.160199143371517}, {
                    0.4743508309157669, 2.156459163413736}}]}, 
                  {RGBColor[0, 0, 1], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{0.09206423692813288, 0.5}, 1.8, {
                    1.7827613002496516`, 1.147327916772258}], 
                   ArrowBox[{{0.8152824120255209, 2.1483189834521728`}, {
                    0.8317291665341706, 2.1410045069745833`}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.003], Arrowheads[0.025], 
                   CircleBox[{0.09206423692813288, 0.5}, 1.5, {
                    1.7827613002496516`, 1.5545260674753634`}], 
                   ArrowBox[{{0.10146956427693687`, 1.9999705129826593`}, {
                    0.11646854915047938`, 1.9998014633760544`}}]}, 
                  {RGBColor[0, 0, 1], ArrowBox[{{0, 2.8}, {-1.5, 2.8}}]}},
                 Frame->True,
                 FrameTicks->Automatic,
                 ImageSize->800,
                 PlotRange->{{-3, 3}, {-1, 3}}], 247->
                GraphicsBox[{InsetBox[
                   TagBox[GridBox[{
                    {"\<\"t\"\>", "12.3`"},
                    {"\<\"driveIntent\"\>", 
                    RowBox[{"-", "0.5`"}]},
                    {"\<\"motorTorque\"\>", "128.93588814472108`"},
                    {"\<\"vw\"\>", "0.05010959843103226`"},
                    {"\<\"controllerState\"\>", 
                    RowBox[{"\[LeftAssociation]", 
                    RowBox[{
                    
                    RowBox[{"\<\"lastError\"\>", "\[Rule]", 
                    "0.07783571547936732`"}], ",", 
                    
                    RowBox[{"\<\"lastInt\"\>", "\[Rule]", 
                    "0.04096796040680263`"}], ",", 
                    RowBox[{"\<\"initialized\"\>", "\[Rule]", "True"}]}], 
                    "\[RightAssociation]"}]}
                    },
                    AutoDelete->False,
                    
                    GridBoxItemSize->{
                    "Columns" -> {{Automatic}}, "Rows" -> {{Automatic}}}],
                    "Grid"], {-1, -0.5}], 
                  {GrayLevel[0], LineBox[{{-5, 0}, {5, 0}}]}, 
                  {GrayLevel[0.5], CircleBox[{0.09491801328966797, 0.5}, 0.5],
                    DiskBox[{0.09491801328966797, 0.5}, 0.1]}, 
                  {GrayLevel[0.5], 
                   DiskBox[{-0.2667753969033213, 1.955739632291079}, 0.1]}, 
                  {GrayLevel[0.5], Thickness[0.01], 
                   LineBox[{{0.09491801328966797, 0.5}, {-0.2667753969033213, 
                    1.955739632291079}}]}, 
                  {RGBColor[0, 1, 0], 
                   DiskBox[{-0.14621092683899156`, 1.4704930881940528`}, 
                    0.05]}, 
                  {RGBColor[0, 0, 1], Arrowheads[0.01], 
                   CircleBox[{0.09491801328966797, 0.5}, 0.3333333333333333, 
                    NCache[{
                    Rational[-1, 2] Pi, 
                    2.297280317546736}, {-1.5707963267948966`, 
                    2.297280317546736}]], 
                   ArrowBox[{{-0.12399440196467375`, 
                    0.751373159980579}, {-0.12649714613961277`, 
                    0.7491715037599952}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{0.09491801328966797, 
                    0.5}, {-2.4837997496047537`, 0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{0.09491801328966797, 0.5}, {1.8160878356882366`,
                     0.5}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.005], Arrowheads[0.04], 
                   ArrowBox[{{0.09491801328966797, 0.5}, {-0.7626299272061849,
                     0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{0.09491801328966797, 0.5}, 1.6, {
                    1.8143252745091818`, 2.922762987343269}], 
                   ArrowBox[{{-1.4633740473969623`, 
                    0.8629405648326125}, {-1.4669254786017485`, 
                    0.8473397570627745}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Dashing[{Small, Small}], 
                   Arrowheads[0.02], 
                   CircleBox[{0.09491801328966797, 0.5}, 1.7, {
                    1.8143252745091818`, 1.3323293676566546`}], 
                   ArrowBox[{{0.4799418092391618, 2.155825074261361}, {
                    0.49648053298293815`, 2.15189210990839}}]}, 
                  {RGBColor[0, 0, 1], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{0.09491801328966797, 0.5}, 1.8, {
                    1.8143252745091818`, 1.0980147848162871`}], 
                   ArrowBox[{{0.8985078424567434, 2.1106655104208416`}, {
                    0.9145740499614119, 2.1025492134557267`}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.003], Arrowheads[0.025], 
                   CircleBox[{0.09491801328966797, 0.5}, 1.5, {
                    1.8143252745091818`, 1.5908242194678697`}], 
                   ArrowBox[{{0.04988294282011494, 1.9993237950582263`}, {
                    0.0648781826194064, 1.9996991726920776`}}]}, 
                  {RGBColor[0, 0, 1], ArrowBox[{{0, 2.8}, {-1.5, 2.8}}]}},
                 Frame->True,
                 FrameTicks->Automatic,
                 ImageSize->800,
                 PlotRange->{{-3, 3}, {-1, 3}}], 248->
                GraphicsBox[{InsetBox[
                   TagBox[GridBox[{
                    {"\<\"t\"\>", "12.35`"},
                    {"\<\"driveIntent\"\>", 
                    RowBox[{"-", "0.5`"}]},
                    {"\<\"motorTorque\"\>", "140.18300583508378`"},
                    {"\<\"vw\"\>", "0.03478133165186023`"},
                    {"\<\"controllerState\"\>", 
                    RowBox[{"\[LeftAssociation]", 
                    RowBox[{
                    
                    RowBox[{"\<\"lastError\"\>", "\[Rule]", 
                    "0.10816247998833226`"}], ",", 
                    
                    RowBox[{"\<\"lastInt\"\>", "\[Rule]", 
                    "0.04565581901906807`"}], ",", 
                    RowBox[{"\<\"initialized\"\>", "\[Rule]", "True"}]}], 
                    "\[RightAssociation]"}]}
                    },
                    AutoDelete->False,
                    
                    GridBoxItemSize->{
                    "Columns" -> {{Automatic}}, "Rows" -> {{Automatic}}}],
                    "Grid"], {-1, -0.5}], 
                  {GrayLevel[0], LineBox[{{-5, 0}, {5, 0}}]}, 
                  {GrayLevel[0.5], CircleBox[{0.09703761954661813, 0.5}, 0.5],
                    DiskBox[{0.09703761954661813, 0.5}, 0.1]}, 
                  {GrayLevel[0.5], 
                   DiskBox[{-0.30182787091339547`, 1.9459966530113728`}, 
                    0.1]}, 
                  {GrayLevel[0.5], Thickness[0.01], 
                   LineBox[{{0.09703761954661813, 
                    0.5}, {-0.30182787091339547`, 1.9459966530113728`}}]}, 
                  {RGBColor[0, 1, 0], 
                   DiskBox[{-0.1688727074267243, 1.4639977686742487`}, 0.05]}, 
                  {RGBColor[0, 0, 1], Arrowheads[0.01], 
                   CircleBox[{0.09703761954661813, 0.5}, 0.3333333333333333, 
                    NCache[{
                    Rational[-1, 2] Pi, 
                    2.6346938482576174`}, {-1.5707963267948966`, 
                    2.6346938482576174`}]], 
                   ArrowBox[{{-0.1927476891731679, 
                    0.6647288255324167}, {-0.19438046082913255`, 
                    0.6618227843698891}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{0.09703761954661813, 
                    0.5}, {-2.7066224971550574`, 0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{0.09703761954661813, 0.5}, {1.982392769886547, 
                    0.5}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.005], Arrowheads[0.04], 
                   ArrowBox[{{0.09703761954661813, 0.5}, {-0.8212673468151286,
                     0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{0.09703761954661813, 0.5}, 1.6, {
                    1.8399444510685987`, 3.062298907949753}], 
                   ArrowBox[{{-1.4965879192305562`, 
                    0.6426802094095808}, {-1.4979350169318053`, 
                    0.6267370856737172}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Dashing[{Small, Small}], 
                   Arrowheads[0.02], 
                   CircleBox[{0.09703761954661813, 0.5}, 1.7, {
                    1.8399444510685987`, 1.327332503683285}], 
                   ArrowBox[{{0.49033050697232916`, 2.1538804989177263`}, {
                    0.506849371835635, 2.1498649422561247`}}]}, 
                  {RGBColor[0, 0, 1], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{0.09703761954661813, 0.5}, 1.8, {
                    1.8399444510685987`, 1.061149974207022}], 
                   ArrowBox[{{0.9594448964382752, 2.079953698297617}, {
                    0.9752010500924422, 2.0712507722359246`}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.003], Arrowheads[0.025], 
                   CircleBox[{0.09703761954661813, 0.5}, 1.5, {
                    1.8399444510685987`, 1.6282756258045823`}], 
                   ArrowBox[{{-0.004104530453650931, 1.9965862038296769`}, {
                    0.010866139220216633`, 1.9975227797861228`}}]}, 
                  {RGBColor[0, 0, 1], ArrowBox[{{0, 2.8}, {-1.5, 2.8}}]}},
                 Frame->True,
                 FrameTicks->Automatic,
                 ImageSize->800,
                 PlotRange->{{-3, 3}, {-1, 3}}], 249->
                GraphicsBox[{InsetBox[
                   TagBox[GridBox[{
                    {"\<\"t\"\>", "12.4`"},
                    {"\<\"driveIntent\"\>", 
                    RowBox[{"-", "0.5`"}]},
                    {"\<\"motorTorque\"\>", "148.5111054448895`"},
                    {"\<\"vw\"\>", "0.01842054653042696`"},
                    {"\<\"controllerState\"\>", 
                    RowBox[{"\[LeftAssociation]", 
                    RowBox[{
                    
                    RowBox[{"\<\"lastError\"\>", "\[Rule]", 
                    "0.13315753044152825`"}], ",", 
                    
                    RowBox[{"\<\"lastInt\"\>", "\[Rule]", 
                    "0.05172292953667976`"}], ",", 
                    RowBox[{"\<\"initialized\"\>", "\[Rule]", "True"}]}], 
                    "\[RightAssociation]"}]}
                    },
                    AutoDelete->False,
                    
                    GridBoxItemSize->{
                    "Columns" -> {{Automatic}}, "Rows" -> {{Automatic}}}],
                    "Grid"], {-1, -0.5}], 
                  {GrayLevel[0], LineBox[{{-5, 0}, {5, 0}}]}, 
                  {GrayLevel[0.5], CircleBox[{0.09836309941014121, 0.5}, 0.5],
                    DiskBox[{0.09836309941014121, 0.5}, 0.1]}, 
                  {GrayLevel[0.5], 
                   DiskBox[{-0.3293163629178698, 1.937737902923486}, 0.1]}, 
                  {GrayLevel[0.5], Thickness[0.01], 
                   LineBox[{{0.09836309941014121, 0.5}, {-0.3293163629178698, 
                    1.937737902923486}}]}, 
                  {RGBColor[0, 1, 0], 
                   DiskBox[{-0.18675654214186616`, 1.458491935282324}, 0.05]}, 
                  {RGBColor[0, 0, 1], Arrowheads[0.01], 
                   CircleBox[{0.09836309941014121, 0.5}, 0.3333333333333333, 
                    NCache[{
                    Rational[-1, 2] Pi, 
                    2.8845368365517885`}, {-1.5707963267948966`, 
                    2.8845368365517885`}]], 
                   ArrowBox[{{-0.22315424202314044`, 
                    0.5879642556313966}, {-0.22401779418571188`, 
                    0.5847447376268896}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{0.09836309941014121, 
                    0.5}, {-2.8718590094876486`, 0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{0.09836309941014121, 0.5}, {2.1083699538179372`,
                     0.5}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.005], Arrowheads[0.04], 
                   ArrowBox[{{0.09836309941014121, 0.5}, {-0.8618521550798531,
                     0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{0.09836309941014121, 0.5}, 1.6, {
                    1.859927575352184, 3.170584700114368}], 
                   ArrowBox[{{-1.5013483509983296`, 
                    0.4696145522983989}, {-1.5009645166795533`, 
                    0.45361922367128105`}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Dashing[{Small, Small}], 
                   Arrowheads[0.02], 
                   CircleBox[{0.09836309941014121, 0.5}, 1.7, {
                    1.859927575352184, 1.3270522887640777`}], 
                   ArrowBox[{{0.49211941337941273`, 2.153770227452814}, {
                    0.5086371523753139, 2.1497500421165414`}}]}, 
                  {RGBColor[0, 0, 1], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{0.09836309941014121, 0.5}, 1.8, {
                    1.859927575352184, 1.034865878436131}], 
                   ArrowBox[{{1.0019953677227127`, 2.056742985744364}, {
                    1.017517356887386, 2.0476289771652167`}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.003], Arrowheads[0.025], 
                   CircleBox[{0.09836309941014121, 0.5}, 1.5, {
                    1.859927575352184, 1.6639624806660631`}], 
                   ArrowBox[{{-0.056111771512819014`, 
                    1.9920246359404843`}, {-0.04118405014376164, 
                    1.9934947582939087`}}]}, 
                  {RGBColor[0, 0, 1], ArrowBox[{{0, 2.8}, {-1.5, 2.8}}]}},
                 Frame->True,
                 FrameTicks->Automatic,
                 ImageSize->800,
                 PlotRange->{{-3, 3}, {-1, 3}}], 250->
                GraphicsBox[{InsetBox[
                   TagBox[GridBox[{
                    {"\<\"t\"\>", "12.450000000000001`"},
                    {"\<\"driveIntent\"\>", 
                    RowBox[{"-", "0.5`"}]},
                    {"\<\"motorTorque\"\>", "154.26354689392568`"},
                    {"\<\"vw\"\>", "0.0013512087184875455`"},
                    {"\<\"controllerState\"\>", 
                    RowBox[{"\[LeftAssociation]", 
                    RowBox[{
                    
                    RowBox[{"\<\"lastError\"\>", "\[Rule]", 
                    "0.1531373244770944`"}], ",", 
                    
                    RowBox[{"\<\"lastInt\"\>", "\[Rule]", 
                    "0.05891041801525079`"}], ",", 
                    RowBox[{"\<\"initialized\"\>", "\[Rule]", "True"}]}], 
                    "\[RightAssociation]"}]}
                    },
                    AutoDelete->False,
                    
                    GridBoxItemSize->{
                    "Columns" -> {{Automatic}}, "Rows" -> {{Automatic}}}],
                    "Grid"], {-1, -0.5}], 
                  {GrayLevel[0], LineBox[{{-5, 0}, {5, 0}}]}, 
                  {GrayLevel[0.5], CircleBox[{0.09885115776468698, 0.5}, 0.5],
                    DiskBox[{0.09885115776468698, 0.5}, 0.1]}, 
                  {GrayLevel[0.5], 
                   DiskBox[{-0.35000544981633375`, 1.9312678805279107`}, 
                    0.1]}, 
                  {GrayLevel[0.5], Thickness[0.01], 
                   LineBox[{{0.09885115776468698, 
                    0.5}, {-0.35000544981633375`, 1.9312678805279107`}}]}, 
                  {RGBColor[0, 1, 0], 
                   DiskBox[{-0.20038658062266015`, 1.4541785870186072`}, 
                    0.05]}, 
                  {RGBColor[0, 0, 1], Arrowheads[0.01], 
                   CircleBox[{0.09885115776468698, 0.5}, 0.3333333333333333, 
                    NCache[{
                    Rational[-1, 2] Pi, 
                    3.0571100800228734`}, {-1.5707963267948966`, 
                    3.0571100800228734`}]], 
                   ArrowBox[{{-0.23299545593465695`, 
                    0.5314473542192264}, {-0.2332933320432337, 
                    0.5281273710351175}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{0.09885115776468698, 
                    0.5}, {-2.9864197801138266`, 0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{0.09885115776468698, 0.5}, {2.1988931015171986`,
                     0.5}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.005], Arrowheads[0.04], 
                   ArrowBox[{{0.09885115776468698, 0.5}, {-0.8863778363613147,
                     0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{0.09885115776468698, 0.5}, 1.6, {
                    1.8746900138107476`, 3.250246154665253}], 
                   ArrowBox[{{-1.49336914435513, 
                    0.34241031278196055`}, {-1.491713663396083, 
                    0.326496254548185}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Dashing[{Small, Small}], 
                   Arrowheads[0.02], 
                   CircleBox[{0.09885115776468698, 0.5}, 1.7, {
                    1.8746900138107476`, 1.3304504927605858`}], 
                   ArrowBox[{{0.48698536044835106`, 2.1550987404705246`}, {
                    0.5035166654562334, 2.151134708884997}}]}, 
                  {RGBColor[0, 0, 1], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{0.09885115776468698, 0.5}, 1.8, {
                    1.8746900138107476`, 1.0176703088444938`}], 
                   ArrowBox[{{1.0291175960874588`, 2.0409751307955184`}, {
                    1.0444805776332031`, 2.031595573341453}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.003], Arrowheads[0.025], 
                   CircleBox[{0.09885115776468698, 0.5}, 1.5, {
                    1.8746900138107476`, 1.6967214609058654`}], 
                   ArrowBox[{{-0.10440929708917354`, 
                    1.9861645896375681`}, {-0.08953773594760653, 
                    1.9881228526992851`}}]}, 
                  {RGBColor[0, 0, 1], ArrowBox[{{0, 2.8}, {-1.5, 2.8}}]}},
                 Frame->True,
                 FrameTicks->Automatic,
                 ImageSize->800,
                 PlotRange->{{-3, 3}, {-1, 3}}], 251->
                GraphicsBox[{InsetBox[
                   TagBox[GridBox[{
                    {"\<\"t\"\>", "12.5`"},
                    {"\<\"driveIntent\"\>", 
                    RowBox[{"-", "0.5`"}]},
                    {"\<\"motorTorque\"\>", "157.76872414062893`"},
                    {"\<\"vw\"\>", 
                    RowBox[{"-", "0.016122000663663743`"}]},
                    {"\<\"controllerState\"\>", 
                    RowBox[{"\[LeftAssociation]", 
                    RowBox[{
                    
                    RowBox[{"\<\"lastError\"\>", "\[Rule]", 
                    "0.1684868795485652`"}], ",", 
                    
                    RowBox[{"\<\"lastInt\"\>", "\[Rule]", 
                    "0.06697711816462945`"}], ",", 
                    RowBox[{"\<\"initialized\"\>", "\[Rule]", "True"}]}], 
                    "\[RightAssociation]"}]}
                    },
                    AutoDelete->False,
                    
                    GridBoxItemSize->{
                    "Columns" -> {{Automatic}}, "Rows" -> {{Automatic}}}],
                    "Grid"], {-1, -0.5}], 
                  {GrayLevel[0], LineBox[{{-5, 0}, {5, 0}}]}, 
                  {GrayLevel[0.5], CircleBox[{0.09847423186656044, 0.5}, 0.5],
                    DiskBox[{0.09847423186656044, 0.5}, 0.1]}, 
                  {GrayLevel[0.5], 
                   DiskBox[{-0.3646998157398341, 1.92669891765008}, 0.1]}, 
                  {GrayLevel[0.5], Thickness[0.01], 
                   LineBox[{{0.09847423186656044, 0.5}, {-0.3646998157398341, 
                    1.92669891765008}}]}, 
                  {RGBColor[0, 1, 0], 
                   DiskBox[{-0.21030846653770258`, 1.4511326117667198`}, 
                    0.05]}, 
                  {RGBColor[0, 0, 1], Arrowheads[0.01], 
                   CircleBox[{0.09847423186656044, 0.5}, 0.3333333333333333, 
                    NCache[{
                    Rational[-1, 2] Pi, 
                    3.1622653974239716`}, {-1.5707963267948966`, 
                    3.1622653974239716`}]], 
                   ArrowBox[{{-0.2348401170701543, 
                    0.4964424862607309}, {-0.23478787694711123`, 
                    0.4931095761976822}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{0.09847423186656044, 
                    0.5}, {-3.0569002509460184`, 0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{0.09847423186656044, 0.5}, {2.2585847410243107`,
                     0.5}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.005], Arrowheads[0.04], 
                   ArrowBox[{{0.09847423186656044, 0.5}, {-0.896789741788268, 
                    0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{0.09847423186656044, 0.5}, 1.6, {
                    1.8847092486478143`, 3.30414230301008}], 
                   ArrowBox[{{-1.482944727564922, 
                    0.256866138206444}, {-1.480434359180045, 
                    0.2410643687724225}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Dashing[{Small, Small}], 
                   Arrowheads[0.02], 
                   CircleBox[{0.09847423186656044, 0.5}, 1.7, {
                    1.8847092486478143`, 1.3367221121751123`}], 
                   ArrowBox[{{0.4762207199467836, 2.1575004044473287`}, {
                    0.4927765605755606, 2.1536401281943602`}}]}, 
                  {RGBColor[0, 0, 1], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{0.09847423186656044, 0.5}, 1.8, {
                    1.8847092486478143`, 1.008216336755431}], 
                   ArrowBox[{{1.043267216877283, 2.0321116850525387`}, {
                    1.0585408391215494`, 2.022587307721267}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.003], Arrowheads[0.025], 
                   CircleBox[{0.09847423186656044, 0.5}, 1.5, {
                    1.8847092486478143`, 1.7259275910276421`}], 
                   ArrowBox[{{-0.1480984846688867, 
                    1.979595179588096}, {-0.13329045093788083`, 
                    1.9819868865157197`}}]}, 
                  {RGBColor[0, 0, 1], ArrowBox[{{0, 2.8}, {-1.5, 2.8}}]}},
                 Frame->True,
                 FrameTicks->Automatic,
                 ImageSize->800,
                 PlotRange->{{-3, 3}, {-1, 3}}], 252->
                GraphicsBox[{InsetBox[
                   TagBox[GridBox[{
                    {"\<\"t\"\>", "12.55`"},
                    {"\<\"driveIntent\"\>", 
                    RowBox[{"-", "0.5`"}]},
                    {"\<\"motorTorque\"\>", "159.33984655212086`"},
                    {"\<\"vw\"\>", 
                    RowBox[{"-", "0.033721979108793834`"}]},
                    {"\<\"controllerState\"\>", 
                    RowBox[{"\[LeftAssociation]", 
                    RowBox[{
                    
                    RowBox[{"\<\"lastError\"\>", "\[Rule]", 
                    "0.17963213003186268`"}], ",", 
                    
                    RowBox[{"\<\"lastInt\"\>", "\[Rule]", 
                    "0.07570225720856283`"}], ",", 
                    RowBox[{"\<\"initialized\"\>", "\[Rule]", "True"}]}], 
                    "\[RightAssociation]"}]}
                    },
                    AutoDelete->False,
                    
                    GridBoxItemSize->{
                    "Columns" -> {{Automatic}}, "Rows" -> {{Automatic}}}],
                    "Grid"], {-1, -0.5}], 
                  {GrayLevel[0], LineBox[{{-5, 0}, {5, 0}}]}, 
                  {GrayLevel[0.5], CircleBox[{0.09721932786775561, 0.5}, 0.5],
                    DiskBox[{0.09721932786775561, 0.5}, 0.1]}, 
                  {GrayLevel[0.5], 
                   DiskBox[{-0.3742013446221694, 1.9239952772215037`}, 0.1]}, 
                  {GrayLevel[0.5], Thickness[0.01], 
                   LineBox[{{0.09721932786775561, 0.5}, {-0.3742013446221694, 
                    1.9239952772215037`}}]}, 
                  {RGBColor[0, 1, 0], 
                   DiskBox[{-0.21706112045886106`, 1.4493301848143356`}, 
                    0.05]}, 
                  {RGBColor[0, 0, 1], Arrowheads[0.01], 
                   CircleBox[{0.09721932786775561, 0.5}, 0.3333333333333333, 
                    NCache[{
                    Rational[-1, 2] Pi, 
                    3.209399069768729}, {-1.5707963267948966`, 
                    3.209399069768729}]], 
                   ArrowBox[{{-0.23555723024207065`, 
                    0.48074192419558687`}, {-0.23534801400444072`, 
                    0.47741517697273705`}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{0.09721932786775561, 
                    0.5}, {-3.0895776031746616`, 0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{0.09721932786775561, 0.5}, {2.2916233612860504`,
                     0.5}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.005], Arrowheads[0.04], 
                   ArrowBox[{{0.09721932786775561, 0.5}, {-0.8951735697563667,
                     0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{0.09721932786775561, 0.5}, 1.6, {
                    1.8904949284266772`, 3.3352004090120904`}], 
                   ArrowBox[{{-1.4758869064950872`, 
                    0.20787541115346503`}, {-1.4728870546375505`, 
                    0.19215921710062256`}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Dashing[{Small, Small}], 
                   Arrowheads[0.02], 
                   CircleBox[{0.09721932786775561, 0.5}, 1.7, {
                    1.8904949284266772`, 1.3451481054450334`}], 
                   ArrowBox[{{0.4609864845310009, 2.1606244174204887`}, {
                    0.4775742637295894, 2.1569037759524665`}}]}, 
                  {RGBColor[0, 0, 1], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{0.09721932786775561, 0.5}, 1.8, {
                    1.8904949284266772`, 1.0052735586926724`}], 
                   ArrowBox[{{1.0465168801038698`, 2.0293247390003613`}, {
                    1.0617624081256187`, 2.0197554560937343`}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.003], Arrowheads[0.025], 
                   CircleBox[{0.09721932786775561, 0.5}, 1.5, {
                    1.8904949284266772`, 1.7511887313244492`}], 
                   ArrowBox[{{-0.18664700713645566`, 
                    1.9728950756422798`}, {-0.17190410866284506`, 
                    1.9756600475414248`}}]}, 
                  {RGBColor[0, 0, 1], ArrowBox[{{0, 2.8}, {-1.5, 2.8}}]}},
                 Frame->True,
                 FrameTicks->Automatic,
                 ImageSize->800,
                 PlotRange->{{-3, 3}, {-1, 3}}], 253->
                GraphicsBox[{InsetBox[
                   TagBox[GridBox[{
                    {"\<\"t\"\>", "12.6`"},
                    {"\<\"driveIntent\"\>", 
                    RowBox[{"-", "0.5`"}]},
                    {"\<\"motorTorque\"\>", "159.26793256238597`"},
                    {"\<\"vw\"\>", 
                    RowBox[{"-", "0.051207180381388995`"}]},
                    {"\<\"controllerState\"\>", 
                    RowBox[{"\[LeftAssociation]", 
                    RowBox[{
                    
                    RowBox[{"\<\"lastError\"\>", "\[Rule]", 
                    "0.18701814447182133`"}], ",", 
                    
                    RowBox[{"\<\"lastInt\"\>", "\[Rule]", 
                    "0.08488693007435358`"}], ",", 
                    RowBox[{"\<\"initialized\"\>", "\[Rule]", "True"}]}], 
                    "\[RightAssociation]"}]}
                    },
                    AutoDelete->False,
                    GridBoxItemSize->{
                    "Columns" -> {{Automatic}}, "Rows" -> {{Automatic}}}],
                    "Grid"], {-1, -0.5}], 
                  {GrayLevel[0], LineBox[{{-5, 0}, {5, 0}}]}, 
                  {GrayLevel[0.5], CircleBox[{0.09508642717371839, 0.5}, 0.5],
                    DiskBox[{0.09508642717371839, 0.5}, 0.1]}, 
                  {GrayLevel[0.5], 
                   DiskBox[{-0.3792806790999649, 1.9230164610732836`}, 0.1]}, 
                  {GrayLevel[0.5], Thickness[0.01], 
                   LineBox[{{0.09508642717371839, 0.5}, {-0.3792806790999649, 
                    1.9230164610732836`}}]}, 
                  {RGBColor[0, 1, 0], 
                   DiskBox[{-0.22115831034207045`, 1.4486776407155224`}, 
                    0.05]}, 
                  {RGBColor[0, 0, 1], Arrowheads[0.01], 
                   CircleBox[{0.09508642717371839, 0.5}, 0.3333333333333333, 
                    NCache[{
                    Rational[-1, 2] Pi, 
                    3.207241650076682}, {-1.5707963267948966`, 
                    3.207241650076682}]], 
                   ArrowBox[{{-0.2377309042086052, 
                    0.4814599071561203}, {-0.23752886564227077`, 
                    0.4781327163084922}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{0.09508642717371839, 0.5}, {-3.090272224074001, 
                    0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{0.09508642717371839, 0.5}, {2.301687937894761, 
                    0.5}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.005], Arrowheads[0.04], 
                   ArrowBox[{{0.09508642717371839, 0.5}, {-0.8836707133529588,
                     0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{0.09508642717371839, 0.5}, 1.6, {
                    1.8925647707755433`, 3.3462998278814235`}], 
                   ArrowBox[{{-1.4746805605234488`, 
                    0.19043319891150717`}, {-1.4715064564114553`, 
                    0.17475126887212766`}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Dashing[{Small, Small}], 
                   Arrowheads[0.02], 
                   CircleBox[{0.09508642717371839, 0.5}, 1.7, {
                    1.8925647707755433`, 1.3550894769671946`}], 
                   ArrowBox[{{0.44232699591066826`, 2.164158642504746}, {
                    0.45895094309357504`, 2.160603087451787}}]}, 
                  {RGBColor[0, 0, 1], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{0.09508642717371839, 0.5}, 1.8, {
                    1.8925647707755433`, 1.0077429232067323`}], 
                   ArrowBox[{{1.0406046287134878`, 2.0316642355806964`}, {
                    1.0558737402770861`, 2.0221326285772903`}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.003], Arrowheads[0.025], 
                   CircleBox[{0.09508642717371839, 0.5}, 1.5, {
                    1.8925647707755433`, 1.7722906149564472`}], 
                   ArrowBox[{{-0.2197952627037969, 
                    1.9665774856378644`}, {-0.2051139883224834, 
                    1.9696529217934102`}}]}, 
                  {RGBColor[0, 0, 1], ArrowBox[{{0, 2.8}, {-1.5, 2.8}}]}},
                 Frame->True,
                 FrameTicks->Automatic,
                 ImageSize->800,
                 PlotRange->{{-3, 3}, {-1, 3}}], 254->
                GraphicsBox[{InsetBox[
                   TagBox[GridBox[{
                    {"\<\"t\"\>", "12.65`"},
                    {"\<\"driveIntent\"\>", 
                    RowBox[{"-", "0.5`"}]},
                    {"\<\"motorTorque\"\>", "157.8161237074895`"},
                    {"\<\"vw\"\>", 
                    RowBox[{"-", "0.06837605152132303`"}]},
                    {"\<\"controllerState\"\>", 
                    RowBox[{"\[LeftAssociation]", 
                    RowBox[{
                    
                    RowBox[{"\<\"lastError\"\>", "\[Rule]", 
                    "0.19109090057824238`"}], ",", 
                    
                    RowBox[{"\<\"lastInt\"\>", "\[Rule]", 
                    "0.09435457686509988`"}], ",", 
                    RowBox[{"\<\"initialized\"\>", "\[Rule]", "True"}]}], 
                    "\[RightAssociation]"}]}
                    },
                    AutoDelete->False,
                    
                    GridBoxItemSize->{
                    "Columns" -> {{Automatic}}, "Rows" -> {{Automatic}}}],
                    "Grid"], {-1, -0.5}], 
                  {GrayLevel[0], LineBox[{{-5, 0}, {5, 0}}]}, 
                  {GrayLevel[0.5], CircleBox[{0.09208657875020784, 0.5}, 0.5],
                    DiskBox[{0.09208657875020784, 0.5}, 0.1]}, 
                  {GrayLevel[0.5], 
                   DiskBox[{-0.38065841558652636`, 1.9235561704160329`}, 
                    0.1]}, 
                  {GrayLevel[0.5], Thickness[0.01], 
                   LineBox[{{0.09208657875020784, 
                    0.5}, {-0.38065841558652636`, 1.9235561704160329`}}]}, 
                  {RGBColor[0, 1, 0], 
                   DiskBox[{-0.223076750807615, 1.449037446944022}, 0.05]}, 
                  {RGBColor[0, 0, 1], Arrowheads[0.01], 
                   CircleBox[{0.09208657875020784, 0.5}, 0.3333333333333333, 
                    NCache[{
                    Rational[-1, 2] Pi, 
                    3.163687384429789}, {-1.5707963267948966`, 
                    3.163687384429789}]], 
                   ArrowBox[{{-0.24122237446131115`, 
                    0.4959685213441511}, {-0.24116539503788006`, 
                    0.4926356889355034}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{0.09208657875020784, 
                    0.5}, {-3.0642358953995825`, 0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{0.09208657875020784, 0.5}, {2.2919765911423684`,
                     0.5}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.005], Arrowheads[0.04], 
                   ArrowBox[{{0.09208657875020784, 0.5}, {-0.8643458830074217,
                     0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{0.09208657875020784, 0.5}, 1.6, {
                    1.891425076152315, 3.34018904449992}], 
                   ArrowBox[{{-1.4795427840653703`, 
                    0.20003142508605964`}, {-1.4764645674974461`, 
                    0.18433039169858084`}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Dashing[{Small, Small}], 
                   Arrowheads[0.02], 
                   CircleBox[{0.09208657875020784, 0.5}, 1.7, {
                    1.891425076152315, 1.3660053314019132`}], 
                   ArrowBox[{{0.4211411070737313, 2.1678498485744404`}, {
                    0.4378028749965803, 2.164475966335859}}]}, 
                  {RGBColor[0, 0, 1], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{0.09208657875020784, 0.5}, 1.8, {
                    1.891425076152315, 1.0146688333329288`}], 
                   ArrowBox[{{1.0269740189326753`, 2.0381760218483036`}, {
                    1.0423087788072958`, 2.0287503951000856`}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.003], Arrowheads[0.025], 
                   CircleBox[{0.09208657875020784, 0.5}, 1.5, {
                    1.891425076152315, 1.7891901069560412`}], 
                   ArrowBox[{{-0.24753338301462982`, 
                    1.9610469813017137`}, {-0.23290617585164494`, 
                    1.9643700725760205`}}]}, 
                  {RGBColor[0, 0, 1], ArrowBox[{{0, 2.8}, {-1.5, 2.8}}]}},
                 Frame->True,
                 FrameTicks->Automatic,
                 ImageSize->800,
                 PlotRange->{{-3, 3}, {-1, 3}}], 255->
                GraphicsBox[{InsetBox[
                   TagBox[GridBox[{
                    {"\<\"t\"\>", "12.700000000000001`"},
                    {"\<\"driveIntent\"\>", 
                    RowBox[{"-", "0.5`"}]},
                    {"\<\"motorTorque\"\>", "155.2171429866292`"},
                    {"\<\"vw\"\>", 
                    RowBox[{"-", "0.08506761688860058`"}]},
                    {"\<\"controllerState\"\>", 
                    RowBox[{"\[LeftAssociation]", 
                    RowBox[{
                    
                    RowBox[{"\<\"lastError\"\>", "\[Rule]", 
                    "0.19228273439558888`"}], ",", 
                    
                    RowBox[{"\<\"lastInt\"\>", "\[Rule]", 
                    "0.10395064117341785`"}], ",", 
                    RowBox[{"\<\"initialized\"\>", "\[Rule]", "True"}]}], 
                    "\[RightAssociation]"}]}
                    },
                    AutoDelete->False,
                    
                    GridBoxItemSize->{
                    "Columns" -> {{Automatic}}, "Rows" -> {{Automatic}}}],
                    "Grid"], {-1, -0.5}], 
                  {GrayLevel[0], LineBox[{{-5, 0}, {5, 0}}]}, 
                  {GrayLevel[0.5], CircleBox[{0.08823986979417742, 0.5}, 0.5],
                    DiskBox[{0.08823986979417742, 0.5}, 0.1]}, 
                  {GrayLevel[0.5], 
                   DiskBox[{-0.37899386529130075`, 1.9253745601764027`}, 
                    0.1]}, 
                  {GrayLevel[0.5], Thickness[0.01], 
                   LineBox[{{0.08823986979417742, 
                    0.5}, {-0.37899386529130075`, 1.9253745601764027`}}]}, 
                  {RGBColor[0, 1, 0], 
                   DiskBox[{-0.2232492869294747, 1.4502497067842683`}, 0.05]}, 
                  {RGBColor[0, 0, 1], Arrowheads[0.01], 
                   CircleBox[{0.08823986979417742, 0.5}, 0.3333333333333333, 
                    NCache[{
                    Rational[-1, 2] Pi, 
                    3.0857179628039795`}, {-1.5707963267948966`, 
                    3.0857179628039795`}]], 
                   ArrowBox[{{-0.2443704792292265, 
                    0.5219423525092572}, {-0.24457326871841467`, 
                    0.5186152073453215}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{0.08823986979417742, 0.5}, {-3.016102989938407, 
                    0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{0.08823986979417742, 0.5}, {2.2652608591343966`,
                     0.5}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.005], Arrowheads[0.04], 
                   ArrowBox[{{0.08823986979417742, 0.5}, {-0.8390820005981873,
                     0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{0.08823986979417742, 0.5}, 1.6, {
                    1.8875560790460162`, 3.3194303628510173`}], 
                   ArrowBox[{{-1.489277384606694, 
                    0.23271866494733157`}, {-1.4865257405974104`, 
                    0.21695711927693673`}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Dashing[{Small, Small}], 
                   Arrowheads[0.02], 
                   CircleBox[{0.08823986979417742, 0.5}, 1.7, {
                    1.8875560790460162`, 1.3774624742216477`}], 
                   ArrowBox[{{0.398164425634559, 2.171510325928961}, {
                    0.41486375421153004`, 2.168327557204556}}]}, 
                  {RGBColor[0, 0, 1], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{0.08823986979417742, 0.5}, 1.8, {
                    1.8875560790460162`, 1.0252386180091873`}], 
                   ArrowBox[{{1.0068172008693312`, 2.0479714748130364`}, {
                    1.0222507291396912`, 2.038708456668987}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.003], Arrowheads[0.025], 
                   CircleBox[{0.08823986979417742, 0.5}, 1.5, {
                    1.8875560790460162`, 1.8020016097262053`}], 
                   ArrowBox[{{-0.2700699160596109, 
                    1.9565761557026164`}, {-0.2554864819240661, 
                    1.9600863656422747`}}]}, 
                  {RGBColor[0, 0, 1], ArrowBox[{{0, 2.8}, {-1.5, 2.8}}]}},
                 Frame->True,
                 FrameTicks->Automatic,
                 ImageSize->800,
                 PlotRange->{{-3, 3}, {-1, 3}}], 256->
                GraphicsBox[{InsetBox[
                   TagBox[GridBox[{
                    {"\<\"t\"\>", "12.75`"},
                    {"\<\"driveIntent\"\>", 
                    RowBox[{"-", "0.5`"}]},
                    {"\<\"motorTorque\"\>", "151.67359439494655`"},
                    {"\<\"vw\"\>", 
                    RowBox[{"-", "0.10115904639066889`"}]},
                    {"\<\"controllerState\"\>", 
                    RowBox[{"\[LeftAssociation]", 
                    RowBox[{
                    
                    RowBox[{"\<\"lastError\"\>", "\[Rule]", 
                    "0.1910017345241255`"}], ",", 
                    
                    RowBox[{"\<\"lastInt\"\>", "\[Rule]", 
                    "0.11354160103621126`"}], ",", 
                    RowBox[{"\<\"initialized\"\>", "\[Rule]", "True"}]}], 
                    "\[RightAssociation]"}]}
                    },
                    AutoDelete->False,
                    
                    GridBoxItemSize->{
                    "Columns" -> {{Automatic}}, "Rows" -> {{Automatic}}}],
                    "Grid"], {-1, -0.5}], 
                  {GrayLevel[0], LineBox[{{-5, 0}, {5, 0}}]}, 
                  {GrayLevel[0.5], CircleBox[{0.08357344867262576, 0.5}, 0.5],
                    DiskBox[{0.08357344867262576, 0.5}, 0.1]}, 
                  {GrayLevel[0.5], 
                   DiskBox[{-0.374879841848606, 1.9282228749079238`}, 0.1]}, 
                  {GrayLevel[0.5], Thickness[0.01], 
                   LineBox[{{0.08357344867262576, 0.5}, {-0.374879841848606, 
                    1.9282228749079238`}}]}, 
                  {RGBColor[0, 1, 0], 
                   DiskBox[{-0.22206207834152877`, 1.4521485832719492`}, 
                    0.05]}, 
                  {RGBColor[0, 0, 1], Arrowheads[0.01], 
                   CircleBox[{0.08357344867262576, 0.5}, 0.3333333333333333, 
                    NCache[{
                    Rational[-1, 2] Pi, 
                    2.9794115050535}, {-1.5707963267948966`, 
                    2.9794115050535}]], 
                   ArrowBox[{{-0.2448310216457598, 
                    0.5571105505665259}, {-0.24538569754636638`, 
                    0.5538237050934145}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{0.08357344867262576, 
                    0.5}, {-2.9498984392263052`, 0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{0.08357344867262576, 0.5}, {2.2239515534365757`,
                     0.5}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.005], Arrowheads[0.04], 
                   ArrowBox[{{0.08357344867262576, 0.5}, {-0.8095203344623552,
                     0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{0.08357344867262576, 0.5}, 1.6, {
                    1.881402150852749, 3.286368076318386}], 
                   ArrowBox[{{-1.5019170027530828`, 
                    0.28501156208320744`}, {-1.49968788034319, 
                    0.26916767114835566`}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Dashing[{Small, Small}], 
                   Arrowheads[0.02], 
                   CircleBox[{0.08357344867262576, 0.5}, 1.7, {
                    1.881402150852749, 1.3891319271287121`}], 
                   ArrowBox[{{0.3739717342536526, 2.1750130852419036`}, {
                    0.3907070661453398, 2.1720254008292232`}}]}, 
                  {RGBColor[0, 0, 1], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{0.08357344867262576, 0.5}, 1.8, {
                    1.881402150852749, 1.0387710708808235`}], 
                   ArrowBox[{{0.981119461065362, 2.0602599641206902`}, {
                    0.9966769237378984, 2.0512066412389176`}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.003], Arrowheads[0.025], 
                   CircleBox[{0.08357344867262576, 0.5}, 1.5, {
                    1.881402150852749, 1.8109689218292093`}], 
                   ArrowBox[{{-0.2877833288894741, 
                    1.9533045598767977`}, {-0.273231957822775, 
                    1.9569454011374798`}}]}, 
                  {RGBColor[0, 0, 1], ArrowBox[{{0, 2.8}, {-1.5, 2.8}}]}},
                 Frame->True,
                 FrameTicks->Automatic,
                 ImageSize->800,
                 PlotRange->{{-3, 3}, {-1, 3}}], 257->
                GraphicsBox[{InsetBox[
                   TagBox[GridBox[{
                    {"\<\"t\"\>", "12.8`"},
                    {"\<\"driveIntent\"\>", 
                    RowBox[{"-", "0.5`"}]},
                    {"\<\"motorTorque\"\>", "147.36019246939725`"},
                    {"\<\"vw\"\>", 
                    RowBox[{"-", "0.11656135161782064`"}]},
                    {"\<\"controllerState\"\>", 
                    RowBox[{"\[LeftAssociation]", 
                    RowBox[{
                    
                    RowBox[{"\<\"lastError\"\>", "\[Rule]", 
                    "0.1876249593832074`"}], ",", 
                    
                    RowBox[{"\<\"lastInt\"\>", "\[Rule]", 
                    "0.12301356848901365`"}], ",", 
                    RowBox[{"\<\"initialized\"\>", "\[Rule]", "True"}]}], 
                    "\[RightAssociation]"}]}
                    },
                    AutoDelete->False,
                    
                    GridBoxItemSize->{
                    "Columns" -> {{Automatic}}, "Rows" -> {{Automatic}}}],
                    "Grid"], {-1, -0.5}], 
                  {GrayLevel[0], LineBox[{{-5, 0}, {5, 0}}]}, 
                  {GrayLevel[0.5], CircleBox[{0.07811972192830989, 0.5}, 0.5],
                    DiskBox[{0.07811972192830989, 0.5}, 0.1]}, 
                  {GrayLevel[0.5], 
                   DiskBox[{-0.3688420194033974, 1.9318607480428145`}, 0.1]}, 
                  {GrayLevel[0.5], Thickness[0.01], 
                   LineBox[{{0.07811972192830989, 0.5}, {-0.3688420194033974, 
                    1.9318607480428145`}}]}, 
                  {RGBColor[0, 1, 0], 
                   DiskBox[{-0.2198547722928283, 1.4545738320285428`}, 0.05]}, 
                  {RGBColor[0, 0, 1], Arrowheads[0.01], 
                   CircleBox[{0.07811972192830989, 0.5}, 0.3333333333333333, 
                    NCache[{
                    Rational[-1, 2] Pi, 
                    2.8500094472870208`}, {-1.5707963267948966`, 
                    2.8500094472870208`}]], 
                   ArrowBox[{{-0.24016941892899601`, 
                    0.5990107768044932}, {-0.24114359587090445`, 
                    0.5958229879462592}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{0.07811972192830989, 0.5}, {-2.869084127459635, 
                    0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{0.07811972192830989, 0.5}, {2.1701624724655777`,
                     0.5}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.005], Arrowheads[0.04], 
                   ArrowBox[{{0.07811972192830989, 0.5}, {-0.7770413769223674,
                     0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{0.07811972192830989, 0.5}, 1.6, {
                    1.873366380667791, 3.2431155558648497`}], 
                   ArrowBox[{{-1.5151838011047742`, 
                    0.3537677070877886}, {-1.5136418380353016`, 
                    0.33784224896043324`}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Dashing[{Small, Small}], 
                   Arrowheads[0.02], 
                   CircleBox[{0.07811972192830989, 0.5}, 1.7, {
                    1.873366380667791, 1.4007760265884046`}], 
                   ArrowBox[{{0.3489947426768211, 2.17828088326552}, {
                    0.36576372815922115`, 2.175488264858764}}]}, 
                  {RGBColor[0, 0, 1], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{0.07811972192830989, 0.5}, 1.8, {
                    1.873366380667791, 1.054698644726695}], 
                   ArrowBox[{{0.9507017835048585, 2.074357184953535}, {
                    0.9664014642233375, 2.0655527925641186`}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.003], Arrowheads[0.025], 
                   CircleBox[{0.07811972192830989, 0.5}, 1.5, {
                    1.873366380667791, 1.8164284831253672`}], 
                   ArrowBox[{{-0.30116588707730624`, 
                    1.9512554657265684`}, {-0.2866346101723269, 
                    1.9549756964340739`}}]}, 
                  {RGBColor[0, 0, 1], ArrowBox[{{0, 2.8}, {-1.5, 2.8}}]}},
                 Frame->True,
                 FrameTicks->Automatic,
                 ImageSize->800,
                 PlotRange->{{-3, 3}, {-1, 3}}], 258->
                GraphicsBox[{InsetBox[
                   TagBox[GridBox[{
                    {"\<\"t\"\>", "12.85`"},
                    {"\<\"driveIntent\"\>", 
                    RowBox[{"-", "0.5`"}]},
                    {"\<\"motorTorque\"\>", "142.42700110744653`"},
                    {"\<\"vw\"\>", 
                    RowBox[{"-", "0.13121424523141786`"}]},
                    {"\<\"controllerState\"\>", 
                    RowBox[{"\[LeftAssociation]", 
                    RowBox[{
                    
                    RowBox[{"\<\"lastError\"\>", "\[Rule]", 
                    "0.1824949777793884`"}], ",", 
                    
                    RowBox[{"\<\"lastInt\"\>", "\[Rule]", 
                    "0.13227063738760492`"}], ",", 
                    RowBox[{"\<\"initialized\"\>", "\[Rule]", "True"}]}], 
                    "\[RightAssociation]"}]}
                    },
                    AutoDelete->False,
                    
                    GridBoxItemSize->{
                    "Columns" -> {{Automatic}}, "Rows" -> {{Automatic}}}],
                    "Grid"], {-1, -0.5}], 
                  {GrayLevel[0], LineBox[{{-5, 0}, {5, 0}}]}, 
                  {GrayLevel[0.5], CircleBox[{0.0719147916157615, 0.5}, 0.5], 
                   DiskBox[{0.0719147916157615, 0.5}, 0.1]}, 
                  {GrayLevel[0.5], 
                   DiskBox[{-0.36134150097970086`, 1.9360671937365728`}, 
                    0.1]}, 
                  {GrayLevel[0.5], Thickness[0.01], 
                   LineBox[{{0.0719147916157615, 0.5}, {-0.36134150097970086`,
                     1.9360671937365728`}}]}, 
                  {RGBColor[0, 1, 0], 
                   DiskBox[{-0.2169227367812134, 1.457378129157715}, 0.05]}, 
                  {RGBColor[0, 0, 1], Arrowheads[0.01], 
                   CircleBox[{0.0719147916157615, 0.5}, 0.3333333333333333, 
                    NCache[{
                    Rational[-1, 2] Pi, 
                    2.702013706428499}, {-1.5707963267948966`, 
                    2.702013706428499}]], 
                   ArrowBox[{{-0.2282952637182515, 
                    0.6448621199191146}, {-0.2297288503961971, 
                    0.6418528263548966}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{0.0719147916157615, 0.5}, {-2.776625230533169, 
                    0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{0.0719147916157615, 0.5}, {2.1057654530636123`, 
                    0.5}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.005], Arrowheads[0.04], 
                   ArrowBox[{{0.0719147916157615, 0.5}, {-0.7427745690853184, 
                    0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{0.0719147916157615, 0.5}, 1.6, {
                    1.8638087176561566`, 3.1915564744628098`}], 
                   ArrowBox[{{-1.5268076928422163`, 
                    0.4360749056433122}, {-1.5260885170946878`, 
                    0.4200911434792304}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Dashing[{Small, Small}], 
                   Arrowheads[0.02], 
                   CircleBox[{0.0719147916157615, 0.5}, 1.7, {
                    1.8638087176561566`, 1.4122311039655409`}], 
                   ArrowBox[{{0.32354762367353695`, 2.181273599932677}, {
                    0.3403474979252419, 2.1786732505712267`}}]}, 
                  {RGBColor[0, 0, 1], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{0.0719147916157615, 0.5}, 1.8, {
                    1.8638087176561566`, 1.072547600392565}], 
                   ArrowBox[{{0.9162587214631892, 2.0896802597157085`}, {
                    0.9321130422702775, 2.0811574777898976`}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.003], Arrowheads[0.025], 
                   CircleBox[{0.0719147916157615, 0.5}, 1.5, {
                    1.8638087176561566`, 1.8187716886842453`}], 
                   ArrowBox[{{-0.31077036289797455`, 
                    1.9503627382537094`}, {-0.2962478431430777, 
                    1.9541170084857098`}}]}, 
                  {RGBColor[0, 0, 1], ArrowBox[{{0, 2.8}, {-1.5, 2.8}}]}},
                 Frame->True,
                 FrameTicks->Automatic,
                 ImageSize->800,
                 PlotRange->{{-3, 3}, {-1, 3}}], 259->
                GraphicsBox[{InsetBox[
                   TagBox[GridBox[{
                    {"\<\"t\"\>", "12.9`"},
                    {"\<\"driveIntent\"\>", 
                    RowBox[{"-", "0.5`"}]},
                    {"\<\"motorTorque\"\>", "137.0029711426102`"},
                    {"\<\"vw\"\>", 
                    RowBox[{"-", "0.14508092638251105`"}]},
                    {"\<\"controllerState\"\>", 
                    RowBox[{"\[LeftAssociation]", 
                    RowBox[{
                    
                    RowBox[{"\<\"lastError\"\>", "\[Rule]", 
                    "0.17591902305020063`"}], ",", 
                    
                    RowBox[{"\<\"lastInt\"\>", "\[Rule]", 
                    "0.1412331280608297`"}], ",", 
                    RowBox[{"\<\"initialized\"\>", "\[Rule]", "True"}]}], 
                    "\[RightAssociation]"}]}
                    },
                    AutoDelete->False,
                    
                    GridBoxItemSize->{
                    "Columns" -> {{Automatic}}, "Rows" -> {{Automatic}}}],
                    "Grid"], {-1, -0.5}], 
                  {GrayLevel[0], LineBox[{{-5, 0}, {5, 0}}]}, 
                  {GrayLevel[0.5], CircleBox[{0.06499715410331394, 0.5}, 0.5],
                    DiskBox[{0.06499715410331394, 0.5}, 0.1]}, 
                  {GrayLevel[0.5], 
                   DiskBox[{-0.3527794155672692, 1.9406466387821408`}, 0.1]}, 
                  {GrayLevel[0.5], Thickness[0.01], 
                   LineBox[{{0.06499715410331394, 0.5}, {-0.3527794155672692, 
                    1.9406466387821408`}}]}, 
                  {RGBColor[0, 1, 0], 
                   DiskBox[{-0.2135205590104082, 1.4604310925214272`}, 0.05]}, 
                  {RGBColor[0, 0, 1], Arrowheads[0.01], 
                   CircleBox[{0.06499715410331394, 0.5}, 0.3333333333333333, 
                    NCache[{
                    Rational[-1, 2] Pi, 
                    2.5392928074834096`}, {-1.5707963267948966`, 
                    2.5392928074834096`}]], 
                   ArrowBox[{{-0.20777896067381843`, 
                    0.691583669236717}, {-0.20968112674365116`, 
                    0.6888463744477686}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{0.06499715410331394, 0.5}, {-2.67506226874889, 
                    0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{0.06499715410331394, 0.5}, {2.032434771216387, 
                    0.5}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.005], Arrowheads[0.04], 
                   ArrowBox[{{0.06499715410331394, 0.5}, {-0.7076246516358168,
                     0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{0.06499715410331394, 0.5}, 1.6, {
                    1.853046734073287, 3.133355670623295}], 
                   ArrowBox[{{-1.5347367832327266`, 
                    0.5291775553350422}, {-1.5349485678928623`, 
                    0.5131790237170625}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Dashing[{Small, Small}], 
                   Arrowheads[0.02], 
                   CircleBox[{0.06499715410331394, 0.5}, 1.7, {
                    1.853046734073287, 1.4233897045269217`}], 
                   ArrowBox[{{0.29785404916246555`, 2.183976741651562}, {
                    0.3146818931698646, 2.181564013374831}}]}, 
                  {RGBColor[0, 0, 1], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{0.06499715410331394, 0.5}, 1.8, {
                    1.853046734073287, 1.0919191166143416`}], 
                   ArrowBox[{{0.8783900754882659, 2.1057371999928423`}, {
                    0.8944065105595091, 2.0975231201528572`}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.003], Arrowheads[0.025], 
                   CircleBox[{0.06499715410331394, 0.5}, 1.5, {
                    1.853046734073287, 1.8184118252900134`}], 
                   ArrowBox[{{-0.31716604318465025`, 
                    1.9505003587171705`}, {-0.3026421733456993, 
                    1.9542494025829398`}}]}, 
                  {RGBColor[0, 0, 1], ArrowBox[{{0, 2.8}, {-1.5, 2.8}}]}},
                 Frame->True,
                 FrameTicks->Automatic,
                 ImageSize->800,
                 PlotRange->{{-3, 3}, {-1, 3}}], 260->
                GraphicsBox[{InsetBox[
                   TagBox[GridBox[{
                    {"\<\"t\"\>", "12.950000000000001`"},
                    {"\<\"driveIntent\"\>", 
                    RowBox[{"-", "0.5`"}]},
                    {"\<\"motorTorque\"\>", "131.19931904200976`"},
                    {"\<\"vw\"\>", 
                    RowBox[{"-", "0.1581432647600916`"}]},
                    {"\<\"controllerState\"\>", 
                    RowBox[{"\[LeftAssociation]", 
                    RowBox[{
                    
                    RowBox[{"\<\"lastError\"\>", "\[Rule]", 
                    "0.16817000803920362`"}], ",", 
                    
                    RowBox[{"\<\"lastInt\"\>", "\[Rule]", 
                    "0.14983584017138613`"}], ",", 
                    RowBox[{"\<\"initialized\"\>", "\[Rule]", "True"}]}], 
                    "\[RightAssociation]"}]}
                    },
                    AutoDelete->False,
                    
                    GridBoxItemSize->{
                    "Columns" -> {{Automatic}}, "Rows" -> {{Automatic}}}],
                    "Grid"], {-1, -0.5}], 
                  {GrayLevel[0], LineBox[{{-5, 0}, {5, 0}}]}, 
                  {GrayLevel[0.5], 
                   CircleBox[{0.057406650859364756`, 0.5}, 0.5], 
                   DiskBox[{0.057406650859364756`, 0.5}, 0.1]}, 
                  {GrayLevel[0.5], 
                   DiskBox[{-0.3435025935735837, 1.9454313465980329`}, 0.1]}, 
                  {GrayLevel[0.5], Thickness[0.01], 
                   LineBox[{{0.057406650859364756`, 
                    0.5}, {-0.3435025935735837, 1.9454313465980329`}}]}, 
                  {RGBColor[0, 1, 0], 
                   DiskBox[{-0.20986617876260089`, 1.4636208977320218`}, 
                    0.05]}, 
                  {RGBColor[0, 0, 1], Arrowheads[0.01], 
                   CircleBox[{0.057406650859364756`, 0.5}, 0.3333333333333333,
                     NCache[{
                    Rational[-1, 2] Pi, 
                    2.3651832444653964`}, {-1.5707963267948966`, 
                    2.3651832444653964`}]], 
                   ArrowBox[{{-0.17805713615645066`, 
                    0.7359404927418814}, {-0.18040472866940935`, 
                    0.7335740971891628}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{0.057406650859364756`, 
                    0.5}, {-2.5665797299808304`, 0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{0.057406650859364756`, 0.5}, {
                    1.9516813574531804`, 0.5}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.005], Arrowheads[0.04], 
                   ArrowBox[{{0.057406650859364756`, 
                    0.5}, {-0.6723050233870149, 0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{0.057406650859364756`, 0.5}, 1.6, {
                    1.8413581149280136`, 3.0699758155275054`}], 
                   ArrowBox[{{-1.53726726008148, 
                    0.6304420092019106}, {-1.5384919254021712`, 
                    0.6144890138240493}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Dashing[{Small, Small}], 
                   Arrowheads[0.02], 
                   CircleBox[{0.057406650859364756`, 0.5}, 1.7, {
                    1.8413581149280136`, 1.4341847435772137`}], 
                   ArrowBox[{{0.2720717366886813, 2.186392273738792}, {
                    0.2889246451972476, 2.184161339746808}}]}, 
                  {RGBColor[0, 0, 1], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{0.057406650859364756`, 0.5}, 1.8, {
                    1.8413581149280136`, 1.1124730091390704`}], 
                   ArrowBox[{{0.8376259382455326, 2.122115243621926}, {
                    0.8538078096912839, 2.11423207569766}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.003], Arrowheads[0.025], 
                   CircleBox[{0.057406650859364756`, 0.5}, 1.5, {
                    1.8413581149280136`, 1.8157583810898332`}], 
                   ArrowBox[{{-0.3209063838207856, 
                    1.9515092999326926`}, {-0.3063726172443616, 
                    1.9552197923674337`}}]}, 
                  {RGBColor[0, 0, 1], ArrowBox[{{0, 2.8}, {-1.5, 2.8}}]}},
                 Frame->True,
                 FrameTicks->Automatic,
                 ImageSize->800,
                 PlotRange->{{-3, 3}, {-1, 3}}], 261->
                GraphicsBox[{InsetBox[
                   TagBox[GridBox[{
                    {"\<\"t\"\>", "13.`"},
                    {"\<\"driveIntent\"\>", 
                    RowBox[{"-", "0.5`"}]},
                    {"\<\"motorTorque\"\>", "125.11250476517509`"},
                    {"\<\"vw\"\>", 
                    RowBox[{"-", "0.17039761854537513`"}]},
                    {"\<\"controllerState\"\>", 
                    RowBox[{"\[LeftAssociation]", 
                    RowBox[{
                    
                    RowBox[{"\<\"lastError\"\>", "\[Rule]", 
                    "0.1594887204938707`"}], ",", 
                    
                    RowBox[{"\<\"lastInt\"\>", "\[Rule]", 
                    "0.158026389013007`"}], ",", 
                    RowBox[{"\<\"initialized\"\>", "\[Rule]", "True"}]}], 
                    "\[RightAssociation]"}]}
                    },
                    AutoDelete->False,
                    
                    GridBoxItemSize->{
                    "Columns" -> {{Automatic}}, "Rows" -> {{Automatic}}}],
                    "Grid"], {-1, -0.5}], 
                  {GrayLevel[0], LineBox[{{-5, 0}, {5, 0}}]}, 
                  {GrayLevel[0.5], CircleBox[{0.04918364445339719, 0.5}, 0.5],
                    DiskBox[{0.04918364445339719, 0.5}, 0.1]}, 
                  {GrayLevel[0.5], 
                   DiskBox[{-0.33380961188213554`, 1.9502814091070413`}, 
                    0.1]}, 
                  {GrayLevel[0.5], Thickness[0.01], 
                   LineBox[{{0.04918364445339719, 
                    0.5}, {-0.33380961188213554`, 1.9502814091070413`}}]}, 
                  {RGBColor[0, 1, 0], 
                   DiskBox[{-0.2061451931036246, 1.4668542727380274`}, 0.05]}, 
                  {RGBColor[0, 0, 1], Arrowheads[0.01], 
                   CircleBox[{0.04918364445339719, 0.5}, 0.3333333333333333, 
                    NCache[{
                    Rational[-1, 2] Pi, 
                    2.1825788161603557`}, {-1.5707963267948966`, 
                    2.1825788161603557`}]], 
                   ArrowBox[{{-0.13952059806740846`, 
                    0.7747759450275079}, {-0.14225887658842207`, 
                    0.772875195370088}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{0.04918364445339719, 
                    0.5}, {-2.4530664508501046`, 0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{0.04918364445339719, 0.5}, {1.86487835247642, 
                    0.5}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.005], Arrowheads[0.04], 
                   ArrowBox[{{0.04918364445339719, 0.5}, {-0.6373717428270819,
                     0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{0.04918364445339719, 0.5}, 1.6, {
                    1.828984125512057, 3.002696880900448}], 
                   ArrowBox[{{-1.5331131375570612`, 
                    0.7373539416975164}, {-1.5354075232354314`, 
                    0.7215193699940362}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Dashing[{Small, Small}], 
                   Arrowheads[0.02], 
                   CircleBox[{0.04918364445339719, 0.5}, 1.7, {
                    1.828984125512057, 1.4445766572638425`}], 
                   ArrowBox[{{0.24631261181622882`, 2.1885319571232476`}, {
                    0.2631877936006442, 2.186476274409979}}]}, 
                  {RGBColor[0, 0, 1], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{0.04918364445339719, 0.5}, 1.8, {
                    1.828984125512057, 1.1339146545944176`}], 
                   ArrowBox[{{0.7944454331241179, 2.138470282412018}, {
                    0.8107926000923142, 2.1309358659035613`}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.003], Arrowheads[0.025], 
                   CircleBox[{0.04918364445339719, 0.5}, 1.5, {
                    1.828984125512057, 1.811199235477197}], 
                   ArrowBox[{{-0.32250783913137165`, 
                    1.953218992798592}, {-0.30795730698603185`, 
                    1.9568631853420346`}}]}, 
                  {RGBColor[0, 0, 1], ArrowBox[{{0, 2.8}, {-1.5, 2.8}}]}},
                 Frame->True,
                 FrameTicks->Automatic,
                 ImageSize->800,
                 PlotRange->{{-3, 3}, {-1, 3}}], 262->
                GraphicsBox[{InsetBox[
                   TagBox[GridBox[{
                    {"\<\"t\"\>", "13.05`"},
                    {"\<\"driveIntent\"\>", 
                    RowBox[{"-", "0.5`"}]},
                    {"\<\"motorTorque\"\>", "118.82671979239836`"},
                    {"\<\"vw\"\>", 
                    RowBox[{"-", "0.18185135704929808`"}]},
                    {"\<\"controllerState\"\>", 
                    RowBox[{"\[LeftAssociation]", 
                    RowBox[{
                    
                    RowBox[{"\<\"lastError\"\>", "\[Rule]", 
                    "0.15008664909179084`"}], ",", 
                    RowBox[{"\<\"lastInt\"\>", "\[Rule]", 
                    "0.16576366964478412`"}], ",", 
                    RowBox[{"\<\"initialized\"\>", "\[Rule]", "True"}]}], 
                    "\[RightAssociation]"}]}
                    },
                    AutoDelete->False,
                    
                    GridBoxItemSize->{
                    "Columns" -> {{Automatic}}, "Rows" -> {{Automatic}}}],
                    "Grid"], {-1, -0.5}], 
                  {GrayLevel[0], LineBox[{{-5, 0}, {5, 0}}]}, 
                  {GrayLevel[0.5], CircleBox[{0.04036838573479649, 0.5}, 0.5],
                    DiskBox[{0.04036838573479649, 0.5}, 0.1]}, 
                  {GrayLevel[0.5], 
                   DiskBox[{-0.3239567206993258, 1.9550832336405246`}, 0.1]}, 
                  {GrayLevel[0.5], Thickness[0.01], 
                   LineBox[{{0.04036838573479649, 0.5}, {-0.3239567206993258, 
                    1.9550832336405246`}}]}, 
                  {RGBColor[0, 1, 0], 
                   DiskBox[{-0.20251501855461837`, 1.4700554890936832`}, 
                    0.05]}, 
                  {RGBColor[0, 0, 1], Arrowheads[0.01], 
                   CircleBox[{0.04036838573479649, 0.5}, 0.3333333333333333, 
                    NCache[{
                    Rational[-1, 2] Pi, 
                    1.9940052669770543`}, {-1.5707963267948966`, 
                    1.9940052669770543`}]], 
                   ArrowBox[{{-0.09348169976396326, 
                    0.8052789965311826}, {-0.09652774640119247, 
                    0.8039252541618036}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{0.04036838573479649, 
                    0.5}, {-2.3361660101131707`, 0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{0.04036838573479649, 0.5}, {1.7732797821269528`,
                     0.5}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.005], Arrowheads[0.04], 
                   ArrowBox[{{0.04036838573479649, 0.5}, {-0.6032546137210145,
                     0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{0.04036838573479649, 0.5}, 1.6, {
                    1.81613348864332, 2.9326362402096278`}], 
                   ArrowBox[{{-1.5214310692927508`, 
                    0.8475377134580595}, {-1.5248282991826652`, 
                    0.831902602465526}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Dashing[{Small, Small}], 
                   Arrowheads[0.02], 
                   CircleBox[{0.04036838573479649, 0.5}, 1.7, {
                    1.81613348864332, 1.454543698848507}], 
                   ArrowBox[{{0.22065817200735718`, 2.1904128469003643`}, {
                    0.23755300432810192`, 2.1885254591477774`}}]}, 
                  {RGBColor[0, 0, 1], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{0.04036838573479649, 0.5}, 1.8, {
                    1.81613348864332, 1.155985045352218}], 
                   ArrowBox[{{0.74928992875332, 2.1545181310110313`}, {
                    0.7657993885300193, 2.147346308516675}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.003], Arrowheads[0.025], 
                   CircleBox[{0.04036838573479649, 0.5}, 1.5, {
                    1.81613348864332, 1.8050898632639378`}], 
                   ArrowBox[{{-0.3224379607801996, 
                    1.955462660094185}, {-0.3078654365889979, 
                    1.9590178905653488`}}]}, 
                  {RGBColor[0, 0, 1], ArrowBox[{{0, 2.8}, {-1.5, 2.8}}]}},
                 Frame->True,
                 FrameTicks->Automatic,
                 ImageSize->800,
                 PlotRange->{{-3, 3}, {-1, 3}}], 263->
                GraphicsBox[{InsetBox[
                   TagBox[GridBox[{
                    {"\<\"t\"\>", "13.1`"},
                    {"\<\"driveIntent\"\>", 
                    RowBox[{"-", "0.5`"}]},
                    {"\<\"motorTorque\"\>", "112.41588958800102`"},
                    {"\<\"vw\"\>", 
                    RowBox[{"-", "0.1925200601466359`"}]},
                    {"\<\"controllerState\"\>", 
                    RowBox[{"\[LeftAssociation]", 
                    RowBox[{
                    
                    RowBox[{"\<\"lastError\"\>", "\[Rule]", 
                    "0.14014903515214497`"}], ",", 
                    
                    RowBox[{"\<\"lastInt\"\>", "\[Rule]", 
                    "0.17301646947669996`"}], ",", 
                    RowBox[{"\<\"initialized\"\>", "\[Rule]", "True"}]}], 
                    "\[RightAssociation]"}]}
                    },
                    AutoDelete->False,
                    
                    GridBoxItemSize->{
                    "Columns" -> {{Automatic}}, "Rows" -> {{Automatic}}}],
                    "Grid"], {-1, -0.5}], 
                  {GrayLevel[0], LineBox[{{-5, 0}, {5, 0}}]}, 
                  {GrayLevel[0.5], 
                   CircleBox[{0.031000537371384907`, 0.5}, 0.5], 
                   DiskBox[{0.031000537371384907`, 0.5}, 0.1]}, 
                  {GrayLevel[0.5], 
                   DiskBox[{-0.314163348021724, 1.9597472014771369`}, 0.1]}, 
                  {GrayLevel[0.5], Thickness[0.01], 
                   LineBox[{{0.031000537371384907`, 0.5}, {-0.314163348021724,
                     1.9597472014771369`}}]}, 
                  {RGBColor[0, 1, 0], 
                   DiskBox[{-0.19910871955735437`, 1.4731648009847578`}, 
                    0.05]}, 
                  {RGBColor[0, 0, 1], Arrowheads[0.01], 
                   CircleBox[{0.031000537371384907`, 0.5}, 0.3333333333333333,
                     NCache[{
                    Rational[-1, 2] Pi, 
                    1.801680360845134}, {-1.5707963267948966`, 
                    1.801680360845134}]], 
                   ArrowBox[{{-0.04203021738470086, 
                    0.8252347151994504}, {-0.0452788588238721, 
                    0.8244881582233751}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{0.031000537371384907`, 
                    0.5}, {-2.2173172543886355`, 0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{0.031000537371384907`, 0.5}, {
                    1.6780339941346207`, 0.5}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.005], Arrowheads[0.04], 
                   ArrowBox[{{0.031000537371384907`, 
                    0.5}, {-0.5702837976253998, 0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{0.031000537371384907`, 0.5}, 1.6, {
                    1.802986277876407, 2.8607679705921356`}], 
                   ArrowBox[{{-1.501811830227734, 
                    0.9587878003338626}, {-1.506322991787111, 
                    0.9434369929264667}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Dashing[{Small, Small}], 
                   Arrowheads[0.02], 
                   CircleBox[{0.031000537371384907`, 0.5}, 1.7, {
                    1.802986277876407, 1.4640750345783324`}], 
                   ArrowBox[{{0.19517048595466305`, 2.192054440017272}, {
                    0.21208253991814724`, 2.1903281658759797`}}]}, 
                  {RGBColor[0, 0, 1], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{0.031000537371384907`, 0.5}, 1.8, {
                    1.802986277876407, 1.178453557943068}], 
                   ArrowBox[{{0.7025717103415874, 2.17002759247727}, {
                    0.7192381296509915, 2.163228491991756}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.003], Arrowheads[0.025], 
                   CircleBox[{0.031000537371384907`, 0.5}, 1.5, {
                    1.802986277876407, 1.7977480771153596`}], 
                   ArrowBox[{{-0.32111043162414343`, 
                    1.9580870569047069`}, {-0.3065121986666534, 
                    1.9615352041644831`}}]}, 
                  {RGBColor[0, 0, 1], ArrowBox[{{0, 2.8}, {-1.5, 2.8}}]}},
                 Frame->True,
                 FrameTicks->Automatic,
                 ImageSize->800,
                 PlotRange->{{-3, 3}, {-1, 3}}], 264->
                GraphicsBox[{InsetBox[
                   TagBox[GridBox[{
                    {"\<\"t\"\>", "13.15`"},
                    {"\<\"driveIntent\"\>", 
                    RowBox[{"-", "0.5`"}]},
                    {"\<\"motorTorque\"\>", "105.94524267364565`"},
                    {"\<\"vw\"\>", 
                    RowBox[{"-", "0.20242531689066434`"}]},
                    {"\<\"controllerState\"\>", 
                    RowBox[{"\[LeftAssociation]", 
                    RowBox[{
                    
                    RowBox[{"\<\"lastError\"\>", "\[Rule]", 
                    "0.12983787734629312`"}], ",", 
                    
                    RowBox[{"\<\"lastInt\"\>", "\[Rule]", 
                    "0.17976223313321324`"}], ",", 
                    RowBox[{"\<\"initialized\"\>", "\[Rule]", "True"}]}], 
                    "\[RightAssociation]"}]}
                    },
                    AutoDelete->False,
                    
                    GridBoxItemSize->{
                    "Columns" -> {{Automatic}}, "Rows" -> {{Automatic}}}],
                    "Grid"], {-1, -0.5}], 
                  {GrayLevel[0], LineBox[{{-5, 0}, {5, 0}}]}, 
                  {GrayLevel[0.5], 
                   CircleBox[{0.021118821758045518`, 0.5}, 0.5], 
                   DiskBox[{0.021118821758045518`, 0.5}, 0.1]}, 
                  {GrayLevel[0.5], 
                   DiskBox[{-0.3046170188511366, 1.9642049590622992`}, 0.1]}, 
                  {GrayLevel[0.5], Thickness[0.01], 
                   LineBox[{{0.021118821758045518`, 
                    0.5}, {-0.3046170188511366, 1.9642049590622992`}}]}, 
                  {RGBColor[0, 1, 0], 
                   DiskBox[{-0.1960384053147426, 1.4761366393748663`}, 0.05]}, 
                  {RGBColor[0, 0, 1], Arrowheads[0.01], 
                   CircleBox[{0.021118821758045518`, 0.5}, 0.3333333333333333,
                     NCache[{
                    Rational[-1, 2] Pi, 
                    1.607560953414473}, {-1.5707963267948966`, 
                    1.607560953414473}]], 
                   ArrowBox[{{0.01219834466408259, 0.8332139495872392}, {
                    0.008866706723728628, 0.8331080857443977}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{0.021118821758045518`, 
                    0.5}, {-2.0977860317148673`, 0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{0.021118821758045518`, 0.5}, {
                    1.5801932227015776`, 0.5}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.005], Arrowheads[0.04], 
                   ArrowBox[{{0.021118821758045518`, 
                    0.5}, {-0.5387116307713355, 0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{0.021118821758045518`, 0.5}, 1.6, {
                    1.7896975796916579`, 2.7879405113510445`}], 
                   ArrowBox[{{-1.4742476252421874`, 
                    1.0691038474530807`}, {-1.4798638011672685`, 
                    1.0541219772543269`}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Dashing[{Small, Small}], 
                   Arrowheads[0.02], 
                   CircleBox[{0.021118821758045518`, 0.5}, 1.7, {
                    1.7896975796916579`, 1.473166095785948}], 
                   ArrowBox[{{0.1698996276853822, 2.1934769770468128`}, {
                    0.18682667623279414`, 2.1919045206409775`}}]}, 
                  {RGBColor[0, 0, 1], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{0.021118821758045518`, 0.5}, 1.8, {
                    1.7896975796916579`, 1.2011128981714043`}], 
                   ArrowBox[{{0.6546791092261632, 2.1848149340931533`}, {
                    0.6714953000156189, 2.1783951967666244`}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.003], Arrowheads[0.025], 
                   CircleBox[{0.021118821758045518`, 0.5}, 1.5, {
                    1.7896975796916579`, 1.789452740544228}], 
                   ArrowBox[{{-0.31888484830070823`, 
                    1.9609577353046796`}, {-0.30425851439756463`, 
                    1.964284668060237}}]}, 
                  {RGBColor[0, 0, 1], ArrowBox[{{0, 2.8}, {-1.5, 2.8}}]}},
                 Frame->True,
                 FrameTicks->Automatic,
                 ImageSize->800,
                 PlotRange->{{-3, 3}, {-1, 3}}], 265->
                GraphicsBox[{InsetBox[
                   TagBox[GridBox[{
                    {"\<\"t\"\>", "13.200000000000001`"},
                    {"\<\"driveIntent\"\>", 
                    RowBox[{"-", "0.5`"}]},
                    {"\<\"motorTorque\"\>", "99.47251683262527`"},
                    {"\<\"vw\"\>", 
                    RowBox[{"-", "0.21159302769793617`"}]},
                    {"\<\"controllerState\"\>", 
                    RowBox[{"\[LeftAssociation]", 
                    RowBox[{
                    
                    RowBox[{"\<\"lastError\"\>", "\[Rule]", 
                    "0.11929472410478266`"}], ",", 
                    
                    RowBox[{"\<\"lastInt\"\>", "\[Rule]", 
                    "0.1859859726401509`"}], ",", 
                    RowBox[{"\<\"initialized\"\>", "\[Rule]", "True"}]}], 
                    "\[RightAssociation]"}]}
                    },
                    AutoDelete->False,
                    
                    GridBoxItemSize->{
                    "Columns" -> {{Automatic}}, "Rows" -> {{Automatic}}}],
                    "Grid"], {-1, -0.5}], 
                  {GrayLevel[0], LineBox[{{-5, 0}, {5, 0}}]}, 
                  {GrayLevel[0.5], 
                   CircleBox[{0.010760765746045035`, 0.5}, 0.5], 
                   DiskBox[{0.010760765746045035`, 0.5}, 0.1]}, 
                  {GrayLevel[0.5], 
                   DiskBox[{-0.29547762764949576`, 1.968406635237841}, 0.1]}, 
                  {GrayLevel[0.5], Thickness[0.01], 
                   LineBox[{{0.010760765746045035`, 
                    0.5}, {-0.29547762764949576`, 1.968406635237841}}]}, 
                  {RGBColor[0, 1, 0], 
                   DiskBox[{-0.19339816318431546`, 1.4789377568252275`}, 
                    0.05]}, 
                  {RGBColor[0, 0, 1], Arrowheads[0.01], 
                   CircleBox[{0.010760765746045035`, 0.5}, 0.3333333333333333,
                     NCache[{
                    Rational[-1, 2] Pi, 
                    1.4133791781838618`}, {-1.5707963267948966`, 
                    1.4133791781838618`}]], 
                   ArrowBox[{{0.06630615527788603, 0.8286728172710167}, {
                    0.06301670463737208, 0.8292118284048995}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{0.010760765746045035`, 
                    0.5}, {-1.9786895709064605`, 0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{0.010760765746045035`, 0.5}, {
                    1.4807203587631323`, 0.5}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.005], Arrowheads[0.04], 
                   ArrowBox[{{0.010760765746045035`, 
                    0.5}, {-0.5087299778893732, 0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{0.010760765746045035`, 0.5}, 1.6, {
                    1.7764007899082466`, 2.7148922713433654`}], 
                   ArrowBox[{{-1.4390827108354391`, 
                    1.1767229074104963`}, {-1.445777335553227, 
                    1.1621908784206494`}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Dashing[{Small, Small}], 
                   Arrowheads[0.02], 
                   CircleBox[{0.010760765746045035`, 0.5}, 1.7, {
                    1.7764007899082466`, 1.4818156273723289`}], 
                   ArrowBox[{{0.14488840627467575`, 2.194700497446738}, {
                    0.1618284224743324, 2.1932745090771975`}}]}, 
                  {RGBColor[0, 0, 1], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{0.010760765746045035`, 0.5}, 1.8, {
                    1.7764007899082466`, 1.2237756963936617`}], 
                   ArrowBox[{{0.6059790085987699, 2.198739310010607}, {
                    0.6229363579129373, 2.1927022905269298`}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.003], Arrowheads[0.025], 
                   CircleBox[{0.010760765746045035`, 0.5}, 1.5, {
                    1.7764007899082466`, 1.780445073680831}], 
                   ArrowBox[{{-0.28672683491729756`, 
                    1.9702044509018355`}, {-0.3014137601373855, 
                    1.9671561148662589`}}]}, 
                  {RGBColor[0, 0, 1], ArrowBox[{{0, 2.8}, {-1.5, 2.8}}]}},
                 Frame->True,
                 FrameTicks->Automatic,
                 ImageSize->800,
                 PlotRange->{{-3, 3}, {-1, 3}}], 266->
                GraphicsBox[{InsetBox[
                   TagBox[GridBox[{
                    {"\<\"t\"\>", "13.25`"},
                    {"\<\"driveIntent\"\>", 
                    RowBox[{"-", "0.5`"}]},
                    {"\<\"motorTorque\"\>", "93.0488745040735`"},
                    {"\<\"vw\"\>", 
                    RowBox[{"-", "0.22005211469378208`"}]},
                    {"\<\"controllerState\"\>", 
                    RowBox[{"\[LeftAssociation]", 
                    RowBox[{
                    
                    RowBox[{"\<\"lastError\"\>", "\[Rule]", 
                    "0.1086431684023622`"}], ",", 
                    
                    RowBox[{"\<\"lastInt\"\>", "\[Rule]", 
                    "0.19167930986557588`"}], ",", 
                    RowBox[{"\<\"initialized\"\>", "\[Rule]", "True"}]}], 
                    "\[RightAssociation]"}]}
                    },
                    AutoDelete->False,
                    
                    GridBoxItemSize->{
                    "Columns" -> {{Automatic}}, "Rows" -> {{Automatic}}}],
                    "Grid"], {-1, -0.5}], 
                  {GrayLevel[0], LineBox[{{-5, 0}, {5, 0}}]}, 
                  {GrayLevel[0.5], 
                   CircleBox[{-0.000037480527453304476`, 0.5}, 0.5], 
                   DiskBox[{-0.000037480527453304476`, 0.5}, 0.1]}, 
                  {GrayLevel[0.5], 
                   DiskBox[{-0.28688107021605214`, 1.9723181568718626`}, 
                    0.1]}, 
                  {GrayLevel[0.5], Thickness[0.01], 
                   LineBox[{{-0.000037480527453304476`, 
                    0.5}, {-0.28688107021605214`, 1.9723181568718626`}}]}, 
                  {RGBColor[0, 1, 0], 
                   DiskBox[{-0.19126654031985255`, 1.481545437914575}, 0.05]}, 
                  {RGBColor[0, 0, 1], Arrowheads[0.01], 
                   CircleBox[{-0.000037480527453304476`, 0.5}, 
                    0.3333333333333333, 
                    NCache[{
                    Rational[-1, 2] Pi, 
                    1.2206699083273085`}, {-1.5707963267948966`, 
                    1.2206699083273085`}]], 
                   ArrowBox[{{0.11742670117324228`, 0.8119507607435781}, {
                    0.1143013723971982, 0.8131097855752611}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-0.000037480527453304476`, 
                    0.5}, {-1.8610149706089232`, 0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-0.000037480527453304476`, 0.5}, {
                    1.3804937558626198`, 0.5}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.005], Arrowheads[0.04], 
                   ArrowBox[{{-0.000037480527453304476`, 
                    0.5}, {-0.4804837342188501, 0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-0.000037480527453304476`, 0.5}, 1.6, {
                    1.763210488959782, 2.6422650792159375`}], 
                   ArrowBox[{{-1.3969535976530465`, 
                    1.2801444492622864`}, {-1.4046850668984345`, 
                    1.2661365140118155`}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Dashing[{Small, Small}], 
                   Arrowheads[0.02], 
                   CircleBox[{-0.000037480527453304476`, 0.5}, 1.7, {
                    1.763210488959782, 1.4900239452465298`}], 
                   ArrowBox[{{0.12017515732544737`, 2.1957443562342904`}, {
                    0.13712630768333983`, 2.1944574633798464`}}]}, 
                  {RGBColor[0, 0, 1], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-0.000037480527453304476`, 0.5}, 1.8, {
                    1.763210488959782, 1.2462722972704845`}], 
                   ArrowBox[{{0.5568175127351621, 2.211698722462131}, {
                    0.5739063721604485, 2.206044681115289}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.003], Arrowheads[0.025], 
                   CircleBox[{-0.000037480527453304476`, 0.5}, 1.5, {
                    1.763210488959782, 1.770931472330839}], 
                   ArrowBox[{{-0.28352489063754305`, 
                    1.9729680540694268`}, {-0.298240151432403, 
                    1.9700595794270224`}}]}, 
                  {RGBColor[0, 0, 1], ArrowBox[{{0, 2.8}, {-1.5, 2.8}}]}},
                 Frame->True,
                 FrameTicks->Automatic,
                 ImageSize->800,
                 PlotRange->{{-3, 3}, {-1, 3}}], 267->
                GraphicsBox[{InsetBox[
                   TagBox[GridBox[{
                    {"\<\"t\"\>", "13.3`"},
                    {"\<\"driveIntent\"\>", 
                    RowBox[{"-", "0.5`"}]},
                    {"\<\"motorTorque\"\>", "86.71959282913406`"},
                    {"\<\"vw\"\>", 
                    RowBox[{"-", "0.2278335540132192`"}]},
                    {"\<\"controllerState\"\>", 
                    RowBox[{"\[LeftAssociation]", 
                    RowBox[{
                    
                    RowBox[{"\<\"lastError\"\>", "\[Rule]", 
                    "0.09799101495757404`"}], ",", 
                    
                    RowBox[{"\<\"lastInt\"\>", "\[Rule]", 
                    "0.19683963538406019`"}], ",", 
                    RowBox[{"\<\"initialized\"\>", "\[Rule]", "True"}]}], 
                    "\[RightAssociation]"}]}
                    },
                    AutoDelete->False,
                    
                    GridBoxItemSize->{
                    "Columns" -> {{Automatic}}, "Rows" -> {{Automatic}}}],
                    "Grid"], {-1, -0.5}], 
                  {GrayLevel[0], LineBox[{{-5, 0}, {5, 0}}]}, 
                  {GrayLevel[0.5], 
                   CircleBox[{-0.011241268870895566`, 0.5}, 0.5], 
                   DiskBox[{-0.011241268870895566`, 0.5}, 0.1]}, 
                  {GrayLevel[0.5], 
                   DiskBox[{-0.27894228264073756`, 1.9759187535994651`}, 
                    0.1]}, 
                  {GrayLevel[0.5], Thickness[0.01], 
                   LineBox[{{-0.011241268870895566`, 
                    0.5}, {-0.27894228264073756`, 1.9759187535994651`}}]}, 
                  {RGBColor[0, 1, 0], 
                   DiskBox[{-0.1897086113841236, 1.483945835732977}, 0.05]}, 
                  {RGBColor[0, 0, 1], Arrowheads[0.01], 
                   CircleBox[{-0.011241268870895566`, 0.5}, 
                    0.3333333333333333, 
                    NCache[{
                    Rational[-1, 2] Pi, 
                    1.0307914580791255`}, {-1.5707963267948966`, 
                    1.0307914580791255`}]], 
                   ArrowBox[{{0.16298919100378775`, 0.7841739924112822}, {
                    0.1601387869913722, 0.7859020593905491}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-0.011241268870895566`, 
                    0.5}, {-1.7456331254535769`, 0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-0.011241268870895566`, 0.5}, {
                    1.280310705923653, 0.5}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.005], Arrowheads[0.04], 
                   ArrowBox[{{-0.011241268870895566`, 
                    0.5}, {-0.4540811506590283, 0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-0.011241268870895566`, 0.5}, 1.6, {
                    1.7502248928187034`, 2.5706155636580847`}], 
                   ArrowBox[{{-1.3487241201381286`, 
                    1.3781455588716902`}, {-1.357438555784703, 
                    1.3647270463596612`}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Dashing[{Small, Small}], 
                   Arrowheads[0.02], 
                   CircleBox[{-0.011241268870895566`, 0.5}, 1.7, {
                    1.7502248928187034`, 1.4977920171973598`}], 
                   ArrowBox[{{0.0957952103178467, 2.1966270044187315`}, {
                    0.11275584581291943`, 2.195471826822873}}]}, 
                  {RGBColor[0, 0, 1], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-0.011241268870895566`, 0.5}, 1.8, {
                    1.7502248928187034`, 1.268449377101292}], 
                   ArrowBox[{{0.5075194253413797, 2.2236262188016283`}, {
                    0.5247294624412349, 2.218352517726424}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.003], Arrowheads[0.025], 
                   CircleBox[{-0.011241268870895566`, 0.5}, 1.5, {
                    1.7502248928187034`, 1.7610870638156273`}], 
                   ArrowBox[{{-0.28021467732998523`, 
                    1.9756874010243157`}, {-0.29495785683520687`, 
                    1.9729239280132187`}}]}, 
                  {RGBColor[0, 0, 1], ArrowBox[{{0, 2.8}, {-1.5, 2.8}}]}},
                 Frame->True,
                 FrameTicks->Automatic,
                 ImageSize->800,
                 PlotRange->{{-3, 3}, {-1, 3}}], 268->
                GraphicsBox[{InsetBox[
                   TagBox[GridBox[{
                    {"\<\"t\"\>", "13.35`"},
                    {"\<\"driveIntent\"\>", 
                    RowBox[{"-", "0.5`"}]},
                    {"\<\"motorTorque\"\>", "80.5245842369351`"},
                    {"\<\"vw\"\>", 
                    RowBox[{"-", "0.23496965648294277`"}]},
                    {"\<\"controllerState\"\>", 
                    RowBox[{"\[LeftAssociation]", 
                    RowBox[{
                    
                    RowBox[{"\<\"lastError\"\>", "\[Rule]", 
                    "0.08743212548359551`"}], ",", 
                    
                    RowBox[{"\<\"lastInt\"\>", "\[Rule]", 
                    "0.2014693674147979`"}], ",", 
                    RowBox[{"\<\"initialized\"\>", "\[Rule]", "True"}]}], 
                    "\[RightAssociation]"}]}
                    },
                    AutoDelete->False,
                    
                    GridBoxItemSize->{
                    "Columns" -> {{Automatic}}, "Rows" -> {{Automatic}}}],
                    "Grid"], {-1, -0.5}], 
                  {GrayLevel[0], LineBox[{{-5, 0}, {5, 0}}]}, 
                  {GrayLevel[0.5], 
                   CircleBox[{-0.022817536497485136`, 0.5}, 0.5], 
                   DiskBox[{-0.022817536497485136`, 0.5}, 0.1]}, 
                  {GrayLevel[0.5], 
                   DiskBox[{-0.2717577573030286, 1.97919869066508}, 0.1]}, 
                  {GrayLevel[0.5], Thickness[0.01], 
                   LineBox[{{-0.022817536497485136`, 
                    0.5}, {-0.2717577573030286, 1.97919869066508}}]}, 
                  {RGBColor[0, 1, 0], 
                   DiskBox[{-0.18877768370118078`, 1.4861324604433868`}, 
                    0.05]}, 
                  {RGBColor[0, 0, 1], Arrowheads[0.01], 
                   CircleBox[{-0.022817536497485136`, 0.5}, 
                    0.3333333333333333, 
                    NCache[{
                    Rational[-1, 2] Pi, 
                    0.8449412003131567}, {-1.5707963267948966`, 
                    0.8449412003131567}]], 
                   ArrowBox[{{0.20092288666994781`, 0.7470856818028295}, {
                    0.1984408841047271, 0.7493106945634816}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-0.022817536497485136`, 
                    0.5}, {-1.6333092212361873`, 0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-0.022817536497485136`, 0.5}, {
                    1.1808900701972225`, 0.5}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.005], Arrowheads[0.04], 
                   ArrowBox[{{-0.022817536497485136`, 
                    0.5}, {-0.42960161454147955`, 0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-0.022817536497485136`, 0.5}, 1.6, {
                    1.73752790782472, 2.5004246629380584`}], 
                   ArrowBox[{{-1.2954197823146951`, 
                    1.4697852978577237`}, {-1.3050538440811557`, 
                    1.4570109986380486`}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Dashing[{Small, Small}], 
                   Arrowheads[0.02], 
                   CircleBox[{-0.022817536497485136`, 0.5}, 1.7, {
                    1.73752790782472, 1.5051210808372097`}], 
                   ArrowBox[{{0.07178149199570828, 2.1973659074601866`}, {
                    0.0887501382653982, 2.196335065353541}}]}, 
                  {RGBColor[0, 0, 1], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-0.022817536497485136`, 0.5}, 1.8, {
                    1.73752790782472, 1.2901691065084138`}], 
                   ArrowBox[{{0.45838704901476895`, 2.234486133378408}, {
                    0.47570756124020175`, 2.22958744413968}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.003], Arrowheads[0.025], 
                   CircleBox[{-0.022817536497485136`, 0.5}, 1.5, {
                    1.73752790782472, 1.7510594811135354`}], 
                   ArrowBox[{{-0.27698009263825113`, 
                    1.9783103175774672`}, {-0.2917502414082985, 
                    1.9756948194763555`}}]}, 
                  {RGBColor[0, 0, 1], ArrowBox[{{0, 2.8}, {-1.5, 2.8}}]}},
                 Frame->True,
                 FrameTicks->Automatic,
                 ImageSize->800,
                 PlotRange->{{-3, 3}, {-1, 3}}], 269->
                GraphicsBox[{InsetBox[
                   TagBox[GridBox[{
                    {"\<\"t\"\>", "13.4`"},
                    {"\<\"driveIntent\"\>", 
                    RowBox[{"-", "0.5`"}]},
                    {"\<\"motorTorque\"\>", "74.49879358869019`"},
                    {"\<\"vw\"\>", 
                    RowBox[{"-", "0.24149353613088212`"}]},
                    {"\<\"controllerState\"\>", 
                    RowBox[{"\[LeftAssociation]", 
                    RowBox[{
                    
                    RowBox[{"\<\"lastError\"\>", "\[Rule]", 
                    "0.07704796856462338`"}], ",", 
                    
                    RowBox[{"\<\"lastInt\"\>", "\[Rule]", 
                    "0.20557529535092106`"}], ",", 
                    RowBox[{"\<\"initialized\"\>", "\[Rule]", "True"}]}], 
                    "\[RightAssociation]"}]}
                    },
                    AutoDelete->False,
                    
                    GridBoxItemSize->{
                    "Columns" -> {{Automatic}}, "Rows" -> {{Automatic}}}],
                    "Grid"], {-1, -0.5}], 
                  {GrayLevel[0], LineBox[{{-5, 0}, {5, 0}}]}, 
                  {GrayLevel[0.5], 
                   CircleBox[{-0.03473485853420891, 0.5}, 0.5], 
                   DiskBox[{-0.03473485853420891, 0.5}, 0.1]}, 
                  {GrayLevel[0.5], 
                   DiskBox[{-0.26540761496158627`, 1.9821572384339645`}, 
                    0.1]}, 
                  {GrayLevel[0.5], Thickness[0.01], 
                   LineBox[{{-0.03473485853420891, 
                    0.5}, {-0.26540761496158627`, 1.9821572384339645`}}]}, 
                  {RGBColor[0, 1, 0], 
                   DiskBox[{-0.18851669615246047`, 1.488104825622643}, 0.05]}, 
                  {RGBColor[0, 0, 1], Arrowheads[0.01], 
                   CircleBox[{-0.03473485853420891, 0.5}, 0.3333333333333333, 
                    NCache[{
                    Rational[-1, 2] Pi, 
                    0.6641674808658093}, {-1.5707963267948966`, 
                    0.6641674808658093}]], 
                   ArrowBox[{{0.22978340478008777`, 0.7028329349102347}, {
                    0.2277418834333554, 0.7054679318949888}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-0.03473485853420891, 
                    0.5}, {-1.5247107303080127`, 0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-0.03473485853420891, 0.5}, {
                    1.0828744373890955`, 0.5}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.005], Arrowheads[0.04], 
                   ArrowBox[{{-0.03473485853420891, 
                    0.5}, {-0.40710143438470847`, 0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-0.03473485853420891, 0.5}, 1.6, {
                    1.7251908344587017`, 2.4321055178394952`}], 
                   ArrowBox[{{-1.238164946658202, 
                    1.5543984175812662`}, {-1.2486485840988448`, 
                    1.5423115977891573`}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Dashing[{Small, Small}], 
                   Arrowheads[0.02], 
                   CircleBox[{-0.03473485853420891, 0.5}, 1.7, {
                    1.7251908344587017`, 1.5120125949187682`}], 
                   ArrowBox[{{0.048164595111469116`, 2.19797752652538}, {
                    0.06513994244374252, 2.197063647636592}}]}, 
                  {RGBColor[0, 0, 1], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-0.03473485853420891, 0.5}, 1.8, {
                    1.7251908344587017`, 1.3113086478548672`}], 
                   ArrowBox[{{0.40969870008983933`, 2.244270280652848}, {
                    0.42711889069335685`, 2.239738806351241}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.003], Arrowheads[0.025], 
                   CircleBox[{-0.03473485853420891, 0.5}, 1.5, {
                    1.7251908344587017`, 1.740972534661689}], 
                   ArrowBox[{{-0.2739731005646634, 
                    1.9807987923921933`}, {-0.28876887987760125`, 
                    1.9783324105221074`}}]}, 
                  {RGBColor[0, 0, 1], ArrowBox[{{0, 2.8}, {-1.5, 2.8}}]}},
                 Frame->True,
                 FrameTicks->Automatic,
                 ImageSize->800,
                 PlotRange->{{-3, 3}, {-1, 3}}], 270->
                GraphicsBox[{InsetBox[
                   TagBox[GridBox[{
                    {"\<\"t\"\>", "13.450000000000001`"},
                    {"\<\"driveIntent\"\>", 
                    RowBox[{"-", "0.5`"}]},
                    {"\<\"motorTorque\"\>", "68.6725089012038`"},
                    {"\<\"vw\"\>", 
                    RowBox[{"-", "0.247438717934092`"}]},
                    {"\<\"controllerState\"\>", 
                    RowBox[{"\[LeftAssociation]", 
                    RowBox[{
                    
                    RowBox[{"\<\"lastError\"\>", "\[Rule]", 
                    "0.06690891141670881`"}], ",", 
                    
                    RowBox[{"\<\"lastInt\"\>", "\[Rule]", 
                    "0.20916799403333217`"}], ",", 
                    RowBox[{"\<\"initialized\"\>", "\[Rule]", "True"}]}], 
                    "\[RightAssociation]"}]}
                    },
                    AutoDelete->False,
                    
                    GridBoxItemSize->{
                    "Columns" -> {{Automatic}}, "Rows" -> {{Automatic}}}],
                    "Grid"], {-1, -0.5}], 
                  {GrayLevel[0], LineBox[{{-5, 0}, {5, 0}}]}, 
                  {GrayLevel[0.5], 
                   CircleBox[{-0.046963477686407115`, 0.5}, 0.5], 
                   DiskBox[{-0.046963477686407115`, 0.5}, 0.1]}, 
                  {GrayLevel[0.5], 
                   DiskBox[{-0.2599573123852556, 1.9848008709521556`}, 0.1]}, 
                  {GrayLevel[0.5], Thickness[0.01], 
                   LineBox[{{-0.046963477686407115`, 
                    0.5}, {-0.2599573123852556, 1.9848008709521556`}}]}, 
                  {RGBColor[0, 1, 0], 
                   DiskBox[{-0.1889593674856394, 1.489867247301437}, 0.05]}, 
                  {RGBColor[0, 0, 1], Arrowheads[0.01], 
                   CircleBox[{-0.046963477686407115`, 0.5}, 
                    0.3333333333333333, 
                    NCache[{
                    Rational[-1, 2] Pi, 
                    0.48937894024121764`}, {-1.5707963267948966`, 
                    0.48937894024121764`}]], 
                   ArrowBox[{{0.2487970370337605, 0.653742736555496}, {
                    0.24724484738936398`, 0.6566926053367564}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-0.046963477686407115`, 
                    0.5}, {-1.4204136557104832`, 0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-0.046963477686407115`, 0.5}, {
                    0.9868320928959838, 0.5}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.005], Arrowheads[0.04], 
                   ArrowBox[{{-0.046963477686407115`, 
                    0.5}, {-0.38661808512809226`, 0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-0.046963477686407115`, 0.5}, 1.6, {
                    1.7132737720321003`, 2.366010018610057}], 
                   ArrowBox[{{-1.1781256025650904`, 
                    1.6315795364179853`}, {-1.1893846516986957`, 
                    1.6202115251899458`}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Dashing[{Small, Small}], 
                   Arrowheads[0.02], 
                   CircleBox[{-0.046963477686407115`, 0.5}, 1.7, {
                    1.7132737720321003`, 1.5184683833503438`}], 
                   ArrowBox[{{0.02497254092631168, 2.1984773207865187`}, {
                    0.041953434285081415`, 2.1976730494313252`}}]}, 
                  {RGBColor[0, 0, 1], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-0.046963477686407115`, 0.5}, 1.8, {
                    1.7132737720321003`, 1.331759833692079}], 
                   ArrowBox[{{0.3617072327436384, 2.2529940816889837`}, {
                    0.3792164480310663, 2.2488197937224546`}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.003], Arrowheads[0.025], 
                   CircleBox[{-0.046963477686407115`, 0.5}, 1.5, {
                    1.7132737720321003`, 1.730929601867905}], 
                   ArrowBox[{{-0.27131834221185397`, 
                    1.9831267291650327`}, {-0.28613814466720666`, 
                    1.9808090621935779`}}]}, 
                  {RGBColor[0, 0, 1], ArrowBox[{{0, 2.8}, {-1.5, 2.8}}]}},
                 Frame->True,
                 FrameTicks->Automatic,
                 ImageSize->800,
                 PlotRange->{{-3, 3}, {-1, 3}}], 271->
                GraphicsBox[{InsetBox[
                   TagBox[GridBox[{
                    {"\<\"t\"\>", "13.5`"},
                    {"\<\"driveIntent\"\>", 
                    RowBox[{"-", "0.5`"}]},
                    {"\<\"motorTorque\"\>", "63.07161489894655`"},
                    {"\<\"vw\"\>", 
                    RowBox[{"-", "0.25283884658308436`"}]},
                    {"\<\"controllerState\"\>", 
                    RowBox[{"\[LeftAssociation]", 
                    RowBox[{
                    
                    RowBox[{"\<\"lastError\"\>", "\[Rule]", 
                    "0.05707529466424245`"}], ",", 
                    
                    RowBox[{"\<\"lastInt\"\>", "\[Rule]", 
                    "0.2122612969065497`"}], ",", 
                    RowBox[{"\<\"initialized\"\>", "\[Rule]", "True"}]}], 
                    "\[RightAssociation]"}]}
                    },
                    AutoDelete->False,
                    
                    GridBoxItemSize->{
                    "Columns" -> {{Automatic}}, "Rows" -> {{Automatic}}}],
                    "Grid"], {-1, -0.5}], 
                  {GrayLevel[0], LineBox[{{-5, 0}, {5, 0}}]}, 
                  {GrayLevel[0.5], 
                   CircleBox[{-0.05947531700549755, 0.5}, 0.5], 
                   DiskBox[{-0.05947531700549755, 0.5}, 0.1]}, 
                  {GrayLevel[0.5], 
                   DiskBox[{-0.25545906003759333`, 1.987141678679987}, 0.1]}, 
                  {GrayLevel[0.5], Thickness[0.01], 
                   LineBox[{{-0.05947531700549755, 
                    0.5}, {-0.25545906003759333`, 1.987141678679987}}]}, 
                  {RGBColor[0, 1, 0], 
                   DiskBox[{-0.19013114569356138`, 1.4914277857866578`}, 
                    0.05]}, 
                  {RGBColor[0, 0, 1], Arrowheads[0.01], 
                   CircleBox[{-0.05947531700549755, 0.5}, 0.3333333333333333, 
                    NCache[{
                    Rational[-1, 2] Pi, 
                    0.3213521201734999}, {-1.5707963267948966`, 
                    0.3213521201734999}]], 
                   ArrowBox[{{0.25783142417223115`, 0.6021153422081178}, {
                    0.2567944225644404, 0.6052832510111401}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-0.05947531700549755, 
                    0.5}, {-1.3209076149844285`, 0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-0.05947531700549755, 0.5}, {0.8932590120734583,
                     0.5}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.005], Arrowheads[0.04], 
                   ArrowBox[{{-0.05947531700549755, 
                    0.5}, {-0.3681732859054727, 0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-0.05947531700549755, 0.5}, 1.6, {
                    1.7018267769162723`, 2.302434268668054}], 
                   ArrowBox[{{-1.1164596385879446`, 
                    1.7011594997871409`}, {-1.1284181846178964`, 
                    1.6905297752599815`}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Dashing[{Small, Small}], 
                   Arrowheads[0.02], 
                   CircleBox[{-0.05947531700549755, 0.5}, 1.7, {
                    1.7018267769162723`, 1.5244908766248007`}], 
                   ArrowBox[{{0.0022303906302478804`, 2.1988797502016357`}, {
                    0.019215819727383787`, 2.198177760129865}}]}, 
                  {RGBColor[0, 0, 1], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-0.05947531700549755, 0.5}, 1.8, {
                    1.7018267769162723`, 1.3514289163665691`}], 
                   ArrowBox[{{0.3146387818310117, 2.2606926594530194`}, {
                    0.3322267094291711, 2.2568635469173395`}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.003], Arrowheads[0.025], 
                   CircleBox[{-0.05947531700549755, 0.5}, 1.5, {
                    1.7018267769162723`, 1.7210166484615281`}], 
                   ArrowBox[{{-0.2691172329468825, 
                    1.9852778417119221`}, {-0.2839592818104859, 
                    1.983107194219431}}]}, 
                  {RGBColor[0, 0, 1], ArrowBox[{{0, 2.8}, {-1.5, 2.8}}]}},
                 Frame->True,
                 FrameTicks->Automatic,
                 ImageSize->800,
                 PlotRange->{{-3, 3}, {-1, 3}}], 272->
                GraphicsBox[{InsetBox[
                   TagBox[GridBox[{
                    {"\<\"t\"\>", "13.55`"},
                    {"\<\"driveIntent\"\>", 
                    RowBox[{"-", "0.5`"}]},
                    {"\<\"motorTorque\"\>", "57.71781207322385`"},
                    {"\<\"vw\"\>", 
                    RowBox[{"-", "0.25772746673698627`"}]},
                    {"\<\"controllerState\"\>", 
                    RowBox[{"\[LeftAssociation]", 
                    RowBox[{
                    
                    RowBox[{"\<\"lastError\"\>", "\[Rule]", 
                    "0.04759833083088749`"}], ",", 
                    
                    RowBox[{"\<\"lastInt\"\>", "\[Rule]", 
                    "0.21487181825338872`"}], ",", 
                    RowBox[{"\<\"initialized\"\>", "\[Rule]", "True"}]}], 
                    "\[RightAssociation]"}]}
                    },
                    AutoDelete->False,
                    
                    GridBoxItemSize->{
                    "Columns" -> {{Automatic}}, "Rows" -> {{Automatic}}}],
                    "Grid"], {-1, -0.5}], 
                  {GrayLevel[0], LineBox[{{-5, 0}, {5, 0}}]}, 
                  {GrayLevel[0.5], CircleBox[{-0.0722439800248975, 0.5}, 0.5],
                    DiskBox[{-0.0722439800248975, 0.5}, 0.1]}, 
                  {GrayLevel[0.5], 
                   DiskBox[{-0.2519530163631193, 1.9891959784589763`}, 0.1]}, 
                  {GrayLevel[0.5], Thickness[0.01], 
                   LineBox[{{-0.0722439800248975, 0.5}, {-0.2519530163631193, 
                    1.9891959784589763`}}]}, 
                  {RGBColor[0, 1, 0], 
                   DiskBox[{-0.1920500042503787, 1.4927973189726509`}, 0.05]}, 
                  {RGBColor[0, 0, 1], Arrowheads[0.01], 
                   CircleBox[{-0.0722439800248975, 0.5}, 0.3333333333333333, 
                    NCache[{
                    Rational[-1, 2] Pi, 
                    0.16073803540181908`}, {-1.5707963267948966`, 
                    0.16073803540181908`}]], 
                   ArrowBox[{{0.257309525983092, 0.5500559465893229}, {
                    0.25679249732182785`, 0.5533489239476201}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-0.0722439800248975, 
                    0.5}, {-1.2266002214893745`, 0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-0.0722439800248975, 0.5}, {0.802581031096148, 
                    0.5}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.005], Arrowheads[0.04], 
                   ArrowBox[{{-0.0722439800248975, 
                    0.5}, {-0.35177521036832904`, 0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-0.0722439800248975, 0.5}, 1.6, {
                    1.690890823828675, 2.2416232054556193`}], 
                   ArrowBox[{{-1.0542757720484046`, 
                    1.763175981189913}, {-1.0668582201516035`, 
                    1.7532926686680872`}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Dashing[{Small, Small}], 
                   Arrowheads[0.02], 
                   CircleBox[{-0.0722439800248975, 0.5}, 1.7, {
                    1.690890823828675, 1.5300833869315855`}], 
                   ArrowBox[{{-0.020040190333022782`, 
                    2.1991982710507347`}, {-0.0030511009885443585`, 
                    2.1985912826488487`}}]}, 
                  {RGBColor[0, 0, 1], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-0.0722439800248975, 0.5}, 1.8, {
                    1.690890823828675, 1.3702363123107648`}], 
                   ArrowBox[{{0.2686918634460323, 2.267416971355873}, {
                    0.28634869194145124`, 2.2639192996313735`}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.003], Arrowheads[0.025], 
                   CircleBox[{-0.0722439800248975, 0.5}, 1.5, {
                    1.690890823828675, 1.7113048552592223`}], 
                   ArrowBox[{{-0.2674515249498246, 
                    1.9872437642849212`}, {-0.28231395442404245`, 
                    1.9852173598015688`}}]}, 
                  {RGBColor[0, 0, 1], ArrowBox[{{0, 2.8}, {-1.5, 2.8}}]}},
                 Frame->True,
                 FrameTicks->Automatic,
                 ImageSize->800,
                 PlotRange->{{-3, 3}, {-1, 3}}], 273->
                GraphicsBox[{InsetBox[
                   TagBox[GridBox[{
                    {"\<\"t\"\>", "13.6`"},
                    {"\<\"driveIntent\"\>", 
                    RowBox[{"-", "0.5`"}]},
                    {"\<\"motorTorque\"\>", "52.628818402696126`"},
                    {"\<\"vw\"\>", 
                    RowBox[{"-", "0.26213785240341164`"}]},
                    {"\<\"controllerState\"\>", 
                    RowBox[{"\[LeftAssociation]", 
                    RowBox[{
                    
                    RowBox[{"\<\"lastError\"\>", "\[Rule]", 
                    "0.03852086424820653`"}], ",", 
                    
                    RowBox[{"\<\"lastInt\"\>", "\[Rule]", 
                    "0.2170185166717277`"}], ",", 
                    RowBox[{"\<\"initialized\"\>", "\[Rule]", "True"}]}], 
                    "\[RightAssociation]"}]}
                    },
                    AutoDelete->False,
                    
                    GridBoxItemSize->{
                    "Columns" -> {{Automatic}}, "Rows" -> {{Automatic}}}],
                    "Grid"], {-1, -0.5}], 
                  {GrayLevel[0], LineBox[{{-5, 0}, {5, 0}}]}, 
                  {GrayLevel[0.5], 
                   CircleBox[{-0.08524474124146832, 0.5}, 0.5], 
                   DiskBox[{-0.08524474124146832, 0.5}, 0.1]}, 
                  {GrayLevel[0.5], 
                   DiskBox[{-0.2494683157974591, 1.990983104384504}, 0.1]}, 
                  {GrayLevel[0.5], Thickness[0.01], 
                   LineBox[{{-0.08524474124146832, 0.5}, {-0.2494683157974591,
                     1.990983104384504}}]}, 
                  {RGBColor[0, 1, 0], 
                   DiskBox[{-0.19472712427879552`, 1.493988736256336}, 0.05]}, 
                  {RGBColor[0, 0, 1], Arrowheads[0.01], 
                   CircleBox[{-0.08524474124146832, 0.5}, 0.3333333333333333, 
                    NCache[{
                    Rational[-1, 2] Pi, 
                    0.008068225285987385}, {-1.5707963267948966`, 
                    0.008068225285987385}]], 
                   ArrowBox[{{0.24808797013313416`, 0.4993560754958239}, {
                    0.24807774277417516`, 0.5026893792503527}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-0.08524474124146832, 0.5}, {-1.137821109295391,
                     0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-0.08524474124146832, 0.5}, {0.7151562980815219,
                     0.5}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.005], Arrowheads[0.04], 
                   ArrowBox[{{-0.08524474124146832, 
                    0.5}, {-0.33742006997240065`, 0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-0.08524474124146832, 0.5}, 1.6, {
                    1.6804986145117697`, 2.1837745887053654`}], 
                   ArrowBox[{{-0.9926015911398283, 
                    1.81784048615245}, {-1.0057344088979387`, 
                    1.8087011774036428`}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Dashing[{Small, Small}], 
                   Arrowheads[0.02], 
                   CircleBox[{-0.08524474124146832, 0.5}, 1.7, {
                    1.6804986145117697`, 1.5352503748105337`}], 
                   ArrowBox[{{-0.04182134621420927, 
                    2.1994453238525526`}, {-0.024829347366813186`, 
                    2.1989261255813832`}}]}, 
                  {RGBColor[0, 0, 1], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-0.08524474124146832, 0.5}, 1.8, {
                    1.6804986145117697`, 1.3881162900523467`}], 
                   ArrowBox[{{0.22403691383516694`, 2.2732300634246694`}, {
                    0.24175345497866052`, 2.2700486376562603`}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.003], Arrowheads[0.025], 
                   CircleBox[{-0.08524474124146832, 0.5}, 1.5, {
                    1.6804986145117697`, 1.701852859305564}], 
                   ArrowBox[{{-0.2663863535605655, 
                    1.9890224028826557`}, {-0.2812672724150919, 
                    1.9871365664498624`}}]}, 
                  {RGBColor[0, 0, 1], ArrowBox[{{0, 2.8}, {-1.5, 2.8}}]}},
                 Frame->True,
                 FrameTicks->Automatic,
                 ImageSize->800,
                 PlotRange->{{-3, 3}, {-1, 3}}], 274->
                GraphicsBox[{InsetBox[
                   TagBox[GridBox[{
                    {"\<\"t\"\>", "13.65`"},
                    {"\<\"driveIntent\"\>", 
                    RowBox[{"-", "0.5`"}]},
                    {"\<\"motorTorque\"\>", "47.818566242078504`"},
                    {"\<\"vw\"\>", 
                    RowBox[{"-", "0.26610286890343116`"}]},
                    {"\<\"controllerState\"\>", 
                    RowBox[{"\[LeftAssociation]", 
                    RowBox[{
                    
                    RowBox[{"\<\"lastError\"\>", "\[Rule]", 
                    "0.02987802569489513`"}], ",", 
                    
                    RowBox[{"\<\"lastInt\"\>", "\[Rule]", 
                    "0.21872229373378385`"}], ",", 
                    RowBox[{"\<\"initialized\"\>", "\[Rule]", "True"}]}], 
                    "\[RightAssociation]"}]}
                    },
                    AutoDelete->False,
                    
                    GridBoxItemSize->{
                    "Columns" -> {{Automatic}}, "Rows" -> {{Automatic}}}],
                    "Grid"], {-1, -0.5}], 
                  {GrayLevel[0], LineBox[{{-5, 0}, {5, 0}}]}, 
                  {GrayLevel[0.5], 
                   CircleBox[{-0.09845452895518815, 0.5}, 0.5], 
                   DiskBox[{-0.09845452895518815, 0.5}, 0.1]}, 
                  {GrayLevel[0.5], 
                   DiskBox[{-0.2480239777658012, 1.9925243649543847`}, 0.1]}, 
                  {GrayLevel[0.5], Thickness[0.01], 
                   LineBox[{{-0.09845452895518815, 0.5}, {-0.2480239777658012,
                     1.9925243649543847`}}]}, 
                  {RGBColor[0, 1, 0], 
                   DiskBox[{-0.1981674948289302, 1.4950162433029233`}, 0.05]}, 
                  {RGBColor[0, 0, 1], Arrowheads[0.01], 
                   CircleBox[{-0.09845452895518815, 0.5}, 0.3333333333333333, 
                    NCache[{
                    Rational[-1, 2] 
                    Pi, -0.13623933953254141`}, {-1.5707963267948966`, \
-0.13623933953254141`}]], 
                   ArrowBox[{{0.23132082798350184`, 0.45142711592801243`}, {
                    0.23179006009834047`, 0.45472724315907964`}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-0.09845452895518815, 
                    0.5}, {-1.0548258537967583`, 0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-0.09845452895518815, 0.5}, {0.6312780631833973,
                     0.5}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.005], Arrowheads[0.04], 
                   ArrowBox[{{-0.09845452895518815, 
                    0.5}, {-0.3250932616581728, 0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-0.09845452895518815, 0.5}, 1.6, {
                    1.6706752719831774`, 2.1290425329764893`}], 
                   ArrowBox[{{-0.9323605667499132, 
                    1.865503833802565}, {-0.9459736825506742, 
                    1.8570966377855305`}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Dashing[{Small, Small}], 
                   Arrowheads[0.02], 
                   CircleBox[{-0.09845452895518815, 0.5}, 1.7, {
                    1.6706752719831774`, 1.5399976800461428`}], 
                   ArrowBox[{{-0.06309937862128592, 
                    2.199632317104163}, {-0.046105106463666196`, 
                    2.1991937905856425`}}]}, 
                  {RGBColor[0, 0, 1], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-0.09845452895518815, 0.5}, 1.8, {
                    1.6706752719831774`, 1.4050165706382969`}], 
                   ArrowBox[{{0.1808162996818678, 2.2782035328590404`}, {
                    0.19858407521961519`, 2.275321961681848}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.003], Arrowheads[0.025], 
                   CircleBox[{-0.09845452895518815, 0.5}, 1.5, {
                    1.6706752719831774`, 1.692708637900214}], 
                   ArrowBox[{{-0.26597280731080997`, 
                    1.9906165256083699`}, {-0.28087034828792956`, 
                    1.988866840539195}}]}, 
                  {RGBColor[0, 0, 1], ArrowBox[{{0, 2.8}, {-1.5, 2.8}}]}},
                 Frame->True,
                 FrameTicks->Automatic,
                 ImageSize->800,
                 PlotRange->{{-3, 3}, {-1, 3}}], 275->
                GraphicsBox[{InsetBox[
                   TagBox[GridBox[{
                    {"\<\"t\"\>", "13.700000000000001`"},
                    {"\<\"driveIntent\"\>", 
                    RowBox[{"-", "0.5`"}]},
                    {"\<\"motorTorque\"\>", "43.297402977853956`"},
                    {"\<\"vw\"\>", 
                    RowBox[{"-", "0.2696548555812811`"}]},
                    {"\<\"controllerState\"\>", 
                    RowBox[{"\[LeftAssociation]", 
                    RowBox[{
                    
                    RowBox[{"\<\"lastError\"\>", "\[Rule]", 
                    "0.021697810059806352`"}], ",", 
                    
                    RowBox[{"\<\"lastInt\"\>", "\[Rule]", 
                    "0.22000562330739104`"}], ",", 
                    RowBox[{"\<\"initialized\"\>", "\[Rule]", "True"}]}], 
                    "\[RightAssociation]"}]}
                    },
                    AutoDelete->False,
                    
                    GridBoxItemSize->{
                    "Columns" -> {{Automatic}}, "Rows" -> {{Automatic}}}],
                    "Grid"], {-1, -0.5}], 
                  {GrayLevel[0], LineBox[{{-5, 0}, {5, 0}}]}, 
                  {GrayLevel[0.5], 
                   CircleBox[{-0.11185190177690306`, 0.5}, 0.5], 
                   DiskBox[{-0.11185190177690306`, 0.5}, 0.1]}, 
                  {GrayLevel[0.5], 
                   DiskBox[{-0.24762973431252422`, 1.9938421537069868`}, 
                    0.1]}, 
                  {GrayLevel[0.5], Thickness[0.01], 
                   LineBox[{{-0.11185190177690306`, 
                    0.5}, {-0.24762973431252422`, 1.9938421537069868`}}]}, 
                  {RGBColor[0, 1, 0], 
                   DiskBox[{-0.20237045680065047`, 1.4958947691379914`}, 
                    0.05]}, 
                  {RGBColor[0, 0, 1], Arrowheads[0.01], 
                   CircleBox[{-0.11185190177690306`, 0.5}, 0.3333333333333333,
                     NCache[{
                    Rational[-1, 2] 
                    Pi, -0.27187423745927797`}, {-1.5707963267948966`, \
-0.27187423745927797`}]], 
                   ArrowBox[{{0.20832669639790574`, 0.40728119725791634`}, {
                    0.20923786017576862`, 0.41048756577833595`}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-0.11185190177690306`, 
                    0.5}, {-0.9777999613339822, 0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-0.11185190177690306`, 0.5}, {
                    0.5511778287803284, 0.5}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.005], Arrowheads[0.04], 
                   ArrowBox[{{-0.11185190177690306`, 
                    0.5}, {-0.31477023077675076`, 0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-0.11185190177690306`, 0.5}, 1.6, {
                    1.661438952099314, 2.077540727547892}], 
                   ArrowBox[{{-0.8743574151012451, 
                    1.9066219613492397`}, {-0.8883852753209591, 
                    1.8989267027876364`}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Dashing[{Small, Small}], 
                   Arrowheads[0.02], 
                   CircleBox[{-0.11185190177690306`, 0.5}, 1.7, {
                    1.661438952099314, 1.544332699956703}], 
                   ArrowBox[{{-0.08386500050492596, 
                    2.199769611846615}, {-0.06686898701338106, 
                    2.199404759725998}}]}, 
                  {RGBColor[0, 0, 1], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-0.11185190177690306`, 0.5}, 1.8, {
                    1.661438952099314, 1.42089782444457}], 
                   ArrowBox[{{0.13914479501593363`, 2.282414277938517}, {
                    0.1569560909974998, 2.279815232831935}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.003], Arrowheads[0.025], 
                   CircleBox[{-0.11185190177690306`, 0.5}, 1.5, {
                    1.661438952099314, 1.683911073630478}], 
                   ArrowBox[{{-0.26625007193222944`, 
                    1.9920325750641932`}, {-0.2811624291678435, 
                    1.9904140180884644`}}]}, 
                  {RGBColor[0, 0, 1], ArrowBox[{{0, 2.8}, {-1.5, 2.8}}]}},
                 Frame->True,
                 FrameTicks->Automatic,
                 ImageSize->800,
                 PlotRange->{{-3, 3}, {-1, 3}}], 276->
                GraphicsBox[{InsetBox[
                   TagBox[GridBox[{
                    {"\<\"t\"\>", "13.75`"},
                    {"\<\"driveIntent\"\>", 
                    RowBox[{"-", "0.5`"}]},
                    {"\<\"motorTorque\"\>", "39.0723007844976`"},
                    {"\<\"vw\"\>", 
                    RowBox[{"-", "0.27282552117115355`"}]},
                    {"\<\"controllerState\"\>", 
                    RowBox[{"\[LeftAssociation]", 
                    RowBox[{
                    
                    RowBox[{"\<\"lastError\"\>", "\[Rule]", 
                    "0.014001600166936981`"}], ",", 
                    
                    RowBox[{"\<\"lastInt\"\>", "\[Rule]", 
                    "0.22089220830177692`"}], ",", 
                    RowBox[{"\<\"initialized\"\>", "\[Rule]", "True"}]}], 
                    "\[RightAssociation]"}]}
                    },
                    AutoDelete->False,
                    
                    GridBoxItemSize->{
                    "Columns" -> {{Automatic}}, "Rows" -> {{Automatic}}}],
                    "Grid"], {-1, -0.5}], 
                  {GrayLevel[0], LineBox[{{-5, 0}, {5, 0}}]}, 
                  {GrayLevel[0.5], 
                   CircleBox[{-0.12541701962129115`, 0.5}, 0.5], 
                   DiskBox[{-0.12541701962129115`, 0.5}, 0.1]}, 
                  {GrayLevel[0.5], 
                   DiskBox[{-0.2482868050455293, 1.9949592020619833`}, 0.1]}, 
                  {GrayLevel[0.5], Thickness[0.01], 
                   LineBox[{{-0.12541701962129115`, 
                    0.5}, {-0.2482868050455293, 1.9949592020619833`}}]}, 
                  {RGBColor[0, 1, 0], 
                   DiskBox[{-0.2073302099041166, 1.4966394680413222`}, 0.05]}, 
                  {RGBColor[0, 0, 1], Arrowheads[0.01], 
                   CircleBox[{-0.12541701962129115`, 0.5}, 0.3333333333333333,
                     NCache[{
                    Rational[-1, 2] 
                    Pi, -0.39862730325996854`}, {-1.5707963267948966`, \
-0.39862730325996854`}]], 
                   ArrowBox[{{0.1804720231474648, 0.3675499919776322}, {
                    0.18178120682811244`, 0.37061545386928135`}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-0.12541701962129115`, 
                    0.5}, {-0.9068630353112431, 0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-0.12541701962129115`, 0.5}, {
                    0.47502884965613834`, 0.5}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.005], Arrowheads[0.04], 
                   ArrowBox[{{-0.12541701962129115`, 
                    0.5}, {-0.30641716603381364`, 0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-0.12541701962129115`, 0.5}, 1.6, {
                    1.652801397894888, 2.029345454529452}], 
                   ArrowBox[{{-0.8192708484765444, 
                    1.9417235741233148`}, {-0.8336531515293784, 
                    1.934713065898499}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Dashing[{Small, Small}], 
                   Arrowheads[0.02], 
                   CircleBox[{-0.12541701962129115`, 0.5}, 1.7, {
                    1.652801397894888, 1.5482645058882867`}], 
                   ArrowBox[{{-0.10411348169807416`, 
                    2.1998665121920467`}, {-0.0871161650538422, 
                    2.199568487746052}}]}, 
                  {RGBColor[0, 0, 1], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-0.12541701962129115`, 0.5}, 1.8, {
                    1.652801397894888, 1.4357330602032343`}], 
                   ArrowBox[{{0.09911049135930519, 2.2859415994967076`}, {
                    0.1169583834168311, 2.283607065472131}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.003], Arrowheads[0.025], 
                   CircleBox[{-0.12541701962129115`, 0.5}, 1.5, {
                    1.652801397894888, 1.675491242134291}], 
                   ArrowBox[{{-0.26724720245672506`, 
                    1.9932796788401252`}, {-0.2821726589163782, 
                    1.9917867372882723`}}]}, 
                  {RGBColor[0, 0, 1], ArrowBox[{{0, 2.8}, {-1.5, 2.8}}]}},
                 Frame->True,
                 FrameTicks->Automatic,
                 ImageSize->800,
                 PlotRange->{{-3, 3}, {-1, 3}}], 277->
                GraphicsBox[{InsetBox[
                   TagBox[GridBox[{
                    {"\<\"t\"\>", "13.8`"},
                    {"\<\"driveIntent\"\>", 
                    RowBox[{"-", "0.5`"}]},
                    {"\<\"motorTorque\"\>", "35.14707812742652`"},
                    {"\<\"vw\"\>", 
                    RowBox[{"-", "0.2756458466887411`"}]},
                    {"\<\"controllerState\"\>", 
                    RowBox[{"\[LeftAssociation]", 
                    RowBox[{
                    
                    RowBox[{"\<\"lastError\"\>", "\[Rule]", 
                    "0.00680465492640836`"}], ",", 
                    
                    RowBox[{"\<\"lastInt\"\>", "\[Rule]", 
                    "0.22140666262916114`"}], ",", 
                    RowBox[{"\<\"initialized\"\>", "\[Rule]", "True"}]}], 
                    "\[RightAssociation]"}]}
                    },
                    AutoDelete->False,
                    
                    GridBoxItemSize->{
                    "Columns" -> {{Automatic}}, "Rows" -> {{Automatic}}}],
                    "Grid"], {-1, -0.5}], 
                  {GrayLevel[0], LineBox[{{-5, 0}, {5, 0}}]}, 
                  {GrayLevel[0.5], CircleBox[{-0.1391316096762019, 0.5}, 0.5],
                    DiskBox[{-0.1391316096762019, 0.5}, 0.1]}, 
                  {GrayLevel[0.5], 
                   DiskBox[{-0.24998864004544719`, 1.995897964039564}, 0.1]}, 
                  {GrayLevel[0.5], Thickness[0.01], 
                   LineBox[{{-0.1391316096762019, 
                    0.5}, {-0.24998864004544719`, 1.995897964039564}}]}, 
                  {RGBColor[0, 1, 0], 
                   DiskBox[{-0.2130362965890321, 1.4972653093597093`}, 0.05]}, 
                  {RGBColor[0, 0, 1], Arrowheads[0.01], 
                   CircleBox[{-0.1391316096762019, 0.5}, 0.3333333333333333, 
                    NCache[{
                    Rational[-1, 2] 
                    Pi, -0.5163839829721011}, {-1.5707963267948966`, \
-0.5163839829721011}]], 
                   ArrowBox[{{0.14907820460915736`, 0.3325299607072661}, {
                    0.15073846671992377`, 0.3354203842475765}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-0.1391316096762019, 0.5}, {-0.8420731722247322,
                     0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-0.1391316096762019, 0.5}, {0.4029499483046607, 
                    0.5}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.005], Arrowheads[0.04], 
                   ArrowBox[{{-0.1391316096762019, 0.5}, {-0.2999916142438697,
                     0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-0.1391316096762019, 0.5}, 1.6, {
                    1.6447684563058766`, 1.9844984865779265`}], 
                   ArrowBox[{{-0.7676525389732638, 
                    1.9713807941642971`}, {-0.7823346759014186, 
                    1.9650221211976562`}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Dashing[{Small, Small}], 
                   Arrowheads[0.02], 
                   CircleBox[{-0.1391316096762019, 0.5}, 1.7, {
                    1.6447684563058766`, 1.5518038945557073`}], 
                   ArrowBox[{{-0.12384468089814477`, 
                    2.199931266201235}, {-0.10684641589666287`, 
                    2.1996934036062554`}}]}, 
                  {RGBColor[0, 0, 1], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-0.1391316096762019, 0.5}, 1.8, {
                    1.6447684563058766`, 1.449506911153507}], 
                   ArrowBox[{{0.06077608749775745, 2.2888646993583963`}, {
                    0.0786544410471515, 2.2867762132148295`}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.003], Arrowheads[0.025], 
                   CircleBox[{-0.1391316096762019, 0.5}, 1.5, {
                    1.6447684563058766`, 1.6674734644296945`}], 
                   ArrowBox[{{-0.2689845775382583, 
                    1.9943688322289836`}, {-0.2839215242060335, 
                    1.9929956063734562`}}]}, 
                  {RGBColor[0, 0, 1], ArrowBox[{{0, 2.8}, {-1.5, 2.8}}]}},
                 Frame->True,
                 FrameTicks->Automatic,
                 ImageSize->800,
                 PlotRange->{{-3, 3}, {-1, 3}}], 278->
                GraphicsBox[{InsetBox[
                   TagBox[GridBox[{
                    {"\<\"t\"\>", "13.85`"},
                    {"\<\"driveIntent\"\>", 
                    RowBox[{"-", "0.5`"}]},
                    {"\<\"motorTorque\"\>", "31.522633505432175`"},
                    {"\<\"vw\"\>", 
                    RowBox[{"-", "0.2781459929942282`"}]},
                    {"\<\"controllerState\"\>", 
                    RowBox[{"\[LeftAssociation]", 
                    RowBox[{
                    
                    RowBox[{"\<\"lastError\"\>", "\[Rule]", 
                    "0.00011657538239541465`"}], ",", 
                    
                    RowBox[{"\<\"lastInt\"\>", "\[Rule]", 
                    "0.2215742169626855`"}], ",", 
                    RowBox[{"\<\"initialized\"\>", "\[Rule]", "True"}]}], 
                    "\[RightAssociation]"}]}
                    },
                    AutoDelete->False,
                    
                    GridBoxItemSize->{
                    "Columns" -> {{Automatic}}, "Rows" -> {{Automatic}}}],
                    "Grid"], {-1, -0.5}], 
                  {GrayLevel[0], LineBox[{{-5, 0}, {5, 0}}]}, 
                  {GrayLevel[0.5], 
                   CircleBox[{-0.15297892764362736`, 0.5}, 0.5], 
                   DiskBox[{-0.15297892764362736`, 0.5}, 0.1]}, 
                  {GrayLevel[0.5], 
                   DiskBox[{-0.2527216444335936, 1.9966801229545865`}, 0.1]}, 
                  {GrayLevel[0.5], Thickness[0.01], 
                   LineBox[{{-0.15297892764362736`, 
                    0.5}, {-0.2527216444335936, 1.9966801229545865`}}]}, 
                  {RGBColor[0, 1, 0], 
                   DiskBox[{-0.21947407217027148`, 1.4977867486363912`}, 
                    0.05]}, 
                  {RGBColor[0, 0, 1], Arrowheads[0.01], 
                   CircleBox[{-0.15297892764362736`, 0.5}, 0.3333333333333333,
                     NCache[{
                    Rational[-1, 2] 
                    Pi, -0.6251173216319313}, {-1.5707963267948966`, \
-0.6251173216319313}]], 
                   ArrowBox[{{0.1153551051493602, 0.30224267913383596`}, {
                    0.11731922880879897`, 0.3049358625232953}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-0.15297892764362736`, 
                    0.5}, {-0.7834315977522708, 0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-0.15297892764362736`, 0.5}, {
                    0.33500959219848525`, 0.5}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.005], Arrowheads[0.04], 
                   ArrowBox[{{-0.15297892764362736`, 
                    0.5}, {-0.2954430779101582, 0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-0.15297892764362736`, 0.5}, 1.6, {
                    1.6373405716159075`, 1.943009919618508}], 
                   ArrowBox[{{-0.7199310284523053, 
                    1.9961835834511175`}, {-0.7348642675553214, 
                    1.9904393483788103`}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Dashing[{Small, Small}], 
                   Arrowheads[0.02], 
                   CircleBox[{-0.15297892764362736`, 0.5}, 1.7, {
                    1.6373405716159075`, 1.554963375419948}], 
                   ArrowBox[{{-0.14306296653542694`, 
                    2.199971080258514}, {-0.12606403485386228`, 
                    2.1997869244543904`}}]}, 
                  {RGBColor[0, 0, 1], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-0.15297892764362736`, 0.5}, 1.8, {
                    1.6373405716159075`, 1.462214829919062}], 
                   ArrowBox[{{0.02418049265015207, 2.2912606007505363`}, {
                    0.04208394221851816, 2.28939947379034}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.003], Arrowheads[0.025], 
                   CircleBox[{-0.15297892764362736`, 0.5}, 1.5, {
                    1.6373405716159075`, 1.6598761637603794`}], 
                   ArrowBox[{{-0.27147508791938835`, 
                    1.995312228265355}, {-0.2864220362259427, 
                    1.99405252142349}}]}, 
                  {RGBColor[0, 0, 1], ArrowBox[{{0, 2.8}, {-1.5, 2.8}}]}},
                 Frame->True,
                 FrameTicks->Automatic,
                 ImageSize->800,
                 PlotRange->{{-3, 3}, {-1, 3}}], 279->
                GraphicsBox[{InsetBox[
                   TagBox[GridBox[{
                    {"\<\"t\"\>", "13.9`"},
                    {"\<\"driveIntent\"\>", 
                    RowBox[{"-", "0.5`"}]},
                    {"\<\"motorTorque\"\>", "28.1971902687712`"},
                    {"\<\"vw\"\>", 
                    RowBox[{"-", "0.2803552118759048`"}]},
                    {"\<\"controllerState\"\>", 
                    RowBox[{"\[LeftAssociation]", 
                    RowBox[{
                    RowBox[{"\<\"lastError\"\>", "\[Rule]", 
                    RowBox[{"-", "0.006058241859972606`"}]}], ",", 
                    
                    RowBox[{"\<\"lastInt\"\>", "\[Rule]", 
                    "0.22142044744318792`"}], ",", 
                    RowBox[{"\<\"initialized\"\>", "\[Rule]", "True"}]}], 
                    "\[RightAssociation]"}]}
                    },
                    AutoDelete->False,
                    
                    GridBoxItemSize->{
                    "Columns" -> {{Automatic}}, "Rows" -> {{Automatic}}}],
                    "Grid"], {-1, -0.5}], 
                  {GrayLevel[0], LineBox[{{-5, 0}, {5, 0}}]}, 
                  {GrayLevel[0.5], CircleBox[{-0.1669437144505245, 0.5}, 0.5],
                    DiskBox[{-0.1669437144505245, 0.5}, 0.1]}, 
                  {GrayLevel[0.5], 
                   DiskBox[{-0.2564658924820012, 1.9973262101628024`}, 0.1]}, 
                  {GrayLevel[0.5], Thickness[0.01], 
                   LineBox[{{-0.1669437144505245, 0.5}, {-0.2564658924820012, 
                    1.9973262101628024`}}]}, 
                  {RGBColor[0, 1, 0], 
                   DiskBox[{-0.22662516647150896`, 1.4982174734418683`}, 
                    0.05]}, 
                  {RGBColor[0, 0, 1], Arrowheads[0.01], 
                   CircleBox[{-0.1669437144505245, 0.5}, 0.3333333333333333, 
                    NCache[{
                    Rational[-1, 2] 
                    Pi, -0.7248806187317605}, {-1.5707963267948966`, \
-0.7248806187317605}]], 
                   ArrowBox[{{0.08035988676980854, 0.27650046994553745`}, {
                    0.08258247974359953, 0.2789846396240583}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-0.1669437144505245, 0.5}, {-0.7308875198259486,
                     0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-0.1669437144505245, 0.5}, {
                    0.27123016782317066`, 0.5}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.005], Arrowheads[0.04], 
                   ArrowBox[{{-0.1669437144505245, 
                    0.5}, {-0.29271363755225344`, 0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-0.1669437144505245, 0.5}, 1.6, {
                    1.6305132653772172`, 1.9048609738748363`}], 
                   ArrowBox[{{-0.6764193279431046, 
                    2.0167183651740883`}, {-0.6915607852413214, 
                    2.0115478586650455`}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Dashing[{Small, Small}], 
                   Arrowheads[0.02], 
                   CircleBox[{-0.1669437144505245, 0.5}, 1.7, {
                    1.6305132653772172`, 1.5577570988257983`}], 
                   ArrowBox[{{-0.16177703485708875`, 
                    2.199992148635393}, {-0.1447776550331698, 
                    2.1998554838014632`}}]}, 
                  {RGBColor[0, 0, 1], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-0.1669437144505245, 0.5}, 1.8, {
                    1.6305132653772172`, 1.4738622083284882`}], 
                   ArrowBox[{{-0.010659329192069661`, 2.293202495794713}, {
                    0.007264582746144527, 2.2915500186117708`}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.003], Arrowheads[0.025], 
                   CircleBox[{-0.1669437144505245, 0.5}, 1.5, {
                    1.6305132653772172`, 1.6527125638912614`}], 
                   ArrowBox[{{-0.27472510742965883`, 
                    1.9961227126567784`}, {-0.28968069617894776`, 
                    1.9949701111782123`}}]}, 
                  {RGBColor[0, 0, 1], ArrowBox[{{0, 2.8}, {-1.5, 2.8}}]}},
                 Frame->True,
                 FrameTicks->Automatic,
                 ImageSize->800,
                 PlotRange->{{-3, 3}, {-1, 3}}], 280->
                GraphicsBox[{InsetBox[
                   TagBox[GridBox[{
                    {"\<\"t\"\>", "13.950000000000001`"},
                    {"\<\"driveIntent\"\>", 
                    RowBox[{"-", "0.5`"}]},
                    {"\<\"motorTorque\"\>", "25.16655015229422`"},
                    {"\<\"vw\"\>", 
                    RowBox[{"-", "0.2823017607148336`"}]},
                    {"\<\"controllerState\"\>", 
                    RowBox[{"\[LeftAssociation]", 
                    RowBox[{
                    RowBox[{"\<\"lastError\"\>", "\[Rule]", 
                    RowBox[{"-", "0.011720157862157847`"}]}], ",", 
                    
                    RowBox[{"\<\"lastInt\"\>", "\[Rule]", 
                    "0.2209710268693113`"}], ",", 
                    RowBox[{"\<\"initialized\"\>", "\[Rule]", "True"}]}], 
                    "\[RightAssociation]"}]}
                    },
                    AutoDelete->False,
                    
                    GridBoxItemSize->{
                    "Columns" -> {{Automatic}}, "Rows" -> {{Automatic}}}],
                    "Grid"], {-1, -0.5}], 
                  {GrayLevel[0], LineBox[{{-5, 0}, {5, 0}}]}, 
                  {GrayLevel[0.5], CircleBox[{-0.1810121486029365, 0.5}, 0.5],
                    DiskBox[{-0.1810121486029365, 0.5}, 0.1]}, 
                  {GrayLevel[0.5], 
                   DiskBox[{-0.26119583448915307`, 1.9978553256298488`}, 
                    0.1]}, 
                  {GrayLevel[0.5], Thickness[0.01], 
                   LineBox[{{-0.1810121486029365, 
                    0.5}, {-0.26119583448915307`, 1.9978553256298488`}}]}, 
                  {RGBColor[0, 1, 0], 
                   DiskBox[{-0.23446793919374756`, 1.498570217086566}, 0.05]}, 
                  {RGBColor[0, 0, 1], Arrowheads[0.01], 
                   CircleBox[{-0.1810121486029365, 0.5}, 0.3333333333333333, 
                    NCache[{
                    Rational[-1, 2] 
                    Pi, -0.81579982222607}, {-1.5707963267948966`, \
-0.81579982222607}]], 
                   ArrowBox[{{0.04497759839786483, 0.2549699092731162}, {
                    0.047416559073801406`, 0.2572420204807955}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-0.1810121486029365, 0.5}, {-0.6843431516488209,
                     0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-0.1810121486029365, 0.5}, {
                    0.21159238000591382`, 0.5}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.005], Arrowheads[0.04], 
                   ArrowBox[{{-0.1810121486029365, 
                    0.5}, {-0.29173862303997056`, 0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-0.1810121486029365, 0.5}, 1.6, {
                    1.6242776086904531`, 1.8700067784392238`}], 
                   ArrowBox[{{-0.6373250384874309, 
                    2.033550959872303}, {-0.6526374770412408, 
                    2.0289112301162096`}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Dashing[{Small, Small}], 
                   Arrowheads[0.02], 
                   CircleBox[{-0.1810121486029365, 0.5}, 1.7, {
                    1.6242776086904531`, 1.5602007322562872`}], 
                   ArrowBox[{{-0.17999963794716228`, 
                    2.1999996984770824`}, {-0.16299997491936877`, 
                    2.199904574262683}}]}, 
                  {RGBColor[0, 0, 1], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-0.1810121486029365, 0.5}, 1.8, {
                    1.6242776086904531`, 1.4844634411777076`}], 
                   ArrowBox[{{-0.04374634648750189, 
                    2.294758507312225}, {-0.025805923772213774`, 
                    2.293296134991038}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.003], Arrowheads[0.025], 
                   CircleBox[{-0.1810121486029365, 0.5}, 1.5, {
                    1.6242776086904531`, 1.6459912621151256`}], 
                   ArrowBox[{{-0.27873529049857654`, 
                    1.9968133442544012`}, {-0.29369828835710027`, 
                    1.995761289079011}}]}, 
                  {RGBColor[0, 0, 1], ArrowBox[{{0, 2.8}, {-1.5, 2.8}}]}},
                 Frame->True,
                 FrameTicks->Automatic,
                 ImageSize->800,
                 PlotRange->{{-3, 3}, {-1, 3}}], 281->
                GraphicsBox[{InsetBox[
                   TagBox[GridBox[{
                    {"\<\"t\"\>", "14.`"},
                    {"\<\"driveIntent\"\>", 
                    RowBox[{"-", "0.5`"}]},
                    {"\<\"motorTorque\"\>", "22.424352381872126`"},
                    {"\<\"vw\"\>", 
                    RowBox[{"-", "0.28401282158746793`"}]},
                    {"\<\"controllerState\"\>", 
                    RowBox[{"\[LeftAssociation]", 
                    RowBox[{
                    RowBox[{"\<\"lastError\"\>", "\[Rule]", 
                    RowBox[{"-", "0.016873848180115433`"}]}], ",", 
                    
                    RowBox[{"\<\"lastInt\"\>", "\[Rule]", 
                    "0.22025149812697173`"}], ",", 
                    RowBox[{"\<\"initialized\"\>", "\[Rule]", "True"}]}], 
                    "\[RightAssociation]"}]}
                    },
                    AutoDelete->False,
                    
                    GridBoxItemSize->{
                    "Columns" -> {{Automatic}}, "Rows" -> {{Automatic}}}],
                    "Grid"], {-1, -0.5}], 
                  {GrayLevel[0], LineBox[{{-5, 0}, {5, 0}}]}, 
                  {GrayLevel[0.5], 
                   CircleBox[{-0.19517179438176593`, 0.5}, 0.5], 
                   DiskBox[{-0.19517179438176593`, 0.5}, 0.1]}, 
                  {GrayLevel[0.5], 
                   DiskBox[{-0.2668809960907532, 1.998284949664202}, 0.1]}, 
                  {GrayLevel[0.5], Thickness[0.01], 
                   LineBox[{{-0.19517179438176593`, 
                    0.5}, {-0.2668809960907532, 1.998284949664202}}]}, 
                  {RGBColor[0, 1, 0], 
                   DiskBox[{-0.24297792885442415`, 1.498856633109468}, 0.05]}, 
                  {RGBColor[0, 0, 1], Arrowheads[0.01], 
                   CircleBox[{-0.19517179438176593`, 0.5}, 0.3333333333333333,
                     NCache[{
                    Rational[-1, 2] 
                    Pi, -0.8980657553387328}, {-1.5707963267948966`, \
-0.8980657553387328}]], 
                   ArrowBox[{{0.009918770226990725, 0.23722829029826575`}, {
                    0.0125361890861658, 0.2392923002387607}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-0.19517179438176593`, 
                    0.5}, {-0.6436588420192084, 0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-0.19517179438176593`, 0.5}, {
                    0.15603970382924234`, 0.5}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.005], Arrowheads[0.04], 
                   ArrowBox[{{-0.19517179438176593`, 
                    0.5}, {-0.29244734380820014`, 0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-0.19517179438176593`, 0.5}, 1.6, {
                    1.618620689588264, 1.838379140506839}], 
                   ArrowBox[{{-0.6027619680288294, 
                    2.0472137054545367`}, {-0.6182134678768596, 
                    2.043060511608819}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Dashing[{Small, Small}], 
                   Arrowheads[0.02], 
                   CircleBox[{-0.19517179438176593`, 0.5}, 1.7, {
                    1.618620689588264, 1.5623112938519912`}], 
                   ArrowBox[{{-0.19774723739366576`, 
                    2.1999980491439666`}, {-0.18074741146274012`, 
                    2.19993880395072}}]}, 
                  {RGBColor[0, 0, 1], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-0.19517179438176593`, 0.5}, 1.8, {
                    1.618620689588264, 1.4940409541334188`}], 
                   ArrowBox[{{-0.07510134789356533, 
                    2.295990837359736}, {-0.05714774232257275, 
                    2.294700354112954}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.003], Arrowheads[0.025], 
                   CircleBox[{-0.19517179438176593`, 0.5}, 1.5, {
                    1.618620689588264, 1.6397167061878208`}], 
                   ArrowBox[{{-0.2835012351859412, 
                    1.9973970448372143`}, {-0.29847053963415066`, 
                    1.9964388959223438`}}]}, 
                  {RGBColor[0, 0, 1], ArrowBox[{{0, 2.8}, {-1.5, 2.8}}]}},
                 Frame->True,
                 FrameTicks->Automatic,
                 ImageSize->800,
                 PlotRange->{{-3, 3}, {-1, 3}}], 282->
                GraphicsBox[{InsetBox[
                   TagBox[GridBox[{
                    {"\<\"t\"\>", "14.05`"},
                    {"\<\"driveIntent\"\>", 
                    RowBox[{"-", "0.5`"}]},
                    {"\<\"motorTorque\"\>", "19.962334796037524`"},
                    {"\<\"vw\"\>", 
                    RowBox[{"-", "0.28551442611586736`"}]},
                    {"\<\"controllerState\"\>", 
                    RowBox[{"\[LeftAssociation]", 
                    RowBox[{
                    RowBox[{"\<\"lastError\"\>", "\[Rule]", 
                    RowBox[{"-", "0.021527862549835325`"}]}], ",", 
                    
                    RowBox[{"\<\"lastInt\"\>", "\[Rule]", 
                    "0.21928706970550849`"}], ",", 
                    RowBox[{"\<\"initialized\"\>", "\[Rule]", "True"}]}], 
                    "\[RightAssociation]"}]}
                    },
                    AutoDelete->False,
                    
                    GridBoxItemSize->{
                    "Columns" -> {{Automatic}}, "Rows" -> {{Automatic}}}],
                    "Grid"], {-1, -0.5}], 
                  {GrayLevel[0], LineBox[{{-5, 0}, {5, 0}}]}, 
                  {GrayLevel[0.5], CircleBox[{-0.2094115461342356, 0.5}, 0.5],
                    DiskBox[{-0.2094115461342356, 0.5}, 0.1]}, 
                  {GrayLevel[0.5], 
                   DiskBox[{-0.273486667140845, 1.9986308347515036`}, 0.1]}, 
                  {GrayLevel[0.5], Thickness[0.01], 
                   LineBox[{{-0.2094115461342356, 0.5}, {-0.273486667140845, 
                    1.9986308347515036`}}]}, 
                  {RGBColor[0, 1, 0], 
                   DiskBox[{-0.25212829347197524`, 1.499087223167669}, 0.05]}, 
                  {RGBColor[0, 0, 1], Arrowheads[0.01], 
                   CircleBox[{-0.2094115461342356, 0.5}, 0.3333333333333333, 
                    NCache[{
                    Rational[-1, 2] 
                    Pi, -0.9719262829137708}, {-1.5707963267948966`, \
-0.9719262829137708}]], 
                   ArrowBox[{{-0.02427096565865347, 
                    0.222810395988097}, {-0.02150837276845291, 
                    0.224675630300949}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-0.2094115461342356, 0.5}, {-0.6086582420549861,
                     0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-0.2094115461342356, 0.5}, {
                    0.10448281930727696`, 0.5}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.005], Arrowheads[0.04], 
                   ArrowBox[{{-0.2094115461342356, 
                    0.5}, {-0.29476387661347353`, 0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-0.2094115461342356, 0.5}, 1.6, {
                    1.613526075832261, 1.8098892900258443`}], 
                   ArrowBox[{{-0.5727623919820555, 
                    2.0581964455169555`}, {-0.5883259293482193, 
                    2.054485088443619}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Dashing[{Small, Small}], 
                   Arrowheads[0.02], 
                   CircleBox[{-0.2094115461342356, 0.5}, 1.7, {
                    1.613526075832261, 1.5641069533354184`}], 
                   ArrowBox[{{-0.21503960097229657`, 
                    2.1999906837976315`}, {-0.19803969606428734`, 
                    2.199961964582145}}]}, 
                  {RGBColor[0, 0, 1], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-0.2094115461342356, 0.5}, 1.8, {
                    1.613526075832261, 1.5026242158542746`}], 
                   ArrowBox[{{-0.104760792680514, 
                    2.2969552637173716`}, {-0.08679677202845511, 
                    2.2958189266100826`}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.003], Arrowheads[0.025], 
                   CircleBox[{-0.2094115461342356, 0.5}, 1.5, {
                    1.613526075832261, 1.6338896001687448`}], 
                   ArrowBox[{{-0.28901404536855296`, 
                    1.9978863248309766`}, {-0.30398867887859604`, 
                    1.9970154194135261`}}]}, 
                  {RGBColor[0, 0, 1], ArrowBox[{{0, 2.8}, {-1.5, 2.8}}]}},
                 Frame->True,
                 FrameTicks->Automatic,
                 ImageSize->800,
                 PlotRange->{{-3, 3}, {-1, 3}}], 283->
                GraphicsBox[{InsetBox[
                   TagBox[GridBox[{
                    {"\<\"t\"\>", "14.1`"},
                    {"\<\"driveIntent\"\>", 
                    RowBox[{"-", "0.5`"}]},
                    {"\<\"motorTorque\"\>", "17.7705933147014`"},
                    {"\<\"vw\"\>", 
                    RowBox[{"-", "0.28683138755152776`"}]},
                    {"\<\"controllerState\"\>", 
                    RowBox[{"\[LeftAssociation]", 
                    RowBox[{
                    RowBox[{"\<\"lastError\"\>", "\[Rule]", 
                    RowBox[{"-", "0.025694188425079492`"}]}], ",", 
                    
                    RowBox[{"\<\"lastInt\"\>", "\[Rule]", 
                    "0.2181024331383704`"}], ",", 
                    RowBox[{"\<\"initialized\"\>", "\[Rule]", "True"}]}], 
                    "\[RightAssociation]"}]}
                    },
                    AutoDelete->False,
                    
                    GridBoxItemSize->{
                    "Columns" -> {{Automatic}}, "Rows" -> {{Automatic}}}],
                    "Grid"], {-1, -0.5}], 
                  {GrayLevel[0], LineBox[{{-5, 0}, {5, 0}}]}, 
                  {GrayLevel[0.5], 
                   CircleBox[{-0.22372156898606388`, 0.5}, 0.5], 
                   DiskBox[{-0.22372156898606388`, 0.5}, 0.1]}, 
                  {GrayLevel[0.5], 
                   DiskBox[{-0.28097457567011197`, 1.9989069661675591`}, 
                    0.1]}, 
                  {GrayLevel[0.5], Thickness[0.01], 
                   LineBox[{{-0.22372156898606388`, 
                    0.5}, {-0.28097457567011197`, 1.9989069661675591`}}]}, 
                  {RGBColor[0, 1, 0], 
                   DiskBox[{-0.2618902401087626, 1.4992713107783726`}, 0.05]}, 
                  {RGBColor[0, 0, 1], Arrowheads[0.01], 
                   CircleBox[{-0.22372156898606388`, 0.5}, 0.3333333333333333,
                     NCache[{
                    Rational[-1, 2] 
                    Pi, -1.0376785273538545`}, {-1.5707963267948966`, \
-1.0376785273538545`}]], 
                   ArrowBox[{{-0.05719376744914845, 
                    0.2112447361061745}, {-0.05431458925653737, 
                    0.21292442400992928`}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-0.22372156898606388`, 
                    0.5}, {-0.5791334352800919, 0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-0.22372156898606388`, 0.5}, {
                    0.0568039650156692, 0.5}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.005], Arrowheads[0.04], 
                   ArrowBox[{{-0.22372156898606388`, 
                    0.5}, {-0.29860790127835884`, 0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-0.22372156898606388`, 0.5}, 1.6, {
                    1.6089742716538038`, 1.7844305835282164`}], 
                   ArrowBox[{{-0.5472892882195344, 
                    2.06694094689942}, {-0.5629422582818675, 
                    2.0636269772403137`}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Dashing[{Small, Small}], 
                   Arrowheads[0.02], 
                   CircleBox[{-0.22372156898606388`, 0.5}, 1.7, {
                    1.6089742716538038`, 1.5656068107472736`}], 
                   ArrowBox[{{-0.23189936016497475`, 
                    2.1999803303954533`}, {-0.21489943130350708`, 
                    2.199977108636087}}]}, 
                  {RGBColor[0, 0, 1], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-0.22372156898606388`, 0.5}, 1.8, {
                    1.6089742716538038`, 1.510248753238796}], 
                   ArrowBox[{{-0.132774677217937, 
                    2.2977009381089273`}, {-0.11480251475886666`, 
                    2.2967016000511196`}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.003], Arrowheads[0.025], 
                   CircleBox[{-0.22372156898606388`, 0.5}, 1.5, {
                    1.6089742716538038`, 1.6285072598611001`}], 
                   ArrowBox[{{-0.2952608201810452, 
                    1.9982930739806752`}, {-0.3102399242738364, 
                    1.997502779362462}}]}, 
                  {RGBColor[0, 0, 1], ArrowBox[{{0, 2.8}, {-1.5, 2.8}}]}},
                 Frame->True,
                 FrameTicks->Automatic,
                 ImageSize->800,
                 PlotRange->{{-3, 3}, {-1, 3}}], 284->
                GraphicsBox[{InsetBox[
                   TagBox[GridBox[{
                    {"\<\"t\"\>", "14.15`"},
                    {"\<\"driveIntent\"\>", 
                    RowBox[{"-", "0.5`"}]},
                    {"\<\"motorTorque\"\>", "15.837836221413806`"},
                    {"\<\"vw\"\>", 
                    RowBox[{"-", "0.28798724154259825`"}]},
                    {"\<\"controllerState\"\>", 
                    RowBox[{"\[LeftAssociation]", 
                    RowBox[{
                    RowBox[{"\<\"lastError\"\>", "\[Rule]", 
                    RowBox[{"-", "0.02938782057191547`"}]}], ",", 
                    
                    RowBox[{"\<\"lastInt\"\>", "\[Rule]", 
                    "0.2167216021235286`"}], ",", 
                    RowBox[{"\<\"initialized\"\>", "\[Rule]", "True"}]}], 
                    "\[RightAssociation]"}]}
                    },
                    AutoDelete->False,
                    
                    GridBoxItemSize->{
                    "Columns" -> {{Automatic}}, "Rows" -> {{Automatic}}}],
                    "Grid"], {-1, -0.5}], 
                  {GrayLevel[0], LineBox[{{-5, 0}, {5, 0}}]}, 
                  {GrayLevel[0.5], CircleBox[{-0.2380932363735195, 0.5}, 0.5],
                    DiskBox[{-0.2380932363735195, 0.5}, 0.1]}, 
                  {GrayLevel[0.5], 
                   DiskBox[{-0.2893035415712383, 1.9991255800104129`}, 0.1]}, 
                  {GrayLevel[0.5], Thickness[0.01], 
                   LineBox[{{-0.2380932363735195, 0.5}, {-0.2893035415712383, 
                    1.9991255800104129`}}]}, 
                  {RGBColor[0, 1, 0], 
                   DiskBox[{-0.2722334398386653, 1.499417053340275}, 0.05]}, 
                  {RGBColor[0, 0, 1], Arrowheads[0.01], 
                   CircleBox[{-0.2380932363735195, 0.5}, 0.3333333333333333, 
                    NCache[{
                    Rational[-1, 2] 
                    Pi, -1.0956612401524823`}, {-1.5707963267948966`, \
-1.0956612401524823`}]], 
                   ArrowBox[{{-0.08857872280419185, 
                    0.20207967282637723`}, {-0.08560704484897619, 
                    0.20358968893533513`}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-0.2380932363735195, 0.5}, {-0.5548499608017956,
                     0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-0.2380932363735195, 0.5}, {
                    0.012861155005114228`, 0.5}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.005], Arrowheads[0.04], 
                   ArrowBox[{{-0.2380932363735195, 
                    0.5}, {-0.30389556942316187`, 0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-0.2380932363735195, 0.5}, 1.6, {
                    1.6049431657798539`, 1.7618811468633566`}], 
                   ArrowBox[{{-0.5262480543808226, 
                    2.07383823846645}, {-0.5419717668395893, 
                    2.07087804705578}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Dashing[{Small, Small}], 
                   Arrowheads[0.02], 
                   CircleBox[{-0.2380932363735195, 0.5}, 1.7, {
                    1.6049431657798539`, 1.5668306630373166`}], 
                   ArrowBox[{{-0.24835154572903348`, 
                    2.1999690488621155`}, {-0.231351625655977, 
                    2.1999866325018362`}}]}, 
                  {RGBColor[0, 0, 1], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-0.2380932363735195, 0.5}, 1.8, {
                    1.6049431657798539`, 1.5169551867719995`}], 
                   ArrowBox[{{-0.15920446130354612`, 
                    2.2982704360490276`}, {-0.14122600105917993`, 
                    2.2973916486738655`}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.003], Arrowheads[0.025], 
                   CircleBox[{-0.2380932363735195, 0.5}, 1.5, {
                    1.6049431657798539`, 1.623563934351289}], 
                   ArrowBox[{{-0.3022250932781706, 
                    1.9986284078883467`}, {-0.3172079210207779, 
                    1.9979121692119217`}}]}, 
                  {RGBColor[0, 0, 1], ArrowBox[{{0, 2.8}, {-1.5, 2.8}}]}},
                 Frame->True,
                 FrameTicks->Automatic,
                 ImageSize->800,
                 PlotRange->{{-3, 3}, {-1, 3}}], 285->
                GraphicsBox[{InsetBox[
                   TagBox[GridBox[{
                    {"\<\"t\"\>", "14.200000000000001`"},
                    {"\<\"driveIntent\"\>", 
                    RowBox[{"-", "0.5`"}]},
                    {"\<\"motorTorque\"\>", "14.151630042164271`"},
                    {"\<\"vw\"\>", 
                    RowBox[{"-", "0.2890041968446845`"}]},
                    {"\<\"controllerState\"\>", 
                    RowBox[{"\[LeftAssociation]", 
                    RowBox[{
                    RowBox[{"\<\"lastError\"\>", "\[Rule]", 
                    RowBox[{"-", "0.03262633966909767`"}]}], ",", 
                    
                    RowBox[{"\<\"lastInt\"\>", "\[Rule]", 
                    "0.21516777294779735`"}], ",", 
                    RowBox[{"\<\"initialized\"\>", "\[Rule]", "True"}]}], 
                    "\[RightAssociation]"}]}
                    },
                    AutoDelete->False,
                    
                    GridBoxItemSize->{
                    "Columns" -> {{Automatic}}, "Rows" -> {{Automatic}}}],
                    "Grid"], {-1, -0.5}], 
                  {GrayLevel[0], LineBox[{{-5, 0}, {5, 0}}]}, 
                  {GrayLevel[0.5], 
                   CircleBox[{-0.25251906486269443`, 0.5}, 0.5], 
                   DiskBox[{-0.25251906486269443`, 0.5}, 0.1]}, 
                  {GrayLevel[0.5], 
                   DiskBox[{-0.2984301044322458, 1.9992972275188277`}, 0.1]}, 
                  {GrayLevel[0.5], Thickness[0.01], 
                   LineBox[{{-0.25251906486269443`, 
                    0.5}, {-0.2984301044322458, 1.9992972275188277`}}]}, 
                  {RGBColor[0, 1, 0], 
                   DiskBox[{-0.2831264245757287, 1.4995314850125516`}, 0.05]}, 
                  {RGBColor[0, 0, 1], Arrowheads[0.01], 
                   CircleBox[{-0.25251906486269443`, 0.5}, 0.3333333333333333,
                     NCache[{
                    Rational[-1, 2] 
                    Pi, -1.1462474255299684`}, {-1.5707963267948966`, \
-1.1462474255299684`}]], 
                   ArrowBox[{{-0.11826003767521587`, 
                    0.19490063138416203`}, {-0.11521580773411633`, 
                    0.1962584541209506}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-0.25251906486269443`, 
                    0.5}, {-0.5355516657059798, 0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-0.25251906486269443`, 
                    0.5}, {-0.027507787335733225`, 0.5}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.005], Arrowheads[0.04], 
                   ArrowBox[{{-0.25251906486269443`, 
                    0.5}, {-0.31054038817901863`, 0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-0.25251906486269443`, 0.5}, 1.6, {
                    1.6014084674061624`, 1.742106434967019}], 
                   ArrowBox[{{-0.509497375343344, 
                    2.0792283393931705`}, {-0.5252765467254185, 
                    2.0765796383589086`}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Dashing[{Small, Small}], 
                   Arrowheads[0.02], 
                   CircleBox[{-0.25251906486269443`, 0.5}, 1.7, {
                    1.6014084674061624`, 1.5677987676871352`}], 
                   ArrowBox[{{-0.2644231170947066, 
                    2.199958321118626}, {-0.247423222010839, 
                    2.199992362449205}}]}, 
                  {RGBColor[0, 0, 1], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-0.25251906486269443`, 0.5}, 1.8, {
                    1.6014084674061624`, 1.5227883005052498`}], 
                   ArrowBox[{{-0.18412108838475727`, 
                    2.2987000074536397`}, {-0.16613780796238137`, 
                    2.297926104837549}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.003], Arrowheads[0.025], 
                   CircleBox[{-0.25251906486269443`, 0.5}, 1.5, {
                    1.6014084674061624`, 1.6190511061750836`}], 
                   ArrowBox[{{-0.3098872391768197, 
                    1.9989025627357717`}, {-0.3248731466035203, 
                    1.998253946050349}}]}, 
                  {RGBColor[0, 0, 1], ArrowBox[{{0, 2.8}, {-1.5, 2.8}}]}},
                 Frame->True,
                 FrameTicks->Automatic,
                 ImageSize->800,
                 PlotRange->{{-3, 3}, {-1, 3}}], 286->
                GraphicsBox[{InsetBox[
                   TagBox[GridBox[{
                    {"\<\"t\"\>", "14.25`"},
                    {"\<\"driveIntent\"\>", 
                    RowBox[{"-", "0.5`"}]},
                    {"\<\"motorTorque\"\>", "12.698634242548508`"},
                    {"\<\"vw\"\>", 
                    RowBox[{"-", "0.2899030969460135`"}]},
                    {"\<\"controllerState\"\>", 
                    RowBox[{"\[LeftAssociation]", 
                    RowBox[{
                    RowBox[{"\<\"lastError\"\>", "\[Rule]", 
                    RowBox[{"-", "0.03542950321333452`"}]}], ",", 
                    
                    RowBox[{"\<\"lastInt\"\>", "\[Rule]", 
                    "0.21346320568138125`"}], ",", 
                    RowBox[{"\<\"initialized\"\>", "\[Rule]", "True"}]}], 
                    "\[RightAssociation]"}]}
                    },
                    AutoDelete->False,
                    
                    GridBoxItemSize->{
                    "Columns" -> {{Automatic}}, "Rows" -> {{Automatic}}}],
                    "Grid"], {-1, -0.5}], 
                  {GrayLevel[0], LineBox[{{-5, 0}, {5, 0}}]}, 
                  {GrayLevel[0.5], CircleBox[{-0.2669926467792796, 0.5}, 0.5],
                    DiskBox[{-0.2669926467792796, 0.5}, 0.1]}, 
                  {GrayLevel[0.5], 
                   DiskBox[{-0.30830912019398127`, 1.9994308750404508`}, 
                    0.1]}, 
                  {GrayLevel[0.5], Thickness[0.01], 
                   LineBox[{{-0.2669926467792796, 
                    0.5}, {-0.30830912019398127`, 1.9994308750404508`}}]}, 
                  {RGBColor[0, 1, 0], 
                   DiskBox[{-0.29453696238908067`, 1.4996205833603007`}, 
                    0.05]}, 
                  {RGBColor[0, 0, 1], Arrowheads[0.01], 
                   CircleBox[{-0.2669926467792796, 0.5}, 0.3333333333333333, 
                    NCache[{
                    Rational[-1, 2] 
                    Pi, -1.1898372995184414`}, {-1.5707963267948966`, \
-1.1898372995184414`}]], 
                   ArrowBox[{{-0.14615618248326098`, 
                    0.18933996071662496`}, {-0.143055675639708, 
                    0.1905638380927981}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-0.2669926467792796, 0.5}, {-0.5209653316302497,
                     0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-0.2669926467792796, 
                    0.5}, {-0.06448141506814672, 0.5}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.005], Arrowheads[0.04], 
                   ArrowBox[{{-0.2669926467792796, 0.5}, {-0.3184540999191169,
                     0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-0.2669926467792796, 0.5}, 1.6, {
                    1.5983441265240337`, 1.7249616871528726`}], 
                   ArrowBox[{{-0.49685904457601615`, 
                    2.083401856498834}, {-0.5126813060179021, 
                    2.0810240613986646`}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Dashing[{Small, Small}], 
                   Arrowheads[0.02], 
                   CircleBox[{-0.2669926467792796, 0.5}, 1.7, {
                    1.5983441265240337`, 1.568531611320138}], 
                   ArrowBox[{{-0.28014249933518803`, 
                    2.1999491408209124`}, {-0.26314263376327035`, 
                    2.199995640406109}}]}, 
                  {RGBColor[0, 0, 1], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-0.2669926467792796, 0.5}, 1.8, {
                    1.5983441265240337`, 1.5277961585098754`}], 
                   ArrowBox[{{-0.2076031245441477, 
                    2.2990199789465047`}, {-0.18961619404121283`, 
                    2.2983361433729996`}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.003], Arrowheads[0.025], 
                   CircleBox[{-0.2669926467792796, 0.5}, 1.5, {
                    1.5983441265240337`, 1.61495777901339}], 
                   ArrowBox[{{-0.31822485904373016`, 
                    1.999124831502197}, {-0.33321329591659626`, 
                    1.9985375623012698`}}]}, 
                  {RGBColor[0, 0, 1], ArrowBox[{{0, 2.8}, {-1.5, 2.8}}]}},
                 Frame->True,
                 FrameTicks->Automatic,
                 ImageSize->800,
                 PlotRange->{{-3, 3}, {-1, 3}}], 287->
                GraphicsBox[{InsetBox[
                   TagBox[GridBox[{
                    {"\<\"t\"\>", "14.3`"},
                    {"\<\"driveIntent\"\>", 
                    RowBox[{"-", "0.5`"}]},
                    {"\<\"motorTorque\"\>", "11.46482247341065`"},
                    {"\<\"vw\"\>", 
                    RowBox[{"-", "0.2907033932348241`"}]},
                    {"\<\"controllerState\"\>", 
                    RowBox[{"\[LeftAssociation]", 
                    RowBox[{
                    RowBox[{"\<\"lastError\"\>", "\[Rule]", 
                    RowBox[{"-", "0.037818852059552516`"}]}], ",", 
                    
                    RowBox[{"\<\"lastInt\"\>", "\[Rule]", 
                    "0.21162912544205187`"}], ",", 
                    RowBox[{"\<\"initialized\"\>", "\[Rule]", "True"}]}], 
                    "\[RightAssociation]"}]}
                    },
                    AutoDelete->False,
                    
                    GridBoxItemSize->{
                    "Columns" -> {{Automatic}}, "Rows" -> {{Automatic}}}],
                    "Grid"], {-1, -0.5}], 
                  {GrayLevel[0], LineBox[{{-5, 0}, {5, 0}}]}, 
                  {GrayLevel[0.5], CircleBox[{-0.2815085812120581, 0.5}, 0.5],
                    DiskBox[{-0.2815085812120581, 0.5}, 0.1]}, 
                  {GrayLevel[0.5], 
                   DiskBox[{-0.31889432191111694`, 1.9995340297547044`}, 
                    0.1]}, 
                  {GrayLevel[0.5], Thickness[0.01], 
                   LineBox[{{-0.2815085812120581, 
                    0.5}, {-0.31889432191111694`, 1.9995340297547044`}}]}, 
                  {RGBColor[0, 1, 0], 
                   DiskBox[{-0.306432408344764, 1.4996893531698028`}, 0.05]}, 
                  {RGBColor[0, 0, 1], Arrowheads[0.01], 
                   CircleBox[{-0.2815085812120581, 0.5}, 0.3333333333333333, 
                    NCache[{
                    Rational[-1, 2] 
                    Pi, -1.2268516525925772`}, {-1.5707963267948966`, \
-1.2268516525925772`}]], 
                   ArrowBox[{{-0.1722511390725932, 
                    0.1850810859153541}, {-0.1691074652445532, 
                    0.18618938794175455`}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-0.2815085812120581, 0.5}, {-0.510805030680271, 
                    0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-0.2815085812120581, 
                    0.5}, {-0.09825109024763279, 0.5}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.005], Arrowheads[0.04], 
                   ArrowBox[{{-0.2815085812120581, 
                    0.5}, {-0.32754753971584577`, 0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-0.2815085812120581, 0.5}, 1.6, {
                    1.5957227350843017`, 1.7102942582175595`}], 
                   ArrowBox[{{-0.48812665198432503`, 
                    2.086602966349914}, {-0.5039820863978706, 
                    2.084457490591129}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Dashing[{Small, Small}], 
                   Arrowheads[0.02], 
                   CircleBox[{-0.2815085812120581, 0.5}, 1.7, {
                    1.5957227350843017`, 1.5690496898143937`}], 
                   ArrowBox[{{-0.29553913905501145`, 
                    2.1999421000277084`}, {-0.278539299854955, 
                    2.199997406871029}}]}, 
                  {RGBColor[0, 0, 1], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-0.2815085812120581, 0.5}, 1.8, {
                    1.5957227350843017`, 1.532029276898687}], 
                   ArrowBox[{{-0.22973503289598962`, 
                    2.299255262516901}, {-0.21174536880104183`, 
                    2.298647573649184}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.003], Arrowheads[0.025], 
                   CircleBox[{-0.2815085812120581, 0.5}, 1.5, {
                    1.5957227350843017`, 1.6112707586311437`}], 
                   ArrowBox[{{-0.32721315396234957`, 
                    1.99930353565571}, {-0.3422036542266395, 
                    1.998771532993523}}]}, 
                  {RGBColor[0, 0, 1], ArrowBox[{{0, 2.8}, {-1.5, 2.8}}]}},
                 Frame->True,
                 FrameTicks->Automatic,
                 ImageSize->800,
                 PlotRange->{{-3, 3}, {-1, 3}}], 288->
                GraphicsBox[{InsetBox[
                   TagBox[GridBox[{
                    {"\<\"t\"\>", "14.35`"},
                    {"\<\"driveIntent\"\>", 
                    RowBox[{"-", "0.5`"}]},
                    {"\<\"motorTorque\"\>", "10.435688627151464`"},
                    {"\<\"vw\"\>", 
                    RowBox[{"-", "0.2914231299786805`"}]},
                    {"\<\"controllerState\"\>", 
                    RowBox[{"\[LeftAssociation]", 
                    RowBox[{
                    RowBox[{"\<\"lastError\"\>", "\[Rule]", 
                    RowBox[{"-", "0.03981733571368931`"}]}], ",", 
                    
                    RowBox[{"\<\"lastInt\"\>", "\[Rule]", 
                    "0.20968564286640995`"}], ",", 
                    RowBox[{"\<\"initialized\"\>", "\[Rule]", "True"}]}], 
                    "\[RightAssociation]"}]}
                    },
                    AutoDelete->False,
                    
                    GridBoxItemSize->{
                    "Columns" -> {{Automatic}}, "Rows" -> {{Automatic}}}],
                    "Grid"], {-1, -0.5}], 
                  {GrayLevel[0], LineBox[{{-5, 0}, {5, 0}}]}, 
                  {GrayLevel[0.5], CircleBox[{-0.2960624039755831, 0.5}, 0.5],
                    DiskBox[{-0.2960624039755831, 0.5}, 0.1]}, 
                  {GrayLevel[0.5], 
                   DiskBox[{-0.3301388407244483, 1.9996128821994363`}, 0.1]}, 
                  {GrayLevel[0.5], Thickness[0.01], 
                   LineBox[{{-0.2960624039755831, 0.5}, {-0.3301388407244483, 
                    1.9996128821994363`}}]}, 
                  {RGBColor[0, 1, 0], 
                   DiskBox[{-0.3187800284748266, 1.499741921466291}, 0.05]}, 
                  {RGBColor[0, 0, 1], Arrowheads[0.01], 
                   CircleBox[{-0.2960624039755831, 0.5}, 0.3333333333333333, 
                    NCache[{
                    Rational[-1, 2] 
                    Pi, -1.2577256679803526`}, {-1.5707963267948966`, \
-1.2577256679803526`}]], 
                   ArrowBox[{{-0.19657829689502715`, 
                    0.1818584850267171}, {-0.19340190893251746`, 
                    0.18286921646011123`}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-0.2960624039755831, 0.5}, {-0.5047761765186124,
                     0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-0.2960624039755831, 
                    0.5}, {-0.12901768153979068`, 0.5}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.005], Arrowheads[0.04], 
                   ArrowBox[{{-0.2960624039755831, 
                    0.5}, {-0.33773145408281996`, 0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-0.2960624039755831, 0.5}, 1.6, {
                    1.5935159058063153`, 1.6979458084448338`}], 
                   ArrowBox[{{-0.4830731328221568, 
                    2.0890333499635156`}, {-0.49895385102570305`, 
                    2.087083822837946}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Dashing[{Small, Small}], 
                   Arrowheads[0.02], 
                   CircleBox[{-0.2960624039755831, 0.5}, 1.7, {
                    1.5935159058063153`, 1.5693733049169576`}], 
                   ArrowBox[{{-0.31064308801130014`, 
                    2.1999374705126806`}, {-0.29364326759954196`, 
                    2.199998278757715}}]}, 
                  {RGBColor[0, 0, 1], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-0.2960624039755831, 0.5}, 1.8, {
                    1.5935159058063153`, 1.535539857877696}], 
                   ArrowBox[{{-0.2506055930192641, 
                    2.299425930217102}, {-0.2326139064415228, 
                    2.2988813991368833`}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.003], Arrowheads[0.025], 
                   CircleBox[{-0.2960624039755831, 0.5}, 1.5, {
                    1.5935159058063153`, 1.6079749300984532`}], 
                   ArrowBox[{{-0.33682529006754913`, 
                    1.9994460267436949`}, {-0.35181746230124444`, 
                    1.9989634329999855`}}]}, 
                  {RGBColor[0, 0, 1], ArrowBox[{{0, 2.8}, {-1.5, 2.8}}]}},
                 Frame->True,
                 FrameTicks->Automatic,
                 ImageSize->800,
                 PlotRange->{{-3, 3}, {-1, 3}}], 289->
                GraphicsBox[{InsetBox[
                   TagBox[GridBox[{
                    {"\<\"t\"\>", "14.4`"},
                    {"\<\"driveIntent\"\>", 
                    RowBox[{"-", "0.5`"}]},
                    {"\<\"motorTorque\"\>", "9.596436487391184`"},
                    {"\<\"vw\"\>", 
                    RowBox[{"-", "0.29207894104163`"}]},
                    {"\<\"controllerState\"\>", 
                    RowBox[{"\[LeftAssociation]", 
                    RowBox[{
                    RowBox[{"\<\"lastError\"\>", "\[Rule]", 
                    RowBox[{"-", "0.04144895911029607`"}]}], ",", 
                    
                    RowBox[{"\<\"lastInt\"\>", "\[Rule]", 
                    "0.2076516927790571`"}], ",", 
                    RowBox[{"\<\"initialized\"\>", "\[Rule]", "True"}]}], 
                    "\[RightAssociation]"}]}
                    },
                    AutoDelete->False,
                    GridBoxItemSize->{
                    "Columns" -> {{Automatic}}, "Rows" -> {{Automatic}}}],
                    "Grid"], {-1, -0.5}], 
                  {GrayLevel[0], LineBox[{{-5, 0}, {5, 0}}]}, 
                  {GrayLevel[0.5], 
                   CircleBox[{-0.31065051712212227`, 0.5}, 0.5], 
                   DiskBox[{-0.31065051712212227`, 0.5}, 0.1]}, 
                  {GrayLevel[0.5], 
                   DiskBox[{-0.34199568410086595`, 1.9996724577410476`}, 
                    0.1]}, 
                  {GrayLevel[0.5], Thickness[0.01], 
                   LineBox[{{-0.31065051712212227`, 
                    0.5}, {-0.34199568410086595`, 1.9996724577410476`}}]}, 
                  {RGBColor[0, 1, 0], 
                   DiskBox[{-0.3315472951079514, 1.499781638494032}, 0.05]}, 
                  {RGBColor[0, 0, 1], Arrowheads[0.01], 
                   CircleBox[{-0.31065051712212227`, 0.5}, 0.3333333333333333,
                     NCache[{
                    Rational[-1, 2] 
                    Pi, -1.282903232173161}, {-1.5707963267948966`, \
-1.282903232173161}]], 
                   ArrowBox[{{-0.21920712251801727`, 
                    0.17945481333454866`}, {-0.2160062962069223, 
                    0.1803852591658734}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-0.31065051712212227`, 
                    0.5}, {-0.5025792468699459, 0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-0.31065051712212227`, 
                    0.5}, {-0.15698853743584015`, 0.5}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.005], Arrowheads[0.04], 
                   ArrowBox[{{-0.31065051712212227`, 
                    0.5}, {-0.34891726718366384`, 0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-0.31065051712212227`, 0.5}, 1.6, {
                    1.5916946259308944`, 1.6877543389784246`}], 
                   ArrowBox[{{-0.48145722949579095`, 
                    2.090856708509003}, {-0.4973569911749725, 
                    2.0890691276803386`}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Dashing[{Small, Small}], 
                   Arrowheads[0.02], 
                   CircleBox[{-0.31065051712212227`, 0.5}, 1.7, {
                    1.5916946259308944`, 1.5695223808737688`}], 
                   ArrowBox[{{-0.32548462079981866`, 
                    2.1999352779938652`}, {-0.3084848096420067, 
                    2.199998620502708}}]}, 
                  {RGBColor[0, 0, 1], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-0.31065051712212227`, 0.5}, 1.8, {
                    1.5916946259308944`, 1.5383810898898322`}], 
                   ArrowBox[{{-0.2703064693207509, 
                    2.2995478203723847`}, {-0.25231330822574427`, 
                    2.2990544099771357`}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.003], Arrowheads[0.025], 
                   CircleBox[{-0.31065051712212227`, 0.5}, 1.5, {
                    1.5916946259308944`, 1.6050535322009098`}], 
                   ArrowBox[{{-0.34703275702835, 
                    1.999558712628287}, {-0.36202627513259433`, 
                    1.9991199189820839`}}]}, 
                  {RGBColor[0, 0, 1], ArrowBox[{{0, 2.8}, {-1.5, 2.8}}]}},
                 Frame->True,
                 FrameTicks->Automatic,
                 ImageSize->800,
                 PlotRange->{{-3, 3}, {-1, 3}}], 290->
                GraphicsBox[{InsetBox[
                   TagBox[GridBox[{
                    {"\<\"t\"\>", "14.450000000000001`"},
                    {"\<\"driveIntent\"\>", 
                    RowBox[{"-", "0.5`"}]},
                    {"\<\"motorTorque\"\>", "8.932152237448022`"},
                    {"\<\"vw\"\>", 
                    RowBox[{"-", "0.2926860579569543`"}]},
                    {"\<\"controllerState\"\>", 
                    RowBox[{"\[LeftAssociation]", 
                    RowBox[{
                    RowBox[{"\<\"lastError\"\>", "\[Rule]", 
                    RowBox[{"-", "0.04273845311291488`"}]}], ",", 
                    
                    RowBox[{"\<\"lastInt\"\>", "\[Rule]", 
                    "0.2055449899261573`"}], ",", 
                    RowBox[{"\<\"initialized\"\>", "\[Rule]", "True"}]}], 
                    "\[RightAssociation]"}]}
                    },
                    AutoDelete->False,
                    
                    GridBoxItemSize->{
                    "Columns" -> {{Automatic}}, "Rows" -> {{Automatic}}}],
                    "Grid"], {-1, -0.5}], 
                  {GrayLevel[0], LineBox[{{-5, 0}, {5, 0}}]}, 
                  {GrayLevel[0.5], 
                   CircleBox[{-0.32527011858119925`, 0.5}, 0.5], 
                   DiskBox[{-0.32527011858119925`, 0.5}, 0.1]}, 
                  {GrayLevel[0.5], 
                   DiskBox[{-0.35441816938079695`, 1.9997167703051748`}, 
                    0.1]}, 
                  {GrayLevel[0.5], Thickness[0.01], 
                   LineBox[{{-0.32527011858119925`, 
                    0.5}, {-0.35441816938079695`, 1.9997167703051748`}}]}, 
                  {RGBColor[0, 1, 0], 
                   DiskBox[{-0.3447021524475977, 1.4998111802034497`}, 0.05]}, 
                  {RGBColor[0, 0, 1], Arrowheads[0.01], 
                   CircleBox[{-0.32527011858119925`, 0.5}, 0.3333333333333333,
                     NCache[{
                    Rational[-1, 2] 
                    Pi, -1.3028317596714558`}, {-1.5707963267948966`, \
-1.3028317596714558`}]], 
                   ArrowBox[{{-0.2402324523222662, 
                    0.17769625129647437`}, {-0.23701372040013452`, 
                    0.17856272883933255`}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-0.32527011858119925`, 
                    0.5}, {-0.5039131633301597, 0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-0.32527011858119925`, 
                    0.5}, {-0.18237473896636777`, 0.5}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.005], Arrowheads[0.04], 
                   ArrowBox[{{-0.32527011858119925`, 
                    0.5}, {-0.3610177837153282, 0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-0.32527011858119925`, 0.5}, 1.6, {
                    1.5902295838045708`, 1.6795560626714066`}], 
                   ArrowBox[{{-0.4830289541166788, 
                    2.092203551626013}, {-0.49894283639096304`, 
                    2.0905463800495006`}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Dashing[{Small, Small}], 
                   Arrowheads[0.02], 
                   CircleBox[{-0.32527011858119925`, 0.5}, 1.7, {
                    1.5902295838045708`, 1.5695163032248498`}], 
                   ArrowBox[{{-0.34009389064881707`, 
                    2.1999353681189433`}, {-0.3230940791063455, 
                    2.199998607308901}}]}, 
                  {RGBColor[0, 0, 1], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-0.32527011858119925`, 0.5}, 1.8, {
                    1.5902295838045708`, 1.5406065158187483`}], 
                   ArrowBox[{{-0.2889309277563221, 
                    2.299633146841376}, {-0.2709367131696664, 
                    2.299179780082131}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.003], Arrowheads[0.025], 
                   CircleBox[{-0.32527011858119925`, 0.5}, 1.5, {
                    1.5902295838045708`, 1.6024884283558585`}], 
                   ArrowBox[{{-0.357805719192885, 
                    1.9996471033856056`}, {-0.3728003135203327, 
                    1.9992467710717432`}}]}, 
                  {RGBColor[0, 0, 1], ArrowBox[{{0, 2.8}, {-1.5, 2.8}}]}},
                 Frame->True,
                 FrameTicks->Automatic,
                 ImageSize->800,
                 PlotRange->{{-3, 3}, {-1, 3}}], 291->
                GraphicsBox[{InsetBox[
                   TagBox[GridBox[{
                    {"\<\"t\"\>", "14.5`"},
                    {"\<\"driveIntent\"\>", 
                    RowBox[{"-", "0.5`"}]},
                    {"\<\"motorTorque\"\>", "8.42795952111807`"},
                    {"\<\"vw\"\>", 
                    RowBox[{"-", "0.29325832871401714`"}]},
                    {"\<\"controllerState\"\>", 
                    RowBox[{"\[LeftAssociation]", 
                    RowBox[{
                    RowBox[{"\<\"lastError\"\>", "\[Rule]", 
                    RowBox[{"-", "0.04371097042605679`"}]}], ",", 
                    
                    RowBox[{"\<\"lastInt\"\>", "\[Rule]", 
                    "0.2033820005418454`"}], ",", 
                    RowBox[{"\<\"initialized\"\>", "\[Rule]", "True"}]}], 
                    "\[RightAssociation]"}]}
                    },
                    AutoDelete->False,
                    
                    GridBoxItemSize->{
                    "Columns" -> {{Automatic}}, "Rows" -> {{Automatic}}}],
                    "Grid"], {-1, -0.5}], 
                  {GrayLevel[0], LineBox[{{-5, 0}, {5, 0}}]}, 
                  {GrayLevel[0.5], 
                   CircleBox[{-0.33991913247912253`, 0.5}, 0.5], 
                   DiskBox[{-0.33991913247912253`, 0.5}, 0.1]}, 
                  {GrayLevel[0.5], 
                   DiskBox[{-0.36736031162594, 1.9997489728908076`}, 0.1]}, 
                  {GrayLevel[0.5], Thickness[0.01], 
                   LineBox[{{-0.33991913247912253`, 0.5}, {-0.36736031162594, 
                    1.9997489728908076`}}]}, 
                  {RGBColor[0, 1, 0], 
                   DiskBox[{-0.3582132519103342, 1.4998326485938716`}, 0.05]}, 
                  {RGBColor[0, 0, 1], Arrowheads[0.01], 
                   CircleBox[{-0.33991913247912253`, 0.5}, 0.3333333333333333,
                     NCache[{
                    Rational[-1, 2] 
                    Pi, -1.3179575411613544`}, {-1.5707963267948966`, \
-1.3179575411613544`}]], 
                   ArrowBox[{{-0.2597661040707805, 
                    0.17644690830084375`}, {-0.25653463469705806`, 
                    0.17726460274592748`}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-0.33991913247912253`, 
                    0.5}, {-0.508478322901484, 0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-0.33991913247912253`, 
                    0.5}, {-0.20538863037610597`, 0.5}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.005], Arrowheads[0.04], 
                   ArrowBox[{{-0.33991913247912253`, 
                    0.5}, {-0.37394782079846733`, 0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-0.33991913247912253`, 0.5}, 1.6, {
                    1.5890914668099638`, 1.673187104147507}], 
                   ArrowBox[{{-0.48753415856476445`, 
                    2.0931760116426985`}, {-0.5034582724633856, 
                    2.0916202278474625`}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Dashing[{Small, Small}], 
                   Arrowheads[0.02], 
                   CircleBox[{-0.33991913247912253`, 0.5}, 1.7, {
                    1.5890914668099638`, 1.5693737807290016`}], 
                   ArrowBox[{{-0.35450062536556043`, 
                    2.1999374635748232`}, {-0.33750080498273743`, 
                    2.199998279908577}}]}, 
                  {RGBColor[0, 0, 1], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-0.33991913247912253`, 0.5}, 1.8, {
                    1.5890914668099638`, 1.542269469470419}], 
                   ArrowBox[{{-0.30657269703456475`, 
                    2.2996910888380655`}, {-0.28857775340107705`, 
                    2.2992676462367583`}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.003], Arrowheads[0.025], 
                   CircleBox[{-0.33991913247912253`, 0.5}, 1.5, {
                    1.5890914668099638`, 1.6002603722707989`}], 
                   ArrowBox[{{-0.369113357244032, 
                    1.9997158721705843`}, {-0.3841088063152682, 
                    1.9993489496198857`}}]}, 
                  {RGBColor[0, 0, 1], ArrowBox[{{0, 2.8}, {-1.5, 2.8}}]}},
                 Frame->True,
                 FrameTicks->Automatic,
                 ImageSize->800,
                 PlotRange->{{-3, 3}, {-1, 3}}], 292->
                GraphicsBox[{InsetBox[
                   TagBox[GridBox[{
                    {"\<\"t\"\>", "14.55`"},
                    {"\<\"driveIntent\"\>", 
                    RowBox[{"-", "0.5`"}]},
                    {"\<\"motorTorque\"\>", "8.069157113301152`"},
                    {"\<\"vw\"\>", 
                    RowBox[{"-", "0.29380824641403874`"}]},
                    {"\<\"controllerState\"\>", 
                    RowBox[{"\[LeftAssociation]", 
                    RowBox[{
                    RowBox[{"\<\"lastError\"\>", "\[Rule]", 
                    RowBox[{"-", "0.044391808047420156`"}]}], ",", 
                    
                    RowBox[{"\<\"lastInt\"\>", "\[Rule]", 
                    "0.20117792844581356`"}], ",", 
                    RowBox[{"\<\"initialized\"\>", "\[Rule]", "True"}]}], 
                    "\[RightAssociation]"}]}
                    },
                    AutoDelete->False,
                    
                    GridBoxItemSize->{
                    "Columns" -> {{Automatic}}, "Rows" -> {{Automatic}}}],
                    "Grid"], {-1, -0.5}], 
                  {GrayLevel[0], LineBox[{{-5, 0}, {5, 0}}]}, 
                  {GrayLevel[0.5], CircleBox[{-0.3545961406533825, 0.5}, 0.5],
                    DiskBox[{-0.3545961406533825, 0.5}, 0.1]}, 
                  {GrayLevel[0.5], 
                   DiskBox[{-0.3807771656365012, 1.9997715005729484`}, 0.1]}, 
                  {GrayLevel[0.5], Thickness[0.01], 
                   LineBox[{{-0.3545961406533825, 0.5}, {-0.3807771656365012, 
                    1.9997715005729484`}}]}, 
                  {RGBColor[0, 1, 0], 
                   DiskBox[{-0.372050157308795, 1.4998476670486323`}, 0.05]}, 
                  {RGBColor[0, 0, 1], Arrowheads[0.01], 
                   CircleBox[{-0.3545961406533825, 0.5}, 0.3333333333333333, 
                    NCache[{
                    Rational[-1, 2] 
                    Pi, -1.328721613395862}, {-1.5707963267948966`, \
-1.328721613395862}]], 
                   ArrowBox[{{-0.27793043727179534`, 
                    0.17560289607316532`}, {-0.2746903535516658, 
                    0.17638576004945916`}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-0.3545961406533825, 0.5}, {-0.5159792829194055,
                     0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-0.3545961406533825, 
                    0.5}, {-0.22624162192965308`, 0.5}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.005], Arrowheads[0.04], 
                   ArrowBox[{{-0.3545961406533825, 
                    0.5}, {-0.38762476419567615`, 0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-0.3545961406533825, 0.5}, 1.6, {
                    1.588251229778273, 1.6684850262316169`}], 
                   ArrowBox[{{-0.49471832475624167`, 
                    2.0938524942798953`}, {-0.5106495780074654, 
                    2.092371603831837}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Dashing[{Small, Small}], 
                   Arrowheads[0.02], 
                   CircleBox[{-0.3545961406533825, 0.5}, 1.7, {
                    1.588251229778273, 1.5691127304084502`}], 
                   ArrowBox[{{-0.36873386382823276`, 
                    2.1999412121551236`}, {-0.3517340281485322, 
                    2.1999975906782954`}}]}, 
                  {RGBColor[0, 0, 1], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-0.3545961406533825, 0.5}, 1.8, {
                    1.588251229778273, 1.543422579148822}], 
                   ArrowBox[{{-0.3233249681313451, 
                    2.2997283444367644`}, {-0.3053295481857982, 
                    2.2993256522560426`}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.003], Arrowheads[0.025], 
                   CircleBox[{-0.3545961406533825, 0.5}, 1.5, {
                    1.588251229778273, 1.5983492659540333`}], 
                   ArrowBox[{{-0.38092419737288213`, 
                    1.9997689266781649`}, {-0.39592032028755986`, 
                    1.9994306626775251`}}]}, 
                  {RGBColor[0, 0, 1], ArrowBox[{{0, 2.8}, {-1.5, 2.8}}]}},
                 Frame->True,
                 FrameTicks->Automatic,
                 ImageSize->800,
                 PlotRange->{{-3, 3}, {-1, 3}}], 293->
                GraphicsBox[{InsetBox[
                   TagBox[GridBox[{
                    {"\<\"t\"\>", "14.6`"},
                    {"\<\"driveIntent\"\>", 
                    RowBox[{"-", "0.5`"}]},
                    {"\<\"motorTorque\"\>", "7.8413395556824135`"},
                    {"\<\"vw\"\>", 
                    RowBox[{"-", "0.2943469868025672`"}]},
                    {"\<\"controllerState\"\>", 
                    RowBox[{"\[LeftAssociation]", 
                    RowBox[{
                    RowBox[{"\<\"lastError\"\>", "\[Rule]", 
                    RowBox[{"-", "0.04480615685103431`"}]}], ",", 
                    
                    RowBox[{"\<\"lastInt\"\>", "\[Rule]", 
                    "0.19894671432784167`"}], ",", 
                    RowBox[{"\<\"initialized\"\>", "\[Rule]", "True"}]}], 
                    "\[RightAssociation]"}]}
                    },
                    AutoDelete->False,
                    
                    GridBoxItemSize->{
                    "Columns" -> {{Automatic}}, "Rows" -> {{Automatic}}}],
                    "Grid"], {-1, -0.5}], 
                  {GrayLevel[0], LineBox[{{-5, 0}, {5, 0}}]}, 
                  {GrayLevel[0.5], 
                   CircleBox[{-0.36930031583055983`, 0.5}, 0.5], 
                   DiskBox[{-0.36930031583055983`, 0.5}, 0.1]}, 
                  {GrayLevel[0.5], 
                   DiskBox[{-0.394625122777571, 1.99978620281462}, 0.1]}, 
                  {GrayLevel[0.5], Thickness[0.01], 
                   LineBox[{{-0.36930031583055983`, 0.5}, {-0.394625122777571,
                     1.99978620281462}}]}, 
                  {RGBColor[0, 1, 0], 
                   DiskBox[{-0.38618352046190063`, 1.4998574685430799`}, 
                    0.05]}, 
                  {RGBColor[0, 0, 1], Arrowheads[0.01], 
                   CircleBox[{-0.36930031583055983`, 0.5}, 0.3333333333333333,
                     NCache[{
                    Rational[-1, 2] 
                    Pi, -1.335556140124424}, {-1.5707963267948966`, \
-1.335556140124424}]], 
                   ArrowBox[{{-0.2948534864155582, 
                    0.17508650274086068`}, {-0.2916081279053969, 
                    0.1758472041667506}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-0.36930031583055983`, 
                    0.5}, {-0.5261271069442082, 0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-0.36930031583055983`, 
                    0.5}, {-0.245142255758954, 0.5}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.005], Arrowheads[0.04], 
                   ArrowBox[{{-0.36930031583055983`, 
                    0.5}, {-0.40196904687260226`, 0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-0.36930031583055983`, 0.5}, 1.6, {
                    1.5876803336012228`, 1.6652901829908813`}], 
                   ArrowBox[{{-0.5043296846132875, 
                    2.0942920276932133`}, {-0.5202655877640332, 
                    2.0928620425730697`}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Dashing[{Small, Small}], 
                   Arrowheads[0.02], 
                   CircleBox[{-0.36930031583055983`, 0.5}, 1.7, {
                    1.5876803336012228`, 1.568750184947742}], 
                   ArrowBox[{{-0.3828217321198481, 
                    2.1999462260029086`}, {-0.36582187711759284`, 
                    2.199996441309252}}]}, 
                  {RGBColor[0, 0, 1], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-0.36930031583055983`, 0.5}, 1.8, {
                    1.5876803336012228`, 1.5441173360696538`}], 
                   ArrowBox[{{-0.33927952447794374`, 
                    2.2997496359456675`}, {-0.32128382910231895`, 
                    2.2993594463036775`}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.003], Arrowheads[0.025], 
                   CircleBox[{-0.36930031583055983`, 0.5}, 1.5, {
                    1.5876803336012228`, 1.5967344074383643`}], 
                   ArrowBox[{{-0.39320642465118444`, 
                    1.9998094872219792`}, {-0.40820307426092606`, 
                    1.999495440268662}}]}, 
                  {RGBColor[0, 0, 1], ArrowBox[{{0, 2.8}, {-1.5, 2.8}}]}},
                 Frame->True,
                 FrameTicks->Automatic,
                 ImageSize->800,
                 PlotRange->{{-3, 3}, {-1, 3}}], 294->
                GraphicsBox[{InsetBox[
                   TagBox[GridBox[{
                    {"\<\"t\"\>", "14.65`"},
                    {"\<\"driveIntent\"\>", 
                    RowBox[{"-", "0.5`"}]},
                    {"\<\"motorTorque\"\>", "7.730501345861075`"},
                    {"\<\"vw\"\>", 
                    RowBox[{"-", "0.2948844535929953`"}]},
                    {"\<\"controllerState\"\>", 
                    RowBox[{"\[LeftAssociation]", 
                    RowBox[{
                    RowBox[{"\<\"lastError\"\>", "\[Rule]", 
                    RowBox[{"-", "0.044978878399617554`"}]}], ",", 
                    
                    RowBox[{"\<\"lastInt\"\>", "\[Rule]", 
                    "0.19670104685828446`"}], ",", 
                    RowBox[{"\<\"initialized\"\>", "\[Rule]", "True"}]}], 
                    "\[RightAssociation]"}]}
                    },
                    AutoDelete->False,
                    
                    GridBoxItemSize->{
                    "Columns" -> {{Automatic}}, "Rows" -> {{Automatic}}}],
                    "Grid"], {-1, -0.5}], 
                  {GrayLevel[0], LineBox[{{-5, 0}, {5, 0}}]}, 
                  {GrayLevel[0.5], CircleBox[{-0.3840313568841795, 0.5}, 0.5],
                    DiskBox[{-0.3840313568841795, 0.5}, 0.1]}, 
                  {GrayLevel[0.5], 
                   DiskBox[{-0.4088621639036254, 1.999794462925758}, 0.1]}, 
                  {GrayLevel[0.5], Thickness[0.01], 
                   LineBox[{{-0.3840313568841795, 0.5}, {-0.4088621639036254, 
                    1.999794462925758}}]}, 
                  {RGBColor[0, 1, 0], 
                   DiskBox[{-0.40058522823047676`, 1.4998629752838388`}, 
                    0.05]}, 
                  {RGBColor[0, 0, 1], Arrowheads[0.01], 
                   CircleBox[{-0.3840313568841795, 0.5}, 0.3333333333333333, 
                    NCache[{
                    Rational[-1, 2] 
                    Pi, -1.3388812864190642`}, {-1.5707963267948966`, \
-1.3388812864190642`}]], 
                   ArrowBox[{{-0.31066532195372076`, 
                    0.17484075281534156`}, {-0.3074174519459886, 
                    0.17559065876391144`}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-0.3840313568841795, 0.5}, {-0.538641383801401, 
                    0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-0.3840313568841795, 
                    0.5}, {-0.26229452329663994`, 0.5}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.005], Arrowheads[0.04], 
                   ArrowBox[{{-0.3840313568841795, 
                    0.5}, {-0.41690455021386147`, 0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-0.3840313568841795, 0.5}, 1.6, {
                    1.5873509542791453`, 1.6634469022970357`}], 
                   ArrowBox[{{-0.5161217702225739, 
                    2.0945382161315838`}, {-0.5320602821632197, 
                    2.0931376077667405`}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Dashing[{Small, Small}], 
                   Arrowheads[0.02], 
                   CircleBox[{-0.3840313568841795, 0.5}, 1.7, {
                    1.5873509542791453`, 1.5683022211363562`}], 
                   ArrowBox[{{-0.39679125745170357`, 
                    2.199952112542441}, {-0.37979138166050286`, 
                    2.1999947125241603`}}]}, 
                  {RGBColor[0, 0, 1], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-0.3840313568841795, 0.5}, 1.8, {
                    1.5873509542791453`, 1.5444037245799171`}], 
                   ArrowBox[{{-0.35452599437273136`, 
                    2.2997581597489893`}, {-0.3365301879892667, 
                    2.2993731238833197`}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.003], Arrowheads[0.025], 
                   CircleBox[{-0.3840313568841795, 0.5}, 1.5, {
                    1.5873509542791453`, 1.5953947256306602`}], 
                   ArrowBox[{{-0.40592817736806264`, 
                    1.9998401679021323`}, {-0.42092523424307193`, 
                    1.9995462119632816`}}]}, 
                  {RGBColor[0, 0, 1], ArrowBox[{{0, 2.8}, {-1.5, 2.8}}]}},
                 Frame->True,
                 FrameTicks->Automatic,
                 ImageSize->800,
                 PlotRange->{{-3, 3}, {-1, 3}}], 295->
                GraphicsBox[{InsetBox[
                   TagBox[GridBox[{
                    {"\<\"t\"\>", "14.700000000000001`"},
                    {"\<\"driveIntent\"\>", 
                    RowBox[{"-", "0.5`"}]},
                    {"\<\"motorTorque\"\>", "7.7231254427788265`"},
                    {"\<\"vw\"\>", 
                    RowBox[{"-", "0.2954293304498805`"}]},
                    {"\<\"controllerState\"\>", 
                    RowBox[{"\[LeftAssociation]", 
                    RowBox[{
                    RowBox[{"\<\"lastError\"\>", "\[Rule]", 
                    RowBox[{"-", "0.04493430865214915`"}]}], ",", 
                    
                    RowBox[{"\<\"lastInt\"\>", "\[Rule]", 
                    "0.19445238426990952`"}], ",", 
                    RowBox[{"\<\"initialized\"\>", "\[Rule]", "True"}]}], 
                    "\[RightAssociation]"}]}
                    },
                    AutoDelete->False,
                    
                    GridBoxItemSize->{
                    "Columns" -> {{Automatic}}, "Rows" -> {{Automatic}}}],
                    "Grid"], {-1, -0.5}], 
                  {GrayLevel[0], LineBox[{{-5, 0}, {5, 0}}]}, 
                  {GrayLevel[0.5], CircleBox[{-0.3987894265333161, 0.5}, 0.5],
                    DiskBox[{-0.3987894265333161, 0.5}, 0.1]}, 
                  {GrayLevel[0.5], 
                   DiskBox[{-0.42344807019447367`, 1.9997973034022938`}, 
                    0.1]}, 
                  {GrayLevel[0.5], Thickness[0.01], 
                   LineBox[{{-0.3987894265333161, 
                    0.5}, {-0.42344807019447367`, 1.9997973034022938`}}]}, 
                  {RGBColor[0, 1, 0], 
                   DiskBox[{-0.41522852230742113`, 1.4998648689348624`}, 
                    0.05]}, 
                  {RGBColor[0, 0, 1], Arrowheads[0.01], 
                   CircleBox[{-0.3987894265333161, 0.5}, 0.3333333333333333, 
                    NCache[{
                    Rational[-1, 2] 
                    Pi, -1.3391025635115317`}, {-1.5707963267948966`, \
-1.3391025635115317`}]], 
                   ArrowBox[{{-0.325495343691207, 
                    0.17482452655305453`}, {-0.32224730782598177`, 
                    0.17557371380403924`}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-0.3987894265333161, 0.5}, {-0.5532519353888926,
                     0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-0.3987894265333161, 0.5}, {-0.277896421211901, 
                    0.5}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.005], Arrowheads[0.04], 
                   ArrowBox[{{-0.3987894265333161, 
                    0.5}, {-0.43235893006747755`, 0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-0.3987894265333161, 0.5}, 1.6, {
                    1.5872361630865357`, 1.6628045030426892`}], 
                   ArrowBox[{{-0.5298554825257697, 
                    2.0946227418943275`}, {-0.545794890927396, 
                    2.0932323727059656`}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Dashing[{Small, Small}], 
                   Arrowheads[0.02], 
                   CircleBox[{-0.3987894265333161, 0.5}, 1.7, {
                    1.5872361630865357`, 1.567783907685955}], 
                   ArrowBox[{{-0.4106682173813568, 
                    2.199958497825164}, {-0.39366832179351186`, 
                    2.1999922865372783`}}]}, 
                  {RGBColor[0, 0, 1], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-0.3987894265333161, 0.5}, 1.8, {
                    1.5872361630865357`, 1.5443299106266535`}], 
                   ArrowBox[{{-0.369151216837679, 
                    2.2997559769385507`}, {-0.3511554389242587, 
                    2.2993696127323204`}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.003], Arrowheads[0.025], 
                   CircleBox[{-0.3987894265333161, 0.5}, 1.5, {
                    1.5872361630865357`, 1.594308999967249}], 
                   ArrowBox[{{-0.4190578194698352, 
                    1.9998630578315377`}, {-0.4340551866610225, 
                    1.9995853847522704`}}]}, 
                  {RGBColor[0, 0, 1], ArrowBox[{{0, 2.8}, {-1.5, 2.8}}]}},
                 Frame->True,
                 FrameTicks->Automatic,
                 ImageSize->800,
                 PlotRange->{{-3, 3}, {-1, 3}}], 296->
                GraphicsBox[{InsetBox[
                   TagBox[GridBox[{
                    {"\<\"t\"\>", "14.75`"},
                    {"\<\"driveIntent\"\>", 
                    RowBox[{"-", "0.5`"}]},
                    {"\<\"motorTorque\"\>", "7.806256974264446`"},
                    {"\<\"vw\"\>", 
                    RowBox[{"-", "0.29598913849564445`"}]},
                    {"\<\"controllerState\"\>", 
                    RowBox[{"\[LeftAssociation]", 
                    RowBox[{
                    RowBox[{"\<\"lastError\"\>", "\[Rule]", 
                    RowBox[{"-", "0.044696087867693586`"}]}], ",", 
                    
                    RowBox[{"\<\"lastInt\"\>", "\[Rule]", 
                    "0.1922109850827716`"}], ",", 
                    RowBox[{"\<\"initialized\"\>", "\[Rule]", "True"}]}], 
                    "\[RightAssociation]"}]}
                    },
                    AutoDelete->False,
                    
                    GridBoxItemSize->{
                    "Columns" -> {{Automatic}}, "Rows" -> {{Automatic}}}],
                    "Grid"], {-1, -0.5}], 
                  {GrayLevel[0], LineBox[{{-5, 0}, {5, 0}}]}, 
                  {GrayLevel[0.5], CircleBox[{-0.4135750917859091, 0.5}, 0.5],
                    DiskBox[{-0.4135750917859091, 0.5}, 0.1]}, 
                  {GrayLevel[0.5], 
                   DiskBox[{-0.4383445941196735, 1.9997954766414445`}, 0.1]}, 
                  {GrayLevel[0.5], Thickness[0.01], 
                   LineBox[{{-0.4135750917859091, 0.5}, {-0.4383445941196735, 
                    1.9997954766414445`}}]}, 
                  {RGBColor[0, 1, 0], 
                   DiskBox[{-0.4300880933417521, 1.4998636510942962`}, 0.05]}, 
                  {RGBColor[0, 0, 1], Arrowheads[0.01], 
                   CircleBox[{-0.4135750917859091, 0.5}, 0.3333333333333333, 
                    NCache[{
                    Rational[-1, 2] 
                    Pi, -1.3366086175669631`}, {-1.5707963267948966`, \
-1.3366086175669631`}]], 
                   ArrowBox[{{-0.3394702676671154, 
                    0.17500832910144642`}, {-0.33622411033346655`, 
                    0.17576561444002425`}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-0.4135750917859091, 0.5}, {-0.5697002312711981,
                     0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-0.4135750917859091, 
                    0.5}, {-0.29213873172794436`, 0.5}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.005], Arrowheads[0.04], 
                   ArrowBox[{{-0.4135750917859091, 0.5}, {-0.4482638712132333,
                     0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-0.4135750917859091, 0.5}, 1.6, {
                    1.5873100789015717`, 1.6632181539207627`}], 
                   ArrowBox[{{-0.5453007536437695, 
                    2.094568389881133}, {-0.5612395855543039, 
                    2.0931714274616287`}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Dashing[{Small, Small}], 
                   Arrowheads[0.02], 
                   CircleBox[{-0.4135750917859091, 0.5}, 1.7, {
                    1.5873100789015717`, 1.5672092705542808`}], 
                   ArrowBox[{{-0.42447702145132066`, 
                    2.199965043149291}, {-0.40747710925397745`, 
                    2.1999890630851247`}}]}, 
                  {RGBColor[0, 0, 1], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-0.4135750917859091, 0.5}, 1.8, {
                    1.5873100789015717`, 1.5439419846001026`}], 
                   ArrowBox[{{-0.3832387121529646, 
                    2.2997443440863385`}, {-0.36524308547433365`, 
                    2.299350998878734}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.003], Arrowheads[0.025], 
                   CircleBox[{-0.4135750917859091, 0.5}, 1.5, {
                    1.5873100789015717`, 1.5934560629666825`}], 
                   ArrowBox[{{-0.43256418880709196`, 
                    1.99987979991542}, {-0.44756178738059044`, 
                    1.9996149187449939`}}]}, 
                  {RGBColor[0, 0, 1], ArrowBox[{{0, 2.8}, {-1.5, 2.8}}]}},
                 Frame->True,
                 FrameTicks->Automatic,
                 ImageSize->800,
                 PlotRange->{{-3, 3}, {-1, 3}}], 297->
                GraphicsBox[{InsetBox[
                   TagBox[GridBox[{
                    {"\<\"t\"\>", "14.8`"},
                    {"\<\"driveIntent\"\>", 
                    RowBox[{"-", "0.5`"}]},
                    {"\<\"motorTorque\"\>", "7.967563112480914`"},
                    {"\<\"vw\"\>", 
                    RowBox[{"-", "0.29657029823148684`"}]},
                    {"\<\"controllerState\"\>", 
                    RowBox[{"\[LeftAssociation]", 
                    RowBox[{
                    RowBox[{"\<\"lastError\"\>", "\[Rule]", 
                    RowBox[{"-", "0.04428701571030637`"}]}], ",", 
                    
                    RowBox[{"\<\"lastInt\"\>", "\[Rule]", 
                    "0.18998594668659624`"}], ",", 
                    RowBox[{"\<\"initialized\"\>", "\[Rule]", "True"}]}], 
                    "\[RightAssociation]"}]}
                    },
                    AutoDelete->False,
                    
                    GridBoxItemSize->{
                    "Columns" -> {{Automatic}}, "Rows" -> {{Automatic}}}],
                    "Grid"], {-1, -0.5}], 
                  {GrayLevel[0], LineBox[{{-5, 0}, {5, 0}}]}, 
                  {GrayLevel[0.5], CircleBox[{-0.4283892673745039, 0.5}, 0.5],
                    DiskBox[{-0.4283892673745039, 0.5}, 0.1]}, 
                  {GrayLevel[0.5], 
                   DiskBox[{-0.45351559304144506`, 1.9997895411551845`}, 
                    0.1]}, 
                  {GrayLevel[0.5], Thickness[0.01], 
                   LineBox[{{-0.4283892673745039, 
                    0.5}, {-0.45351559304144506`, 1.9997895411551845`}}]}, 
                  {RGBColor[0, 1, 0], 
                   DiskBox[{-0.44514015115246464`, 1.4998596941034563`}, 
                    0.05]}, 
                  {RGBColor[0, 0, 1], Arrowheads[0.01], 
                   CircleBox[{-0.4283892673745039, 0.5}, 0.3333333333333333, 
                    NCache[{
                    Rational[-1, 2] 
                    Pi, -1.3317694334204693`}, {-1.5707963267948966`, \
-1.3317694334204693`}]], 
                   ArrowBox[{{-0.3527126225329925, 
                    0.175370739863698}, {-0.3494701678369459, 
                    0.1761437250271471}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-0.4283892673745039, 0.5}, {-0.5877405296241222,
                     0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-0.4283892673745039, 0.5}, {-0.3052040126165932,
                     0.5}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.005], Arrowheads[0.04], 
                   ArrowBox[{{-0.4283892673745039, 
                    0.5}, {-0.46455527486621145`, 0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-0.4283892673745039, 0.5}, 1.6, {
                    1.5875479940327493`, 1.6645495820332832`}], 
                   ArrowBox[{{-0.5622378650316552, 
                    2.094391593337476}, {-0.5781748228603136, 
                    2.0929734107535527`}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Dashing[{Small, Small}], 
                   Arrowheads[0.02], 
                   CircleBox[{-0.4283892673745039, 0.5}, 1.7, {
                    1.5875479940327493`, 1.5665912738337469`}], 
                   ArrowBox[{{-0.4382406222032671, 
                    2.199971455880374}, {-0.42124069840798595`, 
                    2.199984969922302}}]}, 
                  {RGBColor[0, 0, 1], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-0.4283892673745039, 0.5}, 1.8, {
                    1.5875479940327493`, 1.5432837545189664`}], 
                   ArrowBox[{{-0.39686824853334507`, 
                    2.2997239858853957`}, {-0.37887288466479735`, 
                    2.299318795501049}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.003], Arrowheads[0.025], 
                   CircleBox[{-0.4283892673745039, 0.5}, 1.5, {
                    1.5875479940327493`, 1.5928149842579098`}], 
                   ArrowBox[{{-0.44641681955401774`, 
                    1.9998916652086627`}, {-0.46141458485531256`, 
                    1.999636398733137}}]}, 
                  {RGBColor[0, 0, 1], ArrowBox[{{0, 2.8}, {-1.5, 2.8}}]}},
                 Frame->True,
                 FrameTicks->Automatic,
                 ImageSize->800,
                 PlotRange->{{-3, 3}, {-1, 3}}], 298->
                GraphicsBox[{InsetBox[
                   TagBox[GridBox[{
                    {"\<\"t\"\>", "14.85`"},
                    {"\<\"driveIntent\"\>", 
                    RowBox[{"-", "0.5`"}]},
                    {"\<\"motorTorque\"\>", "8.19538012817792`"},
                    {"\<\"vw\"\>", 
                    RowBox[{"-", "0.29717819481510527`"}]},
                    {"\<\"controllerState\"\>", 
                    RowBox[{"\[LeftAssociation]", 
                    RowBox[{
                    RowBox[{"\<\"lastError\"\>", "\[Rule]", 
                    RowBox[{"-", "0.04372893032963607`"}]}], ",", 
                    
                    RowBox[{"\<\"lastInt\"\>", "\[Rule]", 
                    "0.18778525055103562`"}], ",", 
                    RowBox[{"\<\"initialized\"\>", "\[Rule]", "True"}]}], 
                    "\[RightAssociation]"}]}
                    },
                    AutoDelete->False,
                    
                    GridBoxItemSize->{
                    "Columns" -> {{Automatic}}, "Rows" -> {{Automatic}}}],
                    "Grid"], {-1, -0.5}], 
                  {GrayLevel[0], LineBox[{{-5, 0}, {5, 0}}]}, 
                  {GrayLevel[0.5], CircleBox[{-0.4432331623779065, 0.5}, 0.5],
                    DiskBox[{-0.4432331623779065, 0.5}, 0.1]}, 
                  {GrayLevel[0.5], 
                   DiskBox[{-0.46892712816165577`, 1.9997799238962708`}, 
                    0.1]}, 
                  {GrayLevel[0.5], Thickness[0.01], 
                   LineBox[{{-0.4432331623779065, 
                    0.5}, {-0.46892712816165577`, 1.9997799238962708`}}]}, 
                  {RGBColor[0, 1, 0], 
                   DiskBox[{-0.460362472900406, 1.4998532825975137`}, 0.05]}, 
                  {RGBColor[0, 0, 1], Arrowheads[0.01], 
                   CircleBox[{-0.4432331623779065, 0.5}, 0.3333333333333333, 
                    NCache[{
                    Rational[-1, 2] 
                    Pi, -1.3249349229495588`}, {-1.5707963267948966`, \
-1.3249349229495588`}]], 
                   ArrowBox[{{-0.365339620172769, 
                    0.17589553043216521`}, {-0.3621025241388869, 
                    0.17669065796045974`}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-0.4432331623779065, 0.5}, {-0.6071407649414648,
                     0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-0.4432331623779065, 0.5}, {-0.3172657819201341,
                     0.5}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.005], Arrowheads[0.04], 
                   ArrowBox[{{-0.4432331623779065, 
                    0.5}, {-0.48117338448369246`, 0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-0.4432331623779065, 0.5}, 1.6, {
                    1.5879264750892113`, 1.666667640574588}], 
                   ArrowBox[{{-0.5804584720087309, 
                    2.094104518027825}, {-0.5963923902979467, 
                    2.0926525832405938`}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Dashing[{Small, Small}], 
                   Arrowheads[0.02], 
                   CircleBox[{-0.4432331623779065, 0.5}, 1.7, {
                    1.5879264750892113`, 1.5659418142835122`}], 
                   ArrowBox[{{-0.4519804525094977, 
                    2.199977495414381}, {-0.4349805235226579, 
                    2.199979968691374}}]}, 
                  {RGBColor[0, 0, 1], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-0.4432331623779065, 0.5}, 1.8, {
                    1.5879264750892113`, 1.542396585488223}], 
                   ArrowBox[{{-0.4101154967666366, 
                    2.2996953131639977`}, {-0.3921204994521973, 
                    2.2992741580116807`}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.003], Arrowheads[0.025], 
                   CircleBox[{-0.4432331623779065, 0.5}, 1.5, {
                    1.5879264750892113`, 1.5923652351726345`}], 
                   ArrowBox[{{-0.4605861378524676, 
                    1.9998996213887712`}, {-0.47558401644279164`, 
                    1.9996511001700603`}}]}, 
                  {RGBColor[0, 0, 1], ArrowBox[{{0, 2.8}, {-1.5, 2.8}}]}},
                 Frame->True,
                 FrameTicks->Automatic,
                 ImageSize->800,
                 PlotRange->{{-3, 3}, {-1, 3}}], 299->
                GraphicsBox[{InsetBox[
                   TagBox[GridBox[{
                    {"\<\"t\"\>", "14.9`"},
                    {"\<\"driveIntent\"\>", 
                    RowBox[{"-", "0.5`"}]},
                    {"\<\"motorTorque\"\>", "8.478748652679254`"},
                    {"\<\"vw\"\>", 
                    RowBox[{"-", "0.2978172457068898`"}]},
                    {"\<\"controllerState\"\>", 
                    RowBox[{"\[LeftAssociation]", 
                    RowBox[{
                    RowBox[{"\<\"lastError\"\>", "\[Rule]", 
                    RowBox[{"-", "0.043042610021991695`"}]}], ",", 
                    
                    RowBox[{"\<\"lastInt\"\>", "\[Rule]", 
                    "0.18561581290000145`"}], ",", 
                    RowBox[{"\<\"initialized\"\>", "\[Rule]", "True"}]}], 
                    "\[RightAssociation]"}]}
                    },
                    AutoDelete->False,
                    
                    GridBoxItemSize->{
                    "Columns" -> {{Automatic}}, "Rows" -> {{Automatic}}}],
                    "Grid"], {-1, -0.5}], 
                  {GrayLevel[0], LineBox[{{-5, 0}, {5, 0}}]}, 
                  {GrayLevel[0.5], 
                   CircleBox[{-0.45810823017107805`, 0.5}, 0.5], 
                   DiskBox[{-0.45810823017107805`, 0.5}, 0.1]}, 
                  {GrayLevel[0.5], 
                   DiskBox[{-0.48454753163120967`, 1.9997669696783897`}, 
                    0.1]}, 
                  {GrayLevel[0.5], Thickness[0.01], 
                   LineBox[{{-0.45810823017107805`, 
                    0.5}, {-0.48454753163120967`, 1.9997669696783897`}}]}, 
                  {RGBColor[0, 1, 0], 
                   DiskBox[{-0.4757344311444991, 1.49984464645226}, 0.05]}, 
                  {RGBColor[0, 0, 1], Arrowheads[0.01], 
                   CircleBox[{-0.45810823017107805`, 0.5}, 0.3333333333333333,
                     NCache[{
                    Rational[-1, 2] 
                    Pi, -1.3164338672145188`}, {-1.5707963267948966`, \
-1.3164338672145188`}]], 
                   ArrowBox[{{-0.3774623055781563, 
                    0.176569410912227}, {-0.37423208585473455`, 
                    0.17739202811192806`}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-0.45810823017107805`, 
                    0.5}, {-0.6276832032246631, 0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-0.45810823017107805`, 
                    0.5}, {-0.3284878824487938, 0.5}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.005], Arrowheads[0.04], 
                   ArrowBox[{{-0.45810823017107805`, 
                    0.5}, {-0.49806285550237883`, 0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-0.45810823017107805`, 0.5}, 1.6, {
                    1.5884234405893267`, 1.6694487454843265`}], 
                   ArrowBox[{{-0.5997663753013184, 
                    2.0937167157052285`}, {-0.6156961939920137, 
                    2.0922204726917597`}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Dashing[{Small, Small}], 
                   Arrowheads[0.02], 
                   CircleBox[{-0.45810823017107805`, 0.5}, 1.7, {
                    1.5884234405893267`, 1.5652717276694028`}], 
                   ArrowBox[{{-0.46571638626011613`, 
                    2.1999829751973765`}, {-0.44871645943258714`, 
                    2.199974057049811}}]}, 
                  {RGBColor[0, 0, 1], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-0.45810823017107805`, 0.5}, 1.8, {
                    1.5884234405893267`, 1.5413192814077752`}], 
                   ArrowBox[{{-0.42305176504831693`, 
                    2.299658591025836}, {-0.40505723188830667`, 
                    2.299218050037627}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.003], Arrowheads[0.025], 
                   CircleBox[{-0.45810823017107805`, 0.5}, 1.5, {
                    1.5884234405893267`, 1.5920868334835512`}], 
                   ArrowBox[{{-0.47504363039049036`, 
                    1.9999043943596566`}, {-0.49004157758831657`, 
                    1.9996600485852554`}}]}, 
                  {RGBColor[0, 0, 1], ArrowBox[{{0, 2.8}, {-1.5, 2.8}}]}},
                 Frame->True,
                 FrameTicks->Automatic,
                 ImageSize->800,
                 PlotRange->{{-3, 3}, {-1, 3}}], 300->
                GraphicsBox[{InsetBox[
                   TagBox[GridBox[{
                    {"\<\"t\"\>", "14.950000000000001`"},
                    {"\<\"driveIntent\"\>", 
                    RowBox[{"-", "0.5`"}]},
                    {"\<\"motorTorque\"\>", "8.807438174072338`"},
                    {"\<\"vw\"\>", 
                    RowBox[{"-", "0.29849096977632733`"}]},
                    {"\<\"controllerState\"\>", 
                    RowBox[{"\[LeftAssociation]", 
                    RowBox[{
                    RowBox[{"\<\"lastError\"\>", "\[Rule]", 
                    RowBox[{"-", "0.04224769595996963`"}]}], ",", 
                    
                    RowBox[{"\<\"lastInt\"\>", "\[Rule]", 
                    "0.1834835397592191`"}], ",", 
                    RowBox[{"\<\"initialized\"\>", "\[Rule]", "True"}]}], 
                    "\[RightAssociation]"}]}
                    },
                    AutoDelete->False,
                    
                    GridBoxItemSize->{
                    "Columns" -> {{Automatic}}, "Rows" -> {{Automatic}}}],
                    "Grid"], {-1, -0.5}], 
                  {GrayLevel[0], LineBox[{{-5, 0}, {5, 0}}]}, 
                  {GrayLevel[0.5], CircleBox[{-0.4730161217981803, 0.5}, 0.5],
                    DiskBox[{-0.4730161217981803, 0.5}, 0.1]}, 
                  {GrayLevel[0.5], 
                   DiskBox[{-0.5003474446848223, 1.9997509789259236`}, 0.1]}, 
                  {GrayLevel[0.5], Thickness[0.01], 
                   LineBox[{{-0.4730161217981803, 0.5}, {-0.5003474446848223, 
                    1.9997509789259236`}}]}, 
                  {RGBColor[0, 1, 0], 
                   DiskBox[{-0.4912370037226083, 1.4998339859506156`}, 0.05]}, 
                  {RGBColor[0, 0, 1], Arrowheads[0.01], 
                   CircleBox[{-0.4730161217981803, 0.5}, 0.3333333333333333, 
                    NCache[{
                    Rational[-1, 2] 
                    Pi, -1.3065731815727264`}, {-1.5707963267948966`, \
-1.3065731815727264`}]], 
                   ArrowBox[{{-0.38918492221820233`, 
                    0.1773803460914184}, {-0.3859629709738394, 
                    0.17823477496369244`}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-0.4730161217981803, 0.5}, {-0.649164885279627, 
                    0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-0.4730161217981803, 
                    0.5}, {-0.33902401127141535`, 0.5}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.005], Arrowheads[0.04], 
                   ArrowBox[{{-0.4730161217981803, 0.5}, {-0.5151727747528622,
                     0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-0.4730161217981803, 0.5}, 1.6, {
                    1.5890182170937506`, 1.6727771913394653`}], 
                   ArrowBox[{{-0.6199780722477076, 
                    2.0932363870813617`}, {-0.6359028225441626, 
                    2.0916871309148948`}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Dashing[{Small, Small}], 
                   Arrowheads[0.02], 
                   CircleBox[{-0.4730161217981803, 0.5}, 1.7, {
                    1.5890182170937506`, 1.5645908052031199`}], 
                   ArrowBox[{{-0.4794667196127779, 
                    2.199987761658252}, {-0.46246680279886937`, 
                    2.1999672678815467`}}]}, 
                  {RGBColor[0, 0, 1], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-0.4730161217981803, 0.5}, 1.8, {
                    1.5890182170937506`, 1.5400880050155707`}], 
                   ArrowBox[{{-0.43574380667168033`, 
                    2.299614062660411}, {-0.4177498295794799, 
                    2.2991513657678713`}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.003], Arrowheads[0.025], 
                   CircleBox[{-0.4730161217981803, 0.5}, 1.5, {
                    1.5890182170937506`, 1.591960468319316}], 
                   ArrowBox[{{-0.48976198621775324`, 
                    1.9999065224289283`}, {-0.5047599641726285, 
                    1.9996640718745335`}}]}, 
                  {RGBColor[0, 0, 1], ArrowBox[{{0, 2.8}, {-1.5, 2.8}}]}},
                 Frame->True,
                 FrameTicks->Automatic,
                 ImageSize->800,
                 PlotRange->{{-3, 3}, {-1, 3}}], 301->
                GraphicsBox[{InsetBox[
                   TagBox[GridBox[{
                    {"\<\"t\"\>", "15.`"},
                    {"\<\"driveIntent\"\>", 
                    RowBox[{"-", "0.5`"}]},
                    {"\<\"motorTorque\"\>", "9.17196177690202`"},
                    {"\<\"vw\"\>", 
                    RowBox[{"-", "0.29920205704629455`"}]},
                    {"\<\"controllerState\"\>", 
                    RowBox[{"\[LeftAssociation]", 
                    RowBox[{
                    RowBox[{"\<\"lastError\"\>", "\[Rule]", 
                    RowBox[{"-", "0.04136263440731733`"}]}], ",", 
                    
                    RowBox[{"\<\"lastInt\"\>", "\[Rule]", 
                    "0.18139338536373514`"}], ",", 
                    RowBox[{"\<\"initialized\"\>", "\[Rule]", "True"}]}], 
                    "\[RightAssociation]"}]}
                    },
                    AutoDelete->False,
                    
                    GridBoxItemSize->{
                    "Columns" -> {{Automatic}}, "Rows" -> {{Automatic}}}],
                    "Grid"], {-1, -0.5}], 
                  {GrayLevel[0], LineBox[{{-5, 0}, {5, 0}}]}, 
                  {GrayLevel[0.5], 
                   CircleBox[{-0.48795864282044954`, 0.5}, 0.5], 
                   DiskBox[{-0.48795864282044954`, 0.5}, 0.1]}, 
                  {GrayLevel[0.5], 
                   DiskBox[{-0.5162998296545075, 1.9997322351436062`}, 0.1]}, 
                  {GrayLevel[0.5], Thickness[0.01], 
                   LineBox[{{-0.48795864282044954`, 
                    0.5}, {-0.5162998296545075, 1.9997322351436062`}}]}, 
                  {RGBColor[0, 1, 0], 
                   DiskBox[{-0.5068527673764881, 1.4998214900957376`}, 0.05]}, 
                  {RGBColor[0, 0, 1], Arrowheads[0.01], 
                   CircleBox[{-0.48795864282044954`, 0.5}, 0.3333333333333333,
                     NCache[{
                    Rational[-1, 2] 
                    Pi, -1.295637473487836}, {-1.5707963267948966`, \
-1.295637473487836}]], 
                   ArrowBox[{{-0.4006044518272142, 
                    0.17831637214954743`}, {-0.3973920368355318, 
                    0.17920598354787887`}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-0.48795864282044954`, 0.5}, {-0.67139787835849,
                     0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-0.48795864282044954`, 
                    0.5}, {-0.3490173997036118, 0.5}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.005], Arrowheads[0.04], 
                   ArrowBox[{{-0.48795864282044954`, 
                    0.5}, {-0.5324566352416522, 0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-0.48795864282044954`, 0.5}, 1.6, {
                    1.5896915756940067`, 1.6765453569021986`}], 
                   ArrowBox[{{-0.6409231141891385, 
                    2.0926713002056947`}, {-0.6568419135874726, 
                    2.0910620480845594`}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Dashing[{Small, Small}], 
                   Arrowheads[0.02], 
                   CircleBox[{-0.48795864282044954`, 0.5}, 1.7, {
                    1.5896915756940067`, 1.5639078185188044`}], 
                   ArrowBox[{{-0.4932481702160951, 
                    2.1999917708330035`}, {-0.4762482713641444, 
                    2.199959666345162}}]}, 
                  {RGBColor[0, 0, 1], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-0.48795864282044954`, 0.5}, 1.8, {
                    1.5896915756940067`, 1.5387362324889955`}], 
                   ArrowBox[{{-0.44825369364011136`, 
                    2.2995620347769585`}, {-0.43026035844876326`, 
                    2.2990750145506897`}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.003], Arrowheads[0.025], 
                   CircleBox[{-0.48795864282044954`, 0.5}, 1.5, {
                    1.5896915756940067`, 1.591967605671502}], 
                   ArrowBox[{{-0.5047152126006924, 
                    1.9999064028695923`}, {-0.5197131888247306, 
                    1.9996638452693531`}}]}, 
                  {RGBColor[0, 0, 1], ArrowBox[{{0, 2.8}, {-1.5, 2.8}}]}},
                 Frame->True,
                 FrameTicks->Automatic,
                 ImageSize->800,
                 PlotRange->{{-3, 3}, {-1, 3}}], 302->
                GraphicsBox[{InsetBox[
                   TagBox[GridBox[{
                    {"\<\"t\"\>", "15.05`"},
                    {"\<\"driveIntent\"\>", 
                    RowBox[{"-", "0.5`"}]},
                    {"\<\"motorTorque\"\>", "9.563582107048418`"},
                    {"\<\"vw\"\>", 
                    RowBox[{"-", "0.2999524383406843`"}]},
                    {"\<\"controllerState\"\>", 
                    RowBox[{"\[LeftAssociation]", 
                    RowBox[{
                    RowBox[{"\<\"lastError\"\>", "\[Rule]", 
                    RowBox[{"-", "0.04040463680171594`"}]}], ",", 
                    
                    RowBox[{"\<\"lastInt\"\>", "\[Rule]", 
                    "0.17934941299228604`"}], ",", 
                    RowBox[{"\<\"initialized\"\>", "\[Rule]", "True"}]}], 
                    "\[RightAssociation]"}]}
                    },
                    AutoDelete->False,
                    
                    GridBoxItemSize->{
                    "Columns" -> {{Automatic}}, "Rows" -> {{Automatic}}}],
                    "Grid"], {-1, -0.5}], 
                  {GrayLevel[0], LineBox[{{-5, 0}, {5, 0}}]}, 
                  {GrayLevel[0.5], CircleBox[{-0.5029377136516979, 0.5}, 0.5],
                    DiskBox[{-0.5029377136516979, 0.5}, 0.1]}, 
                  {GrayLevel[0.5], 
                   DiskBox[{-0.5323799586633738, 1.9997110235670978`}, 0.1]}, 
                  {GrayLevel[0.5], Thickness[0.01], 
                   LineBox[{{-0.5029377136516979, 0.5}, {-0.5323799586633738, 
                    1.9997110235670978`}}]}, 
                  {RGBColor[0, 1, 0], 
                   DiskBox[{-0.5225658769928152, 1.499807349044732}, 0.05]}, 
                  {RGBColor[0, 0, 1], Arrowheads[0.01], 
                   CircleBox[{-0.5029377136516979, 0.5}, 0.3333333333333333, 
                    NCache[{
                    Rational[-1, 2] 
                    Pi, -1.283888863583444}, {-1.5707963267948966`, \
-1.283888863583444}]], 
                   ArrowBox[{{-0.41181030281805175`, 
                    0.17936483956704474`}, {-0.40860856098522047`, 
                    0.18029213011197942`}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-0.5029377136516979, 0.5}, {-0.6942093557926663,
                     0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-0.5029377136516979, 0.5}, {-0.3586006296649482,
                     0.5}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.005], Arrowheads[0.04], 
                   ArrowBox[{{-0.5029377136516979, 0.5}, {-0.5498722718059166,
                     0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-0.5029377136516979, 0.5}, 1.6, {
                    1.590425750694626, 1.6806538107082982`}], 
                   ArrowBox[{{-0.6624442921055758, 
                    2.092029412865836}, {-0.6783563456348635, 
                    2.0903547728582934`}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Dashing[{Small, Small}], 
                   Arrowheads[0.02], 
                   CircleBox[{-0.5029377136516979, 0.5}, 1.7, {
                    1.590425750694626, 1.5632305517786786`}], 
                   ArrowBox[{{-0.5070758920373621, 
                    2.1999949633689058`}, {-0.49007601882755525`, 
                    2.19995134542323}}]}, 
                  {RGBColor[0, 0, 1], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-0.5029377136516979, 0.5}, 1.8, {
                    1.590425750694626, 1.5372947389888014`}], 
                   ArrowBox[{{-0.46063874964491885`, 
                    2.299502930712799}, {-0.4426461351840222, 
                    2.298989973725778}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.003], Arrowheads[0.025], 
                   CircleBox[{-0.5029377136516979, 0.5}, 1.5, {
                    1.590425750694626, 1.5920905752241465`}], 
                   ArrowBox[{{-0.5198787261241532, 
                    1.9999043309812823`}, {-0.534876672407596, 
                    1.999659929088455}}]}, 
                  {RGBColor[0, 0, 1], ArrowBox[{{0, 2.8}, {-1.5, 2.8}}]}},
                 Frame->True,
                 FrameTicks->Automatic,
                 ImageSize->800,
                 PlotRange->{{-3, 3}, {-1, 3}}], 303->
                GraphicsBox[{InsetBox[
                   TagBox[GridBox[{
                    {"\<\"t\"\>", "15.1`"},
                    {"\<\"driveIntent\"\>", 
                    RowBox[{"-", "0.5`"}]},
                    {"\<\"motorTorque\"\>", "9.974309508816972`"},
                    {"\<\"vw\"\>", 
                    RowBox[{"-", "0.3007433541874168`"}]},
                    {"\<\"controllerState\"\>", 
                    RowBox[{"\[LeftAssociation]", 
                    RowBox[{
                    RowBox[{"\<\"lastError\"\>", "\[Rule]", 
                    RowBox[{"-", "0.0393896560842786`"}]}], ",", 
                    
                    RowBox[{"\<\"lastInt\"\>", "\[Rule]", 
                    "0.17735485737650836`"}], ",", 
                    RowBox[{"\<\"initialized\"\>", "\[Rule]", "True"}]}], 
                    "\[RightAssociation]"}]}
                    },
                    AutoDelete->False,
                    
                    GridBoxItemSize->{
                    "Columns" -> {{Automatic}}, "Rows" -> {{Automatic}}}],
                    "Grid"], {-1, -0.5}], 
                  {GrayLevel[0], LineBox[{{-5, 0}, {5, 0}}]}, 
                  {GrayLevel[0.5], CircleBox[{-0.5179553333597297, 0.5}, 0.5],
                    DiskBox[{-0.5179553333597297, 0.5}, 0.1]}, 
                  {GrayLevel[0.5], 
                   DiskBox[{-0.5485653817179825, 1.9996876424574237`}, 0.1]}, 
                  {GrayLevel[0.5], Thickness[0.01], 
                   LineBox[{{-0.5179553333597297, 0.5}, {-0.5485653817179825, 
                    1.9996876424574237`}}]}, 
                  {RGBColor[0, 1, 0], 
                   DiskBox[{-0.5383620322652315, 1.4997917616382823`}, 0.05]}, 
                  {RGBColor[0, 0, 1], Arrowheads[0.01], 
                   CircleBox[{-0.5179553333597297, 0.5}, 0.3333333333333333, 
                    NCache[{
                    Rational[-1, 2] 
                    Pi, -1.2715670415303872`}, {-1.5707963267948966`, \
-1.2715670415303872`}]], 
                   ArrowBox[{{-0.4228841308347562, 
                    0.1805120071715267}, {-0.4196940576747174, 
                    0.18147867761817715`}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-0.5179553333597297, 0.5}, {-0.7174415235360692,
                     0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-0.5179553333597297, 
                    0.5}, {-0.36789557270064804`, 0.5}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.005], Arrowheads[0.04], 
                   ArrowBox[{{-0.5179553333597297, 0.5}, {-0.5673817628769875,
                     0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-0.5179553333597297, 0.5}, 1.6, {
                    1.591204442304245, 1.6850113269081388`}], 
                   ArrowBox[{{-0.6843976694677846, 
                    2.0913192479043587`}, {-0.700302274680825, 
                    2.0895752869841813`}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Dashing[{Small, Small}], 
                   Arrowheads[0.02], 
                   CircleBox[{-0.5179553333597297, 0.5}, 1.7, {
                    1.591204442304245, 1.5625658396492403`}], 
                   ArrowBox[{{-0.5209635036422617, 
                    2.199997338501314}, {-0.503963663181461, 
                    2.199942420544185}}]}, 
                  {RGBColor[0, 0, 1], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-0.5179553333597297, 0.5}, 1.8, {
                    1.591204442304245, 1.5357916116997063`}], 
                   ArrowBox[{{-0.4729515361945081, 
                    2.2994373171190796`}, {-0.454959713099144, 
                    2.2988973155319297`}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.003], Arrowheads[0.025], 
                   CircleBox[{-0.5179553333597297, 0.5}, 1.5, {
                    1.591204442304245, 1.5923126394806133`}], 
                   ArrowBox[{{-0.535229420551776, 
                    1.9999005320059333`}, {-0.5502273121925012, 
                    1.999652799611369}}]}, 
                  {RGBColor[0, 0, 1], ArrowBox[{{0, 2.8}, {-1.5, 2.8}}]}},
                 Frame->True,
                 FrameTicks->Automatic,
                 ImageSize->800,
                 PlotRange->{{-3, 3}, {-1, 3}}], 304->
                GraphicsBox[{InsetBox[
                   TagBox[GridBox[{
                    {"\<\"t\"\>", "15.15`"},
                    {"\<\"driveIntent\"\>", 
                    RowBox[{"-", "0.5`"}]},
                    {"\<\"motorTorque\"\>", "10.39689324215945`"},
                    {"\<\"vw\"\>", 
                    RowBox[{"-", "0.30157542241228535`"}]},
                    {"\<\"controllerState\"\>", 
                    RowBox[{"\[LeftAssociation]", 
                    RowBox[{
                    RowBox[{"\<\"lastError\"\>", "\[Rule]", 
                    RowBox[{"-", "0.03833237767429158`"}]}], ",", 
                    
                    RowBox[{"\<\"lastInt\"\>", "\[Rule]", 
                    "0.17541218791361823`"}], ",", 
                    RowBox[{"\<\"initialized\"\>", "\[Rule]", "True"}]}], 
                    "\[RightAssociation]"}]}
                    },
                    AutoDelete->False,
                    
                    GridBoxItemSize->{
                    "Columns" -> {{Automatic}}, "Rows" -> {{Automatic}}}],
                    "Grid"], {-1, -0.5}], 
                  {GrayLevel[0], LineBox[{{-5, 0}, {5, 0}}]}, 
                  {GrayLevel[0.5], CircleBox[{-0.5330135468817087, 0.5}, 0.5],
                    DiskBox[{-0.5330135468817087, 0.5}, 0.1]}, 
                  {GrayLevel[0.5], 
                   DiskBox[{-0.5648358768111326, 1.9996624084499361`}, 0.1]}, 
                  {GrayLevel[0.5], Thickness[0.01], 
                   LineBox[{{-0.5330135468817087, 0.5}, {-0.5648358768111326, 
                    1.9996624084499361`}}]}, 
                  {RGBColor[0, 1, 0], 
                   DiskBox[{-0.5542284335013246, 1.499774938966624}, 0.05]}, 
                  {RGBColor[0, 0, 1], Arrowheads[0.01], 
                   CircleBox[{-0.5330135468817087, 0.5}, 0.3333333333333333, 
                    NCache[{
                    Rational[-1, 2] 
                    Pi, -1.258889529530113}, {-1.5707963267948966`, \
-1.258889529530113}]], 
                   ArrowBox[{{-0.4338997797734893, 
                    0.18174291479885518`}, {-0.4307222176111183, 
                    0.18274994867271216`}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-0.5330135468817087, 0.5}, {-0.7409514117248976,
                     0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-0.5330135468817087, 0.5}, {-0.3770134384297007,
                     0.5}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.005], Arrowheads[0.04], 
                   ArrowBox[{{-0.5330135468817087, 0.5}, {-0.5849513032728897,
                     0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-0.5330135468817087, 0.5}, 1.6, {
                    1.5920128051059064`, 1.6895348212941514`}], 
                   ArrowBox[{{-0.7066524792565944, 
                    2.090550068738394}, {-0.7225490329793562, 
                    2.0887341815136127`}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Dashing[{Small, Small}], 
                   Arrowheads[0.02], 
                   CircleBox[{-0.5330135468817087, 0.5}, 1.7, {
                    1.5920128051059064`, 1.5619196100274197`}], 
                   ArrowBox[{{-0.5349231279754241, 
                    2.199998927499675}, {-0.5179233265539063, 
                    2.199933023754306}}]}, 
                  {RGBColor[0, 0, 1], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-0.5330135468817087, 0.5}, 1.8, {
                    1.5920128051059064`, 1.5342522870939095`}], 
                   ArrowBox[{{-0.4852398859902463, 
                    2.2993659097929546`}, {-0.4672489154482745, 
                    2.298798213600521}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.003], Arrowheads[0.025], 
                   CircleBox[{-0.5330135468817087, 0.5}, 1.5, {
                    1.5920128051059064`, 1.5926180463366866`}], 
                   ArrowBox[{{-0.5507457131669304, 
                    1.999895186430983}, {-0.5657435284490333, 
                    1.9996428735891099`}}]}, 
                  {RGBColor[0, 0, 1], ArrowBox[{{0, 2.8}, {-1.5, 2.8}}]}},
                 Frame->True,
                 FrameTicks->Automatic,
                 ImageSize->800,
                 PlotRange->{{-3, 3}, {-1, 3}}], 305->
                GraphicsBox[{InsetBox[
                   TagBox[GridBox[{
                    {"\<\"t\"\>", "15.200000000000001`"},
                    {"\<\"driveIntent\"\>", 
                    RowBox[{"-", "0.5`"}]},
                    {"\<\"motorTorque\"\>", "10.824806645783053`"},
                    {"\<\"vw\"\>", 
                    RowBox[{"-", "0.30244870393796525`"}]},
                    {"\<\"controllerState\"\>", 
                    RowBox[{"\[LeftAssociation]", 
                    RowBox[{
                    RowBox[{"\<\"lastError\"\>", "\[Rule]", 
                    RowBox[{"-", "0.03724622352555882`"}]}], ",", 
                    
                    RowBox[{"\<\"lastInt\"\>", "\[Rule]", 
                    "0.17352317199035758`"}], ",", 
                    RowBox[{"\<\"initialized\"\>", "\[Rule]", "True"}]}], 
                    "\[RightAssociation]"}]}
                    },
                    AutoDelete->False,
                    
                    GridBoxItemSize->{
                    "Columns" -> {{Automatic}}, "Rows" -> {{Automatic}}}],
                    "Grid"], {-1, -0.5}], 
                  {GrayLevel[0], LineBox[{{-5, 0}, {5, 0}}]}, 
                  {GrayLevel[0.5], CircleBox[{-0.5481144155753137, 0.5}, 0.5],
                    DiskBox[{-0.5481144155753137, 0.5}, 0.1]}, 
                  {GrayLevel[0.5], 
                   DiskBox[{-0.5811733845243087, 1.9996356572754694`}, 0.1]}, 
                  {GrayLevel[0.5], Thickness[0.01], 
                   LineBox[{{-0.5481144155753137, 0.5}, {-0.5811733845243087, 
                    1.9996356572754694`}}]}, 
                  {RGBColor[0, 1, 0], 
                   DiskBox[{-0.570153728207977, 1.499757104850313}, 0.05]}, 
                  {RGBColor[0, 0, 1], Arrowheads[0.01], 
                   CircleBox[{-0.5481144155753137, 0.5}, 0.3333333333333333, 
                    NCache[{
                    Rational[-1, 2] 
                    Pi, -1.246052127421405}, {-1.5707963267948966`, \
-1.246052127421405}]], 
                   ArrowBox[{{-0.4449233333137844, 
                    0.18304146698218487`}, {-0.44175896032088124`, 
                    0.18408920840095777`}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-0.5481144155753137, 0.5}, {-0.7646105484909748,
                     0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-0.5481144155753137, 0.5}, {-0.3860549196825138,
                     0.5}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.005], Arrowheads[0.04], 
                   ArrowBox[{{-0.5481144155753137, 0.5}, {-0.6025510525981749,
                     0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-0.5481144155753137, 0.5}, 1.6, {
                    1.5928374240149952`, 1.6941492170911394`}], 
                   ArrowBox[{{-0.7290909008432347, 
                    2.0897318993402854`}, {-0.744978906133789, 
                    2.087842678717622}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Dashing[{Small, Small}], 
                   Arrowheads[0.02], 
                   CircleBox[{-0.5481144155753137, 0.5}, 1.7, {
                    1.5928374240149952`, 1.5612969305220565`}], 
                   ArrowBox[{{-0.5489654418759404, 
                    2.1999997869865267`}, {-0.5319656847869959, 
                    2.199923297826677}}]}, 
                  {RGBColor[0, 0, 1], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-0.5481144155753137, 0.5}, 1.8, {
                    1.5928374240149952`, 1.5326996093162004`}], 
                   ArrowBox[{{-0.4975469779308725, 
                    2.2992895637584505`}, {-0.479556910524195, 
                    2.298693934081384}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.003], Arrowheads[0.025], 
                   CircleBox[{-0.5481144155753137, 0.5}, 1.5, {
                    1.5928374240149952`, 1.592992066365665}], 
                   ArrowBox[{{-0.5664075714482643, 
                    1.9998884493348856`}, {-0.5814052913112833, 
                    1.999630527027485}}]}, 
                  {RGBColor[0, 0, 1], ArrowBox[{{0, 2.8}, {-1.5, 2.8}}]}},
                 Frame->True,
                 FrameTicks->Automatic,
                 ImageSize->800,
                 PlotRange->{{-3, 3}, {-1, 3}}], 306->
                GraphicsBox[{InsetBox[
                   TagBox[GridBox[{
                    {"\<\"t\"\>", "15.25`"},
                    {"\<\"driveIntent\"\>", 
                    RowBox[{"-", "0.5`"}]},
                    {"\<\"motorTorque\"\>", "11.252227068099053`"},
                    {"\<\"vw\"\>", 
                    RowBox[{"-", "0.30336276637616977`"}]},
                    {"\<\"controllerState\"\>", 
                    RowBox[{"\[LeftAssociation]", 
                    RowBox[{
                    RowBox[{"\<\"lastError\"\>", "\[Rule]", 
                    RowBox[{"-", "0.03614336775210546`"}]}], ",", 
                    
                    RowBox[{"\<\"lastInt\"\>", "\[Rule]", 
                    "0.17168893780292815`"}], ",", 
                    RowBox[{"\<\"initialized\"\>", "\[Rule]", "True"}]}], 
                    "\[RightAssociation]"}]}
                    },
                    AutoDelete->False,
                    
                    GridBoxItemSize->{
                    "Columns" -> {{Automatic}}, "Rows" -> {{Automatic}}}],
                    "Grid"], {-1, -0.5}], 
                  {GrayLevel[0], LineBox[{{-5, 0}, {5, 0}}]}, 
                  {GrayLevel[0.5], CircleBox[{-0.563259991005141, 0.5}, 0.5], 
                   DiskBox[{-0.563259991005141, 0.5}, 0.1]}, 
                  {GrayLevel[0.5], 
                   DiskBox[{-0.59756192948541, 1.9996077410498039`}, 0.1]}, 
                  {GrayLevel[0.5], Thickness[0.01], 
                   LineBox[{{-0.563259991005141, 0.5}, {-0.59756192948541, 
                    1.9996077410498039`}}]}, 
                  {RGBColor[0, 1, 0], 
                   DiskBox[{-0.586127949991987, 1.4997384940332026`}, 0.05]}, 
                  {RGBColor[0, 0, 1], Arrowheads[0.01], 
                   CircleBox[{-0.563259991005141, 0.5}, 0.3333333333333333, 
                    NCache[{
                    Rational[-1, 2] 
                    Pi, -1.233229514751925}, {-1.5707963267948966`, \
-1.233229514751925}]], 
                   ArrowBox[{{-0.456013266806198, 
                    0.18439066671008136`}, {-0.45286258836611537`, 
                    0.1854788964128672}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-0.563259991005141, 0.5}, {-0.7883045323671221, 
                    0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-0.563259991005141, 0.5}, {-0.3951104221239313, 
                    0.5}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.005], Arrowheads[0.04], 
                   ArrowBox[{{-0.563259991005141, 0.5}, {-0.6201549634859124, 
                    0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-0.563259991005141, 0.5}, 1.6, {
                    1.5936662793595182`, 1.6987872496712464`}], 
                   ArrowBox[{{-0.7516077316607619, 
                    2.088875428908737}, {-0.767486803830047, 
                    2.086912539783898}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Dashing[{Small, Small}], 
                   Arrowheads[0.02], 
                   CircleBox[{-0.563259991005141, 0.5}, 1.7, {
                    1.5936662793595182`, 1.5607020578053512`}], 
                   ArrowBox[{{-0.5630997337231511, 
                    2.199999992446354}, {-0.5461000251434004, 
                    2.199913390608952}}]}, 
                  {RGBColor[0, 0, 1], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-0.563259991005141, 0.5}, 1.8, {
                    1.5936662793595182`, 1.531153906758968}], 
                   ArrowBox[{{-0.5099114483960131, 
                    2.299209252144253}, {-0.4919223231461819, 
                    2.298585815896602}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.003], Arrowheads[0.025], 
                   CircleBox[{-0.563259991005141, 0.5}, 1.5, {
                    1.5936662793595182`, 1.593421016143312}], 
                   ArrowBox[{{-0.6121883443017058, 
                    1.9992017930364432`}, {-0.5971941298397795, 
                    1.9996161089497377`}}]}, 
                  {RGBColor[0, 0, 1], ArrowBox[{{0, 2.8}, {-1.5, 2.8}}]}},
                 Frame->True,
                 FrameTicks->Automatic,
                 ImageSize->800,
                 PlotRange->{{-3, 3}, {-1, 3}}], 307->
                GraphicsBox[{InsetBox[
                   TagBox[GridBox[{
                    {"\<\"t\"\>", "15.3`"},
                    {"\<\"driveIntent\"\>", 
                    RowBox[{"-", "0.5`"}]},
                    {"\<\"motorTorque\"\>", "11.674011342866734`"},
                    {"\<\"vw\"\>", 
                    RowBox[{"-", "0.3043167450691021`"}]},
                    {"\<\"controllerState\"\>", 
                    RowBox[{"\[LeftAssociation]", 
                    RowBox[{
                    RowBox[{"\<\"lastError\"\>", "\[Rule]", 
                    RowBox[{"-", "0.03503476237233888`"}]}], ",", 
                    
                    RowBox[{"\<\"lastInt\"\>", "\[Rule]", 
                    "0.1699100361317141`"}], ",", 
                    RowBox[{"\<\"initialized\"\>", "\[Rule]", "True"}]}], 
                    "\[RightAssociation]"}]}
                    },
                    AutoDelete->False,
                    
                    GridBoxItemSize->{
                    "Columns" -> {{Automatic}}, "Rows" -> {{Automatic}}}],
                    "Grid"], {-1, -0.5}], 
                  {GrayLevel[0], LineBox[{{-5, 0}, {5, 0}}]}, 
                  {GrayLevel[0.5], CircleBox[{-0.5784522918449676, 0.5}, 0.5],
                    DiskBox[{-0.5784522918449676, 0.5}, 0.1]}, 
                  {GrayLevel[0.5], 
                   DiskBox[{-0.6139875308966806, 1.9995790231880204`}, 0.1]}, 
                  {GrayLevel[0.5], Thickness[0.01], 
                   LineBox[{{-0.5784522918449676, 0.5}, {-0.6139875308966806, 
                    1.9995790231880204`}}]}, 
                  {RGBColor[0, 1, 0], 
                   DiskBox[{-0.6021424512127763, 1.4997193487920137`}, 0.05]}, 
                  {RGBColor[0, 0, 1], Arrowheads[0.01], 
                   CircleBox[{-0.5784522918449676, 0.5}, 0.3333333333333333, 
                    NCache[{
                    Rational[-1, 2] 
                    Pi, -1.2205759865088945`}, {-1.5707963267948966`, \
-1.2205759865088945`}]], 
                   ArrowBox[{{-0.46722068821950324`, 
                    0.1857729459992048}, {-0.46408403158424405`, 
                    0.18690095471872403`}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-0.5784522918449676, 0.5}, {-0.8119325187023023,
                     0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-0.5784522918449676, 
                    0.5}, {-0.40426036672003085`, 0.5}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.005], Arrowheads[0.04], 
                   ArrowBox[{{-0.5784522918449676, 0.5}, {-0.6377405935773656,
                     0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-0.5784522918449676, 0.5}, 1.6, {
                    1.5944887026354775`, 1.703389218900003}], 
                   ArrowBox[{{-0.7741099681017869, 
                    2.087991836793181}, {-0.7899798390034468, 
                    2.085955893709884}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Dashing[{Small, Small}], 
                   Arrowheads[0.02], 
                   CircleBox[{-0.5784522918449676, 0.5}, 1.7, {
                    1.5944887026354775`, 1.5601384890458052`}], 
                   ArrowBox[{{-0.5773339677521715, 
                    2.1999996321620845`}, {-0.5603343106781447, 
                    2.199903449834266}}]}, 
                  {RGBColor[0, 0, 1], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-0.5784522918449676, 0.5}, 1.8, {
                    1.5944887026354775`, 1.5296330840639956`}], 
                   ArrowBox[{{-0.5223675337219196, 
                    2.2991260378045446`}, {-0.504379377411252, 
                    2.29847524401847}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.003], Arrowheads[0.025], 
                   CircleBox[{-0.5784522918449676, 0.5}, 1.5, {
                    1.5944887026354775`, 1.5938922689635646`}], 
                   ArrowBox[{{-0.6280871427554935, 
                    1.9991785689420356`}, {-0.6130931252060542, 
                    1.9995999508749156`}}]}, 
                  {RGBColor[0, 0, 1], ArrowBox[{{0, 2.8}, {-1.5, 2.8}}]}},
                 Frame->True,
                 FrameTicks->Automatic,
                 ImageSize->800,
                 PlotRange->{{-3, 3}, {-1, 3}}], 308->
                GraphicsBox[{InsetBox[
                   TagBox[GridBox[{
                    {"\<\"t\"\>", "15.35`"},
                    {"\<\"driveIntent\"\>", 
                    RowBox[{"-", "0.5`"}]},
                    {"\<\"motorTorque\"\>", "12.085667540635107`"},
                    {"\<\"vw\"\>", 
                    RowBox[{"-", "0.30530940129903417`"}]},
                    {"\<\"controllerState\"\>", 
                    RowBox[{"\[LeftAssociation]", 
                    RowBox[{
                    RowBox[{"\<\"lastError\"\>", "\[Rule]", 
                    RowBox[{"-", "0.033930171789261`"}]}], ",", 
                    
                    RowBox[{"\<\"lastInt\"\>", "\[Rule]", 
                    "0.16818650060041979`"}], ",", 
                    RowBox[{"\<\"initialized\"\>", "\[Rule]", "True"}]}], 
                    "\[RightAssociation]"}]}
                    },
                    AutoDelete->False,
                    
                    GridBoxItemSize->{
                    "Columns" -> {{Automatic}}, "Rows" -> {{Automatic}}}],
                    "Grid"], {-1, -0.5}], 
                  {GrayLevel[0], LineBox[{{-5, 0}, {5, 0}}]}, 
                  {GrayLevel[0.5], CircleBox[{-0.5936932837609222, 0.5}, 0.5],
                    DiskBox[{-0.5936932837609222, 0.5}, 0.1]}, 
                  {GrayLevel[0.5], 
                   DiskBox[{-0.6304381042028878, 1.9995498718517792`}, 0.1]}, 
                  {GrayLevel[0.5], Thickness[0.01], 
                   LineBox[{{-0.5936932837609222, 0.5}, {-0.6304381042028878, 
                    1.9995498718517792`}}]}, 
                  {RGBColor[0, 1, 0], 
                   DiskBox[{-0.6181898307222327, 1.499699914567853}, 0.05]}, 
                  {RGBColor[0, 0, 1], Arrowheads[0.01], 
                   CircleBox[{-0.5936932837609222, 0.5}, 0.3333333333333333, 
                    NCache[{
                    Rational[-1, 2] 
                    Pi, -1.2082263005758433`}, {-1.5707963267948966`, \
-1.2082263005758433`}]], 
                   ArrowBox[{{-0.4785896554696132, 
                    0.18717054827704083`}, {-0.47546716822381974`, 
                    0.18833720671835275`}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-0.5936932837609222, 0.5}, {-0.8354066345736244,
                     0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-0.5936932837609222, 0.5}, {-0.413575553998793, 
                    0.5}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.005], Arrowheads[0.04], 
                   ArrowBox[{{-0.5936932837609222, 0.5}, {-0.6552889048114952,
                     0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-0.5936932837609222, 0.5}, 1.6, {
                    1.5952953244026644`, 1.7079026973361648`}], 
                   ArrowBox[{{-0.7965163096793666, 
                    2.0870925682383135`}, {-0.8123768297808577, 
                    2.084985017815672}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Dashing[{Small, Small}], 
                   Arrowheads[0.02], 
                   CircleBox[{-0.5936932837609222, 0.5}, 1.7, {
                    1.5952953244026644`, 1.5596090147173296`}], 
                   ArrowBox[{{-0.5916748537032919, 
                    2.1999988017466663`}, {-0.5746752499382041, 
                    2.1998936185507367`}}]}, 
                  {RGBColor[0, 0, 1], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-0.5936932837609222, 0.5}, 1.8, {
                    1.5952953244026644`, 1.5281527269546915`}], 
                   ArrowBox[{{-0.5349452390437067, 
                    2.299041041010989}, {-0.5169580658500288, 
                    2.2983636190526564`}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.003], Arrowheads[0.025], 
                   CircleBox[{-0.5936932837609222, 0.5}, 1.5, {
                    1.5952953244026644`, 1.594394254284118}], 
                   ArrowBox[{{-0.6440806940206143, 
                    1.9991534640880904`}, {-0.6290868898878765, 
                    1.9995823727442685`}}]}, 
                  {RGBColor[0, 0, 1], ArrowBox[{{0, 2.8}, {-1.5, 2.8}}]}},
                 Frame->True,
                 FrameTicks->Automatic,
                 ImageSize->800,
                 PlotRange->{{-3, 3}, {-1, 3}}], 309->
                GraphicsBox[{InsetBox[
                   TagBox[GridBox[{
                    {"\<\"t\"\>", "15.4`"},
                    {"\<\"driveIntent\"\>", 
                    RowBox[{"-", "0.5`"}]},
                    {"\<\"motorTorque\"\>", "12.483323680749663`"},
                    {"\<\"vw\"\>", 
                    RowBox[{"-", "0.3063391774422757`"}]},
                    {"\<\"controllerState\"\>", 
                    RowBox[{"\[LeftAssociation]", 
                    RowBox[{
                    RowBox[{"\<\"lastError\"\>", "\[Rule]", 
                    RowBox[{"-", "0.03283821469783828`"}]}], ",", 
                    
                    RowBox[{"\<\"lastInt\"\>", "\[Rule]", 
                    "0.1665179060165109`"}], ",", 
                    RowBox[{"\<\"initialized\"\>", "\[Rule]", "True"}]}], 
                    "\[RightAssociation]"}]}
                    },
                    AutoDelete->False,
                    
                    GridBoxItemSize->{
                    "Columns" -> {{Automatic}}, "Rows" -> {{Automatic}}}],
                    "Grid"], {-1, -0.5}], 
                  {GrayLevel[0], LineBox[{{-5, 0}, {5, 0}}]}, 
                  {GrayLevel[0.5], CircleBox[{-0.6089848621280378, 0.5}, 0.5],
                    DiskBox[{-0.6089848621280378, 0.5}, 0.1]}, 
                  {GrayLevel[0.5], 
                   DiskBox[{-0.6469033558228234, 1.9995206526873575`}, 0.1]}, 
                  {GrayLevel[0.5], Thickness[0.01], 
                   LineBox[{{-0.6089848621280378, 0.5}, {-0.6469033558228234, 
                    1.9995206526873575`}}]}, 
                  {RGBColor[0, 1, 0], 
                   DiskBox[{-0.6342638579245616, 1.4996804351249051`}, 0.05]}, 
                  {RGBColor[0, 0, 1], Arrowheads[0.01], 
                   CircleBox[{-0.6089848621280378, 0.5}, 0.3333333333333333, 
                    NCache[{
                    Rational[-1, 2] 
                    Pi, -1.1962966163724067`}, {-1.5707963267948966`, \
-1.1962966163724067`}]], 
                   ArrowBox[{{-0.49015755631329355`, 
                    0.1885659259106025}, {-0.4870492087935986, 
                    0.1897697507382387}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-0.6089848621280378, 0.5}, {-0.8586513357430311,
                     0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-0.6089848621280378, 0.5}, {-0.4231175796741069,
                     0.5}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.005], Arrowheads[0.04], 
                   ArrowBox[{{-0.6089848621280378, 0.5}, {-0.6727840532891002,
                     0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-0.6089848621280378, 0.5}, 1.6, {
                    1.5960780156953835`, 1.7122822020053738`}], 
                   ArrowBox[{{-0.8187566000645836, 
                    2.086189086446846}, {-0.8346077380660339, 
                    2.08401209523585}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Dashing[{Small, Small}], 
                   Arrowheads[0.02], 
                   CircleBox[{-0.6089848621280378, 0.5}, 1.7, {
                    1.5960780156953835`, 1.5591157721524023`}], 
                   ArrowBox[{{-0.6061279205805907, 
                    2.1999975993762444`}, {-0.5891283707642317, 
                    2.199884031265286}}]}, 
                  {RGBColor[0, 0, 1], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-0.6089848621280378, 0.5}, 1.8, {
                    1.5960780156953835`, 1.5267262174689964`}], 
                   ArrowBox[{{-0.5476705289451042, 
                    2.2989554059359927`}, {-0.5296843404012577, 
                    2.298252325802433}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.003], Arrowheads[0.025], 
                   CircleBox[{-0.6089848621280378, 0.5}, 1.5, {
                    1.5960780156953835`, 1.5949164472056636`}], 
                   ArrowBox[{{-0.660155112809452, 
                    1.9991269477416518`}, {-0.6451615346940643, 
                    1.9995636859973807`}}]}, 
                  {RGBColor[0, 0, 1], ArrowBox[{{0, 2.8}, {-1.5, 2.8}}]}},
                 Frame->True,
                 FrameTicks->Automatic,
                 ImageSize->800,
                 PlotRange->{{-3, 3}, {-1, 3}}], 310->
                GraphicsBox[{InsetBox[
                   TagBox[GridBox[{
                    {"\<\"t\"\>", "15.450000000000001`"},
                    {"\<\"driveIntent\"\>", 
                    RowBox[{"-", "0.5`"}]},
                    {"\<\"motorTorque\"\>", "12.863694041901468`"},
                    {"\<\"vw\"\>", 
                    RowBox[{"-", "0.30740424889625334`"}]},
                    {"\<\"controllerState\"\>", 
                    RowBox[{"\[LeftAssociation]", 
                    RowBox[{
                    RowBox[{"\<\"lastError\"\>", "\[Rule]", 
                    RowBox[{"-", "0.03176641218758223`"}]}], ",", 
                    
                    RowBox[{"\<\"lastInt\"\>", "\[Rule]", 
                    "0.16490342445334144`"}], ",", 
                    RowBox[{"\<\"initialized\"\>", "\[Rule]", "True"}]}], 
                    "\[RightAssociation]"}]}
                    },
                    AutoDelete->False,
                    
                    GridBoxItemSize->{
                    "Columns" -> {{Automatic}}, "Rows" -> {{Automatic}}}],
                    "Grid"], {-1, -0.5}], 
                  {GrayLevel[0], LineBox[{{-5, 0}, {5, 0}}]}, 
                  {GrayLevel[0.5], CircleBox[{-0.6243288374228247, 0.5}, 0.5],
                    DiskBox[{-0.6243288374228247, 0.5}, 0.1]}, 
                  {GrayLevel[0.5], 
                   DiskBox[{-0.6633746727196667, 1.9994917214663013`}, 0.1]}, 
                  {GrayLevel[0.5], Thickness[0.01], 
                   LineBox[{{-0.6243288374228247, 0.5}, {-0.6633746727196667, 
                    1.9994917214663013`}}]}, 
                  {RGBColor[0, 1, 0], 
                   DiskBox[{-0.650359394287386, 1.499661147644201}, 0.05]}, 
                  {RGBColor[0, 0, 1], Arrowheads[0.01], 
                   CircleBox[{-0.6243288374228247, 0.5}, 0.3333333333333333, 
                    NCache[{
                    Rational[-1, 2] 
                    Pi, -1.1848855055378524`}, {-1.5707963267948966`, \
-1.1848855055378524`}]], 
                   ArrowBox[{{-0.5019555363689054, 
                    0.1899421242730997}, {-0.49886112790175474`, 
                    0.19118133965278672`}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-0.6243288374228247, 0.5}, {-0.8816027182608541,
                     0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-0.6243288374228247, 0.5}, {-0.4329392918473036,
                     0.5}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.005], Arrowheads[0.04], 
                   ArrowBox[{{-0.6243288374228247, 0.5}, {-0.690213172685333, 
                    0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-0.6243288374228247, 0.5}, 1.6, {
                    1.5968298242298682`, 1.7164888369529225`}], 
                   ArrowBox[{{-0.8407712180903502, 
                    2.0852926215216465`}, {-0.8566130580626015, 
                    2.0830489698179813`}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Dashing[{Small, Small}], 
                   Arrowheads[0.02], 
                   CircleBox[{-0.6243288374228247, 0.5}, 1.7, {
                    1.5968298242298682`, 1.5586602992727627`}], 
                   ArrowBox[{{-0.6206975933965259, 
                    2.1999961217799355`}, {-0.6036980970706851, 
                    2.1998748108471173`}}]}, 
                  {RGBColor[0, 0, 1], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-0.6243288374228247, 0.5}, 1.8, {
                    1.5968298242298682`, 1.5253648573304155`}], 
                   ArrowBox[{{-0.5605655356325908, 
                    2.298870268069604}, {-0.5425803209006287, 
                    2.298142702914989}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.003], Arrowheads[0.025], 
                   CircleBox[{-0.6243288374228247, 0.5}, 1.5, {
                    1.5968298242298682`, 1.5954493492370636`}], 
                   ArrowBox[{{-0.6762979685964243, 
                    1.9990994661479475`}, {-0.6613046253487016, 
                    1.9995441944495203`}}]}, 
                  {RGBColor[0, 0, 1], ArrowBox[{{0, 2.8}, {-1.5, 2.8}}]}},
                 Frame->True,
                 FrameTicks->Automatic,
                 ImageSize->800,
                 PlotRange->{{-3, 3}, {-1, 3}}], 311->
                GraphicsBox[{InsetBox[
                   TagBox[GridBox[{
                    {"\<\"t\"\>", "15.5`"},
                    {"\<\"driveIntent\"\>", 
                    RowBox[{"-", "0.5`"}]},
                    {"\<\"motorTorque\"\>", "13.224043662275207`"},
                    {"\<\"vw\"\>", 
                    RowBox[{"-", "0.30850257265626363`"}]},
                    {"\<\"controllerState\"\>", 
                    RowBox[{"\[LeftAssociation]", 
                    RowBox[{
                    RowBox[{"\<\"lastError\"\>", "\[Rule]", 
                    RowBox[{"-", "0.03072124088757345`"}]}], ",", 
                    
                    RowBox[{"\<\"lastInt\"\>", "\[Rule]", 
                    "0.16334187879393594`"}], ",", 
                    RowBox[{"\<\"initialized\"\>", "\[Rule]", "True"}]}], 
                    "\[RightAssociation]"}]}
                    },
                    AutoDelete->False,
                    
                    GridBoxItemSize->{
                    "Columns" -> {{Automatic}}, "Rows" -> {{Automatic}}}],
                    "Grid"], {-1, -0.5}], 
                  {GrayLevel[0], LineBox[{{-5, 0}, {5, 0}}]}, 
                  {GrayLevel[0.5], CircleBox[{-0.6397269231271868, 0.5}, 0.5],
                    DiskBox[{-0.6397269231271868, 0.5}, 0.1]}, 
                  {GrayLevel[0.5], 
                   DiskBox[{-0.6798450084388086, 1.9994634171032415`}, 0.1]}, 
                  {GrayLevel[0.5], Thickness[0.01], 
                   LineBox[{{-0.6397269231271868, 0.5}, {-0.6798450084388086, 
                    1.9994634171032415`}}]}, 
                  {RGBColor[0, 1, 0], 
                   DiskBox[{-0.6664723133349346, 1.4996422780688277`}, 0.05]}, 
                  {RGBColor[0, 0, 1], Arrowheads[0.01], 
                   CircleBox[{-0.6397269231271868, 0.5}, 0.3333333333333333, 
                    NCache[{
                    Rational[-1, 2] 
                    Pi, -1.1740750169266403`}, {-1.5707963267948966`, \
-1.1740750169266403`}]], 
                   ArrowBox[{{-0.5140089608357301, 
                    0.19128313122151075`}, {-0.5109281294462317, 
                    0.19255572560634365`}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-0.6397269231271868, 0.5}, {-0.9042077963726909,
                     0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-0.6397269231271868, 
                    0.5}, {-0.44308528066691766`, 0.5}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.005], Arrowheads[0.04], 
                   ArrowBox[{{-0.6397269231271868, 0.5}, {-0.7075661539124218,
                     0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-0.6397269231271868, 0.5}, 1.6, {
                    1.5975449065965892`, 1.7204899132594689`}], 
                   ArrowBox[{{-0.8625104311495041, 
                    2.0844139322012003`}, {-0.8783431673212728, 
                    2.082106914214936}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Dashing[{Small, Small}], 
                   Arrowheads[0.02], 
                   CircleBox[{-0.6397269231271868, 0.5}, 1.7, {
                    1.5975449065965892`, 1.5582435879888936`}], 
                   ArrowBox[{{-0.6353872718701904, 
                    2.1999944609989077`}, {-0.6183878275719495, 
                    2.1998660661948883`}}]}, 
                  {RGBColor[0, 0, 1], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-0.6397269231271868, 0.5}, 1.8, {
                    1.5975449065965892`, 1.5240779973617269`}], 
                   ArrowBox[{{-0.5736487806353041, 
                    2.298786724179613}, {-0.5556645170693886, 
                    2.2980360151809456`}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.003], Arrowheads[0.025], 
                   CircleBox[{-0.6397269231271868, 0.5}, 1.5, {
                    1.5975449065965892`, 1.5959844615331045`}], 
                   ArrowBox[{{-0.6924982333792721, 
                    1.999071442198229}, {-0.6775051302577563, 
                    1.9995241935580765`}}]}, 
                  {RGBColor[0, 0, 1], ArrowBox[{{0, 2.8}, {-1.5, 2.8}}]}},
                 Frame->True,
                 FrameTicks->Automatic,
                 ImageSize->800,
                 PlotRange->{{-3, 3}, {-1, 3}}], 312->
                GraphicsBox[{InsetBox[
                   TagBox[GridBox[{
                    {"\<\"t\"\>", "15.55`"},
                    {"\<\"driveIntent\"\>", 
                    RowBox[{"-", "0.5`"}]},
                    {"\<\"motorTorque\"\>", "13.562151573065623`"},
                    {"\<\"vw\"\>", 
                    RowBox[{"-", "0.3096319324619813`"}]},
                    {"\<\"controllerState\"\>", 
                    RowBox[{"\[LeftAssociation]", 
                    RowBox[{
                    RowBox[{"\<\"lastError\"\>", "\[Rule]", 
                    RowBox[{"-", "0.029708190081679534`"}]}], ",", 
                    
                    RowBox[{"\<\"lastInt\"\>", "\[Rule]", 
                    "0.16183179351198237`"}], ",", 
                    RowBox[{"\<\"initialized\"\>", "\[Rule]", "True"}]}], 
                    "\[RightAssociation]"}]}
                    },
                    AutoDelete->False,
                    
                    GridBoxItemSize->{
                    "Columns" -> {{Automatic}}, "Rows" -> {{Automatic}}}],
                    "Grid"], {-1, -0.5}], 
                  {GrayLevel[0], LineBox[{{-5, 0}, {5, 0}}]}, 
                  {GrayLevel[0.5], CircleBox[{-0.6551807259739489, 0.5}, 0.5],
                    DiskBox[{-0.6551807259739489, 0.5}, 0.1]}, 
                  {GrayLevel[0.5], 
                   DiskBox[{-0.696308767096297, 1.9994360553999755`}, 0.1]}, 
                  {GrayLevel[0.5], Thickness[0.01], 
                   LineBox[{{-0.6551807259739489, 0.5}, {-0.696308767096297, 
                    1.9994360553999755`}}]}, 
                  {RGBColor[0, 1, 0], 
                   DiskBox[{-0.6825994200555142, 1.4996240369333171`}, 0.05]}, 
                  {RGBColor[0, 0, 1], Arrowheads[0.01], 
                   CircleBox[{-0.6551807259739489, 0.5}, 0.3333333333333333, 
                    NCache[{
                    Rational[-1, 2] 
                    Pi, -1.1639317796029278`}, {-1.5707963267948966`, \
-1.1639317796029278`}]], 
                   ArrowBox[{{-0.5263378961193674, 
                    0.1925741775546913}, {-0.5232701312201037, 
                    0.1938779555425681}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-0.6551807259739489, 0.5}, {-0.9264237574352614,
                     0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-0.6551807259739489, 0.5}, {-0.4535923920108268,
                     0.5}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.005], Arrowheads[0.04], 
                   ArrowBox[{{-0.6551807259739489, 0.5}, {-0.7248354234721393,
                     0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-0.6551807259739489, 0.5}, 1.6, {
                    1.5982184575320602`, 1.724258552679708}], 
                   ArrowBox[{{-0.8839337226094964, 
                    2.08356309205862}, {-0.8997576520497019, 
                    2.081196422722786}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Dashing[{Small, Small}], 
                   Arrowheads[0.02], 
                   CircleBox[{-0.6551807259739489, 0.5}, 1.7, {
                    1.5982184575320602`, 1.5578661368156244`}], 
                   ArrowBox[{{-0.6501994101374673, 
                    2.1999927018939047`}, {-0.6332000155129448, 
                    2.199857890638988}}]}, 
                  {RGBColor[0, 0, 1], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-0.6551807259739489, 0.5}, 1.8, {
                    1.5982184575320602`, 1.522873171015029}], 
                   ArrowBox[{{-0.5869354063287362, 
                    2.298705805946743}, {-0.5689520602896176, 
                    2.297933429583615}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.003], Arrowheads[0.025], 
                   CircleBox[{-0.6551807259739489, 0.5}, 1.5, {
                    1.5982184575320602`, 1.59651425172382}], 
                   ArrowBox[{{-0.7087462221282771, 
                    1.999043274099097}, {-0.6937533609740938, 
                    1.9995039685939966`}}]}, 
                  {RGBColor[0, 0, 1], ArrowBox[{{0, 2.8}, {-1.5, 2.8}}]}},
                 Frame->True,
                 FrameTicks->Automatic,
                 ImageSize->800,
                 PlotRange->{{-3, 3}, {-1, 3}}], 313->
                GraphicsBox[{InsetBox[
                   TagBox[GridBox[{
                    {"\<\"t\"\>", "15.6`"},
                    {"\<\"driveIntent\"\>", 
                    RowBox[{"-", "0.5`"}]},
                    {"\<\"motorTorque\"\>", "13.876273262168375`"},
                    {"\<\"vw\"\>", 
                    RowBox[{"-", "0.31078998047331696`"}]},
                    {"\<\"controllerState\"\>", 
                    RowBox[{"\[LeftAssociation]", 
                    RowBox[{
                    RowBox[{"\<\"lastError\"\>", "\[Rule]", 
                    RowBox[{"-", "0.02873182180284315`"}]}], ",", 
                    
                    RowBox[{"\<\"lastInt\"\>", "\[Rule]", 
                    "0.16037144251713928`"}], ",", 
                    RowBox[{"\<\"initialized\"\>", "\[Rule]", "True"}]}], 
                    "\[RightAssociation]"}]}
                    },
                    AutoDelete->False,
                    
                    GridBoxItemSize->{
                    "Columns" -> {{Automatic}}, "Rows" -> {{Automatic}}}],
                    "Grid"], {-1, -0.5}], 
                  {GrayLevel[0], LineBox[{{-5, 0}, {5, 0}}]}, 
                  {GrayLevel[0.5], CircleBox[{-0.6706917383613036, 0.5}, 0.5],
                    DiskBox[{-0.6706917383613036, 0.5}, 0.1]}, 
                  {GrayLevel[0.5], 
                   DiskBox[{-0.7127616866579213, 1.9994099237534477`}, 0.1]}, 
                  {GrayLevel[0.5], Thickness[0.01], 
                   LineBox[{{-0.6706917383613036, 0.5}, {-0.7127616866579213, 
                    1.9994099237534477`}}]}, 
                  {RGBColor[0, 1, 0], 
                   DiskBox[{-0.6987383705590487, 1.4996066158356318`}, 0.05]}, 
                  {RGBColor[0, 0, 1], Arrowheads[0.01], 
                   CircleBox[{-0.6706917383613036, 0.5}, 0.3333333333333333, 
                    NCache[{
                    Rational[-1, 2] 
                    Pi, -1.1545081289298453`}, {-1.5707963267948966`, \
-1.1545081289298453`}]], 
                   ArrowBox[{{-0.5389575987409023, 
                    0.193801979807864}, {-0.5359022562238211, 
                    0.19513460902191537`}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-0.6706917383613036, 0.5}, {-0.9482172036046711,
                     0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-0.6706917383613036, 
                    0.5}, {-0.46449025744951444`, 0.5}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.005], Arrowheads[0.04], 
                   ArrowBox[{{-0.6706917383613036, 0.5}, {-0.742015722692882, 
                    0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-0.6706917383613036, 0.5}, 1.6, {
                    1.5988466372717625`, 1.727773280541458}], 
                   ArrowBox[{{-0.9050091039309173, 
                    2.0827493080688795`}, {-0.9208246174507271, 
                    2.0803270366599556`}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Dashing[{Small, Small}], 
                   Arrowheads[0.02], 
                   CircleBox[{-0.6706917383613036, 0.5}, 1.7, {
                    1.5988466372717625`, 1.5575280023011697`}], 
                   ArrowBox[{{-0.6651355966137136, 
                    2.1999909203548356`}, {-0.648136248545341, 
                    2.199850361025688}}]}, 
                  {RGBColor[0, 0, 1], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-0.6706917383613036, 0.5}, 1.8, {
                    1.5988466372717625`, 1.5217562302597158`}], 
                   ArrowBox[{{-0.600437413882291, 
                    2.2986284579901426`}, {-0.5824549417592517, 
                    2.2978359957808747`}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.003], Arrowheads[0.025], 
                   CircleBox[{-0.6706917383613036, 0.5}, 1.5, {
                    1.5988466372717625`, 1.5970321153783102`}], 
                   ArrowBox[{{-0.7250335273264261, 
                    1.9990153334679637`}, {-0.7100409067595803, 
                    1.9994837921586095`}}]}, 
                  {RGBColor[0, 0, 1], ArrowBox[{{0, 2.8}, {-1.5, 2.8}}]}},
                 Frame->True,
                 FrameTicks->Automatic,
                 ImageSize->800,
                 PlotRange->{{-3, 3}, {-1, 3}}], 314->
                GraphicsBox[{InsetBox[
                   TagBox[GridBox[{
                    {"\<\"t\"\>", "15.65`"},
                    {"\<\"driveIntent\"\>", 
                    RowBox[{"-", "0.5`"}]},
                    {"\<\"motorTorque\"\>", "14.165102817932812`"},
                    {"\<\"vw\"\>", 
                    RowBox[{"-", "0.31197427547092166`"}]},
                    {"\<\"controllerState\"\>", 
                    RowBox[{"\[LeftAssociation]", 
                    RowBox[{
                    RowBox[{"\<\"lastError\"\>", "\[Rule]", 
                    RowBox[{"-", "0.027795832996468577`"}]}], ",", 
                    
                    RowBox[{"\<\"lastInt\"\>", "\[Rule]", 
                    "0.15895889393924548`"}], ",", 
                    RowBox[{"\<\"initialized\"\>", "\[Rule]", "True"}]}], 
                    "\[RightAssociation]"}]}
                    },
                    AutoDelete->False,
                    
                    GridBoxItemSize->{
                    "Columns" -> {{Automatic}}, "Rows" -> {{Automatic}}}],
                    "Grid"], {-1, -0.5}], 
                  {GrayLevel[0], LineBox[{{-5, 0}, {5, 0}}]}, 
                  {GrayLevel[0.5], CircleBox[{-0.6862613327623756, 0.5}, 0.5],
                    DiskBox[{-0.6862613327623756, 0.5}, 0.1]}, 
                  {GrayLevel[0.5], 
                   DiskBox[{-0.7292007227088098, 1.9993852769688742`}, 0.1]}, 
                  {GrayLevel[0.5], Thickness[0.01], 
                   LineBox[{{-0.6862613327623756, 0.5}, {-0.7292007227088098, 
                    1.9993852769688742`}}]}, 
                  {RGBColor[0, 1, 0], 
                   DiskBox[{-0.714887592726665, 1.4995901846459163`}, 0.05]}, 
                  {RGBColor[0, 0, 1], Arrowheads[0.01], 
                   CircleBox[{-0.6862613327623756, 0.5}, 0.3333333333333333, 
                    NCache[{
                    Rational[-1, 2] 
                    Pi, -1.1458432422569123`}, {-1.5707963267948966`, \
-1.1458432422569123`}]], 
                   ArrowBox[{{-0.5518790004833937, 
                    0.19495492155687477`}, {-0.5488353196001762, 
                    0.1963139746095417}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-0.6862613327623756, 0.5}, {-0.9695633891210318,
                     0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-0.6862613327623756, 
                    0.5}, {-0.47580183344838267`, 0.5}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.005], Arrowheads[0.04], 
                   ArrowBox[{{-0.6862613327623756, 0.5}, {-0.7591038898070389,
                     0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-0.6862613327623756, 0.5}, 1.6, {
                    1.5994264978939519`, 1.7310176130246389`}], 
                   ArrowBox[{{-0.9257124211414398, 
                    2.0819807762024425`}, {-0.9415199927866719, 
                    2.079507206847313}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Dashing[{Small, Small}], 
                   Arrowheads[0.02], 
                   CircleBox[{-0.6862613327623756, 0.5}, 1.7, {
                    1.5994264978939519`, 1.5572288489176593`}], 
                   ArrowBox[{{-0.680196633235227, 
                    2.1999891821478292`}, {-0.6631973279763119, 
                    2.199843537412555}}]}, 
                  {RGBColor[0, 0, 1], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-0.6862613327623756, 0.5}, 1.8, {
                    1.5994264978939519`, 1.5207314822387696`}], 
                   ArrowBox[{{-0.6141639045405316, 
                    2.2985555206453308`}, {-0.5961822539332036, 
                    2.2977446313526535`}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.003], Arrowheads[0.025], 
                   CircleBox[{-0.6862613327623756, 0.5}, 1.5, {
                    1.5994264978939519`, 1.5975323330740012`}], 
                   ArrowBox[{{-0.7413529488935008, 
                    1.9989879632045284`}, {-0.7263605645336818, 
                    1.9994639214103656`}}]}, 
                  {RGBColor[0, 0, 1], ArrowBox[{{0, 2.8}, {-1.5, 2.8}}]}},
                 Frame->True,
                 FrameTicks->Automatic,
                 ImageSize->800,
                 PlotRange->{{-3, 3}, {-1, 3}}], 315->
                GraphicsBox[{InsetBox[
                   TagBox[GridBox[{
                    {"\<\"t\"\>", "15.700000000000001`"},
                    {"\<\"driveIntent\"\>", 
                    RowBox[{"-", "0.5`"}]},
                    {"\<\"motorTorque\"\>", "14.427735156728751`"},
                    {"\<\"vw\"\>", 
                    RowBox[{"-", "0.3131823176087656`"}]},
                    {"\<\"controllerState\"\>", 
                    RowBox[{"\[LeftAssociation]", 
                    RowBox[{
                    RowBox[{"\<\"lastError\"\>", "\[Rule]", 
                    RowBox[{"-", "0.0269031189235633`"}]}], ",", 
                    
                    RowBox[{"\<\"lastInt\"\>", "\[Rule]", 
                    "0.1575920517694587`"}], ",", 
                    RowBox[{"\<\"initialized\"\>", "\[Rule]", "True"}]}], 
                    "\[RightAssociation]"}]}
                    },
                    AutoDelete->False,
                    
                    GridBoxItemSize->{
                    "Columns" -> {{Automatic}}, "Rows" -> {{Automatic}}}],
                    "Grid"], {-1, -0.5}], 
                  {GrayLevel[0], LineBox[{{-5, 0}, {5, 0}}]}, 
                  {GrayLevel[0.5], CircleBox[{-0.7018907579567129, 0.5}, 0.5],
                    DiskBox[{-0.7018907579567129, 0.5}, 0.1]}, 
                  {GrayLevel[0.5], 
                   DiskBox[{-0.7456239337769827, 1.99936233423835}, 0.1]}, 
                  {GrayLevel[0.5], Thickness[0.01], 
                   LineBox[{{-0.7018907579567129, 0.5}, {-0.7456239337769827, 
                    1.99936233423835}}]}, 
                  {RGBColor[0, 1, 0], 
                   DiskBox[{-0.7310462085035594, 1.4995748894922332`}, 0.05]}, 
                  {RGBColor[0, 0, 1], Arrowheads[0.01], 
                   CircleBox[{-0.7018907579567129, 0.5}, 0.3333333333333333, 
                    NCache[{
                    Rational[-1, 2] 
                    Pi, -1.137964272093034}, {-1.5707963267948966`, \
-1.137964272093034}]], 
                   ArrowBox[{{-0.5651091805559464, 
                    0.19602317326008345`}, {-0.5620763019729763, 
                    0.19740616495195595`}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-0.7018907579567129, 0.5}, {-0.9904454610912878,
                     0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-0.7018907579567129, 0.5}, {-0.4875439434639399,
                     0.5}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.005], Arrowheads[0.04], 
                   ArrowBox[{{-0.7018907579567129, 0.5}, {-0.7760986465985149,
                     0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-0.7018907579567129, 0.5}, 1.6, {
                    1.5999559094738587`, 1.7339796434294366`}], 
                   ArrowBox[{{-0.9460266642165251, 
                    2.0812645759880604`}, {-0.9618268397400376, 
                    2.078744195044635}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Dashing[{Small, Small}], 
                   Arrowheads[0.02], 
                   CircleBox[{-0.7018907579567129, 0.5}, 1.7, {
                    1.5999559094738587`, 1.55696799710795}], 
                   ArrowBox[{{-0.6953826133862759, 
                    2.199987542323252}, {-0.6783833466974006, 
                    2.1998374632934423`}}]}, 
                  {RGBColor[0, 0, 1], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-0.7018907579567129, 0.5}, 1.8, {
                    1.5999559094738587`, 1.5198018252698102`}], 
                   ArrowBox[{{-0.6281213214574217, 
                    2.298487717566783}, {-0.6101404324692958, 
                    2.297660111860124}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.003], Arrowheads[0.025], 
                   CircleBox[{-0.7018907579567129, 0.5}, 1.5, {
                    1.5999559094738587`, 1.59801002396692}], 
                   ArrowBox[{{-0.7576984206735842, 
                    1.9989614754162564`}, {-0.7427062653852023, 
                    1.9994445952929887`}}]}, 
                  {RGBColor[0, 0, 1], ArrowBox[{{0, 2.8}, {-1.5, 2.8}}]}},
                 Frame->True,
                 FrameTicks->Automatic,
                 ImageSize->800,
                 PlotRange->{{-3, 3}, {-1, 3}}], 316->
                GraphicsBox[{InsetBox[
                   TagBox[GridBox[{
                    {"\<\"t\"\>", "15.75`"},
                    {"\<\"driveIntent\"\>", 
                    RowBox[{"-", "0.5`"}]},
                    {"\<\"motorTorque\"\>", "14.663628692628237`"},
                    {"\<\"vw\"\>", 
                    RowBox[{"-", "0.3144115797748776`"}]},
                    {"\<\"controllerState\"\>", 
                    RowBox[{"\[LeftAssociation]", 
                    RowBox[{
                    RowBox[{"\<\"lastError\"\>", "\[Rule]", 
                    RowBox[{"-", "0.026055837053920804`"}]}], ",", 
                    
                    RowBox[{"\<\"lastInt\"\>", "\[Rule]", 
                    "0.1562686943157827`"}], ",", 
                    RowBox[{"\<\"initialized\"\>", "\[Rule]", "True"}]}], 
                    "\[RightAssociation]"}]}
                    },
                    AutoDelete->False,
                    
                    GridBoxItemSize->{
                    "Columns" -> {{Automatic}}, "Rows" -> {{Automatic}}}],
                    "Grid"], {-1, -0.5}], 
                  {GrayLevel[0], LineBox[{{-5, 0}, {5, 0}}]}, 
                  {GrayLevel[0.5], CircleBox[{-0.717581136912628, 0.5}, 0.5], 
                   DiskBox[{-0.717581136912628, 0.5}, 0.1]}, 
                  {GrayLevel[0.5], 
                   DiskBox[{-0.7620303691422449, 1.999341277279525}, 0.1]}, 
                  {GrayLevel[0.5], Thickness[0.01], 
                   LineBox[{{-0.717581136912628, 0.5}, {-0.7620303691422449, 
                    1.999341277279525}}]}, 
                  {RGBColor[0, 1, 0], 
                   DiskBox[{-0.7472139583990393, 1.4995608515196834`}, 0.05]}, 
                  {RGBColor[0, 0, 1], Arrowheads[0.01], 
                   CircleBox[{-0.717581136912628, 0.5}, 0.3333333333333333, 
                    NCache[{
                    Rational[-1, 2] 
                    Pi, -1.1308874660160495`}, {-1.5707963267948966`, \
-1.1308874660160495`}]], 
                   ArrowBox[{{-0.5786518174901258, 
                    0.19699875360667002`}, {-0.5756288019342319, 
                    0.19840317358219384`}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-0.717581136912628, 0.5}, {-1.0108537107651927`,
                     0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-0.717581136912628, 0.5}, {-0.4997278172763656, 
                    0.5}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.005], Arrowheads[0.04], 
                   ArrowBox[{{-0.717581136912628, 0.5}, {-0.7930003911289304, 
                    0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-0.717581136912628, 0.5}, 1.6, {
                    1.6004334867795897`, 1.7366516315435183`}], 
                   ArrowBox[{{-0.9659412867910124, 
                    2.0806066037924764`}, {-0.9817346714918096, 
                    2.078044014015238}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Dashing[{Small, Small}], 
                   Arrowheads[0.02], 
                   CircleBox[{-0.717581136912628, 0.5}, 1.7, {
                    1.6004334867795897`, 1.5567444692311347`}], 
                   ArrowBox[{{-0.7106929979019964, 
                    2.1999860451018325`}, {-0.6936937651846494, 
                    2.199832166265814}}]}, 
                  {RGBColor[0, 0, 1], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-0.717581136912628, 0.5}, 1.8, {
                    1.6004334867795897`, 1.518968882931655}], 
                   ArrowBox[{{-0.6423136896123001, 
                    2.298425647997963}, {-0.6243334962094335, 
                    2.297583065536413}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.003], Arrowheads[0.025], 
                   CircleBox[{-0.717581136912628, 0.5}, 1.5, {
                    1.6004334867795897`, 1.5984610966886763`}], 
                   ArrowBox[{{-0.7740649345615971, 
                    1.9989361496084983`}, {-0.759072998720607, 
                    1.999426031988143}}]}, 
                  {RGBColor[0, 0, 1], ArrowBox[{{0, 2.8}, {-1.5, 2.8}}]}},
                 Frame->True,
                 FrameTicks->Automatic,
                 ImageSize->800,
                 PlotRange->{{-3, 3}, {-1, 3}}], 317->
                GraphicsBox[{InsetBox[
                   TagBox[GridBox[{
                    {"\<\"t\"\>", "15.8`"},
                    {"\<\"driveIntent\"\>", 
                    RowBox[{"-", "0.5`"}]},
                    {"\<\"motorTorque\"\>", "14.872568763399663`"},
                    {"\<\"vw\"\>", 
                    RowBox[{"-", "0.31565953564168897`"}]},
                    {"\<\"controllerState\"\>", 
                    RowBox[{"\[LeftAssociation]", 
                    RowBox[{
                    RowBox[{"\<\"lastError\"\>", "\[Rule]", 
                    RowBox[{"-", "0.02525547077776453`"}]}], ",", 
                    
                    RowBox[{"\<\"lastInt\"\>", "\[Rule]", 
                    "0.15498650946592757`"}], ",", 
                    RowBox[{"\<\"initialized\"\>", "\[Rule]", "True"}]}], 
                    "\[RightAssociation]"}]}
                    },
                    AutoDelete->False,
                    
                    GridBoxItemSize->{
                    "Columns" -> {{Automatic}}, "Rows" -> {{Automatic}}}],
                    "Grid"], {-1, -0.5}], 
                  {GrayLevel[0], LineBox[{{-5, 0}, {5, 0}}]}, 
                  {GrayLevel[0.5], CircleBox[{-0.7333334661527732, 0.5}, 0.5],
                    DiskBox[{-0.7333334661527732, 0.5}, 0.1]}, 
                  {GrayLevel[0.5], 
                   DiskBox[{-0.7784199599347595, 1.999322249577604}, 0.1]}, 
                  {GrayLevel[0.5], Thickness[0.01], 
                   LineBox[{{-0.7333334661527732, 0.5}, {-0.7784199599347595, 
                    1.999322249577604}}]}, 
                  {RGBColor[0, 1, 0], 
                   DiskBox[{-0.7633911286740974, 1.4995481663850692`}, 0.05]}, 
                  {RGBColor[0, 0, 1], Arrowheads[0.01], 
                   CircleBox[{-0.7333334661527732, 0.5}, 0.3333333333333333, 
                    NCache[{
                    Rational[-1, 2] 
                    Pi, -1.1246192638929067`}, {-1.5707963267948966`, \
-1.1246192638929067`}]], 
                   ArrowBox[{{-0.5925076153938775, 
                    0.19787553745328013`}, {-0.5894934623560963, 
                    0.1992988785872538}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-0.7333334661527732, 
                    0.5}, {-1.0307848414207665`, 0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-0.7333334661527732, 0.5}, {-0.5123596225727663,
                     0.5}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.005], Arrowheads[0.04], 
                   ArrowBox[{{-0.7333334661527732, 0.5}, {-0.8098109978407596,
                     0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-0.7333334661527732, 0.5}, 1.6, {
                    1.6008585171553344`, 1.7390295997325584`}], 
                   ArrowBox[{{-0.9854515425072199, 
                    2.0800115428614228`}, {-1.0012387888031247`, 
                    2.0774114041985405`}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Dashing[{Small, Small}], 
                   Arrowheads[0.02], 
                   CircleBox[{-0.7333334661527732, 0.5}, 1.7, {
                    1.6008585171553344`, 1.5565570331936367`}], 
                   ArrowBox[{{-0.7261266886169094, 
                    2.1999847241541755`}, {-0.7091274850406123, 
                    2.1998276590520573`}}]}, 
                  {RGBColor[0, 0, 1], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-0.7333334661527732, 0.5}, 1.8, {
                    1.6008585171553344`, 1.5182331351364475`}], 
                   ArrowBox[{{-0.656742851638533, 
                    2.2983697833783605`}, {-0.638763283030361, 
                    2.2975139722584057`}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.003], Arrowheads[0.025], 
                   CircleBox[{-0.7333334661527732, 0.5}, 1.5, {
                    1.6008585171553344`, 1.5988821983251182`}], 
                   ArrowBox[{{-0.7904484632405699, 
                    1.9989122312889642`}, {-0.7754567350190799, 
                    1.9994084267537038`}}]}, 
                  {RGBColor[0, 0, 1], ArrowBox[{{0, 2.8}, {-1.5, 2.8}}]}},
                 Frame->True,
                 FrameTicks->Automatic,
                 ImageSize->800,
                 PlotRange->{{-3, 3}, {-1, 3}}], 318->
                GraphicsBox[{InsetBox[
                   TagBox[GridBox[{
                    {"\<\"t\"\>", "15.85`"},
                    {"\<\"driveIntent\"\>", 
                    RowBox[{"-", "0.5`"}]},
                    {"\<\"motorTorque\"\>", "15.054632084346991`"},
                    {"\<\"vw\"\>", 
                    RowBox[{"-", "0.316923684509537`"}]},
                    {"\<\"controllerState\"\>", 
                    RowBox[{"\[LeftAssociation]", 
                    RowBox[{
                    RowBox[{"\<\"lastError\"\>", "\[Rule]", 
                    RowBox[{"-", "0.024502892340463735`"}]}], ",", 
                    
                    RowBox[{"\<\"lastInt\"\>", "\[Rule]", 
                    "0.15374312678207563`"}], ",", 
                    RowBox[{"\<\"initialized\"\>", "\[Rule]", "True"}]}], 
                    "\[RightAssociation]"}]}
                    },
                    AutoDelete->False,
                    
                    GridBoxItemSize->{
                    "Columns" -> {{Automatic}}, "Rows" -> {{Automatic}}}],
                    "Grid"], {-1, -0.5}], 
                  {GrayLevel[0], LineBox[{{-5, 0}, {5, 0}}]}, 
                  {GrayLevel[0.5], CircleBox[{-0.7491486164399837, 0.5}, 0.5],
                    DiskBox[{-0.7491486164399837, 0.5}, 0.1]}, 
                  {GrayLevel[0.5], 
                   DiskBox[{-0.7947934142062734, 1.9993053566358236`}, 0.1]}, 
                  {GrayLevel[0.5], Thickness[0.01], 
                   LineBox[{{-0.7491486164399837, 0.5}, {-0.7947934142062734, 
                    1.9993053566358236`}}]}, 
                  {RGBColor[0, 1, 0], 
                   DiskBox[{-0.7795784816175101, 1.4995369044238824`}, 0.05]}, 
                  {RGBColor[0, 0, 1], Arrowheads[0.01], 
                   CircleBox[{-0.7491486164399837, 0.5}, 0.3333333333333333, 
                    NCache[{
                    Rational[-1, 2] 
                    Pi, -1.1191573642644868`}, {-1.5707963267948966`, \
-1.1191573642644868`}]], 
                   ArrowBox[{{-0.6066747009737534, 
                    0.1986492168206086}, {-0.6036683670032482, 
                    0.2000889996433336}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-0.7491486164399837, 
                    0.5}, {-1.0502412581269236`, 0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-0.7491486164399837, 0.5}, {-0.5254409844453215,
                     0.5}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.005], Arrowheads[0.04], 
                   ArrowBox[{{-0.7491486164399837, 0.5}, {-0.8265336261322613,
                     0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-0.7491486164399837, 0.5}, 1.6, {
                    1.6012308901546994`, 1.7411129389093198`}], 
                   ArrowBox[{{-1.0045578432261337`, 
                    2.0794828669132506`}, {-1.020339638294519, 
                    2.0768498437281506`}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Dashing[{Small, Small}], 
                   Arrowheads[0.02], 
                   CircleBox[{-0.7491486164399837, 0.5}, 1.7, {
                    1.6012308901546994`, 1.5564042435964784`}], 
                   ArrowBox[{{-0.7416820990080717, 
                    2.19998360319076}, {-0.7246829196281085, 
                    2.1998239407890186`}}]}, 
                  {RGBColor[0, 0, 1], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-0.7491486164399837, 0.5}, 1.8, {
                    1.6012308901546994`, 1.5175940452416605`}], 
                   ArrowBox[{{-0.6714086976895107, 
                    2.298320467834549}, {-0.6534296796932906, 
                    2.2974531663295386`}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.003], Arrowheads[0.025], 
                   CircleBox[{-0.7491486164399837, 0.5}, 1.5, {
                    1.6012308901546994`, 1.5992706621646642`}], 
                   ArrowBox[{{-0.8068458824041996, 
                    1.9988899310824175`}, {-0.791854348067857, 
                    1.999391950253878}}]}, 
                  {RGBColor[0, 0, 1], ArrowBox[{{0, 2.8}, {-1.5, 2.8}}]}},
                 Frame->True,
                 FrameTicks->Automatic,
                 ImageSize->800,
                 PlotRange->{{-3, 3}, {-1, 3}}], 319->
                GraphicsBox[{InsetBox[
                   TagBox[GridBox[{
                    {"\<\"t\"\>", "15.9`"},
                    {"\<\"driveIntent\"\>", 
                    RowBox[{"-", "0.5`"}]},
                    {"\<\"motorTorque\"\>", "15.210152460613479`"},
                    {"\<\"vw\"\>", 
                    RowBox[{"-", "0.3182015730658928`"}]},
                    {"\<\"controllerState\"\>", 
                    RowBox[{"\[LeftAssociation]", 
                    RowBox[{
                    RowBox[{"\<\"lastError\"\>", "\[Rule]", 
                    RowBox[{"-", "0.023798424478719497`"}]}], ",", 
                    
                    RowBox[{"\<\"lastInt\"\>", "\[Rule]", 
                    "0.1525361464800018`"}], ",", 
                    RowBox[{"\<\"initialized\"\>", "\[Rule]", "True"}]}], 
                    "\[RightAssociation]"}]}
                    },
                    AutoDelete->False,
                    
                    GridBoxItemSize->{
                    "Columns" -> {{Automatic}}, "Rows" -> {{Automatic}}}],
                    "Grid"], {-1, -0.5}], 
                  {GrayLevel[0], LineBox[{{-5, 0}, {5, 0}}]}, 
                  {GrayLevel[0.5], CircleBox[{-0.765027334626102, 0.5}, 0.5], 
                   DiskBox[{-0.765027334626102, 0.5}, 0.1]}, 
                  {GrayLevel[0.5], 
                   DiskBox[{-0.811152116541815, 1.9992906671133612`}, 0.1]}, 
                  {GrayLevel[0.5], Thickness[0.01], 
                   LineBox[{{-0.765027334626102, 0.5}, {-0.811152116541815, 
                    1.9992906671133612`}}]}, 
                  {RGBColor[0, 1, 0], 
                   DiskBox[{-0.7957771892365773, 1.4995271114089075`}, 0.05]}, 
                  {RGBColor[0, 0, 1], Arrowheads[0.01], 
                   CircleBox[{-0.765027334626102, 0.5}, 0.3333333333333333, 
                    NCache[{
                    Rational[-1, 2] 
                    Pi, -1.1144917529764922`}, {-1.5707963267948966`, \
-1.1144917529764922`}]], 
                   ArrowBox[{{-0.6211489893233206, 
                    0.19931722220246018`}, {-0.6181494055162069, 
                    0.20077101568948946`}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-0.765027334626102, 0.5}, {-1.0692303838383714`,
                     0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-0.765027334626102, 0.5}, {-0.5389694890911092, 
                    0.5}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.005], Arrowheads[0.04], 
                   ArrowBox[{{-0.765027334626102, 0.5}, {-0.8431725383033788, 
                    0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-0.765027334626102, 0.5}, 1.6, {
                    1.6015510294075512`, 1.742904027086841}], 
                   ArrowBox[{{-1.023265143313322, 
                    2.0790228732239515`}, {-1.0390421970935626`, 
                    2.0763615876907617`}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Dashing[{Small, Small}], 
                   Arrowheads[0.02], 
                   CircleBox[{-0.765027334626102, 0.5}, 1.7, {
                    1.6015510294075512`, 1.5562844802717328`}], 
                   ArrowBox[{{-0.7573572215598926, 
                    2.199982696784162}, {-0.7403580614235412, 
                    2.19982099850533}}]}, 
                  {RGBColor[0, 0, 1], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-0.765027334626102, 0.5}, 1.8, {
                    1.6015510294075512`, 1.5170501824041431`}], 
                   ArrowBox[{{-0.6863093877486622, 
                    2.298277922024124}, {-0.6683308441044543, 
                    2.2974008425281207`}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.003], Arrowheads[0.025], 
                   CircleBox[{-0.765027334626102, 0.5}, 1.5, {
                    1.6015510294075512`, 1.5996244548407763`}], 
                   ArrowBox[{{-0.8232548932482083, 
                    1.9988694244052445`}, {-0.8082635374608078, 
                    1.9993767474402275`}}]}, 
                  {RGBColor[0, 0, 1], ArrowBox[{{0, 2.8}, {-1.5, 2.8}}]}},
                 Frame->True,
                 FrameTicks->Automatic,
                 ImageSize->800,
                 PlotRange->{{-3, 3}, {-1, 3}}], 320->
                GraphicsBox[{InsetBox[
                   TagBox[GridBox[{
                    {"\<\"t\"\>", "15.950000000000001`"},
                    {"\<\"driveIntent\"\>", 
                    RowBox[{"-", "0.5`"}]},
                    {"\<\"motorTorque\"\>", "15.339687949866935`"},
                    {"\<\"vw\"\>", 
                    RowBox[{"-", "0.31949081419883824`"}]},
                    {"\<\"controllerState\"\>", 
                    RowBox[{"\[LeftAssociation]", 
                    RowBox[{
                    RowBox[{"\<\"lastError\"\>", "\[Rule]", 
                    RowBox[{"-", "0.023141900307588507`"}]}], ",", 
                    
                    RowBox[{"\<\"lastInt\"\>", "\[Rule]", 
                    "0.15136316536925426`"}], ",", 
                    RowBox[{"\<\"initialized\"\>", "\[Rule]", "True"}]}], 
                    "\[RightAssociation]"}]}
                    },
                    AutoDelete->False,
                    
                    GridBoxItemSize->{
                    "Columns" -> {{Automatic}}, "Rows" -> {{Automatic}}}],
                    "Grid"], {-1, -0.5}], 
                  {GrayLevel[0], LineBox[{{-5, 0}, {5, 0}}]}, 
                  {GrayLevel[0.5], CircleBox[{-0.7809702465130403, 0.5}, 0.5],
                    DiskBox[{-0.7809702465130403, 0.5}, 0.1]}, 
                  {GrayLevel[0.5], 
                   DiskBox[{-0.8274980326712904, 1.9992782147137376`}, 0.1]}, 
                  {GrayLevel[0.5], Thickness[0.01], 
                   LineBox[{{-0.7809702465130403, 0.5}, {-0.8274980326712904, 
                    1.9992782147137376`}}]}, 
                  {RGBColor[0, 1, 0], 
                   DiskBox[{-0.8119887706185404, 1.4995188098091585`}, 0.05]}, 
                  {RGBColor[0, 0, 1], Arrowheads[0.01], 
                   CircleBox[{-0.7809702465130403, 0.5}, 0.3333333333333333, 
                    NCache[{
                    Rational[-1, 2] 
                    Pi, -1.1106056882988886`}, {-1.5707963267948966`, \
-1.1106056882988886`}]], 
                   ArrowBox[{{-0.6359245178171957, 
                    0.19987861172818333`}, {-0.632930606180458, 
                    0.2013440507853379}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-0.7809702465130403, 0.5}, {-1.087764005510379, 
                    0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-0.7809702465130403, 0.5}, {-0.5529391685911879,
                     0.5}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.005], Arrowheads[0.04], 
                   ArrowBox[{{-0.7809702465130403, 0.5}, {-0.8597329275885266,
                     0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-0.7809702465130403, 0.5}, 1.6, {
                    1.6018198271280593`, 1.7444078627933104`}], 
                   ArrowBox[{{-1.0415823532781487`, 
                    2.0786327406358494`}, {-1.0573553870836156`, 
                    2.0759477320240616`}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Dashing[{Small, Small}], 
                   Arrowheads[0.02], 
                   CircleBox[{-0.7809702465130403, 0.5}, 1.7, {
                    1.6018198271280593`, 1.55619598411968}], 
                   ArrowBox[{{-0.7731496915498399, 
                    2.1999820113519046`}, {-0.7561505457897291, 
                    2.199818808713448}}]}, 
                  {RGBColor[0, 0, 1], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-0.7809702465130403, 0.5}, 1.8, {
                    1.6018198271280593`, 1.5165993385176875`}], 
                   ArrowBox[{{-0.7014415650558862, 
                    2.298242249761051}, {-0.6834634186647547, 
                    2.2973570648379704`}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.003], Arrowheads[0.025], 
                   CircleBox[{-0.7809702465130403, 0.5}, 1.5, {
                    1.6018198271280593`, 1.5999421234290512`}], 
                   ArrowBox[{{-0.8396739459232292, 
                    1.9988508517112562`}, {-0.8246827520528321, 
                    1.9993629370033905`}}]}, 
                  {RGBColor[0, 0, 1], ArrowBox[{{0, 2.8}, {-1.5, 2.8}}]}},
                 Frame->True,
                 FrameTicks->Automatic,
                 ImageSize->800,
                 PlotRange->{{-3, 3}, {-1, 3}}], 321->
                GraphicsBox[{InsetBox[
                   TagBox[GridBox[{
                    {"\<\"t\"\>", "16.`"},
                    {"\<\"driveIntent\"\>", 
                    RowBox[{"-", "0.5`"}]},
                    {"\<\"motorTorque\"\>", "15.443989630699498`"},
                    {"\<\"vw\"\>", 
                    RowBox[{"-", "0.32078910301638314`"}]},
                    {"\<\"controllerState\"\>", 
                    RowBox[{"\[LeftAssociation]", 
                    RowBox[{
                    RowBox[{"\<\"lastError\"\>", "\[Rule]", 
                    RowBox[{"-", "0.022532721075467745`"}]}], ",", 
                    
                    RowBox[{"\<\"lastInt\"\>", "\[Rule]", 
                    "0.15022179985189416`"}], ",", 
                    RowBox[{"\<\"initialized\"\>", "\[Rule]", "True"}]}], 
                    "\[RightAssociation]"}]}
                    },
                    AutoDelete->False,
                    
                    GridBoxItemSize->{
                    "Columns" -> {{Automatic}}, "Rows" -> {{Automatic}}}],
                    "Grid"], {-1, -0.5}], 
                  {GrayLevel[0], LineBox[{{-5, 0}, {5, 0}}]}, 
                  {GrayLevel[0.5], CircleBox[{-0.7969778605826118, 0.5}, 0.5],
                    DiskBox[{-0.7969778605826118, 0.5}, 0.1]}, 
                  {GrayLevel[0.5], 
                   DiskBox[{-0.8438336194391596, 1.9992680006796575`}, 0.1]}, 
                  {GrayLevel[0.5], Thickness[0.01], 
                   LineBox[{{-0.7969778605826118, 0.5}, {-0.8438336194391596, 
                    1.9992680006796575`}}]}, 
                  {RGBColor[0, 1, 0], 
                   DiskBox[{-0.8282150331536436, 1.4995120004531048`}, 0.05]}, 
                  {RGBColor[0, 0, 1], Arrowheads[0.01], 
                   CircleBox[{-0.7969778605826118, 0.5}, 0.3333333333333333, 
                    NCache[{
                    Rational[-1, 2] 
                    Pi, -1.1074766378739116`}, {-1.5707963267948966`, \
-1.1074766378739116`}]], 
                   ArrowBox[{{-0.6509937485293247, 
                    0.20033393562312662`}, {-0.6480044369744264, 
                    0.20180873559145418`}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-0.7969778605826118, 
                    0.5}, {-1.1058576531966018`, 0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-0.7969778605826118, 0.5}, {-0.5673409642019613,
                     0.5}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.005], Arrowheads[0.04], 
                   ArrowBox[{{-0.7969778605826118, 0.5}, {-0.8762207568159514,
                     0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-0.7969778605826118, 0.5}, 1.6, {
                    1.6020385816000138`, 1.745631715222065}], 
                   ArrowBox[{{-1.0595217852048526`, 
                    2.078312607706075}, {-1.0752915211442893`, 
                    2.0756082972441976`}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Dashing[{Small, Small}], 
                   Arrowheads[0.02], 
                   CircleBox[{-0.7969778605826118, 0.5}, 1.7, {
                    1.6020385816000138`, 1.5561368901961101`}], 
                   ArrowBox[{{-0.789056847026076, 
                    2.1999815462363808`}, {-0.7720577109399307, 
                    2.1998173390519895`}}]}, 
                  {RGBColor[0, 0, 1], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-0.7969778605826118, 0.5}, 1.8, {
                    1.6020385816000138`, 1.5162386392072387`}], 
                   ArrowBox[{{-0.7168005595735132, 
                    2.2982134468418636`}, {-0.6988227336374805, 
                    2.2973217772715007`}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.003], Arrowheads[0.025], 
                   CircleBox[{-0.7969778605826118, 0.5}, 1.5, {
                    1.6020385816000138`, 1.6002227430011144`}], 
                   ArrowBox[{{-0.856102164560494, 
                    1.9988343192892037`}, {-0.841111114981509, 
                    1.999350611383529}}]}, 
                  {RGBColor[0, 0, 1], ArrowBox[{{0, 2.8}, {-1.5, 2.8}}]}},
                 Frame->True,
                 FrameTicks->Automatic,
                 ImageSize->800,
                 PlotRange->{{-3, 3}, {-1, 3}}], 322->
                GraphicsBox[{InsetBox[
                   TagBox[GridBox[{
                    {"\<\"t\"\>", "16.05`"},
                    {"\<\"driveIntent\"\>", 
                    RowBox[{"-", "0.5`"}]},
                    {"\<\"motorTorque\"\>", "15.523972098157827`"},
                    {"\<\"vw\"\>", 
                    RowBox[{"-", "0.32209423023347866`"}]},
                    {"\<\"controllerState\"\>", 
                    RowBox[{"\[LeftAssociation]", 
                    RowBox[{
                    RowBox[{"\<\"lastError\"\>", "\[Rule]", 
                    RowBox[{"-", "0.02196991146836165`"}]}], ",", 
                    
                    RowBox[{"\<\"lastInt\"\>", "\[Rule]", 
                    "0.14910970609477933`"}], ",", 
                    RowBox[{"\<\"initialized\"\>", "\[Rule]", "True"}]}], 
                    "\[RightAssociation]"}]}
                    },
                    AutoDelete->False,
                    
                    GridBoxItemSize->{
                    "Columns" -> {{Automatic}}, "Rows" -> {{Automatic}}}],
                    "Grid"], {-1, -0.5}], 
                  {GrayLevel[0], LineBox[{{-5, 0}, {5, 0}}]}, 
                  {GrayLevel[0.5], CircleBox[{-0.8130505724595105, 0.5}, 0.5],
                    DiskBox[{-0.8130505724595105, 0.5}, 0.1]}, 
                  {GrayLevel[0.5], 
                   DiskBox[{-0.8601617403966418, 1.9992599967502633`}, 0.1]}, 
                  {GrayLevel[0.5], Thickness[0.01], 
                   LineBox[{{-0.8130505724595105, 0.5}, {-0.8601617403966418, 
                    1.9992599967502633`}}]}, 
                  {RGBColor[0, 1, 0], 
                   DiskBox[{-0.8444580177509313, 1.4995066645001756`}, 0.05]}, 
                  {RGBColor[0, 0, 1], Arrowheads[0.01], 
                   CircleBox[{-0.8130505724595105, 0.5}, 0.3333333333333333, 
                    NCache[{
                    Rational[-1, 2] 
                    Pi, -1.1050771638501617`}, {-1.5707963267948966`, \
-1.1050771638501617`}]], 
                   ArrowBox[{{-0.6663478404087076, 
                    0.200685083031364}, {-0.6633620762000679, 
                    0.20216705152267328`}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-0.8130505724595105, 
                    0.5}, {-1.1235300144226672`, 0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-0.8130505724595105, 0.5}, {-0.5821631661092844,
                     0.5}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.005], Arrowheads[0.04], 
                   ArrowBox[{{-0.8130505724595105, 0.5}, {-0.892642608072441, 
                    0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-0.8130505724595105, 0.5}, 1.6, {
                    1.6022089379082989`, 1.7465847926116327`}], 
                   ArrowBox[{{-1.0770986316723599`, 
                    2.078061666230419}, {-1.0928657830327604`, 
                    2.0753423272202927`}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Dashing[{Small, Small}], 
                   Arrowheads[0.02], 
                   CircleBox[{-0.8130505724595105, 0.5}, 1.7, {
                    1.6022089379082989`, 1.556105258031945}], 
                   ArrowBox[{{-0.8050757848115984, 
                    2.1999812948270847`}, {-0.7880766539281863, 
                    2.1998165499233115`}}]}, 
                  {RGBColor[0, 0, 1], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-0.8130505724595105, 0.5}, 1.8, {
                    1.6022089379082989`, 1.5159646484740887`}], 
                   ArrowBox[{{-0.732380580645408, 
                    2.298191411507883}, {-0.7144029996933776, 
                    2.297294816213342}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.003], Arrowheads[0.025], 
                   CircleBox[{-0.8130505724595105, 0.5}, 1.5, {
                    1.6022089379082989`, 1.6004658650776018`}], 
                   ArrowBox[{{-0.8725392743984492, 
                    1.9988199005689844`}, {-0.8575483507845171, 
                    1.9993398373031175`}}]}, 
                  {RGBColor[0, 0, 1], ArrowBox[{{0, 2.8}, {-1.5, 2.8}}]}},
                 Frame->True,
                 FrameTicks->Automatic,
                 ImageSize->800,
                 PlotRange->{{-3, 3}, {-1, 3}}], 323->
                GraphicsBox[{InsetBox[
                   TagBox[GridBox[{
                    {"\<\"t\"\>", "16.1`"},
                    {"\<\"driveIntent\"\>", 
                    RowBox[{"-", "0.5`"}]},
                    {"\<\"motorTorque\"\>", "15.580685776288222`"},
                    {"\<\"vw\"\>", 
                    RowBox[{"-", "0.32340409309620793`"}]},
                    {"\<\"controllerState\"\>", 
                    RowBox[{"\[LeftAssociation]", 
                    RowBox[{
                    RowBox[{"\<\"lastError\"\>", "\[Rule]", 
                    RowBox[{"-", "0.021452172205103544`"}]}], ",", 
                    
                    RowBox[{"\<\"lastInt\"\>", "\[Rule]", 
                    "0.1480245975047607`"}], ",", 
                    RowBox[{"\<\"initialized\"\>", "\[Rule]", "True"}]}], 
                    "\[RightAssociation]"}]}
                    },
                    AutoDelete->False,
                    
                    GridBoxItemSize->{
                    "Columns" -> {{Automatic}}, "Rows" -> {{Automatic}}}],
                    "Grid"], {-1, -0.5}], 
                  {GrayLevel[0], LineBox[{{-5, 0}, {5, 0}}]}, 
                  {GrayLevel[0.5], CircleBox[{-0.829188669980123, 0.5}, 0.5], 
                   DiskBox[{-0.829188669980123, 0.5}, 0.1]}, 
                  {GrayLevel[0.5], 
                   DiskBox[{-0.876485587194891, 1.9992541484424777`}, 0.1]}, 
                  {GrayLevel[0.5], Thickness[0.01], 
                   LineBox[{{-0.829188669980123, 0.5}, {-0.876485587194891, 
                    1.9992541484424777`}}]}, 
                  {RGBColor[0, 1, 0], 
                   DiskBox[{-0.8607199481233017, 1.4995027656283186`}, 0.05]}, 
                  {RGBColor[0, 0, 1], Arrowheads[0.01], 
                   CircleBox[{-0.829188669980123, 0.5}, 0.3333333333333333, 
                    NCache[{
                    Rational[-1, 2] 
                    Pi, -1.1033757535062498`}, {-1.5707963267948966`, \
-1.1033757535062498`}]], 
                   ArrowBox[{{-0.6819768930164767, 
                    0.20093511768461214`}, {-0.6789936545647308, 
                    0.20242216403858093`}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-0.829188669980123, 0.5}, {-1.1408023855058875`,
                     0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-0.829188669980123, 0.5}, {-0.597391828073494, 
                    0.5}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.005], Arrowheads[0.04], 
                   ArrowBox[{{-0.829188669980123, 0.5}, {-0.9090055435992584, 
                    0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-0.829188669980123, 0.5}, 1.6, {
                    1.6023328321228363`, 1.74727792999915}], 
                   ArrowBox[{{-1.0943304792165114`, 
                    2.077878265581618}, {-1.1100957419139452`, 
                    2.0751479984235024`}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Dashing[{Small, Small}], 
                   Arrowheads[0.02], 
                   CircleBox[{-0.829188669980123, 0.5}, 1.7, {
                    1.6023328321228363`, 1.5560990991993693`}], 
                   ArrowBox[{{-0.8212034124321854, 
                    2.1999812456794614`}, {-0.804204282563732, 
                    2.199816396080891}}]}, 
                  {RGBColor[0, 0, 1], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-0.829188669980123, 0.5}, 1.8, {
                    1.6023328321228363`, 1.5157734666990128`}], 
                   ArrowBox[{{-0.7481748982163995, 
                    2.298175956013375}, {-0.7301974890055924, 
                    2.2972759237494023`}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.003], Arrowheads[0.025], 
                   CircleBox[{-0.829188669980123, 0.5}, 1.5, {
                    1.6023328321228363`, 1.6006714673690536`}], 
                   ArrowBox[{{-0.888985531465559, 
                    1.9988076378763524`}, {-0.8739947150686604, 
                    1.9993306567677205`}}]}, 
                  {RGBColor[0, 0, 1], ArrowBox[{{0, 2.8}, {-1.5, 2.8}}]}},
                 Frame->True,
                 FrameTicks->Automatic,
                 ImageSize->800,
                 PlotRange->{{-3, 3}, {-1, 3}}], 324->
                GraphicsBox[{InsetBox[
                   TagBox[GridBox[{
                    {"\<\"t\"\>", "16.15`"},
                    {"\<\"driveIntent\"\>", 
                    RowBox[{"-", "0.5`"}]},
                    {"\<\"motorTorque\"\>", "15.615291108940832`"},
                    {"\<\"vw\"\>", 
                    RowBox[{"-", "0.3247167040177868`"}]},
                    {"\<\"controllerState\"\>", 
                    RowBox[{"\[LeftAssociation]", 
                    RowBox[{
                    RowBox[{"\<\"lastError\"\>", "\[Rule]", 
                    RowBox[{"-", "0.020977929721466032`"}]}], ",", 
                    
                    RowBox[{"\<\"lastInt\"\>", "\[Rule]", 
                    "0.146964259647621`"}], ",", 
                    RowBox[{"\<\"initialized\"\>", "\[Rule]", "True"}]}], 
                    "\[RightAssociation]"}]}
                    },
                    AutoDelete->False,
                    
                    GridBoxItemSize->{
                    "Columns" -> {{Automatic}}, "Rows" -> {{Automatic}}}],
                    "Grid"], {-1, -0.5}], 
                  {GrayLevel[0], LineBox[{{-5, 0}, {5, 0}}]}, 
                  {GrayLevel[0.5], CircleBox[{-0.8453923387484634, 0.5}, 0.5],
                    DiskBox[{-0.8453923387484634, 0.5}, 0.1]}, 
                  {GrayLevel[0.5], 
                   DiskBox[{-0.892808606879399, 1.9992503785280613`}, 0.1]}, 
                  {GrayLevel[0.5], Thickness[0.01], 
                   LineBox[{{-0.8453923387484634, 0.5}, {-0.892808606879399, 
                    1.9992503785280613`}}]}, 
                  {RGBColor[0, 1, 0], 
                   DiskBox[{-0.8770031841690871, 1.4995002523520409`}, 0.05]}, 
                  {RGBColor[0, 0, 1], Arrowheads[0.01], 
                   CircleBox[{-0.8453923387484634, 0.5}, 0.3333333333333333, 
                    NCache[{
                    Rational[-1, 2] 
                    Pi, -1.1023375935266717`}, {-1.5707963267948966`, \
-1.1023375935266717`}]], 
                   ArrowBox[{{-0.6978701639791345, 
                    0.201088108195009}, {-0.6948884709267551, 
                    0.20257825082584013`}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-0.8453923387484634, 0.5}, {-1.15769816092728, 
                    0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-0.8453923387484634, 0.5}, {-0.6130111558306239,
                     0.5}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.005], Arrowheads[0.04], 
                   ArrowBox[{{-0.8453923387484634, 0.5}, {-0.9253169780094406,
                     0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-0.8453923387484634, 0.5}, 1.6, {
                    1.6024124390835937`, 1.7477232971667933`}], 
                   ArrowBox[{{-1.111236856839956, 
                    2.077760023641334}, {-1.1270009020025438`, 
                    2.0750227354238326`}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Dashing[{Small, Small}], 
                   Arrowheads[0.02], 
                   CircleBox[{-0.8453923387484634, 0.5}, 1.7, {
                    1.6024124390835937`, 1.5561164021648846`}], 
                   ArrowBox[{{-0.8374364959185904, 
                    2.1999813835936166`}, {-0.8204373632002947, 
                    2.1998168281304284`}}]}, 
                  {RGBColor[0, 0, 1], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-0.8453923387484634, 0.5}, 1.8, {
                    1.6024124390835937`, 1.5156608218117}], 
                   ArrowBox[{{-0.7641760121712212, 
                    2.2981668188177924`}, {-0.7461987044585034, 
                    2.297264761496299}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.003], Arrowheads[0.025], 
                   CircleBox[{-0.8453923387484634, 0.5}, 1.5, {
                    1.6024124390835937`, 1.6008399051381974`}], 
                   ArrowBox[{{-0.9054416551993484, 
                    1.9987975445649027`}, {-0.8904509271112391, 
                    1.9993230884685107`}}]}, 
                  {RGBColor[0, 0, 1], ArrowBox[{{0, 2.8}, {-1.5, 2.8}}]}},
                 Frame->True,
                 FrameTicks->Automatic,
                 ImageSize->800,
                 PlotRange->{{-3, 3}, {-1, 3}}], 325->
                GraphicsBox[{InsetBox[
                   TagBox[GridBox[{
                    {"\<\"t\"\>", "16.2`"},
                    {"\<\"driveIntent\"\>", 
                    RowBox[{"-", "0.5`"}]},
                    {"\<\"motorTorque\"\>", "15.629034663992417`"},
                    {"\<\"vw\"\>", 
                    RowBox[{"-", "0.3260301971038498`"}]},
                    {"\<\"controllerState\"\>", 
                    RowBox[{"\[LeftAssociation]", 
                    RowBox[{
                    RowBox[{"\<\"lastError\"\>", "\[Rule]", 
                    RowBox[{"-", "0.020545382793108573`"}]}], ",", 
                    
                    RowBox[{"\<\"lastInt\"\>", "\[Rule]", 
                    "0.14592656276034444`"}], ",", 
                    RowBox[{"\<\"initialized\"\>", "\[Rule]", "True"}]}], 
                    "\[RightAssociation]"}]}
                    },
                    AutoDelete->False,
                    
                    GridBoxItemSize->{
                    "Columns" -> {{Automatic}}, "Rows" -> {{Automatic}}}],
                    "Grid"], {-1, -0.5}], 
                  {GrayLevel[0], LineBox[{{-5, 0}, {5, 0}}]}, 
                  {GrayLevel[0.5], CircleBox[{-0.8616616680693473, 0.5}, 0.5],
                    DiskBox[{-0.8616616680693473, 0.5}, 0.1]}, 
                  {GrayLevel[0.5], 
                   DiskBox[{-0.90913443511558, 1.999248590590958}, 0.1]}, 
                  {GrayLevel[0.5], Thickness[0.01], 
                   LineBox[{{-0.8616616680693473, 0.5}, {-0.90913443511558, 
                    1.999248590590958}}]}, 
                  {RGBColor[0, 1, 0], 
                   DiskBox[{-0.8933101794335024, 1.4994990603939722`}, 0.05]}, 
                  {RGBColor[0, 0, 1], Arrowheads[0.01], 
                   CircleBox[{-0.8616616680693473, 0.5}, 0.3333333333333333, 
                    NCache[{
                    Rational[-1, 2] 
                    Pi, -1.1019252868751241`}, {-1.5707963267948966`, \
-1.1019252868751241`}]], 
                   ArrowBox[{{-0.714016262481439, 
                    0.20114895797422228`}, {-0.7110351840781999, 
                    0.20264032985023722`}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-0.8616616680693473, 
                    0.5}, {-1.1742423613491957`, 0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-0.8616616680693473, 0.5}, {-0.6290038685113514,
                     0.5}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.005], Arrowheads[0.04], 
                   ArrowBox[{{-0.8616616680693473, 0.5}, {-0.9415845617911998,
                     0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-0.8616616680693473, 0.5}, 1.6, {
                    1.6024501238824695`, 1.747934127305575}], 
                   ArrowBox[{{-1.1278388196148166`, 
                    2.077703940539904}, {-1.143602287324203, 
                    2.074963328847433}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Dashing[{Small, Small}], 
                   Arrowheads[0.02], 
                   CircleBox[{-0.8616616680693473, 0.5}, 1.7, {
                    1.6024501238824695`, 1.5561551544942902`}], 
                   ArrowBox[{{-0.8537717034839921, 
                    2.1999816906245906`}, {-0.8367725644015531, 
                    2.199817793917517}}]}, 
                  {RGBColor[0, 0, 1], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-0.8616616680693473, 0.5}, 1.8, {
                    1.6024501238824695`, 1.515622153526956}], 
                   ArrowBox[{{-0.7803758095262741, 
                    2.298163676977409}, {-0.7623985367080058, 
                    2.297260924504937}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.003], Arrowheads[0.025], 
                   CircleBox[{-0.8616616680693473, 0.5}, 1.5, {
                    1.6024501238824695`, 1.6009718644672295`}], 
                   ArrowBox[{{-0.9219087643151667, 
                    1.998789607447939}, {-0.9069181057079968, 
                    1.9993171295133858`}}]}, 
                  {RGBColor[0, 0, 1], ArrowBox[{{0, 2.8}, {-1.5, 2.8}}]}},
                 Frame->True,
                 FrameTicks->Automatic,
                 ImageSize->800,
                 PlotRange->{{-3, 3}, {-1, 3}}], 326->
                GraphicsBox[{InsetBox[
                   TagBox[GridBox[{
                    {"\<\"t\"\>", "16.25`"},
                    {"\<\"driveIntent\"\>", 
                    RowBox[{"-", "0.5`"}]},
                    {"\<\"motorTorque\"\>", "15.623227162843065`"},
                    {"\<\"vw\"\>", 
                    RowBox[{"-", "0.3273428327452492`"}]},
                    {"\<\"controllerState\"\>", 
                    RowBox[{"\[LeftAssociation]", 
                    RowBox[{
                    RowBox[{"\<\"lastError\"\>", "\[Rule]", 
                    RowBox[{"-", "0.020152545994960663`"}]}], ",", 
                    
                    RowBox[{"\<\"lastInt\"\>", "\[Rule]", 
                    "0.1449094720125774`"}], ",", 
                    RowBox[{"\<\"initialized\"\>", "\[Rule]", "True"}]}], 
                    "\[RightAssociation]"}]}
                    },
                    AutoDelete->False,
                    
                    GridBoxItemSize->{
                    "Columns" -> {{Automatic}}, "Rows" -> {{Automatic}}}],
                    "Grid"], {-1, -0.5}], 
                  {GrayLevel[0], LineBox[{{-5, 0}, {5, 0}}]}, 
                  {GrayLevel[0.5], CircleBox[{-0.8779966571578062, 0.5}, 0.5],
                    DiskBox[{-0.8779966571578062, 0.5}, 0.1]}, 
                  {GrayLevel[0.5], 
                   DiskBox[{-0.9254668353129161, 1.9992486725643355`}, 0.1]}, 
                  {GrayLevel[0.5], Thickness[0.01], 
                   LineBox[{{-0.8779966571578062, 0.5}, {-0.9254668353129161, 
                    1.9992486725643355`}}]}, 
                  {RGBColor[0, 1, 0], 
                   DiskBox[{-0.9096434425945461, 1.4994991150428905`}, 0.05]}, 
                  {RGBColor[0, 0, 1], Arrowheads[0.01], 
                   CircleBox[{-0.8779966571578062, 0.5}, 0.3333333333333333, 
                    NCache[{
                    Rational[-1, 2] 
                    Pi, -1.1020995119096046`}, {-1.5707963267948966`, \
-1.1020995119096046`}]], 
                   ArrowBox[{{-0.7304033211435771, 
                    0.20112323898418938`}, {-0.7274219829512671, 
                    0.2026140914590846}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-0.8779966571578062, 
                    0.5}, {-1.1904612004146675`, 0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-0.8779966571578062, 0.5}, {-0.6453515326951843,
                     0.5}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.005], Arrowheads[0.04], 
                   ArrowBox[{{-0.8779966571578062, 0.5}, {-0.9578160759520457,
                     0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-0.8779966571578062, 0.5}, 1.6, {
                    1.602448397090083, 1.7479244666540854`}], 
                   ArrowBox[{{-1.1441585670429313`, 
                    2.0777065119109777`}, {-1.1599220612276768`, 
                    2.074966052504003}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Dashing[{Small, Small}], 
                   Arrowheads[0.02], 
                   CircleBox[{-0.8779966571578062, 0.5}, 1.7, {
                    1.602448397090083, 1.556213362494806}], 
                   ArrowBox[{{-0.870205645120886, 
                    2.1999821470037384`}, {-0.8532064965271455, 
                    2.199819239782838}}]}, 
                  {RGBColor[0, 0, 1], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-0.8779966571578062, 0.5}, 1.8, {
                    1.602448397090083, 1.5156526906298435`}], 
                   ArrowBox[{{-0.7967657093618369, 
                    2.2981661583736273`}, {-0.7787884089845055, 
                    2.297263954875406}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.003], Arrowheads[0.025], 
                   CircleBox[{-0.8779966571578062, 0.5}, 1.5, {
                    1.602448397090083, 1.6010683176647516`}], 
                   ArrowBox[{{-0.9383883161732066, 
                    1.9987837894510894`}, {-0.9233977085169573, 
                    1.9993127574110359`}}]}, 
                  {RGBColor[0, 0, 1], ArrowBox[{{0, 2.8}, {-1.5, 2.8}}]}},
                 Frame->True,
                 FrameTicks->Automatic,
                 ImageSize->800,
                 PlotRange->{{-3, 3}, {-1, 3}}], 327->
                GraphicsBox[{InsetBox[
                   TagBox[GridBox[{
                    {"\<\"t\"\>", "16.3`"},
                    {"\<\"driveIntent\"\>", 
                    RowBox[{"-", "0.5`"}]},
                    {"\<\"motorTorque\"\>", "15.599223426394463`"},
                    {"\<\"vw\"\>", 
                    RowBox[{"-", "0.3286530004554321`"}]},
                    {"\<\"controllerState\"\>", 
                    RowBox[{"\[LeftAssociation]", 
                    RowBox[{
                    RowBox[{"\<\"lastError\"\>", "\[Rule]", 
                    RowBox[{"-", "0.01979728993789205`"}]}], ",", 
                    
                    RowBox[{"\<\"lastInt\"\>", "\[Rule]", 
                    "0.14391105567713752`"}], ",", 
                    RowBox[{"\<\"initialized\"\>", "\[Rule]", "True"}]}], 
                    "\[RightAssociation]"}]}
                    },
                    AutoDelete->False,
                    
                    GridBoxItemSize->{
                    "Columns" -> {{Automatic}}, "Rows" -> {{Automatic}}}],
                    "Grid"], {-1, -0.5}], 
                  {GrayLevel[0], LineBox[{{-5, 0}, {5, 0}}]}, 
                  {GrayLevel[0.5], CircleBox[{-0.8943972215326456, 0.5}, 0.5],
                    DiskBox[{-0.8943972215326456, 0.5}, 0.1]}, 
                  {GrayLevel[0.5], 
                   DiskBox[{-0.941809643560085, 1.9992505001624952`}, 0.1]}, 
                  {GrayLevel[0.5], Thickness[0.01], 
                   LineBox[{{-0.8943972215326456, 0.5}, {-0.941809643560085, 
                    1.9992505001624952`}}]}, 
                  {RGBColor[0, 1, 0], 
                   DiskBox[{-0.9260055028842719, 1.4995003334416634`}, 0.05]}, 
                  {RGBColor[0, 0, 1], Arrowheads[0.01], 
                   CircleBox[{-0.8943972215326456, 0.5}, 0.3333333333333333, 
                    NCache[{
                    Rational[-1, 2] 
                    Pi, -1.1028196240030628`}, {-1.5707963267948966`, \
-1.1028196240030628`}]], 
                   ArrowBox[{{-0.7470191485379801, 
                    0.20101703274017402`}, {-0.7440367375378697, 
                    0.2025057379310185}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-0.8943972215326456, 
                    0.5}, {-1.2063816900605349`, 0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-0.8943972215326456, 0.5}, {-0.6620348690340473,
                     0.5}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.005], Arrowheads[0.04], 
                   ArrowBox[{{-0.8943972215326456, 0.5}, {-0.9740193375619366,
                     0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-0.8943972215326456, 0.5}, 1.6, {
                    1.6024098737326722`, 1.7477089451312313`}], 
                   ArrowBox[{{-1.1602190955287919`, 
                    2.077763838888817}, {-1.175983179975414, 
                    2.075026776917735}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Dashing[{Small, Small}], 
                   Arrowheads[0.02], 
                   CircleBox[{-0.8943972215326456, 0.5}, 1.7, {
                    1.6024098737326722`, 1.5562890683963468`}], 
                   ArrowBox[{{-0.8867349081989714, 
                    2.199982731957703}, {-0.8697357473209071, 
                    2.1998211116731383`}}]}, 
                  {RGBColor[0, 0, 1], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-0.8943972215326456, 0.5}, 1.8, {
                    1.6024098737326722`, 1.515747521363814}], 
                   ArrowBox[{{-0.8133367955182699, 
                    2.298173853478681}, {-0.7953594096651524, 
                    2.297273354785103}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.003], Arrowheads[0.025], 
                   CircleBox[{-0.8943972215326456, 0.5}, 1.5, {
                    1.6024098737326722`, 1.6011304810006701`}], 
                   ArrowBox[{{-0.9548820498314738, 
                    1.9987800324082454`}, {-0.9398914750866014, 
                    1.9993099322333485`}}]}, 
                  {RGBColor[0, 0, 1], ArrowBox[{{0, 2.8}, {-1.5, 2.8}}]}},
                 Frame->True,
                 FrameTicks->Automatic,
                 ImageSize->800,
                 PlotRange->{{-3, 3}, {-1, 3}}], 328->
                GraphicsBox[{InsetBox[
                   TagBox[GridBox[{
                    {"\<\"t\"\>", "16.35`"},
                    {"\<\"driveIntent\"\>", 
                    RowBox[{"-", "0.5`"}]},
                    {"\<\"motorTorque\"\>", "15.558404210567913`"},
                    {"\<\"vw\"\>", 
                    RowBox[{"-", "0.3299592201266351`"}]},
                    {"\<\"controllerState\"\>", 
                    RowBox[{"\[LeftAssociation]", 
                    RowBox[{
                    RowBox[{"\<\"lastError\"\>", "\[Rule]", 
                    RowBox[{"-", "0.019477378262385096`"}]}], ",", 
                    
                    RowBox[{"\<\"lastInt\"\>", "\[Rule]", 
                    "0.14292949137137526`"}], ",", 
                    RowBox[{"\<\"initialized\"\>", "\[Rule]", "True"}]}], 
                    "\[RightAssociation]"}]}
                    },
                    AutoDelete->False,
                    
                    GridBoxItemSize->{
                    "Columns" -> {{Automatic}}, "Rows" -> {{Automatic}}}],
                    "Grid"], {-1, -0.5}], 
                  {GrayLevel[0], LineBox[{{-5, 0}, {5, 0}}]}, 
                  {GrayLevel[0.5], CircleBox[{-0.9108631995108378, 0.5}, 0.5],
                    DiskBox[{-0.9108631995108378, 0.5}, 0.1]}, 
                  {GrayLevel[0.5], 
                   DiskBox[{-0.9581667192358455, 1.99925394013877}, 0.1]}, 
                  {GrayLevel[0.5], Thickness[0.01], 
                   LineBox[{{-0.9108631995108378, 0.5}, {-0.9581667192358455, 
                    1.99925394013877}}]}, 
                  {RGBColor[0, 1, 0], 
                   DiskBox[{-0.9423988793275095, 1.4995026267591798`}, 0.05]}, 
                  {RGBColor[0, 0, 1], Arrowheads[0.01], 
                   CircleBox[{-0.9108631995108378, 0.5}, 0.3333333333333333, 
                    NCache[{
                    Rational[-1, 2] 
                    Pi, -1.1040442004778592`}, {-1.5707963267948966`, \
-1.1040442004778592`}]], 
                   ArrowBox[{{-0.7638513644358843, 
                    0.20083678123972903`}, {-0.7608671326390677, 
                    0.20232183312491564`}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-0.9108631995108378, 0.5}, {-1.222031283722196, 
                    0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-0.9108631995108378, 0.5}, {-0.6790340316582169,
                     0.5}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.005], Arrowheads[0.04], 
                   ArrowBox[{{-0.9108631995108378, 0.5}, {-0.9902021158695752,
                     0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-0.9108631995108378, 0.5}, 1.6, {
                    1.6023372359863874`, 1.747302567771152}], 
                   ArrowBox[{{-1.1760438840717033`, 
                    2.077871732599273}, {-1.1918090794966525`, 
                    2.0751410770210383`}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Dashing[{Small, Small}], 
                   Arrowheads[0.02], 
                   CircleBox[{-0.9108631995108378, 0.5}, 1.7, {
                    1.6023372359863874`, 1.5563803651883228`}], 
                   ArrowBox[{{-0.9033560891787235, 
                    2.199983424417562}, {-0.8863569136160905, 
                    2.199823356102523}}]}, 
                  {RGBColor[0, 0, 1], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-0.9108631995108378, 0.5}, 1.8, {
                    1.6023372359863874`, 1.515901657038788}], 
                   ArrowBox[{{-0.8300799371988999, 
                    2.2981863264217757`}, {-0.8121024127603608, 
                    2.2972885986953866`}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.003], Arrowheads[0.025], 
                   CircleBox[{-0.9108631995108378, 0.5}, 1.5, {
                    1.6023372359863874`, 1.6011597749152102`}], 
                   ArrowBox[{{-0.9713919329178916, 
                    1.998778259927778}, {-0.9564013737022913, 
                    1.999308598885269}}]}, 
                  {RGBColor[0, 0, 1], ArrowBox[{{0, 2.8}, {-1.5, 2.8}}]}},
                 Frame->True,
                 FrameTicks->Automatic,
                 ImageSize->800,
                 PlotRange->{{-3, 3}, {-1, 3}}], 329->
                GraphicsBox[{InsetBox[
                   TagBox[GridBox[{
                    {"\<\"t\"\>", "16.4`"},
                    {"\<\"driveIntent\"\>", 
                    RowBox[{"-", "0.5`"}]},
                    {"\<\"motorTorque\"\>", "15.502159888821486`"},
                    {"\<\"vw\"\>", 
                    RowBox[{"-", "0.3312601418748096`"}]},
                    {"\<\"controllerState\"\>", 
                    RowBox[{"\[LeftAssociation]", 
                    RowBox[{
                    RowBox[{"\<\"lastError\"\>", "\[Rule]", 
                    RowBox[{"-", "0.01919050140348228`"}]}], ",", 
                    
                    RowBox[{"\<\"lastInt\"\>", "\[Rule]", 
                    "0.141963070531311`"}], ",", 
                    RowBox[{"\<\"initialized\"\>", "\[Rule]", "True"}]}], 
                    "\[RightAssociation]"}]}
                    },
                    AutoDelete->False,
                    
                    GridBoxItemSize->{
                    "Columns" -> {{Automatic}}, "Rows" -> {{Automatic}}}],
                    "Grid"], {-1, -0.5}], 
                  {GrayLevel[0], LineBox[{{-5, 0}, {5, 0}}]}, 
                  {GrayLevel[0.5], CircleBox[{-0.9273943587280378, 0.5}, 0.5],
                    DiskBox[{-0.9273943587280378, 0.5}, 0.1]}, 
                  {GrayLevel[0.5], 
                   DiskBox[{-0.9745419011197964, 1.9992588533160034`}, 0.1]}, 
                  {GrayLevel[0.5], Thickness[0.01], 
                   LineBox[{{-0.9273943587280378, 0.5}, {-0.9745419011197964, 
                    1.9992588533160034`}}]}, 
                  {RGBColor[0, 1, 0], 
                   DiskBox[{-0.9588260536558769, 1.499505902210669}, 0.05]}, 
                  {RGBColor[0, 0, 1], Arrowheads[0.01], 
                   CircleBox[{-0.9273943587280378, 0.5}, 0.3333333333333333, 
                    NCache[{
                    Rational[-1, 2] 
                    Pi, -1.105731530130252}, {-1.5707963267948966`, \
-1.105731530130252}]], 
                   ArrowBox[{{-0.7808875196607532, 
                    0.20058914979977116`}, {-0.7779007863412188, 
                    0.20206916419051552`}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-0.9273943587280378, 
                    0.5}, {-1.2374375565044675`, 0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-0.9273943587280378, 0.5}, {-0.6963288608201873,
                     0.5}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.005], Arrowheads[0.04], 
                   ArrowBox[{{-0.9273943587280378, 
                    0.5}, {-1.0063720585966172`, 0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-0.9273943587280378, 0.5}, 1.6, {
                    1.602233199523079, 1.7467205265843728`}], 
                   ArrowBox[{{-1.191656612087043, 
                    2.0780258114015817`}, {-1.207423394195513, 
                    2.075304332278419}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Dashing[{Small, Small}], 
                   Arrowheads[0.02], 
                   CircleBox[{-0.9273943587280378, 0.5}, 1.7, {
                    1.602233199523079, 1.556485409238625}], 
                   ArrowBox[{{-0.9200658215813603, 
                    2.199984203615813}, {-0.9030666292982898, 
                    2.1998259209639066`}}]}, 
                  {RGBColor[0, 0, 1], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-0.9273943587280378, 0.5}, 1.8, {
                    1.602233199523079, 1.5161100890296262`}], 
                   ArrowBox[{{-0.8469858977240701, 
                    2.2982031251777353`}, {-0.8290081865608618, 
                    2.2973091445620284`}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.003], Arrowheads[0.025], 
                   CircleBox[{-0.9273943587280378, 0.5}, 1.5, {
                    1.602233199523079, 1.6011577868072682`}], 
                   ArrowBox[{{-0.9879201124020099, 
                    1.9987783802624715`}, {-0.9729295521320682, 
                    1.9993086894171117`}}]}, 
                  {RGBColor[0, 0, 1], ArrowBox[{{0, 2.8}, {-1.5, 2.8}}]}},
                 Frame->True,
                 FrameTicks->Automatic,
                 ImageSize->800,
                 PlotRange->{{-3, 3}, {-1, 3}}], 330->
                GraphicsBox[{InsetBox[
                   TagBox[GridBox[{
                    {"\<\"t\"\>", "16.45`"},
                    {"\<\"driveIntent\"\>", 
                    RowBox[{"-", "0.5`"}]},
                    {"\<\"motorTorque\"\>", "15.431875925788662`"},
                    {"\<\"vw\"\>", 
                    RowBox[{"-", "0.3325545446376309`"}]},
                    {"\<\"controllerState\"\>", 
                    RowBox[{"\[LeftAssociation]", 
                    RowBox[{
                    RowBox[{"\<\"lastError\"\>", "\[Rule]", 
                    RowBox[{"-", "0.018934307171639513`"}]}], ",", 
                    
                    RowBox[{"\<\"lastInt\"\>", "\[Rule]", 
                    "0.14101020127896283`"}], ",", 
                    RowBox[{"\<\"initialized\"\>", "\[Rule]", "True"}]}], 
                    "\[RightAssociation]"}]}
                    },
                    AutoDelete->False,
                    GridBoxItemSize->{
                    "Columns" -> {{Automatic}}, "Rows" -> {{Automatic}}}],
                    "Grid"], {-1, -0.5}], 
                  {GrayLevel[0], LineBox[{{-5, 0}, {5, 0}}]}, 
                  {GrayLevel[0.5], CircleBox[{-0.9439904026188771, 0.5}, 0.5],
                    DiskBox[{-0.9439904026188771, 0.5}, 0.1]}, 
                  {GrayLevel[0.5], 
                   DiskBox[{-0.9909389687931074, 1.99926509735076}, 0.1]}, 
                  {GrayLevel[0.5], Thickness[0.01], 
                   LineBox[{{-0.9439904026188771, 0.5}, {-0.9909389687931074, 
                    1.99926509735076}}]}, 
                  {RGBColor[0, 1, 0], 
                   DiskBox[{-0.9752894467350306, 1.4995100649005066`}, 0.05]}, 
                  {RGBColor[0, 0, 1], Arrowheads[0.01], 
                   CircleBox[{-0.9439904026188771, 0.5}, 0.3333333333333333, 
                    NCache[{
                    Rational[-1, 2] 
                    Pi, -1.1078400490212368`}, {-1.5707963267948966`, \
-1.1078400490212368`}]], 
                   ArrowBox[{{-0.7981152021913527, 
                    0.2002809031587398}, {-0.7951253548751134, 
                    0.20175461668056277`}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-0.9439904026188771, 
                    0.5}, {-1.2526279211346503`, 0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-0.9439904026188771, 0.5}, {-0.7138991094405941,
                     0.5}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.005], Arrowheads[0.04], 
                   ArrowBox[{{-0.9439904026188771, 
                    0.5}, {-1.0225366279563675`, 0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-0.9439904026188771, 0.5}, 1.6, {
                    1.6021004834130794`, 1.7459780323107403`}], 
                   ArrowBox[{{-1.2070809081133405`, 
                    2.078221589612393}, {-1.222849706558194, 
                    2.075511817983756}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Dashing[{Small, Small}], 
                   Arrowheads[0.02], 
                   CircleBox[{-0.9439904026188771, 0.5}, 1.7, {
                    1.6021004834130794`, 1.5566024308297288`}], 
                   ArrowBox[{{-0.9368608003782826, 
                    2.1999850495730513`}, {-0.9198615896891184, 
                    2.199828756194753}}]}, 
                  {RGBColor[0, 0, 1], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-0.9439904026188771, 0.5}, 1.8, {
                    1.6021004834130794`, 1.5163678393809201`}], 
                   ArrowBox[{{-0.8640454317679677, 
                    2.2982237907545455`}, {-0.846067490778121, 
                    2.2973344439298513`}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.003], Arrowheads[0.025], 
                   CircleBox[{-0.9439904026188771, 0.5}, 1.5, {
                    1.6021004834130794`, 1.6011262364708627`}], 
                   ArrowBox[{{-1.0044688693006383`, 
                    1.998780289124401}, {-0.9894782923067253, 
                    1.9993101253215582`}}]}, 
                  {RGBColor[0, 0, 1], ArrowBox[{{0, 2.8}, {-1.5, 2.8}}]}},
                 Frame->True,
                 FrameTicks->Automatic,
                 ImageSize->800,
                 PlotRange->{{-3, 3}, {-1, 3}}], 331->
                GraphicsBox[{InsetBox[
                   TagBox[GridBox[{
                    {"\<\"t\"\>", "16.5`"},
                    {"\<\"driveIntent\"\>", 
                    RowBox[{"-", "0.5`"}]},
                    {"\<\"motorTorque\"\>", "15.348920075053318`"},
                    {"\<\"vw\"\>", 
                    RowBox[{"-", "0.3338413336833118`"}]},
                    {"\<\"controllerState\"\>", 
                    RowBox[{"\[LeftAssociation]", 
                    RowBox[{
                    RowBox[{"\<\"lastError\"\>", "\[Rule]", 
                    RowBox[{"-", "0.018706428220444337`"}]}], ",", 
                    
                    RowBox[{"\<\"lastInt\"\>", "\[Rule]", 
                    "0.14006940984036167`"}], ",", 
                    RowBox[{"\<\"initialized\"\>", "\[Rule]", "True"}]}], 
                    "\[RightAssociation]"}]}
                    },
                    AutoDelete->False,
                    
                    GridBoxItemSize->{
                    "Columns" -> {{Automatic}}, "Rows" -> {{Automatic}}}],
                    "Grid"], {-1, -0.5}], 
                  {GrayLevel[0], LineBox[{{-5, 0}, {5, 0}}]}, 
                  {GrayLevel[0.5], CircleBox[{-0.9606509767987436, 0.5}, 0.5],
                    DiskBox[{-0.9606509767987436, 0.5}, 0.1]}, 
                  {GrayLevel[0.5], 
                   DiskBox[{-1.0073616090914808`, 1.9992725292056854`}, 0.1]}, 
                  {GrayLevel[0.5], Thickness[0.01], 
                   LineBox[{{-0.9606509767987436, 0.5}, {-1.0073616090914808`,
                     1.9992725292056854`}}]}, 
                  {RGBColor[0, 1, 0], 
                   DiskBox[{-0.9917913983272351, 1.499515019470457}, 0.05]}, 
                  {RGBColor[0, 0, 1], Arrowheads[0.01], 
                   CircleBox[{-0.9606509767987436, 0.5}, 0.3333333333333333, 
                    NCache[{
                    Rational[-1, 2] 
                    Pi, -1.110328724543297}, {-1.5707963267948966`, \
-1.110328724543297}]], 
                   ArrowBox[{{-0.8155221309202833, 
                    0.19991879564841586`}, {-0.8125286252718752, 
                    0.20138506385436455`}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-0.9606509767987436, 0.5}, {-1.26762937829981, 
                    0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-0.9606509767987436, 0.5}, {-0.7317246443968718,
                     0.5}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.005], Arrowheads[0.04], 
                   ArrowBox[{{-0.9606509767987436, 0.5}, {-1.038703045897938, 
                    0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-0.9606509767987436, 0.5}, 1.6, {
                    1.601941783466221, 1.7450901653954622`}], 
                   ArrowBox[{{-1.2223401280447919`, 
                    2.0784545568752124`}, {-1.2381113261905743`, 
                    2.075758786907236}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Dashing[{Small, Small}], 
                   Arrowheads[0.02], 
                   CircleBox[{-0.9606509767987436, 0.5}, 1.7, {
                    1.601941783466221, 1.5567297427520583`}], 
                   ArrowBox[{{-0.9537378029799364, 
                    2.19998594347946}, {-0.9367385725305262, 
                    2.1998318143046136`}}]}, 
                  {RGBColor[0, 0, 1], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-0.9606509767987436, 0.5}, 1.8, {
                    1.601941783466221, 1.5166700052714805`}], 
                   ArrowBox[{{-0.8812493714823826, 
                    2.298247865304776}, {-0.8632711625829957, 
                    2.2973639508411496`}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.003], Arrowheads[0.025], 
                   CircleBox[{-0.9606509767987436, 0.5}, 1.5, {
                    1.601941783466221, 1.6010669442143384`}], 
                   ArrowBox[{{-1.0210405773088722`, 
                    1.9987838723946247`}, {-1.0060499689261257`, 
                    1.9993128197657146`}}]}, 
                  {RGBColor[0, 0, 1], ArrowBox[{{0, 2.8}, {-1.5, 2.8}}]}},
                 Frame->True,
                 FrameTicks->Automatic,
                 ImageSize->800,
                 PlotRange->{{-3, 3}, {-1, 3}}], 332->
                GraphicsBox[{InsetBox[
                   TagBox[GridBox[{
                    {"\<\"t\"\>", "16.55`"},
                    {"\<\"driveIntent\"\>", 
                    RowBox[{"-", "0.5`"}]},
                    {"\<\"motorTorque\"\>", "15.254631224968806`"},
                    {"\<\"vw\"\>", 
                    RowBox[{"-", "0.3351195371804571`"}]},
                    {"\<\"controllerState\"\>", 
                    RowBox[{"\[LeftAssociation]", 
                    RowBox[{
                    RowBox[{"\<\"lastError\"\>", "\[Rule]", 
                    RowBox[{"-", "0.018504506494645283`"}]}], ",", 
                    
                    RowBox[{"\<\"lastInt\"\>", "\[Rule]", 
                    "0.13913934066761363`"}], ",", 
                    RowBox[{"\<\"initialized\"\>", "\[Rule]", "True"}]}], 
                    "\[RightAssociation]"}]}
                    },
                    AutoDelete->False,
                    
                    GridBoxItemSize->{
                    "Columns" -> {{Automatic}}, "Rows" -> {{Automatic}}}],
                    "Grid"], {-1, -0.5}], 
                  {GrayLevel[0], LineBox[{{-5, 0}, {5, 0}}]}, 
                  {GrayLevel[0.5], CircleBox[{-0.9773756752964453, 0.5}, 0.5],
                    DiskBox[{-0.9773756752964453, 0.5}, 0.1]}, 
                  {GrayLevel[0.5], 
                   DiskBox[{-1.023813387350525, 1.9992810073162344`}, 0.1]}, 
                  {GrayLevel[0.5], Thickness[0.01], 
                   LineBox[{{-0.9773756752964453, 0.5}, {-1.023813387350525, 
                    1.9992810073162344`}}]}, 
                  {RGBColor[0, 1, 0], 
                   DiskBox[{-1.0083341499991652`, 1.4995206715441562`}, 
                    0.05]}, 
                  {RGBColor[0, 0, 1], Arrowheads[0.01], 
                   CircleBox[{-0.9773756752964453, 0.5}, 0.3333333333333333, 
                    NCache[{
                    Rational[-1, 2] 
                    Pi, -1.1131573900458323`}, {-1.5707963267948966`, \
-1.1131573900458323`}]], 
                   ArrowBox[{{-0.8330962382495821, 
                    0.19950947576246048`}, {-0.8300985970004464, 
                    0.20096727048748442`}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-0.9773756752964453, 
                    0.5}, {-1.2824682997958214`, 0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-0.9773756752964453, 0.5}, {-0.7497856235421511,
                     0.5}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.005], Arrowheads[0.04], 
                   ArrowBox[{{-0.9773756752964453, 0.5}, {-1.054878248041527, 
                    0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-0.9773756752964453, 0.5}, 1.6, {
                    1.6017597488721544`, 1.744071745408136}], 
                   ArrowBox[{{-1.2374571614412986`, 
                    2.0787202477211357`}, {-1.253231096833844, 
                    2.076040540851797}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Dashing[{Small, Small}], 
                   Arrowheads[0.02], 
                   CircleBox[{-0.9773756752964453, 0.5}, 1.7, {
                    1.6017597488721544`, 1.5568657470980678`}], 
                   ArrowBox[{{-0.9706937070173307, 
                    2.1999868679786667`}, {-0.9536944557629018, 
                    2.1998350507744577`}}]}, 
                  {RGBColor[0, 0, 1], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-0.9773756752964453, 0.5}, 1.8, {
                    1.6017597488721544`, 1.5170117976223279`}], 
                   ArrowBox[{{-0.8985887019713336, 
                    2.298274899128125}, {-0.8806101920068754, 
                    2.297397129530293}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.003], Arrowheads[0.025], 
                   CircleBox[{-0.9773756752964453, 0.5}, 1.5, {
                    1.6017597488721544`, 1.6009818016667778`}], 
                   ArrowBox[{{-1.0376376653107, 
                    1.9987890086865203`}, {-1.0226470119463622`, 
                    1.999316679717107}}]}, 
                  {RGBColor[0, 0, 1], ArrowBox[{{0, 2.8}, {-1.5, 2.8}}]}},
                 Frame->True,
                 FrameTicks->Automatic,
                 ImageSize->800,
                 PlotRange->{{-3, 3}, {-1, 3}}], 333->
                GraphicsBox[{InsetBox[
                   TagBox[GridBox[{
                    {"\<\"t\"\>", "16.6`"},
                    {"\<\"driveIntent\"\>", 
                    RowBox[{"-", "0.5`"}]},
                    {"\<\"motorTorque\"\>", "15.15030980920825`"},
                    {"\<\"vw\"\>", 
                    RowBox[{"-", "0.3363883019710383`"}]},
                    {"\<\"controllerState\"\>", 
                    RowBox[{"\[LeftAssociation]", 
                    RowBox[{
                    RowBox[{"\<\"lastError\"\>", "\[Rule]", 
                    RowBox[{"-", "0.01832621477080218`"}]}], ",", 
                    
                    RowBox[{"\<\"lastInt\"\>", "\[Rule]", 
                    "0.1382187554132018`"}], ",", 
                    RowBox[{"\<\"initialized\"\>", "\[Rule]", "True"}]}], 
                    "\[RightAssociation]"}]}
                    },
                    AutoDelete->False,
                    
                    GridBoxItemSize->{
                    "Columns" -> {{Automatic}}, "Rows" -> {{Automatic}}}],
                    "Grid"], {-1, -0.5}], 
                  {GrayLevel[0], LineBox[{{-5, 0}, {5, 0}}]}, 
                  {GrayLevel[0.5], CircleBox[{-0.9941640465944778, 0.5}, 0.5],
                    DiskBox[{-0.9941640465944778, 0.5}, 0.1]}, 
                  {GrayLevel[0.5], 
                   DiskBox[{-1.0402977231669246`, 1.99929039344815}, 0.1]}, 
                  {GrayLevel[0.5], Thickness[0.01], 
                   LineBox[{{-0.9941640465944778, 0.5}, {-1.0402977231669246`,
                     1.99929039344815}}]}, 
                  {RGBColor[0, 1, 0], 
                   DiskBox[{-1.024919830976109, 1.4995269289654334`}, 0.05]}, 
                  {RGBColor[0, 0, 1], Arrowheads[0.01], 
                   CircleBox[{-0.9941640465944778, 0.5}, 0.3333333333333333, 
                    NCache[{
                    Rational[-1, 2] 
                    Pi, -1.116287032518649}, {-1.5707963267948966`, \
-1.116287032518649}]], 
                   ArrowBox[{{-0.8508257425033361, 
                    0.19905940504580233`}, {-0.8478235535657896, 
                    0.20050781110147203`}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-0.9941640465944778, 0.5}, {-1.297170242778643, 
                    0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-0.9941640465944778, 0.5}, {-0.7680626495612801,
                     0.5}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.005], Arrowheads[0.04], 
                   ArrowBox[{{-0.9941640465944778, 
                    0.5}, {-1.0710688457454451`, 0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-0.9941640465944778, 0.5}, 1.6, {
                    1.6015569619846963`, 1.7429372180344416`}], 
                   ArrowBox[{{-1.2524542644048908`, 
                    2.0790143011966196`}, {-1.2682312298458527`, 
                    2.076352492009531}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Dashing[{Small, Small}], 
                   Arrowheads[0.02], 
                   CircleBox[{-0.9941640465944778, 0.5}, 1.7, {
                    1.6015569619846963`, 1.5570089404016128`}], 
                   ArrowBox[{{-0.9877255051186448, 
                    2.1999878073632364`}, {-0.970726232299288, 
                    2.1998384243395206`}}]}, 
                  {RGBColor[0, 0, 1], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-0.9941640465944778, 0.5}, 1.8, {
                    1.6015569619846963`, 1.5173885741557616`}], 
                   ArrowBox[{{-0.9160546266283107, 
                    2.2983044565680606`}, {-0.8980757872169938, 
                    2.297433460913034}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.003], Arrowheads[0.025], 
                   CircleBox[{-0.9941640465944778, 0.5}, 1.5, {
                    1.6015569619846963`, 1.600872745248875}], 
                   ArrowBox[{{-1.0542625836902197`, 
                    1.9987955717304984`}, {-1.039271872869114, 
                    1.9993216079309928`}}]}, 
                  {RGBColor[0, 0, 1], ArrowBox[{{0, 2.8}, {-1.5, 2.8}}]}},
                 Frame->True,
                 FrameTicks->Automatic,
                 ImageSize->800,
                 PlotRange->{{-3, 3}, {-1, 3}}], 334->
                GraphicsBox[{InsetBox[
                   TagBox[GridBox[{
                    {"\<\"t\"\>", "16.65`"},
                    {"\<\"driveIntent\"\>", 
                    RowBox[{"-", "0.5`"}]},
                    {"\<\"motorTorque\"\>", "15.037209693234377`"},
                    {"\<\"vw\"\>", 
                    RowBox[{"-", "0.33764688867990483`"}]},
                    {"\<\"controllerState\"\>", 
                    RowBox[{"\[LeftAssociation]", 
                    RowBox[{
                    RowBox[{"\<\"lastError\"\>", "\[Rule]", 
                    RowBox[{"-", "0.018169275418346734`"}]}], ",", 
                    
                    RowBox[{"\<\"lastInt\"\>", "\[Rule]", 
                    "0.13730653089869505`"}], ",", 
                    RowBox[{"\<\"initialized\"\>", "\[Rule]", "True"}]}], 
                    "\[RightAssociation]"}]}
                    },
                    AutoDelete->False,
                    
                    GridBoxItemSize->{
                    "Columns" -> {{Automatic}}, "Rows" -> {{Automatic}}}],
                    "Grid"], {-1, -0.5}], 
                  {GrayLevel[0], LineBox[{{-5, 0}, {5, 0}}]}, 
                  {GrayLevel[0.5], 
                   CircleBox[{-1.0110155994404828`, 0.5}, 0.5], 
                   DiskBox[{-1.0110155994404828`, 0.5}, 0.1]}, 
                  {GrayLevel[0.5], 
                   DiskBox[{-1.0568178703868027`, 1.9993005542506013`}, 0.1]}, 
                  {GrayLevel[0.5], Thickness[0.01], 
                   LineBox[{{-1.0110155994404828`, 
                    0.5}, {-1.0568178703868027`, 1.9993005542506013`}}]}, 
                  {RGBColor[0, 1, 0], 
                   DiskBox[{-1.0415504467380294`, 1.499533702833734}, 0.05]}, 
                  {RGBColor[0, 0, 1], Arrowheads[0.01], 
                   CircleBox[{-1.0110155994404828`, 0.5}, 0.3333333333333333, 
                    NCache[{
                    Rational[-1, 2] 
                    Pi, -1.1196800359978654`}, {-1.5707963267948966`, \
-1.1196800359978654`}]], 
                   ArrowBox[{{-0.8686992109642011, 
                    0.19857479089800872`}, {-0.8656921248705903, 
                    0.20001300219835372`}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-1.0110155994404828`, 
                    0.5}, {-1.3117597933051703`, 0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-1.0110155994404828`, 
                    0.5}, {-0.7865369018650428, 0.5}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.005], Arrowheads[0.04], 
                   ArrowBox[{{-1.0110155994404828`, 
                    0.5}, {-1.0872810957297303`, 0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-1.0110155994404828`, 0.5}, 1.6, {
                    1.6013359210821139`, 1.741700558699529}], 
                   ArrowBox[{{-1.2673529174680926`, 
                    2.0793325107102723`}, {-1.283133162595332, 
                    2.076690214285186}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Dashing[{Small, Small}], 
                   Arrowheads[0.02], 
                   CircleBox[{-1.0110155994404828`, 0.5}, 1.7, {
                    1.6013359210821139`, 1.5571579172664591`}], 
                   ArrowBox[{{-1.0048303168889798`, 
                    2.1999887476921014`}, {-0.9878310220036506, 
                    2.199841897168403}}]}, 
                  {RGBColor[0, 0, 1], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-1.0110155994404828`, 0.5}, 1.8, {
                    1.6013359210821139`, 1.5177958672308116`}], 
                   ArrowBox[{{-0.9336386228847512, 
                    2.298336120834783}, {-0.915659430214179, 
                    2.297472447908586}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.003], Arrowheads[0.025], 
                   CircleBox[{-1.0110155994404828`, 0.5}, 1.5, {
                    1.6013359210821139`, 1.6007417322613853`}], 
                   ArrowBox[{{-1.0709177743355203`, 
                    1.998803432556399}, {-1.0559269947254937`, 
                    1.999327504774575}}]}, 
                  {RGBColor[0, 0, 1], ArrowBox[{{0, 2.8}, {-1.5, 2.8}}]}},
                 Frame->True,
                 FrameTicks->Automatic,
                 ImageSize->800,
                 PlotRange->{{-3, 3}, {-1, 3}}], 335->
                GraphicsBox[{InsetBox[
                   TagBox[GridBox[{
                    {"\<\"t\"\>", "16.7`"},
                    {"\<\"driveIntent\"\>", 
                    RowBox[{"-", "0.5`"}]},
                    {"\<\"motorTorque\"\>", "14.916531443912215`"},
                    {"\<\"vw\"\>", 
                    RowBox[{"-", "0.3388946662852475`"}]},
                    {"\<\"controllerState\"\>", 
                    RowBox[{"\[LeftAssociation]", 
                    RowBox[{
                    RowBox[{"\<\"lastError\"\>", "\[Rule]", 
                    RowBox[{"-", "0.018031476521184038`"}]}], ",", 
                    
                    RowBox[{"\<\"lastInt\"\>", "\[Rule]", 
                    "0.13640165621331582`"}], ",", 
                    RowBox[{"\<\"initialized\"\>", "\[Rule]", "True"}]}], 
                    "\[RightAssociation]"}]}
                    },
                    AutoDelete->False,
                    
                    GridBoxItemSize->{
                    "Columns" -> {{Automatic}}, "Rows" -> {{Automatic}}}],
                    "Grid"], {-1, -0.5}], 
                  {GrayLevel[0], LineBox[{{-5, 0}, {5, 0}}]}, 
                  {GrayLevel[0.5], 
                   CircleBox[{-1.0279298083999329`, 0.5}, 0.5], 
                   DiskBox[{-1.0279298083999329`, 0.5}, 0.1]}, 
                  {GrayLevel[0.5], 
                   DiskBox[{-1.0733769010250305`, 1.9993113625167809`}, 0.1]}, 
                  {GrayLevel[0.5], Thickness[0.01], 
                   LineBox[{{-1.0279298083999329`, 
                    0.5}, {-1.0733769010250305`, 1.9993113625167809`}}]}, 
                  {RGBColor[0, 1, 0], 
                   DiskBox[{-1.0582278701499979`, 1.4995409083445206`}, 
                    0.05]}, 
                  {RGBColor[0, 0, 1], Arrowheads[0.01], 
                   CircleBox[{-1.0279298083999329`, 0.5}, 0.3333333333333333, 
                    NCache[{
                    Rational[-1, 2] 
                    Pi, -1.1233003834775301`}, {-1.5707963267948966`, \
-1.1233003834775301`}]], 
                   ArrowBox[{{-0.8867056141993285, 
                    0.19806153262046738`}, {-0.8836933409992263, 
                    0.19948884782278709`}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-1.0279298083999329`, 
                    0.5}, {-1.3262604372781772`, 0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-1.0279298083999329`, 
                    0.5}, {-0.8051902477948937, 0.5}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.005], Arrowheads[0.04], 
                   ArrowBox[{{-1.0279298083999329`, 0.5}, {-1.103520876673138,
                     0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-1.0279298083999329`, 0.5}, 1.6, {
                    1.6010990259256748`, 1.740375191829398}], 
                   ArrowBox[{{-1.2821737069130508`, 
                    2.0796708644742585`}, {-1.2979574401916412`, 
                    2.077049484977869}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Dashing[{Small, Small}], 
                   Arrowheads[0.02], 
                   CircleBox[{-1.0279298083999329`, 0.5}, 1.7, {
                    1.6010990259256748`, 1.5573113726257555`}], 
                   ArrowBox[{{-1.02200539830431, 
                    2.1999896768407208`}, {-1.0050060810841352`, 
                    2.199845434951646}}]}, 
                  {RGBColor[0, 0, 1], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-1.0279298083999329`, 0.5}, 1.8, {
                    1.6010990259256748`, 1.5182294067928293`}], 
                   ArrowBox[{{-0.9513324889457198, 
                    2.298369497809732}, {-0.9333529235284307, 
                    2.297513619655774}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.003], Arrowheads[0.025], 
                   CircleBox[{-1.0279298083999329`, 0.5}, 1.5, {
                    1.6010990259256748`, 1.600590719523656}], 
                   ArrowBox[{{-1.087605644203132, 
                    1.9988124614578}, {-1.0726147856224566`, 
                    1.9993342698713394`}}]}, 
                  {RGBColor[0, 0, 1], ArrowBox[{{0, 2.8}, {-1.5, 2.8}}]}},
                 Frame->True,
                 FrameTicks->Automatic,
                 ImageSize->800,
                 PlotRange->{{-3, 3}, {-1, 3}}], 336->
                GraphicsBox[{InsetBox[
                   TagBox[GridBox[{
                    {"\<\"t\"\>", "16.75`"},
                    {"\<\"driveIntent\"\>", 
                    RowBox[{"-", "0.5`"}]},
                    {"\<\"motorTorque\"\>", "14.789416886939454`"},
                    {"\<\"vw\"\>", 
                    RowBox[{"-", "0.34013110626521836`"}]},
                    {"\<\"controllerState\"\>", 
                    RowBox[{"\[LeftAssociation]", 
                    RowBox[{
                    RowBox[{"\<\"lastError\"\>", "\[Rule]", 
                    RowBox[{"-", "0.017910685509425747`"}]}], ",", 
                    
                    RowBox[{"\<\"lastInt\"\>", "\[Rule]", 
                    "0.13550322907055792`"}], ",", 
                    RowBox[{"\<\"initialized\"\>", "\[Rule]", "True"}]}], 
                    "\[RightAssociation]"}]}
                    },
                    AutoDelete->False,
                    
                    GridBoxItemSize->{
                    "Columns" -> {{Automatic}}, "Rows" -> {{Automatic}}}],
                    "Grid"], {-1, -0.5}], 
                  {GrayLevel[0], LineBox[{{-5, 0}, {5, 0}}]}, 
                  {GrayLevel[0.5], 
                   CircleBox[{-1.0449061191260807`, 0.5}, 0.5], 
                   DiskBox[{-1.0449061191260807`, 0.5}, 0.1]}, 
                  {GrayLevel[0.5], 
                   DiskBox[{-1.0899776928155023`, 1.9993226981691297`}, 0.1]}, 
                  {GrayLevel[0.5], Thickness[0.01], 
                   LineBox[{{-1.0449061191260807`, 
                    0.5}, {-1.0899776928155023`, 1.9993226981691297`}}]}, 
                  {RGBColor[0, 1, 0], 
                   DiskBox[{-1.0749538349190284`, 1.4995484654460864`}, 
                    0.05]}, 
                  {RGBColor[0, 0, 1], Arrowheads[0.01], 
                   CircleBox[{-1.0449061191260807`, 0.5}, 0.3333333333333333, 
                    NCache[{
                    Rational[-1, 2] 
                    Pi, -1.1271138201867128`}, {-1.5707963267948966`, \
-1.1271138201867128`}]], 
                   ArrowBox[{{-0.9048343722311967, 
                    0.19752517983652484`}, {-0.9018166779707615, 
                    0.19894099757527667`}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-1.0449061191260807`, 
                    0.5}, {-1.3406944568648698`, 0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-1.0449061191260807`, 
                    0.5}, {-0.8240053344610211, 0.5}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.005], Arrowheads[0.04], 
                   ArrowBox[{{-1.0449061191260807`, 0.5}, {-1.11979367219981, 
                    0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-1.0449061191260807`, 0.5}, 1.6, {
                    1.6008485659321487`, 1.7389739247200755`}], 
                   ArrowBox[{{-1.2969362279274548`, 
                    2.0800255770896774`}, {-1.3127236189616442`, 
                    2.0774263173859593`}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Dashing[{Small, Small}], 
                   Arrowheads[0.02], 
                   CircleBox[{-1.0449061191260807`, 0.5}, 1.7, {
                    1.6008485659321487`, 1.557468102770788}], 
                   ArrowBox[{{-1.0392481487307366`, 
                    2.1999905844948096`}, {-1.0222488091122983`, 
                    2.199849006912949}}]}, 
                  {RGBColor[0, 0, 1], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-1.0449061191260807`, 0.5}, 1.8, {
                    1.6008485659321487`, 1.5186851387824851`}], 
                   ArrowBox[{{-0.9691283821071827, 
                    2.298404218904164}, {-0.9511484285059558, 
                    2.297556534701922}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.003], Arrowheads[0.025], 
                   CircleBox[{-1.0449061191260807`, 0.5}, 1.5, {
                    1.6008485659321487`, 1.6004216444767987`}], 
                   ArrowBox[{{-1.1043285422903781`, 
                    1.9988225297296818`}, {-1.0893375956991884`, 
                    1.999341803555658}}]}, 
                  {RGBColor[0, 0, 1], ArrowBox[{{0, 2.8}, {-1.5, 2.8}}]}},
                 Frame->True,
                 FrameTicks->Automatic,
                 ImageSize->800,
                 PlotRange->{{-3, 3}, {-1, 3}}], 337->
                GraphicsBox[{InsetBox[
                   TagBox[GridBox[{
                    {"\<\"t\"\>", "16.8`"},
                    {"\<\"driveIntent\"\>", 
                    RowBox[{"-", "0.5`"}]},
                    {"\<\"motorTorque\"\>", "14.656944855402127`"},
                    {"\<\"vw\"\>", 
                    RowBox[{"-", "0.34135577642663023`"}]},
                    {"\<\"controllerState\"\>", 
                    RowBox[{"\[LeftAssociation]", 
                    RowBox[{
                    RowBox[{"\<\"lastError\"\>", "\[Rule]", 
                    RowBox[{"-", "0.017804860457680285`"}]}], ",", 
                    
                    RowBox[{"\<\"lastInt\"\>", "\[Rule]", 
                    "0.13461045154338458`"}], ",", 
                    RowBox[{"\<\"initialized\"\>", "\[Rule]", "True"}]}], 
                    "\[RightAssociation]"}]}
                    },
                    AutoDelete->False,
                    
                    GridBoxItemSize->{
                    "Columns" -> {{Automatic}}, "Rows" -> {{Automatic}}}],
                    "Grid"], {-1, -0.5}], 
                  {GrayLevel[0], LineBox[{{-5, 0}, {5, 0}}]}, 
                  {GrayLevel[0.5], 
                   CircleBox[{-1.0619439533287018`, 0.5}, 0.5], 
                   DiskBox[{-1.0619439533287018`, 0.5}, 0.1]}, 
                  {GrayLevel[0.5], 
                   DiskBox[{-1.1066229200920485`, 1.9993344489902711`}, 0.1]}, 
                  {GrayLevel[0.5], Thickness[0.01], 
                   LineBox[{{-1.0619439533287018`, 
                    0.5}, {-1.1066229200920485`, 1.9993344489902711`}}]}, 
                  {RGBColor[0, 1, 0], 
                   DiskBox[{-1.091729931170933, 1.4995562993268474`}, 0.05]}, 
                  {RGBColor[0, 0, 1], Arrowheads[0.01], 
                   CircleBox[{-1.0619439533287018`, 0.5}, 0.3333333333333333, 
                    NCache[{
                    Rational[-1, 2] 
                    Pi, -1.1310879811328327`}, {-1.5707963267948966`, \
-1.1310879811328327`}]], 
                   ArrowBox[{{-0.9230753930290163, 
                    0.19697090226942937`}, {-0.9200520959264611, 
                    0.19837471605640677`}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-1.0619439533287018`, 
                    0.5}, {-1.3550828504367445`, 0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-1.0619439533287018`, 
                    0.5}, {-0.8429656625682871, 0.5}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.005], Arrowheads[0.04], 
                   ArrowBox[{{-1.0619439533287018`, 
                    0.5}, {-1.1361045596763295`, 0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-1.0619439533287018`, 0.5}, 1.6, {
                    1.600586710771946, 1.7375088949625002`}], 
                   ArrowBox[{{-1.3116590080023518`, 
                    2.0803931129530198`}, {-1.3274501900856608`, 
                    2.0778169850280985`}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Dashing[{Small, Small}], 
                   Arrowheads[0.02], 
                   CircleBox[{-1.0619439533287018`, 0.5}, 1.7, {
                    1.600586710771946, 1.5576270052835417`}], 
                   ArrowBox[{{-1.056556115779187, 
                    2.1999914620981307`}, {-1.0395567538783317`, 
                    2.1998525857558264`}}]}, 
                  {RGBColor[0, 0, 1], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-1.0619439533287018`, 0.5}, 1.8, {
                    1.600586710771946, 1.5191592393530453`}], 
                   ArrowBox[{{-0.9870188492604752, 
                    2.298439943056305}, {-0.9690384957923887, 
                    2.2976007832552723`}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.003], Arrowheads[0.025], 
                   CircleBox[{-1.0619439533287018`, 0.5}, 1.5, {
                    1.600586710771946, 1.6002364086523313`}], 
                   ArrowBox[{{-1.121088739848102, 
                    1.9988335111771336`}, {-1.1060976973259844`, 
                    1.9993500081338655`}}]}, 
                  {RGBColor[0, 0, 1], ArrowBox[{{0, 2.8}, {-1.5, 2.8}}]}},
                 Frame->True,
                 FrameTicks->Automatic,
                 ImageSize->800,
                 PlotRange->{{-3, 3}, {-1, 3}}], 338->
                GraphicsBox[{InsetBox[
                   TagBox[GridBox[{
                    {"\<\"t\"\>", "16.85`"},
                    {"\<\"driveIntent\"\>", 
                    RowBox[{"-", "0.5`"}]},
                    {"\<\"motorTorque\"\>", "14.520128032637286`"},
                    {"\<\"vw\"\>", 
                    RowBox[{"-", "0.34256833451239976`"}]},
                    {"\<\"controllerState\"\>", 
                    RowBox[{"\[LeftAssociation]", 
                    RowBox[{
                    RowBox[{"\<\"lastError\"\>", "\[Rule]", 
                    RowBox[{"-", "0.01771205921078886`"}]}], ",", 
                    
                    RowBox[{"\<\"lastInt\"\>", "\[Rule]", 
                    "0.13372262529058607`"}], ",", 
                    RowBox[{"\<\"initialized\"\>", "\[Rule]", "True"}]}], 
                    "\[RightAssociation]"}]}
                    },
                    AutoDelete->False,
                    
                    GridBoxItemSize->{
                    "Columns" -> {{Automatic}}, "Rows" -> {{Automatic}}}],
                    "Grid"], {-1, -0.5}], 
                  {GrayLevel[0], LineBox[{{-5, 0}, {5, 0}}]}, 
                  {GrayLevel[0.5], CircleBox[{-1.079042713428248, 0.5}, 0.5], 
                   DiskBox[{-1.079042713428248, 0.5}, 0.1]}, 
                  {GrayLevel[0.5], 
                   DiskBox[{-1.1233150477023917`, 1.9993465111234023`}, 0.1]}, 
                  {GrayLevel[0.5], Thickness[0.01], 
                   LineBox[{{-1.079042713428248, 0.5}, {-1.1233150477023917`, 
                    1.9993465111234023`}}]}, 
                  {RGBColor[0, 1, 0], 
                   DiskBox[{-1.1085576029443438`, 1.4995643407489347`}, 
                    0.05]}, 
                  {RGBColor[0, 0, 1], Arrowheads[0.01], 
                   CircleBox[{-1.079042713428248, 0.5}, 0.3333333333333333, 
                    NCache[{
                    Rational[-1, 2] 
                    Pi, -1.135192485815778}, {-1.5707963267948966`, \
-1.135192485815778}]], 
                   ArrowBox[{{-0.9414191037417461, 
                    0.19640346976955014`}, {-0.9383900701617516, 
                    0.19779486262927498`}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-1.079042713428248, 0.5}, {-1.3694452740809937`,
                     0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-1.079042713428248, 0.5}, {-0.8620556435996904, 
                    0.5}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.005], Arrowheads[0.04], 
                   ArrowBox[{{-1.079042713428248, 0.5}, {-1.1524582042524363`,
                     0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-1.079042713428248, 0.5}, 1.6, {
                    1.6003155032029386`, 1.7359915303621667`}], 
                   ArrowBox[{{-1.3263594489865627`, 
                    2.0807702022472396`}, {-1.3421545218149225`, 
                    2.078218038259446}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Dashing[{Small, Small}], 
                   Arrowheads[0.02], 
                   CircleBox[{-1.079042713428248, 0.5}, 1.7, {
                    1.6003155032029386`, 1.5577870780013487`}], 
                   ArrowBox[{{-1.073926998200185, 
                    2.1999923027642523`}, {-1.056927614286806, 
                    2.1998561475577763`}}]}, 
                  {RGBColor[0, 0, 1], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-1.079042713428248, 0.5}, 1.8, {
                    1.6003155032029386`, 1.5196481252438425`}], 
                   ArrowBox[{{-1.0049968501925062`, 
                    2.298476357959057}, {-0.9870160886197858, 
                    2.29764598859908}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.003], Arrowheads[0.025], 
                   CircleBox[{-1.079042713428248, 0.5}, 1.5, {
                    1.6003155032029386`, 1.6000368633933983`}], 
                   ArrowBox[{{-1.1378884136510252`, 
                    1.9988452833982868`}, {-1.1228972683628484`, 
                    1.9993587889532938`}}]}, 
                  {RGBColor[0, 0, 1], ArrowBox[{{0, 2.8}, {-1.5, 2.8}}]}},
                 Frame->True,
                 FrameTicks->Automatic,
                 ImageSize->800,
                 PlotRange->{{-3, 3}, {-1, 3}}], 339->
                GraphicsBox[{InsetBox[
                   TagBox[GridBox[{
                    {"\<\"t\"\>", "16.9`"},
                    {"\<\"driveIntent\"\>", 
                    RowBox[{"-", "0.5`"}]},
                    {"\<\"motorTorque\"\>", "14.37991079327655`"},
                    {"\<\"vw\"\>", 
                    RowBox[{"-", "0.3437685216752817`"}]},
                    {"\<\"controllerState\"\>", 
                    RowBox[{"\[LeftAssociation]", 
                    RowBox[{
                    RowBox[{"\<\"lastError\"\>", "\[Rule]", 
                    RowBox[{"-", "0.017630446500231586`"}]}], ",", 
                    
                    RowBox[{"\<\"lastInt\"\>", "\[Rule]", 
                    "0.1328391463787733`"}], ",", 
                    RowBox[{"\<\"initialized\"\>", "\[Rule]", "True"}]}], 
                    "\[RightAssociation]"}]}
                    },
                    AutoDelete->False,
                    
                    GridBoxItemSize->{
                    "Columns" -> {{Automatic}}, "Rows" -> {{Automatic}}}],
                    "Grid"], {-1, -0.5}], 
                  {GrayLevel[0], LineBox[{{-5, 0}, {5, 0}}]}, 
                  {GrayLevel[0.5], CircleBox[{-1.096201786886622, 0.5}, 0.5], 
                   DiskBox[{-1.096201786886622, 0.5}, 0.1]}, 
                  {GrayLevel[0.5], 
                   DiskBox[{-1.1400563276628557`, 1.99935878936741}, 0.1]}, 
                  {GrayLevel[0.5], Thickness[0.01], 
                   LineBox[{{-1.096201786886622, 0.5}, {-1.1400563276628557`, 
                    1.99935878936741}}]}, 
                  {RGBColor[0, 1, 0], 
                   DiskBox[{-1.125438147404111, 1.4995725262449402`}, 0.05]}, 
                  {RGBColor[0, 0, 1], Arrowheads[0.01], 
                   CircleBox[{-1.096201786886622, 0.5}, 0.3333333333333333, 
                    NCache[{
                    Rational[-1, 2] 
                    Pi, -1.1393990029966}, {-1.5707963267948966`, \
-1.1393990029966}]], 
                   ArrowBox[{{-0.9598564750627074, 
                    0.19582724143219105`}, {-0.9568216153810165, 
                    0.19720588033751535`}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-1.096201786886622, 0.5}, {-1.383800002752153, 
                    0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-1.096201786886622, 0.5}, {-0.8812606417272462, 
                    0.5}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.005], Arrowheads[0.04], 
                   ArrowBox[{{-1.096201786886622, 0.5}, {-1.1688588575927772`,
                     0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-1.096201786886622, 0.5}, 1.6, {
                    1.6000368539504506`, 1.7344325202952167`}], 
                   ArrowBox[{{-1.3410537862312717`, 
                    2.0811538503311207`}, {-1.356852818712313, 
                    2.078626314112432}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Dashing[{Small, Small}], 
                   Arrowheads[0.02], 
                   CircleBox[{-1.096201786886622, 0.5}, 1.7, {
                    1.6000368539504506`, 1.5579474171367327`}], 
                   ArrowBox[{{-1.0913586470191345`, 
                    2.199993101161362}, {-1.0743592414932632`, 
                    2.199859671623143}}]}, 
                  {RGBColor[0, 0, 1], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-1.096201786886622, 0.5}, 1.8, {
                    1.6000368539504506`, 1.5201484606544697`}], 
                   ArrowBox[{{-1.023055774288554, 
                    2.2985131806136434`}, {-1.005074599503268, 
                    2.2976918077689508`}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.003], Arrowheads[0.025], 
                   CircleBox[{-1.096201786886622, 0.5}, 1.5, {
                    1.6000368539504506`, 1.5998247977074773`}], 
                   ArrowBox[{{-1.154729632135468, 
                    1.9988577288490483`}, {-1.1397383782874742`, 
                    1.9993680552850241`}}]}, 
                  {RGBColor[0, 0, 1], ArrowBox[{{0, 2.8}, {-1.5, 2.8}}]}},
                 Frame->True,
                 FrameTicks->Automatic,
                 ImageSize->800,
                 PlotRange->{{-3, 3}, {-1, 3}}], 340->
                GraphicsBox[{InsetBox[
                   TagBox[GridBox[{
                    {"\<\"t\"\>", "16.95`"},
                    {"\<\"driveIntent\"\>", 
                    RowBox[{"-", "0.5`"}]},
                    {"\<\"motorTorque\"\>", "14.237167947909505`"},
                    {"\<\"vw\"\>", 
                    RowBox[{"-", "0.34495615589651635`"}]},
                    {"\<\"controllerState\"\>", 
                    RowBox[{"\[LeftAssociation]", 
                    RowBox[{
                    RowBox[{"\<\"lastError\"\>", "\[Rule]", 
                    RowBox[{"-", "0.01755829921488263`"}]}], ",", 
                    
                    RowBox[{"\<\"lastInt\"\>", "\[Rule]", 
                    "0.1319594997962938`"}], ",", 
                    RowBox[{"\<\"initialized\"\>", "\[Rule]", "True"}]}], 
                    "\[RightAssociation]"}]}
                    },
                    AutoDelete->False,
                    
                    GridBoxItemSize->{
                    "Columns" -> {{Automatic}}, "Rows" -> {{Automatic}}}],
                    "Grid"], {-1, -0.5}], 
                  {GrayLevel[0], LineBox[{{-5, 0}, {5, 0}}]}, 
                  {GrayLevel[0.5], 
                   CircleBox[{-1.1134205502099217`, 0.5}, 0.5], 
                   DiskBox[{-1.1134205502099217`, 0.5}, 0.1]}, 
                  {GrayLevel[0.5], 
                   DiskBox[{-1.1568487982690858`, 1.9993711972925556`}, 0.1]}, 
                  {GrayLevel[0.5], Thickness[0.01], 
                   LineBox[{{-1.1134205502099217`, 
                    0.5}, {-1.1568487982690858`, 1.9993711972925556`}}]}, 
                  {RGBColor[0, 1, 0], 
                   DiskBox[{-1.1423727155826977`, 1.499580798195037}, 0.05]}, 
                  {RGBColor[0, 0, 1], Arrowheads[0.01], 
                   CircleBox[{-1.1134205502099217`, 0.5}, 0.3333333333333333, 
                    NCache[{
                    Rational[-1, 2] 
                    Pi, -1.1436812883576115`}, {-1.5707963267948966`, \
-1.1436812883576115`}]], 
                   ArrowBox[{{-0.9783790391022765, 
                    0.19524616263470146`}, {-0.9753383035399641, 
                    0.19661179280385982`}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-1.1134205502099217`, 
                    0.5}, {-1.3981639091681117`, 0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-1.1134205502099217`, 
                    0.5}, {-0.9005670018073778, 0.5}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.005], Arrowheads[0.04], 
                   ArrowBox[{{-1.1134205502099217`, 0.5}, {-1.185310360765568,
                     0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-1.1134205502099217`, 0.5}, 1.6, {
                    1.5997525384461575`, 1.7328417974553454`}], 
                   ArrowBox[{{-1.3557570632856524`, 
                    2.081541341359845}, {-1.3715600963856647`, 
                    2.07903894021021}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Dashing[{Small, Small}], 
                   Arrowheads[0.02], 
                   CircleBox[{-1.1134205502099217`, 0.5}, 1.7, {
                    1.5997525384461575`, 1.5581072146694936`}], 
                   ArrowBox[{{-1.1088490651063896`, 
                    2.1999938533782846`}, {-1.0918496384758494`, 
                    2.199863140304819}}]}, 
                  {RGBColor[0, 0, 1], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-1.1134205502099217`, 0.5}, 1.8, {
                    1.5997525384461575`, 1.5206571609577715`}], 
                   ArrowBox[{{-1.0411894512370807`, 
                    2.298550157304815}, {-1.0232078609457458`, 
                    2.297737931595071}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.003], Arrowheads[0.025], 
                   CircleBox[{-1.1134205502099217`, 0.5}, 1.5, {
                    1.5997525384461575`, 1.5996019281230087`}], 
                   ArrowBox[{{-1.171614344208763, 
                    1.998870735700721}, {-1.1566229769968435`, 
                    1.9993777210295347`}}]}, 
                  {RGBColor[0, 0, 1], ArrowBox[{{0, 2.8}, {-1.5, 2.8}}]}},
                 Frame->True,
                 FrameTicks->Automatic,
                 ImageSize->800,
                 PlotRange->{{-3, 3}, {-1, 3}}], 341->
                GraphicsBox[{InsetBox[
                   TagBox[GridBox[{
                    {"\<\"t\"\>", "17.`"},
                    {"\<\"driveIntent\"\>", "0"},
                    {"\<\"motorTorque\"\>", "200"},
                    {"\<\"vw\"\>", 
                    RowBox[{"-", "0.34613112541938335`"}]},
                    {"\<\"controllerState\"\>", 
                    RowBox[{"\[LeftAssociation]", 
                    RowBox[{
                    
                    RowBox[{"\<\"lastError\"\>", "\[Rule]", 
                    "0.13250599001141755`"}], ",", 
                    RowBox[{"\<\"lastInt\"\>", "\[Rule]", 
                    "0.1312332537472036`"}], ",", 
                    RowBox[{"\<\"initialized\"\>", "\[Rule]", "True"}]}], 
                    "\[RightAssociation]"}]}
                    },
                    AutoDelete->False,
                    
                    GridBoxItemSize->{
                    "Columns" -> {{Automatic}}, "Rows" -> {{Automatic}}}],
                    "Grid"], {-1, -0.5}], 
                  {GrayLevel[0], LineBox[{{-5, 0}, {5, 0}}]}, 
                  {GrayLevel[0.5], CircleBox[{-1.130698372622222, 0.5}, 0.5], 
                   DiskBox[{-1.130698372622222, 0.5}, 0.1]}, 
                  {GrayLevel[0.5], 
                   DiskBox[{-1.1736942853874817`, 1.999383657202346}, 0.1]}, 
                  {GrayLevel[0.5], Thickness[0.01], 
                   LineBox[{{-1.130698372622222, 0.5}, {-1.1736942853874817`, 
                    1.999383657202346}}]}, 
                  {RGBColor[0, 1, 0], 
                   DiskBox[{-1.1593623144657283`, 1.4995891048015642`}, 
                    0.05]}, 
                  {RGBColor[0, 0, 1], Arrowheads[0.01], 
                   CircleBox[{-1.130698372622222, 0.5}, 0.3333333333333333, 
                    NCache[{
                    Rational[-1, 2] Pi, 
                    4.429203673205103}, {-1.5707963267948966`, 
                    4.429203673205103}]], 
                   ArrowBox[{{-1.2270327294153422`, 
                    0.18089060995895645`}, {-1.223836872021864, 
                    0.1799432377832113}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-1.130698372622222, 0.5}, {-5.130698372622222, 
                    0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-1.130698372622222, 0.5}, {-0.9199620647937894, 
                    0.5}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.005], Arrowheads[0.04], 
                   ArrowBox[{{-1.130698372622222, 0.5}, {-4.919962064793789, 
                    0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-1.130698372622222, 0.5}, 1.6, {
                    1.5994641952424702`, 1.7312285289672933`}], 
                   ArrowBox[{{-1.3704831206394315`, 
                    2.0819302369631614`}, {-1.386290170217851, 
                    2.079453333594201}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Dashing[{Small, Small}], 
                   Arrowheads[0.02], 
                   CircleBox[{-1.130698372622222, 0.5}, 1.7, {
                    1.5994641952424702`, -0.5956515321805758}], 
                   ArrowBox[{{0.2860002539739235, -0.4396621740820448}, {
                    0.2765329545016417, -0.4537819415155696}}]}, 
                  {RGBColor[0, 0, 1], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-1.130698372622222, 0.5}, 1.8, {
                    1.5994641952424702`, 0.488353084131359}], 
                   ArrowBox[{{0.4503693282220613, 1.3603632519738234`}, {
                    0.45889376462237474`, 1.3445098206712311`}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.003], Arrowheads[0.025], 
                   CircleBox[{-1.130698372622222, 0.5}, 1.5, {
                    1.5994641952424702`, -2.0811183398638367`}], 
                   ArrowBox[{{-1.8502605328176107`, -0.816142202657048}, \
{-1.8633857576800508`, -0.808880894419779}}]}, 
                  {RGBColor[0, 0, 1], ArrowBox[{{0, 2.8}, {0, 2.8}}]}},
                 Frame->True,
                 FrameTicks->Automatic,
                 ImageSize->800,
                 PlotRange->{{-3, 3}, {-1, 3}}], 342->
                GraphicsBox[{InsetBox[
                   TagBox[GridBox[{
                    {"\<\"t\"\>", "17.05`"},
                    {"\<\"driveIntent\"\>", "0"},
                    {"\<\"motorTorque\"\>", "35.09304160094388`"},
                    {"\<\"vw\"\>", 
                    RowBox[{"-", "0.3720107303339527`"}]},
                    {"\<\"controllerState\"\>", 
                    RowBox[{"\[LeftAssociation]", 
                    RowBox[{
                    
                    RowBox[{"\<\"lastError\"\>", "\[Rule]", 
                    "0.11557648375520803`"}], ",", 
                    
                    RowBox[{"\<\"lastInt\"\>", "\[Rule]", 
                    "0.13762984390653957`"}], ",", 
                    RowBox[{"\<\"initialized\"\>", "\[Rule]", "True"}]}], 
                    "\[RightAssociation]"}]}
                    },
                    AutoDelete->False,
                    
                    GridBoxItemSize->{
                    "Columns" -> {{Automatic}}, "Rows" -> {{Automatic}}}],
                    "Grid"], {-1, -0.5}], 
                  {GrayLevel[0], LineBox[{{-5, 0}, {5, 0}}]}, 
                  {GrayLevel[0.5], CircleBox[{-1.148811969728682, 0.5}, 0.5], 
                   DiskBox[{-1.148811969728682, 0.5}, 0.1]}, 
                  {GrayLevel[0.5], 
                   DiskBox[{-1.1536572578559428`, 1.999992174373908}, 0.1]}, 
                  {GrayLevel[0.5], Thickness[0.01], 
                   LineBox[{{-1.148811969728682, 0.5}, {-1.1536572578559428`, 
                    1.999992174373908}}]}, 
                  {RGBColor[0, 1, 0], 
                   DiskBox[{-1.1520421618135226`, 1.4999947829159386`}, 
                    0.05]}, 
                  {RGBColor[0, 0, 1], Arrowheads[0.01], 
                   CircleBox[{-1.148811969728682, 0.5}, 0.3333333333333333, 
                    NCache[{
                    Rational[-1, 2] 
                    Pi, -0.5180050787665802}, {-1.5707963267948966`, \
-0.5180050787665802}]], 
                   ArrowBox[{{-0.8608740190014481, 
                    0.3320629652455658}, {-0.8592090734208322, 
                    0.33495069354518375`}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-1.148811969728682, 0.5}, {-1.8506728017475598`,
                     0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-1.148811969728682, 0.5}, {-1.125054071269681, 
                    0.5}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.005], Arrowheads[0.04], 
                   ArrowBox[{{-1.148811969728682, 0.5}, {-1.8269149032885585`,
                     0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-1.148811969728682, 0.5}, 1.6, {
                    1.5740265244971434`, 1.5888752885012691`}], 
                   ArrowBox[{{-1.1617381678431293`, 
                    2.099947784586205}, {-1.1777367327278407`, 
                    2.0997385280368297`}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Dashing[{Small, Small}], 
                   Arrowheads[0.02], 
                   CircleBox[{-1.148811969728682, 0.5}, 1.7, {
                    1.5740265244971434`, 1.181039380238949}], 
                   ArrowBox[{{-0.5186309060828128, 
                    2.0788830947926957`}, {-0.5028738470713603, 
                    2.0725024456890155`}}]}, 
                  {RGBColor[0, 0, 1], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-1.148811969728682, 0.5}, 1.8, {
                    1.5740265244971434`, 1.3790651822696773`}], 
                   ArrowBox[{{-0.823493505847068, 
                    2.2703581267804847`}, {-0.8058064854251208, 
                    2.267016479192449}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.003], Arrowheads[0.025], 
                   CircleBox[{-1.148811969728682, 0.5}, 1.5, {
                    1.5740265244971434`, 0.910526165269298}], 
                   ArrowBox[{{-0.2407097874364349, 
                    1.693880407124624}, {-0.22881658707500163`, 
                    1.6847398431284009`}}]}, 
                  {RGBColor[0, 0, 1], ArrowBox[{{0, 2.8}, {0, 2.8}}]}},
                 Frame->True,
                 FrameTicks->Automatic,
                 ImageSize->800,
                 PlotRange->{{-3, 3}, {-1, 3}}], 343->
                GraphicsBox[{InsetBox[
                   TagBox[GridBox[{
                    {"\<\"t\"\>", "17.1`"},
                    {"\<\"driveIntent\"\>", "0"},
                    {"\<\"motorTorque\"\>", 
                    RowBox[{"-", "19.25668050614681`"}]},
                    {"\<\"vw\"\>", 
                    RowBox[{"-", "0.3761494778026032`"}]},
                    {"\<\"controllerState\"\>", 
                    RowBox[{"\[LeftAssociation]", 
                    RowBox[{
                    
                    RowBox[{"\<\"lastError\"\>", "\[Rule]", 
                    "0.07104044938671704`"}], ",", 
                    
                    RowBox[{"\<\"lastInt\"\>", "\[Rule]", 
                    "0.14232070172436584`"}], ",", 
                    RowBox[{"\<\"initialized\"\>", "\[Rule]", "True"}]}], 
                    "\[RightAssociation]"}]}
                    },
                    AutoDelete->False,
                    
                    GridBoxItemSize->{
                    "Columns" -> {{Automatic}}, "Rows" -> {{Automatic}}}],
                    "Grid"], {-1, -0.5}], 
                  {GrayLevel[0], LineBox[{{-5, 0}, {5, 0}}]}, 
                  {GrayLevel[0.5], CircleBox[{-1.167568847477195, 0.5}, 0.5], 
                   DiskBox[{-1.167568847477195, 0.5}, 0.1]}, 
                  {GrayLevel[0.5], 
                   DiskBox[{-1.1034180489385434`, 1.9986275971857896`}, 0.1]}, 
                  {GrayLevel[0.5], Thickness[0.01], 
                   LineBox[{{-1.167568847477195, 0.5}, {-1.1034180489385434`, 
                    1.9986275971857896`}}]}, 
                  {RGBColor[0, 1, 0], 
                   DiskBox[{-1.1248016484514274`, 1.4990850647905263`}, 
                    0.05]}, 
                  {RGBColor[0, 0, 1], Arrowheads[0.01], 
                   CircleBox[{-1.167568847477195, 0.5}, 0.3333333333333333, 
                    NCache[{
                    Rational[-1, 2] 
                    Pi, -2.148496741979301}, {-1.5707963267948966`, \
-2.148496741979301}]], 
                   ArrowBox[{{-1.3468003811207556`, 
                    0.21895344076312145`}, {-1.3496018383702628`, 
                    0.22075977843864358`}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-1.167568847477195, 0.5}, {-0.7824352373542588, 
                    0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-1.167568847477195, 0.5}, {-1.4818332668256988`,
                     0.5}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.005], Arrowheads[0.04], 
                   ArrowBox[{{-1.167568847477195, 0.5}, {-1.0966996567027625`,
                     0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-1.167568847477195, 0.5}, 1.6, {
                    1.5280160799217635`, 1.33142094601893}], 
                   ArrowBox[{{-0.8037779541107963, 
                    2.0580937667238377`}, {-0.7882154655176433, 
                    2.0543780143825536`}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Dashing[{Small, Small}], 
                   Arrowheads[0.02], 
                   CircleBox[{-1.167568847477195, 0.5}, 1.7, {
                    1.5280160799217635`, 1.5690493530195264`}], 
                   ArrowBox[{{-1.1475994513339243`, 
                    2.199882708664828}, {-1.164598993569691, 
                    2.1999974058708935`}}]}, 
                  {RGBColor[0, 0, 1], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-1.167568847477195, 0.5}, 1.8, {
                    1.5280160799217635`, 1.634997638289246}], 
                   ArrowBox[{{-1.265083445689439, 
                    2.297356643277985}, {-1.2830518368749964`, 
                    2.2962916464649465`}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.003], Arrowheads[0.025], 
                   CircleBox[{-1.167568847477195, 0.5}, 1.5, {
                    1.5280160799217635`, 1.4931968499771844`}], 
                   ArrowBox[{{-1.0662468417582431`, 
                    1.996574037980443}, {-1.0512864168642664`, 
                    1.9954860067318414`}}]}, 
                  {RGBColor[0, 0, 1], ArrowBox[{{0, 2.8}, {0, 2.8}}]}},
                 Frame->True,
                 FrameTicks->Automatic,
                 ImageSize->800,
                 PlotRange->{{-3, 3}, {-1, 3}}], 344->
                GraphicsBox[{InsetBox[
                   TagBox[GridBox[{
                    {"\<\"t\"\>", "17.150000000000002`"},
                    {"\<\"driveIntent\"\>", "0"},
                    {"\<\"motorTorque\"\>", 
                    RowBox[{"-", "52.39323438110192`"}]},
                    {"\<\"vw\"\>", 
                    RowBox[{"-", "0.3721255180778582`"}]},
                    {"\<\"controllerState\"\>", 
                    RowBox[{"\[LeftAssociation]", 
                    RowBox[{
                    
                    RowBox[{"\<\"lastError\"\>", "\[Rule]", 
                    "0.021976036081718178`"}], ",", 
                    
                    RowBox[{"\<\"lastInt\"\>", "\[Rule]", 
                    "0.14461923780413863`"}], ",", 
                    RowBox[{"\<\"initialized\"\>", "\[Rule]", "True"}]}], 
                    "\[RightAssociation]"}]}
                    },
                    AutoDelete->False,
                    
                    GridBoxItemSize->{
                    "Columns" -> {{Automatic}}, "Rows" -> {{Automatic}}}],
                    "Grid"], {-1, -0.5}], 
                  {GrayLevel[0], LineBox[{{-5, 0}, {5, 0}}]}, 
                  {GrayLevel[0.5], 
                   CircleBox[{-1.1862952150022696`, 0.5}, 0.5], 
                   DiskBox[{-1.1862952150022696`, 0.5}, 0.1]}, 
                  {GrayLevel[0.5], 
                   DiskBox[{-1.0505401149926386`, 1.993844219730215}, 0.1]}, 
                  {GrayLevel[0.5], Thickness[0.01], 
                   LineBox[{{-1.1862952150022696`, 
                    0.5}, {-1.0505401149926386`, 1.993844219730215}}]}, 
                  {RGBColor[0, 1, 0], 
                   DiskBox[{-1.0957918149958488`, 1.49589614648681}, 0.05]}, 
                  {RGBColor[0, 0, 1], Arrowheads[0.01], 
                   CircleBox[{-1.1862952150022696`, 0.5}, 0.3333333333333333, 
                    NCache[{
                    Rational[-1, 2] 
                    Pi, -3.142593358227954}, {-1.5707963267948966`, \
-3.142593358227954}]], 
                   ArrowBox[{{-1.5196150505405308`, 
                    0.4970002753697112}, {-1.5196283814339882`, 
                    0.5003335681570472}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-1.1862952150022696`, 
                    0.5}, {-0.13843052738023132`, 0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-1.1862952150022696`, 
                    0.5}, {-1.8492148550200436`, 0.5}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.005], Arrowheads[0.04], 
                   ArrowBox[{{-1.1862952150022696`, 
                    0.5}, {-0.8013501673980051, 0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-1.1862952150022696`, 0.5}, 1.6, {
                    1.4801689189685863`, 1.0641368091218832`}], 
                   ArrowBox[{{-0.4239091028292128, 
                    1.9066866800981843`}, {-0.4098805894627876, 
                    1.898992612292283}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Dashing[{Small, Small}], 
                   Arrowheads[0.02], 
                   CircleBox[{-1.1862952150022696`, 0.5}, 1.7, {
                    1.4801689189685863`, 1.7023249743712492`}], 
                   ArrowBox[{{-1.3923857412858915`, 
                    2.18746161288965}, {-1.4092497717321486`, 
                    2.1853163695975253`}}]}, 
                  {RGBColor[0, 0, 1], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-1.1862952150022696`, 0.5}, 1.8, {
                    1.4801689189685863`, 1.7712424433080414`}], 
                   ArrowBox[{{-1.527029751055272, 
                    2.26745579179247}, {-1.5446869778138739`, 
                    2.2639601311675928`}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.003], Arrowheads[0.025], 
                   CircleBox[{-1.1862952150022696`, 0.5}, 1.5, {
                    1.4801689189685863`, 1.637466427129134}], 
                   ArrowBox[{{-1.2712548737984133`, 
                    1.9975920193354542`}, {-1.2862262964468059`, 
                    1.9966675579303927`}}]}, 
                  {RGBColor[0, 0, 1], ArrowBox[{{0, 2.8}, {0, 2.8}}]}},
                 Frame->True,
                 FrameTicks->Automatic,
                 ImageSize->800,
                 PlotRange->{{-3, 3}, {-1, 3}}], 345->
                GraphicsBox[{InsetBox[
                   TagBox[GridBox[{
                    {"\<\"t\"\>", "17.2`"},
                    {"\<\"driveIntent\"\>", "0"},
                    {"\<\"motorTorque\"\>", 
                    RowBox[{"-", "76.53593236718008`"}]},
                    {"\<\"vw\"\>", 
                    RowBox[{"-", "0.36426401660469454`"}]},
                    {"\<\"controllerState\"\>", 
                    RowBox[{"\[LeftAssociation]", 
                    RowBox[{
                    RowBox[{"\<\"lastError\"\>", "\[Rule]", 
                    RowBox[{"-", "0.024404441223986023`"}]}], ",", 
                    
                    RowBox[{"\<\"lastInt\"\>", "\[Rule]", 
                    "0.1445177190865197`"}], ",", 
                    RowBox[{"\<\"initialized\"\>", "\[Rule]", "True"}]}], 
                    "\[RightAssociation]"}]}
                    },
                    AutoDelete->False,
                    
                    GridBoxItemSize->{
                    "Columns" -> {{Automatic}}, "Rows" -> {{Automatic}}}],
                    "Grid"], {-1, -0.5}], 
                  {GrayLevel[0], LineBox[{{-5, 0}, {5, 0}}]}, 
                  {GrayLevel[0.5], CircleBox[{-1.204712825555364, 0.5}, 0.5], 
                   DiskBox[{-1.204712825555364, 0.5}, 0.1]}, 
                  {GrayLevel[0.5], 
                   DiskBox[{-1.0034736022524657`, 1.986439630460802}, 0.1]}, 
                  {GrayLevel[0.5], Thickness[0.01], 
                   LineBox[{{-1.204712825555364, 0.5}, {-1.0034736022524657`, 
                    1.986439630460802}}]}, 
                  {RGBColor[0, 1, 0], 
                   DiskBox[{-1.0705533433534318`, 1.4909597536405348`}, 
                    0.05]}, 
                  {RGBColor[0, 0, 1], Arrowheads[0.01], 
                   CircleBox[{-1.204712825555364, 0.5}, 0.3333333333333333, 
                    NCache[{
                    Rational[-1, 2] 
                    Pi, -3.8668742978102992`}, {-1.5707963267948966`, \
-3.8668742978102992`}]], 
                   ArrowBox[{{-1.4563490123585012`, 
                    0.7186100194462454}, {-1.4541503668943678`, 
                    0.7211154089652367}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-1.204712825555364, 0.5}, {0.32600582178823756`,
                     0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-1.204712825555364, 0.5}, {-2.182533755579987, 
                    0.5}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.005], Arrowheads[0.04], 
                   ArrowBox[{{-1.204712825555364, 0.5}, {-0.6518151082363851, 
                    0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-1.204712825555364, 0.5}, 1.6, {
                    1.436231098739231, 0.8195177771131785}], 
                   ArrowBox[{{-0.12434237917761637`, 
                    1.6801693516582892`}, {-0.11259490042710807`, 
                    1.669306819279505}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Dashing[{Small, Small}], 
                   Arrowheads[0.02], 
                   CircleBox[{-1.204712825555364, 0.5}, 1.7, {
                    1.436231098739231, 1.753697693920591}], 
                   ArrowBox[{{-1.4971828251212045`, 
                    2.1746525906449845`}, {-1.5139144485421685`, 
                    2.1716442074623195`}}]}, 
                  {RGBColor[0, 0, 1], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-1.204712825555364, 0.5}, 1.8, {
                    1.436231098739231, 1.8614307230013427`}], 
                   ArrowBox[{{-1.703250334661284, 
                    2.229583866718947}, {-1.7205209583982048`, 
                    2.224512096244379}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.003], Arrowheads[0.025], 
                   CircleBox[{-1.204712825555364, 0.5}, 1.5, {
                    1.436231098739231, 1.64843857932485}], 
                   ArrowBox[{{-1.306098848201995, 
                    1.9965697024902973`}, {-1.3210592265409733`, 
                    1.995481031299861}}]}, 
                  {RGBColor[0, 0, 1], ArrowBox[{{0, 2.8}, {0, 2.8}}]}},
                 Frame->True,
                 FrameTicks->Automatic,
                 ImageSize->800,
                 PlotRange->{{-3, 3}, {-1, 3}}], 346->
                GraphicsBox[{InsetBox[
                   TagBox[GridBox[{
                    {"\<\"t\"\>", "17.25`"},
                    {"\<\"driveIntent\"\>", "0"},
                    {"\<\"motorTorque\"\>", 
                    RowBox[{"-", "95.56252932650239`"}]},
                    {"\<\"vw\"\>", 
                    RowBox[{"-", "0.3540571966964894`"}]},
                    {"\<\"controllerState\"\>", 
                    RowBox[{"\[LeftAssociation]", 
                    RowBox[{
                    RowBox[{"\<\"lastError\"\>", "\[Rule]", 
                    RowBox[{"-", "0.06594230208698589`"}]}], ",", 
                    
                    RowBox[{"\<\"lastInt\"\>", "\[Rule]", 
                    "0.14221632119647373`"}], ",", 
                    RowBox[{"\<\"initialized\"\>", "\[Rule]", "True"}]}], 
                    "\[RightAssociation]"}]}
                    },
                    AutoDelete->False,
                    
                    GridBoxItemSize->{
                    "Columns" -> {{Automatic}}, "Rows" -> {{Automatic}}}],
                    "Grid"], {-1, -0.5}], 
                  {GrayLevel[0], LineBox[{{-5, 0}, {5, 0}}]}, 
                  {GrayLevel[0.5], 
                   CircleBox[{-1.2226738914147017`, 0.5}, 0.5], 
                   DiskBox[{-1.2226738914147017`, 0.5}, 0.1]}, 
                  {GrayLevel[0.5], 
                   DiskBox[{-0.9645613253167828, 1.97762576561941}, 0.1]}, 
                  {GrayLevel[0.5], Thickness[0.01], 
                   LineBox[{{-1.2226738914147017`, 0.5}, {-0.9645613253167828,
                     1.97762576561941}}]}, 
                  {RGBColor[0, 1, 0], 
                   DiskBox[{-1.0505988473494225`, 1.4850838437462732`}, 
                    0.05]}, 
                  {RGBColor[0, 0, 1], Arrowheads[0.01], 
                   CircleBox[{-1.2226738914147017`, 0.5}, 0.3333333333333333, 
                    NCache[{
                    Rational[-1, 2] 
                    Pi, -4.437672206589968}, {-1.5707963267948966`, \
-4.437672206589968}]], 
                   ArrowBox[{{-1.3163024376459946`, 
                    0.8199137484412411}, {-1.3130986720919742`, 
                    0.8208340227447493}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-1.2226738914147017`, 0.5}, {0.688576695115346, 
                    0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-1.2226738914147017`, 0.5}, {-2.469405656071147,
                     0.5}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.005], Arrowheads[0.04], 
                   ArrowBox[{{-1.2226738914147017`, 
                    0.5}, {-0.5581550695410991, 0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-1.2226738914147017`, 0.5}, 1.6, {
                    1.3978605790947949`, 0.606854455243496}], 
                   ArrowBox[{{0.08244946748749449, 1.4255555186199527`}, {
                    0.09163961279106658, 1.4124582251601185`}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Dashing[{Small, Small}], 
                   Arrowheads[0.02], 
                   CircleBox[{-1.2226738914147017`, 0.5}, 1.7, {
                    1.3978605790947949`, 1.7771055028262162`}], 
                   ArrowBox[{{-1.5542601368990034`, 
                    2.167348362462273}, {-1.5709167634595083`, 
                    2.1639491885481315`}}]}, 
                  {RGBColor[0, 0, 1], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-1.2226738914147017`, 0.5}, 1.8, {
                    1.3978605790947949`, 1.9287635197975859`}], 
                   ArrowBox[{{-1.8364514577771835`, 
                    2.192122069778108}, {-1.8533417078334515`, 
                    2.1858997910117925`}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.003], Arrowheads[0.025], 
                   CircleBox[{-1.2226738914147017`, 0.5}, 1.5, {
                    1.3978605790947949`, 1.621015156059036}], 
                   ArrowBox[{{-1.2829858725919339`, 
                    1.9987869978507544`}, {-1.2979704771999285`, 
                    1.9981089493655286`}}]}, 
                  {RGBColor[0, 0, 1], ArrowBox[{{0, 2.8}, {0, 2.8}}]}},
                 Frame->True,
                 FrameTicks->Automatic,
                 ImageSize->800,
                 PlotRange->{{-3, 3}, {-1, 3}}], 347->
                GraphicsBox[{InsetBox[
                   TagBox[GridBox[{
                    {"\<\"t\"\>", "17.3`"},
                    {"\<\"driveIntent\"\>", "0"},
                    {"\<\"motorTorque\"\>", 
                    RowBox[{"-", "110.78537739560133`"}]},
                    {"\<\"vw\"\>", 
                    RowBox[{"-", "0.3421348049077366`"}]},
                    {"\<\"controllerState\"\>", 
                    RowBox[{"\[LeftAssociation]", 
                    RowBox[{
                    RowBox[{"\<\"lastError\"\>", "\[Rule]", 
                    RowBox[{"-", "0.10208013284046495`"}]}], ",", 
                    
                    RowBox[{"\<\"lastInt\"\>", "\[Rule]", 
                    "0.13797492723585703`"}], ",", 
                    RowBox[{"\<\"initialized\"\>", "\[Rule]", "True"}]}], 
                    "\[RightAssociation]"}]}
                    },
                    AutoDelete->False,
                    
                    GridBoxItemSize->{
                    "Columns" -> {{Automatic}}, "Rows" -> {{Automatic}}}],
                    "Grid"], {-1, -0.5}], 
                  {GrayLevel[0], LineBox[{{-5, 0}, {5, 0}}]}, 
                  {GrayLevel[0.5], 
                   CircleBox[{-1.2400789976037967`, 0.5}, 0.5], 
                   DiskBox[{-1.2400789976037967`, 0.5}, 0.1]}, 
                  {GrayLevel[0.5], 
                   DiskBox[{-0.9341589928606908, 1.9684729996489476`}, 0.1]}, 
                  {GrayLevel[0.5], Thickness[0.01], 
                   LineBox[{{-1.2400789976037967`, 0.5}, {-0.9341589928606908,
                     1.9684729996489476`}}]}, 
                  {RGBColor[0, 1, 0], 
                   DiskBox[{-1.0361323277750594`, 1.4789819997659652`}, 
                    0.05]}, 
                  {RGBColor[0, 0, 1], Arrowheads[0.01], 
                   CircleBox[{-1.2400789976037967`, 0.5}, 0.3333333333333333, 
                    NCache[{
                    Rational[-1, 2] 
                    Pi, -4.894357648662936}, {-1.5707963267948966`, \
-4.894357648662936}]], 
                   ArrowBox[{{-1.1830382277925522`, 
                    0.8284165977694363}, {-1.1797569685654075`, 
                    0.8278297788850226}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-1.2400789976037967`, 0.5}, {0.9756285503082298,
                     0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-1.2400789976037967`, 
                    0.5}, {-2.7085766747036004`, 0.5}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.005], Arrowheads[0.04], 
                   ArrowBox[{{-1.2400789976037967`, 
                    0.5}, {-0.4928691267915738, 0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-1.2400789976037967`, 0.5}, 1.6, {
                    1.365408684723448, 0.42789293018782937`}], 
                   ArrowBox[{{0.20895642369658773`, 1.178451433646372}, {
                    0.21566837379107562`, 1.1639273986490692`}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Dashing[{Small, Small}], 
                   Arrowheads[0.02], 
                   CircleBox[{-1.2400789976037967`, 0.5}, 1.7, {
                    1.365408684723448, 1.7891450403292222`}], 
                   ArrowBox[{{-1.5917148297739607`, 
                    2.163235473868326}, {-1.6083293256630236`, 
                    2.159636013071623}}]}, 
                  {RGBColor[0, 0, 1], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-1.2400789976037967`, 0.5}, 1.8, {
                    1.365408684723448, 1.9808830035878997`}], 
                   ArrowBox[{{-1.9411757134297212`, 
                    2.1578490266179555`}, {-1.9577188728454424`, 
                    2.1507552845480022`}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.003], Arrowheads[0.025], 
                   CircleBox[{-1.2400789976037967`, 0.5}, 1.5, {
                    1.365408684723448, 1.584195598876008}], 
                   ArrowBox[{{-1.245177895905779, 
                    1.9999913337203339`}, {-1.260177304302886, 
                    1.9998653466454344`}}]}, 
                  {RGBColor[0, 0, 1], ArrowBox[{{0, 2.8}, {0, 2.8}}]}},
                 Frame->True,
                 FrameTicks->Automatic,
                 ImageSize->800,
                 PlotRange->{{-3, 3}, {-1, 3}}], 348->
                GraphicsBox[{InsetBox[
                   TagBox[GridBox[{
                    {"\<\"t\"\>", "17.35`"},
                    {"\<\"driveIntent\"\>", "0"},
                    {"\<\"motorTorque\"\>", 
                    RowBox[{"-", "122.76189246934729`"}]},
                    {"\<\"vw\"\>", 
                    RowBox[{"-", "0.32887662341888213`"}]},
                    {"\<\"controllerState\"\>", 
                    RowBox[{"\[LeftAssociation]", 
                    RowBox[{
                    RowBox[{"\<\"lastError\"\>", "\[Rule]", 
                    RowBox[{"-", "0.1327959861712918`"}]}], ",", 
                    
                    RowBox[{"\<\"lastInt\"\>", "\[Rule]", 
                    "0.1320653691876726`"}], ",", 
                    RowBox[{"\<\"initialized\"\>", "\[Rule]", "True"}]}], 
                    "\[RightAssociation]"}]}
                    },
                    AutoDelete->False,
                    
                    GridBoxItemSize->{
                    "Columns" -> {{Automatic}}, "Rows" -> {{Automatic}}}],
                    "Grid"], {-1, -0.5}], 
                  {GrayLevel[0], LineBox[{{-5, 0}, {5, 0}}]}, 
                  {GrayLevel[0.5], 
                   CircleBox[{-1.2568525437186542`, 0.5}, 0.5], 
                   DiskBox[{-1.2568525437186542`, 0.5}, 0.1]}, 
                  {GrayLevel[0.5], 
                   DiskBox[{-0.9119381307402327, 1.959806167859196}, 0.1]}, 
                  {GrayLevel[0.5], Thickness[0.01], 
                   LineBox[{{-1.2568525437186542`, 0.5}, {-0.9119381307402327,
                     1.959806167859196}}]}, 
                  {RGBColor[0, 1, 0], 
                   DiskBox[{-1.02690960173304, 1.4732041119061305`}, 0.05]}, 
                  {RGBColor[0, 0, 1], Arrowheads[0.01], 
                   CircleBox[{-1.2568525437186542`, 0.5}, 0.3333333333333333, 
                    NCache[{
                    Rational[-1, 2] 
                    Pi, -5.253653100875315}, {-1.5707963267948966`, \
-5.253653100875315}]], 
                   ArrowBox[{{-1.0879780006885167`, 
                    0.7873891087488761}, {-1.0851126011557608`, 
                    0.7856860221284989}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-1.2568525437186542`, 0.5}, {
                    1.1983853056682918`, 0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-1.2568525437186542`, 0.5}, {-2.902762065868984,
                     0.5}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.005], Arrowheads[0.04], 
                   ArrowBox[{{-1.2568525437186542`, 0.5}, {-0.447524216482038,
                     0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-1.2568525437186542`, 0.5}, 1.6, {
                    1.3387772733696512`, 0.2817601083587644}], 
                   ArrowBox[{{0.27552999157809865`, 0.960221430962853}, {
                    0.2800555106962648, 0.9448748501259079}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Dashing[{Small, Small}], 
                   Arrowheads[0.02], 
                   CircleBox[{-1.2568525437186542`, 0.5}, 1.7, {
                    1.3387772733696512`, 1.7949704347258633`}], 
                   ArrowBox[{{-1.6181713572961716`, 
                    2.1611588469965586`}, {-1.6347646031175838`, 
                    2.1574626618300816`}}]}, 
                  {RGBColor[0, 0, 1], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-1.2568525437186542`, 0.5}, 1.8, {
                    1.3387772733696512`, 2.0207877870882474`}], 
                   ArrowBox[{{-2.023529673444668, 
                    2.1285595410530993`}, {-2.039776663892927, 
                    2.1208114702362337`}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.003], Arrowheads[0.025], 
                   CircleBox[{-1.2568525437186542`, 0.5}, 1.5, {
                    1.3387772733696512`, 1.5467238300240609`}], 
                   ArrowBox[{{-1.2057536869772107`, 
                    1.9991293829552261`}, {-1.220747285874661, 
                    1.9995654071616946`}}]}, 
                  {RGBColor[0, 0, 1], ArrowBox[{{0, 2.8}, {0, 2.8}}]}},
                 Frame->True,
                 FrameTicks->Automatic,
                 ImageSize->800,
                 PlotRange->{{-3, 3}, {-1, 3}}], 349->
                GraphicsBox[{InsetBox[
                   TagBox[GridBox[{
                    {"\<\"t\"\>", "17.400000000000002`"},
                    {"\<\"driveIntent\"\>", "0"},
                    {"\<\"motorTorque\"\>", 
                    RowBox[{"-", "131.8466913893104`"}]},
                    {"\<\"vw\"\>", 
                    RowBox[{"-", "0.3145956399238827`"}]},
                    {"\<\"controllerState\"\>", 
                    RowBox[{"\[LeftAssociation]", 
                    RowBox[{
                    RowBox[{"\<\"lastError\"\>", "\[Rule]", 
                    RowBox[{"-", "0.158276081613566`"}]}], ",", 
                    
                    RowBox[{"\<\"lastInt\"\>", "\[Rule]", 
                    "0.12475456488862005`"}], ",", 
                    RowBox[{"\<\"initialized\"\>", "\[Rule]", "True"}]}], 
                    "\[RightAssociation]"}]}
                    },
                    AutoDelete->False,
                    
                    GridBoxItemSize->{
                    "Columns" -> {{Automatic}}, "Rows" -> {{Automatic}}}],
                    "Grid"], {-1, -0.5}], 
                  {GrayLevel[0], LineBox[{{-5, 0}, {5, 0}}]}, 
                  {GrayLevel[0.5], 
                   CircleBox[{-1.2729358488123519`, 0.5}, 0.5], 
                   DiskBox[{-1.2729358488123519`, 0.5}, 0.1]}, 
                  {GrayLevel[0.5], 
                   DiskBox[{-0.8973076160871737, 1.95220640088789}, 0.1]}, 
                  {GrayLevel[0.5], Thickness[0.01], 
                   LineBox[{{-1.2729358488123519`, 0.5}, {-0.8973076160871737,
                     1.95220640088789}}]}, 
                  {RGBColor[0, 1, 0], 
                   DiskBox[{-1.0225170269955663`, 1.4681376005919269`}, 
                    0.05]}, 
                  {RGBColor[0, 0, 1], Arrowheads[0.01], 
                   CircleBox[{-1.2729358488123519`, 0.5}, 0.3333333333333333, 
                    NCache[{
                    Rational[-1, 2] 
                    Pi, -5.526197068474208}, {-1.5707963267948966`, \
-5.526197068474208}]], 
                   ArrowBox[{{-1.0329345142680233`, 
                    0.731323303037226}, {-1.030633319758069, 
                    0.7289117636230378}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-1.2729358488123519`, 0.5}, {1.363997978973856, 
                    0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-1.2729358488123519`, 
                    0.5}, {-3.0560781158315553`, 0.5}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.005], Arrowheads[0.04], 
                   ArrowBox[{{-1.2729358488123519`, 
                    0.5}, {-0.4191442880453472, 0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-1.2729358488123519`, 0.5}, 1.6, {
                    1.3176834901689851`, 0.16654142502699476`}], 
                   ArrowBox[{{0.30219534241964463`, 0.7810012996554849}, {
                    0.30492655267962476`, 0.7652362003158182}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Dashing[{Small, Small}], 
                   Arrowheads[0.02], 
                   CircleBox[{-1.2729358488123519`, 0.5}, 1.7, {
                    1.3176834901689851`, 1.7963767802827277`}], 
                   ArrowBox[{{-1.636590467673979, 
                    2.1606490653297596`}, {-1.6531784989743956`, 
                    2.1569295479886117`}}]}, 
                  {RGBColor[0, 0, 1], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-1.2729358488123519`, 0.5}, 1.8, {
                    1.3176834901689851`, 2.0501651089984874`}], 
                   ArrowBox[{{-2.087118007374006, 
                    2.105337164797441}, {-2.103130402698437, 
                    2.097115212718821}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.003], Arrowheads[0.025], 
                   CircleBox[{-1.2729358488123519`, 0.5}, 1.5, {
                    1.3176834901689851`, 1.5112956254085066`}], 
                   ArrowBox[{{-1.1687687045996744`, 
                    1.9963786974113789`}, {-1.1837374504923934`, 
                    1.9973455331810204`}}]}, 
                  {RGBColor[0, 0, 1], ArrowBox[{{0, 2.8}, {0, 2.8}}]}},
                 Frame->True,
                 FrameTicks->Automatic,
                 ImageSize->800,
                 PlotRange->{{-3, 3}, {-1, 3}}], 350->
                GraphicsBox[{InsetBox[
                   TagBox[GridBox[{
                    {"\<\"t\"\>", "17.45`"},
                    {"\<\"driveIntent\"\>", "0"},
                    {"\<\"motorTorque\"\>", 
                    RowBox[{"-", "138.34924371101047`"}]},
                    {"\<\"vw\"\>", 
                    RowBox[{"-", "0.29958530093402114`"}]},
                    {"\<\"controllerState\"\>", 
                    RowBox[{"\[LeftAssociation]", 
                    RowBox[{
                    RowBox[{"\<\"lastError\"\>", "\[Rule]", 
                    RowBox[{"-", "0.17881020177546714`"}]}], ",", 
                    
                    RowBox[{"\<\"lastInt\"\>", "\[Rule]", 
                    "0.11629723753183356`"}], ",", 
                    RowBox[{"\<\"initialized\"\>", "\[Rule]", "True"}]}], 
                    "\[RightAssociation]"}]}
                    },
                    AutoDelete->False,
                    
                    GridBoxItemSize->{
                    "Columns" -> {{Automatic}}, "Rows" -> {{Automatic}}}],
                    "Grid"], {-1, -0.5}], 
                  {GrayLevel[0], LineBox[{{-5, 0}, {5, 0}}]}, 
                  {GrayLevel[0.5], 
                   CircleBox[{-1.2882853112145347`, 0.5}, 0.5], 
                   DiskBox[{-1.2882853112145347`, 0.5}, 0.1]}, 
                  {GrayLevel[0.5], 
                   DiskBox[{-0.8895640684961044, 1.946036434743285}, 0.1]}, 
                  {GrayLevel[0.5], Thickness[0.01], 
                   LineBox[{{-1.2882853112145347`, 0.5}, {-0.8895640684961044,
                     1.946036434743285}}]}, 
                  {RGBColor[0, 1, 0], 
                   DiskBox[{-1.0224711494022478`, 1.4640242898288567`}, 
                    0.05]}, 
                  {RGBColor[0, 0, 1], Arrowheads[0.01], 
                   CircleBox[{-1.2882853112145347`, 0.5}, 0.3333333333333333, 
                    NCache[{
                    Rational[-1, 2] 
                    Pi, -5.7212736381252105`}, {-1.5707963267948966`, \
-5.7212736381252105`}]], 
                   ArrowBox[{{-1.007996031365131, 
                    0.6804134992527271}, {-1.0062059407885757`, 
                    0.677601632569089}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-1.2882853112145347`, 0.5}, {
                    1.4786995630056747`, 0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-1.2882853112145347`, 0.5}, {-3.173010482587454,
                     0.5}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.005], Arrowheads[0.04], 
                   ArrowBox[{{-1.2882853112145347`, 
                    0.5}, {-0.40602560836724455`, 0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-1.2882853112145347`, 0.5}, 1.6, {
                    1.3017479577709352`, 0.07983555936321807}], 
                   ArrowBox[{{0.3052626875569051, 0.6435436366110999}, {
                    0.30661842326323385`, 0.6276012450913698}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Dashing[{Small, Small}], 
                   Arrowheads[0.02], 
                   CircleBox[{-1.2882853112145347`, 0.5}, 1.7, {
                    1.3017479577709352`, 1.7942527452401924`}], 
                   ArrowBox[{{-1.6484118356021615`, 
                    2.1614177338744422`}, {-1.6650077298631678`, 
                    2.157733458456915}}]}, 
                  {RGBColor[0, 0, 1], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-1.2882853112145347`, 0.5}, 1.8, {
                    1.3017479577709352`, 2.070354867276549}], 
                   ArrowBox[{{-2.1347107017688107`, 
                    2.088572962827028}, {-2.1505538457193922`, 
                    2.08002942200544}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.003], Arrowheads[0.025], 
                   CircleBox[{-1.2882853112145347`, 0.5}, 1.5, {
                    1.3017479577709352`, 1.4788751166745984`}], 
                   ArrowBox[{{-1.1356680462675377`, 
                    1.9922157921829196`}, {-1.1505975862876354`, 
                    1.9936673292284512`}}]}, 
                  {RGBColor[0, 0, 1], ArrowBox[{{0, 2.8}, {0, 2.8}}]}},
                 Frame->True,
                 FrameTicks->Automatic,
                 ImageSize->800,
                 PlotRange->{{-3, 3}, {-1, 3}}], 351->
                GraphicsBox[{InsetBox[
                   TagBox[GridBox[{
                    {"\<\"t\"\>", "17.5`"},
                    {"\<\"driveIntent\"\>", "0"},
                    {"\<\"motorTorque\"\>", 
                    RowBox[{"-", "142.569986092251`"}]},
                    {"\<\"vw\"\>", 
                    RowBox[{"-", "0.2841244397096202`"}]},
                    {"\<\"controllerState\"\>", 
                    RowBox[{"\[LeftAssociation]", 
                    RowBox[{
                    RowBox[{"\<\"lastError\"\>", "\[Rule]", 
                    RowBox[{"-", "0.19475128890870824`"}]}], ",", 
                    
                    RowBox[{"\<\"lastInt\"\>", "\[Rule]", 
                    "0.10693189496063699`"}], ",", 
                    RowBox[{"\<\"initialized\"\>", "\[Rule]", "True"}]}], 
                    "\[RightAssociation]"}]}
                    },
                    AutoDelete->False,
                    
                    GridBoxItemSize->{
                    "Columns" -> {{Automatic}}, "Rows" -> {{Automatic}}}],
                    "Grid"], {-1, -0.5}], 
                  {GrayLevel[0], LineBox[{{-5, 0}, {5, 0}}]}, 
                  {GrayLevel[0.5], CircleBox[{-1.302871644853798, 0.5}, 0.5], 
                   DiskBox[{-1.302871644853798, 0.5}, 0.1]}, 
                  {GrayLevel[0.5], 
                   DiskBox[{-0.8879586972313298, 1.9414739837732884`}, 0.1]}, 
                  {GrayLevel[0.5], Thickness[0.01], 
                   LineBox[{{-1.302871644853798, 0.5}, {-0.8879586972313298, 
                    1.9414739837732884`}}]}, 
                  {RGBColor[0, 1, 0], 
                   DiskBox[{-1.026263013105486, 1.460982655848859}, 0.05]}, 
                  {RGBColor[0, 0, 1], Arrowheads[0.01], 
                   CircleBox[{-1.302871644853798, 0.5}, 0.3333333333333333, 
                    NCache[{
                    Rational[-1, 2] 
                    Pi, -5.847895909562427}, {-1.5707963267948966`, \
-5.847895909562427}]], 
                   ArrowBox[{{-1.0020429601485445`, 
                    0.6435730252158052}, {-1.0006222951339947`, 
                    0.6405576099151773}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-1.302871644853798, 0.5}, {1.5485280769912224`, 
                    0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-1.302871644853798, 0.5}, {-3.2579457197667776`,
                     0.5}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.005], Arrowheads[0.04], 
                   ArrowBox[{{-1.302871644853798, 0.5}, {-0.4065459979217573, 
                    0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-1.302871644853798, 0.5}, 1.6, {
                    1.2905330820824155`, 0.01899993901932895}], 
                   ArrowBox[{{0.2964556051260505, 0.5463933990120944}, {
                    0.29683956568786196`, 0.5303980734148847}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Dashing[{Small, Small}], 
                   Arrowheads[0.02], 
                   CircleBox[{-1.302871644853798, 0.5}, 1.7, {
                    1.2905330820824155`, 1.7892749907618022`}], 
                   ArrowBox[{{-1.654723612223608, 
                    2.1631897645963294`}, {-1.6713376402208993`, 
                    2.1595881447690966`}}]}, 
                  {RGBColor[0, 0, 1], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-1.302871644853798, 0.5}, 1.8, {
                    1.2905330820824155`, 2.082588560372699}], 
                   ArrowBox[{{-2.168667326141578, 
                    2.0780994386487275`}, {-2.1844047680894896`, 
                    2.0693627218200152`}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.003], Arrowheads[0.025], 
                   CircleBox[{-1.302871644853798, 0.5}, 1.5, {
                    1.2905330820824155`, 1.4499384666001014`}], 
                   ArrowBox[{{-1.1071445704124776`, 
                    1.987175481350685}, {-1.122025863636813, 
                    1.9890583613196693`}}]}, 
                  {RGBColor[0, 0, 1], ArrowBox[{{0, 2.8}, {0, 2.8}}]}},
                 Frame->True,
                 FrameTicks->Automatic,
                 ImageSize->800,
                 PlotRange->{{-3, 3}, {-1, 3}}], 352->
                GraphicsBox[{InsetBox[
                   TagBox[GridBox[{
                    {"\<\"t\"\>", "17.55`"},
                    {"\<\"driveIntent\"\>", "0"},
                    {"\<\"motorTorque\"\>", 
                    RowBox[{"-", "144.80117378055087`"}]},
                    {"\<\"vw\"\>", 
                    RowBox[{"-", "0.2684707198186829`"}]},
                    {"\<\"controllerState\"\>", 
                    RowBox[{"\[LeftAssociation]", 
                    RowBox[{
                    RowBox[{"\<\"lastError\"\>", "\[Rule]", 
                    RowBox[{"-", "0.20649230949081449`"}]}], ",", 
                    
                    RowBox[{"\<\"lastInt\"\>", "\[Rule]", 
                    "0.09687829456055887`"}], ",", 
                    RowBox[{"\<\"initialized\"\>", "\[Rule]", "True"}]}], 
                    "\[RightAssociation]"}]}
                    },
                    AutoDelete->False,
                    
                    GridBoxItemSize->{
                    "Columns" -> {{Automatic}}, "Rows" -> {{Automatic}}}],
                    "Grid"], {-1, -0.5}], 
                  {GrayLevel[0], LineBox[{{-5, 0}, {5, 0}}]}, 
                  {GrayLevel[0.5], CircleBox[{-1.316678999042934, 0.5}, 0.5], 
                   DiskBox[{-1.316678999042934, 0.5}, 0.1]}, 
                  {GrayLevel[0.5], 
                   DiskBox[{-0.8917362028903547, 1.938549137151049}, 0.1]}, 
                  {GrayLevel[0.5], Thickness[0.01], 
                   LineBox[{{-1.316678999042934, 0.5}, {-0.8917362028903547, 
                    1.938549137151049}}]}, 
                  {RGBColor[0, 1, 0], 
                   DiskBox[{-1.0333838016078811`, 1.4590327581006992`}, 
                    0.05]}, 
                  {RGBColor[0, 0, 1], Arrowheads[0.01], 
                   CircleBox[{-1.316678999042934, 0.5}, 0.3333333333333333, 
                    NCache[{
                    Rational[-1, 2] 
                    Pi, -5.914831540211423}, {-1.5707963267948966`, \
-5.914831540211423}]], 
                   ArrowBox[{{-1.0069209995593642`, 
                    0.6231303896974583}, {-1.0057052039549275`, 
                    0.620026704860517}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-1.316678999042934, 0.5}, {1.5793444765680833`, 
                    0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-1.316678999042934, 0.5}, {-3.31495095276175, 
                    0.5}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.005], Arrowheads[0.04], 
                   ArrowBox[{{-1.316678999042934, 0.5}, {-0.4189274771507323, 
                    0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-1.316678999042934, 0.5}, 1.6, {
                    1.2835679935666384`, -0.018702403898889886`}], 
                   ArrowBox[{{0.2832604158725236, 0.48607632950751184`}, {
                    0.2830411851840924, 0.47007789819131623`}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Dashing[{Small, Small}], 
                   Arrowheads[0.02], 
                   CircleBox[{-1.316678999042934, 0.5}, 1.7, {
                    1.2835679935666384`, 1.78206638122502}], 
                   ArrowBox[{{-1.6565326428619542`, 
                    2.165682893225158}, {-1.6731722016411927`, 
                    2.16220112997833}}]}, 
                  {RGBColor[0, 0, 1], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-1.316678999042934, 0.5}, 1.8, {
                    1.2835679935666384`, 2.088018959014143}], 
                   ArrowBox[{{-2.191031581467098, 
                    2.073374577654093}, {-2.206721347751045, 
                    2.064552529481241}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.003], Arrowheads[0.025], 
                   CircleBox[{-1.316678999042934, 0.5}, 1.5, {
                    1.2835679935666384`, 1.4247855674865795`}], 
                   ArrowBox[{{-1.083611005874101, 
                    1.981782477477802}, {-1.0984402369889148`, 
                    1.9840390300585526`}}]}, 
                  {RGBColor[0, 0, 1], ArrowBox[{{0, 2.8}, {0, 2.8}}]}},
                 Frame->True,
                 FrameTicks->Automatic,
                 ImageSize->800,
                 PlotRange->{{-3, 3}, {-1, 3}}], 353->
                GraphicsBox[{InsetBox[
                   TagBox[GridBox[{
                    {"\<\"t\"\>", "17.6`"},
                    {"\<\"driveIntent\"\>", "0"},
                    {"\<\"motorTorque\"\>", 
                    RowBox[{"-", "145.31984981305138`"}]},
                    {"\<\"vw\"\>", 
                    RowBox[{"-", "0.25285315435163785`"}]},
                    {"\<\"controllerState\"\>", 
                    RowBox[{"\[LeftAssociation]", 
                    RowBox[{
                    RowBox[{"\<\"lastError\"\>", "\[Rule]", 
                    RowBox[{"-", "0.21444784350951007`"}]}], ",", 
                    
                    RowBox[{"\<\"lastInt\"\>", "\[Rule]", 
                    "0.08633592271596938`"}], ",", 
                    RowBox[{"\<\"initialized\"\>", "\[Rule]", "True"}]}], 
                    "\[RightAssociation]"}]}
                    },
                    AutoDelete->False,
                    
                    GridBoxItemSize->{
                    "Columns" -> {{Automatic}}, "Rows" -> {{Automatic}}}],
                    "Grid"], {-1, -0.5}], 
                  {GrayLevel[0], LineBox[{{-5, 0}, {5, 0}}]}, 
                  {GrayLevel[0.5], 
                   CircleBox[{-1.3297037027601926`, 0.5}, 0.5], 
                   DiskBox[{-1.3297037027601926`, 0.5}, 0.1]}, 
                  {GrayLevel[0.5], 
                   DiskBox[{-0.9001613911985638, 1.9371824527798454`}, 0.1]}, 
                  {GrayLevel[0.5], Thickness[0.01], 
                   LineBox[{{-1.3297037027601926`, 0.5}, {-0.9001613911985638,
                     1.9371824527798454`}}]}, 
                  {RGBColor[0, 1, 0], 
                   DiskBox[{-1.0433421617191068`, 1.4581216351865636`}, 
                    0.05]}, 
                  {RGBColor[0, 0, 1], Arrowheads[0.01], 
                   CircleBox[{-1.3297037027601926`, 0.5}, 0.3333333333333333, 
                    NCache[{
                    Rational[-1, 2] 
                    Pi, -5.930391821186438}, {-1.5707963267948966`, \
-5.930391821186438}]], 
                   ArrowBox[{{-1.0180673360407226`, 
                    0.618295756682136}, {-1.01689997987825, 
                    0.6151735302155317}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-1.3297037027601926`, 0.5}, {
                    1.5766932935008349`, 0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-1.3297037027601926`, 
                    0.5}, {-3.3476856505678234`, 0.5}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.005], Arrowheads[0.04], 
                   ArrowBox[{{-1.3297037027601926`, 
                    0.5}, {-0.44128865430679587`, 0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-1.3297037027601926`, 0.5}, 1.6, {
                    1.2803691473995875`, -0.035996824374114667`}], 
                   ArrowBox[{{0.2697556597872761, 0.4584097660592621}, {
                    0.26925979207785167`, 0.44241751850338273`}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Dashing[{Small, Small}], 
                   Arrowheads[0.02], 
                   CircleBox[{-1.3297037027601926`, 0.5}, 1.7, {
                    1.2803691473995875`, 1.7732037531831248`}], 
                   ArrowBox[{{-1.6547818648994541`, 
                    2.1686294341465273`}, {-1.671451627364746, 
                    2.16529527592811}}]}, 
                  {RGBColor[0, 0, 1], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-1.3297037027601926`, 0.5}, 1.8, {
                    1.2803691473995875`, 2.087701646360984}], 
                   ArrowBox[{{-2.2035569895129776`, 
                    2.0736519415777286`}, {-2.219249554354509, 
                    2.0648348724102945`}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.003], Arrowheads[0.025], 
                   CircleBox[{-1.3297037027601926`, 0.5}, 1.5, {
                    1.2803691473995875`, 1.4035903294826568`}], 
                   ArrowBox[{{-1.0652836783282384`, 
                    1.976510091628027}, {-1.0800617540517807`, 
                    1.979080422913192}}]}, 
                  {RGBColor[0, 0, 1], ArrowBox[{{0, 2.8}, {0, 2.8}}]}},
                 Frame->True,
                 FrameTicks->Automatic,
                 ImageSize->800,
                 PlotRange->{{-3, 3}, {-1, 3}}], 354->
                GraphicsBox[{InsetBox[
                   TagBox[GridBox[{
                    {"\<\"t\"\>", "17.650000000000002`"},
                    {"\<\"driveIntent\"\>", "0"},
                    {"\<\"motorTorque\"\>", 
                    RowBox[{"-", "144.38139857834992`"}]},
                    {"\<\"vw\"\>", 
                    RowBox[{"-", "0.23746717261307315`"}]},
                    {"\<\"controllerState\"\>", 
                    RowBox[{"\[LeftAssociation]", 
                    RowBox[{
                    RowBox[{"\<\"lastError\"\>", "\[Rule]", 
                    RowBox[{"-", "0.21903803891024892`"}]}], ",", 
                    
                    RowBox[{"\<\"lastInt\"\>", "\[Rule]", 
                    "0.07548333295259735`"}], ",", 
                    RowBox[{"\<\"initialized\"\>", "\[Rule]", "True"}]}], 
                    "\[RightAssociation]"}]}
                    },
                    AutoDelete->False,
                    
                    GridBoxItemSize->{
                    "Columns" -> {{Automatic}}, "Rows" -> {{Automatic}}}],
                    "Grid"], {-1, -0.5}], 
                  {GrayLevel[0], LineBox[{{-5, 0}, {5, 0}}]}, 
                  {GrayLevel[0.5], CircleBox[{-1.341952692981371, 0.5}, 0.5], 
                   DiskBox[{-1.341952692981371, 0.5}, 0.1]}, 
                  {GrayLevel[0.5], 
                   DiskBox[{-0.9125377964168653, 1.937220528175302}, 0.1]}, 
                  {GrayLevel[0.5], Thickness[0.01], 
                   LineBox[{{-1.341952692981371, 0.5}, {-0.9125377964168653, 
                    1.937220528175302}}]}, 
                  {RGBColor[0, 1, 0], 
                   DiskBox[{-1.0556760952717004`, 1.4581470187835348`}, 
                    0.05]}, 
                  {RGBColor[0, 0, 1], Arrowheads[0.01], 
                   CircleBox[{-1.341952692981371, 0.5}, 0.3333333333333333, 
                    NCache[{
                    Rational[-1, 2] 
                    Pi, -5.902238284145394}, {-1.5707963267948966`, \
-5.902238284145394}]], 
                   ArrowBox[{{-1.0337698269909885`, 
                    0.6270213848966586}, {-1.0325150433270374`, 
                    0.6239332565839896}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-1.341952692981371, 0.5}, {1.5456752785856276`, 
                    0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-1.341952692981371, 0.5}, {-3.359389494162382, 
                    0.5}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.005], Arrowheads[0.04], 
                   ArrowBox[{{-1.341952692981371, 0.5}, {-0.4717615225953835, 
                    0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-1.341952692981371, 0.5}, 1.6, {
                    1.2804578023292432`, -0.03551769623147871}], 
                   ArrowBox[{{0.25752641302791246`, 0.4591761167973457}, {
                    0.2570382077109794, 0.4431836334929212}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Dashing[{Small, Small}], 
                   Arrowheads[0.02], 
                   CircleBox[{-1.341952692981371, 0.5}, 1.7, {
                    1.2804578023292432`, 1.763196075668148}], 
                   ArrowBox[{{-1.6503157498564966`, 
                    2.1717990983233086`}, {-1.6670180441835805`, 
                    2.168631929889808}}]}, 
                  {RGBColor[0, 0, 1], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-1.341952692981371, 0.5}, 1.8, {
                    1.2804578023292432`, 2.082576683320076}], 
                   ArrowBox[{{-2.207729631038008, 
                    2.0781097216382878`}, {-2.2234671767512553`, 
                    2.069373191724618}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.003], Arrowheads[0.025], 
                   CircleBox[{-1.341952692981371, 0.5}, 1.5, {
                    1.2804578023292432`, 1.3863966375042311`}], 
                   ArrowBox[{{-1.052186342989438, 
                    1.9717457193456867`}, {-1.066918043089932, 
                    1.9745697478787139`}}]}, 
                  {RGBColor[0, 0, 1], ArrowBox[{{0, 2.8}, {0, 2.8}}]}},
                 Frame->True,
                 FrameTicks->Automatic,
                 ImageSize->800,
                 PlotRange->{{-3, 3}, {-1, 3}}], 355->
                GraphicsBox[{InsetBox[
                   TagBox[GridBox[{
                    {"\<\"t\"\>", "17.7`"},
                    {"\<\"driveIntent\"\>", "0"},
                    {"\<\"motorTorque\"\>", 
                    RowBox[{"-", "142.21584147765233`"}]},
                    {"\<\"vw\"\>", 
                    RowBox[{"-", "0.2224729259387889`"}]},
                    {"\<\"controllerState\"\>", 
                    RowBox[{"\[LeftAssociation]", 
                    RowBox[{
                    RowBox[{"\<\"lastError\"\>", "\[Rule]", 
                    RowBox[{"-", "0.22067515669752233`"}]}], ",", 
                    
                    RowBox[{"\<\"lastInt\"\>", "\[Rule]", 
                    "0.06447822261251525`"}], ",", 
                    RowBox[{"\<\"initialized\"\>", "\[Rule]", "True"}]}], 
                    "\[RightAssociation]"}]}
                    },
                    AutoDelete->False,
                    
                    GridBoxItemSize->{
                    "Columns" -> {{Automatic}}, "Rows" -> {{Automatic}}}],
                    "Grid"], {-1, -0.5}], 
                  {GrayLevel[0], LineBox[{{-5, 0}, {5, 0}}]}, 
                  {GrayLevel[0.5], CircleBox[{-1.353441779245186, 0.5}, 0.5], 
                   DiskBox[{-1.353441779245186, 0.5}, 0.1]}, 
                  {GrayLevel[0.5], 
                   DiskBox[{-0.9282198158337189, 1.9384666425859507`}, 0.1]}, 
                  {GrayLevel[0.5], Thickness[0.01], 
                   LineBox[{{-1.353441779245186, 0.5}, {-0.9282198158337189, 
                    1.9384666425859507`}}]}, 
                  {RGBColor[0, 1, 0], 
                   DiskBox[{-1.069960470304208, 1.458977761723967}, 0.05]}, 
                  {RGBColor[0, 0, 1], Arrowheads[0.01], 
                   CircleBox[{-1.353441779245186, 0.5}, 0.3333333333333333, 
                    NCache[{
                    Rational[-1, 2] 
                    Pi, -5.837271571124466}, {-1.5707963267948966`, \
-5.837271571124466}]], 
                   ArrowBox[{{-1.0541554123490833`, 
                    0.6467609679078277}, {-1.0527027913236866`, 
                    0.6437608161324332}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-1.353441779245186, 0.5}, {1.4908750503078607`, 
                    0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-1.353441779245186, 0.5}, {-3.352911835317668, 
                    0.5}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.005], Arrowheads[0.04], 
                   ArrowBox[{{-1.353441779245186, 0.5}, {-0.5085950057646214, 
                    0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-1.353441779245186, 0.5}, 1.6, {
                    1.283373926325174, -0.01975200101515906}], 
                   ArrowBox[{{0.2464821401387256, 0.48439704568927794`}, {
                    0.2462461176667614, 0.46839885328930647`}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Dashing[{Small, Small}], 
                   Arrowheads[0.02], 
                   CircleBox[{-1.353441779245186, 0.5}, 1.7, {
                    1.283373926325174, 1.7524681419879673`}], 
                   ArrowBox[{{-1.6438524859227965`, 
                    2.1750109317395543`}, {-1.6605877956587696`, 
                    2.172023123225653}}]}, 
                  {RGBColor[0, 0, 1], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-1.353441779245186, 0.5}, 1.8, {
                    1.283373926325174, 2.0734619345343535`}], 
                   ArrowBox[{{-2.204798879261539, 
                    2.0859353985114732`}, {-2.220615401425139, 
                    2.077342673294329}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.003], Arrowheads[0.025], 
                   CircleBox[{-1.353441779245186, 0.5}, 1.5, {
                    1.283373926325174, 1.373118658097667}], 
                   ArrowBox[{{-1.0441597373287423`, 
                    1.9677685848075626`}, {-1.0588526425231723`, 
                    1.9707879658623058`}}]}, 
                  {RGBColor[0, 0, 1], ArrowBox[{{0, 2.8}, {0, 2.8}}]}},
                 Frame->True,
                 FrameTicks->Automatic,
                 ImageSize->800,
                 PlotRange->{{-3, 3}, {-1, 3}}], 356->
                GraphicsBox[{InsetBox[
                   TagBox[GridBox[{
                    {"\<\"t\"\>", "17.75`"},
                    {"\<\"driveIntent\"\>", "0"},
                    {"\<\"motorTorque\"\>", 
                    RowBox[{"-", "139.02686139201728`"}]},
                    {"\<\"vw\"\>", 
                    RowBox[{"-", "0.20799636996072352`"}]},
                    {"\<\"controllerState\"\>", 
                    RowBox[{"\[LeftAssociation]", 
                    RowBox[{
                    RowBox[{"\<\"lastError\"\>", "\[Rule]", 
                    RowBox[{"-", "0.21975319096472262`"}]}], ",", 
                    
                    RowBox[{"\<\"lastInt\"\>", "\[Rule]", 
                    "0.05345810530689508`"}], ",", 
                    RowBox[{"\<\"initialized\"\>", "\[Rule]", "True"}]}], 
                    "\[RightAssociation]"}]}
                    },
                    AutoDelete->False,
                    
                    GridBoxItemSize->{
                    "Columns" -> {{Automatic}}, "Rows" -> {{Automatic}}}],
                    "Grid"], {-1, -0.5}], 
                  {GrayLevel[0], LineBox[{{-5, 0}, {5, 0}}]}, 
                  {GrayLevel[0.5], 
                   CircleBox[{-1.3641938974692915`, 0.5}, 0.5], 
                   DiskBox[{-1.3641938974692915`, 0.5}, 0.1]}, 
                  {GrayLevel[0.5], 
                   DiskBox[{-0.9466194086012577, 1.9407052253138393`}, 0.1]}, 
                  {GrayLevel[0.5], Thickness[0.01], 
                   LineBox[{{-1.3641938974692915`, 0.5}, {-0.9466194086012577,
                     1.9407052253138393`}}]}, 
                  {RGBColor[0, 1, 0], 
                   DiskBox[{-1.0858109048906024`, 1.4604701502092263`}, 
                    0.05]}, 
                  {RGBColor[0, 0, 1], Arrowheads[0.01], 
                   CircleBox[{-1.3641938974692915`, 0.5}, 0.3333333333333333, 
                    NCache[{
                    Rational[-1, 2] 
                    Pi, -5.741602168555415}, {-1.5707963267948966`, \
-5.741602168555415}]], 
                   ArrowBox[{{-1.0802952415653224`, 
                    0.6746787459510482}, {-1.078562678033295, 
                    0.6718310728436996}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-1.3641938974692915`, 0.5}, {1.416343330371054, 
                    0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-1.3641938974692915`, 
                    0.5}, {-3.3307598250392907`, 0.5}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.005], Arrowheads[0.04], 
                   ArrowBox[{{-1.3641938974692915`, 
                    0.5}, {-0.5502225971989451, 0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-1.3641938974692915`, 0.5}, 1.6, {
                    1.288686187565498, 0.008996543422465475}], 
                   ArrowBox[{{0.23551741628274891`, 0.5303926414403405}, {
                    0.23574135273259378`, 0.5143942753006308}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Dashing[{Small, Small}], 
                   Arrowheads[0.02], 
                   CircleBox[{-1.3641938974692915`, 0.5}, 1.7, {
                    1.288686187565498, 1.7413566122406285`}], 
                   ArrowBox[{{-1.6359751255866055`, 
                    2.1781343700797753`}, {-1.652742600651576, 
                    2.175332697075963}}]}, 
                  {RGBColor[0, 0, 1], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-1.3641938974692915`, 0.5}, 1.8, {
                    1.288686187565498, 2.061057639743372}], 
                   ArrowBox[{{-2.195813594925504, 
                    2.0963736025137845`}, {-2.2117354842513395`, 
                    2.08797772612684}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.003], Arrowheads[0.025], 
                   CircleBox[{-1.3641938974692915`, 0.5}, 1.5, {
                    1.288686187565498, 1.3635561243915264`}], 
                   ArrowBox[{{-1.0408906234783226`, 
                    1.9647440025570067`}, {-1.0555539844101032`, 
                    1.9679037448234868`}}]}, 
                  {RGBColor[0, 0, 1], ArrowBox[{{0, 2.8}, {0, 2.8}}]}},
                 Frame->True,
                 FrameTicks->Automatic,
                 ImageSize->800,
                 PlotRange->{{-3, 3}, {-1, 3}}], 357->
                GraphicsBox[{InsetBox[
                   TagBox[GridBox[{
                    {"\<\"t\"\>", "17.8`"},
                    {"\<\"driveIntent\"\>", "0"},
                    {"\<\"motorTorque\"\>", 
                    RowBox[{"-", "134.99285258004912`"}]},
                    {"\<\"vw\"\>", 
                    RowBox[{"-", "0.19413226017703317`"}]},
                    {"\<\"controllerState\"\>", 
                    RowBox[{"\[LeftAssociation]", 
                    RowBox[{
                    RowBox[{"\<\"lastError\"\>", "\[Rule]", 
                    RowBox[{"-", "0.21664070028319427`"}]}], ",", 
                    
                    RowBox[{"\<\"lastInt\"\>", "\[Rule]", 
                    "0.0425414201721645`"}], ",", 
                    RowBox[{"\<\"initialized\"\>", "\[Rule]", "True"}]}], 
                    "\[RightAssociation]"}]}
                    },
                    AutoDelete->False,
                    
                    GridBoxItemSize->{
                    "Columns" -> {{Automatic}}, "Rows" -> {{Automatic}}}],
                    "Grid"], {-1, -0.5}], 
                  {GrayLevel[0], LineBox[{{-5, 0}, {5, 0}}]}, 
                  {GrayLevel[0.5], 
                   CircleBox[{-1.3742374704907545`, 0.5}, 0.5], 
                   DiskBox[{-1.3742374704907545`, 0.5}, 0.1]}, 
                  {GrayLevel[0.5], 
                   DiskBox[{-0.9672084496807776, 1.9437199784648238`}, 0.1]}, 
                  {GrayLevel[0.5], Thickness[0.01], 
                   LineBox[{{-1.3742374704907545`, 0.5}, {-0.9672084496807776,
                     1.9437199784648238`}}]}, 
                  {RGBColor[0, 1, 0], 
                   DiskBox[{-1.1028847899507699`, 1.462479985643216}, 0.05]}, 
                  {RGBColor[0, 0, 1], Arrowheads[0.01], 
                   CircleBox[{-1.3742374704907545`, 0.5}, 0.3333333333333333, 
                    NCache[{
                    Rational[-1, 2] 
                    Pi, -5.62058190419637}, {-1.5707963267948966`, \
-5.62058190419637}]], 
                   ArrowBox[{{-1.113503358930569, 
                    0.7076748279884487}, {-1.1114396818599217`, 
                    0.7050571466734463}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-1.3742374704907545`, 0.5}, {
                    1.3256195811102278`, 0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-1.3742374704907545`, 
                    0.5}, {-3.2951507653842937`, 0.5}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.005], Arrowheads[0.04], 
                   ArrowBox[{{-1.3742374704907545`, 
                    0.5}, {-0.5952937137833112, 0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-1.3742374704907545`, 0.5}, 1.6, {
                    1.295998161583599, 0.04862592816917388}], 
                   ArrowBox[{{0.22301371738708498`, 0.593747761695046}, {
                    0.22387131748561173`, 0.5777708286744445}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Dashing[{Small, Small}], 
                   Arrowheads[0.02], 
                   CircleBox[{-1.3742374704907545`, 0.5}, 1.7, {
                    1.295998161583599, 1.7301162198211368`}], 
                   ArrowBox[{{-1.6271390378509152`, 
                    2.18108322138637}, {-1.6439369449126502`, 
                    2.1784701944021982`}}]}, 
                  {RGBColor[0, 0, 1], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-1.3742374704907545`, 0.5}, 1.8, {
                    1.295998161583599, 2.045958453694983}], 
                   ArrowBox[{{-2.181659345018979, 
                    2.1087479344301667`}, {-2.1977061854826636`, 
                    2.100593413528112}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.003], Arrowheads[0.025], 
                   CircleBox[{-1.3742374704907545`, 0.5}, 1.5, {
                    1.295998161583599, 1.3574205958110825`}], 
                   ArrowBox[{{-1.0419533595043875`, 
                    1.962732808679015}, {-1.0565970578706931`, 
                    1.9659824583775063`}}]}, 
                  {RGBColor[0, 0, 1], ArrowBox[{{0, 2.8}, {0, 2.8}}]}},
                 Frame->True,
                 FrameTicks->Automatic,
                 ImageSize->800,
                 PlotRange->{{-3, 3}, {-1, 3}}], 358->
                GraphicsBox[{InsetBox[
                   TagBox[GridBox[{
                    {"\<\"t\"\>", "17.85`"},
                    {"\<\"driveIntent\"\>", "0"},
                    {"\<\"motorTorque\"\>", 
                    RowBox[{"-", "130.26918878842812`"}]},
                    {"\<\"vw\"\>", 
                    RowBox[{"-", "0.18094818626344644`"}]},
                    {"\<\"controllerState\"\>", 
                    RowBox[{"\[LeftAssociation]", 
                    RowBox[{
                    RowBox[{"\<\"lastError\"\>", "\[Rule]", 
                    RowBox[{"-", "0.21167660149525067`"}]}], ",", 
                    
                    RowBox[{"\<\"lastInt\"\>", "\[Rule]", 
                    "0.03182892232895868`"}], ",", 
                    RowBox[{"\<\"initialized\"\>", "\[Rule]", "True"}]}], 
                    "\[RightAssociation]"}]}
                    },
                    AutoDelete->False,
                    
                    GridBoxItemSize->{
                    "Columns" -> {{Automatic}}, "Rows" -> {{Automatic}}}],
                    "Grid"], {-1, -0.5}], 
                  {GrayLevel[0], LineBox[{{-5, 0}, {5, 0}}]}, 
                  {GrayLevel[0.5], CircleBox[{-1.383604948239916, 0.5}, 0.5], 
                   DiskBox[{-1.383604948239916, 0.5}, 0.1]}, 
                  {GrayLevel[0.5], 
                   DiskBox[{-0.9895178530669444, 1.947306243135201}, 0.1]}, 
                  {GrayLevel[0.5], Thickness[0.01], 
                   LineBox[{{-1.383604948239916, 0.5}, {-0.9895178530669444, 
                    1.947306243135201}}]}, 
                  {RGBColor[0, 1, 0], 
                   DiskBox[{-1.1208802181246016`, 1.4648708287568006`}, 
                    0.05]}, 
                  {RGBColor[0, 0, 1], Arrowheads[0.01], 
                   CircleBox[{-1.383604948239916, 0.5}, 0.3333333333333333, 
                    NCache[{
                    Rational[-1, 2] 
                    Pi, -5.4788719904477405`}, {-1.5707963267948966`, \
-5.4788719904477405`}]], 
                   ArrowBox[{{-1.1548156320050045`, 
                    0.7424181509043264}, {-1.1524029302692675`, 
                    0.7401181750668014}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-1.383604948239916, 0.5}, {1.2217788275286465`, 
                    0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-1.383604948239916, 0.5}, {-3.2480606530848073`,
                     0.5}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.005], Arrowheads[0.04], 
                   ArrowBox[{{-1.383604948239916, 0.5}, {-0.6426768773162446, 
                    0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-1.383604948239916, 0.5}, 1.6, {
                    1.304951272482476, 0.0972405812705941}], 
                   ArrowBox[{{0.20720343206174263`, 0.6712562324764073}, {
                    0.20883642608776376`, 0.6553398510665278}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Dashing[{Small, Small}], 
                   Arrowheads[0.02], 
                   CircleBox[{-1.383604948239916, 0.5}, 1.7, {
                    1.304951272482476, 1.7189318896915475`}], 
                   ArrowBox[{{-1.617689300390325, 
                    2.183806555420878}, {-1.6345153811914384`, 
                    2.1813815612870506`}}]}, 
                  {RGBColor[0, 0, 1], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-1.383604948239916, 0.5}, 1.8, {
                    1.304951272482476, 2.0286689879737434`}], 
                   ArrowBox[{{-2.163093139682025, 
                    2.1224666897666378`}, {-2.1792785620851416`, 
                    2.114590815107803}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.003], Arrowheads[0.025], 
                   CircleBox[{-1.383604948239916, 0.5}, 1.5, {
                    1.304951272482476, 1.3543658266162706`}], 
                   ArrowBox[{{-1.0468540834512425`, 
                    1.9617109341672452`}, {-1.0614877865785712`, 
                    1.9650053017526043`}}]}, 
                  {RGBColor[0, 0, 1], ArrowBox[{{0, 2.8}, {0, 2.8}}]}},
                 Frame->True,
                 FrameTicks->Automatic,
                 ImageSize->800,
                 PlotRange->{{-3, 3}, {-1, 3}}], 359->
                GraphicsBox[{InsetBox[
                   TagBox[GridBox[{
                    {"\<\"t\"\>", "17.900000000000002`"},
                    {"\<\"driveIntent\"\>", "0"},
                    {"\<\"motorTorque\"\>", 
                    RowBox[{"-", "124.9910327463933`"}]},
                    {"\<\"vw\"\>", 
                    RowBox[{"-", "0.16848894099246484`"}]},
                    {"\<\"controllerState\"\>", 
                    RowBox[{"\[LeftAssociation]", 
                    RowBox[{
                    RowBox[{"\<\"lastError\"\>", "\[Rule]", 
                    RowBox[{"-", "0.20516842343148525`"}]}], ",", 
                    
                    RowBox[{"\<\"lastInt\"\>", "\[Rule]", 
                    "0.021405218511350212`"}], ",", 
                    RowBox[{"\<\"initialized\"\>", "\[Rule]", "True"}]}], 
                    "\[RightAssociation]"}]}
                    },
                    AutoDelete->False,
                    
                    GridBoxItemSize->{
                    "Columns" -> {{Automatic}}, "Rows" -> {{Automatic}}}],
                    "Grid"], {-1, -0.5}], 
                  {GrayLevel[0], LineBox[{{-5, 0}, {5, 0}}]}, 
                  {GrayLevel[0.5], 
                   CircleBox[{-1.3923315606353952`, 0.5}, 0.5], 
                   DiskBox[{-1.3923315606353952`, 0.5}, 0.1]}, 
                  {GrayLevel[0.5], 
                   DiskBox[{-1.0131345005905061`, 1.951278605111132}, 0.1]}, 
                  {GrayLevel[0.5], Thickness[0.01], 
                   LineBox[{{-1.3923315606353952`, 
                    0.5}, {-1.0131345005905061`, 1.951278605111132}}]}, 
                  {RGBColor[0, 1, 0], 
                   DiskBox[{-1.139533520605469, 1.467519070074088}, 0.05]}, 
                  {RGBColor[0, 0, 1], Arrowheads[0.01], 
                   CircleBox[{-1.3923315606353952`, 0.5}, 0.3333333333333333, 
                    NCache[{
                    Rational[-1, 2] 
                    
                    Pi, -5.320527309186695}, {-1.5707963267948966`, \
-5.320527309186695}]], 
                   ArrowBox[{{-1.2046298965469244`, 
                    0.7754617875668601}, {-1.2018847095863199`, 
                    0.7735710292348267}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-1.3923315606353952`, 0.5}, {
                    1.1074890942924707`, 0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-1.3923315606353952`, 0.5}, {-3.191265333778219,
                     0.5}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.005], Arrowheads[0.04], 
                   ArrowBox[{{-1.3923315606353952`, 
                    0.5}, {-0.6914446788503531, 0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-1.3923315606353952`, 0.5}, 1.6, {
                    1.3152251853972663`, 0.15314617011938747`}], 
                   ArrowBox[{{0.18642228903961255`, 0.7598774367588369}, {
                    0.18894208305983962`, 0.7440771676228567}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Dashing[{Small, Small}], 
                   Arrowheads[0.02], 
                   CircleBox[{-1.3923315606353952`, 0.5}, 1.7, {
                    1.3152251853972663`, 1.707932972105553}], 
                   ArrowBox[{{-1.60788207749032, 
                    2.1862793287838094`}, {-1.6247338122969783`, 
                    2.184039546276342}}]}, 
                  {RGBColor[0, 0, 1], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-1.3923315606353952`, 0.5}, 1.8, {
                    1.3152251853972663`, 2.009619811766118}], 
                   ArrowBox[{{-2.140773544770206, 
                    2.1370200354254516`}, {-2.1571060505017954`, 
                    2.129453890004129}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.003], Arrowheads[0.025], 
                   CircleBox[{-1.3923315606353952`, 0.5}, 1.5, {
                    1.3152251853972663`, 1.3540166123462092`}], 
                   ArrowBox[{{-1.0550702660736666`, 
                    1.9615932468339292`}, {-1.0697028178685524`, 
                    1.9648927245162653`}}]}, 
                  {RGBColor[0, 0, 1], ArrowBox[{{0, 2.8}, {0, 2.8}}]}},
                 Frame->True,
                 FrameTicks->Automatic,
                 ImageSize->800,
                 PlotRange->{{-3, 3}, {-1, 3}}], 360->
                GraphicsBox[{InsetBox[
                   TagBox[GridBox[{
                    {"\<\"t\"\>", "17.95`"},
                    {"\<\"driveIntent\"\>", "0"},
                    {"\<\"motorTorque\"\>", 
                    RowBox[{"-", "119.27621056035522`"}]},
                    {"\<\"vw\"\>", 
                    RowBox[{"-", "0.1567807426233922`"}]},
                    {"\<\"controllerState\"\>", 
                    RowBox[{"\[LeftAssociation]", 
                    RowBox[{
                    RowBox[{"\<\"lastError\"\>", "\[Rule]", 
                    RowBox[{"-", "0.1973924159082393`"}]}], ",", 
                    
                    RowBox[{"\<\"lastInt\"\>", "\[Rule]", 
                    "0.01134034014110316`"}], ",", 
                    RowBox[{"\<\"initialized\"\>", "\[Rule]", "True"}]}], 
                    "\[RightAssociation]"}]}
                    },
                    AutoDelete->False,
                    
                    GridBoxItemSize->{
                    "Columns" -> {{Automatic}}, "Rows" -> {{Automatic}}}],
                    "Grid"], {-1, -0.5}], 
                  {GrayLevel[0], LineBox[{{-5, 0}, {5, 0}}]}, 
                  {GrayLevel[0.5], 
                   CircleBox[{-1.4004542867368772`, 0.5}, 0.5], 
                   DiskBox[{-1.4004542867368772`, 0.5}, 0.1]}, 
                  {GrayLevel[0.5], 
                   DiskBox[{-1.0376968578954666`, 1.95547485303607}, 0.1]}, 
                  {GrayLevel[0.5], Thickness[0.01], 
                   LineBox[{{-1.4004542867368772`, 
                    0.5}, {-1.0376968578954666`, 1.95547485303607}}]}, 
                  {RGBColor[0, 1, 0], 
                   DiskBox[{-1.1586160008426036`, 1.4703165686907131`}, 
                    0.05]}, 
                  {RGBColor[0, 0, 1], Arrowheads[0.01], 
                   CircleBox[{-1.4004542867368772`, 0.5}, 0.3333333333333333, 
                    NCache[{
                    Rational[-1, 2] 
                    Pi, -5.149082643605553}, {-1.5707963267948966`, \
-5.149082643605553}]], 
                   ArrowBox[{{-1.2624998925933433`, 
                    0.8034463646966323}, {-1.2594723771827445`, 
                    0.8020516715556818}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-1.4004542867368772`, 0.5}, {0.9850699244702272,
                     0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-1.4004542867368772`, 0.5}, {-3.126373415709684,
                     0.5}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.005], Arrowheads[0.04], 
                   ArrowBox[{{-1.4004542867368772`, 
                    0.5}, {-0.7408492045025795, 0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-1.4004542867368772`, 0.5}, 1.6, {
                    1.3265363998920932`, 0.2148379187834626}], 
                   ArrowBox[{{0.15927412224760218`, 0.8567173814194435}, {
                    0.16276325083863252`, 0.8411025215625945}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Dashing[{Small, Small}], 
                   Arrowheads[0.02], 
                   CircleBox[{-1.4004542867368772`, 0.5}, 1.7, {
                    1.3265363998920932`, 1.6972069934842475`}], 
                   ArrowBox[{{-1.597905755307628, 
                    2.1884942752521415`}, {-1.6147805441546865`, 
                    2.1864353694646215`}}]}, 
                  {RGBColor[0, 0, 1], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-1.4004542867368772`, 0.5}, 1.8, {
                    1.3265363998920932`, 1.989182014116289}], 
                   ArrowBox[{{-2.115285207776492, 
                    2.151973593713185}, {-2.1317689271378617`, 
                    2.1447428056493143`}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.003], Arrowheads[0.025], 
                   CircleBox[{-1.4004542867368772`, 0.5}, 1.5, {
                    1.3265363998920932`, 1.3559927631830302`}], 
                   ArrowBox[{{-1.066081977544558, 
                    1.9622568717039413`}, {-1.0807210210294758`, 
                    1.9655274268331113`}}]}, 
                  {RGBColor[0, 0, 1], ArrowBox[{{0, 2.8}, {0, 2.8}}]}},
                 Frame->True,
                 FrameTicks->Automatic,
                 ImageSize->800,
                 PlotRange->{{-3, 3}, {-1, 3}}], 361->
                GraphicsBox[{InsetBox[
                   TagBox[GridBox[{
                    {"\<\"t\"\>", "18.`"},
                    {"\<\"driveIntent\"\>", "0"},
                    {"\<\"motorTorque\"\>", 
                    RowBox[{"-", "113.22786476755053`"}]},
                    {"\<\"vw\"\>", 
                    RowBox[{"-", "0.1458350333271492`"}]},
                    {"\<\"controllerState\"\>", 
                    RowBox[{"\[LeftAssociation]", 
                    RowBox[{
                    RowBox[{"\<\"lastError\"\>", "\[Rule]", 
                    RowBox[{"-", "0.18859492357259405`"}]}], ",", 
                    
                    RowBox[{"\<\"lastInt\"\>", "\[Rule]", 
                    "0.0016912763774371146`"}], ",", 
                    RowBox[{"\<\"initialized\"\>", "\[Rule]", "True"}]}], 
                    "\[RightAssociation]"}]}
                    },
                    AutoDelete->False,
                    
                    GridBoxItemSize->{
                    "Columns" -> {{Automatic}}, "Rows" -> {{Automatic}}}],
                    "Grid"], {-1, -0.5}], 
                  {GrayLevel[0], LineBox[{{-5, 0}, {5, 0}}]}, 
                  {GrayLevel[0.5], 
                   CircleBox[{-1.4080110249978652`, 0.5}, 0.5], 
                   DiskBox[{-1.4080110249978652`, 0.5}, 0.1]}, 
                  {GrayLevel[0.5], 
                   DiskBox[{-1.0628899703393777`, 1.9597573283362593`}, 0.1]}, 
                  {GrayLevel[0.5], Thickness[0.01], 
                   LineBox[{{-1.4080110249978652`, 
                    0.5}, {-1.0628899703393777`, 1.9597573283362593`}}]}, 
                  {RGBColor[0, 1, 0], 
                   DiskBox[{-1.177930321892207, 1.4731715522241728`}, 0.05]}, 
                  {RGBColor[0, 0, 1], Arrowheads[0.01], 
                   CircleBox[{-1.4080110249978652`, 0.5}, 0.3333333333333333, 
                    NCache[{
                    Rational[-1, 2] 
                    Pi, -4.967632269821412}, {-1.5707963267948966`, \
-4.967632269821412}]], 
                   ArrowBox[{{-1.3270802444074798`, 
                    0.8233594282901644}, {-1.3238507505228554`, 
                    0.8225339661359747}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-1.4080110249978652`, 0.5}, {0.8565462703531455,
                     0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-1.4080110249978652`, 
                    0.5}, {-3.0548515292123137`, 0.5}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.005], Arrowheads[0.04], 
                   ArrowBox[{{-1.4080110249978652`, 
                    0.5}, {-0.7902942338613032, 0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-1.4080110249978652`, 0.5}, 1.6, {
                    1.3386357168116314`, 0.28098528222830144`}], 
                   ArrowBox[{{0.12472764186657304`, 0.9590339629032092}, {
                    0.12924126869562302`, 0.9436838801828503}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Dashing[{Small, Small}], 
                   Arrowheads[0.02], 
                   CircleBox[{-1.4080110249978652`, 0.5}, 1.7, {
                    1.3386357168116314`, 1.6868115113622737`}], 
                   ArrowBox[{{-1.5878994288535921`, 
                    2.1904556078638207`}, {-1.6047947088457981`, 
                    2.1885722317304745`}}]}, 
                  {RGBColor[0, 0, 1], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-1.4080110249978652`, 0.5}, 1.8, {
                    1.3386357168116314`, 1.9676794099646902`}], 
                   ArrowBox[{{-2.0871577000815993`, 
                    2.166961245417154}, {-2.1037930776595086`, 
                    2.1600865444891575`}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.003], Arrowheads[0.025], 
                   CircleBox[{-1.4080110249978652`, 0.5}, 1.5, {
                    1.3386357168116314`, 1.3599269188638108`}], 
                   ArrowBox[{{-1.0793940347782653`, 
                    1.9635610249453253`}, {-1.0940458318146884`, 
                    1.9667739626368645`}}]}, 
                  {RGBColor[0, 0, 1], ArrowBox[{{0, 2.8}, {0, 2.8}}]}},
                 Frame->True,
                 FrameTicks->Automatic,
                 ImageSize->800,
                 PlotRange->{{-3, 3}, {-1, 3}}], 362->
                GraphicsBox[{InsetBox[
                   TagBox[GridBox[{
                    {"\<\"t\"\>", "18.05`"},
                    {"\<\"driveIntent\"\>", "0"},
                    {"\<\"motorTorque\"\>", 
                    RowBox[{"-", "106.93674722157878`"}]},
                    {"\<\"vw\"\>", 
                    RowBox[{"-", "0.13565173101873976`"}]},
                    {"\<\"controllerState\"\>", 
                    RowBox[{"\[LeftAssociation]", 
                    RowBox[{
                    RowBox[{"\<\"lastError\"\>", "\[Rule]", 
                    RowBox[{"-", "0.17899451750900186`"}]}], ",", 
                    RowBox[{"\<\"lastInt\"\>", "\[Rule]", 
                    RowBox[{"-", "0.0074965827365029705`"}]}], ",", 
                    RowBox[{"\<\"initialized\"\>", "\[Rule]", "True"}]}], 
                    "\[RightAssociation]"}]}
                    },
                    AutoDelete->False,
                    
                    GridBoxItemSize->{
                    "Columns" -> {{Automatic}}, "Rows" -> {{Automatic}}}],
                    "Grid"], {-1, -0.5}], 
                  {GrayLevel[0], LineBox[{{-5, 0}, {5, 0}}]}, 
                  {GrayLevel[0.5], 
                   CircleBox[{-1.4150399397613183`, 0.5}, 0.5], 
                   DiskBox[{-1.4150399397613183`, 0.5}, 0.1]}, 
                  {GrayLevel[0.5], 
                   DiskBox[{-1.0884403430337983`, 1.9640125352664919`}, 0.1]}, 
                  {GrayLevel[0.5], Thickness[0.01], 
                   LineBox[{{-1.4150399397613183`, 
                    0.5}, {-1.0884403430337983`, 1.9640125352664919`}}]}, 
                  {RGBColor[0, 1, 0], 
                   DiskBox[{-1.197306875276305, 1.476008356844328}, 0.05]}, 
                  {RGBColor[0, 0, 1], Arrowheads[0.01], 
                   CircleBox[{-1.4150399397613183`, 0.5}, 0.3333333333333333, 
                    NCache[{
                    Rational[-1, 2] 
                    Pi, -4.77889874344226}, {-1.5707963267948966`, \
-4.77889874344226}]], 
                   ArrowBox[{{-1.3962133757881716`, 
                    0.8328012493968076}, {-1.3928863600811554`, 
                    0.832596346971018}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-1.4150399397613183`, 0.5}, {0.7236950046702573,
                     0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-1.4150399397613183`, 
                    0.5}, {-2.9780431150131097`, 0.5}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.005], Arrowheads[0.04], 
                   ArrowBox[{{-1.4150399397613183`, 
                    0.5}, {-0.8393081705815343, 0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-1.4150399397613183`, 0.5}, 1.6, {
                    1.3513051203889042`, 0.3504151406239253}], 
                   ArrowBox[{{0.08216085974284115, 1.0642603707191443`}, {
                    0.08772850999096415, 1.0492603994728933`}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Dashing[{Small, Small}], 
                   Arrowheads[0.02], 
                   CircleBox[{-1.4150399397613183`, 0.5}, 1.7, {
                    1.3513051203889042`, 1.6767835535511681`}], 
                   ArrowBox[{{-1.5779677655031354`, 
                    2.1921744956117983`}, {-1.5948810821081802`, 
                    2.1904606364891714`}}]}, 
                  {RGBColor[0, 0, 1], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-1.4150399397613183`, 0.5}, 1.8, {
                    1.3513051203889042`, 1.9453981605087862`}], 
                   ArrowBox[{{-2.0568791330471905`, 
                    2.1816784621211456`}, {-2.0736635456978276`, 
                    2.175176093938542}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.003], Arrowheads[0.025], 
                   CircleBox[{-1.4150399397613183`, 0.5}, 1.5, {
                    1.3513051203889042`, 1.3654763477006844`}], 
                   ArrowBox[{{-1.0945498956685606`, 
                    1.9653621162147676`}, {-1.1092192969735777`, 
                    1.9684936957457098`}}]}, 
                  {RGBColor[0, 0, 1], ArrowBox[{{0, 2.8}, {0, 2.8}}]}},
                 Frame->True,
                 FrameTicks->Automatic,
                 ImageSize->800,
                 PlotRange->{{-3, 3}, {-1, 3}}], 363->
                GraphicsBox[{InsetBox[
                   TagBox[GridBox[{
                    {"\<\"t\"\>", "18.1`"},
                    {"\<\"driveIntent\"\>", "0"},
                    {"\<\"motorTorque\"\>", 
                    RowBox[{"-", "100.48311195756322`"}]},
                    {"\<\"vw\"\>", 
                    RowBox[{"-", "0.12622191443688865`"}]},
                    {"\<\"controllerState\"\>", 
                    RowBox[{"\[LeftAssociation]", 
                    RowBox[{
                    RowBox[{"\<\"lastError\"\>", "\[Rule]", 
                    RowBox[{"-", "0.1687844893466422`"}]}], ",", 
                    RowBox[{"\<\"lastInt\"\>", "\[Rule]", 
                    RowBox[{"-", "0.016188120356144215`"}]}], ",", 
                    RowBox[{"\<\"initialized\"\>", "\[Rule]", "True"}]}], 
                    "\[RightAssociation]"}]}
                    },
                    AutoDelete->False,
                    
                    GridBoxItemSize->{
                    "Columns" -> {{Automatic}}, "Rows" -> {{Automatic}}}],
                    "Grid"], {-1, -0.5}], 
                  {GrayLevel[0], LineBox[{{-5, 0}, {5, 0}}]}, 
                  {GrayLevel[0.5], 
                   CircleBox[{-1.4215789558271275`, 0.5}, 0.5], 
                   DiskBox[{-1.4215789558271275`, 0.5}, 0.1]}, 
                  {GrayLevel[0.5], 
                   DiskBox[{-1.1141110422691325`, 1.9681496797439604`}, 0.1]}, 
                  {GrayLevel[0.5], Thickness[0.01], 
                   LineBox[{{-1.4215789558271275`, 
                    0.5}, {-1.1141110422691325`, 1.9681496797439604`}}]}, 
                  {RGBColor[0, 1, 0], 
                   DiskBox[{-1.2166003467884643`, 1.47876645316264}, 0.05]}, 
                  {RGBColor[0, 0, 1], Arrowheads[0.01], 
                   CircleBox[{-1.4215789558271275`, 0.5}, 0.3333333333333333, 
                    NCache[{
                    Rational[-1, 2] 
                    Pi, -4.5852896855217935`}, {-1.5707963267948966`, \
-4.5852896855217935`}]], 
                   ArrowBox[{{-1.4671356914875964`, 
                    0.8302055343980068}, {-1.4638314133597958`, 
                    0.8306445840227255}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-1.4215789558271275`, 0.5}, {0.5880832833241372,
                     0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-1.4215789558271275`, 
                    0.5}, {-2.8971820470814476`, 0.5}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.005], Arrowheads[0.04], 
                   ArrowBox[{{-1.4215789558271275`, 0.5}, {-0.887519807930183,
                     0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-1.4215789558271275`, 0.5}, 1.6, {
                    1.364354474604869, 0.42209503257564696`}], 
                   ArrowBox[{{0.03136568034061771, 1.1700387184569105`}, {
                    0.03799330922620858, 1.1554760125947205`}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Dashing[{Small, Small}], 
                   Arrowheads[0.02], 
                   CircleBox[{-1.4215789558271275`, 0.5}, 1.7, {
                    1.364354474604869, 1.6671466794045586`}], 
                   ArrowBox[{{-1.568192195885534, 
                    2.193666011301985}, {-1.5851212431213826`, 
                    2.192115220741946}}]}, 
                  {RGBColor[0, 0, 1], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-1.4215789558271275`, 0.5}, 1.8, {
                    1.364354474604869, 1.9225939854802203`}], 
                   ArrowBox[{{-2.024905302120329, 
                    2.195876563865571}, {-2.041833619048375, 
                    2.189758607834947}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.003], Arrowheads[0.025], 
                   CircleBox[{-1.4215789558271275`, 0.5}, 1.5, {
                    1.364354474604869, 1.3723296482637686`}], 
                   ArrowBox[{{-1.1111389264532194`, 
                    1.9675241013906133`}, {-1.1258294447531174`, 
                    1.970555074351003}}]}, 
                  {RGBColor[0, 0, 1], ArrowBox[{{0, 2.8}, {0, 2.8}}]}},
                 Frame->True,
                 FrameTicks->Automatic,
                 ImageSize->800,
                 PlotRange->{{-3, 3}, {-1, 3}}], 364->
                GraphicsBox[{InsetBox[
                   TagBox[GridBox[{
                    {"\<\"t\"\>", "18.150000000000002`"},
                    {"\<\"driveIntent\"\>", "0"},
                    {"\<\"motorTorque\"\>", 
                    RowBox[{"-", "93.93822625487516`"}]},
                    {"\<\"vw\"\>", 
                    RowBox[{"-", "0.11752998046910706`"}]},
                    {"\<\"controllerState\"\>", 
                    RowBox[{"\[LeftAssociation]", 
                    RowBox[{
                    RowBox[{"\<\"lastError\"\>", "\[Rule]", 
                    RowBox[{"-", "0.15813542512548645`"}]}], ",", 
                    RowBox[{"\<\"lastInt\"\>", "\[Rule]", 
                    RowBox[{"-", "0.024357294616292467`"}]}], ",", 
                    RowBox[{"\<\"initialized\"\>", "\[Rule]", "True"}]}], 
                    "\[RightAssociation]"}]}
                    },
                    AutoDelete->False,
                    
                    GridBoxItemSize->{
                    "Columns" -> {{Automatic}}, "Rows" -> {{Automatic}}}],
                    "Grid"], {-1, -0.5}], 
                  {GrayLevel[0], LineBox[{{-5, 0}, {5, 0}}]}, 
                  {GrayLevel[0.5], 
                   CircleBox[{-1.4276653735755669`, 0.5}, 0.5], 
                   DiskBox[{-1.4276653735755669`, 0.5}, 0.1]}, 
                  {GrayLevel[0.5], 
                   DiskBox[{-1.1396972199037523`, 1.9720986184596623`}, 0.1]}, 
                  {GrayLevel[0.5], Thickness[0.01], 
                   LineBox[{{-1.4276653735755669`, 
                    0.5}, {-1.1396972199037523`, 1.9720986184596623`}}]}, 
                  {RGBColor[0, 1, 0], 
                   DiskBox[{-1.2356866044610237`, 1.4813990789731082`}, 
                    0.05]}, 
                  {RGBColor[0, 0, 1], Arrowheads[0.01], 
                   CircleBox[{-1.4276653735755669`, 0.5}, 0.3333333333333333, 
                    NCache[{
                    Rational[-1, 2] 
                    Pi, -4.388943114441151}, {-1.5707963267948966`, \
-4.388943114441151}]], 
                   ArrowBox[{{-1.5367657179741998`, 
                    0.8149733734194222}, {-1.5336105817635437`, 
                    0.8160486101426763}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-1.4276653735755669`, 0.5}, {
                    0.45109915152193647`, 0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-1.4276653735755669`, 
                    0.5}, {-2.8134022932710003`, 0.5}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.005], Arrowheads[0.04], 
                   ArrowBox[{{-1.4276653735755669`, 
                    0.5}, {-0.9346377681734972, 0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-1.4276653735755669`, 0.5}, 1.6, {
                    1.3776183027080464`, 0.4951173982687649}], 
                   ArrowBox[{{-0.027477101192462383`, 
                    1.2742562895326182`}, {-0.01980467616941528, 
                    1.2602159276804614`}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Dashing[{Small, Small}], 
                   Arrowheads[0.02], 
                   CircleBox[{-1.4276653735755669`, 0.5}, 1.7, {
                    1.3776183027080464`, 1.657915983623833}], 
                   ArrowBox[{{-1.5586388738414825`, 
                    2.194947179775256}, {-1.5755815145290097`, 
                    2.1935527199486415`}}]}, 
                  {RGBColor[0, 0, 1], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-1.4276653735755669`, 0.5}, 1.8, {
                    1.3776183027080464`, 1.899497337457353}], 
                   ArrowBox[{{-1.991665221560301, 
                    2.209357824293438}, {-2.0087303151542892`, 
                    2.203632452634111}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.003], Arrowheads[0.025], 
                   CircleBox[{-1.4276653735755669`, 0.5}, 1.5, {
                    1.3776183027080464`, 1.3802095513765387`}], 
                   ArrowBox[{{-1.1287988102703184`, 
                    1.9699247522707106`}, {-1.1435127560104297`, 
                    1.9728398724677714`}}]}, 
                  {RGBColor[0, 0, 1], ArrowBox[{{0, 2.8}, {0, 2.8}}]}},
                 Frame->True,
                 FrameTicks->Automatic,
                 ImageSize->800,
                 PlotRange->{{-3, 3}, {-1, 3}}], 365->
                GraphicsBox[{InsetBox[
                   TagBox[GridBox[{
                    {"\<\"t\"\>", "18.2`"},
                    {"\<\"driveIntent\"\>", "0"},
                    {"\<\"motorTorque\"\>", 
                    RowBox[{"-", "87.36554687609997`"}]},
                    {"\<\"vw\"\>", 
                    RowBox[{"-", "0.10955534105914588`"}]},
                    {"\<\"controllerState\"\>", 
                    RowBox[{"\[LeftAssociation]", 
                    RowBox[{
                    RowBox[{"\<\"lastError\"\>", "\[Rule]", 
                    RowBox[{"-", "0.14719767411166804`"}]}], ",", 
                    RowBox[{"\<\"lastInt\"\>", "\[Rule]", 
                    RowBox[{"-", "0.03198606701857751`"}]}], ",", 
                    RowBox[{"\<\"initialized\"\>", "\[Rule]", "True"}]}], 
                    "\[RightAssociation]"}]}
                    },
                    AutoDelete->False,
                    
                    GridBoxItemSize->{
                    "Columns" -> {{Automatic}}, "Rows" -> {{Automatic}}}],
                    "Grid"], {-1, -0.5}], 
                  {GrayLevel[0], LineBox[{{-5, 0}, {5, 0}}]}, 
                  {GrayLevel[0.5], 
                   CircleBox[{-1.4333355799063419`, 0.5}, 0.5], 
                   DiskBox[{-1.4333355799063419`, 0.5}, 0.1]}, 
                  {GrayLevel[0.5], 
                   DiskBox[{-1.1650221596276116`, 1.975807544532257}, 0.1]}, 
                  {GrayLevel[0.5], Thickness[0.01], 
                   LineBox[{{-1.4333355799063419`, 
                    0.5}, {-1.1650221596276116`, 1.975807544532257}}]}, 
                  {RGBColor[0, 1, 0], 
                   DiskBox[{-1.2544599663871883`, 1.483871696354838}, 0.05]}, 
                  {RGBColor[0, 0, 1], Arrowheads[0.01], 
                   CircleBox[{-1.4333355799063419`, 0.5}, 0.3333333333333333, 
                    NCache[{
                    Rational[-1, 2] 
                    Pi, -4.191762733077896}, {-1.5707963267948966`, \
-4.191762733077896}]], 
                   ArrowBox[{{-1.602026770787093, 
                    0.7874967708172478}, {-1.5991434165055531`, 
                    0.7891692798922464}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-1.4333355799063419`, 0.5}, {
                    0.31397535761565765`, 0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-1.4333355799063419`, 
                    0.5}, {-2.7277446351152976`, 0.5}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.005], Arrowheads[0.04], 
                   ArrowBox[{{-1.4333355799063419`, 
                    0.5}, {-0.9804336975932985, 0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-1.4333355799063419`, 0.5}, 1.6, {
                    1.390952812745259, 0.5686853743151308}], 
                   ArrowBox[{{-0.09384378271176708, 
                    1.3750781252256559`}, {-0.08516012133687556, 
                    1.3616396769595802`}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Dashing[{Small, Small}], 
                   Arrowheads[0.02], 
                   CircleBox[{-1.4333355799063419`, 0.5}, 1.7, {
                    1.390952812745259, 1.649101459255634}], 
                   ArrowBox[{{-1.5493640324020852`, 
                    2.196035789189439}, {-1.5663183062484816`, 
                    2.19479072291968}}]}, 
                  {RGBColor[0, 0, 1], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-1.4333355799063419`, 0.5}, 1.8, {
                    1.390952812745259, 1.8763169620569256`}], 
                   ArrowBox[{{-1.9575639000423, 
                    2.2219711578210104`}, {-1.974757113429173, 
                    2.216642864150195}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.003], Arrowheads[0.025], 
                   CircleBox[{-1.4333355799063419`, 0.5}, 1.5, {
                    1.390952812745259, 1.3888729903055468`}], 
                   ArrowBox[{{-1.1767191107229178`, 
                    1.9778863243645746`}, {-1.161953324508964, 
                    1.9752463087415038`}}]}, 
                  {RGBColor[0, 0, 1], ArrowBox[{{0, 2.8}, {0, 2.8}}]}},
                 Frame->True,
                 FrameTicks->Automatic,
                 ImageSize->800,
                 PlotRange->{{-3, 3}, {-1, 3}}], 366->
                GraphicsBox[{InsetBox[
                   TagBox[GridBox[{
                    {"\<\"t\"\>", "18.25`"},
                    {"\<\"driveIntent\"\>", "0"},
                    {"\<\"motorTorque\"\>", 
                    RowBox[{"-", "80.82161870364418`"}]},
                    {"\<\"vw\"\>", 
                    RowBox[{"-", "0.1022737358517674`"}]},
                    {"\<\"controllerState\"\>", 
                    RowBox[{"\[LeftAssociation]", 
                    RowBox[{
                    RowBox[{"\<\"lastError\"\>", "\[Rule]", 
                    RowBox[{"-", "0.13610360518112916`"}]}], ",", 
                    RowBox[{"\<\"lastInt\"\>", "\[Rule]", 
                    RowBox[{"-", "0.03906344915665074`"}]}], ",", 
                    RowBox[{"\<\"initialized\"\>", "\[Rule]", "True"}]}], 
                    "\[RightAssociation]"}]}
                    },
                    AutoDelete->False,
                    
                    GridBoxItemSize->{
                    "Columns" -> {{Automatic}}, "Rows" -> {{Automatic}}}],
                    "Grid"], {-1, -0.5}], 
                  {GrayLevel[0], LineBox[{{-5, 0}, {5, 0}}]}, 
                  {GrayLevel[0.5], 
                   CircleBox[{-1.4386248339058818`, 0.5}, 0.5], 
                   DiskBox[{-1.4386248339058818`, 0.5}, 0.1]}, 
                  {GrayLevel[0.5], 
                   DiskBox[{-1.1899338712880654`, 1.979240617719857}, 0.1]}, 
                  {GrayLevel[0.5], Thickness[0.01], 
                   LineBox[{{-1.4386248339058818`, 
                    0.5}, {-1.1899338712880654`, 1.979240617719857}}]}, 
                  {RGBColor[0, 1, 0], 
                   DiskBox[{-1.2728308588273376`, 1.4861604118132379`}, 
                    0.05]}, 
                  {RGBColor[0, 0, 1], Arrowheads[0.01], 
                   CircleBox[{-1.4386248339058818`, 0.5}, 0.3333333333333333, 
                    NCache[{
                    Rational[-1, 2] 
                    Pi, -3.995444887904222}, {-1.5707963267948966`, \
-3.995444887904222}]], 
                   ArrowBox[{{-1.6601546141346368`, 
                    0.7490696038919856}, {-1.6576528832104025`, 
                    0.7512724113964118}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-1.4386248339058818`, 0.5}, {
                    0.17780754016700184`, 0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-1.4386248339058818`, 0.5}, {-2.641161279772536,
                     0.5}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.005], Arrowheads[0.04], 
                   ArrowBox[{{-1.4386248339058818`, 
                    0.5}, {-1.0247289056996525`, 0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-1.4386248339058818`, 0.5}, 1.6, {
                    1.4042332523059582`, 0.6421003683822053}], 
                   ArrowBox[{{-0.16692734732721415`, 
                    1.4709714221487158`}, {-0.15728137927794505`, 
                    1.458206111064913}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Dashing[{Small, Small}], 
                   Arrowheads[0.02], 
                   CircleBox[{-1.4386248339058818`, 0.5}, 1.7, {
                    1.4042332523059582`, 1.640710126270164}], 
                   ArrowBox[{{-1.5404173673448267`, 
                    2.196949698764251}, {-1.557381491924675, 
                    2.1958469436173207`}}]}, 
                  {RGBColor[0, 0, 1], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-1.4386248339058818`, 0.5}, 1.8, {
                    1.4042332523059582`, 1.8532422451039814`}], 
                   ArrowBox[{{-1.9229831283491734`, 
                    2.233608099486151}, {-1.9402947026978896`, 
                    2.228677917585057}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.003], Arrowheads[0.025], 
                   CircleBox[{-1.4386248339058818`, 0.5}, 1.5, {
                    1.4042332523059582`, 1.3981094246376828`}], 
                   ArrowBox[{{-1.19566951673705, 
                    1.980193471766239}, {-1.1808799763816933`, 
                    1.9776899500300582`}}]}, 
                  {RGBColor[0, 0, 1], ArrowBox[{{0, 2.8}, {0, 2.8}}]}},
                 Frame->True,
                 FrameTicks->Automatic,
                 ImageSize->800,
                 PlotRange->{{-3, 3}, {-1, 3}}], 367->
                GraphicsBox[{InsetBox[
                   TagBox[GridBox[{
                    {"\<\"t\"\>", "18.3`"},
                    {"\<\"driveIntent\"\>", "0"},
                    {"\<\"motorTorque\"\>", 
                    RowBox[{"-", "74.35675285618166`"}]},
                    {"\<\"vw\"\>", 
                    RowBox[{"-", "0.09565823450669604`"}]},
                    {"\<\"controllerState\"\>", 
                    RowBox[{"\[LeftAssociation]", 
                    RowBox[{
                    RowBox[{"\<\"lastError\"\>", "\[Rule]", 
                    RowBox[{"-", "0.124969600188094`"}]}], ",", 
                    RowBox[{"\<\"lastInt\"\>", "\[Rule]", 
                    RowBox[{"-", "0.045584655671017986`"}]}], ",", 
                    RowBox[{"\<\"initialized\"\>", "\[Rule]", "True"}]}], 
                    "\[RightAssociation]"}]}
                    },
                    AutoDelete->False,
                    
                    GridBoxItemSize->{
                    "Columns" -> {{Automatic}}, "Rows" -> {{Automatic}}}],
                    "Grid"], {-1, -0.5}], 
                  {GrayLevel[0], LineBox[{{-5, 0}, {5, 0}}]}, 
                  {GrayLevel[0.5], 
                   CircleBox[{-1.4435671099388654`, 0.5}, 0.5], 
                   DiskBox[{-1.4435671099388654`, 0.5}, 0.1]}, 
                  {GrayLevel[0.5], 
                   DiskBox[{-1.214302211494329, 1.982375663029185}, 0.1]}, 
                  {GrayLevel[0.5], Thickness[0.01], 
                   LineBox[{{-1.4435671099388654`, 0.5}, {-1.214302211494329, 
                    1.982375663029185}}]}, 
                  {RGBColor[0, 1, 0], 
                   DiskBox[{-1.2907238443091744`, 1.4882504420194569`}, 
                    0.05]}, 
                  {RGBColor[0, 0, 1], Arrowheads[0.01], 
                   CircleBox[{-1.4435671099388654`, 0.5}, 0.3333333333333333, 
                    NCache[{
                    Rational[-1, 2] 
                    Pi, -3.8014989124803464`}, {-1.5707963267948966`, \
-3.8014989124803464`}]], 
                   ArrowBox[{{-1.7089472852376377`, 
                    0.7017039257662181}, {-1.7069170106989382`, 
                    0.7043475981771523}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-1.4435671099388654`, 0.5}, {
                    0.04356794718476786, 0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-1.4435671099388654`, 
                    0.5}, {-2.5545190313842054`, 0.5}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.005], Arrowheads[0.04], 
                   ArrowBox[{{-1.4435671099388654`, 
                    0.5}, {-1.0673839742605722`, 0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-1.4435671099388654`, 0.5}, 1.6, {
                    1.417351619942078, 0.714751427338605}], 
                   ArrowBox[{{-0.245704332047606, 
                    1.5607189850957865`}, {-0.2351572116220484, 
                    1.548687521452382}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Dashing[{Small, Small}], 
                   Arrowheads[0.02], 
                   CircleBox[{-1.4435671099388654`, 0.5}, 1.7, {
                    1.417351619942078, 1.63274727302533}], 
                   ArrowBox[{{-1.5318439975966511`, 
                    2.1977064502161303`}, {-1.5488163653415512`, 
                    2.196738811437158}}]}, 
                  {RGBColor[0, 0, 1], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-1.4435671099388654`, 0.5}, 1.8, {
                    1.417351619942078, 1.8304446913653094`}], 
                   ArrowBox[{{-1.8882809419937046`, 
                    2.244198844048207}, {-1.905700404229527, 
                    2.2396645706308043`}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.003], Arrowheads[0.025], 
                   CircleBox[{-1.4435671099388654`, 0.5}, 1.5, {
                    1.417351619942078, 1.4077381737006038`}], 
                   ArrowBox[{{-1.2148752464929666`, 
                    1.9824641754840628`}, {-1.2000622863121375`, 
                    1.980103172373641}}]}, 
                  {RGBColor[0, 0, 1], ArrowBox[{{0, 2.8}, {0, 2.8}}]}},
                 Frame->True,
                 FrameTicks->Automatic,
                 ImageSize->800,
                 PlotRange->{{-3, 3}, {-1, 3}}], 368->
                GraphicsBox[{InsetBox[
                   TagBox[GridBox[{
                    {"\<\"t\"\>", "18.35`"},
                    {"\<\"driveIntent\"\>", "0"},
                    {"\<\"motorTorque\"\>", 
                    RowBox[{"-", "68.01553618476193`"}]},
                    {"\<\"vw\"\>", 
                    RowBox[{"-", "0.08967999498424581`"}]},
                    {"\<\"controllerState\"\>", 
                    RowBox[{"\[LeftAssociation]", 
                    RowBox[{
                    RowBox[{"\<\"lastError\"\>", "\[Rule]", 
                    RowBox[{"-", "0.1138977726734219`"}]}], ",", 
                    RowBox[{"\<\"lastInt\"\>", "\[Rule]", 
                    RowBox[{"-", "0.051550349855949064`"}]}], ",", 
                    RowBox[{"\<\"initialized\"\>", "\[Rule]", "True"}]}], 
                    "\[RightAssociation]"}]}
                    },
                    AutoDelete->False,
                    
                    GridBoxItemSize->{
                    "Columns" -> {{Automatic}}, "Rows" -> {{Automatic}}}],
                    "Grid"], {-1, -0.5}], 
                  {GrayLevel[0], LineBox[{{-5, 0}, {5, 0}}]}, 
                  {GrayLevel[0.5], 
                   CircleBox[{-1.4481949843760107`, 0.5}, 0.5], 
                   DiskBox[{-1.4481949843760107`, 0.5}, 0.1]}, 
                  {GrayLevel[0.5], 
                   DiskBox[{-1.238016479861041, 1.9852020051965493`}, 0.1]}, 
                  {GrayLevel[0.5], Thickness[0.01], 
                   LineBox[{{-1.4481949843760107`, 0.5}, {-1.238016479861041, 
                    1.9852020051965493`}}]}, 
                  {RGBColor[0, 1, 0], 
                   DiskBox[{-1.3080759813660308`, 1.4901346701310327`}, 
                    0.05]}, 
                  {RGBColor[0, 0, 1], Arrowheads[0.01], 
                   CircleBox[{-1.4481949843760107`, 0.5}, 0.3333333333333333, 
                    NCache[{
                    Rational[-1, 2] 
                    Pi, -3.611262412337754}, {-1.5707963267948966`, \
-3.611262412337754}]], 
                   ArrowBox[{{-1.7469280039227184`, 
                    0.6478840564212973}, {-1.7454342514792194`, 
                    0.6508639426869738}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-1.4481949843760107`, 
                    0.5}, {-0.08788426068077215, 0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-1.4481949843760107`, 
                    0.5}, {-2.4686015523407585`, 0.5}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.005], Arrowheads[0.04], 
                   ArrowBox[{{-1.4481949843760107`, 
                    0.5}, {-1.1082908286455198`, 0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-1.4481949843760107`, 0.5}, 1.6, {
                    1.430214724395587, 0.7861062771137965}], 
                   ArrowBox[{{-0.3290036136364378, 
                    1.6434206031299576`}, {-0.3176255572764948, 
                    1.6321717053997915`}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Dashing[{Small, Small}], 
                   Arrowheads[0.02], 
                   CircleBox[{-1.4481949843760107`, 0.5}, 1.7, {
                    1.430214724395587, 1.6252170853709662`}], 
                   ArrowBox[{{-1.5236854418960932`, 
                    2.198323052550194}, {-1.5406646148777468`, 
                    2.1974832451116777`}}]}, 
                  {RGBColor[0, 0, 1], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-1.4481949843760107`, 0.5}, 1.8, {
                    1.430214724395587, 1.808078814310931}], 
                   ArrowBox[{{-1.8537903063249446`, 
                    2.2537081954570266`}, {-1.8713068163991773`, 
                    2.249564625157356}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.003], Arrowheads[0.025], 
                   CircleBox[{-1.4481949843760107`, 0.5}, 1.5, {
                    1.430214724395587, 1.4176052976321867`}], 
                   ArrowBox[{{-1.2341416739614353`, 
                    1.9846485039565969`}, {-1.2193061389383815`, 
                    1.9824337747212282`}}]}, 
                  {RGBColor[0, 0, 1], ArrowBox[{{0, 2.8}, {0, 2.8}}]}},
                 Frame->True,
                 FrameTicks->Automatic,
                 ImageSize->800,
                 PlotRange->{{-3, 3}, {-1, 3}}], 369->
                GraphicsBox[{InsetBox[
                   TagBox[GridBox[{
                    {"\<\"t\"\>", "18.400000000000002`"},
                    {"\<\"driveIntent\"\>", "0"},
                    {"\<\"motorTorque\"\>", 
                    RowBox[{"-", "61.837216942685366`"}]},
                    {"\<\"vw\"\>", 
                    RowBox[{"-", "0.08430883446173476`"}]},
                    {"\<\"controllerState\"\>", 
                    RowBox[{"\[LeftAssociation]", 
                    RowBox[{
                    RowBox[{"\<\"lastError\"\>", "\[Rule]", 
                    RowBox[{"-", "0.10297742509447166`"}]}], ",", 
                    RowBox[{"\<\"lastInt\"\>", "\[Rule]", 
                    RowBox[{"-", "0.0569659684425822`"}]}], ",", 
                    RowBox[{"\<\"initialized\"\>", "\[Rule]", "True"}]}], 
                    "\[RightAssociation]"}]}
                    },
                    AutoDelete->False,
                    
                    GridBoxItemSize->{
                    "Columns" -> {{Automatic}}, "Rows" -> {{Automatic}}}],
                    "Grid"], {-1, -0.5}], 
                  {GrayLevel[0], LineBox[{{-5, 0}, {5, 0}}]}, 
                  {GrayLevel[0.5], 
                   CircleBox[{-1.4525395552430007`, 0.5}, 0.5], 
                   DiskBox[{-1.4525395552430007`, 0.5}, 0.1]}, 
                  {GrayLevel[0.5], 
                   DiskBox[{-1.2609834253837067`, 1.987718471053354}, 0.1]}, 
                  {GrayLevel[0.5], Thickness[0.01], 
                   LineBox[{{-1.4525395552430007`, 
                    0.5}, {-1.2609834253837067`, 1.987718471053354}}]}, 
                  {RGBColor[0, 1, 0], 
                   DiskBox[{-1.324835468670138, 1.4918123140355695`}, 0.05]}, 
                  {RGBColor[0, 0, 1], Arrowheads[0.01], 
                   CircleBox[{-1.4525395552430007`, 0.5}, 0.3333333333333333, 
                    NCache[{
                    Rational[-1, 2] 
                    Pi, -3.4259128350754575`}, {-1.5707963267948966`, \
-3.4259128350754575`}]], 
                   ArrowBox[{{-1.7734094145281116`, 
                    0.5902975332634519}, {-1.7724904108857218`, 
                    0.5935016635392212}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-1.4525395552430007`, 
                    0.5}, {-0.21579521638929333`, 0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-1.4525395552430007`, 
                    0.5}, {-2.3841111337148866`, 0.5}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.005], Arrowheads[0.04], 
                   ArrowBox[{{-1.4525395552430007`, 
                    0.5}, {-1.1473667948611792`, 0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-1.4525395552430007`, 0.5}, 1.6, {
                    1.4427425610759146`, 0.8557038357994629}], 
                   ArrowBox[{{-0.4155727248842276, 
                    1.7184825779368698`}, {-0.40343995009372313`, 
                    1.7080521588390256`}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Dashing[{Small, Small}], 
                   Arrowheads[0.02], 
                   CircleBox[{-1.4525395552430007`, 0.5}, 1.7, {
                    1.4427425610759146`, 1.6181228704019457`}], 
                   ArrowBox[{{-1.515979945372241, 
                    2.19881585726654}, {-1.5329646488172726`, 
                    2.1980965238535672`}}]}, 
                  {RGBColor[0, 0, 1], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-1.4525395552430007`, 0.5}, 1.8, {
                    1.4427425610759146`, 1.7862826552019444`}], 
                   ArrowBox[{{-1.8198174610099478`, 
                    2.2621313628488218`}, {-1.8374201172090876`, 
                    2.258370539169907}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.003], Arrowheads[0.025], 
                   CircleBox[{-1.4525395552430007`, 0.5}, 1.5, {
                    1.4427425610759146`, 1.4275803843017703`}], 
                   ArrowBox[{{-1.2533061460496757`, 
                    1.9867098064724014`}, {-1.2384492573561263`, 
                    1.984643170715007}}]}, 
                  {RGBColor[0, 0, 1], ArrowBox[{{0, 2.8}, {0, 2.8}}]}},
                 Frame->True,
                 FrameTicks->Automatic,
                 ImageSize->800,
                 PlotRange->{{-3, 3}, {-1, 3}}], 370->
                GraphicsBox[{InsetBox[
                   TagBox[GridBox[{
                    {"\<\"t\"\>", "18.45`"},
                    {"\<\"driveIntent\"\>", "0"},
                    {"\<\"motorTorque\"\>", 
                    RowBox[{"-", "55.85600401366604`"}]},
                    {"\<\"vw\"\>", 
                    RowBox[{"-", "0.0795136597335674`"}]},
                    {"\<\"controllerState\"\>", 
                    RowBox[{"\[LeftAssociation]", 
                    RowBox[{
                    RowBox[{"\<\"lastError\"\>", "\[Rule]", 
                    RowBox[{"-", "0.09228627212344653`"}]}], ",", 
                    RowBox[{"\<\"lastInt\"\>", "\[Rule]", 
                    RowBox[{"-", "0.0618411131443745`"}]}], ",", 
                    RowBox[{"\<\"initialized\"\>", "\[Rule]", "True"}]}], 
                    "\[RightAssociation]"}]}
                    },
                    AutoDelete->False,
                    
                    GridBoxItemSize->{
                    "Columns" -> {{Automatic}}, "Rows" -> {{Automatic}}}],
                    "Grid"], {-1, -0.5}], 
                  {GrayLevel[0], LineBox[{{-5, 0}, {5, 0}}]}, 
                  {GrayLevel[0.5], 
                   CircleBox[{-1.4566303866565702`, 0.5}, 0.5], 
                   DiskBox[{-1.4566303866565702`, 0.5}, 0.1]}, 
                  {GrayLevel[0.5], 
                   DiskBox[{-1.2831255923394274`, 1.9899315710290073`}, 0.1]}, 
                  {GrayLevel[0.5], Thickness[0.01], 
                   LineBox[{{-1.4566303866565702`, 
                    0.5}, {-1.2831255923394274`, 1.9899315710290073`}}]}, 
                  {RGBColor[0, 1, 0], 
                   DiskBox[{-1.340960523778475, 1.4932877140193384`}, 0.05]}, 
                  {RGBColor[0, 0, 1], Arrowheads[0.01], 
                   CircleBox[{-1.4566303866565702`, 0.5}, 0.3333333333333333, 
                    NCache[{
                    Rational[-1, 2] 
                    Pi, -3.246476447204878}, {-1.5707963267948966`, \
-3.246476447204878}]], 
                   ArrowBox[{{-1.788464356337355, 
                    0.5315804951988276}, {-1.7881319650885357`, 
                    0.5348972005786488}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-1.4566303866565702`, 
                    0.5}, {-0.3395103063832494, 0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-1.4566303866565702`, 
                    0.5}, {-2.3016703022982785`, 0.5}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.005], Arrowheads[0.04], 
                   ArrowBox[{{-1.4566303866565702`, 0.5}, {-1.184550222024958,
                     0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-1.4566303866565702`, 0.5}, 1.6, {
                    1.4548669639815037`, 0.9231479667345639}], 
                   ArrowBox[{{-0.5041383044584851, 
                    1.785596683781487}, {-0.4913301760929524, 
                    1.776007642408867}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Dashing[{Small, Small}], 
                   Arrowheads[0.02], 
                   CircleBox[{-1.4566303866565702`, 0.5}, 1.7, {
                    1.4548669639815037`, 1.6114670261341517`}], 
                   ArrowBox[{{-1.5087624012764493`, 
                    2.199200474650261}, {-1.5257515162452795`, 
                    2.1985942038769535`}}]}, 
                  {RGBColor[0, 0, 1], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-1.4566303866565702`, 0.5}, 1.8, {
                    1.4548669639815037`, 1.7651780973907596`}], 
                   ArrowBox[{{-1.7866402596349809`, 
                    2.2694896110847256`}, {-1.8043183604762223`, 
                    2.2661010936130426`}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.003], Arrowheads[0.025], 
                   CircleBox[{-1.4566303866565702`, 0.5}, 1.5, {
                    1.4548669639815037`, 1.4375534641155423`}], 
                   ArrowBox[{{-1.272233715261633, 
                    1.988622809034736}, {-1.2573569550305845`, 
                    1.9867044425332172`}}]}, 
                  {RGBColor[0, 0, 1], ArrowBox[{{0, 2.8}, {0, 2.8}}]}},
                 Frame->True,
                 FrameTicks->Automatic,
                 ImageSize->800,
                 PlotRange->{{-3, 3}, {-1, 3}}], 371->
                GraphicsBox[{InsetBox[
                   TagBox[GridBox[{
                    {"\<\"t\"\>", "18.5`"},
                    {"\<\"driveIntent\"\>", "0"},
                    {"\<\"motorTorque\"\>", 
                    RowBox[{"-", "50.10131009401417`"}]},
                    {"\<\"vw\"\>", 
                    RowBox[{"-", "0.07526279487106205`"}]},
                    {"\<\"controllerState\"\>", 
                    RowBox[{"\[LeftAssociation]", 
                    RowBox[{
                    RowBox[{"\<\"lastError\"\>", "\[Rule]", 
                    RowBox[{"-", "0.08189146454501621`"}]}], ",", 
                    RowBox[{"\<\"lastInt\"\>", "\[Rule]", 
                    RowBox[{"-", "0.06618899813173258`"}]}], ",", 
                    RowBox[{"\<\"initialized\"\>", "\[Rule]", "True"}]}], 
                    "\[RightAssociation]"}]}
                    },
                    AutoDelete->False,
                    
                    GridBoxItemSize->{
                    "Columns" -> {{Automatic}}, "Rows" -> {{Automatic}}}],
                    "Grid"], {-1, -0.5}], 
                  {GrayLevel[0], LineBox[{{-5, 0}, {5, 0}}]}, 
                  {GrayLevel[0.5], 
                   CircleBox[{-1.4604954720209196`, 0.5}, 0.5], 
                   DiskBox[{-1.4604954720209196`, 0.5}, 0.1]}, 
                  {GrayLevel[0.5], 
                   DiskBox[{-1.3043799365215076`, 1.9918538599929057`}, 0.1]}, 
                  {GrayLevel[0.5], Thickness[0.01], 
                   LineBox[{{-1.4604954720209196`, 
                    0.5}, {-1.3043799365215076`, 1.9918538599929057`}}]}, 
                  {RGBColor[0, 1, 0], 
                   DiskBox[{-1.3564184483546449`, 1.4945692399952704`}, 
                    0.05]}, 
                  {RGBColor[0, 0, 1], Arrowheads[0.01], 
                   CircleBox[{-1.4604954720209196`, 0.5}, 0.3333333333333333, 
                    NCache[{
                    Rational[-1, 2] 
                    Pi, -3.0738356296153215`}, {-1.5707963267948966`, \
-3.0738356296153215`}]], 
                   ArrowBox[{{-1.7928216205110938`, 
                    0.47410710250285365`}, {-1.7930639290016481`, 
                    0.477431603234427}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-1.4604954720209196`, 
                    0.5}, {-0.4584692701406361, 0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-1.4604954720209196`, 
                    0.5}, {-2.2218235136711977`, 0.5}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.005], Arrowheads[0.04], 
                   ArrowBox[{{-1.4604954720209196`, 
                    0.5}, {-1.2197973117909142`, 0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-1.4604954720209196`, 0.5}, 1.6, {
                    1.4665304871199358`, 0.9881022320557769}], 
                   ArrowBox[{{-0.5934582742365961, 
                    1.8447105627823068`}, {-0.5800547442247046, 
                    1.8359731003420958`}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Dashing[{Small, Small}], 
                   Arrowheads[0.02], 
                   CircleBox[{-1.4604954720209196`, 0.5}, 1.7, {
                    1.4665304871199358`, 1.6052508657048912`}], 
                   ArrowBox[{{-1.5020640447081939`, 
                    2.1994917045295463`}, {-1.519056600094974, 
                    2.198991051853627}}]}, 
                  {RGBColor[0, 0, 1], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-1.4604954720209196`, 0.5}, 1.8, {
                    1.4665304871199358`, 1.7448710987533478`}], 
                   ArrowBox[{{-1.7545067500297287`, 
                    2.2758258271586285`}, {-1.7722500118904274`, 
                    2.272796972828742}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.003], Arrowheads[0.025], 
                   CircleBox[{-1.4604954720209196`, 0.5}, 1.5, {
                    1.4665304871199358`, 1.4474321780745436`}], 
                   ArrowBox[{{-1.290813237834299, 
                    1.9903717453714145`}, {-1.2759182528156423`, 
                    1.988600433343497}}]}, 
                  {RGBColor[0, 0, 1], ArrowBox[{{0, 2.8}, {0, 2.8}}]}},
                 Frame->True,
                 FrameTicks->Automatic,
                 ImageSize->800,
                 PlotRange->{{-3, 3}, {-1, 3}}], 372->
                GraphicsBox[{InsetBox[
                   TagBox[GridBox[{
                    {"\<\"t\"\>", "18.55`"},
                    {"\<\"driveIntent\"\>", "0"},
                    {"\<\"motorTorque\"\>", 
                    RowBox[{"-", "44.59796290867491`"}]},
                    {"\<\"vw\"\>", 
                    RowBox[{"-", "0.07152423591692489`"}]},
                    {"\<\"controllerState\"\>", 
                    RowBox[{"\[LeftAssociation]", 
                    RowBox[{
                    RowBox[{"\<\"lastError\"\>", "\[Rule]", 
                    RowBox[{"-", "0.07185045027599389`"}]}], ",", 
                    RowBox[{"\<\"lastInt\"\>", "\[Rule]", 
                    RowBox[{"-", "0.07002594439668286`"}]}], ",", 
                    RowBox[{"\<\"initialized\"\>", "\[Rule]", "True"}]}], 
                    "\[RightAssociation]"}]}
                    },
                    AutoDelete->False,
                    
                    GridBoxItemSize->{
                    "Columns" -> {{Automatic}}, "Rows" -> {{Automatic}}}],
                    "Grid"], {-1, -0.5}], 
                  {GrayLevel[0], LineBox[{{-5, 0}, {5, 0}}]}, 
                  {GrayLevel[0.5], 
                   CircleBox[{-1.4641612116376452`, 0.5}, 0.5], 
                   DiskBox[{-1.4641612116376452`, 0.5}, 0.1]}, 
                  {GrayLevel[0.5], 
                   DiskBox[{-1.3246966481449705`, 1.9935024725556358`}, 0.1]}, 
                  {GrayLevel[0.5], Thickness[0.01], 
                   LineBox[{{-1.4641612116376452`, 
                    0.5}, {-1.3246966481449705`, 1.9935024725556358`}}]}, 
                  {RGBColor[0, 1, 0], 
                   DiskBox[{-1.371184835975862, 1.4956683150370904`}, 0.05]}, 
                  {RGBColor[0, 0, 1], Arrowheads[0.01], 
                   CircleBox[{-1.4641612116376452`, 0.5}, 0.3333333333333333, 
                    NCache[{
                    Rational[-1, 2] 
                    Pi, -2.908735214055144}, {-1.5707963267948966`, \
-2.908735214055144}]], 
                   ArrowBox[{{-1.7877128081138431`, 
                    0.41984093607822826`}, {-1.7884982079482725`, 
                    0.4230804060377903}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-1.4641612116376452`, 0.5}, {-0.572201953464147,
                     0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-1.4641612116376452`, 
                    0.5}, {-2.1450391126303816`, 0.5}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.005], Arrowheads[0.04], 
                   ArrowBox[{{-1.4641612116376452`, 
                    0.5}, {-1.2530798544568833`, 0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-1.4641612116376452`, 0.5}, 1.6, {
                    1.4776854699694415`, 1.050285419477117}], 
                   ArrowBox[{{-0.6823642921088905, 
                    1.8959919686786704`}, {-0.6684436946064973, 
                    1.8881043307654455`}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Dashing[{Small, Small}], 
                   Arrowheads[0.02], 
                   CircleBox[{-1.4641612116376452`, 0.5}, 1.7, {
                    1.4776854699694415`, 1.5994743733105408`}], 
                   ArrowBox[{{-1.495912044489981, 
                    2.199703469612621}, {-1.5129072083751989`, 
                    2.1993009821106035`}}]}, 
                  {RGBColor[0, 0, 1], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-1.4641612116376452`, 0.5}, 1.8, {
                    1.4776854699694415`, 1.725451930573191}], 
                   ArrowBox[{{-1.72363416193285, 
                    2.281200097705225}, {-1.7414328925053022`, 
                    2.2785163521848273`}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.003], Arrowheads[0.025], 
                   CircleBox[{-1.4641612116376452`, 0.5}, 1.5, {
                    1.4776854699694415`, 1.4571392592887071`}], 
                   ArrowBox[{{-1.3089539041204592`, 
                    1.9919486223369978`}, {-1.2940424268546562`, 
                    1.9903219783201076`}}]}, 
                  {RGBColor[0, 0, 1], ArrowBox[{{0, 2.8}, {0, 2.8}}]}},
                 Frame->True,
                 FrameTicks->Automatic,
                 ImageSize->800,
                 PlotRange->{{-3, 3}, {-1, 3}}], 373->
                GraphicsBox[{InsetBox[
                   TagBox[GridBox[{
                    {"\<\"t\"\>", "18.6`"},
                    {"\<\"driveIntent\"\>", "0"},
                    {"\<\"motorTorque\"\>", 
                    RowBox[{"-", "39.36640293611389`"}]},
                    {"\<\"vw\"\>", 
                    RowBox[{"-", "0.06826585552796585`"}]},
                    {"\<\"controllerState\"\>", 
                    RowBox[{"\[LeftAssociation]", 
                    RowBox[{
                    RowBox[{"\<\"lastError\"\>", "\[Rule]", 
                    RowBox[{"-", "0.06221170778774548`"}]}], ",", 
                    RowBox[{"\<\"lastInt\"\>", "\[Rule]", 
                    RowBox[{"-", "0.0733709137734382`"}]}], ",", 
                    RowBox[{"\<\"initialized\"\>", "\[Rule]", "True"}]}], 
                    "\[RightAssociation]"}]}
                    },
                    AutoDelete->False,
                    
                    GridBoxItemSize->{
                    "Columns" -> {{Automatic}}, "Rows" -> {{Automatic}}}],
                    "Grid"], {-1, -0.5}], 
                  {GrayLevel[0], LineBox[{{-5, 0}, {5, 0}}]}, 
                  {GrayLevel[0.5], 
                   CircleBox[{-1.4676524016914598`, 0.5}, 0.5], 
                   DiskBox[{-1.4676524016914598`, 0.5}, 0.1]}, 
                  {GrayLevel[0.5], 
                   DiskBox[{-1.3440381256776643`, 1.9948978261960197`}, 0.1]}, 
                  {GrayLevel[0.5], Thickness[0.01], 
                   LineBox[{{-1.4676524016914598`, 
                    0.5}, {-1.3440381256776643`, 1.9948978261960197`}}]}, 
                  {RGBColor[0, 1, 0], 
                   DiskBox[{-1.3852428843489295`, 1.4965985507973465`}, 
                    0.05]}, 
                  {RGBColor[0, 0, 1], Arrowheads[0.01], 
                   CircleBox[{-1.4676524016914598`, 0.5}, 0.3333333333333333, 
                    NCache[{
                    Rational[-1, 2] 
                    Pi, -2.7517884148783134`}, {-1.5707963267948966`, \
-2.7517884148783134`}]], 
                   ArrowBox[{{-1.7746981381548566`, 
                    0.3702539910796428}, {-1.7759802244609455`, 
                    0.37333088451662855`}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-1.4676524016914598`, 0.5}, {-0.680324342969182,
                     0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-1.4676524016914598`, 0.5}, {-2.071711681936905,
                     0.5}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.005], Arrowheads[0.04], 
                   ArrowBox[{{-1.4676524016914598`, 0.5}, {-1.284383623214627,
                     0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-1.4676524016914598`, 0.5}, 1.6, {
                    1.4882932448755692`, 1.1094676386659796`}], 
                   ArrowBox[{{-0.7697939005904114, 
                    1.9397893986416896`}, {-0.755431139201976, 
                    1.9327389410698261`}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Dashing[{Small, Small}], 
                   Arrowheads[0.02], 
                   CircleBox[{-1.4676524016914598`, 0.5}, 1.7, {
                    1.4882932448755692`, 1.5941359463336655`}], 
                   ArrowBox[{{-1.490329082358049, 
                    2.1998487486108713`}, {-1.507326152712865, 
                    2.199536993854471}}]}, 
                  {RGBColor[0, 0, 1], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-1.4676524016914598`, 0.5}, 1.8, {
                    1.4882932448755692`, 1.7069954834095353`}], 
                   ArrowBox[{{-1.6942084006324416`, 
                    2.285685408839937}, {-1.7120536294025452`, 
                    2.2833306030832627`}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.003], Arrowheads[0.025], 
                   CircleBox[{-1.4676524016914598`, 0.5}, 1.5, {
                    1.4882932448755692`, 1.4666103461452753`}], 
                   ArrowBox[{{-1.3265822190213508`, 
                    1.9933516677465566`}, {-1.31165600468494, 
                    1.9918663224702757`}}]}, 
                  {RGBColor[0, 0, 1], ArrowBox[{{0, 2.8}, {0, 2.8}}]}},
                 Frame->True,
                 FrameTicks->Automatic,
                 ImageSize->800,
                 PlotRange->{{-3, 3}, {-1, 3}}], 374->
                GraphicsBox[{InsetBox[
                   TagBox[GridBox[{
                    {"\<\"t\"\>", "18.650000000000002`"},
                    {"\<\"driveIntent\"\>", "0"},
                    {"\<\"motorTorque\"\>", 
                    RowBox[{"-", "34.42288118930242`"}]},
                    {"\<\"vw\"\>", 
                    RowBox[{"-", "0.06545557469591023`"}]},
                    {"\<\"controllerState\"\>", 
                    RowBox[{"\[LeftAssociation]", 
                    RowBox[{
                    RowBox[{"\<\"lastError\"\>", "\[Rule]", 
                    RowBox[{"-", "0.053015383985862884`"}]}], ",", 
                    RowBox[{"\<\"lastInt\"\>", "\[Rule]", 
                    RowBox[{"-", "0.07624507706868792`"}]}], ",", 
                    RowBox[{"\<\"initialized\"\>", "\[Rule]", "True"}]}], 
                    "\[RightAssociation]"}]}
                    },
                    AutoDelete->False,
                    
                    GridBoxItemSize->{
                    "Columns" -> {{Automatic}}, "Rows" -> {{Automatic}}}],
                    "Grid"], {-1, -0.5}], 
                  {GrayLevel[0], LineBox[{{-5, 0}, {5, 0}}]}, 
                  {GrayLevel[0.5], 
                   CircleBox[{-1.4709922325679163`, 0.5}, 0.5], 
                   DiskBox[{-1.4709922325679163`, 0.5}, 0.1]}, 
                  {GrayLevel[0.5], 
                   DiskBox[{-1.362378053938503, 1.9960624853931261`}, 0.1]}, 
                  {GrayLevel[0.5], Thickness[0.01], 
                   LineBox[{{-1.4709922325679163`, 0.5}, {-1.362378053938503, 
                    1.9960624853931261`}}]}, 
                  {RGBColor[0, 1, 0], 
                   DiskBox[{-1.3985827801483075`, 1.497374990262084}, 0.05]}, 
                  {RGBColor[0, 0, 1], Arrowheads[0.01], 
                   CircleBox[{-1.4709922325679163`, 0.5}, 0.3333333333333333, 
                    NCache[{
                    Rational[-1, 2] 
                    Pi, -2.6034827624739694`}, {-1.5707963267948966`, \
-2.6034827624739694`}]], 
                   ArrowBox[{{-1.7554958773745502`, 
                    0.3263083559786142}, {-1.757218539802604, 
                    0.32916202951947315`}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-1.4709922325679163`, 
                    0.5}, {-0.7825346087818679, 0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-1.4709922325679163`, 
                    0.5}, {-2.0021648469391073`, 0.5}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.005], Arrowheads[0.04], 
                   ArrowBox[{{-1.4709922325679163`, 
                    0.5}, {-1.3137072231530589`, 0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-1.4709922325679163`, 0.5}, 1.6, {
                    1.4983234492677733`, 1.165466813375231}], 
                   ArrowBox[{{-0.854812540302409, 
                    1.9765915436706203`}, {-0.8400776796909356, 
                    1.9703560204821278`}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Dashing[{Small, Small}], 
                   Arrowheads[0.02], 
                   CircleBox[{-1.4709922325679163`, 0.5}, 1.7, {
                    1.4983234492677733`, 1.5892321614606888`}], 
                   ArrowBox[{{-1.4853329814091667`, 
                    2.199939511548182}, {-1.5023313761713464`, 
                    2.199711110182611}}]}, 
                  {RGBColor[0, 0, 1], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-1.4709922325679163`, 0.5}, 1.8, {
                    1.4983234492677733`, 1.6895616780972313`}], 
                   ArrowBox[{{-1.6663840881454404`, 
                    2.2893635803754284`}, {-1.6842676562120564`, 
                    2.2873202269513464`}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.003], Arrowheads[0.025], 
                   CircleBox[{-1.4709922325679163`, 0.5}, 1.5, {
                    1.4983234492677733`, 1.4757921193963819`}], 
                   ArrowBox[{{-1.3436394200527646`, 
                    1.9945839759426303`}, {-1.3287001969769845`, 
                    1.9932357404667853`}}]}, 
                  {RGBColor[0, 0, 1], ArrowBox[{{0, 2.8}, {0, 2.8}}]}},
                 Frame->True,
                 FrameTicks->Automatic,
                 ImageSize->800,
                 PlotRange->{{-3, 3}, {-1, 3}}], 375->
                GraphicsBox[{InsetBox[
                   TagBox[GridBox[{
                    {"\<\"t\"\>", "18.7`"},
                    {"\<\"driveIntent\"\>", "0"},
                    {"\<\"motorTorque\"\>", 
                    RowBox[{"-", "29.77966630723681`"}]},
                    {"\<\"vw\"\>", 
                    RowBox[{"-", "0.06306151386476296`"}]},
                    {"\<\"controllerState\"\>", 
                    RowBox[{"\[LeftAssociation]", 
                    RowBox[{
                    RowBox[{"\<\"lastError\"\>", "\[Rule]", 
                    RowBox[{"-", "0.04429386426414796`"}]}], ",", 
                    RowBox[{"\<\"lastInt\"\>", "\[Rule]", 
                    RowBox[{"-", "0.07867141225075953`"}]}], ",", 
                    RowBox[{"\<\"initialized\"\>", "\[Rule]", "True"}]}], 
                    "\[RightAssociation]"}]}
                    },
                    AutoDelete->False,
                    
                    GridBoxItemSize->{
                    "Columns" -> {{Automatic}}, "Rows" -> {{Automatic}}}],
                    "Grid"], {-1, -0.5}], 
                  {GrayLevel[0], LineBox[{{-5, 0}, {5, 0}}]}, 
                  {GrayLevel[0.5], 
                   CircleBox[{-1.4742022951893226`, 0.5}, 0.5], 
                   DiskBox[{-1.4742022951893226`, 0.5}, 0.1]}, 
                  {GrayLevel[0.5], 
                   DiskBox[{-1.379700549201265, 1.9970201802264418`}, 0.1]}, 
                  {GrayLevel[0.5], Thickness[0.01], 
                   LineBox[{{-1.4742022951893226`, 0.5}, {-1.379700549201265, 
                    1.9970201802264418`}}]}, 
                  {RGBColor[0, 1, 0], 
                   DiskBox[{-1.4112011311972843`, 1.4980134534842946`}, 
                    0.05]}, 
                  {RGBColor[0, 0, 1], Arrowheads[0.01], 
                   CircleBox[{-1.4742022951893226`, 0.5}, 0.3333333333333333, 
                    NCache[{
                    Rational[-1, 2] 
                    Pi, -2.4641863160120008`}, {-1.5707963267948966`, \
-2.4641863160120008`}]], 
                   ArrowBox[{{-1.7318337557222372`, 
                    0.2884884361204137}, {-1.7339359546436013`, 
                    0.29107528327744514`}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-1.4742022951893226`, 
                    0.5}, {-0.8786089690445864, 0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-1.4742022951893226`, 
                    0.5}, {-1.9366545572672536`, 0.5}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.005], Arrowheads[0.04], 
                   ArrowBox[{{-1.4742022951893226`, 
                    0.5}, {-1.3410612311225174`, 0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-1.4742022951893226`, 0.5}, 1.6, {
                    1.5077534113769098`, 1.218145427847602}], 
                   ArrowBox[{{-0.9366261914873133, 
                    2.0069877015850417`}, {-0.9215834442160198, 
                    2.001536681386421}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Dashing[{Small, Small}], 
                   Arrowheads[0.02], 
                   CircleBox[{-1.4742022951893226`, 0.5}, 1.7, {
                    1.5077534113769098`, 1.584757590058589}], 
                   ArrowBox[{{-1.4809364251260144`, 
                    2.1999866621517934`}, {-1.4979356717141477`, 
                    2.1998343233499935`}}]}, 
                  {RGBColor[0, 0, 1], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-1.4742022951893226`, 0.5}, 1.8, {
                    1.5077534113769098`, 1.6731960019726697`}], 
                   ArrowBox[{{-1.6402851473059012`, 
                    2.292321535392805}, {-1.658199759867662, 
                    2.2905711192220046`}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.003], Arrowheads[0.025], 
                   CircleBox[{-1.4742022951893226`, 0.5}, 1.5, {
                    1.5077534113769098`, 1.4846407401664639`}], 
                   ArrowBox[{{-1.3600793026133033`, 
                    1.99565234682579}, {-1.345128734521268, 
                    1.9944363539262788`}}]}, 
                  {RGBColor[0, 0, 1], ArrowBox[{{0, 2.8}, {0, 2.8}}]}},
                 Frame->True,
                 FrameTicks->Automatic,
                 ImageSize->800,
                 PlotRange->{{-3, 3}, {-1, 3}}], 376->
                GraphicsBox[{InsetBox[
                   TagBox[GridBox[{
                    {"\<\"t\"\>", "18.75`"},
                    {"\<\"driveIntent\"\>", "0"},
                    {"\<\"motorTorque\"\>", 
                    RowBox[{"-", "25.445266525792494`"}]},
                    {"\<\"vw\"\>", 
                    RowBox[{"-", "0.061052131815761655`"}]},
                    {"\<\"controllerState\"\>", 
                    RowBox[{"\[LeftAssociation]", 
                    RowBox[{
                    RowBox[{"\<\"lastError\"\>", "\[Rule]", 
                    RowBox[{"-", "0.036072297606784334`"}]}], ",", 
                    RowBox[{"\<\"lastInt\"\>", "\[Rule]", 
                    RowBox[{"-", "0.08067432991026462`"}]}], ",", 
                    RowBox[{"\<\"initialized\"\>", "\[Rule]", "True"}]}], 
                    "\[RightAssociation]"}]}
                    },
                    AutoDelete->False,
                    
                    GridBoxItemSize->{
                    "Columns" -> {{Automatic}}, "Rows" -> {{Automatic}}}],
                    "Grid"], {-1, -0.5}], 
                  {GrayLevel[0], LineBox[{{-5, 0}, {5, 0}}]}, 
                  {GrayLevel[0.5], 
                   CircleBox[{-1.4773025945669016`, 0.5}, 0.5], 
                   DiskBox[{-1.4773025945669016`, 0.5}, 0.1]}, 
                  {GrayLevel[0.5], 
                   DiskBox[{-1.3959993431460178`, 1.9977949730548545`}, 0.1]}, 
                  {GrayLevel[0.5], Thickness[0.01], 
                   LineBox[{{-1.4773025945669016`, 
                    0.5}, {-1.3959993431460178`, 1.9977949730548545`}}]}, 
                  {RGBColor[0, 1, 0], 
                   DiskBox[{-1.423100426952979, 1.4985299820365698`}, 0.05]}, 
                  {RGBColor[0, 0, 1], Arrowheads[0.01], 
                   CircleBox[{-1.4773025945669016`, 0.5}, 0.3333333333333333, 
                    NCache[{
                    Rational[-1, 2] 
                    Pi, -2.334154322568671}, {-1.5707963267948966`, \
-2.334154322568671}]], 
                   ArrowBox[{{-1.7053332347046355`, 
                    0.2568680640691429}, {-1.7077531121051628`, 
                    0.25916048896109545`}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-1.4773025945669016`, 
                    0.5}, {-0.9683972640510518, 0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-1.4773025945669016`, 
                    0.5}, {-1.8753728279690205`, 0.5}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.005], Arrowheads[0.04], 
                   ArrowBox[{{-1.4773025945669016`, 
                    0.5}, {-1.3664674974531708`, 0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-1.4773025945669016`, 0.5}, 1.6, {
                    1.5165675841673354`, 1.26740741837152}], 
                   ArrowBox[{{-1.0145858304163775`, 
                    2.0316308942346613`}, {-0.9992929123899797, 
                    2.026927222805697}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Dashing[{Small, Small}], 
                   Arrowheads[0.02], 
                   CircleBox[{-1.4773025945669016`, 0.5}, 1.7, {
                    1.5165675841673354`, 1.5807046782909364`}], 
                   ArrowBox[{{-1.4771467921103874`, 
                    2.199999992860469}, {-1.4941465164971346`, 
                    2.199916551567755}}]}, 
                  {RGBColor[0, 0, 1], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-1.4773025945669016`, 0.5}, 1.8, {
                    1.5165675841673354`, 1.6579301759772938`}], 
                   ArrowBox[{{-1.616005888670404, 
                    2.294647986710719}, {-1.6339451343240967`, 
                    2.293171245235216}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.003], Arrowheads[0.025], 
                   CircleBox[{-1.4773025945669016`, 0.5}, 1.5, {
                    1.5165675841673354`, 1.4931205574971633`}], 
                   ArrowBox[{{-1.3758664117980421`, 
                    1.996566303517917}, {-1.3609060699562066`, 
                    1.9954771309045658`}}]}, 
                  {RGBColor[0, 0, 1], ArrowBox[{{0, 2.8}, {0, 2.8}}]}},
                 Frame->True,
                 FrameTicks->Automatic,
                 ImageSize->800,
                 PlotRange->{{-3, 3}, {-1, 3}}], 377->
                GraphicsBox[{InsetBox[
                   TagBox[GridBox[{
                    {"\<\"t\"\>", "18.8`"},
                    {"\<\"driveIntent\"\>", "0"},
                    {"\<\"motorTorque\"\>", 
                    RowBox[{"-", "21.424669004537456`"}]},
                    {"\<\"vw\"\>", 
                    RowBox[{"-", "0.05939635751078064`"}]},
                    {"\<\"controllerState\"\>", 
                    RowBox[{"\[LeftAssociation]", 
                    RowBox[{
                    RowBox[{"\<\"lastError\"\>", "\[Rule]", 
                    RowBox[{"-", "0.028369094684828344`"}]}], ",", 
                    RowBox[{"\<\"lastInt\"\>", "\[Rule]", 
                    RowBox[{"-", "0.08227932422253931`"}]}], ",", 
                    RowBox[{"\<\"initialized\"\>", "\[Rule]", "True"}]}], 
                    "\[RightAssociation]"}]}
                    },
                    AutoDelete->False,
                    
                    GridBoxItemSize->{
                    "Columns" -> {{Automatic}}, "Rows" -> {{Automatic}}}],
                    "Grid"], {-1, -0.5}], 
                  {GrayLevel[0], LineBox[{{-5, 0}, {5, 0}}]}, 
                  {GrayLevel[0.5], 
                   CircleBox[{-1.4803115701017175`, 0.5}, 0.5], 
                   DiskBox[{-1.4803115701017175`, 0.5}, 0.1]}, 
                  {GrayLevel[0.5], 
                   DiskBox[{-1.4112769858922776`, 1.9984105666281955`}, 0.1]}, 
                  {GrayLevel[0.5], Thickness[0.01], 
                   LineBox[{{-1.4803115701017175`, 
                    0.5}, {-1.4112769858922776`, 1.9984105666281955`}}]}, 
                  {RGBColor[0, 1, 0], 
                   DiskBox[{-1.434288513962091, 1.4989403777521302`}, 0.05]}, 
                  {RGBColor[0, 0, 1], Arrowheads[0.01], 
                   CircleBox[{-1.4803115701017175`, 0.5}, 0.3333333333333333, 
                    NCache[{
                    Rational[-1, 2] 
                    Pi, -2.21353639693102}, {-1.5707963267948966`, \
-2.21353639693102}]], 
                   ArrowBox[{{-1.6774304352927256`, 
                    0.23119660698398853`}, {-1.6801085685614172`, 
                    0.23318120284056842`}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-1.4803115701017175`, 
                    0.5}, {-1.0518181900109684`, 0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-1.4803115701017175`, 
                    0.5}, {-1.8184518916464139`, 0.5}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.005], Arrowheads[0.04], 
                   ArrowBox[{{-1.4803115701017175`, 
                    0.5}, {-1.3899585115556645`, 0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-1.4803115701017175`, 0.5}, 1.6, {
                    1.524757008077452, 1.3131951314407315`}], 
                   ArrowBox[{{-1.0881851412193453`, 
                    2.0512049715533918`}, {-1.0726929561947391`, 
                    2.047206213016401}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Dashing[{Small, Small}], 
                   Arrowheads[0.02], 
                   CircleBox[{-1.4803115701017175`, 0.5}, 1.7, {
                    1.524757008077452, 1.5770636995685112`}], 
                   ArrowBox[{{-1.47396611855156, 
                    2.19998815738364}, {-1.490966034065387, 
                    2.1999666121420285`}}]}, 
                  {RGBColor[0, 0, 1], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-1.4803115701017175`, 0.5}, 1.8, {
                    1.524757008077452, 1.6437829469915488`}], 
                   ArrowBox[{{-1.5936125350088797`, 
                    2.2964305974212045`}, {-1.6115708765784524`, 
                    2.295207785874172}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.003], Arrowheads[0.025], 
                   CircleBox[{-1.4803115701017175`, 0.5}, 1.5, {
                    1.524757008077452, 1.5012030500517337`}], 
                   ArrowBox[{{-1.3909745548672345`, 
                    1.9973372691912112`}, {-1.3760058985438244`, 
                    1.9963690476887233`}}]}, 
                  {RGBColor[0, 0, 1], ArrowBox[{{0, 2.8}, {0, 2.8}}]}},
                 Frame->True,
                 FrameTicks->Automatic,
                 ImageSize->800,
                 PlotRange->{{-3, 3}, {-1, 3}}], 378->
                GraphicsBox[{InsetBox[
                   TagBox[GridBox[{
                    {"\<\"t\"\>", "18.85`"},
                    {"\<\"driveIntent\"\>", "0"},
                    {"\<\"motorTorque\"\>", 
                    RowBox[{"-", "17.71959647923742`"}]},
                    {"\<\"vw\"\>", 
                    RowBox[{"-", "0.058063717588830624`"}]},
                    {"\<\"controllerState\"\>", 
                    RowBox[{"\[LeftAssociation]", 
                    RowBox[{
                    RowBox[{"\<\"lastError\"\>", "\[Rule]", 
                    RowBox[{"-", "0.021196412174120466`"}]}], ",", 
                    RowBox[{"\<\"lastInt\"\>", "\[Rule]", 
                    RowBox[{"-", "0.08351264841781031`"}]}], ",", 
                    RowBox[{"\<\"initialized\"\>", "\[Rule]", "True"}]}], 
                    "\[RightAssociation]"}]}
                    },
                    AutoDelete->False,
                    
                    GridBoxItemSize->{
                    "Columns" -> {{Automatic}}, "Rows" -> {{Automatic}}}],
                    "Grid"], {-1, -0.5}], 
                  {GrayLevel[0], LineBox[{{-5, 0}, {5, 0}}]}, 
                  {GrayLevel[0.5], 
                   CircleBox[{-1.4832461223600628`, 0.5}, 0.5], 
                   DiskBox[{-1.4832461223600628`, 0.5}, 0.1]}, 
                  {GrayLevel[0.5], 
                   DiskBox[{-1.4255440557496815`, 1.9988897462818576`}, 0.1]}, 
                  {GrayLevel[0.5], Thickness[0.01], 
                   LineBox[{{-1.4832461223600628`, 
                    0.5}, {-1.4255440557496815`, 1.9988897462818576`}}]}, 
                  {RGBColor[0, 1, 0], 
                   DiskBox[{-1.444778077953142, 1.4992598308545717`}, 0.05]}, 
                  {RGBColor[0, 0, 1], Arrowheads[0.01], 
                   CircleBox[{-1.4832461223600628`, 0.5}, 0.3333333333333333, 
                    NCache[{
                    Rational[-1, 2] 
                    Pi, -2.102384221172019}, {-1.5707963267948966`, \
-2.102384221172019}]], 
                   ArrowBox[{{-1.6493319610406085`, 
                    0.21099030241029837`}, {-1.6522137056257318`, 
                    0.21266558348072823`}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-1.4832461223600628`, 
                    0.5}, {-1.1288541927753144`, 0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-1.4832461223600628`, 
                    0.5}, {-1.7659686905962944`, 0.5}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.005], Arrowheads[0.04], 
                   ArrowBox[{{-1.4832461223600628`, 0.5}, {-1.411576761011546,
                     0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-1.4832461223600628`, 0.5}, 1.6, {
                    1.532318788623935, 1.3554862975224675`}], 
                   ArrowBox[{{-1.1570530220698863`, 
                    2.066396521102841}, {-1.1414056274427404`, 
                    2.063056325291792}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Dashing[{Small, Small}], 
                   Arrowheads[0.02], 
                   CircleBox[{-1.4832461223600628`, 0.5}, 1.7, {
                    1.532318788623935, 1.5738227805041725`}], 
                   ArrowBox[{{-1.471391189751293, 
                    2.1999586643718256`}, {-1.4883910858116747`, 
                    2.1999922144971964`}}]}, 
                  {RGBColor[0, 0, 1], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-1.4832461223600628`, 0.5}, 1.8, {
                    1.532318788623935, 1.630760991286365}], 
                   ArrowBox[{{-1.5731451025641086`, 
                    2.297753646459457}, {-1.5911178444930412`, 
                    2.2967647847072428`}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.003], Arrowheads[0.025], 
                   CircleBox[{-1.4832461223600628`, 0.5}, 1.5, {
                    1.532318788623935, 1.5088659654415553`}], 
                   ArrowBox[{{-1.4053855865714138`, 
                    1.997977882669402}, {-1.3904099504007996`, 
                    1.9971243920182284`}}]}, 
                  {RGBColor[0, 0, 1], ArrowBox[{{0, 2.8}, {0, 2.8}}]}},
                 Frame->True,
                 FrameTicks->Automatic,
                 ImageSize->800,
                 PlotRange->{{-3, 3}, {-1, 3}}], 379->
                GraphicsBox[{InsetBox[
                   TagBox[GridBox[{
                    {"\<\"t\"\>", "18.900000000000002`"},
                    {"\<\"driveIntent\"\>", "0"},
                    {"\<\"motorTorque\"\>", 
                    RowBox[{"-", "14.328779275383129`"}]},
                    {"\<\"vw\"\>", 
                    RowBox[{"-", "0.05702446032437459`"}]},
                    {"\<\"controllerState\"\>", 
                    RowBox[{"\[LeftAssociation]", 
                    RowBox[{
                    RowBox[{"\<\"lastError\"\>", "\[Rule]", 
                    RowBox[{"-", "0.014560632213266356`"}]}], ",", 
                    RowBox[{"\<\"lastInt\"\>", "\[Rule]", 
                    RowBox[{"-", "0.08440101431447601`"}]}], ",", 
                    RowBox[{"\<\"initialized\"\>", "\[Rule]", "True"}]}], 
                    "\[RightAssociation]"}]}
                    },
                    AutoDelete->False,
                    
                    GridBoxItemSize->{
                    "Columns" -> {{Automatic}}, "Rows" -> {{Automatic}}}],
                    "Grid"], {-1, -0.5}], 
                  {GrayLevel[0], LineBox[{{-5, 0}, {5, 0}}]}, 
                  {GrayLevel[0.5], 
                   CircleBox[{-1.4861216461332736`, 0.5}, 0.5], 
                   DiskBox[{-1.4861216461332736`, 0.5}, 0.1]}, 
                  {GrayLevel[0.5], 
                   DiskBox[{-1.4388183695661902`, 1.9992539478107163`}, 0.1]}, 
                  {GrayLevel[0.5], Thickness[0.01], 
                   LineBox[{{-1.4861216461332736`, 
                    0.5}, {-1.4388183695661902`, 1.9992539478107163`}}]}, 
                  {RGBColor[0, 1, 0], 
                   DiskBox[{-1.4545861284218848`, 1.4995026318738107`}, 
                    0.05]}, 
                  {RGBColor[0, 0, 1], Arrowheads[0.01], 
                   CircleBox[{-1.4861216461332736`, 0.5}, 0.3333333333333333, 
                    NCache[{
                    Rational[-1, 2] 
                    Pi, -2.0006597050563903`}, {-1.5707963267948966`, \
-2.0006597050563903`}]], 
                   ArrowBox[{{-1.6220002132852722`, 
                    0.19561845308916814`}, {-1.6250371841526345`, 
                    0.19699243506489383`}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-1.4861216461332736`, 0.5}, {-1.199546060625611,
                     0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-1.4861216461332736`, 
                    0.5}, {-1.7179496234828786`, 0.5}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.005], Arrowheads[0.04], 
                   ArrowBox[{{-1.4861216461332736`, 
                    0.5}, {-1.4313740379752162`, 0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-1.4861216461332736`, 0.5}, 1.6, {
                    1.539255579789355, 1.3942909931810468`}], 
                   ArrowBox[{{-1.2209423932740082`, 
                    2.077871973213607}, {-1.2051771953713715`, 
                    2.0751413319401193`}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Dashing[{Small, Small}], 
                   Arrowheads[0.02], 
                   CircleBox[{-1.4861216461332736`, 0.5}, 1.7, {
                    1.539255579789355, 1.5709679963845982`}], 
                   ArrowBox[{{-1.4694137534249505`, 
                    2.1999178939940736`}, {-1.4864134844343329`, 
                    2.1999999749501193`}}]}, 
                  {RGBColor[0, 0, 1], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-1.4861216461332736`, 0.5}, 1.8, {
                    1.539255579789355, 1.6188599090970393`}], 
                   ArrowBox[{{-1.5546195511056955`, 
                    2.2986962047589885`}, {-1.5726027885053755`, 
                    2.297921303064741}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.003], Arrowheads[0.025], 
                   CircleBox[{-1.4861216461332736`, 0.5}, 1.5, {
                    1.539255579789355, 1.5160926213843227`}], 
                   ArrowBox[{{-1.419088419994879, 
                    1.9985014336307718`}, {-1.404107007040938, 
                    1.9977561880942285`}}]}, 
                  {RGBColor[0, 0, 1], ArrowBox[{{0, 2.8}, {0, 2.8}}]}},
                 Frame->True,
                 FrameTicks->Automatic,
                 ImageSize->800,
                 PlotRange->{{-3, 3}, {-1, 3}}], 380->
                GraphicsBox[{InsetBox[
                   TagBox[GridBox[{
                    {"\<\"t\"\>", "18.95`"},
                    {"\<\"driveIntent\"\>", "0"},
                    {"\<\"motorTorque\"\>", 
                    RowBox[{"-", "11.248239331176618`"}]},
                    {"\<\"vw\"\>", 
                    RowBox[{"-", "0.05624967551135602`"}]},
                    {"\<\"controllerState\"\>", 
                    RowBox[{"\[LeftAssociation]", 
                    RowBox[{
                    RowBox[{"\<\"lastError\"\>", "\[Rule]", 
                    RowBox[{"-", "0.008462842155222212`"}]}], ",", 
                    RowBox[{"\<\"lastInt\"\>", "\[Rule]", 
                    RowBox[{"-", "0.08497131581834913`"}]}], ",", 
                    RowBox[{"\<\"initialized\"\>", "\[Rule]", "True"}]}], 
                    "\[RightAssociation]"}]}
                    },
                    AutoDelete->False,
                    
                    GridBoxItemSize->{
                    "Columns" -> {{Automatic}}, "Rows" -> {{Automatic}}}],
                    "Grid"], {-1, -0.5}], 
                  {GrayLevel[0], LineBox[{{-5, 0}, {5, 0}}]}, 
                  {GrayLevel[0.5], CircleBox[{-1.488952069595941, 0.5}, 0.5], 
                   DiskBox[{-1.488952069595941, 0.5}, 0.1]}, 
                  {GrayLevel[0.5], 
                   DiskBox[{-1.4511241925660334`, 1.9995229413781606`}, 0.1]}, 
                  {GrayLevel[0.5], Thickness[0.01], 
                   LineBox[{{-1.488952069595941, 0.5}, {-1.4511241925660334`, 
                    1.9995229413781606`}}]}, 
                  {RGBColor[0, 1, 0], 
                   DiskBox[{-1.463733484909336, 1.4996819609187737`}, 0.05]}, 
                  {RGBColor[0, 0, 1], Arrowheads[0.01], 
                   CircleBox[{-1.488952069595941, 0.5}, 0.3333333333333333, 
                    NCache[{
                    Rational[-1, 2] 
                    Pi, -1.908243506730195}, {-1.5707963267948966`, \
-1.908243506730195}]], 
                   ArrowBox[{{-1.5961610360177823`, 
                    0.18437783881693032`}, {-1.5993118446225316`, 
                    0.18546569158962734`}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-1.488952069595941, 0.5}, {-1.2639872829724088`,
                     0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-1.488952069595941, 0.5}, {-1.6743754540894404`,
                     0.5}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.005], Arrowheads[0.04], 
                   ArrowBox[{{-1.488952069595941, 0.5}, {-1.449410667465908, 
                    0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-1.488952069595941, 0.5}, 1.6, {
                    1.5455750682696332`, 1.4296485838085884`}], 
                   ArrowBox[{{-1.2797166852848858`, 
                    2.0862599263526156`}, {-1.2638648120787266`, 
                    2.0840882950465165`}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Dashing[{Small, Small}], 
                   Arrowheads[0.02], 
                   CircleBox[{-1.488952069595941, 0.5}, 1.7, {
                    1.5455750682696332`, 1.5684835282226202`}], 
                   ArrowBox[{{-1.4680208409122768`, 
                    2.1998711373706517`}, {-1.4850203155282564`, 
                    2.1999954533203763`}}]}, 
                  {RGBColor[0, 0, 1], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-1.488952069595941, 0.5}, 1.8, {
                    1.5455750682696332`, 1.60806528677617}], 
                   ArrowBox[{{-1.5380301146600688`, 
                    2.2993308049085037`}, {-1.5560206689403817`, 
                    2.298750066846969}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.003], Arrowheads[0.025], 
                   CircleBox[{-1.488952069595941, 0.5}, 1.5, {
                    1.5455750682696332`, 1.5228713349990812`}], 
                   ArrowBox[{{-1.4320782178479055`, 
                    1.998921400536847}, {-1.4170920973304115`, 
                    1.998277726052816}}]}, 
                  {RGBColor[0, 0, 1], ArrowBox[{{0, 2.8}, {0, 2.8}}]}},
                 Frame->True,
                 FrameTicks->Automatic,
                 ImageSize->800,
                 PlotRange->{{-3, 3}, {-1, 3}}], 381->
                GraphicsBox[{InsetBox[
                   TagBox[GridBox[{
                    {"\<\"t\"\>", "19.`"},
                    {"\<\"driveIntent\"\>", "0"},
                    {"\<\"motorTorque\"\>", 
                    RowBox[{"-", "8.471581998073333`"}]},
                    {"\<\"vw\"\>", 
                    RowBox[{"-", "0.05571140886418904`"}]},
                    {"\<\"controllerState\"\>", 
                    RowBox[{"\[LeftAssociation]", 
                    RowBox[{
                    RowBox[{"\<\"lastError\"\>", "\[Rule]", 
                    RowBox[{"-", "0.0028993166215471704`"}]}], ",", 
                    RowBox[{"\<\"lastInt\"\>", "\[Rule]", 
                    RowBox[{"-", "0.085250376465521`"}]}], ",", 
                    RowBox[{"\<\"initialized\"\>", "\[Rule]", "True"}]}], 
                    "\[RightAssociation]"}]}
                    },
                    AutoDelete->False,
                    
                    GridBoxItemSize->{
                    "Columns" -> {{Automatic}}, "Rows" -> {{Automatic}}}],
                    "Grid"], {-1, -0.5}], 
                  {GrayLevel[0], LineBox[{{-5, 0}, {5, 0}}]}, 
                  {GrayLevel[0.5], 
                   CircleBox[{-1.4917498993256453`, 0.5}, 0.5], 
                   DiskBox[{-1.4917498993256453`, 0.5}, 0.1]}, 
                  {GrayLevel[0.5], 
                   DiskBox[{-1.4624914503533468`, 1.9997146205741063`}, 0.1]}, 
                  {GrayLevel[0.5], Thickness[0.01], 
                   LineBox[{{-1.4917498993256453`, 
                    0.5}, {-1.4624914503533468`, 1.9997146205741063`}}]}, 
                  {RGBColor[0, 1, 0], 
                   DiskBox[{-1.4722442666774462`, 1.499809747049404}, 0.05]}, 
                  {RGBColor[0, 0, 1], Arrowheads[0.01], 
                   CircleBox[{-1.4917498993256453`, 0.5}, 0.3333333333333333, 
                    NCache[{
                    Rational[-1, 2] 
                    Pi, -1.8249437867370966`}, {-1.5707963267948966`, \
-1.8249437867370966`}]], 
                   ArrowBox[{{-1.5723262845955506`, 
                    0.17655207954301688`}, {-1.575556681106713, 
                    0.17737400222763888`}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-1.4917498993256453`, 
                    0.5}, {-1.3223182593641787`, 0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-1.4917498993256453`, 
                    0.5}, {-1.6351862892434983`, 0.5}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.005], Arrowheads[0.04], 
                   ArrowBox[{{-1.4917498993256453`, 
                    0.5}, {-1.4657546492820317`, 0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-1.4917498993256453`, 0.5}, 1.6, {
                    1.5512894570511644`, 1.4616246543592295`}], 
                   ArrowBox[{{-1.3333351903413653`, 
                    2.092138429904079}, {-1.3174219920668462`, 
                    2.090474702958449}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Dashing[{Small, Small}], 
                   Arrowheads[0.02], 
                   CircleBox[{-1.4917498993256453`, 0.5}, 1.7, {
                    1.5512894570511644`, 1.5663518706923751`}], 
                   ArrowBox[{{-1.4671951778311212`, 
                    2.199822657118185}, {-1.484194348825787, 
                    2.1999832098160983`}}]}, 
                  {RGBColor[0, 0, 1], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-1.4917498993256453`, 0.5}, 1.8, {
                    1.5512894570511644`, 1.5983538014849052`}], 
                   ArrowBox[{{-1.5233517300866413`, 
                    2.29972256870123}, {-1.5413470757430212`, 
                    2.2993165702820124`}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.003], Arrowheads[0.025], 
                   CircleBox[{-1.4917498993256453`, 0.5}, 1.5, {
                    1.5512894570511644`, 1.5291949496269615`}], 
                   ArrowBox[{{-1.4443557228352364`, 
                    1.999251077049737}, {-1.4293658316277462`, 
                    1.998702181254656}}]}, 
                  {RGBColor[0, 0, 1], ArrowBox[{{0, 2.8}, {0, 2.8}}]}},
                 Frame->True,
                 FrameTicks->Automatic,
                 ImageSize->800,
                 PlotRange->{{-3, 3}, {-1, 3}}], 382->
                GraphicsBox[{InsetBox[
                   TagBox[GridBox[{
                    {"\<\"t\"\>", "19.05`"},
                    {"\<\"driveIntent\"\>", "0"},
                    {"\<\"motorTorque\"\>", 
                    RowBox[{"-", "5.990290956040658`"}]},
                    {"\<\"vw\"\>", 
                    RowBox[{"-", "0.055382769055582544`"}]},
                    {"\<\"controllerState\"\>", 
                    RowBox[{"\[LeftAssociation]", 
                    RowBox[{
                    
                    RowBox[{"\<\"lastError\"\>", "\[Rule]", 
                    "0.0021379986244663036`"}], ",", 
                    RowBox[{"\<\"lastInt\"\>", "\[Rule]", 
                    RowBox[{"-", "0.08526472112066616`"}]}], ",", 
                    RowBox[{"\<\"initialized\"\>", "\[Rule]", "True"}]}], 
                    "\[RightAssociation]"}]}
                    },
                    AutoDelete->False,
                    
                    GridBoxItemSize->{
                    "Columns" -> {{Automatic}}, "Rows" -> {{Automatic}}}],
                    "Grid"], {-1, -0.5}], 
                  {GrayLevel[0], LineBox[{{-5, 0}, {5, 0}}]}, 
                  {GrayLevel[0.5], 
                   CircleBox[{-1.4945262708633198`, 0.5}, 0.5], 
                   DiskBox[{-1.4945262708633198`, 0.5}, 0.1]}, 
                  {GrayLevel[0.5], 
                   DiskBox[{-1.472954948379817, 1.9998448846618482`}, 0.1]}, 
                  {GrayLevel[0.5], Thickness[0.01], 
                   LineBox[{{-1.4945262708633198`, 0.5}, {-1.472954948379817, 
                    1.9998448846618482`}}]}, 
                  {RGBColor[0, 1, 0], 
                   DiskBox[{-1.4801453892076513`, 1.4998965897745653`}, 
                    0.05]}, 
                  {RGBColor[0, 0, 1], Arrowheads[0.01], 
                   CircleBox[{-1.4945262708633198`, 0.5}, 0.3333333333333333, 
                    NCache[{
                    Rational[-1, 2] 
                    Pi, -1.7505050554761163`}, {-1.5707963267948966`, \
-1.7505050554761163`}]], 
                   ArrowBox[{{-1.5508246939698527`, 
                    0.1714553323110708}, {-1.5541072709918735`, 
                    0.17203473425560384`}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-1.4945262708633198`, 
                    0.5}, {-1.3747204517425067`, 0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-1.4945262708633198`, 
                    0.5}, {-1.6002865378556062`, 0.5}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.005], Arrowheads[0.04], 
                   ArrowBox[{{-1.4945262708633198`, 0.5}, {-1.480480718734793,
                     0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-1.4945262708633198`, 0.5}, 1.6, {
                    1.5564149494086512`, 1.4903079463983882`}], 
                   ArrowBox[{{-1.3818382335353459`, 
                    2.0960267561175687`}, {-1.3658838663322128`, 
                    2.0948200938527375`}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Dashing[{Small, Small}], 
                   Arrowheads[0.02], 
                   CircleBox[{-1.4945262708633198`, 0.5}, 1.7, {
                    1.5564149494086512`, 1.564554078633594}], 
                   ArrowBox[{{-1.4669156630217575`, 
                    2.199775765898143}, {-1.483914517905082, 
                    2.1999668792947564`}}]}, 
                  {RGBColor[0, 0, 1], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-1.4945262708633198`, 0.5}, 1.8, {
                    1.5564149494086512`, 1.589694343608877}], 
                   ArrowBox[{{-1.510542489779969, 
                    2.2999287432372464`}, {-1.5285406764214453`, 
                    2.2996785880302424`}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.003], Arrowheads[0.025], 
                   CircleBox[{-1.4945262708633198`, 0.5}, 1.5, {
                    1.5564149494086512`, 1.535060432359577}], 
                   ArrowBox[{{-1.455926690523249, 
                    1.999503275220688}, {-1.440933837649939, 
                    1.9990423113115485`}}]}, 
                  {RGBColor[0, 0, 1], ArrowBox[{{0, 2.8}, {0, 2.8}}]}},
                 Frame->True,
                 FrameTicks->Automatic,
                 ImageSize->800,
                 PlotRange->{{-3, 3}, {-1, 3}}], 383->
                GraphicsBox[{InsetBox[
                   TagBox[GridBox[{
                    {"\<\"t\"\>", "19.1`"},
                    {"\<\"driveIntent\"\>", "0"},
                    {"\<\"motorTorque\"\>", 
                    RowBox[{"-", "3.79402152901058`"}]},
                    {"\<\"vw\"\>", 
                    RowBox[{"-", "0.05523802536791609`"}]},
                    {"\<\"controllerState\"\>", 
                    RowBox[{"\[LeftAssociation]", 
                    RowBox[{
                    
                    RowBox[{"\<\"lastError\"\>", "\[Rule]", 
                    "0.006661003318806143`"}], ",", 
                    RowBox[{"\<\"lastInt\"\>", "\[Rule]", 
                    RowBox[{"-", "0.0850403718686436`"}]}], ",", 
                    RowBox[{"\<\"initialized\"\>", "\[Rule]", "True"}]}], 
                    "\[RightAssociation]"}]}
                    },
                    AutoDelete->False,
                    
                    GridBoxItemSize->{
                    "Columns" -> {{Automatic}}, "Rows" -> {{Automatic}}}],
                    "Grid"], {-1, -0.5}], 
                  {GrayLevel[0], LineBox[{{-5, 0}, {5, 0}}]}, 
                  {GrayLevel[0.5], 
                   CircleBox[{-1.4972910043945218`, 0.5}, 0.5], 
                   DiskBox[{-1.4972910043945218`, 0.5}, 0.1]}, 
                  {GrayLevel[0.5], 
                   DiskBox[{-1.4825536057523747`, 1.999927601279896}, 0.1]}, 
                  {GrayLevel[0.5], Thickness[0.01], 
                   LineBox[{{-1.4972910043945218`, 
                    0.5}, {-1.4825536057523747`, 1.999927601279896}}]}, 
                  {RGBColor[0, 1, 0], 
                   DiskBox[{-1.4874660719664237`, 1.4999517341865973`}, 
                    0.05]}, 
                  {RGBColor[0, 0, 1], Arrowheads[0.01], 
                   CircleBox[{-1.4972910043945218`, 0.5}, 0.3333333333333333, 
                    NCache[{
                    Rational[-1, 2] 
                    Pi, -1.684616972665214}, {-1.5707963267948966`, \
-1.684616972665214}]], 
                   ArrowBox[{{-1.531835750162209, 
                    0.1684615080402323}, {-1.5351493526027729`, 
                    0.16882352652693777`}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-1.4972910043945218`, 
                    0.5}, {-1.4214105738143101`, 0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-1.4972910043945218`, 
                    0.5}, {-1.5695497717950024`, 0.5}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.005], Arrowheads[0.04], 
                   ArrowBox[{{-1.4972910043945218`, 
                    0.5}, {-1.4936693412147908`, 0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-1.4972910043945218`, 0.5}, 1.6, {
                    1.560971236294295, 1.515807326796166}], 
                   ArrowBox[{{-1.4253328841236554`, 
                    2.0983810649926644`}, {-1.4093529377452723`, 
                    2.0975815773956556`}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Dashing[{Small, Small}], 
                   Arrowheads[0.02], 
                   CircleBox[{-1.4972910043945218`, 0.5}, 1.7, {
                    1.560971236294295, 1.563070039251283}], 
                   ArrowBox[{{-1.467157893705617, 
                    2.1997329189140893`}, {-1.4841564462503336`, 
                    2.1999492590610923`}}]}, 
                  {RGBColor[0, 0, 1], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-1.4972910043945218`, 0.5}, 1.8, {
                    1.560971236294295, 1.582049133677687}], 
                   ArrowBox[{{-1.499546056193651, 
                    2.299998587427608}, {-1.517545629318012, 
                    2.299886038106082}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.003], Arrowheads[0.025], 
                   CircleBox[{-1.4972910043945218`, 0.5}, 1.5, {
                    1.560971236294295, 1.5404685197076147`}], 
                   ArrowBox[{{-1.4668013936830069`, 
                    1.9996900958660295`}, {-1.4518062671392777`, 
                    1.999310220960566}}]}, 
                  {RGBColor[0, 0, 1], ArrowBox[{{0, 2.8}, {0, 2.8}}]}},
                 Frame->True,
                 FrameTicks->Automatic,
                 ImageSize->800,
                 PlotRange->{{-3, 3}, {-1, 3}}], 384->
                GraphicsBox[{InsetBox[
                   TagBox[GridBox[{
                    {"\<\"t\"\>", "19.150000000000002`"},
                    {"\<\"driveIntent\"\>", "0"},
                    {"\<\"motorTorque\"\>", 
                    RowBox[{"-", "1.8708879353069854`"}]},
                    {"\<\"vw\"\>", 
                    RowBox[{"-", "0.05525269404633832`"}]},
                    {"\<\"controllerState\"\>", 
                    RowBox[{"\[LeftAssociation]", 
                    RowBox[{
                    
                    RowBox[{"\<\"lastError\"\>", "\[Rule]", 
                    "0.010684963430421288`"}], ",", 
                    RowBox[{"\<\"lastInt\"\>", "\[Rule]", 
                    RowBox[{"-", "0.08460266798644198`"}]}], ",", 
                    RowBox[{"\<\"initialized\"\>", "\[Rule]", "True"}]}], 
                    "\[RightAssociation]"}]}
                    },
                    AutoDelete->False,
                    
                    GridBoxItemSize->{
                    "Columns" -> {{Automatic}}, "Rows" -> {{Automatic}}}],
                    "Grid"], {-1, -0.5}], 
                  {GrayLevel[0], LineBox[{{-5, 0}, {5, 0}}]}, 
                  {GrayLevel[0.5], 
                   CircleBox[{-1.5000526650326216`, 0.5}, 0.5], 
                   DiskBox[{-1.5000526650326216`, 0.5}, 0.1]}, 
                  {GrayLevel[0.5], 
                   DiskBox[{-1.4913297109405017`, 1.9999746364761994`}, 0.1]}, 
                  {GrayLevel[0.5], Thickness[0.01], 
                   LineBox[{{-1.5000526650326216`, 
                    0.5}, {-1.4913297109405017`, 1.9999746364761994`}}]}, 
                  {RGBColor[0, 1, 0], 
                   DiskBox[{-1.4942373623045415`, 1.499983090984133}, 0.05]}, 
                  {RGBColor[0, 0, 1], Arrowheads[0.01], 
                   CircleBox[{-1.5000526650326216`, 0.5}, 0.3333333333333333, 
                    NCache[{
                    Rational[-1, 2] 
                    Pi, -1.6269229648541061`}, {-1.5707963267948966`, \
-1.6269229648541061`}]], 
                   ArrowBox[{{-1.5154227592927243`, 
                    0.16702121491970895`}, {-1.5187517231490317`, 
                    0.1671915621011537}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-1.5000526650326216`, 
                    0.5}, {-1.4626349063264819`, 0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-1.5000526650326216`, 
                    0.5}, {-1.5428234206827507`, 0.5}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.005], Arrowheads[0.04], 
                   ArrowBox[{{-1.5000526650326216`, 0.5}, {-1.505405661976611,
                     0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-1.5000526650326216`, 0.5}, 1.6, {
                    1.5649809912895796`, 1.5382488169934894`}], 
                   ArrowBox[{{-1.4639797060547965`, 
                    2.099593305071819}, {-1.447985843234548, 
                    2.099152602495412}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Dashing[{Small, Small}], 
                   Arrowheads[0.02], 
                   CircleBox[{-1.5000526650326216`, 0.5}, 1.7, {
                    1.5649809912895796`, 1.5618787574885862`}], 
                   ArrowBox[{{-1.5018927968525602`, 
                    2.1999990040923216`}, {-1.4848929981375634`, 
                    2.1999324058619596`}}]}, 
                  {RGBColor[0, 0, 1], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-1.5000526650326216`, 0.5}, 1.8, {
                    1.5649809912895796`, 1.5753748131523961`}], 
                   ArrowBox[{{-1.490293988282107, 
                    2.299973546535637}, {-1.5082939116831435`, 
                    2.299981133749364}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.003], Arrowheads[0.025], 
                   CircleBox[{-1.5000526650326216`, 0.5}, 1.5, {
                    1.5649809912895796`, 1.5454233933013375`}], 
                   ArrowBox[{{-1.4769941730416876`, 
                    1.9998227581774801`}, {-1.4619973483441147`, 
                    1.9995171865876489`}}]}, 
                  {RGBColor[0, 0, 1], ArrowBox[{{0, 2.8}, {0, 2.8}}]}},
                 Frame->True,
                 FrameTicks->Automatic,
                 ImageSize->800,
                 PlotRange->{{-3, 3}, {-1, 3}}], 385->
                GraphicsBox[{InsetBox[
                   TagBox[GridBox[{
                    {"\<\"t\"\>", "19.2`"},
                    {"\<\"driveIntent\"\>", "0"},
                    {"\<\"motorTorque\"\>", 
                    RowBox[{"-", "0.20774047760044922`"}]},
                    {"\<\"vw\"\>", 
                    RowBox[{"-", "0.055403611736280056`"}]},
                    {"\<\"controllerState\"\>", 
                    RowBox[{"\[LeftAssociation]", 
                    RowBox[{
                    
                    RowBox[{"\<\"lastError\"\>", "\[Rule]", 
                    "0.014228040272153412`"}], ",", 
                    RowBox[{"\<\"lastInt\"\>", "\[Rule]", 
                    RowBox[{"-", "0.08397610968349876`"}]}], ",", 
                    RowBox[{"\<\"initialized\"\>", "\[Rule]", "True"}]}], 
                    "\[RightAssociation]"}]}
                    },
                    AutoDelete->False,
                    GridBoxItemSize->{
                    "Columns" -> {{Automatic}}, "Rows" -> {{Automatic}}}],
                    "Grid"], {-1, -0.5}], 
                  {GrayLevel[0], LineBox[{{-5, 0}, {5, 0}}]}, 
                  {GrayLevel[0.5], 
                   CircleBox[{-1.5028186270962482`, 0.5}, 0.5], 
                   DiskBox[{-1.5028186270962482`, 0.5}, 0.1]}, 
                  {GrayLevel[0.5], 
                   DiskBox[{-1.4993282069016995`, 1.9999959389834578`}, 0.1]}, 
                  {GrayLevel[0.5], Thickness[0.01], 
                   LineBox[{{-1.5028186270962482`, 
                    0.5}, {-1.4993282069016995`, 1.9999959389834578`}}]}, 
                  {RGBColor[0, 1, 0], 
                   DiskBox[{-1.5004916802998824`, 1.4999972926556386`}, 
                    0.05]}, 
                  {RGBColor[0, 0, 1], Arrowheads[0.01], 
                   CircleBox[{-1.5028186270962482`, 0.5}, 0.3333333333333333, 
                    NCache[{
                    Rational[-1, 2] 
                    Pi, -1.5770285411229101`}, {-1.5707963267948966`, \
-1.5770285411229101`}]], 
                   ArrowBox[{{-1.501562701510488, 
                    0.16666903269867928`}, {-1.5048960184243738`, 
                    0.1666731400616192}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-1.5028186270962482`, 
                    0.5}, {-1.4986638175442393`, 0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-1.5028186270962482`, 0.5}, {-1.519933245361393,
                     0.5}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.005], Arrowheads[0.04], 
                   ArrowBox[{{-1.5028186270962482`, 0.5}, {-1.515778435809384,
                     0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-1.5028186270962482`, 0.5}, 1.6, {
                    1.5684693778985797`, 1.5577727125233076`}], 
                   ArrowBox[{{-1.497980851633067, 
                    2.0999926862109612`}, {-1.4819814333218293`, 
                    2.099864310294972}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Dashing[{Small, Small}], 
                   Arrowheads[0.02], 
                   CircleBox[{-1.5028186270962482`, 0.5}, 1.7, {
                    1.5684693778985797`, 1.5609586431268105`}], 
                   ArrowBox[{{-1.5030945648592902`, 
                    2.199999977605397}, {-1.486094834618376, 
                    2.199917737646489}}]}, 
                  {RGBColor[0, 0, 1], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-1.5028186270962482`, 0.5}, 1.8, {
                    1.5684693778985797`, 1.5696234916630267`}], 
                   ArrowBox[{{-1.4827079422752982`, 
                    2.299887652148331}, {-1.5007075243428676`, 
                    2.29999876201212}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.003], Arrowheads[0.025], 
                   CircleBox[{-1.5028186270962482`, 0.5}, 1.5, {
                    1.5684693778985797`, 1.5499323719409541`}], 
                   ArrowBox[{{-1.4865230153694096`, 
                    1.9999114817343218`}, {-1.47152496530986, 
                    1.9996735333838491`}}]}, 
                  {RGBColor[0, 0, 1], ArrowBox[{{0, 2.8}, {0, 2.8}}]}},
                 Frame->True,
                 FrameTicks->Automatic,
                 ImageSize->800,
                 PlotRange->{{-3, 3}, {-1, 3}}], 386->
                GraphicsBox[{InsetBox[
                   TagBox[GridBox[{
                    {"\<\"t\"\>", "19.25`"},
                    {"\<\"driveIntent\"\>", "0"},
                    {"\<\"motorTorque\"\>", "1.20957071005704`"},
                    {"\<\"vw\"\>", 
                    RowBox[{"-", "0.05566899479921575`"}]},
                    {"\<\"controllerState\"\>", 
                    RowBox[{"\[LeftAssociation]", 
                    RowBox[{
                    
                    RowBox[{"\<\"lastError\"\>", "\[Rule]", 
                    "0.017310832521050163`"}], ",", 
                    RowBox[{"\<\"lastInt\"\>", "\[Rule]", 
                    RowBox[{"-", "0.08318422507615336`"}]}], ",", 
                    RowBox[{"\<\"initialized\"\>", "\[Rule]", "True"}]}], 
                    "\[RightAssociation]"}]}
                    },
                    AutoDelete->False,
                    
                    GridBoxItemSize->{
                    "Columns" -> {{Automatic}}, "Rows" -> {{Automatic}}}],
                    "Grid"], {-1, -0.5}], 
                  {GrayLevel[0], LineBox[{{-5, 0}, {5, 0}}]}, 
                  {GrayLevel[0.5], 
                   CircleBox[{-1.5055951417025777`, 0.5}, 0.5], 
                   DiskBox[{-1.5055951417025777`, 0.5}, 0.1]}, 
                  {GrayLevel[0.5], 
                   DiskBox[{-1.5065960125524482`, 1.9999996660858101`}, 0.1]}, 
                  {GrayLevel[0.5], Thickness[0.01], 
                   LineBox[{{-1.5055951417025777`, 
                    0.5}, {-1.5065960125524482`, 1.9999996660858101`}}]}, 
                  {RGBColor[0, 1, 0], 
                   DiskBox[{-1.5062623889358246`, 1.4999997773905402`}, 
                    0.05]}, 
                  {RGBColor[0, 0, 1], Arrowheads[0.01], 
                   CircleBox[{-1.5055951417025777`, 0.5}, 0.3333333333333333, 
                    NCache[{
                    Rational[-1, 2] 
                    Pi, -1.5345092054931853`}, {-1.5707963267948966`, \
-1.5345092054931853`}]], 
                   ArrowBox[{{-1.4968337770524678`, 
                    0.16678182882594939`}, {-1.4935020889413935`, 
                    0.16688610178194863`}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-1.5055951417025777`, 
                    0.5}, {-1.5297865559037185`, 0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-1.5055951417025777`, 
                    0.5}, {-1.5006875477351136`, 0.5}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.005], Arrowheads[0.04], 
                   ArrowBox[{{-1.5055951417025777`, 
                    0.5}, {-1.5248789619362544`, 0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-1.5055951417025777`, 0.5}, 1.6, {
                    1.5714635740776555`, 1.5745308209901185`}], 
                   ArrowBox[{{-1.4955703980047301`, 
                    2.0999685948523465`}, {-1.5115703185261624`, 
                    2.099988842855452}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Dashing[{Small, Small}], 
                   Arrowheads[0.02], 
                   CircleBox[{-1.5055951417025777`, 0.5}, 1.7, {
                    1.5714635740776555`, 1.5602877897567304`}], 
                   ArrowBox[{{-1.5047306287749571`, 
                    2.1999997801815736`}, {-1.4877309575303075`, 
                    2.199906135915704}}]}, 
                  {RGBColor[0, 0, 1], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-1.5055951417025777`, 0.5}, 1.8, {
                    1.5714635740776555`, 1.5647437367995607`}], 
                   ArrowBox[{{-1.5127004612583703`, 
                    2.2999859761770396`}, {-1.4947005462297456`, 
                    2.2999670296395665`}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.003], Arrowheads[0.025], 
                   CircleBox[{-1.5055951417025777`, 0.5}, 1.5, {
                    1.5714635740776555`, 1.5540056104868543`}], 
                   ArrowBox[{{-1.4954091455268146`, 
                    1.999965414761923}, {-1.480410250667746, 
                    1.999788558852068}}]}, 
                  {RGBColor[0, 0, 1], ArrowBox[{{0, 2.8}, {0, 2.8}}]}},
                 Frame->True,
                 FrameTicks->Automatic,
                 ImageSize->800,
                 PlotRange->{{-3, 3}, {-1, 3}}], 387->
                GraphicsBox[{InsetBox[
                   TagBox[GridBox[{
                    {"\<\"t\"\>", "19.3`"},
                    {"\<\"driveIntent\"\>", "0"},
                    {"\<\"motorTorque\"\>", "2.39594805367069`"},
                    {"\<\"vw\"\>", 
                    RowBox[{"-", "0.056028483769690565`"}]},
                    {"\<\"controllerState\"\>", 
                    RowBox[{"\[LeftAssociation]", 
                    RowBox[{
                    
                    RowBox[{"\<\"lastError\"\>", "\[Rule]", 
                    "0.019955935559055258`"}], ",", 
                    RowBox[{"\<\"lastInt\"\>", "\[Rule]", 
                    RowBox[{"-", "0.08224945963713529`"}]}], ",", 
                    RowBox[{"\<\"initialized\"\>", "\[Rule]", "True"}]}], 
                    "\[RightAssociation]"}]}
                    },
                    AutoDelete->False,
                    
                    GridBoxItemSize->{
                    "Columns" -> {{Automatic}}, "Rows" -> {{Automatic}}}],
                    "Grid"], {-1, -0.5}], 
                  {GrayLevel[0], LineBox[{{-5, 0}, {5, 0}}]}, 
                  {GrayLevel[0.5], CircleBox[{-1.508387406949571, 0.5}, 0.5], 
                   DiskBox[{-1.508387406949571, 0.5}, 0.1]}, 
                  {GrayLevel[0.5], 
                   DiskBox[{-1.513181386538085, 1.9999923392336725`}, 0.1]}, 
                  {GrayLevel[0.5], Thickness[0.01], 
                   LineBox[{{-1.508387406949571, 0.5}, {-1.513181386538085, 
                    1.9999923392336725`}}]}, 
                  {RGBColor[0, 1, 0], 
                   DiskBox[{-1.5115833933419136`, 1.4999948928224485`}, 
                    0.05]}, 
                  {RGBColor[0, 0, 1], Arrowheads[0.01], 
                   CircleBox[{-1.508387406949571, 0.5}, 0.3333333333333333, 
                    NCache[{
                    Rational[-1, 2] 
                    Pi, -1.4989178851847758`}, {-1.5707963267948966`, \
-1.4989178851847758`}]], 
                   ArrowBox[{{-1.4877744199450893`, 
                    0.16730461999320734`}, {-1.4844485522347357`, 
                    0.16752738105814924`}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-1.508387406949571, 0.5}, {-1.556306368022985, 
                    0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-1.508387406949571, 0.5}, {-1.484881087035284, 
                    0.5}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.005], Arrowheads[0.04], 
                   ArrowBox[{{-1.508387406949571, 0.5}, {-1.5328000481086979`,
                     0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-1.508387406949571, 0.5}, 1.6, {
                    1.5739923186280735`, 1.5886838436067294`}], 
                   ArrowBox[{{-1.5210073029941202`, 
                    2.099950229920864}, {-1.5370059076467466`, 
                    2.0997440362188717`}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Dashing[{Small, Small}], 
                   Arrowheads[0.02], 
                   CircleBox[{-1.508387406949571, 0.5}, 1.7, {
                    1.5739923186280735`, 1.5598442376569934`}], 
                   ArrowBox[{{-1.5067688556596648`, 
                    2.1999992294973905`}, {-1.4897692276233692`, 
                    2.1998980450011043`}}]}, 
                  {RGBColor[0, 0, 1], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-1.508387406949571, 0.5}, 1.8, {
                    1.5739923186280735`, 1.560681496107681}], 
                   ArrowBox[{{-1.508180711713037, 
                    2.299999988132522}, {-1.4901810221648857`, 
                    2.2999079219651968`}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.003], Arrowheads[0.025], 
                   CircleBox[{-1.508387406949571, 0.5}, 1.5, {
                    1.5739923186280735`, 1.5576557998136114`}], 
                   ArrowBox[{{-1.5036766242213229`, 
                    1.9999926028237895`}, {-1.4886771837277757`, 
                    1.9998704967764875`}}]}, 
                  {RGBColor[0, 0, 1], ArrowBox[{{0, 2.8}, {0, 2.8}}]}},
                 Frame->True,
                 FrameTicks->Automatic,
                 ImageSize->800,
                 PlotRange->{{-3, 3}, {-1, 3}}], 388->
                GraphicsBox[{InsetBox[
                   TagBox[GridBox[{
                    {"\<\"t\"\>", "19.35`"},
                    {"\<\"driveIntent\"\>", "0"},
                    {"\<\"motorTorque\"\>", "3.366817062105838`"},
                    {"\<\"vw\"\>", 
                    RowBox[{"-", "0.05646317269519534`"}]},
                    {"\<\"controllerState\"\>", 
                    RowBox[{"\[LeftAssociation]", 
                    RowBox[{
                    
                    RowBox[{"\<\"lastError\"\>", "\[Rule]", 
                    "0.02218752214688513`"}], ",", 
                    RowBox[{"\<\"lastInt\"\>", "\[Rule]", 
                    RowBox[{"-", "0.08119308714962332`"}]}], ",", 
                    RowBox[{"\<\"initialized\"\>", "\[Rule]", "True"}]}], 
                    "\[RightAssociation]"}]}
                    },
                    AutoDelete->False,
                    
                    GridBoxItemSize->{
                    "Columns" -> {{Automatic}}, "Rows" -> {{Automatic}}}],
                    "Grid"], {-1, -0.5}], 
                  {GrayLevel[0], LineBox[{{-5, 0}, {5, 0}}]}, 
                  {GrayLevel[0.5], 
                   CircleBox[{-1.5111996399356864`, 0.5}, 0.5], 
                   DiskBox[{-1.5111996399356864`, 0.5}, 0.1]}, 
                  {GrayLevel[0.5], 
                   DiskBox[{-1.5191333380496894`, 1.999979018664673}, 0.1]}, 
                  {GrayLevel[0.5], Thickness[0.01], 
                   LineBox[{{-1.5111996399356864`, 
                    0.5}, {-1.5191333380496894`, 1.999979018664673}}]}, 
                  {RGBColor[0, 1, 0], 
                   DiskBox[{-1.5164887720116884`, 1.4999860124431155`}, 
                    0.05]}, 
                  {RGBColor[0, 0, 1], Arrowheads[0.01], 
                   CircleBox[{-1.5111996399356864`, 0.5}, 0.3333333333333333, 
                    NCache[{
                    Rational[-1, 2] 
                    Pi, -1.4697918149317215`}, {-1.5707963267948966`, \
-1.4697918149317215`}]], 
                   ArrowBox[{{-1.4809066565958435`, 
                    0.16804601784060025`}, {-1.4775886867361814`, 
                    0.1683655401859876}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-1.5111996399356864`, 
                    0.5}, {-1.5785359811778032`, 0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-1.5111996399356864`, 0.5}, {-1.472298683767791,
                     0.5}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.005], Arrowheads[0.04], 
                   ArrowBox[{{-1.5111996399356864`, 
                    0.5}, {-1.5396350250099078`, 0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-1.5111996399356864`, 0.5}, 1.6, {
                    1.576085483531715, 1.6003989212222425`}], 
                   ArrowBox[{{-1.5425617823843198`, 
                    2.0996926004770513`}, {-1.5585568736809352`, 
                    2.0992990003160754`}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Dashing[{Small, Small}], 
                   Arrowheads[0.02], 
                   CircleBox[{-1.5111996399356864`, 0.5}, 1.7, {
                    1.576085483531715, 1.559606214615038}], 
                   ArrowBox[{{-1.5091764497075235`, 
                    2.1999987960881917`}, {-1.492176846237026, 
                    2.1998935652916325`}}]}, 
                  {RGBColor[0, 0, 1], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-1.5111996399356864`, 0.5}, 1.8, {
                    1.576085483531715, 1.5573809442977937`}], 
                   ArrowBox[{{-1.505051963392859, 
                    2.2999895016563636`}, {-1.4870526757543112`, 
                    2.2998380271904537`}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.003], Arrowheads[0.025], 
                   CircleBox[{-1.5111996399356864`, 0.5}, 1.5, {
                    1.576085483531715, 1.5608978652153378`}], 
                   ArrowBox[{{-1.5113519475660864`, 
                    1.9999999922674618`}, {-1.4963521900268424`, 
                    1.999926515943766}}]}, 
                  {RGBColor[0, 0, 1], ArrowBox[{{0, 2.8}, {0, 2.8}}]}},
                 Frame->True,
                 FrameTicks->Automatic,
                 ImageSize->800,
                 PlotRange->{{-3, 3}, {-1, 3}}], 389->
                GraphicsBox[{InsetBox[
                   TagBox[GridBox[{
                    {"\<\"t\"\>", "19.400000000000002`"},
                    {"\<\"driveIntent\"\>", "0"},
                    {"\<\"motorTorque\"\>", "4.137916301124495`"},
                    {"\<\"vw\"\>", 
                    RowBox[{"-", "0.056955623550794074`"}]},
                    {"\<\"controllerState\"\>", 
                    RowBox[{"\[LeftAssociation]", 
                    RowBox[{
                    
                    RowBox[{"\<\"lastError\"\>", "\[Rule]", 
                    "0.024030947856052115`"}], ",", 
                    RowBox[{"\<\"lastInt\"\>", "\[Rule]", 
                    RowBox[{"-", "0.0800351410091451`"}]}], ",", 
                    RowBox[{"\<\"initialized\"\>", "\[Rule]", "True"}]}], 
                    "\[RightAssociation]"}]}
                    },
                    AutoDelete->False,
                    
                    GridBoxItemSize->{
                    "Columns" -> {{Automatic}}, "Rows" -> {{Automatic}}}],
                    "Grid"], {-1, -0.5}], 
                  {GrayLevel[0], LineBox[{{-5, 0}, {5, 0}}]}, 
                  {GrayLevel[0.5], 
                   CircleBox[{-1.5140351498643156`, 0.5}, 0.5], 
                   DiskBox[{-1.5140351498643156`, 0.5}, 0.1]}, 
                  {GrayLevel[0.5], 
                   DiskBox[{-1.5245010881281016`, 1.9999634876010344`}, 0.1]}, 
                  {GrayLevel[0.5], Thickness[0.01], 
                   LineBox[{{-1.5140351498643156`, 
                    0.5}, {-1.5245010881281016`, 1.9999634876010344`}}]}, 
                  {RGBColor[0, 1, 0], 
                   DiskBox[{-1.521012442040173, 1.4999756584006896`}, 0.05]}, 
                  {RGBColor[0, 0, 1], Arrowheads[0.01], 
                   CircleBox[{-1.5140351498643156`, 0.5}, 0.3333333333333333, 
                    NCache[{
                    Rational[-1, 2] 
                    Pi, -1.4466588377611618`}, {-1.5707963267948966`, \
-1.4466588377611618`}]], 
                   ArrowBox[{{-1.4760718725566075`, 
                    0.1688355383088802}, {-1.4727621812815448`, 
                    0.16923172283987586`}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-1.5140351498643156`, 
                    0.5}, {-1.5967934758868054`, 0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-1.5140351498643156`, 
                    0.5}, {-1.4627185022867175`, 0.5}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.005], Arrowheads[0.04], 
                   ArrowBox[{{-1.5140351498643156`, 
                    0.5}, {-1.5454768283092073`, 0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-1.5140351498643156`, 0.5}, 1.6, {
                    1.5777736755841218`, 1.6098473610449207`}], 
                   ArrowBox[{{-1.560510266810609, 
                    2.0993248774107247`}, {-1.5765009252954199`, 
                    2.098780168409588}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Dashing[{Small, Small}], 
                   Arrowheads[0.02], 
                   CircleBox[{-1.5140351498643156`, 0.5}, 1.7, {
                    1.5777736755841218`, 1.5595523506704845`}], 
                   ArrowBox[{{-1.5119203909982384`, 
                    2.1999986846450614`}, {-1.4949207932205475`, 
                    2.1998925381829566`}}]}, 
                  {RGBColor[0, 0, 1], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-1.5140351498643156`, 0.5}, 1.8, {
                    1.5777736755841218`, 1.5547852516889857`}], 
                   ArrowBox[{{-1.5032152798330545`, 
                    2.2999674803763837`}, {-1.4852164460113642`, 
                    2.299769285855339}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.003], Arrowheads[0.025], 
                   CircleBox[{-1.5140351498643156`, 0.5}, 1.5, {
                    1.5777736755841218`, 1.5637486631660216`}], 
                   ArrowBox[{{-1.5184636479876508`, 
                    1.999993462787212}, {-1.5034637419343782`, 
                    1.999962747982222}}]}, 
                  {RGBColor[0, 0, 1], ArrowBox[{{0, 2.8}, {0, 2.8}}]}},
                 Frame->True,
                 FrameTicks->Automatic,
                 ImageSize->800,
                 PlotRange->{{-3, 3}, {-1, 3}}], 390->
                GraphicsBox[{InsetBox[
                   TagBox[GridBox[{
                    {"\<\"t\"\>", "19.45`"},
                    {"\<\"driveIntent\"\>", "0"},
                    {"\<\"motorTorque\"\>", "4.7251075272103105`"},
                    {"\<\"vw\"\>", 
                    RowBox[{"-", "0.057489866316042866`"}]},
                    {"\<\"controllerState\"\>", 
                    RowBox[{"\[LeftAssociation]", 
                    RowBox[{
                    
                    RowBox[{"\<\"lastError\"\>", "\[Rule]", 
                    "0.025512383984996133`"}], ",", 
                    RowBox[{"\<\"lastInt\"\>", "\[Rule]", 
                    RowBox[{"-", "0.07879436456284748`"}]}], ",", 
                    RowBox[{"\<\"initialized\"\>", "\[Rule]", "True"}]}], 
                    "\[RightAssociation]"}]}
                    },
                    AutoDelete->False,
                    
                    GridBoxItemSize->{
                    "Columns" -> {{Automatic}}, "Rows" -> {{Automatic}}}],
                    "Grid"], {-1, -0.5}], 
                  {GrayLevel[0], LineBox[{{-5, 0}, {5, 0}}]}, 
                  {GrayLevel[0.5], 
                   CircleBox[{-1.5168964115001597`, 0.5}, 0.5], 
                   DiskBox[{-1.5168964115001597`, 0.5}, 0.1]}, 
                  {GrayLevel[0.5], 
                   DiskBox[{-1.5293335835879813`, 1.999948438030607}, 0.1]}, 
                  {GrayLevel[0.5], Thickness[0.01], 
                   LineBox[{{-1.5168964115001597`, 
                    0.5}, {-1.5293335835879813`, 1.999948438030607}}]}, 
                  {RGBColor[0, 1, 0], 
                   DiskBox[{-1.5251878595587074`, 1.4999656253537381`}, 
                    0.05]}, 
                  {RGBColor[0, 0, 1], Arrowheads[0.01], 
                   CircleBox[{-1.5168964115001597`, 0.5}, 0.3333333333333333, 
                    NCache[{
                    Rational[-1, 2] 
                    Pi, -1.4290431009785873`}, {-1.5707963267948966`, \
-1.4290431009785873`}]], 
                   ArrowBox[{{-1.4731056200318553`, 
                    0.16955563600858525`}, {-1.4698034209870539`, 
                    0.17001005870535446`}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-1.5168964115001597`, 0.5}, {-1.611398562044366,
                     0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-1.5168964115001597`, 
                    0.5}, {-1.4559150109587737`, 0.5}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.005], Arrowheads[0.04], 
                   ArrowBox[{{-1.5168964115001597`, 0.5}, {-1.55041716150298, 
                    0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-1.5168964115001597`, 0.5}, 1.6, {
                    1.5790878698599486`, 1.6172025553771467`}], 
                   ArrowBox[{{-1.57513351053618, 
                    2.098939786325886}, {-1.5911197300801203`, 
                    2.0982774787185665`}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Dashing[{Small, Small}], 
                   Arrowheads[0.02], 
                   CircleBox[{-1.5168964115001597`, 0.5}, 1.7, {
                    1.5790878698599486`, 1.5596618645304094`}], 
                   ArrowBox[{{-1.5149678260642148`, 
                    2.199998906046182}, {-1.4979682167639554`, 
                    2.19989462127628}}]}, 
                  {RGBColor[0, 0, 1], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-1.5168964115001597`, 0.5}, 1.8, {
                    1.5790878698599486`, 1.552837272486558}], 
                   ArrowBox[{{-1.50257026499825, 
                    2.2999429884100238`}, {-1.4845718514045037`, 
                    2.2997097329332368`}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.003], Arrowheads[0.025], 
                   CircleBox[{-1.5168964115001597`, 0.5}, 1.5, {
                    1.5790878698599486`, 1.5662266781900804`}], 
                   ArrowBox[{{-1.5250418985594716`, 
                    1.9999778835171425`}, {-1.5100419624484052`, 
                    1.9999843387609741`}}]}, 
                  {RGBColor[0, 0, 1], ArrowBox[{{0, 2.8}, {0, 2.8}}]}},
                 Frame->True,
                 FrameTicks->Automatic,
                 ImageSize->800,
                 PlotRange->{{-3, 3}, {-1, 3}}], 391->
                GraphicsBox[{InsetBox[
                   TagBox[GridBox[{
                    {"\<\"t\"\>", "19.5`"},
                    {"\<\"driveIntent\"\>", "0"},
                    {"\<\"motorTorque\"\>", "5.1442061601829`"},
                    {"\<\"vw\"\>", 
                    RowBox[{"-", "0.05805138562665568`"}]},
                    {"\<\"controllerState\"\>", 
                    RowBox[{"\[LeftAssociation]", 
                    RowBox[{
                    
                    RowBox[{"\<\"lastError\"\>", "\[Rule]", 
                    "0.026658479941705084`"}], ",", 
                    RowBox[{"\<\"lastInt\"\>", "\[Rule]", 
                    RowBox[{"-", "0.07748817905826691`"}]}], ",", 
                    RowBox[{"\<\"initialized\"\>", "\[Rule]", "True"}]}], 
                    "\[RightAssociation]"}]}
                    },
                    AutoDelete->False,
                    
                    GridBoxItemSize->{
                    "Columns" -> {{Automatic}}, "Rows" -> {{Automatic}}}],
                    "Grid"], {-1, -0.5}], 
                  {GrayLevel[0], LineBox[{{-5, 0}, {5, 0}}]}, 
                  {GrayLevel[0.5], 
                   CircleBox[{-1.5197851382827308`, 0.5}, 0.5], 
                   DiskBox[{-1.5197851382827308`, 0.5}, 0.1]}, 
                  {GrayLevel[0.5], 
                   DiskBox[{-1.5336790644654328`, 1.9999356515581692`}, 0.1]}, 
                  {GrayLevel[0.5], Thickness[0.01], 
                   LineBox[{{-1.5197851382827308`, 
                    0.5}, {-1.5336790644654328`, 1.9999356515581692`}}]}, 
                  {RGBColor[0, 1, 0], 
                   DiskBox[{-1.5290477557378654`, 1.4999571010387793`}, 
                    0.05]}, 
                  {RGBColor[0, 0, 1], Arrowheads[0.01], 
                   CircleBox[{-1.5197851382827308`, 0.5}, 0.3333333333333333, 
                    NCache[{
                    Rational[-1, 2] 
                    Pi, -1.4164701419894095`}, {-1.5707963267948966`, \
-1.4164701419894095`}]], 
                   ArrowBox[{{-1.4718432540011248`, 
                    0.17013231919049693`}, {-1.46854702924494, 
                    0.17062822328963495`}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-1.5197851382827308`, 
                    0.5}, {-1.6226692614863887`, 0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-1.5197851382827308`, 
                    0.5}, {-1.4516616252362518`, 0.5}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.005], Arrowheads[0.04], 
                   ArrowBox[{{-1.5197851382827308`, 0.5}, {-1.55454574843991, 
                    0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-1.5197851382827308`, 0.5}, 1.6, {
                    1.5800590767045266`, 1.6226380989544003`}], 
                   ArrowBox[{{-1.5867124410988112`, 
                    2.0985996172080625`}, {-1.6026948245017, 
                    2.097850426019617}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Dashing[{Small, Small}], 
                   Arrowheads[0.02], 
                   CircleBox[{-1.5197851382827308`, 0.5}, 1.7, {
                    1.5800590767045266`, 1.5599147209194688`}], 
                   ArrowBox[{{-1.5182864084886463`, 
                    2.199999339355461}, {-1.5012867733627633`, 
                    2.19989935304867}}]}, 
                  {RGBColor[0, 0, 1], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-1.5197851382827308`, 0.5}, 1.8, {
                    1.5800590767045266`, 1.5514801535923992`}], 
                   ArrowBox[{{-1.5030162690854116`, 
                    2.2999218885901254`}, {-1.4850181886214648`, 
                    2.299664207348485}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.003], Arrowheads[0.025], 
                   CircleBox[{-1.5197851382827308`, 0.5}, 1.5, {
                    1.5800590767045266`, 1.5683517228134403`}], 
                   ArrowBox[{{-1.5311181244877805`, 
                    1.9999571871969133`}, {-1.5161182359628376`, 
                    1.9999955179357625`}}]}, 
                  {RGBColor[0, 0, 1], ArrowBox[{{0, 2.8}, {0, 2.8}}]}},
                 Frame->True,
                 FrameTicks->Automatic,
                 ImageSize->800,
                 PlotRange->{{-3, 3}, {-1, 3}}], 392->
                GraphicsBox[{InsetBox[
                   TagBox[GridBox[{
                    {"\<\"t\"\>", "19.55`"},
                    {"\<\"driveIntent\"\>", "0"},
                    {"\<\"motorTorque\"\>", "5.410831835089057`"},
                    {"\<\"vw\"\>", 
                    RowBox[{"-", "0.05862709516059325`"}]},
                    {"\<\"controllerState\"\>", 
                    RowBox[{"\[LeftAssociation]", 
                    RowBox[{
                    
                    RowBox[{"\<\"lastError\"\>", "\[Rule]", 
                    "0.027496056340102812`"}], ",", 
                    RowBox[{"\<\"lastInt\"\>", "\[Rule]", 
                    RowBox[{"-", "0.07613266769348288`"}]}], ",", 
                    RowBox[{"\<\"initialized\"\>", "\[Rule]", "True"}]}], 
                    "\[RightAssociation]"}]}
                    },
                    AutoDelete->False,
                    
                    GridBoxItemSize->{
                    "Columns" -> {{Automatic}}, "Rows" -> {{Automatic}}}],
                    "Grid"], {-1, -0.5}], 
                  {GrayLevel[0], LineBox[{{-5, 0}, {5, 0}}]}, 
                  {GrayLevel[0.5], 
                   CircleBox[{-1.5227023544544225`, 0.5}, 0.5], 
                   DiskBox[{-1.5227023544544225`, 0.5}, 0.1]}, 
                  {GrayLevel[0.5], 
                   DiskBox[{-1.5375846847935974`, 1.999926170264282}, 0.1]}, 
                  {GrayLevel[0.5], Thickness[0.01], 
                   LineBox[{{-1.5227023544544225`, 
                    0.5}, {-1.5375846847935974`, 1.999926170264282}}]}, 
                  {RGBColor[0, 1, 0], 
                   DiskBox[{-1.5326239080138724`, 1.499950780176188}, 0.05]}, 
                  {RGBColor[0, 0, 1], Arrowheads[0.01], 
                   CircleBox[{-1.5227023544544225`, 0.5}, 0.3333333333333333, 
                    NCache[{
                    Rational[-1, 2] 
                    Pi, -1.4084713717422248`}, {-1.5707963267948966`, \
-1.4084713717422248`}]], 
                   ArrowBox[{{-1.4721234961781924`, 
                    0.1705263436834682}, {-1.4688313434488678`, 
                    0.1710485973819985}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-1.5227023544544225`, 
                    0.5}, {-1.6309189911562036`, 0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-1.5227023544544225`, 
                    0.5}, {-1.4497330437574871`, 0.5}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.005], Arrowheads[0.04], 
                   ArrowBox[{{-1.5227023544544225`, 
                    0.5}, {-1.5579496804592683`, 0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-1.5227023544544225`, 0.5}, 1.6, {
                    1.5807180431365908`, 1.6263261071430501`}], 
                   ArrowBox[{{-1.5955248372321584`, 
                    2.098341917739157}, {-1.6115043489267662`, 
                    2.0975337886184864`}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Dashing[{Small, Small}], 
                   Arrowheads[0.02], 
                   CircleBox[{-1.5227023544544225`, 0.5}, 1.7, {
                    1.5807180431365908`, 1.5602917593463472`}], 
                   ArrowBox[{{-1.5218445898282846`, 
                    2.199999783599941}, {-1.5048449182120398`, 
                    2.1999062068157906`}}]}, 
                  {RGBColor[0, 0, 1], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-1.5227023544544225`, 0.5}, 1.8, {
                    1.5807180431365908`, 1.5506578662749848`}], 
                   ArrowBox[{{-1.5044534381517594`, 
                    2.299907491248864}, {-1.4864555756605642`, 
                    2.299635010502705}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.003], Arrowheads[0.025], 
                   CircleBox[{-1.5227023544544225`, 0.5}, 1.5, {
                    1.5807180431365908`, 1.5701446442139444`}], 
                   ArrowBox[{{-1.5367246263440946`, 
                    1.9999344571983977`}, {-1.521724830652185, 
                    1.9999996814823715`}}]}, 
                  {RGBColor[0, 0, 1], ArrowBox[{{0, 2.8}, {0, 2.8}}]}},
                 Frame->True,
                 FrameTicks->Automatic,
                 ImageSize->800,
                 PlotRange->{{-3, 3}, {-1, 3}}], 393->
                GraphicsBox[{InsetBox[
                   TagBox[GridBox[{
                    {"\<\"t\"\>", "19.6`"},
                    {"\<\"driveIntent\"\>", "0"},
                    {"\<\"motorTorque\"\>", "5.540278427268337`"},
                    {"\<\"vw\"\>", 
                    RowBox[{"-", "0.059205301088057036`"}]},
                    {"\<\"controllerState\"\>", 
                    RowBox[{"\[LeftAssociation]", 
                    RowBox[{
                    
                    RowBox[{"\<\"lastError\"\>", "\[Rule]", 
                    "0.02805182937019763`"}], ",", 
                    RowBox[{"\<\"lastInt\"\>", "\[Rule]", 
                    RowBox[{"-", "0.07474257421596253`"}]}], ",", 
                    RowBox[{"\<\"initialized\"\>", "\[Rule]", "True"}]}], 
                    "\[RightAssociation]"}]}
                    },
                    AutoDelete->False,
                    
                    GridBoxItemSize->{
                    "Columns" -> {{Automatic}}, "Rows" -> {{Automatic}}}],
                    "Grid"], {-1, -0.5}], 
                  {GrayLevel[0], LineBox[{{-5, 0}, {5, 0}}]}, 
                  {GrayLevel[0.5], 
                   CircleBox[{-1.5256484656232252`, 0.5}, 0.5], 
                   DiskBox[{-1.5256484656232252`, 0.5}, 0.1]}, 
                  {GrayLevel[0.5], 
                   DiskBox[{-1.54109618557242, 1.999920453873595}, 0.1]}, 
                  {GrayLevel[0.5], Thickness[0.01], 
                   LineBox[{{-1.5256484656232252`, 0.5}, {-1.54109618557242, 
                    1.999920453873595}}]}, 
                  {RGBColor[0, 1, 0], 
                   DiskBox[{-1.535946945589355, 1.4999469692490632`}, 0.05]}, 
                  {RGBColor[0, 0, 1], Arrowheads[0.01], 
                   CircleBox[{-1.5256484656232252`, 0.5}, 0.3333333333333333, 
                    NCache[{
                    Rational[-1, 2] 
                    Pi, -1.4045879739768465`}, {-1.5707963267948966`, \
-1.4045879739768465`}]], 
                   ArrowBox[{{-1.4737905146860133`, 
                    0.17072524537143025`}, {-1.4705004148945184`, 
                    0.17126027983838782`}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-1.5256484656232252`, 0.5}, {-1.636454034168592,
                     0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-1.5256484656232252`, 
                    0.5}, {-1.4499072910277206`, 0.5}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.005], Arrowheads[0.04], 
                   ArrowBox[{{-1.5256484656232252`, 
                    0.5}, {-1.5607128595730875`, 0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-1.5256484656232252`, 0.5}, 1.6, {
                    1.5810949888102634`, 1.6284357334476918`}], 
                   ArrowBox[{{-1.6018426880043004`, 
                    2.098184732900344}, {-1.6178204592911418`, 
                    2.09734289480477}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Dashing[{Small, Small}], 
                   Arrowheads[0.02], 
                   CircleBox[{-1.5256484656232252`, 0.5}, 1.7, {
                    1.5810949888102634`, 1.5607747956844904`}], 
                   ArrowBox[{{-1.5256118627355375`, 
                    2.1999999996059496`}, {-1.5086121479015238`, 
                    2.199914634291524}}]}, 
                  {RGBColor[0, 0, 1], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-1.5256484656232252`, 0.5}, 1.8, {
                    1.5810949888102634`, 1.5503156642143283`}], 
                   ArrowBox[{{-1.5067836183485814`, 
                    2.2999011410456145`}, {-1.4887858501546523`, 
                    2.2996225014098974`}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.003], Arrowheads[0.025], 
                   CircleBox[{-1.5256484656232252`, 0.5}, 1.5, {
                    1.5810949888102634`, 1.5716270413658553`}], 
                   ArrowBox[{{-1.541894219858967, 
                    1.9999120225764282`}, {-1.526894537336347, 
                    1.999999482435006}}]}, 
                  {RGBColor[0, 0, 1], ArrowBox[{{0, 2.8}, {0, 2.8}}]}},
                 Frame->True,
                 FrameTicks->Automatic,
                 ImageSize->800,
                 PlotRange->{{-3, 3}, {-1, 3}}], 394->
                GraphicsBox[{InsetBox[
                   TagBox[GridBox[{
                    {"\<\"t\"\>", "19.650000000000002`"},
                    {"\<\"driveIntent\"\>", "0"},
                    {"\<\"motorTorque\"\>", "5.547402684613099`"},
                    {"\<\"vw\"\>", 
                    RowBox[{"-", "0.05977565601288455`"}]},
                    {"\<\"controllerState\"\>", 
                    RowBox[{"\[LeftAssociation]", 
                    RowBox[{
                    RowBox[{"\<\"lastError\"\>", "\[Rule]", 
                    "0.028352166388958294`"}], ",", 
                    RowBox[{"\<\"lastInt\"\>", "\[Rule]", 
                    RowBox[{"-", "0.07333131450538656`"}]}], ",", 
                    RowBox[{"\<\"initialized\"\>", "\[Rule]", "True"}]}], 
                    "\[RightAssociation]"}]}
                    },
                    AutoDelete->False,
                    
                    GridBoxItemSize->{
                    "Columns" -> {{Automatic}}, "Rows" -> {{Automatic}}}],
                    "Grid"], {-1, -0.5}], 
                  {GrayLevel[0], LineBox[{{-5, 0}, {5, 0}}]}, 
                  {GrayLevel[0.5], 
                   CircleBox[{-1.5286233272494032`, 0.5}, 0.5], 
                   DiskBox[{-1.5286233272494032`, 0.5}, 0.1]}, 
                  {GrayLevel[0.5], 
                   DiskBox[{-1.5442576180341383`, 1.9999185207709314`}, 0.1]}, 
                  {GrayLevel[0.5], Thickness[0.01], 
                   LineBox[{{-1.5286233272494032`, 
                    0.5}, {-1.5442576180341383`, 1.9999185207709314`}}]}, 
                  {RGBColor[0, 1, 0], 
                   DiskBox[{-1.5390461877725599`, 1.4999456805139542`}, 
                    0.05]}, 
                  {RGBColor[0, 0, 1], Arrowheads[0.01], 
                   CircleBox[{-1.5286233272494032`, 0.5}, 0.3333333333333333, 
                    NCache[{
                    Rational[-1, 2] 
                    Pi, -1.4043742462565036`}, {-1.5707963267948966`, \
-1.4043742462565036`}]], 
                   ArrowBox[{{-1.4766950023544774`, 
                    0.1707363363735407}, {-1.4734050169898238`, 
                    0.171272074013801}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-1.5286233272494032`, 
                    0.5}, {-1.6395713809416652`, 0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-1.5286233272494032`, 
                    0.5}, {-1.4519674825196238`, 0.5}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.005], Arrowheads[0.04], 
                   ArrowBox[{{-1.5286233272494032`, 
                    0.5}, {-1.5629155362118858`, 0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-1.5286233272494032`, 0.5}, 1.6, {
                    1.5812193760436295`, 1.6291318815824176`}], 
                   ArrowBox[{{-1.6059301043986651`, 
                    2.098131303180935}, {-1.6219072857699428`, 
                    2.097278342394568}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Dashing[{Small, Small}], 
                   Arrowheads[0.02], 
                   CircleBox[{-1.5286233272494032`, 0.5}, 1.7, {
                    1.5812193760436295`, 1.5613466985963924`}], 
                   ArrowBox[{{-1.5295589592647107`, 
                    2.199999742527255}, {-1.512559198390101, 
                    2.1999240994126743`}}]}, 
                  {RGBColor[0, 0, 1], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-1.5286233272494032`, 0.5}, 1.8, {
                    1.5812193760436295`, 1.5504004722402234`}], 
                   ArrowBox[{{-1.5099111261049967`, 
                    2.299902734463263}, {-1.4919133343449136`, 
                    2.299625621183737}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.003], Arrowheads[0.025], 
                   CircleBox[{-1.5286233272494032`, 0.5}, 1.5, {
                    1.5812193760436295`, 1.5728209962807143`}], 
                   ArrowBox[{{-1.5466598968114866`, 
                    1.9998915567995015`}, {-1.531660329403205, 
                    1.9999969255361552`}}]}, 
                  {RGBColor[0, 0, 1], ArrowBox[{{0, 2.8}, {0, 2.8}}]}},
                 Frame->True,
                 FrameTicks->Automatic,
                 ImageSize->800,
                 PlotRange->{{-3, 3}, {-1, 3}}], 395->
                GraphicsBox[{InsetBox[
                   TagBox[GridBox[{
                    {"\<\"t\"\>", "19.7`"},
                    {"\<\"driveIntent\"\>", "0"},
                    {"\<\"motorTorque\"\>", "5.446530418399915`"},
                    {"\<\"vw\"\>", 
                    RowBox[{"-", "0.06032910486818206`"}]},
                    {"\<\"controllerState\"\>", 
                    RowBox[{"\[LeftAssociation]", 
                    RowBox[{
                    
                    RowBox[{"\<\"lastError\"\>", "\[Rule]", 
                    "0.028422872155242966`"}], ",", 
                    RowBox[{"\<\"lastInt\"\>", "\[Rule]", 
                    RowBox[{"-", "0.07191099959214031`"}]}], ",", 
                    RowBox[{"\<\"initialized\"\>", "\[Rule]", "True"}]}], 
                    "\[RightAssociation]"}]}
                    },
                    AutoDelete->False,
                    
                    GridBoxItemSize->{
                    "Columns" -> {{Automatic}}, "Rows" -> {{Automatic}}}],
                    "Grid"], {-1, -0.5}], 
                  {GrayLevel[0], LineBox[{{-5, 0}, {5, 0}}]}, 
                  {GrayLevel[0.5], CircleBox[{-1.531626310617987, 0.5}, 0.5], 
                   DiskBox[{-1.531626310617987, 0.5}, 0.1]}, 
                  {GrayLevel[0.5], 
                   DiskBox[{-1.5471111147136634`, 1.999920071484517}, 0.1]}, 
                  {GrayLevel[0.5], Thickness[0.01], 
                   LineBox[{{-1.531626310617987, 0.5}, {-1.5471111147136634`, 
                    1.999920071484517}}]}, 
                  {RGBColor[0, 1, 0], 
                   DiskBox[{-1.541949513348438, 1.4999467143230114`}, 0.05]}, 
                  {RGBColor[0, 0, 1], Arrowheads[0.01], 
                   CircleBox[{-1.531626310617987, 0.5}, 0.3333333333333333, 
                    NCache[{
                    Rational[-1, 2] 
                    Pi, -1.407400414242899}, {-1.5707963267948966`, \
-1.407400414242899}]], 
                   ArrowBox[{{-1.4806946291318521`, 
                    0.1705807004255731}, {-1.477403037601903, 
                    0.17110647957958447`}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-1.531626310617987, 0.5}, {-1.6405569189859854`,
                     0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-1.531626310617987, 0.5}, {-1.4557033293985926`,
                     0.5}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.005], Arrowheads[0.04], 
                   ArrowBox[{{-1.531626310617987, 0.5}, {-1.5646339377665908`,
                     0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-1.531626310617987, 0.5}, 1.6, {
                    1.5811197128888712`, 1.6285741047903912`}], 
                   ArrowBox[{{-1.6080416752362985`, 
                    2.098174174503596}, {-1.6240193298839125`, 
                    2.097330125550422}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Dashing[{Small, Small}], 
                   Arrowheads[0.02], 
                   CircleBox[{-1.531626310617987, 0.5}, 1.7, {
                    1.5811197128888712`, 1.561991443212855}], 
                   ArrowBox[{{-1.5336580080448696`, 
                    2.19999878594238}, {-1.5166582019331352`, 
                    2.1999341033470676`}}]}, 
                  {RGBColor[0, 0, 1], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-1.531626310617987, 0.5}, 1.8, {
                    1.5811197128888712`, 1.550861210564427}], 
                   ArrowBox[{{-1.5137433955999577`, 
                    2.2999111648496595`}, {-1.4957454780734551`, 
                    2.299642343871668}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.003], Arrowheads[0.025], 
                   CircleBox[{-1.531626310617987, 0.5}, 1.5, {
                    1.5811197128888712`, 1.5737488224950082`}], 
                   ArrowBox[{{-1.5510545109219058`, 
                    1.9998741764004575`}, {-1.5360550477337607`, 
                    1.999993462081605}}]}, 
                  {RGBColor[0, 0, 1], ArrowBox[{{0, 2.8}, {0, 2.8}}]}},
                 Frame->True,
                 FrameTicks->Automatic,
                 ImageSize->800,
                 PlotRange->{{-3, 3}, {-1, 3}}], 396->
                GraphicsBox[{InsetBox[
                   TagBox[GridBox[{
                    {"\<\"t\"\>", "19.75`"},
                    {"\<\"driveIntent\"\>", "0"},
                    {"\<\"motorTorque\"\>", "5.251379087215787`"},
                    {"\<\"vw\"\>", 
                    RowBox[{"-", "0.060857824212672176`"}]},
                    {"\<\"controllerState\"\>", 
                    RowBox[{"\[LeftAssociation]", 
                    RowBox[{
                    
                    RowBox[{"\<\"lastError\"\>", "\[Rule]", 
                    "0.028289004703656147`"}], ",", 
                    RowBox[{"\<\"lastInt\"\>", "\[Rule]", 
                    RowBox[{"-", "0.07049246860331267`"}]}], ",", 
                    RowBox[{"\<\"initialized\"\>", "\[Rule]", "True"}]}], 
                    "\[RightAssociation]"}]}
                    },
                    AutoDelete->False,
                    
                    GridBoxItemSize->{
                    "Columns" -> {{Automatic}}, "Rows" -> {{Automatic}}}],
                    "Grid"], {-1, -0.5}], 
                  {GrayLevel[0], LineBox[{{-5, 0}, {5, 0}}]}, 
                  {GrayLevel[0.5], 
                   CircleBox[{-1.5346563659318255`, 0.5}, 0.5], 
                   DiskBox[{-1.5346563659318255`, 0.5}, 0.1]}, 
                  {GrayLevel[0.5], 
                   DiskBox[{-1.54969670540162, 1.9999245941675312`}, 0.1]}, 
                  {GrayLevel[0.5], Thickness[0.01], 
                   LineBox[{{-1.5346563659318255`, 0.5}, {-1.54969670540162, 
                    1.9999245941675312`}}]}, 
                  {RGBColor[0, 1, 0], 
                   DiskBox[{-1.5446832589116886`, 1.4999497294450208`}, 
                    0.05]}, 
                  {RGBColor[0, 0, 1], Arrowheads[0.01], 
                   CircleBox[{-1.5346563659318255`, 0.5}, 0.3333333333333333, 
                    NCache[{
                    Rational[-1, 2] 
                    Pi, -1.413254954178423}, {-1.5707963267948966`, \
-1.413254954178423}]], 
                   ArrowBox[{{-1.4856541447287186`, 
                    0.17028816607790243`}, {-1.4823595314318379`, 
                    0.1707946655772541}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-1.5346563659318255`, 
                    0.5}, {-1.6396839476761413`, 0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-1.5346563659318255`, 
                    0.5}, {-1.4609124007375218`, 0.5}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.005], Arrowheads[0.04], 
                   ArrowBox[{{-1.5346563659318255`, 
                    0.5}, {-1.5659399824818376`, 0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-1.5346563659318255`, 0.5}, 1.6, {
                    1.5808233877972966`, 1.6269156831290028`}], 
                   ArrowBox[{{-1.6084211800104675`, 
                    2.0982987055628692`}, {-1.624400212474342, 
                    2.0974811554468342`}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Dashing[{Small, Small}], 
                   Arrowheads[0.02], 
                   CircleBox[{-1.5346563659318255`, 0.5}, 1.7, {
                    1.5808233877972966`, 1.5626941446540328`}], 
                   ArrowBox[{{-1.5378826543556623`, 
                    2.199996938545186}, {-1.5208828069885398`, 
                    2.199944201753115}}]}, 
                  {RGBColor[0, 0, 1], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-1.5346563659318255`, 0.5}, 1.8, {
                    1.5808233877972966`, 1.5516490595349866`}], 
                   ArrowBox[{{-1.518191514475438, 
                    2.299924695276588}, {-1.500193390744334, 
                    2.299670054021355}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.003], Arrowheads[0.025], 
                   CircleBox[{-1.5346563659318255`, 0.5}, 1.5, {
                    1.5808233877972966`, 1.5744328333406459`}], 
                   ArrowBox[{{-1.5551104918145495`, 
                    1.9998605364280952`}, {-1.5401111137280032`, 
                    1.999990081876037}}]}, 
                  {RGBColor[0, 0, 1], ArrowBox[{{0, 2.8}, {0, 2.8}}]}},
                 Frame->True,
                 FrameTicks->Automatic,
                 ImageSize->800,
                 PlotRange->{{-3, 3}, {-1, 3}}], 397->
                GraphicsBox[{InsetBox[
                   TagBox[GridBox[{
                    {"\<\"t\"\>", "19.8`"},
                    {"\<\"driveIntent\"\>", "0"},
                    {"\<\"motorTorque\"\>", "4.974995545144239`"},
                    {"\<\"vw\"\>", 
                    RowBox[{"-", "0.061355156317680085`"}]},
                    {"\<\"controllerState\"\>", 
                    RowBox[{"\[LeftAssociation]", 
                    RowBox[{
                    
                    RowBox[{"\<\"lastError\"\>", "\[Rule]", 
                    "0.027974719517586802`"}], ",", 
                    RowBox[{"\<\"lastInt\"\>", "\[Rule]", 
                    RowBox[{"-", "0.06908533018721538`"}]}], ",", 
                    RowBox[{"\<\"initialized\"\>", "\[Rule]", "True"}]}], 
                    "\[RightAssociation]"}]}
                    },
                    AutoDelete->False,
                    
                    GridBoxItemSize->{
                    "Columns" -> {{Automatic}}, "Rows" -> {{Automatic}}}],
                    "Grid"], {-1, -0.5}], 
                  {GrayLevel[0], LineBox[{{-5, 0}, {5, 0}}]}, 
                  {GrayLevel[0.5], CircleBox[{-1.537712082230994, 0.5}, 0.5], 
                   DiskBox[{-1.537712082230994, 0.5}, 0.1]}, 
                  {GrayLevel[0.5], 
                   DiskBox[{-1.5520521747698777`, 1.9999314523490652`}, 0.1]}, 
                  {GrayLevel[0.5], Thickness[0.01], 
                   LineBox[{{-1.537712082230994, 0.5}, {-1.5520521747698777`, 
                    1.9999314523490652`}}]}, 
                  {RGBColor[0, 1, 0], 
                   DiskBox[{-1.547272143923583, 1.4999543015660435`}, 0.05]}, 
                  {RGBColor[0, 0, 1], Arrowheads[0.01], 
                   CircleBox[{-1.537712082230994, 0.5}, 0.3333333333333333, 
                    NCache[{
                    Rational[-1, 2] 
                    Pi, -1.4215464604405694`}, {-1.5707963267948966`, \
-1.4215464604405694`}]], 
                   ArrowBox[{{-1.4914453218582713`, 
                    0.1698932021364538}, {-1.4881466222159012`, 
                    0.170372367231442}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-1.537712082230994, 0.5}, {-1.6372119931338787`,
                     0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-1.537712082230994, 0.5}, {-1.4674011612285927`,
                     0.5}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.005], Arrowheads[0.04], 
                   ArrowBox[{{-1.537712082230994, 0.5}, {-1.5669010721314776`,
                     0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-1.537712082230994, 0.5}, 1.6, {
                    1.5803565341167636`, 1.624302868021902}], 
                   ArrowBox[{{-1.6073005902688262`, 
                    2.0984859835322514`}, {-1.6232817042947434`, 
                    2.097710186416757}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Dashing[{Small, Small}], 
                   Arrowheads[0.02], 
                   CircleBox[{-1.537712082230994, 0.5}, 1.7, {
                    1.5803565341167636`, 1.5634410739948446`}], 
                   ArrowBox[{{-1.5422081472294726`, 
                    2.1999940545188768`}, {-1.5252082652138332`, 
                    2.1999540154251234`}}]}, 
                  {RGBColor[0, 0, 1], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-1.537712082230994, 0.5}, 1.8, {
                    1.5803565341167636`, 1.5527176699770733`}], 
                   ArrowBox[{{-1.52317065813372, 
                    2.2999412620930784`}, {-1.5051722725666457`, 
                    2.299705853962532}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.003], Arrowheads[0.025], 
                   CircleBox[{-1.537712082230994, 0.5}, 1.5, {
                    1.5803565341167636`, 1.5748951318431081`}], 
                   ArrowBox[{{-1.5588595891831858`, 
                    1.9998509202416441`}, {-1.5438602725881367`, 
                    1.999987399865523}}]}, 
                  {RGBColor[0, 0, 1], ArrowBox[{{0, 2.8}, {0, 2.8}}]}},
                 Frame->True,
                 FrameTicks->Automatic,
                 ImageSize->800,
                 PlotRange->{{-3, 3}, {-1, 3}}], 398->
                GraphicsBox[{InsetBox[
                   TagBox[GridBox[{
                    {"\<\"t\"\>", "19.85`"},
                    {"\<\"driveIntent\"\>", "0"},
                    {"\<\"motorTorque\"\>", "4.629707703685796`"},
                    {"\<\"vw\"\>", 
                    RowBox[{"-", "0.06181553934897903`"}]},
                    {"\<\"controllerState\"\>", 
                    RowBox[{"\[LeftAssociation]", 
                    RowBox[{
                    
                    RowBox[{"\<\"lastError\"\>", "\[Rule]", 
                    "0.02750314041444772`"}], ",", 
                    RowBox[{"\<\"lastInt\"\>", "\[Rule]", 
                    RowBox[{"-", "0.06769801104107044`"}]}], ",", 
                    RowBox[{"\<\"initialized\"\>", "\[Rule]", "True"}]}], 
                    "\[RightAssociation]"}]}
                    },
                    AutoDelete->False,
                    
                    GridBoxItemSize->{
                    "Columns" -> {{Automatic}}, "Rows" -> {{Automatic}}}],
                    "Grid"], {-1, -0.5}], 
                  {GrayLevel[0], LineBox[{{-5, 0}, {5, 0}}]}, 
                  {GrayLevel[0.5], CircleBox[{-1.540791743911742, 0.5}, 0.5], 
                   DiskBox[{-1.540791743911742, 0.5}, 0.1]}, 
                  {GrayLevel[0.5], 
                   DiskBox[{-1.5542129582938868`, 1.9999399557997342`}, 0.1]}, 
                  {GrayLevel[0.5], Thickness[0.01], 
                   LineBox[{{-1.540791743911742, 0.5}, {-1.5542129582938868`, 
                    1.9999399557997342`}}]}, 
                  {RGBColor[0, 1, 0], 
                   DiskBox[{-1.5497392201665052`, 1.499959970533156}, 0.05]}, 
                  {RGBColor[0, 0, 1], Arrowheads[0.01], 
                   CircleBox[{-1.540791743911742, 0.5}, 0.3333333333333333, 
                    NCache[{
                    Rational[-1, 2] 
                    Pi, -1.4319050956843227`}, {-1.5707963267948966`, \
-1.4319050956843227`}]], 
                   ArrowBox[{{-1.4979468605180044`, 
                    0.16943166050257003`}, {-1.4946433744437955`, 
                    0.1698766304749677}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-1.540791743911742, 0.5}, {-1.6333858979854579`,
                     0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-1.540791743911742, 0.5}, {-1.4749858021836235`,
                     0.5}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.005], Arrowheads[0.04], 
                   ArrowBox[{{-1.540791743911742, 0.5}, {-1.5675799562573394`,
                     0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-1.540791743911742, 0.5}, 1.6, {
                    1.5797439224391394`, 1.6208742824455957`}], 
                   ArrowBox[{{-1.604899307687005, 
                    2.098715177968484}, {-1.6208829876640152`, 
                    2.0979941779225024`}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Dashing[{Small, Small}], 
                   Arrowheads[0.02], 
                   CircleBox[{-1.540791743911742, 0.5}, 1.7, {
                    1.5797439224391394`, 1.5642196591779585`}], 
                   ArrowBox[{{-1.5466113975959797`, 
                    2.199990038685814}, {-1.5296114895591184`, 
                    2.1999632354591117`}}]}, 
                  {RGBColor[0, 0, 1], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-1.540791743911742, 0.5}, 1.8, {
                    1.5797439224391394`, 1.5540233240853294`}], 
                   ArrowBox[{{-1.5286004322448428`, 
                    2.299958713948695}, {-1.5106017546574784`, 
                    2.2997468056781845`}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.003], Arrowheads[0.025], 
                   CircleBox[{-1.540791743911742, 0.5}, 1.5, {
                    1.5797439224391394`, 1.5751574233921157`}], 
                   ArrowBox[{{-1.5623326483531328`, 
                    1.9998453218368375`}, {-1.547333368071488, 
                    1.9999857356499604`}}]}, 
                  {RGBColor[0, 0, 1], ArrowBox[{{0, 2.8}, {0, 2.8}}]}},
                 Frame->True,
                 FrameTicks->Automatic,
                 ImageSize->800,
                 PlotRange->{{-3, 3}, {-1, 3}}], 399->
                GraphicsBox[{InsetBox[
                   TagBox[GridBox[{
                    {"\<\"t\"\>", "19.900000000000002`"},
                    {"\<\"driveIntent\"\>", "0"},
                    {"\<\"motorTorque\"\>", "4.227088866202475`"},
                    {"\<\"vw\"\>", 
                    RowBox[{"-", "0.062234434842776146`"}]},
                    {"\<\"controllerState\"\>", 
                    RowBox[{"\[LeftAssociation]", 
                    RowBox[{
                    
                    RowBox[{"\<\"lastError\"\>", "\[Rule]", 
                    "0.026896255386642833`"}], ",", 
                    RowBox[{"\<\"lastInt\"\>", "\[Rule]", 
                    RowBox[{"-", "0.06633781025048865`"}]}], ",", 
                    RowBox[{"\<\"initialized\"\>", "\[Rule]", "True"}]}], 
                    "\[RightAssociation]"}]}
                    },
                    AutoDelete->False,
                    
                    GridBoxItemSize->{
                    "Columns" -> {{Automatic}}, "Rows" -> {{Automatic}}}],
                    "Grid"], {-1, -0.5}], 
                  {GrayLevel[0], LineBox[{{-5, 0}, {5, 0}}]}, 
                  {GrayLevel[0.5], CircleBox[{-1.543893383680778, 0.5}, 0.5], 
                   DiskBox[{-1.543893383680778, 0.5}, 0.1]}, 
                  {GrayLevel[0.5], 
                   DiskBox[{-1.5562120730312328`, 1.9999494157779745`}, 0.1]}, 
                  {GrayLevel[0.5], Thickness[0.01], 
                   LineBox[{{-1.543893383680778, 0.5}, {-1.5562120730312328`, 
                    1.9999494157779745`}}]}, 
                  {RGBColor[0, 1, 0], 
                   DiskBox[{-1.5521058432477477`, 1.4999662771853162`}, 
                    0.05]}, 
                  {RGBColor[0, 0, 1], Arrowheads[0.01], 
                   CircleBox[{-1.543893383680778, 0.5}, 0.3333333333333333, 
                    NCache[{
                    Rational[-1, 2] 
                    Pi, -1.4439836608088223`}, {-1.5707963267948966`, \
-1.4439836608088223`}]], 
                   ArrowBox[{{-1.505044319737793, 
                    0.1689382816726389}, {-1.5017357001682066`, 
                    0.16934331878523162`}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-1.543893383680778, 0.5}, {-1.6284351610048275`,
                     0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-1.543893383680778, 0.5}, {-1.4834928831650367`,
                     0.5}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.005], Arrowheads[0.04], 
                   ArrowBox[{{-1.543893383680778, 0.5}, {-1.5680346604890862`,
                     0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-1.543893383680778, 0.5}, 1.6, {
                    1.5790088786788623`, 1.6167604645909364`}], 
                   ArrowBox[{{-1.6014236005015305`, 
                    2.098965376158145}, {-1.617410111283345, 
                    2.098310135975684}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Dashing[{Small, Small}], 
                   Arrowheads[0.02], 
                   CircleBox[{-1.543893383680778, 0.5}, 1.7, {
                    1.5790088786788623`, 1.5650184732023602`}], 
                   ArrowBox[{{-1.551071011248068, 
                    2.1999848474802666`}, {-1.5340710872242687`, 
                    2.1999716240256246`}}]}, 
                  {RGBColor[0, 0, 1], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-1.543893383680778, 0.5}, 1.8, {
                    1.5790088786788623`, 1.5555250516444041`}], 
                   ArrowBox[{{-1.5344051323506662`, 
                    2.299974992350365}, {-1.5164061568301077`, 
                    2.299790113418801}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.003], Arrowheads[0.025], 
                   CircleBox[{-1.543893383680778, 0.5}, 1.5, {
                    1.5790088786788623`, 1.5752408516703558`}], 
                   ArrowBox[{{-1.5655594175598855`, 
                    1.9998435194966004`}, {-1.550560149044923, 
                    1.9999851846733618`}}]}, 
                  {RGBColor[0, 0, 1], ArrowBox[{{0, 2.8}, {0, 2.8}}]}},
                 Frame->True,
                 FrameTicks->Automatic,
                 ImageSize->800,
                 PlotRange->{{-3, 3}, {-1, 3}}], 400->
                GraphicsBox[{InsetBox[
                   TagBox[GridBox[{
                    {"\<\"t\"\>", "19.95`"},
                    {"\<\"driveIntent\"\>", "0"},
                    {"\<\"motorTorque\"\>", "3.7779335248101287`"},
                    {"\<\"vw\"\>", 
                    RowBox[{"-", "0.06260825355935276`"}]},
                    {"\<\"controllerState\"\>", 
                    RowBox[{"\[LeftAssociation]", 
                    RowBox[{
                    
                    RowBox[{"\<\"lastError\"\>", "\[Rule]", 
                    "0.026174835538751173`"}], ",", 
                    RowBox[{"\<\"lastInt\"\>", "\[Rule]", 
                    RowBox[{"-", "0.06501095824006314`"}]}], ",", 
                    RowBox[{"\<\"initialized\"\>", "\[Rule]", "True"}]}], 
                    "\[RightAssociation]"}]}
                    },
                    AutoDelete->False,
                    
                    GridBoxItemSize->{
                    "Columns" -> {{Automatic}}, "Rows" -> {{Automatic}}}],
                    "Grid"], {-1, -0.5}], 
                  {GrayLevel[0], LineBox[{{-5, 0}, {5, 0}}]}, 
                  {GrayLevel[0.5], 
                   CircleBox[{-1.5470148318377157`, 0.5}, 0.5], 
                   DiskBox[{-1.5470148318377157`, 0.5}, 0.1]}, 
                  {GrayLevel[0.5], 
                   DiskBox[{-1.5580800798303165`, 1.9999591862070323`}, 0.1]}, 
                  {GrayLevel[0.5], Thickness[0.01], 
                   LineBox[{{-1.5470148318377157`, 
                    0.5}, {-1.5580800798303165`, 1.9999591862070323`}}]}, 
                  {RGBColor[0, 1, 0], 
                   DiskBox[{-1.5543916638327828`, 1.499972790804688}, 0.05]}, 
                  {RGBColor[0, 0, 1], Arrowheads[0.01], 
                   CircleBox[{-1.5470148318377157`, 0.5}, 0.3333333333333333, 
                    NCache[{
                    Rational[-1, 2] 
                    Pi, -1.4574583210505927`}, {-1.5707963267948966`, \
-1.4574583210505927`}]], 
                   ArrowBox[{{-1.512630103866806, 
                    0.16844487397496377`}, {-1.5093163270875387`, 
                    0.16880529314206727`}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-1.5470148318377157`, 
                    0.5}, {-1.6225735023339183`, 0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-1.5470148318377157`, 
                    0.5}, {-1.4927598010002954`, 0.5}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.005], Arrowheads[0.04], 
                   ArrowBox[{{-1.5470148318377157`, 0.5}, {-1.568318471496498,
                     0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-1.5470148318377157`, 0.5}, 1.6, {
                    1.5781732256965795`, 1.612083542642404}], 
                   ArrowBox[{{-1.5970662104631008`, 
                    2.0992169519792174`}, {-1.6130556108999905`, 
                    2.0986364863535574`}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Dashing[{Small, Small}], 
                   Arrowheads[0.02], 
                   CircleBox[{-1.5470148318377157`, 0.5}, 1.7, {
                    1.5781732256965795`, 1.5658272116980834`}], 
                   ArrowBox[{{-1.5555673000961456`, 
                    2.1999784867129017`}, {-1.5385673709374983`, 
                    2.1999790117540687`}}]}, 
                  {RGBColor[0, 0, 1], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-1.5470148318377157`, 0.5}, 1.8, {
                    1.5781732256965795`, 1.557184706114301}], 
                   ArrowBox[{{-1.540513928745416, 
                    2.299988260589214}, {-1.5225146711785131`, 
                    2.2998332539787327`}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.003], Arrowheads[0.025], 
                   CircleBox[{-1.5470148318377157`, 0.5}, 1.5, {
                    1.5781732256965795`, 1.5751658577417627`}], 
                   ArrowBox[{{-1.5685683864982007`, 
                    1.9998451400999695`}, {-1.5535691074013889`, 
                    1.9999856804222618`}}]}, 
                  {RGBColor[0, 0, 1], ArrowBox[{{0, 2.8}, {0, 2.8}}]}},
                 Frame->True,
                 FrameTicks->Automatic,
                 ImageSize->800,
                 PlotRange->{{-3, 3}, {-1, 3}}], 401->
                GraphicsBox[{InsetBox[
                   TagBox[GridBox[{
                    {"\<\"t\"\>", "20.`"},
                    {"\<\"driveIntent\"\>", "0"},
                    {"\<\"motorTorque\"\>", "3.292243455228668`"},
                    {"\<\"vw\"\>", 
                    RowBox[{"-", "0.06293428067803182`"}]},
                    {"\<\"controllerState\"\>", 
                    RowBox[{"\[LeftAssociation]", 
                    RowBox[{
                    
                    RowBox[{"\<\"lastError\"\>", "\[Rule]", 
                    "0.02535837521313615`"}], ",", 
                    RowBox[{"\<\"lastInt\"\>", "\[Rule]", 
                    RowBox[{"-", "0.06372267922881361`"}]}], ",", 
                    RowBox[{"\<\"initialized\"\>", "\[Rule]", "True"}]}], 
                    "\[RightAssociation]"}]}
                    },
                    AutoDelete->False,
                    
                    GridBoxItemSize->{
                    "Columns" -> {{Automatic}}, "Rows" -> {{Automatic}}}],
                    "Grid"], {-1, -0.5}], 
                  {GrayLevel[0], LineBox[{{-5, 0}, {5, 0}}]}, 
                  {GrayLevel[0.5], 
                   CircleBox[{-1.5501537618296448`, 0.5}, 0.5], 
                   DiskBox[{-1.5501537618296448`, 0.5}, 0.1]}, 
                  {GrayLevel[0.5], 
                   DiskBox[{-1.5598450736178435`, 1.9999686924985547`}, 0.1]}, 
                  {GrayLevel[0.5], Thickness[0.01], 
                   LineBox[{{-1.5501537618296448`, 
                    0.5}, {-1.5598450736178435`, 1.9999686924985547`}}]}, 
                  {RGBColor[0, 1, 0], 
                   DiskBox[{-1.5566146363551105`, 1.4999791283323698`}, 
                    0.05]}, 
                  {RGBColor[0, 0, 1], Arrowheads[0.01], 
                   CircleBox[{-1.5501537618296448`, 0.5}, 0.3333333333333333, 
                    NCache[{
                    Rational[-1, 2] 
                    Pi, -1.4720290231380364`}, {-1.5707963267948966`, \
-1.4720290231380364`}]], 
                   ArrowBox[{{-1.5206035038529837`, 
                    0.16797907691739683`}, {-1.5172848274592878`, 
                    0.16829117547995764`}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-1.5501537618296448`, 
                    0.5}, {-1.6159986309342182`, 0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-1.5501537618296448`, 
                    0.5}, {-1.5026351022701452`, 0.5}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.005], Arrowheads[0.04], 
                   ArrowBox[{{-1.5501537618296448`, 
                    0.5}, {-1.5684799713747186`, 0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-1.5501537618296448`, 0.5}, 1.6, {
                    1.5772572462704795`, 1.6069570283791474`}], 
                   ArrowBox[{{-1.5920061101488017`, 
                    2.09945252537866}, {-1.6079982762285228`, 
                    2.098954036910992}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Dashing[{Small, Small}], 
                   Arrowheads[0.02], 
                   CircleBox[{-1.5501537618296448`, 0.5}, 1.7, {
                    1.5772572462704795`, 1.5666366617630954`}], 
                   ArrowBox[{{-1.5600822748325682`, 
                    2.1999710069967517`}, {-1.543082351668172, 
                    2.1999852926300063`}}]}, 
                  {RGBColor[0, 0, 1], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-1.5501537618296448`, 0.5}, 1.8, {
                    1.5772572462704795`, 1.5589670048525424`}], 
                   ArrowBox[{{-1.5468609841699101`, 
                    2.299996988223948}, {-1.5288614789231796`, 
                    2.2998740618967295`}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.003], Arrowheads[0.025], 
                   CircleBox[{-1.5501537618296448`, 0.5}, 1.5, {
                    1.5772572462704795`, 1.5749520617098325`}], 
                   ArrowBox[{{-1.5713866550605167`, 
                    1.999849713886377}, {-1.5563873462595408`, 
                    1.9999870474191286`}}]}, 
                  {RGBColor[0, 0, 1], ArrowBox[{{0, 2.8}, {0, 2.8}}]}},
                 Frame->True,
                 FrameTicks->Automatic,
                 ImageSize->800,
                 PlotRange->{{-3, 3}, {-1, 3}}], 402->
                GraphicsBox[{InsetBox[
                   TagBox[GridBox[{
                    {"\<\"t\"\>", "20.05`"},
                    {"\<\"driveIntent\"\>", "0"},
                    {"\<\"motorTorque\"\>", "2.779222999336614`"},
                    {"\<\"vw\"\>", 
                    RowBox[{"-", "0.06321060117835434`"}]},
                    {"\<\"controllerState\"\>", 
                    RowBox[{"\[LeftAssociation]", 
                    RowBox[{
                    
                    RowBox[{"\<\"lastError\"\>", "\[Rule]", 
                    "0.024465051391449662`"}], ",", 
                    RowBox[{"\<\"lastInt\"\>", "\[Rule]", 
                    RowBox[{"-", "0.06247725617988749`"}]}], ",", 
                    RowBox[{"\<\"initialized\"\>", "\[Rule]", "True"}]}], 
                    "\[RightAssociation]"}]}
                    },
                    AutoDelete->False,
                    
                    GridBoxItemSize->{
                    "Columns" -> {{Automatic}}, "Rows" -> {{Automatic}}}],
                    "Grid"], {-1, -0.5}], 
                  {GrayLevel[0], LineBox[{{-5, 0}, {5, 0}}]}, 
                  {GrayLevel[0.5], 
                   CircleBox[{-1.5533077320669173`, 0.5}, 0.5], 
                   DiskBox[{-1.5533077320669173`, 0.5}, 0.1]}, 
                  {GrayLevel[0.5], 
                   DiskBox[{-1.5615326985321552`, 1.9999774498060447`}, 0.1]}, 
                  {GrayLevel[0.5], Thickness[0.01], 
                   LineBox[{{-1.5533077320669173`, 
                    0.5}, {-1.5615326985321552`, 1.9999774498060447`}}]}, 
                  {RGBColor[0, 1, 0], 
                   DiskBox[{-1.5587910430437426`, 1.4999849665373632`}, 
                    0.05]}, 
                  {RGBColor[0, 0, 1], Arrowheads[0.01], 
                   CircleBox[{-1.5533077320669173`, 0.5}, 0.3333333333333333, 
                    NCache[{
                    Rational[-1, 2] 
                    Pi, -1.487419636814798}, {-1.5707963267948966`, \
-1.487419636814798}]], 
                   ArrowBox[{{-1.5288707778473967`, 
                    0.16756362055336943`}, {-1.5255476912962456`, 
                    0.16782460770321694`}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-1.5533077320669173`, 
                    0.5}, {-1.6088921920536496`, 0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-1.5533077320669173`, 
                    0.5}, {-1.5129786546685478`, 0.5}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.005], Arrowheads[0.04], 
                   ArrowBox[{{-1.5533077320669173`, 0.5}, {-1.56856311465528, 
                    0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-1.5533077320669173`, 0.5}, 1.6, {
                    1.576279665249604, 1.6014857175578305`}], 
                   ArrowBox[{{-1.586408395708254, 
                    2.099657571502884}, {-1.6024030497956303`, 
                    2.099246588171167}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Dashing[{Small, Small}], 
                   Arrowheads[0.02], 
                   CircleBox[{-1.5533077320669173`, 0.5}, 1.7, {
                    1.576279665249604, 1.5674386637061548`}], 
                   ArrowBox[{{-1.564599621781396, 
                    2.19996249759419}, {-1.5475997155413397`, 
                    2.199990417192798}}]}, 
                  {RGBColor[0, 0, 1], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-1.5533077320669173`, 0.5}, 1.8, {
                    1.576279665249604, 1.5608395374755117`}], 
                   ArrowBox[{{-1.5533855112920003`, 
                    2.2999999983195534`}, {-1.5353858074183757`, 
                    2.2999107768489226`}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.003], Arrowheads[0.025], 
                   CircleBox[{-1.5533077320669173`, 0.5}, 1.5, {
                    1.576279665249604, 1.5746181659668925`}], 
                   ArrowBox[{{-1.574039830688989, 
                    1.999856719852508}, {-1.5590404768690411`, 
                    1.9999890451723419`}}]}, 
                  {RGBColor[0, 0, 1], ArrowBox[{{0, 2.8}, {0, 2.8}}]}},
                 Frame->True,
                 FrameTicks->Automatic,
                 ImageSize->800,
                 PlotRange->{{-3, 3}, {-1, 3}}], 403->
                GraphicsBox[{InsetBox[
                   TagBox[GridBox[{
                    {"\<\"t\"\>", "20.1`"},
                    {"\<\"driveIntent\"\>", "0"},
                    {"\<\"motorTorque\"\>", "2.247282483921074`"},
                    {"\<\"vw\"\>", 
                    RowBox[{"-", "0.06343602613824248`"}]},
                    {"\<\"controllerState\"\>", 
                    RowBox[{"\[LeftAssociation]", 
                    RowBox[{
                    
                    RowBox[{"\<\"lastError\"\>", "\[Rule]", 
                    "0.023511700488146096`"}], ",", 
                    RowBox[{"\<\"lastInt\"\>", "\[Rule]", 
                    RowBox[{"-", "0.061278097328937804`"}]}], ",", 
                    RowBox[{"\<\"initialized\"\>", "\[Rule]", "True"}]}], 
                    "\[RightAssociation]"}]}
                    },
                    AutoDelete->False,
                    
                    GridBoxItemSize->{
                    "Columns" -> {{Automatic}}, "Rows" -> {{Automatic}}}],
                    "Grid"], {-1, -0.5}], 
                  {GrayLevel[0], LineBox[{{-5, 0}, {5, 0}}]}, 
                  {GrayLevel[0.5], 
                   CircleBox[{-1.5564742240285716`, 0.5}, 0.5], 
                   DiskBox[{-1.5564742240285716`, 0.5}, 0.1]}, 
                  {GrayLevel[0.5], 
                   DiskBox[{-1.5631661848176905`, 1.999985072479322}, 0.1]}, 
                  {GrayLevel[0.5], Thickness[0.01], 
                   LineBox[{{-1.5564742240285716`, 
                    0.5}, {-1.5631661848176905`, 1.999985072479322}}]}, 
                  {RGBColor[0, 1, 0], 
                   DiskBox[{-1.5609355312213176`, 1.4999900483195479`}, 
                    0.05]}, 
                  {RGBColor[0, 0, 1], Arrowheads[0.01], 
                   CircleBox[{-1.5564742240285716`, 0.5}, 0.3333333333333333, 
                    NCache[{
                    Rational[-1, 2] 
                    Pi, -1.5033778522772643`}, {-1.5707963267948966`, \
-1.5033778522772643`}]], 
                   ArrowBox[{{-1.5373452475608742`, 
                    0.16721599592166486`}, {-1.5340184194246673`, 
                    0.1674239215597394}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-1.5564742240285716`, 
                    0.5}, {-1.6014198737069931`, 0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-1.5564742240285716`, 
                    0.5}, {-1.5236616919353487`, 0.5}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.005], Arrowheads[0.04], 
                   ArrowBox[{{-1.5564742240285716`, 
                    0.5}, {-1.5686073416137702`, 0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-1.5564742240285716`, 0.5}, 1.6, {
                    1.5752576487868692`, 1.595765685434561}], 
                   ArrowBox[{{-1.5804243033662424`, 
                    2.099820737989016}, {-1.5964210466166884`, 
                    2.0995012508169895`}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Dashing[{Small, Small}], 
                   Arrowheads[0.02], 
                   CircleBox[{-1.5564742240285716`, 0.5}, 1.7, {
                    1.5752576487868692`, 1.5682260671214512`}], 
                   ArrowBox[{{-1.5691046663807042`, 
                    2.199953079330718}, {-1.5521047873946388`, 
                    2.199994384703521}}]}, 
                  {RGBColor[0, 0, 1], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-1.5564742240285716`, 0.5}, 1.8, {
                    1.5752576487868692`, 1.5627727460984189`}], 
                   ArrowBox[{{-1.5600317764588056`, 
                    2.2999964843912073`}, {-1.5420319337366684`, 
                    2.299942060248364}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.003], Arrowheads[0.025], 
                   CircleBox[{-1.5564742240285716`, 0.5}, 1.5, {
                    1.5752576487868692`, 1.5741818787641262`}], 
                   ArrowBox[{{-1.5765519524054066`, 
                    1.9998656222552826`}, {-1.5615525422811545`, 
                    1.9999914035366086`}}]}, 
                  {RGBColor[0, 0, 1], ArrowBox[{{0, 2.8}, {0, 2.8}}]}},
                 Frame->True,
                 FrameTicks->Automatic,
                 ImageSize->800,
                 PlotRange->{{-3, 3}, {-1, 3}}], 404->
                GraphicsBox[{InsetBox[
                   TagBox[GridBox[{
                    {"\<\"t\"\>", "20.150000000000002`"},
                    {"\<\"driveIntent\"\>", "0"},
                    {"\<\"motorTorque\"\>", "1.704048784441369`"},
                    {"\<\"vw\"\>", 
                    RowBox[{"-", "0.06361002057287922`"}]},
                    {"\<\"controllerState\"\>", 
                    RowBox[{"\[LeftAssociation]", 
                    RowBox[{
                    
                    RowBox[{"\<\"lastError\"\>", "\[Rule]", 
                    "0.022513810705386852`"}], ",", 
                    RowBox[{"\<\"lastInt\"\>", "\[Rule]", 
                    RowBox[{"-", "0.060127803468521986`"}]}], ",", 
                    RowBox[{"\<\"initialized\"\>", "\[Rule]", "True"}]}], 
                    "\[RightAssociation]"}]}
                    },
                    AutoDelete->False,
                    
                    GridBoxItemSize->{
                    "Columns" -> {{Automatic}}, "Rows" -> {{Automatic}}}],
                    "Grid"], {-1, -0.5}], 
                  {GrayLevel[0], LineBox[{{-5, 0}, {5, 0}}]}, 
                  {GrayLevel[0.5], 
                   CircleBox[{-1.5596506767195548`, 0.5}, 0.5], 
                   DiskBox[{-1.5596506767195548`, 0.5}, 0.1]}, 
                  {GrayLevel[0.5], 
                   DiskBox[{-1.5647664045629581`, 1.9999912764175105`}, 0.1]}, 
                  {GrayLevel[0.5], Thickness[0.01], 
                   LineBox[{{-1.5596506767195548`, 
                    0.5}, {-1.5647664045629581`, 1.9999912764175105`}}]}, 
                  {RGBColor[0, 1, 0], 
                   DiskBox[{-1.5630611619484902`, 1.4999941842783402`}, 
                    0.05]}, 
                  {RGBColor[0, 0, 1], Arrowheads[0.01], 
                   CircleBox[{-1.5596506767195548`, 0.5}, 0.3333333333333333, 
                    NCache[{
                    Rational[-1, 2] 
                    Pi, -1.5196748632616555`}, {-1.5707963267948966`, \
-1.5196748632616555`}]], 
                   ArrowBox[{{-1.5459473849572616`, 
                    0.1669484560822626}, {-1.5426176101852758`, 
                    0.16710213915443983`}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-1.5596506767195548`, 
                    0.5}, {-1.5937316524083822`, 0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-1.5596506767195548`, 0.5}, {-1.534566746373805,
                     0.5}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.005], Arrowheads[0.04], 
                   ArrowBox[{{-1.5596506767195548`, 
                    0.5}, {-1.5686477220626325`, 0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-1.5596506767195548`, 0.5}, 1.6, {
                    1.574206818635325, 1.589884366277672}], 
                   ArrowBox[{{-1.5741913397318748`, 
                    2.099933926485454}, {-1.5901896853153161`, 
                    2.099708526249075}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Dashing[{Small, Small}], 
                   Arrowheads[0.02], 
                   CircleBox[{-1.5596506767195548`, 0.5}, 1.7, {
                    1.574206818635325, 1.5689926825247091`}], 
                   ArrowBox[{{-1.5735843254480189`, 
                    2.199942897109521}, {-1.5565844831226925`, 
                    2.199997234837994}}]}, 
                  {RGBColor[0, 0, 1], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-1.5596506767195548`, 0.5}, 1.8, {
                    1.574206818635325, 1.564739880943984}], 
                   ArrowBox[{{-1.5667490557893307`, 
                    2.299986003560745}, {-1.5487491408338947`, 
                    2.299966987618199}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.003], Arrowheads[0.025], 
                   CircleBox[{-1.5596506767195548`, 0.5}, 1.5, {
                    1.574206818635325, 1.573659856632188}], 
                   ArrowBox[{{-1.578945439346538, 
                    1.9998758989113627`}, {-1.5639459656053991`, 
                    1.9999938501518555`}}]}, 
                  {RGBColor[0, 0, 1], ArrowBox[{{0, 2.8}, {0, 2.8}}]}},
                 Frame->True,
                 FrameTicks->Automatic,
                 ImageSize->800,
                 PlotRange->{{-3, 3}, {-1, 3}}], 405->
                GraphicsBox[{InsetBox[
                   TagBox[GridBox[{
                    {"\<\"t\"\>", "20.2`"},
                    {"\<\"driveIntent\"\>", "0"},
                    {"\<\"motorTorque\"\>", "1.1563821026657521`"},
                    {"\<\"vw\"\>", 
                    RowBox[{"-", "0.06373263333936631`"}]},
                    {"\<\"controllerState\"\>", 
                    RowBox[{"\[LeftAssociation]", 
                    RowBox[{
                    
                    RowBox[{"\<\"lastError\"\>", "\[Rule]", 
                    "0.021485528189341633`"}], ",", 
                    RowBox[{"\<\"lastInt\"\>", "\[Rule]", 
                    RowBox[{"-", "0.05902823525565344`"}]}], ",", 
                    RowBox[{"\<\"initialized\"\>", "\[Rule]", "True"}]}], 
                    "\[RightAssociation]"}]}
                    },
                    AutoDelete->False,
                    
                    GridBoxItemSize->{
                    "Columns" -> {{Automatic}}, "Rows" -> {{Automatic}}}],
                    "Grid"], {-1, -0.5}], 
                  {GrayLevel[0], LineBox[{{-5, 0}, {5, 0}}]}, 
                  {GrayLevel[0.5], 
                   CircleBox[{-1.5628345175705811`, 0.5}, 0.5], 
                   DiskBox[{-1.5628345175705811`, 0.5}, 0.1]}, 
                  {GrayLevel[0.5], 
                   DiskBox[{-1.566351943542154, 1.9999958758991756`}, 0.1]}, 
                  {GrayLevel[0.5], Thickness[0.01], 
                   LineBox[{{-1.5628345175705811`, 0.5}, {-1.566351943542154, 
                    1.9999958758991756`}}]}, 
                  {RGBColor[0, 1, 0], 
                   DiskBox[{-1.5651794682182965`, 1.4999972505994503`}, 
                    0.05]}, 
                  {RGBColor[0, 0, 1], Arrowheads[0.01], 
                   CircleBox[{-1.5628345175705811`, 0.5}, 0.3333333333333333, 
                    NCache[{
                    Rational[-1, 2] 
                    Pi, -1.536104863714924}, {-1.5707963267948966`, \
-1.536104863714924}]], 
                   ArrowBox[{{-1.554604866163071, 
                    0.16676827289583906`}, {-1.5512730159095143`, 
                    0.1668672294858216}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-1.5628345175705811`, 
                    0.5}, {-1.5859621596238962`, 0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-1.5628345175705811`, 
                    0.5}, {-1.5455874822876565`, 0.5}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.005], Arrowheads[0.04], 
                   ArrowBox[{{-1.5628345175705811`, 
                    0.5}, {-1.5687151243409716`, 0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-1.5628345175705811`, 0.5}, 1.6, {
                    1.5731412795916837`, 1.5839207062804732`}], 
                   ArrowBox[{{-1.5678335166143336`, 
                    2.099992190608617}, {-1.5838329219085188`, 
                    2.0998622025084726`}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Dashing[{Small, Small}], 
                   Arrowheads[0.02], 
                   CircleBox[{-1.5628345175705811`, 0.5}, 1.7, {
                    1.5731412795916837`, 1.5697332296097424`}], 
                   ArrowBox[{{-1.5780270501204823`, 
                    2.1999321124547064`}, {-1.56102725269624, 
                    2.199999039350809}}]}, 
                  {RGBColor[0, 0, 1], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-1.5628345175705811`, 0.5}, 1.8, {
                    1.5731412795916837`, 1.566716934576874}], 
                   ArrowBox[{{-1.5734915493166708`, 
                    2.2999684518552996`}, {-1.555491631944213, 
                    2.2999850227239884`}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.003], Arrowheads[0.025], 
                   CircleBox[{-1.5628345175705811`, 0.5}, 1.5, {
                    1.5731412795916837`, 1.5730676640624206`}], 
                   ArrowBox[{{-1.581241061503311, 
                    1.9998870621284965`}, {-1.566241520542426, 
                    1.9999961307719263`}}]}, 
                  {RGBColor[0, 0, 1], ArrowBox[{{0, 2.8}, {0, 2.8}}]}},
                 Frame->True,
                 FrameTicks->Automatic,
                 ImageSize->800,
                 PlotRange->{{-3, 3}, {-1, 3}}], 406->
                GraphicsBox[{InsetBox[
                   TagBox[GridBox[{
                    {"\<\"t\"\>", "20.25`"},
                    {"\<\"driveIntent\"\>", "0"},
                    {"\<\"motorTorque\"\>", "0.6103980858594316`"},
                    {"\<\"vw\"\>", 
                    RowBox[{"-", "0.06380442954247408`"}]},
                    {"\<\"controllerState\"\>", 
                    RowBox[{"\[LeftAssociation]", 
                    RowBox[{
                    
                    RowBox[{"\<\"lastError\"\>", "\[Rule]", 
                    "0.020439675310107673`"}], ",", 
                    RowBox[{"\<\"lastInt\"\>", "\[Rule]", 
                    RowBox[{"-", "0.05798057989558008`"}]}], ",", 
                    RowBox[{"\<\"initialized\"\>", "\[Rule]", "True"}]}], 
                    "\[RightAssociation]"}]}
                    },
                    AutoDelete->False,
                    
                    GridBoxItemSize->{
                    "Columns" -> {{Automatic}}, "Rows" -> {{Automatic}}}],
                    "Grid"], {-1, -0.5}], 
                  {GrayLevel[0], LineBox[{{-5, 0}, {5, 0}}]}, 
                  {GrayLevel[0.5], CircleBox[{-1.56602318989536, 0.5}, 0.5], 
                   DiskBox[{-1.56602318989536, 0.5}, 0.1]}, 
                  {GrayLevel[0.5], 
                   DiskBox[{-1.5679391866027195`, 1.9999987763183733`}, 0.1]}, 
                  {GrayLevel[0.5], Thickness[0.01], 
                   LineBox[{{-1.56602318989536, 0.5}, {-1.5679391866027195`, 
                    1.9999987763183733`}}]}, 
                  {RGBColor[0, 1, 0], 
                   DiskBox[{-1.5673005210335997`, 1.4999991842122489`}, 
                    0.05]}, 
                  {RGBColor[0, 0, 1], Arrowheads[0.01], 
                   CircleBox[{-1.56602318989536, 0.5}, 0.3333333333333333, 
                    NCache[{
                    Rational[-1, 2] 
                    Pi, -1.5524843842191136`}, {-1.5707963267948966`, \
-1.5524843842191136`}]], 
                   ArrowBox[{{-1.5632525742732513`, 
                    0.16667818133193618`}, {-1.559919550169556, 
                    0.16672255297844008`}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-1.56602318989536, 0.5}, {-1.5782311516125487`, 
                    0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-1.56602318989536, 0.5}, {-1.5566284430043722`, 
                    0.5}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.005], Arrowheads[0.04], 
                   ArrowBox[{{-1.56602318989536, 0.5}, {-1.5688364047215608`, 
                    0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-1.56602318989536, 0.5}, 1.6, {1.57207365828048,
                     1.577945379877426}], 
                   ArrowBox[{{-1.5614616810066595`, 
                    2.0999934976857433`}, {-1.5774615773928111`, 
                    2.09995911300616}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Dashing[{Small, Small}], 
                   Arrowheads[0.02], 
                   CircleBox[{-1.56602318989536, 0.5}, 1.7, {1.57207365828048,
                     1.5704432830383237`}], 
                   ArrowBox[{{-1.5824227611387842`, 
                    2.1999208964134276`}, {-1.5654230155216537`, 
                    2.199999894056091}}]}, 
                  {RGBColor[0, 0, 1], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-1.56602318989536, 0.5}, 1.8, {1.57207365828048,
                     1.568682557803483}], 
                   ArrowBox[{{-1.5802182585716165`, 
                    2.299944026914525}, {-1.5622184085441233`, 
                    2.299995978784083}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.003], Arrowheads[0.025], 
                   CircleBox[{-1.56602318989536, 0.5}, 1.5, {1.57207365828048,
                     1.5724197488037157`}], 
                   ArrowBox[{{-1.5534584698480334`, 
                    1.9999473750135812`}, {-1.568458321838957, 
                    1.99999802337617}}]}, 
                  {RGBColor[0, 0, 1], ArrowBox[{{0, 2.8}, {0, 2.8}}]}},
                 Frame->True,
                 FrameTicks->Automatic,
                 ImageSize->800,
                 PlotRange->{{-3, 3}, {-1, 3}}], 407->
                GraphicsBox[{InsetBox[
                   TagBox[GridBox[{
                    {"\<\"t\"\>", "20.3`"},
                    {"\<\"driveIntent\"\>", "0"},
                    {"\<\"motorTorque\"\>", "0.07149447221591387`"},
                    {"\<\"vw\"\>", 
                    RowBox[{"-", "0.06382642579593613`"}]},
                    {"\<\"controllerState\"\>", 
                    RowBox[{"\[LeftAssociation]", 
                    RowBox[{
                    
                    RowBox[{"\<\"lastError\"\>", "\[Rule]", 
                    "0.0193877794768517`"}], ",", 
                    RowBox[{"\<\"lastInt\"\>", "\[Rule]", 
                    RowBox[{"-", "0.056985416636495485`"}]}], ",", 
                    RowBox[{"\<\"initialized\"\>", "\[Rule]", "True"}]}], 
                    "\[RightAssociation]"}]}
                    },
                    AutoDelete->False,
                    
                    GridBoxItemSize->{
                    "Columns" -> {{Automatic}}, "Rows" -> {{Automatic}}}],
                    "Grid"], {-1, -0.5}], 
                  {GrayLevel[0], LineBox[{{-5, 0}, {5, 0}}]}, 
                  {GrayLevel[0.5], CircleBox[{-1.569214177039611, 0.5}, 0.5], 
                   DiskBox[{-1.569214177039611, 0.5}, 0.1]}, 
                  {GrayLevel[0.5], 
                   DiskBox[{-1.5695424142235916`, 1.9999999640867832`}, 0.1]}, 
                  {GrayLevel[0.5], Thickness[0.01], 
                   LineBox[{{-1.569214177039611, 0.5}, {-1.5695424142235916`, 
                    1.9999999640867832`}}]}, 
                  {RGBColor[0, 1, 0], 
                   DiskBox[{-1.5694330018289315`, 1.4999999760578555`}, 
                    0.05]}, 
                  {RGBColor[0, 0, 1], Arrowheads[0.01], 
                   CircleBox[{-1.569214177039611, 0.5}, 0.3333333333333333, 
                    NCache[{
                    Rational[-1, 2] 
                    Pi, -1.5686514926284192`}, {-1.5707963267948966`, \
-1.5686514926284192`}]], 
                   ArrowBox[{{-1.5718325387235212`, 
                    0.1666769505521658}, {-1.5684992328656135`, 
                    0.16666743338530637`}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-1.569214177039611, 0.5}, {-1.5706440664839294`,
                     0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-1.569214177039611, 0.5}, {-1.567604723488003, 
                    0.5}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.005], Arrowheads[0.04], 
                   ArrowBox[{{-1.569214177039611, 0.5}, {-1.5690346129323214`,
                     0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-1.569214177039611, 0.5}, 1.6, {
                    1.5710151515859634`, 1.572021060079802}], 
                   ArrowBox[{{-1.5551739304926586`, 
                    2.0999383961505833`}, {-1.5711737498055756`, 
                    2.0999988000228544`}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Dashing[{Small, Small}], 
                   Arrowheads[0.02], 
                   CircleBox[{-1.569214177039611, 0.5}, 1.7, {
                    1.5710151515859634`, 1.5711192165543175`}], 
                   ArrowBox[{{-1.5527633463937813`, 
                    2.199920401127965}, {-1.5697630896210886`, 
                    2.199999911380874}}]}, 
                  {RGBColor[0, 0, 1], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-1.569214177039611, 0.5}, 1.8, {
                    1.5710151515859634`, 1.5706179600736527`}], 
                   ArrowBox[{{-1.586892832711118, 
                    2.2999131826656662`}, {-1.5688931169430744`, 
                    2.2999999713667814`}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.003], Arrowheads[0.025], 
                   CircleBox[{-1.569214177039611, 0.5}, 1.5, {
                    1.5710151515859634`, 1.571729431128753}], 
                   ArrowBox[{{-1.5556140198838229`, 
                    1.999938343974624}, {-1.570613833337286, 
                    1.9999993469872739`}}]}, 
                  {RGBColor[0, 0, 1], ArrowBox[{{0, 2.8}, {0, 2.8}}]}},
                 Frame->True,
                 FrameTicks->Automatic,
                 ImageSize->800,
                 PlotRange->{{-3, 3}, {-1, 3}}], 408->
                GraphicsBox[{InsetBox[
                   TagBox[GridBox[{
                    {"\<\"t\"\>", "20.35`"},
                    {"\<\"driveIntent\"\>", "0"},
                    {"\<\"motorTorque\"\>", 
                    RowBox[{"-", "0.45561849755052286`"}]},
                    {"\<\"vw\"\>", 
                    RowBox[{"-", "0.06380002862135824`"}]},
                    {"\<\"controllerState\"\>", 
                    RowBox[{"\[LeftAssociation]", 
                    RowBox[{
                    
                    RowBox[{"\<\"lastError\"\>", "\[Rule]", 
                    "0.018340110993083927`"}], ",", 
                    RowBox[{"\<\"lastInt\"\>", "\[Rule]", 
                    RowBox[{"-", "0.056042780586939275`"}]}], ",", 
                    RowBox[{"\<\"initialized\"\>", "\[Rule]", "True"}]}], 
                    "\[RightAssociation]"}]}
                    },
                    AutoDelete->False,
                    
                    GridBoxItemSize->{
                    "Columns" -> {{Automatic}}, "Rows" -> {{Automatic}}}],
                    "Grid"], {-1, -0.5}], 
                  {GrayLevel[0], LineBox[{{-5, 0}, {5, 0}}]}, 
                  {GrayLevel[0.5], 
                   CircleBox[{-1.5724050233721159`, 0.5}, 0.5], 
                   DiskBox[{-1.5724050233721159`, 0.5}, 0.1]}, 
                  {GrayLevel[0.5], 
                   DiskBox[{-1.5711739080488434`, 1.999999494784935}, 0.1]}, 
                  {GrayLevel[0.5], Thickness[0.01], 
                   LineBox[{{-1.5724050233721159`, 
                    0.5}, {-1.5711739080488434`, 1.999999494784935}}]}, 
                  {RGBColor[0, 1, 0], 
                   DiskBox[{-1.5715842798232675`, 1.4999996631899568`}, 
                    0.05]}, 
                  {RGBColor[0, 0, 1], Arrowheads[0.01], 
                   CircleBox[{-1.5724050233721159`, 0.5}, 0.3333333333333333, 
                    NCache[{
                    Rational[-1, 2] 
                    Pi, -1.5844648817214122`}, {-1.5707963267948966`, \
-1.5844648817214122`}]], 
                   ArrowBox[{{-1.5736278722713732`, 
                    0.16666890971335918`}, {-1.5769610664774, 
                    0.16669780441417148`}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-1.5724050233721159`, 
                    0.5}, {-1.5632926534211053`, 0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-1.5724050233721159`, 
                    0.5}, {-1.5784415798814269`, 0.5}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.005], Arrowheads[0.04], 
                   ArrowBox[{{-1.5724050233721159`, 
                    0.5}, {-1.5693292099304164`, 0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-1.5724050233721159`, 0.5}, 1.6, {
                    1.5699755831539035`, 1.5662027340648506`}], 
                   ArrowBox[{{-1.5810552328640575`, 
                    2.0999766166653018`}, {-1.5650553008519703`, 
                    2.099983119154348}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Dashing[{Small, Small}], 
                   Arrowheads[0.02], 
                   CircleBox[{-1.5724050233721159`, 0.5}, 1.7, {
                    1.5699755831539035`, 1.5717581461111738`}], 
                   ArrowBox[{{-1.5570403253988403`, 
                    2.1999305650691117`}, {-1.574040115957684, 
                    2.1999992136681232`}}]}, 
                  {RGBColor[0, 0, 1], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-1.5724050233721159`, 0.5}, 1.8, {
                    1.5699755831539035`, 1.5725067970291842`}], 
                   ArrowBox[{{-1.5574840406809998`, 
                    2.299938155680781}, {-1.5754838682925325`, 
                    2.299997366863062}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.003], Arrowheads[0.025], 
                   CircleBox[{-1.5724050233721159`, 0.5}, 1.5, {
                    1.5699755831539035`, 1.5710089054608236`}], 
                   ArrowBox[{{-1.5577241257630048`, 
                    1.9999281553612462`}, {-1.5727238913686048`, 
                    1.9999999661077332`}}]}, 
                  {RGBColor[0, 0, 1], ArrowBox[{{0, 2.8}, {0, 2.8}}]}},
                 Frame->True,
                 FrameTicks->Automatic,
                 ImageSize->800,
                 PlotRange->{{-3, 3}, {-1, 3}}], 409->
                GraphicsBox[{InsetBox[
                   TagBox[GridBox[{
                    {"\<\"t\"\>", "20.400000000000002`"},
                    {"\<\"driveIntent\"\>", "0"},
                    {"\<\"motorTorque\"\>", 
                    RowBox[{"-", "0.9668846090125438`"}]},
                    {"\<\"vw\"\>", 
                    RowBox[{"-", "0.0637269762022708`"}]},
                    {"\<\"controllerState\"\>", 
                    RowBox[{"\[LeftAssociation]", 
                    RowBox[{
                    
                    RowBox[{"\<\"lastError\"\>", "\[Rule]", 
                    "0.01730572855190707`"}], ",", 
                    RowBox[{"\<\"lastInt\"\>", "\[Rule]", 
                    RowBox[{"-", "0.05515222443998154`"}]}], ",", 
                    RowBox[{"\<\"initialized\"\>", "\[Rule]", "True"}]}], 
                    "\[RightAssociation]"}]}
                    },
                    AutoDelete->False,
                    
                    GridBoxItemSize->{
                    "Columns" -> {{Automatic}}, "Rows" -> {{Automatic}}}],
                    "Grid"], {-1, -0.5}], 
                  {GrayLevel[0], LineBox[{{-5, 0}, {5, 0}}]}, 
                  {GrayLevel[0.5], 
                   CircleBox[{-1.5755933522804908`, 0.5}, 0.5], 
                   DiskBox[{-1.5755933522804908`, 0.5}, 0.1]}, 
                  {GrayLevel[0.5], 
                   DiskBox[{-1.5728440633771552`, 1.9999974804680591`}, 0.1]}, 
                  {GrayLevel[0.5], Thickness[0.01], 
                   LineBox[{{-1.5755933522804908`, 
                    0.5}, {-1.5728440633771552`, 1.9999974804680591`}}]}, 
                  {RGBColor[0, 1, 0], 
                   DiskBox[{-1.5737604930116005`, 1.4999983203120395`}, 
                    0.05]}, 
                  {RGBColor[0, 0, 1], Arrowheads[0.01], 
                   CircleBox[{-1.5755933522804908`, 0.5}, 0.3333333333333333, 
                    NCache[{
                    Rational[-1, 2] 
                    Pi, -1.599802865065273}, {-1.5707963267948966`, \
-1.599802865065273}]], 
                   ArrowBox[{{-1.5819284835950957`, 
                    0.1667268729370191}, {-1.5852608425665562`, 
                    0.166806886711805}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-1.5755933522804908`, 0.5}, {-1.55625566010024, 
                    0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-1.5755933522804908`, 
                    0.5}, {-1.5890739866491415`, 0.5}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.005], Arrowheads[0.04], 
                   ArrowBox[{{-1.5755933522804908`, 
                    0.5}, {-1.5697362944688906`, 0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-1.5755933522804908`, 0.5}, 1.6, {
                    1.568963466499795, 1.5605380558673274`}], 
                   ArrowBox[{{-1.5751801188009742`, 
                    2.0999999466369026`}, {-1.5591804066614314`, 
                    2.099915815040312}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Dashing[{Small, Small}], 
                   Arrowheads[0.02], 
                   CircleBox[{-1.5755933522804908`, 0.5}, 1.7, {
                    1.568963466499795, 1.5723578726171443`}], 
                   ArrowBox[{{-1.5612481504274076`, 
                    2.1999394739765865`}, {-1.5782479790994601`, 
                    2.1999979273388695`}}]}, 
                  {RGBColor[0, 0, 1], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-1.5755933522804908`, 0.5}, 1.8, {
                    1.568963466499795, 1.5743350476609757`}], 
                   ArrowBox[{{-1.5639631307631567`, 
                    2.299962426815476}, {-1.5819630365453037`, 
                    2.29998872972093}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.003], Arrowheads[0.025], 
                   CircleBox[{-1.5755933522804908`, 0.5}, 1.5, {
                    1.568963466499795, 1.5702692528182425`}], 
                   ArrowBox[{{-1.5598030329646004`, 
                    1.999916886302672}, {-1.5748027413521157`, 
                    1.999999791644772}}]}, 
                  {RGBColor[0, 0, 1], ArrowBox[{{0, 2.8}, {0, 2.8}}]}},
                 Frame->True,
                 FrameTicks->Automatic,
                 ImageSize->800,
                 PlotRange->{{-3, 3}, {-1, 3}}], 410->
                GraphicsBox[{InsetBox[
                   TagBox[GridBox[{
                    {"\<\"t\"\>", "20.45`"},
                    {"\<\"driveIntent\"\>", "0"},
                    {"\<\"motorTorque\"\>", 
                    RowBox[{"-", "1.45886579775431`"}]},
                    {"\<\"vw\"\>", 
                    RowBox[{"-", "0.06360928365343456`"}]},
                    {"\<\"controllerState\"\>", 
                    RowBox[{"\[LeftAssociation]", 
                    RowBox[{
                    
                    RowBox[{"\<\"lastError\"\>", "\[Rule]", 
                    "0.016292531066105412`"}], ",", 
                    RowBox[{"\<\"lastInt\"\>", "\[Rule]", 
                    RowBox[{"-", "0.054312877755867066`"}]}], ",", 
                    RowBox[{"\<\"initialized\"\>", "\[Rule]", "True"}]}], 
                    "\[RightAssociation]"}]}
                    },
                    AutoDelete->False,
                    
                    GridBoxItemSize->{
                    "Columns" -> {{Automatic}}, "Rows" -> {{Automatic}}}],
                    "Grid"], {-1, -0.5}], 
                  {GrayLevel[0], LineBox[{{-5, 0}, {5, 0}}]}, 
                  {GrayLevel[0.5], 
                   CircleBox[{-1.5787768813437453`, 0.5}, 0.5], 
                   DiskBox[{-1.5787768813437453`, 0.5}, 0.1]}, 
                  {GrayLevel[0.5], 
                   DiskBox[{-1.5745615067581913`, 1.9999940768606734`}, 0.1]}, 
                  {GrayLevel[0.5], Thickness[0.01], 
                   LineBox[{{-1.5787768813437453`, 
                    0.5}, {-1.5745615067581913`, 1.9999940768606734`}}]}, 
                  {RGBColor[0, 1, 0], 
                   DiskBox[{-1.5759666316200427`, 1.4999960512404489`}, 
                    0.05]}, 
                  {RGBColor[0, 0, 1], Arrowheads[0.01], 
                   CircleBox[{-1.5787768813437453`, 0.5}, 0.3333333333333333, 
                    NCache[{
                    Rational[-1, 2] 
                    Pi, -1.6145623007275258`}, {-1.5707963267948966`, \
-1.6145623007275258`}]], 
                   ArrowBox[{{-1.5900300673337027`, 
                    0.1668566721121575}, {-1.5933608824343601`, 
                    0.16698585912412062`}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-1.5787768813437453`, 
                    0.5}, {-1.5495995653886592`, 0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-1.5787768813437453`, 
                    0.5}, {-1.5994461513151572`, 0.5}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.005], Arrowheads[0.04], 
                   ArrowBox[{{-1.5787768813437453`, 0.5}, {-1.570268835360071,
                     0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-1.5787768813437453`, 0.5}, 1.6, {
                    1.567986073372188, 1.5550677286286183`}], 
                   ArrowBox[{{-1.569611174409477, 
                    2.0999737466022355`}, {-1.55361216188594, 
                    2.099802093039827}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Dashing[{Small, Small}], 
                   Arrowheads[0.02], 
                   CircleBox[{-1.5787768813437453`, 0.5}, 1.7, {
                    1.567986073372188, 1.5729168248313028`}], 
                   ArrowBox[{{-1.5653818666150177`, 
                    2.199947226704528}, {-1.582381725304097, 
                    2.199996177966298}}]}, 
                  {RGBColor[0, 0, 1], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-1.5787768813437453`, 0.5}, 1.8, {
                    1.567986073372188, 1.5760908833597118`}], 
                   ArrowBox[{{-1.5703071144156244`, 
                    2.299980072958638}, {-1.5883070386348492`, 
                    2.2999747709626397`}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.003], Arrowheads[0.025], 
                   CircleBox[{-1.5787768813437453`, 0.5}, 1.5, {
                    1.567986073372188, 1.5695204626177393`}], 
                   ArrowBox[{{-1.5618634434934868`, 
                    1.999904642175657}, {-1.5768630855972314`, 
                    1.9999987791281169`}}]}, 
                  {RGBColor[0, 0, 1], ArrowBox[{{0, 2.8}, {0, 2.8}}]}},
                 Frame->True,
                 FrameTicks->Automatic,
                 ImageSize->800,
                 PlotRange->{{-3, 3}, {-1, 3}}], 411->
                GraphicsBox[{InsetBox[
                   TagBox[GridBox[{
                    {"\<\"t\"\>", "20.5`"},
                    {"\<\"driveIntent\"\>", "0"},
                    {"\<\"motorTorque\"\>", 
                    RowBox[{"-", "1.928704471728274`"}]},
                    {"\<\"vw\"\>", 
                    RowBox[{"-", "0.06344919191446495`"}]},
                    {"\<\"controllerState\"\>", 
                    RowBox[{"\[LeftAssociation]", 
                    RowBox[{
                    
                    RowBox[{"\<\"lastError\"\>", "\[Rule]", 
                    "0.015307314622055802`"}], ",", 
                    RowBox[{"\<\"lastInt\"\>", "\[Rule]", 
                    RowBox[{"-", "0.0535235035176483`"}]}], ",", 
                    RowBox[{"\<\"initialized\"\>", "\[Rule]", "True"}]}], 
                    "\[RightAssociation]"}]}
                    },
                    AutoDelete->False,
                    
                    GridBoxItemSize->{
                    "Columns" -> {{Automatic}}, "Rows" -> {{Automatic}}}],
                    "Grid"], {-1, -0.5}], 
                  {GrayLevel[0], LineBox[{{-5, 0}, {5, 0}}]}, 
                  {GrayLevel[0.5], 
                   CircleBox[{-1.5819534348603725`, 0.5}, 0.5], 
                   DiskBox[{-1.5819534348603725`, 0.5}, 0.1]}, 
                  {GrayLevel[0.5], 
                   DiskBox[{-1.5763332170120203`, 1.9999894710134924`}, 0.1]}, 
                  {GrayLevel[0.5], Thickness[0.01], 
                   LineBox[{{-1.5819534348603725`, 
                    0.5}, {-1.5763332170120203`, 1.9999894710134924`}}]}, 
                  {RGBColor[0, 1, 0], 
                   DiskBox[{-1.578206622961471, 1.4999929806756618`}, 0.05]}, 
                  {RGBColor[0, 0, 1], Arrowheads[0.01], 
                   CircleBox[{-1.5819534348603725`, 0.5}, 0.3333333333333333, 
                    NCache[{
                    Rational[-1, 2] 
                    Pi, -1.6286574609467448`}, {-1.5707963267948966`, \
-1.6286574609467448`}]], 
                   ArrowBox[{{-1.5979010561122582`, 
                    0.16704837515411114`}, {-1.6012297194946392`, 
                    0.16722449615121937`}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-1.5819534348603725`, 
                    0.5}, {-1.5433793454258071`, 0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-1.5819534348603725`, 
                    0.5}, {-1.6095109961048255`, 0.5}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.005], Arrowheads[0.04], 
                   ArrowBox[{{-1.5819534348603725`, 0.5}, {-1.57093690667026, 
                    0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-1.5819534348603725`, 0.5}, 1.6, {
                    1.5670495061292746`, 1.54982590945348}], 
                   ArrowBox[{{-1.5644011191894138`, 
                    2.099903720920289}, {-1.548403226237968, 
                    2.0996482061695296`}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Dashing[{Small, Small}], 
                   Arrowheads[0.02], 
                   CircleBox[{-1.5819534348603725`, 0.5}, 1.7, {
                    1.5670495061292746`, 1.5734340028823013`}], 
                   ArrowBox[{{-1.5694375972773629`, 
                    2.1999539269667268`}, {-1.5864374790094566`, 
                    2.199994086268558}}]}, 
                  {RGBColor[0, 0, 1], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-1.5819534348603725`, 0.5}, 1.8, {
                    1.5670495061292746`, 1.5777645309722095`}], 
                   ArrowBox[{{-1.5764962107398173`, 
                    2.299991727399017}, {-1.5944961008756184`, 
                    2.299956299894314}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.003], Arrowheads[0.025], 
                   CircleBox[{-1.5819534348603725`, 0.5}, 1.5, {
                    1.5670495061292746`, 1.5687714624733793`}], 
                   ArrowBox[{{-1.5639165730658684`, 
                    1.9998915532853054`}, {-1.5789161404536203`, 
                    1.9999969249444103`}}]}, 
                  {RGBColor[0, 0, 1], ArrowBox[{{0, 2.8}, {0, 2.8}}]}},
                 Frame->True,
                 FrameTicks->Automatic,
                 ImageSize->800,
                 PlotRange->{{-3, 3}, {-1, 3}}], 412->
                GraphicsBox[{InsetBox[
                   TagBox[GridBox[{
                    {"\<\"t\"\>", "20.55`"},
                    {"\<\"driveIntent\"\>", "0"},
                    {"\<\"motorTorque\"\>", 
                    RowBox[{"-", "2.3740845862001443`"}]},
                    {"\<\"vw\"\>", 
                    RowBox[{"-", "0.06324912033148104`"}]},
                    {"\<\"controllerState\"\>", 
                    RowBox[{"\[LeftAssociation]", 
                    RowBox[{
                    
                    RowBox[{"\<\"lastError\"\>", "\[Rule]", 
                    "0.01435583343904829`"}], ",", 
                    RowBox[{"\<\"lastInt\"\>", "\[Rule]", 
                    RowBox[{"-", "0.05278255173251672`"}]}], ",", 
                    RowBox[{"\<\"initialized\"\>", "\[Rule]", "True"}]}], 
                    "\[RightAssociation]"}]}
                    },
                    AutoDelete->False,
                    
                    GridBoxItemSize->{
                    "Columns" -> {{Automatic}}, "Rows" -> {{Automatic}}}],
                    "Grid"], {-1, -0.5}], 
                  {GrayLevel[0], LineBox[{{-5, 0}, {5, 0}}]}, 
                  {GrayLevel[0.5], CircleBox[{-1.585120953914982, 0.5}, 0.5], 
                   DiskBox[{-1.585120953914982, 0.5}, 0.1]}, 
                  {GrayLevel[0.5], 
                   DiskBox[{-1.5781646481470306`, 1.9999838698499603`}, 0.1]}, 
                  {GrayLevel[0.5], Thickness[0.01], 
                   LineBox[{{-1.585120953914982, 0.5}, {-1.5781646481470306`, 
                    1.9999838698499603`}}]}, 
                  {RGBColor[0, 1, 0], 
                   DiskBox[{-1.5804834167363477`, 1.4999892465666402`}, 
                    0.05]}, 
                  {RGBColor[0, 0, 1], Arrowheads[0.01], 
                   CircleBox[{-1.585120953914982, 0.5}, 0.3333333333333333, 
                    NCache[{
                    Rational[-1, 2] 
                    Pi, -1.6420188643809008`}, {-1.5707963267948966`, \
-1.6420188643809008`}]], 
                   ArrowBox[{{-1.6055157202626014`, 
                    0.16729117141750338`}, {-1.6088417333674125`, 
                    0.16751175098466975`}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-1.585120953914982, 0.5}, {-1.5376392621909791`,
                     0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-1.585120953914982, 0.5}, {-1.6192296151054624`,
                     0.5}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.005], Arrowheads[0.04], 
                   ArrowBox[{{-1.585120953914982, 0.5}, {-1.5717479233814595`,
                     0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-1.585120953914982, 0.5}, 1.6, {
                    1.566158772993042, 1.5448406305057671`}], 
                   ArrowBox[{{-1.559592923056954, 
                    2.0997963369255825`}, {-1.543596502709994, 
                    2.099461071721386}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Dashing[{Small, Small}], 
                   Arrowheads[0.02], 
                   CircleBox[{-1.585120953914982, 0.5}, 1.7, {
                    1.566158772993042, 1.5739089228277061`}], 
                   ArrowBox[{{-1.573412459739276, 
                    2.199959679275993}, {-1.5904123586266696`, 
                    2.1999917649906946`}}]}, 
                  {RGBColor[0, 0, 1], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-1.585120953914982, 0.5}, 1.8, {
                    1.566158772993042, 1.579348131805265}], 
                   ArrowBox[{{-1.582514203844821, 
                    2.299998112458475}, {-1.6005140153076376`, 
                    2.299934180369094}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.003], Arrowheads[0.025], 
                   CircleBox[{-1.585120953914982, 0.5}, 1.5, {
                    1.566158772993042, 1.5680301547272357`}], 
                   ArrowBox[{{-1.5659722159514518`, 
                    1.9998777702981012`}, {-1.5809717011049738`, 
                    1.9999942612227284`}}]}, 
                  {RGBColor[0, 0, 1], ArrowBox[{{0, 2.8}, {0, 2.8}}]}},
                 Frame->True,
                 FrameTicks->Automatic,
                 ImageSize->800,
                 PlotRange->{{-3, 3}, {-1, 3}}], 413->
                GraphicsBox[{InsetBox[
                   TagBox[GridBox[{
                    {"\<\"t\"\>", "20.6`"},
                    {"\<\"driveIntent\"\>", "0"},
                    {"\<\"motorTorque\"\>", 
                    RowBox[{"-", "2.7931919764756152`"}]},
                    {"\<\"vw\"\>", 
                    RowBox[{"-", "0.06301162295019079`"}]},
                    {"\<\"controllerState\"\>", 
                    RowBox[{"\[LeftAssociation]", 
                    RowBox[{
                    
                    RowBox[{"\<\"lastError\"\>", "\[Rule]", 
                    "0.013442863806332892`"}], ",", 
                    RowBox[{"\<\"lastInt\"\>", "\[Rule]", 
                    RowBox[{"-", "0.0520882099051424`"}]}], ",", 
                    RowBox[{"\<\"initialized\"\>", "\[Rule]", "True"}]}], 
                    "\[RightAssociation]"}]}
                    },
                    AutoDelete->False,
                    
                    GridBoxItemSize->{
                    "Columns" -> {{Automatic}}, "Rows" -> {{Automatic}}}],
                    "Grid"], {-1, -0.5}], 
                  {GrayLevel[0], LineBox[{{-5, 0}, {5, 0}}]}, 
                  {GrayLevel[0.5], 
                   CircleBox[{-1.5882775041686434`, 0.5}, 0.5], 
                   DiskBox[{-1.5882775041686434`, 0.5}, 0.1]}, 
                  {GrayLevel[0.5], 
                   DiskBox[{-1.5800598528053633`, 1.999977489899789}, 0.1]}, 
                  {GrayLevel[0.5], Thickness[0.01], 
                   LineBox[{{-1.5882775041686434`, 
                    0.5}, {-1.5800598528053633`, 1.999977489899789}}]}, 
                  {RGBColor[0, 1, 0], 
                   DiskBox[{-1.5827990699264567`, 1.499984993266526}, 0.05]}, 
                  {RGBColor[0, 0, 1], Arrowheads[0.01], 
                   CircleBox[{-1.5882775041686434`, 0.5}, 0.3333333333333333, 
                    NCache[{
                    Rational[-1, 2] 
                    Pi, -1.654592086089165}, {-1.5707963267948966`, \
-1.654592086089165}]], 
                   ArrowBox[{{-1.6128537701238284`, 
                    0.16757389052180166`}, {-1.616176747011478, 
                    0.16783627025229314`}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-1.5882775041686434`, 0.5}, {-1.532413664639131,
                     0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-1.5882775041686434`, 
                    0.5}, {-1.6285707148609512`, 0.5}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.005], Arrowheads[0.04], 
                   ArrowBox[{{-1.5882775041686434`, 
                    0.5}, {-1.5727068753314388`, 0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-1.5882775041686434`, 0.5}, 1.6, {
                    1.5653178651480777`, 1.540134230541293}], 
                   ArrowBox[{{-1.555220502408142, 
                    2.0996584743733915`}, {-1.539225837109135, 
                    2.0992479276080624`}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Dashing[{Small, Small}], 
                   Arrowheads[0.02], 
                   CircleBox[{-1.5882775041686434`, 0.5}, 1.7, {
                    1.5653178651480777`, 1.5743415626231165`}], 
                   ArrowBox[{{-1.577304481273623, 
                    2.1999645857395222`}, {-1.5943043924515765`, 
                    2.1999893166186735`}}]}, 
                  {RGBColor[0, 0, 1], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-1.5882775041686434`, 0.5}, 1.8, {
                    1.5653178651480777`, 1.58083559835072}], 
                   ArrowBox[{{-1.5883481929691075`, 
                    2.2999999986119706`}, {-1.6063478894223169`, 
                    2.299909292485814}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.003], Arrowheads[0.025], 
                   CircleBox[{-1.5882775041686434`, 0.5}, 1.5, {
                    1.5653178651480777`, 1.5673034585490555`}], 
                   ArrowBox[{{-1.5680388159137348`, 
                    1.9998634589514208`}, {-1.5830382124532356`, 
                    1.9999908499128656`}}]}, 
                  {RGBColor[0, 0, 1], ArrowBox[{{0, 2.8}, {0, 2.8}}]}},
                 Frame->True,
                 FrameTicks->Automatic,
                 ImageSize->800,
                 PlotRange->{{-3, 3}, {-1, 3}}], 414->
                GraphicsBox[{InsetBox[
                   TagBox[GridBox[{
                    {"\<\"t\"\>", "20.650000000000002`"},
                    {"\<\"driveIntent\"\>", "0"},
                    {"\<\"motorTorque\"\>", 
                    RowBox[{"-", "3.184674405725181`"}]},
                    {"\<\"vw\"\>", 
                    RowBox[{"-", "0.06273934850808106`"}]},
                    {"\<\"controllerState\"\>", 
                    RowBox[{"\[LeftAssociation]", 
                    RowBox[{
                    
                    RowBox[{"\<\"lastError\"\>", "\[Rule]", 
                    "0.012572270058839378`"}], ",", 
                    RowBox[{"\<\"lastInt\"\>", "\[Rule]", 
                    RowBox[{"-", "0.0514384502584455`"}]}], ",", 
                    RowBox[{"\<\"initialized\"\>", "\[Rule]", "True"}]}], 
                    "\[RightAssociation]"}]}
                    },
                    AutoDelete->False,
                    
                    GridBoxItemSize->{
                    "Columns" -> {{Automatic}}, "Rows" -> {{Automatic}}}],
                    "Grid"], {-1, -0.5}], 
                  {GrayLevel[0], LineBox[{{-5, 0}, {5, 0}}]}, 
                  {GrayLevel[0.5], 
                   CircleBox[{-1.5914212815583273`, 0.5}, 0.5], 
                   DiskBox[{-1.5914212815583273`, 0.5}, 0.1]}, 
                  {GrayLevel[0.5], 
                   DiskBox[{-1.5820216050125981`, 1.9999705484044799`}, 0.1]}, 
                  {GrayLevel[0.5], Thickness[0.01], 
                   LineBox[{{-1.5914212815583273`, 
                    0.5}, {-1.5820216050125981`, 1.9999705484044799`}}]}, 
                  {RGBColor[0, 1, 0], 
                   DiskBox[{-1.5851548305278411`, 1.4999803656029864`}, 
                    0.05]}, 
                  {RGBColor[0, 0, 1], Arrowheads[0.01], 
                   CircleBox[{-1.5914212815583273`, 0.5}, 0.3333333333333333, 
                    NCache[{
                    Rational[-1, 2] 
                    Pi, -1.666336558966652}, {-1.5707963267948966`, \
-1.666336558966652}]], 
                   ArrowBox[{{-1.6198999322734193`, 
                    0.16788544511786463`}, {-1.6232195985494218`, 
                    0.168186832467961}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-1.5914212815583273`, 
                    0.5}, {-1.5277277934438236`, 0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-1.5914212815583273`, 
                    0.5}, {-1.6375100456140552`, 0.5}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.005], Arrowheads[0.04], 
                   ArrowBox[{{-1.5914212815583273`, 
                    0.5}, {-1.5738165574995515`, 0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-1.5914212815583273`, 0.5}, 1.6, {
                    1.5645298347514258`, 1.5357237916273088`}], 
                   ArrowBox[{{-1.5513094281976596`, 
                    2.099497120728879}, {-1.5353167291478462`, 
                    2.0990160346909654`}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Dashing[{Small, Small}], 
                   Arrowheads[0.02], 
                   CircleBox[{-1.5914212815583273`, 0.5}, 1.7, {
                    1.5645298347514258`, 1.574732309824312}], 
                   ArrowBox[{{-1.5811125158881116`, 
                    2.1999687436392343`}, {-1.5981124354317848`, 
                    2.1999868318489533`}}]}, 
                  {RGBColor[0, 0, 1], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-1.5914212815583273`, 0.5}, 1.8, {
                    1.5645298347514258`, 1.582222470338788}], 
                   ArrowBox[{{-1.5939883390671479`, 
                    2.2999981695034433`}, {-1.6119878924121818`, 
                    2.2998825001977172`}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.003], Arrowheads[0.025], 
                   CircleBox[{-1.5914212815583273`, 0.5}, 1.5, {
                    1.5645298347514258`, 1.5665973565398064`}], 
                   ArrowBox[{{-1.5701235418347605`, 
                    1.9998487944731853`}, {-1.5851228446840555`, 
                    1.9999867765060266`}}]}, 
                  {RGBColor[0, 0, 1], ArrowBox[{{0, 2.8}, {0, 2.8}}]}},
                 Frame->True,
                 FrameTicks->Automatic,
                 ImageSize->800,
                 PlotRange->{{-3, 3}, {-1, 3}}], 415->
                GraphicsBox[{InsetBox[
                   TagBox[GridBox[{
                    {"\<\"t\"\>", "20.7`"},
                    {"\<\"driveIntent\"\>", "0"},
                    {"\<\"motorTorque\"\>", 
                    RowBox[{"-", "3.547601738404407`"}]},
                    {"\<\"vw\"\>", 
                    RowBox[{"-", "0.06243500408175698`"}]},
                    {"\<\"controllerState\"\>", 
                    RowBox[{"\[LeftAssociation]", 
                    RowBox[{
                    
                    RowBox[{"\<\"lastError\"\>", "\[Rule]", 
                    "0.011747071738884389`"}], ",", 
                    RowBox[{"\<\"lastInt\"\>", "\[Rule]", 
                    RowBox[{"-", "0.05083107362196425`"}]}], ",", 
                    RowBox[{"\<\"initialized\"\>", "\[Rule]", "True"}]}], 
                    "\[RightAssociation]"}]}
                    },
                    AutoDelete->False,
                    
                    GridBoxItemSize->{
                    "Columns" -> {{Automatic}}, "Rows" -> {{Automatic}}}],
                    "Grid"], {-1, -0.5}], 
                  {GrayLevel[0], LineBox[{{-5, 0}, {5, 0}}]}, 
                  {GrayLevel[0.5], 
                   CircleBox[{-1.5945506160894116`, 0.5}, 0.5], 
                   DiskBox[{-1.5945506160894116`, 0.5}, 0.1]}, 
                  {GrayLevel[0.5], 
                   DiskBox[{-1.5840515211503259`, 1.9999632558851101`}, 0.1]}, 
                  {GrayLevel[0.5], Thickness[0.01], 
                   LineBox[{{-1.5945506160894116`, 
                    0.5}, {-1.5840515211503259`, 1.9999632558851101`}}]}, 
                  {RGBColor[0, 1, 0], 
                   DiskBox[{-1.5875512194633545`, 1.4999755039234066`}, 
                    0.05]}, 
                  {RGBColor[0, 0, 1], Arrowheads[0.01], 
                   CircleBox[{-1.5945506160894116`, 0.5}, 0.3333333333333333, 
                    NCache[{
                    Rational[-1, 2] 
                    Pi, -1.6772243789470287`}, {-1.5707963267948966`, \
-1.6772243789470287`}]], 
                   ArrowBox[{{-1.6266435108695902`, 
                    0.1682151944170145}, {-1.6299596989968619`, 
                    0.16855270711806314`}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-1.5945506160894116`, 
                    0.5}, {-1.5235985813213235`, 0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-1.5945506160894116`, 0.5}, {-1.646029829711953,
                     0.5}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.005], Arrowheads[0.04], 
                   ArrowBox[{{-1.5945506160894116`, 
                    0.5}, {-1.5750777949438648`, 0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-1.5945506160894116`, 0.5}, 1.6, {
                    1.5637968730156941`, 1.5316215763330738`}], 
                   ArrowBox[{{-1.5478776370840215`, 
                    2.099319115445933}, {-1.5318870461109184`, 
                    2.0987724281452786`}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Dashing[{Small, Small}], 
                   Arrowheads[0.02], 
                   CircleBox[{-1.5945506160894116`, 0.5}, 1.7, {
                    1.5637968730156941`, 1.575081911302435}], 
                   ArrowBox[{{-1.5848361626226037`, 
                    2.1999722437127733`}, {-1.6018360874510664`, 
                    2.1999843887245074`}}]}, 
                  {RGBColor[0, 0, 1], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-1.5945506160894116`, 0.5}, 1.8, {
                    1.5637968730156941`, 1.5835057715623853`}], 
                   ArrowBox[{{-1.5994276107038092`, 
                    2.2999933930221887`}, {-1.6174270007889335`, 
                    2.2998546249692167`}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.003], Arrowheads[0.025], 
                   CircleBox[{-1.5945506160894116`, 0.5}, 1.5, {
                    1.5637968730156941`, 1.5659169448685681`}], 
                   ArrowBox[{{-1.5722323667497367`, 
                    1.9998339560586071`}, {-1.5872315722424144`, 
                    1.99998214375944}}]}, 
                  {RGBColor[0, 0, 1], ArrowBox[{{0, 2.8}, {0, 2.8}}]}},
                 Frame->True,
                 FrameTicks->Automatic,
                 ImageSize->800,
                 PlotRange->{{-3, 3}, {-1, 3}}], 416->
                GraphicsBox[{InsetBox[
                   TagBox[GridBox[{
                    {"\<\"t\"\>", "20.75`"},
                    {"\<\"driveIntent\"\>", "0"},
                    {"\<\"motorTorque\"\>", 
                    RowBox[{"-", "3.881426604053299`"}]},
                    {"\<\"vw\"\>", 
                    RowBox[{"-", "0.062101322317630554`"}]},
                    {"\<\"controllerState\"\>", 
                    RowBox[{"\[LeftAssociation]", 
                    RowBox[{
                    
                    RowBox[{"\<\"lastError\"\>", "\[Rule]", 
                    "0.010969511175137614`"}], ",", 
                    RowBox[{"\<\"lastInt\"\>", "\[Rule]", 
                    RowBox[{"-", "0.05026374994846739`"}]}], ",", 
                    RowBox[{"\<\"initialized\"\>", "\[Rule]", "True"}]}], 
                    "\[RightAssociation]"}]}
                    },
                    AutoDelete->False,
                    
                    GridBoxItemSize->{
                    "Columns" -> {{Automatic}}, "Rows" -> {{Automatic}}}],
                    "Grid"], {-1, -0.5}], 
                  {GrayLevel[0], LineBox[{{-5, 0}, {5, 0}}]}, 
                  {GrayLevel[0.5], 
                   CircleBox[{-1.5976639739022707`, 0.5}, 0.5], 
                   DiskBox[{-1.5976639739022707`, 0.5}, 0.1]}, 
                  {GrayLevel[0.5], 
                   DiskBox[{-1.5861501782060805`, 1.999955810185309}, 0.1]}, 
                  {GrayLevel[0.5], Thickness[0.01], 
                   LineBox[{{-1.5976639739022707`, 
                    0.5}, {-1.5861501782060805`, 1.999955810185309}}]}, 
                  {RGBColor[0, 1, 0], 
                   DiskBox[{-1.5899881101048106`, 1.4999705401235393`}, 
                    0.05]}, 
                  {RGBColor[0, 0, 1], Arrowheads[0.01], 
                   CircleBox[{-1.5976639739022707`, 0.5}, 0.3333333333333333, 
                    NCache[{
                    Rational[-1, 2] 
                    Pi, -1.6872391249164955`}, {-1.5707963267948966`, \
-1.6872391249164955`}]], 
                   ArrowBox[{{-1.6330779443175318`, 
                    0.1685532292953541}, {-1.636390586099961, 
                    0.16892393529764071`}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-1.5976639739022707`, 
                    0.5}, {-1.5200354418212048`, 0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-1.5976639739022707`, 
                    0.5}, {-1.6541181930009752`, 0.5}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.005], Arrowheads[0.04], 
                   ArrowBox[{{-1.5976639739022707`, 
                    0.5}, {-1.5764896609199093`, 0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-1.5976639739022707`, 0.5}, 1.6, {
                    1.5631203876198818`, 1.527835461193618}], 
                   ArrowBox[{{-1.5449361375680473`, 
                    2.099130943755236}, {-1.528947731020833, 
                    2.098523718298937}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Dashing[{Small, Small}], 
                   Arrowheads[0.02], 
                   CircleBox[{-1.5976639739022707`, 0.5}, 1.7, {
                    1.5631203876198818`, 1.5753914252100354`}], 
                   ArrowBox[{{-1.5884756859443423`, 
                    2.199975169043478}, {-1.6054756137175683`, 
                    2.199982052341552}}]}, 
                  {RGBColor[0, 0, 1], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-1.5976639739022707`, 0.5}, 1.8, {
                    1.5631203876198818`, 1.5846838687535112`}], 
                   ArrowBox[{{-1.604661531802084, 
                    2.2999863983328983`}, {-1.6226607459142006`, 
                    2.299826425350229}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.003], Arrowheads[0.025], 
                   CircleBox[{-1.5976639739022707`, 0.5}, 1.5, {
                    1.5631203876198818`, 1.5652664860610632`}], 
                   ArrowBox[{{-1.574370149146265, 
                    1.9998191216704222`}, {-1.5893692550758973`, 
                    1.9999770657045364`}}]}, 
                  {RGBColor[0, 0, 1], ArrowBox[{{0, 2.8}, {0, 2.8}}]}},
                 Frame->True,
                 FrameTicks->Automatic,
                 ImageSize->800,
                 PlotRange->{{-3, 3}, {-1, 3}}], 417->
                GraphicsBox[{InsetBox[
                   TagBox[GridBox[{
                    {"\<\"t\"\>", "20.8`"},
                    {"\<\"driveIntent\"\>", "0"},
                    {"\<\"motorTorque\"\>", 
                    RowBox[{"-", "4.18594587190359`"}]},
                    {"\<\"vw\"\>", 
                    RowBox[{"-", "0.061741032149826965`"}]},
                    {"\<\"controllerState\"\>", 
                    RowBox[{"\[LeftAssociation]", 
                    RowBox[{
                    
                    RowBox[{"\<\"lastError\"\>", "\[Rule]", 
                    "0.01024112079150085`"}], ",", 
                    RowBox[{"\<\"lastInt\"\>", "\[Rule]", 
                    RowBox[{"-", "0.0497340554558086`"}]}], ",", 
                    RowBox[{"\<\"initialized\"\>", "\[Rule]", "True"}]}], 
                    "\[RightAssociation]"}]}
                    },
                    AutoDelete->False,
                    
                    GridBoxItemSize->{
                    "Columns" -> {{Automatic}}, "Rows" -> {{Automatic}}}],
                    "Grid"], {-1, -0.5}], 
                  {GrayLevel[0], LineBox[{{-5, 0}, {5, 0}}]}, 
                  {GrayLevel[0.5], 
                   CircleBox[{-1.6007599577897824`, 0.5}, 0.5], 
                   DiskBox[{-1.6007599577897824`, 0.5}, 0.1]}, 
                  {GrayLevel[0.5], 
                   DiskBox[{-1.5883172284848417`, 1.9999483919413508`}, 0.1]}, 
                  {GrayLevel[0.5], Thickness[0.01], 
                   LineBox[{{-1.6007599577897824`, 
                    0.5}, {-1.5883172284848417`, 1.9999483919413508`}}]}, 
                  {RGBColor[0, 1, 0], 
                   DiskBox[{-1.592464804919822, 1.4999655946275672`}, 0.05]}, 
                  {RGBColor[0, 0, 1], Arrowheads[0.01], 
                   CircleBox[{-1.6007599577897824`, 0.5}, 0.3333333333333333, 
                    NCache[{
                    Rational[-1, 2] 
                    Pi, -1.6963747029520042`}, {-1.5707963267948966`, \
-1.6963747029520042`}]], 
                   ArrowBox[{{-1.639200366132116, 
                    0.1688905828618193}, {-1.6425094831144706`, 
                    0.16929153587143508`}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-1.6007599577897824`, 
                    0.5}, {-1.5170410403517107`, 0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-1.6007599577897824`, 
                    0.5}, {-1.6617686043615034`, 0.5}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.005], Arrowheads[0.04], 
                   ArrowBox[{{-1.6007599577897824`, 
                    0.5}, {-1.5780496869234315`, 0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-1.6007599577897824`, 0.5}, 1.6, {
                    1.562501078791019, 1.5243693627478014`}], 
                   ArrowBox[{{-1.5424897040065557`, 
                    2.098938578408826}, {-1.5265034981993082`, 
                    2.098275939319956}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Dashing[{Small, Small}], 
                   Arrowheads[0.02], 
                   CircleBox[{-1.6007599577897824`, 0.5}, 1.7, {
                    1.562501078791019, 1.5756621753951061`}], 
                   ArrowBox[{{-1.5920319387546416`, 
                    2.19997759446521}, {-1.6090318677684263`, 
                    2.199979875029497}}]}, 
                  {RGBColor[0, 0, 1], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-1.6007599577897824`, 0.5}, 1.8, {
                    1.562501078791019, 1.5857563336349279`}], 
                   ArrowBox[{{-1.6096879334945513`, 
                    2.2999778585443256`}, {-1.6276869656901196`, 
                    2.2997985821323272`}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.003], Arrowheads[0.025], 
                   CircleBox[{-1.6007599577897824`, 0.5}, 1.5, {
                    1.562501078791019, 1.5646494636422428`}], 
                   ArrowBox[{{-1.5765407155044244`, 
                    1.9998044633561813`}, {-1.5915397211238484`, 
                    1.9999716621442634`}}]}, 
                  {RGBColor[0, 0, 1], ArrowBox[{{0, 2.8}, {0, 2.8}}]}},
                 Frame->True,
                 FrameTicks->Automatic,
                 ImageSize->800,
                 PlotRange->{{-3, 3}, {-1, 3}}], 418->
                GraphicsBox[{InsetBox[
                   TagBox[GridBox[{
                    {"\<\"t\"\>", "20.85`"},
                    {"\<\"driveIntent\"\>", "0"},
                    {"\<\"motorTorque\"\>", 
                    RowBox[{"-", "4.46126321388092`"}]},
                    {"\<\"vw\"\>", 
                    RowBox[{"-", "0.061356832888134524`"}]},
                    {"\<\"controllerState\"\>", 
                    RowBox[{"\[LeftAssociation]", 
                    RowBox[{
                    
                    RowBox[{"\<\"lastError\"\>", "\[Rule]", 
                    "0.009562789537168884`"}], ",", 
                    RowBox[{"\<\"lastInt\"\>", "\[Rule]", 
                    RowBox[{"-", "0.04923950642337118`"}]}], ",", 
                    RowBox[{"\<\"initialized\"\>", "\[Rule]", "True"}]}], 
                    "\[RightAssociation]"}]}
                    },
                    AutoDelete->False,
                    
                    GridBoxItemSize->{
                    "Columns" -> {{Automatic}}, "Rows" -> {{Automatic}}}],
                    "Grid"], {-1, -0.5}], 
                  {GrayLevel[0], LineBox[{{-5, 0}, {5, 0}}]}, 
                  {GrayLevel[0.5], 
                   CircleBox[{-1.6038373063371894`, 0.5}, 0.5], 
                   DiskBox[{-1.6038373063371894`, 0.5}, 0.1]}, 
                  {GrayLevel[0.5], 
                   DiskBox[{-1.5905515100895196`, 1.999941161385361}, 0.1]}, 
                  {GrayLevel[0.5], Thickness[0.01], 
                   LineBox[{{-1.6038373063371894`, 
                    0.5}, {-1.5905515100895196`, 1.999941161385361}}]}, 
                  {RGBColor[0, 1, 0], 
                   DiskBox[{-1.5949801088387427`, 1.4999607742569072`}, 
                    0.05]}, 
                  {RGBColor[0, 0, 1], Arrowheads[0.01], 
                   CircleBox[{-1.6038373063371894`, 0.5}, 0.3333333333333333, 
                    NCache[{
                    Rational[-1, 2] 
                    Pi, -1.704634223211324}, {-1.5707963267948966`, \
-1.704634223211324}]], 
                   ArrowBox[{{-1.6450111773353233`, 
                    0.1692193726075541}, {-1.6483168698030246`, 
                    0.16964764334880272`}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-1.6038373063371894`, 0.5}, {-1.514612042059571,
                     0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-1.6038373063371894`, 
                    0.5}, {-1.6689793278788583`, 0.5}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.005], Arrowheads[0.04], 
                   ArrowBox[{{-1.6038373063371894`, 0.5}, {-1.57975406360124, 
                    0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-1.6038373063371894`, 0.5}, 1.6, {
                    1.5619390134845816`, 1.5212236529307657`}], 
                   ArrowBox[{{-1.5405375523725313`, 
                    2.0987473662677334`}, {-1.5245535101277397`, 
                    2.0980344425758224`}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Dashing[{Small, Small}], 
                   Arrowheads[0.02], 
                   CircleBox[{-1.6038373063371894`, 0.5}, 1.7, {
                    1.5619390134845816`, 1.5758957084203247`}], 
                   ArrowBox[{{-1.595506288446882, 
                    2.1999795863894693`}, {-1.6125062175296856`, 
                    2.1999778969088792`}}]}, 
                  {RGBColor[0, 0, 1], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-1.6038373063371894`, 0.5}, 1.8, {
                    1.5619390134845816`, 1.5867238091172535`}], 
                   ArrowBox[{{-1.614506712038831, 
                    2.299968378550572}, {-1.6325055623652676`, 
                    2.2997716886028376`}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.003], Arrowheads[0.025], 
                   CircleBox[{-1.6038373063371894`, 0.5}, 1.5, {
                    1.5619390134845816`, 1.5640686379135593`}], 
                   ArrowBox[{{-1.578746943165784, 
                    1.9997901432120186`}, {-1.5937458491418344`, 
                    1.9999660537797763`}}]}, 
                  {RGBColor[0, 0, 1], ArrowBox[{{0, 2.8}, {0, 2.8}}]}},
                 Frame->True,
                 FrameTicks->Automatic,
                 ImageSize->800,
                 PlotRange->{{-3, 3}, {-1, 3}}], 419->
                GraphicsBox[{InsetBox[
                   TagBox[GridBox[{
                    {"\<\"t\"\>", "20.900000000000002`"},
                    {"\<\"driveIntent\"\>", "0"},
                    {"\<\"motorTorque\"\>", 
                    RowBox[{"-", "4.707752992567215`"}]},
                    {"\<\"vw\"\>", 
                    RowBox[{"-", "0.0609513715409054`"}]},
                    {"\<\"controllerState\"\>", 
                    RowBox[{"\[LeftAssociation]", 
                    RowBox[{
                    
                    RowBox[{"\<\"lastError\"\>", "\[Rule]", 
                    "0.008934827904772227`"}], ",", 
                    RowBox[{"\<\"lastInt\"\>", "\[Rule]", 
                    RowBox[{"-", "0.04877758970094698`"}]}], ",", 
                    RowBox[{"\<\"initialized\"\>", "\[Rule]", "True"}]}], 
                    "\[RightAssociation]"}]}
                    },
                    AutoDelete->False,
                    
                    GridBoxItemSize->{
                    "Columns" -> {{Automatic}}, "Rows" -> {{Automatic}}}],
                    "Grid"], {-1, -0.5}], 
                  {GrayLevel[0], LineBox[{{-5, 0}, {5, 0}}]}, 
                  {GrayLevel[0.5], 
                   CircleBox[{-1.6068948918494401`, 0.5}, 0.5], 
                   DiskBox[{-1.6068948918494401`, 0.5}, 0.1]}, 
                  {GrayLevel[0.5], 
                   DiskBox[{-1.592851152594381, 1.9999342563551696`}, 0.1]}, 
                  {GrayLevel[0.5], Thickness[0.01], 
                   LineBox[{{-1.6068948918494401`, 0.5}, {-1.592851152594381, 
                    1.9999342563551696`}}]}, 
                  {RGBColor[0, 1, 0], 
                   DiskBox[{-1.597532399012734, 1.4999561709034466`}, 0.05]}, 
                  {RGBColor[0, 0, 1], Arrowheads[0.01], 
                   CircleBox[{-1.6068948918494401`, 0.5}, 0.3333333333333333, 
                    NCache[{
                    Rational[-1, 2] 
                    Pi, -1.7120289165719131`}, {-1.5707963267948966`, \
-1.7120289165719131`}]], 
                   ArrowBox[{{-1.6505136361456345`, 
                    0.16953288173082287`}, {-1.6538160713317083`, 
                    0.1699855851222497}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-1.6068948918494401`, 
                    0.5}, {-1.5127398319980958`, 0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-1.6068948918494401`, 
                    0.5}, {-1.6757528915160567`, 0.5}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.005], Arrowheads[0.04], 
                   ArrowBox[{{-1.6068948918494401`, 
                    0.5}, {-1.5815978316647124`, 0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-1.6068948918494401`, 0.5}, 1.6, {
                    1.5614336971725915`, 1.5183955610583328`}], 
                   ArrowBox[{{-1.5390739926181023`, 
                    2.09856195551735}, {-1.5230920305052922`, 
                    2.097803830396752}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Dashing[{Small, Small}], 
                   Arrowheads[0.02], 
                   CircleBox[{-1.6068948918494401`, 0.5}, 1.7, {
                    1.5614336971725915`, 1.576093753306649}], 
                   ArrowBox[{{-1.598900546384251, 
                    2.19998120296684}, {-1.6159004747990777`, 
                    2.1999761467372823`}}]}, 
                  {RGBColor[0, 0, 1], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-1.6068948918494401`, 0.5}, 1.8, {
                    1.5614336971725915`, 1.5875878804646315`}], 
                   ArrowBox[{{-1.6191195944766457`, 
                    2.299958487478441}, {-1.637118268129811, 
                    2.299746245315215}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.003], Arrowheads[0.025], 
                   CircleBox[{-1.6068948918494401`, 0.5}, 1.5, {
                    1.5614336971725915`, 1.5635261022197624`}], 
                   ArrowBox[{{-1.5809908427246662`, 
                    1.9997763100672517`}, {-1.595989651055533, 
                    1.999960358050581}}]}, 
                  {RGBColor[0, 0, 1], ArrowBox[{{0, 2.8}, {0, 2.8}}]}},
                 Frame->True,
                 FrameTicks->Automatic,
                 ImageSize->800,
                 PlotRange->{{-3, 3}, {-1, 3}}], 420->
                GraphicsBox[{InsetBox[
                   TagBox[GridBox[{
                    {"\<\"t\"\>", "20.95`"},
                    {"\<\"driveIntent\"\>", "0"},
                    {"\<\"motorTorque\"\>", 
                    RowBox[{"-", "4.9260256716262445`"}]},
                    {"\<\"vw\"\>", 
                    RowBox[{"-", "0.060527223222853976`"}]},
                    {"\<\"controllerState\"\>", 
                    RowBox[{"\[LeftAssociation]", 
                    RowBox[{
                    
                    RowBox[{"\<\"lastError\"\>", "\[Rule]", 
                    "0.008357031075910473`"}], ",", 
                    RowBox[{"\<\"lastInt\"\>", "\[Rule]", 
                    RowBox[{"-", "0.04834579001268676`"}]}], ",", 
                    RowBox[{"\<\"initialized\"\>", "\[Rule]", "True"}]}], 
                    "\[RightAssociation]"}]}
                    },
                    AutoDelete->False,
                    GridBoxItemSize->{
                    "Columns" -> {{Automatic}}, "Rows" -> {{Automatic}}}],
                    "Grid"], {-1, -0.5}], 
                  {GrayLevel[0], LineBox[{{-5, 0}, {5, 0}}]}, 
                  {GrayLevel[0.5], 
                   CircleBox[{-1.6099317172240029`, 0.5}, 0.5], 
                   DiskBox[{-1.6099317172240029`, 0.5}, 0.1]}, 
                  {GrayLevel[0.5], 
                   DiskBox[{-1.59521367744493, 1.9999277913636582`}, 0.1]}, 
                  {GrayLevel[0.5], Thickness[0.01], 
                   LineBox[{{-1.6099317172240029`, 0.5}, {-1.59521367744493, 
                    1.9999277913636582`}}]}, 
                  {RGBColor[0, 1, 0], 
                   DiskBox[{-1.600119690704621, 1.4999518609091056`}, 0.05]}, 
                  {RGBColor[0, 0, 1], Arrowheads[0.01], 
                   CircleBox[{-1.6099317172240029`, 0.5}, 0.3333333333333333, 
                    NCache[{
                    Rational[-1, 2] 
                    Pi, -1.718577096943684}, {-1.5707963267948966`, \
-1.718577096943684}]], 
                   ArrowBox[{{-1.6557134691995423`, 
                    0.16982558806418496`}, {-1.659012869221584, 
                    0.17029990653671112`}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-1.6099317172240029`, 0.5}, {-1.511411203791478,
                     0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-1.6099317172240029`, 
                    0.5}, {-1.6820955755539515`, 0.5}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.005], Arrowheads[0.04], 
                   ArrowBox[{{-1.6099317172240029`, 
                    0.5}, {-1.5835750621214264`, 0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-1.6099317172240029`, 0.5}, 1.6, {
                    1.5609841428251374`, 1.5158795600753108`}], 
                   ArrowBox[{{-1.5380890534464096`, 
                    2.098386258593754}, {-1.5221090493601042`, 
                    2.097587925282758}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Dashing[{Small, Small}], 
                   Arrowheads[0.02], 
                   CircleBox[{-1.6099317172240029`, 0.5}, 1.7, {
                    1.5609841428251374`, 1.5762581840835586`}], 
                   ArrowBox[{{-1.602216901095563, 
                    2.1999824945016653`}, {-1.6192168284491724`, 
                    2.199974642960752}}]}, 
                  {RGBColor[0, 0, 1], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-1.6099317172240029`, 0.5}, 1.8, {
                    1.5609841428251374`, 1.5883509521119499`}], 
                   ArrowBox[{{-1.6235299134469707`, 
                    2.299948635116981}, {-1.641528419904056, 
                    2.299722658739326}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.003], Arrowheads[0.025], 
                   CircleBox[{-1.6099317172240029`, 0.5}, 1.5, {
                    1.5609841428251374`, 1.5630233391290411`}], 
                   ArrowBox[{{-1.5832736392318476`, 
                    1.9997630969182314`}, {-1.5982723531345557`, 
                    1.9999546857252153`}}]}, 
                  {RGBColor[0, 0, 1], ArrowBox[{{0, 2.8}, {0, 2.8}}]}},
                 Frame->True,
                 FrameTicks->Automatic,
                 ImageSize->800,
                 PlotRange->{{-3, 3}, {-1, 3}}], 421->
                GraphicsBox[{InsetBox[
                   TagBox[GridBox[{
                    {"\<\"t\"\>", "21.`"},
                    {"\<\"driveIntent\"\>", "0"},
                    {"\<\"motorTorque\"\>", 
                    RowBox[{"-", "5.116894909349172`"}]},
                    {"\<\"vw\"\>", 
                    RowBox[{"-", "0.0600868744855606`"}]},
                    {"\<\"controllerState\"\>", 
                    RowBox[{"\[LeftAssociation]", 
                    RowBox[{
                    
                    RowBox[{"\<\"lastError\"\>", "\[Rule]", 
                    "0.00782873980233023`"}], ",", 
                    RowBox[{"\<\"lastInt\"\>", "\[Rule]", 
                    RowBox[{"-", "0.04794161416002253`"}]}], ",", 
                    RowBox[{"\<\"initialized\"\>", "\[Rule]", "True"}]}], 
                    "\[RightAssociation]"}]}
                    },
                    AutoDelete->False,
                    
                    GridBoxItemSize->{
                    "Columns" -> {{Automatic}}, "Rows" -> {{Automatic}}}],
                    "Grid"], {-1, -0.5}], 
                  {GrayLevel[0], LineBox[{{-5, 0}, {5, 0}}]}, 
                  {GrayLevel[0.5], 
                   CircleBox[{-1.6129469119192845`, 0.5}, 0.5], 
                   DiskBox[{-1.6129469119192845`, 0.5}, 0.1]}, 
                  {GrayLevel[0.5], 
                   DiskBox[{-1.59763609271998, 1.9999218575697357`}, 0.1]}, 
                  {GrayLevel[0.5], Thickness[0.01], 
                   LineBox[{{-1.6129469119192845`, 0.5}, {-1.59763609271998, 
                    1.9999218575697357`}}]}, 
                  {RGBColor[0, 1, 0], 
                   DiskBox[{-1.6027396991197482`, 1.4999479050464903`}, 
                    0.05]}, 
                  {RGBColor[0, 0, 1], Arrowheads[0.01], 
                   CircleBox[{-1.6129469119192845`, 0.5}, 0.3333333333333333, 
                    NCache[{
                    Rational[-1, 2] 
                    Pi, -1.7243031740753718`}, {-1.5707963267948966`, \
-1.7243031740753718`}]], 
                   ArrowBox[{{-1.6606185071699266`, 
                    0.17009314933246678`}, {-1.6639151371325023`, 
                    0.17058635254481969`}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-1.6129469119192845`, 
                    0.5}, {-1.5106090137323012`, 0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-1.6129469119192845`, 
                    0.5}, {-1.6880169234972098`, 0.5}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.005], Arrowheads[0.04], 
                   ArrowBox[{{-1.6129469119192845`, 
                    0.5}, {-1.5856790253102262`, 0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-1.6129469119192845`, 0.5}, 1.6, {
                    1.5605889367435746`, 1.5136677351495555`}], 
                   ArrowBox[{{-1.5375690761897098`, 
                    2.0982234455421827`}, {-1.5215908769639097`, 
                    2.0973897692414436`}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Dashing[{Small, Small}], 
                   Arrowheads[0.02], 
                   CircleBox[{-1.6129469119192845`, 0.5}, 1.7, {
                    1.5605889367435746`, 1.5763909851935691`}], 
                   ArrowBox[{{-1.605457855420479, 
                    2.1999835040472475`}, {-1.6224577815814887`, 
                    2.1999733948971874`}}]}, 
                  {RGBColor[0, 0, 1], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-1.6129469119192845`, 0.5}, 1.8, {
                    1.5605889367435746`, 1.5890161306844033`}], 
                   ArrowBox[{{-1.6277423923092105`, 
                    2.299939191683994}, {-1.6457407444698304`, 
                    2.2997012431363846`}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.003], Arrowheads[0.025], 
                   CircleBox[{-1.6129469119192845`, 0.5}, 1.5, {
                    1.5605889367435746`, 1.5625612760156093`}], 
                   ArrowBox[{{-1.5855958515915969`, 
                    1.9997506191027066`}, {-1.6005944753670573`, 
                    1.9999491382414347`}}]}, 
                  {RGBColor[0, 0, 1], ArrowBox[{{0, 2.8}, {0, 2.8}}]}},
                 Frame->True,
                 FrameTicks->Automatic,
                 ImageSize->800,
                 PlotRange->{{-3, 3}, {-1, 3}}], 422->
                GraphicsBox[{InsetBox[
                   TagBox[GridBox[{
                    {"\<\"t\"\>", "21.05`"},
                    {"\<\"driveIntent\"\>", "0"},
                    {"\<\"motorTorque\"\>", 
                    RowBox[{"-", "5.281346461519707`"}]},
                    {"\<\"vw\"\>", 
                    RowBox[{"-", "0.05963270939902401`"}]},
                    {"\<\"controllerState\"\>", 
                    RowBox[{"\[LeftAssociation]", 
                    RowBox[{
                    
                    RowBox[{"\<\"lastError\"\>", "\[Rule]", 
                    "0.007348898696247287`"}], ",", 
                    RowBox[{"\<\"lastInt\"\>", "\[Rule]", 
                    RowBox[{"-", "0.04756261224536846`"}]}], ",", 
                    RowBox[{"\<\"initialized\"\>", "\[Rule]", "True"}]}], 
                    "\[RightAssociation]"}]}
                    },
                    AutoDelete->False,
                    
                    GridBoxItemSize->{
                    "Columns" -> {{Automatic}}, "Rows" -> {{Automatic}}}],
                    "Grid"], {-1, -0.5}], 
                  {GrayLevel[0], LineBox[{{-5, 0}, {5, 0}}]}, 
                  {GrayLevel[0.5], 
                   CircleBox[{-1.6159397271604683`, 0.5}, 0.5], 
                   DiskBox[{-1.6159397271604683`, 0.5}, 0.1]}, 
                  {GrayLevel[0.5], 
                   DiskBox[{-1.600114981986481, 1.9999165234906169`}, 0.1]}, 
                  {GrayLevel[0.5], Thickness[0.01], 
                   LineBox[{{-1.6159397271604683`, 0.5}, {-1.600114981986481, 
                    1.9999165234906169`}}]}, 
                  {RGBColor[0, 1, 0], 
                   DiskBox[{-1.6053898970444767`, 1.4999443489937447`}, 
                    0.05]}, 
                  {RGBColor[0, 0, 1], Arrowheads[0.01], 
                   CircleBox[{-1.6159397271604683`, 0.5}, 0.3333333333333333, 
                    NCache[{
                    Rational[-1, 2] 
                    Pi, -1.7292367206404877`}, {-1.5707963267948966`, \
-1.7292367206404877`}]], 
                   ArrowBox[{{-1.665238346459077, 
                    0.17033235335210412`}, {-1.6685325030707991`, 
                    0.1708418145736661}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-1.6159397271604683`, 
                    0.5}, {-1.5103127979300741`, 0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-1.6159397271604683`, 0.5}, {-1.693529277613841,
                     0.5}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.005], Arrowheads[0.04], 
                   ArrowBox[{{-1.6159397271604683`, 0.5}, {-1.587902348383447,
                     0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-1.6159397271604683`, 0.5}, 1.6, {
                    1.5602463009716612`, 1.5117501330777605`}], 
                   ArrowBox[{{-1.537497275308376, 
                    2.0980759624459138`}, {-1.5215207041184873`, 
                    2.09721164786881}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Dashing[{Small, Small}], 
                   Arrowheads[0.02], 
                   CircleBox[{-1.6159397271604683`, 0.5}, 1.7, {
                    1.5602463009716612`, 1.576494219766565}], 
                   ArrowBox[{{-1.6086261677723983`, 
                    2.199984268118113}, {-1.6256260927992068`, 
                    2.199972403987992}}]}, 
                  {RGBColor[0, 0, 1], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-1.6159397271604683`, 0.5}, 1.8, {
                    1.5602463009716612`, 1.5895871146467706`}], 
                   ArrowBox[{{-1.6317629414944068`, 
                    2.299930450291939}, {-1.6497611548562943`, 
                    2.2996822250132425`}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.003], Arrowheads[0.025], 
                   CircleBox[{-1.6159397271604683`, 0.5}, 1.5, {
                    1.5602463009716612`, 1.562140339595856}], 
                   ArrowBox[{{-1.5879573696184284`, 
                    1.9997389731771291`}, {-1.6029559085011718`, 
                    1.999943805765077}}]}, 
                  {RGBColor[0, 0, 1], ArrowBox[{{0, 2.8}, {0, 2.8}}]}},
                 Frame->True,
                 FrameTicks->Automatic,
                 ImageSize->800,
                 PlotRange->{{-3, 3}, {-1, 3}}], 423->
                GraphicsBox[{InsetBox[
                   TagBox[GridBox[{
                    {"\<\"t\"\>", "21.1`"},
                    {"\<\"driveIntent\"\>", "0"},
                    {"\<\"motorTorque\"\>", 
                    RowBox[{"-", "5.420508987797594`"}]},
                    {"\<\"vw\"\>", 
                    RowBox[{"-", "0.059166998205670625`"}]},
                    {"\<\"controllerState\"\>", 
                    RowBox[{"\[LeftAssociation]", 
                    RowBox[{
                    
                    RowBox[{"\<\"lastError\"\>", "\[Rule]", 
                    "0.006916111664647296`"}], ",", 
                    RowBox[{"\<\"lastInt\"\>", "\[Rule]", 
                    RowBox[{"-", "0.047206396053152355`"}]}], ",", 
                    RowBox[{"\<\"initialized\"\>", "\[Rule]", "True"}]}], 
                    "\[RightAssociation]"}]}
                    },
                    AutoDelete->False,
                    
                    GridBoxItemSize->{
                    "Columns" -> {{Automatic}}, "Rows" -> {{Automatic}}}],
                    "Grid"], {-1, -0.5}], 
                  {GrayLevel[0], LineBox[{{-5, 0}, {5, 0}}]}, 
                  {GrayLevel[0.5], 
                   CircleBox[{-1.6189095305157708`, 0.5}, 0.5], 
                   DiskBox[{-1.6189095305157708`, 0.5}, 0.1]}, 
                  {GrayLevel[0.5], 
                   DiskBox[{-1.602646587064772, 1.999911836299157}, 0.1]}, 
                  {GrayLevel[0.5], Thickness[0.01], 
                   LineBox[{{-1.6189095305157708`, 0.5}, {-1.602646587064772, 
                    1.999911836299157}}]}, 
                  {RGBColor[0, 1, 0], 
                   DiskBox[{-1.6080675682151049`, 1.499941224199438}, 0.05]}, 
                  {RGBColor[0, 0, 1], Arrowheads[0.01], 
                   CircleBox[{-1.6189095305157708`, 0.5}, 0.3333333333333333, 
                    NCache[{
                    Rational[-1, 2] 
                    Pi, -1.7334115964288244`}, {-1.5707963267948966`, \
-1.7334115964288244`}]], 
                   ArrowBox[{{-1.669584037665768, 
                    0.17054104134776366`}, {-1.6728760386384953`, 
                    0.17106425078421317`}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-1.6189095305157708`, 0.5}, {-1.510499350759819,
                     0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-1.6189095305157708`, 
                    0.5}, {-1.6986473407835287`, 0.5}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.005], Arrowheads[0.04], 
                   ArrowBox[{{-1.6189095305157708`, 
                    0.5}, {-1.5902371610275767`, 0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-1.6189095305157708`, 0.5}, 1.6, {
                    1.559954152074232, 1.5101150913261885`}], 
                   ArrowBox[{{-1.5378542637596027`, 
                    2.097945569702262}, {-1.521879127115075, 
                    2.097055133931167}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Dashing[{Small, Small}], 
                   Arrowheads[0.02], 
                   CircleBox[{-1.6189095305157708`, 0.5}, 1.7, {
                    1.559954152074232, 1.5765700007495158`}], 
                   ArrowBox[{{-1.6117247976274165`, 
                    2.199984817465533}, {-1.6287247217063363`, 
                    2.1999716650644188`}}]}, 
                  {RGBColor[0, 0, 1], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-1.6189095305157708`, 0.5}, 1.8, {
                    1.559954152074232, 1.5900680908953295`}], 
                   ArrowBox[{{-1.635598466781722, 
                    2.299922631505674}, {-1.653596558671314, 
                    2.2996657495428803`}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.003], Arrowheads[0.025], 
                   CircleBox[{-1.6189095305157708`, 0.5}, 1.5, {
                    1.559954152074232, 1.561760509026958}], 
                   ArrowBox[{{-1.5903575282985727`, 
                    1.9997282364379851`}, {-1.6053559882977104`, 
                    1.9999387659145769`}}]}, 
                  {RGBColor[0, 0, 1], ArrowBox[{{0, 2.8}, {0, 2.8}}]}},
                 Frame->True,
                 FrameTicks->Automatic,
                 ImageSize->800,
                 PlotRange->{{-3, 3}, {-1, 3}}], 424->
                GraphicsBox[{InsetBox[
                   TagBox[GridBox[{
                    {"\<\"t\"\>", "21.150000000000002`"},
                    {"\<\"driveIntent\"\>", "0"},
                    {"\<\"motorTorque\"\>", 
                    RowBox[{"-", "5.535626826475865`"}]},
                    {"\<\"vw\"\>", 
                    RowBox[{"-", "0.0586918883636595`"}]},
                    {"\<\"controllerState\"\>", 
                    RowBox[{"\[LeftAssociation]", 
                    RowBox[{
                    
                    RowBox[{"\<\"lastError\"\>", "\[Rule]", 
                    "0.006528694279640333`"}], ",", 
                    RowBox[{"\<\"lastInt\"\>", "\[Rule]", 
                    RowBox[{"-", "0.046870654736510364`"}]}], ",", 
                    RowBox[{"\<\"initialized\"\>", "\[Rule]", "True"}]}], 
                    "\[RightAssociation]"}]}
                    },
                    AutoDelete->False,
                    
                    GridBoxItemSize->{
                    "Columns" -> {{Automatic}}, "Rows" -> {{Automatic}}}],
                    "Grid"], {-1, -0.5}], 
                  {GrayLevel[0], LineBox[{{-5, 0}, {5, 0}}]}, 
                  {GrayLevel[0.5], 
                   CircleBox[{-1.6218557999671068`, 0.5}, 0.5], 
                   DiskBox[{-1.6218557999671068`, 0.5}, 0.1]}, 
                  {GrayLevel[0.5], 
                   DiskBox[{-1.605226884601373, 1.9999078235590877`}, 0.1]}, 
                  {GrayLevel[0.5], Thickness[0.01], 
                   LineBox[{{-1.6218557999671068`, 0.5}, {-1.605226884601373, 
                    1.9999078235590877`}}]}, 
                  {RGBColor[0, 1, 0], 
                   DiskBox[{-1.610769856389951, 1.499938549039392}, 0.05]}, 
                  {RGBColor[0, 0, 1], Arrowheads[0.01], 
                   CircleBox[{-1.6218557999671068`, 0.5}, 0.3333333333333333, 
                    NCache[{
                    Rational[-1, 2] 
                    Pi, -1.7368651315891726`}, {-1.5707963267948966`, \
-1.7368651315891726`}]], 
                   ArrowBox[{{-1.673667800758249, 
                    0.1707180119029743}, {-1.6769579751807109`, 
                    0.1712525872377999}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-1.6218557999671068`, 
                    0.5}, {-1.5111432634375894`, 0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-1.6218557999671068`, 0.5}, {-1.703387765877266,
                     0.5}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.005], Arrowheads[0.04], 
                   ArrowBox[{{-1.6218557999671068`, 
                    0.5}, {-1.5926752293477486`, 0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-1.6218557999671068`, 0.5}, 1.6, {
                    1.5597101561315316`, 1.5087495458592277`}], 
                   ArrowBox[{{-1.5386185421323877`, 
                    2.097833395228726}, {-1.5226446363125443`, 
                    2.0969211455191856`}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Dashing[{Small, Small}], 
                   Arrowheads[0.02], 
                   CircleBox[{-1.6218557999671068`, 0.5}, 1.7, {
                    1.5597101561315316`, 1.5766204648508315`}], 
                   ArrowBox[{{-1.6147568552939282`, 
                    2.1999851778720094`}, {-1.6317567786874778`, 
                    2.1999711675850206`}}]}, 
                  {RGBColor[0, 0, 1], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-1.6218557999671068`, 0.5}, 1.8, {
                    1.5597101561315316`, 1.590463638500842}], 
                   ArrowBox[{{-1.6392566899959347`, 
                    2.299915889431005}, {-1.6572546788685119`, 
                    2.2996518883863413`}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.003], Arrowheads[0.025], 
                   CircleBox[{-1.6218557999671068`, 0.5}, 1.5, {
                    1.5597101561315316`, 1.5614213672327864`}], 
                   ArrowBox[{{-1.5927951788767682`, 
                    1.9997184670136738`}, {-1.607793566614023, 
                    1.9999340830826946`}}]}, 
                  {RGBColor[0, 0, 1], ArrowBox[{{0, 2.8}, {0, 2.8}}]}},
                 Frame->True,
                 FrameTicks->Automatic,
                 ImageSize->800,
                 PlotRange->{{-3, 3}, {-1, 3}}], 425->
                GraphicsBox[{InsetBox[
                   TagBox[GridBox[{
                    {"\<\"t\"\>", "21.2`"},
                    {"\<\"driveIntent\"\>", "0"},
                    {"\<\"motorTorque\"\>", 
                    RowBox[{"-", "5.628034775698921`"}]},
                    {"\<\"vw\"\>", 
                    RowBox[{"-", "0.05820939779395225`"}]},
                    {"\<\"controllerState\"\>", 
                    RowBox[{"\[LeftAssociation]", 
                    RowBox[{
                    
                    RowBox[{"\<\"lastError\"\>", "\[Rule]", 
                    "0.006184722929955729`"}], ",", 
                    RowBox[{"\<\"lastInt\"\>", "\[Rule]", 
                    RowBox[{"-", "0.04655316796700431`"}]}], ",", 
                    RowBox[{"\<\"initialized\"\>", "\[Rule]", "True"}]}], 
                    "\[RightAssociation]"}]}
                    },
                    AutoDelete->False,
                    
                    GridBoxItemSize->{
                    "Columns" -> {{Automatic}}, "Rows" -> {{Automatic}}}],
                    "Grid"], {-1, -0.5}], 
                  {GrayLevel[0], LineBox[{{-5, 0}, {5, 0}}]}, 
                  {GrayLevel[0.5], 
                   CircleBox[{-1.6247781175898794`, 0.5}, 0.5], 
                   DiskBox[{-1.6247781175898794`, 0.5}, 0.1]}, 
                  {GrayLevel[0.5], 
                   DiskBox[{-1.607851656418081, 1.9999044952636815`}, 0.1]}, 
                  {GrayLevel[0.5], Thickness[0.01], 
                   LineBox[{{-1.6247781175898794`, 0.5}, {-1.607851656418081, 
                    1.9999044952636815`}}]}, 
                  {RGBColor[0, 1, 0], 
                   DiskBox[{-1.6134938101420138`, 1.4999363301757878`}, 
                    0.05]}, 
                  {RGBColor[0, 0, 1], Arrowheads[0.01], 
                   CircleBox[{-1.6247781175898794`, 0.5}, 0.3333333333333333, 
                    NCache[{
                    Rational[-1, 2] 
                    Pi, -1.7396373700658643`}, {-1.5707963267948966`, \
-1.7396373700658643`}]], 
                   ArrowBox[{{-1.6775027663134359`, 
                    0.17086291225677913`}, {-1.6807914461244935`, 
                    0.17140660667386431`}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-1.6247781175898794`, 0.5}, {-1.512217422075901,
                     0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-1.6247781175898794`, 
                    0.5}, {-1.7077687734781641`, 0.5}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.005], Arrowheads[0.04], 
                   ArrowBox[{{-1.6247781175898794`, 
                    0.5}, {-1.5952080779641857`, 0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-1.6247781175898794`, 0.5}, 1.6, {
                    1.5595117798509741`, 1.5076393172102185`}], 
                   ArrowBox[{{-1.5397669510073018`, 
                    2.0977399981086626`}, {-1.5237940678377913`, 
                    2.096810014277111}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Dashing[{Small, Small}], 
                   Arrowheads[0.02], 
                   CircleBox[{-1.6247781175898794`, 0.5}, 1.7, {
                    1.5595117798509741`, 1.5766477492345572`}], 
                   ArrowBox[{{-1.6177255559672585`, 
                    2.199985370929573}, {-1.6347254789722183`, 
                    2.1999708968101572`}}]}, 
                  {RGBColor[0, 0, 1], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-1.6247781175898794`, 0.5}, 1.8, {
                    1.5595117798509741`, 1.5907786397159682`}], 
                   ArrowBox[{{-1.6427459824383184`, 
                    2.2999103188305767`}, {-1.660743887257315, 
                    2.2996406484107403`}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.003], Arrowheads[0.025], 
                   CircleBox[{-1.6247781175898794`, 0.5}, 1.5, {
                    1.5595117798509741`, 1.5611221501735548`}], 
                   ArrowBox[{{-1.595268756457773, 
                    1.9997097044446885`}, {-1.6102670790076132`, 
                    1.9999298082774621`}}]}, 
                  {RGBColor[0, 0, 1], ArrowBox[{{0, 2.8}, {0, 2.8}}]}},
                 Frame->True,
                 FrameTicks->Automatic,
                 ImageSize->800,
                 PlotRange->{{-3, 3}, {-1, 3}}], 426->
                GraphicsBox[{InsetBox[
                   TagBox[GridBox[{
                    {"\<\"t\"\>", "21.25`"},
                    {"\<\"driveIntent\"\>", "0"},
                    {"\<\"motorTorque\"\>", 
                    RowBox[{"-", "5.699134895174124`"}]},
                    {"\<\"vw\"\>", 
                    RowBox[{"-", "0.057721410145265224`"}]},
                    {"\<\"controllerState\"\>", 
                    RowBox[{"\[LeftAssociation]", 
                    RowBox[{
                    
                    RowBox[{"\<\"lastError\"\>", "\[Rule]", 
                    "0.005882080647352217`"}], ",", 
                    RowBox[{"\<\"lastInt\"\>", "\[Rule]", 
                    RowBox[{"-", "0.046251816711207226`"}]}], ",", 
                    RowBox[{"\<\"initialized\"\>", "\[Rule]", "True"}]}], 
                    "\[RightAssociation]"}]}
                    },
                    AutoDelete->False,
                    
                    GridBoxItemSize->{
                    "Columns" -> {{Automatic}}, "Rows" -> {{Automatic}}}],
                    "Grid"], {-1, -0.5}], 
                  {GrayLevel[0], LineBox[{{-5, 0}, {5, 0}}]}, 
                  {GrayLevel[0.5], CircleBox[{-1.627676162947364, 0.5}, 0.5], 
                   DiskBox[{-1.627676162947364, 0.5}, 0.1]}, 
                  {GrayLevel[0.5], 
                   DiskBox[{-1.6105165536701818`, 1.9999018460584195`}, 0.1]}, 
                  {GrayLevel[0.5], Thickness[0.01], 
                   LineBox[{{-1.627676162947364, 0.5}, {-1.6105165536701818`, 
                    1.9999018460584195`}}]}, 
                  {RGBColor[0, 1, 0], 
                   DiskBox[{-1.6162364234292426`, 1.4999345640389463`}, 
                    0.05]}, 
                  {RGBColor[0, 0, 1], Arrowheads[0.01], 
                   CircleBox[{-1.627676162947364, 0.5}, 0.3333333333333333, 
                    NCache[{
                    Rational[-1, 2] 
                    Pi, -1.7417703736501202`}, {-1.5707963267948966`, \
-1.7417703736501202`}]], 
                   ArrowBox[{{-1.6811027417856999`, 
                    0.17097612277413332`}, {-1.6843902544142388`, 
                    0.17152683071489988`}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-1.627676162947364, 0.5}, {-1.5136934650438816`,
                     0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-1.627676162947364, 0.5}, {-1.7118097983808396`,
                     0.5}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.005], Arrowheads[0.04], 
                   ArrowBox[{{-1.627676162947364, 0.5}, {-1.597827100477357, 
                    0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-1.627676162947364, 0.5}, 1.6, {
                    1.5593563377467938`, 1.5067693745223942`}], 
                   ArrowBox[{{-1.5412750864799407`, 
                    2.097665438690238}, {-1.525303018387103, 
                    2.0967215597194286`}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Dashing[{Small, Small}], 
                   Arrowheads[0.02], 
                   CircleBox[{-1.627676162947364, 0.5}, 1.7, {
                    1.5593563377467938`, 1.5766539708794496`}], 
                   ArrowBox[{{-1.6206341780301796`, 
                    2.199985414775205}, {-1.6376341009447575`, 
                    2.1999708348883047`}}]}, 
                  {RGBColor[0, 0, 1], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-1.627676162947364, 0.5}, 1.8, {
                    1.5593563377467938`, 1.5910181982755391`}], 
                   ArrowBox[{{-1.646075211199442, 
                    2.2999059628278964`}, {-1.6640730509001473`, 
                    2.2996319808636856`}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.003], Arrowheads[0.025], 
                   CircleBox[{-1.627676162947364, 0.5}, 1.5, {
                    1.5593563377467938`, 1.5608617938256046`}], 
                   ArrowBox[{{-1.5977763438781865`, 
                    1.9997019706660488`}, {-1.6127746086142634`, 
                    1.9999259793998028`}}]}, 
                  {RGBColor[0, 0, 1], ArrowBox[{{0, 2.8}, {0, 2.8}}]}},
                 Frame->True,
                 FrameTicks->Automatic,
                 ImageSize->800,
                 PlotRange->{{-3, 3}, {-1, 3}}], 427->
                GraphicsBox[{InsetBox[
                   TagBox[GridBox[{
                    {"\<\"t\"\>", "21.3`"},
                    {"\<\"driveIntent\"\>", "0"},
                    {"\<\"motorTorque\"\>", 
                    RowBox[{"-", "5.750375320922553`"}]},
                    {"\<\"vw\"\>", 
                    RowBox[{"-", "0.05722967189249881`"}]},
                    {"\<\"controllerState\"\>", 
                    RowBox[{"\[LeftAssociation]", 
                    RowBox[{
                    
                    RowBox[{"\<\"lastError\"\>", "\[Rule]", 
                    "0.00561849954604487`"}], ",", 
                    RowBox[{"\<\"lastInt\"\>", "\[Rule]", 
                    RowBox[{"-", "0.045964591802165194`"}]}], ",", 
                    RowBox[{"\<\"initialized\"\>", "\[Rule]", "True"}]}], 
                    "\[RightAssociation]"}]}
                    },
                    AutoDelete->False,
                    
                    GridBoxItemSize->{
                    "Columns" -> {{Automatic}}, "Rows" -> {{Automatic}}}],
                    "Grid"], {-1, -0.5}], 
                  {GrayLevel[0], LineBox[{{-5, 0}, {5, 0}}]}, 
                  {GrayLevel[0.5], 
                   CircleBox[{-1.6305497062958456`, 0.5}, 0.5], 
                   DiskBox[{-1.6305497062958456`, 0.5}, 0.1]}, 
                  {GrayLevel[0.5], 
                   DiskBox[{-1.6132171549030834`, 1.9998998575445686`}, 0.1]}, 
                  {GrayLevel[0.5], Thickness[0.01], 
                   LineBox[{{-1.6305497062958456`, 
                    0.5}, {-1.6132171549030834`, 1.9998998575445686`}}]}, 
                  {RGBColor[0, 1, 0], 
                   DiskBox[{-1.618994672034004, 1.4999332383630457`}, 0.05]}, 
                  {RGBColor[0, 0, 1], Arrowheads[0.01], 
                   CircleBox[{-1.6305497062958456`, 0.5}, 0.3333333333333333, 
                    NCache[{
                    Rational[-1, 2] 
                    Pi, -1.743307586422573}, {-1.5707963267948966`, \
-1.743307586422573}]], 
                   ArrowBox[{{-1.6844820015173925`, 
                    0.17105863950660294`}, {-1.6877686637067513`, 
                    0.1716144004011137}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-1.6305497062958456`, 
                    0.5}, {-1.5155421998773946`, 0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-1.6305497062958456`, 0.5}, {-1.715531164973672,
                     0.5}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.005], Arrowheads[0.04], 
                   ArrowBox[{{-1.6305497062958456`, 0.5}, {-1.600523658555221,
                     0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-1.6305497062958456`, 0.5}, 1.6, {
                    1.5592410353818538`, 1.5061240775331564`}], 
                   ArrowBox[{{-1.5431176791916348`, 
                    2.0976093516991092`}, {-1.5271462235064628`, 
                    2.096655166198082}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Dashing[{Small, Small}], 
                   Arrowheads[0.02], 
                   CircleBox[{-1.6305497062958456`, 0.5}, 1.7, {
                    1.5592410353818538`, 1.576641208500318}], 
                   ArrowBox[{{-1.6234860255208539`, 
                    2.1999853247642784`}, {-1.6404859486201215`, 
                    2.199970961836841}}]}, 
                  {RGBColor[0, 0, 1], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-1.6305497062958456`, 0.5}, 1.8, {
                    1.5592410353818538`, 1.5911875649425347`}], 
                   ArrowBox[{{-1.6492535983564067`, 
                    2.2999028208272208`}, {-1.6672513913955656`, 
                    2.299625790632831}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.003], Arrowheads[0.025], 
                   CircleBox[{-1.6305497062958456`, 0.5}, 1.5, {
                    1.5592410353818538`, 1.560638978683653}], 
                   ArrowBox[{{-1.6003157316641659`, 
                    1.9996952713061313`}, {-1.6153139461153991`, 
                    1.9999226218747832`}}]}, 
                  {RGBColor[0, 0, 1], ArrowBox[{{0, 2.8}, {0, 2.8}}]}},
                 Frame->True,
                 FrameTicks->Automatic,
                 ImageSize->800,
                 PlotRange->{{-3, 3}, {-1, 3}}], 428->
                GraphicsBox[{InsetBox[
                   TagBox[GridBox[{
                    {"\<\"t\"\>", "21.35`"},
                    {"\<\"driveIntent\"\>", "0"},
                    {"\<\"motorTorque\"\>", 
                    RowBox[{"-", "5.783231066727782`"}]},
                    {"\<\"vw\"\>", 
                    RowBox[{"-", "0.05673579108734607`"}]},
                    {"\<\"controllerState\"\>", 
                    RowBox[{"\[LeftAssociation]", 
                    RowBox[{
                    
                    RowBox[{"\<\"lastError\"\>", "\[Rule]", 
                    "0.005391599853222873`"}], ",", 
                    RowBox[{"\<\"lastInt\"\>", "\[Rule]", 
                    RowBox[{"-", "0.045689600475801236`"}]}], ",", 
                    RowBox[{"\<\"initialized\"\>", "\[Rule]", "True"}]}], 
                    "\[RightAssociation]"}]}
                    },
                    AutoDelete->False,
                    
                    GridBoxItemSize->{
                    "Columns" -> {{Automatic}}, "Rows" -> {{Automatic}}}],
                    "Grid"], {-1, -0.5}], 
                  {GrayLevel[0], LineBox[{{-5, 0}, {5, 0}}]}, 
                  {GrayLevel[0.5], 
                   CircleBox[{-1.6333986016875608`, 0.5}, 0.5], 
                   DiskBox[{-1.6333986016875608`, 0.5}, 0.1]}, 
                  {GrayLevel[0.5], 
                   DiskBox[{-1.6159490181459284`, 1.9998985005773635`}, 0.1]}, 
                  {GrayLevel[0.5], Thickness[0.01], 
                   LineBox[{{-1.6333986016875608`, 
                    0.5}, {-1.6159490181459284`, 1.9998985005773635`}}]}, 
                  {RGBColor[0, 1, 0], 
                   DiskBox[{-1.6217655459931393`, 1.4999323337182424`}, 
                    0.05]}, 
                  {RGBColor[0, 0, 1], Arrowheads[0.01], 
                   CircleBox[{-1.6333986016875608`, 0.5}, 0.3333333333333333, 
                    NCache[{
                    Rational[-1, 2] 
                    Pi, -1.7442932587967301`}, {-1.5707963267948966`, \
-1.7442932587967301`}]], 
                   ArrowBox[{{-1.6876550990694037`, 
                    0.17111195886295105`}, {-1.6909412118641123`, 
                    0.1716709590590858}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-1.6333986016875608`, 
                    0.5}, {-1.5177339803530052`, 0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-1.6333986016875608`, 
                    0.5}, {-1.7189537913135082`, 0.5}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.005], Arrowheads[0.04], 
                   ArrowBox[{{-1.6333986016875608`, 
                    0.5}, {-1.6032891699789527`, 0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-1.6333986016875608`, 0.5}, 1.6, {
                    1.5591630087048314`, 1.5056873966927846`}], 
                   ArrowBox[{{-1.54526893754745, 
                    2.0975710194850654`}, {-1.529297900059591, 
                    2.0966098596465463`}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Dashing[{Small, Small}], 
                   Arrowheads[0.02], 
                   CircleBox[{-1.6333986016875608`, 0.5}, 1.7, {
                    1.5591630087048314`, 1.576611486914913}], 
                   ArrowBox[{{-1.6262843946566792`, 
                    2.1999851140696265`}, {-1.6432843181753272`, 
                    2.1999712564068616`}}]}, 
                  {RGBColor[0, 0, 1], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-1.6333986016875608`, 0.5}, 1.8, {
                    1.5591630087048314`, 1.5912920701866524`}], 
                   ArrowBox[{{-1.6522905929293226`, 
                    2.299900856343738}, {-1.6702883569190934`, 
                    2.2996219452871087`}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.003], Arrowheads[0.025], 
                   CircleBox[{-1.6333986016875608`, 0.5}, 1.5, {
                    1.5591630087048314`, 1.5604521716415924`}], 
                   ArrowBox[{{-1.602884473947421, 
                    1.9996895972194575`}, {-1.617882645666272, 
                    1.999919749556204}}]}, 
                  {RGBColor[0, 0, 1], ArrowBox[{{0, 2.8}, {0, 2.8}}]}},
                 Frame->True,
                 FrameTicks->Automatic,
                 ImageSize->800,
                 PlotRange->{{-3, 3}, {-1, 3}}], 429->
                GraphicsBox[{InsetBox[
                   TagBox[GridBox[{
                    {"\<\"t\"\>", "21.400000000000002`"},
                    {"\<\"driveIntent\"\>", "0"},
                    {"\<\"motorTorque\"\>", 
                    RowBox[{"-", "5.799186769490511`"}]},
                    {"\<\"vw\"\>", 
                    RowBox[{"-", "0.05624123758432931`"}]},
                    {"\<\"controllerState\"\>", 
                    RowBox[{"\[LeftAssociation]", 
                    RowBox[{
                    
                    RowBox[{"\<\"lastError\"\>", "\[Rule]", 
                    "0.005198925544410732`"}], ",", 
                    RowBox[{"\<\"lastInt\"\>", "\[Rule]", 
                    RowBox[{"-", "0.04542507104249571`"}]}], ",", 
                    RowBox[{"\<\"initialized\"\>", "\[Rule]", "True"}]}], 
                    "\[RightAssociation]"}]}
                    },
                    AutoDelete->False,
                    
                    GridBoxItemSize->{
                    "Columns" -> {{Automatic}}, "Rows" -> {{Automatic}}}],
                    "Grid"], {-1, -0.5}], 
                  {GrayLevel[0], LineBox[{{-5, 0}, {5, 0}}]}, 
                  {GrayLevel[0.5], 
                   CircleBox[{-1.6362227800495155`, 0.5}, 0.5], 
                   DiskBox[{-1.6362227800495155`, 0.5}, 0.1]}, 
                  {GrayLevel[0.5], 
                   DiskBox[{-1.6187077272229877`, 1.9998977374889544`}, 0.1]}, 
                  {GrayLevel[0.5], Thickness[0.01], 
                   LineBox[{{-1.6362227800495155`, 
                    0.5}, {-1.6187077272229877`, 1.9998977374889544`}}]}, 
                  {RGBColor[0, 1, 0], 
                   DiskBox[{-1.6245460781651635`, 1.499931824992636}, 0.05]}, 
                  {RGBColor[0, 0, 1], Arrowheads[0.01], 
                   CircleBox[{-1.6362227800495155`, 0.5}, 0.3333333333333333, 
                    NCache[{
                    Rational[-1, 2] 
                    Pi, -1.7447719298796118`}, {-1.5707963267948966`, \
-1.7447719298796118`}]], 
                   ArrowBox[{{-1.6906367004043572`, 
                    0.1711379675567153}, {-1.69392254524538, 
                    0.1716985406559191}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-1.6362227800495155`, 
                    0.5}, {-1.5202390446597054`, 0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-1.6362227800495155`, 
                    0.5}, {-1.7220989214444131`, 0.5}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.005], Arrowheads[0.04], 
                   ArrowBox[{{-1.6362227800495155`, 
                    0.5}, {-1.6061151860546028`, 0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-1.6362227800495155`, 0.5}, 1.6, {
                    1.5591193595502293`, 1.5054431117998326`}], 
                   ArrowBox[{{-1.5477028560773816`, 
                    2.097549443072097}, {-1.5317320538628851`, 
                    2.0965843817791123`}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Dashing[{Small, Small}], 
                   Arrowheads[0.02], 
                   CircleBox[{-1.6362227800495155`, 0.5}, 1.7, {
                    1.5591193595502293`, 1.5765667637280771`}], 
                   ArrowBox[{{-1.6290325442738993`, 
                    2.1999847941994926`}, {-1.6460324683953047`, 
                    2.1999716968274967`}}]}, 
                  {RGBColor[0, 0, 1], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-1.6362227800495155`, 0.5}, 1.8, {
                    1.5591193595502293`, 1.5913370638251765`}], 
                   ArrowBox[{{-1.6551957553606367`, 
                    2.2999000045024287`}, {-1.6731935067829666`, 
                    2.299620283661195}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.003], Arrowheads[0.025], 
                   CircleBox[{-1.6362227800495155`, 0.5}, 1.5, {
                    1.5591193595502293`, 1.5602996651479706`}], 
                   ArrowBox[{{-1.6054799402631215`, 
                    1.9996849261767848`}, {-1.6204780767078313`, 
                    1.9999173658294254`}}]}, 
                  {RGBColor[0, 0, 1], ArrowBox[{{0, 2.8}, {0, 2.8}}]}},
                 Frame->True,
                 FrameTicks->Automatic,
                 ImageSize->800,
                 PlotRange->{{-3, 3}, {-1, 3}}], 430->
                GraphicsBox[{InsetBox[
                   TagBox[GridBox[{
                    {"\<\"t\"\>", "21.45`"},
                    {"\<\"driveIntent\"\>", "0"},
                    {"\<\"motorTorque\"\>", 
                    RowBox[{"-", "5.7997213216293755`"}]},
                    {"\<\"vw\"\>", 
                    RowBox[{"-", "0.0557473445712931`"}]},
                    {"\<\"controllerState\"\>", 
                    RowBox[{"\[LeftAssociation]", 
                    RowBox[{
                    
                    RowBox[{"\<\"lastError\"\>", "\[Rule]", 
                    "0.005037976628909854`"}], ",", 
                    RowBox[{"\<\"lastInt\"\>", "\[Rule]", 
                    RowBox[{"-", "0.045169355862570454`"}]}], ",", 
                    RowBox[{"\<\"initialized\"\>", "\[Rule]", "True"}]}], 
                    "\[RightAssociation]"}]}
                    },
                    AutoDelete->False,
                    
                    GridBoxItemSize->{
                    "Columns" -> {{Automatic}}, "Rows" -> {{Automatic}}}],
                    "Grid"], {-1, -0.5}], 
                  {GrayLevel[0], LineBox[{{-5, 0}, {5, 0}}]}, 
                  {GrayLevel[0.5], 
                   CircleBox[{-1.6390222423075724`, 0.5}, 0.5], 
                   DiskBox[{-1.6390222423075724`, 0.5}, 0.1]}, 
                  {GrayLevel[0.5], 
                   DiskBox[{-1.6214889324992439`, 1.9998975241819572`}, 0.1]}, 
                  {GrayLevel[0.5], Thickness[0.01], 
                   LineBox[{{-1.6390222423075724`, 
                    0.5}, {-1.6214889324992439`, 1.9998975241819572`}}]}, 
                  {RGBColor[0, 1, 0], 
                   DiskBox[{-1.6273333691020202`, 1.4999316827879716`}, 
                    0.05]}, 
                  {RGBColor[0, 0, 1], Arrowheads[0.01], 
                   CircleBox[{-1.6390222423075724`, 0.5}, 0.3333333333333333, 
                    NCache[{
                    Rational[-1, 2] 
                    Pi, -1.7447879664437778`}, {-1.5707963267948966`, \
-1.7447879664437778`}]], 
                   ArrowBox[{{-1.6934414364725021`, 
                    0.17113884021132753`}, {-1.696727272323436, 
                    0.1716994660041209}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-1.6390222423075724`, 
                    0.5}, {-1.5230278158749848`, 0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-1.6390222423075724`, 0.5}, {-1.724987885289186,
                     0.5}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.005], Arrowheads[0.04], 
                   ArrowBox[{{-1.6390222423075724`, 
                    0.5}, {-1.6089934588565986`, 0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-1.6390222423075724`, 0.5}, 1.6, {
                    1.559107187398327, 1.5053749897008757`}], 
                   ArrowBox[{{-1.5503934901196668`, 
                    2.097543409202271}, {-1.5344227536842285`, 
                    2.0965772599469576`}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Dashing[{Small, Small}], 
                   Arrowheads[0.02], 
                   CircleBox[{-1.6390222423075724`, 0.5}, 1.7, {
                    1.559107187398327, 1.57650891819639}], 
                   ArrowBox[{{-1.6317336700197602`, 
                    2.199984375432317}, {-1.6487335948703483`, 
                    2.199972261429992}}]}, 
                  {RGBColor[0, 0, 1], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-1.6390222423075724`, 0.5}, 1.8, {
                    1.559107187398327, 1.5913278614073791`}], 
                   ArrowBox[{{-1.6579786541860555`, 
                    2.2999001790234628`}, {-1.6759764081817314`, 
                    2.299620623805069}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.003], Arrowheads[0.025], 
                   CircleBox[{-1.6390222423075724`, 0.5}, 1.5, {
                    1.559107187398327, 1.560179613569713}], 
                   ArrowBox[{{-1.608099363200873, 
                    1.9996812246433415`}, {-1.6230974716327564`, 
                    1.999915464844254}}]}, 
                  {RGBColor[0, 0, 1], ArrowBox[{{0, 2.8}, {0, 2.8}}]}},
                 Frame->True,
                 FrameTicks->Automatic,
                 ImageSize->800,
                 PlotRange->{{-3, 3}, {-1, 3}}], 431->
                GraphicsBox[{InsetBox[
                   TagBox[GridBox[{
                    {"\<\"t\"\>", "21.5`"},
                    {"\<\"driveIntent\"\>", "0"},
                    {"\<\"motorTorque\"\>", 
                    RowBox[{"-", "5.786294321799549`"}]},
                    {"\<\"vw\"\>", 
                    RowBox[{"-", "0.05525531124020842`"}]},
                    {"\<\"controllerState\"\>", 
                    RowBox[{"\[LeftAssociation]", 
                    RowBox[{
                    
                    RowBox[{"\<\"lastError\"\>", "\[Rule]", 
                    "0.004906238157990253`"}], ",", 
                    RowBox[{"\<\"lastInt\"\>", "\[Rule]", 
                    RowBox[{"-", "0.044920932791424964`"}]}], ",", 
                    RowBox[{"\<\"initialized\"\>", "\[Rule]", "True"}]}], 
                    "\[RightAssociation]"}]}
                    },
                    AutoDelete->False,
                    
                    GridBoxItemSize->{
                    "Columns" -> {{Automatic}}, "Rows" -> {{Automatic}}}],
                    "Grid"], {-1, -0.5}], 
                  {GrayLevel[0], LineBox[{{-5, 0}, {5, 0}}]}, 
                  {GrayLevel[0.5], 
                   CircleBox[{-1.6417970526168013`, 0.5}, 0.5], 
                   DiskBox[{-1.6417970526168013`, 0.5}, 0.1]}, 
                  {GrayLevel[0.5], 
                   DiskBox[{-1.6242883863059012`, 1.999897812053879}, 0.1]}, 
                  {GrayLevel[0.5], Thickness[0.01], 
                   LineBox[{{-1.6417970526168013`, 
                    0.5}, {-1.6242883863059012`, 1.999897812053879}}]}, 
                  {RGBColor[0, 1, 0], 
                   DiskBox[{-1.6301246084095347`, 1.499931874702586}, 0.05]}, 
                  {RGBColor[0, 0, 1], Arrowheads[0.01], 
                   CircleBox[{-1.6417970526168013`, 0.5}, 0.3333333333333333, 
                    NCache[{
                    Rational[-1, 2] 
                    Pi, -1.744385156448883}, {-1.5707963267948966`, \
-1.744385156448883}]], 
                   ArrowBox[{{-1.6960837738083003`, 
                    0.1711169462964266}, {-1.6993698352183277`, 
                    0.17167624847625096`}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-1.6417970526168013`, 
                    0.5}, {-1.5260711661808104`, 0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-1.6417970526168013`, 
                    0.5}, {-1.7276418852525028`, 0.5}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.005], Arrowheads[0.04], 
                   ArrowBox[{{-1.6417970526168013`, 0.5}, {-1.611915998816512,
                     0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-1.6417970526168013`, 0.5}, 1.6, {
                    1.5591236175174956`, 1.5054669417431867`}], 
                   ArrowBox[{{-1.5533151981825313`, 
                    2.0975515520432735`}, {-1.5373443729752136`, 
                    2.096586871333875}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Dashing[{Small, Small}], 
                   Arrowheads[0.02], 
                   CircleBox[{-1.6417970526168013`, 0.5}, 1.7, {
                    1.5591236175174956`, 1.5764397421302503`}], 
                   ArrowBox[{{-1.6343908821149304`, 
                    2.1999838671700673`}, {-1.6513908077628425`, 
                    2.199972929155696}}]}, 
                  {RGBColor[0, 0, 1], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-1.6417970526168013`, 0.5}, 1.8, {
                    1.5591236175174956`, 1.5912696970830487`}], 
                   ArrowBox[{{-1.6606487744855027`, 
                    2.2999012785657396`}, {-1.678646544710875, 
                    2.299622770175019}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.003], Arrowheads[0.025], 
                   CircleBox[{-1.6417970526168013`, 0.5}, 1.5, {
                    1.5591236175174956`, 1.5600900667313518`}], 
                   ArrowBox[{{-1.6107398819220446`, 
                    1.999678449584589}, {-1.6257379693183263`, 
                    1.999914032817752}}]}, 
                  {RGBColor[0, 0, 1], ArrowBox[{{0, 2.8}, {0, 2.8}}]}},
                 Frame->True,
                 FrameTicks->Automatic,
                 ImageSize->800,
                 PlotRange->{{-3, 3}, {-1, 3}}], 432->
                GraphicsBox[{InsetBox[
                   TagBox[GridBox[{
                    {"\<\"t\"\>", "21.55`"},
                    {"\<\"driveIntent\"\>", "0"},
                    {"\<\"motorTorque\"\>", 
                    RowBox[{"-", "5.760334265447316`"}]},
                    {"\<\"vw\"\>", 
                    RowBox[{"-", "0.054766206441823624`"}]},
                    {"\<\"controllerState\"\>", 
                    RowBox[{"\[LeftAssociation]", 
                    RowBox[{
                    
                    RowBox[{"\<\"lastError\"\>", "\[Rule]", 
                    "0.004801206052056079`"}], ",", 
                    RowBox[{"\<\"lastInt\"\>", "\[Rule]", 
                    RowBox[{"-", "0.044678405255824345`"}]}], ",", 
                    RowBox[{"\<\"initialized\"\>", "\[Rule]", "True"}]}], 
                    "\[RightAssociation]"}]}
                    },
                    AutoDelete->False,
                    
                    GridBoxItemSize->{
                    "Columns" -> {{Automatic}}, "Rows" -> {{Automatic}}}],
                    "Grid"], {-1, -0.5}], 
                  {GrayLevel[0], LineBox[{{-5, 0}, {5, 0}}]}, 
                  {GrayLevel[0.5], 
                   CircleBox[{-1.6445473317510964`, 0.5}, 0.5], 
                   DiskBox[{-1.6445473317510964`, 0.5}, 0.1]}, 
                  {GrayLevel[0.5], 
                   DiskBox[{-1.6271019733150767`, 1.999898549725627}, 0.1]}, 
                  {GrayLevel[0.5], Thickness[0.01], 
                   LineBox[{{-1.6445473317510964`, 
                    0.5}, {-1.6271019733150767`, 1.999898549725627}}]}, 
                  {RGBColor[0, 1, 0], 
                   DiskBox[{-1.6329170927937497`, 1.4999323664837512`}, 
                    0.05]}, 
                  {RGBColor[0, 0, 1], Arrowheads[0.01], 
                   CircleBox[{-1.6445473317510964`, 0.5}, 0.3333333333333333, 
                    NCache[{
                    Rational[-1, 2] 
                    Pi, -1.743606354758316}, {-1.5707963267948966`, \
-1.743606354758316}]], 
                   ArrowBox[{{-1.6985779018269467`, 
                    0.17107476744951639`}, {-1.7018643978258623`, 
                    0.17163151026980067`}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-1.6445473317510964`, 0.5}, {-1.52934064644215, 
                    0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-1.6445473317510964`, 
                    0.5}, {-1.7300818085156826`, 0.5}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.005], Arrowheads[0.04], 
                   ArrowBox[{{-1.6445473317510964`, 
                    0.5}, {-1.6148751232067364`, 0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-1.6445473317510964`, 0.5}, 1.6, {
                    1.5591658256324714`, 1.505703161786661}], 
                   ArrowBox[{{-1.5564428534790324`, 
                    2.097572408658965}, {-1.540471800840383, 
                    2.096611500605471}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Dashing[{Small, Small}], 
                   Arrowheads[0.02], 
                   CircleBox[{-1.6445473317510964`, 0.5}, 1.7, {
                    1.5591658256324714`, 1.576360932686728}], 
                   ArrowBox[{{-1.6370071864897955`, 
                    2.199983278214653}, {-1.6540071129469336`, 
                    2.1999736799549945`}}]}, 
                  {RGBColor[0, 0, 1], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-1.6445473317510964`, 0.5}, 1.8, {
                    1.5591658256324714`, 1.5911676826627341`}], 
                   ArrowBox[{{-1.6632154376364656`, 
                    2.2999031923474806`}, {-1.681213236180059, 
                    2.2996265199903023`}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.003], Arrowheads[0.025], 
                   CircleBox[{-1.6445473317510964`, 0.5}, 1.5, {
                    1.5591658256324714`, 1.5600290006280801`}], 
                   ArrowBox[{{-1.6133985815952272`, 
                    1.999676550247995}, {-1.6283966545773942`, 
                    1.999913049355472}}]}, 
                  {RGBColor[0, 0, 1], ArrowBox[{{0, 2.8}, {0, 2.8}}]}},
                 Frame->True,
                 FrameTicks->Automatic,
                 ImageSize->800,
                 PlotRange->{{-3, 3}, {-1, 3}}], 433->
                GraphicsBox[{InsetBox[
                   TagBox[GridBox[{
                    {"\<\"t\"\>", "21.6`"},
                    {"\<\"driveIntent\"\>", "0"},
                    {"\<\"motorTorque\"\>", 
                    RowBox[{"-", "5.723228388867204`"}]},
                    {"\<\"vw\"\>", 
                    RowBox[{"-", "0.054280973176061356`"}]},
                    {"\<\"controllerState\"\>", 
                    RowBox[{"\[LeftAssociation]", 
                    RowBox[{
                    
                    RowBox[{"\<\"lastError\"\>", "\[Rule]", 
                    "0.004720409862885851`"}], ",", 
                    RowBox[{"\<\"lastInt\"\>", "\[Rule]", 
                    RowBox[{"-", "0.04444050111750465`"}]}], ",", 
                    RowBox[{"\<\"initialized\"\>", "\[Rule]", "True"}]}], 
                    "\[RightAssociation]"}]}
                    },
                    AutoDelete->False,
                    GridBoxItemSize->{
                    "Columns" -> {{Automatic}}, "Rows" -> {{Automatic}}}],
                    "Grid"], {-1, -0.5}], 
                  {GrayLevel[0], LineBox[{{-5, 0}, {5, 0}}]}, 
                  {GrayLevel[0.5], CircleBox[{-1.647273250697439, 0.5}, 0.5], 
                   DiskBox[{-1.647273250697439, 0.5}, 0.1]}, 
                  {GrayLevel[0.5], 
                   DiskBox[{-1.629925736151012, 1.9998996845586245`}, 0.1]}, 
                  {GrayLevel[0.5], Thickness[0.01], 
                   LineBox[{{-1.647273250697439, 0.5}, {-1.629925736151012, 
                    1.9998996845586245`}}]}, 
                  {RGBColor[0, 1, 0], 
                   DiskBox[{-1.635708240999821, 1.499933123039083}, 0.05]}, 
                  {RGBColor[0, 0, 1], Arrowheads[0.01], 
                   CircleBox[{-1.647273250697439, 0.5}, 0.3333333333333333, 
                    NCache[{
                    Rational[-1, 2] 
                    Pi, -1.7424931784609128`}, {-1.5707963267948966`, \
-1.7424931784609128`}]], 
                   ArrowBox[{{-1.7009376356001196`, 
                    0.17101482570771054`}, {-1.704224749315569, 
                    0.17156790973435587`}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-1.647273250697439, 0.5}, {-1.5328086829200949`,
                     0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-1.647273250697439, 0.5}, {-1.7323280638729466`,
                     0.5}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.005], Arrowheads[0.04], 
                   ArrowBox[{{-1.647273250697439, 0.5}, {-1.6178634960956024`,
                     0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-1.647273250697439, 0.5}, 1.6, {
                    1.5592310592794811`, 1.5060682456773815`}], 
                   ArrowBox[{{-1.5597520262346372`, 
                    2.097604467716754}, {-1.5437806238483058`, 
                    2.096649390501204}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Dashing[{Small, Small}], 
                   Arrowheads[0.02], 
                   CircleBox[{-1.647273250697439, 0.5}, 1.7, {
                    1.5592310592794811`, 1.5762740869045762`}], 
                   ArrowBox[{{-1.6395854690873166`, 
                    2.199982616974043}, {-1.6565853963139148`, 
                    2.1999744950863285`}}]}, 
                  {RGBColor[0, 0, 1], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-1.647273250697439, 0.5}, 1.8, {
                    1.5592310592794811`, 1.5910267725509657`}], 
                   ArrowBox[{{-1.6656877318383077`, 
                    2.2999058050032266`}, {-1.6836855691891548`, 
                    2.2996316687205907`}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.003], Arrowheads[0.025], 
                   CircleBox[{-1.647273250697439, 0.5}, 1.5, {
                    1.5592310592794811`, 1.5599943453381304`}], 
                   ArrowBox[{{-1.6160725288346052`, 
                    1.9996754698784796`}, {-1.631070593611821, 
                    1.9999124887483823`}}]}, 
                  {RGBColor[0, 0, 1], ArrowBox[{{0, 2.8}, {0, 2.8}}]}},
                 Frame->True,
                 FrameTicks->Automatic,
                 ImageSize->800,
                 PlotRange->{{-3, 3}, {-1, 3}}], 434->
                GraphicsBox[{InsetBox[
                   TagBox[GridBox[{
                    {"\<\"t\"\>", "21.650000000000002`"},
                    {"\<\"driveIntent\"\>", "0"},
                    {"\<\"motorTorque\"\>", 
                    RowBox[{"-", "5.676314074380031`"}]},
                    {"\<\"vw\"\>", 
                    RowBox[{"-", "0.05380043377893573`"}]},
                    {"\<\"controllerState\"\>", 
                    RowBox[{"\[LeftAssociation]", 
                    RowBox[{
                    
                    RowBox[{"\<\"lastError\"\>", "\[Rule]", 
                    "0.004661432603468114`"}], ",", 
                    RowBox[{"\<\"lastInt\"\>", "\[Rule]", 
                    RowBox[{"-", "0.04420607047402462`"}]}], ",", 
                    RowBox[{"\<\"initialized\"\>", "\[Rule]", "True"}]}], 
                    "\[RightAssociation]"}]}
                    },
                    AutoDelete->False,
                    
                    GridBoxItemSize->{
                    "Columns" -> {{Automatic}}, "Rows" -> {{Automatic}}}],
                    "Grid"], {-1, -0.5}], 
                  {GrayLevel[0], LineBox[{{-5, 0}, {5, 0}}]}, 
                  {GrayLevel[0.5], 
                   CircleBox[{-1.6499750244929972`, 0.5}, 0.5], 
                   DiskBox[{-1.6499750244929972`, 0.5}, 0.1]}, 
                  {GrayLevel[0.5], 
                   DiskBox[{-1.632755896538271, 1.9999011639546382`}, 0.1]}, 
                  {GrayLevel[0.5], Thickness[0.01], 
                   LineBox[{{-1.6499750244929972`, 0.5}, {-1.632755896538271, 
                    1.9999011639546382`}}]}, 
                  {RGBColor[0, 1, 0], 
                   DiskBox[{-1.638495605856513, 1.4999341093030922`}, 0.05]}, 
                  {RGBColor[0, 0, 1], Arrowheads[0.01], 
                   CircleBox[{-1.6499750244929972`, 0.5}, 0.3333333333333333, 
                    NCache[{
                    Rational[-1, 2] 
                    Pi, -1.7410857490262974`}, {-1.5707963267948966`, \
-1.7410857490262974`}]], 
                   ArrowBox[{{-1.7031763329799465`, 
                    0.17093962273408314`}, {-1.7064642218662267`, 
                    0.1714880798338675}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-1.6499750244929972`, 
                    0.5}, {-1.5364487430053966`, 0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-1.6499750244929972`, 0.5}, {-1.734400441855454,
                     0.5}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.005], Arrowheads[0.04], 
                   ArrowBox[{{-1.6499750244929972`, 
                    0.5}, {-1.6208741603678534`, 0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-1.6499750244929972`, 0.5}, 1.6, {
                    1.5593166560228024`, 1.5065472931611725`}], 
                   ArrowBox[{{-1.5632191384437455`, 
                    2.0976462112231884`}, {-1.547247280362705, 
                    2.0966987851770607`}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Dashing[{Small, Small}], 
                   Arrowheads[0.02], 
                   CircleBox[{-1.6499750244929972`, 0.5}, 1.7, {
                    1.5593166560228024`, 1.5761806978329198`}], 
                   ArrowBox[{{-1.6421284831181988`, 
                    2.1999818916060407`}, {-1.65912841102916, 
                    2.1999753573257816`}}]}, 
                  {RGBColor[0, 0, 1], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-1.6499750244929972`, 0.5}, 1.8, {
                    1.5593166560228024`, 1.5908517342138024`}], 
                   ArrowBox[{{-1.668074452834221, 
                    2.299909000670234}, {-1.6860723378937141`, 
                    2.299638014703304}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.003], Arrowheads[0.025], 
                   CircleBox[{-1.6499750244929972`, 0.5}, 1.5, {
                    1.5593166560228024`, 1.5599840101843971`}], 
                   ArrowBox[{{-1.6187588032552989`, 
                    1.999675147334128}, {-1.633756865582087, 
                    1.9999123212113235`}}]}, 
                  {RGBColor[0, 0, 1], ArrowBox[{{0, 2.8}, {0, 2.8}}]}},
                 Frame->True,
                 FrameTicks->Automatic,
                 ImageSize->800,
                 PlotRange->{{-3, 3}, {-1, 3}}], 435->
                GraphicsBox[{InsetBox[
                   TagBox[GridBox[{
                    {"\<\"t\"\>", "21.7`"},
                    {"\<\"driveIntent\"\>", "0"},
                    {"\<\"motorTorque\"\>", 
                    RowBox[{"-", "5.6208717198227225`"}]},
                    {"\<\"vw\"\>", 
                    RowBox[{"-", "0.0533252956760023`"}]},
                    {"\<\"controllerState\"\>", 
                    RowBox[{"\[LeftAssociation]", 
                    RowBox[{
                    
                    RowBox[{"\<\"lastError\"\>", "\[Rule]", 
                    "0.004621927791152726`"}], ",", 
                    RowBox[{"\<\"lastInt\"\>", "\[Rule]", 
                    RowBox[{"-", "0.0439740825398161`"}]}], ",", 
                    RowBox[{"\<\"initialized\"\>", "\[Rule]", "True"}]}], 
                    "\[RightAssociation]"}]}
                    },
                    AutoDelete->False,
                    
                    GridBoxItemSize->{
                    "Columns" -> {{Automatic}}, "Rows" -> {{Automatic}}}],
                    "Grid"], {-1, -0.5}], 
                  {GrayLevel[0], LineBox[{{-5, 0}, {5, 0}}]}, 
                  {GrayLevel[0.5], CircleBox[{-1.652652906336544, 0.5}, 0.5], 
                   DiskBox[{-1.652652906336544, 0.5}, 0.1]}, 
                  {GrayLevel[0.5], 
                   DiskBox[{-1.635588872296039, 1.9999029364403098`}, 0.1]}, 
                  {GrayLevel[0.5], Thickness[0.01], 
                   LineBox[{{-1.652652906336544, 0.5}, {-1.635588872296039, 
                    1.9999029364403098`}}]}, 
                  {RGBColor[0, 1, 0], 
                   DiskBox[{-1.6412768836428742`, 1.4999352909602066`}, 
                    0.05]}, 
                  {RGBColor[0, 0, 1], Arrowheads[0.01], 
                   CircleBox[{-1.652652906336544, 0.5}, 0.3333333333333333, 
                    NCache[{
                    Rational[-1, 2] 
                    Pi, -1.7394224783895782`}, {-1.5707963267948966`, \
-1.7394224783895782`}]], 
                   ArrowBox[{{-1.705306825022761, 
                    0.1708515897682379}, {-1.7085956215932894`, 
                    0.1713945774628577}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-1.652652906336544, 0.5}, {-1.5402354719400897`,
                     0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-1.652652906336544, 0.5}, {-1.736317996810184, 
                    0.5}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.005], Arrowheads[0.04], 
                   ArrowBox[{{-1.652652906336544, 0.5}, {-1.6239005624137297`,
                     0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-1.652652906336544, 0.5}, 1.6, {
                    1.5594200587177067`, 1.50712599327292}], 
                   ArrowBox[{{-1.56682159280366, 
                    2.0976961493404245`}, {-1.5508491891215295`, 
                    2.0967579663684806`}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Dashing[{Small, Small}], 
                   Arrowheads[0.02], 
                   CircleBox[{-1.652652906336544, 0.5}, 1.7, {
                    1.5594200587177067`, 1.5760821521068027`}], 
                   ArrowBox[{{-1.6446388390502225`, 
                    2.1999811101084417`}, {-1.6616387675225637`, 
                    2.1999762510984517`}}]}, 
                  {RGBColor[0, 0, 1], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-1.652652906336544, 0.5}, 1.8, {
                    1.5594200587177067`, 1.590647123827833}], 
                   ArrowBox[{{-1.6703840542261308`, 
                    2.2999126663242633`}, {-1.6883819943554224`, 
                    2.2996453629171887`}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.003], Arrowheads[0.025], 
                   CircleBox[{-1.652652906336544, 0.5}, 1.5, {
                    1.5594200587177067`, 1.5599959062163944`}], 
                   ArrowBox[{{-1.6214545252845922`, 
                    1.9996755185771813`}, {-1.6364525904317473`, 
                    1.9999125140369305`}}]}, 
                  {RGBColor[0, 0, 1], ArrowBox[{{0, 2.8}, {0, 2.8}}]}},
                 Frame->True,
                 FrameTicks->Automatic,
                 ImageSize->800,
                 PlotRange->{{-3, 3}, {-1, 3}}], 436->
                GraphicsBox[{InsetBox[
                   TagBox[GridBox[{
                    {"\<\"t\"\>", "21.75`"},
                    {"\<\"driveIntent\"\>", "0"},
                    {"\<\"motorTorque\"\>", 
                    RowBox[{"-", "5.558118972576086`"}]},
                    {"\<\"vw\"\>", 
                    RowBox[{"-", "0.052856157581812564`"}]},
                    {"\<\"controllerState\"\>", 
                    RowBox[{"\[LeftAssociation]", 
                    RowBox[{
                    
                    RowBox[{"\<\"lastError\"\>", "\[Rule]", 
                    "0.004599633860060638`"}], ",", 
                    RowBox[{"\<\"lastInt\"\>", "\[Rule]", 
                    RowBox[{"-", "0.04374362174282831`"}]}], ",", 
                    RowBox[{"\<\"initialized\"\>", "\[Rule]", "True"}]}], 
                    "\[RightAssociation]"}]}
                    },
                    AutoDelete->False,
                    
                    GridBoxItemSize->{
                    "Columns" -> {{Automatic}}, "Rows" -> {{Automatic}}}],
                    "Grid"], {-1, -0.5}], 
                  {GrayLevel[0], LineBox[{{-5, 0}, {5, 0}}]}, 
                  {GrayLevel[0.5], 
                   CircleBox[{-1.6553071819993976`, 0.5}, 0.5], 
                   DiskBox[{-1.6553071819993976`, 0.5}, 0.1]}, 
                  {GrayLevel[0.5], 
                   DiskBox[{-1.6384212904921953`, 1.9999049525446626`}, 0.1]}, 
                  {GrayLevel[0.5], Thickness[0.01], 
                   LineBox[{{-1.6553071819993976`, 
                    0.5}, {-1.6384212904921953`, 1.9999049525446626`}}]}, 
                  {RGBColor[0, 1, 0], 
                   DiskBox[{-1.644049920994596, 1.4999366350297751`}, 0.05]}, 
                  {RGBColor[0, 0, 1], Arrowheads[0.01], 
                   CircleBox[{-1.6553071819993976`, 0.5}, 0.3333333333333333, 
                    NCache[{
                    Rational[-1, 2] 
                    Pi, -1.737539895972179}, {-1.5707963267948966`, \
-1.737539895972179}]], 
                   ArrowBox[{{-1.7073413587360304`, 
                    0.17075304775525718`}, {-1.710631171697104, 
                    0.17128984306073053`}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-1.6553071819993976`, 0.5}, {-1.544144802547876,
                     0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-1.6553071819993976`, 
                    0.5}, {-1.7380989495436252`, 0.5}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.005], Arrowheads[0.04], 
                   ArrowBox[{{-1.6553071819993976`, 
                    0.5}, {-1.6269365700921035`, 0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-1.6553071819993976`, 0.5}, 1.6, {
                    1.5595388280120654`, 1.50779069427797}], 
                   ArrowBox[{{-1.5705378775858554`, 
                    2.0977528485436174`}, {-1.5545648538211294`, 
                    2.09682528265093}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Dashing[{Small, Small}], 
                   Arrowheads[0.02], 
                   CircleBox[{-1.6553071819993976`, 0.5}, 1.7, {
                    1.5595388280120654`, 1.575979728825634}], 
                   ArrowBox[{{-1.6471189971121944`, 
                    2.1999802803645263`}, {-1.6641189259930425`, 
                    2.199977162543012}}]}, 
                  {RGBColor[0, 0, 1], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-1.6553071819993976`, 0.5}, 1.8, {
                    1.5595388280120654`, 1.5904172667485994`}], 
                   ArrowBox[{{-1.6726246067558643`, 
                    2.2999166944055505`}, {-1.6906226078512818`, 
                    2.2996535279594514`}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.003], Arrowheads[0.025], 
                   CircleBox[{-1.6553071819993976`, 0.5}, 1.5, {
                    1.5595388280120654`, 1.560027966101582}], 
                   ArrowBox[{{-1.624156880388415, 
                    1.9996765180229852`}, {-1.6391549531259098`, 
                    1.9999130326463659`}}]}, 
                  {RGBColor[0, 0, 1], ArrowBox[{{0, 2.8}, {0, 2.8}}]}},
                 Frame->True,
                 FrameTicks->Automatic,
                 ImageSize->800,
                 PlotRange->{{-3, 3}, {-1, 3}}], 437->
                GraphicsBox[{InsetBox[
                   TagBox[GridBox[{
                    {"\<\"t\"\>", "21.8`"},
                    {"\<\"driveIntent\"\>", "0"},
                    {"\<\"motorTorque\"\>", 
                    RowBox[{"-", "5.489206226722182`"}]},
                    {"\<\"vw\"\>", 
                    RowBox[{"-", "0.052393516034401455`"}]},
                    {"\<\"controllerState\"\>", 
                    RowBox[{"\[LeftAssociation]", 
                    RowBox[{
                    
                    RowBox[{"\<\"lastError\"\>", "\[Rule]", 
                    "0.004592386106190461`"}], ",", 
                    RowBox[{"\<\"lastInt\"\>", "\[Rule]", 
                    RowBox[{"-", "0.043513883164160494`"}]}], ",", 
                    RowBox[{"\<\"initialized\"\>", "\[Rule]", "True"}]}], 
                    "\[RightAssociation]"}]}
                    },
                    AutoDelete->False,
                    
                    GridBoxItemSize->{
                    "Columns" -> {{Automatic}}, "Rows" -> {{Automatic}}}],
                    "Grid"], {-1, -0.5}], 
                  {GrayLevel[0], LineBox[{{-5, 0}, {5, 0}}]}, 
                  {GrayLevel[0.5], CircleBox[{-1.657938164555327, 0.5}, 0.5], 
                   DiskBox[{-1.657938164555327, 0.5}, 0.1]}, 
                  {GrayLevel[0.5], 
                   DiskBox[{-1.641249997071814, 1.9999071654825982`}, 0.1]}, 
                  {GrayLevel[0.5], Thickness[0.01], 
                   LineBox[{{-1.657938164555327, 0.5}, {-1.641249997071814, 
                    1.9999071654825982`}}]}, 
                  {RGBColor[0, 1, 0], 
                   DiskBox[{-1.6468127195663183`, 1.4999381103217322`}, 
                    0.05]}, 
                  {RGBColor[0, 0, 1], Arrowheads[0.01], 
                   CircleBox[{-1.657938164555327, 0.5}, 0.3333333333333333, 
                    NCache[{
                    Rational[-1, 2] 
                    Pi, -1.735472513596562}, {-1.5707963267948966`, \
-1.735472513596562}]], 
                   ArrowBox[{{-1.7092915512317275`, 
                    0.17064617690396366`}, {-1.7125824669227219`, 
                    0.1711761697657972}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-1.657938164555327, 0.5}, {-1.5481540400208833`,
                     0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-1.657938164555327, 0.5}, {-1.739760609103229, 
                    0.5}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.005], Arrowheads[0.04], 
                   ArrowBox[{{-1.657938164555327, 0.5}, {-1.6299764845687854`,
                     0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-1.657938164555327, 0.5}, 1.6, {
                    1.559670652283469, 1.5085284592671584`}], 
                   ArrowBox[{{-1.5743476492177289`, 
                    2.0978149535367336`}, {-1.5583739454744598`, 
                    2.096899172233118}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Dashing[{Small, Small}], 
                   Arrowheads[0.02], 
                   CircleBox[{-1.657938164555327, 0.5}, 1.7, {
                    1.559670652283469, 1.5758745995946863`}], 
                   ArrowBox[{{-1.6495712620942007`, 
                    2.1999794101527246`}, {-1.6665711912088796`, 
                    2.1999780795206743`}}]}, 
                  {RGBColor[0, 0, 1], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-1.657938164555327, 0.5}, 1.8, {
                    1.559670652283469, 1.5901662424319254`}], 
                   ArrowBox[{{-1.6748037659126422`, 
                    2.299920984790959}, {-1.6928018325021805`, 
                    2.299662336289031}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.003], Arrowheads[0.025], 
                   CircleBox[{-1.657938164555327, 0.5}, 1.5, {
                    1.559670652283469, 1.5600781615302655`}], 
                   ArrowBox[{{-1.6268631398892646`, 
                    1.9996780797364493`}, {-1.6418612244798176`, 
                    1.9999138415248419`}}]}, 
                  {RGBColor[0, 0, 1], ArrowBox[{{0, 2.8}, {0, 2.8}}]}},
                 Frame->True,
                 FrameTicks->Automatic,
                 ImageSize->800,
                 PlotRange->{{-3, 3}, {-1, 3}}], 438->
                GraphicsBox[{InsetBox[
                   TagBox[GridBox[{
                    {"\<\"t\"\>", "21.85`"},
                    {"\<\"driveIntent\"\>", "0"},
                    {"\<\"motorTorque\"\>", 
                    RowBox[{"-", "5.415213281479538`"}]},
                    {"\<\"vw\"\>", 
                    RowBox[{"-", "0.05193777216337872`"}]},
                    {"\<\"controllerState\"\>", 
                    RowBox[{"\[LeftAssociation]", 
                    RowBox[{
                    
                    RowBox[{"\<\"lastError\"\>", "\[Rule]", 
                    "0.004598126333675762`"}], ",", 
                    RowBox[{"\<\"lastInt\"\>", "\[Rule]", 
                    RowBox[{"-", "0.04328416743976906`"}]}], ",", 
                    RowBox[{"\<\"initialized\"\>", "\[Rule]", "True"}]}], 
                    "\[RightAssociation]"}]}
                    },
                    AutoDelete->False,
                    
                    GridBoxItemSize->{
                    "Columns" -> {{Automatic}}, "Rows" -> {{Automatic}}}],
                    "Grid"], {-1, -0.5}], 
                  {GrayLevel[0], LineBox[{{-5, 0}, {5, 0}}]}, 
                  {GrayLevel[0.5], 
                   CircleBox[{-1.6605461894435385`, 0.5}, 0.5], 
                   DiskBox[{-1.6605461894435385`, 0.5}, 0.1]}, 
                  {GrayLevel[0.5], 
                   DiskBox[{-1.6440720632724883`, 1.9999095316607933`}, 0.1]}, 
                  {GrayLevel[0.5], Thickness[0.01], 
                   LineBox[{{-1.6605461894435385`, 
                    0.5}, {-1.6440720632724883`, 1.9999095316607933`}}]}, 
                  {RGBColor[0, 1, 0], 
                   DiskBox[{-1.6495634386628384`, 1.4999396877738622`}, 
                    0.05]}, 
                  {RGBColor[0, 0, 1], Arrowheads[0.01], 
                   CircleBox[{-1.6605461894435385`, 0.5}, 0.3333333333333333, 
                    NCache[{
                    Rational[-1, 2] 
                    Pi, -1.7332527252392826`}, {-1.5707963267948966`, \
-1.7332527252392826`}]], 
                   ArrowBox[{{-1.711168354417576, 
                    0.17053299478634826`}, {-1.7144604384716633`, 
                    0.17105568121208653`}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-1.6605461894435385`, 
                    0.5}, {-1.5522419238139478`, 0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-1.6605461894435385`, 
                    0.5}, {-1.7413193122505197`, 0.5}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.005], Arrowheads[0.04], 
                   ArrowBox[{{-1.6605461894435385`, 0.5}, {-1.633015046620929,
                     0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-1.6605461894435385`, 0.5}, 1.6, {
                    1.55981335521082, 1.5093271085185294`}], 
                   ArrowBox[{{-1.5782317943454662`, 
                    2.0978812034565144`}, {-1.5622573643085607`, 
                    2.0969781798301397`}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Dashing[{Small, Small}], 
                   Arrowheads[0.02], 
                   CircleBox[{-1.6605461894435385`, 0.5}, 1.7, {
                    1.55981335521082, 1.5757678295949524`}], 
                   ArrowBox[{{-1.651997780229278, 
                    2.1999785071288125`}, {-1.6689977093891304`, 
                    2.1999789915791927`}}]}, 
                  {RGBColor[0, 0, 1], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-1.6605461894435385`, 0.5}, 1.8, {
                    1.55981335521082, 1.5898978734412619`}], 
                   ArrowBox[{{-1.6769287472213432`, 
                    2.29992544617844}, {-1.6949268825759902`, 
                    2.299671627808732}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.003], Arrowheads[0.025], 
                   CircleBox[{-1.6605461894435385`, 0.5}, 1.5, {
                    1.55981335521082, 1.5601445182509237`}], 
                   ArrowBox[{{-1.6295706785652324`, 
                    1.9996801384714769`}, {-1.6445687787671446`, 
                    1.9999149050356415`}}]}, 
                  {RGBColor[0, 0, 1], ArrowBox[{{0, 2.8}, {0, 2.8}}]}},
                 Frame->True,
                 FrameTicks->Automatic,
                 ImageSize->800,
                 PlotRange->{{-3, 3}, {-1, 3}}], 439->
                GraphicsBox[{InsetBox[
                   TagBox[GridBox[{
                    {"\<\"t\"\>", "21.900000000000002`"},
                    {"\<\"driveIntent\"\>", "0"},
                    {"\<\"motorTorque\"\>", 
                    RowBox[{"-", "5.337147059621059`"}]},
                    {"\<\"vw\"\>", 
                    RowBox[{"-", "0.05148923859962796`"}]},
                    {"\<\"controllerState\"\>", 
                    RowBox[{"\[LeftAssociation]", 
                    RowBox[{
                    
                    RowBox[{"\<\"lastError\"\>", "\[Rule]", 
                    "0.004614910373429353`"}], ",", 
                    RowBox[{"\<\"lastInt\"\>", "\[Rule]", 
                    RowBox[{"-", "0.04305387523483229`"}]}], ",", 
                    RowBox[{"\<\"initialized\"\>", "\[Rule]", "True"}]}], 
                    "\[RightAssociation]"}]}
                    },
                    AutoDelete->False,
                    
                    GridBoxItemSize->{
                    "Columns" -> {{Automatic}}, "Rows" -> {{Automatic}}}],
                    "Grid"], {-1, -0.5}], 
                  {GrayLevel[0], LineBox[{{-5, 0}, {5, 0}}]}, 
                  {GrayLevel[0.5], CircleBox[{-1.66313160987404, 0.5}, 0.5], 
                   DiskBox[{-1.66313160987404, 0.5}, 0.1]}, 
                  {GrayLevel[0.5], 
                   DiskBox[{-1.6468847891338603`, 1.9999120110245923`}, 0.1]}, 
                  {GrayLevel[0.5], Thickness[0.01], 
                   LineBox[{{-1.66313160987404, 0.5}, {-1.6468847891338603`, 
                    1.9999120110245923`}}]}, 
                  {RGBColor[0, 1, 0], 
                   DiskBox[{-1.6523003960472535`, 1.4999413406830615`}, 
                    0.05]}, 
                  {RGBColor[0, 0, 1], Arrowheads[0.01], 
                   CircleBox[{-1.66313160987404, 0.5}, 0.3333333333333333, 
                    NCache[{
                    Rational[-1, 2] 
                    Pi, -1.7309107385835283`}, {-1.5707963267948966`, \
-1.7309107385835283`}]], 
                   ArrowBox[{{-1.7129820293949898`, 
                    0.1704153420064939}, {-1.7162753285442183`, 
                    0.17093031698891487`}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-1.66313160987404, 0.5}, {-1.5563886686816188`, 
                    0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-1.66313160987404, 0.5}, {-1.742790379176772, 
                    0.5}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.005], Arrowheads[0.04], 
                   ArrowBox[{{-1.66313160987404, 0.5}, {-1.6360474379843508`, 
                    0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-1.66313160987404, 0.5}, 1.6, {
                    1.5599649011792724`, 1.5101752497381207`}], 
                   ArrowBox[{{-1.5821724731323121`, 
                    2.0979504429668756`}, {-1.5661972829494897`, 
                    2.0970609682362427`}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Dashing[{Small, Small}], 
                   Arrowheads[0.02], 
                   CircleBox[{-1.66313160987404, 0.5}, 1.7, {
                    1.5599649011792724`, 1.5756603795521262`}], 
                   ArrowBox[{{-1.6544005379460842`, 
                    2.199977578788317}, {-1.671400466955746, 
                    2.1999798898818073`}}]}, 
                  {RGBColor[0, 0, 1], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-1.66313160987404, 0.5}, 1.8, {
                    1.5599649011792724`, 1.589615718177167}], 
                   ArrowBox[{{-1.6790063085668452`, 
                    2.2999299969558296`}, {-1.6970045148212503`, 
                    2.299681256864792}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.003], Arrowheads[0.025], 
                   CircleBox[{-1.66313160987404, 0.5}, 1.5, {
                    1.5599649011792724`, 1.5602251288620856`}], 
                   ArrowBox[{{-1.6322769892287532`, 
                    1.9996826305538233`}, {-1.6472751083066122`, 
                    1.9999161881112033`}}]}, 
                  {RGBColor[0, 0, 1], ArrowBox[{{0, 2.8}, {0, 2.8}}]}},
                 Frame->True,
                 FrameTicks->Automatic,
                 ImageSize->800,
                 PlotRange->{{-3, 3}, {-1, 3}}], 440->
                GraphicsBox[{InsetBox[
                   TagBox[GridBox[{
                    {"\<\"t\"\>", "21.95`"},
                    {"\<\"driveIntent\"\>", "0"},
                    {"\<\"motorTorque\"\>", 
                    RowBox[{"-", "5.255940286103884`"}]},
                    {"\<\"vw\"\>", 
                    RowBox[{"-", "0.051048146443853706`"}]},
                    {"\<\"controllerState\"\>", 
                    RowBox[{"\[LeftAssociation]", 
                    RowBox[{
                    
                    RowBox[{"\<\"lastError\"\>", "\[Rule]", 
                    "0.004640913646196954`"}], ",", 
                    RowBox[{"\<\"lastInt\"\>", "\[Rule]", 
                    RowBox[{"-", "0.04282250139276602`"}]}], ",", 
                    RowBox[{"\<\"initialized\"\>", "\[Rule]", "True"}]}], 
                    "\[RightAssociation]"}]}
                    },
                    AutoDelete->False,
                    
                    GridBoxItemSize->{
                    "Columns" -> {{Automatic}}, "Rows" -> {{Automatic}}}],
                    "Grid"], {-1, -0.5}], 
                  {GrayLevel[0], LineBox[{{-5, 0}, {5, 0}}]}, 
                  {GrayLevel[0.5], 
                   CircleBox[{-1.6656947925803178`, 0.5}, 0.5], 
                   DiskBox[{-1.6656947925803178`, 0.5}, 0.1]}, 
                  {GrayLevel[0.5], 
                   DiskBox[{-1.6496857044013076`, 1.9999145672656415`}, 0.1]}, 
                  {GrayLevel[0.5], Thickness[0.01], 
                   LineBox[{{-1.6656947925803178`, 
                    0.5}, {-1.6496857044013076`, 1.9999145672656415`}}]}, 
                  {RGBColor[0, 1, 0], 
                   DiskBox[{-1.6550220671276443`, 1.4999430448437612`}, 
                    0.05]}, 
                  {RGBColor[0, 0, 1], Arrowheads[0.01], 
                   CircleBox[{-1.6656947925803178`, 0.5}, 0.3333333333333333, 
                    NCache[{
                    Rational[-1, 2] 
                    Pi, -1.728474535378013}, {-1.5707963267948966`, \
-1.728474535378013}]], 
                   ArrowBox[{{-1.7147421297620316`, 
                    0.17029487443096308`}, {-1.7180366737207196`, 
                    0.17080182474716665`}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-1.6656947925803178`, 0.5}, {-1.56057598685824, 
                    0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-1.6656947925803178`, 
                    0.5}, {-1.7441880840233475`, 0.5}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.005], Arrowheads[0.04], 
                   ArrowBox[{{-1.6656947925803178`, 
                    0.5}, {-1.6390692783012697`, 0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-1.6656947925803178`, 0.5}, 1.6, {
                    1.5601233987153582`, 1.5110622972807672`}], 
                   ArrowBox[{{-1.5861531455178877`, 
                    2.0980216288844766`}, {-1.5701771726138603`, 
                    2.0971463252551232`}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Dashing[{Small, Small}], 
                   Arrowheads[0.02], 
                   CircleBox[{-1.6656947925803178`, 0.5}, 1.7, {
                    1.5601233987153582`, 1.575553108482179}], 
                   ArrowBox[{{-1.6567813622891834`, 
                    2.199976632416}, {-1.673781290953122, 
                    2.1999807671100475`}}]}, 
                  {RGBColor[0, 0, 1], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-1.6656947925803178`, 0.5}, 1.8, {
                    1.5601233987153582`, 1.5893230669714908`}], 
                   ArrowBox[{{-1.6810427389171076`, 
                    2.2999345656282184`}, {-1.6990410171948738`, 
                    2.2996910927445176`}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.003], Arrowheads[0.025], 
                   CircleBox[{-1.6656947925803178`, 0.5}, 1.5, {
                    1.5601233987153582`, 1.560318163494884}], 
                   ArrowBox[{{-1.6349796944912074`, 
                    1.9996854946119123`}, {-1.6499778352331005`, 
                    1.9999176568237824`}}]}, 
                  {RGBColor[0, 0, 1], ArrowBox[{{0, 2.8}, {0, 2.8}}]}},
                 Frame->True,
                 FrameTicks->Automatic,
                 ImageSize->800,
                 PlotRange->{{-3, 3}, {-1, 3}}], 441->
                GraphicsBox[{InsetBox[
                   TagBox[GridBox[{
                    {"\<\"t\"\>", "22.`"},
                    {"\<\"driveIntent\"\>", "0"},
                    {"\<\"motorTorque\"\>", 
                    RowBox[{"-", "5.172451029397719`"}]},
                    {"\<\"vw\"\>", 
                    RowBox[{"-", "0.050614652220191854`"}]},
                    {"\<\"controllerState\"\>", 
                    RowBox[{"\[LeftAssociation]", 
                    RowBox[{
                    
                    RowBox[{"\<\"lastError\"\>", "\[Rule]", 
                    "0.004674434941064582`"}], ",", 
                    RowBox[{"\<\"lastInt\"\>", "\[Rule]", 
                    RowBox[{"-", "0.042589628852300734`"}]}], ",", 
                    RowBox[{"\<\"initialized\"\>", "\[Rule]", "True"}]}], 
                    "\[RightAssociation]"}]}
                    },
                    AutoDelete->False,
                    
                    GridBoxItemSize->{
                    "Columns" -> {{Automatic}}, "Rows" -> {{Automatic}}}],
                    "Grid"], {-1, -0.5}], 
                  {GrayLevel[0], LineBox[{{-5, 0}, {5, 0}}]}, 
                  {GrayLevel[0.5], 
                   CircleBox[{-1.6682361139203488`, 0.5}, 0.5], 
                   DiskBox[{-1.6682361139203488`, 0.5}, 0.1]}, 
                  {GrayLevel[0.5], 
                   DiskBox[{-1.6524725671142622`, 1.9999171679103123`}, 0.1]}, 
                  {GrayLevel[0.5], Thickness[0.01], 
                   LineBox[{{-1.6682361139203488`, 
                    0.5}, {-1.6524725671142622`, 1.9999171679103123`}}]}, 
                  {RGBColor[0, 1, 0], 
                   DiskBox[{-1.657727082716291, 1.499944778606875}, 0.05]}, 
                  {RGBColor[0, 0, 1], Arrowheads[0.01], 
                   CircleBox[{-1.6682361139203488`, 0.5}, 0.3333333333333333, 
                    NCache[{
                    Rational[-1, 2] 
                    Pi, -1.725969857676828}, {-1.5707963267948966`, \
-1.725969857676828}]], 
                   ArrowBox[{{-1.7164574930425618`, 
                    0.17017306097490625`}, {-1.7197532964130673`, 
                    0.1706717579387959}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-1.6682361139203488`, 
                    0.5}, {-1.5647870933323944`, 0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-1.6682361139203488`, 
                    0.5}, {-1.7455256387923657`, 0.5}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.005], Arrowheads[0.04], 
                   ArrowBox[{{-1.6682361139203488`, 
                    0.5}, {-1.6420766182044113`, 0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-1.6682361139203488`, 0.5}, 1.6, {
                    1.5602871021454514`, 1.511978481431105}], 
                   ArrowBox[{{-1.5901585821247404`, 
                    2.098093832986257}, {-1.5741818139865758`, 
                    2.0972331666553785`}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Dashing[{Small, Small}], 
                   Arrowheads[0.02], 
                   CircleBox[{-1.6682361139203488`, 0.5}, 1.7, {
                    1.5602871021454514`, 1.5754467770975404`}], 
                   ArrowBox[{{-1.6591419228107662`, 
                    2.1999756750283406`}, {-1.6761418509389534`, 
                    2.199981617348315}}]}, 
                  {RGBColor[0, 0, 1], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-1.6682361139203488`, 0.5}, 1.8, {
                    1.5602871021454514`, 1.589022941197661}], 
                   ArrowBox[{{-1.6830438528196596`, 
                    2.2999390908774355`}, {-1.7010422033593104`, 
                    2.299701019751815}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.003], Arrowheads[0.025], 
                   CircleBox[{-1.6682361139203488`, 0.5}, 1.5, {
                    1.5602871021454514`, 1.5604218785251676`}], 
                   ArrowBox[{{-1.6376765559226472`, 
                    1.999688672163321}, {-1.652674720662585, 
                    1.9999192788413238`}}]}, 
                  {RGBColor[0, 0, 1], ArrowBox[{{0, 2.8}, {0, 2.8}}]}},
                 Frame->True,
                 FrameTicks->Automatic,
                 ImageSize->800,
                 PlotRange->{{-3, 3}, {-1, 3}}], 442->
                GraphicsBox[{InsetBox[
                   TagBox[GridBox[{
                    {"\<\"t\"\>", "22.05`"},
                    {"\<\"driveIntent\"\>", "0"},
                    {"\<\"motorTorque\"\>", 
                    RowBox[{"-", "5.087463010960931`"}]},
                    {"\<\"vw\"\>", 
                    RowBox[{"-", "0.05018884474974601`"}]},
                    {"\<\"controllerState\"\>", 
                    RowBox[{"\[LeftAssociation]", 
                    RowBox[{
                    
                    RowBox[{"\<\"lastError\"\>", "\[Rule]", 
                    "0.004713898577939021`"}], ",", 
                    RowBox[{"\<\"lastInt\"\>", "\[Rule]", 
                    RowBox[{"-", "0.0423549224175371`"}]}], ",", 
                    RowBox[{"\<\"initialized\"\>", "\[Rule]", "True"}]}], 
                    "\[RightAssociation]"}]}
                    },
                    AutoDelete->False,
                    
                    GridBoxItemSize->{
                    "Columns" -> {{Automatic}}, "Rows" -> {{Automatic}}}],
                    "Grid"], {-1, -0.5}], 
                  {GrayLevel[0], LineBox[{{-5, 0}, {5, 0}}]}, 
                  {GrayLevel[0.5], 
                   CircleBox[{-1.6707559563235064`, 0.5}, 0.5], 
                   DiskBox[{-1.6707559563235064`, 0.5}, 0.1]}, 
                  {GrayLevel[0.5], 
                   DiskBox[{-1.6552433601584466`, 1.9999197843085543`}, 0.1]}, 
                  {GrayLevel[0.5], Thickness[0.01], 
                   LineBox[{{-1.6707559563235064`, 
                    0.5}, {-1.6552433601584466`, 1.9999197843085543`}}]}, 
                  {RGBColor[0, 1, 0], 
                   DiskBox[{-1.6604142255468, 1.4999465228723694`}, 0.05]}, 
                  {RGBColor[0, 0, 1], Arrowheads[0.01], 
                   CircleBox[{-1.6707559563235064`, 0.5}, 0.3333333333333333, 
                    NCache[{
                    Rational[-1, 2] 
                    Pi, -1.7234202171237245`}, {-1.5707963267948966`, \
-1.7234202171237245`}]], 
                   ArrowBox[{{-1.718136239482109, 
                    0.17005118597133623`}, {-1.7214333036367853`, 
                    0.17054147820947108`}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-1.6707559563235064`, 
                    0.5}, {-1.5690066961042877`, 0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-1.6707559563235064`, 
                    0.5}, {-1.7468151892677137`, 0.5}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.005], Arrowheads[0.04], 
                   ArrowBox[{{-1.6707559563235064`, 0.5}, {-1.645065929048495,
                     0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-1.6707559563235064`, 0.5}, 1.6, {
                    1.5604544116655612`, 1.5129148487961606`}], 
                   ArrowBox[{{-1.5941748614492892`, 
                    2.0981662416369167`}, {-1.5781972944154634`, 
                    2.0973205358054448`}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Dashing[{Small, Small}], 
                   Arrowheads[0.02], 
                   CircleBox[{-1.6707559563235064`, 0.5}, 1.7, {
                    1.5604544116655612`, 1.5753420517651529`}], 
                   ArrowBox[{{-1.6614837347465008`, 
                    2.199974713314003}, {-1.6784836621591541`, 
                    2.1999824359570654`}}]}, 
                  {RGBColor[0, 0, 1], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-1.6707559563235064`, 0.5}, 1.8, {
                    1.5604544116655612`, 1.5887180950597886`}], 
                   ArrowBox[{{-1.6850149900630058`, 
                    2.2999435213241597`}, {-1.703013412341415, 
                    2.299710936937166}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.003], Arrowheads[0.025], 
                   CircleBox[{-1.6707559563235064`, 0.5}, 1.5, {
                    1.5604544116655612`, 1.5605346234574773`}], 
                   ArrowBox[{{-1.640365480817503, 
                    1.9996921080670256`}, {-1.655363671461851, 
                    1.999921023776498}}]}, 
                  {RGBColor[0, 0, 1], ArrowBox[{{0, 2.8}, {0, 2.8}}]}},
                 Frame->True,
                 FrameTicks->Automatic,
                 ImageSize->800,
                 PlotRange->{{-3, 3}, {-1, 3}}], 443->
                GraphicsBox[{InsetBox[
                   TagBox[GridBox[{
                    {"\<\"t\"\>", "22.1`"},
                    {"\<\"driveIntent\"\>", "0"},
                    {"\<\"motorTorque\"\>", 
                    RowBox[{"-", "5.001686591819985`"}]},
                    {"\<\"vw\"\>", 
                    RowBox[{"-", "0.04977075188719201`"}]},
                    {"\<\"controllerState\"\>", 
                    RowBox[{"\[LeftAssociation]", 
                    RowBox[{
                    
                    RowBox[{"\<\"lastError\"\>", "\[Rule]", 
                    "0.004757855118660493`"}], ",", 
                    RowBox[{"\<\"lastInt\"\>", "\[Rule]", 
                    RowBox[{"-", "0.04211812245756303`"}]}], ",", 
                    RowBox[{"\<\"initialized\"\>", "\[Rule]", "True"}]}], 
                    "\[RightAssociation]"}]}
                    },
                    AutoDelete->False,
                    
                    GridBoxItemSize->{
                    "Columns" -> {{Automatic}}, "Rows" -> {{Automatic}}}],
                    "Grid"], {-1, -0.5}], 
                  {GrayLevel[0], LineBox[{{-5, 0}, {5, 0}}]}, 
                  {GrayLevel[0.5], 
                   CircleBox[{-1.6732547050778708`, 0.5}, 0.5], 
                   DiskBox[{-1.6732547050778708`, 0.5}, 0.1]}, 
                  {GrayLevel[0.5], 
                   DiskBox[{-1.6579962860487345`, 1.9999223915418862`}, 0.1]}, 
                  {GrayLevel[0.5], Thickness[0.01], 
                   LineBox[{{-1.6732547050778708`, 
                    0.5}, {-1.6579962860487345`, 1.9999223915418862`}}]}, 
                  {RGBColor[0, 1, 0], 
                   DiskBox[{-1.6630824257251131`, 1.499948261027924}, 0.05]}, 
                  {RGBColor[0, 0, 1], Arrowheads[0.01], 
                   CircleBox[{-1.6732547050778708`, 0.5}, 0.3333333333333333, 
                    NCache[{
                    Rational[-1, 2] 
                    Pi, -1.7208469245494962`}, {-1.5707963267948966`, \
-1.7208469245494962`}]], 
                   ArrowBox[{{-1.7197857774683933`, 
                    0.16993035520772432`}, {-1.723084092370752, 
                    0.17041216152120064`}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-1.6732547050778708`, 0.5}, {-1.573220973241471,
                     0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-1.6732547050778708`, 
                    0.5}, {-1.7480678216098844`, 0.5}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.005], Arrowheads[0.04], 
                   ArrowBox[{{-1.6732547050778708`, 
                    0.5}, {-1.6480340897734849`, 0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-1.6732547050778708`, 0.5}, 1.6, {
                    1.5606238720040824`, 1.5138632548253073`}], 
                   ArrowBox[{{-1.5981893549169262`, 
                    2.0982381528436918`}, {-1.5822109929964139`, 
                    2.0974076006111995`}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Dashing[{Small, Small}], 
                   Arrowheads[0.02], 
                   CircleBox[{-1.6732547050778708`, 0.5}, 1.7, {
                    1.5606238720040824`, 1.575239508914893}], 
                   ArrowBox[{{-1.6638081632974466`, 
                    2.1999737535763284`}, {-1.6808080898288207`, 
                    2.199983219440358}}]}, 
                  {RGBColor[0, 0, 1], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-1.6732547050778708`, 0.5}, 1.8, {
                    1.5606238720040824`, 1.5884110197364156`}], 
                   ArrowBox[{{-1.6869610199149134`, 
                    2.299947815058422}, {-1.7049595127656643`, 
                    2.299720757553649}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.003], Arrowheads[0.025], 
                   CircleBox[{-1.6732547050778708`, 0.5}, 1.5, {
                    1.5606238720040824`, 1.560654846124441}], 
                   ArrowBox[{{-1.6430445267759135`, 
                    1.99969575085314}, {-1.6580427448327304`, 
                    1.9999228634384836`}}]}, 
                  {RGBColor[0, 0, 1], ArrowBox[{{0, 2.8}, {0, 2.8}}]}},
                 Frame->True,
                 FrameTicks->Automatic,
                 ImageSize->800,
                 PlotRange->{{-3, 3}, {-1, 3}}], 444->
                GraphicsBox[{InsetBox[
                   TagBox[GridBox[{
                    {"\<\"t\"\>", "22.150000000000002`"},
                    {"\<\"driveIntent\"\>", "0"},
                    {"\<\"motorTorque\"\>", 
                    RowBox[{"-", "4.915760349196269`"}]},
                    {"\<\"vw\"\>", 
                    RowBox[{"-", "0.049360347071458865`"}]},
                    {"\<\"controllerState\"\>", 
                    RowBox[{"\[LeftAssociation]", 
                    RowBox[{
                    
                    RowBox[{"\<\"lastError\"\>", "\[Rule]", 
                    "0.004804980786403247`"}], ",", 
                    RowBox[{"\<\"lastInt\"\>", "\[Rule]", 
                    RowBox[{"-", "0.04187903860410745`"}]}], ",", 
                    RowBox[{"\<\"initialized\"\>", "\[Rule]", "True"}]}], 
                    "\[RightAssociation]"}]}
                    },
                    AutoDelete->False,
                    
                    GridBoxItemSize->{
                    "Columns" -> {{Automatic}}, "Rows" -> {{Automatic}}}],
                    "Grid"], {-1, -0.5}], 
                  {GrayLevel[0], LineBox[{{-5, 0}, {5, 0}}]}, 
                  {GrayLevel[0.5], CircleBox[{-1.675732745449815, 0.5}, 0.5], 
                   DiskBox[{-1.675732745449815, 0.5}, 0.1]}, 
                  {GrayLevel[0.5], 
                   DiskBox[{-1.6607297601956457`, 1.9999249682679008`}, 0.1]}, 
                  {GrayLevel[0.5], Thickness[0.01], 
                   LineBox[{{-1.675732745449815, 0.5}, {-1.6607297601956457`, 
                    1.9999249682679008`}}]}, 
                  {RGBColor[0, 1, 0], 
                   DiskBox[{-1.6657307552803688`, 1.499949978845267}, 0.05]}, 
                  {RGBColor[0, 0, 1], Arrowheads[0.01], 
                   CircleBox[{-1.675732745449815, 0.5}, 0.3333333333333333, 
                    NCache[{
                    Rational[-1, 2] 
                    Pi, -1.7182691372707846`}, {-1.5707963267948966`, \
-1.7182691372707846`}]], 
                   ArrowBox[{{-1.7214128148521108`, 
                    0.1698115047877765}, {-1.7247123607886554`, 
                    0.17028480715567518`}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-1.675732745449815, 0.5}, {-1.5774175384658897`,
                     0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-1.675732745449815, 0.5}, {-1.7492935783393437`,
                     0.5}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.005], Arrowheads[0.04], 
                   ArrowBox[{{-1.675732745449815, 0.5}, {-1.6509783713554183`,
                     0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-1.675732745449815, 0.5}, 1.6, {
                    1.5607941698517476`, 1.5148163494321225`}], 
                   ArrowBox[{{-1.6021907013164105`, 
                    2.09830897130207}, {-1.5862115550584506`, 
                    2.0974936483350763`}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Dashing[{Small, Small}], 
                   Arrowheads[0.02], 
                   CircleBox[{-1.675732745449815, 0.5}, 1.7, {
                    1.5607941698517476`, 1.5751396398044148`}], 
                   ArrowBox[{{-1.6661164288501764`, 
                    2.199972801680973}, {-1.6831163543514258`, 
                    2.199983965312493}}]}, 
                  {RGBColor[0, 0, 1], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-1.675732745449815, 0.5}, 1.8, {
                    1.5607941698517476`, 1.5881039495695046`}], 
                   ArrowBox[{{-1.6888863493733441`, 
                    2.2999519389983227`}, {-1.7068849110981221`, 
                    2.299730408304372}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.003], Arrowheads[0.025], 
                   CircleBox[{-1.675732745449815, 0.5}, 1.5, {
                    1.5607941698517476`, 1.5607810963451458`}], 
                   ArrowBox[{{-1.6757098997751898`, 
                    1.999999999826025}, {-1.6607101509179536`, 
                    1.999924771998093}}]}, 
                  {RGBColor[0, 0, 1], ArrowBox[{{0, 2.8}, {0, 2.8}}]}},
                 Frame->True,
                 FrameTicks->Automatic,
                 ImageSize->800,
                 PlotRange->{{-3, 3}, {-1, 3}}], 445->
                GraphicsBox[{InsetBox[
                   TagBox[GridBox[{
                    {"\<\"t\"\>", "22.2`"},
                    {"\<\"driveIntent\"\>", "0"},
                    {"\<\"motorTorque\"\>", 
                    RowBox[{"-", "4.830253160462227`"}]},
                    {"\<\"vw\"\>", 
                    RowBox[{"-", "0.048957555648924386`"}]},
                    {"\<\"controllerState\"\>", 
                    RowBox[{"\[LeftAssociation]", 
                    RowBox[{
                    
                    RowBox[{"\<\"lastError\"\>", "\[Rule]", 
                    "0.00485407574706021`"}], ",", 
                    RowBox[{"\<\"lastInt\"\>", "\[Rule]", 
                    RowBox[{"-", "0.04163754350786985`"}]}], ",", 
                    RowBox[{"\<\"initialized\"\>", "\[Rule]", "True"}]}], 
                    "\[RightAssociation]"}]}
                    },
                    AutoDelete->False,
                    
                    GridBoxItemSize->{
                    "Columns" -> {{Automatic}}, "Rows" -> {{Automatic}}}],
                    "Grid"], {-1, -0.5}], 
                  {GrayLevel[0], LineBox[{{-5, 0}, {5, 0}}]}, 
                  {GrayLevel[0.5], 
                   CircleBox[{-1.6781904601254831`, 0.5}, 0.5], 
                   DiskBox[{-1.6781904601254831`, 0.5}, 0.1]}, 
                  {GrayLevel[0.5], 
                   DiskBox[{-1.663442402893957, 1.999927496517047}, 0.1]}, 
                  {GrayLevel[0.5], Thickness[0.01], 
                   LineBox[{{-1.6781904601254831`, 0.5}, {-1.663442402893957, 
                    1.999927496517047}}]}, 
                  {RGBColor[0, 1, 0], 
                   DiskBox[{-1.6683584219711323`, 1.499951664344698}, 0.05]}, 
                  {RGBColor[0, 0, 1], Arrowheads[0.01], 
                   CircleBox[{-1.6781904601254831`, 0.5}, 0.3333333333333333, 
                    NCache[{
                    Rational[-1, 2] 
                    Pi, -1.7157039216087635`}, {-1.5707963267948966`, \
-1.7157039216087635`}]], 
                   ArrowBox[{{-1.7230233754624547`, 
                    0.16969541206108107`}, {-1.726324124664268, 
                    0.17016024883410608`}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-1.6781904601254831`, 
                    0.5}, {-1.5815853969162386`, 0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-1.6781904601254831`, 
                    0.5}, {-1.7505014824806542`, 0.5}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.005], Arrowheads[0.04], 
                   ArrowBox[{{-1.6781904601254831`, 
                    0.5}, {-1.6538964192714096`, 0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-1.6781904601254831`, 0.5}, 1.6, {
                    1.5609641302248138`, 1.5157675566468307`}], 
                   ArrowBox[{{-1.606168772057891, 
                    2.0983782019434245`}, {-1.5901888574878862`, 
                    2.0975780788221945`}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Dashing[{Small, Small}], 
                   Arrowheads[0.02], 
                   CircleBox[{-1.6781904601254831`, 0.5}, 1.7, {
                    1.5609641302248138`, 1.5750428555537834`}], 
                   ArrowBox[{{-1.668409612977436, 
                    2.199971863010993}, {-1.6854095373186007`, 
                    2.199984671967509}}]}, 
                  {RGBColor[0, 0, 1], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-1.6781904601254831`, 0.5}, 1.8, {
                    1.5609641302248138`, 1.5877988700051595`}], 
                   ArrowBox[{{-1.6907949348920126`, 
                    2.2999558681300662`}, {-1.708793563363682, 
                    2.2997398284397086`}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.003], Arrowheads[0.025], 
                   CircleBox[{-1.6781904601254831`, 0.5}, 1.5, {
                    1.5609641302248138`, 1.5609120281842122`}], 
                   ArrowBox[{{-1.6783640122090693`, 
                    1.9999999899598915`}, {-1.6633642536306867`, 
                    1.999926726077302}}]}, 
                  {RGBColor[0, 0, 1], ArrowBox[{{0, 2.8}, {0, 2.8}}]}},
                 Frame->True,
                 FrameTicks->Automatic,
                 ImageSize->800,
                 PlotRange->{{-3, 3}, {-1, 3}}], 446->
                GraphicsBox[{InsetBox[
                   TagBox[GridBox[{
                    {"\<\"t\"\>", "22.25`"},
                    {"\<\"driveIntent\"\>", "0"},
                    {"\<\"motorTorque\"\>", 
                    RowBox[{"-", "4.7456667163576975`"}]},
                    {"\<\"vw\"\>", 
                    RowBox[{"-", "0.048562260934532205`"}]},
                    {"\<\"controllerState\"\>", 
                    RowBox[{"\[LeftAssociation]", 
                    RowBox[{
                    
                    RowBox[{"\<\"lastError\"\>", "\[Rule]", 
                    "0.00490406139955955`"}], ",", 
                    RowBox[{"\<\"lastInt\"\>", "\[Rule]", 
                    RowBox[{"-", "0.04139356670664991`"}]}], ",", 
                    RowBox[{"\<\"initialized\"\>", "\[Rule]", "True"}]}], 
                    "\[RightAssociation]"}]}
                    },
                    AutoDelete->False,
                    GridBoxItemSize->{
                    "Columns" -> {{Automatic}}, "Rows" -> {{Automatic}}}],
                    "Grid"], {-1, -0.5}], 
                  {GrayLevel[0], LineBox[{{-5, 0}, {5, 0}}]}, 
                  {GrayLevel[0.5], 
                   CircleBox[{-1.6806282269618782`, 0.5}, 0.5], 
                   DiskBox[{-1.6806282269618782`, 0.5}, 0.1]}, 
                  {GrayLevel[0.5], 
                   DiskBox[{-1.666133030256752, 1.999929961455694}, 0.1]}, 
                  {GrayLevel[0.5], Thickness[0.01], 
                   LineBox[{{-1.6806282269618782`, 0.5}, {-1.666133030256752, 
                    1.999929961455694}}]}, 
                  {RGBColor[0, 1, 0], 
                   DiskBox[{-1.670964762491794, 1.4999533076371292`}, 0.05]}, 
                  {RGBColor[0, 0, 1], Arrowheads[0.01], 
                   CircleBox[{-1.6806282269618782`, 0.5}, 0.3333333333333333, 
                    NCache[{
                    Rational[-1, 2] 
                    Pi, -1.7131663282856275`}, {-1.5707963267948966`, \
-1.7131663282856275`}]], 
                   ArrowBox[{{-1.7246228201334441`, 
                    0.16958270795435132`}, {-1.7279247382732998`, 
                    0.1700391672806003}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-1.6806282269618782`, 
                    0.5}, {-1.5857148926347242`, 0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-1.6806282269618782`, 
                    0.5}, {-1.7516995687022072`, 0.5}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.005], Arrowheads[0.04], 
                   ArrowBox[{{-1.6806282269618782`, 
                    0.5}, {-1.6567862343750532`, 0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-1.6806282269618782`, 0.5}, 1.6, {
                    1.5611327119186735`, 1.5167110491785718`}], 
                   ArrowBox[{{-1.6101146286230417`, 
                    2.0984454424375296`}, {-1.5941339662554446`, 
                    2.0976603966002436`}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Dashing[{Small, Small}], 
                   Arrowheads[0.02], 
                   CircleBox[{-1.6806282269618782`, 0.5}, 1.7, {
                    1.5611327119186735`, 1.574949492370815}], 
                   ArrowBox[{{-1.6706886650726023`, 
                    2.1999709424309137`}, {-1.6876885881437906`, 
                    2.1999853385544186`}}]}, 
                  {RGBColor[0, 0, 1], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-1.6806282269618782`, 0.5}, 1.8, {
                    1.5611327119186735`, 1.5874975270095495`}], 
                   ArrowBox[{{-1.6926902970710576`, 
                    2.2999595846753564`}, {-1.7106889898275672`, 
                    2.2997489687553463`}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.003], Arrowheads[0.025], 
                   CircleBox[{-1.6806282269618782`, 0.5}, 1.5, {
                    1.5611327119186735`, 1.5610464009507101`}], 
                   ArrowBox[{{-1.6810033381916887`, 
                    1.9999999530971877`}, {-1.6660035699040538`, 
                    1.9999287047743106`}}]}, 
                  {RGBColor[0, 0, 1], ArrowBox[{{0, 2.8}, {0, 2.8}}]}},
                 Frame->True,
                 FrameTicks->Automatic,
                 ImageSize->800,
                 PlotRange->{{-3, 3}, {-1, 3}}], 447->
                GraphicsBox[{InsetBox[
                   TagBox[GridBox[{
                    {"\<\"t\"\>", "22.3`"},
                    {"\<\"driveIntent\"\>", "0"},
                    {"\<\"motorTorque\"\>", 
                    RowBox[{"-", "4.662438390231425`"}]},
                    {"\<\"vw\"\>", 
                    RowBox[{"-", "0.048174309982733116`"}]},
                    {"\<\"controllerState\"\>", 
                    RowBox[{"\[LeftAssociation]", 
                    RowBox[{
                    
                    RowBox[{"\<\"lastError\"\>", "\[Rule]", 
                    "0.004953976814682751`"}], ",", 
                    RowBox[{"\<\"lastInt\"\>", "\[Rule]", 
                    RowBox[{"-", "0.04114708865123992`"}]}], ",", 
                    RowBox[{"\<\"initialized\"\>", "\[Rule]", "True"}]}], 
                    "\[RightAssociation]"}]}
                    },
                    AutoDelete->False,
                    
                    GridBoxItemSize->{
                    "Columns" -> {{Automatic}}, "Rows" -> {{Automatic}}}],
                    "Grid"], {-1, -0.5}], 
                  {GrayLevel[0], LineBox[{{-5, 0}, {5, 0}}]}, 
                  {GrayLevel[0.5], 
                   CircleBox[{-1.6830464170337076`, 0.5}, 0.5], 
                   DiskBox[{-1.6830464170337076`, 0.5}, 0.1]}, 
                  {GrayLevel[0.5], 
                   DiskBox[{-1.6688006443026995`, 1.9999323511276421`}, 0.1]}, 
                  {GrayLevel[0.5], Thickness[0.01], 
                   LineBox[{{-1.6830464170337076`, 
                    0.5}, {-1.6688006443026995`, 1.9999323511276421`}}]}, 
                  {RGBColor[0, 1, 0], 
                   DiskBox[{-1.6735492352130354`, 1.4999549007517614`}, 
                    0.05]}, 
                  {RGBColor[0, 0, 1], Arrowheads[0.01], 
                   CircleBox[{-1.6830464170337076`, 0.5}, 0.3333333333333333, 
                    NCache[{
                    Rational[-1, 2] 
                    Pi, -1.7106694785018393`}, {-1.5707963267948966`, \
-1.7106694785018393`}]], 
                   ArrowBox[{{-1.7262158715815128`, 
                    0.16947389013096698`}, {-1.7295189191380533`, 
                    0.16992210364934662`}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-1.6830464170337076`, 0.5}, {-1.589797649229079,
                     0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-1.6830464170337076`, 
                    0.5}, {-1.7528949203529136`, 0.5}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.005], Arrowheads[0.04], 
                   ArrowBox[{{-1.6830464170337076`, 
                    0.5}, {-1.6596461525482853`, 0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-1.6830464170337076`, 0.5}, 1.6, {
                    1.5612990021997286`, 1.5176417187145597`}], 
                   ArrowBox[{{-1.6140204734958623`, 
                    2.0985103750425615`}, {-1.5980390874309165`, 
                    2.0977402022587412`}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Dashing[{Small, Small}], 
                   Arrowheads[0.02], 
                   CircleBox[{-1.6830464170337076`, 0.5}, 1.7, {
                    1.5612990021997286`, 1.5748598168951782`}], 
                   ArrowBox[{{-1.6729544094817703`, 
                    2.1999700442606542`}, {-1.6899543311936254`, 
                    2.1999859648602866`}}]}, 
                  {RGBColor[0, 0, 1], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-1.6830464170337076`, 0.5}, 1.8, {
                    1.5612990021997286`, 1.5872014377010144`}], 
                   ArrowBox[{{-1.6945755378331224`, 
                    2.2999630772250836`}, {-1.7125742921617908`, 
                    2.299757790534721}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.003], Arrowheads[0.025], 
                   CircleBox[{-1.6830464170337076`, 0.5}, 1.5, {
                    1.5612990021997286`, 1.5611830790719339`}], 
                   ArrowBox[{{-1.6836265454348012`, 
                    1.9999998878170087`}, {-1.668626767549184, 
                    1.999930689634939}}]}, 
                  {RGBColor[0, 0, 1], ArrowBox[{{0, 2.8}, {0, 2.8}}]}},
                 Frame->True,
                 FrameTicks->Automatic,
                 ImageSize->800,
                 PlotRange->{{-3, 3}, {-1, 3}}], 448->
                GraphicsBox[{InsetBox[
                   TagBox[GridBox[{
                    {"\<\"t\"\>", "22.35`"},
                    {"\<\"driveIntent\"\>", "0"},
                    {"\<\"motorTorque\"\>", 
                    RowBox[{"-", "4.580944395079651`"}]},
                    {"\<\"vw\"\>", 
                    RowBox[{"-", "0.04779351904616806`"}]},
                    {"\<\"controllerState\"\>", 
                    RowBox[{"\[LeftAssociation]", 
                    RowBox[{
                    
                    RowBox[{"\<\"lastError\"\>", "\[Rule]", 
                    "0.005002974454089216`"}], ",", 
                    RowBox[{"\<\"lastInt\"\>", "\[Rule]", 
                    RowBox[{"-", "0.04089813492826249`"}]}], ",", 
                    RowBox[{"\<\"initialized\"\>", "\[Rule]", "True"}]}], 
                    "\[RightAssociation]"}]}
                    },
                    AutoDelete->False,
                    
                    GridBoxItemSize->{
                    "Columns" -> {{Automatic}}, "Rows" -> {{Automatic}}}],
                    "Grid"], {-1, -0.5}], 
                  {GrayLevel[0], LineBox[{{-5, 0}, {5, 0}}]}, 
                  {GrayLevel[0.5], 
                   CircleBox[{-1.6854453929608935`, 0.5}, 0.5], 
                   DiskBox[{-1.6854453929608935`, 0.5}, 0.1]}, 
                  {GrayLevel[0.5], 
                   DiskBox[{-1.671444422388611, 1.9999346561844065`}, 0.1]}, 
                  {GrayLevel[0.5], Thickness[0.01], 
                   LineBox[{{-1.6854453929608935`, 0.5}, {-1.671444422388611, 
                    1.9999346561844065`}}]}, 
                  {RGBColor[0, 1, 0], 
                   DiskBox[{-1.6761114125793717`, 1.499956437456271}, 0.05]}, 
                  {RGBColor[0, 0, 1], Arrowheads[0.01], 
                   CircleBox[{-1.6854453929608935`, 0.5}, 0.3333333333333333, 
                    NCache[{
                    Rational[-1, 2] 
                    Pi, -1.708224658647286}, {-1.5707963267948966`, \
-1.708224658647286}]], 
                   ArrowBox[{{-1.7278066425028793`, 
                    0.169369336497123}, {-1.7311107759882465`, 
                    0.16980947332778273`}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-1.6854453929608935`, 
                    0.5}, {-1.5938265050593003`, 0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-1.6854453929608935`, 
                    0.5}, {-1.7540937113665795`, 0.5}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.005], Arrowheads[0.04], 
                   ArrowBox[{{-1.6854453929608935`, 
                    0.5}, {-1.6624748234649867`, 0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-1.6854453929608935`, 0.5}, 1.6, {
                    1.5614622108737035`, 1.5185551427291173`}], 
                   ArrowBox[{{-1.617879595779761, 
                    2.0985727581349805`}, {-1.6018975128875788`, 
                    2.097817183452242}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Dashing[{Small, Small}], 
                   Arrowheads[0.02], 
                   CircleBox[{-1.6854453929608935`, 0.5}, 1.7, {
                    1.5614622108737035`, 1.5747740315966539`}], 
                   ArrowBox[{{-1.6752075530085397`, 
                    2.1999691722596353`}, {-1.6922074732920893`, 
                    2.1999865512025663`}}]}, 
                  {RGBColor[0, 0, 1], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-1.6854453929608935`, 0.5}, 1.8, {
                    1.5614622108737035`, 1.5869119019574793`}], 
                   ArrowBox[{{-1.696453359636441, 
                    2.299966339871296}, {-1.714452172648517, 
                    2.2997662644721824`}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.003], Arrowheads[0.025], 
                   CircleBox[{-1.6854453929608935`, 0.5}, 1.5, {
                    1.5614622108737035`, 1.5613210309726864`}], 
                   ArrowBox[{{-1.6862324491914635`, 
                    1.9999997935141491`}, {-1.671232661902554, 
                    1.9999326645806015`}}]}, 
                  {RGBColor[0, 0, 1], ArrowBox[{{0, 2.8}, {0, 2.8}}]}},
                 Frame->True,
                 FrameTicks->Automatic,
                 ImageSize->800,
                 PlotRange->{{-3, 3}, {-1, 3}}], 449->
                GraphicsBox[{InsetBox[
                   TagBox[GridBox[{
                    {"\<\"t\"\>", "22.400000000000002`"},
                    {"\<\"driveIntent\"\>", "0"},
                    {"\<\"motorTorque\"\>", 
                    RowBox[{"-", "4.501503165199795`"}]},
                    {"\<\"vw\"\>", 
                    RowBox[{"-", "0.04741967870554318`"}]},
                    {"\<\"controllerState\"\>", 
                    RowBox[{"\[LeftAssociation]", 
                    RowBox[{
                    
                    RowBox[{"\<\"lastError\"\>", "\[Rule]", 
                    "0.0050503152930336265`"}], ",", 
                    RowBox[{"\<\"lastInt\"\>", "\[Rule]", 
                    RowBox[{"-", "0.040646770712759926`"}]}], ",", 
                    RowBox[{"\<\"initialized\"\>", "\[Rule]", "True"}]}], 
                    "\[RightAssociation]"}]}
                    },
                    AutoDelete->False,
                    
                    GridBoxItemSize->{
                    "Columns" -> {{Automatic}}, "Rows" -> {{Automatic}}}],
                    "Grid"], {-1, -0.5}], 
                  {GrayLevel[0], LineBox[{{-5, 0}, {5, 0}}]}, 
                  {GrayLevel[0.5], CircleBox[{-1.687825507500692, 0.5}, 0.5], 
                   DiskBox[{-1.687825507500692, 0.5}, 0.1]}, 
                  {GrayLevel[0.5], 
                   DiskBox[{-1.6740637061635746`, 1.9999368696128372`}, 0.1]}, 
                  {GrayLevel[0.5], Thickness[0.01], 
                   LineBox[{{-1.687825507500692, 0.5}, {-1.6740637061635746`, 
                    1.9999368696128372`}}]}, 
                  {RGBColor[0, 1, 0], 
                   DiskBox[{-1.6786509732759471`, 1.499957913075225}, 0.05]}, 
                  {RGBColor[0, 0, 1], Arrowheads[0.01], 
                   CircleBox[{-1.687825507500692, 0.5}, 0.3333333333333333, 
                    NCache[{
                    Rational[-1, 2] 
                    Pi, -1.7058414217508904`}, {-1.5707963267948966`, \
-1.7058414217508904`}]], 
                   ArrowBox[{{-1.729398666290221, 
                    0.1692693186603763}, {-1.7327038393414953`, 
                    0.16970157971570943`}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-1.687825507500692, 0.5}, {-1.5977954441966962`,
                     0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-1.687825507500692, 0.5}, {-1.7553012520759022`,
                     0.5}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.005], Arrowheads[0.04], 
                   ArrowBox[{{-1.687825507500692, 0.5}, {-1.6652711887719063`,
                     0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-1.687825507500692, 0.5}, 1.6, {
                    1.5616216638586742`, 1.519447548520348}], 
                   ArrowBox[{{-1.621686312621568, 
                    2.098632417693555}, {-1.6057035618142226`, 
                    2.097891105813118}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Dashing[{Small, Small}], 
                   Arrowheads[0.02], 
                   CircleBox[{-1.687825507500692, 0.5}, 1.7, {
                    1.5616216638586742`, 1.5746922801696457`}], 
                   ArrowBox[{{-1.6774486926770023`, 
                    2.199968329620912}, {-1.6944486114829906`, 
                    2.1999870983315253`}}]}, 
                  {RGBColor[0, 0, 1], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-1.687825507500692, 0.5}, 1.8, {
                    1.5616216638586742`, 1.5866300147764507`}], 
                   ArrowBox[{{-1.6983260863081178`, 
                    2.299969371362943}, {-1.716324955003787, 
                    2.2997743696063733`}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.003], Arrowheads[0.025], 
                   CircleBox[{-1.687825507500692, 0.5}, 1.5, {
                    1.5616216638586742`, 1.5614593270848927`}], 
                   ArrowBox[{{-1.6888200078628275`, 
                    1.9999996703229737`}, {-1.673820211433689, 
                    1.9999346158023275`}}]}, 
                  {RGBColor[0, 0, 1], ArrowBox[{{0, 2.8}, {0, 2.8}}]}},
                 Frame->True,
                 FrameTicks->Automatic,
                 ImageSize->800,
                 PlotRange->{{-3, 3}, {-1, 3}}], 450->
                GraphicsBox[{InsetBox[
                   TagBox[GridBox[{
                    {"\<\"t\"\>", "22.45`"},
                    {"\<\"driveIntent\"\>", "0"},
                    {"\<\"motorTorque\"\>", 
                    RowBox[{"-", "4.424378904375075`"}]},
                    {"\<\"vw\"\>", 
                    RowBox[{"-", "0.04705255865919924`"}]},
                    {"\<\"controllerState\"\>", 
                    RowBox[{"\[LeftAssociation]", 
                    RowBox[{
                    
                    RowBox[{"\<\"lastError\"\>", "\[Rule]", 
                    "0.005095363461808116`"}], ",", 
                    RowBox[{"\<\"lastInt\"\>", "\[Rule]", 
                    RowBox[{"-", "0.040393095477381326`"}]}], ",", 
                    RowBox[{"\<\"initialized\"\>", "\[Rule]", "True"}]}], 
                    "\[RightAssociation]"}]}
                    },
                    AutoDelete->False,
                    
                    GridBoxItemSize->{
                    "Columns" -> {{Automatic}}, "Rows" -> {{Automatic}}}],
                    "Grid"], {-1, -0.5}], 
                  {GrayLevel[0], LineBox[{{-5, 0}, {5, 0}}]}, 
                  {GrayLevel[0.5], 
                   CircleBox[{-1.6901871023876656`, 0.5}, 0.5], 
                   DiskBox[{-1.6901871023876656`, 0.5}, 0.1]}, 
                  {GrayLevel[0.5], 
                   DiskBox[{-1.6766579902053178`, 1.9999389864669688`}, 0.1]}, 
                  {GrayLevel[0.5], Thickness[0.01], 
                   LineBox[{{-1.6901871023876656`, 
                    0.5}, {-1.6766579902053178`, 1.9999389864669688`}}]}, 
                  {RGBColor[0, 1, 0], 
                   DiskBox[{-1.6811676942661002`, 1.4999593243113125`}, 
                    0.05]}, 
                  {RGBColor[0, 0, 1], Arrowheads[0.01], 
                   CircleBox[{-1.6901871023876656`, 0.5}, 0.3333333333333333, 
                    NCache[{
                    Rational[-1, 2] 
                    Pi, -1.7035276939261488`}, {-1.5707963267948966`, \
-1.7035276939261488`}]], 
                   ArrowBox[{{-1.7309949298025271`, 
                    0.1691740150278549}, {-1.7343010941404924`, 
                    0.16959862766213785`}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-1.6901871023876656`, 
                    0.5}, {-1.6016995243001642`, 0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-1.6901871023876656`, 
                    0.5}, {-1.7565220380502278`, 0.5}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.005], Arrowheads[0.04], 
                   ArrowBox[{{-1.6901871023876656`, 
                    0.5}, {-1.6680344599627261`, 0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-1.6901871023876656`, 0.5}, 1.6, {
                    1.5617767963811293`, 1.5203157751364351`}], 
                   ArrowBox[{{-1.625435907476981, 
                    2.0986892389572276`}, {-1.609452519067049, 
                    2.097961804004103}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Dashing[{Small, Small}], 
                   Arrowheads[0.02], 
                   CircleBox[{-1.6901871023876656`, 0.5}, 1.7, {
                    1.5617767963811293`, 1.5746146528728269`}], 
                   ArrowBox[{{-1.6796783236492676`, 
                    2.1999675189748267`}, {-1.696678240947071, 
                    2.1999876073431253`}}]}, 
                  {RGBColor[0, 0, 1], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-1.6901871023876656`, 0.5}, 1.8, {
                    1.5617767963811293`, 1.586356679183213}], 
                   ArrowBox[{{-1.7001956851130253`, 
                    2.299972174304878}, {-1.7181946064372455`, 
                    2.299782092286984}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.003], Arrowheads[0.025], 
                   CircleBox[{-1.6901871023876656`, 0.5}, 1.5, {
                    1.5617767963811293`, 1.5615971371065467`}], 
                   ArrowBox[{{-1.6913883177267512`, 
                    1.9999995190271593`}, {-1.6763885124748832`, 
                    1.9999365316293949`}}]}, 
                  {RGBColor[0, 0, 1], ArrowBox[{{0, 2.8}, {0, 2.8}}]}},
                 Frame->True,
                 FrameTicks->Automatic,
                 ImageSize->800,
                 PlotRange->{{-3, 3}, {-1, 3}}], 451->
                GraphicsBox[{InsetBox[
                   TagBox[GridBox[{
                    {"\<\"t\"\>", "22.5`"},
                    {"\<\"driveIntent\"\>", "0"},
                    {"\<\"motorTorque\"\>", 
                    RowBox[{"-", "4.3497852475636485`"}]},
                    {"\<\"vw\"\>", 
                    RowBox[{"-", "0.04669191216545951`"}]},
                    {"\<\"controllerState\"\>", 
                    RowBox[{"\[LeftAssociation]", 
                    RowBox[{
                    
                    RowBox[{"\<\"lastError\"\>", "\[Rule]", 
                    "0.005137580512360159`"}], ",", 
                    RowBox[{"\<\"lastInt\"\>", "\[Rule]", 
                    RowBox[{"-", "0.04013723797948121`"}]}], ",", 
                    RowBox[{"\<\"initialized\"\>", "\[Rule]", "True"}]}], 
                    "\[RightAssociation]"}]}
                    },
                    AutoDelete->False,
                    
                    GridBoxItemSize->{
                    "Columns" -> {{Automatic}}, "Rows" -> {{Automatic}}}],
                    "Grid"], {-1, -0.5}], 
                  {GrayLevel[0], LineBox[{{-5, 0}, {5, 0}}]}, 
                  {GrayLevel[0.5], 
                   CircleBox[{-1.6925305074042973`, 0.5}, 0.5], 
                   DiskBox[{-1.6925305074042973`, 0.5}, 0.1]}, 
                  {GrayLevel[0.5], 
                   DiskBox[{-1.6792269104840838`, 1.9999410036094702`}, 0.1]}, 
                  {GrayLevel[0.5], Thickness[0.01], 
                   LineBox[{{-1.6925305074042973`, 
                    0.5}, {-1.6792269104840838`, 1.9999410036094702`}}]}, 
                  {RGBColor[0, 1, 0], 
                   DiskBox[{-1.6836614427908216`, 1.49996066907298}, 0.05]}, 
                  {RGBColor[0, 0, 1], Arrowheads[0.01], 
                   CircleBox[{-1.6925305074042973`, 0.5}, 0.3333333333333333, 
                    NCache[{
                    Rational[-1, 2] 
                    Pi, -1.701289884221806}, {-1.5707963267948966`, \
-1.701289884221806}]], 
                   ArrowBox[{{-1.73259790765892, 
                    0.16908352330543142`}, {-1.7359050139200773`, 
                    0.16950073631606433`}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-1.6925305074042973`, 
                    0.5}, {-1.6055348024530243`, 0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-1.6925305074042973`, 
                    0.5}, {-1.7577598011436029`, 0.5}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.005], Arrowheads[0.04], 
                   ArrowBox[{{-1.6925305074042973`, 
                    0.5}, {-1.6707640961923298`, 0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-1.6925305074042973`, 0.5}, 1.6, {
                    1.5619271459034139`, 1.5211572337979102`}], 
                   ArrowBox[{{-1.6291245661656184`, 
                    2.0987431584265304`}, {-1.6131405713078568`, 
                    2.0980291730899667`}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Dashing[{Small, Small}], 
                   Arrowheads[0.02], 
                   CircleBox[{-1.6925305074042973`, 0.5}, 1.7, {
                    1.5619271459034139`, 1.5745411917692032`}], 
                   ArrowBox[{{-1.6818968472042988`, 
                    2.1999667424013776`}, {-1.6988967629805183`, 
                    2.1999880796023064`}}]}, 
                  {RGBColor[0, 0, 1], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-1.6925305074042973`, 0.5}, 1.8, {
                    1.5619271459034139`, 1.5860926195009897`}], 
                   ArrowBox[{{-1.702063789705886, 
                    2.2999747544142273`}, {-1.720062760595593, 
                    2.2997894251923503`}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.003], Arrowheads[0.025], 
                   CircleBox[{-1.6925305074042973`, 0.5}, 1.5, {
                    1.5619271459034139`, 1.5617337266225289`}], 
                   ArrowBox[{{-1.693936606939818, 
                    1.9999993409612207`}, {-1.6789367932244552`, 
                    1.9999384023801767`}}]}, 
                  {RGBColor[0, 0, 1], ArrowBox[{{0, 2.8}, {0, 2.8}}]}},
                 Frame->True,
                 FrameTicks->Automatic,
                 ImageSize->800,
                 PlotRange->{{-3, 3}, {-1, 3}}], 452->
                GraphicsBox[{InsetBox[
                   TagBox[GridBox[{
                    {"\<\"t\"\>", "22.55`"},
                    {"\<\"driveIntent\"\>", "0"},
                    {"\<\"motorTorque\"\>", 
                    RowBox[{"-", "4.277888988042129`"}]},
                    {"\<\"vw\"\>", 
                    RowBox[{"-", "0.04633748013496024`"}]},
                    {"\<\"controllerState\"\>", 
                    RowBox[{"\[LeftAssociation]", 
                    RowBox[{
                    
                    RowBox[{"\<\"lastError\"\>", "\[Rule]", 
                    "0.0051765194079247245`"}], ",", 
                    RowBox[{"\<\"lastInt\"\>", "\[Rule]", 
                    RowBox[{"-", "0.03987935154233988`"}]}], ",", 
                    RowBox[{"\<\"initialized\"\>", "\[Rule]", "True"}]}], 
                    "\[RightAssociation]"}]}
                    },
                    AutoDelete->False,
                    
                    GridBoxItemSize->{
                    "Columns" -> {{Automatic}}, "Rows" -> {{Automatic}}}],
                    "Grid"], {-1, -0.5}], 
                  {GrayLevel[0], LineBox[{{-5, 0}, {5, 0}}]}, 
                  {GrayLevel[0.5], 
                   CircleBox[{-1.6948560396648056`, 0.5}, 0.5], 
                   DiskBox[{-1.6948560396648056`, 0.5}, 0.1]}, 
                  {GrayLevel[0.5], 
                   DiskBox[{-1.6817702327843205`, 1.9999429194666998`}, 0.1]}, 
                  {GrayLevel[0.5], Thickness[0.01], 
                   LineBox[{{-1.6948560396648056`, 
                    0.5}, {-1.6817702327843205`, 1.9999429194666998`}}]}, 
                  {RGBColor[0, 1, 0], 
                   DiskBox[{-1.6861321684111488`, 1.4999619463111333`}, 
                    0.05]}, 
                  {RGBColor[0, 0, 1], Arrowheads[0.01], 
                   CircleBox[{-1.6948560396648056`, 0.5}, 0.3333333333333333, 
                    NCache[{
                    Rational[-1, 2] 
                    Pi, -1.6991329964361603`}, {-1.5707963267948966`, \
-1.6991329964361603`}]], 
                   ArrowBox[{{-1.7342095975661493`, 
                    0.16899787222494067`}, {-1.7375175960156568`, 
                    0.16940795121353225`}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-1.6948560396648056`, 0.5}, {-1.609298259903963,
                     0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-1.6948560396648056`, 
                    0.5}, {-1.7590175620115374`, 0.5}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.005], Arrowheads[0.04], 
                   ArrowBox[{{-1.6948560396648056`, 
                    0.5}, {-1.6734597822506947`, 0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-1.6948560396648056`, 0.5}, 1.6, {
                    1.5620723448810612`, 1.5219698673671518`}], 
                   ArrowBox[{{-1.6327493115780622`, 
                    2.0987941563335664`}, {-1.6167647417896136`, 
                    2.0980931603621134`}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Dashing[{Small, Small}], 
                   Arrowheads[0.02], 
                   CircleBox[{-1.6948560396648056`, 0.5}, 1.7, {
                    1.5620723448810612`, 1.5744718958278536`}], 
                   ArrowBox[{{-1.6841045786947548`, 
                    2.199966001450326}, {-1.7011044929515766`, 
                    2.1999885166763695`}}]}, 
                  {RGBColor[0, 0, 1], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-1.6948560396648056`, 0.5}, 1.8, {
                    1.5620723448810612`, 1.5858383948146286`}], 
                   ArrowBox[{{-1.7039317236458555`, 
                    2.2999771198435486`}, {-1.7219307410691869`, 
                    2.29979636641034}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.003], Arrowheads[0.025], 
                   CircleBox[{-1.6948560396648056`, 0.5}, 1.5, {
                    1.5620723448810612`, 1.5618684531928702`}], 
                   ArrowBox[{{-1.6964642289536755`, 
                    1.9999991379088227`}, {-1.6814644071643994`, 
                    1.9999402202017833`}}]}, 
                  {RGBColor[0, 0, 1], ArrowBox[{{0, 2.8}, {0, 2.8}}]}},
                 Frame->True,
                 FrameTicks->Automatic,
                 ImageSize->800,
                 PlotRange->{{-3, 3}, {-1, 3}}], 453->
                GraphicsBox[{InsetBox[
                   TagBox[GridBox[{
                    {"\<\"t\"\>", "22.6`"},
                    {"\<\"driveIntent\"\>", "0"},
                    {"\<\"motorTorque\"\>", 
                    RowBox[{"-", "4.208813826830834`"}]},
                    {"\<\"vw\"\>", 
                    RowBox[{"-", "0.045988994873839686`"}]},
                    {"\<\"controllerState\"\>", 
                    RowBox[{"\[LeftAssociation]", 
                    RowBox[{
                    
                    RowBox[{"\<\"lastError\"\>", "\[Rule]", 
                    "0.00521181832495035`"}], ",", 
                    RowBox[{"\<\"lastInt\"\>", "\[Rule]", 
                    RowBox[{"-", "0.03961960964204292`"}]}], ",", 
                    RowBox[{"\<\"initialized\"\>", "\[Rule]", "True"}]}], 
                    "\[RightAssociation]"}]}
                    },
                    AutoDelete->False,
                    
                    GridBoxItemSize->{
                    "Columns" -> {{Automatic}}, "Rows" -> {{Automatic}}}],
                    "Grid"], {-1, -0.5}], 
                  {GrayLevel[0], LineBox[{{-5, 0}, {5, 0}}]}, 
                  {GrayLevel[0.5], 
                   CircleBox[{-1.6971640030946606`, 0.5}, 0.5], 
                   DiskBox[{-1.6971640030946606`, 0.5}, 0.1]}, 
                  {GrayLevel[0.5], 
                   DiskBox[{-1.6842878411999456`, 1.9999447338001695`}, 0.1]}, 
                  {GrayLevel[0.5], Thickness[0.01], 
                   LineBox[{{-1.6971640030946606`, 
                    0.5}, {-1.6842878411999456`, 1.9999447338001695`}}]}, 
                  {RGBColor[0, 1, 0], 
                   DiskBox[{-1.6885798951648505`, 1.4999631558667796`}, 
                    0.05]}, 
                  {RGBColor[0, 0, 1], Arrowheads[0.01], 
                   CircleBox[{-1.6971640030946606`, 0.5}, 0.3333333333333333, 
                    NCache[{
                    Rational[-1, 2] 
                    Pi, -1.6970607415998216`}, {-1.5707963267948966`, \
-1.6970607415998216`}]], 
                   ArrowBox[{{-1.735831556230019, 
                    0.16891703238367073`}, {-1.7391403973644184`, 
                    0.16932025548089435`}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-1.6971640030946606`, 0.5}, {-1.612987726558044,
                     0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-1.6971640030946606`, 0.5}, {-1.760297683425629,
                     0.5}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.005], Arrowheads[0.04], 
                   ArrowBox[{{-1.6971640030946606`, 
                    0.5}, {-1.6761214068890122`, 0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-1.6971640030946606`, 0.5}, 1.6, {
                    1.5622121134388607`, 1.5227521093623584`}], 
                   ArrowBox[{{-1.6363079378127783`, 
                    2.098842249666429}, {-1.6203228245663972`, 
                    2.0981537577099356`}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Dashing[{Small, Small}], 
                   Arrowheads[0.02], 
                   CircleBox[{-1.6971640030946606`, 0.5}, 1.7, {
                    1.5622121134388607`, 1.574406725854436}], 
                   ArrowBox[{{-1.6863017554083946`, 
                    2.1999652971679167`}, {-1.7033016681617992`, 
                    2.1999889202778715`}}]}, 
                  {RGBColor[0, 0, 1], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-1.6971640030946606`, 0.5}, 1.8, {
                    1.5622121134388607`, 1.5855944124768098`}], 
                   ArrowBox[{{-1.705800524184222, 
                    2.29997928057616}, {-1.7237995851724808`, 
                    2.299802918590637}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.003], Arrowheads[0.025], 
                   CircleBox[{-1.6971640030946606`, 0.5}, 1.5, {
                    1.5622121134388607`, 1.562000762007116}], 
                   ArrowBox[{{-1.6989706554761819`, 
                    1.999998912001996}, {-1.6839708260228643`, 
                    1.9999419789041015`}}]}, 
                  {RGBColor[0, 0, 1], ArrowBox[{{0, 2.8}, {0, 2.8}}]}},
                 Frame->True,
                 FrameTicks->Automatic,
                 ImageSize->800,
                 PlotRange->{{-3, 3}, {-1, 3}}], 454->
                GraphicsBox[{InsetBox[
                   TagBox[GridBox[{
                    {"\<\"t\"\>", "22.650000000000002`"},
                    {"\<\"driveIntent\"\>", "0"},
                    {"\<\"motorTorque\"\>", 
                    RowBox[{"-", "4.142644105951895`"}]},
                    {"\<\"vw\"\>", 
                    RowBox[{"-", "0.04564618348190581`"}]},
                    {"\<\"controllerState\"\>", 
                    RowBox[{"\[LeftAssociation]", 
                    RowBox[{
                    
                    RowBox[{"\<\"lastError\"\>", "\[Rule]", 
                    "0.005243194348169828`"}], ",", 
                    RowBox[{"\<\"lastInt\"\>", "\[Rule]", 
                    RowBox[{"-", "0.03935820180730845`"}]}], ",", 
                    RowBox[{"\<\"initialized\"\>", "\[Rule]", "True"}]}], 
                    "\[RightAssociation]"}]}
                    },
                    AutoDelete->False,
                    
                    GridBoxItemSize->{
                    "Columns" -> {{Automatic}}, "Rows" -> {{Automatic}}}],
                    "Grid"], {-1, -0.5}], 
                  {GrayLevel[0], LineBox[{{-5, 0}, {5, 0}}]}, 
                  {GrayLevel[0.5], 
                   CircleBox[{-1.6994546880884327`, 0.5}, 0.5], 
                   DiskBox[{-1.6994546880884327`, 0.5}, 0.1]}, 
                  {GrayLevel[0.5], 
                   DiskBox[{-1.6867797268049973`, 1.9999464474961974`}, 0.1]}, 
                  {GrayLevel[0.5], Thickness[0.01], 
                   LineBox[{{-1.6994546880884327`, 
                    0.5}, {-1.6867797268049973`, 1.9999464474961974`}}]}, 
                  {RGBColor[0, 1, 0], 
                   DiskBox[{-1.6910047138994757`, 1.4999642983307984`}, 
                    0.05]}, 
                  {RGBColor[0, 0, 1], Arrowheads[0.01], 
                   CircleBox[{-1.6994546880884327`, 0.5}, 0.3333333333333333, 
                    NCache[{
                    Rational[-1, 2] 
                    Pi, -1.6950756499734534`}, {-1.5707963267948966`, \
-1.6950756499734534`}]], 
                   ArrowBox[{{-1.7374649354423506`, 
                    0.168840926128839}, {-1.740774570491629, 
                    0.16923758008307965`}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-1.6994546880884327`, 
                    0.5}, {-1.6166018059693947`, 0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-1.6994546880884327`, 
                    0.5}, {-1.7616019237842664`, 0.5}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.005], Arrowheads[0.04], 
                   ArrowBox[{{-1.6994546880884327`, 
                    0.5}, {-1.6787490416652284`, 0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-1.6994546880884327`, 0.5}, 1.6, {
                    1.5623462520451092`, 1.5235028429606714`}], 
                   ArrowBox[{{-1.6397989444382952`, 
                    2.0988874858005326`}, {-1.6238133188225312`, 
                    2.0982109945985163`}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Dashing[{Small, Small}], 
                   Arrowheads[0.02], 
                   CircleBox[{-1.6994546880884327`, 0.5}, 1.7, {
                    1.5623462520451092`, 1.5743456092229078`}], 
                   ArrowBox[{{-1.6884885442698399`, 
                    2.1999646301290356`}, {-1.7054884555477303`, 
                    2.19998929221635}}]}, 
                  {RGBColor[0, 0, 1], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-1.6994546880884327`, 0.5}, 1.8, {
                    1.5623462520451092`, 1.5853609415226197`}], 
                   ArrowBox[{{-1.7076709660662708`, 
                    2.2999812478957082`}, {-1.7256700677393777`, 
                    2.2998090881728976`}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.003], Arrowheads[0.025], 
                   CircleBox[{-1.6994546880884327`, 0.5}, 1.5, {
                    1.5623462520451092`, 1.5621301811962265`}], 
                   ArrowBox[{{-1.70145546909714, 
                    1.999998665624525}, {-1.6864556324012054`, 
                    1.9999436737928662`}}]}, 
                  {RGBColor[0, 0, 1], ArrowBox[{{0, 2.8}, {0, 2.8}}]}},
                 Frame->True,
                 FrameTicks->Automatic,
                 ImageSize->800,
                 PlotRange->{{-3, 3}, {-1, 3}}], 455->
                GraphicsBox[{InsetBox[
                   TagBox[GridBox[{
                    {"\<\"t\"\>", "22.7`"},
                    {"\<\"driveIntent\"\>", "0"},
                    {"\<\"motorTorque\"\>", 
                    RowBox[{"-", "4.079428491621279`"}]},
                    {"\<\"vw\"\>", 
                    RowBox[{"-", "0.04530877091273254`"}]},
                    {"\<\"controllerState\"\>", 
                    RowBox[{"\[LeftAssociation]", 
                    RowBox[{
                    
                    RowBox[{"\<\"lastError\"\>", "\[Rule]", 
                    "0.005270437131430904`"}], ",", 
                    RowBox[{"\<\"lastInt\"\>", "\[Rule]", 
                    RowBox[{"-", "0.03909532983571987`"}]}], ",", 
                    RowBox[{"\<\"initialized\"\>", "\[Rule]", "True"}]}], 
                    "\[RightAssociation]"}]}
                    },
                    AutoDelete->False,
                    
                    GridBoxItemSize->{
                    "Columns" -> {{Automatic}}, "Rows" -> {{Automatic}}}],
                    "Grid"], {-1, -0.5}], 
                  {GrayLevel[0], LineBox[{{-5, 0}, {5, 0}}]}, 
                  {GrayLevel[0.5], 
                   CircleBox[{-1.7017283713289004`, 0.5}, 0.5], 
                   DiskBox[{-1.7017283713289004`, 0.5}, 0.1]}, 
                  {GrayLevel[0.5], 
                   DiskBox[{-1.689245976588157, 1.9999480623746733`}, 0.1]}, 
                  {GrayLevel[0.5], Thickness[0.01], 
                   LineBox[{{-1.7017283713289004`, 0.5}, {-1.689245976588157, 
                    1.9999480623746733`}}]}, 
                  {RGBColor[0, 1, 0], 
                   DiskBox[{-1.6934067748350714`, 1.4999653749164488`}, 
                    0.05]}, 
                  {RGBColor[0, 0, 1], Arrowheads[0.01], 
                   CircleBox[{-1.7017283713289004`, 0.5}, 0.3333333333333333, 
                    NCache[{
                    Rational[-1, 2] 
                    Pi, -1.6931791815435349`}, {-1.5707963267948966`, \
-1.6931791815435349`}]], 
                   ArrowBox[{{-1.7391105179767385`, 
                    0.1687694364598722}, {-1.7424208993155654`, 
                    0.16915981308618872`}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-1.7017283713289004`, 0.5}, {-1.620139801496475,
                     0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-1.7017283713289004`, 0.5}, {-1.76293149028457, 
                    0.5}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.005], Arrowheads[0.04], 
                   ArrowBox[{{-1.7017283713289004`, 
                    0.5}, {-1.6813429204521446`, 0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-1.7017283713289004`, 0.5}, 1.6, {
                    1.5624746342544127`, 1.5242213603843153`}], 
                   ArrowBox[{{-1.6432214714960602`, 
                    2.0989299367614427`}, {-1.6272353639360502`, 
                    2.0982649316835955`}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Dashing[{Small, Small}], 
                   Arrowheads[0.02], 
                   CircleBox[{-1.7017283713289004`, 0.5}, 1.7, {
                    1.5624746342544127`, 1.5742884443859413`}], 
                   ArrowBox[{{-1.690665049327348, 
                    2.1999640004736247`}, {-1.707664959167658, 
                    2.1999896343580545`}}]}, 
                  {RGBColor[0, 0, 1], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-1.7017283713289004`, 0.5}, 1.8, {
                    1.5624746342544127`, 1.585138125874531}], 
                   ArrowBox[{{-1.7095435851178034`, 
                    2.2999830339293297`}, {-1.7275427247039914`, 
                    2.299814884692264}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.003], Arrowheads[0.025], 
                   CircleBox[{-1.7017283713289004`, 0.5}, 1.5, {
                    1.5624746342544127`, 1.5622563168861832`}], 
                   ArrowBox[{{-1.7039183556878121`, 
                    1.999998401321984}, {-1.6889185121747705`, 
                    1.9999453015055086`}}]}, 
                  {RGBColor[0, 0, 1], ArrowBox[{{0, 2.8}, {0, 2.8}}]}},
                 Frame->True,
                 FrameTicks->Automatic,
                 ImageSize->800,
                 PlotRange->{{-3, 3}, {-1, 3}}], 456->
                GraphicsBox[{InsetBox[
                   TagBox[GridBox[{
                    {"\<\"t\"\>", "22.75`"},
                    {"\<\"driveIntent\"\>", "0"},
                    {"\<\"motorTorque\"\>", 
                    RowBox[{"-", "4.019183577837985`"}]},
                    {"\<\"vw\"\>", 
                    RowBox[{"-", "0.04497648270507549`"}]},
                    {"\<\"controllerState\"\>", 
                    RowBox[{"\[LeftAssociation]", 
                    RowBox[{
                    
                    RowBox[{"\<\"lastError\"\>", "\[Rule]", 
                    "0.005293402588916902`"}], ",", 
                    RowBox[{"\<\"lastInt\"\>", "\[Rule]", 
                    RowBox[{"-", "0.038831204326394794`"}]}], ",", 
                    RowBox[{"\<\"initialized\"\>", "\[Rule]", "True"}]}], 
                    "\[RightAssociation]"}]}
                    },
                    AutoDelete->False,
                    
                    GridBoxItemSize->{
                    "Columns" -> {{Automatic}}, "Rows" -> {{Automatic}}}],
                    "Grid"], {-1, -0.5}], 
                  {GrayLevel[0], LineBox[{{-5, 0}, {5, 0}}]}, 
                  {GrayLevel[0.5], 
                   CircleBox[{-1.7039853157507387`, 0.5}, 0.5], 
                   DiskBox[{-1.7039853157507387`, 0.5}, 0.1]}, 
                  {GrayLevel[0.5], 
                   DiskBox[{-1.691686762726966, 1.9999495810171493`}, 0.1]}, 
                  {GrayLevel[0.5], Thickness[0.01], 
                   LineBox[{{-1.7039853157507387`, 0.5}, {-1.691686762726966, 
                    1.9999495810171493`}}]}, 
                  {RGBColor[0, 1, 0], 
                   DiskBox[{-1.6957862804015569`, 1.4999663873447662`}, 
                    0.05]}, 
                  {RGBColor[0, 0, 1], Arrowheads[0.01], 
                   CircleBox[{-1.7039853157507387`, 0.5}, 0.3333333333333333, 
                    NCache[{
                    Rational[-1, 2] 
                    Pi, -1.6913718341300361`}, {-1.5707963267948966`, \
-1.6913718341300361`}]], 
                   ArrowBox[{{-1.7407687529677438`, 
                    0.168702414953544}, {-1.7440798344456858`, 
                    0.16908680793638398`}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-1.7039853157507387`, 
                    0.5}, {-1.6236016441939791`, 0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-1.7039853157507387`, 
                    0.5}, {-1.7642870912851065`, 0.5}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.005], Arrowheads[0.04], 
                   ArrowBox[{{-1.7039853157507387`, 
                    0.5}, {-1.6839034197283467`, 0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-1.7039853157507387`, 0.5}, 1.6, {
                    1.5625971995806969`, 1.5249073230148902`}], 
                   ArrowBox[{{-1.6465752357824646`, 
                    2.0989696941211977`}, {-1.6305886758149468`, 
                    2.098315655071343}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Dashing[{Small, Small}], 
                   Arrowheads[0.02], 
                   CircleBox[{-1.7039853157507387`, 0.5}, 1.7, {
                    1.5625971995806969`, 1.5742351051462558`}], 
                   ArrowBox[{{-1.6928313189776893`, 
                    2.199963407946179}, {-1.7098312274265244`, 
                    2.1999899485928376`}}]}, 
                  {RGBColor[0, 0, 1], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-1.7039853157507387`, 0.5}, 1.8, {
                    1.5625971995806969`, 1.5849259972353522`}], 
                   ArrowBox[{{-1.7114187014151365`, 
                    2.2999846512616613`}, {-1.7294178762656252`, 
                    2.299820320161337}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.003], Arrowheads[0.025], 
                   CircleBox[{-1.7039853157507387`, 0.5}, 1.5, {
                    1.5625971995806969`, 1.5623788480703897`}], 
                   ArrowBox[{{-1.7063590966731725`, 
                    1.9999981217202014`}, {-1.6913592467663503`, 
                    1.9999468598527088`}}]}, 
                  {RGBColor[0, 0, 1], ArrowBox[{{0, 2.8}, {0, 2.8}}]}},
                 Frame->True,
                 FrameTicks->Automatic,
                 ImageSize->800,
                 PlotRange->{{-3, 3}, {-1, 3}}], 457->
                GraphicsBox[{InsetBox[
                   TagBox[GridBox[{
                    {"\<\"t\"\>", "22.8`"},
                    {"\<\"driveIntent\"\>", "0"},
                    {"\<\"motorTorque\"\>", 
                    RowBox[{"-", "3.96189738496435`"}]},
                    {"\<\"vw\"\>", 
                    RowBox[{"-", "0.04464904739706717`"}]},
                    {"\<\"controllerState\"\>", 
                    RowBox[{"\[LeftAssociation]", 
                    RowBox[{
                    
                    RowBox[{"\<\"lastError\"\>", "\[Rule]", 
                    "0.005312006673699181`"}], ",", 
                    RowBox[{"\<\"lastInt\"\>", "\[Rule]", 
                    RowBox[{"-", "0.03856604152608773`"}]}], ",", 
                    RowBox[{"\<\"initialized\"\>", "\[Rule]", "True"}]}], 
                    "\[RightAssociation]"}]}
                    },
                    AutoDelete->False,
                    
                    GridBoxItemSize->{
                    "Columns" -> {{Automatic}}, "Rows" -> {{Automatic}}}],
                    "Grid"], {-1, -0.5}], 
                  {GrayLevel[0], LineBox[{{-5, 0}, {5, 0}}]}, 
                  {GrayLevel[0.5], 
                   CircleBox[{-1.7062257706326316`, 0.5}, 0.5], 
                   DiskBox[{-1.7062257706326316`, 0.5}, 0.1]}, 
                  {GrayLevel[0.5], 
                   DiskBox[{-1.6941023322656301`, 1.9999510066139365`}, 0.1]}, 
                  {GrayLevel[0.5], Thickness[0.01], 
                   LineBox[{{-1.7062257706326316`, 
                    0.5}, {-1.6941023322656301`, 1.9999510066139365`}}]}, 
                  {RGBColor[0, 1, 0], 
                   DiskBox[{-1.6981434783879639`, 1.4999673377426244`}, 
                    0.05]}, 
                  {RGBColor[0, 0, 1], Arrowheads[0.01], 
                   CircleBox[{-1.7062257706326316`, 0.5}, 0.3333333333333333, 
                    NCache[{
                    Rational[-1, 2] 
                    Pi, -1.6896532483438271`}, {-1.5707963267948966`, \
-1.6896532483438271`}]], 
                   ArrowBox[{{-1.742439790488633, 
                    0.16863968874202778`}, {-1.7457515276888669`, 
                    0.16901839078244457`}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-1.7062257706326316`, 
                    0.5}, {-1.6269878229333445`, 0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-1.7062257706326316`, 
                    0.5}, {-1.7656689874504448`, 0.5}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.005], Arrowheads[0.04], 
                   ArrowBox[{{-1.7062257706326316`, 
                    0.5}, {-1.6864310397511577`, 0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-1.7062257706326316`, 0.5}, 1.6, {
                    1.562713946553776, 1.5255607225344776`}], 
                   ArrowBox[{{-1.6498604688750595`, 
                    2.0990068645124005`}, {-1.6338734849713497`, 
                    2.09836327121202}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Dashing[{Small, Small}], 
                   Arrowheads[0.02], 
                   CircleBox[{-1.7062257706326316`, 0.5}, 1.7, {
                    1.562713946553776, 1.5741854446753112`}], 
                   ArrowBox[{{-1.6949873528899435`, 
                    2.1999628519372534`}, {-1.7119872599997952`, 
                    2.1999902368073387`}}]}, 
                  {RGBColor[0, 0, 1], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-1.7062257706326316`, 0.5}, 1.8, {
                    1.562713946553776, 1.5847244875813558`}], 
                   ArrowBox[{{-1.7132964418642889`, 
                    2.2999861126154095`}, {-1.731295649463642, 
                    2.2998254085258933`}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.003], Arrowheads[0.025], 
                   CircleBox[{-1.7062257706326316`, 0.5}, 1.5, {
                    1.562713946553776, 1.562497521370693}], 
                   ArrowBox[{{-1.7087775612654854`, 
                    1.999997829453285}, {-1.6937777053808725`, 
                    1.9999483476678415`}}]}, 
                  {RGBColor[0, 0, 1], ArrowBox[{{0, 2.8}, {0, 2.8}}]}},
                 Frame->True,
                 FrameTicks->Automatic,
                 ImageSize->800,
                 PlotRange->{{-3, 3}, {-1, 3}}], 458->
                GraphicsBox[{InsetBox[
                   TagBox[GridBox[{
                    {"\<\"t\"\>", "22.85`"},
                    {"\<\"driveIntent\"\>", "0"},
                    {"\<\"motorTorque\"\>", 
                    RowBox[{"-", "3.9075327318076414`"}]},
                    {"\<\"vw\"\>", 
                    RowBox[{"-", "0.04432619863637721`"}]},
                    {"\<\"controllerState\"\>", 
                    RowBox[{"\[LeftAssociation]", 
                    RowBox[{
                    
                    RowBox[{"\<\"lastError\"\>", "\[Rule]", 
                    "0.005326219293211497`"}], ",", 
                    RowBox[{"\<\"lastInt\"\>", "\[Rule]", 
                    RowBox[{"-", "0.038300060483065224`"}]}], ",", 
                    RowBox[{"\<\"initialized\"\>", "\[Rule]", "True"}]}], 
                    "\[RightAssociation]"}]}
                    },
                    AutoDelete->False,
                    
                    GridBoxItemSize->{
                    "Columns" -> {{Automatic}}, "Rows" -> {{Automatic}}}],
                    "Grid"], {-1, -0.5}], 
                  {GrayLevel[0], LineBox[{{-5, 0}, {5, 0}}]}, 
                  {GrayLevel[0.5], CircleBox[{-1.708449971802289, 0.5}, 0.5], 
                   DiskBox[{-1.708449971802289, 0.5}, 0.1]}, 
                  {GrayLevel[0.5], 
                   DiskBox[{-1.6964929972491465`, 1.999952342829443}, 0.1]}, 
                  {GrayLevel[0.5], Thickness[0.01], 
                   LineBox[{{-1.708449971802289, 0.5}, {-1.6964929972491465`, 
                    1.999952342829443}}]}, 
                  {RGBColor[0, 1, 0], 
                   DiskBox[{-1.7004786554335274`, 1.499968228552962}, 0.05]}, 
                  {RGBColor[0, 0, 1], Arrowheads[0.01], 
                   CircleBox[{-1.708449971802289, 0.5}, 0.3333333333333333, 
                    NCache[{
                    Rational[-1, 2] 
                    Pi, -1.6880223087491257`}, {-1.5707963267948966`, \
-1.6880223087491257`}]], 
                   ArrowBox[{{-1.7441235150821508`, 
                    0.16858106659249095`}, {-1.7474358655177131`, 
                    0.16895436688830834`}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-1.708449971802289, 0.5}, {-1.6302993171661362`,
                     0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-1.708449971802289, 0.5}, {-1.7670770413270604`,
                     0.5}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.005], Arrowheads[0.04], 
                   ArrowBox[{{-1.708449971802289, 0.5}, {-1.6889263866909077`,
                     0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-1.708449971802289, 0.5}, 1.6, {
                    1.5628249260049767`, 1.5261818433482346`}], 
                   ArrowBox[{{-1.6530778572983607`, 
                    2.099041565730974}, {-1.6370904767292995`, 
                    2.0984079024025526`}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Dashing[{Small, Small}], 
                   Arrowheads[0.02], 
                   CircleBox[{-1.708449971802289, 0.5}, 1.7, {
                    1.5628249260049767`, 1.574139299269835}], 
                   ArrowBox[{{-1.6971331085964974`, 
                    2.1999623315259607`}, {-1.7141330144245632`, 
                    2.199990500863623}}]}, 
                  {RGBColor[0, 0, 1], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-1.708449971802289, 0.5}, 1.8, {
                    1.5628249260049767`, 1.5845334411816858`}], 
                   ArrowBox[{{-1.7151767620407317`, 
                    2.2999874305930827`}, {-1.7331760000135483`, 
                    2.2998301651902873`}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.003], Arrowheads[0.025], 
                   CircleBox[{-1.708449971802289, 0.5}, 1.5, {
                    1.5628249260049767`, 1.5626121457500088`}], 
                   ArrowBox[{{-1.7111736987381787`, 
                    1.999997527101821}, {-1.6961738372802864`, 
                    1.9999497646658697`}}]}, 
                  {RGBColor[0, 0, 1], ArrowBox[{{0, 2.8}, {0, 2.8}}]}},
                 Frame->True,
                 FrameTicks->Automatic,
                 ImageSize->800,
                 PlotRange->{{-3, 3}, {-1, 3}}], 459->
                GraphicsBox[{InsetBox[
                   TagBox[GridBox[{
                    {"\<\"t\"\>", "22.900000000000002`"},
                    {"\<\"driveIntent\"\>", "0"},
                    {"\<\"motorTorque\"\>", 
                    RowBox[{"-", "3.8560304633809186`"}]},
                    {"\<\"vw\"\>", 
                    RowBox[{"-", "0.04400767700092206`"}]},
                    {"\<\"controllerState\"\>", 
                    RowBox[{"\[LeftAssociation]", 
                    RowBox[{
                    
                    RowBox[{"\<\"lastError\"\>", "\[Rule]", 
                    "0.005336058404251399`"}], ",", 
                    RowBox[{"\<\"lastInt\"\>", "\[Rule]", 
                    RowBox[{"-", "0.03803348050079279`"}]}], ",", 
                    RowBox[{"\<\"initialized\"\>", "\[Rule]", "True"}]}], 
                    "\[RightAssociation]"}]}
                    },
                    AutoDelete->False,
                    
                    GridBoxItemSize->{
                    "Columns" -> {{Automatic}}, "Rows" -> {{Automatic}}}],
                    "Grid"], {-1, -0.5}], 
                  {GrayLevel[0], LineBox[{{-5, 0}, {5, 0}}]}, 
                  {GrayLevel[0.5], CircleBox[{-1.710658141939525, 0.5}, 0.5], 
                   DiskBox[{-1.710658141939525, 0.5}, 0.1]}, 
                  {GrayLevel[0.5], 
                   DiskBox[{-1.6988591253560714`, 1.999953593684706}, 0.1]}, 
                  {GrayLevel[0.5], Thickness[0.01], 
                   LineBox[{{-1.710658141939525, 0.5}, {-1.6988591253560714`, 
                    1.999953593684706}}]}, 
                  {RGBColor[0, 1, 0], 
                   DiskBox[{-1.7027921308838894`, 1.4999690624564705`}, 
                    0.05]}, 
                  {RGBColor[0, 0, 1], Arrowheads[0.01], 
                   CircleBox[{-1.710658141939525, 0.5}, 0.3333333333333333, 
                    NCache[{
                    Rational[-1, 2] 
                    Pi, -1.6864772406963242`}, {-1.5707963267948966`, \
-1.6864772406963242`}]], 
                   ArrowBox[{{-1.7458195780364598`, 
                    0.16852634414977669`}, {-1.7491325012924746`, 
                    0.16889452619521494`}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-1.710658141939525, 0.5}, {-1.6335375326719066`,
                     0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-1.710658141939525, 0.5}, {-1.7685107650551675`,
                     0.5}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.005], Arrowheads[0.04], 
                   ArrowBox[{{-1.710658141939525, 0.5}, {-1.6913901557875493`,
                     0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-1.710658141939525, 0.5}, 1.6, {
                    1.5629302346199039`, 1.5267712265017397`}], 
                   ArrowBox[{{-1.6562284851590043`, 
                    2.0990739233890205`}, {-1.6402407338962621`, 
                    2.0984496828628885`}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Dashing[{Small, Small}], 
                   Arrowheads[0.02], 
                   CircleBox[{-1.710658141939525, 0.5}, 1.7, {
                    1.5629302346199039`, 1.5740964918402416`}], 
                   ArrowBox[{{-1.699268507726306, 
                    2.1999618455225662`}, {-1.7162684123329393`, 
                    2.1999907425824743`}}]}, 
                  {RGBColor[0, 0, 1], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-1.710658141939525, 0.5}, 1.8, {
                    1.5629302346199039`, 1.584352626083131}], 
                   ArrowBox[{{-1.7170594671651183`, 
                    2.299988617473832}, {-1.7350587332796594`, 
                    2.299834606607577}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.003], Arrowheads[0.025], 
                   CircleBox[{-1.710658141939525, 0.5}, 1.5, {
                    1.5629302346199039`, 1.5627225872327843`}], 
                   ArrowBox[{{-1.7135475308095196`, 
                    1.9999972171414047`}, {-1.698547664168152, 
                    1.9999511113126818`}}]}, 
                  {RGBColor[0, 0, 1], ArrowBox[{{0, 2.8}, {0, 2.8}}]}},
                 Frame->True,
                 FrameTicks->Automatic,
                 ImageSize->800,
                 PlotRange->{{-3, 3}, {-1, 3}}], 460->
                GraphicsBox[{InsetBox[
                   TagBox[GridBox[{
                    {"\<\"t\"\>", "22.95`"},
                    {"\<\"driveIntent\"\>", "0"},
                    {"\<\"motorTorque\"\>", 
                    RowBox[{"-", "3.8073125199359414`"}]},
                    {"\<\"vw\"\>", 
                    RowBox[{"-", "0.04369323154580994`"}]},
                    {"\<\"controllerState\"\>", 
                    RowBox[{"\[LeftAssociation]", 
                    RowBox[{
                    
                    RowBox[{"\<\"lastError\"\>", "\[Rule]", 
                    "0.005341584323521221`"}], ",", 
                    RowBox[{"\<\"lastInt\"\>", "\[Rule]", 
                    RowBox[{"-", "0.037766518881512404`"}]}], ",", 
                    RowBox[{"\<\"initialized\"\>", "\[Rule]", "True"}]}], 
                    "\[RightAssociation]"}]}
                    },
                    AutoDelete->False,
                    GridBoxItemSize->{
                    "Columns" -> {{Automatic}}, "Rows" -> {{Automatic}}}],
                    "Grid"], {-1, -0.5}], 
                  {GrayLevel[0], LineBox[{{-5, 0}, {5, 0}}]}, 
                  {GrayLevel[0.5], 
                   CircleBox[{-1.7128504909633235`, 0.5}, 0.5], 
                   DiskBox[{-1.7128504909633235`, 0.5}, 0.1]}, 
                  {GrayLevel[0.5], 
                   DiskBox[{-1.7012011310626438`, 1.999954763455853}, 0.1]}, 
                  {GrayLevel[0.5], Thickness[0.01], 
                   LineBox[{{-1.7128504909633235`, 
                    0.5}, {-1.7012011310626438`, 1.999954763455853}}]}, 
                  {RGBColor[0, 1, 0], 
                   DiskBox[{-1.7050842510295372`, 1.499969842303902}, 0.05]}, 
                  {RGBColor[0, 0, 1], Arrowheads[0.01], 
                   CircleBox[{-1.7128504909633235`, 0.5}, 0.3333333333333333, 
                    NCache[{
                    Rational[-1, 2] 
                    Pi, -1.6850157023929748`}, {-1.5707963267948966`, \
-1.6850157023929748`}]], 
                   ArrowBox[{{-1.7475274282340945`, 
                    0.16847530841183167`}, {-1.7508408860637221`, 
                    0.16883864810152333`}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-1.7128504909633235`, 
                    0.5}, {-1.6367042405646046`, 0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-1.7128504909633235`, 0.5}, {-1.769969365972699,
                     0.5}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.005], Arrowheads[0.04], 
                   ArrowBox[{{-1.7128504909633235`, 
                    0.5}, {-1.6938231155739802`, 0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-1.7128504909633235`, 0.5}, 1.6, {
                    1.5630300087895337`, 1.5273296352676589`}], 
                   ArrowBox[{{-1.659313779518707, 
                    2.099104068073024}, {-1.6433256821672542`, 
                    2.0984887553442064`}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Dashing[{Small, Small}], 
                   Arrowheads[0.02], 
                   CircleBox[{-1.7128504909633235`, 0.5}, 1.7, {
                    1.5630300087895337`, 1.5740568351281947`}], 
                   ArrowBox[{{-1.7013934418616798`, 
                    2.1999613925103954`}, {-1.718393345308983, 
                    2.1999909637306025`}}]}, 
                  {RGBColor[0, 0, 1], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-1.7128504909633235`, 0.5}, 1.8, {
                    1.5630300087895337`, 1.5841817450114002`}], 
                   ArrowBox[{{-1.7189442321128958`, 
                    2.299989685059001}, {-1.7369435242821865`, 
                    2.299838749928864}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.003], Arrowheads[0.025], 
                   CircleBox[{-1.7128504909633235`, 0.5}, 1.5, {
                    1.5630300087895337`, 1.5628287636829226`}], 
                   ArrowBox[{{-1.7158991441964655`, 
                    1.9999969019012886`}, {-1.7008992727442949`, 
                    1.9999523887054154`}}]}, 
                  {RGBColor[0, 0, 1], ArrowBox[{{0, 2.8}, {0, 2.8}}]}},
                 Frame->True,
                 FrameTicks->Automatic,
                 ImageSize->800,
                 PlotRange->{{-3, 3}, {-1, 3}}], 461->
                GraphicsBox[{InsetBox[
                   TagBox[GridBox[{
                    {"\<\"t\"\>", "23.`"},
                    {"\<\"driveIntent\"\>", "0"},
                    {"\<\"motorTorque\"\>", 
                    RowBox[{"-", "3.7612848360310105`"}]},
                    {"\<\"vw\"\>", 
                    RowBox[{"-", "0.04338262109303191`"}]},
                    {"\<\"controllerState\"\>", 
                    RowBox[{"\[LeftAssociation]", 
                    RowBox[{
                    
                    RowBox[{"\<\"lastError\"\>", "\[Rule]", 
                    "0.005342894283538541`"}], ",", 
                    RowBox[{"\<\"lastInt\"\>", "\[Rule]", 
                    RowBox[{"-", "0.03749938894814783`"}]}], ",", 
                    RowBox[{"\<\"initialized\"\>", "\[Rule]", "True"}]}], 
                    "\[RightAssociation]"}]}
                    },
                    AutoDelete->False,
                    
                    GridBoxItemSize->{
                    "Columns" -> {{Automatic}}, "Rows" -> {{Automatic}}}],
                    "Grid"], {-1, -0.5}], 
                  {GrayLevel[0], LineBox[{{-5, 0}, {5, 0}}]}, 
                  {GrayLevel[0.5], 
                   CircleBox[{-1.7150272164896203`, 0.5}, 0.5], 
                   DiskBox[{-1.7150272164896203`, 0.5}, 0.1]}, 
                  {GrayLevel[0.5], 
                   DiskBox[{-1.703519467362148, 1.999955856587126}, 0.1]}, 
                  {GrayLevel[0.5], Thickness[0.01], 
                   LineBox[{{-1.7150272164896203`, 0.5}, {-1.703519467362148, 
                    1.999955856587126}}]}, 
                  {RGBColor[0, 1, 0], 
                   DiskBox[{-1.7073553837379722`, 1.4999705710580842`}, 
                    0.05]}, 
                  {RGBColor[0, 0, 1], Arrowheads[0.01], 
                   CircleBox[{-1.7150272164896203`, 0.5}, 0.3333333333333333, 
                    NCache[{
                    Rational[-1, 2] 
                    Pi, -1.683634871875827}, {-1.5707963267948966`, \
-1.683634871875827}]], 
                   ArrowBox[{{-1.7492463414354023`, 
                    0.16842774151166373`}, {-1.7525602978165293`, 
                    0.16878650553273233`}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-1.7150272164896203`, 
                    0.5}, {-1.6398015197690001`, 0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-1.7150272164896203`, 0.5}, {-1.77145178991572, 
                    0.5}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.005], Arrowheads[0.04], 
                   ArrowBox[{{-1.7150272164896203`, 
                    0.5}, {-1.6962260931950999`, 0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-1.7150272164896203`, 0.5}, 1.6, {
                    1.5631244187843891`, 1.52785802254035}], 
                   ArrowBox[{{-1.6623354587165053`, 
                    2.0991321329592436`}, {-1.6463470384735364`, 
                    2.0985252682230833`}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Dashing[{Small, Small}], 
                   Arrowheads[0.02], 
                   CircleBox[{-1.7150272164896203`, 0.5}, 1.7, {
                    1.5631244187843891`, 1.5740201346535023`}], 
                   ArrowBox[{{-1.7035077780056425`, 
                    2.1999609708863948`}, {-1.7205076803562191`, 
                    2.199991166011108}}]}, 
                  {RGBColor[0, 0, 1], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-1.7150272164896203`, 0.5}, 1.8, {
                    1.5631244187843891`, 1.584020445651228}], 
                   ArrowBox[{{-1.7208306203766628`, 
                    2.2999906445599443`}, {-1.7388299366575453`, 
                    2.299842612706069}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.003], Arrowheads[0.025], 
                   CircleBox[{-1.7150272164896203`, 0.5}, 1.5, {
                    1.5631244187843891`, 1.562930639682616}], 
                   ArrowBox[{{-1.7182286833905955`, 
                    1.99999658353267}, {-1.7032288074814383`, 
                    1.9999535984639243`}}]}, 
                  {RGBColor[0, 0, 1], ArrowBox[{{0, 2.8}, {0, 2.8}}]}},
                 Frame->True,
                 FrameTicks->Automatic,
                 ImageSize->800,
                 PlotRange->{{-3, 3}, {-1, 3}}], 462->
                GraphicsBox[{InsetBox[
                   TagBox[GridBox[{
                    {"\<\"t\"\>", "23.05`"},
                    {"\<\"driveIntent\"\>", "0"},
                    {"\<\"motorTorque\"\>", 
                    RowBox[{"-", "3.7178400613100697`"}]},
                    {"\<\"vw\"\>", 
                    RowBox[{"-", "0.04307561528098249`"}]},
                    {"\<\"controllerState\"\>", 
                    RowBox[{"\[LeftAssociation]", 
                    RowBox[{
                    
                    RowBox[{"\<\"lastError\"\>", "\[Rule]", 
                    "0.005340117257984643`"}], ",", 
                    RowBox[{"\<\"lastInt\"\>", "\[Rule]", 
                    RowBox[{"-", "0.03723229833163259`"}]}], ",", 
                    RowBox[{"\<\"initialized\"\>", "\[Rule]", "True"}]}], 
                    "\[RightAssociation]"}]}
                    },
                    AutoDelete->False,
                    
                    GridBoxItemSize->{
                    "Columns" -> {{Automatic}}, "Rows" -> {{Automatic}}}],
                    "Grid"], {-1, -0.5}], 
                  {GrayLevel[0], LineBox[{{-5, 0}, {5, 0}}]}, 
                  {GrayLevel[0.5], 
                   CircleBox[{-1.7171885043473765`, 0.5}, 0.5], 
                   DiskBox[{-1.7171885043473765`, 0.5}, 0.1]}, 
                  {GrayLevel[0.5], 
                   DiskBox[{-1.7058146180553548`, 1.9999568776170253`}, 0.1]}, 
                  {GrayLevel[0.5], Thickness[0.01], 
                   LineBox[{{-1.7171885043473765`, 
                    0.5}, {-1.7058146180553548`, 1.9999568776170253`}}]}, 
                  {RGBColor[0, 1, 0], 
                   DiskBox[{-1.7096059134860286`, 1.4999712517446837`}, 
                    0.05]}, 
                  {RGBColor[0, 0, 1], Arrowheads[0.01], 
                   CircleBox[{-1.7171885043473765`, 0.5}, 0.3333333333333333, 
                    NCache[{
                    Rational[-1, 2] 
                    Pi, -1.6823315286341987`}, {-1.5707963267948966`, \
-1.6823315286341987`}]], 
                   ArrowBox[{{-1.7509754478892245`, 
                    0.1683834238805137}, {-1.7542898690481672`, 
                    0.16873786837543592`}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-1.7171885043473765`, 0.5}, {-1.642831703121175,
                     0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-1.7171885043473765`, 
                    0.5}, {-1.7729567620640005`, 0.5}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.005], Arrowheads[0.04], 
                   ArrowBox[{{-1.7171885043473765`, 
                    0.5}, {-1.6985999608377993`, 0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-1.7171885043473765`, 0.5}, 1.6, {
                    1.5632136632706273`, 1.5283575001438603`}], 
                   ArrowBox[{{-1.665295483799231, 
                    2.0991582518370064`}, {-1.649306762435309, 
                    2.09855937303404}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Dashing[{Small, Small}], 
                   Arrowheads[0.02], 
                   CircleBox[{-1.7171885043473765`, 0.5}, 1.7, {
                    1.5632136632706273`, 1.5739861913931112`}], 
                   ArrowBox[{{-1.7056113636521564`, 
                    2.199960578899794}, {-1.7226112649680188`, 
                    2.1999913510566023`}}]}, 
                  {RGBColor[0, 0, 1], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-1.7171885043473765`, 0.5}, 1.8, {
                    1.5632136632706273`, 1.5838683302779053`}], 
                   ArrowBox[{{-1.7227181019194577`, 
                    2.29999150652182}, {-1.740717440510017, 
                    2.299846212642362}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.003], Arrowheads[0.025], 
                   CircleBox[{-1.7171885043473765`, 0.5}, 1.5, {
                    1.5632136632706273`, 1.563028221549625}], 
                   ArrowBox[{{-1.7205363437000052`, 
                    1.9999962639859037`}, {-1.7055364636677006`, 
                    1.9999547426332567`}}]}, 
                  {RGBColor[0, 0, 1], ArrowBox[{{0, 2.8}, {0, 2.8}}]}},
                 Frame->True,
                 FrameTicks->Automatic,
                 ImageSize->800,
                 PlotRange->{{-3, 3}, {-1, 3}}], 463->
                GraphicsBox[{InsetBox[
                   TagBox[GridBox[{
                    {"\<\"t\"\>", "23.1`"},
                    {"\<\"driveIntent\"\>", "0"},
                    {"\<\"motorTorque\"\>", 
                    RowBox[{"-", "3.6768600973242247`"}]},
                    {"\<\"vw\"\>", 
                    RowBox[{"-", "0.04277199539123706`"}]},
                    {"\<\"controllerState\"\>", 
                    RowBox[{"\[LeftAssociation]", 
                    RowBox[{
                    
                    RowBox[{"\<\"lastError\"\>", "\[Rule]", 
                    "0.005333409075227206`"}], ",", 
                    RowBox[{"\<\"lastInt\"\>", "\[Rule]", 
                    RowBox[{"-", "0.03696544750968962`"}]}], ",", 
                    RowBox[{"\<\"initialized\"\>", "\[Rule]", "True"}]}], 
                    "\[RightAssociation]"}]}
                    },
                    AutoDelete->False,
                    
                    GridBoxItemSize->{
                    "Columns" -> {{Automatic}}, "Rows" -> {{Automatic}}}],
                    "Grid"], {-1, -0.5}], 
                  {GrayLevel[0], LineBox[{{-5, 0}, {5, 0}}]}, 
                  {GrayLevel[0.5], 
                   CircleBox[{-1.7193345291413689`, 0.5}, 0.5], 
                   DiskBox[{-1.7193345291413689`, 0.5}, 0.1]}, 
                  {GrayLevel[0.5], 
                   DiskBox[{-1.7080870906205914`, 1.999957831116169}, 0.1]}, 
                  {GrayLevel[0.5], Thickness[0.01], 
                   LineBox[{{-1.7193345291413689`, 
                    0.5}, {-1.7080870906205914`, 1.999957831116169}}]}, 
                  {RGBColor[0, 1, 0], 
                   DiskBox[{-1.7118362367941837`, 1.4999718874107792`}, 
                    0.05]}, 
                  {RGBColor[0, 0, 1], Arrowheads[0.01], 
                   CircleBox[{-1.7193345291413689`, 0.5}, 0.3333333333333333, 
                    NCache[{
                    Rational[-1, 2] 
                    Pi, -1.6811021297146234`}, {-1.5707963267948966`, \
-1.6811021297146234`}]], 
                   ArrowBox[{{-1.7527137581923642`, 
                    0.1683421368653047}, {-1.7560286126001323`, 
                    0.16869250634760402`}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-1.7193345291413689`, 
                    0.5}, {-1.6457973271948845`, 0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-1.7193345291413689`, 
                    0.5}, {-1.7744828252212808`, 0.5}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.005], Arrowheads[0.04], 
                   ArrowBox[{{-1.7193345291413689`, 
                    0.5}, {-1.7009456232747961`, 0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-1.7193345291413689`, 0.5}, 1.6, {
                    1.5632979641814504`, 1.5288293101283932`}], 
                   ArrowBox[{{-1.66819601317111, 
                    2.099182557491283}, {-1.6522070110297815`, 
                    2.0985912223930727`}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Dashing[{Small, Small}], 
                   Arrowheads[0.02], 
                   CircleBox[{-1.7193345291413689`, 0.5}, 1.7, {
                    1.5632979641814504`, 1.5739548041971698`}], 
                   ArrowBox[{{-1.7077040314560776`, 
                    2.1999602146884474`}, {-1.7247039317977144`, 
                    2.199991520424474}}]}, 
                  {RGBColor[0, 0, 1], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-1.7193345291413689`, 0.5}, 1.8, {
                    1.5632979641814504`, 1.5837249647221405`}], 
                   ArrowBox[{{-1.7246060698748031`, 
                    2.2999922807774196`}, {-1.7426054291105235`, 
                    2.2998495673846264`}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.003], Arrowheads[0.025], 
                   CircleBox[{-1.7193345291413689`, 0.5}, 1.5, {
                    1.5632979641814504`, 1.563121552524833}], 
                   ArrowBox[{{-1.7228223645933396`, 
                    1.9999959449958056`}, {-1.7078224807511362`, 
                    1.999955823596769}}]}, 
                  {RGBColor[0, 0, 1], ArrowBox[{{0, 2.8}, {0, 2.8}}]}},
                 Frame->True,
                 FrameTicks->Automatic,
                 ImageSize->800,
                 PlotRange->{{-3, 3}, {-1, 3}}], 464->
                GraphicsBox[{InsetBox[
                   TagBox[GridBox[{
                    {"\<\"t\"\>", "23.150000000000002`"},
                    {"\<\"driveIntent\"\>", "0"},
                    {"\<\"motorTorque\"\>", 
                    RowBox[{"-", "3.6382184471390406`"}]},
                    {"\<\"vw\"\>", 
                    RowBox[{"-", "0.04247155497015165`"}]},
                    {"\<\"controllerState\"\>", 
                    RowBox[{"\[LeftAssociation]", 
                    RowBox[{
                    
                    RowBox[{"\<\"lastError\"\>", "\[Rule]", 
                    "0.005322947833851416`"}], ",", 
                    RowBox[{"\<\"lastInt\"\>", "\[Rule]", 
                    RowBox[{"-", "0.03669902858228153`"}]}], ",", 
                    RowBox[{"\<\"initialized\"\>", "\[Rule]", "True"}]}], 
                    "\[RightAssociation]"}]}
                    },
                    AutoDelete->False,
                    
                    GridBoxItemSize->{
                    "Columns" -> {{Automatic}}, "Rows" -> {{Automatic}}}],
                    "Grid"], {-1, -0.5}], 
                  {GrayLevel[0], LineBox[{{-5, 0}, {5, 0}}]}, 
                  {GrayLevel[0.5], 
                   CircleBox[{-1.7214654548510235`, 0.5}, 0.5], 
                   DiskBox[{-1.7214654548510235`, 0.5}, 0.1]}, 
                  {GrayLevel[0.5], 
                   DiskBox[{-1.7103374096654864`, 1.999958721635482}, 0.1]}, 
                  {GrayLevel[0.5], Thickness[0.01], 
                   LineBox[{{-1.7214654548510235`, 
                    0.5}, {-1.7103374096654864`, 1.999958721635482}}]}, 
                  {RGBColor[0, 1, 0], 
                   DiskBox[{-1.7140467580606655`, 1.4999724810903212`}, 
                    0.05]}, 
                  {RGBColor[0, 0, 1], Arrowheads[0.01], 
                   CircleBox[{-1.7214654548510235`, 0.5}, 0.3333333333333333, 
                    NCache[{
                    Rational[-1, 2] 
                    Pi, -1.6799428802090677`}, {-1.5707963267948966`, \
-1.6799428802090677`}]], 
                   ArrowBox[{{-1.7544601873457863`, 
                    0.1683036648699444}, {-1.7577754456917638`, 
                    0.1686501913743476}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-1.7214654548510235`, 
                    0.5}, {-1.6487010859082427`, 0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-1.7214654548510235`, 0.5}, {-1.776028375456992,
                     0.5}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.005], Arrowheads[0.04], 
                   ArrowBox[{{-1.7214654548510235`, 
                    0.5}, {-1.7032640065142113`, 0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-1.7214654548510235`, 0.5}, 1.6, {
                    1.5633775619523071`, 1.5292747981026986`}], 
                   ArrowBox[{{-1.6710393605297447`, 
                    2.099205180397907}, {-1.655050096542335, 
                    2.0986209682663146`}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Dashing[{Small, Small}], 
                   Arrowheads[0.02], 
                   CircleBox[{-1.7214654548510235`, 0.5}, 1.7, {
                    1.5633775619523071`, 1.5739257719491}], 
                   ArrowBox[{{-1.709785603503979, 
                    2.19995987631253}, {-1.7267855029295756`, 
                    2.1999916755938664`}}]}, 
                  {RGBColor[0, 0, 1], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-1.7214654548510235`, 0.5}, 1.8, {
                    1.5633775619523071`, 1.583589886658635}], 
                   ArrowBox[{{-1.7264938560654925`, 
                    2.2999929764255267`}, {-1.7444932344144528`, 
                    2.2998526943526176`}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.003], Arrowheads[0.025], 
                   CircleBox[{-1.7214654548510235`, 0.5}, 1.5, {
                    1.5633775619523071`, 1.5632107081566835`}], 
                   ArrowBox[{{-1.7250870233752045`, 
                    1.9999956280741036`}, {-1.7100871360155674`, 
                    1.999956843999346}}]}, 
                  {RGBColor[0, 0, 1], ArrowBox[{{0, 2.8}, {0, 2.8}}]}},
                 Frame->True,
                 FrameTicks->Automatic,
                 ImageSize->800,
                 PlotRange->{{-3, 3}, {-1, 3}}], 465->
                GraphicsBox[{InsetBox[
                   TagBox[GridBox[{
                    {"\<\"t\"\>", "23.2`"},
                    {"\<\"driveIntent\"\>", "0"},
                    {"\<\"motorTorque\"\>", 
                    RowBox[{"-", "3.6017823766381603`"}]},
                    {"\<\"vw\"\>", 
                    RowBox[{"-", "0.04217410026280459`"}]},
                    {"\<\"controllerState\"\>", 
                    RowBox[{"\[LeftAssociation]", 
                    RowBox[{
                    
                    RowBox[{"\<\"lastError\"\>", "\[Rule]", 
                    "0.005308929629559622`"}], ",", 
                    RowBox[{"\<\"lastInt\"\>", "\[Rule]", 
                    RowBox[{"-", "0.03643322426837456`"}]}], ",", 
                    RowBox[{"\<\"initialized\"\>", "\[Rule]", "True"}]}], 
                    "\[RightAssociation]"}]}
                    },
                    AutoDelete->False,
                    
                    GridBoxItemSize->{
                    "Columns" -> {{Automatic}}, "Rows" -> {{Automatic}}}],
                    "Grid"], {-1, -0.5}], 
                  {GrayLevel[0], LineBox[{{-5, 0}, {5, 0}}]}, 
                  {GrayLevel[0.5], 
                   CircleBox[{-1.7235814354554597`, 0.5}, 0.5], 
                   DiskBox[{-1.7235814354554597`, 0.5}, 0.1]}, 
                  {GrayLevel[0.5], 
                   DiskBox[{-1.7125661109565968`, 1.9999595536634263`}, 0.1]}, 
                  {GrayLevel[0.5], Thickness[0.01], 
                   LineBox[{{-1.7235814354554597`, 
                    0.5}, {-1.7125661109565968`, 1.9999595536634263`}}]}, 
                  {RGBColor[0, 1, 0], 
                   DiskBox[{-1.716237885789551, 1.4999730357756174`}, 0.05]}, 
                  {RGBColor[0, 0, 1], Arrowheads[0.01], 
                   CircleBox[{-1.7235814354554597`, 0.5}, 0.3333333333333333, 
                    NCache[{
                    Rational[-1, 2] 
                    Pi, -1.6788497980940413`}, {-1.5707963267948966`, \
-1.6788497980940413`}]], 
                   ArrowBox[{{-1.7562135769793523`, 
                    0.16826779708524564`}, {-1.7595292121265962`, 
                    0.16861069953374597`}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-1.7235814354554597`, 
                    0.5}, {-1.6515457879226965`, 0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-1.7235814354554597`, 0.5}, {-1.777591695069817,
                     0.5}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.005], Arrowheads[0.04], 
                   ArrowBox[{{-1.7235814354554597`, 
                    0.5}, {-1.7055560475370537`, 0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-1.7235814354554597`, 0.5}, 1.6, {
                    1.5634527111239025`, 1.5296953886249107`}], 
                   ArrowBox[{{-1.6738279561163951`, 
                    2.0992262476878176`}, {-1.6578384478291286`, 
                    2.09864876054059}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Dashing[{Small, Small}], 
                   Arrowheads[0.02], 
                   CircleBox[{-1.7235814354554597`, 0.5}, 1.7, {
                    1.5634527111239025`, 1.5738988954782978`}], 
                   ArrowBox[{{-1.7118558951906584`, 
                    2.1999595617853673`}, {-1.7288557937554627`, 
                    2.1999918179639932`}}]}, 
                  {RGBColor[0, 0, 1], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-1.7235814354554597`, 0.5}, 1.8, {
                    1.5634527111239025`, 1.5834626132163367`}], 
                   ArrowBox[{{-1.7283807453275983`, 
                    2.299993601828837}, {-1.7463801413849591`, 
                    2.2998556105999004`}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.003], Arrowheads[0.025], 
                   CircleBox[{-1.7235814354554597`, 0.5}, 1.5, {
                    1.5634527111239025`, 1.5632957919042136`}], 
                   ArrowBox[{{-1.7273306292156931`, 
                    1.9999953145080647`}, {-1.7123307386104558`, 
                    1.9999578066800752`}}]}, 
                  {RGBColor[0, 0, 1], ArrowBox[{{0, 2.8}, {0, 2.8}}]}},
                 Frame->True,
                 FrameTicks->Automatic,
                 ImageSize->800,
                 PlotRange->{{-3, 3}, {-1, 3}}], 466->
                GraphicsBox[{InsetBox[
                   TagBox[GridBox[{
                    {"\<\"t\"\>", "23.25`"},
                    {"\<\"driveIntent\"\>", "0"},
                    {"\<\"motorTorque\"\>", 
                    RowBox[{"-", "3.5674148883555237`"}]},
                    {"\<\"vw\"\>", 
                    RowBox[{"-", "0.04187945047659239`"}]},
                    {"\<\"controllerState\"\>", 
                    RowBox[{"\[LeftAssociation]", 
                    RowBox[{
                    
                    RowBox[{"\<\"lastError\"\>", "\[Rule]", 
                    "0.0052915645987479405`"}], ",", 
                    RowBox[{"\<\"lastInt\"\>", "\[Rule]", 
                    RowBox[{"-", "0.036168207108296616`"}]}], ",", 
                    RowBox[{"\<\"initialized\"\>", "\[Rule]", "True"}]}], 
                    "\[RightAssociation]"}]}
                    },
                    AutoDelete->False,
                    
                    GridBoxItemSize->{
                    "Columns" -> {{Automatic}}, "Rows" -> {{Automatic}}}],
                    "Grid"], {-1, -0.5}], 
                  {GrayLevel[0], LineBox[{{-5, 0}, {5, 0}}]}, 
                  {GrayLevel[0.5], 
                   CircleBox[{-1.7256826155758251`, 0.5}, 0.5], 
                   DiskBox[{-1.7256826155758251`, 0.5}, 0.1]}, 
                  {GrayLevel[0.5], 
                   DiskBox[{-1.7147737360180628`, 1.9999603315910706`}, 0.1]}, 
                  {GrayLevel[0.5], Thickness[0.01], 
                   LineBox[{{-1.7256826155758251`, 
                    0.5}, {-1.7147737360180628`, 1.9999603315910706`}}]}, 
                  {RGBColor[0, 1, 0], 
                   DiskBox[{-1.7184100292039837`, 1.499973554394047}, 0.05]}, 
                  {RGBColor[0, 0, 1], Arrowheads[0.01], 
                   CircleBox[{-1.7256826155758251`, 0.5}, 0.3333333333333333, 
                    NCache[{
                    Rational[-1, 2] 
                    Pi, -1.6778187734455623`}, {-1.5707963267948966`, \
-1.6778187734455623`}]], 
                   ArrowBox[{{-1.7579727157382397`, 
                    0.16823432886657436`}, {-1.7612887026640178`, 
                    0.1685738126318635}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-1.7256826155758251`, 
                    0.5}, {-1.6543343178087146`, 0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-1.7256826155758251`, 
                    0.5}, {-1.7791709828636966`, 0.5}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.005], Arrowheads[0.04], 
                   ArrowBox[{{-1.7256826155758251`, 0.5}, {-1.707822685096586,
                     0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-1.7256826155758251`, 0.5}, 1.6, {
                    1.563523676313059, 1.530092562652081}], 
                   ArrowBox[{{-1.6765643112734478`, 
                    2.099245882340317}, {-1.6605745748844416`, 
                    2.0986747458558694`}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Dashing[{Small, Small}], 
                   Arrowheads[0.02], 
                   CircleBox[{-1.7256826155758251`, 0.5}, 1.7, {
                    1.563523676313059, 1.5738739792353929`}], 
                   ArrowBox[{{-1.713914718709298, 
                    2.1999592691012744`}, {-1.7309146164651226`, 
                    2.1999919488534925`}}]}, 
                  {RGBColor[0, 0, 1], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-1.7256826155758251`, 0.5}, 1.8, {
                    1.563523676313059, 1.583342647915034}], 
                   ArrowBox[{{-1.7302659886391603`, 
                    2.299994164627031}, {-1.7482654011211598`, 
                    2.299858332702053}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.003], Arrowheads[0.025], 
                   CircleBox[{-1.7256826155758251`, 0.5}, 1.5, {
                    1.563523676313059, 1.5633769309756975`}], 
                   ArrowBox[{{-1.729553517550633, 
                    1.9999950053643183`}, {-1.7145536239514214`, 
                    1.9999587146136468`}}]}, 
                  {RGBColor[0, 0, 1], ArrowBox[{{0, 2.8}, {0, 2.8}}]}},
                 Frame->True,
                 FrameTicks->Automatic,
                 ImageSize->800,
                 PlotRange->{{-3, 3}, {-1, 3}}], 467->
                GraphicsBox[{InsetBox[
                   TagBox[GridBox[{
                    {"\<\"t\"\>", "23.3`"},
                    {"\<\"driveIntent\"\>", "0"},
                    {"\<\"motorTorque\"\>", 
                    RowBox[{"-", "3.5349765103763438`"}]},
                    {"\<\"vw\"\>", 
                    RowBox[{"-", "0.0415874378914412`"}]},
                    {"\<\"controllerState\"\>", 
                    RowBox[{"\[LeftAssociation]", 
                    RowBox[{
                    
                    RowBox[{"\<\"lastError\"\>", "\[Rule]", 
                    "0.005271073280428029`"}], ",", 
                    RowBox[{"\<\"lastInt\"\>", "\[Rule]", 
                    RowBox[{"-", "0.035904138855798524`"}]}], ",", 
                    RowBox[{"\<\"initialized\"\>", "\[Rule]", "True"}]}], 
                    "\[RightAssociation]"}]}
                    },
                    AutoDelete->False,
                    
                    GridBoxItemSize->{
                    "Columns" -> {{Automatic}}, "Rows" -> {{Automatic}}}],
                    "Grid"], {-1, -0.5}], 
                  {GrayLevel[0], LineBox[{{-5, 0}, {5, 0}}]}, 
                  {GrayLevel[0.5], 
                   CircleBox[{-1.7277691311268226`, 0.5}, 0.5], 
                   DiskBox[{-1.7277691311268226`, 0.5}, 0.1]}, 
                  {GrayLevel[0.5], 
                   DiskBox[{-1.716960827285968, 1.999961059683912}, 0.1]}, 
                  {GrayLevel[0.5], Thickness[0.01], 
                   LineBox[{{-1.7277691311268226`, 0.5}, {-1.716960827285968, 
                    1.999961059683912}}]}, 
                  {RGBColor[0, 1, 0], 
                   DiskBox[{-1.7205635952329195`, 1.4999740397892745`}, 
                    0.05]}, 
                  {RGBColor[0, 0, 1], Arrowheads[0.01], 
                   CircleBox[{-1.7277691311268226`, 0.5}, 0.3333333333333333, 
                    NCache[{
                    Rational[-1, 2] 
                    Pi, -1.676845622106187}, {-1.5707963267948966`, \
-1.676845622106187}]], 
                   ArrowBox[{{-1.7597363578432328`, 
                    0.1682030628122485}, {-1.763052673567881, 
                    0.1685393194601792}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-1.7277691311268226`, 
                    0.5}, {-1.6570696009192958`, 0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-1.7277691311268226`, 
                    0.5}, {-1.7807643817536918`, 0.5}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.005], Arrowheads[0.04], 
                   ArrowBox[{{-1.7277691311268226`, 0.5}, {-1.710064851546165,
                     0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-1.7277691311268226`, 0.5}, 1.6, {
                    1.5635907285479358`, 1.5304678370288989`}], 
                   ArrowBox[{{-1.6792509862709188`, 
                    2.0992642025693384`}, {-1.6632610366749536`, 
                    2.098699066663326}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Dashing[{Small, Small}], 
                   Arrowheads[0.02], 
                   CircleBox[{-1.7277691311268226`, 0.5}, 1.7, {
                    1.5635907285479358`, 1.5738508327412168`}], 
                   ArrowBox[{{-1.7159618861661297`, 
                    2.1999589962603325`}, {-1.7329617831609787`, 
                    2.1999920695005764`}}]}, 
                  {RGBColor[0, 0, 1], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-1.7277691311268226`, 0.5}, 1.8, {
                    1.5635907285479358`, 1.5832294869389156`}], 
                   ArrowBox[{{-1.7321488150645699`, 
                    2.299994671761171}, {-1.7501482428021975`, 
                    2.299860876668144}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.003], Arrowheads[0.025], 
                   CircleBox[{-1.7277691311268226`, 0.5}, 1.5, {
                    1.5635907285479358`, 1.563454272415857}], 
                   ArrowBox[{{-1.7317560448638847`, 
                    1.9999947014969264`}, {-1.7167561485027565`, 
                    1.9999595708597355`}}]}, 
                  {RGBColor[0, 0, 1], ArrowBox[{{0, 2.8}, {0, 2.8}}]}},
                 Frame->True,
                 FrameTicks->Automatic,
                 ImageSize->800,
                 PlotRange->{{-3, 3}, {-1, 3}}], 468->
                GraphicsBox[{InsetBox[
                   TagBox[GridBox[{
                    {"\<\"t\"\>", "23.35`"},
                    {"\<\"driveIntent\"\>", "0"},
                    {"\<\"motorTorque\"\>", 
                    RowBox[{"-", "3.5043269043246212`"}]},
                    {"\<\"vw\"\>", 
                    RowBox[{"-", "0.04129790783312476`"}]},
                    {"\<\"controllerState\"\>", 
                    RowBox[{"\[LeftAssociation]", 
                    RowBox[{
                    
                    RowBox[{"\<\"lastError\"\>", "\[Rule]", 
                    "0.005247683294923344`"}], ",", 
                    RowBox[{"\<\"lastInt\"\>", "\[Rule]", 
                    RowBox[{"-", "0.03564117004392875`"}]}], ",", 
                    RowBox[{"\<\"initialized\"\>", "\[Rule]", "True"}]}], 
                    "\[RightAssociation]"}]}
                    },
                    AutoDelete->False,
                    
                    GridBoxItemSize->{
                    "Columns" -> {{Automatic}}, "Rows" -> {{Automatic}}}],
                    "Grid"], {-1, -0.5}], 
                  {GrayLevel[0], LineBox[{{-5, 0}, {5, 0}}]}, 
                  {GrayLevel[0.5], 
                   CircleBox[{-1.7298411099702005`, 0.5}, 0.5], 
                   DiskBox[{-1.7298411099702005`, 0.5}, 0.1]}, 
                  {GrayLevel[0.5], 
                   DiskBox[{-1.7191279238013142`, 1.9999617420594802`}, 0.1]}, 
                  {GrayLevel[0.5], Thickness[0.01], 
                   LineBox[{{-1.7298411099702005`, 
                    0.5}, {-1.7191279238013142`, 1.9999617420594802`}}]}, 
                  {RGBColor[0, 1, 0], 
                   DiskBox[{-1.7226989858576096`, 1.4999744947063203`}, 
                    0.05]}, 
                  {RGBColor[0, 0, 1], Arrowheads[0.01], 
                   CircleBox[{-1.7298411099702005`, 0.5}, 0.3333333333333333, 
                    NCache[{
                    Rational[-1, 2] 
                    Pi, -1.6759261339246352`}, {-1.5707963267948966`, \
-1.6759261339246352`}]], 
                   ArrowBox[{{-1.7615032398536996`, 
                    0.1681738095894908}, {-1.7648198633604144`, 
                    0.16850701678259056`}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-1.7298411099702005`, 
                    0.5}, {-1.6597545718837081`, 0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-1.7298411099702005`, 
                    0.5}, {-1.7823700037427206`, 0.5}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.005], Arrowheads[0.04], 
                   ArrowBox[{{-1.7298411099702005`, 
                    0.5}, {-1.7122834656562282`, 0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-1.7298411099702005`, 0.5}, 1.6, {
                    1.5636541419610286`, 1.5308227459788069`}], 
                   ArrowBox[{{-1.6818905613368218`, 
                    2.0992813213708708`}, {-1.6659004121761196`, 
                    2.0987218604765516`}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Dashing[{Small, Small}], 
                   Arrowheads[0.02], 
                   CircleBox[{-1.7298411099702005`, 0.5}, 1.7, {
                    1.5636541419610286`, 1.5738292718214846`}], 
                   ArrowBox[{{-1.7179972123327882`, 
                    2.199958741290139}, {-1.7349971086105964`, 
                    2.1999921810637897`}}]}, 
                  {RGBColor[0, 0, 1], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-1.7298411099702005`, 0.5}, 1.8, {
                    1.5636541419610286`, 1.583122624762832}], 
                   ArrowBox[{{-1.7340284425357435`, 
                    2.2999951295061845`}, {-1.7520278844682335`, 
                    2.2998632578719342`}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.003], Arrowheads[0.025], 
                   CircleBox[{-1.7298411099702005`, 0.5}, 1.5, {
                    1.5636541419610286`, 1.5635279794506731`}], 
                   ArrowBox[{{-1.7339385838580195`, 
                    1.9999944035588062`}, {-1.7189386849482615`, 
                    1.9999603785195932`}}]}, 
                  {RGBColor[0, 0, 1], ArrowBox[{{0, 2.8}, {0, 2.8}}]}},
                 Frame->True,
                 FrameTicks->Automatic,
                 ImageSize->800,
                 PlotRange->{{-3, 3}, {-1, 3}}], 469->
                GraphicsBox[{InsetBox[
                   TagBox[GridBox[{
                    {"\<\"t\"\>", "23.400000000000002`"},
                    {"\<\"driveIntent\"\>", "0"},
                    {"\<\"motorTorque\"\>", 
                    RowBox[{"-", "3.475326297725897`"}]},
                    {"\<\"vw\"\>", 
                    RowBox[{"-", "0.04101071852560268`"}]},
                    {"\<\"controllerState\"\>", 
                    RowBox[{"\[LeftAssociation]", 
                    RowBox[{
                    
                    RowBox[{"\<\"lastError\"\>", "\[Rule]", 
                    "0.005221626334914581`"}], ",", 
                    RowBox[{"\<\"lastInt\"\>", "\[Rule]", 
                    RowBox[{"-", "0.03537943970898439`"}]}], ",", 
                    RowBox[{"\<\"initialized\"\>", "\[Rule]", "True"}]}], 
                    "\[RightAssociation]"}]}
                    },
                    AutoDelete->False,
                    
                    GridBoxItemSize->{
                    "Columns" -> {{Automatic}}, "Rows" -> {{Automatic}}}],
                    "Grid"], {-1, -0.5}], 
                  {GrayLevel[0], LineBox[{{-5, 0}, {5, 0}}]}, 
                  {GrayLevel[0.5], CircleBox[{-1.731898672563741, 0.5}, 0.5], 
                   DiskBox[{-1.731898672563741, 0.5}, 0.1]}, 
                  {GrayLevel[0.5], 
                   DiskBox[{-1.7212755574212728`, 1.9999623826698687`}, 0.1]}, 
                  {GrayLevel[0.5], Thickness[0.01], 
                   LineBox[{{-1.731898672563741, 0.5}, {-1.7212755574212728`, 
                    1.9999623826698687`}}]}, 
                  {RGBColor[0, 1, 0], 
                   DiskBox[{-1.7248165958020956`, 1.4999749217799125`}, 
                    0.05]}, 
                  {RGBColor[0, 0, 1], Arrowheads[0.01], 
                   CircleBox[{-1.731898672563741, 0.5}, 0.3333333333333333, 
                    NCache[{
                    Rational[-1, 2] 
                    Pi, -1.6750561157266735`}, {-1.5707963267948966`, \
-1.6750561157266735`}]], 
                   ArrowBox[{{-1.7632720956764096`, 
                    0.1681463885486485}, {-1.766589007824181, 
                    0.16847671009319815`}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-1.731898672563741, 0.5}, {-1.662392146609223, 
                    0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-1.731898672563741, 0.5}, {-1.7839859523306587`,
                     0.5}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.005], Arrowheads[0.04], 
                   ArrowBox[{{-1.731898672563741, 0.5}, {-1.714479426376141, 
                    0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-1.731898672563741, 0.5}, 1.6, {
                    1.5637141908306973`, 1.531158824545733}], 
                   ArrowBox[{{-1.6844856108028212`, 
                    2.0992973462037807`}, {-1.668495274522341, 
                    2.0987432592873705`}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Dashing[{Small, Small}], 
                   Arrowheads[0.02], 
                   CircleBox[{-1.731898672563741, 0.5}, 1.7, {
                    1.5637141908306973`, 1.5738091196398476`}], 
                   ArrowBox[{{-1.7200205170514062`, 
                    2.199958502264577}, {-1.737020412651878, 
                    2.1999922846232183`}}]}, 
                  {RGBColor[0, 0, 1], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-1.731898672563741, 0.5}, 1.8, {
                    1.5637141908306973`, 1.5830215591513968`}], 
                   ArrowBox[{{-1.7359040874998646`, 
                    2.299995543508703}, {-1.7539035426681164`, 
                    2.2998654909997276`}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.003], Arrowheads[0.025], 
                   CircleBox[{-1.731898672563741, 0.5}, 1.5, {
                    1.5637141908306973`, 1.5635982280952785`}], 
                   ArrowBox[{{-1.7361015190151288`, 
                    1.9999941120156792`}, {-1.7211016177521692`, 
                    1.999961140699117}}]}, 
                  {RGBColor[0, 0, 1], ArrowBox[{{0, 2.8}, {0, 2.8}}]}},
                 Frame->True,
                 FrameTicks->Automatic,
                 ImageSize->800,
                 PlotRange->{{-3, 3}, {-1, 3}}], 470->
                GraphicsBox[{InsetBox[
                   TagBox[GridBox[{
                    {"\<\"t\"\>", "23.45`"},
                    {"\<\"driveIntent\"\>", "0"},
                    {"\<\"motorTorque\"\>", 
                    RowBox[{"-", "3.447836747116255`"}]},
                    {"\<\"vw\"\>", 
                    RowBox[{"-", "0.040725740837631516`"}]},
                    {"\<\"controllerState\"\>", 
                    RowBox[{"\[LeftAssociation]", 
                    RowBox[{
                    
                    RowBox[{"\<\"lastError\"\>", "\[Rule]", 
                    "0.005193135461923604`"}], ",", 
                    RowBox[{"\<\"lastInt\"\>", "\[Rule]", 
                    RowBox[{"-", "0.03511907525708777`"}]}], ",", 
                    RowBox[{"\<\"initialized\"\>", "\[Rule]", "True"}]}], 
                    "\[RightAssociation]"}]}
                    },
                    AutoDelete->False,
                    
                    GridBoxItemSize->{
                    "Columns" -> {{Automatic}}, "Rows" -> {{Automatic}}}],
                    "Grid"], {-1, -0.5}], 
                  {GrayLevel[0], LineBox[{{-5, 0}, {5, 0}}]}, 
                  {GrayLevel[0.5], CircleBox[{-1.733941932600126, 0.5}, 0.5], 
                   DiskBox[{-1.733941932600126, 0.5}, 0.1]}, 
                  {GrayLevel[0.5], 
                   DiskBox[{-1.723404249525795, 1.9999629852884453`}, 0.1]}, 
                  {GrayLevel[0.5], Thickness[0.01], 
                   LineBox[{{-1.733941932600126, 0.5}, {-1.723404249525795, 
                    1.9999629852884453`}}]}, 
                  {RGBColor[0, 1, 0], 
                   DiskBox[{-1.7269168105505721`, 1.4999753235256303`}, 
                    0.05]}, 
                  {RGBColor[0, 0, 1], Arrowheads[0.01], 
                   CircleBox[{-1.733941932600126, 0.5}, 0.3333333333333333, 
                    NCache[{
                    Rational[-1, 2] 
                    Pi, -1.6742314292083842`}, {-1.5707963267948966`, \
-1.6742314292083842`}]], 
                   ArrowBox[{{-1.7650416698757512`, 
                    0.1681206281606249}, {-1.7683588533072874`, 
                    0.16844821418042688`}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-1.733941932600126, 0.5}, {-1.664985197657801, 
                    0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-1.733941932600126, 0.5}, {-1.7856103424349152`,
                     0.5}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.005], Arrowheads[0.04], 
                   ArrowBox[{{-1.733941932600126, 0.5}, {-1.7166536074925902`,
                     0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-1.733941932600126, 0.5}, 1.6, {
                    1.5637711469596909`, 1.5314775939219136`}], 
                   ArrowBox[{{-1.6870386802568622`, 
                    2.0993123787802133`}, {-1.6710481681628646`, 
                    2.0987633891213893`}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Dashing[{Small, Small}], 
                   Arrowheads[0.02], 
                   CircleBox[{-1.733941932600126, 0.5}, 1.7, {
                    1.5637711469596909`, 1.5737902075425068`}], 
                   ArrowBox[{{-1.722031627309247, 
                    2.199958277319734}, {-1.7390315222677835`, 
                    2.1999923811820494`}}]}, 
                  {RGBColor[0, 0, 1], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-1.733941932600126, 0.5}, 1.8, {
                    1.5637711469596909`, 1.582925795554781}], 
                   ArrowBox[{{-1.7377749734710082`, 
                    2.2999959188280625`}, {-1.7557744410110225`, 
                    2.299867590012245}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.003], Arrowheads[0.025], 
                   CircleBox[{-1.733941932600126, 0.5}, 1.5, {
                    1.5637711469596909`, 1.5636652040273615`}], 
                   ArrowBox[{{-1.7382452425457835`, 
                    1.9999938271618025`}, {-1.7232453391081823`, 
                    1.9999618604776817`}}]}, 
                  {RGBColor[0, 0, 1], ArrowBox[{{0, 2.8}, {0, 2.8}}]}},
                 Frame->True,
                 FrameTicks->Automatic,
                 ImageSize->800,
                 PlotRange->{{-3, 3}, {-1, 3}}], 471->
                GraphicsBox[{InsetBox[
                   TagBox[GridBox[{
                    {"\<\"t\"\>", "23.5`"},
                    {"\<\"driveIntent\"\>", "0"},
                    {"\<\"motorTorque\"\>", 
                    RowBox[{"-", "3.4217232391570627`"}]},
                    {"\<\"vw\"\>", 
                    RowBox[{"-", "0.04044285793816835`"}]},
                    {"\<\"controllerState\"\>", 
                    RowBox[{"\[LeftAssociation]", 
                    RowBox[{
                    
                    RowBox[{"\<\"lastError\"\>", "\[Rule]", 
                    "0.005162442699190162`"}], ",", 
                    RowBox[{"\<\"lastInt\"\>", "\[Rule]", 
                    RowBox[{"-", "0.03486019245833922`"}]}], ",", 
                    RowBox[{"\<\"initialized\"\>", "\[Rule]", "True"}]}], 
                    "\[RightAssociation]"}]}
                    },
                    AutoDelete->False,
                    
                    GridBoxItemSize->{
                    "Columns" -> {{Automatic}}, "Rows" -> {{Automatic}}}],
                    "Grid"], {-1, -0.5}], 
                  {GrayLevel[0], LineBox[{{-5, 0}, {5, 0}}]}, 
                  {GrayLevel[0.5], 
                   CircleBox[{-1.7359709976307345`, 0.5}, 0.5], 
                   DiskBox[{-1.7359709976307345`, 0.5}, 0.1]}, 
                  {GrayLevel[0.5], 
                   DiskBox[{-1.725514508194481, 1.9999635535001075`}, 0.1]}, 
                  {GrayLevel[0.5], Thickness[0.01], 
                   LineBox[{{-1.7359709976307345`, 0.5}, {-1.725514508194481, 
                    1.9999635535001075`}}]}, 
                  {RGBColor[0, 1, 0], 
                   DiskBox[{-1.7290000046732321`, 1.4999757023334048`}, 
                    0.05]}, 
                  {RGBColor[0, 0, 1], Arrowheads[0.01], 
                   CircleBox[{-1.7359709976307345`, 0.5}, 0.3333333333333333, 
                    NCache[{
                    Rational[-1, 2] 
                    Pi, -1.6734480239696086`}, {-1.5707963267948966`, \
-1.6734480239696086`}]], 
                   ArrowBox[{{-1.76681072935109, 
                    0.16809636630713487`}, {-1.7701281683972867`, 
                    0.16842135352780074`}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-1.7359709976307345`, 
                    0.5}, {-1.6675365328475933`, 0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-1.7359709976307345`, 0.5}, {-1.787241317916335,
                     0.5}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.005], Arrowheads[0.04], 
                   ArrowBox[{{-1.7359709976307345`, 
                    0.5}, {-1.7188068531331937`, 0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-1.7359709976307345`, 0.5}, 1.6, {
                    1.5638252773772245`, 1.5317805485865879`}], 
                   ArrowBox[{{-1.6895522665779077`, 
                    2.099326514945415}, {-1.6735615888987516`, 
                    2.098782369711939}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Dashing[{Small, Small}], 
                   Arrowheads[0.02], 
                   CircleBox[{-1.7359709976307345`, 0.5}, 1.7, {
                    1.5638252773772245`, 1.5737723757278268`}], 
                   ArrowBox[{{-1.724030379000786, 
                    2.1999580646671064`}, {-1.741030273348486, 
                    2.199992471668393}}]}, 
                  {RGBColor[0, 0, 1], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-1.7359709976307345`, 0.5}, 1.8, {
                    1.5638252773772245`, 1.582834850928097}], 
                   ArrowBox[{{-1.7396403385291201`, 
                    2.299996259978718}, {-1.7576398176655146`, 
                    2.2998695681182846`}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.003], Arrowheads[0.025], 
                   CircleBox[{-1.7359709976307345`, 0.5}, 1.5, {
                    1.5638252773772245`, 1.5637290997256086`}], 
                   ArrowBox[{{-1.740370150720495, 
                    1.9999935491368264`}, {-1.72537024527098, 
                    1.9999625408820738`}}]}, 
                  {RGBColor[0, 0, 1], ArrowBox[{{0, 2.8}, {0, 2.8}}]}},
                 Frame->True,
                 FrameTicks->Automatic,
                 ImageSize->800,
                 PlotRange->{{-3, 3}, {-1, 3}}], 472->
                GraphicsBox[{InsetBox[
                   TagBox[GridBox[{
                    {"\<\"t\"\>", "23.55`"},
                    {"\<\"driveIntent\"\>", "0"},
                    {"\<\"motorTorque\"\>", 
                    RowBox[{"-", "3.3968546377439774`"}]},
                    {"\<\"vw\"\>", 
                    RowBox[{"-", "0.04016196487429922`"}]},
                    {"\<\"controllerState\"\>", 
                    RowBox[{"\[LeftAssociation]", 
                    RowBox[{
                    
                    RowBox[{"\<\"lastError\"\>", "\[Rule]", 
                    "0.005129776910091001`"}], ",", 
                    RowBox[{"\<\"lastInt\"\>", "\[Rule]", 
                    RowBox[{"-", "0.03460289555399604`"}]}], ",", 
                    RowBox[{"\<\"initialized\"\>", "\[Rule]", "True"}]}], 
                    "\[RightAssociation]"}]}
                    },
                    AutoDelete->False,
                    
                    GridBoxItemSize->{
                    "Columns" -> {{Automatic}}, "Rows" -> {{Automatic}}}],
                    "Grid"], {-1, -0.5}], 
                  {GrayLevel[0], LineBox[{{-5, 0}, {5, 0}}]}, 
                  {GrayLevel[0.5], 
                   CircleBox[{-1.7379859696701903`, 0.5}, 0.5], 
                   DiskBox[{-1.7379859696701903`, 0.5}, 0.1]}, 
                  {GrayLevel[0.5], 
                   DiskBox[{-1.7276068258270196`, 1.9999640906945348`}, 0.1]}, 
                  {GrayLevel[0.5], Thickness[0.01], 
                   LineBox[{{-1.7379859696701903`, 
                    0.5}, {-1.7276068258270196`, 1.9999640906945348`}}]}, 
                  {RGBColor[0, 1, 0], 
                   DiskBox[{-1.7310665404414098`, 1.4999760604630232`}, 
                    0.05]}, 
                  {RGBColor[0, 0, 1], Arrowheads[0.01], 
                   CircleBox[{-1.7379859696701903`, 0.5}, 0.3333333333333333, 
                    NCache[{
                    Rational[-1, 2] 
                    Pi, -1.672701965927216}, {-1.5707963267948966`, \
-1.672701965927216}]], 
                   ArrowBox[{{-1.7685780734555638`, 
                    0.16807345044860295`}, {-1.7718957540378202`, 
                    0.16839596257697353`}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-1.7379859696701903`, 
                    0.5}, {-1.6700488769153108`, 0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-1.7379859696701903`, 
                    0.5}, {-1.7888770668165364`, 0.5}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.005], Arrowheads[0.04], 
                   ArrowBox[{{-1.7379859696701903`, 
                    0.5}, {-1.7209399740616569`, 0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-1.7379859696701903`, 0.5}, 1.6, {
                    1.56387684234961, 1.532069145171601}], 
                   ArrowBox[{{-1.6920288007145523`, 
                    2.0993398446301472`}, {-1.6760379666628578`, 
                    2.0988003142742384`}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Dashing[{Small, Small}], 
                   Arrowheads[0.02], 
                   CircleBox[{-1.7379859696701903`, 0.5}, 1.7, {
                    1.56387684234961, 1.573755473754616}], 
                   ArrowBox[{{-1.7260166183962802`, 
                    2.199957862604271}, {-1.7430165121600056`, 
                    2.1999925569373113`}}]}, 
                  {RGBColor[0, 0, 1], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-1.7379859696701903`, 0.5}, 1.8, {
                    1.56387684234961, 1.5827482570037432`}], 
                   ArrowBox[{{-1.7414994418150398`, 
                    2.2999965709727026`}, {-1.7594989318546952`, 
                    2.299871437758277}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.003], Arrowheads[0.025], 
                   CircleBox[{-1.7379859696701903`, 0.5}, 1.5, {
                    1.56387684234961, 1.5637901118703668`}], 
                   ArrowBox[{{-1.7424766405752623`, 
                    1.9999932779432121`}, {-1.7274767332617857`, 
                    1.9999631848649193`}}]}, 
                  {RGBColor[0, 0, 1], ArrowBox[{{0, 2.8}, {0, 2.8}}]}},
                 Frame->True,
                 FrameTicks->Automatic,
                 ImageSize->800,
                 PlotRange->{{-3, 3}, {-1, 3}}], 473->
                GraphicsBox[{InsetBox[
                   TagBox[GridBox[{
                    {"\<\"t\"\>", "23.6`"},
                    {"\<\"driveIntent\"\>", "0"},
                    {"\<\"motorTorque\"\>", 
                    RowBox[{"-", "3.373104485646569`"}]},
                    {"\<\"vw\"\>", 
                    RowBox[{"-", "0.03988296808459561`"}]},
                    {"\<\"controllerState\"\>", 
                    RowBox[{"\[LeftAssociation]", 
                    RowBox[{
                    
                    RowBox[{"\<\"lastError\"\>", "\[Rule]", 
                    "0.005095361949756889`"}], ",", 
                    RowBox[{"\<\"lastInt\"\>", "\[Rule]", 
                    RowBox[{"-", "0.03434727746270995`"}]}], ",", 
                    RowBox[{"\<\"initialized\"\>", "\[Rule]", "True"}]}], 
                    "\[RightAssociation]"}]}
                    },
                    AutoDelete->False,
                    
                    GridBoxItemSize->{
                    "Columns" -> {{Automatic}}, "Rows" -> {{Automatic}}}],
                    "Grid"], {-1, -0.5}], 
                  {GrayLevel[0], LineBox[{{-5, 0}, {5, 0}}]}, 
                  {GrayLevel[0.5], 
                   CircleBox[{-1.7399869457780988`, 0.5}, 0.5], 
                   DiskBox[{-1.7399869457780988`, 0.5}, 0.1]}, 
                  {GrayLevel[0.5], 
                   DiskBox[{-1.729681677179284, 1.9999646000619835`}, 0.1]}, 
                  {GrayLevel[0.5], Thickness[0.01], 
                   LineBox[{{-1.7399869457780988`, 0.5}, {-1.729681677179284, 
                    1.9999646000619835`}}]}, 
                  {RGBColor[0, 1, 0], 
                   DiskBox[{-1.7331167667122223`, 1.4999764000413225`}, 
                    0.05]}, 
                  {RGBColor[0, 0, 1], Arrowheads[0.01], 
                   CircleBox[{-1.7399869457780988`, 0.5}, 0.3333333333333333, 
                    NCache[{
                    Rational[-1, 2] 
                    Pi, -1.6719894613642936`}, {-1.5707963267948966`, \
-1.6719894613642936`}]], 
                   ArrowBox[{{-1.7703425426371366`, 
                    0.16805173769028003`}, {-1.7736604521686052`, 
                    0.16837188587443375`}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-1.7399869457780988`, 
                    0.5}, {-1.6725248560651673`, 0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-1.7399869457780988`, 
                    0.5}, {-1.7905158344225554`, 0.5}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.005], Arrowheads[0.04], 
                   ArrowBox[{{-1.7399869457780988`, 0.5}, {-1.723053744709624,
                     0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-1.7399869457780988`, 0.5}, 1.6, {
                    1.5639260936831958`, 1.5323447929630185`}], 
                   ArrowBox[{{-1.694470633058728, 
                    2.0993524518618254`}, {-1.67847965089419, 
                    2.098817329364447}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Dashing[{Small, Small}], 
                   Arrowheads[0.02], 
                   CircleBox[{-1.7399869457780988`, 0.5}, 1.7, {
                    1.5639260936831958`, 1.573739360902552}], 
                   ArrowBox[{{-1.7279902033361887`, 
                    2.1999576695232097`}, {-1.7449900965386826`, 
                    2.1999926377730192`}}]}, 
                  {RGBColor[0, 0, 1], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-1.7399869457780988`, 0.5}, 1.8, {
                    1.5639260936831958`, 1.582665563047899}], 
                   ArrowBox[{{-1.743351569074145, 
                    2.2999968553611634`}, {-1.761351069400018, 
                    2.2998732105961985`}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.003], Arrowheads[0.025], 
                   CircleBox[{-1.7399869457780988`, 0.5}, 1.5, {
                    1.5639260936831958`, 1.5638484390012544`}], 
                   ArrowBox[{{-1.7445651069797357`, 
                    1.9999930134637334`}, {-1.7295651979365314`, 
                    1.9999637952870486`}}]}, 
                  {RGBColor[0, 0, 1], ArrowBox[{{0, 2.8}, {0, 2.8}}]}},
                 Frame->True,
                 FrameTicks->Automatic,
                 ImageSize->800,
                 PlotRange->{{-3, 3}, {-1, 3}}], 474->
                GraphicsBox[{InsetBox[
                   TagBox[GridBox[{
                    {"\<\"t\"\>", "23.650000000000002`"},
                    {"\<\"driveIntent\"\>", "0"},
                    {"\<\"motorTorque\"\>", 
                    RowBox[{"-", "3.350351669651163`"}]},
                    {"\<\"vw\"\>", 
                    RowBox[{"-", "0.0396057848599444`"}]},
                    {"\<\"controllerState\"\>", 
                    RowBox[{"\[LeftAssociation]", 
                    RowBox[{
                    
                    RowBox[{"\<\"lastError\"\>", "\[Rule]", 
                    "0.005059415076337875`"}], ",", 
                    RowBox[{"\<\"lastInt\"\>", "\[Rule]", 
                    RowBox[{"-", "0.034093420072503926`"}]}], ",", 
                    RowBox[{"\<\"initialized\"\>", "\[Rule]", "True"}]}], 
                    "\[RightAssociation]"}]}
                    },
                    AutoDelete->False,
                    GridBoxItemSize->{
                    "Columns" -> {{Automatic}}, "Rows" -> {{Automatic}}}],
                    "Grid"], {-1, -0.5}], 
                  {GrayLevel[0], LineBox[{{-5, 0}, {5, 0}}]}, 
                  {GrayLevel[0.5], 
                   CircleBox[{-1.7419740186150656`, 0.5}, 0.5], 
                   DiskBox[{-1.7419740186150656`, 0.5}, 0.1]}, 
                  {GrayLevel[0.5], 
                   DiskBox[{-1.731739517786389, 1.999965084591234}, 0.1]}, 
                  {GrayLevel[0.5], Thickness[0.01], 
                   LineBox[{{-1.7419740186150656`, 0.5}, {-1.731739517786389, 
                    1.999965084591234}}]}, 
                  {RGBColor[0, 1, 0], 
                   DiskBox[{-1.7351510180626146`, 1.4999767230608225`}, 
                    0.05]}, 
                  {RGBColor[0, 0, 1], Arrowheads[0.01], 
                   CircleBox[{-1.7419740186150656`, 0.5}, 0.3333333333333333, 
                    NCache[{
                    Rational[-1, 2] 
                    Pi, -1.6713068768844315`}, {-1.5707963267948966`, \
-1.6713068768844315`}]], 
                   ArrowBox[{{-1.7721030256880652`, 
                    0.16803109476352118`}, {-1.775421152974756, 
                    0.16834897811971705`}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-1.7419740186150656`, 
                    0.5}, {-1.6749669852220423`, 0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-1.7419740186150656`, 0.5}, {-1.792155934282247,
                     0.5}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.005], Arrowheads[0.04], 
                   ArrowBox[{{-1.7419740186150656`, 
                    0.5}, {-1.7251489008892238`, 0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-1.7419740186150656`, 0.5}, 1.6, {
                    1.5639732733024296`, 1.5326088459425733`}], 
                   ArrowBox[{{-1.696880021257849, 
                    2.0993644148230723`}, {-1.6808888983501, 
                    2.098833514810787}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Dashing[{Small, Small}], 
                   Arrowheads[0.02], 
                   CircleBox[{-1.7419740186150656`, 0.5}, 1.7, {
                    1.5639732733024296`, 1.5737239063983384`}], 
                   ArrowBox[{{-1.7299510041716224`, 
                    2.199957483916493}, {-1.7469508968316694`, 
                    2.199992714891219}}]}, 
                  {RGBColor[0, 0, 1], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-1.7419740186150656`, 0.5}, 1.8, {
                    1.5639732733024296`, 1.582586338133825}], 
                   ArrowBox[{{-1.7451960373045026`, 
                    2.299997116274236}, {-1.7631955473696335`, 
                    2.2998748975185244`}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.003], Arrowheads[0.025], 
                   CircleBox[{-1.7419740186150656`, 0.5}, 1.5, {
                    1.5639732733024296`, 1.5639042794249538`}], 
                   ArrowBox[{{-1.7466359400549394`, 
                    1.9999927554786685`}, {-1.7316360294035655`, 
                    1.9999643749033051`}}]}, 
                  {RGBColor[0, 0, 1], ArrowBox[{{0, 2.8}, {0, 2.8}}]}},
                 Frame->True,
                 FrameTicks->Automatic,
                 ImageSize->800,
                 PlotRange->{{-3, 3}, {-1, 3}}], 475->
                GraphicsBox[{InsetBox[
                   TagBox[GridBox[{
                    {"\<\"t\"\>", "23.7`"},
                    {"\<\"driveIntent\"\>", "0"},
                    {"\<\"motorTorque\"\>", 
                    RowBox[{"-", "3.328480958441313`"}]},
                    {"\<\"vw\"\>", 
                    RowBox[{"-", "0.0393303427630222`"}]},
                    {"\<\"controllerState\"\>", 
                    RowBox[{"\[LeftAssociation]", 
                    RowBox[{
                    
                    RowBox[{"\<\"lastError\"\>", "\[Rule]", 
                    "0.005022145607429612`"}], ",", 
                    RowBox[{"\<\"lastInt\"\>", "\[Rule]", 
                    RowBox[{"-", "0.033841394605872405`"}]}], ",", 
                    RowBox[{"\<\"initialized\"\>", "\[Rule]", "True"}]}], 
                    "\[RightAssociation]"}]}
                    },
                    AutoDelete->False,
                    
                    GridBoxItemSize->{
                    "Columns" -> {{Automatic}}, "Rows" -> {{Automatic}}}],
                    "Grid"], {-1, -0.5}], 
                  {GrayLevel[0], LineBox[{{-5, 0}, {5, 0}}]}, 
                  {GrayLevel[0.5], 
                   CircleBox[{-1.7439472769706654`, 0.5}, 0.5], 
                   DiskBox[{-1.7439472769706654`, 0.5}, 0.1]}, 
                  {GrayLevel[0.5], 
                   DiskBox[{-1.7337807827436098`, 1.9999655470693756`}, 0.1]}, 
                  {GrayLevel[0.5], Thickness[0.01], 
                   LineBox[{{-1.7439472769706654`, 
                    0.5}, {-1.7337807827436098`, 1.9999655470693756`}}]}, 
                  {RGBColor[0, 1, 0], 
                   DiskBox[{-1.7371696141526283`, 1.4999770313795837`}, 
                    0.05]}, 
                  {RGBColor[0, 0, 1], Arrowheads[0.01], 
                   CircleBox[{-1.7439472769706654`, 0.5}, 0.3333333333333333, 
                    NCache[{
                    Rational[-1, 2] 
                    Pi, -1.670650755548136}, {-1.5707963267948966`, \
-1.670650755548136}]], 
                   ArrowBox[{{-1.7738584656923841`, 
                    0.16801139793606645`}, {-1.7771768008348934`, 
                    0.1683271041298855}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-1.7439472769706654`, 
                    0.5}, {-1.6773776578018391`, 0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-1.7439472769706654`, 
                    0.5}, {-1.7937957572994212`, 0.5}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.005], Arrowheads[0.04], 
                   ArrowBox[{{-1.7439472769706654`, 
                    0.5}, {-1.7272261381305951`, 0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-1.7439472769706654`, 0.5}, 1.6, {
                    1.5640186120851944`, 1.5328625962690212`}], 
                   ArrowBox[{{-1.6992591203032479`, 
                    2.0993758059485796`}, {-1.683267863194277, 
                    2.0988489637060015`}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Dashing[{Small, Small}], 
                   Arrowheads[0.02], 
                   CircleBox[{-1.7439472769706654`, 0.5}, 1.7, {
                    1.5640186120851944`, 1.5737089895206657`}], 
                   ArrowBox[{{-1.7318989044707245`, 
                    2.1999573043815257`}, {-1.7488987966033438`, 
                    2.199992788941567}}]}, 
                  {RGBColor[0, 0, 1], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-1.7439472769706654`, 0.5}, 1.8, {
                    1.5640186120851944`, 1.582510172965424}], 
                   ArrowBox[{{-1.747032198567407, 
                    2.2999973564588205`}, {-1.7650317178891415`, 
                    2.2998765086391777`}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.003], Arrowheads[0.025], 
                   CircleBox[{-1.7439472769706654`, 0.5}, 1.5, {
                    1.5640186120851944`, 1.563957829364419}], 
                   ArrowBox[{{-1.7486895229250705`, 
                    1.9999925036823711`}, {-1.7336896107754256`, 
                    1.9999649263513555`}}]}, 
                  {RGBColor[0, 0, 1], ArrowBox[{{0, 2.8}, {0, 2.8}}]}},
                 Frame->True,
                 FrameTicks->Automatic,
                 ImageSize->800,
                 PlotRange->{{-3, 3}, {-1, 3}}], 476->
                GraphicsBox[{InsetBox[
                   TagBox[GridBox[{
                    {"\<\"t\"\>", "23.75`"},
                    {"\<\"driveIntent\"\>", "0"},
                    {"\<\"motorTorque\"\>", 
                    RowBox[{"-", "3.307383422631951`"}]},
                    {"\<\"vw\"\>", 
                    RowBox[{"-", "0.039056579016702564`"}]},
                    {"\<\"controllerState\"\>", 
                    RowBox[{"\[LeftAssociation]", 
                    RowBox[{
                    
                    RowBox[{"\<\"lastError\"\>", "\[Rule]", 
                    "0.004983753806481799`"}], ",", 
                    RowBox[{"\<\"lastInt\"\>", "\[Rule]", 
                    RowBox[{"-", "0.03359126204613098`"}]}], ",", 
                    RowBox[{"\<\"initialized\"\>", "\[Rule]", "True"}]}], 
                    "\[RightAssociation]"}]}
                    },
                    AutoDelete->False,
                    
                    GridBoxItemSize->{
                    "Columns" -> {{Automatic}}, "Rows" -> {{Automatic}}}],
                    "Grid"], {-1, -0.5}], 
                  {GrayLevel[0], LineBox[{{-5, 0}, {5, 0}}]}, 
                  {GrayLevel[0.5], 
                   CircleBox[{-1.7459068062615373`, 0.5}, 0.5], 
                   DiskBox[{-1.7459068062615373`, 0.5}, 0.1]}, 
                  {GrayLevel[0.5], 
                   DiskBox[{-1.7358058858159129`, 1.999965990083159}, 0.1]}, 
                  {GrayLevel[0.5], Thickness[0.01], 
                   LineBox[{{-1.7459068062615373`, 
                    0.5}, {-1.7358058858159129`, 1.999965990083159}}]}, 
                  {RGBColor[0, 1, 0], 
                   DiskBox[{-1.7391728592977878`, 1.499977326722106}, 0.05]}, 
                  {RGBColor[0, 0, 1], Arrowheads[0.01], 
                   CircleBox[{-1.7459068062615373`, 0.5}, 0.3333333333333333, 
                    NCache[{
                    Rational[-1, 2] 
                    Pi, -1.670017829473855}, {-1.5707963267948966`, \
-1.670017829473855}]], 
                   ArrowBox[{{-1.7756078647635325`, 
                    0.1679925328626295}, {-1.7789263990600552`, 
                    0.16830613873251865`}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-1.7459068062615373`, 
                    0.5}, {-1.6797591378088983`, 0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-1.7459068062615373`, 
                    0.5}, {-1.7954337790412997`, 0.5}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.005], Arrowheads[0.04], 
                   ArrowBox[{{-1.7459068062615373`, 
                    0.5}, {-1.7292861105886606`, 0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-1.7459068062615373`, 0.5}, 1.6, {
                    1.5640623289371351`, 1.5331072690971097`}], 
                   ArrowBox[{{-1.7016099747294566`, 
                    2.099386692052994}, {-1.6856185891951623`, 
                    2.0988637624522477`}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Dashing[{Small, Small}], 
                   Arrowheads[0.02], 
                   CircleBox[{-1.7459068062615373`, 0.5}, 1.7, {
                    1.5640623289371351`, 1.5736944995969484`}], 
                   ArrowBox[{{-1.733833801511199, 
                    2.1999571296230673`}, {-1.750833693127865, 
                    2.1999928605102452`}}]}, 
                  {RGBColor[0, 0, 1], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-1.7459068062615373`, 0.5}, 1.8, {
                    1.5640623289371351`, 1.5824366812850905`}], 
                   ArrowBox[{{-1.748859443019745, 
                    2.2999975783139748`}, {-1.7668589711741807`, 
                    2.2998780533095773`}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.003], Arrowheads[0.025], 
                   CircleBox[{-1.7459068062615373`, 0.5}, 1.5, {
                    1.5640623289371351`, 1.564009281339776}], 
                   ArrowBox[{{-1.7507262297869663`, 
                    1.9999922576989797`}, {-1.735726316238268, 
                    1.9999654521431105`}}]}, 
                  {RGBColor[0, 0, 1], ArrowBox[{{0, 2.8}, {0, 2.8}}]}},
                 Frame->True,
                 FrameTicks->Automatic,
                 ImageSize->800,
                 PlotRange->{{-3, 3}, {-1, 3}}], 477->
                GraphicsBox[{InsetBox[
                   TagBox[GridBox[{
                    {"\<\"t\"\>", "23.8`"},
                    {"\<\"driveIntent\"\>", "0"},
                    {"\<\"motorTorque\"\>", 
                    RowBox[{"-", "3.2869567464077107`"}]},
                    {"\<\"vw\"\>", 
                    RowBox[{"-", "0.03878443987080569`"}]},
                    {"\<\"controllerState\"\>", 
                    RowBox[{"\[LeftAssociation]", 
                    RowBox[{
                    
                    RowBox[{"\<\"lastError\"\>", "\[Rule]", 
                    "0.0049444299835430345`"}], ",", 
                    RowBox[{"\<\"lastInt\"\>", "\[Rule]", 
                    RowBox[{"-", "0.033343073613914594`"}]}], ",", 
                    RowBox[{"\<\"initialized\"\>", "\[Rule]", "True"}]}], 
                    "\[RightAssociation]"}]}
                    },
                    AutoDelete->False,
                    
                    GridBoxItemSize->{
                    "Columns" -> {{Automatic}}, "Rows" -> {{Automatic}}}],
                    "Grid"], {-1, -0.5}], 
                  {GrayLevel[0], LineBox[{{-5, 0}, {5, 0}}]}, 
                  {GrayLevel[0.5], 
                   CircleBox[{-1.7478526889983248`, 0.5}, 0.5], 
                   DiskBox[{-1.7478526889983248`, 0.5}, 0.1]}, 
                  {GrayLevel[0.5], 
                   DiskBox[{-1.7378152188471125`, 1.9999664160216932`}, 0.1]}, 
                  {GrayLevel[0.5], Thickness[0.01], 
                   LineBox[{{-1.7478526889983248`, 
                    0.5}, {-1.7378152188471125`, 1.9999664160216932`}}]}, 
                  {RGBColor[0, 1, 0], 
                   DiskBox[{-1.7411610422308499`, 1.499977610681129}, 0.05]}, 
                  {RGBColor[0, 0, 1], Arrowheads[0.01], 
                   CircleBox[{-1.7478526889983248`, 0.5}, 0.3333333333333333, 
                    NCache[{
                    Rational[-1, 2] 
                    Pi, -1.6694050291871279`}, {-1.5707963267948966`, \
-1.6694050291871279`}]], 
                   ArrowBox[{{-1.7773502876652636`, 
                    0.16797439438501838`}, {-1.7806690135164462`, 
                    0.1682859665973831}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-1.7478526889983248`, 
                    0.5}, {-1.6821135540701706`, 0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-1.7478526889983248`, 0.5}, {-1.797068565392896,
                     0.5}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.005], Arrowheads[0.04], 
                   ArrowBox[{{-1.7478526889983248`, 
                    0.5}, {-1.7313294304647417`, 0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-1.7478526889983248`, 0.5}, 1.6, {
                    1.5641046300865034`, 1.5333440186307579`}], 
                   ArrowBox[{{-1.7039345127575998`, 
                    2.0993971344840183`}, {-1.6879430038681287`, 
                    2.0988779908509594`}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Dashing[{Small, Small}], 
                   Arrowheads[0.02], 
                   CircleBox[{-1.7478526889983248`, 0.5}, 1.7, {
                    1.5641046300865034`, 1.5736803359041462`}], 
                   ArrowBox[{{-1.7357556065786373`, 
                    2.1999569584542225`}, {-1.752755497687517, 
                    2.1999929301226393`}}]}, 
                  {RGBColor[0, 0, 1], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-1.7478526889983248`, 0.5}, 1.8, {
                    1.5641046300865034`, 1.5823655008998796`}], 
                   ArrowBox[{{-1.7506772012281575`, 
                    2.2999977839238204`}, {-1.7686767378448305`, 
                    2.299879540133072}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.003], Arrowheads[0.025], 
                   CircleBox[{-1.7478526889983248`, 0.5}, 1.5, {
                    1.5641046300865034`, 1.564058822769755}], 
                   ArrowBox[{{-1.752746424279213, 
                    1.9999920170970913`}, {-1.7377465094209366`, 
                    1.9999659546584214`}}]}, 
                  {RGBColor[0, 0, 1], ArrowBox[{{0, 2.8}, {0, 2.8}}]}},
                 Frame->True,
                 FrameTicks->Automatic,
                 ImageSize->800,
                 PlotRange->{{-3, 3}, {-1, 3}}], 478->
                GraphicsBox[{InsetBox[
                   TagBox[GridBox[{
                    {"\<\"t\"\>", "23.85`"},
                    {"\<\"driveIntent\"\>", "0"},
                    {"\<\"motorTorque\"\>", 
                    RowBox[{"-", "3.267105440190599`"}]},
                    {"\<\"vw\"\>", 
                    RowBox[{"-", "0.038513879955730475`"}]},
                    {"\<\"controllerState\"\>", 
                    RowBox[{"\[LeftAssociation]", 
                    RowBox[{
                    
                    RowBox[{"\<\"lastError\"\>", "\[Rule]", 
                    "0.004904353794432214`"}], ",", 
                    RowBox[{"\<\"lastInt\"\>", "\[Rule]", 
                    RowBox[{"-", "0.03309687128351183`"}]}], ",", 
                    RowBox[{"\<\"initialized\"\>", "\[Rule]", "True"}]}], 
                    "\[RightAssociation]"}]}
                    },
                    AutoDelete->False,
                    
                    GridBoxItemSize->{
                    "Columns" -> {{Automatic}}, "Rows" -> {{Automatic}}}],
                    "Grid"], {-1, -0.5}], 
                  {GrayLevel[0], LineBox[{{-5, 0}, {5, 0}}]}, 
                  {GrayLevel[0.5], 
                   CircleBox[{-1.7497850052205743`, 0.5}, 0.5], 
                   DiskBox[{-1.7497850052205743`, 0.5}, 0.1]}, 
                  {GrayLevel[0.5], 
                   DiskBox[{-1.739809151440047, 1.9999668270803024`}, 0.1]}, 
                  {GrayLevel[0.5], Thickness[0.01], 
                   LineBox[{{-1.7497850052205743`, 0.5}, {-1.739809151440047, 
                    1.9999668270803024`}}]}, 
                  {RGBColor[0, 1, 0], 
                   DiskBox[{-1.7431344360335561`, 1.4999778847202014`}, 
                    0.05]}, 
                  {RGBColor[0, 0, 1], Arrowheads[0.01], 
                   CircleBox[{-1.7497850052205743`, 0.5}, 0.3333333333333333, 
                    NCache[{
                    Rational[-1, 2] 
                    Pi, -1.6688094900006145`}, {-1.5707963267948966`, \
-1.6688094900006145`}]], 
                   ArrowBox[{{-1.7790848644092208`, 
                    0.16795688628939076`}, {-1.7824037752253332`, 
                    0.16826648201532646`}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-1.7497850052205743`, 
                    0.5}, {-1.6844428964167624`, 0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-1.7497850052205743`, 
                    0.5}, {-1.7986987766933127`, 0.5}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.005], Arrowheads[0.04], 
                   ArrowBox[{{-1.7497850052205743`, 
                    0.5}, {-1.7333566678895007`, 0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-1.7497850052205743`, 0.5}, 1.6, {
                    1.5641457085810453`, 1.5335739253070426`}], 
                   ArrowBox[{{-1.7062345422165517`, 
                    2.0994071892961266`}, {-1.6902429143951267`, 
                    2.0988917222314134`}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Dashing[{Small, Small}], 
                   Arrowheads[0.02], 
                   CircleBox[{-1.7497850052205743`, 0.5}, 1.7, {
                    1.5641457085810453`, 1.573666407485693}], 
                   ArrowBox[{{-1.7376642450901911`, 
                    2.1999567897961}, {-1.7546641356963935`, 
                    2.1999929982461106`}}]}, 
                  {RGBColor[0, 0, 1], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-1.7497850052205743`, 0.5}, 1.8, {
                    1.5641457085810453`, 1.582296294359882}], 
                   ArrowBox[{{-1.7524849458251137`, 
                    2.299997975087953}, {-1.7704844905819255`, 
                    2.299880976983138}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.003], Arrowheads[0.025], 
                   CircleBox[{-1.7497850052205743`, 0.5}, 1.5, {
                    1.5641457085810453`, 1.5641066347819805`}], 
                   ArrowBox[{{-1.7547504581325015`, 
                    1.999991781403278}, {-1.7397505420452721`, 
                    1.9999664361407503`}}]}, 
                  {RGBColor[0, 0, 1], ArrowBox[{{0, 2.8}, {0, 2.8}}]}},
                 Frame->True,
                 FrameTicks->Automatic,
                 ImageSize->800,
                 PlotRange->{{-3, 3}, {-1, 3}}], 479->
                GraphicsBox[{InsetBox[
                   TagBox[GridBox[{
                    {"\<\"t\"\>", "23.900000000000002`"},
                    {"\<\"driveIntent\"\>", "0"},
                    {"\<\"motorTorque\"\>", 
                    RowBox[{"-", "3.247740963617473`"}]},
                    {"\<\"vw\"\>", 
                    RowBox[{"-", "0.03824486163065827`"}]},
                    {"\<\"controllerState\"\>", 
                    RowBox[{"\[LeftAssociation]", 
                    RowBox[{
                    
                    RowBox[{"\<\"lastError\"\>", "\[Rule]", 
                    "0.0048636937223458425`"}], ",", 
                    RowBox[{"\<\"lastInt\"\>", "\[Rule]", 
                    RowBox[{"-", "0.03285268832952191`"}]}], ",", 
                    RowBox[{"\<\"initialized\"\>", "\[Rule]", "True"}]}], 
                    "\[RightAssociation]"}]}
                    },
                    AutoDelete->False,
                    
                    GridBoxItemSize->{
                    "Columns" -> {{Automatic}}, "Rows" -> {{Automatic}}}],
                    "Grid"], {-1, -0.5}], 
                  {GrayLevel[0], LineBox[{{-5, 0}, {5, 0}}]}, 
                  {GrayLevel[0.5], CircleBox[{-1.751703832899153, 0.5}, 0.5], 
                   DiskBox[{-1.751703832899153, 0.5}, 0.1]}, 
                  {GrayLevel[0.5], 
                   DiskBox[{-1.7417880308798954`, 1.9999672252653773`}, 0.1]}, 
                  {GrayLevel[0.5], Thickness[0.01], 
                   LineBox[{{-1.751703832899153, 0.5}, {-1.7417880308798954`, 
                    1.9999672252653773`}}]}, 
                  {RGBColor[0, 1, 0], 
                   DiskBox[{-1.7450932982196479`, 1.499978150176918}, 0.05]}, 
                  {RGBColor[0, 0, 1], Arrowheads[0.01], 
                   CircleBox[{-1.751703832899153, 0.5}, 0.3333333333333333, 
                    NCache[{
                    Rational[-1, 2] 
                    Pi, -1.6682285557034207`}, {-1.5707963267948966`, \
-1.6682285557034207`}]], 
                   ArrowBox[{{-1.7808107919216214`, 
                    0.1679399210269692}, {-1.7841298820324587`, 
                    0.16824758863164907`}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-1.751703832899153, 0.5}, {-1.6867490136268035`,
                     0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-1.751703832899153, 0.5}, {-1.8003231704880924`,
                     0.5}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.005], Arrowheads[0.04], 
                   ArrowBox[{{-1.751703832899153, 0.5}, {-1.735368351215743, 
                    0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-1.751703832899153, 0.5}, 1.6, {
                    1.5641857439686329`, 1.5337979940085853`}], 
                   ArrowBox[{{-1.7085117480775396`, 
                    2.099416907441197}, {-1.6925200051575238`, 
                    2.098905023612675}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Dashing[{Small, Small}], 
                   Arrowheads[0.02], 
                   CircleBox[{-1.751703832899153, 0.5}, 1.7, {
                    1.5641857439686329`, 1.573652632895811}], 
                   ArrowBox[{{-1.739559656562324, 
                    2.199956622676326}, {-1.7565595466681567`, 
                    2.199993065292854}}]}, 
                  {RGBColor[0, 0, 1], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-1.751703832899153, 0.5}, 1.8, {
                    1.5641857439686329`, 1.5822287493220633`}], 
                   ArrowBox[{{-1.754282192566325, 
                    2.299998153349449}, {-1.7722817451847181`, 
                    2.299882371024831}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.003], Arrowheads[0.025], 
                   CircleBox[{-1.751703832899153, 0.5}, 1.5, {
                    1.5641857439686329`, 1.5641528912195004`}], 
                   ArrowBox[{{-1.75673867008186, 
                    1.999991550114381}, {-1.7417387528382966`, 
                    1.9999668986945616`}}]}, 
                  {RGBColor[0, 0, 1], ArrowBox[{{0, 2.8}, {0, 2.8}}]}},
                 Frame->True,
                 FrameTicks->Automatic,
                 ImageSize->800,
                 PlotRange->{{-3, 3}, {-1, 3}}], 480->
                GraphicsBox[{InsetBox[
                   TagBox[GridBox[{
                    {"\<\"t\"\>", "23.95`"},
                    {"\<\"driveIntent\"\>", "0"},
                    {"\<\"motorTorque\"\>", 
                    RowBox[{"-", "3.2287817679138358`"}]},
                    {"\<\"vw\"\>", 
                    RowBox[{"-", "0.037977354333189`"}]},
                    {"\<\"controllerState\"\>", 
                    RowBox[{"\[LeftAssociation]", 
                    RowBox[{
                    
                    RowBox[{"\<\"lastError\"\>", "\[Rule]", 
                    "0.004822606725991569`"}], ",", 
                    RowBox[{"\<\"lastInt\"\>", "\[Rule]", 
                    RowBox[{"-", "0.03261054989511915`"}]}], ",", 
                    RowBox[{"\<\"initialized\"\>", "\[Rule]", "True"}]}], 
                    "\[RightAssociation]"}]}
                    },
                    AutoDelete->False,
                    
                    GridBoxItemSize->{
                    "Columns" -> {{Automatic}}, "Rows" -> {{Automatic}}}],
                    "Grid"], {-1, -0.5}], 
                  {GrayLevel[0], LineBox[{{-5, 0}, {5, 0}}]}, 
                  {GrayLevel[0.5], CircleBox[{-1.753609248306086, 0.5}, 0.5], 
                   DiskBox[{-1.753609248306086, 0.5}, 0.1]}, 
                  {GrayLevel[0.5], 
                   DiskBox[{-1.743752182273584, 1.9999676124000914`}, 0.1]}, 
                  {GrayLevel[0.5], Thickness[0.01], 
                   LineBox[{{-1.753609248306086, 0.5}, {-1.743752182273584, 
                    1.9999676124000914`}}]}, 
                  {RGBColor[0, 1, 0], 
                   DiskBox[{-1.7470378709510845`, 1.4999784082667276`}, 
                    0.05]}, 
                  {RGBColor[0, 0, 1], Arrowheads[0.01], 
                   CircleBox[{-1.753609248306086, 0.5}, 0.3333333333333333, 
                    NCache[{
                    Rational[-1, 2] 
                    Pi, -1.6676597798323116`}, {-1.5707963267948966`, \
-1.6676597798323116`}]], 
                   ArrowBox[{{-1.7825273348699215`, 
                    0.16792341940359956`}, {-1.7858465994377866`, 
                    0.16822919914024598`}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-1.753609248306086, 0.5}, {-1.6890336129478094`,
                     0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-1.753609248306086, 0.5}, {-1.8019406030296525`,
                     0.5}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.005], Arrowheads[0.04], 
                   ArrowBox[{{-1.753609248306086, 0.5}, {-1.7373649676713758`,
                     0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-1.753609248306086, 0.5}, 1.6, {
                    1.5642249021436776`, 1.5340171532037907`}], 
                   ArrowBox[{{-1.7107676914411736`, 
                    2.099426334973071}, {-1.6947758367211585`, 
                    2.0989179558943225`}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Dashing[{Small, Small}], 
                   Arrowheads[0.02], 
                   CircleBox[{-1.753609248306086, 0.5}, 1.7, {
                    1.5642249021436776`, 1.573638939881995}], 
                   ArrowBox[{{-1.7414417944408753`, 
                    2.1999564562265816`}, {-1.7584416840461052`, 
                    2.1999931316228363`}}]}, 
                  {RGBColor[0, 0, 1], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-1.753609248306086, 0.5}, 1.8, {
                    1.5642249021436776`, 1.5821625786320879`}], 
                   ArrowBox[{{-1.7560685008479386`, 
                    2.2999983200205874`}, {-1.7740680610883222`, 
                    2.2998837287390375`}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.003], Arrowheads[0.025], 
                   CircleBox[{-1.753609248306086, 0.5}, 1.5, {
                    1.5642249021436776`, 1.5641977578306503`}], 
                   ArrowBox[{{-1.75871138502131, 
                    1.999991322708548}, {-1.7437114666868185`, 
                    1.9999673442842072`}}]}, 
                  {RGBColor[0, 0, 1], ArrowBox[{{0, 2.8}, {0, 2.8}}]}},
                 Frame->True,
                 FrameTicks->Automatic,
                 ImageSize->800,
                 PlotRange->{{-3, 3}, {-1, 3}}], 481->
                GraphicsBox[{InsetBox[
                   TagBox[GridBox[{
                    {"\<\"t\"\>", "24.`"},
                    {"\<\"driveIntent\"\>", "0"},
                    {"\<\"motorTorque\"\>", 
                    RowBox[{"-", "3.210153266482861`"}]},
                    {"\<\"vw\"\>", 
                    RowBox[{"-", "0.03771133393647263`"}]},
                    {"\<\"controllerState\"\>", 
                    RowBox[{"\[LeftAssociation]", 
                    RowBox[{
                    
                    RowBox[{"\<\"lastError\"\>", "\[Rule]", 
                    "0.004781238038563093`"}], ",", 
                    RowBox[{"\<\"lastInt\"\>", "\[Rule]", 
                    RowBox[{"-", "0.0323704735739996`"}]}], ",", 
                    RowBox[{"\<\"initialized\"\>", "\[Rule]", "True"}]}], 
                    "\[RightAssociation]"}]}
                    },
                    AutoDelete->False,
                    
                    GridBoxItemSize->{
                    "Columns" -> {{Automatic}}, "Rows" -> {{Automatic}}}],
                    "Grid"], {-1, -0.5}], 
                  {GrayLevel[0], LineBox[{{-5, 0}, {5, 0}}]}, 
                  {GrayLevel[0.5], 
                   CircleBox[{-1.7555013263520374`, 0.5}, 0.5], 
                   DiskBox[{-1.7555013263520374`, 0.5}, 0.1]}, 
                  {GrayLevel[0.5], 
                   DiskBox[{-1.74570190887926, 1.9999679901308542`}, 0.1]}, 
                  {GrayLevel[0.5], Thickness[0.01], 
                   LineBox[{{-1.7555013263520374`, 0.5}, {-1.74570190887926, 
                    1.9999679901308542`}}]}, 
                  {RGBColor[0, 1, 0], 
                   DiskBox[{-1.748968381370186, 1.499978660087236}, 0.05]}, 
                  {RGBColor[0, 0, 1], Arrowheads[0.01], 
                   CircleBox[{-1.7555013263520374`, 0.5}, 0.3333333333333333, 
                    NCache[{
                    Rational[-1, 2] 
                    Pi, -1.6671009247893824`}, {-1.5707963267948966`, \
-1.6671009247893824`}]], 
                   ArrowBox[{{-1.7842338257379948`, 
                    0.1679073102428374}, {-1.787553260674064, 
                    0.1682112349440874}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-1.7555013263520374`, 
                    0.5}, {-1.6912982610223801`, 0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-1.7555013263520374`, 
                    0.5}, {-1.8035500296547247`, 0.5}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.005], Arrowheads[0.04], 
                   ArrowBox[{{-1.7555013263520374`, 
                    0.5}, {-1.7393469643250674`, 0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-1.7555013263520374`, 0.5}, 1.6, {
                    1.564263335341823, 1.534232254917007}], 
                   ArrowBox[{{-1.7130038098204006`, 
                    2.0994355132635523`}, {-1.6970118461171368`, 
                    2.098930574071886}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Dashing[{Small, Small}], 
                   Arrowheads[0.02], 
                   CircleBox[{-1.7555013263520374`, 0.5}, 1.7, {
                    1.564263335341823, 1.5736252650157958`}], 
                   ArrowBox[{{-1.7433106258108817`, 
                    2.1999562896793305`}, {-1.760310514912991, 
                    2.1999931975467977`}}]}, 
                  {RGBColor[0, 0, 1], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-1.7555013263520374`, 0.5}, 1.8, {
                    1.564263335341823, 1.5820975201556167`}], 
                   ArrowBox[{{-1.7578434737404167`, 
                    2.2999984762064694`}, {-1.7758430413978419`, 
                    2.2998850559491277`}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.003], Arrowheads[0.025], 
                   CircleBox[{-1.7555013263520374`, 0.5}, 1.5, {
                    1.564263335341823, 1.564241391629074}], 
                   ArrowBox[{{-1.760668913381396, 
                    1.99999109865502}, {-1.745668994014914, 
                    1.999967774734115}}]}, 
                  {RGBColor[0, 0, 1], ArrowBox[{{0, 2.8}, {0, 2.8}}]}},
                 Frame->True,
                 FrameTicks->Automatic,
                 ImageSize->800,
                 PlotRange->{{-3, 3}, {-1, 3}}], 482->
                GraphicsBox[{InsetBox[
                   TagBox[GridBox[{
                    {"\<\"t\"\>", "24.05`"},
                    {"\<\"driveIntent\"\>", "0"},
                    {"\<\"motorTorque\"\>", 
                    RowBox[{"-", "3.1917877422054306`"}]},
                    {"\<\"vw\"\>", 
                    RowBox[{"-", "0.03744678211913268`"}]},
                    {"\<\"controllerState\"\>", 
                    RowBox[{"\[LeftAssociation]", 
                    RowBox[{
                    
                    RowBox[{"\<\"lastError\"\>", "\[Rule]", 
                    "0.004739721102220395`"}], ",", 
                    RowBox[{"\<\"lastInt\"\>", "\[Rule]", 
                    RowBox[{"-", "0.032132469998860436`"}]}], ",", 
                    RowBox[{"\<\"initialized\"\>", "\[Rule]", "True"}]}], 
                    "\[RightAssociation]"}]}
                    },
                    AutoDelete->False,
                    
                    GridBoxItemSize->{
                    "Columns" -> {{Automatic}}, "Rows" -> {{Automatic}}}],
                    "Grid"], {-1, -0.5}], 
                  {GrayLevel[0], LineBox[{{-5, 0}, {5, 0}}]}, 
                  {GrayLevel[0.5], 
                   CircleBox[{-1.7573801408919627`, 0.5}, 0.5], 
                   DiskBox[{-1.7573801408919627`, 0.5}, 0.1]}, 
                  {GrayLevel[0.5], 
                   DiskBox[{-1.7476374926009899`, 1.9999683599343947`}, 0.1]}, 
                  {GrayLevel[0.5], Thickness[0.01], 
                   LineBox[{{-1.7573801408919627`, 
                    0.5}, {-1.7476374926009899`, 1.9999683599343947`}}]}, 
                  {RGBColor[0, 1, 0], 
                   DiskBox[{-1.7508850420313142`, 1.4999789066229299`}, 
                    0.05]}, 
                  {RGBColor[0, 0, 1], Arrowheads[0.01], 
                   CircleBox[{-1.7573801408919627`, 0.5}, 0.3333333333333333, 
                    NCache[{
                    Rational[-1, 2] 
                    Pi, -1.6665499590610595`}, {-1.5707963267948966`, \
-1.6665499590610595`}]], 
                   ArrowBox[{{-1.7859296642354296`, 
                    0.16789153002675172`}, {-1.7892492661197557`, 
                    0.16819362578707692`}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-1.7573801408919627`, 0.5}, {-1.693544386047854,
                     0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-1.7573801408919627`, 
                    0.5}, {-1.8051505041637566`, 0.5}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.005], Arrowheads[0.04], 
                   ArrowBox[{{-1.7573801408919627`, 0.5}, {-1.741314749319648,
                     0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-1.7573801408919627`, 0.5}, 1.6, {
                    1.5643011822660062`, 1.5344440754339226`}], 
                   ArrowBox[{{-1.7152214185683725`, 
                    2.0994444792277234`}, {-1.6992293482673924`, 
                    2.0989429274733773`}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Dashing[{Small, Small}], 
                   Arrowheads[0.02], 
                   CircleBox[{-1.7573801408919627`, 0.5}, 1.7, {
                    1.5643011822660062`, 1.5736115532813555`}], 
                   ArrowBox[{{-1.7451661310027495`, 
                    2.199956122363876}, {-1.7621660195971898`, 
                    2.1999932633293047`}}]}, 
                  {RGBColor[0, 0, 1], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-1.7573801408919627`, 0.5}, 1.8, {
                    1.5643011822660062`, 1.5820333363893697`}], 
                   ArrowBox[{{-1.7596067575941554`, 
                    2.2999986228267124`}, {-1.7776063324942446`, 
                    2.2998863578496467`}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.003], Arrowheads[0.025], 
                   CircleBox[{-1.7573801408919627`, 0.5}, 1.5, {
                    1.5643011822660062`, 1.5642839404105489`}], 
                   ArrowBox[{{-1.7626115507100952`, 
                    1.9999908774226978`}, {-1.7476116303647866`, 
                    1.9999681917301049`}}]}, 
                  {RGBColor[0, 0, 1], ArrowBox[{{0, 2.8}, {0, 2.8}}]}},
                 Frame->True,
                 FrameTicks->Automatic,
                 ImageSize->800,
                 PlotRange->{{-3, 3}, {-1, 3}}], 483->
                GraphicsBox[{InsetBox[
                   TagBox[GridBox[{
                    {"\<\"t\"\>", "24.1`"},
                    {"\<\"driveIntent\"\>", "0"},
                    {"\<\"motorTorque\"\>", 
                    RowBox[{"-", "3.1736241995797614`"}]},
                    {"\<\"vw\"\>", 
                    RowBox[{"-", "0.03718368575254961`"}]},
                    {"\<\"controllerState\"\>", 
                    RowBox[{"\[LeftAssociation]", 
                    RowBox[{
                    
                    RowBox[{"\<\"lastError\"\>", "\[Rule]", 
                    "0.004698177623195366`"}], ",", 
                    RowBox[{"\<\"lastInt\"\>", "\[Rule]", 
                    RowBox[{"-", "0.03189654343001859`"}]}], ",", 
                    RowBox[{"\<\"initialized\"\>", "\[Rule]", "True"}]}], 
                    "\[RightAssociation]"}]}
                    },
                    AutoDelete->False,
                    
                    GridBoxItemSize->{
                    "Columns" -> {{Automatic}}, "Rows" -> {{Automatic}}}],
                    "Grid"], {-1, -0.5}], 
                  {GrayLevel[0], LineBox[{{-5, 0}, {5, 0}}]}, 
                  {GrayLevel[0.5], 
                   CircleBox[{-1.7592457649996704`, 0.5}, 0.5], 
                   DiskBox[{-1.7592457649996704`, 0.5}, 0.1]}, 
                  {GrayLevel[0.5], 
                   DiskBox[{-1.7495591946250497`, 1.9999687231253782`}, 0.1]}, 
                  {GrayLevel[0.5], Thickness[0.01], 
                   LineBox[{{-1.7592457649996704`, 
                    0.5}, {-1.7495591946250497`, 1.9999687231253782`}}]}, 
                  {RGBColor[0, 1, 0], 
                   DiskBox[{-1.7527880514165899`, 1.4999791487502523`}, 
                    0.05]}, 
                  {RGBColor[0, 0, 1], Arrowheads[0.01], 
                   CircleBox[{-1.7592457649996704`, 0.5}, 0.3333333333333333, 
                    NCache[{
                    Rational[-1, 2] 
                    Pi, -1.6660050527822894`}, {-1.5707963267948966`, \
-1.6660050527822894`}]], 
                   ArrowBox[{{-1.7876143161230689`, 
                    0.1678760225182918}, {-1.7909340821284305`, 
                    0.1681763093619471}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-1.7592457649996704`, 
                    0.5}, {-1.6957732810080752`, 0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-1.7592457649996704`, 
                    0.5}, {-1.8067411773224538`, 0.5}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.005], Arrowheads[0.04], 
                   ArrowBox[{{-1.7592457649996704`, 
                    0.5}, {-1.7432686933308585`, 0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-1.7592457649996704`, 0.5}, 1.6, {
                    1.564338568327642, 1.534653316651306}], 
                   ArrowBox[{{-1.7174217133065035`, 
                    2.0994532655566926`}, {-1.701429538410306, 
                    2.09895506001356}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Dashing[{Small, Small}], 
                   Arrowheads[0.02], 
                   CircleBox[{-1.7592457649996704`, 0.5}, 1.7, {
                    1.564338568327642, 1.5735977576304438`}], 
                   ArrowBox[{{-1.7470083031103911`, 
                    2.199955953701892}, {-1.7640081911908299`, 
                    2.1999933291918454`}}]}, 
                  {RGBColor[0, 0, 1], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-1.7592457649996704`, 0.5}, 1.8, {
                    1.564338568327642, 1.5819698138808629`}], 
                   ArrowBox[{{-1.7613580412696168`, 
                    2.2999987606353955`}, {-1.7793576232647441`, 
                    2.2998876390367053`}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.003], Arrowheads[0.025], 
                   CircleBox[{-1.7592457649996704`, 0.5}, 1.5, {
                    1.564338568327642, 1.5643255424132574`}], 
                   ArrowBox[{{-1.7645395774378032`, 
                    1.9999906584875353`}, {-1.7495396561617047`, 
                    1.9999685968216885`}}]}, 
                  {RGBColor[0, 0, 1], ArrowBox[{{0, 2.8}, {0, 2.8}}]}},
                 Frame->True,
                 FrameTicks->Automatic,
                 ImageSize->800,
                 PlotRange->{{-3, 3}, {-1, 3}}], 484->
                GraphicsBox[{InsetBox[
                   TagBox[GridBox[{
                    {"\<\"t\"\>", "24.150000000000002`"},
                    {"\<\"driveIntent\"\>", "0"},
                    {"\<\"motorTorque\"\>", 
                    RowBox[{"-", "3.1556081694333855`"}]},
                    {"\<\"vw\"\>", 
                    RowBox[{"-", "0.0369220363093847`"}]},
                    {"\<\"controllerState\"\>", 
                    RowBox[{"\[LeftAssociation]", 
                    RowBox[{
                    
                    RowBox[{"\<\"lastError\"\>", "\[Rule]", 
                    "0.004656717733190063`"}], ",", 
                    RowBox[{"\<\"lastInt\"\>", "\[Rule]", 
                    RowBox[{"-", "0.03166269233850515`"}]}], ",", 
                    RowBox[{"\<\"initialized\"\>", "\[Rule]", "True"}]}], 
                    "\[RightAssociation]"}]}
                    },
                    AutoDelete->False,
                    
                    GridBoxItemSize->{
                    "Columns" -> {{Automatic}}, "Rows" -> {{Automatic}}}],
                    "Grid"], {-1, -0.5}], 
                  {GrayLevel[0], LineBox[{{-5, 0}, {5, 0}}]}, 
                  {GrayLevel[0.5], 
                   CircleBox[{-1.7610982712122878`, 0.5}, 0.5], 
                   DiskBox[{-1.7610982712122878`, 0.5}, 0.1]}, 
                  {GrayLevel[0.5], 
                   DiskBox[{-1.7514672561755849`, 1.9999690808644566`}, 0.1]}, 
                  {GrayLevel[0.5], Thickness[0.01], 
                   LineBox[{{-1.7610982712122878`, 
                    0.5}, {-1.7514672561755849`, 1.9999690808644566`}}]}, 
                  {RGBColor[0, 1, 0], 
                   DiskBox[{-1.7546775945211526`, 1.4999793872429712`}, 
                    0.05]}, 
                  {RGBColor[0, 0, 1], Arrowheads[0.01], 
                   CircleBox[{-1.7610982712122878`, 0.5}, 0.3333333333333333, 
                    NCache[{
                    Rational[-1, 2] 
                    Pi, -1.6654645718778982`}, {-1.5707963267948966`, \
-1.6654645718778982`}]], 
                   ArrowBox[{{-1.7892873115332015`, 
                    0.1678607383688327}, {-1.7926072393529757`, 
                    0.16815923089858265`}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-1.7610982712122878`, 0.5}, {-1.69798610782362, 
                    0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-1.7610982712122878`, 0.5}, {-1.808321294600357,
                     0.5}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.005], Arrowheads[0.04], 
                   ArrowBox[{{-1.7610982712122878`, 
                    0.5}, {-1.7452091312116893`, 0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-1.7610982712122878`, 0.5}, 1.6, {
                    1.564375605987448, 1.5348606079844223`}], 
                   ArrowBox[{{-1.7196057732150039`, 
                    2.099461900956052}, {-1.7036134953887054`, 
                    2.098967010462853}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Dashing[{Small, Small}], 
                   Arrowheads[0.02], 
                   CircleBox[{-1.7610982712122878`, 0.5}, 1.7, {
                    1.564375605987448, 1.5735838385121346`}], 
                   ArrowBox[{{-1.7488371474361546`, 
                    2.199955783202536}, {-1.7658370349947126`, 
                    2.199993395315939}}]}, 
                  {RGBColor[0, 0, 1], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-1.7610982712122878`, 0.5}, 1.8, {
                    1.564375605987448, 1.5819067624843002`}], 
                   ArrowBox[{{-1.7630970550424416`, 
                    2.2999988902394355`}, {-1.7810966440081621`, 
                    2.299888903539748}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.003], Arrowheads[0.025], 
                   CircleBox[{-1.7610982712122878`, 0.5}, 1.5, {
                    1.564375605987448, 1.564366326108421}], 
                   ArrowBox[{{-1.766453258807765, 
                    1.999990441338828}, {-1.7514533366443852`, 
                    1.9999689914252166`}}]}, 
                  {RGBColor[0, 0, 1], ArrowBox[{{0, 2.8}, {0, 2.8}}]}},
                 Frame->True,
                 FrameTicks->Automatic,
                 ImageSize->800,
                 PlotRange->{{-3, 3}, {-1, 3}}], 485->
                GraphicsBox[{InsetBox[
                   TagBox[GridBox[{
                    {"\<\"t\"\>", "24.2`"},
                    {"\<\"driveIntent\"\>", "0"},
                    {"\<\"motorTorque\"\>", 
                    RowBox[{"-", "3.137691473503328`"}]},
                    {"\<\"vw\"\>", 
                    RowBox[{"-", "0.036661829296574903`"}]},
                    {"\<\"controllerState\"\>", 
                    RowBox[{"\[LeftAssociation]", 
                    RowBox[{
                    
                    RowBox[{"\<\"lastError\"\>", "\[Rule]", 
                    "0.004615440243355098`"}], ",", 
                    RowBox[{"\<\"lastInt\"\>", "\[Rule]", 
                    RowBox[{"-", "0.03143090997867339`"}]}], ",", 
                    RowBox[{"\<\"initialized\"\>", "\[Rule]", "True"}]}], 
                    "\[RightAssociation]"}]}
                    },
                    AutoDelete->False,
                    
                    GridBoxItemSize->{
                    "Columns" -> {{Automatic}}, "Rows" -> {{Automatic}}}],
                    "Grid"], {-1, -0.5}], 
                  {GrayLevel[0], LineBox[{{-5, 0}, {5, 0}}]}, 
                  {GrayLevel[0.5], 
                   CircleBox[{-1.7629377317457566`, 0.5}, 0.5], 
                   DiskBox[{-1.7629377317457566`, 0.5}, 0.1]}, 
                  {GrayLevel[0.5], 
                   DiskBox[{-1.7533618993687579`, 1.9999694341666725`}, 0.1]}, 
                  {GrayLevel[0.5], Thickness[0.01], 
                   LineBox[{{-1.7629377317457566`, 
                    0.5}, {-1.7533618993687579`, 1.9999694341666725`}}]}, 
                  {RGBColor[0, 1, 0], 
                   DiskBox[{-1.7565538434944241`, 1.4999796227777815`}, 
                    0.05]}, 
                  {RGBColor[0, 0, 1], Arrowheads[0.01], 
                   CircleBox[{-1.7629377317457566`, 0.5}, 0.3333333333333333, 
                    NCache[{
                    Rational[-1, 2] 
                    Pi, -1.6649270709999964`}, {-1.5707963267948966`, \
-1.6649270709999964`}]], 
                   ArrowBox[{{-1.790948242858546, 
                    0.16784563471435243`}, {-1.7942683306387337`, 
                    0.16814234273695222`}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-1.7629377317457566`, 
                    0.5}, {-1.7001839022756902`, 0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-1.7629377317457566`, 0.5}, {-1.809890193255466,
                     0.5}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.005], Arrowheads[0.04], 
                   ArrowBox[{{-1.7629377317457566`, 
                    0.5}, {-1.7471363637853996`, 0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-1.7629377317457566`, 0.5}, 1.6, {
                    1.5644123951812408`, 1.5350665087500233`}], 
                   ArrowBox[{{-1.7217745650556422`, 
                    2.0994704103884017`}, {-1.7057821856700102`, 
                    2.0989788127279185`}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Dashing[{Small, Small}], 
                   Arrowheads[0.02], 
                   CircleBox[{-1.7629377317457566`, 0.5}, 1.7, {
                    1.5644123951812408`, 1.5735697633844916`}], 
                   ArrowBox[{{-1.7506526808762026`, 
                    2.199955610457265}, {-1.7676525679036816`, 
                    2.199993461846252}}]}, 
                  {RGBColor[0, 0, 1], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-1.7629377317457566`, 0.5}, 1.8, {
                    1.5644123951812408`, 1.5818440144784813`}], 
                   ArrowBox[{{-1.7648235692312109`, 
                    2.2999990121155562`}, {-1.7828231650629427`, 
                    2.2998901548543977`}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.003], Arrowheads[0.025], 
                   CircleBox[{-1.7629377317457566`, 0.5}, 1.5, {
                    1.5644123951812408`, 1.5644064101083153`}], 
                   ArrowBox[{{-1.7683528449536077`, 
                    1.9999902254844684`}, {-1.75335292194248, 
                    1.9999693768277522`}}]}, 
                  {RGBColor[0, 0, 1], ArrowBox[{{0, 2.8}, {0, 2.8}}]}},
                 Frame->True,
                 FrameTicks->Automatic,
                 ImageSize->800,
                 PlotRange->{{-3, 3}, {-1, 3}}], 486->
                GraphicsBox[{InsetBox[
                   TagBox[GridBox[{
                    {"\<\"t\"\>", "24.25`"},
                    {"\<\"driveIntent\"\>", "0"},
                    {"\<\"motorTorque\"\>", 
                    RowBox[{"-", "3.119831955735885`"}]},
                    {"\<\"vw\"\>", 
                    RowBox[{"-", "0.03640306371542718`"}]},
                    {"\<\"controllerState\"\>", 
                    RowBox[{"\[LeftAssociation]", 
                    RowBox[{
                    
                    RowBox[{"\<\"lastError\"\>", "\[Rule]", 
                    "0.004574432977818518`"}], ",", 
                    RowBox[{"\<\"lastInt\"\>", "\[Rule]", 
                    RowBox[{"-", "0.031201184946026316`"}]}], ",", 
                    RowBox[{"\<\"initialized\"\>", "\[Rule]", "True"}]}], 
                    "\[RightAssociation]"}]}
                    },
                    AutoDelete->False,
                    
                    GridBoxItemSize->{
                    "Columns" -> {{Automatic}}, "Rows" -> {{Automatic}}}],
                    "Grid"], {-1, -0.5}], 
                  {GrayLevel[0], LineBox[{{-5, 0}, {5, 0}}]}, 
                  {GrayLevel[0.5], CircleBox[{-1.76476421868266, 0.5}, 0.5], 
                   DiskBox[{-1.76476421868266, 0.5}, 0.1]}, 
                  {GrayLevel[0.5], 
                   DiskBox[{-1.755243328145976, 1.9999697839101254`}, 0.1]}, 
                  {GrayLevel[0.5], Thickness[0.01], 
                   LineBox[{{-1.76476421868266, 0.5}, {-1.755243328145976, 
                    1.9999697839101254`}}]}, 
                  {RGBColor[0, 1, 0], 
                   DiskBox[{-1.7584169583248705`, 1.4999798559400837`}, 
                    0.05]}, 
                  {RGBColor[0, 0, 1], Arrowheads[0.01], 
                   CircleBox[{-1.76476421868266, 0.5}, 0.3333333333333333, 
                    NCache[{
                    Rational[-1, 2] 
                    Pi, -1.6643912854669731`}, {-1.5707963267948966`, \
-1.6643912854669731`}]], 
                   ArrowBox[{{-1.7925967622798789`, 
                    0.16783067476357783`}, {-1.7959170085553826`, 
                    0.16812560388867437`}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-1.76476421868266, 0.5}, {-1.7023675795679423`, 
                    0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-1.76476421868266, 0.5}, {-1.8114472988677381`, 
                    0.5}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.005], Arrowheads[0.04], 
                   ArrowBox[{{-1.76476421868266, 0.5}, {-1.7490506597530204`, 
                    0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-1.76476421868266, 0.5}, 1.6, {
                    1.5644490238168995`, 1.5352715109476582`}], 
                   ArrowBox[{{-1.72392894780439, 
                    2.099478815318383}, {-1.7079364679762052`, 
                    2.098990496141126}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Dashing[{Small, Small}], 
                   Arrowheads[0.02], 
                   CircleBox[{-1.76476421868266, 0.5}, 1.7, {
                    1.5644490238168995`, 1.5735555062150375`}], 
                   ArrowBox[{{-1.7524549312591526`, 
                    2.1999554351344406`}, {-1.76945481774525, 
                    2.199993528893694}}]}, 
                  {RGBColor[0, 0, 1], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-1.76476421868266, 0.5}, 1.8, {
                    1.5644490238168995`, 1.5817814235709877`}], 
                   ArrowBox[{{-1.766537392592838, 
                    2.299999126625978}, {-1.7845369952027865`, 
                    2.29989139597607}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.003], Arrowheads[0.025], 
                   CircleBox[{-1.76476421868266, 0.5}, 1.5, {
                    1.5644490238168995`, 1.564445903179171}], 
                   ArrowBox[{{-1.7702385711065307`, 
                    1.9999900104552495`}, {-1.7552386472837231`, 
                    1.9999697541915717`}}]}, 
                  {RGBColor[0, 0, 1], ArrowBox[{{0, 2.8}, {0, 2.8}}]}},
                 Frame->True,
                 FrameTicks->Automatic,
                 ImageSize->800,
                 PlotRange->{{-3, 3}, {-1, 3}}], 487->
                GraphicsBox[{InsetBox[
                   TagBox[GridBox[{
                    {"\<\"t\"\>", "24.3`"},
                    {"\<\"driveIntent\"\>", "0"},
                    {"\<\"motorTorque\"\>", 
                    RowBox[{"-", "3.1019931866889427`"}]},
                    {"\<\"vw\"\>", 
                    RowBox[{"-", "0.036145741550879226`"}]},
                    {"\<\"controllerState\"\>", 
                    RowBox[{"\[LeftAssociation]", 
                    RowBox[{
                    
                    RowBox[{"\<\"lastError\"\>", "\[Rule]", 
                    "0.004533773174463432`"}], ",", 
                    RowBox[{"\<\"lastInt\"\>", "\[Rule]", 
                    RowBox[{"-", "0.030973501716603628`"}]}], ",", 
                    RowBox[{"\<\"initialized\"\>", "\[Rule]", "True"}]}], 
                    "\[RightAssociation]"}]}
                    },
                    AutoDelete->False,
                    GridBoxItemSize->{
                    "Columns" -> {{Automatic}}, "Rows" -> {{Automatic}}}],
                    "Grid"], {-1, -0.5}], 
                  {GrayLevel[0], LineBox[{{-5, 0}, {5, 0}}]}, 
                  {GrayLevel[0.5], 
                   CircleBox[{-1.7665778041337945`, 0.5}, 0.5], 
                   DiskBox[{-1.7665778041337945`, 0.5}, 0.1]}, 
                  {GrayLevel[0.5], 
                   DiskBox[{-1.7571117292682599`, 1.9999701308448246`}, 0.1]}, 
                  {GrayLevel[0.5], Thickness[0.01], 
                   LineBox[{{-1.7665778041337945`, 
                    0.5}, {-1.7571117292682599`, 1.9999701308448246`}}]}, 
                  {RGBColor[0, 1, 0], 
                   DiskBox[{-1.7602670875567714`, 1.4999800872298832`}, 
                    0.05]}, 
                  {RGBColor[0, 0, 1], Arrowheads[0.01], 
                   CircleBox[{-1.7665778041337945`, 0.5}, 0.3333333333333333, 
                    NCache[{
                    Rational[-1, 2] 
                    Pi, -1.6638561223955648`}, {-1.5707963267948966`, \
-1.6638561223955648`}]], 
                   ArrowBox[{{-1.7942325789975642`, 
                    0.1678158273813376}, {-1.7975529826327783`, 
                    0.1681089795910905}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-1.7665778041337945`, 
                    0.5}, {-1.7045379404000156`, 0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-1.7665778041337945`, 0.5}, {-1.812992121417807,
                     0.5}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.005], Arrowheads[0.04], 
                   ArrowBox[{{-1.7665778041337945`, 
                    0.5}, {-1.7509522576840282`, 0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-1.7665778041337945`, 0.5}, 1.6, {
                    1.5644855683295904`, 1.5354760423670608`}], 
                   ArrowBox[{{-1.7260696777797782`, 
                    2.0994871339586596`}, {-1.710077098409487, 
                    2.099002085756193}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Dashing[{Small, Small}], 
                   Arrowheads[0.02], 
                   CircleBox[{-1.7665778041337945`, 0.5}, 1.7, {
                    1.5644855683295904`, 1.5735410469760625`}], 
                   ArrowBox[{{-1.7542439366496914`, 
                    2.1999552569738077`}, {-1.771243822583205, 
                    2.1999935965384783`}}]}, 
                  {RGBColor[0, 0, 1], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-1.7665778041337945`, 0.5}, 1.8, {
                    1.5644855683295904`, 1.5817188638111956`}], 
                   ArrowBox[{{-1.7682383705275884`, 
                    2.2999992340329625`}, {-1.786237979841918, 
                    2.29989262943408}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.003], Arrowheads[0.025], 
                   CircleBox[{-1.7665778041337945`, 0.5}, 1.5, {
                    1.5644855683295904`, 1.56448490434682}], 
                   ArrowBox[{{-1.7721106579153565`, 
                    1.9999897958083022`}, {-1.757110733313939, 
                    1.9999701245591832`}}]}, 
                  {RGBColor[0, 0, 1], ArrowBox[{{0, 2.8}, {0, 2.8}}]}},
                 Frame->True,
                 FrameTicks->Automatic,
                 ImageSize->800,
                 PlotRange->{{-3, 3}, {-1, 3}}], 488->
                GraphicsBox[{InsetBox[
                   TagBox[GridBox[{
                    {"\<\"t\"\>", "24.35`"},
                    {"\<\"driveIntent\"\>", "0"},
                    {"\<\"motorTorque\"\>", 
                    RowBox[{"-", "3.084144146947203`"}]},
                    {"\<\"vw\"\>", 
                    RowBox[{"-", "0.03588986729147762`"}]},
                    {"\<\"controllerState\"\>", 
                    RowBox[{"\[LeftAssociation]", 
                    RowBox[{
                    
                    RowBox[{"\<\"lastError\"\>", "\[Rule]", 
                    "0.004493527941415969`"}], ",", 
                    RowBox[{"\<\"lastInt\"\>", "\[Rule]", 
                    RowBox[{"-", "0.030747841164862462`"}]}], ",", 
                    RowBox[{"\<\"initialized\"\>", "\[Rule]", "True"}]}], 
                    "\[RightAssociation]"}]}
                    },
                    AutoDelete->False,
                    
                    GridBoxItemSize->{
                    "Columns" -> {{Automatic}}, "Rows" -> {{Automatic}}}],
                    "Grid"], {-1, -0.5}], 
                  {GrayLevel[0], LineBox[{{-5, 0}, {5, 0}}]}, 
                  {GrayLevel[0.5], 
                   CircleBox[{-1.7683785603749762`, 0.5}, 0.5], 
                   DiskBox[{-1.7683785603749762`, 0.5}, 0.1]}, 
                  {GrayLevel[0.5], 
                   DiskBox[{-1.758967273355253, 1.9999704756016474`}, 0.1]}, 
                  {GrayLevel[0.5], Thickness[0.01], 
                   LineBox[{{-1.7683785603749762`, 0.5}, {-1.758967273355253, 
                    1.9999704756016474`}}]}, 
                  {RGBColor[0, 1, 0], 
                   DiskBox[{-1.762104369028494, 1.4999803170677648`}, 0.05]}, 
                  {RGBColor[0, 0, 1], Arrowheads[0.01], 
                   CircleBox[{-1.7683785603749762`, 0.5}, 0.3333333333333333, 
                    NCache[{
                    Rational[-1, 2] 
                    Pi, -1.6633206512033127`}, {-1.5707963267948966`, \
-1.6633206512033127`}]], 
                   ArrowBox[{{-1.7958554562275684`, 
                    0.16780106667026018`}, {-1.7991760163613097`, 
                    0.16809244085757713`}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-1.7683785603749762`, 
                    0.5}, {-1.7066956774360322`, 0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-1.7683785603749762`, 
                    0.5}, {-1.8145242510005803`, 0.5}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.005], Arrowheads[0.04], 
                   ArrowBox[{{-1.7683785603749762`, 
                    0.5}, {-1.7528413680616364`, 0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-1.7683785603749762`, 0.5}, 1.6, {
                    1.5645220942832627`, 1.5356804699545228`}], 
                   ArrowBox[{{-1.7281974141610843`, 
                    2.099495381515351}, {-1.71220473596773, 
                    2.099013602647415}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Dashing[{Small, Small}], 
                   Arrowheads[0.02], 
                   CircleBox[{-1.7683785603749762`, 0.5}, 1.7, {
                    1.5645220942832627`, 1.5735263711401932`}], 
                   ArrowBox[{{-1.7560197446278658`, 
                    2.1999550757809248`}, {-1.7730196299968835`, 
                    2.1999936648331264`}}]}, 
                  {RGBColor[0, 0, 1], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-1.7683785603749762`, 0.5}, 1.8, {
                    1.5645220942832627`, 1.5816562284329694`}], 
                   ArrowBox[{{-1.7699263831327559`, 
                    2.2999993345122967`}, {-1.7879259990889969`, 
                    2.2998938573259595`}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.003], Arrowheads[0.025], 
                   CircleBox[{-1.7683785603749762`, 0.5}, 1.5, {
                    1.5645220942832627`, 1.5645235030834428`}], 
                   ArrowBox[{{-1.7439704020739548`, 
                    1.9998014007889018`}, {-1.7589693865139384`, 
                    1.9999704888587817`}}]}, 
                  {RGBColor[0, 0, 1], ArrowBox[{{0, 2.8}, {0, 2.8}}]}},
                 Frame->True,
                 FrameTicks->Automatic,
                 ImageSize->800,
                 PlotRange->{{-3, 3}, {-1, 3}}], 489->
                GraphicsBox[{InsetBox[
                   TagBox[GridBox[{
                    {"\<\"t\"\>", "24.400000000000002`"},
                    {"\<\"driveIntent\"\>", "0"},
                    {"\<\"motorTorque\"\>", 
                    RowBox[{"-", "3.0662588949867295`"}]},
                    {"\<\"vw\"\>", 
                    RowBox[{"-", "0.035635447481153146`"}]},
                    {"\<\"controllerState\"\>", 
                    RowBox[{"\[LeftAssociation]", 
                    RowBox[{
                    
                    RowBox[{"\<\"lastError\"\>", "\[Rule]", 
                    "0.004453754758491838`"}], ",", 
                    RowBox[{"\<\"lastInt\"\>", "\[Rule]", 
                    RowBox[{"-", "0.030524181057544255`"}]}], ",", 
                    RowBox[{"\<\"initialized\"\>", "\[Rule]", "True"}]}], 
                    "\[RightAssociation]"}]}
                    },
                    AutoDelete->False,
                    
                    GridBoxItemSize->{
                    "Columns" -> {{Automatic}}, "Rows" -> {{Automatic}}}],
                    "Grid"], {-1, -0.5}], 
                  {GrayLevel[0], LineBox[{{-5, 0}, {5, 0}}]}, 
                  {GrayLevel[0.5], 
                   CircleBox[{-1.7701665599606642`, 0.5}, 0.5], 
                   DiskBox[{-1.7701665599606642`, 0.5}, 0.1]}, 
                  {GrayLevel[0.5], 
                   DiskBox[{-1.7608101159538696`, 1.9999708187013332`}, 0.1]}, 
                  {GrayLevel[0.5], Thickness[0.01], 
                   LineBox[{{-1.7701665599606642`, 
                    0.5}, {-1.7608101159538696`, 1.9999708187013332`}}]}, 
                  {RGBColor[0, 1, 0], 
                   DiskBox[{-1.7639289306228012`, 1.4999805458008888`}, 
                    0.05]}, 
                  {RGBColor[0, 0, 1], Arrowheads[0.01], 
                   CircleBox[{-1.7701665599606642`, 0.5}, 0.3333333333333333, 
                    NCache[{
                    Rational[-1, 2] 
                    Pi, -1.6627840936444984`}, {-1.5707963267948966`, \
-1.6627840936444984`}]], 
                   ArrowBox[{{-1.797465208017884, 
                    0.16778637155384618`}, {-1.8007859240126558`, 
                    0.1680759640276669}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-1.7701665599606642`, 
                    0.5}, {-1.7088413820609296`, 0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-1.7701665599606642`, 
                    0.5}, {-1.8160433532566453`, 0.5}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.005], Arrowheads[0.04], 
                   ArrowBox[{{-1.7701665599606642`, 
                    0.5}, {-1.7547181753569108`, 0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-1.7701665599606642`, 0.5}, 1.6, {
                    1.5645586570073577`, 1.5358851033763479`}], 
                   ArrowBox[{{-1.7303127247988903`, 
                    2.0995035704314318`}, {-1.714319948352324, 
                    2.0990250642100228`}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Dashing[{Small, Small}], 
                   Arrowheads[0.02], 
                   CircleBox[{-1.7701665599606642`, 0.5}, 1.7, {
                    1.5645586570073577`, 1.5735114691810346`}], 
                   ArrowBox[{{-1.757782411553816, 
                    2.199954891421604}, {-1.7747822963458937`, 
                    2.1999937338053988`}}]}, 
                  {RGBColor[0, 0, 1], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-1.7701665599606642`, 0.5}, 1.8, {
                    1.5645586570073577`, 1.5815934286461728`}], 
                   ArrowBox[{{-1.7716013431410247`, 
                    2.2999994281658047`}, {-1.7896009656857168`, 
                    2.29989508135172}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.003], Arrowheads[0.025], 
                   CircleBox[{-1.7701665599606642`, 0.5}, 1.5, {
                    1.5645586570073577`, 1.5645617795643592`}], 
                   ArrowBox[{{-1.7458158087972044`, 
                    1.999802333948636}, {-1.760814799698297, 
                    1.9999708479100504`}}]}, 
                  {RGBColor[0, 0, 1], ArrowBox[{{0, 2.8}, {0, 2.8}}]}},
                 Frame->True,
                 FrameTicks->Automatic,
                 ImageSize->800,
                 PlotRange->{{-3, 3}, {-1, 3}}], 490->
                GraphicsBox[{InsetBox[
                   TagBox[GridBox[{
                    {"\<\"t\"\>", "24.45`"},
                    {"\<\"driveIntent\"\>", "0"},
                    {"\<\"motorTorque\"\>", 
                    RowBox[{"-", "3.0483162244530853`"}]},
                    {"\<\"vw\"\>", 
                    RowBox[{"-", "0.035382490303443856`"}]},
                    {"\<\"controllerState\"\>", 
                    RowBox[{"\[LeftAssociation]", 
                    RowBox[{
                    
                    RowBox[{"\<\"lastError\"\>", "\[Rule]", 
                    "0.004414502013648744`"}], ",", 
                    RowBox[{"\<\"lastInt\"\>", "\[Rule]", 
                    RowBox[{"-", "0.030302496521536926`"}]}], ",", 
                    RowBox[{"\<\"initialized\"\>", "\[Rule]", "True"}]}], 
                    "\[RightAssociation]"}]}
                    },
                    AutoDelete->False,
                    
                    GridBoxItemSize->{
                    "Columns" -> {{Automatic}}, "Rows" -> {{Automatic}}}],
                    "Grid"], {-1, -0.5}], 
                  {GrayLevel[0], LineBox[{{-5, 0}, {5, 0}}]}, 
                  {GrayLevel[0.5], CircleBox[{-1.771941875815988, 0.5}, 0.5], 
                   DiskBox[{-1.771941875815988, 0.5}, 0.1]}, 
                  {GrayLevel[0.5], 
                   DiskBox[{-1.7626403986229688`, 1.999971160563438}, 0.1]}, 
                  {GrayLevel[0.5], Thickness[0.01], 
                   LineBox[{{-1.771941875815988, 0.5}, {-1.7626403986229688`, 
                    1.999971160563438}}]}, 
                  {RGBColor[0, 1, 0], 
                   DiskBox[{-1.7657408910206418`, 1.4999807737089588`}, 
                    0.05]}, 
                  {RGBColor[0, 0, 1], Arrowheads[0.01], 
                   CircleBox[{-1.771941875815988, 0.5}, 0.3333333333333333, 
                    NCache[{
                    Rational[-1, 2] 
                    Pi, -1.6622458135284892`}, {-1.5707963267948966`, \
-1.6622458135284892`}]], 
                   ArrowBox[{{-1.7990616959365535`, 
                    0.16777172536381063`}, {-1.802382567332107, 
                    0.16805953032037269`}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-1.771941875815988, 0.5}, {-1.7109755513269262`,
                     0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-1.771941875815988, 0.5}, {-1.8175491645975819`,
                     0.5}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.005], Arrowheads[0.04], 
                   ArrowBox[{{-1.771941875815988, 0.5}, {-1.7565828401085202`,
                     0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-1.771941875815988, 0.5}, 1.6, {
                    1.5645953022585988`, 1.5360901987226856`}], 
                   ArrowBox[{{-1.732416092228868, 
                    2.0995117106266603`}, {-1.716423217979264, 
                    2.099036484459316}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Dashing[{Small, Small}], 
                   Arrowheads[0.02], 
                   CircleBox[{-1.771941875815988, 0.5}, 1.7, {
                    1.5645953022585988`, 1.5734963360831022`}], 
                   ArrowBox[{{-1.7595320018267064`, 
                    2.1999547038164193`}, {-1.776531886029032, 
                    2.1999938034611315`}}]}, 
                  {RGBColor[0, 0, 1], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-1.771941875815988, 0.5}, 1.8, {
                    1.5645953022585988`, 1.581530392394449}], 
                   ArrowBox[{{-1.7732631937765164`, 
                    2.2999995150329475`}, {-1.7912628228630791`, 
                    2.2998963028478068`}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.003], Arrowheads[0.025], 
                   CircleBox[{-1.771941875815988, 0.5}, 1.5, {
                    1.5645953022585988`, 1.564599804984416}], 
                   ArrowBox[{{-1.74764815528387, 
                    1.9998032588118706`}, {-1.762647152581933, 
                    1.999971202430234}}]}, 
                  {RGBColor[0, 0, 1], ArrowBox[{{0, 2.8}, {0, 2.8}}]}},
                 Frame->True,
                 FrameTicks->Automatic,
                 ImageSize->800,
                 PlotRange->{{-3, 3}, {-1, 3}}], 491->
                GraphicsBox[{InsetBox[
                   TagBox[GridBox[{
                    {"\<\"t\"\>", "24.5`"},
                    {"\<\"driveIntent\"\>", "0"},
                    {"\<\"motorTorque\"\>", 
                    RowBox[{"-", "3.0302993153459146`"}]},
                    {"\<\"vw\"\>", 
                    RowBox[{"-", "0.035131005198431875`"}]},
                    {"\<\"controllerState\"\>", 
                    RowBox[{"\[LeftAssociation]", 
                    RowBox[{
                    
                    RowBox[{"\<\"lastError\"\>", "\[Rule]", 
                    "0.004375809565295931`"}], ",", 
                    RowBox[{"\<\"lastInt\"\>", "\[Rule]", 
                    RowBox[{"-", "0.030082760484218516`"}]}], ",", 
                    RowBox[{"\<\"initialized\"\>", "\[Rule]", "True"}]}], 
                    "\[RightAssociation]"}]}
                    },
                    AutoDelete->False,
                    
                    GridBoxItemSize->{
                    "Columns" -> {{Automatic}}, "Rows" -> {{Automatic}}}],
                    "Grid"], {-1, -0.5}], 
                  {GrayLevel[0], LineBox[{{-5, 0}, {5, 0}}]}, 
                  {GrayLevel[0.5], 
                   CircleBox[{-1.7737045813088195`, 0.5}, 0.5], 
                   DiskBox[{-1.7737045813088195`, 0.5}, 0.1]}, 
                  {GrayLevel[0.5], 
                   DiskBox[{-1.7644582500218766`, 1.9999715015151895`}, 0.1]}, 
                  {GrayLevel[0.5], Thickness[0.01], 
                   LineBox[{{-1.7737045813088195`, 
                    0.5}, {-1.7644582500218766`, 1.9999715015151895`}}]}, 
                  {RGBColor[0, 1, 0], 
                   DiskBox[{-1.7675403604508577`, 1.4999810010101262`}, 
                    0.05]}, 
                  {RGBColor[0, 0, 1], Arrowheads[0.01], 
                   CircleBox[{-1.7737045813088195`, 0.5}, 0.3333333333333333, 
                    NCache[{
                    Rational[-1, 2] 
                    Pi, -1.661705306255274}, {-1.5707963267948966`, \
-1.661705306255274}]], 
                   ArrowBox[{{-1.800644825677818, 
                    0.1677571154344314}, {-1.8039658521489432`, 
                    0.16804312539389732`}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-1.7737045813088195`, 0.5}, {-1.713098595001901,
                     0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-1.7737045813088195`, 
                    0.5}, {-1.8190414872945853`, 0.5}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.005], Arrowheads[0.04], 
                   ArrowBox[{{-1.7737045813088195`, 
                    0.5}, {-1.7584355009876669`, 0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-1.7737045813088195`, 0.5}, 1.6, {
                    1.5646320668986482`, 1.5362959623001802`}], 
                   ArrowBox[{{-1.7345079198079343`, 
                    2.099519809732654}, {-1.718514948112653, 
                    2.099047874326367}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Dashing[{Small, Small}], 
                   Arrowheads[0.02], 
                   CircleBox[{-1.7737045813088195`, 0.5}, 1.7, {
                    1.5646320668986482`, 1.5734809708646458`}], 
                   ArrowBox[{{-1.7612685871456137`, 
                    2.1999545129353235`}, {-1.778268470745158, 
                    2.1999938737869655`}}]}, 
                  {RGBColor[0, 0, 1], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-1.7737045813088195`, 0.5}, 1.8, {
                    1.5646320668986482`, 1.5814670630950143`}], 
                   ArrowBox[{{-1.7749119065585046`, 
                    2.299999595101549}, {-1.792911542145328, 
                    2.2998975228205145`}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.003], Arrowheads[0.025], 
                   CircleBox[{-1.7737045813088195`, 0.5}, 1.5, {
                    1.5646320668986482`, 1.5646376419240524`}], 
                   ArrowBox[{{-1.7494676087594478`, 
                    1.9998041769383232`}, {-1.764466612401247, 
                    1.9999715530404112`}}]}, 
                  {RGBColor[0, 0, 1], ArrowBox[{{0, 2.8}, {0, 2.8}}]}},
                 Frame->True,
                 FrameTicks->Automatic,
                 ImageSize->800,
                 PlotRange->{{-3, 3}, {-1, 3}}], 492->
                GraphicsBox[{InsetBox[
                   TagBox[GridBox[{
                    {"\<\"t\"\>", "24.55`"},
                    {"\<\"driveIntent\"\>", "0"},
                    {"\<\"motorTorque\"\>", 
                    RowBox[{"-", "3.012195383154662`"}]},
                    {"\<\"vw\"\>", 
                    RowBox[{"-", "0.0348810025123156`"}]},
                    {"\<\"controllerState\"\>", 
                    RowBox[{"\[LeftAssociation]", 
                    RowBox[{
                    
                    RowBox[{"\<\"lastError\"\>", "\[Rule]", 
                    "0.00433770932211084`"}], ",", 
                    RowBox[{"\<\"lastInt\"\>", "\[Rule]", 
                    RowBox[{"-", "0.029864944085205923`"}]}], ",", 
                    RowBox[{"\<\"initialized\"\>", "\[Rule]", "True"}]}], 
                    "\[RightAssociation]"}]}
                    },
                    AutoDelete->False,
                    
                    GridBoxItemSize->{
                    "Columns" -> {{Automatic}}, "Rows" -> {{Automatic}}}],
                    "Grid"], {-1, -0.5}], 
                  {GrayLevel[0], LineBox[{{-5, 0}, {5, 0}}]}, 
                  {GrayLevel[0.5], 
                   CircleBox[{-1.7754547503035345`, 0.5}, 0.5], 
                   DiskBox[{-1.7754547503035345`, 0.5}, 0.1]}, 
                  {GrayLevel[0.5], 
                   DiskBox[{-1.7662637869919364`, 1.9999718418001735`}, 0.1]}, 
                  {GrayLevel[0.5], Thickness[0.01], 
                   LineBox[{{-1.7754547503035345`, 
                    0.5}, {-1.7662637869919364`, 1.9999718418001735`}}]}, 
                  {RGBColor[0, 1, 0], 
                   DiskBox[{-1.7693274414291358`, 1.4999812278667823`}, 
                    0.05]}, 
                  {RGBColor[0, 0, 1], Arrowheads[0.01], 
                   CircleBox[{-1.7754547503035345`, 0.5}, 0.3333333333333333, 
                    NCache[{
                    Rational[-1, 2] 
                    Pi, -1.6611621882895364`}, {-1.5707963267948966`, \
-1.6611621882895364`}]], 
                   ArrowBox[{{-1.8022145436284305`, 
                    0.16774253270645678`}, {-1.805535724946882, 
                    0.168026738914687}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-1.7754547503035345`, 
                    0.5}, {-1.7152108426404413`, 0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-1.7754547503035345`, 
                    0.5}, {-1.8205201844931607`, 0.5}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.005], Arrowheads[0.04], 
                   ArrowBox[{{-1.7754547503035345`, 
                    0.5}, {-1.7602762768300675`, 0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-1.7754547503035345`, 0.5}, 1.6, {
                    1.5646689795793238`, 1.5365025544669229`}], 
                   ArrowBox[{{-1.7365885379000023`, 
                    2.0995278733218137`}, {-1.7205954690478562`, 
                    2.0990592419482494`}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Dashing[{Small, Small}], 
                   Arrowheads[0.02], 
                   CircleBox[{-1.7754547503035345`, 0.5}, 1.7, {
                    1.5646689795793238`, 1.5734653761165072`}], 
                   ArrowBox[{{-1.7629922457793656`, 
                    2.1999543187924155`}, {-1.77999212876302, 
                    2.199993944752956}}]}, 
                  {RGBColor[0, 0, 1], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-1.7754547503035345`, 0.5}, 1.8, {
                    1.5646689795793238`, 1.5814033983746274`}], 
                   ArrowBox[{{-1.7765474790799316`, 
                    2.2999996683176978`}, {-1.7945471211286803`, 
                    2.2998987419786356`}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.003], Arrowheads[0.025], 
                   CircleBox[{-1.7754547503035345`, 0.5}, 1.5, {
                    1.5646689795793238`, 1.564675344755935}], 
                   ArrowBox[{{-1.7512743246361158`, 
                    1.9998050896748358`}, {-1.7662733345778072`, 
                    1.999971900271892}}]}, 
                  {RGBColor[0, 0, 1], ArrowBox[{{0, 2.8}, {0, 2.8}}]}},
                 Frame->True,
                 FrameTicks->Automatic,
                 ImageSize->800,
                 PlotRange->{{-3, 3}, {-1, 3}}], 493->
                GraphicsBox[{InsetBox[
                   TagBox[GridBox[{
                    {"\<\"t\"\>", "24.6`"},
                    {"\<\"driveIntent\"\>", "0"},
                    {"\<\"motorTorque\"\>", 
                    RowBox[{"-", "2.993995329538442`"}]},
                    {"\<\"vw\"\>", 
                    RowBox[{"-", "0.034632493179234926`"}]},
                    {"\<\"controllerState\"\>", 
                    RowBox[{"\[LeftAssociation]", 
                    RowBox[{
                    
                    RowBox[{"\<\"lastError\"\>", "\[Rule]", 
                    "0.0043002258328008035`"}], ",", 
                    RowBox[{"\<\"lastInt\"\>", "\[Rule]", 
                    RowBox[{"-", "0.029649017058829198`"}]}], ",", 
                    RowBox[{"\<\"initialized\"\>", "\[Rule]", "True"}]}], 
                    "\[RightAssociation]"}]}
                    },
                    AutoDelete->False,
                    
                    GridBoxItemSize->{
                    "Columns" -> {{Automatic}}, "Rows" -> {{Automatic}}}],
                    "Grid"], {-1, -0.5}], 
                  {GrayLevel[0], LineBox[{{-5, 0}, {5, 0}}]}, 
                  {GrayLevel[0.5], 
                   CircleBox[{-1.7771924571980713`, 0.5}, 0.5], 
                   DiskBox[{-1.7771924571980713`, 0.5}, 0.1]}, 
                  {GrayLevel[0.5], 
                   DiskBox[{-1.768057115621545, 1.9999721815868052`}, 0.1]}, 
                  {GrayLevel[0.5], Thickness[0.01], 
                   LineBox[{{-1.7771924571980713`, 0.5}, {-1.768057115621545, 
                    1.9999721815868052`}}]}, 
                  {RGBColor[0, 1, 0], 
                   DiskBox[{-1.771102229480387, 1.4999814543912036`}, 0.05]}, 
                  {RGBColor[0, 0, 1], Arrowheads[0.01], 
                   CircleBox[{-1.7771924571980713`, 0.5}, 0.3333333333333333, 
                    NCache[{
                    Rational[-1, 2] 
                    Pi, -1.6606161866810498`}, {-1.5707963267948966`, \
-1.6606161866810498`}]], 
                   ArrowBox[{{-1.8037708334316225`, 
                    0.16772797134290873`}, {-1.8070921694320614`, 
                    0.16801036413852366`}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-1.7771924571980713`, 
                    0.5}, {-1.7173125506073024`, 0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-1.7771924571980713`, 
                    0.5}, {-1.8219851752101601`, 0.5}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.005], Arrowheads[0.04], 
                   ArrowBox[{{-1.7771924571980713`, 
                    0.5}, {-1.7621052686193914`, 0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-1.7771924571980713`, 0.5}, 1.6, {
                    1.5647060614279396`, 1.5367100934681144`}], 
                   ArrowBox[{{-1.7386582100463945`, 
                    2.0995359051288762`}, {-1.722665044279392, 
                    2.0990705929509166`}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Dashing[{Small, Small}], 
                   Arrowheads[0.02], 
                   CircleBox[{-1.7771924571980713`, 0.5}, 1.7, {
                    1.5647060614279396`, 1.5734495575595524`}], 
                   ArrowBox[{{-1.7647030618512505`, 
                    2.1999541214408906`}, {-1.7817029442059524`, 
                    2.1999940163150433`}}]}, 
                  {RGBColor[0, 0, 1], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-1.7771924571980713`, 0.5}, 1.8, {
                    1.5647060614279396`, 1.5813393688142643`}], 
                   ArrowBox[{{-1.778169932784891, 
                    2.2999997345948353`}, {-1.7961695812590115`, 
                    2.2998999607651465`}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.003], Arrowheads[0.025], 
                   CircleBox[{-1.7771924571980713`, 0.5}, 1.5, {
                    1.5647060614279396`, 1.5647129600835434`}], 
                   ArrowBox[{{-1.7530684472075428`, 
                    1.9998059981684222`}, {-1.7680674634132585`, 
                    1.999972244572688}}]}, 
                  {RGBColor[0, 0, 1], ArrowBox[{{0, 2.8}, {0, 2.8}}]}},
                 Frame->True,
                 FrameTicks->Automatic,
                 ImageSize->800,
                 PlotRange->{{-3, 3}, {-1, 3}}], 494->
                GraphicsBox[{InsetBox[
                   TagBox[GridBox[{
                    {"\<\"t\"\>", "24.650000000000002`"},
                    {"\<\"driveIntent\"\>", "0"},
                    {"\<\"motorTorque\"\>", 
                    RowBox[{"-", "2.975693397724017`"}]},
                    {"\<\"vw\"\>", 
                    RowBox[{"-", "0.03438548843470225`"}]},
                    {"\<\"controllerState\"\>", 
                    RowBox[{"\[LeftAssociation]", 
                    RowBox[{
                    
                    RowBox[{"\<\"lastError\"\>", "\[Rule]", 
                    "0.004263376879017881`"}], ",", 
                    RowBox[{"\<\"lastInt\"\>", "\[Rule]", 
                    RowBox[{"-", "0.029434948087010745`"}]}], ",", 
                    RowBox[{"\<\"initialized\"\>", "\[Rule]", "True"}]}], 
                    "\[RightAssociation]"}]}
                    },
                    AutoDelete->False,
                    
                    GridBoxItemSize->{
                    "Columns" -> {{Automatic}}, "Rows" -> {{Automatic}}}],
                    "Grid"], {-1, -0.5}], 
                  {GrayLevel[0], LineBox[{{-5, 0}, {5, 0}}]}, 
                  {GrayLevel[0.5], CircleBox[{-1.778917776945886, 0.5}, 0.5], 
                   DiskBox[{-1.778917776945886, 0.5}, 0.1]}, 
                  {GrayLevel[0.5], 
                   DiskBox[{-1.769838332286404, 1.9999725209765262`}, 0.1]}, 
                  {GrayLevel[0.5], Thickness[0.01], 
                   LineBox[{{-1.778917776945886, 0.5}, {-1.769838332286404, 
                    1.9999725209765262`}}]}, 
                  {RGBColor[0, 1, 0], 
                   DiskBox[{-1.7728648138395646`, 1.4999816806510173`}, 
                    0.05]}, 
                  {RGBColor[0, 0, 1], Arrowheads[0.01], 
                   CircleBox[{-1.778917776945886, 0.5}, 0.3333333333333333, 
                    NCache[{
                    Rational[-1, 2] 
                    Pi, -1.660067128726617}, {-1.5707963267948966`, \
-1.660067128726617}]], 
                   ArrowBox[{{-1.8053137125820136`, 
                    0.16771342835888703`}, {-1.808635203131823, 
                    0.16799399750607746`}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-1.778917776945886, 0.5}, {-1.7194039089914055`,
                     0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-1.778917776945886, 0.5}, {-1.82343642936317, 
                    0.5}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.005], Arrowheads[0.04], 
                   ArrowBox[{{-1.778917776945886, 0.5}, {-1.7639225614086897`,
                     0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-1.778917776945886, 0.5}, 1.6, {
                    1.5647433267261899`, 1.5369186592355932`}], 
                   ArrowBox[{{-1.7407171390635674`, 
                    2.099543907264}, {-1.724723876596223, 
                    2.0990819307230293`}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Dashing[{Small, Small}], 
                   Arrowheads[0.02], 
                   CircleBox[{-1.778917776945886, 0.5}, 1.7, {
                    1.5647433267261899`, 1.573433523622826}], 
                   ArrowBox[{{-1.766401124643851, 
                    2.1999539209681975`}, {-1.7834010063566959`, 
                    2.199994088417383}}]}, 
                  {RGBColor[0, 0, 1], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-1.778917776945886, 0.5}, 1.8, {
                    1.5647433267261899`, 1.5812749567135456`}], 
                   ArrowBox[{{-1.7797793107665598`, 
                    2.2999997938220647`}, {-1.7977789656299827`, 
                    2.299901179387753}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.003], Arrowheads[0.025], 
                   CircleBox[{-1.778917776945886, 0.5}, 1.5, {
                    1.5647433267261899`, 1.5647505272039013`}], 
                   ArrowBox[{{-1.7548501103648138`, 
                    1.9998069033796793`}, {-1.769849132805344, 
                    1.9999725863139806`}}]}, 
                  {RGBColor[0, 0, 1], ArrowBox[{{0, 2.8}, {0, 2.8}}]}},
                 Frame->True,
                 FrameTicks->Automatic,
                 ImageSize->800,
                 PlotRange->{{-3, 3}, {-1, 3}}], 495->
                GraphicsBox[{InsetBox[
                   TagBox[GridBox[{
                    {"\<\"t\"\>", "24.7`"},
                    {"\<\"driveIntent\"\>", "0"},
                    {"\<\"motorTorque\"\>", 
                    RowBox[{"-", "2.95728683538476`"}]},
                    {"\<\"vw\"\>", 
                    RowBox[{"-", "0.03413999955976298`"}]},
                    {"\<\"controllerState\"\>", 
                    RowBox[{"\[LeftAssociation]", 
                    RowBox[{
                    
                    RowBox[{"\<\"lastError\"\>", "\[Rule]", 
                    "0.0042271740653819275`"}], ",", 
                    RowBox[{"\<\"lastInt\"\>", "\[Rule]", 
                    RowBox[{"-", "0.02922270512254882`"}]}], ",", 
                    RowBox[{"\<\"initialized\"\>", "\[Rule]", "True"}]}], 
                    "\[RightAssociation]"}]}
                    },
                    AutoDelete->False,
                    
                    GridBoxItemSize->{
                    "Columns" -> {{Automatic}}, "Rows" -> {{Automatic}}}],
                    "Grid"], {-1, -0.5}], 
                  {GrayLevel[0], LineBox[{{-5, 0}, {5, 0}}]}, 
                  {GrayLevel[0.5], 
                   CircleBox[{-1.7806307850643783`, 0.5}, 0.5], 
                   DiskBox[{-1.7806307850643783`, 0.5}, 0.1]}, 
                  {GrayLevel[0.5], 
                   DiskBox[{-1.7716075246579055`, 1.999972860011686}, 0.1]}, 
                  {GrayLevel[0.5], Thickness[0.01], 
                   LineBox[{{-1.7806307850643783`, 
                    0.5}, {-1.7716075246579055`, 1.999972860011686}}]}, 
                  {RGBColor[0, 1, 0], 
                   DiskBox[{-1.77461527812673, 1.4999819066744573`}, 0.05]}, 
                  {RGBColor[0, 0, 1], Arrowheads[0.01], 
                   CircleBox[{-1.7806307850643783`, 0.5}, 0.3333333333333333, 
                    NCache[{
                    Rational[-1, 2] 
                    Pi, -1.6595149318564393`}, {-1.5707963267948966`, \
-1.6595149318564393`}]], 
                   ArrowBox[{{-1.806843229080626, 
                    0.16769890326722336`}, {-1.8101648740534355`, 
                    0.1679776382550453}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-1.7806307850643783`, 0.5}, {-1.721485048356683,
                     0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-1.7806307850643783`, 0.5}, {-1.824873962876243,
                     0.5}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.005], Arrowheads[0.04], 
                   ArrowBox[{{-1.7806307850643783`, 
                    0.5}, {-1.7657282261685479`, 0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-1.7806307850643783`, 0.5}, 1.6, {
                    1.5647807835768104`, 1.5371282971189555`}], 
                   ArrowBox[{{-1.742765473018089, 
                    2.099551880416399}, {-1.7267721140543975`, 
                    2.0990932566792466`}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Dashing[{Small, Small}], 
                   Arrowheads[0.02], 
                   CircleBox[{-1.7806307850643783`, 0.5}, 1.7, {
                    1.5647807835768104`, 1.5734172850441022`}], 
                   ArrowBox[{{-1.7680865279284221`, 
                    2.1999537174914225`}, {-1.7850864089867633`, 
                    2.19999416099452}}]}, 
                  {RGBColor[0, 0, 1], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-1.7806307850643783`, 0.5}, 1.8, {
                    1.5647807835768104`, 1.5812101548845034`}], 
                   ArrowBox[{{-1.7813756756044097`, 
                    2.2999998458716835`}, {-1.7993753368204357`, 
                    2.2999023978481343`}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.003], Arrowheads[0.025], 
                   CircleBox[{-1.7806307850643783`, 0.5}, 1.5, {
                    1.5647807835768104`, 1.5647880785872608`}], 
                   ArrowBox[{{-1.756619438324256, 
                    1.9998078060964097`}, {-1.7716184669758346`, 
                    1.9999729257965528`}}]}, 
                  {RGBColor[0, 0, 1], ArrowBox[{{0, 2.8}, {0, 2.8}}]}},
                 Frame->True,
                 FrameTicks->Automatic,
                 ImageSize->800,
                 PlotRange->{{-3, 3}, {-1, 3}}], 496->
                GraphicsBox[{InsetBox[
                   TagBox[GridBox[{
                    {"\<\"t\"\>", "24.75`"},
                    {"\<\"driveIntent\"\>", "0"},
                    {"\<\"motorTorque\"\>", 
                    RowBox[{"-", "2.9387755673711924`"}]},
                    {"\<\"vw\"\>", 
                    RowBox[{"-", "0.03389603765481492`"}]},
                    {"\<\"controllerState\"\>", 
                    RowBox[{"\[LeftAssociation]", 
                    RowBox[{
                    
                    RowBox[{"\<\"lastError\"\>", "\[Rule]", 
                    "0.004191623401286563`"}], ",", 
                    RowBox[{"\<\"lastInt\"\>", "\[Rule]", 
                    RowBox[{"-", "0.02901225568308871`"}]}], ",", 
                    RowBox[{"\<\"initialized\"\>", "\[Rule]", "True"}]}], 
                    "\[RightAssociation]"}]}
                    },
                    AutoDelete->False,
                    
                    GridBoxItemSize->{
                    "Columns" -> {{Automatic}}, "Rows" -> {{Automatic}}}],
                    "Grid"], {-1, -0.5}], 
                  {GrayLevel[0], LineBox[{{-5, 0}, {5, 0}}]}, 
                  {GrayLevel[0.5], 
                   CircleBox[{-1.7823315576312737`, 0.5}, 0.5], 
                   DiskBox[{-1.7823315576312737`, 0.5}, 0.1]}, 
                  {GrayLevel[0.5], 
                   DiskBox[{-1.773364772673657, 1.999973198683071}, 0.1]}, 
                  {GrayLevel[0.5], Thickness[0.01], 
                   LineBox[{{-1.7823315576312737`, 0.5}, {-1.773364772673657, 
                    1.999973198683071}}]}, 
                  {RGBColor[0, 1, 0], 
                   DiskBox[{-1.7763537009928625`, 1.4999821324553806`}, 
                    0.05]}, 
                  {RGBColor[0, 0, 1], Arrowheads[0.01], 
                   CircleBox[{-1.7823315576312737`, 0.5}, 0.3333333333333333, 
                    NCache[{
                    Rational[-1, 2] 
                    Pi, -1.6589595938160324`}, {-1.5707963267948966`, \
-1.6589595938160324`}]], 
                   ArrowBox[{{-1.808359458175159, 
                    0.16768439774156146`}, {-1.8116812574279042`, 
                    0.1679612880506871}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-1.7823315576312737`, 0.5}, {-1.72355604628385, 
                    0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-1.7823315576312737`, 
                    0.5}, {-1.8262978329001587`, 0.5}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.005], Arrowheads[0.04], 
                   ArrowBox[{{-1.7823315576312737`, 
                    0.5}, {-1.7675223215527347`, 0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-1.7823315576312737`, 0.5}, 1.6, {
                    1.564818434553024, 1.5373390215203326`}], 
                   ArrowBox[{{-1.7448033110356627`, 
                    2.0995598240476845`}, {-1.72880985578385, 
                    2.0991045705116838`}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Dashing[{Small, Small}], 
                   Arrowheads[0.02], 
                   CircleBox[{-1.7823315576312737`, 0.5}, 1.7, {
                    1.564818434553024, 1.5734008544940463`}], 
                   ArrowBox[{{-1.769759369322364, 
                    2.1999535111529154`}, {-1.7867592497139018`, 
                    2.1999942339734044`}}]}, 
                  {RGBColor[0, 0, 1], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-1.7823315576312737`, 0.5}, 1.8, {
                    1.564818434553024, 1.5811449654828642`}], 
                   ArrowBox[{{-1.7829591072569024`, 
                    2.2999998906059598`}, {-1.8009587747872604`, 
                    2.299903615969762}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.003], Arrowheads[0.025], 
                   CircleBox[{-1.7823315576312737`, 0.5}, 1.5, {
                    1.564818434553024, 1.5648256403671197`}], 
                   ArrowBox[{{-1.75837654635872, 
                    1.9998087069473}, {-1.7733755812019074`, 
                    1.9999732632571143`}}]}, 
                  {RGBColor[0, 0, 1], ArrowBox[{{0, 2.8}, {0, 2.8}}]}},
                 Frame->True,
                 FrameTicks->Automatic,
                 ImageSize->800,
                 PlotRange->{{-3, 3}, {-1, 3}}], 497->
                GraphicsBox[{InsetBox[
                   TagBox[GridBox[{
                    {"\<\"t\"\>", "24.8`"},
                    {"\<\"driveIntent\"\>", "0"},
                    {"\<\"motorTorque\"\>", 
                    RowBox[{"-", "2.920161880308668`"}]},
                    {"\<\"vw\"\>", 
                    RowBox[{"-", "0.0336536134418542`"}]},
                    {"\<\"controllerState\"\>", 
                    RowBox[{"\[LeftAssociation]", 
                    RowBox[{
                    RowBox[{"\<\"lastError\"\>", "\[Rule]", 
                    "0.004156725869850843`"}], ",", 
                    RowBox[{"\<\"lastInt\"\>", "\[Rule]", 
                    RowBox[{"-", "0.02880356711631343`"}]}], ",", 
                    RowBox[{"\<\"initialized\"\>", "\[Rule]", "True"}]}], 
                    "\[RightAssociation]"}]}
                    },
                    AutoDelete->False,
                    
                    GridBoxItemSize->{
                    "Columns" -> {{Automatic}}, "Rows" -> {{Automatic}}}],
                    "Grid"], {-1, -0.5}], 
                  {GrayLevel[0], LineBox[{{-5, 0}, {5, 0}}]}, 
                  {GrayLevel[0.5], 
                   CircleBox[{-1.7840201712704329`, 0.5}, 0.5], 
                   DiskBox[{-1.7840201712704329`, 0.5}, 0.1]}, 
                  {GrayLevel[0.5], 
                   DiskBox[{-1.7751101494652302`, 1.9999735369370457`}, 0.1]}, 
                  {GrayLevel[0.5], Thickness[0.01], 
                   LineBox[{{-1.7840201712704329`, 
                    0.5}, {-1.7751101494652302`, 1.9999735369370457`}}]}, 
                  {RGBColor[0, 1, 0], 
                   DiskBox[{-1.778080156733631, 1.4999823579580305`}, 0.05]}, 
                  {RGBColor[0, 0, 1], Arrowheads[0.01], 
                   CircleBox[{-1.7840201712704329`, 0.5}, 0.3333333333333333, 
                    NCache[{
                    Rational[-1, 2] 
                    Pi, -1.6584011832041565`}, {-1.5707963267948966`, \
-1.6584011832041565`}]], 
                   ArrowBox[{{-1.809862499207128, 
                    0.16766991529817316`}, {-1.8131844525604464`, 
                    0.16794495063627168`}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-1.7840201712704329`, 
                    0.5}, {-1.7256169336642595`, 0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-1.7840201712704329`, 
                    0.5}, {-1.8277081331799734`, 0.5}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.005], Arrowheads[0.04], 
                   ArrowBox[{{-1.7840201712704329`, 0.5}, {-1.7693048955738, 
                    0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-1.7840201712704329`, 0.5}, 1.6, {
                    1.5648562773265196`, 1.5375508194090222`}], 
                   ArrowBox[{{-1.7468307089075825`, 
                    2.0995677365742167`}, {-1.7308371575927526`, 
                    2.0991158704284567`}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Dashing[{Small, Small}], 
                   Arrowheads[0.02], 
                   CircleBox[{-1.7840201712704329`, 0.5}, 1.7, {
                    1.5648562773265196`, 1.5733842462249708`}], 
                   ArrowBox[{{-1.7714197496779294`, 
                    2.1999533021161763`}, {-1.7884196293907868`, 
                    2.1999943072752473`}}]}, 
                  {RGBColor[0, 0, 1], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-1.7840201712704329`, 0.5}, 1.8, {
                    1.5648562773265196`, 1.58107939888379}], 
                   ArrowBox[{{-1.784529701023636, 
                    2.2999999278831735`}, {-1.8025293748277045`, 
                    2.2999048334241663`}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.003], Arrowheads[0.025], 
                   CircleBox[{-1.7840201712704329`, 0.5}, 1.5, {
                    1.5648562773265196`, 1.5648632328348249`}], 
                   ArrowBox[{{-1.7601215415251723`, 
                    1.9998096064155273`}, {-1.7751205825438392`, 
                    1.9999735988744927`}}]}, 
                  {RGBColor[0, 0, 1], ArrowBox[{{0, 2.8}, {0, 2.8}}]}},
                 Frame->True,
                 FrameTicks->Automatic,
                 ImageSize->800,
                 PlotRange->{{-3, 3}, {-1, 3}}], 498->
                GraphicsBox[{InsetBox[
                   TagBox[GridBox[{
                    {"\<\"t\"\>", "24.85`"},
                    {"\<\"driveIntent\"\>", "0"},
                    {"\<\"motorTorque\"\>", 
                    RowBox[{"-", "2.9014501207204333`"}]},
                    {"\<\"vw\"\>", 
                    RowBox[{"-", "0.033412737093784346`"}]},
                    {"\<\"controllerState\"\>", 
                    RowBox[{"\[LeftAssociation]", 
                    RowBox[{
                    
                    RowBox[{"\<\"lastError\"\>", "\[Rule]", 
                    "0.004122477980033602`"}], ",", 
                    RowBox[{"\<\"lastInt\"\>", "\[Rule]", 
                    RowBox[{"-", "0.02859660683710095`"}]}], ",", 
                    RowBox[{"\<\"initialized\"\>", "\[Rule]", "True"}]}], 
                    "\[RightAssociation]"}]}
                    },
                    AutoDelete->False,
                    
                    GridBoxItemSize->{
                    "Columns" -> {{Automatic}}, "Rows" -> {{Automatic}}}],
                    "Grid"], {-1, -0.5}], 
                  {GrayLevel[0], LineBox[{{-5, 0}, {5, 0}}]}, 
                  {GrayLevel[0.5], CircleBox[{-1.785696703128473, 0.5}, 0.5], 
                   DiskBox[{-1.785696703128473, 0.5}, 0.1]}, 
                  {GrayLevel[0.5], 
                   DiskBox[{-1.7768437222391824`, 1.9999738746822806`}, 0.1]}, 
                  {GrayLevel[0.5], Thickness[0.01], 
                   LineBox[{{-1.785696703128473, 0.5}, {-1.7768437222391824`, 
                    1.9999738746822806`}}]}, 
                  {RGBColor[0, 1, 0], 
                   DiskBox[{-1.779794715868946, 1.4999825831215203`}, 0.05]}, 
                  {RGBColor[0, 0, 1], Arrowheads[0.01], 
                   CircleBox[{-1.785696703128473, 0.5}, 0.3333333333333333, 
                    NCache[{
                    Rational[-1, 2] 
                    Pi, -1.6578398304165096`}, {-1.5707963267948966`, \
-1.6578398304165096`}]], 
                   ArrowBox[{{-1.8113524725838217`, 
                    0.16765546099752882`}, {-1.8146745798055839`, 
                    0.1679286315046158}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-1.785696703128473, 0.5}, {-1.7276677007140644`,
                     0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-1.785696703128473, 0.5}, {-1.8291049895974347`,
                     0.5}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.005], Arrowheads[0.04], 
                   ArrowBox[{{-1.785696703128473, 0.5}, {-1.771075987183026, 
                    0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-1.785696703128473, 0.5}, 1.6, {
                    1.5648943052703992`, 1.5377636536960366`}], 
                   ArrowBox[{{-1.7488476844641094`, 
                    2.0995756155379066`}, {-1.732854037338913, 
                    2.099127153378446}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Dashing[{Small, Small}], 
                   Arrowheads[0.02], 
                   CircleBox[{-1.785696703128473, 0.5}, 1.7, {
                    1.5648943052703992`, 1.573367475744654}], 
                   ArrowBox[{{-1.7730677725043502`, 
                    2.1999530905619986`}, {-1.790067651527141, 
                    2.1999943808172118`}}]}, 
                  {RGBColor[0, 0, 1], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-1.785696703128473, 0.5}, 1.8, {
                    1.5648943052703992`, 1.581013472607735}], 
                   ArrowBox[{{-1.7860875655885102`, 
                    2.2999999575629264`}, {-1.8040872456226864`, 
                    2.299906049755589}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.003], Arrowheads[0.025], 
                   CircleBox[{-1.785696703128473, 0.5}, 1.5, {
                    1.5648943052703992`, 1.564900870933371}], 
                   ArrowBox[{{-1.7618545233818934`, 
                    1.9998105048521737`}, {-1.7768535705623008`, 
                    1.999973932775639}}]}, 
                  {RGBColor[0, 0, 1], ArrowBox[{{0, 2.8}, {0, 2.8}}]}},
                 Frame->True,
                 FrameTicks->Automatic,
                 ImageSize->800,
                 PlotRange->{{-3, 3}, {-1, 3}}], 499->
                GraphicsBox[{InsetBox[
                   TagBox[GridBox[{
                    {"\<\"t\"\>", "24.900000000000002`"},
                    {"\<\"driveIntent\"\>", "0"},
                    {"\<\"motorTorque\"\>", 
                    RowBox[{"-", "2.882646408025983`"}]},
                    {"\<\"vw\"\>", 
                    RowBox[{"-", "0.03317341808932132`"}]},
                    {"\<\"controllerState\"\>", 
                    RowBox[{"\[LeftAssociation]", 
                    RowBox[{
                    
                    RowBox[{"\<\"lastError\"\>", "\[Rule]", 
                    "0.004088872298544401`"}], ",", 
                    RowBox[{"\<\"lastInt\"\>", "\[Rule]", 
                    RowBox[{"-", "0.028391342537578114`"}]}], ",", 
                    RowBox[{"\<\"initialized\"\>", "\[Rule]", "True"}]}], 
                    "\[RightAssociation]"}]}
                    },
                    AutoDelete->False,
                    
                    GridBoxItemSize->{
                    "Columns" -> {{Automatic}}, "Rows" -> {{Automatic}}}],
                    "Grid"], {-1, -0.5}], 
                  {GrayLevel[0], LineBox[{{-5, 0}, {5, 0}}]}, 
                  {GrayLevel[0.5], 
                   CircleBox[{-1.7873612308435076`, 0.5}, 0.5], 
                   DiskBox[{-1.7873612308435076`, 0.5}, 0.1]}, 
                  {GrayLevel[0.5], 
                   DiskBox[{-1.778565553108282, 1.9999742117960488`}, 0.1]}, 
                  {GrayLevel[0.5], Thickness[0.01], 
                   LineBox[{{-1.7873612308435076`, 0.5}, {-1.778565553108282, 
                    1.9999742117960488`}}]}, 
                  {RGBColor[0, 1, 0], 
                   DiskBox[{-1.7814974456866906`, 1.4999828078640327`}, 
                    0.05]}, 
                  {RGBColor[0, 0, 1], Arrowheads[0.01], 
                   CircleBox[{-1.7873612308435076`, 0.5}, 0.3333333333333333, 
                    NCache[{
                    Rational[-1, 2] 
                    Pi, -1.6572757190356762`}, {-1.5707963267948966`, \
-1.6572757190356762`}]], 
                   ArrowBox[{{-1.81282951688988, 
                    0.16764104116637496`}, {-1.816151777681643, 
                    0.16791233759160484`}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-1.7873612308435076`, 0.5}, {-1.729708302682988,
                     0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-1.7873612308435076`, 0.5}, {-1.830488555910999,
                     0.5}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.005], Arrowheads[0.04], 
                   ArrowBox[{{-1.7873612308435076`, 
                    0.5}, {-1.7728356277504793`, 0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-1.7873612308435076`, 0.5}, 1.6, {
                    1.5649325080341845`, 1.5379774664522625`}], 
                   ArrowBox[{{-1.750854222689981, 
                    2.0995834577650765`}, {-1.7348604800434368`, 
                    2.0991384152616153`}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Dashing[{Small, Small}], 
                   Arrowheads[0.02], 
                   CircleBox[{-1.7873612308435076`, 0.5}, 1.7, {
                    1.5649325080341845`, 1.5733505595155441`}], 
                   ArrowBox[{{-1.7747035434252372`, 
                    2.1999528766848866`}, {-1.7917034217471202`, 
                    2.1999944545139423`}}]}, 
                  {RGBColor[0, 0, 1], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-1.7873612308435076`, 0.5}, 1.8, {
                    1.5649325080341845`, 1.5809472103009956`}], 
                   ArrowBox[{{-1.7876328211534553`, 
                    2.299999979510751}, {-1.8056325073706638`, 
                    2.299907264403938}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.003], Arrowheads[0.025], 
                   CircleBox[{-1.7873612308435076`, 0.5}, 1.5, {
                    1.5649325080341845`, 1.5649385647458485`}], 
                   ArrowBox[{{-1.7635755846897743`, 
                    1.9998114024893427`}, {-1.7785746380197474`, 
                    1.9999742650414205`}}]}, 
                  {RGBColor[0, 0, 1], ArrowBox[{{0, 2.8}, {0, 2.8}}]}},
                 Frame->True,
                 FrameTicks->Automatic,
                 ImageSize->800,
                 PlotRange->{{-3, 3}, {-1, 3}}], 500->
                GraphicsBox[{InsetBox[
                   TagBox[GridBox[{
                    {"\<\"t\"\>", "24.95`"},
                    {"\<\"driveIntent\"\>", "0"},
                    {"\<\"motorTorque\"\>", 
                    RowBox[{"-", "2.8637583634591115`"}]},
                    {"\<\"vw\"\>", 
                    RowBox[{"-", "0.03293566509195026`"}]},
                    {"\<\"controllerState\"\>", 
                    RowBox[{"\[LeftAssociation]", 
                    RowBox[{
                    
                    RowBox[{"\<\"lastError\"\>", "\[Rule]", 
                    "0.004055897958764177`"}], ",", 
                    RowBox[{"\<\"lastInt\"\>", "\[Rule]", 
                    RowBox[{"-", "0.028187742371154765`"}]}], ",", 
                    RowBox[{"\<\"initialized\"\>", "\[Rule]", "True"}]}], 
                    "\[RightAssociation]"}]}
                    },
                    AutoDelete->False,
                    
                    GridBoxItemSize->{
                    "Columns" -> {{Automatic}}, "Rows" -> {{Automatic}}}],
                    "Grid"], {-1, -0.5}], 
                  {GrayLevel[0], LineBox[{{-5, 0}, {5, 0}}]}, 
                  {GrayLevel[0.5], 
                   CircleBox[{-1.7890138325072518`, 0.5}, 0.5], 
                   DiskBox[{-1.7890138325072518`, 0.5}, 0.1]}, 
                  {GrayLevel[0.5], 
                   DiskBox[{-1.7802756998707157`, 1.9999745481300761`}, 0.1]}, 
                  {GrayLevel[0.5], Thickness[0.01], 
                   LineBox[{{-1.7890138325072518`, 
                    0.5}, {-1.7802756998707157`, 1.9999745481300761`}}]}, 
                  {RGBColor[0, 1, 0], 
                   DiskBox[{-1.783188410749561, 1.4999830320867176`}, 0.05]}, 
                  {RGBColor[0, 0, 1], Arrowheads[0.01], 
                   CircleBox[{-1.7890138325072518`, 0.5}, 0.3333333333333333, 
                    NCache[{
                    Rational[-1, 2] 
                    Pi, -1.65670907769867}, {-1.5707963267948966`, \
-1.65670907769867}]], 
                   ArrowBox[{{-1.8142937861501938`, 
                    0.1676266631407961}, {-1.8176162001363576`, 
                    0.16789607699227554`}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-1.7890138325072518`, 
                    0.5}, {-1.7317386652380695`, 0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-1.7890138325072518`, 
                    0.5}, {-1.8318590097117078`, 0.5}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.005], Arrowheads[0.04], 
                   ArrowBox[{{-1.7890138325072518`, 
                    0.5}, {-1.7745838424425255`, 0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-1.7890138325072518`, 0.5}, 1.6, {
                    1.5649708720885651`, 1.5381921819572881`}], 
                   ArrowBox[{{-1.7528502805625876`, 
                    2.0995912595131125`}, {-1.7368564427271957`, 
                    2.0991496511244128`}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Dashing[{Small, Small}], 
                   Arrowheads[0.02], 
                   CircleBox[{-1.7890138325072518`, 0.5}, 1.7, {
                    1.5649708720885651`, 1.5733335146792746`}], 
                   ArrowBox[{{-1.7763271696723724`, 
                    2.1999526606897364`}, {-1.7933270472830987`, 
                    2.1999945282789284`}}]}, 
                  {RGBColor[0, 0, 1], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-1.7890138325072518`, 0.5}, 1.8, {
                    1.5649708720885651`, 1.5808806407744491`}], 
                   ArrowBox[{{-1.7891655976702665`, 
                    2.2999999936020377`}, {-1.807165290019593, 
                    2.299908476726019}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.003], Arrowheads[0.025], 
                   CircleBox[{-1.7890138325072518`, 0.5}, 1.5, {
                    1.5649708720885651`, 1.5649763199744011`}], 
                   ArrowBox[{{-1.7652848120927445`, 
                    1.9998122994528909`}, {-1.7802838715609404`, 
                    1.999974595712166}}]}, 
                  {RGBColor[0, 0, 1], ArrowBox[{{0, 2.8}, {0, 2.8}}]}},
                 Frame->True,
                 FrameTicks->Automatic,
                 ImageSize->800,
                 PlotRange->{{-3, 3}, {-1, 3}}], 501->
                GraphicsBox[{InsetBox[
                   TagBox[GridBox[{
                    {"\<\"t\"\>", "25.`"},
                    {"\<\"driveIntent\"\>", "0"},
                    {"\<\"motorTorque\"\>", 
                    RowBox[{"-", "2.8447948556849654`"}]},
                    {"\<\"vw\"\>", 
                    RowBox[{"-", "0.03269948585133561`"}]},
                    {"\<\"controllerState\"\>", 
                    RowBox[{"\[LeftAssociation]", 
                    RowBox[{
                    
                    RowBox[{"\<\"lastError\"\>", "\[Rule]", 
                    "0.004023541144427775`"}], ",", 
                    RowBox[{"\<\"lastInt\"\>", "\[Rule]", 
                    RowBox[{"-", "0.027985775111747196`"}]}], ",", 
                    RowBox[{"\<\"initialized\"\>", "\[Rule]", "True"}]}], 
                    "\[RightAssociation]"}]}
                    },
                    AutoDelete->False,
                    
                    GridBoxItemSize->{
                    "Columns" -> {{Automatic}}, "Rows" -> {{Automatic}}}],
                    "Grid"], {-1, -0.5}], 
                  {GrayLevel[0], LineBox[{{-5, 0}, {5, 0}}]}, 
                  {GrayLevel[0.5], 
                   CircleBox[{-1.7906545866216608`, 0.5}, 0.5], 
                   DiskBox[{-1.7906545866216608`, 0.5}, 0.1]}, 
                  {GrayLevel[0.5], 
                   DiskBox[{-1.7819742167358137`, 1.9999748835159359`}, 0.1]}, 
                  {GrayLevel[0.5], Thickness[0.01], 
                   LineBox[{{-1.7906545866216608`, 
                    0.5}, {-1.7819742167358137`, 1.9999748835159359`}}]}, 
                  {RGBColor[0, 1, 0], 
                   DiskBox[{-1.7848676733644293`, 1.4999832556772905`}, 
                    0.05]}, 
                  {RGBColor[0, 0, 1], Arrowheads[0.01], 
                   CircleBox[{-1.7906545866216608`, 0.5}, 0.3333333333333333, 
                    NCache[{
                    Rational[-1, 2] 
                    Pi, -1.6561401724654454`}, {-1.5707963267948966`, \
-1.6561401724654454`}]], 
                   ArrowBox[{{-1.8157454472531107`, 
                    0.16761233503048845`}, {-1.8190680139725615`, 
                    0.16787985869976785`}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-1.7906545866216608`, 
                    0.5}, {-1.7337586895079615`, 0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-1.7906545866216608`, 0.5}, {-1.833216548609022,
                     0.5}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.005], Arrowheads[0.04], 
                   ArrowBox[{{-1.7906545866216608`, 
                    0.5}, {-1.7763206514953227`, 0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-1.7906545866216608`, 0.5}, 1.6, {
                    1.5650093812381345`, 1.5384077095690587`}], 
                   ArrowBox[{{-1.7548357915982369`, 
                    2.0995990166048086`}, {-1.7388418589555186`, 
                    2.0991608553400107`}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Dashing[{Small, Small}], 
                   Arrowheads[0.02], 
                   CircleBox[{-1.7906545866216608`, 0.5}, 1.7, {
                    1.5650093812381345`, 1.5733163588062833`}], 
                   ArrowBox[{{-1.7779387596167118`, 
                    2.199952442788792}, {-1.794938636506661, 
                    2.199994602025719}}]}, 
                  {RGBColor[0, 0, 1], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-1.7906545866216608`, 0.5}, 1.8, {
                    1.5650093812381345`, 1.580813797103051}], 
                   ArrowBox[{{-1.7906860331763368`, 
                    2.2999999997253093`}, {-1.8086857316027753`, 
                    2.299909686015015}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.003], Arrowheads[0.025], 
                   CircleBox[{-1.7906545866216608`, 0.5}, 1.5, {
                    1.5650093812381345`, 1.5650141384062553`}], 
                   ArrowBox[{{-1.7669822867733498`, 
                    1.999813195774691}, {-1.7819813523686097`, 
                    1.999974924792942}}]}, 
                  {RGBColor[0, 0, 1], ArrowBox[{{0, 2.8}, {0, 2.8}}]}},
                 Frame->True,
                 FrameTicks->Automatic,
                 ImageSize->800,
                 PlotRange->{{-3, 3}, {-1, 3}}], 502->
                GraphicsBox[{InsetBox[
                   TagBox[GridBox[{
                    {"\<\"t\"\>", "25.05`"},
                    {"\<\"driveIntent\"\>", "0"},
                    {"\<\"motorTorque\"\>", 
                    RowBox[{"-", "2.825765763639903`"}]},
                    {"\<\"vw\"\>", 
                    RowBox[{"-", "0.03246488712555441`"}]},
                    {"\<\"controllerState\"\>", 
                    RowBox[{"\[LeftAssociation]", 
                    RowBox[{
                    
                    RowBox[{"\<\"lastError\"\>", "\[Rule]", 
                    "0.003991785546319985`"}], ",", 
                    RowBox[{"\<\"lastInt\"\>", "\[Rule]", 
                    RowBox[{"-", "0.02778541028949923`"}]}], ",", 
                    RowBox[{"\<\"initialized\"\>", "\[Rule]", "True"}]}], 
                    "\[RightAssociation]"}]}
                    },
                    AutoDelete->False,
                    
                    GridBoxItemSize->{
                    "Columns" -> {{Automatic}}, "Rows" -> {{Automatic}}}],
                    "Grid"], {-1, -0.5}], 
                  {GrayLevel[0], LineBox[{{-5, 0}, {5, 0}}]}, 
                  {GrayLevel[0.5], 
                   CircleBox[{-1.7922835720511694`, 0.5}, 0.5], 
                   DiskBox[{-1.7922835720511694`, 0.5}, 0.1]}, 
                  {GrayLevel[0.5], 
                   DiskBox[{-1.7836611549954793`, 1.999975217769986}, 0.1]}, 
                  {GrayLevel[0.5], Thickness[0.01], 
                   LineBox[{{-1.7922835720511694`, 
                    0.5}, {-1.7836611549954793`, 1.999975217769986}}]}, 
                  {RGBColor[0, 1, 0], 
                   DiskBox[{-1.7865352940140427`, 1.4999834785133241`}, 
                    0.05]}, 
                  {RGBColor[0, 0, 1], Arrowheads[0.01], 
                   CircleBox[{-1.7922835720511694`, 0.5}, 0.3333333333333333, 
                    NCache[{
                    Rational[-1, 2] 
                    Pi, -1.6555692997040936`}, {-1.5707963267948966`, \
-1.6555692997040936`}]], 
                   ArrowBox[{{-1.8171846775402836`, 
                    0.16759806550422862`}, {-1.820507396440297, 
                    0.16786369236718068`}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-1.7922835720511694`, 
                    0.5}, {-1.7357682567783714`, 0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-1.7922835720511694`, 
                    0.5}, {-1.8345613866568702`, 0.5}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.005], Arrowheads[0.04], 
                   ArrowBox[{{-1.7922835720511694`, 
                    0.5}, {-1.7780460713840722`, 0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-1.7922835720511694`, 0.5}, 1.6, {
                    1.565048017100861, 1.538623946407366}], 
                   ArrowBox[{{-1.7568106700963744`, 
                    2.099606724550415}, {-1.7408166430809753`, 
                    2.0991720217732603`}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Dashing[{Small, Small}], 
                   Arrowheads[0.02], 
                   CircleBox[{-1.7922835720511694`, 0.5}, 1.7, {
                    1.565048017100861, 1.5732991096700186`}], 
                   ArrowBox[{{-1.7795384223367843`, 
                    2.1999522231988635`}, {-1.7965382984969944`, 
                    2.199994675668977}}]}, 
                  {RGBColor[0, 0, 1], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-1.7922835720511694`, 0.5}, 1.8, {
                    1.565048017100861, 1.5807467157877493`}], 
                   ArrowBox[{{-1.7921942722383408`, 
                    2.299999997784873}, {-1.8101939766826434`, 
                    2.2999108915182265`}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.003], Arrowheads[0.025], 
                   CircleBox[{-1.7922835720511694`, 0.5}, 1.5, {
                    1.565048017100861, 1.5650520183637777`}], 
                   ArrowBox[{{-1.768668085079976, 
                    1.9998140914043692`}, {-1.7836671567907632`, 
                    1.9999752522585366`}}]}, 
                  {RGBColor[0, 0, 1], ArrowBox[{{0, 2.8}, {0, 2.8}}]}},
                 Frame->True,
                 FrameTicks->Automatic,
                 ImageSize->800,
                 PlotRange->{{-3, 3}, {-1, 3}}], 503->
                GraphicsBox[{InsetBox[
                   TagBox[GridBox[{
                    {"\<\"t\"\>", "25.1`"},
                    {"\<\"driveIntent\"\>", "0"},
                    {"\<\"motorTorque\"\>", 
                    RowBox[{"-", "2.8066817568970355`"}]},
                    {"\<\"vw\"\>", 
                    RowBox[{"-", "0.03223187462250928`"}]},
                    {"\<\"controllerState\"\>", 
                    RowBox[{"\[LeftAssociation]", 
                    RowBox[{
                    
                    RowBox[{"\<\"lastError\"\>", "\[Rule]", 
                    "0.003960612790695302`"}], ",", 
                    RowBox[{"\<\"lastInt\"\>", "\[Rule]", 
                    RowBox[{"-", "0.027586618304385156`"}]}], ",", 
                    RowBox[{"\<\"initialized\"\>", "\[Rule]", "True"}]}], 
                    "\[RightAssociation]"}]}
                    },
                    AutoDelete->False,
                    
                    GridBoxItemSize->{
                    "Columns" -> {{Automatic}}, "Rows" -> {{Automatic}}}],
                    "Grid"], {-1, -0.5}], 
                  {GrayLevel[0], LineBox[{{-5, 0}, {5, 0}}]}, 
                  {GrayLevel[0.5], CircleBox[{-1.793900867971543, 0.5}, 0.5], 
                   DiskBox[{-1.793900867971543, 0.5}, 0.1]}, 
                  {GrayLevel[0.5], 
                   DiskBox[{-1.7853365636411562`, 1.999975550697856}, 0.1]}, 
                  {GrayLevel[0.5], Thickness[0.01], 
                   LineBox[{{-1.793900867971543, 0.5}, {-1.7853365636411562`, 
                    1.999975550697856}}]}, 
                  {RGBColor[0, 1, 0], 
                   DiskBox[{-1.788191331751285, 1.4999837004652372`}, 0.05]}, 
                  {RGBColor[0, 0, 1], Arrowheads[0.01], 
                   CircleBox[{-1.793900867971543, 0.5}, 0.3333333333333333, 
                    NCache[{
                    Rational[-1, 2] 
                    Pi, -1.6549967795018077`}, {-1.5707963267948966`, \
-1.6549967795018077`}]], 
                   ArrowBox[{{-1.8186116625672415`, 
                    0.1675838635963045}, {-1.8219345329994323`, 
                    0.16784758809213024`}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-1.793900867971543, 0.5}, {-1.7377672328336022`,
                     0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-1.793900867971543, 0.5}, {-1.8358937510267228`,
                     0.5}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.005], Arrowheads[0.04], 
                   ArrowBox[{{-1.793900867971543, 0.5}, {-1.7797601158887821`,
                     0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-1.793900867971543, 0.5}, 1.6, {
                    1.5650867595535085`, 1.5388407798467625`}], 
                   ArrowBox[{{-1.7587748150757059`, 
                    2.0996143786575434`}, {-1.742780694178203, 
                    2.099183143930469}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Dashing[{Small, Small}], 
                   Arrowheads[0.02], 
                   CircleBox[{-1.793900867971543, 0.5}, 1.7, {
                    1.5650867595535085`, 1.5732817850451049`}], 
                   ArrowBox[{{-1.781126267224434, 
                    2.199952002138811}, {-1.7981261426466195`, 
                    2.1999947491253966`}}]}, 
                  {RGBColor[0, 0, 1], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-1.793900867971543, 0.5}, 1.8, {
                    1.5650867595535085`, 1.5806794359807141`}], 
                   ArrowBox[{{-1.7936904645064935`, 
                    2.2999999877028836`}, {-1.81169017490511, 
                    2.29991209245308}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.003], Arrowheads[0.025], 
                   CircleBox[{-1.793900867971543, 0.5}, 1.5, {
                    1.5650867595535085`, 1.5650899551361024`}], 
                   ArrowBox[{{-1.7703422791230445`, 
                    1.9998149862204562`}, {-1.785341356936961, 
                    1.9999755780581392`}}]}, 
                  {RGBColor[0, 0, 1], ArrowBox[{{0, 2.8}, {0, 2.8}}]}},
                 Frame->True,
                 FrameTicks->Automatic,
                 ImageSize->800,
                 PlotRange->{{-3, 3}, {-1, 3}}], 504->
                GraphicsBox[{InsetBox[
                   TagBox[GridBox[{
                    {"\<\"t\"\>", "25.150000000000002`"},
                    {"\<\"driveIntent\"\>", "0"},
                    {"\<\"motorTorque\"\>", 
                    RowBox[{"-", "2.7875540936574046`"}]},
                    {"\<\"vw\"\>", 
                    RowBox[{"-", "0.03200045295888203`"}]},
                    {"\<\"controllerState\"\>", 
                    RowBox[{"\[LeftAssociation]", 
                    RowBox[{
                    
                    RowBox[{"\<\"lastError\"\>", "\[Rule]", 
                    "0.003930002838550375`"}], ",", 
                    RowBox[{"\<\"lastInt\"\>", "\[Rule]", 
                    RowBox[{"-", "0.027389370519132122`"}]}], ",", 
                    RowBox[{"\<\"initialized\"\>", "\[Rule]", "True"}]}], 
                    "\[RightAssociation]"}]}
                    },
                    AutoDelete->False,
                    
                    GridBoxItemSize->{
                    "Columns" -> {{Automatic}}, "Rows" -> {{Automatic}}}],
                    "Grid"], {-1, -0.5}], 
                  {GrayLevel[0], LineBox[{{-5, 0}, {5, 0}}]}, 
                  {GrayLevel[0.5], 
                   CircleBox[{-1.7955065538162516`, 0.5}, 0.5], 
                   DiskBox[{-1.7955065538162516`, 0.5}, 0.1]}, 
                  {GrayLevel[0.5], 
                   DiskBox[{-1.7870004899266858`, 1.9999758820984779`}, 0.1]}, 
                  {GrayLevel[0.5], Thickness[0.01], 
                   LineBox[{{-1.7955065538162516`, 
                    0.5}, {-1.7870004899266858`, 1.9999758820984779`}}]}, 
                  {RGBColor[0, 1, 0], 
                   DiskBox[{-1.7898358445565412`, 1.4999839213989852`}, 
                    0.05]}, 
                  {RGBColor[0, 0, 1], Arrowheads[0.01], 
                   CircleBox[{-1.7955065538162516`, 0.5}, 0.3333333333333333, 
                    NCache[{
                    Rational[-1, 2] 
                    Pi, -1.6544229496046188`}, {-1.5707963267948966`, \
-1.6544229496046188`}]], 
                   ArrowBox[{{-1.8200265940366476`, 
                    0.16756973853347634`}, {-1.823349615254752, 
                    0.16783155622358858`}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-1.7955065538162516`, 
                    0.5}, {-1.7397554719431034`, 0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-1.7955065538162516`, 0.5}, {-1.837213878931334,
                     0.5}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.005], Arrowheads[0.04], 
                   ArrowBox[{{-1.7955065538162516`, 
                    0.5}, {-1.7814627970581858`, 0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-1.7955065538162516`, 0.5}, 1.6, {
                    1.5651255871426335`, 1.539058089816818}], 
                   ArrowBox[{{-1.7607281138997948`, 
                    2.0996219741291933`}, {-1.744733899668337, 
                    2.099194215094207}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Dashing[{Small, Small}], 
                   Arrowheads[0.02], 
                   CircleBox[{-1.7955065538162516`, 0.5}, 1.7, {
                    1.5651255871426335`, 1.5732644025286662`}], 
                   ArrowBox[{{-1.7827024036274517`, 
                    2.1999517798272814`}, {-1.7997022783040189`, 
                    2.1999948223144745`}}]}, 
                  {RGBColor[0, 0, 1], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-1.7955065538162516`, 0.5}, 1.8, {
                    1.5651255871426335`, 1.5806119987740634`}], 
                   ArrowBox[{{-1.7951747633806308`, 
                    2.299999969420863}, {-1.8131744796657279`, 
                    2.2999132880214477`}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.003], Arrowheads[0.025], 
                   CircleBox[{-1.7955065538162516`, 0.5}, 1.5, {
                    1.5651255871426335`, 1.5651279413903167`}], 
                   ArrowBox[{{-1.7720049373380775`, 
                    1.9998158800409178`}, {-1.7870040212414549`, 
                    1.9999759021197023`}}]}, 
                  {RGBColor[0, 0, 1], ArrowBox[{{0, 2.8}, {0, 2.8}}]}},
                 Frame->True,
                 FrameTicks->Automatic,
                 ImageSize->800,
                 PlotRange->{{-3, 3}, {-1, 3}}], 505->
                GraphicsBox[{InsetBox[
                   TagBox[GridBox[{
                    {"\<\"t\"\>", "25.2`"},
                    {"\<\"driveIntent\"\>", "0"},
                    {"\<\"motorTorque\"\>", 
                    RowBox[{"-", "2.768394436283057`"}]},
                    {"\<\"vw\"\>", 
                    RowBox[{"-", "0.03177062563500819`"}]},
                    {"\<\"controllerState\"\>", 
                    RowBox[{"\[LeftAssociation]", 
                    RowBox[{
                    
                    RowBox[{"\<\"lastError\"\>", "\[Rule]", 
                    "0.003899934355257142`"}], ",", 
                    RowBox[{"\<\"lastInt\"\>", "\[Rule]", 
                    RowBox[{"-", "0.027193639332933325`"}]}], ",", 
                    RowBox[{"\<\"initialized\"\>", "\[Rule]", "True"}]}], 
                    "\[RightAssociation]"}]}
                    },
                    AutoDelete->False,
                    
                    GridBoxItemSize->{
                    "Columns" -> {{Automatic}}, "Rows" -> {{Automatic}}}],
                    "Grid"], {-1, -0.5}], 
                  {GrayLevel[0], LineBox[{{-5, 0}, {5, 0}}]}, 
                  {GrayLevel[0.5], 
                   CircleBox[{-1.7971007092212083`, 0.5}, 0.5], 
                   DiskBox[{-1.7971007092212083`, 0.5}, 0.1]}, 
                  {GrayLevel[0.5], 
                   DiskBox[{-1.7886529798778978`, 1.9999762117676874`}, 0.1]}, 
                  {GrayLevel[0.5], Thickness[0.01], 
                   LineBox[{{-1.7971007092212083`, 
                    0.5}, {-1.7886529798778978`, 1.9999762117676874`}}]}, 
                  {RGBColor[0, 1, 0], 
                   DiskBox[{-1.7914688896590014`, 1.4999841411784582`}, 
                    0.05]}, 
                  {RGBColor[0, 0, 1], Arrowheads[0.01], 
                   CircleBox[{-1.7971007092212083`, 0.5}, 0.3333333333333333, 
                    NCache[{
                    Rational[-1, 2] 
                    Pi, -1.6538481598833883`}, {-1.5707963267948966`, \
-1.6538481598833883`}]], 
                   ArrowBox[{{-1.821429667904304, 
                    0.16755569958185879`}, {-1.824752839063582, 
                    0.16781560719038646`}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-1.7971007092212083`, 
                    0.5}, {-1.7417328204955471`, 0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-1.7971007092212083`, 
                    0.5}, {-1.8385220147999854`, 0.5}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.005], Arrowheads[0.04], 
                   ArrowBox[{{-1.7971007092212083`, 
                    0.5}, {-1.7831541260743242`, 0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-1.7971007092212083`, 0.5}, 1.6, {
                    1.5651644774611602`, 1.5392757509097303`}], 
                   ArrowBox[{{-1.7626704455929758`, 
                    2.0996295061502495`}, {-1.7466761386338934`, 
                    2.099205228443518}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Dashing[{Small, Small}], 
                   Arrowheads[0.02], 
                   CircleBox[{-1.7971007092212083`, 0.5}, 1.7, {
                    1.5651644774611602`, 1.5732469793838424`}], 
                   ArrowBox[{{-1.7842669405283, 
                    2.199951556480695}, {-1.8012668144523516`, 
                    2.1999948951591595`}}]}, 
                  {RGBColor[0, 0, 1], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-1.7971007092212083`, 0.5}, 1.8, {
                    1.5651644774611602`, 1.5805444465516216`}], 
                   ArrowBox[{{-1.7966473247881074`, 
                    2.299999942900709}, {-1.8146470468876568`, 
                    2.2999144774223286`}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.003], Arrowheads[0.025], 
                   CircleBox[{-1.7971007092212083`, 0.5}, 1.5, {
                    1.5651644774611602`, 1.5651659675605618`}], 
                   ArrowBox[{{-1.7736561250140028`, 
                    1.9998167726330278`}, {-1.7886552149915627`, 
                    1.9999762243539785`}}]}, 
                  {RGBColor[0, 0, 1], ArrowBox[{{0, 2.8}, {0, 2.8}}]}},
                 Frame->True,
                 FrameTicks->Automatic,
                 ImageSize->800,
                 PlotRange->{{-3, 3}, {-1, 3}}], 506->
                GraphicsBox[{InsetBox[
                   TagBox[GridBox[{
                    {"\<\"t\"\>", "25.25`"},
                    {"\<\"driveIntent\"\>", "0"},
                    {"\<\"motorTorque\"\>", 
                    RowBox[{"-", "2.7492146841332716`"}]},
                    {"\<\"vw\"\>", 
                    RowBox[{"-", "0.03154239502408567`"}]},
                    {"\<\"controllerState\"\>", 
                    RowBox[{"\[LeftAssociation]", 
                    RowBox[{
                    
                    RowBox[{"\<\"lastError\"\>", "\[Rule]", 
                    "0.003870385050405107`"}], ",", 
                    RowBox[{"\<\"lastInt\"\>", "\[Rule]", 
                    RowBox[{"-", "0.02699939823743933`"}]}], ",", 
                    RowBox[{"\<\"initialized\"\>", "\[Rule]", "True"}]}], 
                    "\[RightAssociation]"}]}
                    },
                    AutoDelete->False,
                    
                    GridBoxItemSize->{
                    "Columns" -> {{Automatic}}, "Rows" -> {{Automatic}}}],
                    "Grid"], {-1, -0.5}], 
                  {GrayLevel[0], LineBox[{{-5, 0}, {5, 0}}]}, 
                  {GrayLevel[0.5], CircleBox[{-1.798683413968616, 0.5}, 0.5], 
                   DiskBox[{-1.798683413968616, 0.5}, 0.1]}, 
                  {GrayLevel[0.5], 
                   DiskBox[{-1.7902940787501738`, 1.9999765395013993`}, 0.1]}, 
                  {GrayLevel[0.5], Thickness[0.01], 
                   LineBox[{{-1.798683413968616, 0.5}, {-1.7902940787501738`, 
                    1.9999765395013993`}}]}, 
                  {RGBColor[0, 1, 0], 
                   DiskBox[{-1.793090523822988, 1.4999843596675995`}, 0.05]}, 
                  {RGBColor[0, 0, 1], Arrowheads[0.01], 
                   CircleBox[{-1.798683413968616, 0.5}, 0.3333333333333333, 
                    NCache[{
                    Rational[-1, 2] 
                    Pi, -1.6532727673188947`}, {-1.5707963267948966`, \
-1.6532727673188947`}]], 
                   ArrowBox[{{-1.8228210826563236`, 
                    0.1675417559129665}, {-1.8261444028143865`, 
                    0.16779975135059944`}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-1.798683413968616, 0.5}, {-1.7436991202859506`,
                     0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-1.798683413968616, 0.5}, {-1.8398184077035464`,
                     0.5}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.005], Arrowheads[0.04], 
                   ArrowBox[{{-1.798683413968616, 0.5}, {-1.784834114020881, 
                    0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-1.798683413968616, 0.5}, 1.6, {
                    1.5652034074908656`, 1.5394936342971355`}], 
                   ArrowBox[{{-1.7646016838502758`, 
                    2.0996369699629165`}, {-1.7486072848277807`, 
                    2.099216177160008}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Dashing[{Small, Small}], 
                   Arrowheads[0.02], 
                   CircleBox[{-1.798683413968616, 0.5}, 1.7, {
                    1.5652034074908656`, 1.573229532404476}], 
                   ArrowBox[{{-1.785819986257933, 
                    2.199951332311467}, {-1.802819859423268, 
                    2.1999949675863752`}}]}, 
                  {RGBColor[0, 0, 1], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-1.798683413968616, 0.5}, 1.8, {
                    1.5652034074908656`, 1.5804768224027173`}], 
                   ArrowBox[{{-1.798108306072478, 
                    2.29999990812525}, {-1.816108033910401, 
                    2.2999156598629513`}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.003], Arrowheads[0.025], 
                   CircleBox[{-1.798683413968616, 0.5}, 1.5, {
                    1.5652034074908656`, 1.565204022213867}], 
                   ArrowBox[{{-1.7752959047856764`, 
                    1.9998176637225666`}, {-1.7902950008202556`, 
                    1.9999765446582332`}}]}, 
                  {RGBColor[0, 0, 1], ArrowBox[{{0, 2.8}, {0, 2.8}}]}},
                 Frame->True,
                 FrameTicks->Automatic,
                 ImageSize->800,
                 PlotRange->{{-3, 3}, {-1, 3}}], 507->
                GraphicsBox[{InsetBox[
                   TagBox[GridBox[{
                    {"\<\"t\"\>", "25.3`"},
                    {"\<\"driveIntent\"\>", "0"},
                    {"\<\"motorTorque\"\>", 
                    RowBox[{"-", "2.7300268233286396`"}]},
                    {"\<\"vw\"\>", 
                    RowBox[{"-", "0.031315762374175375`"}]},
                    {"\<\"controllerState\"\>", 
                    RowBox[{"\[LeftAssociation]", 
                    RowBox[{
                    
                    RowBox[{"\<\"lastError\"\>", "\[Rule]", 
                    "0.0038413319880041766`"}], ",", 
                    RowBox[{"\<\"lastInt\"\>", "\[Rule]", 
                    RowBox[{"-", "0.026806621856514234`"}]}], ",", 
                    RowBox[{"\<\"initialized\"\>", "\[Rule]", "True"}]}], 
                    "\[RightAssociation]"}]}
                    },
                    AutoDelete->False,
                    
                    GridBoxItemSize->{
                    "Columns" -> {{Automatic}}, "Rows" -> {{Automatic}}}],
                    "Grid"], {-1, -0.5}], 
                  {GrayLevel[0], LineBox[{{-5, 0}, {5, 0}}]}, 
                  {GrayLevel[0.5], 
                   CircleBox[{-1.8002547479306155`, 0.5}, 0.5], 
                   DiskBox[{-1.8002547479306155`, 0.5}, 0.1]}, 
                  {GrayLevel[0.5], 
                   DiskBox[{-1.7919238314356083`, 1.9999768650983767`}, 0.1]}, 
                  {GrayLevel[0.5], Thickness[0.01], 
                   LineBox[{{-1.8002547479306155`, 
                    0.5}, {-1.7919238314356083`, 1.9999768650983767`}}]}, 
                  {RGBColor[0, 1, 0], 
                   DiskBox[{-1.7947008036006107`, 1.499984576732251}, 0.05]}, 
                  {RGBColor[0, 0, 1], Arrowheads[0.01], 
                   CircleBox[{-1.8002547479306155`, 0.5}, 0.3333333333333333, 
                    NCache[{
                    Rational[-1, 2] 
                    Pi, -1.6526971314947558`}, {-1.5707963267948966`, \
-1.6526971314947558`}]], 
                   ArrowBox[{{-1.8242010377544773`, 
                    0.1675279164880415}, {-1.8275245058733462`, 
                    0.1677839988608978}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-1.8002547479306155`, 
                    0.5}, {-1.7456542114640428`, 0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-1.8002547479306155`, 0.5}, {-1.841103309025484,
                     0.5}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.005], Arrowheads[0.04], 
                   ArrowBox[{{-1.8002547479306155`, 
                    0.5}, {-1.7865027725589113`, 0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-1.8002547479306155`, 0.5}, 1.6, {
                    1.5652423539113922`, 1.5397116094595917`}], 
                   ArrowBox[{{-1.7665216997475783`, 
                    2.0996443609316042`}, {-1.7505272093826765`, 
                    2.0992270545203904`}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Dashing[{Small, Small}], 
                   Arrowheads[0.02], 
                   CircleBox[{-1.8002547479306155`, 0.5}, 1.7, {
                    1.5652423539113922`, 1.5732120777998735`}], 
                   ArrowBox[{{-1.7873616482435442`, 
                    2.1999511075264664`}, {-1.8043615206446533`, 
                    2.1999950395274324`}}]}, 
                  {RGBColor[0, 0, 1], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-1.8002547479306155`, 0.5}, 1.8, {
                    1.5652423539113922`, 1.5804091695965512`}], 
                   ArrowBox[{{-1.7995578649910031`, 
                    2.299999865098375}, {-1.817557598487366, 
                    2.299916834568367}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.003], Arrowheads[0.025], 
                   CircleBox[{-1.8002547479306155`, 0.5}, 1.5, {
                    1.5652423539113922`, 1.5652420923919417`}], 
                   ArrowBox[{{-1.8069233743587525`, 
                    1.9999851764006076`}, {-1.791923439162483, 
                    1.9999768629196286`}}]}, 
                  {RGBColor[0, 0, 1], ArrowBox[{{0, 2.8}, {0, 2.8}}]}},
                 Frame->True,
                 FrameTicks->Automatic,
                 ImageSize->800,
                 PlotRange->{{-3, 3}, {-1, 3}}], 508->
                GraphicsBox[{InsetBox[
                   TagBox[GridBox[{
                    {"\<\"t\"\>", "25.35`"},
                    {"\<\"driveIntent\"\>", "0"},
                    {"\<\"motorTorque\"\>", 
                    RowBox[{"-", "2.710842792936903`"}]},
                    {"\<\"vw\"\>", 
                    RowBox[{"-", "0.031090727821504367`"}]},
                    {"\<\"controllerState\"\>", 
                    RowBox[{"\[LeftAssociation]", 
                    RowBox[{
                    
                    RowBox[{"\<\"lastError\"\>", "\[Rule]", 
                    "0.0038127518674660197`"}], ",", 
                    RowBox[{"\<\"lastInt\"\>", "\[Rule]", 
                    RowBox[{"-", "0.026615285971229065`"}]}], ",", 
                    RowBox[{"\<\"initialized\"\>", "\[Rule]", "True"}]}], 
                    "\[RightAssociation]"}]}
                    },
                    AutoDelete->False,
                    
                    GridBoxItemSize->{
                    "Columns" -> {{Automatic}}, "Rows" -> {{Automatic}}}],
                    "Grid"], {-1, -0.5}], 
                  {GrayLevel[0], LineBox[{{-5, 0}, {5, 0}}]}, 
                  {GrayLevel[0.5], 
                   CircleBox[{-1.8018147910133184`, 0.5}, 0.5], 
                   DiskBox[{-1.8018147910133184`, 0.5}, 0.1]}, 
                  {GrayLevel[0.5], 
                   DiskBox[{-1.7935422828216534`, 1.999977188362616}, 0.1]}, 
                  {GrayLevel[0.5], Thickness[0.01], 
                   LineBox[{{-1.8018147910133184`, 
                    0.5}, {-1.7935422828216534`, 1.999977188362616}}]}, 
                  {RGBColor[0, 1, 0], 
                   DiskBox[{-1.7962997855522085`, 1.499984792241744}, 0.05]}, 
                  {RGBColor[0, 0, 1], Arrowheads[0.01], 
                   CircleBox[{-1.8018147910133184`, 0.5}, 0.3333333333333333, 
                    NCache[{
                    Rational[-1, 2] 
                    Pi, -1.6521216105830037`}, {-1.5707963267948966`, \
-1.6521216105830037`}]], 
                   ArrowBox[{{-1.8255697322453115`, 
                    0.16751418995966816`}, {-1.8288933471945266`, 
                    0.16776835956481773`}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-1.8018147910133184`, 
                    0.5}, {-1.7475979351545803`, 0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-1.8018147910133184`, 0.5}, {-1.842376970373015,
                     0.5}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.005], Arrowheads[0.04], 
                   ArrowBox[{{-1.8018147910133184`, 
                    0.5}, {-1.7881601145142771`, 0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-1.8018147910133184`, 0.5}, 1.6, {
                    1.5652812933766598`, 1.5399295457336}], 
                   ArrowBox[{{-1.7684303641603776`, 
                    2.0996516745978484`}, {-1.7524357832291109`, 
                    2.0992378539761516`}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Dashing[{Small, Small}], 
                   Arrowheads[0.02], 
                   CircleBox[{-1.8018147910133184`, 0.5}, 1.7, {
                    1.5652812933766598`, 1.5731946310983576`}], 
                   ArrowBox[{{-1.7888920327886462`, 
                    2.1999508823256826`}, {-1.8058919044206994`, 
                    2.199995110918341}}]}, 
                  {RGBColor[0, 0, 1], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-1.8018147910133184`, 0.5}, 1.8, {
                    1.5652812933766598`, 1.5803415311151983`}], 
                   ArrowBox[{{-1.8009961588180823`, 
                    2.2999998138448037`}, {-1.81899589788933, 
                    2.2999180007896234`}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.003], Arrowheads[0.025], 
                   CircleBox[{-1.8018147910133184`, 0.5}, 1.5, {
                    1.5652812933766598`, 1.5652801639282325`}], 
                   ArrowBox[{{-1.8085405241766879`, 
                    1.999984921428684}, {-1.7935405886747822`, 
                    1.999977179018288}}]}, 
                  {RGBColor[0, 0, 1], ArrowBox[{{0, 2.8}, {0, 2.8}}]}},
                 Frame->True,
                 FrameTicks->Automatic,
                 ImageSize->800,
                 PlotRange->{{-3, 3}, {-1, 3}}], 509->
                GraphicsBox[{InsetBox[
                   TagBox[GridBox[{
                    {"\<\"t\"\>", "25.400000000000002`"},
                    {"\<\"driveIntent\"\>", "0"},
                    {"\<\"motorTorque\"\>", 
                    RowBox[{"-", "2.691674366993208`"}]},
                    {"\<\"vw\"\>", 
                    RowBox[{"-", "0.03086729041364603`"}]},
                    {"\<\"controllerState\"\>", 
                    RowBox[{"\[LeftAssociation]", 
                    RowBox[{
                    
                    RowBox[{"\<\"lastError\"\>", "\[Rule]", 
                    "0.0037846212760145465`"}], ",", 
                    RowBox[{"\<\"lastInt\"\>", "\[Rule]", 
                    RowBox[{"-", "0.026425367531537968`"}]}], ",", 
                    RowBox[{"\<\"initialized\"\>", "\[Rule]", "True"}]}], 
                    "\[RightAssociation]"}]}
                    },
                    AutoDelete->False,
                    
                    GridBoxItemSize->{
                    "Columns" -> {{Automatic}}, "Rows" -> {{Automatic}}}],
                    "Grid"], {-1, -0.5}], 
                  {GrayLevel[0], LineBox[{{-5, 0}, {5, 0}}]}, 
                  {GrayLevel[0.5], 
                   CircleBox[{-1.8033636231017758`, 0.5}, 0.5], 
                   DiskBox[{-1.8033636231017758`, 0.5}, 0.1]}, 
                  {GrayLevel[0.5], 
                   DiskBox[{-1.7951494781034207`, 1.9999775091053684`}, 0.1]}, 
                  {GrayLevel[0.5], Thickness[0.01], 
                   LineBox[{{-1.8033636231017758`, 
                    0.5}, {-1.7951494781034207`, 1.9999775091053684`}}]}, 
                  {RGBColor[0, 1, 0], 
                   DiskBox[{-1.7978875264362058`, 1.4999850060702458`}, 
                    0.05]}, 
                  {RGBColor[0, 0, 1], Arrowheads[0.01], 
                   CircleBox[{-1.8033636231017758`, 0.5}, 0.3333333333333333, 
                    NCache[{
                    Rational[-1, 2] 
                    Pi, -1.6515465578046928`}, {-1.5707963267948966`, \
-1.6515465578046928`}]], 
                   ArrowBox[{{-1.8269273635277845`, 
                    0.16750058458961647`}, {-1.830251124088393, 
                    0.16775284289883569`}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-1.8033636231017758`, 
                    0.5}, {-1.7495301357619115`, 0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-1.8033636231017758`, 
                    0.5}, {-1.8436396417209866`, 0.5}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.005], Arrowheads[0.04], 
                   ArrowBox[{{-1.8033636231017758`, 
                    0.5}, {-1.7898061543811226`, 0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-1.8033636231017758`, 0.5}, 1.6, {
                    1.5653202027597595`, 1.5401473136822221`}], 
                   ArrowBox[{{-1.7703275499013615`, 
                    2.0996589067258986`}, {-1.7543328792324824`, 
                    2.099248569221065}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Dashing[{Small, Small}], 
                   Arrowheads[0.02], 
                   CircleBox[{-1.8033636231017758`, 0.5}, 1.7, {
                    1.5653202027597595`, 1.5731772070684933`}], 
                   ArrowBox[{{-1.7904112448841254`, 
                    2.1999506569011076`}, {-1.8074111157429575`, 
                    2.1999951817000305`}}]}, 
                  {RGBColor[0, 0, 1], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-1.8033636231017758`, 0.5}, 1.8, {
                    1.5653202027597595`, 1.580273949243055}], 
                   ArrowBox[{{-1.8024233435512247`, 
                    2.2999997544095296`}, {-1.8204230881104466`, 
                    2.2999191578105993`}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.003], Arrowheads[0.025], 
                   CircleBox[{-1.8033636231017758`, 0.5}, 1.5, {
                    1.5653202027597595`, 1.5653182217402393`}], 
                   ArrowBox[{{-1.8101464424044134`, 
                    1.9999846643757089`}, {-1.7951465066187113`, 
                    1.9999774928300438`}}]}, 
                  {RGBColor[0, 0, 1], ArrowBox[{{0, 2.8}, {0, 2.8}}]}},
                 Frame->True,
                 FrameTicks->Automatic,
                 ImageSize->800,
                 PlotRange->{{-3, 3}, {-1, 3}}], 510->
                GraphicsBox[{InsetBox[
                   TagBox[GridBox[{
                    {"\<\"t\"\>", "25.45`"},
                    {"\<\"driveIntent\"\>", "0"},
                    {"\<\"motorTorque\"\>", 
                    RowBox[{"-", "2.6725330516668255`"}]},
                    {"\<\"vw\"\>", 
                    RowBox[{"-", "0.030645448141218567`"}]},
                    {"\<\"controllerState\"\>", 
                    RowBox[{"\[LeftAssociation]", 
                    RowBox[{
                    
                    RowBox[{"\<\"lastError\"\>", "\[Rule]", 
                    "0.003756916913375097`"}], ",", 
                    RowBox[{"\<\"lastInt\"\>", "\[Rule]", 
                    RowBox[{"-", "0.026236844656044443`"}]}], ",", 
                    RowBox[{"\<\"initialized\"\>", "\[Rule]", "True"}]}], 
                    "\[RightAssociation]"}]}
                    },
                    AutoDelete->False,
                    
                    GridBoxItemSize->{
                    "Columns" -> {{Automatic}}, "Rows" -> {{Automatic}}}],
                    "Grid"], {-1, -0.5}], 
                  {GrayLevel[0], LineBox[{{-5, 0}, {5, 0}}]}, 
                  {GrayLevel[0.5], 
                   CircleBox[{-1.8049013240063152`, 0.5}, 0.5], 
                   DiskBox[{-1.8049013240063152`, 0.5}, 0.1]}, 
                  {GrayLevel[0.5], 
                   DiskBox[{-1.7967454630519684`, 1.9999778271468194`}, 0.1]}, 
                  {GrayLevel[0.5], Thickness[0.01], 
                   LineBox[{{-1.8049013240063152`, 
                    0.5}, {-1.7967454630519684`, 1.9999778271468194`}}]}, 
                  {RGBColor[0, 1, 0], 
                   DiskBox[{-1.799464083370084, 1.4999852180978794`}, 0.05]}, 
                  {RGBColor[0, 0, 1], Arrowheads[0.01], 
                   CircleBox[{-1.8049013240063152`, 0.5}, 0.3333333333333333, 
                    NCache[{
                    Rational[-1, 2] 
                    Pi, -1.6509723183449014`}, {-1.5707963267948966`, \
-1.6509723183449014`}]], 
                   ArrowBox[{{-1.8282741262730482`, 
                    0.16748710818178597`}, {-1.8315980311423175`, 
                    0.16773745781505012`}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-1.8049013240063152`, 
                    0.5}, {-1.7514506629729787`, 0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-1.8049013240063152`, 
                    0.5}, {-1.8448915697796509`, 0.5}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.005], Arrowheads[0.04], 
                   ArrowBox[{{-1.8049013240063152`, 
                    0.5}, {-1.7914409087463143`, 0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-1.8049013240063152`, 0.5}, 1.6, {
                    1.5653590593675872`, 1.5403647862963543`}], 
                   ArrowBox[{{-1.7722131335885338`, 
                    2.099666053339637}, {-1.7562183740607142`, 
                    2.099259194247322}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Dashing[{Small, Small}], 
                   Arrowheads[0.02], 
                   CircleBox[{-1.8049013240063152`, 0.5}, 1.7, {
                    1.5653590593675872`, 1.5731598196566539`}], 
                   ArrowBox[{{-1.7919193880484459`, 
                    2.1999504314358065`}, {-1.8089192581305373`, 
                    2.199995251818479}}]}, 
                  {RGBColor[0, 0, 1], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-1.8049013240063152`, 0.5}, 1.8, {
                    1.5653590593675872`, 1.5802064652101808`}], 
                   ArrowBox[{{-1.8038395732153971`, 
                    2.2999996868569887`}, {-1.8218393231726164`, 
                    2.299920304953595}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.003], Arrowheads[0.025], 
                   CircleBox[{-1.8049013240063152`, 0.5}, 1.5, {
                    1.5653590593675872`, 1.565356250096993}], 
                   ArrowBox[{{-1.8117411852559766`, 
                    1.999984405351631}, {-1.7967412492083987`, 
                    1.99997780422888}}]}, 
                  {RGBColor[0, 0, 1], ArrowBox[{{0, 2.8}, {0, 2.8}}]}},
                 Frame->True,
                 FrameTicks->Automatic,
                 ImageSize->800,
                 PlotRange->{{-3, 3}, {-1, 3}}], 511->
                GraphicsBox[{InsetBox[
                   TagBox[GridBox[{
                    {"\<\"t\"\>", "25.5`"},
                    {"\<\"driveIntent\"\>", "0"},
                    {"\<\"motorTorque\"\>", 
                    RowBox[{"-", "2.6534299968255617`"}]},
                    {"\<\"vw\"\>", 
                    RowBox[{"-", "0.030425197976818228`"}]},
                    {"\<\"controllerState\"\>", 
                    RowBox[{"\[LeftAssociation]", 
                    RowBox[{
                    
                    RowBox[{"\<\"lastError\"\>", "\[Rule]", 
                    "0.0037296157897607927`"}], ",", 
                    RowBox[{"\<\"lastInt\"\>", "\[Rule]", 
                    RowBox[{"-", "0.026049696621218796`"}]}], ",", 
                    RowBox[{"\<\"initialized\"\>", "\[Rule]", "True"}]}], 
                    "\[RightAssociation]"}]}
                    },
                    AutoDelete->False,
                    
                    GridBoxItemSize->{
                    "Columns" -> {{Automatic}}, "Rows" -> {{Automatic}}}],
                    "Grid"], {-1, -0.5}], 
                  {GrayLevel[0], LineBox[{{-5, 0}, {5, 0}}]}, 
                  {GrayLevel[0.5], CircleBox[{-1.80642797341066, 0.5}, 0.5], 
                   DiskBox[{-1.80642797341066, 0.5}, 0.1]}, 
                  {GrayLevel[0.5], 
                   DiskBox[{-1.798330284241093, 1.9999781423174516`}, 0.1]}, 
                  {GrayLevel[0.5], Thickness[0.01], 
                   LineBox[{{-1.80642797341066, 0.5}, {-1.798330284241093, 
                    1.9999781423174516`}}]}, 
                  {RGBColor[0, 1, 0], 
                   DiskBox[{-1.801029513964282, 1.4999854282116343`}, 0.05]}, 
                  {RGBColor[0, 0, 1], Arrowheads[0.01], 
                   CircleBox[{-1.80642797341066, 0.5}, 0.3333333333333333, 
                    NCache[{
                    Rational[-1, 2] 
                    Pi, -1.6503992266996634`}, {-1.5707963267948966`, \
-1.6503992266996634`}]], 
                   ArrowBox[{{-1.8296102114893775`, 
                    0.16747376802908803`}, {-1.8329342592860804`, 
                    0.1677222127192346}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-1.80642797341066, 0.5}, {-1.7533593734741488`, 
                    0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-1.80642797341066, 0.5}, {-1.846132996576402, 
                    0.5}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.005], Arrowheads[0.04], 
                   ArrowBox[{{-1.80642797341066, 0.5}, {-1.7930643966398907`, 
                    0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-1.80642797341066, 0.5}, 1.6, {
                    1.5653978411266294`, 1.5405818400345248`}], 
                   ArrowBox[{{-1.7740869972568625`, 
                    2.099673110751512}, {-1.7580921497955317`, 
                    2.0992697233911004`}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Dashing[{Small, Small}], 
                   Arrowheads[0.02], 
                   CircleBox[{-1.80642797341066, 0.5}, 1.7, {
                    1.5653978411266294`, 1.573142481939644}], 
                   ArrowBox[{{-1.7934165641952324`, 
                    2.1999502061031757`}, {-1.8104164334976858`, 
                    2.199995321224778}}]}, 
                  {RGBColor[0, 0, 1], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-1.80642797341066, 0.5}, 1.8, {
                    1.5653978411266294`, 1.5801391188867715`}], 
                   ArrowBox[{{-1.8052449992611936`, 
                    2.299999611270003}, {-1.823244754523673, 
                    2.29992144158377}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.003], Arrowheads[0.025], 
                   CircleBox[{-1.80642797341066, 0.5}, 1.5, {
                    1.5653978411266294`, 1.5653942328619712`}], 
                   ArrowBox[{{-1.8133248082105133`, 
                    1.999984144472782}, {-1.7983248719230265`, 
                    1.9999781130890812`}}]}, 
                  {RGBColor[0, 0, 1], ArrowBox[{{0, 2.8}, {0, 2.8}}]}},
                 Frame->True,
                 FrameTicks->Automatic,
                 ImageSize->800,
                 PlotRange->{{-3, 3}, {-1, 3}}], 512->
                GraphicsBox[{InsetBox[
                   TagBox[GridBox[{
                    {"\<\"t\"\>", "25.55`"},
                    {"\<\"driveIntent\"\>", "0"},
                    {"\<\"motorTorque\"\>", 
                    RowBox[{"-", "2.6343759212025186`"}]},
                    {"\<\"vw\"\>", 
                    RowBox[{"-", "0.030206535919980522`"}]},
                    {"\<\"controllerState\"\>", 
                    RowBox[{"\[LeftAssociation]", 
                    RowBox[{
                    
                    RowBox[{"\<\"lastError\"\>", "\[Rule]", 
                    "0.003702695398313567`"}], ",", 
                    RowBox[{"\<\"lastInt\"\>", "\[Rule]", 
                    RowBox[{"-", "0.02586390384137261`"}]}], ",", 
                    RowBox[{"\<\"initialized\"\>", "\[Rule]", "True"}]}], 
                    "\[RightAssociation]"}]}
                    },
                    AutoDelete->False,
                    
                    GridBoxItemSize->{
                    "Columns" -> {{Automatic}}, "Rows" -> {{Automatic}}}],
                    "Grid"], {-1, -0.5}], 
                  {GrayLevel[0], LineBox[{{-5, 0}, {5, 0}}]}, 
                  {GrayLevel[0.5], CircleBox[{-1.807943650822152, 0.5}, 0.5], 
                   DiskBox[{-1.807943650822152, 0.5}, 0.1]}, 
                  {GrayLevel[0.5], 
                   DiskBox[{-1.7999039892352313`, 1.999978454459119}, 0.1]}, 
                  {GrayLevel[0.5], Thickness[0.01], 
                   LineBox[{{-1.807943650822152, 0.5}, {-1.7999039892352313`, 
                    1.999978454459119}}]}, 
                  {RGBColor[0, 1, 0], 
                   DiskBox[{-1.8025838764308717`, 1.4999856363060795`}, 
                    0.05]}, 
                  {RGBColor[0, 0, 1], Arrowheads[0.01], 
                   CircleBox[{-1.807943650822152, 0.5}, 0.3333333333333333, 
                    NCache[{
                    Rational[-1, 2] 
                    Pi, -1.649827604430972}, {-1.5707963267948966`, \
-1.649827604430972}]], 
                   ArrowBox[{{-1.830935805724681, 
                    0.16746057087309296`}, {-1.8342599949948237`, 
                    0.1677071154230103}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-1.807943650822152, 0.5}, {-1.7552561323981017`,
                     0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-1.807943650822152, 0.5}, {-1.8473641582406035`,
                     0.5}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.005], Arrowheads[0.04], 
                   ArrowBox[{{-1.807943650822152, 0.5}, {-1.7946766398165532`,
                     0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-1.807943650822152, 0.5}, 1.6, {
                    1.5654365267414159`, 1.540798355709736}], 
                   ArrowBox[{{-1.7759490297273968`, 
                    2.0996800755841782`}, {-1.7599540953012915`, 
                    2.099280151368393}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Dashing[{Small, Small}], 
                   Arrowheads[0.02], 
                   CircleBox[{-1.807943650822152, 0.5}, 1.7, {
                    1.5654365267414159`, 1.5731252060911727`}], 
                   ArrowBox[{{-1.794902873526498, 
                    2.199949981066362}, {-1.8119027420470126`, 
                    2.1999953898751237`}}]}, 
                  {RGBColor[0, 0, 1], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-1.807943650822152, 0.5}, 1.8, {
                    1.5654365267414159`, 1.5800719485258743`}], 
                   ArrowBox[{{-1.8066397700519414`, 
                    2.2999995277485317`}, {-1.8246395305245007`, 
                    2.299922567112531}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.003], Arrowheads[0.025], 
                   CircleBox[{-1.807943650822152, 0.5}, 1.5, {
                    1.5654365267414159`, 1.565432153712032}], 
                   ArrowBox[{{-1.8148973662908687`, 
                    1.9999838818604618`}, {-1.7998974297854518`, 
                    1.9999784192871004`}}]}, 
                  {RGBColor[0, 0, 1], ArrowBox[{{0, 2.8}, {0, 2.8}}]}},
                 Frame->True,
                 FrameTicks->Automatic,
                 ImageSize->800,
                 PlotRange->{{-3, 3}, {-1, 3}}], 513->
                GraphicsBox[{InsetBox[
                   TagBox[GridBox[{
                    {"\<\"t\"\>", "25.6`"},
                    {"\<\"driveIntent\"\>", "0"},
                    {"\<\"motorTorque\"\>", 
                    RowBox[{"-", "2.615381050320309`"}]},
                    {"\<\"vw\"\>", 
                    RowBox[{"-", "0.029989457047043585`"}]},
                    {"\<\"controllerState\"\>", 
                    RowBox[{"\[LeftAssociation]", 
                    RowBox[{
                    
                    RowBox[{"\<\"lastError\"\>", "\[Rule]", 
                    "0.0036761338632700646`"}], ",", 
                    RowBox[{"\<\"lastInt\"\>", "\[Rule]", 
                    RowBox[{"-", "0.02567944784063582`"}]}], ",", 
                    RowBox[{"\<\"initialized\"\>", "\[Rule]", "True"}]}], 
                    "\[RightAssociation]"}]}
                    },
                    AutoDelete->False,
                    
                    GridBoxItemSize->{
                    "Columns" -> {{Automatic}}, "Rows" -> {{Automatic}}}],
                    "Grid"], {-1, -0.5}], 
                  {GrayLevel[0], LineBox[{{-5, 0}, {5, 0}}]}, 
                  {GrayLevel[0.5], 
                   CircleBox[{-1.8094484355243514`, 0.5}, 0.5], 
                   DiskBox[{-1.8094484355243514`, 0.5}, 0.1]}, 
                  {GrayLevel[0.5], 
                   DiskBox[{-1.801466626741163, 1.9999787634258521`}, 0.1]}, 
                  {GrayLevel[0.5], Thickness[0.01], 
                   LineBox[{{-1.8094484355243514`, 0.5}, {-1.801466626741163, 
                    1.9999787634258521`}}]}, 
                  {RGBColor[0, 1, 0], 
                   DiskBox[{-1.8041272296688924`, 1.4999858422839014`}, 
                    0.05]}, 
                  {RGBColor[0, 0, 1], Arrowheads[0.01], 
                   CircleBox[{-1.8094484355243514`, 0.5}, 0.3333333333333333, 
                    NCache[{
                    Rational[-1, 2] 
                    Pi, -1.6492577583045058`}, {-1.5707963267948966`, \
-1.6492577583045058`}]], 
                   ArrowBox[{{-1.8322510903984917`, 
                    0.16744752287525816`}, {-1.8355754196213605`, 
                    0.16769217310886925`}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-1.8094484355243514`, 
                    0.5}, {-1.7571408145179452`, 0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-1.8094484355243514`, 
                    0.5}, {-1.8485852839799448`, 0.5}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.005], Arrowheads[0.04], 
                   ArrowBox[{{-1.8094484355243514`, 
                    0.5}, {-1.7962776629735386`, 0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-1.8094484355243514`, 0.5}, 1.6, {
                    1.5654750958272547`, 1.5410142192323626`}], 
                   ArrowBox[{{-1.7777991277485332`, 
                    2.099686944785545}, {-1.761804107366037, 
                    2.0992904733019397`}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Dashing[{Small, Small}], 
                   Arrowheads[0.02], 
                   CircleBox[{-1.8094484355243514`, 0.5}, 1.7, {
                    1.5654750958272547`, 1.5731080033608549`}], 
                   ArrowBox[{{-1.7963784144495505`, 
                    2.1999497564778507`}, {-1.813378282186394, 
                    2.199995457730759}}]}, 
                  {RGBColor[0, 0, 1], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-1.8094484355243514`, 0.5}, 1.8, {
                    1.5654750958272547`, 1.5800049905512565`}], 
                   ArrowBox[{{-1.8080240304344628`, 
                    2.299999436408284}, {-1.8260237960198005`, 
                    2.29992368099996}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.003], Arrowheads[0.025], 
                   CircleBox[{-1.8094484355243514`, 0.5}, 1.5, {
                    1.5654750958272547`, 1.5654699963329186`}], 
                   ArrowBox[{{-1.816458914309453, 
                    1.9999836176396073`}, {-1.8014589776080583`, 
                    1.99997872270316}}]}, 
                  {RGBColor[0, 0, 1], ArrowBox[{{0, 2.8}, {0, 2.8}}]}},
                 Frame->True,
                 FrameTicks->Automatic,
                 ImageSize->800,
                 PlotRange->{{-3, 3}, {-1, 3}}], 514->
                GraphicsBox[{InsetBox[
                   TagBox[GridBox[{
                    {"\<\"t\"\>", "25.650000000000002`"},
                    {"\<\"driveIntent\"\>", "0"},
                    {"\<\"motorTorque\"\>", 
                    RowBox[{"-", "2.596455066309305`"}]},
                    {"\<\"vw\"\>", 
                    RowBox[{"-", "0.029773955564869288`"}]},
                    {"\<\"controllerState\"\>", 
                    RowBox[{"\[LeftAssociation]", 
                    RowBox[{
                    
                    RowBox[{"\<\"lastError\"\>", "\[Rule]", 
                    "0.0036499100652096763`"}], ",", 
                    RowBox[{"\<\"lastInt\"\>", "\[Rule]", 
                    RowBox[{"-", "0.02549631121811556`"}]}], ",", 
                    RowBox[{"\<\"initialized\"\>", "\[Rule]", "True"}]}], 
                    "\[RightAssociation]"}]}
                    },
                    AutoDelete->False,
                    
                    GridBoxItemSize->{
                    "Columns" -> {{Automatic}}, "Rows" -> {{Automatic}}}],
                    "Grid"], {-1, -0.5}], 
                  {GrayLevel[0], LineBox[{{-5, 0}, {5, 0}}]}, 
                  {GrayLevel[0.5], 
                   CircleBox[{-1.8109424065322346`, 0.5}, 0.5], 
                   DiskBox[{-1.8109424065322346`, 0.5}, 0.1]}, 
                  {GrayLevel[0.5], 
                   DiskBox[{-1.803018246726238, 1.999979069084422}, 0.1]}, 
                  {GrayLevel[0.5], Thickness[0.01], 
                   LineBox[{{-1.8109424065322346`, 0.5}, {-1.803018246726238, 
                    1.999979069084422}}]}, 
                  {RGBColor[0, 1, 0], 
                   DiskBox[{-1.8056596333282369`, 1.4999860460562813`}, 
                    0.05]}, 
                  {RGBColor[0, 0, 1], Arrowheads[0.01], 
                   CircleBox[{-1.8109424065322346`, 0.5}, 0.3333333333333333, 
                    NCache[{
                    Rational[-1, 2] 
                    Pi, -1.6486899787841758`}, {-1.5707963267948966`, \
-1.6486899787841758`}]], 
                   ArrowBox[{{-1.8335562412550865`, 
                    0.16743462959857258`}, {-1.8368807088495025`, 
                    0.16767739230679912`}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-1.8109424065322346`, 
                    0.5}, {-1.7590133052060486`, 0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-1.8109424065322346`, 
                    0.5}, {-1.8497965952362452`, 0.5}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.005], Arrowheads[0.04], 
                   ArrowBox[{{-1.8109424065322346`, 
                    0.5}, {-1.7978674939100592`, 0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-1.8109424065322346`, 0.5}, 1.6, {
                    1.5655135290189217`, 1.54122932221846}], 
                   ArrowBox[{{-1.7796371969246032`, 
                    2.09969371563791}, {-1.7636420916299467`, 
                    2.0993006847400975`}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Dashing[{Small, Small}], 
                   Arrowheads[0.02], 
                   CircleBox[{-1.8109424065322346`, 0.5}, 1.7, {
                    1.5655135290189217`, 1.5730908840645312`}], 
                   ArrowBox[{{-1.7978432835157339`, 
                    2.1999495324791845`}, {-1.814843150467713, 
                    2.1999955247578593`}}]}, 
                  {RGBColor[0, 0, 1], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-1.8109424065322346`, 0.5}, 1.8, {
                    1.5655135290189217`, 1.5799382793873067`}], 
                   ArrowBox[{{-1.8093979213880926`, 
                    2.2999993373792225`}, {-1.827397691987109, 
                    2.2999247827563796`}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.003], Arrowheads[0.025], 
                   CircleBox[{-1.8109424065322346`, 0.5}, 1.5, {
                    1.5655135290189217`, 1.5655077445921821`}], 
                   ArrowBox[{{-1.8180095070828175`, 
                    1.99998335193755}, {-1.8030095702073345`, 
                    1.9999790232226058`}}]}, 
                  {RGBColor[0, 0, 1], ArrowBox[{{0, 2.8}, {0, 2.8}}]}},
                 Frame->True,
                 FrameTicks->Automatic,
                 ImageSize->800,
                 PlotRange->{{-3, 3}, {-1, 3}}], 515->
                GraphicsBox[{InsetBox[
                   TagBox[GridBox[{
                    {"\<\"t\"\>", "25.7`"},
                    {"\<\"driveIntent\"\>", "0"},
                    {"\<\"motorTorque\"\>", 
                    RowBox[{"-", "2.577607068739178`"}]},
                    {"\<\"vw\"\>", 
                    RowBox[{"-", "0.029560024867460812`"}]},
                    {"\<\"controllerState\"\>", 
                    RowBox[{"\[LeftAssociation]", 
                    RowBox[{
                    
                    RowBox[{"\<\"lastError\"\>", "\[Rule]", 
                    "0.0036240037448065053`"}], ",", 
                    RowBox[{"\<\"lastInt\"\>", "\[Rule]", 
                    RowBox[{"-", "0.025314477607346845`"}]}], ",", 
                    RowBox[{"\<\"initialized\"\>", "\[Rule]", "True"}]}], 
                    "\[RightAssociation]"}]}
                    },
                    AutoDelete->False,
                    
                    GridBoxItemSize->{
                    "Columns" -> {{Automatic}}, "Rows" -> {{Automatic}}}],
                    "Grid"], {-1, -0.5}], 
                  {GrayLevel[0], LineBox[{{-5, 0}, {5, 0}}]}, 
                  {GrayLevel[0.5], 
                   CircleBox[{-1.8124256425501495`, 0.5}, 0.5], 
                   DiskBox[{-1.8124256425501495`, 0.5}, 0.1]}, 
                  {GrayLevel[0.5], 
                   DiskBox[{-1.804558900505861, 1.9999793713146887`}, 0.1]}, 
                  {GrayLevel[0.5], Thickness[0.01], 
                   LineBox[{{-1.8124256425501495`, 0.5}, {-1.804558900505861, 
                    1.9999793713146887`}}]}, 
                  {RGBColor[0, 1, 0], 
                   DiskBox[{-1.8071811478539572`, 1.4999862475431258`}, 
                    0.05]}, 
                  {RGBColor[0, 0, 1], Arrowheads[0.01], 
                   CircleBox[{-1.8124256425501495`, 0.5}, 0.3333333333333333, 
                    NCache[{
                    Rational[-1, 2] 
                    Pi, -1.648124538857072}, {-1.5707963267948966`, \
-1.648124538857072}]], 
                   ArrowBox[{{-1.834851427929151, 
                    0.16742189599847968`}, {-1.8381760322598346`, 
                    0.16766277888128384`}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-1.8124256425501495`, 0.5}, {-1.760873501175366,
                     0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-1.8124256425501495`, 
                    0.5}, {-1.8509983050082877`, 0.5}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.005], Arrowheads[0.04], 
                   ArrowBox[{{-1.8124256425501495`, 
                    0.5}, {-1.7994461636335042`, 0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-1.8124256425501495`, 0.5}, 1.6, {
                    1.5655518080570097`, 1.5414435624730656`}], 
                   ArrowBox[{{-1.7814631524472624`, 
                    2.0997003857618557`}, {-1.765467963316646, 
                    2.0993107816684673`}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Dashing[{Small, Small}], 
                   Arrowheads[0.02], 
                   CircleBox[{-1.8124256425501495`, 0.5}, 1.7, {
                    1.5655518080570097`, 1.5730738575847014`}], 
                   ArrowBox[{{-1.7992975753791223`, 
                    2.1999493092008224`}, {-1.8162974415455508`, 
                    2.199995590927382}}]}, 
                  {RGBColor[0, 0, 1], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-1.8124256425501495`, 0.5}, 1.8, {
                    1.5655518080570097`, 1.5798718473277829`}], 
                   ArrowBox[{{-1.81076157974638, 
                    2.299999230803998}, {-1.828761355258495, 
                    2.2999258719431506`}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.003], Arrowheads[0.025], 
                   CircleBox[{-1.8124256425501495`, 0.5}, 1.5, {
                    1.5655518080570097`, 1.5655453826904477`}], 
                   ArrowBox[{{-1.8195491996165105`, 
                    1.999983084882867}, {-1.8045492625887274`, 
                    1.9999793207370238`}}]}, 
                  {RGBColor[0, 0, 1], ArrowBox[{{0, 2.8}, {0, 2.8}}]}},
                 Frame->True,
                 FrameTicks->Automatic,
                 ImageSize->800,
                 PlotRange->{{-3, 3}, {-1, 3}}], 516->
                GraphicsBox[{InsetBox[
                   TagBox[GridBox[{
                    {"\<\"t\"\>", "25.75`"},
                    {"\<\"driveIntent\"\>", "0"},
                    {"\<\"motorTorque\"\>", 
                    RowBox[{"-", "2.5588455455776744`"}]},
                    {"\<\"vw\"\>", 
                    RowBox[{"-", "0.02934765759459696`"}]},
                    {"\<\"controllerState\"\>", 
                    RowBox[{"\[LeftAssociation]", 
                    RowBox[{
                    
                    RowBox[{"\<\"lastError\"\>", "\[Rule]", 
                    "0.003598395586550165`"}], ",", 
                    RowBox[{"\<\"lastInt\"\>", "\[Rule]", 
                    RowBox[{"-", "0.02513393163107264`"}]}], ",", 
                    RowBox[{"\<\"initialized\"\>", "\[Rule]", "True"}]}], 
                    "\[RightAssociation]"}]}
                    },
                    AutoDelete->False,
                    
                    GridBoxItemSize->{
                    "Columns" -> {{Automatic}}, "Rows" -> {{Automatic}}}],
                    "Grid"], {-1, -0.5}], 
                  {GrayLevel[0], LineBox[{{-5, 0}, {5, 0}}]}, 
                  {GrayLevel[0.5], 
                   CircleBox[{-1.8138982219326671`, 0.5}, 0.5], 
                   DiskBox[{-1.8138982219326671`, 0.5}, 0.1]}, 
                  {GrayLevel[0.5], 
                   DiskBox[{-1.8060886408029615`, 1.9999796700097567`}, 0.1]}, 
                  {GrayLevel[0.5], Thickness[0.01], 
                   LineBox[{{-1.8138982219326671`, 
                    0.5}, {-1.8060886408029615`, 1.9999796700097567`}}]}, 
                  {RGBColor[0, 1, 0], 
                   DiskBox[{-1.8086918345128633`, 1.4999864466731712`}, 
                    0.05]}, 
                  {RGBColor[0, 0, 1], Arrowheads[0.01], 
                   CircleBox[{-1.8138982219326671`, 0.5}, 0.3333333333333333, 
                    NCache[{
                    Rational[-1, 2] 
                    Pi, -1.6475616931622268`}, {-1.5707963267948966`, \
-1.6475616931622268`}]], 
                   ArrowBox[{{-1.8361368136153238`, 
                    0.1674093264219768}, {-1.8394615529992842`, 
                    0.1676483380274899}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-1.8138982219326671`, 
                    0.5}, {-1.7627213110211137`, 0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-1.8138982219326671`, 
                    0.5}, {-1.8521906173290976`, 0.5}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.005], Arrowheads[0.04], 
                   ArrowBox[{{-1.8138982219326671`, 0.5}, {-1.801013706417544,
                     0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-1.8138982219326671`, 0.5}, 1.6, {
                    1.565589915853675, 1.5416568443581662`}], 
                   ArrowBox[{{-1.7832769196452938`, 
                    2.099706953115547}, {-1.7672816477829862`, 
                    2.0993207605150843`}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Dashing[{Small, Small}], 
                   Arrowheads[0.02], 
                   CircleBox[{-1.8138982219326671`, 0.5}, 1.7, {
                    1.565589915853675, 1.5730569323799048`}], 
                   ArrowBox[{{-1.8007413827733196`, 
                    2.199949086762111}, {-1.8177412481539856`, 
                    2.1999956562148806`}}]}, 
                  {RGBColor[0, 0, 1], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-1.8138982219326671`, 0.5}, 1.8, {
                    1.565589915853675, 1.5798057244402177`}], 
                   ArrowBox[{{-1.8121151379858664`, 
                    2.2999991168363496`}, {-1.8301149183093315`, 
                    2.2999269481727938`}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.003], Arrowheads[0.025], 
                   CircleBox[{-1.8138982219326671`, 0.5}, 1.5, {
                    1.565589915853675, 1.5655828952920436`}], 
                   ArrowBox[{{-1.8210780472618666`, 
                    1.9999828166043243`}, {-1.8060781101034347`, 
                    1.9999796151451454`}}]}, 
                  {RGBColor[0, 0, 1], ArrowBox[{{0, 2.8}, {0, 2.8}}]}},
                 Frame->True,
                 FrameTicks->Automatic,
                 ImageSize->800,
                 PlotRange->{{-3, 3}, {-1, 3}}], 517->
                GraphicsBox[{InsetBox[
                   TagBox[GridBox[{
                    {"\<\"t\"\>", "25.8`"},
                    {"\<\"driveIntent\"\>", "0"},
                    {"\<\"motorTorque\"\>", 
                    RowBox[{"-", "2.5401783533971414`"}]},
                    {"\<\"vw\"\>", 
                    RowBox[{"-", "0.02913684569168585`"}]},
                    {"\<\"controllerState\"\>", 
                    RowBox[{"\[LeftAssociation]", 
                    RowBox[{
                    
                    RowBox[{"\<\"lastError\"\>", "\[Rule]", 
                    "0.0035730672839249626`"}], ",", 
                    RowBox[{"\<\"lastInt\"\>", "\[Rule]", 
                    RowBox[{"-", "0.024954658852316933`"}]}], ",", 
                    RowBox[{"\<\"initialized\"\>", "\[Rule]", "True"}]}], 
                    "\[RightAssociation]"}]}
                    },
                    AutoDelete->False,
                    
                    GridBoxItemSize->{
                    "Columns" -> {{Automatic}}, "Rows" -> {{Automatic}}}],
                    "Grid"], {-1, -0.5}], 
                  {GrayLevel[0], LineBox[{{-5, 0}, {5, 0}}]}, 
                  {GrayLevel[0.5], CircleBox[{-1.81536022264839, 0.5}, 0.5], 
                   DiskBox[{-1.81536022264839, 0.5}, 0.1]}, 
                  {GrayLevel[0.5], 
                   DiskBox[{-1.8076075217821035`, 1.9999799650759598`}, 0.1]}, 
                  {GrayLevel[0.5], Thickness[0.01], 
                   LineBox[{{-1.81536022264839, 0.5}, {-1.8076075217821035`, 
                    1.9999799650759598`}}]}, 
                  {RGBColor[0, 1, 0], 
                   DiskBox[{-1.810191755404199, 1.4999866433839733`}, 0.05]}, 
                  {RGBColor[0, 0, 1], Arrowheads[0.01], 
                   CircleBox[{-1.81536022264839, 0.5}, 0.3333333333333333, 
                    NCache[{
                    Rational[-1, 2] 
                    Pi, -1.6470016773968108`}, {-1.5707963267948966`, \
-1.6470016773968108`}]], 
                   ArrowBox[{{-1.8374125548329407`, 
                    0.16739692461383748`}, {-1.840737427545813, 
                    0.1676340742754978}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-1.81536022264839, 0.5}, {-1.7645566555804473`, 
                    0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-1.81536022264839, 0.5}, {-1.8533737268850132`, 
                    0.5}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.005], Arrowheads[0.04], 
                   ArrowBox[{{-1.81536022264839, 0.5}, {-1.8025701598170702`, 
                    0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-1.81536022264839, 0.5}, 1.6, {
                    1.565627836539505, 1.5418690790550147`}], 
                   ArrowBox[{{-1.785078434368355, 
                    2.0997134159900526`}, {-1.7690830809028206`, 
                    2.0993306181499376`}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Dashing[{Small, Small}], 
                   Arrowheads[0.02], 
                   CircleBox[{-1.81536022264839, 0.5}, 1.7, {
                    1.565627836539505, 1.5730401160019492`}], 
                   ArrowBox[{{-1.8021747965045558`, 
                    2.199948865271366}, {-1.8191746610996884`, 
                    2.1999957206002905`}}]}, 
                  {RGBColor[0, 0, 1], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-1.81536022264839, 0.5}, 1.8, {
                    1.565627836539505, 1.5797399385028226`}], 
                   ArrowBox[{{-1.8134587240763216`, 
                    2.299998995639492}, {-1.8314585091085203`, 
                    2.2999280111085127`}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.003], Arrowheads[0.025], 
                   CircleBox[{-1.81536022264839, 0.5}, 1.5, {
                    1.565627836539505, 1.5656202676361328`}], 
                   ArrowBox[{{-1.8225961058465125`, 
                    1.9999825472299146`}, {-1.8075961685789095`, 
                    1.9999799063535504`}}]}, 
                  {RGBColor[0, 0, 1], ArrowBox[{{0, 2.8}, {0, 2.8}}]}},
                 Frame->True,
                 FrameTicks->Automatic,
                 ImageSize->800,
                 PlotRange->{{-3, 3}, {-1, 3}}], 518->
                GraphicsBox[{InsetBox[
                   TagBox[GridBox[{
                    {"\<\"t\"\>", "25.85`"},
                    {"\<\"driveIntent\"\>", "0"},
                    {"\<\"motorTorque\"\>", 
                    RowBox[{"-", "2.521612705957591`"}]},
                    {"\<\"vw\"\>", 
                    RowBox[{"-", "0.028927580470119087`"}]},
                    {"\<\"controllerState\"\>", 
                    RowBox[{"\[LeftAssociation]", 
                    RowBox[{
                    
                    RowBox[{"\<\"lastError\"\>", "\[Rule]", 
                    "0.003548001587543902`"}], ",", 
                    RowBox[{"\<\"lastInt\"\>", "\[Rule]", 
                    RowBox[{"-", "0.024776645722639587`"}]}], ",", 
                    RowBox[{"\<\"initialized\"\>", "\[Rule]", "True"}]}], 
                    "\[RightAssociation]"}]}
                    },
                    AutoDelete->False,
                    
                    GridBoxItemSize->{
                    "Columns" -> {{Automatic}}, "Rows" -> {{Automatic}}}],
                    "Grid"], {-1, -0.5}], 
                  {GrayLevel[0], LineBox[{{-5, 0}, {5, 0}}]}, 
                  {GrayLevel[0.5], CircleBox[{-1.81681172224677, 0.5}, 0.5], 
                   DiskBox[{-1.81681172224677, 0.5}, 0.1]}, 
                  {GrayLevel[0.5], 
                   DiskBox[{-1.8091155990608598`, 1.9999802564326996`}, 0.1]}, 
                  {GrayLevel[0.5], Thickness[0.01], 
                   LineBox[{{-1.81681172224677, 0.5}, {-1.8091155990608598`, 
                    1.9999802564326996`}}]}, 
                  {RGBColor[0, 1, 0], 
                   DiskBox[{-1.8116809734561632`, 1.4999868376217997`}, 
                    0.05]}, 
                  {RGBColor[0, 0, 1], Arrowheads[0.01], 
                   CircleBox[{-1.81681172224677, 0.5}, 0.3333333333333333, 
                    NCache[{
                    Rational[-1, 2] 
                    Pi, -1.6464447079736242`}, {-1.5707963267948966`, \
-1.6464447079736242`}]], 
                   ArrowBox[{{-1.8386788012773694`, 
                    0.1673846937289557}, {-1.8420038055596324`, 
                    0.16761999150149104`}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-1.81681172224677, 0.5}, {-1.7663794681276181`, 
                    0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-1.81681172224677, 0.5}, {-1.8545478187640108`, 
                    0.5}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.005], Arrowheads[0.04], 
                   ArrowBox[{{-1.81681172224677, 0.5}, {-1.804115564644859, 
                    0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-1.81681172224677, 0.5}, 1.6, {
                    1.5656655554932193`, 1.5420801847303738`}], 
                   ArrowBox[{{-1.7868676432200532`, 
                    2.099719773001272}, {-1.7708722093001443`, 
                    2.0993403518795573`}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Dashing[{Small, Small}], 
                   Arrowheads[0.02], 
                   CircleBox[{-1.81681172224677, 0.5}, 1.7, {
                    1.5656655554932193`, 1.573023415119901}], 
                   ArrowBox[{{-1.803597905459289, 
                    2.199948644826045}, {-1.8205977692695257`, 
                    2.1999957840676965`}}]}, 
                  {RGBColor[0, 0, 1], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-1.81681172224677, 0.5}, 1.8, {
                    1.5656655554932193`, 1.5796745149707614`}], 
                   ArrowBox[{{-1.8147924613868536`, 
                    2.299998867384527}, {-1.8327922510245893`, 
                    2.299929060463212}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.003], Arrowheads[0.025], 
                   CircleBox[{-1.81681172224677, 0.5}, 1.5, {
                    1.5656655554932193`, 1.5656574856294738`}], 
                   ArrowBox[{{-1.8241034317803224`, 
                    1.9999822768859898`}, {-1.80910349442482, 
                    1.9999801942771926`}}]}, 
                  {RGBColor[0, 0, 1], ArrowBox[{{0, 2.8}, {0, 2.8}}]}},
                 Frame->True,
                 FrameTicks->Automatic,
                 ImageSize->800,
                 PlotRange->{{-3, 3}, {-1, 3}}], 519->
                GraphicsBox[{InsetBox[
                   TagBox[GridBox[{
                    {"\<\"t\"\>", "25.900000000000002`"},
                    {"\<\"driveIntent\"\>", "0"},
                    {"\<\"motorTorque\"\>", 
                    RowBox[{"-", "2.5031551703171653`"}]},
                    {"\<\"vw\"\>", 
                    RowBox[{"-", "0.02871985266748619`"}]},
                    {"\<\"controllerState\"\>", 
                    RowBox[{"\[LeftAssociation]", 
                    RowBox[{
                    
                    RowBox[{"\<\"lastError\"\>", "\[Rule]", 
                    "0.003523182337726441`"}], ",", 
                    RowBox[{"\<\"lastInt\"\>", "\[Rule]", 
                    RowBox[{"-", "0.0245998795283876`"}]}], ",", 
                    RowBox[{"\<\"initialized\"\>", "\[Rule]", "True"}]}], 
                    "\[RightAssociation]"}]}
                    },
                    AutoDelete->False,
                    
                    GridBoxItemSize->{
                    "Columns" -> {{Automatic}}, "Rows" -> {{Automatic}}}],
                    "Grid"], {-1, -0.5}], 
                  {GrayLevel[0], LineBox[{{-5, 0}, {5, 0}}]}, 
                  {GrayLevel[0.5], 
                   CircleBox[{-1.8182527978279486`, 0.5}, 0.5], 
                   DiskBox[{-1.8182527978279486`, 0.5}, 0.1]}, 
                  {GrayLevel[0.5], 
                   DiskBox[{-1.81061292970098, 1.9999805440121556`}, 0.1]}, 
                  {GrayLevel[0.5], Thickness[0.01], 
                   LineBox[{{-1.8182527978279486`, 0.5}, {-1.81061292970098, 
                    1.9999805440121556`}}]}, 
                  {RGBColor[0, 1, 0], 
                   DiskBox[{-1.8131595524099695`, 1.4999870293414372`}, 
                    0.05]}, 
                  {RGBColor[0, 0, 1], Arrowheads[0.01], 
                   CircleBox[{-1.8182527978279486`, 0.5}, 0.3333333333333333, 
                    NCache[{
                    Rational[-1, 2] 
                    Pi, -1.6458909819044116`}, {-1.5707963267948966`, \
-1.6458909819044116`}]], 
                   ArrowBox[{{-1.8399356957494981`, 
                    0.1673726363498716}, {-1.843260829812521, 
                    0.16760609294487705`}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-1.8182527978279486`, 
                    0.5}, {-1.7681896944216053`, 0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-1.8182527978279486`, 0.5}, {-1.855713068320936,
                     0.5}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.005], Arrowheads[0.04], 
                   ArrowBox[{{-1.8182527978279486`, 
                    0.5}, {-1.8056499649145927`, 0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-1.8182527978279486`, 0.5}, 1.6, {
                    1.5657030593558876`, 1.5422900866160951`}], 
                   ArrowBox[{{-1.788644503655395, 
                    2.099726023079012}, {-1.7726489904466476`, 
                    2.099349959437374}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Dashing[{Small, Small}], 
                   Arrowheads[0.02], 
                   CircleBox[{-1.8182527978279486`, 0.5}, 1.7, {
                    1.5657030593558876`, 1.5730068355498374`}], 
                   ArrowBox[{{-1.8050107966246238`, 
                    2.199948425513001}, {-1.8220106596509755`, 
                    2.199995846605079}}]}, 
                  {RGBColor[0, 0, 1], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-1.8182527978279486`, 0.5}, 1.8, {
                    1.5657030593558876`, 1.5796094769687608`}], 
                   ArrowBox[{{-1.8161164686424474`, 
                    2.29999873224889}, {-1.8341162627822167`, 
                    2.2999300959980764`}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.003], Arrowheads[0.025], 
                   CircleBox[{-1.8182527978279486`, 0.5}, 1.5, {
                    1.5657030593558876`, 1.5656945359220333`}], 
                   ArrowBox[{{-1.8256000821386762`, 
                    1.9999820056964874`}, {-1.8106001447163083`, 
                    1.999980478839759}}]}, 
                  {RGBColor[0, 0, 1], ArrowBox[{{0, 2.8}, {0, 2.8}}]}},
                 Frame->True,
                 FrameTicks->Automatic,
                 ImageSize->800,
                 PlotRange->{{-3, 3}, {-1, 3}}], 520->
                GraphicsBox[{InsetBox[
                   TagBox[GridBox[{
                    {"\<\"t\"\>", "25.95`"},
                    {"\<\"driveIntent\"\>", "0"},
                    {"\<\"motorTorque\"\>", 
                    RowBox[{"-", "2.484811669647891`"}]},
                    {"\<\"vw\"\>", 
                    RowBox[{"-", "0.028513652507082807`"}]},
                    {"\<\"controllerState\"\>", 
                    RowBox[{"\[LeftAssociation]", 
                    RowBox[{
                    
                    RowBox[{"\<\"lastError\"\>", "\[Rule]", 
                    "0.00349859448298734`"}], ",", 
                    RowBox[{"\<\"lastInt\"\>", "\[Rule]", 
                    RowBox[{"-", "0.024424348335682595`"}]}], ",", 
                    RowBox[{"\<\"initialized\"\>", "\[Rule]", "True"}]}], 
                    "\[RightAssociation]"}]}
                    },
                    AutoDelete->False,
                    
                    GridBoxItemSize->{
                    "Columns" -> {{Automatic}}, "Rows" -> {{Automatic}}}],
                    "Grid"], {-1, -0.5}], 
                  {GrayLevel[0], LineBox[{{-5, 0}, {5, 0}}]}, 
                  {GrayLevel[0.5], CircleBox[{-1.819683526015585, 0.5}, 0.5], 
                   DiskBox[{-1.819683526015585, 0.5}, 0.1]}, 
                  {GrayLevel[0.5], 
                   DiskBox[{-1.8120995721817743`, 1.9999808277588906`}, 0.1]}, 
                  {GrayLevel[0.5], Thickness[0.01], 
                   LineBox[{{-1.819683526015585, 0.5}, {-1.8120995721817743`, 
                    1.9999808277588906`}}]}, 
                  {RGBColor[0, 1, 0], 
                   DiskBox[{-1.8146275567930445`, 1.499987218505927}, 0.05]}, 
                  {RGBColor[0, 0, 1], Arrowheads[0.01], 
                   CircleBox[{-1.819683526015585, 0.5}, 0.3333333333333333, 
                    NCache[{
                    Rational[-1, 2] 
                    Pi, -1.6453406768843333`}, {-1.5707963267948966`, \
-1.6453406768843333`}]], 
                   ArrowBox[{{-1.8411833741551644`, 
                    0.16736075450860266`}, {-1.8445086361870326`, 
                    0.16759238123038367`}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-1.819683526015585, 0.5}, {-1.7699872926226272`,
                     0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-1.819683526015585, 0.5}, {-1.8568696411475696`,
                     0.5}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.005], Arrowheads[0.04], 
                   ArrowBox[{{-1.819683526015585, 0.5}, {-1.8071734077546118`,
                     0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-1.819683526015585, 0.5}, 1.6, {
                    1.5657403360312991`, 1.542498717011193}], 
                   ArrowBox[{{-1.7904089839572284`, 
                    2.099732165453728}, {-1.7744133926382908`, 
                    2.099359438970491}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Dashing[{Small, Small}], 
                   Arrowheads[0.02], 
                   CircleBox[{-1.819683526015585, 0.5}, 1.7, {
                    1.5657403360312991`, 1.5729903822894473`}], 
                   ArrowBox[{{-1.8064135551199592`, 
                    2.199948207408811}, {-1.8234134173637784`, 
                    2.1999959082040554`}}]}, 
                  {RGBColor[0, 0, 1], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-1.819683526015585, 0.5}, 1.8, {
                    1.5657403360312991`, 1.5795448453071208`}], 
                   ArrowBox[{{-1.8174308599256117`, 
                    2.2999985904148614`}, {-1.8354306584638613`, 
                    2.299931117520795}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.003], Arrowheads[0.025], 
                   CircleBox[{-1.819683526015585, 0.5}, 1.5, {
                    1.5657403360312991`, 1.56573140596674}], 
                   ArrowBox[{{-1.8270861147249153`, 
                    1.9999817337822487`}, {-1.8120861772564476`, 
                    1.9999807599738846`}}]}, 
                  {RGBColor[0, 0, 1], ArrowBox[{{0, 2.8}, {0, 2.8}}]}},
                 Frame->True,
                 FrameTicks->Automatic,
                 ImageSize->800,
                 PlotRange->{{-3, 3}, {-1, 3}}], 521->
                GraphicsBox[{InsetBox[
                   TagBox[GridBox[{
                    {"\<\"t\"\>", "26.`"},
                    {"\<\"driveIntent\"\>", "0"},
                    {"\<\"motorTorque\"\>", 
                    RowBox[{"-", "2.466587491958093`"}]},
                    {"\<\"vw\"\>", 
                    RowBox[{"-", "0.028308969756218345`"}]},
                    {"\<\"controllerState\"\>", 
                    RowBox[{"\[LeftAssociation]", 
                    RowBox[{
                    
                    RowBox[{"\<\"lastError\"\>", "\[Rule]", 
                    "0.0034742240858711866`"}], ",", 
                    RowBox[{"\<\"lastInt\"\>", "\[Rule]", 
                    RowBox[{"-", "0.02425004093481246`"}]}], ",", 
                    RowBox[{"\<\"initialized\"\>", "\[Rule]", "True"}]}], 
                    "\[RightAssociation]"}]}
                    },
                    AutoDelete->False,
                    
                    GridBoxItemSize->{
                    "Columns" -> {{Automatic}}, "Rows" -> {{Automatic}}}],
                    "Grid"], {-1, -0.5}], 
                  {GrayLevel[0], LineBox[{{-5, 0}, {5, 0}}]}, 
                  {GrayLevel[0.5], 
                   CircleBox[{-1.8211039829326383`, 0.5}, 0.5], 
                   DiskBox[{-1.8211039829326383`, 0.5}, 0.1]}, 
                  {GrayLevel[0.5], 
                   DiskBox[{-1.8135755863580538`, 1.9999811076293648`}, 0.1]}, 
                  {GrayLevel[0.5], Thickness[0.01], 
                   LineBox[{{-1.8211039829326383`, 
                    0.5}, {-1.8135755863580538`, 1.9999811076293648`}}]}, 
                  {RGBColor[0, 1, 0], 
                   DiskBox[{-1.8160850518829155`, 1.4999874050862432`}, 
                    0.05]}, 
                  {RGBColor[0, 0, 1], Arrowheads[0.01], 
                   CircleBox[{-1.8211039829326383`, 0.5}, 0.3333333333333333, 
                    NCache[{
                    Rational[-1, 2] 
                    Pi, -1.6447939515536394`}, {-1.5707963267948966`, \
-1.6447939515536394`}]], 
                   ArrowBox[{{-1.8424219655665401`, 
                    0.16734904971196646`}, {-1.8457473537376234`, 
                    0.16757885839423647`}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-1.8211039829326383`, 
                    0.5}, {-1.7717722330934764`, 0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-1.8211039829326383`, 0.5}, {-1.85801769313586, 
                    0.5}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.005], Arrowheads[0.04], 
                   ArrowBox[{{-1.8211039829326383`, 
                    0.5}, {-1.8086859432966982`, 0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-1.8211039829326383`, 0.5}, 1.6, {
                    1.5657773746740662`, 1.54270601521527}], 
                   ArrowBox[{{-1.7921610631058102`, 
                    2.099738199641397}, {-1.7761653948650282`, 
                    2.0993687890234978`}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Dashing[{Small, Small}], 
                   Arrowheads[0.02], 
                   CircleBox[{-1.8211039829326383`, 0.5}, 1.7, {
                    1.5657773746740662`, 1.5729740595565442`}], 
                   ArrowBox[{{-1.8078062642382429`, 
                    2.19994799058016}, {-1.82480612570119, 
                    2.199995968859609}}]}, 
                  {RGBColor[0, 0, 1], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-1.8211039829326383`, 0.5}, 1.8, {
                    1.5657773746740662`, 1.5794806385182778`}], 
                   ArrowBox[{{-1.8187357447179753`, 
                    2.299998442068148}, {-1.8367355475513405`, 
                    2.299932124883483}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.003], Arrowheads[0.025], 
                   CircleBox[{-1.8211039829326383`, 0.5}, 1.5, {
                    1.5657773746740662`, 1.565768084064546}], 
                   ArrowBox[{{-1.828561588113716, 
                    1.999981461260426}, {-1.8135616506196204`, 
                    1.9999810376212361`}}]}, 
                  {RGBColor[0, 0, 1], ArrowBox[{{0, 2.8}, {0, 2.8}}]}},
                 Frame->True,
                 FrameTicks->Automatic,
                 ImageSize->800,
                 PlotRange->{{-3, 3}, {-1, 3}}], 522->
                GraphicsBox[{InsetBox[
                   TagBox[GridBox[{
                    {"\<\"t\"\>", "26.05`"},
                    {"\<\"driveIntent\"\>", "0"},
                    {"\<\"motorTorque\"\>", 
                    RowBox[{"-", "2.448487303963535`"}]},
                    {"\<\"vw\"\>", 
                    RowBox[{"-", "0.02810579378289662`"}]},
                    {"\<\"controllerState\"\>", 
                    RowBox[{"\[LeftAssociation]", 
                    RowBox[{
                    
                    RowBox[{"\<\"lastError\"\>", "\[Rule]", 
                    "0.0034500583175243703`"}], ",", 
                    RowBox[{"\<\"lastInt\"\>", "\[Rule]", 
                    RowBox[{"-", "0.02407694678462425`"}]}], ",", 
                    RowBox[{"\<\"initialized\"\>", "\[Rule]", "True"}]}], 
                    "\[RightAssociation]"}]}
                    },
                    AutoDelete->False,
                    
                    GridBoxItemSize->{
                    "Columns" -> {{Automatic}}, "Rows" -> {{Automatic}}}],
                    "Grid"], {-1, -0.5}], 
                  {GrayLevel[0], LineBox[{{-5, 0}, {5, 0}}]}, 
                  {GrayLevel[0.5], 
                   CircleBox[{-1.8225142441800175`, 0.5}, 0.5], 
                   DiskBox[{-1.8225142441800175`, 0.5}, 0.1]}, 
                  {GrayLevel[0.5], 
                   DiskBox[{-1.8150410334048181`, 1.9999813835913796`}, 0.1]}, 
                  {GrayLevel[0.5], Thickness[0.01], 
                   LineBox[{{-1.8225142441800175`, 
                    0.5}, {-1.8150410334048181`, 1.9999813835913796`}}]}, 
                  {RGBColor[0, 1, 0], 
                   DiskBox[{-1.817532103663218, 1.4999875890609198`}, 0.05]}, 
                  {RGBColor[0, 0, 1], Arrowheads[0.01], 
                   CircleBox[{-1.8225142441800175`, 0.5}, 0.3333333333333333, 
                    NCache[{
                    Rational[-1, 2] 
                    Pi, -1.6442509459138026`}, {-1.5707963267948966`, \
-1.6442509459138026`}]], 
                   ArrowBox[{{-1.8436515923378345`, 
                    0.1673375229696527}, {-1.8469771048060684`, 
                    0.16756552591359974`}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-1.8225142441800175`, 
                    0.5}, {-1.7735444981007469`, 0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-1.8225142441800175`, 
                    0.5}, {-1.8591573706230677`, 0.5}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.005], Arrowheads[0.04], 
                   ArrowBox[{{-1.8225142441800175`, 
                    0.5}, {-1.8101876245437971`, 0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-1.8225142441800175`, 0.5}, 1.6, {
                    1.5658141656669804`, 1.5429119274017902`}], 
                   ArrowBox[{{-1.793900730554997, 
                    2.099744125426948}, {-1.7779049865871743`, 
                    2.0993780085198797`}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Dashing[{Small, Small}], 
                   Arrowheads[0.02], 
                   CircleBox[{-1.8225142441800175`, 0.5}, 1.7, {
                    1.5658141656669804`, 1.5729578708307037`}], 
                   ArrowBox[{{-1.809189005495403, 
                    2.199947775084281}, {-1.8261888661794154`, 
                    2.1999960285698203`}}]}, 
                  {RGBColor[0, 0, 1], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-1.8225142441800175`, 0.5}, 1.8, {
                    1.5658141656669804`, 1.5794168729112223`}], 
                   ArrowBox[{{-1.8200312279768898`, 
                    2.299998287396556}, {-1.8380310350024163`, 
                    2.2999331179803804`}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.003], Arrowheads[0.025], 
                   CircleBox[{-1.8225142441800175`, 0.5}, 1.5, {
                    1.5658141656669804`, 1.5658045593960874`}], 
                   ArrowBox[{{-1.8300265616772275`, 
                    1.9999811882439795`}, {-1.8150266241776578`, 
                    1.999981311732483}}]}, 
                  {RGBColor[0, 0, 1], ArrowBox[{{0, 2.8}, {0, 2.8}}]}},
                 Frame->True,
                 FrameTicks->Automatic,
                 ImageSize->800,
                 PlotRange->{{-3, 3}, {-1, 3}}], 523->
                GraphicsBox[{InsetBox[
                   TagBox[GridBox[{
                    {"\<\"t\"\>", "26.1`"},
                    {"\<\"driveIntent\"\>", "0"},
                    {"\<\"motorTorque\"\>", 
                    RowBox[{"-", "2.430515169386136`"}]},
                    {"\<\"vw\"\>", 
                    RowBox[{"-", "0.027904113610507614`"}]},
                    {"\<\"controllerState\"\>", 
                    RowBox[{"\[LeftAssociation]", 
                    RowBox[{
                    
                    RowBox[{"\<\"lastError\"\>", "\[Rule]", 
                    "0.003426085442344723`"}], ",", 
                    RowBox[{"\<\"lastInt\"\>", "\[Rule]", 
                    RowBox[{"-", "0.02390505595744692`"}]}], ",", 
                    RowBox[{"\<\"initialized\"\>", "\[Rule]", "True"}]}], 
                    "\[RightAssociation]"}]}
                    },
                    AutoDelete->False,
                    
                    GridBoxItemSize->{
                    "Columns" -> {{Automatic}}, "Rows" -> {{Automatic}}}],
                    "Grid"], {-1, -0.5}], 
                  {GrayLevel[0], LineBox[{{-5, 0}, {5, 0}}]}, 
                  {GrayLevel[0.5], 
                   CircleBox[{-1.8239143848180188`, 0.5}, 0.5], 
                   DiskBox[{-1.8239143848180188`, 0.5}, 0.1]}, 
                  {GrayLevel[0.5], 
                   DiskBox[{-1.8164959757507797`, 1.999981655623465}, 0.1]}, 
                  {GrayLevel[0.5], Thickness[0.01], 
                   LineBox[{{-1.8239143848180188`, 
                    0.5}, {-1.8164959757507797`, 1.999981655623465}}]}, 
                  {RGBColor[0, 1, 0], 
                   DiskBox[{-1.8189687787731927`, 1.4999877704156432`}, 
                    0.05]}, 
                  {RGBColor[0, 0, 1], Arrowheads[0.01], 
                   CircleBox[{-1.8239143848180188`, 0.5}, 0.3333333333333333, 
                    NCache[{
                    Rational[-1, 2] 
                    Pi, -1.6437117818764806`}, {-1.5707963267948966`, \
-1.6437117818764806`}]], 
                   ArrowBox[{{-1.8448723702680536`, 
                    0.16732617482436762`}, {-1.8481980051839095`, 
                    0.16755238473853296`}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-1.8239143848180188`, 0.5}, {-1.775304081430296,
                     0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-1.8239143848180188`, 
                    0.5}, {-1.8602888106080901`, 0.5}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.005], Arrowheads[0.04], 
                   ArrowBox[{{-1.8239143848180188`, 
                    0.5}, {-1.8116785072203674`, 0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-1.8239143848180188`, 0.5}, 1.6, {
                    1.5658507005890698`, 1.543116406439307}], 
                   ArrowBox[{{-1.7956279859279487`, 
                    2.0997499428466413`}, {-1.7796321674312983`, 
                    2.099387096741535}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Dashing[{Small, Small}], 
                   Arrowheads[0.02], 
                   CircleBox[{-1.8239143848180188`, 0.5}, 1.7, {
                    1.5658507005890698`, 1.5729418188972812`}], 
                   ArrowBox[{{-1.810561858686615, 
                    2.199947560969429}, {-1.8275617185938753`, 
                    2.1999960873355935`}}]}, 
                  {RGBColor[0, 0, 1], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-1.8239143848180188`, 0.5}, 1.8, {
                    1.5658507005890698`, 1.579353562641215}], 
                   ArrowBox[{{-1.8213174102423555`, 
                    2.299998126588762}, {-1.839317221357702, 
                    2.299934096745359}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.003], Arrowheads[0.025], 
                   CircleBox[{-1.8239143848180188`, 0.5}, 1.5, {
                    1.5658507005890698`, 1.5658408220412035`}], 
                   ArrowBox[{{-1.8314810955957876`, 
                    1.9999809148412542`}, {-1.8164811581105582`, 
                    1.9999815822671674`}}]}, 
                  {RGBColor[0, 0, 1], ArrowBox[{{0, 2.8}, {0, 2.8}}]}},
                 Frame->True,
                 FrameTicks->Automatic,
                 ImageSize->800,
                 PlotRange->{{-3, 3}, {-1, 3}}], 524->
                GraphicsBox[{InsetBox[
                   TagBox[GridBox[{
                    {"\<\"t\"\>", "26.150000000000002`"},
                    {"\<\"driveIntent\"\>", "0"},
                    {"\<\"motorTorque\"\>", 
                    RowBox[{"-", "2.41267457099336`"}]},
                    {"\<\"vw\"\>", 
                    RowBox[{"-", "0.02770391797022894`"}]},
                    {"\<\"controllerState\"\>", 
                    RowBox[{"\[LeftAssociation]", 
                    RowBox[{
                    
                    RowBox[{"\<\"lastError\"\>", "\[Rule]", 
                    "0.003402294793991076`"}], ",", 
                    RowBox[{"\<\"lastInt\"\>", "\[Rule]", 
                    RowBox[{"-", "0.023734359085007067`"}]}], ",", 
                    RowBox[{"\<\"initialized\"\>", "\[Rule]", "True"}]}], 
                    "\[RightAssociation]"}]}
                    },
                    AutoDelete->False,
                    
                    GridBoxItemSize->{
                    "Columns" -> {{Automatic}}, "Rows" -> {{Automatic}}}],
                    "Grid"], {-1, -0.5}], 
                  {GrayLevel[0], LineBox[{{-5, 0}, {5, 0}}]}, 
                  {GrayLevel[0.5], 
                   CircleBox[{-1.8253044793504467`, 0.5}, 0.5], 
                   DiskBox[{-1.8253044793504467`, 0.5}, 0.1]}, 
                  {GrayLevel[0.5], 
                   DiskBox[{-1.817940477002682, 1.9999819237142233`}, 0.1]}, 
                  {GrayLevel[0.5], Thickness[0.01], 
                   LineBox[{{-1.8253044793504467`, 0.5}, {-1.817940477002682, 
                    1.9999819237142233`}}]}, 
                  {RGBColor[0, 1, 0], 
                   DiskBox[{-1.8203951444519368`, 1.4999879491428154`}, 
                    0.05]}, 
                  {RGBColor[0, 0, 1], Arrowheads[0.01], 
                   CircleBox[{-1.8253044793504467`, 0.5}, 0.3333333333333333, 
                    NCache[{
                    Rational[-1, 2] 
                    Pi, -1.6431765639246974`}, {-1.5707963267948966`, \
-1.6431765639246974`}]], 
                   ArrowBox[{{-1.846084408803866, 
                    0.1673150053834378}, {-1.8494101643149954`, 
                    0.1675394353257803}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-1.8253044793504467`, 
                    0.5}, {-1.7770509879305796`, 0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-1.8253044793504467`, 
                    0.5}, {-1.8614121410287825`, 0.5}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.005], Arrowheads[0.04], 
                   ArrowBox[{{-1.8253044793504467`, 
                    0.5}, {-1.8131586496089152`, 0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-1.8253044793504467`, 0.5}, 1.6, {
                    1.5658869721757271`, 1.5433194116683187`}], 
                   ArrowBox[{{-1.797342838644512, 
                    2.0997556521697405`}, {-1.7813469468178083`, 
                    2.0993960533068856`}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Dashing[{Small, Small}], 
                   Arrowheads[0.02], 
                   CircleBox[{-1.8253044793504467`, 0.5}, 1.7, {
                    1.5658869721757271`, 1.5729259058930667`}], 
                   ArrowBox[{{-1.8119249019480324`, 
                    2.1999473482753906`}, {-1.82892476108094, 
                    2.1999961451603918`}}]}, 
                  {RGBColor[0, 0, 1], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-1.8253044793504467`, 0.5}, 1.8, {
                    1.5658869721757271`, 1.579290719792357}], 
                   ArrowBox[{{-1.8225943877697717`, 
                    2.299997959833184}, {-1.8405942028733935`, 
                    2.2999350611493163`}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.003], Arrowheads[0.025], 
                   CircleBox[{-1.8253044793504467`, 0.5}, 1.5, {
                    1.5658869721757271`, 1.565876862987432}], 
                   ArrowBox[{{-1.832925250854785, 
                    1.999980641155638}, {-1.8179253134033524`, 
                    1.999981849193491}}]}, 
                  {RGBColor[0, 0, 1], ArrowBox[{{0, 2.8}, {0, 2.8}}]}},
                 Frame->True,
                 FrameTicks->Automatic,
                 ImageSize->800,
                 PlotRange->{{-3, 3}, {-1, 3}}], 525->
                GraphicsBox[{InsetBox[
                   TagBox[GridBox[{
                    {"\<\"t\"\>", "26.2`"},
                    {"\<\"driveIntent\"\>", "0"},
                    {"\<\"motorTorque\"\>", 
                    RowBox[{"-", "2.394968435743837`"}]},
                    {"\<\"vw\"\>", 
                    RowBox[{"-", "0.02750519535089199`"}]},
                    {"\<\"controllerState\"\>", 
                    RowBox[{"\[LeftAssociation]", 
                    RowBox[{
                    
                    RowBox[{"\<\"lastError\"\>", "\[Rule]", 
                    "0.0033786767439707398`"}], ",", 
                    RowBox[{"\<\"lastInt\"\>", "\[Rule]", 
                    RowBox[{"-", "0.023564847305738497`"}]}], ",", 
                    RowBox[{"\<\"initialized\"\>", "\[Rule]", "True"}]}], 
                    "\[RightAssociation]"}]}
                    },
                    AutoDelete->False,
                    
                    GridBoxItemSize->{
                    "Columns" -> {{Automatic}}, "Rows" -> {{Automatic}}}],
                    "Grid"], {-1, -0.5}], 
                  {GrayLevel[0], LineBox[{{-5, 0}, {5, 0}}]}, 
                  {GrayLevel[0.5], 
                   CircleBox[{-1.8266846017112794`, 0.5}, 0.5], 
                   DiskBox[{-1.8266846017112794`, 0.5}, 0.1]}, 
                  {GrayLevel[0.5], 
                   DiskBox[{-1.8193746018622037`, 1.9999821878616446`}, 0.1]}, 
                  {GrayLevel[0.5], Thickness[0.01], 
                   LineBox[{{-1.8266846017112794`, 
                    0.5}, {-1.8193746018622037`, 1.9999821878616446`}}]}, 
                  {RGBColor[0, 1, 0], 
                   DiskBox[{-1.8218112684785623`, 1.4999881252410965`}, 
                    0.05]}, 
                  {RGBColor[0, 0, 1], Arrowheads[0.01], 
                   CircleBox[{-1.8266846017112794`, 0.5}, 0.3333333333333333, 
                    NCache[{
                    Rational[-1, 2] 
                    Pi, -1.6426453798672116`}, {-1.5707963267948966`, \
-1.6426453798672116`}]], 
                   ArrowBox[{{-1.847287811276108, 
                    0.1673040143513318}, {-1.8506136855316473`, 
                    0.16752667767378854`}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-1.8266846017112794`, 
                    0.5}, {-1.7787852329964027`, 0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-1.8266846017112794`, 
                    0.5}, {-1.8625274810906436`, 0.5}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.005], Arrowheads[0.04], 
                   ArrowBox[{{-1.8266846017112794`, 
                    0.5}, {-1.8146281123757668`, 0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-1.8266846017112794`, 0.5}, 1.6, {
                    1.565922974272202, 1.5435209086409711`}], 
                   ArrowBox[{{-1.7990453074916952`, 
                    2.099761253879792}, {-1.7830493435316344`, 
                    2.099404878147993}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Dashing[{Small, Small}], 
                   Arrowheads[0.02], 
                   CircleBox[{-1.8266846017112794`, 0.5}, 1.7, {
                    1.565922974272202, 1.5729101333530116`}], 
                   ArrowBox[{{-1.813278211822888, 
                    2.1999471370340196`}, {-1.8302780701840304`, 
                    2.1999962020499737`}}]}, 
                  {RGBColor[0, 0, 1], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-1.8266846017112794`, 0.5}, 1.8, {
                    1.565922974272202, 1.5792283544707788`}], 
                   ArrowBox[{{-1.8238622526843427`, 
                    2.299997787316965}, {-1.8418620716756566`, 
                    2.299936011197476}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.003], Arrowheads[0.025], 
                   CircleBox[{-1.8266846017112794`, 0.5}, 1.5, {
                    1.565922974272202, 1.5659126741287654`}], 
                   ArrowBox[{{-1.8343590892294577`, 
                    1.9999803672852965`}, {-1.8193591518309042`, 
                    1.999982112488029}}]}, 
                  {RGBColor[0, 0, 1], ArrowBox[{{0, 2.8}, {0, 2.8}}]}},
                 Frame->True,
                 FrameTicks->Automatic,
                 ImageSize->800,
                 PlotRange->{{-3, 3}, {-1, 3}}], 526->
                GraphicsBox[{InsetBox[
                   TagBox[GridBox[{
                    {"\<\"t\"\>", "26.25`"},
                    {"\<\"driveIntent\"\>", "0"},
                    {"\<\"motorTorque\"\>", 
                    RowBox[{"-", "2.3773991624377837`"}]},
                    {"\<\"vw\"\>", 
                    RowBox[{"-", "0.027307934046120256`"}]},
                    {"\<\"controllerState\"\>", 
                    RowBox[{"\[LeftAssociation]", 
                    RowBox[{
                    
                    RowBox[{"\<\"lastError\"\>", "\[Rule]", 
                    "0.0033552226639549125`"}], ",", 
                    RowBox[{"\<\"lastInt\"\>", "\[Rule]", 
                    RowBox[{"-", "0.023396512213826878`"}]}], ",", 
                    RowBox[{"\<\"initialized\"\>", "\[Rule]", "True"}]}], 
                    "\[RightAssociation]"}]}
                    },
                    AutoDelete->False,
                    
                    GridBoxItemSize->{
                    "Columns" -> {{Automatic}}, "Rows" -> {{Automatic}}}],
                    "Grid"], {-1, -0.5}], 
                  {GrayLevel[0], LineBox[{{-5, 0}, {5, 0}}]}, 
                  {GrayLevel[0.5], 
                   CircleBox[{-1.8280548252537714`, 0.5}, 0.5], 
                   DiskBox[{-1.8280548252537714`, 0.5}, 0.1]}, 
                  {GrayLevel[0.5], 
                   DiskBox[{-1.8207984160371535`, 1.9999824480724038`}, 0.1]}, 
                  {GrayLevel[0.5], Thickness[0.01], 
                   LineBox[{{-1.8280548252537714`, 
                    0.5}, {-1.8207984160371535`, 1.9999824480724038`}}]}, 
                  {RGBColor[0, 1, 0], 
                   DiskBox[{-1.8232172191093596`, 1.4999882987149358`}, 
                    0.05]}, 
                  {RGBColor[0, 0, 1], Arrowheads[0.01], 
                   CircleBox[{-1.8280548252537714`, 0.5}, 0.3333333333333333, 
                    NCache[{
                    Rational[-1, 2] 
                    Pi, -1.64211830166803}, {-1.5707963267948966`, \
-1.64211830166803}]], 
                   ArrowBox[{{-1.848482675163825, 
                    0.16729320106261214`}, {-1.851808666318359, 
                    0.16751411135840755`}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-1.8280548252537714`, 
                    0.5}, {-1.7805068420050156`, 0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-1.8280548252537714`, 
                    0.5}, {-1.8636349416378875`, 0.5}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.005], Arrowheads[0.04], 
                   ArrowBox[{{-1.8280548252537714`, 0.5}, {-1.816086958389132,
                     0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-1.8280548252537714`, 0.5}, 1.6, {
                    1.5659587017816603`, 1.543720868830365}], 
                   ArrowBox[{{-1.800735420147932, 
                    2.0997667486557727`}, {-1.7847393852467048`, 
                    2.0994135714870605`}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Dashing[{Small, Small}], 
                   Arrowheads[0.02], 
                   CircleBox[{-1.8280548252537714`, 0.5}, 1.7, {
                    1.5659587017816603`, 1.5728945022574043`}], 
                   ArrowBox[{{-1.8146218633307927`, 
                    2.1999469272697825`}, {-1.8316217209229184`, 
                    2.199996258012142}}]}, 
                  {RGBColor[0, 0, 1], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-1.8280548252537714`, 0.5}, 1.8, {
                    1.5659587017816603`, 1.5791664749063146`}], 
                   ArrowBox[{{-1.8251210931531934`, 
                    2.2999976092250685`}, {-1.8431209159327253`, 
                    2.2999369469266564`}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.003], Arrowheads[0.025], 
                   CircleBox[{-1.8280548252537714`, 0.5}, 1.5, {
                    1.5659587017816603`, 1.5659482482557072`}], 
                   ArrowBox[{{-1.8357826732590783`, 
                    1.999980093322977}, {-1.8207827359321003`, 
                    1.9999823721353853`}}]}, 
                  {RGBColor[0, 0, 1], ArrowBox[{{0, 2.8}, {0, 2.8}}]}},
                 Frame->True,
                 FrameTicks->Automatic,
                 ImageSize->800,
                 PlotRange->{{-3, 3}, {-1, 3}}], 527->
                GraphicsBox[{InsetBox[
                   TagBox[GridBox[{
                    {"\<\"t\"\>", "26.3`"},
                    {"\<\"driveIntent\"\>", "0"},
                    {"\<\"motorTorque\"\>", 
                    RowBox[{"-", "2.3599686513233795`"}]},
                    {"\<\"vw\"\>", 
                    RowBox[{"-", "0.02711212219859554`"}]},
                    {"\<\"controllerState\"\>", 
                    RowBox[{"\[LeftAssociation]", 
                    RowBox[{
                    
                    RowBox[{"\<\"lastError\"\>", "\[Rule]", 
                    "0.003331924882900409`"}], ",", 
                    RowBox[{"\<\"lastInt\"\>", "\[Rule]", 
                    RowBox[{"-", "0.023229345810275488`"}]}], ",", 
                    RowBox[{"\<\"initialized\"\>", "\[Rule]", "True"}]}], 
                    "\[RightAssociation]"}]}
                    },
                    AutoDelete->False,
                    
                    GridBoxItemSize->{
                    "Columns" -> {{Automatic}}, "Rows" -> {{Automatic}}}],
                    "Grid"], {-1, -0.5}], 
                  {GrayLevel[0], LineBox[{{-5, 0}, {5, 0}}]}, 
                  {GrayLevel[0.5], 
                   CircleBox[{-1.8294152227418288`, 0.5}, 0.5], 
                   DiskBox[{-1.8294152227418288`, 0.5}, 0.1]}, 
                  {GrayLevel[0.5], 
                   DiskBox[{-1.8222119861484714`, 1.9999827043611471`}, 0.1]}, 
                  {GrayLevel[0.5], Thickness[0.01], 
                   LineBox[{{-1.8294152227418288`, 
                    0.5}, {-1.8222119861484714`, 1.9999827043611471`}}]}, 
                  {RGBColor[0, 1, 0], 
                   DiskBox[{-1.824613065012924, 1.499988469574098}, 0.05]}, 
                  {RGBColor[0, 0, 1], Arrowheads[0.01], 
                   CircleBox[{-1.8294152227418288`, 0.5}, 0.3333333333333333, 
                    NCache[{
                    Rational[-1, 2] 
                    Pi, -1.641595386334598}, {-1.5707963267948966`, \
-1.641595386334598}]], 
                   ArrowBox[{{-1.8496690923802097`, 
                    0.16728256451489865`}, {-1.8529951985973891`, 
                    0.16750173556879688`}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-1.8294152227418288`, 0.5}, {-1.782215849715361,
                     0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-1.8294152227418288`, 
                    0.5}, {-1.8647346255584998`, 0.5}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.005], Arrowheads[0.04], 
                   ArrowBox[{{-1.8294152227418288`, 
                    0.5}, {-1.8175352525320323`, 0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-1.8294152227418288`, 0.5}, 1.6, {
                    1.5659941506089319`, 1.5439192693157295`}], 
                   ArrowBox[{{-1.802413212670994, 
                    2.099772137353359}, {-1.7864171080140696`, 
                    2.099422133812665}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Dashing[{Small, Small}], 
                   Arrowheads[0.02], 
                   CircleBox[{-1.8294152227418288`, 0.5}, 1.7, {
                    1.5659941506089319`, 1.5728790130790298`}], 
                   ArrowBox[{{-1.8159559300392787`, 
                    2.199946719000318}, {-1.8329557868652726`, 
                    2.1999963130564986`}}]}, 
                  {RGBColor[0, 0, 1], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-1.8294152227418288`, 0.5}, 1.8, {
                    1.5659941506089319`, 1.5791050875607284`}], 
                   ArrowBox[{{-1.8263709935715562`, 
                    2.2999974257394813`}, {-1.8443708200410702`, 
                    2.299937868402524}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.003], Arrowheads[0.025], 
                   CircleBox[{-1.8294152227418288`, 0.5}, 1.5, {
                    1.5659941506089319`, 1.5659835790377827`}], 
                   ArrowBox[{{-1.837196066212096, 
                    1.9999798193558775`}, {-1.8221961289749924`, 
                    1.9999826281278013`}}]}, 
                  {RGBColor[0, 0, 1], ArrowBox[{{0, 2.8}, {0, 2.8}}]}},
                 Frame->True,
                 FrameTicks->Automatic,
                 ImageSize->800,
                 PlotRange->{{-3, 3}, {-1, 3}}], 528->
                GraphicsBox[{InsetBox[
                   TagBox[GridBox[{
                    {"\<\"t\"\>", "26.35`"},
                    {"\<\"driveIntent\"\>", "0"},
                    {"\<\"motorTorque\"\>", 
                    RowBox[{"-", "2.3426783351487876`"}]},
                    {"\<\"vw\"\>", 
                    RowBox[{"-", "0.026917747841351706`"}]},
                    {"\<\"controllerState\"\>", 
                    RowBox[{"\[LeftAssociation]", 
                    RowBox[{
                    
                    RowBox[{"\<\"lastError\"\>", "\[Rule]", 
                    "0.0033087766399825134`"}], ",", 
                    RowBox[{"\<\"lastInt\"\>", "\[Rule]", 
                    RowBox[{"-", "0.02306334045622571`"}]}], ",", 
                    RowBox[{"\<\"initialized\"\>", "\[Rule]", "True"}]}], 
                    "\[RightAssociation]"}]}
                    },
                    AutoDelete->False,
                    GridBoxItemSize->{
                    "Columns" -> {{Automatic}}, "Rows" -> {{Automatic}}}],
                    "Grid"], {-1, -0.5}], 
                  {GrayLevel[0], LineBox[{{-5, 0}, {5, 0}}]}, 
                  {GrayLevel[0.5], 
                   CircleBox[{-1.8307658663435311`, 0.5}, 0.5], 
                   DiskBox[{-1.8307658663435311`, 0.5}, 0.1]}, 
                  {GrayLevel[0.5], 
                   DiskBox[{-1.8236153796344576`, 1.9999829567497838`}, 0.1]}, 
                  {GrayLevel[0.5], Thickness[0.01], 
                   LineBox[{{-1.8307658663435311`, 
                    0.5}, {-1.8236153796344576`, 1.9999829567497838`}}]}, 
                  {RGBColor[0, 1, 0], 
                   DiskBox[{-1.8259988752041487`, 1.499988637833189}, 0.05]}, 
                  {RGBColor[0, 0, 1], Arrowheads[0.01], 
                   CircleBox[{-1.8307658663435311`, 0.5}, 0.3333333333333333, 
                    NCache[{
                    Rational[-1, 2] 
                    Pi, -1.64107667684936}, {-1.5707963267948966`, \
-1.64107667684936}]], 
                   ArrowBox[{{-1.8508471495752128`, 
                    0.16727210340147303`}, {-1.8541733690310316`, 
                    0.16748954914311975`}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-1.8307658663435311`, 
                    0.5}, {-1.7839122996405554`, 0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-1.8307658663435311`, 0.5}, {-1.865826628215548,
                     0.5}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.005], Arrowheads[0.04], 
                   ArrowBox[{{-1.8307658663435311`, 
                    0.5}, {-1.8189730615125723`, 0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-1.8307658663435311`, 0.5}, 1.6, {
                    1.5660293176009825`, 1.5441160924492525`}], 
                   ArrowBox[{{-1.8040787289586802`, 
                    2.099777420986495}, {-1.7880825557228017`, 
                    2.0994305658560033`}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Dashing[{Small, Small}], 
                   Arrowheads[0.02], 
                   CircleBox[{-1.8307658663435311`, 0.5}, 1.7, {
                    1.5660293176009825`, 1.5728636658298334`}], 
                   ArrowBox[{{-1.81728048413665, 
                    2.1999465122369983`}, {-1.8342803401995096`, 
                    2.199996367194211}}]}, 
                  {RGBColor[0, 0, 1], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-1.8307658663435311`, 0.5}, 1.8, {
                    1.5660293176009825`, 1.579044197240698}], 
                   ArrowBox[{{-1.8276120347596627`, 
                    2.2999972370385295`}, {-1.8456118648222737`, 
                    2.2999387757168765`}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.003], Arrowheads[0.025], 
                   CircleBox[{-1.8307658663435311`, 0.5}, 1.5, {
                    1.5660293176009825`, 1.5660186609994933`}], 
                   ArrowBox[{{-1.8385993320435887`, 
                    1.9999795454655793`}, {-1.8235993949142528`, 
                    1.9999828804647248`}}]}, 
                  {RGBColor[0, 0, 1], ArrowBox[{{0, 2.8}, {0, 2.8}}]}},
                 Frame->True,
                 FrameTicks->Automatic,
                 ImageSize->800,
                 PlotRange->{{-3, 3}, {-1, 3}}], 529->
                GraphicsBox[{InsetBox[
                   TagBox[GridBox[{
                    {"\<\"t\"\>", "26.400000000000002`"},
                    {"\<\"driveIntent\"\>", "0"},
                    {"\<\"motorTorque\"\>", 
                    RowBox[{"-", "2.3255292111985417`"}]},
                    {"\<\"vw\"\>", 
                    RowBox[{"-", "0.026724798936035736`"}]},
                    {"\<\"controllerState\"\>", 
                    RowBox[{"\[LeftAssociation]", 
                    RowBox[{
                    
                    RowBox[{"\<\"lastError\"\>", "\[Rule]", 
                    "0.0032857720342688657`"}], ",", 
                    RowBox[{"\<\"lastInt\"\>", "\[Rule]", 
                    RowBox[{"-", "0.022898488828717665`"}]}], ",", 
                    RowBox[{"\<\"initialized\"\>", "\[Rule]", "True"}]}], 
                    "\[RightAssociation]"}]}
                    },
                    AutoDelete->False,
                    
                    GridBoxItemSize->{
                    "Columns" -> {{Automatic}}, "Rows" -> {{Automatic}}}],
                    "Grid"], {-1, -0.5}], 
                  {GrayLevel[0], LineBox[{{-5, 0}, {5, 0}}]}, 
                  {GrayLevel[0.5], 
                   CircleBox[{-1.8321068276266366`, 0.5}, 0.5], 
                   DiskBox[{-1.8321068276266366`, 0.5}, 0.1]}, 
                  {GrayLevel[0.5], 
                   DiskBox[{-1.82500866465349, 1.9999832052667812`}, 0.1]}, 
                  {GrayLevel[0.5], Thickness[0.01], 
                   LineBox[{{-1.8321068276266366`, 0.5}, {-1.82500866465349, 
                    1.9999832052667812`}}]}, 
                  {RGBColor[0, 1, 0], 
                   DiskBox[{-1.8273747189778722`, 1.4999888035111875`}, 
                    0.05]}, 
                  {RGBColor[0, 0, 1], Arrowheads[0.01], 
                   CircleBox[{-1.8321068276266366`, 0.5}, 0.3333333333333333, 
                    NCache[{
                    Rational[-1, 2] 
                    Pi, -1.6405622031308529`}, {-1.5707963267948966`, \
-1.6405622031308529`}]], 
                   ArrowBox[{{-1.8520169284500663`, 
                    0.16726181614321456`}, {-1.8553432593358021`, 
                    0.1674775506036677}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-1.8321068276266366`, 
                    0.5}, {-1.7855962434026658`, 0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-1.8321068276266366`, 
                    0.5}, {-1.8669110378976264`, 0.5}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.005], Arrowheads[0.04], 
                   ArrowBox[{{-1.8321068276266366`, 
                    0.5}, {-1.8204004536736555`, 0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-1.8321068276266366`, 0.5}, 1.6, {
                    1.5660642004850525`, 1.5443113255098626`}], 
                   ArrowBox[{{-1.8057320201905984`, 
                    2.0997826007094567`}, {-1.7897357795419864`, 
                    2.0994388685674137`}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Dashing[{Small, Small}], 
                   Arrowheads[0.02], 
                   CircleBox[{-1.8321068276266366`, 0.5}, 1.7, {
                    1.5660642004850525`, 1.572848460106726}], 
                   ArrowBox[{{-1.8185955965053533`, 
                    2.1999463069854848`}, {-1.8355954518081667`, 
                    2.199996420437796}}]}, 
                  {RGBColor[0, 0, 1], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-1.8321068276266366`, 0.5}, 1.8, {
                    1.5660642004850525`, 1.5789838072139333`}], 
                   ArrowBox[{{-1.8288442941672638`, 
                    2.299997043296301}, {-1.8468441277275331`, 
                    2.2999396689849734`}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.003], Arrowheads[0.025], 
                   CircleBox[{-1.8321068276266366`, 0.5}, 1.5, {
                    1.5660642004850525`, 1.566053489490737}], 
                   ArrowBox[{{-1.8399925353463975`, 
                    1.9999792717280325`}, {-1.8249925983423123`, 
                    1.9999831291523549`}}]}, 
                  {RGBColor[0, 0, 1], ArrowBox[{{0, 2.8}, {0, 2.8}}]}},
                 Frame->True,
                 FrameTicks->Automatic,
                 ImageSize->800,
                 PlotRange->{{-3, 3}, {-1, 3}}], 530->
                GraphicsBox[{InsetBox[
                   TagBox[GridBox[{
                    {"\<\"t\"\>", "26.45`"},
                    {"\<\"driveIntent\"\>", "0"},
                    {"\<\"motorTorque\"\>", 
                    RowBox[{"-", "2.308521873890674`"}]},
                    {"\<\"vw\"\>", 
                    RowBox[{"-", "0.02653326340811178`"}]},
                    {"\<\"controllerState\"\>", 
                    RowBox[{"\[LeftAssociation]", 
                    RowBox[{
                    
                    RowBox[{"\<\"lastError\"\>", "\[Rule]", 
                    "0.0032629059719892113`"}], ",", 
                    RowBox[{"\<\"lastInt\"\>", "\[Rule]", 
                    RowBox[{"-", "0.022734783879031863`"}]}], ",", 
                    RowBox[{"\<\"initialized\"\>", "\[Rule]", "True"}]}], 
                    "\[RightAssociation]"}]}
                    },
                    AutoDelete->False,
                    
                    GridBoxItemSize->{
                    "Columns" -> {{Automatic}}, "Rows" -> {{Automatic}}}],
                    "Grid"], {-1, -0.5}], 
                  {GrayLevel[0], LineBox[{{-5, 0}, {5, 0}}]}, 
                  {GrayLevel[0.5], 
                   CircleBox[{-1.8334381775559216`, 0.5}, 0.5], 
                   DiskBox[{-1.8334381775559216`, 0.5}, 0.1]}, 
                  {GrayLevel[0.5], 
                   DiskBox[{-1.826391909986369, 1.9999834499464781`}, 0.1]}, 
                  {GrayLevel[0.5], Thickness[0.01], 
                   LineBox[{{-1.8334381775559216`, 0.5}, {-1.826391909986369, 
                    1.9999834499464781`}}]}, 
                  {RGBColor[0, 1, 0], 
                   DiskBox[{-1.8287406658428866`, 1.4999889666309854`}, 
                    0.05]}, 
                  {RGBColor[0, 0, 1], Arrowheads[0.01], 
                   CircleBox[{-1.8334381775559216`, 0.5}, 0.3333333333333333, 
                    NCache[{
                    Rational[-1, 2] 
                    Pi, -1.6400519830116167`}, {-1.5707963267948966`, \
-1.6400519830116167`}]], 
                   ArrowBox[{{-1.8531785060793313`, 
                    0.16725170091960057`}, {-1.8565049466041617`, 
                    0.16746573819110572`}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-1.8334381775559216`, 0.5}, {-1.787267740078108,
                     0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-1.8334381775559216`, 
                    0.5}, {-1.8679879362819545`, 0.5}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.005], Arrowheads[0.04], 
                   ArrowBox[{{-1.8334381775559216`, 
                    0.5}, {-1.8218174988041411`, 0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-1.8334381775559216`, 0.5}, 1.6, {
                    1.5660987978053251`, 1.5445049603487775`}], 
                   ArrowBox[{{-1.8073731442585892`, 
                    2.0997876777995286`}, {-1.7913768373513448`, 
                    2.099447043093392}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Dashing[{Small, Small}], 
                   Arrowheads[0.02], 
                   CircleBox[{-1.8334381775559216`, 0.5}, 1.7, {
                    1.5660987978053251`, 1.572833395136148}], 
                   ArrowBox[{{-1.8199013367950734`, 
                    2.199946103246281}, {-1.836901191341, 
                    2.1999964728009065`}}]}, 
                  {RGBColor[0, 0, 1], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-1.8334381775559216`, 0.5}, 1.8, {
                    1.5660987978053251`, 1.57892391932694}], 
                   ArrowBox[{{-1.8300678460829467`, 
                    2.2999968446821684`}, {-1.8480676830469638`, 
                    2.299940548342941}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.003], Arrowheads[0.025], 
                   CircleBox[{-1.8334381775559216`, 0.5}, 1.5, {
                    1.5660987978053251`, 1.5660880606525447`}], 
                   ArrowBox[{{-1.841375741297071, 
                    1.9999789982135934`}, {-1.8263758044353051`, 
                    1.9999833742031627`}}]}, 
                  {RGBColor[0, 0, 1], ArrowBox[{{0, 2.8}, {0, 2.8}}]}},
                 Frame->True,
                 FrameTicks->Automatic,
                 ImageSize->800,
                 PlotRange->{{-3, 3}, {-1, 3}}], 531->
                GraphicsBox[{InsetBox[
                   TagBox[GridBox[{
                    {"\<\"t\"\>", "26.5`"},
                    {"\<\"driveIntent\"\>", "0"},
                    {"\<\"motorTorque\"\>", 
                    RowBox[{"-", "2.2916565475633726`"}]},
                    {"\<\"vw\"\>", 
                    RowBox[{"-", "0.026343129179016187`"}]},
                    {"\<\"controllerState\"\>", 
                    RowBox[{"\[LeftAssociation]", 
                    RowBox[{
                    
                    RowBox[{"\<\"lastError\"\>", "\[Rule]", 
                    "0.003240174112180946`"}], ",", 
                    RowBox[{"\<\"lastInt\"\>", "\[Rule]", 
                    RowBox[{"-", "0.022572218793711938`"}]}], ",", 
                    RowBox[{"\<\"initialized\"\>", "\[Rule]", "True"}]}], 
                    "\[RightAssociation]"}]}
                    },
                    AutoDelete->False,
                    
                    GridBoxItemSize->{
                    "Columns" -> {{Automatic}}, "Rows" -> {{Automatic}}}],
                    "Grid"], {-1, -0.5}], 
                  {GrayLevel[0], LineBox[{{-5, 0}, {5, 0}}]}, 
                  {GrayLevel[0.5], 
                   CircleBox[{-1.8347599864922066`, 0.5}, 0.5], 
                   DiskBox[{-1.8347599864922066`, 0.5}, 0.1]}, 
                  {GrayLevel[0.5], 
                   DiskBox[{-1.8277651849393062`, 1.9999836908284154`}, 0.1]}, 
                  {GrayLevel[0.5], Thickness[0.01], 
                   LineBox[{{-1.8347599864922066`, 
                    0.5}, {-1.8277651849393062`, 1.9999836908284154`}}]}, 
                  {RGBColor[0, 1, 0], 
                   DiskBox[{-1.8300967854569397`, 1.4999891272189436`}, 
                    0.05]}, 
                  {RGBColor[0, 0, 1], Arrowheads[0.01], 
                   CircleBox[{-1.8347599864922066`, 0.5}, 0.3333333333333333, 
                    NCache[{
                    Rational[-1, 2] 
                    Pi, -1.6395460232217978`}, {-1.5707963267948966`, \
-1.6395460232217978`}]], 
                   ArrowBox[{{-1.8543319552366184`, 
                    0.16724175569855693`}, {-1.8576585036299205`, 
                    0.1674541098975888}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-1.8347599864922066`, 
                    0.5}, {-1.7889268555409392`, 0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-1.8347599864922066`, 0.5}, {-1.869057398904272,
                     0.5}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.005], Arrowheads[0.04], 
                   ArrowBox[{{-1.8347599864922066`, 
                    0.5}, {-1.8232242679530046`, 0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-1.8347599864922066`, 0.5}, 1.6, {
                    1.5661331088589014`, 1.5446969930311663`}], 
                   ArrowBox[{{-1.8090021651926007`, 
                    2.0997926536404328`}, {-1.7930057931673047`, 
                    2.0994550907542813`}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Dashing[{Small, Small}], 
                   Arrowheads[0.02], 
                   CircleBox[{-1.8347599864922066`, 0.5}, 1.7, {
                    1.5661331088589014`, 1.5728184698171719`}], 
                   ArrowBox[{{-1.821197773495034, 
                    2.19994590101527}, {-1.8381976272872855`, 
                    2.1999965242981423`}}]}, 
                  {RGBColor[0, 0, 1], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-1.8347599864922066`, 0.5}, 1.8, {
                    1.5661331088589014`, 1.5788645341231424`}], 
                   ArrowBox[{{-1.8312827618457812`, 
                    2.2999966413604103`}, {-1.8492826021212285`, 
                    2.2999414139452683`}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.003], Arrowheads[0.025], 
                   CircleBox[{-1.8347599864922066`, 0.5}, 1.5, {
                    1.5661331088589014`, 1.5661223713791133`}], 
                   ArrowBox[{{-1.8427490155979402`, 
                    1.9999787249871073`}, {-1.8277490788951467`, 
                    1.999983615635406}}]}, 
                  {RGBColor[0, 0, 1], ArrowBox[{{0, 2.8}, {0, 2.8}}]}},
                 Frame->True,
                 FrameTicks->Automatic,
                 ImageSize->800,
                 PlotRange->{{-3, 3}, {-1, 3}}], 532->
                GraphicsBox[{InsetBox[
                   TagBox[GridBox[{
                    {"\<\"t\"\>", "26.55`"},
                    {"\<\"driveIntent\"\>", "0"},
                    {"\<\"motorTorque\"\>", 
                    RowBox[{"-", "2.27493311910747`"}]},
                    {"\<\"vw\"\>", 
                    RowBox[{"-", "0.02615438419529954`"}]},
                    {"\<\"controllerState\"\>", 
                    RowBox[{"\[LeftAssociation]", 
                    RowBox[{
                    
                    RowBox[{"\<\"lastError\"\>", "\[Rule]", 
                    "0.0032175728114168045`"}], ",", 
                    RowBox[{"\<\"lastInt\"\>", "\[Rule]", 
                    RowBox[{"-", "0.022410786958331246`"}]}], ",", 
                    RowBox[{"\<\"initialized\"\>", "\[Rule]", "True"}]}], 
                    "\[RightAssociation]"}]}
                    },
                    AutoDelete->False,
                    
                    GridBoxItemSize->{
                    "Columns" -> {{Automatic}}, "Rows" -> {{Automatic}}}],
                    "Grid"], {-1, -0.5}], 
                  {GrayLevel[0], LineBox[{{-5, 0}, {5, 0}}]}, 
                  {GrayLevel[0.5], 
                   CircleBox[{-1.8360723241929089`, 0.5}, 0.5], 
                   DiskBox[{-1.8360723241929089`, 0.5}, 0.1]}, 
                  {GrayLevel[0.5], 
                   DiskBox[{-1.8291285592484383`, 1.999983927956695}, 0.1]}, 
                  {GrayLevel[0.5], Thickness[0.01], 
                   LineBox[{{-1.8360723241929089`, 
                    0.5}, {-1.8291285592484383`, 1.999983927956695}}]}, 
                  {RGBColor[0, 1, 0], 
                   DiskBox[{-1.831443147563262, 1.4999892853044634`}, 0.05]}, 
                  {RGBColor[0, 0, 1], Arrowheads[0.01], 
                   CircleBox[{-1.8360723241929089`, 0.5}, 0.3333333333333333, 
                    NCache[{
                    Rational[-1, 2] 
                    Pi, -1.6390443203681206`}, {-1.5707963267948966`, \
-1.6390443203681206`}]], 
                   ArrowBox[{{-1.855477344720385, 
                    0.16723197826497915`}, {-1.8588039992337346`, 
                    0.16744266349853382`}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-1.8360723241929089`, 
                    0.5}, {-1.7905736618107595`, 0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-1.8360723241929089`, 
                    0.5}, {-1.8701194956302698`, 0.5}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.005], Arrowheads[0.04], 
                   ArrowBox[{{-1.8360723241929089`, 
                    0.5}, {-1.8246208332481204`, 0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-1.8360723241929089`, 0.5}, 1.6, {
                    1.5661671336317726`, 1.5448874234778067`}], 
                   ArrowBox[{{-1.8106191525880642`, 
                    2.0997975297065734`}, {-1.7946227165705617`, 
                    2.0994630130227936`}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Dashing[{Small, Small}], 
                   Arrowheads[0.02], 
                   CircleBox[{-1.8360723241929089`, 0.5}, 1.7, {
                    1.5661671336317726`, 1.5728036827628167`}], 
                   ArrowBox[{{-1.8224849740047808`, 
                    2.1999457002842373`}, {-1.8394848270466044`, 
                    2.1999965749448656`}}]}, 
                  {RGBColor[0, 0, 1], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-1.8360723241929089`, 0.5}, 1.8, {
                    1.5661671336317726`, 1.5788056509601476`}], 
                   ArrowBox[{{-1.8324891100569494`, 
                    2.2999964334899268`}, {-1.8504889535531568`, 
                    2.2999422659624082`}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.003], Arrowheads[0.025], 
                   CircleBox[{-1.8360723241929089`, 0.5}, 1.5, {
                    1.5661671336317726`, 1.5661564192767752`}], 
                   ArrowBox[{{-1.8441124244161253`, 
                    1.9999784521080297`}, {-1.8291124878885425`, 
                    1.9999838534726353`}}]}, 
                  {RGBColor[0, 0, 1], ArrowBox[{{0, 2.8}, {0, 2.8}}]}},
                 Frame->True,
                 FrameTicks->Automatic,
                 ImageSize->800,
                 PlotRange->{{-3, 3}, {-1, 3}}], 533->
                GraphicsBox[{InsetBox[
                   TagBox[GridBox[{
                    {"\<\"t\"\>", "26.6`"},
                    {"\<\"driveIntent\"\>", "0"},
                    {"\<\"motorTorque\"\>", 
                    RowBox[{"-", "2.258351170155863`"}]},
                    {"\<\"vw\"\>", 
                    RowBox[{"-", "0.025967016454816742`"}]},
                    {"\<\"controllerState\"\>", 
                    RowBox[{"\[LeftAssociation]", 
                    RowBox[{
                    
                    RowBox[{"\<\"lastError\"\>", "\[Rule]", 
                    "0.00319509906824893`"}], ",", 
                    RowBox[{"\<\"lastInt\"\>", "\[Rule]", 
                    RowBox[{"-", "0.02225048192403282`"}]}], ",", 
                    RowBox[{"\<\"initialized\"\>", "\[Rule]", "True"}]}], 
                    "\[RightAssociation]"}]}
                    },
                    AutoDelete->False,
                    
                    GridBoxItemSize->{
                    "Columns" -> {{Automatic}}, "Rows" -> {{Automatic}}}],
                    "Grid"], {-1, -0.5}], 
                  {GrayLevel[0], LineBox[{{-5, 0}, {5, 0}}]}, 
                  {GrayLevel[0.5], 
                   CircleBox[{-1.8373752598139772`, 0.5}, 0.5], 
                   DiskBox[{-1.8373752598139772`, 0.5}, 0.1]}, 
                  {GrayLevel[0.5], 
                   DiskBox[{-1.830482102986647, 1.999984161379364}, 0.1]}, 
                  {GrayLevel[0.5], Thickness[0.01], 
                   LineBox[{{-1.8373752598139772`, 0.5}, {-1.830482102986647, 
                    1.999984161379364}}]}, 
                  {RGBColor[0, 1, 0], 
                   DiskBox[{-1.8327798219290903`, 1.499989440919576}, 0.05]}, 
                  {RGBColor[0, 0, 1], Arrowheads[0.01], 
                   CircleBox[{-1.8373752598139772`, 0.5}, 0.3333333333333333, 
                    NCache[{
                    Rational[-1, 2] 
                    Pi, -1.6385468618995724`}, {-1.5707963267948966`, \
-1.6385468618995724`}]], 
                   ArrowBox[{{-1.8566147396767758`, 
                    0.1672223662477899}, {-1.8599414985856595`, 
                    0.1674313965828847}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-1.8373752598139772`, 0.5}, {-1.79220823641086, 
                    0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-1.8373752598139772`, 
                    0.5}, {-1.8711742911238882`, 0.5}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.005], Arrowheads[0.04], 
                   ArrowBox[{{-1.8373752598139772`, 0.5}, {-1.826007267720771,
                     0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-1.8373752598139772`, 0.5}, 1.6, {
                    1.5662008727354086`, 1.5450762551101778`}], 
                   ArrowBox[{{-1.8122241810401316`, 
                    2.0998023075481895`}, {-1.7962276821404937`, 
                    2.0994708115034815`}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Dashing[{Small, Small}], 
                   Arrowheads[0.02], 
                   CircleBox[{-1.8373752598139772`, 0.5}, 1.7, {
                    1.5662008727354086`, 1.5727890323394524`}], 
                   ArrowBox[{{-1.8237630047031028`, 
                    2.1999455010413707`}, {-1.8407628569977668`, 
                    2.1999966247570377`}}]}, 
                  {RGBColor[0, 0, 1], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-1.8373752598139772`, 0.5}, 1.8, {
                    1.5662008727354086`, 1.5787472681251635`}], 
                   ArrowBox[{{-1.8336869567894358`, 
                    2.299996221224033}, {-1.8516868034174205`, 
                    2.2999431045784995`}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.003], Arrowheads[0.025], 
                   CircleBox[{-1.8373752598139772`, 0.5}, 1.5, {
                    1.5662008727354086`, 1.5661902026207846`}], 
                   ArrowBox[{{-1.8454660343206502`, 
                    1.9999781796305844`}, {-1.8304660979841032`, 
                    1.999984087743203}}]}, 
                  {RGBColor[0, 0, 1], ArrowBox[{{0, 2.8}, {0, 2.8}}]}},
                 Frame->True,
                 FrameTicks->Automatic,
                 ImageSize->800,
                 PlotRange->{{-3, 3}, {-1, 3}}], 534->
                GraphicsBox[{InsetBox[
                   TagBox[GridBox[{
                    {"\<\"t\"\>", "26.650000000000002`"},
                    {"\<\"driveIntent\"\>", "0"},
                    {"\<\"motorTorque\"\>", 
                    RowBox[{"-", "2.2419100085675385`"}]},
                    {"\<\"vw\"\>", 
                    RowBox[{"-", "0.02578101403004764`"}]},
                    {"\<\"controllerState\"\>", 
                    RowBox[{"\[LeftAssociation]", 
                    RowBox[{
                    
                    RowBox[{"\<\"lastError\"\>", "\[Rule]", 
                    "0.0031727504679337753`"}], ",", 
                    RowBox[{"\<\"lastInt\"\>", "\[Rule]", 
                    RowBox[{"-", "0.022091297376842182`"}]}], ",", 
                    RowBox[{"\<\"initialized\"\>", "\[Rule]", "True"}]}], 
                    "\[RightAssociation]"}]}
                    },
                    AutoDelete->False,
                    
                    GridBoxItemSize->{
                    "Columns" -> {{Automatic}}, "Rows" -> {{Automatic}}}],
                    "Grid"], {-1, -0.5}], 
                  {GrayLevel[0], LineBox[{{-5, 0}, {5, 0}}]}, 
                  {GrayLevel[0.5], 
                   CircleBox[{-1.8386688619130673`, 0.5}, 0.5], 
                   DiskBox[{-1.8386688619130673`, 0.5}, 0.1]}, 
                  {GrayLevel[0.5], 
                   DiskBox[{-1.8318258864733423`, 1.9999843911478317`}, 0.1]}, 
                  {GrayLevel[0.5], Thickness[0.01], 
                   LineBox[{{-1.8386688619130673`, 
                    0.5}, {-1.8318258864733423`, 1.9999843911478317`}}]}, 
                  {RGBColor[0, 1, 0], 
                   DiskBox[{-1.834106878286584, 1.4999895940985546`}, 0.05]}, 
                  {RGBColor[0, 0, 1], Arrowheads[0.01], 
                   CircleBox[{-1.8386688619130673`, 0.5}, 0.3333333333333333, 
                    NCache[{
                    Rational[-1, 2] 
                    Pi, -1.6380536270519228`}, {-1.5707963267948966`, \
-1.6380536270519228`}]], 
                   ArrowBox[{{-1.85774420191674, 
                    0.16721291714542857`}, {-1.861071063521997, 
                    0.1674203065817398}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-1.8386688619130673`, 
                    0.5}, {-1.7938306617417166`, 0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-1.8386688619130673`, 
                    0.5}, {-1.8722218453083852`, 0.5}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.005], Arrowheads[0.04], 
                   ArrowBox[{{-1.8386688619130673`, 
                    0.5}, {-1.8273836451370344`, 0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-1.8386688619130673`, 0.5}, 1.6, {
                    1.5662343273444967`, 1.5452634945020032`}], 
                   ArrowBox[{{-1.8138173295894584`, 
                    2.0998069887774493`}, {-1.7978207689011103`, 
                    2.0994784879132595`}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Dashing[{Small, Small}], 
                   Arrowheads[0.02], 
                   CircleBox[{-1.8386688619130673`, 0.5}, 1.7, {
                    1.5662343273444967`, 1.572774516704096}], 
                   ArrowBox[{{-1.825031931014609, 
                    2.1999453032717464`}, {-1.8420317825653887`, 
                    2.1999966737510652`}}]}, 
                  {RGBColor[0, 0, 1], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-1.8386688619130673`, 0.5}, 1.8, {
                    1.5662343273444967`, 1.5786893829476496`}], 
                   ArrowBox[{{-1.8348763657939735`, 
                    2.29999600471034}, {-1.8528762154664684`, 
                    2.2999439299892117`}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.003], Arrowheads[0.025], 
                   CircleBox[{-1.8386688619130673`, 0.5}, 1.5, {
                    1.5662343273444967`, 1.5662237203104998`}], 
                   ArrowBox[{{-1.8468099122183905`, 
                    1.9999779076039508`}, {-1.8318099760882958`, 
                    1.9999843184797774`}}]}, 
                  {RGBColor[0, 0, 1], ArrowBox[{{0, 2.8}, {0, 2.8}}]}},
                 Frame->True,
                 FrameTicks->Automatic,
                 ImageSize->800,
                 PlotRange->{{-3, 3}, {-1, 3}}], 535->
                GraphicsBox[{InsetBox[
                   TagBox[GridBox[{
                    {"\<\"t\"\>", "26.7`"},
                    {"\<\"driveIntent\"\>", "0"},
                    {"\<\"motorTorque\"\>", 
                    RowBox[{"-", "2.22560869898644`"}]},
                    {"\<\"vw\"\>", 
                    RowBox[{"-", "0.02559636508864824`"}]},
                    {"\<\"controllerState\"\>", 
                    RowBox[{"\[LeftAssociation]", 
                    RowBox[{
                    
                    RowBox[{"\<\"lastError\"\>", "\[Rule]", 
                    "0.0031505251279348835`"}], ",", 
                    RowBox[{"\<\"lastInt\"\>", "\[Rule]", 
                    RowBox[{"-", "0.021933227109725964`"}]}], ",", 
                    RowBox[{"\<\"initialized\"\>", "\[Rule]", "True"}]}], 
                    "\[RightAssociation]"}]}
                    },
                    AutoDelete->False,
                    
                    GridBoxItemSize->{
                    "Columns" -> {{Automatic}}, "Rows" -> {{Automatic}}}],
                    "Grid"], {-1, -0.5}], 
                  {GrayLevel[0], LineBox[{{-5, 0}, {5, 0}}]}, 
                  {GrayLevel[0.5], CircleBox[{-1.839953198453804, 0.5}, 0.5], 
                   DiskBox[{-1.839953198453804, 0.5}, 0.1]}, 
                  {GrayLevel[0.5], 
                   DiskBox[{-1.8331599801877654`, 1.999984617316321}, 0.1]}, 
                  {GrayLevel[0.5], Thickness[0.01], 
                   LineBox[{{-1.839953198453804, 0.5}, {-1.8331599801877654`, 
                    1.999984617316321}}]}, 
                  {RGBColor[0, 1, 0], 
                   DiskBox[{-1.835424386276445, 1.4999897448775474`}, 0.05]}, 
                  {RGBColor[0, 0, 1], Arrowheads[0.01], 
                   CircleBox[{-1.839953198453804, 0.5}, 0.3333333333333333, 
                    NCache[{
                    Rational[-1, 2] 
                    Pi, -1.6375645877644898`}, {-1.5707963267948966`, \
-1.6375645877644898`}]], 
                   ArrowBox[{{-1.8588657902250738`, 
                    0.16720362834970032`}, {-1.862192752854084, 
                    0.16740939079524797`}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-1.839953198453804, 0.5}, {-1.7954410244740753`,
                     0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-1.839953198453804, 0.5}, {-1.8732622138165935`,
                     0.5}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.005], Arrowheads[0.04], 
                   ArrowBox[{{-1.839953198453804, 0.5}, {-1.8287500398368648`,
                     0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-1.839953198453804, 0.5}, 1.6, {
                    1.5662674991362995`, 1.545449151039847}], 
                   ArrowBox[{{-1.8153986811835392`, 
                    2.0998115750555204`}, {-1.7994020597825457`, 
                    2.09948604406305}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Dashing[{Small, Small}], 
                   Arrowheads[0.02], 
                   CircleBox[{-1.839953198453804, 0.5}, 1.7, {
                    1.5662674991362995`, 1.572760133839492}], 
                   ArrowBox[{{-1.8262918174736258`, 
                    2.1999451069577844`}, {-1.8432916682837928`, 
                    2.1999967219436614`}}]}, 
                  {RGBColor[0, 0, 1], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-1.839953198453804, 0.5}, 1.8, {
                    1.5662674991362995`, 1.5786319919084464`}], 
                   ArrowBox[{{-1.8360573986997406`, 
                    2.2999957840906955`}, {-1.8540572513312146`, 
                    2.29994474239973}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.003], Arrowheads[0.025], 
                   CircleBox[{-1.839953198453804, 0.5}, 1.5, {
                    1.5662674991362995`, 1.5662569718236115`}], 
                   ArrowBox[{{-1.8481441252896793`, 
                    1.9999776360724746`}, {-1.8331441893810483`, 
                    1.999984545718871}}]}, 
                  {RGBColor[0, 0, 1], ArrowBox[{{0, 2.8}, {0, 2.8}}]}},
                 Frame->True,
                 FrameTicks->Automatic,
                 ImageSize->800,
                 PlotRange->{{-3, 3}, {-1, 3}}], 536->
                GraphicsBox[{InsetBox[
                   TagBox[GridBox[{
                    {"\<\"t\"\>", "26.75`"},
                    {"\<\"driveIntent\"\>", "0"},
                    {"\<\"motorTorque\"\>", 
                    RowBox[{"-", "2.2094460922876142`"}]},
                    {"\<\"vw\"\>", 
                    RowBox[{"-", "0.02541305791134854`"}]},
                    {"\<\"controllerState\"\>", 
                    RowBox[{"\[LeftAssociation]", 
                    RowBox[{
                    
                    RowBox[{"\<\"lastError\"\>", "\[Rule]", 
                    "0.0031284216446364477`"}], ",", 
                    RowBox[{"\<\"lastInt\"\>", "\[Rule]", 
                    RowBox[{"-", "0.021776264997346247`"}]}], ",", 
                    RowBox[{"\<\"initialized\"\>", "\[Rule]", "True"}]}], 
                    "\[RightAssociation]"}]}
                    },
                    AutoDelete->False,
                    
                    GridBoxItemSize->{
                    "Columns" -> {{Automatic}}, "Rows" -> {{Automatic}}}],
                    "Grid"], {-1, -0.5}], 
                  {GrayLevel[0], LineBox[{{-5, 0}, {5, 0}}]}, 
                  {GrayLevel[0.5], 
                   CircleBox[{-1.8412283368110092`, 0.5}, 0.5], 
                   DiskBox[{-1.8412283368110092`, 0.5}, 0.1]}, 
                  {GrayLevel[0.5], 
                   DiskBox[{-1.8344844546862746`, 1.9999848399413533`}, 0.1]}, 
                  {GrayLevel[0.5], Thickness[0.01], 
                   LineBox[{{-1.8412283368110092`, 
                    0.5}, {-1.8344844546862746`, 1.9999848399413533`}}]}, 
                  {RGBColor[0, 1, 0], 
                   DiskBox[{-1.8367324153945195`, 1.4999898932942357`}, 
                    0.05]}, 
                  {RGBColor[0, 0, 1], Arrowheads[0.01], 
                   CircleBox[{-1.8412283368110092`, 0.5}, 0.3333333333333333, 
                    NCache[{
                    Rational[-1, 2] 
                    Pi, -1.637079709563525}, {-1.5707963267948966`, \
-1.637079709563525}]], 
                   ArrowBox[{{-1.8599795606593885`, 
                    0.16719449716794188`}, {-1.8633066226670234`, 
                    0.1673986464177104}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-1.8412283368110092`, 
                    0.5}, {-1.7970394149652569`, 0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-1.8412283368110092`, 
                    0.5}, {-1.8742954484273278`, 0.5}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.005], Arrowheads[0.04], 
                   ArrowBox[{{-1.8412283368110092`, 
                    0.5}, {-1.8301065265815755`, 0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-1.8412283368110092`, 0.5}, 1.6, {
                    1.5663003902320274`, 1.5456332365949874`}], 
                   ArrowBox[{{-1.8169683221570407`, 
                    2.099816068080637}, {-1.8009716411015368`, 
                    2.0994934818405966`}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Dashing[{Small, Small}], 
                   Arrowheads[0.02], 
                   CircleBox[{-1.8412283368110092`, 0.5}, 1.7, {
                    1.5663003902320274`, 1.5727458815869026`}], 
                   ArrowBox[{{-1.8275427277851668`, 
                    2.19994491207968}, {-1.8445425778579772`, 
                    2.1999967693517193`}}]}, 
                  {RGBColor[0, 0, 1], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-1.8412283368110092`, 0.5}, 1.8, {
                    1.5663003902320274`, 1.5785750907447362`}], 
                   ArrowBox[{{-1.8372301152085198`, 
                    2.2999955595011943`}, {-1.8552299707151858`, 
                    2.299945542022873}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.003], Arrowheads[0.025], 
                   CircleBox[{-1.8412283368110092`, 0.5}, 1.5, {
                    1.5663003902320274`, 1.566289957169961}], 
                   ArrowBox[{{-1.8494687409242616`, 
                    1.9999773650759034`}, {-1.834468805251708, 
                    1.9999847695003767`}}]}, 
                  {RGBColor[0, 0, 1], ArrowBox[{{0, 2.8}, {0, 2.8}}]}},
                 Frame->True,
                 FrameTicks->Automatic,
                 ImageSize->800,
                 PlotRange->{{-3, 3}, {-1, 3}}], 537->
                GraphicsBox[{InsetBox[
                   TagBox[GridBox[{
                    {"\<\"t\"\>", "26.8`"},
                    {"\<\"driveIntent\"\>", "0"},
                    {"\<\"motorTorque\"\>", 
                    RowBox[{"-", "2.193420853754844`"}]},
                    {"\<\"vw\"\>", 
                    RowBox[{"-", "0.025231080907324187`"}]},
                    {"\<\"controllerState\"\>", 
                    RowBox[{"\[LeftAssociation]", 
                    RowBox[{
                    
                    RowBox[{"\<\"lastError\"\>", "\[Rule]", 
                    "0.0031064390416391634`"}], ",", 
                    RowBox[{"\<\"lastInt\"\>", "\[Rule]", 
                    RowBox[{"-", "0.021620404973440272`"}]}], ",", 
                    RowBox[{"\<\"initialized\"\>", "\[Rule]", "True"}]}], 
                    "\[RightAssociation]"}]}
                    },
                    AutoDelete->False,
                    
                    GridBoxItemSize->{
                    "Columns" -> {{Automatic}}, "Rows" -> {{Automatic}}}],
                    "Grid"], {-1, -0.5}], 
                  {GrayLevel[0], LineBox[{{-5, 0}, {5, 0}}]}, 
                  {GrayLevel[0.5], 
                   CircleBox[{-1.8424943437767585`, 0.5}, 0.5], 
                   DiskBox[{-1.8424943437767585`, 0.5}, 0.1]}, 
                  {GrayLevel[0.5], 
                   DiskBox[{-1.8357993805239838`, 1.9999850590812709`}, 0.1]}, 
                  {GrayLevel[0.5], Thickness[0.01], 
                   LineBox[{{-1.8424943437767585`, 
                    0.5}, {-1.8357993805239838`, 1.9999850590812709`}}]}, 
                  {RGBColor[0, 1, 0], 
                   DiskBox[{-1.8380310349415754`, 1.499990039387514}, 0.05]}, 
                  {RGBColor[0, 0, 1], Arrowheads[0.01], 
                   CircleBox[{-1.8424943437767585`, 0.5}, 0.3333333333333333, 
                    NCache[{
                    Rational[-1, 2] 
                    Pi, -1.636598952407542}, {-1.5707963267948966`, \
-1.636598952407542}]], 
                   ArrowBox[{{-1.861085566837303, 
                    0.16718552084348226`}, {-1.864412726606659, 
                    0.1673880705608516}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-1.8424943437767585`, 
                    0.5}, {-1.7986259267016615`, 0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-1.8424943437767585`, 
                    0.5}, {-1.8753215974853594`, 0.5}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.005], Arrowheads[0.04], 
                   ArrowBox[{{-1.8424943437767585`, 
                    0.5}, {-1.8314531804102625`, 0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-1.8424943437767585`, 0.5}, 1.6, {
                    1.5663330031405578`, 1.545815765208426}], 
                   ArrowBox[{{-1.8185263417339637`, 
                    2.099820469577158}, {-1.8025296020637194`, 
                    2.0995008031944877`}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Dashing[{Small, Small}], 
                   Arrowheads[0.02], 
                   CircleBox[{-1.8424943437767585`, 0.5}, 1.7, {
                    1.5663330031405578`, 1.5727317576765407`}], 
                   ArrowBox[{{-1.8287847248827236`, 
                    2.199944718615808}, {-1.8457845742214107`, 
                    2.199996815992201}}]}, 
                  {RGBColor[0, 0, 1], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-1.8424943437767585`, 0.5}, 1.8, {
                    1.5663330031405578`, 1.578518674550307}], 
                   ArrowBox[{{-1.838394573281229, 
                    2.299995331072246}, {-1.8563944315810463`, 
                    2.2999463290773514`}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.003], Arrowheads[0.025], 
                   CircleBox[{-1.8424943437767585`, 0.5}, 1.5, {
                    1.5663330031405578`, 1.5663226768454301`}], 
                   ArrowBox[{{-1.8507838266581877`, 
                    1.9999770946496345`}, {-1.8357838912359337`, 
                    1.999984989867131}}]}, 
                  {RGBColor[0, 0, 1], ArrowBox[{{0, 2.8}, {0, 2.8}}]}},
                 Frame->True,
                 FrameTicks->Automatic,
                 ImageSize->800,
                 PlotRange->{{-3, 3}, {-1, 3}}], 538->
                GraphicsBox[{InsetBox[
                   TagBox[GridBox[{
                    {"\<\"t\"\>", "26.85`"},
                    {"\<\"driveIntent\"\>", "0"},
                    {"\<\"motorTorque\"\>", 
                    RowBox[{"-", "2.1775314898649047`"}]},
                    {"\<\"vw\"\>", 
                    RowBox[{"-", "0.02505042262717984`"}]},
                    {"\<\"controllerState\"\>", 
                    RowBox[{"\[LeftAssociation]", 
                    RowBox[{
                    
                    RowBox[{"\<\"lastError\"\>", "\[Rule]", 
                    "0.003084576719952478`"}], ",", 
                    RowBox[{"\<\"lastInt\"\>", "\[Rule]", 
                    RowBox[{"-", "0.02146564101073818`"}]}], ",", 
                    RowBox[{"\<\"initialized\"\>", "\[Rule]", "True"}]}], 
                    "\[RightAssociation]"}]}
                    },
                    AutoDelete->False,
                    
                    GridBoxItemSize->{
                    "Columns" -> {{Automatic}}, "Rows" -> {{Automatic}}}],
                    "Grid"], {-1, -0.5}], 
                  {GrayLevel[0], LineBox[{{-5, 0}, {5, 0}}]}, 
                  {GrayLevel[0.5], CircleBox[{-1.843751285567143, 0.5}, 0.5], 
                   DiskBox[{-1.843751285567143, 0.5}, 0.1]}, 
                  {GrayLevel[0.5], 
                   DiskBox[{-1.8371048281810312`, 1.9999852747957942`}, 0.1]}, 
                  {GrayLevel[0.5], Thickness[0.01], 
                   LineBox[{{-1.843751285567143, 0.5}, {-1.8371048281810312`, 
                    1.9999852747957942`}}]}, 
                  {RGBColor[0, 1, 0], 
                   DiskBox[{-1.8393203139764018`, 1.4999901831971962`}, 
                    0.05]}, 
                  {RGBColor[0, 0, 1], Arrowheads[0.01], 
                   CircleBox[{-1.843751285567143, 0.5}, 0.3333333333333333, 
                    NCache[{
                    Rational[-1, 2] 
                    Pi, -1.6361222714908437`}, {-1.5707963267948966`, \
-1.6361222714908437`}]], 
                   ArrowBox[{{-1.8621838602104883`, 
                    0.1671766965744006}, {-1.8655111161534192`, 
                    0.16737766027524903`}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-1.843751285567143, 0.5}, {-1.800200655769845, 
                    0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-1.843751285567143, 0.5}, {-1.8763407063028423`,
                     0.5}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.005], Arrowheads[0.04], 
                   ArrowBox[{{-1.843751285567143, 0.5}, {-1.8327900765055443`,
                     0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-1.843751285567143, 0.5}, 1.6, {
                    1.5663653407047735`, 1.5459967527905505`}], 
                   ArrowBox[{{-1.8200728315539436`, 
                    2.0998247812856086`}, {-1.8040760342900526`, 
                    2.0995080101193926`}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Dashing[{Small, Small}], 
                   Arrowheads[0.02], 
                   CircleBox[{-1.843751285567143, 0.5}, 1.7, {
                    1.5663653407047735`, 1.5727177597556345`}], 
                   ArrowBox[{{-1.830017870982736, 
                    2.199944526543103}, {-1.8470177195905029`, 
                    2.1999968618820365`}}]}, 
                  {RGBColor[0, 0, 1], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-1.843751285567143, 0.5}, 1.8, {
                    1.5663653407047735`, 1.5784627378706897`}], 
                   ArrowBox[{{-1.839550829315933, 
                    2.299995098928684}, {-1.8575506903285992`, 
                    2.299947103786172}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.003], Arrowheads[0.025], 
                   CircleBox[{-1.843751285567143, 0.5}, 1.5, {
                    1.5663653407047735`, 1.566355131786344}], 
                   ArrowBox[{{-1.852089450112177, 
                    1.9999768248249772`}, {-1.8370895149540616`, 
                    1.9999852068644872`}}]}, 
                  {RGBColor[0, 0, 1], ArrowBox[{{0, 2.8}, {0, 2.8}}]}},
                 Frame->True,
                 FrameTicks->Automatic,
                 ImageSize->800,
                 PlotRange->{{-3, 3}, {-1, 3}}], 539->
                GraphicsBox[{InsetBox[
                   TagBox[GridBox[{
                    {"\<\"t\"\>", "26.900000000000002`"},
                    {"\<\"driveIntent\"\>", "0"},
                    {"\<\"motorTorque\"\>", 
                    RowBox[{"-", "2.161776373580469`"}]},
                    {"\<\"vw\"\>", 
                    RowBox[{"-", "0.024871071773688525`"}]},
                    {"\<\"controllerState\"\>", 
                    RowBox[{"\[LeftAssociation]", 
                    RowBox[{
                    
                    RowBox[{"\<\"lastError\"\>", "\[Rule]", 
                    "0.003062834410342978`"}], ",", 
                    RowBox[{"\<\"lastInt\"\>", "\[Rule]", 
                    RowBox[{"-", "0.021311967103317137`"}]}], ",", 
                    RowBox[{"\<\"initialized\"\>", "\[Rule]", "True"}]}], 
                    "\[RightAssociation]"}]}
                    },
                    AutoDelete->False,
                    
                    GridBoxItemSize->{
                    "Columns" -> {{Automatic}}, "Rows" -> {{Automatic}}}],
                    "Grid"], {-1, -0.5}], 
                  {GrayLevel[0], LineBox[{{-5, 0}, {5, 0}}]}, 
                  {GrayLevel[0.5], 
                   CircleBox[{-1.8449992278296234`, 0.5}, 0.5], 
                   DiskBox[{-1.8449992278296234`, 0.5}, 0.1]}, 
                  {GrayLevel[0.5], 
                   DiskBox[{-1.8384008679936925`, 1.9999854871456175`}, 0.1]}, 
                  {GrayLevel[0.5], Thickness[0.01], 
                   LineBox[{{-1.8449992278296234`, 
                    0.5}, {-1.8384008679936925`, 1.9999854871456175`}}]}, 
                  {RGBColor[0, 1, 0], 
                   DiskBox[{-1.840600321272336, 1.4999903247637452`}, 0.05]}, 
                  {RGBColor[0, 0, 1], Arrowheads[0.01], 
                   CircleBox[{-1.8449992278296234`, 0.5}, 0.3333333333333333, 
                    NCache[{
                    Rational[-1, 2] 
                    Pi, -1.6356496180023106`}, {-1.5707963267948966`, \
-1.6356496180023106`}]], 
                   ArrowBox[{{-1.863274490324466, 
                    0.16716802153059918`}, {-1.866601840881931, 
                    0.1673674125699297}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-1.8449992278296234`, 0.5}, {-1.801763700358014,
                     0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-1.8449992278296234`, 
                    0.5}, {-1.8773528175404868`, 0.5}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.005], Arrowheads[0.04], 
                   ArrowBox[{{-1.8449992278296234`, 
                    0.5}, {-1.8341172900688774`, 0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-1.8449992278296234`, 0.5}, 1.6, {
                    1.5663974060507342`, 1.5461762168366617`}], 
                   ArrowBox[{{-1.8216078852245114`, 
                    2.099829004953695}, {-1.8056110313691927`, 
                    2.099515104642527}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Dashing[{Small, Small}], 
                   Arrowheads[0.02], 
                   CircleBox[{-1.8449992278296234`, 0.5}, 1.7, {
                    1.5663974060507342`, 1.572703885414117}], 
                   ArrowBox[{{-1.831242227635617, 
                    2.199944335837401}, {-1.8482420755156292`, 
                    2.199996907038035}}]}, 
                  {RGBColor[0, 0, 1], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-1.8449992278296234`, 0.5}, 1.8, {
                    1.5663974060507342`, 1.578407274792848}], 
                   ArrowBox[{{-1.8406989383166388`, 
                    2.2999948631899216`}, {-1.858698801963578, 
                    2.299947866375176}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.003], Arrowheads[0.025], 
                   CircleBox[{-1.8449992278296234`, 0.5}, 1.5, {
                    1.5663974060507342`, 1.5663873233247605`}], 
                   ArrowBox[{{-1.853385678931909, 
                    1.9999765556294236`}, {-1.8383857440513964`, 
                    1.9999854205399181`}}]}, 
                  {RGBColor[0, 0, 1], ArrowBox[{{0, 2.8}, {0, 2.8}}]}},
                 Frame->True,
                 FrameTicks->Automatic,
                 ImageSize->800,
                 PlotRange->{{-3, 3}, {-1, 3}}], 540->
                GraphicsBox[{InsetBox[
                   TagBox[GridBox[{
                    {"\<\"t\"\>", "26.95`"},
                    {"\<\"driveIntent\"\>", "0"},
                    {"\<\"motorTorque\"\>", 
                    RowBox[{"-", "2.1461537680806`"}]},
                    {"\<\"vw\"\>", 
                    RowBox[{"-", "0.02469301721043667`"}]},
                    {"\<\"controllerState\"\>", 
                    RowBox[{"\[LeftAssociation]", 
                    RowBox[{
                    
                    RowBox[{"\<\"lastError\"\>", "\[Rule]", 
                    "0.003041212128048449`"}], ",", 
                    RowBox[{"\<\"lastInt\"\>", "\[Rule]", 
                    RowBox[{"-", "0.021159377251278526`"}]}], ",", 
                    RowBox[{"\<\"initialized\"\>", "\[Rule]", "True"}]}], 
                    "\[RightAssociation]"}]}
                    },
                    AutoDelete->False,
                    
                    GridBoxItemSize->{
                    "Columns" -> {{Automatic}}, "Rows" -> {{Automatic}}}],
                    "Grid"], {-1, -0.5}], 
                  {GrayLevel[0], LineBox[{{-5, 0}, {5, 0}}]}, 
                  {GrayLevel[0.5], CircleBox[{-1.846238235650869, 0.5}, 0.5], 
                   DiskBox[{-1.846238235650869, 0.5}, 0.1]}, 
                  {GrayLevel[0.5], 
                   DiskBox[{-1.839687570090476, 1.9999856961920388`}, 0.1]}, 
                  {GrayLevel[0.5], Thickness[0.01], 
                   LineBox[{{-1.846238235650869, 0.5}, {-1.839687570090476, 
                    1.9999856961920388`}}]}, 
                  {RGBColor[0, 1, 0], 
                   DiskBox[{-1.8418711252772737`, 1.499990464128026}, 0.05]}, 
                  {RGBColor[0, 0, 1], Arrowheads[0.01], 
                   CircleBox[{-1.846238235650869, 0.5}, 0.3333333333333333, 
                    NCache[{
                    Rational[-1, 2] 
                    Pi, -1.6351809398373145`}, {-1.5707963267948966`, \
-1.6351809398373145`}]], 
                   ArrowBox[{{-1.8643575050633365`, 
                    0.16715949286922793`}, {-1.8676849487055829`, 
                    0.16735732443016293`}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-1.846238235650869, 0.5}, {-1.803315160289257, 
                    0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-1.846238235650869, 0.5}, {-1.8783579715671732`,
                     0.5}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.005], Arrowheads[0.04], 
                   ArrowBox[{{-1.846238235650869, 0.5}, {-1.835434896205561, 
                    0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-1.846238235650869, 0.5}, 1.6, {
                    1.5664292025398465`, 1.546354176159275}], 
                   ArrowBox[{{-1.8231315979006693`, 
                    2.09983314232825}, {-1.8071346884371704`, 
                    2.0995220888113133`}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Dashing[{Small, Small}], 
                   Arrowheads[0.02], 
                   CircleBox[{-1.846238235650869, 0.5}, 1.7, {
                    1.5664292025398465`, 1.5726901322079823`}], 
                   ArrowBox[{{-1.832457855773296, 
                    2.1999441464737686`}, {-1.8494577029286778`, 
                    2.19999695147681}}]}, 
                  {RGBColor[0, 0, 1], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-1.846238235650869, 0.5}, 1.8, {
                    1.5664292025398465`, 1.578352279029183}], 
                   ArrowBox[{{-1.8418389540523434`, 
                    2.299994623970143}, {-1.8598388202566878`, 
                    2.299948617071715}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.003], Arrowheads[0.025], 
                   CircleBox[{-1.846238235650869, 0.5}, 1.5, {
                    1.5664292025398465`, 1.5664192531449947`}], 
                   ArrowBox[{{-1.8546725807306481`, 
                    1.9999762870869244`}, {-1.839672646140837, 
                    1.9999856309426387`}}]}, 
                  {RGBColor[0, 0, 1], ArrowBox[{{0, 2.8}, {0, 2.8}}]}},
                 Frame->True,
                 FrameTicks->Automatic,
                 ImageSize->800,
                 PlotRange->{{-3, 3}, {-1, 3}}], 541->
                GraphicsBox[{InsetBox[
                   TagBox[GridBox[{
                    {"\<\"t\"\>", "27.`"},
                    {"\<\"driveIntent\"\>", "0"},
                    {"\<\"motorTorque\"\>", 
                    RowBox[{"-", "2.1306618488772533`"}]},
                    {"\<\"vw\"\>", 
                    RowBox[{"-", "0.024516247968527378`"}]},
                    {"\<\"controllerState\"\>", 
                    RowBox[{"\[LeftAssociation]", 
                    RowBox[{
                    
                    RowBox[{"\<\"lastError\"\>", "\[Rule]", 
                    "0.0030197101300203835`"}], ",", 
                    RowBox[{"\<\"lastInt\"\>", "\[Rule]", 
                    RowBox[{"-", "0.021007865447625496`"}]}], ",", 
                    RowBox[{"\<\"initialized\"\>", "\[Rule]", "True"}]}], 
                    "\[RightAssociation]"}]}
                    },
                    AutoDelete->False,
                    
                    GridBoxItemSize->{
                    "Columns" -> {{Automatic}}, "Rows" -> {{Automatic}}}],
                    "Grid"], {-1, -0.5}], 
                  {GrayLevel[0], LineBox[{{-5, 0}, {5, 0}}]}, 
                  {GrayLevel[0.5], 
                   CircleBox[{-1.8474683735649764`, 0.5}, 0.5], 
                   DiskBox[{-1.8474683735649764`, 0.5}, 0.1]}, 
                  {GrayLevel[0.5], 
                   DiskBox[{-1.8409650043332664`, 1.9999859019966275`}, 0.1]}, 
                  {GrayLevel[0.5], Thickness[0.01], 
                   LineBox[{{-1.8474683735649764`, 
                    0.5}, {-1.8409650043332664`, 1.9999859019966275`}}]}, 
                  {RGBColor[0, 1, 0], 
                   DiskBox[{-1.8431327940771698`, 1.4999906013310849`}, 
                    0.05]}, 
                  {RGBColor[0, 0, 1], Arrowheads[0.01], 
                   CircleBox[{-1.8474683735649764`, 0.5}, 0.3333333333333333, 
                    NCache[{
                    Rational[-1, 2] 
                    Pi, -1.6347161822612142`}, {-1.5707963267948966`, \
-1.6347161822612142`}]], 
                   ArrowBox[{{-1.8654329508788132`, 
                    0.16715110774850656`}, {-1.8687604861054097`, 
                    0.1673473928334897}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-1.8474683735649764`, 
                    0.5}, {-1.8048551365874312`, 0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-1.8474683735649764`, 
                    0.5}, {-1.8793562067970473`, 0.5}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.005], Arrowheads[0.04], 
                   ArrowBox[{{-1.8474683735649764`, 
                    0.5}, {-1.8367429698195024`, 0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-1.8474683735649764`, 0.5}, 1.6, {
                    1.5664607337241467`, 1.54653065063785}], 
                   ArrowBox[{{-1.8246440658927183`, 
                    2.099837195148082}, {-1.8086471017853103`, 
                    2.0995289646822313`}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Dashing[{Small, Small}], 
                   Arrowheads[0.02], 
                   CircleBox[{-1.8474683735649764`, 0.5}, 1.7, {
                    1.5664607337241467`, 1.5726764976803176`}], 
                   ArrowBox[{{-1.8336648157531914`, 
                    2.199943958426788}, {-1.850664662187022, 
                    2.1999969952147103`}}]}, 
                  {RGBColor[0, 0, 1], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-1.8474683735649764`, 0.5}, 1.8, {
                    1.5664607337241467`, 1.5782977439956871`}], 
                   ArrowBox[{{-1.8429709292059306`, 
                    2.299994381378519}, {-1.8609707978924963`, 
                    2.2999493561034656`}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.003], Arrowheads[0.025], 
                   CircleBox[{-1.8474683735649764`, 0.5}, 1.5, {
                    1.5664607337241467`, 1.5664509232416057`}], 
                   ArrowBox[{{-1.855950223034444, 
                    1.9999760192181666`}, {-1.8409502887480762`, 
                    1.9999858381232536`}}]}, 
                  {RGBColor[0, 0, 1], ArrowBox[{{0, 2.8}, {0, 2.8}}]}},
                 Frame->True,
                 FrameTicks->Automatic,
                 ImageSize->800,
                 PlotRange->{{-3, 3}, {-1, 3}}], 542->
                GraphicsBox[{InsetBox[
                   TagBox[GridBox[{
                    {"\<\"t\"\>", "27.05`"},
                    {"\<\"driveIntent\"\>", "0"},
                    {"\<\"motorTorque\"\>", 
                    RowBox[{"-", "2.1152987242943975`"}]},
                    {"\<\"vw\"\>", 
                    RowBox[{"-", "0.024340753251495616`"}]},
                    {"\<\"controllerState\"\>", 
                    RowBox[{"\[LeftAssociation]", 
                    RowBox[{
                    
                    RowBox[{"\<\"lastError\"\>", "\[Rule]", 
                    "0.002998328874814866`"}], ",", 
                    RowBox[{"\<\"lastInt\"\>", "\[Rule]", 
                    RowBox[{"-", "0.02085742566721127`"}]}], ",", 
                    RowBox[{"\<\"initialized\"\>", "\[Rule]", "True"}]}], 
                    "\[RightAssociation]"}]}
                    },
                    AutoDelete->False,
                    
                    GridBoxItemSize->{
                    "Columns" -> {{Automatic}}, "Rows" -> {{Automatic}}}],
                    "Grid"], {-1, -0.5}], 
                  {GrayLevel[0], LineBox[{{-5, 0}, {5, 0}}]}, 
                  {GrayLevel[0.5], 
                   CircleBox[{-1.8486897055619749`, 0.5}, 0.5], 
                   DiskBox[{-1.8486897055619749`, 0.5}, 0.1]}, 
                  {GrayLevel[0.5], 
                   DiskBox[{-1.8422332402635366`, 1.9999861046209229`}, 0.1]}, 
                  {GrayLevel[0.5], Thickness[0.01], 
                   LineBox[{{-1.8486897055619749`, 
                    0.5}, {-1.8422332402635366`, 1.9999861046209229`}}]}, 
                  {RGBColor[0, 1, 0], 
                   DiskBox[{-1.844385395363016, 1.4999907364139484`}, 0.05]}, 
                  {RGBColor[0, 0, 1], Arrowheads[0.01], 
                   CircleBox[{-1.8486897055619749`, 0.5}, 0.3333333333333333, 
                    NCache[{
                    Rational[-1, 2] 
                    Pi, -1.6342552885237285`}, {-1.5707963267948966`, \
-1.6342552885237285`}]], 
                   ArrowBox[{{-1.86650087300323, 
                    0.1671428633400049}, {-1.8698284983429672`, 
                    0.16733761476404735`}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-1.8486897055619749`, 
                    0.5}, {-1.8063837310760869`, 0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-1.8486897055619749`, 
                    0.5}, {-1.8803475600034645`, 0.5}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.005], Arrowheads[0.04], 
                   ArrowBox[{{-1.8486897055619749`, 
                    0.5}, {-1.8380415855175765`, 0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-1.8486897055619749`, 0.5}, 1.6, {
                    1.5664920033047725`, 1.5467056609863568`}], 
                   ArrowBox[{{-1.826145386302903, 
                    2.0998411651376974`}, {-1.8101483684969566`, 
                    2.099535734310816}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Dashing[{Small, Small}], 
                   Arrowheads[0.02], 
                   CircleBox[{-1.8486897055619749`, 0.5}, 1.7, {
                    1.5664920033047725`, 1.5726629793801283`}], 
                   ArrowBox[{{-1.8348631673987168`, 
                    2.1999437716708217`}, {-1.8518630131140268`, 
                    2.199997038267767}}]}, 
                  {RGBColor[0, 0, 1], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-1.8486897055619749`, 0.5}, 1.8, {
                    1.5664920033047725`, 1.5782436628841858`}], 
                   ArrowBox[{{-1.8440949155127124`, 
                    2.2999941355194475`}, {-1.8620947866079722`, 
                    2.2999500836973645`}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.003], Arrowheads[0.025], 
                   CircleBox[{-1.8486897055619749`, 0.5}, 1.5, {
                    1.5664920033047725`, 1.5664823358791615`}], 
                   ArrowBox[{{-1.857218673230284, 
                    1.9999757520408497`}, {-1.8422187392597544`, 
                    1.9999860421334312`}}]}, 
                  {RGBColor[0, 0, 1], ArrowBox[{{0, 2.8}, {0, 2.8}}]}},
                 Frame->True,
                 FrameTicks->Automatic,
                 ImageSize->800,
                 PlotRange->{{-3, 3}, {-1, 3}}], 543->
                GraphicsBox[{InsetBox[
                   TagBox[GridBox[{
                    {"\<\"t\"\>", "27.1`"},
                    {"\<\"driveIntent\"\>", "0"},
                    {"\<\"motorTorque\"\>", 
                    RowBox[{"-", "2.1000624542948696`"}]},
                    {"\<\"vw\"\>", 
                    RowBox[{"-", "0.02416652243858875`"}]},
                    {"\<\"controllerState\"\>", 
                    RowBox[{"\[LeftAssociation]", 
                    RowBox[{
                    
                    RowBox[{"\<\"lastError\"\>", "\[Rule]", 
                    "0.002977068985212766`"}], ",", 
                    RowBox[{"\<\"lastInt\"\>", "\[Rule]", 
                    RowBox[{"-", "0.020708051857623613`"}]}], ",", 
                    RowBox[{"\<\"initialized\"\>", "\[Rule]", "True"}]}], 
                    "\[RightAssociation]"}]}
                    },
                    AutoDelete->False,
                    
                    GridBoxItemSize->{
                    "Columns" -> {{Automatic}}, "Rows" -> {{Automatic}}}],
                    "Grid"], {-1, -0.5}], 
                  {GrayLevel[0], LineBox[{{-5, 0}, {5, 0}}]}, 
                  {GrayLevel[0.5], 
                   CircleBox[{-1.8499022950965447`, 0.5}, 0.5], 
                   DiskBox[{-1.8499022950965447`, 0.5}, 0.1]}, 
                  {GrayLevel[0.5], 
                   DiskBox[{-1.843492347053603, 1.9999863041261698`}, 0.1]}, 
                  {GrayLevel[0.5], Thickness[0.01], 
                   LineBox[{{-1.8499022950965447`, 0.5}, {-1.843492347053603, 
                    1.9999863041261698`}}]}, 
                  {RGBColor[0, 1, 0], 
                   DiskBox[{-1.8456289964012502`, 1.4999908694174466`}, 
                    0.05]}, 
                  {RGBColor[0, 0, 1], Arrowheads[0.01], 
                   CircleBox[{-1.8499022950965447`, 0.5}, 0.3333333333333333, 
                    NCache[{
                    Rational[-1, 2] 
                    Pi, -1.6337982004237426`}, {-1.5707963267948966`, \
-1.6337982004237426`}]], 
                   ArrowBox[{{-1.8675613156462958`, 
                    0.16713475683944623`}, {-1.8708890296569687`, 
                    0.16732798722525305`}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-1.8499022950965447`, 
                    0.5}, {-1.8079010460106473`, 0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-1.8499022950965447`, 
                    0.5}, {-1.8813320666094577`, 0.5}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.005], Arrowheads[0.04], 
                   ArrowBox[{{-1.8499022950965447`, 
                    0.5}, {-1.8393308175235603`, 0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-1.8499022950965447`, 0.5}, 1.6, {
                    1.5665230150936524`, 1.546879228538867}], 
                   ArrowBox[{{-1.8276356566910985`, 
                    2.0998450540018307`}, {-1.8116385861132318`, 
                    2.099542399742764}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Dashing[{Small, Small}], 
                   Arrowheads[0.02], 
                   CircleBox[{-1.8499022950965447`, 0.5}, 1.7, {
                    1.5665230150936524`, 1.5726495748789593`}], 
                   ArrowBox[{{-1.836052970036255, 
                    2.199943586180251}, {-1.8530528150360255`, 
                    2.199997080651644}}]}, 
                  {RGBColor[0, 0, 1], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-1.8499022950965447`, 0.5}, 1.8, {
                    1.5665230150936524`, 1.578190028728624}], 
                   ArrowBox[{{-1.8452109638884624`, 
                    2.2999938864928113`}, {-1.863210837320515, 
                    2.299950800078678}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.003], Arrowheads[0.025], 
                   CircleBox[{-1.8499022950965447`, 0.5}, 1.5, {
                    1.5665230150936524`, 1.5665134935538616`}], 
                   ArrowBox[{{-1.8584779985172464`, 
                    1.999975485569961}, {-1.8434780648746132`, 
                    1.9999862430256004`}}]}, 
                  {RGBColor[0, 0, 1], ArrowBox[{{0, 2.8}, {0, 2.8}}]}},
                 Frame->True,
                 FrameTicks->Automatic,
                 ImageSize->800,
                 PlotRange->{{-3, 3}, {-1, 3}}], 544->
                GraphicsBox[{InsetBox[
                   TagBox[GridBox[{
                    {"\<\"t\"\>", "27.150000000000002`"},
                    {"\<\"driveIntent\"\>", "0"},
                    {"\<\"motorTorque\"\>", 
                    RowBox[{"-", "2.0849510676691243`"}]},
                    {"\<\"vw\"\>", 
                    RowBox[{"-", "0.02399354508656392`"}]},
                    {"\<\"controllerState\"\>", 
                    RowBox[{"\[LeftAssociation]", 
                    RowBox[{
                    
                    RowBox[{"\<\"lastError\"\>", "\[Rule]", 
                    "0.0029559312136145906`"}], ",", 
                    RowBox[{"\<\"lastInt\"\>", "\[Rule]", 
                    RowBox[{"-", "0.020559737931867605`"}]}], ",", 
                    RowBox[{"\<\"initialized\"\>", "\[Rule]", "True"}]}], 
                    "\[RightAssociation]"}]}
                    },
                    AutoDelete->False,
                    
                    GridBoxItemSize->{
                    "Columns" -> {{Automatic}}, "Rows" -> {{Automatic}}}],
                    "Grid"], {-1, -0.5}], 
                  {GrayLevel[0], LineBox[{{-5, 0}, {5, 0}}]}, 
                  {GrayLevel[0.5], CircleBox[{-1.851106205096846, 0.5}, 0.5], 
                   DiskBox[{-1.851106205096846, 0.5}, 0.1]}, 
                  {GrayLevel[0.5], 
                   DiskBox[{-1.8447423934628224`, 1.9999865005730841`}, 0.1]}, 
                  {GrayLevel[0.5], Thickness[0.01], 
                   LineBox[{{-1.851106205096846, 0.5}, {-1.8447423934628224`, 
                    1.9999865005730841`}}]}, 
                  {RGBColor[0, 1, 0], 
                   DiskBox[{-1.8468636640074971`, 1.499991000382056}, 0.05]}, 
                  {RGBColor[0, 0, 1], Arrowheads[0.01], 
                   CircleBox[{-1.851106205096846, 0.5}, 0.3333333333333333, 
                    NCache[{
                    Rational[-1, 2] 
                    Pi, -1.6333448588249704`}, {-1.5707963267948966`, \
-1.6333448588249704`}]], 
                   ArrowBox[{{-1.8686143221756273`, 
                    0.16712678547611137`}, {-1.8719421234437157`, 
                    0.16731850725092373`}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-1.851106205096846, 0.5}, {-1.8094071837434635`,
                     0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-1.851106205096846, 0.5}, {-1.8823097609546362`,
                     0.5}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.005], Arrowheads[0.04], 
                   ArrowBox[{{-1.851106205096846, 0.5}, {-1.8406107396012537`,
                     0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-1.851106205096846, 0.5}, 1.6, {
                    1.5665537729784187`, 1.5470513750531696`}], 
                   ArrowBox[{{-1.8291149747694395`, 
                    2.099848863420757}, {-1.8131178523277294`, 
                    2.0995489630061}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Dashing[{Small, Small}], 
                   Arrowheads[0.02], 
                   CircleBox[{-1.851106205096846, 0.5}, 1.7, {
                    1.5665537729784187`, 1.572636281785477}], 
                   ArrowBox[{{-1.837234282528775, 
                    2.1999434019296835`}, {-1.8542341268159366`, 
                    2.1999971223815997`}}]}, 
                  {RGBColor[0, 0, 1], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-1.851106205096846, 0.5}, 1.8, {
                    1.5665537729784187`, 1.5781368344654694`}], 
                   ArrowBox[{{-1.8463191245469717`, 
                    2.2999936343942466`}, {-1.864319000245508, 
                    2.299951505470178}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.003], Arrowheads[0.025], 
                   CircleBox[{-1.851106205096846, 0.5}, 1.5, {
                    1.5665537729784187`, 1.5665443989572798`}], 
                   ArrowBox[{{-1.859728265860944, 
                    1.999975219818041}, {-1.8447283325579378`, 
                    1.9999864408526755`}}]}, 
                  {RGBColor[0, 0, 1], ArrowBox[{{0, 2.8}, {0, 2.8}}]}},
                 Frame->True,
                 FrameTicks->Automatic,
                 ImageSize->800,
                 PlotRange->{{-3, 3}, {-1, 3}}], 545->
                GraphicsBox[{InsetBox[
                   TagBox[GridBox[{
                    {"\<\"t\"\>", "27.2`"},
                    {"\<\"driveIntent\"\>", "0"},
                    {"\<\"motorTorque\"\>", 
                    RowBox[{"-", "2.069962577603663`"}]},
                    {"\<\"vw\"\>", 
                    RowBox[{"-", "0.023821810930150344`"}]},
                    {"\<\"controllerState\"\>", 
                    RowBox[{"\[LeftAssociation]", 
                    RowBox[{
                    
                    RowBox[{"\<\"lastError\"\>", "\[Rule]", 
                    "0.002934916410223639`"}], ",", 
                    RowBox[{"\<\"lastInt\"\>", "\[Rule]", 
                    RowBox[{"-", "0.02041247776270763`"}]}], ",", 
                    RowBox[{"\<\"initialized\"\>", "\[Rule]", "True"}]}], 
                    "\[RightAssociation]"}]}
                    },
                    AutoDelete->False,
                    
                    GridBoxItemSize->{
                    "Columns" -> {{Automatic}}, "Rows" -> {{Automatic}}}],
                    "Grid"], {-1, -0.5}], 
                  {GrayLevel[0], LineBox[{{-5, 0}, {5, 0}}]}, 
                  {GrayLevel[0.5], 
                   CircleBox[{-1.8523014979734187`, 0.5}, 0.5], 
                   DiskBox[{-1.8523014979734187`, 0.5}, 0.1]}, 
                  {GrayLevel[0.5], 
                   DiskBox[{-1.8459834477986417`, 1.9999866940216466`}, 0.1]}, 
                  {GrayLevel[0.5], Thickness[0.01], 
                   LineBox[{{-1.8523014979734187`, 
                    0.5}, {-1.8459834477986417`, 1.9999866940216466`}}]}, 
                  {RGBColor[0, 1, 0], 
                   DiskBox[{-1.8480894645235673`, 1.4999911293477646`}, 
                    0.05]}, 
                  {RGBColor[0, 0, 1], Arrowheads[0.01], 
                   CircleBox[{-1.8523014979734187`, 0.5}, 0.3333333333333333, 
                    NCache[{
                    Rational[-1, 2] 
                    Pi, -1.6328952041230065`}, {-1.5707963267948966`, \
-1.6328952041230065`}]], 
                   ArrowBox[{{-1.869659935281198, 
                    0.1671189465209187}, {-1.8729878224214578`, 
                    0.167309171914912}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-1.8523014979734187`, 
                    0.5}, {-1.8109022464213453`, 0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-1.8523014979734187`, 0.5}, {-1.8832806765387, 
                    0.5}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.005], Arrowheads[0.04], 
                   ArrowBox[{{-1.8523014979734187`, 
                    0.5}, {-1.8418814249866267`, 0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-1.8523014979734187`, 0.5}, 1.6, {
                    1.5665842808905064`, 1.5472221225322322`}], 
                   ArrowBox[{{-1.830583438125569, 
                    2.09985259504632}, {-1.814586264709815, 
                    2.0995554261043514`}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Dashing[{Small, Small}], 
                   Arrowheads[0.02], 
                   CircleBox[{-1.8523014979734187`, 0.5}, 1.7, {
                    1.5665842808905064`, 1.572623097758049}], 
                   ArrowBox[{{-1.8384071633061168`, 
                    2.199943218894135}, {-1.855407006883549, 
                    2.1999971634724598`}}]}, 
                  {RGBColor[0, 0, 1], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-1.8523014979734187`, 0.5}, 1.8, {
                    1.5665842808905064`, 1.5780840729883046`}], 
                   ArrowBox[{{-1.8474194471072134`, 
                    2.2999933793154184`}, {-1.8654193250034803`, 
                    2.2999522000914383`}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.003], Arrowheads[0.025], 
                   CircleBox[{-1.8523014979734187`, 0.5}, 1.5, {
                    1.5665842808905064`, 1.5665750549422535`}], 
                   ArrowBox[{{-1.8609695419512575`, 
                    1.9999749547954453`}, {-1.8459696089992916`, 
                    1.9999866356678047`}}]}, 
                  {RGBColor[0, 0, 1], ArrowBox[{{0, 2.8}, {0, 2.8}}]}},
                 Frame->True,
                 FrameTicks->Automatic,
                 ImageSize->800,
                 PlotRange->{{-3, 3}, {-1, 3}}], 546->
                GraphicsBox[{InsetBox[
                   TagBox[GridBox[{
                    {"\<\"t\"\>", "27.25`"},
                    {"\<\"driveIntent\"\>", "0"},
                    {"\<\"motorTorque\"\>", 
                    RowBox[{"-", "2.055094995667086`"}]},
                    {"\<\"vw\"\>", 
                    RowBox[{"-", "0.02365130988132119`"}]},
                    {"\<\"controllerState\"\>", 
                    RowBox[{"\[LeftAssociation]", 
                    RowBox[{
                    
                    RowBox[{"\<\"lastError\"\>", "\[Rule]", 
                    "0.0029140254940029026`"}], ",", 
                    RowBox[{"\<\"lastInt\"\>", "\[Rule]", 
                    RowBox[{"-", "0.02026626517852932`"}]}], ",", 
                    RowBox[{"\<\"initialized\"\>", "\[Rule]", "True"}]}], 
                    "\[RightAssociation]"}]}
                    },
                    AutoDelete->False,
                    
                    GridBoxItemSize->{
                    "Columns" -> {{Automatic}}, "Rows" -> {{Automatic}}}],
                    "Grid"], {-1, -0.5}], 
                  {GrayLevel[0], LineBox[{{-5, 0}, {5, 0}}]}, 
                  {GrayLevel[0.5], 
                   CircleBox[{-1.8534882356280642`, 0.5}, 0.5], 
                   DiskBox[{-1.8534882356280642`, 0.5}, 0.1]}, 
                  {GrayLevel[0.5], 
                   DiskBox[{-1.847215577882326, 1.9999868845309297`}, 0.1]}, 
                  {GrayLevel[0.5], Thickness[0.01], 
                   LineBox[{{-1.8534882356280642`, 0.5}, {-1.847215577882326, 
                    1.9999868845309297`}}]}, 
                  {RGBColor[0, 1, 0], 
                   DiskBox[{-1.8493064637975722`, 1.4999912563539533`}, 
                    0.05]}, 
                  {RGBColor[0, 0, 1], Arrowheads[0.01], 
                   CircleBox[{-1.8534882356280642`, 0.5}, 0.3333333333333333, 
                    NCache[{
                    Rational[-1, 2] 
                    Pi, -1.6324491766649092`}, {-1.5707963267948966`, \
-1.6324491766649092`}]], 
                   ArrowBox[{{-1.8706981971239862`, 
                    0.16711123729326643`}, {-1.8740261687789666`, 
                    0.16729997833934518`}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-1.8534882356280642`, 
                    0.5}, {-1.8123863357147225`, 0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-1.8534882356280642`, 
                    0.5}, {-1.8842448462419172`, 0.5}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.005], Arrowheads[0.04], 
                   ArrowBox[{{-1.8534882356280642`, 
                    0.5}, {-1.8431429463285756`, 0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-1.8534882356280642`, 0.5}, 1.6, {
                    1.5666145427763842`, 1.5473914930631836`}], 
                   ArrowBox[{{-1.8320411439738922`, 
                    2.099856250498643}, {-1.8160439204559273`, 
                    2.0995617910106787`}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Dashing[{Small, Small}], 
                   Arrowheads[0.02], 
                   CircleBox[{-1.8534882356280642`, 0.5}, 1.7, {
                    1.5666145427763842`, 1.5726100205154565`}], 
                   ArrowBox[{{-1.839571670392084, 
                    2.1999430370491924`}, {-1.8565715132626162`, 
                    2.199997203938591}}]}, 
                  {RGBColor[0, 0, 1], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-1.8534882356280642`, 0.5}, 1.8, {
                    1.5666145427763842`, 1.578031737196757}], 
                   ArrowBox[{{-1.848511980690301, 
                    2.2999931213442997`}, {-1.8665118607170639`, 
                    2.299952884158233}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.003], Arrowheads[0.025], 
                   CircleBox[{-1.8534882356280642`, 0.5}, 1.5, {
                    1.5666145427763842`, 1.5666054644910323`}], 
                   ArrowBox[{{-1.862201893163443, 
                    1.9999746905105953`}, {-1.8472019605736227`, 
                    1.999986827524142}}]}, 
                  {RGBColor[0, 0, 1], ArrowBox[{{0, 2.8}, {0, 2.8}}]}},
                 Frame->True,
                 FrameTicks->Automatic,
                 ImageSize->800,
                 PlotRange->{{-3, 3}, {-1, 3}}], 547->
                GraphicsBox[{InsetBox[
                   TagBox[GridBox[{
                    {"\<\"t\"\>", "27.3`"},
                    {"\<\"driveIntent\"\>", "0"},
                    {"\<\"motorTorque\"\>", 
                    RowBox[{"-", "2.040346344260402`"}]},
                    {"\<\"vw\"\>", 
                    RowBox[{"-", "0.02348203202751405`"}]},
                    {"\<\"controllerState\"\>", 
                    RowBox[{"\[LeftAssociation]", 
                    RowBox[{
                    
                    RowBox[{"\<\"lastError\"\>", "\[Rule]", 
                    "0.0028932594263661312`"}], ",", 
                    RowBox[{"\<\"lastInt\"\>", "\[Rule]", 
                    RowBox[{"-", "0.020121093960583657`"}]}], ",", 
                    RowBox[{"\<\"initialized\"\>", "\[Rule]", "True"}]}], 
                    "\[RightAssociation]"}]}
                    },
                    AutoDelete->False,
                    
                    GridBoxItemSize->{
                    "Columns" -> {{Automatic}}, "Rows" -> {{Automatic}}}],
                    "Grid"], {-1, -0.5}], 
                  {GrayLevel[0], LineBox[{{-5, 0}, {5, 0}}]}, 
                  {GrayLevel[0.5], 
                   CircleBox[{-1.8546664794626653`, 0.5}, 0.5], 
                   DiskBox[{-1.8546664794626653`, 0.5}, 0.1]}, 
                  {GrayLevel[0.5], 
                   DiskBox[{-1.8484388510191978`, 1.999987072158947}, 0.1]}, 
                  {GrayLevel[0.5], Thickness[0.01], 
                   LineBox[{{-1.8546664794626653`, 
                    0.5}, {-1.8484388510191978`, 1.999987072158947}}]}, 
                  {RGBColor[0, 1, 0], 
                   DiskBox[{-1.8505147271670204`, 1.499991381439298}, 0.05]}, 
                  {RGBColor[0, 0, 1], Arrowheads[0.01], 
                   CircleBox[{-1.8546664794626653`, 0.5}, 0.3333333333333333, 
                    NCache[{
                    Rational[-1, 2] 
                    Pi, -1.6320067171227086`}, {-1.5707963267948966`, \
-1.6320067171227086`}]], 
                   ArrowBox[{{-1.8717291494692414`, 
                    0.16710365516672043`}, {-1.8750572043087366`, 
                    0.1672909237015574}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-1.8546664794626653`, 
                    0.5}, {-1.8138595525774572`, 0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-1.8546664794626653`, 
                    0.5}, {-1.8852023025231175`, 0.5}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.005], Arrowheads[0.04], 
                   ArrowBox[{{-1.8546664794626653`, 
                    0.5}, {-1.8443953756379095`, 0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-1.8546664794626653`, 0.5}, 1.6, {
                    1.5666445625718342`, 1.547559508673356}], 
                   ArrowBox[{{-1.833488188934062, 
                    2.0998598313634496`}, {-1.8174909161681017`, 
                    2.099568059662899}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Dashing[{Small, Small}], 
                   Arrowheads[0.02], 
                   CircleBox[{-1.8546664794626653`, 0.5}, 1.7, {
                    1.5666445625718342`, 1.5725970478458482`}], 
                   ArrowBox[{{-1.8407278614284839`, 
                    2.199942856371148}, {-1.8577277035948967`, 
                    2.199997243793887}}]}, 
                  {RGBColor[0, 0, 1], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-1.8546664794626653`, 0.5}, 1.8, {
                    1.5666445625718342`, 1.5779798200399475`}], 
                   ArrowBox[{{-1.8495967740065138`, 
                    2.299992860565449}, {-1.8675966560980186`, 
                    2.29995355788203}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.003], Arrowheads[0.025], 
                   CircleBox[{-1.8546664794626653`, 0.5}, 1.5, {
                    1.5666445625718342`, 1.5666356306857427`}], 
                   ArrowBox[{{-1.863425385522647, 
                    1.9999744269702175`}, {-1.8484254533057793`, 
                    1.9999870164746456`}}]}, 
                  {RGBColor[0, 0, 1], ArrowBox[{{0, 2.8}, {0, 2.8}}]}},
                 Frame->True,
                 FrameTicks->Automatic,
                 ImageSize->800,
                 PlotRange->{{-3, 3}, {-1, 3}}], 548->
                GraphicsBox[{InsetBox[
                   TagBox[GridBox[{
                    {"\<\"t\"\>", "27.35`"},
                    {"\<\"driveIntent\"\>", "0"},
                    {"\<\"motorTorque\"\>", 
                    RowBox[{"-", "2.0257146675874287`"}]},
                    {"\<\"vw\"\>", 
                    RowBox[{"-", "0.02331396762893303`"}]},
                    {"\<\"controllerState\"\>", 
                    RowBox[{"\[LeftAssociation]", 
                    RowBox[{
                    
                    RowBox[{"\<\"lastError\"\>", "\[Rule]", 
                    "0.002872619187541719`"}], ",", 
                    RowBox[{"\<\"lastInt\"\>", "\[Rule]", 
                    RowBox[{"-", "0.019976957841478224`"}]}], ",", 
                    RowBox[{"\<\"initialized\"\>", "\[Rule]", "True"}]}], 
                    "\[RightAssociation]"}]}
                    },
                    AutoDelete->False,
                    
                    GridBoxItemSize->{
                    "Columns" -> {{Automatic}}, "Rows" -> {{Automatic}}}],
                    "Grid"], {-1, -0.5}], 
                  {GrayLevel[0], LineBox[{{-5, 0}, {5, 0}}]}, 
                  {GrayLevel[0.5], 
                   CircleBox[{-1.8558362903878949`, 0.5}, 0.5], 
                   DiskBox[{-1.8558362903878949`, 0.5}, 0.1]}, 
                  {GrayLevel[0.5], 
                   DiskBox[{-1.8496533339731838`, 1.9999872569625297`}, 0.1]}, 
                  {GrayLevel[0.5], Thickness[0.01], 
                   LineBox[{{-1.8558362903878949`, 
                    0.5}, {-1.8496533339731838`, 1.9999872569625297`}}]}, 
                  {RGBColor[0, 1, 0], 
                   DiskBox[{-1.851714319444754, 1.4999915046416863`}, 0.05]}, 
                  {RGBColor[0, 0, 1], Arrowheads[0.01], 
                   CircleBox[{-1.8558362903878949`, 0.5}, 0.3333333333333333, 
                    NCache[{
                    Rational[-1, 2] 
                    Pi, -1.6315677668225195`}, {-1.5707963267948966`, \
-1.6315677668225195`}]], 
                   ArrowBox[{{-1.872752833804872, 
                    0.16709619757363448`}, {-1.8760809705253236`, 
                    0.1672820052398063}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-1.8558362903878949`, 
                    0.5}, {-1.8153219970361463`, 0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-1.8558362903878949`, 
                    0.5}, {-1.8861530775959001`, 0.5}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.005], Arrowheads[0.04], 
                   ArrowBox[{{-1.8558362903878949`, 
                    0.5}, {-1.8456387842441515`, 0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-1.8558362903878949`, 0.5}, 1.6, {
                    1.5666743441791762`, 1.547726191202824}], 
                   ArrowBox[{{-1.8349246688357415`, 
                    2.0998633391899633`}, {-1.8189273476587628`, 
                    2.0995742339593426`}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Dashing[{Small, Small}], 
                   Arrowheads[0.02], 
                   CircleBox[{-1.8558362903878949`, 0.5}, 1.7, {
                    1.5666743441791762`, 1.572584177614049}], 
                   ArrowBox[{{-1.8418757936962618`, 
                    2.199942676837111}, {-1.8588756351612876`, 
                    2.199997283051755}}]}, 
                  {RGBColor[0, 0, 1], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-1.8558362903878949`, 0.5}, 1.8, {
                    1.5666743441791762`, 1.5779283145546619`}], 
                   ArrowBox[{{-1.8506738754327148`, 
                    2.2999925970602852`}, {-1.8686737595246483`, 
                    2.2999542214695805`}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.003], Arrowheads[0.025], 
                   CircleBox[{-1.8558362903878949`, 0.5}, 1.5, {
                    1.5666743441791762`, 1.5666655566811725`}], 
                   ArrowBox[{{-1.864640084671794, 
                    1.9999741641795725`}, {-1.8496401528383966`, 
                    1.9999872025718979`}}]}, 
                  {RGBColor[0, 0, 1], ArrowBox[{{0, 2.8}, {0, 2.8}}]}},
                 Frame->True,
                 FrameTicks->Automatic,
                 ImageSize->800,
                 PlotRange->{{-3, 3}, {-1, 3}}], 549->
                GraphicsBox[{InsetBox[
                   TagBox[GridBox[{
                    {"\<\"t\"\>", "27.400000000000002`"},
                    {"\<\"driveIntent\"\>", "0"},
                    {"\<\"motorTorque\"\>", 
                    RowBox[{"-", "2.0111980412080066`"}]},
                    {"\<\"vw\"\>", 
                    RowBox[{"-", "0.023147107115059958`"}]},
                    {"\<\"controllerState\"\>", 
                    RowBox[{"\[LeftAssociation]", 
                    RowBox[{
                    
                    RowBox[{"\<\"lastError\"\>", "\[Rule]", 
                    "0.002852105755529824`"}], ",", 
                    RowBox[{"\<\"lastInt\"\>", "\[Rule]", 
                    RowBox[{"-", "0.01983385050478363`"}]}], ",", 
                    RowBox[{"\<\"initialized\"\>", "\[Rule]", "True"}]}], 
                    "\[RightAssociation]"}]}
                    },
                    AutoDelete->False,
                    
                    GridBoxItemSize->{
                    "Columns" -> {{Automatic}}, "Rows" -> {{Automatic}}}],
                    "Grid"], {-1, -0.5}], 
                  {GrayLevel[0], LineBox[{{-5, 0}, {5, 0}}]}, 
                  {GrayLevel[0.5], 
                   CircleBox[{-1.8569977288317685`, 0.5}, 0.5], 
                   DiskBox[{-1.8569977288317685`, 0.5}, 0.1]}, 
                  {GrayLevel[0.5], 
                   DiskBox[{-1.8508590929454498`, 1.9999874389972254`}, 0.1]}, 
                  {GrayLevel[0.5], Thickness[0.01], 
                   LineBox[{{-1.8569977288317685`, 
                    0.5}, {-1.8508590929454498`, 1.9999874389972254`}}]}, 
                  {RGBColor[0, 1, 0], 
                   DiskBox[{-1.8529053049075561`, 1.4999916259981503`}, 
                    0.05]}, 
                  {RGBColor[0, 0, 1], Arrowheads[0.01], 
                   CircleBox[{-1.8569977288317685`, 0.5}, 0.3333333333333333, 
                    NCache[{
                    Rational[-1, 2] 
                    Pi, -1.6311322680311369`}, {-1.5707963267948966`, \
-1.6311322680311369`}]], 
                   ArrowBox[{{-1.873769291445536, 
                    0.16708886200878836`}, {-1.8770975087693933`, 
                    0.16727322025786662`}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-1.8569977288317685`, 
                    0.5}, {-1.8167737680076084`, 0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-1.8569977288317685`, 0.5}, {-1.887097203583884,
                     0.5}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.005], Arrowheads[0.04], 
                   ArrowBox[{{-1.8569977288317685`, 
                    0.5}, {-1.8468732427597239`, 0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-1.8569977288317685`, 0.5}, 1.6, {
                    1.5667038914473241`, 1.547891562192772}], 
                   ArrowBox[{{-1.8363506785485335`, 
                    2.099866775489322}, {-1.820353309780681, 
                    2.0995803157554827`}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Dashing[{Small, Small}], 
                   Arrowheads[0.02], 
                   CircleBox[{-1.8569977288317685`, 0.5}, 1.7, {
                    1.5667038914473241`, 1.5725714077673274`}], 
                   ArrowBox[{{-1.8430155241338555`, 
                    2.199942498425104}, {-1.8600153649001825`, 
                    2.199997321725113}}]}, 
                  {RGBColor[0, 0, 1], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-1.8569977288317685`, 0.5}, 1.8, {
                    1.5667038914473241`, 1.5778772138984796`}], 
                   ArrowBox[{{-1.8517433330805357`, 
                    2.299992330907354}, {-1.8697432191099856`, 
                    2.2999548751225873`}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.003], Arrowheads[0.025], 
                   CircleBox[{-1.8569977288317685`, 0.5}, 1.5, {
                    1.5667038914473241`, 1.5666952456798588`}], 
                   ArrowBox[{{-1.8658460558427712`, 
                    1.9999739021426695`}, {-1.8508461244030812`, 
                    1.9999873858679456`}}]}, 
                  {RGBColor[0, 0, 1], ArrowBox[{{0, 2.8}, {0, 2.8}}]}},
                 Frame->True,
                 FrameTicks->Automatic,
                 ImageSize->800,
                 PlotRange->{{-3, 3}, {-1, 3}}], 550->
                GraphicsBox[{InsetBox[
                   TagBox[GridBox[{
                    {"\<\"t\"\>", "27.45`"},
                    {"\<\"driveIntent\"\>", "0"},
                    {"\<\"motorTorque\"\>", 
                    RowBox[{"-", "1.9967945802468856`"}]},
                    {"\<\"vw\"\>", 
                    RowBox[{"-", "0.022981441080494103`"}]},
                    {"\<\"controllerState\"\>", 
                    RowBox[{"\[LeftAssociation]", 
                    RowBox[{
                    
                    RowBox[{"\<\"lastError\"\>", "\[Rule]", 
                    "0.002831720087556916`"}], ",", 
                    RowBox[{"\<\"lastInt\"\>", "\[Rule]", 
                    RowBox[{"-", "0.019691765585628046`"}]}], ",", 
                    RowBox[{"\<\"initialized\"\>", "\[Rule]", "True"}]}], 
                    "\[RightAssociation]"}]}
                    },
                    AutoDelete->False,
                    
                    GridBoxItemSize->{
                    "Columns" -> {{Automatic}}, "Rows" -> {{Automatic}}}],
                    "Grid"], {-1, -0.5}], 
                  {GrayLevel[0], LineBox[{{-5, 0}, {5, 0}}]}, 
                  {GrayLevel[0.5], CircleBox[{-1.858150854748006, 0.5}, 0.5], 
                   DiskBox[{-1.858150854748006, 0.5}, 0.1]}, 
                  {GrayLevel[0.5], 
                   DiskBox[{-1.8520561935568904`, 1.9999876183172198`}, 0.1]}, 
                  {GrayLevel[0.5], Thickness[0.01], 
                   LineBox[{{-1.858150854748006, 0.5}, {-1.8520561935568904`, 
                    1.9999876183172198`}}]}, 
                  {RGBColor[0, 1, 0], 
                   DiskBox[{-1.8540877472872623`, 1.499991745544813}, 0.05]}, 
                  {RGBColor[0, 0, 1], Arrowheads[0.01], 
                   CircleBox[{-1.858150854748006, 0.5}, 0.3333333333333333, 
                    NCache[{
                    Rational[-1, 2] 
                    Pi, -1.6307001642023031`}, {-1.5707963267948966`, \
-1.6307001642023031`}]], 
                   ArrowBox[{{-1.874778563623122, 
                    0.16708164603213077`}, {-1.8781068602981703`, 
                    0.16726456612859386`}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-1.858150854748006, 0.5}, {-1.8182149631430682`,
                     0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-1.858150854748006, 0.5}, {-1.8880347126559296`,
                     0.5}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.005], Arrowheads[0.04], 
                   ArrowBox[{{-1.858150854748006, 0.5}, {-1.8480988210509919`,
                     0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-1.858150854748006, 0.5}, 1.6, {
                    1.5667332081545366`, 1.548055642788958}], 
                   ArrowBox[{{-1.837766311835868, 
                    2.099870141733467}, {-1.8217688962808758`, 
                    2.0995863068612763`}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Dashing[{Small, Small}], 
                   Arrowheads[0.02], 
                   CircleBox[{-1.858150854748006, 0.5}, 1.7, {
                    1.5667332081545366`, 1.5725587363397733`}], 
                   ArrowBox[{{-1.844147109353001, 
                    2.1999423211141345`}, {-1.8611469494232737`, 
                    2.1999973598263898`}}]}, 
                  {RGBColor[0, 0, 1], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-1.858150854748006, 0.5}, 1.8, {
                    1.5667332081545366`, 1.5778265113781305`}], 
                   ArrowBox[{{-1.85280519485578, 
                    2.299992062182586}, {-1.870805082761196, 
                    2.2999555190374537`}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.003], Arrowheads[0.025], 
                   CircleBox[{-1.858150854748006, 0.5}, 1.5, {
                    1.5667332081545366`, 1.5667247009095002`}], 
                   ArrowBox[{{-1.8670433638309074`, 
                    1.9999736408624689`}, {-1.8520434327948907`, 
                    1.9999875664141642`}}]}, 
                  {RGBColor[0, 0, 1], ArrowBox[{{0, 2.8}, {0, 2.8}}]}},
                 Frame->True,
                 FrameTicks->Automatic,
                 ImageSize->800,
                 PlotRange->{{-3, 3}, {-1, 3}}], 551->
                GraphicsBox[{InsetBox[
                   TagBox[GridBox[{
                    {"\<\"t\"\>", "27.5`"},
                    {"\<\"driveIntent\"\>", "0"},
                    {"\<\"motorTorque\"\>", 
                    RowBox[{"-", "1.9825024463304846`"}]},
                    {"\<\"vw\"\>", 
                    RowBox[{"-", "0.02281696028023373`"}]},
                    {"\<\"controllerState\"\>", 
                    RowBox[{"\[LeftAssociation]", 
                    RowBox[{
                    
                    RowBox[{"\<\"lastError\"\>", "\[Rule]", 
                    "0.0028114631039194166`"}], ",", 
                    RowBox[{"\<\"lastInt\"\>", "\[Rule]", 
                    RowBox[{"-", "0.019550696672157516`"}]}], ",", 
                    RowBox[{"\<\"initialized\"\>", "\[Rule]", "True"}]}], 
                    "\[RightAssociation]"}]}
                    },
                    AutoDelete->False,
                    
                    GridBoxItemSize->{
                    "Columns" -> {{Automatic}}, "Rows" -> {{Automatic}}}],
                    "Grid"], {-1, -0.5}], 
                  {GrayLevel[0], LineBox[{{-5, 0}, {5, 0}}]}, 
                  {GrayLevel[0.5], 
                   CircleBox[{-1.8592957276241695`, 0.5}, 0.5], 
                   DiskBox[{-1.8592957276241695`, 0.5}, 0.1]}, 
                  {GrayLevel[0.5], 
                   DiskBox[{-1.8532447008342252`, 1.999987794975275}, 0.1]}, 
                  {GrayLevel[0.5], Thickness[0.01], 
                   LineBox[{{-1.8592957276241695`, 
                    0.5}, {-1.8532447008342252`, 1.999987794975275}}]}, 
                  {RGBColor[0, 1, 0], 
                   DiskBox[{-1.8552617097642066`, 1.49999186331685}, 0.05]}, 
                  {RGBColor[0, 0, 1], Arrowheads[0.01], 
                   CircleBox[{-1.8592957276241695`, 0.5}, 0.3333333333333333, 
                    NCache[{
                    Rational[-1, 2] 
                    Pi, -1.630271400184811}, {-1.5707963267948966`, \
-1.630271400184811}]], 
                   ArrowBox[{{-1.875780691564308, 
                    0.16707454727070892`}, {-1.8791090663629746`, 
                    0.16725604029654806`}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-1.8592957276241695`, 
                    0.5}, {-1.8196456786975599`, 0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-1.8592957276241695`, 
                    0.5}, {-1.8889656371423558`, 0.5}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.005], Arrowheads[0.04], 
                   ArrowBox[{{-1.8592957276241695`, 
                    0.5}, {-1.8493155882157462`, 0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-1.8592957276241695`, 0.5}, 1.6, {
                    1.566762297993723, 1.548218453659471}], 
                   ArrowBox[{{-1.8391716612315345`, 
                    2.0998734393544463`}, {-1.8231741996771647`, 
                    2.099592209039158}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Dashing[{Small, Small}], 
                   Arrowheads[0.02], 
                   CircleBox[{-1.8592957276241695`, 0.5}, 1.7, {
                    1.566762297993723, 1.5725461614553722`}], 
                   ArrowBox[{{-1.8452706056520958`, 
                    2.1999421448842513`}, {-1.862270445028919, 
                    2.1999973973675258`}}]}, 
                  {RGBColor[0, 0, 1], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-1.8592957276241695`, 0.5}, 1.8, {
                    1.566762297993723, 1.5777762004733369`}], 
                   ArrowBox[{{-1.8538595085094776`, 
                    2.2999917909595418`}, {-1.8718593982306317`, 
                    2.2999561534051023`}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.003], Arrowheads[0.025], 
                   CircleBox[{-1.8592957276241695`, 0.5}, 1.5, {
                    1.566762297993723, 1.5667539256025929`}], 
                   ArrowBox[{{-1.8682320729725685`, 
                    1.9999733803410693`}, {-1.8532321423499274`, 
                    1.9999877442611396`}}]}, 
                  {RGBColor[0, 0, 1], ArrowBox[{{0, 2.8}, {0, 2.8}}]}},
                 Frame->True,
                 FrameTicks->Automatic,
                 ImageSize->800,
                 PlotRange->{{-3, 3}, {-1, 3}}], 552->
                GraphicsBox[{InsetBox[
                   TagBox[GridBox[{
                    {"\<\"t\"\>", "27.55`"},
                    {"\<\"driveIntent\"\>", "0"},
                    {"\<\"motorTorque\"\>", 
                    RowBox[{"-", "1.968319853331662`"}]},
                    {"\<\"vw\"\>", 
                    RowBox[{"-", "0.02265365562450406`"}]},
                    {"\<\"controllerState\"\>", 
                    RowBox[{"\[LeftAssociation]", 
                    RowBox[{
                    
                    RowBox[{"\<\"lastError\"\>", "\[Rule]", 
                    "0.00279133567409723`"}], ",", 
                    RowBox[{"\<\"lastInt\"\>", "\[Rule]", 
                    RowBox[{"-", "0.019410637307745858`"}]}], ",", 
                    RowBox[{"\<\"initialized\"\>", "\[Rule]", "True"}]}], 
                    "\[RightAssociation]"}]}
                    },
                    AutoDelete->False,
                    
                    GridBoxItemSize->{
                    "Columns" -> {{Automatic}}, "Rows" -> {{Automatic}}}],
                    "Grid"], {-1, -0.5}], 
                  {GrayLevel[0], LineBox[{{-5, 0}, {5, 0}}]}, 
                  {GrayLevel[0.5], CircleBox[{-1.860432406489548, 0.5}, 0.5], 
                   DiskBox[{-1.860432406489548, 0.5}, 0.1]}, 
                  {GrayLevel[0.5], 
                   DiskBox[{-1.8544246791994463`, 1.9999879690226878`}, 0.1]}, 
                  {GrayLevel[0.5], Thickness[0.01], 
                   LineBox[{{-1.860432406489548, 0.5}, {-1.8544246791994463`, 
                    1.9999879690226878`}}]}, 
                  {RGBColor[0, 1, 0], 
                   DiskBox[{-1.8564272549628136`, 1.4999919793484584`}, 
                    0.05]}, 
                  {RGBColor[0, 0, 1], Arrowheads[0.01], 
                   CircleBox[{-1.860432406489548, 0.5}, 0.3333333333333333, 
                    NCache[{
                    Rational[-1, 2] 
                    Pi, -1.6298459223948465`}, {-1.5707963267948966`, \
-1.6298459223948465`}]], 
                   ArrowBox[{{-1.8767757165559626`, 
                    0.16706756341986723`}, {-1.8801041682746085`, 
                    0.16724764027976757`}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-1.860432406489548, 0.5}, {-1.821066009422915, 
                    0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-1.860432406489548, 0.5}, {-1.8898900096332212`,
                     0.5}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.005], Arrowheads[0.04], 
                   ArrowBox[{{-1.860432406489548, 0.5}, {-1.850523612566588, 
                    0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-1.860432406489548, 0.5}, 1.6, {
                    1.5667911645601529`, 1.5483800149259415`}], 
                   ArrowBox[{{-1.840566817937483, 
                    2.099876669744103}, {-1.8245693111559707`, 
                    2.0995980240026224`}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Dashing[{Small, Small}], 
                   Arrowheads[0.02], 
                   CircleBox[{-1.860432406489548, 0.5}, 1.7, {
                    1.5667911645601529`, 1.5725336813299102`}], 
                   ArrowBox[{{-1.846386069027321, 
                    2.1999419697165834`}, {-1.8633859077132624`, 
                    2.1999974343599824`}}]}, 
                  {RGBColor[0, 0, 1], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-1.860432406489548, 0.5}, 1.8, {
                    1.5667911645601529`, 1.5777262748564398`}], 
                   ArrowBox[{{-1.8549063216810953`, 
                    2.2999915173096483`}, {-1.8729062131590433`, 
                    2.2999567784108517`}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.003], Arrowheads[0.025], 
                   CircleBox[{-1.860432406489548, 0.5}, 1.5, {
                    1.5667911645601529`, 1.5667829229782635`}], 
                   ArrowBox[{{-1.8694122471257897`, 
                    1.9999731205798814`}, {-1.8544123169259707`, 
                    1.999987919458569}}]}, 
                  {RGBColor[0, 0, 1], ArrowBox[{{0, 2.8}, {0, 2.8}}]}},
                 Frame->True,
                 FrameTicks->Automatic,
                 ImageSize->800,
                 PlotRange->{{-3, 3}, {-1, 3}}], 553->
                GraphicsBox[{InsetBox[
                   TagBox[GridBox[{
                    {"\<\"t\"\>", "27.6`"},
                    {"\<\"driveIntent\"\>", "0"},
                    {"\<\"motorTorque\"\>", 
                    RowBox[{"-", "1.954245072003418`"}]},
                    {"\<\"vw\"\>", 
                    RowBox[{"-", "0.022491518173229863`"}]},
                    {"\<\"controllerState\"\>", 
                    RowBox[{"\[LeftAssociation]", 
                    RowBox[{
                    
                    RowBox[{"\<\"lastError\"\>", "\[Rule]", 
                    "0.0027713386050101773`"}], ",", 
                    RowBox[{"\<\"lastInt\"\>", "\[Rule]", 
                    RowBox[{"-", "0.019271580993843525`"}]}], ",", 
                    RowBox[{"\<\"initialized\"\>", "\[Rule]", "True"}]}], 
                    "\[RightAssociation]"}]}
                    },
                    AutoDelete->False,
                    
                    GridBoxItemSize->{
                    "Columns" -> {{Automatic}}, "Rows" -> {{Automatic}}}],
                    "Grid"], {-1, -0.5}], 
                  {GrayLevel[0], LineBox[{{-5, 0}, {5, 0}}]}, 
                  {GrayLevel[0.5], 
                   CircleBox[{-1.8615609499227854`, 0.5}, 0.5], 
                   DiskBox[{-1.8615609499227854`, 0.5}, 0.1]}, 
                  {GrayLevel[0.5], 
                   DiskBox[{-1.8555961924623816`, 1.9999881405092637`}, 0.1]}, 
                  {GrayLevel[0.5], Thickness[0.01], 
                   LineBox[{{-1.8615609499227854`, 
                    0.5}, {-1.8555961924623816`, 1.9999881405092637`}}]}, 
                  {RGBColor[0, 1, 0], 
                   DiskBox[{-1.857584444949183, 1.4999920936728426`}, 0.05]}, 
                  {RGBColor[0, 0, 1], Arrowheads[0.01], 
                   CircleBox[{-1.8615609499227854`, 0.5}, 0.3333333333333333, 
                    NCache[{
                    Rational[-1, 2] 
                    Pi, -1.6294236789549992`}, {-1.5707963267948966`, \
-1.6294236789549992`}]], 
                   ArrowBox[{{-1.8777636799991986`, 
                    0.1670606922437935}, {-1.8810922074574008`, 
                    0.16723936367077963`}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-1.8615609499227854`, 
                    0.5}, {-1.8224760484827172`, 0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-1.8615609499227854`, 
                    0.5}, {-1.8908078630598277`, 0.5}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.005], Arrowheads[0.04], 
                   ArrowBox[{{-1.8615609499227854`, 
                    0.5}, {-1.8517229616197592`, 0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-1.8615609499227854`, 0.5}, 1.6, {
                    1.5668198113414111`, 1.5485403461073275`}], 
                   ArrowBox[{{-1.8419518717414867`, 
                    2.0998798342540854`}, {-1.8259543204899902`, 
                    2.0996037534153373`}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Dashing[{Small, Small}], 
                   Arrowheads[0.02], 
                   CircleBox[{-1.8615609499227854`, 0.5}, 1.7, {
                    1.5668198113414111`, 1.5725212942718272`}], 
                   ArrowBox[{{-1.8474935551817198`, 
                    2.1999417955933662`}, {-1.864493393179313, 
                    2.19999747081475}}]}, 
                  {RGBColor[0, 0, 1], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-1.8615609499227854`, 0.5}, 1.8, {
                    1.5668198113414111`, 1.5776767284080968`}], 
                   ArrowBox[{{-1.8559456819344216`, 
                    2.2999912413024175`}, {-1.8739455751114653`, 
                    2.299957394234357}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.003], Arrowheads[0.025], 
                   CircleBox[{-1.8615609499227854`, 0.5}, 1.5, {
                    1.5668198113414111`, 1.5668116962262162`}], 
                   ArrowBox[{{-1.8705839496538248`, 
                    1.9999728615797867`}, {-1.8555840198860267`, 
                    1.9999880920551787`}}]}, 
                  {RGBColor[0, 0, 1], ArrowBox[{{0, 2.8}, {0, 2.8}}]}},
                 Frame->True,
                 FrameTicks->Automatic,
                 ImageSize->800,
                 PlotRange->{{-3, 3}, {-1, 3}}], 554->
                GraphicsBox[{InsetBox[
                   TagBox[GridBox[{
                    {"\<\"t\"\>", "27.650000000000002`"},
                    {"\<\"driveIntent\"\>", "0"},
                    {"\<\"motorTorque\"\>", 
                    RowBox[{"-", "1.9402764335831246`"}]},
                    {"\<\"vw\"\>", 
                    RowBox[{"-", "0.022330539130241914`"}]},
                    {"\<\"controllerState\"\>", 
                    RowBox[{"\[LeftAssociation]", 
                    RowBox[{
                    
                    RowBox[{"\<\"lastError\"\>", "\[Rule]", 
                    "0.002751472631284063`"}], ",", 
                    RowBox[{"\<\"lastInt\"\>", "\[Rule]", 
                    RowBox[{"-", "0.01913352119336195`"}]}], ",", 
                    RowBox[{"\<\"initialized\"\>", "\[Rule]", "True"}]}], 
                    "\[RightAssociation]"}]}
                    },
                    AutoDelete->False,
                    
                    GridBoxItemSize->{
                    "Columns" -> {{Automatic}}, "Rows" -> {{Automatic}}}],
                    "Grid"], {-1, -0.5}], 
                  {GrayLevel[0], LineBox[{{-5, 0}, {5, 0}}]}, 
                  {GrayLevel[0.5], 
                   CircleBox[{-1.8626814160592136`, 0.5}, 0.5], 
                   DiskBox[{-1.8626814160592136`, 0.5}, 0.1]}, 
                  {GrayLevel[0.5], 
                   DiskBox[{-1.8567593038160923`, 1.999988309483304}, 0.1]}, 
                  {GrayLevel[0.5], Thickness[0.01], 
                   LineBox[{{-1.8626814160592136`, 
                    0.5}, {-1.8567593038160923`, 1.999988309483304}}]}, 
                  {RGBColor[0, 1, 0], 
                   DiskBox[{-1.858733341230466, 1.4999922063222026`}, 0.05]}, 
                  {RGBColor[0, 0, 1], Arrowheads[0.01], 
                   CircleBox[{-1.8626814160592136`, 0.5}, 0.3333333333333333, 
                    NCache[{
                    Rational[-1, 2] 
                    Pi, -1.6290046198023902`}, {-1.5707963267948966`, \
-1.6290046198023902`}]], 
                   ArrowBox[{{-1.8787446234528464`, 
                    0.16705393157548737`}, {-1.8820732254926809`, 
                    0.16723120813693004`}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-1.8626814160592136`, 
                    0.5}, {-1.8238758873875511`, 0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-1.8626814160592136`, 
                    0.5}, {-1.8917192307605801`, 0.5}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.005], Arrowheads[0.04], 
                   ArrowBox[{{-1.8626814160592136`, 
                    0.5}, {-1.8529137020889175`, 0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-1.8626814160592136`, 0.5}, 1.6, {
                    1.5668482417094427`, 1.548699466075379}], 
                   ArrowBox[{{-1.843326910953196, 
                    2.099882934196156}, {-1.8273293159742479`, 
                    2.0996093988907365`}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Dashing[{Small, Small}], 
                   Arrowheads[0.02], 
                   CircleBox[{-1.8626814160592136`, 0.5}, 1.7, {
                    1.5668482417094427`, 1.572508998682117}], 
                   ArrowBox[{{-1.848593119532362, 
                    2.1999416224979527`}, {-1.8655929568441108`, 
                    2.199997506742365}}]}, 
                  {RGBColor[0, 0, 1], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-1.8626814160592136`, 0.5}, 1.8, {
                    1.5668482417094427`, 1.577627555229349}], 
                   ArrowBox[{{-1.8569776367866213`, 
                    2.2999909630056505`}, {-1.8749775316062711`, 
                    2.2999580010495952`}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.003], Arrowheads[0.025], 
                   CircleBox[{-1.8626814160592136`, 0.5}, 1.5, {
                    1.5668482417094427`, 1.566840248492693}], 
                   ArrowBox[{{-1.871747243411421, 
                    1.999972603341281}, {-1.856747314084602, 
                    1.9999882620986589`}}]}, 
                  {RGBColor[0, 0, 1], ArrowBox[{{0, 2.8}, {0, 2.8}}]}},
                 Frame->True,
                 FrameTicks->Automatic,
                 ImageSize->800,
                 PlotRange->{{-3, 3}, {-1, 3}}], 555->
                GraphicsBox[{InsetBox[
                   TagBox[GridBox[{
                    {"\<\"t\"\>", "27.7`"},
                    {"\<\"driveIntent\"\>", "0"},
                    {"\<\"motorTorque\"\>", 
                    RowBox[{"-", "1.9264123324514646`"}]},
                    {"\<\"vw\"\>", 
                    RowBox[{"-", "0.022170709837299785`"}]},
                    {"\<\"controllerState\"\>", 
                    RowBox[{"\[LeftAssociation]", 
                    RowBox[{
                    RowBox[{"\<\"lastError\"\>", "\[Rule]", 
                    "0.0027317384073892873`"}], ",", 
                    RowBox[{"\<\"lastInt\"\>", "\[Rule]", 
                    RowBox[{"-", "0.018996451334496223`"}]}], ",", 
                    RowBox[{"\<\"initialized\"\>", "\[Rule]", "True"}]}], 
                    "\[RightAssociation]"}]}
                    },
                    AutoDelete->False,
                    
                    GridBoxItemSize->{
                    "Columns" -> {{Automatic}}, "Rows" -> {{Automatic}}}],
                    "Grid"], {-1, -0.5}], 
                  {GrayLevel[0], LineBox[{{-5, 0}, {5, 0}}]}, 
                  {GrayLevel[0.5], 
                   CircleBox[{-1.8637938625979014`, 0.5}, 0.5], 
                   DiskBox[{-1.8637938625979014`, 0.5}, 0.1]}, 
                  {GrayLevel[0.5], 
                   DiskBox[{-1.8579140758348727`, 1.9999884759916062`}, 0.1]}, 
                  {GrayLevel[0.5], Thickness[0.01], 
                   LineBox[{{-1.8637938625979014`, 
                    0.5}, {-1.8579140758348727`, 1.9999884759916062`}}]}, 
                  {RGBColor[0, 1, 0], 
                   DiskBox[{-1.8598740047558822`, 1.4999923173277374`}, 
                    0.05]}, 
                  {RGBColor[0, 0, 1], Arrowheads[0.01], 
                   CircleBox[{-1.8637938625979014`, 0.5}, 0.3333333333333333, 
                    NCache[{
                    Rational[-1, 2] 
                    Pi, -1.6285886967684404`}, {-1.5707963267948966`, \
-1.6285886967684404`}]], 
                   ArrowBox[{{-1.879718588667203, 
                    0.16704727931622365`}, {-1.88304726415253, 
                    0.16722317142011328`}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-1.8637938625979014`, 0.5}, {-1.825265615948872,
                     0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-1.8637938625979014`, 
                    0.5}, {-1.8926241465324045`, 0.5}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.005], Arrowheads[0.04], 
                   ArrowBox[{{-1.8637938625979014`, 
                    0.5}, {-1.8540958998833752`, 0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-1.8637938625979014`, 0.5}, 1.6, {
                    1.566876458914519, 1.5488573930208767`}], 
                   ArrowBox[{{-1.8446920223571446`, 
                    2.09988597084274}, {-1.8286943843790986`, 
                    2.0996149619920317`}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Dashing[{Small, Small}], 
                   Arrowheads[0.02], 
                   CircleBox[{-1.8637938625979014`, 0.5}, 1.7, {
                    1.566876458914519, 1.5724967930534006`}], 
                   ArrowBox[{{-1.849684817215816, 
                    2.1999414504148094`}, {-1.8666846538441961`, 
                    2.1999975421529205`}}]}, 
                  {RGBColor[0, 0, 1], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-1.8637938625979014`, 0.5}, 1.8, {
                    1.566876458914519, 1.5775787496503604`}], 
                   ArrowBox[{{-1.8580022337310134`, 
                    2.2999906824856255`}, {-1.8760021301379526`, 
                    2.2999585990248974`}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.003], Arrowheads[0.025], 
                   CircleBox[{-1.8637938625979014`, 0.5}, 1.5, {
                    1.566876458914519, 1.5668685828683764`}], 
                   ArrowBox[{{-1.8729021907337218`, 
                    1.9999723458646064`}, {-1.857902261856605, 
                    1.9999884296356105`}}]}, 
                  {RGBColor[0, 0, 1], ArrowBox[{{0, 2.8}, {0, 2.8}}]}},
                 Frame->True,
                 FrameTicks->Automatic,
                 ImageSize->800,
                 PlotRange->{{-3, 3}, {-1, 3}}], 556->
                GraphicsBox[{InsetBox[
                   TagBox[GridBox[{
                    {"\<\"t\"\>", "27.75`"},
                    {"\<\"driveIntent\"\>", "0"},
                    {"\<\"motorTorque\"\>", 
                    RowBox[{"-", "1.9126512279271384`"}]},
                    {"\<\"vw\"\>", 
                    RowBox[{"-", "0.022012021768005305`"}]},
                    {"\<\"controllerState\"\>", 
                    RowBox[{"\[LeftAssociation]", 
                    RowBox[{
                    
                    RowBox[{"\<\"lastError\"\>", "\[Rule]", 
                    "0.0027121365015125576`"}], ",", 
                    RowBox[{"\<\"lastInt\"\>", "\[Rule]", 
                    RowBox[{"-", "0.018860364814896037`"}]}], ",", 
                    RowBox[{"\<\"initialized\"\>", "\[Rule]", "True"}]}], 
                    "\[RightAssociation]"}]}
                    },
                    AutoDelete->False,
                    
                    GridBoxItemSize->{
                    "Columns" -> {{Automatic}}, "Rows" -> {{Automatic}}}],
                    "Grid"], {-1, -0.5}], 
                  {GrayLevel[0], LineBox[{{-5, 0}, {5, 0}}]}, 
                  {GrayLevel[0.5], 
                   CircleBox[{-1.8648983468083917`, 0.5}, 0.5], 
                   DiskBox[{-1.8648983468083917`, 0.5}, 0.1]}, 
                  {GrayLevel[0.5], 
                   DiskBox[{-1.8590605704745835`, 1.9999886400794762`}, 0.1]}, 
                  {GrayLevel[0.5], Thickness[0.01], 
                   LineBox[{{-1.8648983468083917`, 
                    0.5}, {-1.8590605704745835`, 1.9999886400794762`}}]}, 
                  {RGBColor[0, 1, 0], 
                   DiskBox[{-1.8610064959191863`, 1.499992426719651}, 0.05]}, 
                  {RGBColor[0, 0, 1], Arrowheads[0.01], 
                   CircleBox[{-1.8648983468083917`, 0.5}, 0.3333333333333333, 
                    NCache[{
                    Rational[-1, 2] 
                    Pi, -1.6281758636327108`}, {-1.5707963267948966`, \
-1.6281758636327108`}]], 
                   ArrowBox[{{-1.8806856176088331`, 
                    0.1670407334345767}, {-1.8840143654245918`, 
                    0.16721525133597814`}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-1.8648983468083917`, 
                    0.5}, {-1.8266453222498489`, 0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-1.8648983468083917`, 
                    0.5}, {-1.8935226446688873`, 0.5}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.005], Arrowheads[0.04], 
                   ArrowBox[{{-1.8648983468083917`, 
                    0.5}, {-1.8552696201103445`, 0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-1.8648983468083917`, 0.5}, 1.6, {
                    1.5669044660809688`, 1.5490141444297376`}], 
                   ArrowBox[{{-1.84604729118123, 
                    2.099888945427695}, {-1.830049610918704, 
                    2.099620444232597}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Dashing[{Small, Small}], 
                   Arrowheads[0.02], 
                   CircleBox[{-1.8648983468083917`, 0.5}, 1.7, {
                    1.5669044660809688`, 1.572484675968262}], 
                   ArrowBox[{{-1.8507687030920574`, 
                    2.1999412793295097`}, {-1.867768539039521, 
                    2.199997577056084}}]}, 
                  {RGBColor[0, 0, 1], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-1.8648983468083917`, 0.5}, 1.8, {
                    1.5669044660809688`, 1.5775303062361195`}], 
                   ArrowBox[{{-1.859019520254072, 
                    2.2999903998072724`}, {-1.877019418194123, 
                    2.299959188323019}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.003], Arrowheads[0.025], 
                   CircleBox[{-1.8648983468083917`, 0.5}, 1.5, {
                    1.5669044660809688`, 1.566896702378102}], 
                   ArrowBox[{{-1.8740488534275732`, 
                    1.9999720891498656`}, {-1.8590489250086544`, 
                    1.9999885947115095`}}]}, 
                  {RGBColor[0, 0, 1], ArrowBox[{{0, 2.8}, {0, 2.8}}]}},
                 Frame->True,
                 FrameTicks->Automatic,
                 ImageSize->800,
                 PlotRange->{{-3, 3}, {-1, 3}}], 557->
                GraphicsBox[{InsetBox[
                   TagBox[GridBox[{
                    {"\<\"t\"\>", "27.8`"},
                    {"\<\"driveIntent\"\>", "0"},
                    {"\<\"motorTorque\"\>", 
                    RowBox[{"-", "1.8989916452802342`"}]},
                    {"\<\"vw\"\>", 
                    RowBox[{"-", "0.02185446652167384`"}]},
                    {"\<\"controllerState\"\>", 
                    RowBox[{"\[LeftAssociation]", 
                    RowBox[{
                    
                    RowBox[{"\<\"lastError\"\>", "\[Rule]", 
                    "0.0026926673910216705`"}], ",", 
                    RowBox[{"\<\"lastInt\"\>", "\[Rule]", 
                    RowBox[{"-", "0.01872525500610185`"}]}], ",", 
                    RowBox[{"\<\"initialized\"\>", "\[Rule]", "True"}]}], 
                    "\[RightAssociation]"}]}
                    },
                    AutoDelete->False,
                    
                    GridBoxItemSize->{
                    "Columns" -> {{Automatic}}, "Rows" -> {{Automatic}}}],
                    "Grid"], {-1, -0.5}], 
                  {GrayLevel[0], LineBox[{{-5, 0}, {5, 0}}]}, 
                  {GrayLevel[0.5], CircleBox[{-1.865994925537138, 0.5}, 0.5], 
                   DiskBox[{-1.865994925537138, 0.5}, 0.1]}, 
                  {GrayLevel[0.5], 
                   DiskBox[{-1.8601988490750896`, 1.9999888017907488`}, 0.1]}, 
                  {GrayLevel[0.5], Thickness[0.01], 
                   LineBox[{{-1.865994925537138, 0.5}, {-1.8601988490750896`, 
                    1.9999888017907488`}}]}, 
                  {RGBColor[0, 1, 0], 
                   DiskBox[{-1.862130874562439, 1.4999925345271659`}, 0.05]}, 
                  {RGBColor[0, 0, 1], Arrowheads[0.01], 
                   CircleBox[{-1.865994925537138, 0.5}, 0.3333333333333333, 
                    NCache[{
                    Rational[-1, 2] 
                    Pi, -1.6277660761533035`}, {-1.5707963267948966`, \
-1.6277660761533035`}]], 
                   ArrowBox[{{-1.881645752477262, 
                    0.1670342919650713}, {-1.8849745715287785`, 
                    0.16720744577268093`}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-1.865994925537138, 0.5}, {-1.8280150926315333`,
                     0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-1.865994925537138, 0.5}, {-1.8944147599863255`,
                     0.5}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.005], Arrowheads[0.04], 
                   ArrowBox[{{-1.865994925537138, 0.5}, {-1.8564349270807208`,
                     0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-1.865994925537138, 0.5}, 1.6, {
                    1.566932266204513, 1.5491697370680921`}], 
                   ArrowBox[{{-1.8473928010792484`, 
                    2.0998918591472533`}, {-1.8313950792335578`, 
                    2.0996258470766747`}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Dashing[{Small, Small}], 
                   Arrowheads[0.02], 
                   CircleBox[{-1.865994925537138, 0.5}, 1.7, {
                    1.566932266204513, 1.57247264609696}], 
                   ArrowBox[{{-1.8518448317470153`, 
                    2.1999411092287082`}, {-1.8688446670159944`, 
                    2.199997611461117}}]}, 
                  {RGBColor[0, 0, 1], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-1.865994925537138, 0.5}, 1.8, {
                    1.566932266204513, 1.5774822197894032`}], 
                   ArrowBox[{{-1.8600295438471983`, 
                    2.299990115034328}, {-1.8780294432672882`, 
                    2.299959769101244}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.003], Arrowheads[0.025], 
                   CircleBox[{-1.865994925537138, 0.5}, 1.5, {
                    1.566932266204513, 1.5669246099723055`}], 
                   ArrowBox[{{-1.8751872927651312`, 
                    1.9999718331971255`}, {-1.8601873648126843`, 
                    1.9999887573706783`}}]}, 
                  {RGBColor[0, 0, 1], ArrowBox[{{0, 2.8}, {0, 2.8}}]}},
                 Frame->True,
                 FrameTicks->Automatic,
                 ImageSize->800,
                 PlotRange->{{-3, 3}, {-1, 3}}], 558->
                GraphicsBox[{InsetBox[
                   TagBox[GridBox[{
                    {"\<\"t\"\>", "27.85`"},
                    {"\<\"driveIntent\"\>", "0"},
                    {"\<\"motorTorque\"\>", 
                    RowBox[{"-", "1.8854321760416428`"}]},
                    {"\<\"vw\"\>", 
                    RowBox[{"-", "0.02169803581722319`"}]},
                    {"\<\"controllerState\"\>", 
                    RowBox[{"\[LeftAssociation]", 
                    RowBox[{
                    
                    RowBox[{"\<\"lastError\"\>", "\[Rule]", 
                    "0.0026733314593840323`"}], ",", 
                    RowBox[{"\<\"lastInt\"\>", "\[Rule]", 
                    RowBox[{"-", "0.018591115258170254`"}]}], ",", 
                    RowBox[{"\<\"initialized\"\>", "\[Rule]", "True"}]}], 
                    "\[RightAssociation]"}]}
                    },
                    AutoDelete->False,
                    
                    GridBoxItemSize->{
                    "Columns" -> {{Automatic}}, "Rows" -> {{Automatic}}}],
                    "Grid"], {-1, -0.5}], 
                  {GrayLevel[0], LineBox[{{-5, 0}, {5, 0}}]}, 
                  {GrayLevel[0.5], 
                   CircleBox[{-1.8670836552136345`, 0.5}, 0.5], 
                   DiskBox[{-1.8670836552136345`, 0.5}, 0.1]}, 
                  {GrayLevel[0.5], 
                   DiskBox[{-1.8613289723645587`, 1.999988961167817}, 0.1]}, 
                  {GrayLevel[0.5], Thickness[0.01], 
                   LineBox[{{-1.8670836552136345`, 
                    0.5}, {-1.8613289723645587`, 1.999988961167817}}]}, 
                  {RGBColor[0, 1, 0], 
                   DiskBox[{-1.8632471999809173`, 1.4999926407785447`}, 
                    0.05]}, 
                  {RGBColor[0, 0, 1], Arrowheads[0.01], 
                   CircleBox[{-1.8670836552136345`, 0.5}, 0.3333333333333333, 
                    NCache[{
                    Rational[-1, 2] 
                    Pi, -1.627359292076146}, {-1.5707963267948966`, \
-1.627359292076146}]], 
                   ArrowBox[{{-1.8825990357143292`, 
                    0.16702795300651774`}, {-1.88592792492664, 
                    0.16719975268925263`}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-1.8670836552136345`, 
                    0.5}, {-1.8293750116928016`, 0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-1.8670836552136345`, 
                    0.5}, {-1.8953005278388366`, 0.5}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.005], Arrowheads[0.04], 
                   ArrowBox[{{-1.8670836552136345`, 
                    0.5}, {-1.8575918843180037`, 0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-1.8670836552136345`, 0.5}, 1.6, {
                    1.5669598621510437`, 1.5493241869754533`}], 
                   ArrowBox[{{-1.8487286341260705`, 
                    2.0998947131611114`}, {-1.8327308713856514`, 
                    2.099631171940353}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Dashing[{Small, Small}], 
                   Arrowheads[0.02], 
                   CircleBox[{-1.8670836552136345`, 0.5}, 1.7, {
                    1.5669598621510437`, 1.572460702194592}], 
                   ArrowBox[{{-1.8529132574938816`, 
                    2.1999409401001153`}, {-1.869913092086791, 
                    2.199997645376893}}]}, 
                  {RGBColor[0, 0, 1], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-1.8670836552136345`, 0.5}, 1.8, {
                    1.5669598621510437`, 1.5774344853512752`}], 
                   ArrowBox[{{-1.8610323520137468`, 
                    2.299989828229477}, {-1.8790322528618755`, 
                    2.2999603415115124`}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.003], Arrowheads[0.025], 
                   CircleBox[{-1.8670836552136345`, 0.5}, 1.5, {
                    1.5669598621510437`, 1.5669523085200463`}], 
                   ArrowBox[{{-1.8763175694795278`, 
                    1.9999715780065068`}, {-1.86131764200161, 
                    1.9999889176562735`}}]}, 
                  {RGBColor[0, 0, 1], ArrowBox[{{0, 2.8}, {0, 2.8}}]}},
                 Frame->True,
                 FrameTicks->Automatic,
                 ImageSize->800,
                 PlotRange->{{-3, 3}, {-1, 3}}], 559->
                GraphicsBox[{InsetBox[
                   TagBox[GridBox[{
                    {"\<\"t\"\>", "27.900000000000002`"},
                    {"\<\"driveIntent\"\>", "0"},
                    {"\<\"motorTorque\"\>", 
                    RowBox[{"-", "1.8719714776875194`"}]},
                    {"\<\"vw\"\>", 
                    RowBox[{"-", "0.02154272148713283`"}]},
                    {"\<\"controllerState\"\>", 
                    RowBox[{"\[LeftAssociation]", 
                    RowBox[{
                    
                    RowBox[{"\<\"lastError\"\>", "\[Rule]", 
                    "0.0026541289944016265`"}], ",", 
                    RowBox[{"\<\"lastInt\"\>", "\[Rule]", 
                    RowBox[{"-", "0.018457938904419095`"}]}], ",", 
                    RowBox[{"\<\"initialized\"\>", "\[Rule]", "True"}]}], 
                    "\[RightAssociation]"}]}
                    },
                    AutoDelete->False,
                    
                    GridBoxItemSize->{
                    "Columns" -> {{Automatic}}, "Rows" -> {{Automatic}}}],
                    "Grid"], {-1, -0.5}], 
                  {GrayLevel[0], LineBox[{{-5, 0}, {5, 0}}]}, 
                  {GrayLevel[0.5], 
                   CircleBox[{-1.8681645918562437`, 0.5}, 0.5], 
                   DiskBox[{-1.8681645918562437`, 0.5}, 0.1]}, 
                  {GrayLevel[0.5], 
                   DiskBox[{-1.8624510004653971`, 1.9999891182516687`}, 0.1]}, 
                  {GrayLevel[0.5], Thickness[0.01], 
                   LineBox[{{-1.8681645918562437`, 
                    0.5}, {-1.8624510004653971`, 1.9999891182516687`}}]}, 
                  {RGBColor[0, 1, 0], 
                   DiskBox[{-1.8643555309290125`, 1.4999927455011124`}, 
                    0.05]}, 
                  {RGBColor[0, 0, 1], Arrowheads[0.01], 
                   CircleBox[{-1.8681645918562437`, 0.5}, 0.3333333333333333, 
                    NCache[{
                    Rational[-1, 2] 
                    Pi, -1.626955471125522}, {-1.5707963267948966`, \
-1.626955471125522}]], 
                   ArrowBox[{{-1.8835455100069878`, 
                    0.16702171472008687`}, {-1.8868744683241845`, 
                    0.1671921701136443}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-1.8681645918562437`, 
                    0.5}, {-1.8307251623024934`, 0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-1.8681645918562437`, 
                    0.5}, {-1.8961799841236606`, 0.5}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.005], Arrowheads[0.04], 
                   ArrowBox[{{-1.8681645918562437`, 0.5}, {-1.85874055456991, 
                    0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-1.8681645918562437`, 0.5}, 1.6, {
                    1.566987256656696, 1.549477509465119}], 
                   ArrowBox[{{-1.8500548708240865`, 
                    2.0998975085936404`}, {-1.8340570678649075`, 
                    2.0996364201927826`}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Dashing[{Small, Small}], 
                   Arrowheads[0.02], 
                   CircleBox[{-1.8681645918562437`, 0.5}, 1.7, {
                    1.566987256656696, 1.5724488430978136`}], 
                   ArrowBox[{{-1.8539740343733615`, 
                    2.199940771932459}, {-1.870973868292602, 
                    2.1999976788119167`}}]}, 
                  {RGBColor[0, 0, 1], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-1.8681645918562437`, 0.5}, 1.8, {
                    1.566987256656696, 1.5773870981994045`}], 
                   ArrowBox[{{-1.8620279922718261`, 
                    2.299989539454477}, {-1.8800278944970366`, 
                    2.2999609057005803`}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.003], Arrowheads[0.025], 
                   CircleBox[{-1.8681645918562437`, 0.5}, 1.5, {
                    1.566987256656696, 1.5669798008035307`}], 
                   ArrowBox[{{-1.8774397437624835`, 
                    1.9999713235782597`}, {-1.8624398167669407`, 
                    1.9999890756102783`}}]}, 
                  {RGBColor[0, 0, 1], ArrowBox[{{0, 2.8}, {0, 2.8}}]}},
                 Frame->True,
                 FrameTicks->Automatic,
                 ImageSize->800,
                 PlotRange->{{-3, 3}, {-1, 3}}], 560->
                GraphicsBox[{InsetBox[
                   TagBox[GridBox[{
                    {"\<\"t\"\>", "27.95`"},
                    {"\<\"driveIntent\"\>", "0"},
                    {"\<\"motorTorque\"\>", 
                    RowBox[{"-", "1.8586082727725424`"}]},
                    {"\<\"vw\"\>", 
                    RowBox[{"-", "0.02138851547151982`"}]},
                    {"\<\"controllerState\"\>", 
                    RowBox[{"\[LeftAssociation]", 
                    RowBox[{
                    
                    RowBox[{"\<\"lastError\"\>", "\[Rule]", 
                    "0.002635060187628278`"}], ",", 
                    RowBox[{"\<\"lastInt\"\>", "\[Rule]", 
                    RowBox[{"-", "0.018325719266230365`"}]}], ",", 
                    RowBox[{"\<\"initialized\"\>", "\[Rule]", "True"}]}], 
                    "\[RightAssociation]"}]}
                    },
                    AutoDelete->False,
                    
                    GridBoxItemSize->{
                    "Columns" -> {{Automatic}}, "Rows" -> {{Automatic}}}],
                    "Grid"], {-1, -0.5}], 
                  {GrayLevel[0], LineBox[{{-5, 0}, {5, 0}}]}, 
                  {GrayLevel[0.5], CircleBox[{-1.869237791077721, 0.5}, 0.5], 
                   DiskBox[{-1.869237791077721, 0.5}, 0.1]}, 
                  {GrayLevel[0.5], 
                   DiskBox[{-1.8635649929015992`, 1.9999892730819286`}, 0.1]}, 
                  {GrayLevel[0.5], Thickness[0.01], 
                   LineBox[{{-1.869237791077721, 0.5}, {-1.8635649929015992`, 
                    1.9999892730819286`}}]}, 
                  {RGBColor[0, 1, 0], 
                   DiskBox[{-1.8654559256269732`, 1.499992848721286}, 0.05]}, 
                  {RGBColor[0, 0, 1], Arrowheads[0.01], 
                   CircleBox[{-1.869237791077721, 0.5}, 0.3333333333333333, 
                    NCache[{
                    Rational[-1, 2] 
                    Pi, -1.6265545749780728`}, {-1.5707963267948966`, \
-1.6265545749780728`}]], 
                   ArrowBox[{{-1.8844852182842946`, 
                    0.16701557532717637`}, {-1.8878142446688888`, 
                    0.16718469614050757`}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-1.869237791077721, 0.5}, {-1.83206562562227, 
                    0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-1.869237791077721, 0.5}, {-1.8970531652777372`,
                     0.5}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.005], Arrowheads[0.04], 
                   ArrowBox[{{-1.869237791077721, 0.5}, {-1.8598809998222863`,
                     0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-1.869237791077721, 0.5}, 1.6, {
                    1.567014452329065, 1.5496297191309825`}], 
                   ArrowBox[{{-1.8513715901196033`, 
                    2.0999002465351784`}, {-1.835373747605563, 
                    2.0996415931575783`}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Dashing[{Small, Small}], 
                   Arrowheads[0.02], 
                   CircleBox[{-1.869237791077721, 0.5}, 1.7, {
                    1.567014452329065, 1.5724370677211943`}], 
                   ArrowBox[{{-1.8550272161530028`, 
                    2.1999406047154437`}, {-1.872027049400965, 
                    2.199997711774344}}]}, 
                  {RGBColor[0, 0, 1], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-1.869237791077721, 0.5}, 1.8, {
                    1.567014452329065, 1.577340053844468}], 
                   ArrowBox[{{-1.863016512153349, 
                    2.299989248770266}, {-1.8810164157056977`, 
                    2.2999614618101893`}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.003], Arrowheads[0.025], 
                   CircleBox[{-1.869237791077721, 0.5}, 1.5, {
                    1.567014452329065, 1.567007089514007}], 
                   ArrowBox[{{-1.8785538752636737`, 
                    1.9999710699128301`}, {-1.863553948758146, 
                    1.999989231273507}}]}, 
                  {RGBColor[0, 0, 1], ArrowBox[{{0, 2.8}, {0, 2.8}}]}},
                 Frame->True,
                 FrameTicks->Automatic,
                 ImageSize->800,
                 PlotRange->{{-3, 3}, {-1, 3}}], 561->
                GraphicsBox[{InsetBox[
                   TagBox[GridBox[{
                    {"\<\"t\"\>", "28.`"},
                    {"\<\"driveIntent\"\>", "0"},
                    {"\<\"motorTorque\"\>", 
                    RowBox[{"-", "1.8453413475824307`"}]},
                    {"\<\"vw\"\>", 
                    RowBox[{"-", "0.021235409812370928`"}]},
                    {"\<\"controllerState\"\>", 
                    RowBox[{"\[LeftAssociation]", 
                    RowBox[{
                    
                    RowBox[{"\<\"lastError\"\>", "\[Rule]", 
                    "0.002616125134838991`"}], ",", 
                    RowBox[{"\<\"lastInt\"\>", "\[Rule]", 
                    RowBox[{"-", "0.018194449657854658`"}]}], ",", 
                    RowBox[{"\<\"initialized\"\>", "\[Rule]", "True"}]}], 
                    "\[RightAssociation]"}]}
                    },
                    AutoDelete->False,
                    
                    GridBoxItemSize->{
                    "Columns" -> {{Automatic}}, "Rows" -> {{Automatic}}}],
                    "Grid"], {-1, -0.5}], 
                  {GrayLevel[0], LineBox[{{-5, 0}, {5, 0}}]}, 
                  {GrayLevel[0.5], CircleBox[{-1.870303308090447, 0.5}, 0.5], 
                   DiskBox[{-1.870303308090447, 0.5}, 0.1]}, 
                  {GrayLevel[0.5], 
                   DiskBox[{-1.8646710086073093`, 1.9999894256969055`}, 0.1]}, 
                  {GrayLevel[0.5], Thickness[0.01], 
                   LineBox[{{-1.870303308090447, 0.5}, {-1.8646710086073093`, 
                    1.9999894256969055`}}]}, 
                  {RGBColor[0, 1, 0], 
                   DiskBox[{-1.866548441768355, 1.4999929504646037`}, 0.05]}, 
                  {RGBColor[0, 0, 1], Arrowheads[0.01], 
                   CircleBox[{-1.870303308090447, 0.5}, 0.3333333333333333, 
                    NCache[{
                    Rational[-1, 2] 
                    Pi, -1.6261565672223695`}, {-1.5707963267948966`, \
-1.6261565672223695`}]], 
                   ArrowBox[{{-1.8854182037092977`, 
                    0.1670095331071138}, {-1.8887472971416097`, 
                    0.1671773289287647}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-1.870303308090447, 0.5}, {-1.8333964811387984`,
                     0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-1.870303308090447, 0.5}, {-1.8979201082666204`,
                     0.5}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.005], Arrowheads[0.04], 
                   ArrowBox[{{-1.870303308090447, 0.5}, {-1.8610132813149718`,
                     0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-1.870303308090447, 0.5}, 1.6, {
                    1.5670414516494255`, 1.549780829859953}], 
                   ArrowBox[{{-1.8526788694279255`, 
                    2.099902928043396}, {-1.836680988011236, 
                    2.0996466921143844`}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Dashing[{Small, Small}], 
                   Arrowheads[0.02], 
                   CircleBox[{-1.870303308090447, 0.5}, 1.7, {
                    1.5670414516494255`, 1.572425375053277}], 
                   ArrowBox[{{-1.856072856325731, 
                    2.199940438439704}, {-1.873072688904797, 
                    2.1999977442720047`}}]}, 
                  {RGBColor[0, 0, 1], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-1.870303308090447, 0.5}, 1.8, {
                    1.5670414516494255`, 1.5772933480248834`}], 
                   ArrowBox[{{-1.8639979591997842`, 
                    2.2999889562370566`}, {-1.8819978640303128`, 
                    2.2999620099772575`}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.003], Arrowheads[0.025], 
                   CircleBox[{-1.870303308090447, 0.5}, 1.5, {
                    1.5670414516494255`, 1.5670341772489025`}], 
                   ArrowBox[{{-1.8796600230916698`, 
                    1.999970817010913}, {-1.8646600970835954`, 
                    1.9999893846856158`}}]}, 
                  {RGBColor[0, 0, 1], ArrowBox[{{0, 2.8}, {0, 2.8}}]}},
                 Frame->True,
                 FrameTicks->Automatic,
                 ImageSize->800,
                 PlotRange->{{-3, 3}, {-1, 3}}], 562->
                GraphicsBox[{InsetBox[
                   TagBox[GridBox[{
                    {"\<\"t\"\>", "28.05`"},
                    {"\<\"driveIntent\"\>", "0"},
                    {"\<\"motorTorque\"\>", 
                    RowBox[{"-", "1.8321695503740971`"}]},
                    {"\<\"vw\"\>", 
                    RowBox[{"-", "0.021083396647964887`"}]},
                    {"\<\"controllerState\"\>", 
                    RowBox[{"\[LeftAssociation]", 
                    RowBox[{
                    
                    RowBox[{"\<\"lastError\"\>", "\[Rule]", 
                    "0.0025973238374260403`"}], ",", 
                    RowBox[{"\<\"lastInt\"\>", "\[Rule]", 
                    RowBox[{"-", "0.0180641233911682`"}]}], ",", 
                    RowBox[{"\<\"initialized\"\>", "\[Rule]", "True"}]}], 
                    "\[RightAssociation]"}]}
                    },
                    AutoDelete->False,
                    
                    GridBoxItemSize->{
                    "Columns" -> {{Automatic}}, "Rows" -> {{Automatic}}}],
                    "Grid"], {-1, -0.5}], 
                  {GrayLevel[0], LineBox[{{-5, 0}, {5, 0}}]}, 
                  {GrayLevel[0.5], 
                   CircleBox[{-1.8713611977113829`, 0.5}, 0.5], 
                   DiskBox[{-1.8713611977113829`, 0.5}, 0.1]}, 
                  {GrayLevel[0.5], 
                   DiskBox[{-1.8657691059364057`, 1.999989576133641}, 0.1]}, 
                  {GrayLevel[0.5], Thickness[0.01], 
                   LineBox[{{-1.8713611977113829`, 
                    0.5}, {-1.8657691059364057`, 1.999989576133641}}]}, 
                  {RGBColor[0, 1, 0], 
                   DiskBox[{-1.8676331365280647`, 1.4999930507557608`}, 
                    0.05]}, 
                  {RGBColor[0, 0, 1], Arrowheads[0.01], 
                   CircleBox[{-1.8713611977113829`, 0.5}, 0.3333333333333333, 
                    NCache[{
                    Rational[-1, 2] 
                    Pi, -1.6257614133061196`}, {-1.5707963267948966`, \
-1.6257614133061196`}]], 
                   ArrowBox[{{-1.8863445096665215`, 
                    0.16700358639473895`}, {-1.8896736691440945`, 
                    0.16717006669901635`}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-1.8713611977113829`, 0.5}, {-1.834717806703901,
                     0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-1.8713611977113829`, 
                    0.5}, {-1.8987808505667338`, 0.5}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.005], Arrowheads[0.04], 
                   ArrowBox[{{-1.8713611977113829`, 0.5}, {-1.862137459559252,
                     0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-1.8713611977113829`, 0.5}, 1.6, {
                    1.5670682569758185`, 1.5499308548492312`}], 
                   ArrowBox[{{-1.853976784665928, 
                    2.099905554144702}, {-1.8379788649874822`, 
                    2.099651718300553}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Dashing[{Small, Small}], 
                   Arrowheads[0.02], 
                   CircleBox[{-1.8713611977113829`, 0.5}, 1.7, {
                    1.5670682569758185`, 1.5724137641524225`}], 
                   ArrowBox[{{-1.857111008107737, 
                    2.199940273096752}, {-1.874110840020285, 
                    2.1999977763124203`}}]}, 
                  {RGBColor[0, 0, 1], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-1.8713611977113829`, 0.5}, 1.8, {
                    1.5670682569758185`, 1.5772469767001192`}], 
                   ArrowBox[{{-1.8649723809550673`, 
                    2.2999886619144174`}, {-1.882972287015777, 
                    2.2999625503340804`}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.003], Arrowheads[0.025], 
                   CircleBox[{-1.8713611977113829`, 0.5}, 1.5, {
                    1.5670682569758185`, 1.567061066510104}], 
                   ArrowBox[{{-1.8807582458163115`, 
                    1.9999705648734956`}, {-1.8657583203129309`, 
                    1.99998953588512}}]}, 
                  {RGBColor[0, 0, 1], ArrowBox[{{0, 2.8}, {0, 2.8}}]}},
                 Frame->True,
                 FrameTicks->Automatic,
                 ImageSize->800,
                 PlotRange->{{-3, 3}, {-1, 3}}], 563->
                GraphicsBox[{InsetBox[
                   TagBox[GridBox[{
                    {"\<\"t\"\>", "28.1`"},
                    {"\<\"driveIntent\"\>", "0"},
                    {"\<\"motorTorque\"\>", 
                    RowBox[{"-", "1.8190917892664982`"}]},
                    {"\<\"vw\"\>", 
                    RowBox[{"-", "0.020932468207512763`"}]},
                    {"\<\"controllerState\"\>", 
                    RowBox[{"\[LeftAssociation]", 
                    RowBox[{
                    
                    RowBox[{"\<\"lastError\"\>", "\[Rule]", 
                    "0.002578656204601935`"}], ",", 
                    RowBox[{"\<\"lastInt\"\>", "\[Rule]", 
                    RowBox[{"-", "0.017934733780338857`"}]}], ",", 
                    RowBox[{"\<\"initialized\"\>", "\[Rule]", "True"}]}], 
                    "\[RightAssociation]"}]}
                    },
                    AutoDelete->False,
                    
                    GridBoxItemSize->{
                    "Columns" -> {{Automatic}}, "Rows" -> {{Automatic}}}],
                    "Grid"], {-1, -0.5}], 
                  {GrayLevel[0], LineBox[{{-5, 0}, {5, 0}}]}, 
                  {GrayLevel[0.5], 
                   CircleBox[{-1.8724115143667415`, 0.5}, 0.5], 
                   DiskBox[{-1.8724115143667415`, 0.5}, 0.1]}, 
                  {GrayLevel[0.5], 
                   DiskBox[{-1.8668593426729116`, 1.999989724427965}, 0.1]}, 
                  {GrayLevel[0.5], Thickness[0.01], 
                   LineBox[{{-1.8724115143667415`, 
                    0.5}, {-1.8668593426729116`, 1.999989724427965}}]}, 
                  {RGBColor[0, 1, 0], 
                   DiskBox[{-1.868710066570855, 1.4999931496186432`}, 0.05]}, 
                  {RGBColor[0, 0, 1], Arrowheads[0.01], 
                   CircleBox[{-1.8724115143667415`, 0.5}, 0.3333333333333333, 
                    NCache[{
                    Rational[-1, 2] 
                    Pi, -1.6253690804728915`}, {-1.5707963267948966`, \
-1.6253690804728915`}]], 
                   ArrowBox[{{-1.8872641797456735`, 
                    0.16699773357790404`}, {-1.890593404282714, 
                    0.16716290773083214`}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-1.8724115143667415`, 
                    0.5}, {-1.8360296785814114`, 0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-1.8724115143667415`, 
                    0.5}, {-1.8996354301419074`, 0.5}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.005], Arrowheads[0.04], 
                   ArrowBox[{{-1.8724115143667415`, 
                    0.5}, {-1.8632535943565773`, 0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-1.8724115143667415`, 0.5}, 1.6, {
                    1.5670948705468772`, 1.5500798066277217`}], 
                   ArrowBox[{{-1.855265410290963, 
                    2.0999081258356753`}, {-1.8392674529806867`, 
                    2.0996566729129214`}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Dashing[{Small, Small}], 
                   Arrowheads[0.02], 
                   CircleBox[{-1.8724115143667415`, 0.5}, 1.7, {
                    1.5670948705468772`, 1.5724022341424941`}], 
                   ArrowBox[{{-1.858141724435817, 
                    2.1999401086789288`}, {-1.8751415556842221`, 
                    2.1999978079028235`}}]}, 
                  {RGBColor[0, 0, 1], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-1.8724115143667415`, 0.5}, 1.8, {
                    1.5670948705468772`, 1.577200936042802}], 
                   ArrowBox[{{-1.8659398249560688`, 
                    2.2999883658613385`}, {-1.8839397321998952`, 
                    2.299963083008535}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.003], Arrowheads[0.025], 
                   CircleBox[{-1.8724115143667415`, 0.5}, 1.5, {
                    1.5670948705468772`, 1.567087759703254}], 
                   ArrowBox[{{-1.881848601472325, 
                    1.9999703135018911`}, {-1.866848676480685, 
                    1.9999896849094174`}}]}, 
                  {RGBColor[0, 0, 1], ArrowBox[{{0, 2.8}, {0, 2.8}}]}},
                 Frame->True,
                 FrameTicks->Automatic,
                 ImageSize->800,
                 PlotRange->{{-3, 3}, {-1, 3}}], 564->
                GraphicsBox[{InsetBox[
                   TagBox[GridBox[{
                    {"\<\"t\"\>", "28.150000000000002`"},
                    {"\<\"driveIntent\"\>", "0"},
                    {"\<\"motorTorque\"\>", 
                    RowBox[{"-", "1.8061070298431832`"}]},
                    {"\<\"vw\"\>", 
                    RowBox[{"-", "0.020782616806039256`"}]},
                    {"\<\"controllerState\"\>", 
                    RowBox[{"\[LeftAssociation]", 
                    RowBox[{
                    
                    RowBox[{"\<\"lastError\"\>", "\[Rule]", 
                    "0.002560122056295336`"}], ",", 
                    RowBox[{"\<\"lastInt\"\>", "\[Rule]", 
                    RowBox[{"-", "0.01780627414636398`"}]}], ",", 
                    RowBox[{"\<\"initialized\"\>", "\[Rule]", "True"}]}], 
                    "\[RightAssociation]"}]}
                    },
                    AutoDelete->False,
                    
                    GridBoxItemSize->{
                    "Columns" -> {{Automatic}}, "Rows" -> {{Automatic}}}],
                    "Grid"], {-1, -0.5}], 
                  {GrayLevel[0], LineBox[{{-5, 0}, {5, 0}}]}, 
                  {GrayLevel[0.5], CircleBox[{-1.873454312096408, 0.5}, 0.5], 
                   DiskBox[{-1.873454312096408, 0.5}, 0.1]}, 
                  {GrayLevel[0.5], 
                   DiskBox[{-1.867941776042081, 1.9999898706145485`}, 0.1]}, 
                  {GrayLevel[0.5], Thickness[0.01], 
                   LineBox[{{-1.873454312096408, 0.5}, {-1.867941776042081, 
                    1.9999898706145485`}}]}, 
                  {RGBColor[0, 1, 0], 
                   DiskBox[{-1.8697792880601898`, 1.4999932470763655`}, 
                    0.05]}, 
                  {RGBColor[0, 0, 1], Arrowheads[0.01], 
                   CircleBox[{-1.873454312096408, 0.5}, 0.3333333333333333, 
                    NCache[{
                    Rational[-1, 2] 
                    Pi, -1.624979537690192}, {-1.5707963267948966`, \
-1.624979537690192}]], 
                   ArrowBox[{{-1.8881772577222151`, 
                    0.1669919730949258}, {-1.8915065463490588`, 
                    0.16715585035996428`}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-1.873454312096408, 0.5}, {-1.8373321714995443`,
                     0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-1.873454312096408, 0.5}, {-1.9004838854151147`,
                     0.5}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.005], Arrowheads[0.04], 
                   ArrowBox[{{-1.873454312096408, 0.5}, {-1.864361744818251, 
                    0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-1.873454312096408, 0.5}, 1.6, {
                    1.5671212944862705`, 1.5502276970809057`}], 
                   ArrowBox[{{-1.8565448193450558`, 
                    2.099910644084504}, {-1.8405468250222434`, 
                    2.0996615571096484`}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Dashing[{Small, Small}], 
                   Arrowheads[0.02], 
                   CircleBox[{-1.873454312096408, 0.5}, 1.7, {
                    1.5671212944862705`, 1.5723907842084552`}], 
                   ArrowBox[{{-1.8591650579643049`, 
                    2.1999399451793433`}, {-1.8761648885509434`, 
                    2.1999978390501806`}}]}, 
                  {RGBColor[0, 0, 1], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-1.873454312096408, 0.5}, 1.8, {
                    1.5671212944862705`, 1.5771552224298437`}], 
                   ArrowBox[{{-1.8669003387210397`, 
                    2.299988068136285}, {-1.884900247101829, 
                    2.299963608124301}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.003], Arrowheads[0.025], 
                   CircleBox[{-1.873454312096408, 0.5}, 1.5, {
                    1.5671212944862705`, 1.5671142591379785`}], 
                   ArrowBox[{{-1.8829311475640849`, 
                    1.9999700628977628`}, {-1.867931223091043, 
                    1.9999898317948155`}}]}, 
                  {RGBColor[0, 0, 1], ArrowBox[{{0, 2.8}, {0, 2.8}}]}},
                 Frame->True,
                 FrameTicks->Automatic,
                 ImageSize->800,
                 PlotRange->{{-3, 3}, {-1, 3}}], 565->
                GraphicsBox[{InsetBox[
                   TagBox[GridBox[{
                    {"\<\"t\"\>", "28.2`"},
                    {"\<\"driveIntent\"\>", "0"},
                    {"\<\"motorTorque\"\>", 
                    RowBox[{"-", "1.7932142925202492`"}]},
                    {"\<\"vw\"\>", 
                    RowBox[{"-", "0.020633834839522892`"}]},
                    {"\<\"controllerState\"\>", 
                    RowBox[{"\[LeftAssociation]", 
                    RowBox[{
                    
                    RowBox[{"\<\"lastError\"\>", "\[Rule]", 
                    "0.0025417211266324556`"}], ",", 
                    RowBox[{"\<\"lastInt\"\>", "\[Rule]", 
                    RowBox[{"-", "0.017678737821448038`"}]}], ",", 
                    RowBox[{"\<\"initialized\"\>", "\[Rule]", "True"}]}], 
                    "\[RightAssociation]"}]}
                    },
                    AutoDelete->False,
                    
                    GridBoxItemSize->{
                    "Columns" -> {{Automatic}}, "Rows" -> {{Automatic}}}],
                    "Grid"], {-1, -0.5}], 
                  {GrayLevel[0], LineBox[{{-5, 0}, {5, 0}}]}, 
                  {GrayLevel[0.5], 
                   CircleBox[{-1.8744896445581003`, 0.5}, 0.5], 
                   DiskBox[{-1.8744896445581003`, 0.5}, 0.1]}, 
                  {GrayLevel[0.5], 
                   DiskBox[{-1.8690164627219839`, 1.9999900147269611`}, 0.1]}, 
                  {GrayLevel[0.5], Thickness[0.01], 
                   LineBox[{{-1.8744896445581003`, 
                    0.5}, {-1.8690164627219839`, 1.9999900147269611`}}]}, 
                  {RGBColor[0, 1, 0], 
                   DiskBox[{-1.870840856667356, 1.4999933431513073`}, 0.05]}, 
                  {RGBColor[0, 0, 1], Arrowheads[0.01], 
                   CircleBox[{-1.8744896445581003`, 0.5}, 0.3333333333333333, 
                    NCache[{
                    Rational[-1, 2] 
                    Pi, -1.624592755570504}, {-1.5707963267948966`, \
-1.624592755570504}]], 
                   ArrowBox[{{-1.8890837875353201`, 
                    0.16698630343201865`}, {-1.8924131392979266`, 
                    0.167148892975519}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-1.8744896445581003`, 
                    0.5}, {-1.8386253587076953`, 0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-1.8744896445581003`, 
                    0.5}, {-1.9013262552362302`, 0.5}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.005], Arrowheads[0.04], 
                   ArrowBox[{{-1.8744896445581003`, 
                    0.5}, {-1.8654619693858252`, 0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-1.8744896445581003`, 0.5}, 1.6, {
                    1.5671475308076543`, 1.5503745374785445`}], 
                   ArrowBox[{{-1.8578150835033698`, 
                    2.099913109832416}, {-1.8418170527770024`, 
                    2.099666372012085}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Dashing[{Small, Small}], 
                   Arrowheads[0.02], 
                   CircleBox[{-1.8744896445581003`, 0.5}, 1.7, {
                    1.5671475308076543`, 1.57237941359191}], 
                   ArrowBox[{{-1.8601810610616547`, 
                    2.1999397825918203`}, {-1.8771808909889054`, 
                    2.199997869761209}}]}, 
                  {RGBColor[0, 0, 1], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-1.8744896445581003`, 0.5}, 1.8, {
                    1.5671475308076543`, 1.5771098324327668`}], 
                   ArrowBox[{{-1.86785396973636, 
                    2.299987768797238}, {-1.8858538792088473`, 
                    2.299964125801068}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.003], Arrowheads[0.025], 
                   CircleBox[{-1.8744896445581003`, 0.5}, 1.5, {
                    1.5671475308076543`, 1.5671405670288951`}], 
                   ArrowBox[{{-1.884005941071286, 
                    1.9999698130631407`}, {-1.869006017123513, 
                    1.9999899765765632`}}]}, 
                  {RGBColor[0, 0, 1], ArrowBox[{{0, 2.8}, {0, 2.8}}]}},
                 Frame->True,
                 FrameTicks->Automatic,
                 ImageSize->800,
                 PlotRange->{{-3, 3}, {-1, 3}}], 566->
                GraphicsBox[{InsetBox[
                   TagBox[GridBox[{
                    {"\<\"t\"\>", "28.25`"},
                    {"\<\"driveIntent\"\>", "0"},
                    {"\<\"motorTorque\"\>", 
                    RowBox[{"-", "1.780412649733754`"}]},
                    {"\<\"vw\"\>", 
                    RowBox[{"-", "0.020486114780308456`"}]},
                    {"\<\"controllerState\"\>", 
                    RowBox[{"\[LeftAssociation]", 
                    RowBox[{
                    
                    RowBox[{"\<\"lastError\"\>", "\[Rule]", 
                    "0.0025234530679031043`"}], ",", 
                    RowBox[{"\<\"lastInt\"\>", "\[Rule]", 
                    RowBox[{"-", "0.017552118153193415`"}]}], ",", 
                    RowBox[{"\<\"initialized\"\>", "\[Rule]", "True"}]}], 
                    "\[RightAssociation]"}]}
                    },
                    AutoDelete->False,
                    
                    GridBoxItemSize->{
                    "Columns" -> {{Automatic}}, "Rows" -> {{Automatic}}}],
                    "Grid"], {-1, -0.5}], 
                  {GrayLevel[0], LineBox[{{-5, 0}, {5, 0}}]}, 
                  {GrayLevel[0.5], 
                   CircleBox[{-1.8755175650313014`, 0.5}, 0.5], 
                   DiskBox[{-1.8755175650313014`, 0.5}, 0.1]}, 
                  {GrayLevel[0.5], 
                   DiskBox[{-1.870083458855466, 1.9999901567977272`}, 0.1]}, 
                  {GrayLevel[0.5], Thickness[0.01], 
                   LineBox[{{-1.8755175650313014`, 0.5}, {-1.870083458855466, 
                    1.9999901567977272`}}]}, 
                  {RGBColor[0, 1, 0], 
                   DiskBox[{-1.8718948275807445`, 1.4999934378651514`}, 
                    0.05]}, 
                  {RGBColor[0, 0, 1], Arrowheads[0.01], 
                   CircleBox[{-1.8755175650313014`, 0.5}, 0.3333333333333333, 
                    NCache[{
                    Rational[-1, 2] 
                    Pi, -1.6242087062869093`}, {-1.5707963267948966`, \
-1.6242087062869093`}]], 
                   ArrowBox[{{-1.8899838132637978`, 
                    0.16698072312073753`}, {-1.8933132272232711`, 
                    0.16714203401711958`}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-1.8755175650313014`, 
                    0.5}, {-1.8399093120366263`, 0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-1.8755175650313014`, 
                    0.5}, {-1.9021625788466086`, 0.5}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.005], Arrowheads[0.04], 
                   ArrowBox[{{-1.8755175650313014`, 
                    0.5}, {-1.8665543258519335`, 0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-1.8755175650313014`, 0.5}, 1.6, {
                    1.5671735814200216`, 1.550520338504629}], 
                   ArrowBox[{{-1.8590762731260373`, 
                    2.099915523995091}, {-1.8430782065950853`, 
                    2.0996711187066692`}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Dashing[{Small, Small}], 
                   Arrowheads[0.02], 
                   CircleBox[{-1.8755175650313014`, 0.5}, 1.7, {
                    1.5671735814200216`, 1.5723681215866614`}], 
                   ArrowBox[{{-1.8611897858068212`, 
                    2.199939620910841}, {-1.8781896150770674`, 
                    2.199997900042395}}]}, 
                  {RGBColor[0, 0, 1], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-1.8755175650313014`, 0.5}, 1.8, {
                    1.5671735814200216`, 1.5770647628074312`}], 
                   ArrowBox[{{-1.8688007654419798`, 
                    2.2999874679017287`}, {-1.8868006759617666`, 
                    2.2999646361547583`}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.003], Arrowheads[0.025], 
                   CircleBox[{-1.8755175650313014`, 0.5}, 1.5, {
                    1.5671735814200216`, 1.567166685497353}], 
                   ArrowBox[{{-1.8850730384554868`, 
                    1.9999695640004298`}, {-1.8700731150394703`, 
                    1.999990119288886}}]}, 
                  {RGBColor[0, 0, 1], ArrowBox[{{0, 2.8}, {0, 2.8}}]}},
                 Frame->True,
                 FrameTicks->Automatic,
                 ImageSize->800,
                 PlotRange->{{-3, 3}, {-1, 3}}], 567->
                GraphicsBox[{InsetBox[
                   TagBox[GridBox[{
                    {"\<\"t\"\>", "28.3`"},
                    {"\<\"driveIntent\"\>", "0"},
                    {"\<\"motorTorque\"\>", 
                    RowBox[{"-", "1.7677012229925024`"}]},
                    {"\<\"vw\"\>", 
                    RowBox[{"-", "0.020339449172801043`"}]},
                    {"\<\"controllerState\"\>", 
                    RowBox[{"\[LeftAssociation]", 
                    RowBox[{
                    
                    RowBox[{"\<\"lastError\"\>", "\[Rule]", 
                    "0.0025053174549174925`"}], ",", 
                    RowBox[{"\<\"lastInt\"\>", "\[Rule]", 
                    RowBox[{"-", "0.01742640850858259`"}]}], ",", 
                    RowBox[{"\<\"initialized\"\>", "\[Rule]", "True"}]}], 
                    "\[RightAssociation]"}]}
                    },
                    AutoDelete->False,
                    
                    GridBoxItemSize->{
                    "Columns" -> {{Automatic}}, "Rows" -> {{Automatic}}}],
                    "Grid"], {-1, -0.5}], 
                  {GrayLevel[0], LineBox[{{-5, 0}, {5, 0}}]}, 
                  {GrayLevel[0.5], 
                   CircleBox[{-1.8765381264209688`, 0.5}, 0.5], 
                   DiskBox[{-1.8765381264209688`, 0.5}, 0.1]}, 
                  {GrayLevel[0.5], 
                   DiskBox[{-1.8711428200623403`, 1.9999902968583818`}, 0.1]}, 
                  {GrayLevel[0.5], Thickness[0.01], 
                   LineBox[{{-1.8765381264209688`, 
                    0.5}, {-1.8711428200623403`, 1.9999902968583818`}}]}, 
                  {RGBColor[0, 1, 0], 
                   DiskBox[{-1.8729412555152165`, 1.4999935312389212`}, 
                    0.05]}, 
                  {RGBColor[0, 0, 1], Arrowheads[0.01], 
                   CircleBox[{-1.8765381264209688`, 0.5}, 0.3333333333333333, 
                    NCache[{
                    Rational[-1, 2] 
                    Pi, -1.6238273634846716`}, {-1.5707963267948966`, \
-1.6238273634846716`}]], 
                   ArrowBox[{{-1.8908773791004412`, 
                    0.16697523073545212`}, {-1.8942068543325767`, 
                    0.16713527197208672`}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-1.8765381264209688`, 
                    0.5}, {-1.8411841019611188`, 0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-1.8765381264209688`, 
                    0.5}, {-1.9029928958412017`, 0.5}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.005], Arrowheads[0.04], 
                   ArrowBox[{{-1.8765381264209688`, 
                    0.5}, {-1.8676388713813517`, 0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-1.8765381264209688`, 0.5}, 1.6, {
                    1.567199448133358, 1.5506651102890312`}], 
                   ArrowBox[{{-1.8603284573124927`, 
                    2.099917887464039}, {-1.8443303555662018`, 
                    2.0996757982468095`}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Dashing[{Small, Small}], 
                   Arrowheads[0.02], 
                   CircleBox[{-1.8765381264209688`, 0.5}, 1.7, {
                    1.567199448133358, 1.5723569075343105`}], 
                   ArrowBox[{{-1.8621912839854875`, 
                    2.1999394601314872`}, {-1.8791911126011196`, 
                    2.1999979299000127`}}]}, 
                  {RGBColor[0, 0, 1], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-1.8765381264209688`, 0.5}, 1.8, {
                    1.567199448133358, 1.5770200104833163`}], 
                   ArrowBox[{{-1.869740773215834, 
                    2.299987165506856}, {-1.8877406847393696`, 
                    2.299965139297737}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.003], Arrowheads[0.025], 
                   CircleBox[{-1.8765381264209688`, 0.5}, 1.5, {
                    1.567199448133358, 1.5671926165737724`}], 
                   ArrowBox[{{-1.8861324956673282`, 
                    1.999969315712413}, {-1.8711325727893755`, 
                    1.9999902599650226`}}]}, 
                  {RGBColor[0, 0, 1], ArrowBox[{{0, 2.8}, {0, 2.8}}]}},
                 Frame->True,
                 FrameTicks->Automatic,
                 ImageSize->800,
                 PlotRange->{{-3, 3}, {-1, 3}}], 568->
                GraphicsBox[{InsetBox[
                   TagBox[GridBox[{
                    {"\<\"t\"\>", "28.35`"},
                    {"\<\"driveIntent\"\>", "0"},
                    {"\<\"motorTorque\"\>", 
                    RowBox[{"-", "1.7550791798397372`"}]},
                    {"\<\"vw\"\>", 
                    RowBox[{"-", "0.02019383062944715`"}]},
                    {"\<\"controllerState\"\>", 
                    RowBox[{"\[LeftAssociation]", 
                    RowBox[{
                    
                    RowBox[{"\<\"lastError\"\>", "\[Rule]", 
                    "0.002487313789666867`"}], ",", 
                    RowBox[{"\<\"lastInt\"\>", "\[Rule]", 
                    RowBox[{"-", "0.01730160227773431`"}]}], ",", 
                    RowBox[{"\<\"initialized\"\>", "\[Rule]", "True"}]}], 
                    "\[RightAssociation]"}]}
                    },
                    AutoDelete->False,
                    
                    GridBoxItemSize->{
                    "Columns" -> {{Automatic}}, "Rows" -> {{Automatic}}}],
                    "Grid"], {-1, -0.5}], 
                  {GrayLevel[0], LineBox[{{-5, 0}, {5, 0}}]}, 
                  {GrayLevel[0.5], 
                   CircleBox[{-1.8775513812610296`, 0.5}, 0.5], 
                   DiskBox[{-1.8775513812610296`, 0.5}, 0.1]}, 
                  {GrayLevel[0.5], 
                   DiskBox[{-1.872194601451689, 1.9999904349395279`}, 0.1]}, 
                  {GrayLevel[0.5], Thickness[0.01], 
                   LineBox[{{-1.8775513812610296`, 0.5}, {-1.872194601451689, 
                    1.9999904349395279`}}]}, 
                  {RGBColor[0, 1, 0], 
                   DiskBox[{-1.8739801947214694`, 1.4999936232930187`}, 
                    0.05]}, 
                  {RGBColor[0, 0, 1], Arrowheads[0.01], 
                   CircleBox[{-1.8775513812610296`, 0.5}, 0.3333333333333333, 
                    NCache[{
                    Rational[-1, 2] 
                    Pi, -1.6234487021900887`}, {-1.5707963267948966`, \
-1.6234487021900887`}]], 
                   ArrowBox[{{-1.8917645293252452`, 
                    0.16696982489087303`}, {-1.8950940649201038`, 
                    0.1671286053726624}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-1.8775513812610296`, 0.5}, {-1.842449797664235,
                     0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-1.8775513812610296`, 0.5}, {-1.903817246128872,
                     0.5}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.005], Arrowheads[0.04], 
                   ArrowBox[{{-1.8775513812610296`, 
                    0.5}, {-1.8687156625320773`, 0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-1.8775513812610296`, 0.5}, 1.6, {
                    1.5672251326645135`, 1.550808862440367}], 
                   ArrowBox[{{-1.861571703957529, 
                    2.099920201107942}, {-1.8455735675756901`, 
                    2.0996804116547487`}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Dashing[{Small, Small}], 
                   Arrowheads[0.02], 
                   CircleBox[{-1.8775513812610296`, 0.5}, 1.7, {
                    1.5672251326645135`, 1.572345770819938}], 
                   ArrowBox[{{-1.8631856070862203`, 
                    2.199939300249382}, {-1.8801854350496372`, 
                    2.1999979593401395`}}]}, 
                  {RGBColor[0, 0, 1], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-1.8775513812610296`, 0.5}, 1.8, {
                    1.5672251326645135`, 1.576975572552512}], 
                   ArrowBox[{{-1.870674040357523, 
                    2.2999868616693004`}, {-1.8886739528420855`, 
                    2.299965635339026}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.003], Arrowheads[0.025], 
                   CircleBox[{-1.8775513812610296`, 0.5}, 1.5, {
                    1.5672251326645135`, 1.5672183622005034`}], 
                   ArrowBox[{{-1.8871843681543152`, 
                    1.9999690682022457`}, {-1.8721844458205565`, 
                    1.9999903986372638`}}]}, 
                  {RGBColor[0, 0, 1], ArrowBox[{{0, 2.8}, {0, 2.8}}]}},
                 Frame->True,
                 FrameTicks->Automatic,
                 ImageSize->800,
                 PlotRange->{{-3, 3}, {-1, 3}}], 569->
                GraphicsBox[{InsetBox[
                   TagBox[GridBox[{
                    {"\<\"t\"\>", "28.400000000000002`"},
                    {"\<\"driveIntent\"\>", "0"},
                    {"\<\"motorTorque\"\>", 
                    RowBox[{"-", "1.74254573076261`"}]},
                    {"\<\"vw\"\>", 
                    RowBox[{"-", "0.020049251827004972`"}]},
                    {"\<\"controllerState\"\>", 
                    RowBox[{"\[LeftAssociation]", 
                    RowBox[{
                    
                    RowBox[{"\<\"lastError\"\>", "\[Rule]", 
                    "0.002469441506208071`"}], ",", 
                    RowBox[{"\<\"lastInt\"\>", "\[Rule]", 
                    RowBox[{"-", "0.017177692877420842`"}]}], ",", 
                    RowBox[{"\<\"initialized\"\>", "\[Rule]", "True"}]}], 
                    "\[RightAssociation]"}]}
                    },
                    AutoDelete->False,
                    
                    GridBoxItemSize->{
                    "Columns" -> {{Automatic}}, "Rows" -> {{Automatic}}}],
                    "Grid"], {-1, -0.5}], 
                  {GrayLevel[0], LineBox[{{-5, 0}, {5, 0}}]}, 
                  {GrayLevel[0.5], 
                   CircleBox[{-1.8785573817176877`, 0.5}, 0.5], 
                   DiskBox[{-1.8785573817176877`, 0.5}, 0.1]}, 
                  {GrayLevel[0.5], 
                   DiskBox[{-1.873238857634183, 1.9999905710708896`}, 0.1]}, 
                  {GrayLevel[0.5], Thickness[0.01], 
                   LineBox[{{-1.8785573817176877`, 0.5}, {-1.873238857634183, 
                    1.9999905710708896`}}]}, 
                  {RGBColor[0, 1, 0], 
                   DiskBox[{-1.8750116989953511`, 1.4999937140472597`}, 
                    0.05]}, 
                  {RGBColor[0, 0, 1], Arrowheads[0.01], 
                   CircleBox[{-1.8785573817176877`, 0.5}, 0.3333333333333333, 
                    NCache[{
                    Rational[-1, 2] 
                    Pi, -1.6230726987177748`}, {-1.5707963267948966`, \
-1.6230726987177748`}]], 
                   ArrowBox[{{-1.89264530827791, 
                    0.16696450423964582`}, {-1.895974903339417, 
                    0.16712203279329574`}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-1.8785573817176877`, 
                    0.5}, {-1.8437064671024355`, 0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-1.8785573817176877`, 
                    0.5}, {-1.9046356698915214`, 0.5}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.005], Arrowheads[0.04], 
                   ArrowBox[{{-1.8785573817176877`, 
                    0.5}, {-1.8697847552762692`, 0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-1.8785573817176877`, 0.5}, 1.6, {
                    1.56725063664321, 1.5509516040796154`}], 
                   ArrowBox[{{-1.8628060798083785`, 
                    2.0999224657739393`}, {-1.8468079093615826`, 
                    2.0996849599233935`}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Dashing[{Small, Small}], 
                   Arrowheads[0.02], 
                   CircleBox[{-1.8785573817176877`, 0.5}, 1.7, {
                    1.56725063664321, 1.5723347108678942`}], 
                   ArrowBox[{{-1.8641728062966194`, 
                    2.199939141260638}, {-1.8811726336102297`, 
                    2.199997988368674}}]}, 
                  {RGBColor[0, 0, 1], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-1.8785573817176877`, 0.5}, 1.8, {
                    1.56725063664321, 1.5769314462585577`}], 
                   ArrowBox[{{-1.8716006140715307`, 
                    2.2999865564453303`}, {-1.8896005274752081`, 
                    2.2999661243845058`}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.003], Arrowheads[0.025], 
                   CircleBox[{-1.8785573817176877`, 0.5}, 1.5, {
                    1.56725063664321, 1.567243924235102}], 
                   ArrowBox[{{-1.88822871086904, 
                    1.9999688214734486`}, {-1.8732287890854316`, 
                    1.9999905353369931`}}]}, 
                  {RGBColor[0, 0, 1], ArrowBox[{{0, 2.8}, {0, 2.8}}]}},
                 Frame->True,
                 FrameTicks->Automatic,
                 ImageSize->800,
                 PlotRange->{{-3, 3}, {-1, 3}}], 570->
                GraphicsBox[{InsetBox[
                   TagBox[GridBox[{
                    {"\<\"t\"\>", "28.45`"},
                    {"\<\"driveIntent\"\>", "0"},
                    {"\<\"motorTorque\"\>", 
                    RowBox[{"-", "1.7301001260866458`"}]},
                    {"\<\"vw\"\>", 
                    RowBox[{"-", "0.01990570550310345`"}]},
                    {"\<\"controllerState\"\>", 
                    RowBox[{"\[LeftAssociation]", 
                    RowBox[{
                    
                    RowBox[{"\<\"lastError\"\>", "\[Rule]", 
                    "0.0024516999756992812`"}], ",", 
                    RowBox[{"\<\"lastInt\"\>", "\[Rule]", 
                    RowBox[{"-", "0.017054673754336728`"}]}], ",", 
                    RowBox[{"\<\"initialized\"\>", "\[Rule]", "True"}]}], 
                    "\[RightAssociation]"}]}
                    },
                    AutoDelete->False,
                    
                    GridBoxItemSize->{
                    "Columns" -> {{Automatic}}, "Rows" -> {{Automatic}}}],
                    "Grid"], {-1, -0.5}], 
                  {GrayLevel[0], LineBox[{{-5, 0}, {5, 0}}]}, 
                  {GrayLevel[0.5], 
                   CircleBox[{-1.8795561795925526`, 0.5}, 0.5], 
                   DiskBox[{-1.8795561795925526`, 0.5}, 0.1]}, 
                  {GrayLevel[0.5], 
                   DiskBox[{-1.8742756427343135`, 1.9999907052813657`}, 0.1]}, 
                  {GrayLevel[0.5], Thickness[0.01], 
                   LineBox[{{-1.8795561795925526`, 
                    0.5}, {-1.8742756427343135`, 1.9999907052813657`}}]}, 
                  {RGBColor[0, 1, 0], 
                   DiskBox[{-1.8760358216870598`, 1.4999938035209106`}, 
                    0.05]}, 
                  {RGBColor[0, 0, 1], Arrowheads[0.01], 
                   CircleBox[{-1.8795561795925526`, 0.5}, 0.3333333333333333, 
                    NCache[{
                    Rational[-1, 2] 
                    Pi, -1.622699330577496}, {-1.5707963267948966`, \
-1.622699330577496}]], 
                   ArrowBox[{{-1.89351976033001, 
                    0.1669592674700291}, {-1.8968494139755798`, 
                    0.16711555284801183`}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-1.8795561795925526`, 
                    0.5}, {-1.8449541770708198`, 0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-1.8795561795925526`, 
                    0.5}, {-1.9054482075425745`, 0.5}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.005], Arrowheads[0.04], 
                   ArrowBox[{{-1.8795561795925526`, 
                    0.5}, {-1.8708462050208416`, 0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-1.8795561795925526`, 0.5}, 1.6, {
                    1.567275961618111, 1.5510933438740948`}], 
                   ArrowBox[{{-1.8640316505221892`, 
                    2.0999246822888704`}, {-1.8480334465720658`, 
                    2.099689444018094}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Dashing[{Small, Small}], 
                   Arrowheads[0.02], 
                   CircleBox[{-1.8795561795925526`, 0.5}, 1.7, {
                    1.567275961618111, 1.5723237271377462`}], 
                   ArrowBox[{{-1.8651529324995566`, 
                    2.1999389831618013`}, {-1.8821527591657807`, 
                    2.199998016991349}}]}, 
                  {RGBColor[0, 0, 1], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-1.8795561795925526`, 0.5}, 1.8, {
                    1.567275961618111, 1.5768876289852591`}], 
                   ArrowBox[{{-1.8725205414502215`, 
                    2.2999862498907957`}, {-1.8905204557318966`, 
                    2.299966606537116}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.003], Arrowheads[0.025], 
                   CircleBox[{-1.8795561795925526`, 0.5}, 1.5, {
                    1.567275961618111, 1.5672693044539914`}], 
                   ArrowBox[{{-1.8892655782778025`, 
                    1.999968575529891}, {-1.8742656570501275`, 
                    1.999990670094727}}]}, 
                  {RGBColor[0, 0, 1], ArrowBox[{{0, 2.8}, {0, 2.8}}]}},
                 Frame->True,
                 FrameTicks->Automatic,
                 ImageSize->800,
                 PlotRange->{{-3, 3}, {-1, 3}}], 571->
                GraphicsBox[{InsetBox[
                   TagBox[GridBox[{
                    {"\<\"t\"\>", "28.5`"},
                    {"\<\"driveIntent\"\>", "0"},
                    {"\<\"motorTorque\"\>", 
                    RowBox[{"-", "1.7177416528834342`"}]},
                    {"\<\"vw\"\>", 
                    RowBox[{"-", "0.019763184453086054`"}]},
                    {"\<\"controllerState\"\>", 
                    RowBox[{"\[LeftAssociation]", 
                    RowBox[{
                    
                    RowBox[{"\<\"lastError\"\>", "\[Rule]", 
                    "0.002434088511520776`"}], ",", 
                    RowBox[{"\<\"lastInt\"\>", "\[Rule]", 
                    RowBox[{"-", "0.016932538388113327`"}]}], ",", 
                    RowBox[{"\<\"initialized\"\>", "\[Rule]", "True"}]}], 
                    "\[RightAssociation]"}]}
                    },
                    AutoDelete->False,
                    GridBoxItemSize->{
                    "Columns" -> {{Automatic}}, "Rows" -> {{Automatic}}}],
                    "Grid"], {-1, -0.5}], 
                  {GrayLevel[0], LineBox[{{-5, 0}, {5, 0}}]}, 
                  {GrayLevel[0.5], CircleBox[{-1.880547826325597, 0.5}, 0.5], 
                   DiskBox[{-1.880547826325597, 0.5}, 0.1]}, 
                  {GrayLevel[0.5], 
                   DiskBox[{-1.8753050104024547`, 1.9999908375990822`}, 0.1]}, 
                  {GrayLevel[0.5], Thickness[0.01], 
                   LineBox[{{-1.880547826325597, 0.5}, {-1.8753050104024547`, 
                    1.9999908375990822`}}]}, 
                  {RGBColor[0, 1, 0], 
                   DiskBox[{-1.8770526157101688`, 1.4999938917327214`}, 
                    0.05]}, 
                  {RGBColor[0, 0, 1], Arrowheads[0.01], 
                   CircleBox[{-1.880547826325597, 0.5}, 0.3333333333333333, 
                    NCache[{
                    Rational[-1, 2] 
                    Pi, -1.6223285763813995`}, {-1.5707963267948966`, \
-1.6223285763813995`}]], 
                   ArrowBox[{{-1.8943879298571191`, 
                    0.16695411330366527`}, {-1.8977176412173022`, 
                    0.16710916418787425`}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-1.880547826325597, 0.5}, {-1.8461929932679284`,
                     0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-1.880547826325597, 0.5}, {-1.906254899685305, 
                    0.5}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.005], Arrowheads[0.04], 
                   ArrowBox[{{-1.880547826325597, 0.5}, {-1.8719000666276364`,
                     0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-1.880547826325597, 0.5}, 1.6, {
                    1.5673011090628908`, 1.5512340900714268`}], 
                   ArrowBox[{{-1.8652484807233063`, 
                    2.099926851460448}, {-1.8492502438227492`, 
                    2.099693864878364}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Dashing[{Small, Small}], 
                   Arrowheads[0.02], 
                   CircleBox[{-1.880547826325597, 0.5}, 1.7, {
                    1.5673011090628908`, 1.572312819120372}], 
                   ArrowBox[{{-1.8661260362694947`, 
                    2.199938825949798}, {-1.8831258622907654`, 
                    2.1999980452137473`}}]}, 
                  {RGBColor[0, 0, 1], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-1.880547826325597, 0.5}, 1.8, {
                    1.5673011090628908`, 1.5768441182455764`}], 
                   ArrowBox[{{-1.873433869456799, 
                    2.2999859420611233`}, {-1.8914337845761324`, 
                    2.299967081897046}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.003], Arrowheads[0.025], 
                   CircleBox[{-1.880547826325597, 0.5}, 1.5, {
                    1.5673011090628908`, 1.567294504556364}], 
                   ArrowBox[{{-1.890295024369419, 
                    1.9999683303757765`}, {-1.875295103703292, 
                    1.9999908029401559`}}]}, 
                  {RGBColor[0, 0, 1], ArrowBox[{{0, 2.8}, {0, 2.8}}]}},
                 Frame->True,
                 FrameTicks->Automatic,
                 ImageSize->800,
                 PlotRange->{{-3, 3}, {-1, 3}}], 572->
                GraphicsBox[{InsetBox[
                   TagBox[GridBox[{
                    {"\<\"t\"\>", "28.55`"},
                    {"\<\"driveIntent\"\>", "0"},
                    {"\<\"motorTorque\"\>", 
                    RowBox[{"-", "1.7054696319217093`"}]},
                    {"\<\"vw\"\>", 
                    RowBox[{"-", "0.019621681527134054`"}]},
                    {"\<\"controllerState\"\>", 
                    RowBox[{"\[LeftAssociation]", 
                    RowBox[{
                    
                    RowBox[{"\<\"lastError\"\>", "\[Rule]", 
                    "0.0024166063744216113`"}], ",", 
                    RowBox[{"\<\"lastInt\"\>", "\[Rule]", 
                    RowBox[{"-", "0.01681128029407651`"}]}], ",", 
                    RowBox[{"\<\"initialized\"\>", "\[Rule]", "True"}]}], 
                    "\[RightAssociation]"}]}
                    },
                    AutoDelete->False,
                    
                    GridBoxItemSize->{
                    "Columns" -> {{Automatic}}, "Rows" -> {{Automatic}}}],
                    "Grid"], {-1, -0.5}], 
                  {GrayLevel[0], LineBox[{{-5, 0}, {5, 0}}]}, 
                  {GrayLevel[0.5], CircleBox[{-1.881532372997972, 0.5}, 0.5], 
                   DiskBox[{-1.881532372997972, 0.5}, 0.1]}, 
                  {GrayLevel[0.5], 
                   DiskBox[{-1.8763270138266923`, 1.9999909680514407`}, 0.1]}, 
                  {GrayLevel[0.5], Thickness[0.01], 
                   LineBox[{{-1.881532372997972, 0.5}, {-1.8763270138266923`, 
                    1.9999909680514407`}}]}, 
                  {RGBColor[0, 1, 0], 
                   DiskBox[{-1.8780621335504522`, 1.4999939787009604`}, 
                    0.05]}, 
                  {RGBColor[0, 0, 1], Arrowheads[0.01], 
                   CircleBox[{-1.881532372997972, 0.5}, 0.3333333333333333, 
                    NCache[{
                    Rational[-1, 2] 
                    Pi, -1.6219604157525478`}, {-1.5707963267948966`, \
-1.6219604157525478`}]], 
                   ArrowBox[{{-1.895249861211222, 
                    0.16694904049345471`}, {-1.8985796294293766`, 
                    0.1671028654985552}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-1.881532372997972, 0.5}, {-1.847422980359538, 
                    0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-1.881532372997972, 0.5}, {-1.9070557870714533`,
                     0.5}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.005], Arrowheads[0.04], 
                   ArrowBox[{{-1.881532372997972, 0.5}, {-1.872946394433019, 
                    0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-1.881532372997972, 0.5}, 1.6, {
                    1.5673260803822435`, 1.5513738505331716`}], 
                   ArrowBox[{{-1.8664566340598923`, 
                    2.099928974078372}, {-1.8504583647532697`, 
                    2.0996982234195327`}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Dashing[{Small, Small}], 
                   Arrowheads[0.02], 
                   CircleBox[{-1.881532372997972, 0.5}, 1.7, {
                    1.5673260803822435`, 1.5723019863342442`}], 
                   ArrowBox[{{-1.8670921678689847`, 
                    2.199938669621887}, {-1.8840919932477482`, 
                    2.1999980730413125`}}]}, 
                  {RGBColor[0, 0, 1], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-1.881532372997972, 0.5}, 1.8, {
                    1.5673260803822435`, 1.5768009116706974`}], 
                   ArrowBox[{{-1.8743406449084514`, 
                    2.299985633011299}, {-1.892340560825867, 
                    2.2999675505619197`}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.003], Arrowheads[0.025], 
                   CircleBox[{-1.881532372997972, 0.5}, 1.5, {
                    1.5673260803822435`, 1.5673195261683124`}], 
                   ArrowBox[{{-1.8913171026642364`, 
                    1.9999680860156186`}, {-1.8763171825651048`, 
                    1.9999909339021849`}}]}, 
                  {RGBColor[0, 0, 1], ArrowBox[{{0, 2.8}, {0, 2.8}}]}},
                 Frame->True,
                 FrameTicks->Automatic,
                 ImageSize->800,
                 PlotRange->{{-3, 3}, {-1, 3}}], 573->
                GraphicsBox[{InsetBox[
                   TagBox[GridBox[{
                    {"\<\"t\"\>", "28.6`"},
                    {"\<\"driveIntent\"\>", "0"},
                    {"\<\"motorTorque\"\>", 
                    RowBox[{"-", "1.6932834146851627`"}]},
                    {"\<\"vw\"\>", 
                    RowBox[{"-", "0.019481189627661138`"}]},
                    {"\<\"controllerState\"\>", 
                    RowBox[{"\[LeftAssociation]", 
                    RowBox[{
                    
                    RowBox[{"\<\"lastError\"\>", "\[Rule]", 
                    "0.002399252777639433`"}], ",", 
                    RowBox[{"\<\"lastInt\"\>", "\[Rule]", 
                    RowBox[{"-", "0.01669089302574745`"}]}], ",", 
                    RowBox[{"\<\"initialized\"\>", "\[Rule]", "True"}]}], 
                    "\[RightAssociation]"}]}
                    },
                    AutoDelete->False,
                    
                    GridBoxItemSize->{
                    "Columns" -> {{Automatic}}, "Rows" -> {{Automatic}}}],
                    "Grid"], {-1, -0.5}], 
                  {GrayLevel[0], LineBox[{{-5, 0}, {5, 0}}]}, 
                  {GrayLevel[0.5], 
                   CircleBox[{-1.8825098703346808`, 0.5}, 0.5], 
                   DiskBox[{-1.8825098703346808`, 0.5}, 0.1]}, 
                  {GrayLevel[0.5], 
                   DiskBox[{-1.877341705744348, 1.999991096665166}, 0.1]}, 
                  {GrayLevel[0.5], Thickness[0.01], 
                   LineBox[{{-1.8825098703346808`, 0.5}, {-1.877341705744348, 
                    1.999991096665166}}]}, 
                  {RGBColor[0, 1, 0], 
                   DiskBox[{-1.879064427274459, 1.499994064443444}, 0.05]}, 
                  {RGBColor[0, 0, 1], Arrowheads[0.01], 
                   CircleBox[{-1.8825098703346808`, 0.5}, 0.3333333333333333, 
                    NCache[{
                    Rational[-1, 2] 
                    Pi, -1.6215948292354514`}, {-1.5707963267948966`, \
-1.6215948292354514`}]], 
                   ArrowBox[{{-1.8961055986936461`, 
                    0.1669440478215405}, {-1.8994354229256298`, 
                    0.16709665549802277`}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-1.8825098703346808`, 
                    0.5}, {-1.8486442020409775`, 0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-1.8825098703346808`, 
                    0.5}, {-1.9078509105605037`, 0.5}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.005], Arrowheads[0.04], 
                   ArrowBox[{{-1.8825098703346808`, 
                    0.5}, {-1.8739852422668004`, 0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-1.8825098703346808`, 0.5}, 1.6, {
                    1.5673508769177844`, 1.5515126327678512`}], 
                   ArrowBox[{{-1.8676561732594228`, 
                    2.0999310509153815`}, {-1.8516578720827757`, 
                    2.0997025205343225`}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Dashing[{Small, Small}], 
                   Arrowheads[0.02], 
                   CircleBox[{-1.8825098703346808`, 0.5}, 1.7, {
                    1.5673508769177844`, 1.572291228321916}], 
                   ArrowBox[{{-1.8680513772453642`, 
                    2.199938514175612}, {-1.8850512019840815`, 
                    2.1999981004793643`}}]}, 
                  {RGBColor[0, 0, 1], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-1.8825098703346808`, 0.5}, 1.8, {
                    1.5673508769177844`, 1.5767580069993687`}], 
                   ArrowBox[{{-1.8752409144598225`, 
                    2.2999853227958527`}, {-1.8932408311364923`, 
                    2.299968012626966}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.003], Arrowheads[0.025], 
                   CircleBox[{-1.8825098703346808`, 0.5}, 1.5, {
                    1.5673508769177844`, 1.5673443708471055`}], 
                   ArrowBox[{{-1.8923318662232185`, 
                    1.9999678424542193`}, {-1.8773319466963647`, 
                    1.9999910630089752`}}]}, 
                  {RGBColor[0, 0, 1], ArrowBox[{{0, 2.8}, {0, 2.8}}]}},
                 Frame->True,
                 FrameTicks->Automatic,
                 ImageSize->800,
                 PlotRange->{{-3, 3}, {-1, 3}}], 574->
                GraphicsBox[{InsetBox[
                   TagBox[GridBox[{
                    {"\<\"t\"\>", "28.650000000000002`"},
                    {"\<\"driveIntent\"\>", "0"},
                    {"\<\"motorTorque\"\>", 
                    RowBox[{"-", "1.6811823804769777`"}]},
                    {"\<\"vw\"\>", 
                    RowBox[{"-", "0.01934170170697012`"}]},
                    {"\<\"controllerState\"\>", 
                    RowBox[{"\[LeftAssociation]", 
                    RowBox[{
                    
                    RowBox[{"\<\"lastError\"\>", "\[Rule]", 
                    "0.002382026891947034`"}], ",", 
                    RowBox[{"\<\"lastInt\"\>", "\[Rule]", 
                    RowBox[{"-", "0.0165713701770891`"}]}], ",", 
                    RowBox[{"\<\"initialized\"\>", "\[Rule]", "True"}]}], 
                    "\[RightAssociation]"}]}
                    },
                    AutoDelete->False,
                    
                    GridBoxItemSize->{
                    "Columns" -> {{Automatic}}, "Rows" -> {{Automatic}}}],
                    "Grid"], {-1, -0.5}], 
                  {GrayLevel[0], LineBox[{{-5, 0}, {5, 0}}]}, 
                  {GrayLevel[0.5], 
                   CircleBox[{-1.8834803687071273`, 0.5}, 0.5], 
                   DiskBox[{-1.8834803687071273`, 0.5}, 0.1]}, 
                  {GrayLevel[0.5], 
                   DiskBox[{-1.8783491384531494`, 1.999991223466351}, 0.1]}, 
                  {GrayLevel[0.5], Thickness[0.01], 
                   LineBox[{{-1.8834803687071273`, 
                    0.5}, {-1.8783491384531494`, 1.999991223466351}}]}, 
                  {RGBColor[0, 1, 0], 
                   DiskBox[{-1.8800595485378087`, 1.4999941489775672`}, 
                    0.05]}, 
                  {RGBColor[0, 0, 1], Arrowheads[0.01], 
                   CircleBox[{-1.8834803687071273`, 0.5}, 0.3333333333333333, 
                    NCache[{
                    Rational[-1, 2] 
                    Pi, -1.621231798209206}, {-1.5707963267948966`, \
-1.621231798209206}]], 
                   ArrowBox[{{-1.8969551865287313`, 
                    0.16693913409740602`}, {-1.9002850659426138`, 
                    0.16709053293435056`}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-1.8834803687071273`, 
                    0.5}, {-1.8498567210975878`, 0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-1.8834803687071273`, 0.5}, {-1.908640311079965,
                     0.5}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.005], Arrowheads[0.04], 
                   ArrowBox[{{-1.8834803687071273`, 
                    0.5}, {-1.8750166634704255`, 0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-1.8834803687071273`, 0.5}, 1.6, {
                    1.5673754999537972`, 1.5516504439631182`}], 
                   ArrowBox[{{-1.868847160182693, 
                    2.0999330827282376`}, {-1.8528488276639203`, 
                    2.0997067570943493`}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Dashing[{Small, Small}], 
                   Arrowheads[0.02], 
                   CircleBox[{-1.8834803687071273`, 0.5}, 1.7, {
                    1.5673754999537972`, 1.5722805446467032`}], 
                   ArrowBox[{{-1.8690037140276692`, 
                    2.199938359608751}, {-1.8860035381288156`, 
                    2.1999981275331066`}}]}, 
                  {RGBColor[0, 0, 1], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-1.8834803687071273`, 0.5}, 1.8, {
                    1.5673754999537972`, 1.5767154020675582`}], 
                   ArrowBox[{{-1.876134724586952, 
                    2.2999850114688343`}, {-1.8941346419847838`, 
                    2.299968468185186}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.003], Arrowheads[0.025], 
                   CircleBox[{-1.8834803687071273`, 0.5}, 1.5, {
                    1.5673754999537972`, 1.5673690400855464`}], 
                   ArrowBox[{{-1.8933393676570316`, 
                    1.999967599696642}, {-1.8783394487075757`, 
                    1.9999911902879823`}}]}, 
                  {RGBColor[0, 0, 1], ArrowBox[{{0, 2.8}, {0, 2.8}}]}},
                 Frame->True,
                 FrameTicks->Automatic,
                 ImageSize->800,
                 PlotRange->{{-3, 3}, {-1, 3}}], 575->
                GraphicsBox[{InsetBox[
                   TagBox[GridBox[{
                    {"\<\"t\"\>", "28.7`"},
                    {"\<\"driveIntent\"\>", "0"},
                    {"\<\"motorTorque\"\>", 
                    RowBox[{"-", "1.6691659336292157`"}]},
                    {"\<\"vw\"\>", 
                    RowBox[{"-", "0.019203210765160954`"}]},
                    {"\<\"controllerState\"\>", 
                    RowBox[{"\[LeftAssociation]", 
                    RowBox[{
                    
                    RowBox[{"\<\"lastError\"\>", "\[Rule]", 
                    "0.0023649278505853`"}], ",", 
                    RowBox[{"\<\"lastInt\"\>", "\[Rule]", 
                    RowBox[{"-", "0.016452705384503094`"}]}], ",", 
                    RowBox[{"\<\"initialized\"\>", "\[Rule]", "True"}]}], 
                    "\[RightAssociation]"}]}
                    },
                    AutoDelete->False,
                    
                    GridBoxItemSize->{
                    "Columns" -> {{Automatic}}, "Rows" -> {{Automatic}}}],
                    "Grid"], {-1, -0.5}], 
                  {GrayLevel[0], LineBox[{{-5, 0}, {5, 0}}]}, 
                  {GrayLevel[0.5], 
                   CircleBox[{-1.8844439181355583`, 0.5}, 0.5], 
                   DiskBox[{-1.8844439181355583`, 0.5}, 0.1]}, 
                  {GrayLevel[0.5], 
                   DiskBox[{-1.8793493638220085`, 1.9999913484804992`}, 0.1]}, 
                  {GrayLevel[0.5], Thickness[0.01], 
                   LineBox[{{-1.8844439181355583`, 
                    0.5}, {-1.8793493638220085`, 1.9999913484804992`}}]}, 
                  {RGBColor[0, 1, 0], 
                   DiskBox[{-1.8810475485931917`, 1.4999942323203328`}, 
                    0.05]}, 
                  {RGBColor[0, 0, 1], Arrowheads[0.01], 
                   CircleBox[{-1.8844439181355583`, 0.5}, 0.3333333333333333, 
                    NCache[{
                    Rational[-1, 2] 
                    Pi, -1.620871304803773}, {-1.5707963267948966`, \
-1.620871304803773}]], 
                   ArrowBox[{{-1.8977986688384338`, 
                    0.16693429815609184`}, {-1.9011286026142296`, 
                    0.1670844965836552}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-1.8844439181355583`, 0.5}, {-1.851060599462974,
                     0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-1.8844439181355583`, 
                    0.5}, {-1.9094240295869334`, 0.5}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.005], Arrowheads[0.04], 
                   ArrowBox[{{-1.8844439181355583`, 0.5}, {-1.876040710914349,
                     0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-1.8844439181355583`, 0.5}, 1.6, {
                    1.5673999507227911`, 1.5517872910168617`}], 
                   ArrowBox[{{-1.8700296558760177`, 
                    2.0999350702586383`}, {-1.8540312925350502`, 
                    2.099710933951532}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Dashing[{Small, Small}], 
                   Arrowheads[0.02], 
                   CircleBox[{-1.8844439181355583`, 0.5}, 1.7, {
                    1.5673999507227911`, 1.5722699348895892`}], 
                   ArrowBox[{{-1.8699492275238112`, 
                    2.199938205919283}, {-1.8869490509898774`, 
                    2.19999815420764}}]}, 
                  {RGBColor[0, 0, 1], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-1.8844439181355583`, 0.5}, 1.8, {
                    1.5673999507227911`, 1.5766730947985088`}], 
                   ArrowBox[{{-1.8770221215718146`, 
                    2.299984699083791}, {-1.8950220396534385`, 
                    2.2999689173275057`}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.003], Arrowheads[0.025], 
                   CircleBox[{-1.8844439181355583`, 0.5}, 1.5, {
                    1.5673999507227911`, 1.5673935353163657`}], 
                   ArrowBox[{{-1.8943396591350772`, 
                    1.9999673577481847`}, {-1.8793397407679784`, 
                    1.9999913157659948`}}]}, 
                  {RGBColor[0, 0, 1], ArrowBox[{{0, 2.8}, {0, 2.8}}]}},
                 Frame->True,
                 FrameTicks->Automatic,
                 ImageSize->800,
                 PlotRange->{{-3, 3}, {-1, 3}}], 576->
                GraphicsBox[{InsetBox[
                   TagBox[GridBox[{
                    {"\<\"t\"\>", "28.75`"},
                    {"\<\"driveIntent\"\>", "0"},
                    {"\<\"motorTorque\"\>", 
                    RowBox[{"-", "1.657233500831696`"}]},
                    {"\<\"vw\"\>", 
                    RowBox[{"-", "0.01906570984827768`"}]},
                    {"\<\"controllerState\"\>", 
                    RowBox[{"\[LeftAssociation]", 
                    RowBox[{
                    
                    RowBox[{"\<\"lastError\"\>", "\[Rule]", 
                    "0.002347954754047726`"}], ",", 
                    RowBox[{"\<\"lastInt\"\>", "\[Rule]", 
                    RowBox[{"-", "0.016334892328583664`"}]}], ",", 
                    RowBox[{"\<\"initialized\"\>", "\[Rule]", "True"}]}], 
                    "\[RightAssociation]"}]}
                    },
                    AutoDelete->False,
                    
                    GridBoxItemSize->{
                    "Columns" -> {{Automatic}}, "Rows" -> {{Automatic}}}],
                    "Grid"], {-1, -0.5}], 
                  {GrayLevel[0], LineBox[{{-5, 0}, {5, 0}}]}, 
                  {GrayLevel[0.5], 
                   CircleBox[{-1.8854005682914046`, 0.5}, 0.5], 
                   DiskBox[{-1.8854005682914046`, 0.5}, 0.1]}, 
                  {GrayLevel[0.5], 
                   DiskBox[{-1.8803424333013687`, 1.9999914717325638`}, 0.1]}, 
                  {GrayLevel[0.5], Thickness[0.01], 
                   LineBox[{{-1.8854005682914046`, 
                    0.5}, {-1.8803424333013687`, 1.9999914717325638`}}]}, 
                  {RGBColor[0, 1, 0], 
                   DiskBox[{-1.8820284782980474`, 1.4999943144883758`}, 
                    0.05]}, 
                  {RGBColor[0, 0, 1], Arrowheads[0.01], 
                   CircleBox[{-1.8854005682914046`, 0.5}, 0.3333333333333333, 
                    NCache[{
                    Rational[-1, 2] 
                    Pi, -1.6205133318198475`}, {-1.5707963267948966`, \
-1.6205133318198475`}]], 
                   ArrowBox[{{-1.898636089618024, 
                    0.16692953885653006`}, {-1.9019660769474414`, 
                    0.16707854524816534`}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-1.8854005682914046`, 
                    0.5}, {-1.8522558982747706`, 0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-1.8854005682914046`, 
                    0.5}, {-1.9102021070311836`, 0.5}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.005], Arrowheads[0.04], 
                   ArrowBox[{{-1.8854005682914046`, 
                    0.5}, {-1.8770574370145499`, 0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-1.8854005682914046`, 0.5}, 1.6, {
                    1.5674242304108392`, 1.5519231805670781`}], 
                   ArrowBox[{{-1.8712037206213543`, 
                    2.099937014234071}, {-1.8552053269703175`, 
                    2.0997150519394263`}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Dashing[{Small, Small}], 
                   Arrowheads[0.02], 
                   CircleBox[{-1.8854005682914046`, 0.5}, 1.7, {
                    1.5674242304108392`, 1.5722593986463482`}], 
                   ArrowBox[{{-1.87088796671803, 
                    2.1999380531053396`}, {-1.8878877895515231`, 
                    2.1999981805079685`}}]}, 
                  {RGBColor[0, 0, 1], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-1.8854005682914046`, 0.5}, 1.8, {
                    1.5674242304108392`, 1.5766310831932375`}], 
                   ArrowBox[{{-1.8779031514875604`, 
                    2.2999843856937394`}, {-1.8959030702163164`, 
                    2.2999693601429216`}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.003], Arrowheads[0.025], 
                   CircleBox[{-1.8854005682914046`, 0.5}, 1.5, {
                    1.5674242304108392`, 1.5674178579166156`}], 
                   ArrowBox[{{-1.8953327923944232`, 
                    1.99996711661435}, {-1.8803328746144823`, 
                    1.9999914394691714`}}]}, 
                  {RGBColor[0, 0, 1], ArrowBox[{{0, 2.8}, {0, 2.8}}]}},
                 Frame->True,
                 FrameTicks->Automatic,
                 ImageSize->800,
                 PlotRange->{{-3, 3}, {-1, 3}}], 577->
                GraphicsBox[{InsetBox[
                   TagBox[GridBox[{
                    {"\<\"t\"\>", "28.8`"},
                    {"\<\"driveIntent\"\>", "0"},
                    {"\<\"motorTorque\"\>", 
                    RowBox[{"-", "1.645384528591346`"}]},
                    {"\<\"vw\"\>", 
                    RowBox[{"-", "0.018929192046681993`"}]},
                    {"\<\"controllerState\"\>", 
                    RowBox[{"\[LeftAssociation]", 
                    RowBox[{
                    
                    RowBox[{"\<\"lastError\"\>", "\[Rule]", 
                    "0.0023311066746874104`"}], ",", 
                    RowBox[{"\<\"lastInt\"\>", "\[Rule]", 
                    RowBox[{"-", "0.016217924735636744`"}]}], ",", 
                    RowBox[{"\<\"initialized\"\>", "\[Rule]", "True"}]}], 
                    "\[RightAssociation]"}]}
                    },
                    AutoDelete->False,
                    
                    GridBoxItemSize->{
                    "Columns" -> {{Automatic}}, "Rows" -> {{Automatic}}}],
                    "Grid"], {-1, -0.5}], 
                  {GrayLevel[0], LineBox[{{-5, 0}, {5, 0}}]}, 
                  {GrayLevel[0.5], 
                   CircleBox[{-1.8863503684995344`, 0.5}, 0.5], 
                   DiskBox[{-1.8863503684995344`, 0.5}, 0.1]}, 
                  {GrayLevel[0.5], 
                   DiskBox[{-1.8813283979330953`, 1.9999915932469854`}, 0.1]}, 
                  {GrayLevel[0.5], Thickness[0.01], 
                   LineBox[{{-1.8863503684995344`, 
                    0.5}, {-1.8813283979330953`, 1.9999915932469854`}}]}, 
                  {RGBColor[0, 1, 0], 
                   DiskBox[{-1.8830023881219085`, 1.4999943954979904`}, 
                    0.05]}, 
                  {RGBColor[0, 0, 1], Arrowheads[0.01], 
                   CircleBox[{-1.8863503684995344`, 0.5}, 0.3333333333333333, 
                    NCache[{
                    Rational[-1, 2] 
                    Pi, -1.620157862652637}, {-1.5707963267948966`, \
-1.620157862652637}]], 
                   ArrowBox[{{-1.899467492712993, 
                    0.16692485507999882`}, {-1.9027975327992013`, 
                    0.16707267775442214`}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-1.8863503684995344`, 
                    0.5}, {-1.8534426779277076`, 0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-1.8863503684995344`, 
                    0.5}, {-1.9109745843199788`, 0.5}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.005], Arrowheads[0.04], 
                   ArrowBox[{{-1.8863503684995344`, 0.5}, {-1.878066893748152,
                     0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-1.8863503684995344`, 0.5}, 1.6, {
                    1.5674483401626689`, 1.5520581190203657`}], 
                   ArrowBox[{{-1.8723694139841383`, 
                    2.099938915368596}, {-1.8563709905275052`, 
                    2.09971911187446}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Dashing[{Small, Small}], 
                   Arrowheads[0.02], 
                   CircleBox[{-1.8863503684995344`, 0.5}, 1.7, {
                    1.5674483401626689`, 1.5722489355248894`}], 
                   ArrowBox[{{-1.8718199802686326`, 
                    2.1999379011651747`}, {-1.8888198024720746`, 
                    2.1999982064390116`}}]}, 
                  {RGBColor[0, 0, 1], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-1.8863503684995344`, 0.5}, 1.8, {
                    1.5674483401626689`, 1.5765893653215097`}], 
                   ArrowBox[{{-1.878777860184522, 
                    2.299984071351138}, {-1.8967777795244485`, 
                    2.299969796718633}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.003], Arrowheads[0.025], 
                   CircleBox[{-1.8863503684995344`, 0.5}, 1.5, {
                    1.5674483401626689`, 1.5674420092120045`}], 
                   ArrowBox[{{-1.8963188187485598`, 
                    1.9999668763008178`}, {-1.8813189015604221`, 
                    1.999991561423077}}]}, 
                  {RGBColor[0, 0, 1], ArrowBox[{{0, 2.8}, {0, 2.8}}]}},
                 Frame->True,
                 FrameTicks->Automatic,
                 ImageSize->800,
                 PlotRange->{{-3, 3}, {-1, 3}},
                 ImageCache->GraphicsData["CompressedBitmap", "\<\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\
\>"]], 578->
                GraphicsBox[{InsetBox[
                   TagBox[GridBox[{
                    {"\<\"t\"\>", "28.85`"},
                    {"\<\"driveIntent\"\>", "0"},
                    {"\<\"motorTorque\"\>", 
                    RowBox[{"-", "1.6336184808321732`"}]},
                    {"\<\"vw\"\>", 
                    RowBox[{"-", "0.018793650493639288`"}]},
                    {"\<\"controllerState\"\>", 
                    RowBox[{"\[LeftAssociation]", 
                    RowBox[{
                    
                    RowBox[{"\<\"lastError\"\>", "\[Rule]", 
                    "0.002314382661122094`"}], ",", 
                    RowBox[{"\<\"lastInt\"\>", "\[Rule]", 
                    RowBox[{"-", "0.016101796378973746`"}]}], ",", 
                    RowBox[{"\<\"initialized\"\>", "\[Rule]", "True"}]}], 
                    "\[RightAssociation]"}]}
                    },
                    AutoDelete->False,
                    
                    GridBoxItemSize->{
                    "Columns" -> {{Automatic}}, "Rows" -> {{Automatic}}}],
                    "Grid"], {-1, -0.5}], 
                  {GrayLevel[0], LineBox[{{-5, 0}, {5, 0}}]}, 
                  {GrayLevel[0.5], 
                   CircleBox[{-1.8872933677404344`, 0.5}, 0.5], 
                   DiskBox[{-1.8872933677404344`, 0.5}, 0.1]}, 
                  {GrayLevel[0.5], 
                   DiskBox[{-1.882307308359893, 1.9999917130477267`}, 0.1]}, 
                  {GrayLevel[0.5], Thickness[0.01], 
                   LineBox[{{-1.8872933677404344`, 0.5}, {-1.882307308359893, 
                    1.9999917130477267`}}]}, 
                  {RGBColor[0, 1, 0], 
                   DiskBox[{-1.883969328153407, 1.4999944753651513`}, 0.05]}, 
                  {RGBColor[0, 0, 1], Arrowheads[0.01], 
                   CircleBox[{-1.8872933677404344`, 0.5}, 0.3333333333333333, 
                    NCache[{
                    Rational[-1, 2] 
                    Pi, -1.6198048812198618`}, {-1.5707963267948966`, \
-1.6198048812198618`}]], 
                   ArrowBox[{{-1.900292921797289, 
                    0.16692024572869335`}, {-1.903623013854701, 
                    0.1670668929516112}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-1.8872933677404344`, 
                    0.5}, {-1.8546209981237909`, 0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-1.8872933677404344`, 
                    0.5}, {-1.9117415022847655`, 0.5}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.005], Arrowheads[0.04], 
                   ArrowBox[{{-1.8872933677404344`, 0.5}, {-1.879069132668122,
                     0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-1.8872933677404344`, 0.5}, 1.6, {
                    1.5674722810864872`, 1.5521921125789289`}], 
                   ArrowBox[{{-1.8735267948586596`, 
                    2.099940774363567}, {-1.8575283420933943`, 
                    2.0997231145570883`}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Dashing[{Small, Small}], 
                   Arrowheads[0.02], 
                   CircleBox[{-1.8872933677404344`, 0.5}, 1.7, {
                    1.5674722810864872`, 1.5722385451428287`}], 
                   ArrowBox[{{-1.8727453165060446`, 
                    2.199937750097127}, {-1.8897451380819743`, 
                    2.1999982320056075`}}]}, 
                  {RGBColor[0, 0, 1], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-1.8872933677404344`, 0.5}, 1.8, {
                    1.5674722810864872`, 1.5765479393133326`}], 
                   ArrowBox[{{-1.8796462932770823`, 
                    2.2999837561078578`}, {-1.897646213192905, 
                    2.2999702271401707`}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.003], Arrowheads[0.025], 
                   CircleBox[{-1.8872933677404344`, 0.5}, 1.5, {
                    1.5674722810864872`, 1.5674659904811599`}], 
                   ArrowBox[{{-1.8972977890959746`, 
                    1.9999666368134128`}, {-1.8822978725041308`, 
                    1.9999916816527163`}}]}, 
                  {RGBColor[0, 0, 1], ArrowBox[{{0, 2.8}, {0, 2.8}}]}},
                 Frame->True,
                 FrameTicks->Automatic,
                 ImageSize->800,
                 PlotRange->{{-3, 3}, {-1, 3}}], 579->
                GraphicsBox[{InsetBox[
                   TagBox[GridBox[{
                    {"\<\"t\"\>", "28.900000000000002`"},
                    {"\<\"driveIntent\"\>", "0"},
                    {"\<\"motorTorque\"\>", 
                    RowBox[{"-", "1.6219348366415411`"}]},
                    {"\<\"vw\"\>", 
                    RowBox[{"-", "0.01865907836410356`"}]},
                    {"\<\"controllerState\"\>", 
                    RowBox[{"\[LeftAssociation]", 
                    RowBox[{
                    
                    RowBox[{"\<\"lastError\"\>", "\[Rule]", 
                    "0.0022977817424178755`"}], ",", 
                    RowBox[{"\<\"lastInt\"\>", "\[Rule]", 
                    RowBox[{"-", "0.01598650107999068`"}]}], ",", 
                    RowBox[{"\<\"initialized\"\>", "\[Rule]", "True"}]}], 
                    "\[RightAssociation]"}]}
                    },
                    AutoDelete->False,
                    
                    GridBoxItemSize->{
                    "Columns" -> {{Automatic}}, "Rows" -> {{Automatic}}}],
                    "Grid"], {-1, -0.5}], 
                  {GrayLevel[0], LineBox[{{-5, 0}, {5, 0}}]}, 
                  {GrayLevel[0.5], CircleBox[{-1.888229614652324, 0.5}, 0.5], 
                   DiskBox[{-1.888229614652324, 0.5}, 0.1]}, 
                  {GrayLevel[0.5], 
                   DiskBox[{-1.883279214834236, 1.9999918311583038`}, 0.1]}, 
                  {GrayLevel[0.5], Thickness[0.01], 
                   LineBox[{{-1.888229614652324, 0.5}, {-1.883279214834236, 
                    1.9999918311583038`}}]}, 
                  {RGBColor[0, 1, 0], 
                   DiskBox[{-1.884929348106932, 1.4999945541055357`}, 0.05]}, 
                  {RGBColor[0, 0, 1], Arrowheads[0.01], 
                   CircleBox[{-1.888229614652324, 0.5}, 0.3333333333333333, 
                    NCache[{
                    Rational[-1, 2] 
                    Pi, -1.6194543718941428`}, {-1.5707963267948966`, \
-1.6194543718941428`}]], 
                   ArrowBox[{{-1.9011124203529473`, 
                    0.1669157097244136}, {-1.904442563607015, 
                    0.16706118971002543`}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-1.888229614652324, 0.5}, {-1.8557909179194934`,
                     0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-1.888229614652324, 0.5}, {-1.9125029016498714`,
                     0.5}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.005], Arrowheads[0.04], 
                   ArrowBox[{{-1.888229614652324, 0.5}, {-1.8800642049170404`,
                     0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-1.888229614652324, 0.5}, 1.6, {
                    1.5674960542585237`, 1.5523251672660072`}], 
                   ArrowBox[{{-1.8746759215108566`, 
                    2.0999425919082935`}, {-1.8586774399265487`, 
                    2.099727060772861}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Dashing[{Small, Small}], 
                   Arrowheads[0.02], 
                   CircleBox[{-1.888229614652324, 0.5}, 1.7, {
                    1.5674960542585237`, 1.572228227125276}], 
                   ArrowBox[{{-1.8736640234311652`, 
                    2.19993759989959}, {-1.8906638443821364`, 
                    2.199998257212525}}]}, 
                  {RGBColor[0, 0, 1], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-1.888229614652324, 0.5}, 1.8, {
                    1.5674960542585237`, 1.5765068033509768`}], 
                   ArrowBox[{{-1.8805084961314307`, 
                    2.2999834400151533`}, {-1.898508416588551, 
                    2.2999706514915057`}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.003], Arrowheads[0.025], 
                   CircleBox[{-1.888229614652324, 0.5}, 1.5, {
                    1.5674960542585237`, 1.567489802959769}], 
                   ArrowBox[{{-1.898269753928478, 
                    1.9999663981580773`}, {-1.8832698379372665`, 
                    1.9999918001825665`}}]}, 
                  {RGBColor[0, 0, 1], ArrowBox[{{0, 2.8}, {0, 2.8}}]}},
                 Frame->True,
                 FrameTicks->Automatic,
                 ImageSize->800,
                 PlotRange->{{-3, 3}, {-1, 3}},
                 ImageCache->GraphicsData["CompressedBitmap", "\<\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\
\>"]], 580->
                GraphicsBox[{InsetBox[
                   TagBox[GridBox[{
                    {"\<\"t\"\>", "28.95`"},
                    {"\<\"driveIntent\"\>", "0"},
                    {"\<\"motorTorque\"\>", 
                    RowBox[{"-", "1.6103330881670095`"}]},
                    {"\<\"vw\"\>", 
                    RowBox[{"-", "0.018525468873686623`"}]},
                    {"\<\"controllerState\"\>", 
                    RowBox[{"\[LeftAssociation]", 
                    RowBox[{
                    
                    RowBox[{"\<\"lastError\"\>", "\[Rule]", 
                    "0.0022813029320363537`"}], ",", 
                    RowBox[{"\<\"lastInt\"\>", "\[Rule]", 
                    RowBox[{"-", "0.01587203270904415`"}]}], ",", 
                    RowBox[{"\<\"initialized\"\>", "\[Rule]", "True"}]}], 
                    "\[RightAssociation]"}]}
                    },
                    AutoDelete->False,
                    
                    GridBoxItemSize->{
                    "Columns" -> {{Automatic}}, "Rows" -> {{Automatic}}}],
                    "Grid"], {-1, -0.5}], 
                  {GrayLevel[0], LineBox[{{-5, 0}, {5, 0}}]}, 
                  {GrayLevel[0.5], 
                   CircleBox[{-1.8891591575332078`, 0.5}, 0.5], 
                   DiskBox[{-1.8891591575332078`, 0.5}, 0.1]}, 
                  {GrayLevel[0.5], 
                   DiskBox[{-1.8842441672267978`, 1.9999919476018155`}, 0.1]}, 
                  {GrayLevel[0.5], Thickness[0.01], 
                   LineBox[{{-1.8891591575332078`, 
                    0.5}, {-1.8842441672267978`, 1.9999919476018155`}}]}, 
                  {RGBColor[0, 1, 0], 
                   DiskBox[{-1.8858824973289345`, 1.4999946317345438`}, 
                    0.05]}, 
                  {RGBColor[0, 0, 1], Arrowheads[0.01], 
                   CircleBox[{-1.8891591575332078`, 0.5}, 0.3333333333333333, 
                    NCache[{
                    Rational[-1, 2] 
                    Pi, -1.619106319439907}, {-1.5707963267948966`, \
-1.619106319439907}]], 
                   ArrowBox[{{-1.901926031651159, 
                    0.16691124600736296`}, {-1.9052562253381842`, 
                    0.16705556691965384`}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-1.8891591575332078`, 
                    0.5}, {-1.8569524957698678`, 0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-1.8891591575332078`, 
                    0.5}, {-1.9132588230032899`, 0.5}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.005], Arrowheads[0.04], 
                   ArrowBox[{{-1.8891591575332078`, 
                    0.5}, {-1.8810521612399498`, 0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-1.8891591575332078`, 0.5}, 1.6, {
                    1.5675196607272832`, 1.552457288949672}], 
                   ArrowBox[{{-1.8758168516184326`, 
                    2.099944368680635}, {-1.8598183416974243`, 
                    2.0997309512934015`}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Dashing[{Small, Small}], 
                   Arrowheads[0.02], 
                   CircleBox[{-1.8891591575332078`, 0.5}, 1.7, {
                    1.5675196607272832`, 1.5722179811028323`}], 
                   ArrowBox[{{-1.8745761487140122`, 
                    2.1999374505709848`}, {-1.8915759690425948`, 
                    2.1999982820644637`}}]}, 
                  {RGBColor[0, 0, 1], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-1.8891591575332078`, 0.5}, 1.8, {
                    1.5675196607272832`, 1.5764659556615446`}], 
                   ArrowBox[{{-1.8813645138542354`, 
                    2.299983123123636}, {-1.8993644348187209`, 
                    2.299971069855159}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.003], Arrowheads[0.025], 
                   CircleBox[{-1.8891591575332078`, 0.5}, 1.5, {
                    1.5675196607272832`, 1.5675134478445667`}], 
                   ArrowBox[{{-1.8992347633392368`, 
                    1.9999661603408396`}, {-1.8842348479528463`, 
                    1.9999919170366076`}}]}, 
                  {RGBColor[0, 0, 1], ArrowBox[{{0, 2.8}, {0, 2.8}}]}},
                 Frame->True,
                 FrameTicks->Automatic,
                 ImageSize->800,
                 PlotRange->{{-3, 3}, {-1, 3}}], 581->
                GraphicsBox[{InsetBox[
                   TagBox[GridBox[{
                    {"\<\"t\"\>", "29.`"},
                    {"\<\"driveIntent\"\>", "0"},
                    {"\<\"motorTorque\"\>", 
                    RowBox[{"-", "1.5988127386680915`"}]},
                    {"\<\"vw\"\>", 
                    RowBox[{"-", "0.018392815277797306`"}]},
                    {"\<\"controllerState\"\>", 
                    RowBox[{"\[LeftAssociation]", 
                    RowBox[{
                    
                    RowBox[{"\<\"lastError\"\>", "\[Rule]", 
                    "0.002264945231534063`"}], ",", 
                    RowBox[{"\<\"lastInt\"\>", "\[Rule]", 
                    RowBox[{"-", "0.01575838518613611`"}]}], ",", 
                    RowBox[{"\<\"initialized\"\>", "\[Rule]", "True"}]}], 
                    "\[RightAssociation]"}]}
                    },
                    AutoDelete->False,
                    
                    GridBoxItemSize->{
                    "Columns" -> {{Automatic}}, "Rows" -> {{Automatic}}}],
                    "Grid"], {-1, -0.5}], 
                  {GrayLevel[0], LineBox[{{-5, 0}, {5, 0}}]}, 
                  {GrayLevel[0.5], 
                   CircleBox[{-1.8900820443428903`, 0.5}, 0.5], 
                   DiskBox[{-1.8900820443428903`, 0.5}, 0.1]}, 
                  {GrayLevel[0.5], 
                   DiskBox[{-1.8852022150343941`, 1.9999920624009715`}, 0.1]}, 
                  {GrayLevel[0.5], Thickness[0.01], 
                   LineBox[{{-1.8900820443428903`, 
                    0.5}, {-1.8852022150343941`, 1.9999920624009715`}}]}, 
                  {RGBColor[0, 1, 0], 
                   DiskBox[{-1.886828824803893, 1.4999947082673142`}, 0.05]}, 
                  {RGBColor[0, 0, 1], Arrowheads[0.01], 
                   CircleBox[{-1.8900820443428903`, 0.5}, 0.3333333333333333, 
                    NCache[{
                    Rational[-1, 2] 
                    Pi, -1.6187607089549394`}, {-1.5707963267948966`, \
-1.6187607089549394`}]], 
                   ArrowBox[{{-1.9027337987348487`, 
                    0.16690685353505597`}, {-1.9060640421018034`, 
                    0.16705002348889525`}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-1.8900820443428903`, 
                    0.5}, {-1.8581057895695285`, 0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-1.8900820443428903`, 0.5}, {-1.914009306769625,
                     0.5}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.005], Arrowheads[0.04], 
                   ArrowBox[{{-1.8900820443428903`, 
                    0.5}, {-1.8820330519962631`, 0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-1.8900820443428903`, 0.5}, 1.6, {
                    1.5675431015174977`, 1.5525884833649468`}], 
                   ArrowBox[{{-1.8769496423082581`, 
                    2.099946105347552}, {-1.8609511045257634`, 
                    2.0997347868773035`}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Dashing[{Small, Small}], 
                   Arrowheads[0.02], 
                   CircleBox[{-1.8900820443428903`, 0.5}, 1.7, {
                    1.5675431015174977`, 1.5722078067098064`}], 
                   ArrowBox[{{-1.8754817396927081`, 
                    2.199937302109734}, {-1.8924815594014877`, 
                    2.199998306566063}}]}, 
                  {RGBColor[0, 0, 1], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-1.8900820443428903`, 0.5}, 1.8, {
                    1.5675431015174977`, 1.576425394510098}], 
                   ArrowBox[{{-1.8822143912822917`, 
                    2.299982805483241}, {-1.9002143127208648`, 
                    2.299971482312294}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.003], Arrowheads[0.025], 
                   CircleBox[{-1.8900820443428903`, 0.5}, 1.5, {
                    1.5675431015174977`, 1.5675369262971806`}], 
                   ArrowBox[{{-1.9001928670305568`, 
                    1.9999659233677871`}, {-1.885192952253028, 
                    1.9999920322383504`}}]}, 
                  {RGBColor[0, 0, 1], ArrowBox[{{0, 2.8}, {0, 2.8}}]}},
                 Frame->True,
                 FrameTicks->Automatic,
                 ImageSize->800,
                 PlotRange->{{-3, 3}, {-1, 3}},
                 ImageCache->GraphicsData["CompressedBitmap", "\<\
eJztXQ2UllWdf3y/Zxhg+EgIqSYlY02LjAIn0qEMKN2kwo2QbPwa0FBHJSJz
BRRlFISJWGEJgggXVkQwacDCWOUU6/Fsc3bN5XSmXdqlzhwPu3GK3cMW1d3n
f7/vff73eZ/7vDPK5MPhfeZ973Pv/f3v/+v+78dzn7+8YeH8W26/YeGtN93Q
NP3uG9rn33rTPU3T7rw7TMqfEwT5u4IgN7spgO8kCMSFhP+1y0K4wL3+SW2k
fwpB9+a2oK2NfZZ19a6kyXkjOUevebgd8EqCR3g+QrqDzW2bg26ZVghI92ZZ
dnM3WYVkTQrTgVOZZ20YSv+UGODmblkKcofAjwrg3q5gGQdejRDTIaqhacuC
rl6il1kZQ79e9ap4anKM5jpRUmOSaHnbsq6gl5fK8Xy9XcskRx50li2qspQi
1oiHnPlLBpbWMskLqEZm6Q26ltEqMV6sMnjBkNcg1J/DGDBCE33IFvrLrJZV
galNRzyLMeoeQ6pejZDAqRumEc71F9UZDCoFdXFV6yRwhR8eU4XOakZJkUsO
8iFU4hJ+JKYoJlWsVbk485TZHuW3aTvlbUDVrLfAS4DymgqKG6uj1Mp4rEfi
Kc25XaaUa1HWJ25BRZgSe7tJvOqVODX5OKePp1YQZCvnrb45/bxOcrQizUTC
2tSlFS6FKmWCPOSSNvlWJG1MDWm11DcQy55NtJxNZbmrKNE/uBfN1DFTx9dZ
Hcv0T6RXxDRxFZL2WB/nw9I4pUn99xtMZiNyy+rE5mals9JZ6ax0VnqAlKZ3
rDLXwSWnDTDecCqz0m/q0jtYWnjtCtqC5qCzh9d4qbzRE3Q2azfuqx3gbr3u
zi766xYkbVY/0nAPT+sK83XxfLcgaZ/tIxo+Q69h5NzTGVIg/imcns7msDj8
GxK0dXHRQt7mzqCH5tHwaTvCf3BP5WHog2hVEsCuSq+yq41DFiFjBJLVN9Ko
j6H28GZpZIjvFLSNfmfYeh6KySDzCpISzbjDoEQriSExRk69RqILkgSyZSWF
rFPSFfpf+FdhJQw+G78ZaCWep4SIafgQV7KnIOjtaosWpcQo6Uu2FLQsBre0
lgUFngZIovK8lmbqU14TmOBikd7X06EejARNR0y5M+LzaYQyDBcKpqIurSeE
T4qYPK9Pw/OyyXMra44LU6sR4MSCCkXimpyPy2omh0WYfpQsSUhgVtkY3HJ1
tticN/I6pGHXKexB0xUprYphqcoQQqqbaXVlRBAlzaB1NYv6l7LiS3NIYbNm
/JKEQQYJJKIYmnJAPkhZ7OeeP02v0wyHI2iQHJXKo+WpKhWHe4mrS/8raYG6
A9xdUQZqiu3Kq6kqK+OgNYo5iItY98/SAuJ8tAOyh9WrawDVEoRaZibD4px0
Mq6G6XK1VONQXuqrVUY3w6Gu2+507XtewEbSSqZQuB8OycprfblukkY/LaB4
ek4TkvBoxTQCacQFktgRSY+g87kuCZ+dt3PGbcOJK7pLRtMjvXhDhNkyu7By
TQhh8/MaQ03HNjY5NzAOVjNErYyhDw0u+0RU0nDNelbWgpGxrjmiYjosTkLO
gLJdBMsmvfNX0njnvEEi48hbXAaUzjMbllSJ1mUKsICQVeIWJLlru7RqnQir
enAgQkjOOZZVepKIz62Y2hVFYdnqjAGIq4eQoWs1n6tbXxFlveCRHkAX6bXO
7NQd4spzfiYSpx0UKE01XbDOuLIiW/MKIDr4GwmLiNJqvRzmKoo+ghkeL5jk
vjfnZHjBZLiI+cr2OBLhf8nFZzMIYG0ras4YCWxLXLV1NiMBtYv7TKHeVrsD
ThKBWXW6g367tdIPucVhjoCFnifyysboiZPO8EqyMluEBdkv8JbT+r48EBxx
OdYRF7hyW95Wxo1mqx3dnByuoQ4Z08OKzopAjyUx99yARNkIphodCp/az+65
3ukXq4SmFbe8ddSod9V5gwbXvCo7HrMpF4YiZr3U6DoMWto0I+rpkfWIeK9U
i0yGpPXMFYTXlSS8dt7GvK9SnGi79Z5L8EeE/Fh+My7GR5VcDizSbKrNLVcz
R1u/o4MB3VJNDTH76zIiiyLXRj8XrDVNeoqSWV7lMcMQLgRIW5TGD18cMRw+
jEs2QtW555wCwMrETFe4jFYPdfSZMG1wNzymR4ihNIooJivEJKCu+VHPPNyh
XhqkYXwOUrra2AzsCGUwKB8T8pV/r0PITjRzkNA9m0aBuRNZnZZGy3QirrHR
DupUN2R5EXqPpjfr1TCzdI2uPUWi1xGv82WEzXVJ2FzFM9dhLspuM2dxHd59
W3mFBSk3aRiWxX7WrjHJWVG1Y4stk9O783gbRWYXWDomitgZC9uiqo4Rkd4T
m7FgUaN00J4zFsmX+4ZpkUqYl8/SC3mr8X+gDdNgnlzVp+qqmHXZ0hMLHtrI
264uEteXoiQwSLF2glLJtTduqc+Y+LJcoKHlNqQ+jKWwghLWdosSNGS2Gxzh
p1klwieiNDU6DeAcLRuzeMknM/Q5FEGRGNRZFl8U3OWl9UjYjIe0JUDNFOtS
cH4YznldA6tpNTGX8wQ765Kwc2jMHBB8h4ZZTlhrcIlDGMMUdV/Mw8dOM1m8
ZooxzLTBCAdspQoi0zC6XUfW52U+6qfs1d+qq3cc1lq9EwyITGLqjHO4iwhX
hLdJvngnna3n4t1NwiIiA9si6gjrdGcQ0YuIKZaQ6s9Fuqkckm+QTYbo10OW
jJO8jlkRebvrtjNdjgGUnVLButKjqq31joPNoYqg3mlsVtga4YRly+p7ESky
IgmD4pbmlGpF10f10KJklbM8DOaAzAhC49EwpCEVJK1RapAShUjT7W6EkS/9
BiRpJFpwFmck9cIDG9sW9LkcVU8JqdvPQgpGWTmayhshQpPIau7uKOLJYgIK
3wvCmB7brxVs14mFM6n30+QNncTUX1oMHk4XEaaP9OKQOV2pTx+VzM0qyrPY
M6F6M6341s0tbEXVbHpk6SSOBbj+DkHsa7Blr0IOtqKZtpne5uhVXQbQvtAB
VJreuYNLjC3sRFdmkPmHZ+ntft8rJidHHPvCTFPT9sq5Tc20KXr9P8mJmI3l
tws3QK2pjX661B5Hoe3NzdqwHOJWxrtndHeUcgNHA8ZZbOuFOU01yJ69YkUj
UwjI1ppo1fTHGT+VlJzj8MYMgnJTXdq8eXNnl/Adu4Q6apKWw+eYufJEa6/R
PTGOBVVlEthIEhsBGSiRsYCWPyU3u9o0NRGyrigkysUemY2nSz2sYWKuvooe
yi6jwGnSmYOsNJm1ItsHzVrTMG0BvcZshLC3ylJLjnSZe/uAe1jEEJmSFDuD
xZBTqhY+G4zNMytNlLXSxD+k45ytwegWHSFfIyCk0JJzcZMRw6vIxzEZ4aqu
aPDDXtCKWnHF0ShZbRrNo9dLknj/i1y3A9VTqnASbz3i5+2Jf4S7noPb1Gx4
r1D2GgPsgvlUBR4jaIMYLEZwD2Kc8Te9/j5FuwdGKNinpZdz4Rn9oqzsp29g
ZdfSq7X/w/Fdm8w2g9tAn8yXz3hgjx5hQUPC52LEJJK+4NuGdaXOrf+m/n6b
XpFlNeu3Piy1RuWxg/KC3emZqxxE+2iAUo75NHIcU12OcbN+xmpzbB5GXZNj
Ga3asw2275V7jJE4PHaj8DbEKbIwa3T8Bm3X1nfX6CGajoUvNUovgRU69kMh
M8JxO7Vc+wyNgRKyZVnjgBb+4tki2wm/JRgWGbmUDP5qDBVzndhilUA1Rihq
VGUklyPK8GaXWl31LZXItAF8tiJ6z0KWUQG674ugoxz3fjBHej/Y2xi3i+kr
b6mPXzy9pWvoxwcxibfDKtG7veUwdINI1FFrG240UsynuByjt37xlGrBk02R
iO9ykTp2PZ/o8Up97AK90Gg5uhxlhCtxRVVc0hl0VjM64o5XtnDXgA3+jKdH
jKmrUQmHqkmegkAGnWnFiC2zNiDe/odCpczxnmxvZIZrlHtR0TU8dKVLo9TU
oOb2aosPxsom1xFsBeGg0+OaYzldV11LRuj+C2QPDtE+fSBuVZBeBtDIbeCW
ZtmHYk9ZR8Nxx7SVnIgy5wzMMk5/YQ4N5bpHQ7SK+KLfFU7AnBFRnBlVhTP0
Dj0vR+y4NzrdJDwJwytY4jgf3+CBuZMk7rVaHe6yjCfOOTTFm5GSATGrOYw3
0V2RrlgSC12N6Qj2gyEO9uOJscvO3smKl5HMiJmopJdz/UyIsWVEKrZYCxKS
LSymHmx2wQn5UtAjFQ8dwSfga2TLpchgXR/EI3tp/cLVmCg88f59xNojZfuL
TXO59iMb5aKBorbqp6blC6bmCQc6FPfBCTZGqtjScsLVy8qlsj62setZmuH/
5GHiGha2q6UBSUvsGPuATuyh7UFxtGvRXwOS9mytHd1ZFH68WUvfLWRoxjA/
4uLWAyn7TCUVtGuT52j4PSwNTaZO/lhPw3Y7RKdIraCiNvwjBr6+g9K1sjGi
dtB/1I3OmLUo4pNUfdnil2iateDgml5JC3wXbw3m/V/iqul4Wt7oa/qS60zV
GtwLq/iIvXbma671xxhfeG/nWC9V+G95PR1IVvqsL03vWGXmyBvZcaxZ6ax0
VjornZWWpeV7LsxXK2EvkEheKYuX4Yq+capQW+WHWlpawsqnTp8VtLbOCqZP
nRpMnTo9mNXaGsyaPrUAtwtQ8Os0Y/7wlVde2T1z5syfzZkzJ8gdaw2z5X8V
3j7Z3t4On6BMryVIg7vw6Zk7dy6U+dG0adMIVEMvz8iqKUhISsuhQ4dCUg5t
eyhoe2gb/KIp2x5qCx7adqhAf9BGyFuP0Wtl0e7du7du2bLl+ccff7y7s7OT
LF269NTy5cvh+zMbN27Mwc0gv/qJJ55o37NnD3yCMr02QBrchc+uTZs2QZH/
fvjhh6GKV1evXn143bp1cO++Xbt2TXv++ecJINLL0wg9OXpVEsNfxIZJjL2U
qzL5+PHjiw8fPnxo69atQML+7du3Lw3rnP3KK6+0HDt2bJTKNvPo0aNwa8/O
nTtPh409ElK/4uDBg5BNiX5XAr3CX8SGUcle6FeunDlzprW7uxugT3Z0dOjQ
wTtZHr0lQB60ZP7LL7/cePr0aU0/d/Sp3j/Iiszo6enZsnfvXqAN/gKrAPfd
7DaQD1nW79u3D7LseOqpp4DBiqon+sTElzK48SdOnFhz4MCB3lWrVgEXgHNA
AtsZWgRkoEAQC4zTGLTNz1fdzyqd0NvbC9I5Gqpv+5EjR0afOnUq+AC7B5wA
GkBqIB0qNQW41Q+QPeBUAKoBsHvDhg3A7stYMtQNGIAFyQrmW34w9+L1fQRH
VzCb/GDYc8QFYBfIRLBnKksWXAUpglQVzEY/GH42BNgHSAhUIPg4wwDFhGaA
PYGWKIwNfhhfVqwBDKgPJD+dJUPVAH8s9Hcmx9b7wdyj6gPtBcEAa65SjBQc
g+8K5m/8YPhoH6oG7oD+BtewNMExS4/X+QGwKZU8VAyVAZA4vA4IB00Ab6sB
fMMP4E6TUfCB7/yVBVA3YFBDVRhf98PgD7U1nTx5UnLpcywNuINwqdMPgD2I
kwdDkFz6vOISiBm0SQNY4wewQMkU2EF7C7YSlQdugS5BL6IBrPYD+BJLA5OD
yigAn/WH/oDJmOghlOdruG5jaeB1wYUA1UGrEjLI3QJY6Qdwq6oMuhX660al
vQjAo34A81ka8BlYQn99kKVNPnKErGlvJ+1r1rDTxNygj/iBzmNpoD0gF/rr
ffRaHt3bS7a0thIAh5SZe/bQ372jR4dVEnitaD4MvIis6z1wzQGDUpPTpqwG
lJpKcQKz4GuffBJK0c8fczlSZBSM506Ufi5WGsXU1WDNCj9ablFGJ22CvW0F
QmbSePIkmdDdTWWyY/ZscqypiXCuUDGJatgDOXngLrNRg6aH/Wi6maWBpwGa
4G9wGU8L48/ZO3aQxStWCBFS2wU+Wt7HE/QmVRmA0kiJeyTwPuCRLP/5UCoA
MCuwX/rrDqUJAGqxbXkVAHrnRkS32xEdfUCWiJldvEFxHhpMtYH3jiqaM2h8
wI8JrQiX74nlcjoAiEggBKG/FiIsUgDL/AC+yNJEsECNd1Gs9i9NBQAaDT0V
/bVIyQXiLQqaGuB6hB2LEY3xrPYLSrQQ+FExflWlAd3UkhXAklQAhmjvjRXt
/X4Ac5UiQgsol++LFe1fpwIwXAAHEFFPbQDXIS24P7YFnmdIXIe0gAOAsoLS
1gYwB2kBV0RgD7DJAvhaKgCjBUtjtcgTgIes4DthVE5b8EBsC+5NBWCERhwA
WiTjtloBDI//gJILtKo2gNlIZbyzM9yeAvhqKgBwE+AudAAxJdInAIZAOQAI
HOZbLCftCcBHVtDNgOekv3hEA0Knoy+zBZ7HlFEA6vFXxGp/mmot3XlYWQQw
pjaAv0K43BFrXp6HZXIAw51xAKPrqRXA8EAcwBEIpgMw2NERq5yeANciAn0U
EbwCqHaiNL1zLcL5K+h1ENALlgu6DxDgNmAOG+DhN1g02ASNOlpipbVQ4sUE
wpwUw5+y+UDqtWDsBfBBPb3SzlWymuczhrjJOYETYijLFJYGtguz35SEwYwQ
AGSyDXN8mOWDIBumzvqEEMM/NyulAg5QEhoZIaB8LNxT+RydRzpCwCvSKSL4
dZkiBFpKSRjOCIEowiYEtIatFRiEeL5+YpYCBTHQX+wUxUIr2bJlC2ltnU12
7GgkJ0/y7aigpUAVpWwEow/olRK7DLHcWukzuhBGX3k06e09TSoVwucausmE
CSvI4sWTSahiQ5T+SmLfwohVsUX4mawaRSeZ+4JYwwG+iwOQ2bMFoWEpuBaA
+D1k5sz5ZP368eTo0TolC+EagnMZzSAbyUxKszXO8dQ/TqkRR3BKgX2S0jKj
9BhpalpP5s+fSfbsAVUoKs8iuTtKeREJM6nPKDUCxQu5hoUCl5TWM84eIZMn
LyVLllAlyCkrIx2LF+tkAkOl9UxSzUECz3RKYHhOTvBismKFJHgwvVZOksbG
535/1VVk39VXk65PfIJ875OfJN+ZO5esveMO8sD99w+j9BbhK1m3YAG99ezV
V7e++tJLQnkd84tp6KayYquA+SZy7JikdhhXh/+trye/GjOGvHLxxeSFyy8n
uz/zGbLp5pvJqoULg0GMtWT13XeTra2t5JlPfSr8dQhWxcmuWbMkNyb1D9nj
Ea0YSa9DyG+GDCH/dv755J8uvZQ8N20anQbce8015JfnnUdONTSQlydOJMvv
uw+aQZsTkkue/+hHS5z6UBbSoCchRl6rshjToO9GPMYo7jfOFArkv0aMIH86
5xzuRfLkl2PHApmU1DInOKQdNIa6zMn9QLARvnOCwewkwW/jxP2+WCQ/v+AC
YLAkbv+MGaAp0Kn1C3FGnMGJg44sws3TYefx62HDiPLJefLr4cNd3KRTG31P
sFN5R3Cb+12pRF4791zyswsvJEcmTwa7It++/nry6kUXBTmap0iVGNyDbnuC
/Mdvu02Golx7HfMctboM6JIl+Y3MwZE/5PPkxMiRpGfcOGpnofeiXgzIDbma
514DzA4oB09SoGkFapKQ7Yk5c0BrNGciR8R970zwzqadrFlj+e7yIdLSsvTM
gw9KlwF0htwmD997Lw/dirTj2djWRu+HrucT//Hqq/3guBP0kHVE9pAQJs0g
+/eLg4jEbhq9h6QjLGL2kMZcRK0EG2N1HnxAgCEJLtFr4SgZP14PPkqquE2z
oo30ZfCBBXT7SejFBKXMCmmYBI4GImcwBf4QFPgNOwiVs2qQg6uDY6yXjrsw
XKAeFn7xc9EnkO5uKAYxB3QvIvAcqpopx2KcTjVHo+h0zMJ60smHQAAoGXEB
S4PgDWilv/iTF0YkPwKJ5PmwwxhS1UqcMYbh4y6qUIKQYUjQ24w0rFZCDJZ/
WBm8FNdQxACaEU2odcRq6Cgfw8PYXHiQoIERAmKQQ2yez9HppCPEiK8vNz0K
JaGihqZSiXk+R6DvM79jzEfwuZptTz/9tNSMIoN/beXKlbKyK1g+Iz6xNCOX
bE7HmALj8OD8JXwJrlMRLU7Ndz4hZ8yNXeFC5iQ5ZkvTTAVSR+5saYupAH2A
Z4m4Wksd89meyHzi2ZjaubwKsjE3VCuy4WecyFfEOpZ0k+2YPTuRHYFTOmRj
Rq8asmNpxxN5NlLZR1zIlyNkKuR0az7GxE81ZEfUlw7Z8FzVkB2eyxP584iJ
TnEhc5IcS9npFiGNjq8asiMK9ETmK8SgMGwvCpFxQxR5ihKN3PWgkNOtTRt9
bDVkR4ecDtlohhP5w7FtTrejwGhGcxVkR5vTIRvNqIbs6DE8kecizbjMhcxJ
cvTP6XagGG12IlurKsybSGTPzTV8947R/UyuguyQsycy3+9kRF98UC7W8RQy
J8nYf6KQl6RCBsbJZR6ObAyFdGTHcoon8hdZmhEuf8iFHD8tm277mqFhTuQP
qcxyy7dC9tyZ14oYygdjkbEtksnxbkB4PNGF98FYHqdDNnZROpEnKo1Gtlc+
WAWZ3uF7TI1toR+ogufYCLBMVhozUrsJkeSlLrwPIJL03Jp7MyLJ97vwLo2V
ZDpksYJOI1on8vsRMSjkh/2Qb1HI0rlPcCFPUDoEma19nJ5bztk3OsCVuwTe
50J+n8oM3Lb6oA4/5HmKgdAMagrvdSG/VzEI2V3uueWfP/ZgbLO/pAqyY67Q
E/lWU11pmy92IV9sqiML+SWy57Mkt6k2S6V5jwuZSUWWHW+K3Xrux/OpmS+Z
dkN16KKEhLxblZWuVhHymB8hC1SrgCPU2v8iISEXmhpkGYLng0q3m3KmTm98
QkLEckMoEvl0hiLE85Es9gAEfXANvK94aurChLRcgBdX5HT6kdOu2gaaQqX9
roS0nG9qCutIJCFr/Qi5C6lsXEJC3ok0QhHi+WTh3SaHxbOW5yek5R2sOAgF
Yu2ogDyfpGSPctAHWSG6hME+kNaUkJy3s3zg12CW3Xra73E/Wr6sWAMTD8Br
oOsdCWkZy4oL9wiDIPMpVs+HZdmsZQFiaJASjCOgvrclJGeMkhQIWmiOIudv
/chhE00FGMqApKBOIOe8hOSMVuRAXqiCbpmS5HzTjxw2B1QEYUE9MLkDHKeY
b01IEzvvsQjNAIEDTVDGisM8nwpnA/gCqCRUKIgalZCmkYomYLEwCosmzwfi
l7BKBU2gnvJ4hZEJCWMHDJbExl2wfGA6MN989NzzTAA2KCyCbtpnLIhTP6vS
xhZti/pRDlCHFWV8x48wtgG0ZB+5AL9BOkMS0sYWvsvAa6AH2mgf7qDRGHOY
hHzNMz3igh5rEWAnSbBxQ0Wc3AC0AyD4R9AoIAJA6xPSz2DLwFvhFYEV4uFq
a/i5s3b62cbzevB7oFz6kSXwHQiGdkEADO0oJGwHo6IOioEIIb84tQFygzzA
0Kw47Kna27NaKbg4hkXseg9y0DI6tA7yPWvXrrVaUC+O3YAPcB1SxfZ4/agX
RfDeFAQXIfda5VChRvHZvnv3br5+XBBo8DF7fXpMbUGbBUjKnA0Jpaeg9rsb
WGJNCNjBNAD6OoLLs14oKJz1QoH7HfcN4TTmLzmAsRG6PwDA+DOADCADyABS
AKg19H4CUJPkGUAGkAG86QDoHf9q6Sno1fY8DkB2ZAAZQAaQAWQA/QcAs3P9
CkAnmTKADGDgAagt6v0EYDwXdhYB6HGoR7Xfk2Wrx6EDiR0ZwJsVoJ4CLFy1
kFy972oy8eWJpOlYE5nQPYHMODCDLFi3ACBpHy1xnkvVkI6VJ1aSyukKUBwU
6TVPphyecstL//yS1ZADfgB878CWVtJKa4Gqywzg55f8zyWV08Q+/MaTU3yv
xMw9ZI8EqGMAS5fw3comQJcfAD/DGyjtHU1G06rrGUDIsMr4o8TeGJcOYEI3
6ZYtGMQAWreQLUgLvpcKYP56st4GWNNO2vsMYMVistgG2D+DzOgzAEPIHABA
+wyg6Rg5ZgMAaO0AXE0PtZAWW01BqxA78NQiDgDqCGqpA8zYz42qJgBuyaCS
vyv/qaxb8t55vfMQgHROdfbh44fJuJ5xtOoC90UjT4wkN2+62doHmsrZlaXX
XtyxmLRubSUt/9BC5m6fC442yA3AbiID+PMCeB3iX1it6ld2ZABnJ4B69UMG
kAFkABlABpABZAAZwBsAoAe6HtX6bDgYSOzIADKADOANA6BPBvQLAH1irdIP
JNMHmCySfyhfsFgLAJ9OrPqOAQWQbsa16rsDFMCzfgDfUCyKfSeAAviuH8Ba
BRB71r8CeCYVQNXD+hXAXj8A/uLNqgfsK4A9fgCdikWxp+GnBuAnPsAZt3DU
Of1VYmlVD7ZXoE/7gdLDHnJwpG3IJ9LeDofa0lN2+Um89Kx6C9k4q7425MA8
x5idW0vPsoWz6T/52xdfDM5joOr4HgN0dypQ/VUNQZFeC/TI6u4JE8jemTPh
0OdRyoD9cLnM6MvD4TZ8cvSaD5Z19VrZzSr4ORTGYb6Myjx9e+bhKVPI9rlz
+YHUebJh/nwpHe5x1PkpvgRr7+rmT1KhxMEpvZK4QfRaogfBHx0/nnx/2jTy
28H0PGo4YHsIo0m+KiMxTWWDJvYucUonfU6Om0oLOXRIUiKOTYe3nQqWDFIu
DIF/Ei7uB+YYKPYYF2cEvIhEwjdwXTrZ2Ej+5ZJL6EHzcO/42LHBEOWq/bhQ
ZxDEJFNwE2QcKl/hag2HysPbEeB9DusWLOCLQvQoczjeHH5x+tRZuEnp48+v
Br1dy6Rq591mB44Fzno2zA40B45iPxCqfKjO2MnZHlpcorS0bRYzz/AQIn3f
O8Y2ThYcj227IPHejit/98ILwgU5dklVpako6RDuAOjDFIt5YzglnXljOBNd
98auc9ERqjy9sdb5wJnmnJc0rerB5qlBeZ9d9Tzy1F0qB6h6kHhqAB7WVD0M
PHVYw45+KY4lx4/LoCD80Mh7RGwg4BmgMaByAzl16mtk+fIlYd3Xke3bp5DD
h+/66YED9GVMjbHC9ww5eTz1WfLUUwAGnyBHr4UNv1m06F9fnDdPvMllkOpF
kLNu96XCBW5K3AK9lnacWLDg6CuzZpEfXHklfT8C4D+6aFFRBarImcqeew+4
Tn6FdHRI/CK91j34x/vv3/XabbcdOtba+sIvvvCF1/590qSfn5g06cf/ee21
G3+5dOnI08Zz/p5bBjjwnaSzUwLXseavJbffvpdccw0QRV+zIzG+n4q5oEUS
o0SvZaj7c2Tnzonk5ZeFiwEs9k4HCfiDVIBQqQTMs+tNZNOmj5GDB8F2NIAX
0wCU5pENG6Dae8iqVdeQvXsvJq+8ApSren+Uxt7yI8mJE/B6AfirEfmSX2Wq
IL3MfbOU3q3UDalxl19lf0+vRRgbgNcD7weekFUuK/UccLCzRAogXtBR8HZg
CFxFwR4QwmMOvMAwvs0wgFBQS1BPMPIywxhH0K3XnhjsZKCSGJxeRF59FcYB
wCQwjXqGBfwKIzq7167GM3rnm4r9t5LHH99MbrzxOTJt2ndIONwZxKoH43PJ
OV9lDp2d1lMA4t9BfvELIWFOuIx5amLSNoYBHACuTycHDnyJrFvHHGAeIGvH
+DslbHglCQgb2MJ6Turca8d4kmGAKMA3AZ/gtSypjWDgOpg3sjTzR4VK2MmM
D5l/1b59pOHUKa0+z2ibu6LCmTNkXE8P+djBg2ReGHjnloRxa9hbbtpEEMI9
Jc0sID/2+HHSEo7NvxgOV8L+nwKM5wpUGwA7CKwEjLhm717y/p/8hEw+coRc
t307+VoY4FcY1lc6OkiBn4eVfniykV6Lo3t7ycqFC+nLEGFkD+8KCz0SBaIN
jDbKE2iTkvS0554jn376aXJnZyepZxjwtXYMNidUGHniBJkYDlY+t3MnZViZ
YUAbaxfOdiV90C6qUgUGMP3AgT4A2MGkr5sEcEeIOzVzeKTQePIkqfCYLj2R
A9fjZKVfh9L0jlWGvjWh2i6AAdTGrHRWOiudlc5K/3mUxvqsOfJG1mdlpbPS
WemsdFb67C7Nx/kw27GltZXOHLEbiWao9Spm8veCw24Tre5E/SGfjYUJMphx
iNwYMOzMSmels9JZ6aw0v4W9Qwt7nVUt+WLe0yWOv9d3ni7r6sV2mn0WSfsQ
kjYmYVksH1Zf0rJ9jZG0vZyTJfonvHZv1jYTEspZx2bCtyakOGlaLfUNxLJn
Ey1nU1mukUm32mYiyHjwOqghd4RKDQnkVpdWuFjb6unO7DZejpaRF/ZCTXS/
u5WTvfTT4aBlKbwQ1j9u7rYzc1qs1IX+qcE5/w+oCMEW\
\>"]], 582->
                GraphicsBox[{InsetBox[
                   TagBox[GridBox[{
                    {"\<\"t\"\>", "29.05`"},
                    {"\<\"driveIntent\"\>", "0"},
                    {"\<\"motorTorque\"\>", 
                    RowBox[{"-", "1.5873733007225719`"}]},
                    {"\<\"vw\"\>", 
                    RowBox[{"-", "0.01826111087093646`"}]},
                    {"\<\"controllerState\"\>", 
                    RowBox[{"\[LeftAssociation]", 
                    RowBox[{
                    
                    RowBox[{"\<\"lastError\"\>", "\[Rule]", 
                    "0.0022487076340067555`"}], ",", 
                    RowBox[{"\<\"lastInt\"\>", "\[Rule]", 
                    RowBox[{"-", "0.01564555248142025`"}]}], ",", 
                    RowBox[{"\<\"initialized\"\>", "\[Rule]", "True"}]}], 
                    "\[RightAssociation]"}]}
                    },
                    AutoDelete->False,
                    
                    GridBoxItemSize->{
                    "Columns" -> {{Automatic}}, "Rows" -> {{Automatic}}}],
                    "Grid"], {-1, -0.5}], 
                  {GrayLevel[0], LineBox[{{-5, 0}, {5, 0}}]}, 
                  {GrayLevel[0.5], 
                   CircleBox[{-1.8909983227049472`, 0.5}, 0.5], 
                   DiskBox[{-1.8909983227049472`, 0.5}, 0.1]}, 
                  {GrayLevel[0.5], 
                   DiskBox[{-1.8861534073874313`, 1.9999921755781147`}, 0.1]}, 
                  {GrayLevel[0.5], Thickness[0.01], 
                   LineBox[{{-1.8909983227049472`, 
                    0.5}, {-1.8861534073874313`, 1.9999921755781147`}}]}, 
                  {RGBColor[0, 1, 0], 
                   DiskBox[{-1.8877683791599367`, 1.4999947837187433`}, 
                    0.05]}, 
                  {RGBColor[0, 0, 1], Arrowheads[0.01], 
                   CircleBox[{-1.8909983227049472`, 0.5}, 0.3333333333333333, 
                    NCache[{
                    Rational[-1, 2] 
                    Pi, -1.6184175258165736`}, {-1.5707963267948966`, \
-1.6184175258165736`}]], 
                   ArrowBox[{{-1.9035357644027493`, 
                    0.16690253128133004`}, {-1.9068660567071078`, 
                    0.16704455834339071`}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-1.8909983227049472`, 
                    0.5}, {-1.8592508566904957`, 0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-1.8909983227049472`, 
                    0.5}, {-1.9147543931852136`, 0.5}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.005], Arrowheads[0.04], 
                   ArrowBox[{{-1.8909983227049472`, 
                    0.5}, {-1.8830069271707621`, 0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-1.8909983227049472`, 0.5}, 1.6, {
                    1.5675663776337763`, 1.5527187561342402`}], 
                   ArrowBox[{{-1.878074350191024, 
                    2.0999478025655898`}, {-1.8620757850152427`, 
                    2.0997385682709497`}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Dashing[{Small, Small}], 
                   Arrowheads[0.02], 
                   CircleBox[{-1.8909983227049472`, 0.5}, 1.7, {
                    1.5675663776337763`, 1.5721977035826367`}], 
                   ArrowBox[{{-1.8763808433727667`, 
                    2.1999371545142408`}, {-1.8933806624643426`, 
                    2.199998330721907}}]}, 
                  {RGBColor[0, 0, 1], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-1.8909983227049472`, 0.5}, 1.8, {
                    1.5675663776337763`, 1.5763851181933461`}], 
                   ArrowBox[{{-1.883058172973128, 
                    2.2999824871432044`}, {-1.9010580948531568`, 
                    2.299971888942804}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.003], Arrowheads[0.025], 
                   CircleBox[{-1.8909983227049472`, 0.5}, 1.5, {
                    1.5675663776337763`, 1.5675602394477965`}], 
                   ArrowBox[{{-1.9011441143213532`, 
                    1.999965687245037}, {-1.8861442001565807`, 
                    1.9999921458108658`}}]}, 
                  {RGBColor[0, 0, 1], ArrowBox[{{0, 2.8}, {0, 2.8}}]}},
                 Frame->True,
                 FrameTicks->Automatic,
                 ImageSize->800,
                 PlotRange->{{-3, 3}, {-1, 3}}], 583->
                GraphicsBox[{InsetBox[
                   TagBox[GridBox[{
                    {"\<\"t\"\>", "29.1`"},
                    {"\<\"driveIntent\"\>", "0"},
                    {"\<\"motorTorque\"\>", 
                    RowBox[{"-", "1.5760142945847562`"}]},
                    {"\<\"vw\"\>", 
                    RowBox[{"-", "0.018130348986133152`"}]},
                    {"\<\"controllerState\"\>", 
                    RowBox[{"\[LeftAssociation]", 
                    RowBox[{
                    
                    RowBox[{"\<\"lastError\"\>", "\[Rule]", 
                    "0.0022325891272742`"}], ",", 
                    RowBox[{"\<\"lastInt\"\>", "\[Rule]", 
                    RowBox[{"-", "0.015533528615542724`"}]}], ",", 
                    RowBox[{"\<\"initialized\"\>", "\[Rule]", "True"}]}], 
                    "\[RightAssociation]"}]}
                    },
                    AutoDelete->False,
                    
                    GridBoxItemSize->{
                    "Columns" -> {{Automatic}}, "Rows" -> {{Automatic}}}],
                    "Grid"], {-1, -0.5}], 
                  {GrayLevel[0], LineBox[{{-5, 0}, {5, 0}}]}, 
                  {GrayLevel[0.5], 
                   CircleBox[{-1.8919080399086607`, 0.5}, 0.5], 
                   DiskBox[{-1.8919080399086607`, 0.5}, 0.1]}, 
                  {GrayLevel[0.5], 
                   DiskBox[{-1.8870977930568635`, 1.9999922871552456`}, 0.1]}, 
                  {GrayLevel[0.5], Thickness[0.01], 
                   LineBox[{{-1.8919080399086607`, 
                    0.5}, {-1.8870977930568635`, 1.9999922871552456`}}]}, 
                  {RGBColor[0, 1, 0], 
                   DiskBox[{-1.8887012086741293`, 1.4999948581034972`}, 
                    0.05]}, 
                  {RGBColor[0, 0, 1], Arrowheads[0.01], 
                   CircleBox[{-1.8919080399086607`, 0.5}, 0.3333333333333333, 
                    NCache[{
                    Rational[-1, 2] 
                    Pi, -1.6180767556324391`}, {-1.5707963267948966`, \
-1.6180767556324391`}]], 
                   ArrowBox[{{-1.904331971194959, 
                    0.1668982782354549}, {-1.9076623117045406`, 
                    0.1670391704249693}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-1.8919080399086607`, 
                    0.5}, {-1.8603877540169655`, 0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-1.8919080399086607`, 
                    0.5}, {-1.9154941222754316`, 0.5}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.005], Arrowheads[0.04], 
                   ArrowBox[{{-1.8919080399086607`, 
                    0.5}, {-1.8839738363837366`, 0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-1.8919080399086607`, 0.5}, 1.6, {
                    1.5675894900639555`, 1.5528481127860876`}], 
                   ArrowBox[{{-1.8791910313931603`, 
                    2.0999494609813203`}, {-1.8631924392853845`, 
                    2.09974229620925}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Dashing[{Small, Small}], 
                   Arrowheads[0.02], 
                   CircleBox[{-1.8919080399086607`, 0.5}, 1.7, {
                    1.5675894900639555`, 1.5721876713584833`}], 
                   ArrowBox[{{-1.8772735064266353`, 
                    2.1999370077828657`}, {-1.8942733249036223`, 
                    2.199998354536525}}]}, 
                  {RGBColor[0, 0, 1], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-1.8919080399086607`, 0.5}, 1.8, {
                    1.5675894900639555`, 1.576345125033871}], 
                   ArrowBox[{{-1.8838959031965496`, 
                    2.2999821681520367`}, {-1.9018958254860392`, 
                    2.2999722898253907`}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.003], Arrowheads[0.025], 
                   CircleBox[{-1.8919080399086607`, 0.5}, 1.5, {
                    1.5675894900639555`, 1.5675833883985926`}], 
                   ArrowBox[{{-1.9020885541542352`, 
                    1.99996545197871}, {-1.8870886406059695`, 
                    1.9999922577768063`}}]}, 
                  {RGBColor[0, 0, 1], ArrowBox[{{0, 2.8}, {0, 2.8}}]}},
                 Frame->True,
                 FrameTicks->Automatic,
                 ImageSize->800,
                 PlotRange->{{-3, 3}, {-1, 3}}], 584->
                GraphicsBox[{InsetBox[
                   TagBox[GridBox[{
                    {"\<\"t\"\>", "29.150000000000002`"},
                    {"\<\"driveIntent\"\>", "0"},
                    {"\<\"motorTorque\"\>", 
                    RowBox[{"-", "1.5647352466985505`"}]},
                    {"\<\"vw\"\>", 
                    RowBox[{"-", "0.01800052299450854`"}]},
                    {"\<\"controllerState\"\>", 
                    RowBox[{"\[LeftAssociation]", 
                    RowBox[{
                    
                    RowBox[{"\<\"lastError\"\>", "\[Rule]", 
                    "0.0022165886968044148`"}], ",", 
                    RowBox[{"\<\"lastInt\"\>", "\[Rule]", 
                    RowBox[{"-", "0.015422307659830308`"}]}], ",", 
                    RowBox[{"\<\"initialized\"\>", "\[Rule]", "True"}]}], 
                    "\[RightAssociation]"}]}
                    },
                    AutoDelete->False,
                    
                    GridBoxItemSize->{
                    "Columns" -> {{Automatic}}, "Rows" -> {{Automatic}}}],
                    "Grid"], {-1, -0.5}], 
                  {GrayLevel[0], LineBox[{{-5, 0}, {5, 0}}]}, 
                  {GrayLevel[0.5], CircleBox[{-1.892811242910941, 0.5}, 0.5], 
                   DiskBox[{-1.892811242910941, 0.5}, 0.1]}, 
                  {GrayLevel[0.5], 
                   DiskBox[{-1.8880354204606886`, 1.99999239715404}, 0.1]}, 
                  {GrayLevel[0.5], Thickness[0.01], 
                   LineBox[{{-1.892811242910941, 0.5}, {-1.8880354204606886`, 
                    1.99999239715404}}]}, 
                  {RGBColor[0, 1, 0], 
                   DiskBox[{-1.8896273612774395`, 1.4999949314360268`}, 
                    0.05]}, 
                  {RGBColor[0, 0, 1], Arrowheads[0.01], 
                   CircleBox[{-1.892811242910941, 0.5}, 0.3333333333333333, 
                    NCache[{
                    Rational[-1, 2] 
                    Pi, -1.617738384195853}, {-1.5707963267948966`, \
-1.617738384195853}]], 
                   ArrowBox[{{-1.9051224613800268`, 
                    0.16689409340133649`}, {-1.908452849372846, 
                    0.16703385869070414`}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-1.892811242910941, 0.5}, {-1.86151653797697, 
                    0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-1.892811242910941, 0.5}, {-1.9162285338341858`,
                     0.5}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.005], Arrowheads[0.04], 
                   ArrowBox[{{-1.892811242910941, 0.5}, {-1.8849338289002147`,
                     0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-1.892811242910941, 0.5}, 1.6, {
                    1.5676124397821483`, 1.552976558772221}], 
                   ArrowBox[{{-1.8802997415860598`, 
                    2.099951081231735}, {-1.864301123000776, 
                    2.0997459714163087`}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Dashing[{Small, Small}], 
                   Arrowheads[0.02], 
                   CircleBox[{-1.892811242910941, 0.5}, 1.7, {
                    1.5676124397821483`, 1.5721777096740428`}], 
                   ArrowBox[{{-1.8781597751936019`, 
                    2.1999368619139146`}, {-1.8951595930586282`, 
                    2.199998378014398}}]}, 
                  {RGBColor[0, 0, 1], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-1.892811242910941, 0.5}, 1.8, {
                    1.5676124397821483`, 1.576305413374918}], 
                   ArrowBox[{{-1.884727625927183, 
                    2.299981848557496}, {-1.9027275485947643`, 
                    2.2999726850376327`}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.003], Arrowheads[0.025], 
                   CircleBox[{-1.892811242910941, 0.5}, 1.5, {
                    1.5676124397821483`, 1.56760637422703}], 
                   ArrowBox[{{-1.9030262351023641`, 
                    1.999965217574904}, {-1.8880263221742148`, 
                    1.9999923681584328`}}]}, 
                  {RGBColor[0, 0, 1], ArrowBox[{{0, 2.8}, {0, 2.8}}]}},
                 Frame->True,
                 FrameTicks->Automatic,
                 ImageSize->800,
                 PlotRange->{{-3, 3}, {-1, 3}}], 585->
                GraphicsBox[{InsetBox[
                   TagBox[GridBox[{
                    {"\<\"t\"\>", "29.2`"},
                    {"\<\"driveIntent\"\>", "0"},
                    {"\<\"motorTorque\"\>", 
                    RowBox[{"-", "1.553535688356303`"}]},
                    {"\<\"vw\"\>", 
                    RowBox[{"-", "0.01787162630495369`"}]},
                    {"\<\"controllerState\"\>", 
                    RowBox[{"\[LeftAssociation]", 
                    RowBox[{
                    
                    RowBox[{"\<\"lastError\"\>", "\[Rule]", 
                    "0.0022007053283786414`"}], ",", 
                    RowBox[{"\<\"lastInt\"\>", "\[Rule]", 
                    RowBox[{"-", "0.015311883736338976`"}]}], ",", 
                    RowBox[{"\<\"initialized\"\>", "\[Rule]", "True"}]}], 
                    "\[RightAssociation]"}]}
                    },
                    AutoDelete->False,
                    
                    GridBoxItemSize->{
                    "Columns" -> {{Automatic}}, "Rows" -> {{Automatic}}}],
                    "Grid"], {-1, -0.5}], 
                  {GrayLevel[0], LineBox[{{-5, 0}, {5, 0}}]}, 
                  {GrayLevel[0.5], 
                   CircleBox[{-1.8937079783382151`, 0.5}, 0.5], 
                   DiskBox[{-1.8937079783382151`, 0.5}, 0.1]}, 
                  {GrayLevel[0.5], 
                   DiskBox[{-1.8889663376699737`, 1.999992505595869}, 0.1]}, 
                  {GrayLevel[0.5], Thickness[0.01], 
                   LineBox[{{-1.8937079783382151`, 
                    0.5}, {-1.8889663376699737`, 1.999992505595869}}]}, 
                  {RGBColor[0, 1, 0], 
                   DiskBox[{-1.8905468845593874`, 1.4999950037305794`}, 
                    0.05]}, 
                  {RGBColor[0, 0, 1], Arrowheads[0.01], 
                   CircleBox[{-1.8937079783382151`, 0.5}, 0.3333333333333333, 
                    NCache[{
                    Rational[-1, 2] 
                    Pi, -1.6174023974455856`}, {-1.5707963267948966`, \
-1.6174023974455856`}]], 
                   ArrowBox[{{-1.9059072769434648`, 
                    0.1668899757968082}, {-1.9092377117075898`, 
                    0.16702862211206915`}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-1.8937079783382151`, 
                    0.5}, {-1.8626372645710891`, 0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-1.8937079783382151`, 
                    0.5}, {-1.9169576674055298`, 0.5}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.005], Arrowheads[0.04], 
                   ArrowBox[{{-1.8937079783382151`, 
                    0.5}, {-1.8858869536384035`, 0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-1.8937079783382151`, 0.5}, 1.6, {
                    1.5676352277514995`, 1.553104099482996}], 
                   ArrowBox[{{-1.8814005360126465`, 
                    2.099952663944594}, {-1.8654018913976327`, 
                    2.099749594606016}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Dashing[{Small, Small}], 
                   Arrowheads[0.02], 
                   CircleBox[{-1.8937079783382151`, 0.5}, 1.7, {
                    1.5676352277514995`, 1.57216781816451}], 
                   ArrowBox[{{-1.8790396956799182`, 
                    2.1999367169056194`}, {-1.896039512935626, 
                    2.1999984011599603`}}]}, 
                  {RGBColor[0, 0, 1], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-1.8937079783382151`, 0.5}, 1.8, {
                    1.5676352277514995`, 1.5762659815757012`}], 
                   ArrowBox[{{-1.885553384837914, 
                    2.299981528406568}, {-1.9035533078528357`, 
                    2.2999730746560485`}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.003], Arrowheads[0.025], 
                   CircleBox[{-1.8937079783382151`, 0.5}, 1.5, {
                    1.5676352277514995`, 1.5676291979888832`}], 
                   ArrowBox[{{-1.9039572053758886`, 
                    1.999964984039671}, {-1.8889572930713248`, 
                    1.9999924769776332`}}]}, 
                  {RGBColor[0, 0, 1], ArrowBox[{{0, 2.8}, {0, 2.8}}]}},
                 Frame->True,
                 FrameTicks->Automatic,
                 ImageSize->800,
                 PlotRange->{{-3, 3}, {-1, 3}}], 586->
                GraphicsBox[{InsetBox[
                   TagBox[GridBox[{
                    {"\<\"t\"\>", "29.25`"},
                    {"\<\"driveIntent\"\>", "0"},
                    {"\<\"motorTorque\"\>", 
                    RowBox[{"-", "1.5424151545021925`"}]},
                    {"\<\"vw\"\>", 
                    RowBox[{"-", "0.01774365236390842`"}]},
                    {"\<\"controllerState\"\>", 
                    RowBox[{"\[LeftAssociation]", 
                    RowBox[{
                    RowBox[{"\<\"lastError\"\>", "\[Rule]", 
                    "0.002184938010500889`"}], ",", 
                    RowBox[{"\<\"lastInt\"\>", "\[Rule]", 
                    RowBox[{"-", "0.015202251017775687`"}]}], ",", 
                    RowBox[{"\<\"initialized\"\>", "\[Rule]", "True"}]}], 
                    "\[RightAssociation]"}]}
                    },
                    AutoDelete->False,
                    
                    GridBoxItemSize->{
                    "Columns" -> {{Automatic}}, "Rows" -> {{Automatic}}}],
                    "Grid"], {-1, -0.5}], 
                  {GrayLevel[0], LineBox[{{-5, 0}, {5, 0}}]}, 
                  {GrayLevel[0.5], 
                   CircleBox[{-1.8945982924883171`, 0.5}, 0.5], 
                   DiskBox[{-1.8945982924883171`, 0.5}, 0.1]}, 
                  {GrayLevel[0.5], 
                   DiskBox[{-1.889890592414449, 1.999992612501813}, 0.1]}, 
                  {GrayLevel[0.5], Thickness[0.01], 
                   LineBox[{{-1.8945982924883171`, 0.5}, {-1.889890592414449, 
                    1.999992612501813}}]}, 
                  {RGBColor[0, 1, 0], 
                   DiskBox[{-1.891459825772405, 1.4999950750012088`}, 0.05]}, 
                  {RGBColor[0, 0, 1], Arrowheads[0.01], 
                   CircleBox[{-1.8945982924883171`, 0.5}, 0.3333333333333333, 
                    NCache[{
                    Rational[-1, 2] 
                    Pi, -1.6170687814299622`}, {-1.5707963267948966`, \
-1.6170687814299622`}]], 
                   ArrowBox[{{-1.9066864595777007`, 
                    0.16688592445300326`}, {-1.9100169404111178`, 
                    0.16702345967419296`}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-1.8945982924883171`, 
                    0.5}, {-1.8637499893982732`, 0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-1.8945982924883171`, 
                    0.5}, {-1.9176815622673817`, 0.5}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.005], Arrowheads[0.04], 
                   ArrowBox[{{-1.8945982924883171`, 
                    0.5}, {-1.8868332591773378`, 0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-1.8945982924883171`, 0.5}, 1.6, {
                    1.5676578549266593`, 1.5532307402612218`}], 
                   ArrowBox[{{-1.8824934695113857`, 
                    2.0999542097387343`}, {-1.8664947993078052`, 
                    2.099753166482584}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Dashing[{Small, Small}], 
                   Arrowheads[0.02], 
                   CircleBox[{-1.8945982924883171`, 0.5}, 1.7, {
                    1.5676578549266593`, 1.5721579964627088`}], 
                   ArrowBox[{{-1.8799133135592137`, 
                    2.1999365727561284`}, {-1.8969131302082571`, 
                    2.1999984239776023`}}]}, 
                  {RGBColor[0, 0, 1], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-1.8945982924883171`, 0.5}, 1.8, {
                    1.5676578549266593`, 1.5762268280072271`}], 
                   ArrowBox[{{-1.88637322329426, 
                    2.2999812077454456`}, {-1.9043731466263791`, 
                    2.2999734587561504`}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.003], Arrowheads[0.025], 
                   CircleBox[{-1.8945982924883171`, 0.5}, 1.5, {
                    1.5676578549266593`, 1.5676518607210632`}], 
                   ArrowBox[{{-1.9048815128280685`, 
                    1.999964751378993}, {-1.8898816011504214`, 
                    1.9999925842559432`}}]}, 
                  {RGBColor[0, 0, 1], ArrowBox[{{0, 2.8}, {0, 2.8}}]}},
                 Frame->True,
                 FrameTicks->Automatic,
                 ImageSize->800,
                 PlotRange->{{-3, 3}, {-1, 3}}], 587->
                GraphicsBox[{InsetBox[
                   TagBox[GridBox[{
                    {"\<\"t\"\>", "29.3`"},
                    {"\<\"driveIntent\"\>", "0"},
                    {"\<\"motorTorque\"\>", 
                    RowBox[{"-", "1.5313731826750285`"}]},
                    {"\<\"vw\"\>", 
                    RowBox[{"-", "0.01761659465522861`"}]},
                    {"\<\"controllerState\"\>", 
                    RowBox[{"\[LeftAssociation]", 
                    RowBox[{
                    
                    RowBox[{"\<\"lastError\"\>", "\[Rule]", 
                    "0.002169285736557734`"}], ",", 
                    RowBox[{"\<\"lastInt\"\>", "\[Rule]", 
                    RowBox[{"-", "0.015093403727306196`"}]}], ",", 
                    RowBox[{"\<\"initialized\"\>", "\[Rule]", "True"}]}], 
                    "\[RightAssociation]"}]}
                    },
                    AutoDelete->False,
                    
                    GridBoxItemSize->{
                    "Columns" -> {{Automatic}}, "Rows" -> {{Automatic}}}],
                    "Grid"], {-1, -0.5}], 
                  {GrayLevel[0], LineBox[{{-5, 0}, {5, 0}}]}, 
                  {GrayLevel[0.5], 
                   CircleBox[{-1.8954822313323543`, 0.5}, 0.5], 
                   DiskBox[{-1.8954822313323543`, 0.5}, 0.1]}, 
                  {GrayLevel[0.5], 
                   DiskBox[{-1.8908082320876676`, 1.999992717892677}, 0.1]}, 
                  {GrayLevel[0.5], Thickness[0.01], 
                   LineBox[{{-1.8954822313323543`, 
                    0.5}, {-1.8908082320876676`, 1.999992717892677}}]}, 
                  {RGBColor[0, 1, 0], 
                   DiskBox[{-1.8923662318358965`, 1.4999951452617848`}, 
                    0.05]}, 
                  {RGBColor[0, 0, 1], Arrowheads[0.01], 
                   CircleBox[{-1.8954822313323543`, 0.5}, 0.3333333333333333, 
                    NCache[{
                    Rational[-1, 2] 
                    Pi, -1.6167375222751474`}, {-1.5707963267948966`, \
-1.6167375222751474`}]], 
                   ArrowBox[{{-1.9074600506734, 
                    0.16688193841380444`}, {-1.9107905768838864`, 
                    0.16701837037520229`}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-1.8954822313323543`, 
                    0.5}, {-1.8648547676788536`, 0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-1.8954822313323543`, 
                    0.5}, {-1.9184002574172505`, 0.5}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.005], Arrowheads[0.04], 
                   ArrowBox[{{-1.8954822313323543`, 
                    0.5}, {-1.8877727937637498`, 0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-1.8954822313323543`, 0.5}, 1.6, {
                    1.567680322255975, 1.5533564864144418`}], 
                   ArrowBox[{{-1.8835785965377967`, 
                    2.099955719224341}, {-1.8675799011802865`, 
                    2.099756687741009}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Dashing[{Small, Small}], 
                   Arrowheads[0.02], 
                   CircleBox[{-1.8954822313323543`, 0.5}, 1.7, {
                    1.567680322255975, 1.5721482441983883`}], 
                   ArrowBox[{{-1.8807806741731636`, 
                    2.1999364294634947`}, {-1.8977804902182094`, 
                    2.1999984464716706`}}]}, 
                  {RGBColor[0, 0, 1], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-1.8954822313323543`, 0.5}, 1.8, {
                    1.567680322255975, 1.576187951048614}], 
                   ArrowBox[{{-1.8871871843496129`, 
                    2.299980886619509}, {-1.9051871079693863`, 
                    2.2999738374124945`}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.003], Arrowheads[0.025], 
                   CircleBox[{-1.8954822313323543`, 0.5}, 1.5, {
                    1.567680322255975, 1.5676743634442414`}], 
                   ArrowBox[{{-1.9057992049610766`, 
                    1.9999645195987619`}, {-1.8907992939135427`, 
                    1.9999926900145653`}}]}, 
                  {RGBColor[0, 0, 1], ArrowBox[{{0, 2.8}, {0, 2.8}}]}},
                 Frame->True,
                 FrameTicks->Automatic,
                 ImageSize->800,
                 PlotRange->{{-3, 3}, {-1, 3}}], 588->
                GraphicsBox[{InsetBox[
                   TagBox[GridBox[{
                    {"\<\"t\"\>", "29.35`"},
                    {"\<\"driveIntent\"\>", "0"},
                    {"\<\"motorTorque\"\>", 
                    RowBox[{"-", "1.5204093120833384`"}]},
                    {"\<\"vw\"\>", 
                    RowBox[{"-", "0.01749044670013016`"}]},
                    {"\<\"controllerState\"\>", 
                    RowBox[{"\[LeftAssociation]", 
                    RowBox[{
                    
                    RowBox[{"\<\"lastError\"\>", "\[Rule]", 
                    "0.0021537475067359337`"}], ",", 
                    RowBox[{"\<\"lastInt\"\>", "\[Rule]", 
                    RowBox[{"-", "0.014985336138261019`"}]}], ",", 
                    RowBox[{"\<\"initialized\"\>", "\[Rule]", "True"}]}], 
                    "\[RightAssociation]"}]}
                    },
                    AutoDelete->False,
                    
                    GridBoxItemSize->{
                    "Columns" -> {{Automatic}}, "Rows" -> {{Automatic}}}],
                    "Grid"], {-1, -0.5}], 
                  {GrayLevel[0], LineBox[{{-5, 0}, {5, 0}}]}, 
                  {GrayLevel[0.5], 
                   CircleBox[{-1.8963598405165802`, 0.5}, 0.5], 
                   DiskBox[{-1.8963598405165802`, 0.5}, 0.1]}, 
                  {GrayLevel[0.5], 
                   DiskBox[{-1.891719303751773, 1.9999928217890026`}, 0.1]}, 
                  {GrayLevel[0.5], Thickness[0.01], 
                   LineBox[{{-1.8963598405165802`, 0.5}, {-1.891719303751773, 
                    1.9999928217890026`}}]}, 
                  {RGBColor[0, 1, 0], 
                   DiskBox[{-1.8932661493400422`, 1.4999952145260016`}, 
                    0.05]}, 
                  {RGBColor[0, 0, 1], Arrowheads[0.01], 
                   CircleBox[{-1.8963598405165802`, 0.5}, 0.3333333333333333, 
                    NCache[{
                    Rational[-1, 2] 
                    Pi, -1.6164086061573968`}, {-1.5707963267948966`, \
-1.6164086061573968`}]], 
                   ArrowBox[{{-1.908228091312117, 
                    0.16687801673536262`}, {-1.911558662217116, 
                    0.16701335322564898`}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-1.8963598405165802`, 
                    0.5}, {-1.8659516542749135`, 0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-1.8963598405165802`, 
                    0.5}, {-1.9191137915599161`, 0.5}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.005], Arrowheads[0.04], 
                   ArrowBox[{{-1.8963598405165802`, 
                    0.5}, {-1.8887056053182494`, 0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-1.8963598405165802`, 0.5}, 1.6, {
                    1.5677026306834227`, 1.5534813432257364`}], 
                   ArrowBox[{{-1.8846559711836028`, 
                    2.0999571930031866`}, {-1.8686572511003599`, 
                    2.099760159067489}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Dashing[{Small, Small}], 
                   Arrowheads[0.02], 
                   CircleBox[{-1.8963598405165802`, 0.5}, 1.7, {
                    1.5677026306834227`, 1.572138560997657}], 
                   ArrowBox[{{-1.8816418225324023`, 
                    2.199936287025669}, {-1.898641637976128, 
                    2.199998468646473}}]}, 
                  {RGBColor[0, 0, 1], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-1.8963598405165802`, 0.5}, 1.8, {
                    1.5677026306834227`, 1.5761493490838858`}], 
                   ArrowBox[{{-1.8879953107413394`, 
                    2.299980565073312}, {-1.9059952346198152`, 
                    2.2999742106987187`}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.003], Arrowheads[0.025], 
                   CircleBox[{-1.8963598405165802`, 0.5}, 1.5, {
                    1.5677026306834227`, 1.5676967071652452`}], 
                   ArrowBox[{{-1.906710328931466, 
                    1.9999642887047588`}, {-1.8917104185171083`, 
                    1.999992794274383}}]}, 
                  {RGBColor[0, 0, 1], ArrowBox[{{0, 2.8}, {0, 2.8}}]}},
                 Frame->True,
                 FrameTicks->Automatic,
                 ImageSize->800,
                 PlotRange->{{-3, 3}, {-1, 3}}], 589->
                GraphicsBox[{InsetBox[
                   TagBox[GridBox[{
                    {"\<\"t\"\>", "29.400000000000002`"},
                    {"\<\"driveIntent\"\>", "0"},
                    {"\<\"motorTorque\"\>", 
                    RowBox[{"-", "1.509523082805597`"}]},
                    {"\<\"vw\"\>", 
                    RowBox[{"-", "0.01736520205719826`"}]},
                    {"\<\"controllerState\"\>", 
                    RowBox[{"\[LeftAssociation]", 
                    RowBox[{
                    
                    RowBox[{"\<\"lastError\"\>", "\[Rule]", 
                    "0.002138322329706776`"}], ",", 
                    RowBox[{"\<\"lastInt\"\>", "\[Rule]", 
                    RowBox[{"-", "0.01487804257375151`"}]}], ",", 
                    RowBox[{"\<\"initialized\"\>", "\[Rule]", "True"}]}], 
                    "\[RightAssociation]"}]}
                    },
                    AutoDelete->False,
                    GridBoxItemSize->{
                    "Columns" -> {{Automatic}}, "Rows" -> {{Automatic}}}],
                    "Grid"], {-1, -0.5}], 
                  {GrayLevel[0], LineBox[{{-5, 0}, {5, 0}}]}, 
                  {GrayLevel[0.5], 
                   CircleBox[{-1.8972311653642617`, 0.5}, 0.5], 
                   DiskBox[{-1.8972311653642617`, 0.5}, 0.1]}, 
                  {GrayLevel[0.5], 
                   DiskBox[{-1.8926238541418807`, 1.9999929242110777`}, 0.1]}, 
                  {GrayLevel[0.5], Thickness[0.01], 
                   LineBox[{{-1.8972311653642617`, 
                    0.5}, {-1.8926238541418807`, 1.9999929242110777`}}]}, 
                  {RGBColor[0, 1, 0], 
                   DiskBox[{-1.894159624549341, 1.4999952828073853`}, 0.05]}, 
                  {RGBColor[0, 0, 1], Arrowheads[0.01], 
                   CircleBox[{-1.8972311653642617`, 0.5}, 0.3333333333333333, 
                    NCache[{
                    Rational[-1, 2] 
                    Pi, -1.6160820192790644`}, {-1.5707963267948966`, \
-1.6160820192790644`}]], 
                   ArrowBox[{{-1.9089906222601865`, 
                    0.16687415848567977`}, {-1.9123212371866887`, 
                    0.16700840724801308`}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-1.8972311653642617`, 
                    0.5}, {-1.8670407037081498`, 0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-1.8972311653642617`, 0.5}, {-1.919822203096964,
                     0.5}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.005], Arrowheads[0.04], 
                   ArrowBox[{{-1.8972311653642617`, 0.5}, {-1.889631741440852,
                     0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-1.8972311653642617`, 0.5}, 1.6, {
                    1.5677247811502835`, 1.5536053159631076`}], 
                   ArrowBox[{{-1.8857256471936095`, 
                    2.0999586316688394`}, {-1.8697269028064742`, 
                    2.0997635811397735`}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Dashing[{Small, Small}], 
                   Arrowheads[0.02], 
                   CircleBox[{-1.8972311653642617`, 0.5}, 1.7, {
                    1.5677247811502835`, 1.5721289464825376`}], 
                   ArrowBox[{{-1.8824968033176308`, 
                    2.199936145440492}, {-1.8994966181627244`, 
                    2.1999984905062764`}}]}, 
                  {RGBColor[0, 0, 1], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-1.8972311653642617`, 0.5}, 1.8, {
                    1.5677247811502835`, 1.5761110204992035`}], 
                   ArrowBox[{{-1.888797644887667, 
                    2.299980243150566}, {-1.9067975689964753`, 
                    2.299974578687584}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.003], Arrowheads[0.025], 
                   CircleBox[{-1.8972311653642617`, 0.5}, 1.5, {
                    1.5677247811502835`, 1.5677188928792212`}], 
                   ArrowBox[{{-1.9076149315552806`, 
                    1.999964058702638}, {-1.8926150217770312`, 
                    1.9999928970559768`}}]}, 
                  {RGBColor[0, 0, 1], ArrowBox[{{0, 2.8}, {0, 2.8}}]}},
                 Frame->True,
                 FrameTicks->Automatic,
                 ImageSize->800,
                 PlotRange->{{-3, 3}, {-1, 3}}], 590->
                GraphicsBox[{InsetBox[
                   TagBox[GridBox[{
                    {"\<\"t\"\>", "29.45`"},
                    {"\<\"driveIntent\"\>", "0"},
                    {"\<\"motorTorque\"\>", 
                    RowBox[{"-", "1.4987140351110837`"}]},
                    {"\<\"vw\"\>", 
                    RowBox[{"-", "0.017240854322451517`"}]},
                    {"\<\"controllerState\"\>", 
                    RowBox[{"\[LeftAssociation]", 
                    RowBox[{
                    
                    RowBox[{"\<\"lastError\"\>", "\[Rule]", 
                    "0.00212300922408738`"}], ",", 
                    RowBox[{"\<\"lastInt\"\>", "\[Rule]", 
                    RowBox[{"-", "0.014771517406207496`"}]}], ",", 
                    RowBox[{"\<\"initialized\"\>", "\[Rule]", "True"}]}], 
                    "\[RightAssociation]"}]}
                    },
                    AutoDelete->False,
                    
                    GridBoxItemSize->{
                    "Columns" -> {{Automatic}}, "Rows" -> {{Automatic}}}],
                    "Grid"], {-1, -0.5}], 
                  {GrayLevel[0], LineBox[{{-5, 0}, {5, 0}}]}, 
                  {GrayLevel[0.5], 
                   CircleBox[{-1.8980962508775505`, 0.5}, 0.5], 
                   DiskBox[{-1.8980962508775505`, 0.5}, 0.1]}, 
                  {GrayLevel[0.5], 
                   DiskBox[{-1.8935219296701067`, 1.9999930251789477`}, 0.1]}, 
                  {GrayLevel[0.5], Thickness[0.01], 
                   LineBox[{{-1.8980962508775505`, 
                    0.5}, {-1.8935219296701067`, 1.9999930251789477`}}]}, 
                  {RGBColor[0, 1, 0], 
                   DiskBox[{-1.8950467034059213`, 1.4999953501192984`}, 
                    0.05]}, 
                  {RGBColor[0, 0, 1], Arrowheads[0.01], 
                   CircleBox[{-1.8980962508775505`, 0.5}, 0.3333333333333333, 
                    NCache[{
                    Rational[-1, 2] 
                    Pi, -1.615757747848229}, {-1.5707963267948966`, \
-1.615757747848229}]], 
                   ArrowBox[{{-1.9097476839638272`, 
                    0.1668703627442507}, {-1.9130783422482567`, 
                    0.1670035314762769}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-1.8980962508775505`, 
                    0.5}, {-1.8681219701753289`, 0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-1.8980962508775505`, 
                    0.5}, {-1.9205255301180821`, 0.5}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.005], Arrowheads[0.04], 
                   ArrowBox[{{-1.8980962508775505`, 
                    0.5}, {-1.8905512494158605`, 0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-1.8980962508775505`, 0.5}, 1.6, {
                    1.5677467745965812`, 1.5537284098875255`}], 
                   ArrowBox[{{-1.8867876779804456`, 
                    2.099960035806842}, {-1.8707889097049828`, 
                    2.099766954627481}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Dashing[{Small, Small}], 
                   Arrowheads[0.02], 
                   CircleBox[{-1.8980962508775505`, 0.5}, 1.7, {
                    1.5677467745965812`, 1.5721194002706604`}], 
                   ArrowBox[{{-1.883345660880969, 
                    2.199936004705693}, {-1.900345475130129, 
                    2.1999985120553083`}}]}, 
                  {RGBColor[0, 0, 1], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-1.8980962508775505`, 0.5}, 1.8, {
                    1.5677467745965812`, 1.5760729636805317`}], 
                   ArrowBox[{{-1.8895942288853533`, 
                    2.2999799208941316`}, {-1.9075941531966982`, 
                    2.2999749414510013`}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.003], Arrowheads[0.025], 
                   CircleBox[{-1.8980962508775505`, 0.5}, 1.5, {
                    1.5677467745965812`, 1.5677409215716185`}], 
                   ArrowBox[{{-1.9085130593129036`, 
                    1.9999638295979076`}, {-1.8935131501735647`, 
                    1.9999929983796383`}}]}, 
                  {RGBColor[0, 0, 1], ArrowBox[{{0, 2.8}, {0, 2.8}}]}},
                 Frame->True,
                 FrameTicks->Automatic,
                 ImageSize->800,
                 PlotRange->{{-3, 3}, {-1, 3}}], 591->
                GraphicsBox[{InsetBox[
                   TagBox[GridBox[{
                    {"\<\"t\"\>", "29.5`"},
                    {"\<\"driveIntent\"\>", "0"},
                    {"\<\"motorTorque\"\>", 
                    RowBox[{"-", "1.48798170889118`"}]},
                    {"\<\"vw\"\>", 
                    RowBox[{"-", "0.017117397129450946`"}]},
                    {"\<\"controllerState\"\>", 
                    RowBox[{"\[LeftAssociation]", 
                    RowBox[{
                    
                    RowBox[{"\<\"lastError\"\>", "\[Rule]", 
                    "0.0021078072196900615`"}], ",", 
                    RowBox[{"\<\"lastInt\"\>", "\[Rule]", 
                    RowBox[{"-", "0.014665755056847222`"}]}], ",", 
                    RowBox[{"\<\"initialized\"\>", "\[Rule]", "True"}]}], 
                    "\[RightAssociation]"}]}
                    },
                    AutoDelete->False,
                    
                    GridBoxItemSize->{
                    "Columns" -> {{Automatic}}, "Rows" -> {{Automatic}}}],
                    "Grid"], {-1, -0.5}], 
                  {GrayLevel[0], LineBox[{{-5, 0}, {5, 0}}]}, 
                  {GrayLevel[0.5], 
                   CircleBox[{-1.8989551417393538`, 0.5}, 0.5], 
                   DiskBox[{-1.8989551417393538`, 0.5}, 0.1]}, 
                  {GrayLevel[0.5], 
                   DiskBox[{-1.8944135764292571`, 1.9999931247124216`}, 0.1]}, 
                  {GrayLevel[0.5], Thickness[0.01], 
                   LineBox[{{-1.8989551417393538`, 
                    0.5}, {-1.8944135764292571`, 1.9999931247124216`}}]}, 
                  {RGBColor[0, 1, 0], 
                   DiskBox[{-1.8959274315326227`, 1.4999954164749476`}, 
                    0.05]}, 
                  {RGBColor[0, 0, 1], Arrowheads[0.01], 
                   CircleBox[{-1.8989551417393538`, 0.5}, 0.3333333333333333, 
                    NCache[{
                    Rational[-1, 2] 
                    Pi, -1.615435778061632}, {-1.5707963267948966`, \
-1.615435778061632}]], 
                   ArrowBox[{{-1.910499316545347, 
                    0.16686662860175694`}, {-1.9138300175334484`, 
                    0.16699872495556212`}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-1.8989551417393538`, 0.5}, {-1.86919550756153, 
                    0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-1.8989551417393538`, 
                    0.5}, {-1.9212238103940145`, 0.5}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.005], Arrowheads[0.04], 
                   ArrowBox[{{-1.8989551417393538`, 0.5}, {-1.891464176216191,
                     0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-1.8989551417393538`, 0.5}, 1.6, {
                    1.5677686119622951`, 1.5538506302597144`}], 
                   ArrowBox[{{-1.8878421166372885`, 
                    2.09996140599487}, {-1.8718433248828654`, 
                    2.0997702801923595`}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Dashing[{Small, Small}], 
                   Arrowheads[0.02], 
                   CircleBox[{-1.8989551417393538`, 0.5}, 1.7, {
                    1.5677686119622951`, 1.572109921975053}], 
                   ArrowBox[{{-1.8841884392474686`, 
                    2.1999358648188805`}, {-1.9011882529034019`, 
                    2.199998533297758}}]}, 
                  {RGBColor[0, 0, 1], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-1.8989551417393538`, 0.5}, 1.8, {
                    1.5677686119622951`, 1.5760351770116907`}], 
                   ArrowBox[{{-1.890385104508051, 
                    2.2999795983460074`}, {-1.9083850289947022`, 
                    2.2999752990600597`}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.003], Arrowheads[0.025], 
                   CircleBox[{-1.8989551417393538`, 0.5}, 1.5, {
                    1.5677686119622951`, 1.5677627942199421`}], 
                   ArrowBox[{{-1.9094047583535554`, 
                    1.9999636013959192`}, {-1.894404849855803, 
                    1.9999930982653802`}}]}, 
                  {RGBColor[0, 0, 1], ArrowBox[{{0, 2.8}, {0, 2.8}}]}},
                 Frame->True,
                 FrameTicks->Automatic,
                 ImageSize->800,
                 PlotRange->{{-3, 3}, {-1, 3}}], 592->
                GraphicsBox[{InsetBox[
                   TagBox[GridBox[{
                    {"\<\"t\"\>", "29.55`"},
                    {"\<\"driveIntent\"\>", "0"},
                    {"\<\"motorTorque\"\>", 
                    RowBox[{"-", "1.4773256431961768`"}]},
                    {"\<\"vw\"\>", 
                    RowBox[{"-", "0.0169948241494449`"}]},
                    {"\<\"controllerState\"\>", 
                    RowBox[{"\[LeftAssociation]", 
                    RowBox[{
                    
                    RowBox[{"\<\"lastError\"\>", "\[Rule]", 
                    "0.002092715358571731`"}], ",", 
                    RowBox[{"\<\"lastInt\"\>", "\[Rule]", 
                    RowBox[{"-", "0.014560749995090078`"}]}], ",", 
                    RowBox[{"\<\"initialized\"\>", "\[Rule]", "True"}]}], 
                    "\[RightAssociation]"}]}
                    },
                    AutoDelete->False,
                    
                    GridBoxItemSize->{
                    "Columns" -> {{Automatic}}, "Rows" -> {{Automatic}}}],
                    "Grid"], {-1, -0.5}], 
                  {GrayLevel[0], LineBox[{{-5, 0}, {5, 0}}]}, 
                  {GrayLevel[0.5], 
                   CircleBox[{-1.8998078823152194`, 0.5}, 0.5], 
                   DiskBox[{-1.8998078823152194`, 0.5}, 0.1]}, 
                  {GrayLevel[0.5], 
                   DiskBox[{-1.8952988401962163`, 1.9999932228310797`}, 0.1]}, 
                  {GrayLevel[0.5], Thickness[0.01], 
                   LineBox[{{-1.8998078823152194`, 
                    0.5}, {-1.8952988401962163`, 1.9999932228310797`}}]}, 
                  {RGBColor[0, 1, 0], 
                   DiskBox[{-1.896801854235884, 1.4999954818873864`}, 0.05]}, 
                  {RGBColor[0, 0, 1], Arrowheads[0.01], 
                   CircleBox[{-1.8998078823152194`, 0.5}, 0.3333333333333333, 
                    NCache[{
                    Rational[-1, 2] 
                    Pi, -1.6151160960907818`}, {-1.5707963267948966`, \
-1.6151160960907818`}]], 
                   ArrowBox[{{-1.9112455598004146`, 
                    0.16686295515980853`}, {-1.9145763028471448`, 
                    0.16699398674182575`}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-1.8998078823152194`, 
                    0.5}, {-1.8702613694512957`, 0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-1.8998078823152194`, 0.5}, {-1.921917081371077,
                     0.5}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.005], Arrowheads[0.04], 
                   ArrowBox[{{-1.8998078823152194`, 
                    0.5}, {-1.8923705685071535`, 0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-1.8998078823152194`, 0.5}, 1.6, {
                    1.567790294188362, 1.5539719823457618`}], 
                   ArrowBox[{{-1.8888890159487175`, 
                    2.099962742802866}, {-1.8728902011185815`, 
                    2.099773558488512}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Dashing[{Small, Small}], 
                   Arrowheads[0.02], 
                   CircleBox[{-1.8998078823152194`, 0.5}, 1.7, {
                    1.567790294188362, 1.5721005112040438`}], 
                   ArrowBox[{{-1.8850251821168345`, 
                    2.19993572577755}, {-1.902024995182256, 
                    2.199998554237778}}]}, 
                  {RGBColor[0, 0, 1], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-1.8998078823152194`, 0.5}, 1.8, {
                    1.567790294188362, 1.575997658872785}], 
                   ArrowBox[{{-1.891170313205366, 
                    2.299979275547325}, {-1.9091702378406499`, 
                    2.299975651585047}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.003], Arrowheads[0.025], 
                   CircleBox[{-1.8998078823152194`, 0.5}, 1.5, {
                    1.567790294188362, 1.5677845117953357`}], 
                   ArrowBox[{{-1.9102900744995552`, 
                    1.9999633741018514`}, {-1.8952901666459407`, 
                    1.9999931967329492`}}]}, 
                  {RGBColor[0, 0, 1], ArrowBox[{{0, 2.8}, {0, 2.8}}]}},
                 Frame->True,
                 FrameTicks->Automatic,
                 ImageSize->800,
                 PlotRange->{{-3, 3}, {-1, 3}}], 593->
                GraphicsBox[{InsetBox[
                   TagBox[GridBox[{
                    {"\<\"t\"\>", "29.6`"},
                    {"\<\"driveIntent\"\>", "0"},
                    {"\<\"motorTorque\"\>", 
                    RowBox[{"-", "1.4667453758659934`"}]},
                    {"\<\"vw\"\>", 
                    RowBox[{"-", "0.016873129091541424`"}]},
                    {"\<\"controllerState\"\>", 
                    RowBox[{"\[LeftAssociation]", 
                    RowBox[{
                    
                    RowBox[{"\<\"lastError\"\>", "\[Rule]", 
                    "0.0020777326958957777`"}], ",", 
                    RowBox[{"\<\"lastInt\"\>", "\[Rule]", 
                    RowBox[{"-", "0.01445649673792159`"}]}], ",", 
                    RowBox[{"\<\"initialized\"\>", "\[Rule]", "True"}]}], 
                    "\[RightAssociation]"}]}
                    },
                    AutoDelete->False,
                    
                    GridBoxItemSize->{
                    "Columns" -> {{Automatic}}, "Rows" -> {{Automatic}}}],
                    "Grid"], {-1, -0.5}], 
                  {GrayLevel[0], LineBox[{{-5, 0}, {5, 0}}]}, 
                  {GrayLevel[0.5], 
                   CircleBox[{-1.9006545166552204`, 0.5}, 0.5], 
                   DiskBox[{-1.9006545166552204`, 0.5}, 0.1]}, 
                  {GrayLevel[0.5], 
                   DiskBox[{-1.8961777664350412`, 1.9999933195542792`}, 0.1]}, 
                  {GrayLevel[0.5], Thickness[0.01], 
                   LineBox[{{-1.9006545166552204`, 
                    0.5}, {-1.8961777664350412`, 1.9999933195542792`}}]}, 
                  {RGBColor[0, 1, 0], 
                   DiskBox[{-1.8976700165084344`, 1.4999955463695196`}, 
                    0.05]}, 
                  {RGBColor[0, 0, 1], Arrowheads[0.01], 
                   CircleBox[{-1.9006545166552204`, 0.5}, 0.3333333333333333, 
                    NCache[{
                    Rational[-1, 2] 
                    Pi, -1.6147986880708765`}, {-1.5707963267948966`, \
-1.6147986880708765`}]], 
                   ArrowBox[{{-1.9119864531962707`, 
                    0.16685934153072712`}, {-1.9153172376656924`, 
                    0.16698931590160632`}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-1.9006545166552204`, 
                    0.5}, {-1.8713196091379005`, 0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-1.9006545166552204`, 
                    0.5}, {-1.9226053801671101`, 0.5}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.005], Arrowheads[0.04], 
                   ArrowBox[{{-1.9006545166552204`, 
                    0.5}, {-1.8932704726497902`, 0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-1.9006545166552204`, 0.5}, 1.6, {
                    1.5678118222174824`, 1.5540924714216324`}], 
                   ArrowBox[{{-1.8899284283998177`, 
                    2.0999640467931577`}, {-1.8739295908911708`, 
                    2.099776790162586}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Dashing[{Small, Small}], 
                   Arrowheads[0.02], 
                   CircleBox[{-1.9006545166552204`, 0.5}, 1.7, {
                    1.5678118222174824`, 1.5720911675612268`}], 
                   ArrowBox[{{-1.8858559328652476`, 
                    2.199935587579074}, {-1.9028557453428803`, 
                    2.199998574879481}}]}, 
                  {RGBColor[0, 0, 1], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-1.9006545166552204`, 0.5}, 1.8, {
                    1.5678118222174824`, 1.57596040763896}], 
                   ArrowBox[{{-1.8919498961025019`, 
                    2.2999789525383436`}, {-1.9099498208602939`, 
                    2.2999759990954702`}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.003], Arrowheads[0.025], 
                   CircleBox[{-1.9006545166552204`, 0.5}, 1.5, {
                    1.5678118222174824`, 1.5678060752639262`}], 
                   ArrowBox[{{-1.911169053250221, 
                    1.9999631477207007`}, {-1.8961691460431733`, 
                    1.9999932938018332`}}]}, 
                  {RGBColor[0, 0, 1], ArrowBox[{{0, 2.8}, {0, 2.8}}]}},
                 Frame->True,
                 FrameTicks->Automatic,
                 ImageSize->800,
                 PlotRange->{{-3, 3}, {-1, 3}}], 594->
                GraphicsBox[{InsetBox[
                   TagBox[GridBox[{
                    {"\<\"t\"\>", "29.650000000000002`"},
                    {"\<\"driveIntent\"\>", "0"},
                    {"\<\"motorTorque\"\>", 
                    RowBox[{"-", "1.4562404432528973`"}]},
                    {"\<\"vw\"\>", 
                    RowBox[{"-", "0.016752305702900174`"}]},
                    {"\<\"controllerState\"\>", 
                    RowBox[{"\[LeftAssociation]", 
                    RowBox[{
                    
                    RowBox[{"\<\"lastError\"\>", "\[Rule]", 
                    "0.0020628583006191757`"}], ",", 
                    RowBox[{"\<\"lastInt\"\>", "\[Rule]", 
                    RowBox[{"-", "0.014352989849219894`"}]}], ",", 
                    RowBox[{"\<\"initialized\"\>", "\[Rule]", "True"}]}], 
                    "\[RightAssociation]"}]}
                    },
                    AutoDelete->False,
                    
                    GridBoxItemSize->{
                    "Columns" -> {{Automatic}}, "Rows" -> {{Automatic}}}],
                    "Grid"], {-1, -0.5}], 
                  {GrayLevel[0], LineBox[{{-5, 0}, {5, 0}}]}, 
                  {GrayLevel[0.5], CircleBox[{-1.901495088495856, 0.5}, 0.5], 
                   DiskBox[{-1.901495088495856, 0.5}, 0.1]}, 
                  {GrayLevel[0.5], 
                   DiskBox[{-1.897050400299801, 1.9999934149011587`}, 0.1]}, 
                  {GrayLevel[0.5], Thickness[0.01], 
                   LineBox[{{-1.901495088495856, 0.5}, {-1.897050400299801, 
                    1.9999934149011587`}}]}, 
                  {RGBColor[0, 1, 0], 
                   DiskBox[{-1.8985319630318194`, 1.4999956099341059`}, 
                    0.05]}, 
                  {RGBColor[0, 0, 1], Arrowheads[0.01], 
                   CircleBox[{-1.901495088495856, 0.5}, 0.3333333333333333, 
                    NCache[{
                    Rational[-1, 2] 
                    Pi, -1.6144835400924835`}, {-1.5707963267948966`, \
-1.6144835400924835`}]], 
                   ArrowBox[{{-1.9127220358708727`, 
                    0.16685578683736724`}, {-1.9160528611360503`, 
                    0.16698471151181737`}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-1.901495088495856, 0.5}, {-1.872370279630798, 
                    0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-1.901495088495856, 0.5}, {-1.9232887435687747`,
                     0.5}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.005], Arrowheads[0.04], 
                   ArrowBox[{{-1.901495088495856, 0.5}, {-1.8941639347037167`,
                     0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-1.901495088495856, 0.5}, 1.6, {
                    1.567833196994747, 1.5542121027766718`}], 
                   ArrowBox[{{-1.8909604061836829`, 
                    2.0999653185205553`}, {-1.8749615463877565`, 
                    2.099779975853929}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Dashing[{Small, Small}], 
                   Arrowheads[0.02], 
                   CircleBox[{-1.901495088495856, 0.5}, 1.7, {
                    1.567833196994747, 1.5720818906455423`}], 
                   ArrowBox[{{-1.8866807345474048`, 
                    2.1999354502207105`}, {-1.9036805464399784`, 
                    2.1999985952269414`}}]}, 
                  {RGBColor[0, 0, 1], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-1.901495088495856, 0.5}, 1.8, {
                    1.567833196994747, 1.5759234216794853`}], 
                   ArrowBox[{{-1.8927238940005124`, 
                    2.29997862935845}, {-1.910723818855229, 
                    2.2999763416600656`}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.003], Arrowheads[0.025], 
                   CircleBox[{-1.901495088495856, 0.5}, 1.5, {
                    1.567833196994747, 1.5678274855880538`}], 
                   ArrowBox[{{-1.9120417397856162`, 
                    1.999962922257271}, {-1.897041833227444, 
                    1.999993389491272}}]}, 
                  {RGBColor[0, 0, 1], ArrowBox[{{0, 2.8}, {0, 2.8}}]}},
                 Frame->True,
                 FrameTicks->Automatic,
                 ImageSize->800,
                 PlotRange->{{-3, 3}, {-1, 3}}], 595->
                GraphicsBox[{InsetBox[
                   TagBox[GridBox[{
                    {"\<\"t\"\>", "29.7`"},
                    {"\<\"driveIntent\"\>", "0"},
                    {"\<\"motorTorque\"\>", 
                    RowBox[{"-", "1.44581038002158`"}]},
                    {"\<\"vw\"\>", 
                    RowBox[{"-", "0.01663234776893711`"}]},
                    {"\<\"controllerState\"\>", 
                    RowBox[{"\[LeftAssociation]", 
                    RowBox[{
                    
                    RowBox[{"\<\"lastError\"\>", "\[Rule]", 
                    "0.00204809125601802`"}], ",", 
                    RowBox[{"\<\"lastInt\"\>", "\[Rule]", 
                    RowBox[{"-", "0.01425022393905202`"}]}], ",", 
                    RowBox[{"\<\"initialized\"\>", "\[Rule]", "True"}]}], 
                    "\[RightAssociation]"}]}
                    },
                    AutoDelete->False,
                    
                    GridBoxItemSize->{
                    "Columns" -> {{Automatic}}, "Rows" -> {{Automatic}}}],
                    "Grid"], {-1, -0.5}], 
                  {GrayLevel[0], LineBox[{{-5, 0}, {5, 0}}]}, 
                  {GrayLevel[0.5], 
                   CircleBox[{-1.9023296412619572`, 0.5}, 0.5], 
                   DiskBox[{-1.9023296412619572`, 0.5}, 0.1]}, 
                  {GrayLevel[0.5], 
                   DiskBox[{-1.8979167866371756`, 1.999993508890642}, 0.1]}, 
                  {GrayLevel[0.5], Thickness[0.01], 
                   LineBox[{{-1.9023296412619572`, 
                    0.5}, {-1.8979167866371756`, 1.999993508890642}}]}, 
                  {RGBColor[0, 1, 0], 
                   DiskBox[{-1.8993877381787694`, 1.4999956725937613`}, 
                    0.05]}, 
                  {RGBColor[0, 0, 1], Arrowheads[0.01], 
                   CircleBox[{-1.9023296412619572`, 0.5}, 0.3333333333333333, 
                    NCache[{
                    Rational[-1, 2] 
                    Pi, -1.614170638195544}, {-1.5707963267948966`, \
-1.614170638195544}]], 
                   ArrowBox[{{-1.9134523466328204`, 
                    0.16685229021296866`}, {-1.916783212075716, 
                    0.1669801726595806}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-1.9023296412619572`, 
                    0.5}, {-1.8734134336615256`, 0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-1.9023296412619572`, 
                    0.5}, {-1.9239672080300683`, 0.5}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.005], Arrowheads[0.04], 
                   ArrowBox[{{-1.9023296412619572`, 
                    0.5}, {-1.8950510004296368`, 0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-1.9023296412619572`, 0.5}, 1.6, {
                    1.5678544194680981`, 1.5543308817161872`}], 
                   ArrowBox[{{-1.8919850012074466`, 
                    2.099966558532441}, {-1.8759861195095744`, 
                    2.099783116194718}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Dashing[{Small, Small}], 
                   Arrowheads[0.02], 
                   CircleBox[{-1.9023296412619572`, 0.5}, 1.7, {
                    1.5678544194680981`, 1.572072680051385}], 
                   ArrowBox[{{-1.887499629898609, 
                    2.1999353136996014`}, {-1.9044994412088585`, 
                    2.1999986152841977`}}]}, 
                  {RGBColor[0, 0, 1], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-1.9023296412619572`, 0.5}, 1.8, {
                    1.5678544194680981`, 1.575886699357107}], 
                   ArrowBox[{{-1.8934923473770415`, 
                    2.2999783060461567`}, {-1.9114922723036307`, 
                    2.299976679346817}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.003], Arrowheads[0.025], 
                   CircleBox[{-1.9023296412619572`, 0.5}, 1.5, {
                    1.5678544194680981`, 1.567848743727262}], 
                   ArrowBox[{{-1.9129081789699405`, 
                    1.9999626977161669`}, {-1.897908273062834, 
                    1.9999934838202624`}}]}, 
                  {RGBColor[0, 0, 1], ArrowBox[{{0, 2.8}, {0, 2.8}}]}},
                 Frame->True,
                 FrameTicks->Automatic,
                 ImageSize->800,
                 PlotRange->{{-3, 3}, {-1, 3}}], 596->
                GraphicsBox[{InsetBox[
                   TagBox[GridBox[{
                    {"\<\"t\"\>", "29.75`"},
                    {"\<\"driveIntent\"\>", "0"},
                    {"\<\"motorTorque\"\>", 
                    RowBox[{"-", "1.4354547190257074`"}]},
                    {"\<\"vw\"\>", 
                    RowBox[{"-", "0.01651324911353586`"}]},
                    {"\<\"controllerState\"\>", 
                    RowBox[{"\[LeftAssociation]", 
                    RowBox[{
                    
                    RowBox[{"\<\"lastError\"\>", "\[Rule]", 
                    "0.0020334306600645617`"}], ",", 
                    RowBox[{"\<\"lastInt\"\>", "\[Rule]", 
                    RowBox[{"-", "0.014148193662947714`"}]}], ",", 
                    RowBox[{"\<\"initialized\"\>", "\[Rule]", "True"}]}], 
                    "\[RightAssociation]"}]}
                    },
                    AutoDelete->False,
                    
                    GridBoxItemSize->{
                    "Columns" -> {{Automatic}}, "Rows" -> {{Automatic}}}],
                    "Grid"], {-1, -0.5}], 
                  {GrayLevel[0], LineBox[{{-5, 0}, {5, 0}}]}, 
                  {GrayLevel[0.5], 
                   CircleBox[{-1.9031582180686073`, 0.5}, 0.5], 
                   DiskBox[{-1.9031582180686073`, 0.5}, 0.1]}, 
                  {GrayLevel[0.5], 
                   DiskBox[{-1.8987769699888402`, 1.999993601541441}, 0.1]}, 
                  {GrayLevel[0.5], Thickness[0.01], 
                   LineBox[{{-1.9031582180686073`, 
                    0.5}, {-1.8987769699888402`, 1.999993601541441}}]}, 
                  {RGBColor[0, 1, 0], 
                   DiskBox[{-1.9002373860154291`, 1.4999957343609607`}, 
                    0.05]}, 
                  {RGBColor[0, 0, 1], Arrowheads[0.01], 
                   CircleBox[{-1.9031582180686073`, 0.5}, 0.3333333333333333, 
                    NCache[{
                    Rational[-1, 2] 
                    Pi, -1.6138599683656678`}, {-1.5707963267948966`, \
-1.6138599683656678`}]], 
                   ArrowBox[{{-1.9141774239620541`, 
                    0.16684885080103817`}, {-1.9175083289734265`, 
                    0.16697569844209492`}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-1.9031582180686073`, 
                    0.5}, {-1.8744491236880931`, 0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-1.9031582180686073`, 0.5}, {-1.924640809671959,
                     0.5}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.005], Arrowheads[0.04], 
                   ArrowBox[{{-1.9031582180686073`, 
                    0.5}, {-1.8959317152914448`, 0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-1.9031582180686073`, 0.5}, 1.6, {
                    1.5678754905886396`, 1.5544488135631873`}], 
                   ArrowBox[{{-1.8930022650969816`, 
                    2.0999677673688426`}, {-1.8770033618766713`, 
                    2.099786211810059}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Dashing[{Small, Small}], 
                   Arrowheads[0.02], 
                   CircleBox[{-1.9031582180686073`, 0.5}, 1.7, {
                    1.5678754905886396`, 1.5720635353688095`}], 
                   ArrowBox[{{-1.888312661337037, 
                    2.199935178012776}, {-1.9053124720677026`, 
                    2.1999986350552483`}}]}, 
                  {RGBColor[0, 0, 1], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-1.9031582180686073`, 0.5}, 1.8, {
                    1.5678754905886396`, 1.5758502390276714`}], 
                   ArrowBox[{{-1.8942552963875647`, 
                    2.2999779826391045`}, {-1.9122552213614998`, 
                    2.299977012222959}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.003], Arrowheads[0.025], 
                   CircleBox[{-1.9031582180686073`, 0.5}, 1.5, {
                    1.5678754905886396`, 1.5678698506391875`}], 
                   ArrowBox[{{-1.913768415354784, 
                    1.9999624741017852`}, {-1.8987685101008185`, 
                    1.9999935768075667`}}]}, 
                  {RGBColor[0, 0, 1], ArrowBox[{{0, 2.8}, {0, 2.8}}]}},
                 Frame->True,
                 FrameTicks->Automatic,
                 ImageSize->800,
                 PlotRange->{{-3, 3}, {-1, 3}}], 597->
                GraphicsBox[{InsetBox[
                   TagBox[GridBox[{
                    {"\<\"t\"\>", "29.8`"},
                    {"\<\"driveIntent\"\>", "0"},
                    {"\<\"motorTorque\"\>", 
                    RowBox[{"-", "1.4251729912479594`"}]},
                    {"\<\"vw\"\>", 
                    RowBox[{"-", "0.01639500359925961`"}]},
                    {"\<\"controllerState\"\>", 
                    RowBox[{"\[LeftAssociation]", 
                    RowBox[{
                    
                    RowBox[{"\<\"lastError\"\>", "\[Rule]", 
                    "0.002018875625668645`"}], ",", 
                    RowBox[{"\<\"lastInt\"\>", "\[Rule]", 
                    RowBox[{"-", "0.01404689372115794`"}]}], ",", 
                    RowBox[{"\<\"initialized\"\>", "\[Rule]", "True"}]}], 
                    "\[RightAssociation]"}]}
                    },
                    AutoDelete->False,
                    
                    GridBoxItemSize->{
                    "Columns" -> {{Automatic}}, "Rows" -> {{Automatic}}}],
                    "Grid"], {-1, -0.5}], 
                  {GrayLevel[0], LineBox[{{-5, 0}, {5, 0}}]}, 
                  {GrayLevel[0.5], CircleBox[{-1.903980861723067, 0.5}, 0.5], 
                   DiskBox[{-1.903980861723067, 0.5}, 0.1]}, 
                  {GrayLevel[0.5], 
                   DiskBox[{-1.8996309945936556`, 1.999993692872059}, 0.1]}, 
                  {GrayLevel[0.5], Thickness[0.01], 
                   LineBox[{{-1.903980861723067, 0.5}, {-1.8996309945936556`, 
                    1.999993692872059}}]}, 
                  {RGBColor[0, 1, 0], 
                   DiskBox[{-1.9010809503034594`, 1.4999957952480392`}, 
                    0.05]}, 
                  {RGBColor[0, 0, 1], Arrowheads[0.01], 
                   CircleBox[{-1.903980861723067, 0.5}, 0.3333333333333333, 
                    NCache[{
                    Rational[-1, 2] 
                    Pi, -1.6135515165323353`}, {-1.5707963267948966`, \
-1.6135515165323353`}]], 
                   ArrowBox[{{-1.9148973060111985`, 
                    0.16684546775525444`}, {-1.9182282499905021`, 
                    0.16697128796653576`}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-1.903980861723067, 0.5}, {-1.8754774018981077`,
                     0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-1.903980861723067, 0.5}, {-1.9253095842830223`,
                     0.5}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.005], Arrowheads[0.04], 
                   ArrowBox[{{-1.903980861723067, 0.5}, {-1.8968061244580632`,
                     0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-1.903980861723067, 0.5}, 1.6, {
                    1.5678964113108127`, 1.5545659036593624`}], 
                   ArrowBox[{{-1.894012249200391, 
                    2.099968945562498}, {-1.8780133248313962`, 
                    2.099789263318072}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Dashing[{Small, Small}], 
                   Arrowheads[0.02], 
                   CircleBox[{-1.903980861723067, 0.5}, 1.7, {
                    1.5678964113108127`, 1.5720544561837566`}], 
                   ArrowBox[{{-1.8891198709660495`, 
                    2.199935043157155}, {-1.9061196811198764`, 
                    2.199998654544052}}]}, 
                  {RGBColor[0, 0, 1], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-1.903980861723067, 0.5}, 1.8, {
                    1.5678964113108127`, 1.5758140390399682`}], 
                   ArrowBox[{{-1.8950127808670343`, 
                    2.2999776591740684`}, {-1.9130127058643045`, 
                    2.299977340354986}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.003], Arrowheads[0.025], 
                   CircleBox[{-1.903980861723067, 0.5}, 1.5, {
                    1.5678964113108127`, 1.567890807280255}], 
                   ArrowBox[{{-1.914622493182096, 
                    1.9999622514183115`}, {-1.8996225885832327`, 
                    1.9999936684717168`}}]}, 
                  {RGBColor[0, 0, 1], ArrowBox[{{0, 2.8}, {0, 2.8}}]}},
                 Frame->True,
                 FrameTicks->Automatic,
                 ImageSize->800,
                 PlotRange->{{-3, 3}, {-1, 3}}], 598->
                GraphicsBox[{InsetBox[
                   TagBox[GridBox[{
                    {"\<\"t\"\>", "29.85`"},
                    {"\<\"driveIntent\"\>", "0"},
                    {"\<\"motorTorque\"\>", 
                    RowBox[{"-", "1.4149647258002371`"}]},
                    {"\<\"vw\"\>", 
                    RowBox[{"-", "0.016277605127559155`"}]},
                    {"\<\"controllerState\"\>", 
                    RowBox[{"\[LeftAssociation]", 
                    RowBox[{
                    
                    RowBox[{"\<\"lastError\"\>", "\[Rule]", 
                    "0.0020044252807965763`"}], ",", 
                    RowBox[{"\<\"lastInt\"\>", "\[Rule]", 
                    RowBox[{"-", "0.01394631885790449`"}]}], ",", 
                    RowBox[{"\<\"initialized\"\>", "\[Rule]", "True"}]}], 
                    "\[RightAssociation]"}]}
                    },
                    AutoDelete->False,
                    
                    GridBoxItemSize->{
                    "Columns" -> {{Automatic}}, "Rows" -> {{Automatic}}}],
                    "Grid"], {-1, -0.5}], 
                  {GrayLevel[0], LineBox[{{-5, 0}, {5, 0}}]}, 
                  {GrayLevel[0.5], 
                   CircleBox[{-1.9047976147267178`, 0.5}, 0.5], 
                   DiskBox[{-1.9047976147267178`, 0.5}, 0.1]}, 
                  {GrayLevel[0.5], 
                   DiskBox[{-1.9004789043896915`, 1.9999937829007908`}, 0.1]}, 
                  {GrayLevel[0.5], Thickness[0.01], 
                   LineBox[{{-1.9047976147267178`, 
                    0.5}, {-1.9004789043896915`, 1.9999937829007908`}}]}, 
                  {RGBColor[0, 1, 0], 
                   DiskBox[{-1.9019184745020337`, 1.499995855267194}, 0.05]}, 
                  {RGBColor[0, 0, 1], Arrowheads[0.01], 
                   CircleBox[{-1.9047976147267178`, 0.5}, 0.3333333333333333, 
                    NCache[{
                    Rational[-1, 2] 
                    Pi, -1.6132452685689036`}, {-1.5707963267948966`, \
-1.6132452685689036`}]], 
                   ArrowBox[{{-1.9156120306075215`, 
                    0.16684214023939326`}, {-1.9189430129628071`, 
                    0.1669669403499804}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-1.9047976147267178`, 0.5}, {-1.876498320210713,
                     0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-1.9047976147267178`, 0.5}, {-1.925973567320988,
                     0.5}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.005], Arrowheads[0.04], 
                   ArrowBox[{{-1.9047976147267178`, 
                    0.5}, {-1.8976742728049831`, 0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-1.9047976147267178`, 0.5}, 1.6, {
                    1.56791718259245, 1.5546821573653882`}], 
                   ArrowBox[{{-1.8950150045904333`, 
                    2.099970093638916}, {-1.8790160594408236`, 
                    2.0997922713299504`}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Dashing[{Small, Small}], 
                   Arrowheads[0.02], 
                   CircleBox[{-1.9047976147267178`, 0.5}, 1.7, {
                    1.56791718259245, 1.5720454420783412`}], 
                   ArrowBox[{{-1.889921300576626, 
                    2.1999349091295546`}, {-1.9069211101563623`, 
                    2.19999867375453}}]}, 
                  {RGBColor[0, 0, 1], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-1.9047976147267178`, 0.5}, 1.8, {
                    1.56791718259245, 1.5757780977357847`}], 
                   ArrowBox[{{-1.8957648403319196`, 
                    2.2999773356869615`}, {-1.9137647653290226`, 
                    2.2999776638086584`}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.003], Arrowheads[0.025], 
                   CircleBox[{-1.9047976147267178`, 0.5}, 1.5, {
                    1.56791718259245, 1.5679116146062622`}], 
                   ArrowBox[{{-1.9154704563870049`, 
                    1.9999620296697163`}, {-1.900470552445093, 
                    1.9999937588310195`}}]}, 
                  {RGBColor[0, 0, 1], ArrowBox[{{0, 2.8}, {0, 2.8}}]}},
                 Frame->True,
                 FrameTicks->Automatic,
                 ImageSize->800,
                 PlotRange->{{-3, 3}, {-1, 3}}], 599->
                GraphicsBox[{InsetBox[
                   TagBox[GridBox[{
                    {"\<\"t\"\>", "29.900000000000002`"},
                    {"\<\"driveIntent\"\>", "0"},
                    {"\<\"motorTorque\"\>", 
                    RowBox[{"-", "1.4048294499747727`"}]},
                    {"\<\"vw\"\>", 
                    RowBox[{"-", "0.016161047638972383`"}]},
                    {"\<\"controllerState\"\>", 
                    RowBox[{"\[LeftAssociation]", 
                    RowBox[{
                    
                    RowBox[{"\<\"lastError\"\>", "\[Rule]", 
                    "0.00199007876847984`"}], ",", 
                    RowBox[{"\<\"lastInt\"\>", "\[Rule]", 
                    RowBox[{"-", "0.013846463860626499`"}]}], ",", 
                    RowBox[{"\<\"initialized\"\>", "\[Rule]", "True"}]}], 
                    "\[RightAssociation]"}]}
                    },
                    AutoDelete->False,
                    
                    GridBoxItemSize->{
                    "Columns" -> {{Automatic}}, "Rows" -> {{Automatic}}}],
                    "Grid"], {-1, -0.5}], 
                  {GrayLevel[0], LineBox[{{-5, 0}, {5, 0}}]}, 
                  {GrayLevel[0.5], 
                   CircleBox[{-1.9056085192770096`, 0.5}, 0.5], 
                   DiskBox[{-1.9056085192770096`, 0.5}, 0.1]}, 
                  {GrayLevel[0.5], 
                   DiskBox[{-1.901320743016094, 1.9999938716457266`}, 0.1]}, 
                  {GrayLevel[0.5], Thickness[0.01], 
                   LineBox[{{-1.9056085192770096`, 0.5}, {-1.901320743016094, 
                    1.9999938716457266`}}]}, 
                  {RGBColor[0, 1, 0], 
                   DiskBox[{-1.9027500017697325`, 1.4999959144304844`}, 
                    0.05]}, 
                  {RGBColor[0, 0, 1], Arrowheads[0.01], 
                   CircleBox[{-1.9056085192770096`, 0.5}, 0.3333333333333333, 
                    NCache[{
                    Rational[-1, 2] 
                    Pi, -1.6129412102941398`}, {-1.5707963267948966`, \
-1.6129412102941398`}]], 
                   ArrowBox[{{-1.916321635255407, 
                    0.1668388674272695}, {-1.9196526554032214`, 
                    0.166962654719355}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-1.9056085192770096`, 0.5}, {-1.877511930277514,
                     0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-1.9056085192770096`, 
                    0.5}, {-1.9266327939150765`, 0.5}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.005], Arrowheads[0.04], 
                   ArrowBox[{{-1.9056085192770096`, 0.5}, {-1.898536204915581,
                     0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-1.9056085192770096`, 0.5}, 1.6, {
                    1.5679378053947226`, 1.5547975800606268`}], 
                   ArrowBox[{{-1.8960105820659878`, 
                    2.0999712121164222`}, {-1.8800116164982204`, 
                    2.099795236450007}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Dashing[{Small, Small}], 
                   Arrowheads[0.02], 
                   CircleBox[{-1.9056085192770096`, 0.5}, 1.7, {
                    1.5679378053947226`, 1.572036492631164}], 
                   ArrowBox[{{-1.8907169916498399`, 
                    2.1999347759266907`}, {-1.9077168006582375`, 
                    2.1999986926905613`}}]}, 
                  {RGBColor[0, 0, 1], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-1.9056085192770096`, 0.5}, 1.8, {
                    1.5679378053947226`, 1.5757424134501379`}], 
                   ArrowBox[{{-1.8965115139825717`, 
                    2.2999770122128433`}, {-1.9145114389565057`, 
                    2.2999779826490045`}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.003], Arrowheads[0.025], 
                   CircleBox[{-1.9056085192770096`, 0.5}, 1.5, {
                    1.5679378053947226`, 1.5679322735728192`}], 
                   ArrowBox[{{-1.9163123486004245`, 
                    1.9999618088597506`}, {-1.9013124453172054`, 
                    1.9999938479035613`}}]}, 
                  {RGBColor[0, 0, 1], ArrowBox[{{0, 2.8}, {0, 2.8}}]}},
                 Frame->True,
                 FrameTicks->Automatic,
                 ImageSize->800,
                 PlotRange->{{-3, 3}, {-1, 3}}], 600->
                GraphicsBox[{InsetBox[
                   TagBox[GridBox[{
                    {"\<\"t\"\>", "29.95`"},
                    {"\<\"driveIntent\"\>", "0"},
                    {"\<\"motorTorque\"\>", 
                    RowBox[{"-", "1.3947666893405601`"}]},
                    {"\<\"vw\"\>", 
                    RowBox[{"-", "0.01604532511331164`"}]},
                    {"\<\"controllerState\"\>", 
                    RowBox[{"\[LeftAssociation]", 
                    RowBox[{
                    RowBox[{"\<\"lastError\"\>", "\[Rule]", 
                    "0.0019758352467259053`"}], ",", 
                    RowBox[{"\<\"lastInt\"\>", "\[Rule]", 
                    RowBox[{"-", "0.01374732355922908`"}]}], ",", 
                    RowBox[{"\<\"initialized\"\>", "\[Rule]", "True"}]}], 
                    "\[RightAssociation]"}]}
                    },
                    AutoDelete->False,
                    
                    GridBoxItemSize->{
                    "Columns" -> {{Automatic}}, "Rows" -> {{Automatic}}}],
                    "Grid"], {-1, -0.5}], 
                  {GrayLevel[0], LineBox[{{-5, 0}, {5, 0}}]}, 
                  {GrayLevel[0.5], CircleBox[{-1.906413617269418, 0.5}, 0.5], 
                   DiskBox[{-1.906413617269418, 0.5}, 0.1]}, 
                  {GrayLevel[0.5], 
                   DiskBox[{-1.9021565538148233`, 1.9999939591247504`}, 0.1]}, 
                  {GrayLevel[0.5], Thickness[0.01], 
                   LineBox[{{-1.906413617269418, 0.5}, {-1.9021565538148233`, 
                    1.9999939591247504`}}]}, 
                  {RGBColor[0, 1, 0], 
                   DiskBox[{-1.9035755749663548`, 1.4999959727498338`}, 
                    0.05]}, 
                  {RGBColor[0, 0, 1], Arrowheads[0.01], 
                   CircleBox[{-1.906413617269418, 0.5}, 0.3333333333333333, 
                    NCache[{
                    Rational[-1, 2] 
                    Pi, -1.6126393274751134`}, {-1.5707963267948966`, \
-1.6126393274751134`}]], 
                   ArrowBox[{{-1.9170261571392915`, 
                    0.1668356485026925}, {-1.9203572145045786`, 
                    0.16695843021140033`}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-1.906413617269418, 0.5}, {-1.8785182834826066`,
                     0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   ArrowBox[{{-1.906413617269418, 0.5}, {-1.9272872988690426`,
                     0.5}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.005], Arrowheads[0.04], 
                   ArrowBox[{{-1.906413617269418, 0.5}, {-1.8993919650822313`,
                     0.5}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-1.906413617269418, 0.5}, 1.6, {
                    1.5679582806819918`, 1.5549121771423038`}], 
                   ArrowBox[{{-1.896999032152693, 
                    2.0999723015062104`}, {-1.881000046523681, 
                    2.0997981592757107`}}]}, 
                  {RGBColor[1, 0, 0], Opacity[0.5], Dashing[{Small, Small}], 
                   Arrowheads[0.02], 
                   CircleBox[{-1.906413617269418, 0.5}, 1.7, {
                    1.5679582806819918`, 1.5720276074176542`}], 
                   ArrowBox[{{-1.8915069853593993`, 
                    2.199934643545186}, {-1.9085067937992117`, 
                    2.1999987113559865`}}]}, 
                  {RGBColor[0, 0, 1], Opacity[0.5], Arrowheads[0.02], 
                   CircleBox[{-1.906413617269418, 0.5}, 1.8, {
                    1.5679582806819918`, 1.5757069845116616`}], 
                   ArrowBox[{{-1.8972528407058782`, 
                    2.299976688785928}, {-1.9152527656341338`, 
                    2.2999782969403233`}}]}, 
                  {RGBColor[1, 0, 0], Thickness[0.003], Arrowheads[0.025], 
                   CircleBox[{-1.906413617269418, 0.5}, 1.5, {
                    1.5679582806819918`, 1.5679527851356792`}], 
                   ArrowBox[{{-1.917148213151509, 
                    1.9999615889919475`}, {-1.9021483105286163`, 
                    1.9999939357072103`}}]}, 
                  {RGBColor[0, 0, 1], ArrowBox[{{0, 2.8}, {0, 2.8}}]}},
                 Frame->True,
                 FrameTicks->Automatic,
                 ImageSize->800,
                 PlotRange->{{-3, 3}, {-1, 3}}]}, Dynamic[$CellContext`i3$$],
                Alignment->Automatic,
                ImageSize->All,
                TransitionDirection->Horizontal,
                TransitionDuration->0.5,
                TransitionEffect->Automatic],
               Identity,
               Editable->True,
               Selectable->True],
              ImageMargins->10],
             Deployed->False,
             StripOnInput->False,
             ScriptLevel->0,
             GraphicsBoxOptions->{PreserveImageOptions->True},
             Graphics3DBoxOptions->{PreserveImageOptions->True}],
            Identity,
            Editable->False,
            Selectable->False],
           Alignment->{Left, Center},
           Background->GrayLevel[1],
           Frame->1,
           FrameStyle->GrayLevel[0, 0.2],
           ItemSize->Automatic,
           StripOnInput->False]}
        },
        AutoDelete->False,
        GridBoxAlignment->{
         "Columns" -> {{Left}}, "ColumnsIndexed" -> {}, "Rows" -> {{Top}}, 
          "RowsIndexed" -> {}},
        GridBoxDividers->{
         "Columns" -> {{False}}, "ColumnsIndexed" -> {}, "Rows" -> {{False}}, 
          "RowsIndexed" -> {}},
        GridBoxItemSize->{"Columns" -> {{Automatic}}, "Rows" -> {{Automatic}}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.7], {
             Offset[0.5599999999999999]}, 
            Offset[0.7]}, "ColumnsIndexed" -> {}, "Rows" -> {
            Offset[0.4], {
             Offset[0.8]}, 
            Offset[0.4]}, "RowsIndexed" -> {}}], If[
        CurrentValue["SelectionOver"], 
        Manipulate`Dump`ReadControllerState[
         Map[Manipulate`Dump`updateOneVar[#, 
           CurrentValue["PreviousFormatTime"], 
           CurrentValue["CurrentFormatTime"]]& , {
           
           Manipulate`Dump`controllerLink[{$CellContext`i3$$, \
$CellContext`i3$91829$$}, "X1", 
            If["DefaultAbsolute", True, "JB1"], False, {1, 600, 1}, 600, 
            1.]}], 
         CurrentValue[{
          "ControllerData", {
           "Gamepad", "Joystick", "Multi-Axis Controller"}}], {}]],
       ImageSizeCache->{915., {328.75, 335.75}}],
      DefaultBaseStyle->{},
      FrameMargins->{{5, 5}, {5, 5}}],
     BaselinePosition->Automatic,
     ImageMargins->0],
    Deinitialization:>None,
    DynamicModuleValues:>{},
    SynchronousInitialization->True,
    UndoTrackedVariables:>{Typeset`show$$, Typeset`bookmarkMode$$},
    UnsavedVariables:>{Typeset`initDone$$},
    UntrackedVariables:>{Typeset`size$$}], "ListAnimate",
   Deployed->True,
   StripOnInput->False],
  Manipulate`InterpretManipulate[1]]], "Output",
 CellChangeTimes->{
  3.6419816173941536`*^9, {3.641981727031945*^9, 3.6419817533445253`*^9}, {
   3.641981849341352*^9, 3.6419818778987346`*^9}, 3.641984560432622*^9, 
   3.64230705163087*^9, 3.642322012877305*^9, 3.642441840809929*^9, 
   3.6426603763028016`*^9, {3.642667570338751*^9, 3.6426675941847167`*^9}, 
   3.642667642734553*^9, 3.642667687239071*^9, 3.6455975979250174`*^9, 
   3.645597737185563*^9, 3.6497453960197115`*^9, 3.649745448550289*^9, 
   3.6497683075494814`*^9, 3.6497683791248064`*^9, 3.6497690246072845`*^9, 
   3.6497705789473896`*^9, 3.654163561474964*^9, 3.654166283376115*^9, 
   3.6543209304330792`*^9, 3.654595936191128*^9, 3.654596104562488*^9}]
}, Open  ]],

Cell[BoxData[
 RowBox[{
  RowBox[{"exportAsGif", "[", 
   RowBox[{"\[CapitalDelta]t_", ",", "frameskip_"}], "]"}], ":=", 
  RowBox[{"Export", "[", 
   RowBox[{"\"\<.\\\\Unicycle.gif\>\"", ",", 
    RowBox[{"drawFrame", "/@", 
     RowBox[{"keepEveryNthFrame", "[", 
      RowBox[{"results", ",", "frameskip"}], "]"}]}], ",", 
    RowBox[{"\"\<DisplayDurations\>\"", "\[Rule]", 
     RowBox[{"\[CapitalDelta]t", "*", "frameskip"}]}]}], "]"}]}]], "Input",
 CellChangeTimes->{{3.641979535866313*^9, 3.6419795893224897`*^9}, 
   3.6541662129570837`*^9}],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"exportAsGif", "[", 
  RowBox[{"0.001", ",", "50"}], "]"}]], "Input",
 CellChangeTimes->{{3.64976826828242*^9, 3.6497682863752203`*^9}, {
  3.649768752489232*^9, 3.649768753568252*^9}}],

Cell[BoxData["\<\".\\\\Unicycle.gif\"\>"], "Output",
 CellChangeTimes->{3.6497682870292215`*^9, 3.6497685490940475`*^9, 
  3.6497688823058915`*^9, 3.6497692602428093`*^9, 3.6497708296032157`*^9, 
  3.6541636855390997`*^9, 3.65416652598624*^9, 3.65432115169569*^9, 
  3.654596280764471*^9}]
}, Open  ]]
},
WindowSize->{1280, 941},
WindowMargins->{{Automatic, -277}, {856, Automatic}},
Magnification:>1.1 Inherited,
FrontEndVersion->"10.0 for Microsoft Windows (64-bit) (December 4, 2014)",
StyleDefinitions->"Default.nb"
]
(* End of Notebook Content *)

(* Internal cache information *)
(*CellTagsOutline
CellTagsIndex->{}
*)
(*CellTagsIndex
CellTagsIndex->{}
*)
(*NotebookFileOutline
Notebook[{
Cell[558, 20, 193, 4, 32, "Input"],
Cell[754, 26, 273, 4, 32, "Input"],
Cell[1030, 32, 796, 17, 32, "Input"],
Cell[1829, 51, 484, 11, 32, "Input"],
Cell[2316, 64, 177, 4, 32, "Input"],
Cell[2496, 70, 838, 19, 32, "Input"],
Cell[3337, 91, 1472, 37, 32, "Input"],
Cell[CellGroupData[{
Cell[4834, 132, 146, 2, 32, "Input"],
Cell[4983, 136, 2007, 29, 33, "Output"]
}, Open  ]],
Cell[7005, 168, 3366, 74, 353, "Input"],
Cell[10374, 244, 1952, 30, 32, "Input"],
Cell[CellGroupData[{
Cell[12351, 278, 700, 15, 32, "Input"],
Cell[13054, 295, 805790, 13221, 264, 796621, 13070, "CachedBoxData", \
"BoxData", "Output"]
}, Open  ]],
Cell[CellGroupData[{
Cell[818881, 13521, 1068, 17, 32, "Input"],
Cell[819952, 13540, 2693858, 49405, 703, "Output"]
}, Open  ]],
Cell[3513825, 62948, 549, 12, 32, "Input"],
Cell[CellGroupData[{
Cell[3514399, 62964, 208, 4, 32, "Input"],
Cell[3514610, 62970, 289, 4, 33, "Output"]
}, Open  ]]
}
]
*)

(* End of internal cache information *)
